// TypeScript declaration file for JVM types. Automatically generated by doppioh.
// http://github.com/plasma-umass/doppio
import * as DoppioJVM from '../src/doppiojvm';
import JVMThread = DoppioJVM.VM.Threading.JVMThread;
import Long = DoppioJVM.VM.Long;
import ClassData = DoppioJVM.VM.ClassFile.ClassData;
import ArrayClassData = DoppioJVM.VM.ClassFile.ArrayClassData;
import ReferenceClassData = DoppioJVM.VM.ClassFile.ReferenceClassData;
import Monitor = DoppioJVM.VM.Monitor;
import ClassLoader = DoppioJVM.VM.ClassFile.ClassLoader;
import Interfaces = DoppioJVM.VM.Interfaces;

declare module JVMTypes {
  export class JVMArray<T> extends java_lang_Object {
    /**
     * NOTE: Our arrays are either JS arrays, or TypedArrays for primitive
     * types.
     */
    public array: T[];
    public getClass(): ArrayClassData<T>;
    /**
     * Create a new JVM array of this type that starts at start, and ends at
     * end. End defaults to the end of the array.
     */
    public slice(start: number, end?: number): JVMArray<T>;
  }
  // Basic, valid JVM types.
  export type BasicType = number | java_lang_Object | Long;
  export type JVMFunction = (thread: JVMThread, args: BasicType[], cb: (e?: JVMTypes.java_lang_Object, rv?: BasicType) => void) => void;
  export interface java_lang_reflect_Member extends java_lang_Object {
    "getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_reflect_Reflection extends java_lang_Object {
    public static "sun/reflect/Reflection/fieldFilterMap": java_util_Map;
    public static "sun/reflect/Reflection/methodFilterMap": java_util_Map;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/Reflection/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getCallerClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "sun/reflect/Reflection/getCallerClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "getCallerClass(I)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "sun/reflect/Reflection/getCallerClass(I)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "getClassAccessFlags(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/Reflection/getClassAccessFlags(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "quickCheckMemberAccess(Ljava/lang/Class;I)Z"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/Reflection/quickCheckMemberAccess(Ljava/lang/Class;I)Z"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ensureMemberAccess(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/Reflection/ensureMemberAccess(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "verifyMemberAccess(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/Object;I)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/Reflection/verifyMemberAccess(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/Object;I)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSameClassPackage(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/Reflection/isSameClassPackage(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSameClassPackage(Ljava/lang/ClassLoader;Ljava/lang/String;Ljava/lang/ClassLoader;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String, java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/Reflection/isSameClassPackage(Ljava/lang/ClassLoader;Ljava/lang/String;Ljava/lang/ClassLoader;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String, java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSubclassOf(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/Reflection/isSubclassOf(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "registerFieldsToFilter(Ljava/lang/Class;[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/Reflection/registerFieldsToFilter(Ljava/lang/Class;[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "registerMethodsToFilter(Ljava/lang/Class;[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/Reflection/registerMethodsToFilter(Ljava/lang/Class;[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "registerFilter(Ljava/util/Map;Ljava/lang/Class;[Ljava/lang/String;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_Map, java_lang_Class, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "sun/reflect/Reflection/registerFilter(Ljava/util/Map;Ljava/lang/Class;[Ljava/lang/String;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_Map, java_lang_Class, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "filterFields(Ljava/lang/Class;[Ljava/lang/reflect/Field;)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_reflect_Field>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public static "sun/reflect/Reflection/filterFields(Ljava/lang/Class;[Ljava/lang/reflect/Field;)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_reflect_Field>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public static "filterMethods(Ljava/lang/Class;[Ljava/lang/reflect/Method;)[Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_reflect_Method>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public static "sun/reflect/Reflection/filterMethods(Ljava/lang/Class;[Ljava/lang/reflect/Method;)[Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_reflect_Method>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public static "filter([Ljava/lang/reflect/Member;[Ljava/lang/String;)[Ljava/lang/reflect/Member;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Member>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Member>) => void): void;
    public static "sun/reflect/Reflection/filter([Ljava/lang/reflect/Member;[Ljava/lang/String;)[Ljava/lang/reflect/Member;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Member>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Member>) => void): void;
    public static "isCallerSensitive(Ljava/lang/reflect/Method;)Z"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/Reflection/isCallerSensitive(Ljava/lang/reflect/Method;)Z"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isExtClassLoader(Ljava/lang/ClassLoader;)Z"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/Reflection/isExtClassLoader(Ljava/lang/ClassLoader;)Z"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/Reflection/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_Map extends java_lang_Object {
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_function_Function extends java_lang_Object {
    "apply(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compose(Ljava/util/function/Function;)Ljava/util/function/Function;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Function) => void): void;
    "andThen(Ljava/util/function/Function;)Ljava/util/function/Function;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Function) => void): void;
    "lambda$andThen$1(Ljava/util/function/Function;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Function, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "lambda$compose$0(Ljava/util/function/Function;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Function, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_function_BiFunction extends java_lang_Object {
    "apply(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "andThen(Ljava/util/function/Function;)Ljava/util/function/BiFunction;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_function_BiFunction) => void): void;
    "lambda$andThen$0(Ljava/util/function/Function;Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Function, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_function_BiConsumer extends java_lang_Object {
    "accept(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "andThen(Ljava/util/function/BiConsumer;)Ljava/util/function/BiConsumer;"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_BiConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/BiConsumer;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_function_BiConsumer, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_Collection extends java_lang_Object, java_lang_Iterable {
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
  }
  export interface java_util_function_Consumer extends java_lang_Object {
    "accept(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "andThen(Ljava/util/function/Consumer;)Ljava/util/function/Consumer;"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Consumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/Consumer;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_function_Consumer, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_stream_Stream extends java_lang_Object, java_util_stream_BaseStream {
    "filter(Ljava/util/function/Predicate;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "map(Ljava/util/function/Function;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "mapToInt(Ljava/util/function/ToIntFunction;)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [java_util_function_ToIntFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "mapToLong(Ljava/util/function/ToLongFunction;)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [java_util_function_ToLongFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "mapToDouble(Ljava/util/function/ToDoubleFunction;)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [java_util_function_ToDoubleFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "flatMap(Ljava/util/function/Function;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "flatMapToInt(Ljava/util/function/Function;)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "flatMapToLong(Ljava/util/function/Function;)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "flatMapToDouble(Ljava/util/function/Function;)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "distinct()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "sorted()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "sorted(Ljava/util/Comparator;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "peek(Ljava/util/function/Consumer;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "limit(J)Ljava/util/stream/Stream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "skip(J)Ljava/util/stream/Stream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachOrdered(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray(Ljava/util/function/IntFunction;)[Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_IntFunction], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "reduce(Ljava/lang/Object;Ljava/util/function/BinaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "reduce(Ljava/util/function/BinaryOperator;)Ljava/util/Optional;"(thread: JVMThread, args: [java_util_function_BinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    "reduce(Ljava/lang/Object;Ljava/util/function/BiFunction;Ljava/util/function/BinaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction, java_util_function_BinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "collect(Ljava/util/function/Supplier;Ljava/util/function/BiConsumer;Ljava/util/function/BiConsumer;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Supplier, java_util_function_BiConsumer, java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "collect(Ljava/util/stream/Collector;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_stream_Collector], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "min(Ljava/util/Comparator;)Ljava/util/Optional;"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    "max(Ljava/util/Comparator;)Ljava/util/Optional;"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    "count()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "anyMatch(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "allMatch(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "noneMatch(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "findFirst()Ljava/util/Optional;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    "findAny()Ljava/util/Optional;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "isParallel()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "sequential()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "parallel()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "unordered()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "onClose(Ljava/lang/Runnable;)Ljava/util/stream/BaseStream;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_lang_Runnable extends java_lang_Object {
    "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_UnaryOperator extends java_lang_Object, java_util_function_Function {
    "apply(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compose(Ljava/util/function/Function;)Ljava/util/function/Function;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Function) => void): void;
    "andThen(Ljava/util/function/Function;)Ljava/util/function/Function;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Function) => void): void;
    "lambda$andThen$1(Ljava/util/function/Function;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Function, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "lambda$compose$0(Ljava/util/function/Function;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Function, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_stream_Stream$Builder extends java_lang_Object, java_util_function_Consumer {
    "accept(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "add(Ljava/lang/Object;)Ljava/util/stream/Stream$Builder;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream$Builder) => void): void;
    "build()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "andThen(Ljava/util/function/Consumer;)Ljava/util/function/Consumer;"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Consumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/Consumer;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_function_Consumer, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_stream_Collector extends java_lang_Object {
    "supplier()Ljava/util/function/Supplier;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Supplier) => void): void;
    "accumulator()Ljava/util/function/BiConsumer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_BiConsumer) => void): void;
    "combiner()Ljava/util/function/BinaryOperator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_BinaryOperator) => void): void;
    "finisher()Ljava/util/function/Function;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Function) => void): void;
    "characteristics()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
  }
  export class java_util_stream_Collector$Characteristics extends java_lang_Enum {
    public static "java/util/stream/Collector$Characteristics/CONCURRENT": java_util_stream_Collector$Characteristics;
    public static "java/util/stream/Collector$Characteristics/UNORDERED": java_util_stream_Collector$Characteristics;
    public static "java/util/stream/Collector$Characteristics/IDENTITY_FINISH": java_util_stream_Collector$Characteristics;
    public static "java/util/stream/Collector$Characteristics/$VALUES": JVMArray<java_util_stream_Collector$Characteristics>;
    public static "values()[Ljava/util/stream/Collector$Characteristics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_stream_Collector$Characteristics>) => void): void;
    public static "java/util/stream/Collector$Characteristics/values()[Ljava/util/stream/Collector$Characteristics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_stream_Collector$Characteristics>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/util/stream/Collector$Characteristics;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Collector$Characteristics) => void): void;
    public static "java/util/stream/Collector$Characteristics/valueOf(Ljava/lang/String;)Ljava/util/stream/Collector$Characteristics;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Collector$Characteristics) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/stream/Collector$Characteristics/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/stream/Collector$Characteristics/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Enum extends java_lang_Object implements java_lang_Comparable, java_io_Serializable {
    public "java/lang/Enum/name": java_lang_String;
    public "java/lang/Enum/ordinal": number;
    public "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Enum/name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "ordinal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Enum/ordinal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Enum/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Enum/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Enum/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Enum/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/Enum/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compareTo(Ljava/lang/Enum;)I"(thread: JVMThread, args: [java_lang_Enum], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Enum/compareTo(Ljava/lang/Enum;)I"(thread: JVMThread, args: [java_lang_Enum], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Enum/getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "valueOf(Ljava/lang/Class;Ljava/lang/String;)Ljava/lang/Enum;"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Enum) => void): void;
    public static "java/lang/Enum/valueOf(Ljava/lang/Class;Ljava/lang/String;)Ljava/lang/Enum;"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Enum) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Enum/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Enum/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObjectNoData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Enum/readObjectNoData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Enum/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_io_Serializable extends java_lang_Object {
  }
  export interface java_lang_Comparable extends java_lang_Object {
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_Supplier extends java_lang_Object {
    "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_Optional extends java_lang_Object {
    public static "java/util/Optional/EMPTY": java_util_Optional;
    public "java/util/Optional/value": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Optional/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "empty()Ljava/util/Optional;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public static "java/util/Optional/empty()Ljava/util/Optional;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Optional/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "of(Ljava/lang/Object;)Ljava/util/Optional;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public static "java/util/Optional/of(Ljava/lang/Object;)Ljava/util/Optional;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public static "ofNullable(Ljava/lang/Object;)Ljava/util/Optional;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public static "java/util/Optional/ofNullable(Ljava/lang/Object;)Ljava/util/Optional;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Optional/get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isPresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Optional/isPresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ifPresent(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Optional/ifPresent(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "filter(Ljava/util/function/Predicate;)Ljava/util/Optional;"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public "java/util/Optional/filter(Ljava/util/function/Predicate;)Ljava/util/Optional;"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public "map(Ljava/util/function/Function;)Ljava/util/Optional;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public "java/util/Optional/map(Ljava/util/function/Function;)Ljava/util/Optional;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public "flatMap(Ljava/util/function/Function;)Ljava/util/Optional;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public "java/util/Optional/flatMap(Ljava/util/function/Function;)Ljava/util/Optional;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_Optional) => void): void;
    public "orElse(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Optional/orElse(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "orElseGet(Ljava/util/function/Supplier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Optional/orElseGet(Ljava/util/function/Supplier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "orElseThrow(Ljava/util/function/Supplier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Optional/orElseThrow(Ljava/util/function/Supplier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Optional/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Optional/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Optional/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Optional/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_BinaryOperator extends java_lang_Object, java_util_function_BiFunction {
    "apply(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "andThen(Ljava/util/function/Function;)Ljava/util/function/BiFunction;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_function_BiFunction) => void): void;
    "lambda$andThen$0(Ljava/util/function/Function;Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Function, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_function_IntFunction extends java_lang_Object {
    "apply(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_Comparator extends java_lang_Object {
    "compare(Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "reversed()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparing(Ljava/util/Comparator;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparing(Ljava/util/function/Function;Ljava/util/Comparator;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_Function, java_util_Comparator], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparing(Ljava/util/function/Function;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparingInt(Ljava/util/function/ToIntFunction;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_ToIntFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparingLong(Ljava/util/function/ToLongFunction;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_ToLongFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparingDouble(Ljava/util/function/ToDoubleFunction;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_ToDoubleFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "lambda$thenComparing$36697e65$1(Ljava/util/Comparator;Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_util_Comparator, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_lang_invoke_SerializedLambda extends java_lang_Object implements java_io_Serializable {
    public static "java/lang/invoke/SerializedLambda/serialVersionUID": Long;
    public "java/lang/invoke/SerializedLambda/capturingClass": java_lang_Class;
    public "java/lang/invoke/SerializedLambda/functionalInterfaceClass": java_lang_String;
    public "java/lang/invoke/SerializedLambda/functionalInterfaceMethodName": java_lang_String;
    public "java/lang/invoke/SerializedLambda/functionalInterfaceMethodSignature": java_lang_String;
    public "java/lang/invoke/SerializedLambda/implClass": java_lang_String;
    public "java/lang/invoke/SerializedLambda/implMethodName": java_lang_String;
    public "java/lang/invoke/SerializedLambda/implMethodSignature": java_lang_String;
    public "java/lang/invoke/SerializedLambda/implMethodKind": number;
    public "java/lang/invoke/SerializedLambda/instantiatedMethodType": java_lang_String;
    public "java/lang/invoke/SerializedLambda/capturedArgs": JVMArray<java_lang_Object>;
    public "<init>(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/SerializedLambda/<init>(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCapturingClass()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/SerializedLambda/getCapturingClass()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFunctionalInterfaceClass()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/SerializedLambda/getFunctionalInterfaceClass()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFunctionalInterfaceMethodName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/SerializedLambda/getFunctionalInterfaceMethodName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFunctionalInterfaceMethodSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/SerializedLambda/getFunctionalInterfaceMethodSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getImplClass()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/SerializedLambda/getImplClass()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getImplMethodName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/SerializedLambda/getImplMethodName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getImplMethodSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/SerializedLambda/getImplMethodSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getImplMethodKind()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/SerializedLambda/getImplMethodKind()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInstantiatedMethodType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/SerializedLambda/getInstantiatedMethodType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCapturedArgCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/SerializedLambda/getCapturedArgCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCapturedArg(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/SerializedLambda/getCapturedArg(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/SerializedLambda/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/SerializedLambda/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000(Ljava/lang/invoke/SerializedLambda;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_invoke_SerializedLambda], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/invoke/SerializedLambda/access$000(Ljava/lang/invoke/SerializedLambda;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_invoke_SerializedLambda], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
  }
  export interface java_util_function_ToDoubleFunction extends java_lang_Object {
    "applyAsDouble(Ljava/lang/Object;)D"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_stream_DoubleStream extends java_lang_Object, java_util_stream_BaseStream {
    "filter(Ljava/util/function/DoublePredicate;)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [java_util_function_DoublePredicate], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "map(Ljava/util/function/DoubleUnaryOperator;)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [java_util_function_DoubleUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "mapToObj(Ljava/util/function/DoubleFunction;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_util_function_DoubleFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "mapToInt(Ljava/util/function/DoubleToIntFunction;)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [java_util_function_DoubleToIntFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "mapToLong(Ljava/util/function/DoubleToLongFunction;)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [java_util_function_DoubleToLongFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "flatMap(Ljava/util/function/DoubleFunction;)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [java_util_function_DoubleFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "distinct()Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "sorted()Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "peek(Ljava/util/function/DoubleConsumer;)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "limit(J)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "skip(J)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "forEach(Ljava/util/function/DoubleConsumer;)V"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachOrdered(Ljava/util/function/DoubleConsumer;)V"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "toArray()[D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    "reduce(DLjava/util/function/DoubleBinaryOperator;)D"(thread: JVMThread, args: [number, any, java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "reduce(Ljava/util/function/DoubleBinaryOperator;)Ljava/util/OptionalDouble;"(thread: JVMThread, args: [java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    "collect(Ljava/util/function/Supplier;Ljava/util/function/ObjDoubleConsumer;Ljava/util/function/BiConsumer;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Supplier, java_util_function_ObjDoubleConsumer, java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "sum()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "min()Ljava/util/OptionalDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    "max()Ljava/util/OptionalDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    "count()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "average()Ljava/util/OptionalDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    "summaryStatistics()Ljava/util/DoubleSummaryStatistics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_DoubleSummaryStatistics) => void): void;
    "anyMatch(Ljava/util/function/DoublePredicate;)Z"(thread: JVMThread, args: [java_util_function_DoublePredicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "allMatch(Ljava/util/function/DoublePredicate;)Z"(thread: JVMThread, args: [java_util_function_DoublePredicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "noneMatch(Ljava/util/function/DoublePredicate;)Z"(thread: JVMThread, args: [java_util_function_DoublePredicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "findFirst()Ljava/util/OptionalDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    "findAny()Ljava/util/OptionalDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    "boxed()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "sequential()Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "parallel()Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "iterator()Ljava/util/PrimitiveIterator$OfDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_PrimitiveIterator$OfDouble) => void): void;
    "spliterator()Ljava/util/Spliterator$OfDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfDouble) => void): void;
    "parallel()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "sequential()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "isParallel()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "unordered()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "onClose(Ljava/lang/Runnable;)Ljava/util/stream/BaseStream;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_DoubleSupplier extends java_lang_Object {
    "getAsDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_stream_DoubleStream$Builder extends java_lang_Object, java_util_function_DoubleConsumer {
    "accept(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    "add(D)Ljava/util/stream/DoubleStream$Builder;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream$Builder) => void): void;
    "build()Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "andThen(Ljava/util/function/DoubleConsumer;)Ljava/util/function/DoubleConsumer;"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_DoubleConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/DoubleConsumer;D)V"(thread: JVMThread, args: [java_util_function_DoubleConsumer, number, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_Spliterator$OfDouble extends java_lang_Object, java_util_Spliterator$OfPrimitive {
    "trySplit()Ljava/util/Spliterator$OfDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfDouble) => void): void;
    "tryAdvance(Ljava/util/function/DoubleConsumer;)Z"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/DoubleConsumer;)V"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "tryAdvance(Ljava/util/function/Consumer;)Z"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "tryAdvance(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "trySplit()Ljava/util/Spliterator$OfPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfPrimitive) => void): void;
    "trySplit()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "estimateSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getExactSizeIfKnown()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "characteristics()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hasCharacteristics(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getComparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "trySplit()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_Spliterator$OfPrimitive extends java_lang_Object, java_util_Spliterator {
    "trySplit()Ljava/util/Spliterator$OfPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfPrimitive) => void): void;
    "tryAdvance(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "trySplit()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "tryAdvance(Ljava/util/function/Consumer;)Z"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "estimateSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getExactSizeIfKnown()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "characteristics()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hasCharacteristics(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getComparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
  }
  export interface java_util_PrimitiveIterator$OfDouble extends java_lang_Object, java_util_PrimitiveIterator {
    "nextDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/DoubleConsumer;)V"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "next()Ljava/lang/Double;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Double) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "next()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_PrimitiveIterator extends java_lang_Object, java_util_Iterator {
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "next()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_DoubleSummaryStatistics extends java_lang_Object implements java_util_function_DoubleConsumer {
    public "java/util/DoubleSummaryStatistics/count": Long;
    public "java/util/DoubleSummaryStatistics/sum": number;
    public "java/util/DoubleSummaryStatistics/sumCompensation": number;
    public "java/util/DoubleSummaryStatistics/simpleSum": number;
    public "java/util/DoubleSummaryStatistics/min": number;
    public "java/util/DoubleSummaryStatistics/max": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/DoubleSummaryStatistics/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "accept(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/DoubleSummaryStatistics/accept(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "combine(Ljava/util/DoubleSummaryStatistics;)V"(thread: JVMThread, args: [java_util_DoubleSummaryStatistics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/DoubleSummaryStatistics/combine(Ljava/util/DoubleSummaryStatistics;)V"(thread: JVMThread, args: [java_util_DoubleSummaryStatistics], cb?: (e?: java_lang_Throwable) => void): void;
    public "sumWithCompensation(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/DoubleSummaryStatistics/sumWithCompensation(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/DoubleSummaryStatistics/getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getSum()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/DoubleSummaryStatistics/getSum()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMin()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/DoubleSummaryStatistics/getMin()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMax()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/DoubleSummaryStatistics/getMax()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/DoubleSummaryStatistics/getAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/DoubleSummaryStatistics/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "andThen(Ljava/util/function/DoubleConsumer;)Ljava/util/function/DoubleConsumer;"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_DoubleConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/DoubleConsumer;D)V"(thread: JVMThread, args: [java_util_function_DoubleConsumer, number, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_ObjDoubleConsumer extends java_lang_Object {
    "accept(Ljava/lang/Object;D)V"(thread: JVMThread, args: [java_lang_Object, number, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_OptionalDouble extends java_lang_Object {
    public static "java/util/OptionalDouble/EMPTY": java_util_OptionalDouble;
    public "java/util/OptionalDouble/isPresent": number;
    public "java/util/OptionalDouble/value": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/OptionalDouble/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "empty()Ljava/util/OptionalDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    public static "java/util/OptionalDouble/empty()Ljava/util/OptionalDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    public "<init>(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/OptionalDouble/<init>(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "of(D)Ljava/util/OptionalDouble;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    public static "java/util/OptionalDouble/of(D)Ljava/util/OptionalDouble;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    public "getAsDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalDouble/getAsDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalDouble/isPresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ifPresent(Ljava/util/function/DoubleConsumer;)V"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/OptionalDouble/ifPresent(Ljava/util/function/DoubleConsumer;)V"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "orElse(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalDouble/orElse(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "orElseGet(Ljava/util/function/DoubleSupplier;)D"(thread: JVMThread, args: [java_util_function_DoubleSupplier], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalDouble/orElseGet(Ljava/util/function/DoubleSupplier;)D"(thread: JVMThread, args: [java_util_function_DoubleSupplier], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "orElseThrow(Ljava/util/function/Supplier;)D"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalDouble/orElseThrow(Ljava/util/function/Supplier;)D"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalDouble/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalDouble/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/OptionalDouble/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/OptionalDouble/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_DoubleBinaryOperator extends java_lang_Object {
    "applyAsDouble(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_DoubleConsumer extends java_lang_Object {
    "accept(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    "andThen(Ljava/util/function/DoubleConsumer;)Ljava/util/function/DoubleConsumer;"(thread: JVMThread, args: [java_util_function_DoubleConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_DoubleConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/DoubleConsumer;D)V"(thread: JVMThread, args: [java_util_function_DoubleConsumer, number, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_DoubleToLongFunction extends java_lang_Object {
    "applyAsLong(D)J"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export interface java_util_function_DoubleToIntFunction extends java_lang_Object {
    "applyAsInt(D)I"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_DoubleFunction extends java_lang_Object {
    "apply(D)Ljava/lang/Object;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_function_DoubleUnaryOperator extends java_lang_Object {
    "applyAsDouble(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "compose(Ljava/util/function/DoubleUnaryOperator;)Ljava/util/function/DoubleUnaryOperator;"(thread: JVMThread, args: [java_util_function_DoubleUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_function_DoubleUnaryOperator) => void): void;
    "andThen(Ljava/util/function/DoubleUnaryOperator;)Ljava/util/function/DoubleUnaryOperator;"(thread: JVMThread, args: [java_util_function_DoubleUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_function_DoubleUnaryOperator) => void): void;
    "lambda$andThen$1(Ljava/util/function/DoubleUnaryOperator;D)D"(thread: JVMThread, args: [java_util_function_DoubleUnaryOperator, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$compose$0(Ljava/util/function/DoubleUnaryOperator;D)D"(thread: JVMThread, args: [java_util_function_DoubleUnaryOperator, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_DoublePredicate extends java_lang_Object {
    "test(D)Z"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "and(Ljava/util/function/DoublePredicate;)Ljava/util/function/DoublePredicate;"(thread: JVMThread, args: [java_util_function_DoublePredicate], cb?: (e?: java_lang_Throwable, rv?: java_util_function_DoublePredicate) => void): void;
    "negate()Ljava/util/function/DoublePredicate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_DoublePredicate) => void): void;
    "or(Ljava/util/function/DoublePredicate;)Ljava/util/function/DoublePredicate;"(thread: JVMThread, args: [java_util_function_DoublePredicate], cb?: (e?: java_lang_Throwable, rv?: java_util_function_DoublePredicate) => void): void;
    "lambda$or$2(Ljava/util/function/DoublePredicate;D)Z"(thread: JVMThread, args: [java_util_function_DoublePredicate, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$negate$1(D)Z"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$and$0(Ljava/util/function/DoublePredicate;D)Z"(thread: JVMThread, args: [java_util_function_DoublePredicate, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_ToLongFunction extends java_lang_Object {
    "applyAsLong(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export interface java_util_stream_LongStream extends java_lang_Object, java_util_stream_BaseStream {
    "filter(Ljava/util/function/LongPredicate;)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [java_util_function_LongPredicate], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "map(Ljava/util/function/LongUnaryOperator;)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [java_util_function_LongUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "mapToObj(Ljava/util/function/LongFunction;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_util_function_LongFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "mapToInt(Ljava/util/function/LongToIntFunction;)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [java_util_function_LongToIntFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "mapToDouble(Ljava/util/function/LongToDoubleFunction;)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [java_util_function_LongToDoubleFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "flatMap(Ljava/util/function/LongFunction;)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [java_util_function_LongFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "distinct()Ljava/util/stream/LongStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "sorted()Ljava/util/stream/LongStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "peek(Ljava/util/function/LongConsumer;)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "limit(J)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "skip(J)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "forEach(Ljava/util/function/LongConsumer;)V"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachOrdered(Ljava/util/function/LongConsumer;)V"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "toArray()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "reduce(JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "reduce(Ljava/util/function/LongBinaryOperator;)Ljava/util/OptionalLong;"(thread: JVMThread, args: [java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalLong) => void): void;
    "collect(Ljava/util/function/Supplier;Ljava/util/function/ObjLongConsumer;Ljava/util/function/BiConsumer;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Supplier, java_util_function_ObjLongConsumer, java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "sum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "min()Ljava/util/OptionalLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalLong) => void): void;
    "max()Ljava/util/OptionalLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalLong) => void): void;
    "count()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "average()Ljava/util/OptionalDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    "summaryStatistics()Ljava/util/LongSummaryStatistics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_LongSummaryStatistics) => void): void;
    "anyMatch(Ljava/util/function/LongPredicate;)Z"(thread: JVMThread, args: [java_util_function_LongPredicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "allMatch(Ljava/util/function/LongPredicate;)Z"(thread: JVMThread, args: [java_util_function_LongPredicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "noneMatch(Ljava/util/function/LongPredicate;)Z"(thread: JVMThread, args: [java_util_function_LongPredicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "findFirst()Ljava/util/OptionalLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalLong) => void): void;
    "findAny()Ljava/util/OptionalLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalLong) => void): void;
    "asDoubleStream()Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "boxed()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "sequential()Ljava/util/stream/LongStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "parallel()Ljava/util/stream/LongStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "iterator()Ljava/util/PrimitiveIterator$OfLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_PrimitiveIterator$OfLong) => void): void;
    "spliterator()Ljava/util/Spliterator$OfLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfLong) => void): void;
    "parallel()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "sequential()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "isParallel()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "unordered()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "onClose(Ljava/lang/Runnable;)Ljava/util/stream/BaseStream;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_LongSupplier extends java_lang_Object {
    "getAsLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export interface java_util_stream_LongStream$Builder extends java_lang_Object, java_util_function_LongConsumer {
    "accept(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    "add(J)Ljava/util/stream/LongStream$Builder;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream$Builder) => void): void;
    "build()Ljava/util/stream/LongStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "andThen(Ljava/util/function/LongConsumer;)Ljava/util/function/LongConsumer;"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_LongConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/LongConsumer;J)V"(thread: JVMThread, args: [java_util_function_LongConsumer, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_Spliterator$OfLong extends java_lang_Object, java_util_Spliterator$OfPrimitive {
    "trySplit()Ljava/util/Spliterator$OfLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfLong) => void): void;
    "tryAdvance(Ljava/util/function/LongConsumer;)Z"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/LongConsumer;)V"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "tryAdvance(Ljava/util/function/Consumer;)Z"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "tryAdvance(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "trySplit()Ljava/util/Spliterator$OfPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfPrimitive) => void): void;
    "trySplit()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "estimateSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getExactSizeIfKnown()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "characteristics()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hasCharacteristics(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getComparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "trySplit()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_PrimitiveIterator$OfLong extends java_lang_Object, java_util_PrimitiveIterator {
    "nextLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "forEachRemaining(Ljava/util/function/LongConsumer;)V"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "next()Ljava/lang/Long;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "next()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Long extends java_lang_Number implements java_lang_Comparable {
    public unbox(): Long;
    public static box(val: Long): java_lang_Long;
    public static "java/lang/Long/MIN_VALUE": Long;
    public static "java/lang/Long/MAX_VALUE": Long;
    public static "java/lang/Long/TYPE": java_lang_Class;
    public "java/lang/Long/value": Long;
    public static "java/lang/Long/SIZE": number;
    public static "java/lang/Long/BYTES": number;
    public static "java/lang/Long/serialVersionUID": Long;
    public static "toString(JI)Ljava/lang/String;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Long/toString(JI)Ljava/lang/String;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toUnsignedString(JI)Ljava/lang/String;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Long/toUnsignedString(JI)Ljava/lang/String;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toUnsignedBigInteger(J)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/lang/Long/toUnsignedBigInteger(J)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "toHexString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Long/toHexString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toOctalString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Long/toOctalString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toBinaryString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Long/toBinaryString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toUnsignedString0(JI)Ljava/lang/String;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Long/toUnsignedString0(JI)Ljava/lang/String;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "formatUnsignedLong(JI[CII)I"(thread: JVMThread, args: [Long, any, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Long/formatUnsignedLong(JI[CII)I"(thread: JVMThread, args: [Long, any, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Long/toString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toUnsignedString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Long/toUnsignedString(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getChars(JI[C)V"(thread: JVMThread, args: [Long, any, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Long/getChars(JI[C)V"(thread: JVMThread, args: [Long, any, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "stringSize(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Long/stringSize(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "parseLong(Ljava/lang/String;I)J"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/parseLong(Ljava/lang/String;I)J"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "parseLong(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/parseLong(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "parseUnsignedLong(Ljava/lang/String;I)J"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/parseUnsignedLong(Ljava/lang/String;I)J"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "parseUnsignedLong(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/parseUnsignedLong(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "valueOf(Ljava/lang/String;I)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/lang/Long/valueOf(Ljava/lang/String;I)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/lang/Long/valueOf(Ljava/lang/String;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "valueOf(J)Ljava/lang/Long;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/lang/Long/valueOf(J)Ljava/lang/Long;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "decode(Ljava/lang/String;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/lang/Long/decode(Ljava/lang/String;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Long/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Long/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Long/byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Long/shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Long/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/Long/longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Long/floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Long/doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Long/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Long/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hashCode(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Long/hashCode(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Long/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getLong(Ljava/lang/String;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/lang/Long/getLong(Ljava/lang/String;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "getLong(Ljava/lang/String;J)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/lang/Long/getLong(Ljava/lang/String;J)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "getLong(Ljava/lang/String;Ljava/lang/Long;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String, java_lang_Long], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/lang/Long/getLong(Ljava/lang/String;Ljava/lang/Long;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_String, java_lang_Long], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public "compareTo(Ljava/lang/Long;)I"(thread: JVMThread, args: [java_lang_Long], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Long/compareTo(Ljava/lang/Long;)I"(thread: JVMThread, args: [java_lang_Long], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Long/compare(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compareUnsigned(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Long/compareUnsigned(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "divideUnsigned(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/divideUnsigned(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "remainderUnsigned(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/remainderUnsigned(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "highestOneBit(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/highestOneBit(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "lowestOneBit(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/lowestOneBit(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "numberOfLeadingZeros(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Long/numberOfLeadingZeros(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "numberOfTrailingZeros(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Long/numberOfTrailingZeros(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "bitCount(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Long/bitCount(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "rotateLeft(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/rotateLeft(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "rotateRight(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/rotateRight(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "reverse(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/reverse(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "signum(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Long/signum(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "reverseBytes(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/reverseBytes(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sum(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/sum(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "max(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/max(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "min(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Long/min(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Long/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Long/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_math_BigInteger extends java_lang_Number implements java_lang_Comparable {
    public "java/math/BigInteger/signum": number;
    public "java/math/BigInteger/mag": JVMArray<number>;
    public "java/math/BigInteger/bitCount": number;
    public "java/math/BigInteger/bitLength": number;
    public "java/math/BigInteger/lowestSetBit": number;
    public "java/math/BigInteger/firstNonzeroIntNum": number;
    public static "java/math/BigInteger/LONG_MASK": Long;
    public static "java/math/BigInteger/MAX_MAG_LENGTH": number;
    public static "java/math/BigInteger/PRIME_SEARCH_BIT_LENGTH_LIMIT": number;
    public static "java/math/BigInteger/KARATSUBA_THRESHOLD": number;
    public static "java/math/BigInteger/TOOM_COOK_THRESHOLD": number;
    public static "java/math/BigInteger/KARATSUBA_SQUARE_THRESHOLD": number;
    public static "java/math/BigInteger/TOOM_COOK_SQUARE_THRESHOLD": number;
    public static "java/math/BigInteger/BURNIKEL_ZIEGLER_THRESHOLD": number;
    public static "java/math/BigInteger/BURNIKEL_ZIEGLER_OFFSET": number;
    public static "java/math/BigInteger/SCHOENHAGE_BASE_CONVERSION_THRESHOLD": number;
    public static "java/math/BigInteger/MULTIPLY_SQUARE_THRESHOLD": number;
    public static "java/math/BigInteger/bitsPerDigit": JVMArray<Long>;
    public static "java/math/BigInteger/SMALL_PRIME_THRESHOLD": number;
    public static "java/math/BigInteger/DEFAULT_PRIME_CERTAINTY": number;
    public static "java/math/BigInteger/SMALL_PRIME_PRODUCT": java_math_BigInteger;
    public static "java/math/BigInteger/MAX_CONSTANT": number;
    public static "java/math/BigInteger/posConst": JVMArray<java_math_BigInteger>;
    public static "java/math/BigInteger/negConst": JVMArray<java_math_BigInteger>;
    public static "java/math/BigInteger/powerCache": JVMArray<JVMArray<java_math_BigInteger>>;
    public static "java/math/BigInteger/logCache": JVMArray<number>;
    public static "java/math/BigInteger/LOG_TWO": number;
    public static "java/math/BigInteger/ZERO": java_math_BigInteger;
    public static "java/math/BigInteger/ONE": java_math_BigInteger;
    public static "java/math/BigInteger/TWO": java_math_BigInteger;
    public static "java/math/BigInteger/NEGATIVE_ONE": java_math_BigInteger;
    public static "java/math/BigInteger/TEN": java_math_BigInteger;
    public static "java/math/BigInteger/bnExpModThreshTable": JVMArray<number>;
    public static "java/math/BigInteger/zeros": JVMArray<java_lang_String>;
    public static "java/math/BigInteger/digitsPerLong": JVMArray<number>;
    public static "java/math/BigInteger/longRadix": JVMArray<java_math_BigInteger>;
    public static "java/math/BigInteger/digitsPerInt": JVMArray<number>;
    public static "java/math/BigInteger/intRadix": JVMArray<number>;
    public static "java/math/BigInteger/serialVersionUID": Long;
    public static "java/math/BigInteger/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/math/BigInteger/$assertionsDisabled": number;
    public "<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I[B)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>(I[B)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I[I)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>(I[I)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "parseInt([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/parseInt([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "destructiveMulAdd([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/BigInteger/destructiveMulAdd([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/util/Random;)V"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>(ILjava/util/Random;)V"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable) => void): void;
    public static "randomBits(ILjava/util/Random;)[B"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/randomBits(ILjava/util/Random;)[B"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "<init>(IILjava/util/Random;)V"(thread: JVMThread, args: [number, number, java_util_Random], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>(IILjava/util/Random;)V"(thread: JVMThread, args: [number, number, java_util_Random], cb?: (e?: java_lang_Throwable) => void): void;
    public static "probablePrime(ILjava/util/Random;)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/probablePrime(ILjava/util/Random;)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "smallPrime(IILjava/util/Random;)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/smallPrime(IILjava/util/Random;)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "largePrime(IILjava/util/Random;)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/largePrime(IILjava/util/Random;)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "nextProbablePrime()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/nextProbablePrime()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "getPrimeSearchLen(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigInteger/getPrimeSearchLen(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "primeToCertainty(ILjava/util/Random;)Z"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/primeToCertainty(ILjava/util/Random;)Z"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "passesLucasLehmer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/passesLucasLehmer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jacobiSymbol(ILjava/math/BigInteger;)I"(thread: JVMThread, args: [number, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigInteger/jacobiSymbol(ILjava/math/BigInteger;)I"(thread: JVMThread, args: [number, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "lucasLehmerSequence(ILjava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/lucasLehmerSequence(ILjava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "passesMillerRabin(ILjava/util/Random;)Z"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/passesMillerRabin(ILjava/util/Random;)Z"(thread: JVMThread, args: [number, java_util_Random], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>([II)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>([II)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>([BI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkRange()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/checkRange()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "reportOverflow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/BigInteger/reportOverflow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "valueOf(J)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/valueOf(J)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "valueOf([I)Ljava/math/BigInteger;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/valueOf([I)Ljava/math/BigInteger;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "add(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/add(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "add(J)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/add(J)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "add([IJ)[I"(thread: JVMThread, args: [JVMArray<number>, Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/add([IJ)[I"(thread: JVMThread, args: [JVMArray<number>, Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "add([I[I)[I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/add([I[I)[I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "subtract(J[I)[I"(thread: JVMThread, args: [Long, any, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/subtract(J[I)[I"(thread: JVMThread, args: [Long, any, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "subtract([IJ)[I"(thread: JVMThread, args: [JVMArray<number>, Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/subtract([IJ)[I"(thread: JVMThread, args: [JVMArray<number>, Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "subtract(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/subtract(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "subtract([I[I)[I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/subtract([I[I)[I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "multiply(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/multiply(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "multiplyByInt([III)Ljava/math/BigInteger;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/multiplyByInt([III)Ljava/math/BigInteger;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "multiply(J)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/multiply(J)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "multiplyToLen([II[II[I)[I"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/math/BigInteger/multiplyToLen([II[II[I)[I"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "multiplyKaratsuba(Ljava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/multiplyKaratsuba(Ljava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "multiplyToomCook3(Ljava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/multiplyToomCook3(Ljava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "getToomSlice(IIII)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/getToomSlice(IIII)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "exactDivideBy3()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/exactDivideBy3()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "getLower(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/getLower(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "getUpper(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/getUpper(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "square()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/square()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "squareToLen([II[I)[I"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/squareToLen([II[I)[I"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "squareKaratsuba()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/squareKaratsuba()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "squareToomCook3()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/squareToomCook3()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "divide(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/divide(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "divideKnuth(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/divideKnuth(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "divideAndRemainder(Ljava/math/BigInteger;)[Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigInteger>) => void): void;
    public "java/math/BigInteger/divideAndRemainder(Ljava/math/BigInteger;)[Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigInteger>) => void): void;
    public "divideAndRemainderKnuth(Ljava/math/BigInteger;)[Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigInteger>) => void): void;
    public "java/math/BigInteger/divideAndRemainderKnuth(Ljava/math/BigInteger;)[Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigInteger>) => void): void;
    public "remainder(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/remainder(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "remainderKnuth(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/remainderKnuth(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "divideBurnikelZiegler(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/divideBurnikelZiegler(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "remainderBurnikelZiegler(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/remainderBurnikelZiegler(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "divideAndRemainderBurnikelZiegler(Ljava/math/BigInteger;)[Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigInteger>) => void): void;
    public "java/math/BigInteger/divideAndRemainderBurnikelZiegler(Ljava/math/BigInteger;)[Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigInteger>) => void): void;
    public "pow(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/pow(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "gcd(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/gcd(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "bitLengthForInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigInteger/bitLengthForInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "leftShift([III)[I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/leftShift([III)[I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "primitiveRightShift([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/BigInteger/primitiveRightShift([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "primitiveLeftShift([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/BigInteger/primitiveLeftShift([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "bitLength([II)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigInteger/bitLength([II)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "abs()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/abs()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "negate()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/negate()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "signum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/signum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mod(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/mod(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "modPow(Ljava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/modPow(Ljava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "oddModPow(Ljava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/oddModPow(Ljava/math/BigInteger;Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "montReduce([I[III)[I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/montReduce([I[III)[I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "intArrayCmpToLen([I[II)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigInteger/intArrayCmpToLen([I[II)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "subN([I[II)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigInteger/subN([I[II)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "mulAdd([I[IIII)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigInteger/mulAdd([I[IIII)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "addOne([IIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigInteger/addOne([IIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "modPow2(Ljava/math/BigInteger;I)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/modPow2(Ljava/math/BigInteger;I)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "mod2(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/mod2(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "modInverse(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/modInverse(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "shiftLeft(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/shiftLeft(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "shiftLeft([II)[I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/shiftLeft([II)[I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "shiftRight(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/shiftRight(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "shiftRightImpl(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/shiftRightImpl(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "javaIncrement([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/math/BigInteger/javaIncrement([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "and(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/and(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "or(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/or(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "xor(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/xor(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "not()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/not()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "andNot(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/andNot(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "testBit(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/testBit(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setBit(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/setBit(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "clearBit(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/clearBit(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "flipBit(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/flipBit(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "getLowestSetBit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/getLowestSetBit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "bitLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/bitLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "bitCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/bitCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isProbablePrime(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/isProbablePrime(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/math/BigInteger;)I"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/compareTo(Ljava/math/BigInteger;)I"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareMagnitude(Ljava/math/BigInteger;)I"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/compareMagnitude(Ljava/math/BigInteger;)I"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareMagnitude(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/compareMagnitude(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "min(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/min(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "max(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigInteger/max(Ljava/math/BigInteger;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/BigInteger/toString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "smallToString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/BigInteger/smallToString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toString(Ljava/math/BigInteger;Ljava/lang/StringBuilder;II)V"(thread: JVMThread, args: [java_math_BigInteger, java_lang_StringBuilder, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/BigInteger/toString(Ljava/math/BigInteger;Ljava/lang/StringBuilder;II)V"(thread: JVMThread, args: [java_math_BigInteger, java_lang_StringBuilder, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getRadixConversionCache(II)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigInteger/getRadixConversionCache(II)Ljava/math/BigInteger;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/BigInteger/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/math/BigInteger/toByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/math/BigInteger/longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "stripLeadingZeroInts([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/stripLeadingZeroInts([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "trustedStripLeadingZeroInts([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/trustedStripLeadingZeroInts([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "stripLeadingZeroBytes([B)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/stripLeadingZeroBytes([B)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "makePositive([B)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/makePositive([B)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "makePositive([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/math/BigInteger/makePositive([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "intLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/intLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "signBit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/signBit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "signInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/signInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "firstNonzeroIntNum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/firstNonzeroIntNum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigInteger/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "magSerializedForm()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/math/BigInteger/magSerializedForm()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "longValueExact()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/math/BigInteger/longValueExact()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "intValueExact()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/intValueExact()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shortValueExact()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/shortValueExact()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "byteValueExact()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/byteValueExact()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigInteger/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/BigInteger/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_StringBuilder extends java_lang_AbstractStringBuilder implements java_io_Serializable, java_lang_CharSequence {
    public static "java/lang/StringBuilder/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/CharSequence;)V"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/<init>(Ljava/lang/CharSequence;)V"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(Ljava/lang/Object;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/Object;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(Ljava/lang/String;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/String;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(Ljava/lang/StringBuffer;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/StringBuffer;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/CharSequence;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/CharSequence;II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append([C)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append([C)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append([CII)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append([CII)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(Z)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Z)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(C)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(C)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(I)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(I)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(J)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(J)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(F)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(F)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "append(D)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/append(D)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "appendCodePoint(I)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/appendCodePoint(I)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "delete(II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/delete(II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "deleteCharAt(I)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/deleteCharAt(I)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "replace(IILjava/lang/String;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/replace(IILjava/lang/String;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(I[CII)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(I[CII)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(ILjava/lang/Object;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ILjava/lang/Object;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(ILjava/lang/String;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ILjava/lang/String;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(I[C)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(I[C)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(ILjava/lang/CharSequence;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ILjava/lang/CharSequence;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(ILjava/lang/CharSequence;II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ILjava/lang/CharSequence;II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(IZ)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(IZ)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(IC)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(IC)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(IJ)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(IJ)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(IF)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(IF)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "insert(ID)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ID)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "indexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/indexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/indexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/lastIndexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/lastIndexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reverse()Ljava/lang/StringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/lang/StringBuilder/reverse()Ljava/lang/StringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StringBuilder/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "reverse()Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/reverse()Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ID)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ID)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IF)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(IF)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IJ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(IJ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IC)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(IC)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IZ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(IZ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ILjava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ILjava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(I[C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(I[C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(ILjava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(I[CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/insert(I[CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "substring(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StringBuilder/substring(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "java/lang/StringBuilder/subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "substring(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StringBuilder/substring(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "replace(IILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/replace(IILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "deleteCharAt(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/deleteCharAt(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "appendCodePoint(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/appendCodePoint(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "delete(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/delete(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(D)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(D)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(F)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(F)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(J)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(J)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Z)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Z)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append([CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append([CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append([C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append([C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/StringBuffer;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/StringBuffer;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "setCharAt(IC)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/setCharAt(IC)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChars(II[CI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/getChars(II[CI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "offsetByCodePoints(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/offsetByCodePoints(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointCount(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/codePointCount(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointBefore(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/codePointBefore(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/codePointAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLength(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/setLength(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuilder/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuilder/length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/lang/StringBuilder/append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/lang/StringBuilder/append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    "chars()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "codePoints()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
  }
  export interface java_lang_Appendable extends java_lang_Object {
    "append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    "append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    "append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
  }
  export class java_lang_StringBuffer extends java_lang_AbstractStringBuilder implements java_io_Serializable, java_lang_CharSequence {
    public "java/lang/StringBuffer/toStringCache": JVMArray<number>;
    public static "java/lang/StringBuffer/serialVersionUID": Long;
    public static "java/lang/StringBuffer/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/CharSequence;)V"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/<init>(Ljava/lang/CharSequence;)V"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable) => void): void;
    public "length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLength(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/setLength(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/codePointAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointBefore(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/codePointBefore(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointCount(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/codePointCount(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offsetByCodePoints(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/offsetByCodePoints(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getChars(II[CI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/getChars(II[CI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCharAt(IC)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/setCharAt(IC)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(Ljava/lang/Object;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/Object;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(Ljava/lang/String;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/String;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(Ljava/lang/StringBuffer;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/StringBuffer;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(Ljava/lang/AbstractStringBuilder;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_AbstractStringBuilder], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/AbstractStringBuilder;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_AbstractStringBuilder], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/CharSequence;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/CharSequence;II)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append([C)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append([C)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append([CII)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append([CII)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(Z)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(Z)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(C)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(C)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(I)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(I)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "appendCodePoint(I)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/appendCodePoint(I)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(J)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(J)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(F)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(F)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(D)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/append(D)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "delete(II)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/delete(II)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "deleteCharAt(I)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/deleteCharAt(I)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "replace(IILjava/lang/String;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/replace(IILjava/lang/String;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "substring(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StringBuffer/substring(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "java/lang/StringBuffer/subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "substring(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StringBuffer/substring(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "insert(I[CII)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(I[CII)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(ILjava/lang/Object;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(ILjava/lang/Object;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(ILjava/lang/String;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(ILjava/lang/String;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(I[C)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(I[C)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(ILjava/lang/CharSequence;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(ILjava/lang/CharSequence;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(ILjava/lang/CharSequence;II)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(ILjava/lang/CharSequence;II)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(IZ)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(IZ)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(IC)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(IC)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(II)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(II)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(IJ)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(IJ)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(IF)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(IF)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "insert(ID)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/insert(ID)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "indexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/indexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/indexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/lastIndexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StringBuffer/lastIndexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reverse()Ljava/lang/StringBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/lang/StringBuffer/reverse()Ljava/lang/StringBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StringBuffer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StringBuffer/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "reverse()Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/reverse()Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ID)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(ID)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IF)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(IF)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IJ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(IJ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IC)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(IC)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IZ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(IZ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(ILjava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(ILjava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(I[C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(I[C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(ILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(ILjava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(I[CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/insert(I[CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "replace(IILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/replace(IILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "deleteCharAt(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/deleteCharAt(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "appendCodePoint(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/appendCodePoint(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "delete(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/delete(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(D)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(D)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(F)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(F)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(J)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(J)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Z)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(Z)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append([CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append([CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append([C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append([C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/AbstractStringBuilder;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_AbstractStringBuilder], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/AbstractStringBuilder;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_AbstractStringBuilder], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/StringBuffer;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/StringBuffer;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/lang/StringBuffer/append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/lang/StringBuffer/append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/StringBuffer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "chars()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "codePoints()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
  }
  export interface java_lang_CharSequence extends java_lang_Object {
    "length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "chars()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "codePoints()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "lambda$codePoints$1()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    "lambda$chars$0()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
  }
  export interface java_util_Spliterator$OfInt extends java_lang_Object, java_util_Spliterator$OfPrimitive {
    "trySplit()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    "tryAdvance(Ljava/util/function/IntConsumer;)Z"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/IntConsumer;)V"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "tryAdvance(Ljava/util/function/Consumer;)Z"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "tryAdvance(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "trySplit()Ljava/util/Spliterator$OfPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfPrimitive) => void): void;
    "trySplit()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "estimateSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getExactSizeIfKnown()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "characteristics()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hasCharacteristics(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getComparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "trySplit()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_IntConsumer extends java_lang_Object {
    "accept(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "andThen(Ljava/util/function/IntConsumer;)Ljava/util/function/IntConsumer;"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_IntConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/IntConsumer;I)V"(thread: JVMThread, args: [java_util_function_IntConsumer, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_AbstractStringBuilder extends java_lang_Object implements java_lang_Appendable, java_lang_CharSequence {
    public "java/lang/AbstractStringBuilder/value": JVMArray<number>;
    public "java/lang/AbstractStringBuilder/count": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureCapacityInternal(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/ensureCapacityInternal(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "expandCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/expandCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLength(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/setLength(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/codePointAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointBefore(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/codePointBefore(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointCount(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/codePointCount(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offsetByCodePoints(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/offsetByCodePoints(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getChars(II[CI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/getChars(II[CI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCharAt(IC)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/setCharAt(IC)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(Ljava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(Ljava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(Ljava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/StringBuffer;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(Ljava/lang/StringBuffer;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/AbstractStringBuilder;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_AbstractStringBuilder], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(Ljava/lang/AbstractStringBuilder;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_AbstractStringBuilder], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(Ljava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "appendNull()Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/appendNull()Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(Ljava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append([C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append([C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append([CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append([CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(Z)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(Z)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(J)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(J)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(F)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(F)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "append(D)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/append(D)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "delete(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/delete(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "appendCodePoint(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/appendCodePoint(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "deleteCharAt(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/deleteCharAt(I)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "replace(IILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/replace(IILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "substring(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/AbstractStringBuilder/substring(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "java/lang/AbstractStringBuilder/subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "substring(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/AbstractStringBuilder/substring(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "insert(I[CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(I[CII)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(ILjava/lang/Object;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(ILjava/lang/String;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(I[C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(I[C)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(ILjava/lang/CharSequence;)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ILjava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(ILjava/lang/CharSequence;II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IZ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(IZ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IC)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(IC)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(II)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IJ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(IJ)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(IF)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(IF)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "insert(ID)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/insert(ID)Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "indexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/indexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/indexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/lastIndexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/AbstractStringBuilder/lastIndexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reverse()Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "java/lang/AbstractStringBuilder/reverse()Ljava/lang/AbstractStringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AbstractStringBuilder) => void): void;
    public "reverseAllValidSurrogatePairs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AbstractStringBuilder/reverseAllValidSurrogatePairs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/AbstractStringBuilder/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getValue()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/AbstractStringBuilder/getValue()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/lang/AbstractStringBuilder/append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/lang/AbstractStringBuilder/append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/lang/AbstractStringBuilder/append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    "chars()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "codePoints()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "lambda$codePoints$1()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    "lambda$chars$0()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
  }
  export class java_util_Random extends java_lang_Object implements java_io_Serializable {
    public static "java/util/Random/serialVersionUID": Long;
    public "java/util/Random/seed": java_util_concurrent_atomic_AtomicLong;
    public static "java/util/Random/multiplier": Long;
    public static "java/util/Random/addend": Long;
    public static "java/util/Random/mask": Long;
    public static "java/util/Random/DOUBLE_UNIT": number;
    public static "java/util/Random/BadBound": java_lang_String;
    public static "java/util/Random/BadRange": java_lang_String;
    public static "java/util/Random/BadSize": java_lang_String;
    public static "java/util/Random/seedUniquifier": java_util_concurrent_atomic_AtomicLong;
    public "java/util/Random/nextNextGaussian": number;
    public "java/util/Random/haveNextNextGaussian": number;
    public static "java/util/Random/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/util/Random/unsafe": sun_misc_Unsafe;
    public static "java/util/Random/seedOffset": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Random/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "seedUniquifier()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/Random/seedUniquifier()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Random/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initialScramble(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/Random/initialScramble(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setSeed(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Random/setSeed(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "next(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Random/next(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextBytes([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Random/nextBytes([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "internalNextLong(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/Random/internalNextLong(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "internalNextInt(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Random/internalNextInt(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "internalNextDouble(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Random/internalNextDouble(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Random/nextInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Random/nextInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/Random/nextLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "nextBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Random/nextBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Random/nextFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Random/nextDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextGaussian()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Random/nextGaussian()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ints(J)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "java/util/Random/ints(J)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "ints()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "java/util/Random/ints()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "ints(JII)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "java/util/Random/ints(JII)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "ints(II)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "java/util/Random/ints(II)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "longs(J)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    public "java/util/Random/longs(J)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    public "longs()Ljava/util/stream/LongStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    public "java/util/Random/longs()Ljava/util/stream/LongStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    public "longs(JJJ)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    public "java/util/Random/longs(JJJ)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    public "longs(JJ)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    public "java/util/Random/longs(JJ)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    public "doubles(J)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    public "java/util/Random/doubles(J)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    public "doubles()Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    public "java/util/Random/doubles()Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    public "doubles(JDD)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [Long, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    public "java/util/Random/doubles(JDD)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [Long, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    public "doubles(DD)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    public "java/util/Random/doubles(DD)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Random/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Random/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "resetSeed(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Random/resetSeed(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Random/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectStreamField extends java_lang_Object implements java_lang_Comparable {
    public "java/io/ObjectStreamField/name": java_lang_String;
    public "java/io/ObjectStreamField/signature": java_lang_String;
    public "java/io/ObjectStreamField/type": java_lang_Class;
    public "java/io/ObjectStreamField/unshared": number;
    public "java/io/ObjectStreamField/field": java_lang_reflect_Field;
    public "java/io/ObjectStreamField/offset": number;
    public "<init>(Ljava/lang/String;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamField/<init>(Ljava/lang/String;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Class;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamField/<init>(Ljava/lang/String;Ljava/lang/Class;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamField/<init>(Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/reflect/Field;ZZ)V"(thread: JVMThread, args: [java_lang_reflect_Field, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamField/<init>(Ljava/lang/reflect/Field;ZZ)V"(thread: JVMThread, args: [java_lang_reflect_Field, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectStreamField/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/io/ObjectStreamField/getType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getTypeCode()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamField/getTypeCode()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTypeString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectStreamField/getTypeString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamField/getOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setOffset(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamField/setOffset(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isPrimitive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamField/isPrimitive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isUnshared()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamField/isUnshared()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamField/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectStreamField/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getField()Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "java/io/ObjectStreamField/getField()Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectStreamField/getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getClassSignature(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/ObjectStreamField/getClassSignature(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_lang_Number extends java_lang_Object implements java_io_Serializable {
    public static "java/lang/Number/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Number/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Number/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/Number/longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Number/floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Number/doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Number/byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Number/shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_LongSummaryStatistics extends java_lang_Object implements java_util_function_LongConsumer, java_util_function_IntConsumer {
    public "java/util/LongSummaryStatistics/count": Long;
    public "java/util/LongSummaryStatistics/sum": Long;
    public "java/util/LongSummaryStatistics/min": Long;
    public "java/util/LongSummaryStatistics/max": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LongSummaryStatistics/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "accept(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LongSummaryStatistics/accept(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "accept(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LongSummaryStatistics/accept(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "combine(Ljava/util/LongSummaryStatistics;)V"(thread: JVMThread, args: [java_util_LongSummaryStatistics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LongSummaryStatistics/combine(Ljava/util/LongSummaryStatistics;)V"(thread: JVMThread, args: [java_util_LongSummaryStatistics], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/LongSummaryStatistics/getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getSum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/LongSummaryStatistics/getSum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMin()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/LongSummaryStatistics/getMin()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMax()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/LongSummaryStatistics/getMax()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LongSummaryStatistics/getAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/LongSummaryStatistics/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "andThen(Ljava/util/function/LongConsumer;)Ljava/util/function/LongConsumer;"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_LongConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/LongConsumer;J)V"(thread: JVMThread, args: [java_util_function_LongConsumer, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    "andThen(Ljava/util/function/IntConsumer;)Ljava/util/function/IntConsumer;"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_IntConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/IntConsumer;I)V"(thread: JVMThread, args: [java_util_function_IntConsumer, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_ObjLongConsumer extends java_lang_Object {
    "accept(Ljava/lang/Object;J)V"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_OptionalLong extends java_lang_Object {
    public static "java/util/OptionalLong/EMPTY": java_util_OptionalLong;
    public "java/util/OptionalLong/isPresent": number;
    public "java/util/OptionalLong/value": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/OptionalLong/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "empty()Ljava/util/OptionalLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalLong) => void): void;
    public static "java/util/OptionalLong/empty()Ljava/util/OptionalLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalLong) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/OptionalLong/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "of(J)Ljava/util/OptionalLong;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalLong) => void): void;
    public static "java/util/OptionalLong/of(J)Ljava/util/OptionalLong;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalLong) => void): void;
    public "getAsLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/OptionalLong/getAsLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "isPresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalLong/isPresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ifPresent(Ljava/util/function/LongConsumer;)V"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/OptionalLong/ifPresent(Ljava/util/function/LongConsumer;)V"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "orElse(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/OptionalLong/orElse(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "orElseGet(Ljava/util/function/LongSupplier;)J"(thread: JVMThread, args: [java_util_function_LongSupplier], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/OptionalLong/orElseGet(Ljava/util/function/LongSupplier;)J"(thread: JVMThread, args: [java_util_function_LongSupplier], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "orElseThrow(Ljava/util/function/Supplier;)J"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/OptionalLong/orElseThrow(Ljava/util/function/Supplier;)J"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalLong/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalLong/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/OptionalLong/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/OptionalLong/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_LongBinaryOperator extends java_lang_Object {
    "applyAsLong(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export interface java_util_function_LongConsumer extends java_lang_Object {
    "accept(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    "andThen(Ljava/util/function/LongConsumer;)Ljava/util/function/LongConsumer;"(thread: JVMThread, args: [java_util_function_LongConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_LongConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/LongConsumer;J)V"(thread: JVMThread, args: [java_util_function_LongConsumer, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_LongToDoubleFunction extends java_lang_Object {
    "applyAsDouble(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_LongToIntFunction extends java_lang_Object {
    "applyAsInt(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_LongFunction extends java_lang_Object {
    "apply(J)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_function_LongUnaryOperator extends java_lang_Object {
    "applyAsLong(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "compose(Ljava/util/function/LongUnaryOperator;)Ljava/util/function/LongUnaryOperator;"(thread: JVMThread, args: [java_util_function_LongUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_function_LongUnaryOperator) => void): void;
    "andThen(Ljava/util/function/LongUnaryOperator;)Ljava/util/function/LongUnaryOperator;"(thread: JVMThread, args: [java_util_function_LongUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_function_LongUnaryOperator) => void): void;
    "lambda$andThen$1(Ljava/util/function/LongUnaryOperator;J)J"(thread: JVMThread, args: [java_util_function_LongUnaryOperator, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "lambda$compose$0(Ljava/util/function/LongUnaryOperator;J)J"(thread: JVMThread, args: [java_util_function_LongUnaryOperator, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export interface java_util_function_LongPredicate extends java_lang_Object {
    "test(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "and(Ljava/util/function/LongPredicate;)Ljava/util/function/LongPredicate;"(thread: JVMThread, args: [java_util_function_LongPredicate], cb?: (e?: java_lang_Throwable, rv?: java_util_function_LongPredicate) => void): void;
    "negate()Ljava/util/function/LongPredicate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_LongPredicate) => void): void;
    "or(Ljava/util/function/LongPredicate;)Ljava/util/function/LongPredicate;"(thread: JVMThread, args: [java_util_function_LongPredicate], cb?: (e?: java_lang_Throwable, rv?: java_util_function_LongPredicate) => void): void;
    "lambda$or$2(Ljava/util/function/LongPredicate;J)Z"(thread: JVMThread, args: [java_util_function_LongPredicate, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$negate$1(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$and$0(Ljava/util/function/LongPredicate;J)Z"(thread: JVMThread, args: [java_util_function_LongPredicate, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_ToIntFunction extends java_lang_Object {
    "applyAsInt(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_stream_IntStream extends java_lang_Object, java_util_stream_BaseStream {
    "filter(Ljava/util/function/IntPredicate;)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [java_util_function_IntPredicate], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "map(Ljava/util/function/IntUnaryOperator;)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [java_util_function_IntUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "mapToObj(Ljava/util/function/IntFunction;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_util_function_IntFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "mapToLong(Ljava/util/function/IntToLongFunction;)Ljava/util/stream/LongStream;"(thread: JVMThread, args: [java_util_function_IntToLongFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "mapToDouble(Ljava/util/function/IntToDoubleFunction;)Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: [java_util_function_IntToDoubleFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "flatMap(Ljava/util/function/IntFunction;)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [java_util_function_IntFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "distinct()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "sorted()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "peek(Ljava/util/function/IntConsumer;)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "limit(J)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "skip(J)Ljava/util/stream/IntStream;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "forEach(Ljava/util/function/IntConsumer;)V"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachOrdered(Ljava/util/function/IntConsumer;)V"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "toArray()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    "reduce(ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "reduce(Ljava/util/function/IntBinaryOperator;)Ljava/util/OptionalInt;"(thread: JVMThread, args: [java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalInt) => void): void;
    "collect(Ljava/util/function/Supplier;Ljava/util/function/ObjIntConsumer;Ljava/util/function/BiConsumer;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_function_Supplier, java_util_function_ObjIntConsumer, java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "sum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "min()Ljava/util/OptionalInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalInt) => void): void;
    "max()Ljava/util/OptionalInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalInt) => void): void;
    "count()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "average()Ljava/util/OptionalDouble;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalDouble) => void): void;
    "summaryStatistics()Ljava/util/IntSummaryStatistics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_IntSummaryStatistics) => void): void;
    "anyMatch(Ljava/util/function/IntPredicate;)Z"(thread: JVMThread, args: [java_util_function_IntPredicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "allMatch(Ljava/util/function/IntPredicate;)Z"(thread: JVMThread, args: [java_util_function_IntPredicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "noneMatch(Ljava/util/function/IntPredicate;)Z"(thread: JVMThread, args: [java_util_function_IntPredicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "findFirst()Ljava/util/OptionalInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalInt) => void): void;
    "findAny()Ljava/util/OptionalInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalInt) => void): void;
    "asLongStream()Ljava/util/stream/LongStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_LongStream) => void): void;
    "asDoubleStream()Ljava/util/stream/DoubleStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_DoubleStream) => void): void;
    "boxed()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "sequential()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "parallel()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "iterator()Ljava/util/PrimitiveIterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_PrimitiveIterator$OfInt) => void): void;
    "spliterator()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    "parallel()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "sequential()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "isParallel()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "unordered()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "onClose(Ljava/lang/Runnable;)Ljava/util/stream/BaseStream;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_IntSupplier extends java_lang_Object {
    "getAsInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_stream_IntStream$Builder extends java_lang_Object, java_util_function_IntConsumer {
    "accept(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "add(I)Ljava/util/stream/IntStream$Builder;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream$Builder) => void): void;
    "build()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "andThen(Ljava/util/function/IntConsumer;)Ljava/util/function/IntConsumer;"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_IntConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/IntConsumer;I)V"(thread: JVMThread, args: [java_util_function_IntConsumer, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_PrimitiveIterator$OfInt extends java_lang_Object, java_util_PrimitiveIterator {
    "nextInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/IntConsumer;)V"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "next()Ljava/lang/Integer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "next()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Integer extends java_lang_Number implements java_lang_Comparable {
    public unbox(): number;
    public static box(val: number): java_lang_Integer;
    public static "java/lang/Integer/MIN_VALUE": number;
    public static "java/lang/Integer/MAX_VALUE": number;
    public static "java/lang/Integer/TYPE": java_lang_Class;
    public static "java/lang/Integer/digits": JVMArray<number>;
    public static "java/lang/Integer/DigitTens": JVMArray<number>;
    public static "java/lang/Integer/DigitOnes": JVMArray<number>;
    public static "java/lang/Integer/sizeTable": JVMArray<number>;
    public "java/lang/Integer/value": number;
    public static "java/lang/Integer/SIZE": number;
    public static "java/lang/Integer/BYTES": number;
    public static "java/lang/Integer/serialVersionUID": Long;
    public static "toString(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Integer/toString(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toUnsignedString(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Integer/toUnsignedString(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toHexString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Integer/toHexString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toOctalString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Integer/toOctalString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toBinaryString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Integer/toBinaryString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toUnsignedString0(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Integer/toUnsignedString0(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "formatUnsignedInt(II[CII)I"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/formatUnsignedInt(II[CII)I"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Integer/toString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toUnsignedString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Integer/toUnsignedString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getChars(II[C)V"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Integer/getChars(II[C)V"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "stringSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/stringSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "parseInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/parseInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "parseInt(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/parseInt(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "parseUnsignedInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/parseUnsignedInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "parseUnsignedInt(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/parseUnsignedInt(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "valueOf(Ljava/lang/String;I)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "java/lang/Integer/valueOf(Ljava/lang/String;I)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "java/lang/Integer/valueOf(Ljava/lang/String;)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "valueOf(I)Ljava/lang/Integer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "java/lang/Integer/valueOf(I)Ljava/lang/Integer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Integer/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Integer/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Integer/byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Integer/shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Integer/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/Integer/longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Integer/floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Integer/doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Integer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Integer/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hashCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/hashCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Integer/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getInteger(Ljava/lang/String;)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "java/lang/Integer/getInteger(Ljava/lang/String;)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "getInteger(Ljava/lang/String;I)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "java/lang/Integer/getInteger(Ljava/lang/String;I)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "getInteger(Ljava/lang/String;Ljava/lang/Integer;)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String, java_lang_Integer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "java/lang/Integer/getInteger(Ljava/lang/String;Ljava/lang/Integer;)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String, java_lang_Integer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "decode(Ljava/lang/String;)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public static "java/lang/Integer/decode(Ljava/lang/String;)Ljava/lang/Integer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    public "compareTo(Ljava/lang/Integer;)I"(thread: JVMThread, args: [java_lang_Integer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Integer/compareTo(Ljava/lang/Integer;)I"(thread: JVMThread, args: [java_lang_Integer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/compare(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compareUnsigned(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/compareUnsigned(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toUnsignedLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Integer/toUnsignedLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "divideUnsigned(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/divideUnsigned(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "remainderUnsigned(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/remainderUnsigned(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "highestOneBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/highestOneBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "lowestOneBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/lowestOneBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "numberOfLeadingZeros(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/numberOfLeadingZeros(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "numberOfTrailingZeros(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/numberOfTrailingZeros(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "bitCount(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/bitCount(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "rotateLeft(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/rotateLeft(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "rotateRight(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/rotateRight(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "reverse(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/reverse(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "signum(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/signum(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "reverseBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/reverseBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sum(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/sum(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "max(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/max(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "min(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Integer/min(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Integer/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Integer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_IntSummaryStatistics extends java_lang_Object implements java_util_function_IntConsumer {
    public "java/util/IntSummaryStatistics/count": Long;
    public "java/util/IntSummaryStatistics/sum": Long;
    public "java/util/IntSummaryStatistics/min": number;
    public "java/util/IntSummaryStatistics/max": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/IntSummaryStatistics/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "accept(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/IntSummaryStatistics/accept(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "combine(Ljava/util/IntSummaryStatistics;)V"(thread: JVMThread, args: [java_util_IntSummaryStatistics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/IntSummaryStatistics/combine(Ljava/util/IntSummaryStatistics;)V"(thread: JVMThread, args: [java_util_IntSummaryStatistics], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/IntSummaryStatistics/getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getSum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/IntSummaryStatistics/getSum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMin()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/IntSummaryStatistics/getMin()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMax()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/IntSummaryStatistics/getMax()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/IntSummaryStatistics/getAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/IntSummaryStatistics/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "andThen(Ljava/util/function/IntConsumer;)Ljava/util/function/IntConsumer;"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable, rv?: java_util_function_IntConsumer) => void): void;
    "lambda$andThen$0(Ljava/util/function/IntConsumer;I)V"(thread: JVMThread, args: [java_util_function_IntConsumer, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_ObjIntConsumer extends java_lang_Object {
    "accept(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_OptionalInt extends java_lang_Object {
    public static "java/util/OptionalInt/EMPTY": java_util_OptionalInt;
    public "java/util/OptionalInt/isPresent": number;
    public "java/util/OptionalInt/value": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/OptionalInt/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "empty()Ljava/util/OptionalInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalInt) => void): void;
    public static "java/util/OptionalInt/empty()Ljava/util/OptionalInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalInt) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/OptionalInt/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "of(I)Ljava/util/OptionalInt;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalInt) => void): void;
    public static "java/util/OptionalInt/of(I)Ljava/util/OptionalInt;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_OptionalInt) => void): void;
    public "getAsInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalInt/getAsInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalInt/isPresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ifPresent(Ljava/util/function/IntConsumer;)V"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/OptionalInt/ifPresent(Ljava/util/function/IntConsumer;)V"(thread: JVMThread, args: [java_util_function_IntConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "orElse(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalInt/orElse(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "orElseGet(Ljava/util/function/IntSupplier;)I"(thread: JVMThread, args: [java_util_function_IntSupplier], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalInt/orElseGet(Ljava/util/function/IntSupplier;)I"(thread: JVMThread, args: [java_util_function_IntSupplier], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "orElseThrow(Ljava/util/function/Supplier;)I"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalInt/orElseThrow(Ljava/util/function/Supplier;)I"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalInt/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/OptionalInt/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/OptionalInt/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/OptionalInt/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_function_IntBinaryOperator extends java_lang_Object {
    "applyAsInt(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_IntToDoubleFunction extends java_lang_Object {
    "applyAsDouble(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_IntToLongFunction extends java_lang_Object {
    "applyAsLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export interface java_util_function_IntUnaryOperator extends java_lang_Object {
    "applyAsInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "compose(Ljava/util/function/IntUnaryOperator;)Ljava/util/function/IntUnaryOperator;"(thread: JVMThread, args: [java_util_function_IntUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_function_IntUnaryOperator) => void): void;
    "andThen(Ljava/util/function/IntUnaryOperator;)Ljava/util/function/IntUnaryOperator;"(thread: JVMThread, args: [java_util_function_IntUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_util_function_IntUnaryOperator) => void): void;
    "lambda$andThen$1(Ljava/util/function/IntUnaryOperator;I)I"(thread: JVMThread, args: [java_util_function_IntUnaryOperator, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$compose$0(Ljava/util/function/IntUnaryOperator;I)I"(thread: JVMThread, args: [java_util_function_IntUnaryOperator, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_IntPredicate extends java_lang_Object {
    "test(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "and(Ljava/util/function/IntPredicate;)Ljava/util/function/IntPredicate;"(thread: JVMThread, args: [java_util_function_IntPredicate], cb?: (e?: java_lang_Throwable, rv?: java_util_function_IntPredicate) => void): void;
    "negate()Ljava/util/function/IntPredicate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_IntPredicate) => void): void;
    "or(Ljava/util/function/IntPredicate;)Ljava/util/function/IntPredicate;"(thread: JVMThread, args: [java_util_function_IntPredicate], cb?: (e?: java_lang_Throwable, rv?: java_util_function_IntPredicate) => void): void;
    "lambda$or$2(Ljava/util/function/IntPredicate;I)Z"(thread: JVMThread, args: [java_util_function_IntPredicate, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$negate$1(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$and$0(Ljava/util/function/IntPredicate;I)Z"(thread: JVMThread, args: [java_util_function_IntPredicate, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_stream_BaseStream extends java_lang_Object, java_lang_AutoCloseable {
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "isParallel()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "sequential()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "parallel()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "unordered()Ljava/util/stream/BaseStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "onClose(Ljava/lang/Runnable;)Ljava/util/stream/BaseStream;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_BaseStream) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_lang_AutoCloseable extends java_lang_Object {
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_Spliterator extends java_lang_Object {
    "tryAdvance(Ljava/util/function/Consumer;)Z"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "trySplit()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "estimateSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getExactSizeIfKnown()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "characteristics()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hasCharacteristics(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getComparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
  }
  export interface java_util_function_Predicate extends java_lang_Object {
    "test(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "and(Ljava/util/function/Predicate;)Ljava/util/function/Predicate;"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Predicate) => void): void;
    "negate()Ljava/util/function/Predicate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Predicate) => void): void;
    "or(Ljava/util/function/Predicate;)Ljava/util/function/Predicate;"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Predicate) => void): void;
    "lambda$or$2(Ljava/util/function/Predicate;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_util_function_Predicate, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$negate$1(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lambda$and$0(Ljava/util/function/Predicate;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_util_function_Predicate, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_Iterator extends java_lang_Object {
    "hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "next()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_lang_Iterable extends java_lang_Object {
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
  }
  export interface java_util_Set extends java_lang_Object, java_util_Collection {
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
  }
  export class sun_reflect_NativeConstructorAccessorImpl extends sun_reflect_ConstructorAccessorImpl {
    public "sun/reflect/NativeConstructorAccessorImpl/c": java_lang_reflect_Constructor;
    public "sun/reflect/NativeConstructorAccessorImpl/parent": sun_reflect_DelegatingConstructorAccessorImpl;
    public "sun/reflect/NativeConstructorAccessorImpl/numInvocations": number;
    public "<init>(Ljava/lang/reflect/Constructor;)V"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/NativeConstructorAccessorImpl/<init>(Ljava/lang/reflect/Constructor;)V"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable) => void): void;
    public "newInstance([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/reflect/NativeConstructorAccessorImpl/newInstance([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setParent(Lsun/reflect/DelegatingConstructorAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_DelegatingConstructorAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/NativeConstructorAccessorImpl/setParent(Lsun/reflect/DelegatingConstructorAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_DelegatingConstructorAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public static "newInstance0(Ljava/lang/reflect/Constructor;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_reflect_Constructor, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/reflect/NativeConstructorAccessorImpl/newInstance0(Ljava/lang/reflect/Constructor;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_reflect_Constructor, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_reflect_DelegatingConstructorAccessorImpl extends sun_reflect_ConstructorAccessorImpl {
    public "sun/reflect/DelegatingConstructorAccessorImpl/delegate": sun_reflect_ConstructorAccessorImpl;
    public "<init>(Lsun/reflect/ConstructorAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_ConstructorAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/DelegatingConstructorAccessorImpl/<init>(Lsun/reflect/ConstructorAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_ConstructorAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "newInstance([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/reflect/DelegatingConstructorAccessorImpl/newInstance([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setDelegate(Lsun/reflect/ConstructorAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_ConstructorAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/DelegatingConstructorAccessorImpl/setDelegate(Lsun/reflect/ConstructorAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_ConstructorAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_ConstructorAccessorImpl extends sun_reflect_MagicAccessorImpl implements sun_reflect_ConstructorAccessor {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/ConstructorAccessorImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "newInstance([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/reflect/ConstructorAccessorImpl/newInstance([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface sun_reflect_ConstructorAccessor extends java_lang_Object {
    "newInstance([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_reflect_MagicAccessorImpl extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/MagicAccessorImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_NativeMethodAccessorImpl extends sun_reflect_MethodAccessorImpl {
    public "sun/reflect/NativeMethodAccessorImpl/method": java_lang_reflect_Method;
    public "sun/reflect/NativeMethodAccessorImpl/parent": sun_reflect_DelegatingMethodAccessorImpl;
    public "sun/reflect/NativeMethodAccessorImpl/numInvocations": number;
    public "<init>(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/NativeMethodAccessorImpl/<init>(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "invoke(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/reflect/NativeMethodAccessorImpl/invoke(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setParent(Lsun/reflect/DelegatingMethodAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_DelegatingMethodAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/NativeMethodAccessorImpl/setParent(Lsun/reflect/DelegatingMethodAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_DelegatingMethodAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public static "invoke0(Ljava/lang/reflect/Method;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/reflect/NativeMethodAccessorImpl/invoke0(Ljava/lang/reflect/Method;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_reflect_DelegatingMethodAccessorImpl extends sun_reflect_MethodAccessorImpl {
    public "sun/reflect/DelegatingMethodAccessorImpl/delegate": sun_reflect_MethodAccessorImpl;
    public "<init>(Lsun/reflect/MethodAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_MethodAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/DelegatingMethodAccessorImpl/<init>(Lsun/reflect/MethodAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_MethodAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "invoke(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/reflect/DelegatingMethodAccessorImpl/invoke(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setDelegate(Lsun/reflect/MethodAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_MethodAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/DelegatingMethodAccessorImpl/setDelegate(Lsun/reflect/MethodAccessorImpl;)V"(thread: JVMThread, args: [sun_reflect_MethodAccessorImpl], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_MethodAccessorImpl extends sun_reflect_MagicAccessorImpl implements sun_reflect_MethodAccessor {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/MethodAccessorImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "invoke(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/reflect/MethodAccessorImpl/invoke(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface sun_reflect_MethodAccessor extends java_lang_Object {
    "invoke(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_nio_ch_sctp_SctpServerChannelImpl extends com_sun_nio_sctp_SctpServerChannel implements sun_nio_ch_SelChImpl {
    public "sun/nio/ch/sctp/SctpServerChannelImpl/fd": java_io_FileDescriptor;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/fdVal": number;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/thread": Long;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/lock": java_lang_Object;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/stateLock": java_lang_Object;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/state": sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/port": number;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/localAddresses": java_util_HashSet;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/wildcard": number;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl/$assertionsDisabled": number;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(Ljava/net/SocketAddress;I)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/bind(Ljava/net/SocketAddress;I)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "bindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/bindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "unbindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/unbindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "bindUnbindAddress(Ljava/net/InetAddress;Z)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/bindUnbindAddress(Ljava/net/InetAddress;Z)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "acceptCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/acceptCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "accept()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/accept()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOption(Lcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/setOption(Lcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "getOption(Lcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/getOption(Lcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getAllLocalAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl/getAllLocalAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "accept0(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl/accept0(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface com_sun_nio_sctp_SctpSocketOption extends java_lang_Object, java_net_SocketOption {
    "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "type()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
  }
  export interface java_net_SocketOption extends java_lang_Object {
    "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "type()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
  }
  export class sun_nio_ch_SelectionKeyImpl extends java_nio_channels_spi_AbstractSelectionKey {
    public "sun/nio/ch/SelectionKeyImpl/channel": sun_nio_ch_SelChImpl;
    public "sun/nio/ch/SelectionKeyImpl/selector": sun_nio_ch_SelectorImpl;
    public "sun/nio/ch/SelectionKeyImpl/index": number;
    public "sun/nio/ch/SelectionKeyImpl/interestOps": number;
    public "sun/nio/ch/SelectionKeyImpl/readyOps": number;
    public "<init>(Lsun/nio/ch/SelChImpl;Lsun/nio/ch/SelectorImpl;)V"(thread: JVMThread, args: [sun_nio_ch_SelChImpl, sun_nio_ch_SelectorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/<init>(Lsun/nio/ch/SelChImpl;Lsun/nio/ch/SelectorImpl;)V"(thread: JVMThread, args: [sun_nio_ch_SelChImpl, sun_nio_ch_SelectorImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "channel()Ljava/nio/channels/SelectableChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectableChannel) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/channel()Ljava/nio/channels/SelectableChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectableChannel) => void): void;
    public "selector()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/selector()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
    public "getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/setIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureValid()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/ensureValid()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interestOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/interestOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "interestOps(I)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/interestOps(I)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "readyOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/readyOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nioReadyOps(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/nioReadyOps(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "nioReadyOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/nioReadyOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nioInterestOps(I)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/nioInterestOps(I)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "nioInterestOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectionKeyImpl/nioInterestOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_channels_SelectionKey extends java_lang_Object {
    public static "java/nio/channels/SelectionKey/OP_READ": number;
    public static "java/nio/channels/SelectionKey/OP_WRITE": number;
    public static "java/nio/channels/SelectionKey/OP_CONNECT": number;
    public static "java/nio/channels/SelectionKey/OP_ACCEPT": number;
    public "java/nio/channels/SelectionKey/attachment": java_lang_Object;
    public static "java/nio/channels/SelectionKey/attachmentUpdater": java_util_concurrent_atomic_AtomicReferenceFieldUpdater;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/SelectionKey/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "channel()Ljava/nio/channels/SelectableChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectableChannel) => void): void;
    public "java/nio/channels/SelectionKey/channel()Ljava/nio/channels/SelectableChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectableChannel) => void): void;
    public "selector()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
    public "java/nio/channels/SelectionKey/selector()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectionKey/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "cancel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/SelectionKey/cancel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interestOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectionKey/interestOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "interestOps(I)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "java/nio/channels/SelectionKey/interestOps(I)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "readyOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectionKey/readyOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isReadable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectionKey/isReadable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isWritable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectionKey/isWritable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConnectable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectionKey/isConnectable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAcceptable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectionKey/isAcceptable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "attach(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/channels/SelectionKey/attach(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "attachment()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/channels/SelectionKey/attachment()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/channels/SelectionKey/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_atomic_AtomicReferenceFieldUpdater extends java_lang_Object {
    public static "newUpdater(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/String;)Ljava/util/concurrent/atomic/AtomicReferenceFieldUpdater;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicReferenceFieldUpdater) => void): void;
    public static "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/newUpdater(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/String;)Ljava/util/concurrent/atomic/AtomicReferenceFieldUpdater;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicReferenceFieldUpdater) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "compareAndSet(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/compareAndSet(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "weakCompareAndSet(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/weakCompareAndSet(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "set(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/set(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "lazySet(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/lazySet(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getAndSet(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/getAndSet(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getAndUpdate(Ljava/lang/Object;Ljava/util/function/UnaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/getAndUpdate(Ljava/lang/Object;Ljava/util/function/UnaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "updateAndGet(Ljava/lang/Object;Ljava/util/function/UnaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/updateAndGet(Ljava/lang/Object;Ljava/util/function/UnaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getAndAccumulate(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BinaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/getAndAccumulate(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BinaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "accumulateAndGet(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BinaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/atomic/AtomicReferenceFieldUpdater/accumulateAndGet(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BinaryOperator;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BinaryOperator], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_nio_channels_Selector extends java_lang_Object implements java_io_Closeable {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/Selector/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
    public static "java/nio/channels/Selector/open()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
    public "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/Selector/isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public "java/nio/channels/Selector/provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public "keys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/nio/channels/Selector/keys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "selectedKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/nio/channels/Selector/selectedKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "selectNow()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/Selector/selectNow()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "select(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/Selector/select(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "select()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/Selector/select()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "wakeup()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
    public "java/nio/channels/Selector/wakeup()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/Selector/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_io_Closeable extends java_lang_Object, java_lang_AutoCloseable {
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_channels_SelectableChannel extends java_nio_channels_spi_AbstractInterruptibleChannel implements java_nio_channels_Channel {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/SelectableChannel/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public "java/nio/channels/SelectableChannel/provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public "validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectableChannel/validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRegistered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectableChannel/isRegistered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keyFor(Ljava/nio/channels/Selector;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "java/nio/channels/SelectableChannel/keyFor(Ljava/nio/channels/Selector;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "register(Ljava/nio/channels/Selector;ILjava/lang/Object;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "java/nio/channels/SelectableChannel/register(Ljava/nio/channels/Selector;ILjava/lang/Object;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "register(Ljava/nio/channels/Selector;I)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "java/nio/channels/SelectableChannel/register(Ljava/nio/channels/Selector;I)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "configureBlocking(Z)Ljava/nio/channels/SelectableChannel;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectableChannel) => void): void;
    public "java/nio/channels/SelectableChannel/configureBlocking(Z)Ljava/nio/channels/SelectableChannel;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectableChannel) => void): void;
    public "isBlocking()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SelectableChannel/isBlocking()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "blockingLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/channels/SelectableChannel/blockingLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_nio_channels_Channel extends java_lang_Object, java_io_Closeable {
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_channels_spi_AbstractInterruptibleChannel extends java_lang_Object implements java_nio_channels_Channel, java_nio_channels_InterruptibleChannel {
    public "java/nio/channels/spi/AbstractInterruptibleChannel/closeLock": java_lang_Object;
    public "java/nio/channels/spi/AbstractInterruptibleChannel/open": number;
    public "java/nio/channels/spi/AbstractInterruptibleChannel/interruptor": sun_nio_ch_Interruptible;
    public "java/nio/channels/spi/AbstractInterruptibleChannel/interrupted": java_lang_Thread;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractInterruptibleChannel/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractInterruptibleChannel/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractInterruptibleChannel/implCloseChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/spi/AbstractInterruptibleChannel/isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "begin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractInterruptibleChannel/begin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "end(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractInterruptibleChannel/end(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "blockedOn(Lsun/nio/ch/Interruptible;)V"(thread: JVMThread, args: [sun_nio_ch_Interruptible], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/channels/spi/AbstractInterruptibleChannel/blockedOn(Lsun/nio/ch/Interruptible;)V"(thread: JVMThread, args: [sun_nio_ch_Interruptible], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/nio/channels/spi/AbstractInterruptibleChannel;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_channels_spi_AbstractInterruptibleChannel], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/nio/channels/spi/AbstractInterruptibleChannel/access$000(Ljava/nio/channels/spi/AbstractInterruptibleChannel;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_channels_spi_AbstractInterruptibleChannel], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$100(Ljava/nio/channels/spi/AbstractInterruptibleChannel;)Z"(thread: JVMThread, args: [java_nio_channels_spi_AbstractInterruptibleChannel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/channels/spi/AbstractInterruptibleChannel/access$100(Ljava/nio/channels/spi/AbstractInterruptibleChannel;)Z"(thread: JVMThread, args: [java_nio_channels_spi_AbstractInterruptibleChannel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$102(Ljava/nio/channels/spi/AbstractInterruptibleChannel;Z)Z"(thread: JVMThread, args: [java_nio_channels_spi_AbstractInterruptibleChannel, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/channels/spi/AbstractInterruptibleChannel/access$102(Ljava/nio/channels/spi/AbstractInterruptibleChannel;Z)Z"(thread: JVMThread, args: [java_nio_channels_spi_AbstractInterruptibleChannel, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$202(Ljava/nio/channels/spi/AbstractInterruptibleChannel;Ljava/lang/Thread;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_nio_channels_spi_AbstractInterruptibleChannel, java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "java/nio/channels/spi/AbstractInterruptibleChannel/access$202(Ljava/nio/channels/spi/AbstractInterruptibleChannel;Ljava/lang/Thread;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_nio_channels_spi_AbstractInterruptibleChannel, java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
  }
  export interface sun_nio_ch_Interruptible extends java_lang_Object {
    "interrupt(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_channels_InterruptibleChannel extends java_lang_Object, java_nio_channels_Channel {
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_nio_ch_SelectorImpl extends java_nio_channels_spi_AbstractSelector {
    public "sun/nio/ch/SelectorImpl/selectedKeys": java_util_Set;
    public "sun/nio/ch/SelectorImpl/keys": java_util_HashSet;
    public "sun/nio/ch/SelectorImpl/publicKeys": java_util_Set;
    public "sun/nio/ch/SelectorImpl/publicSelectedKeys": java_util_Set;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectorImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "keys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/SelectorImpl/keys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "selectedKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/SelectorImpl/selectedKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "doSelect(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectorImpl/doSelect(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lockAndDoSelect(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectorImpl/lockAndDoSelect(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "select(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectorImpl/select(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "select()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectorImpl/select()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "selectNow()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SelectorImpl/selectNow()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "implCloseSelector()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectorImpl/implCloseSelector()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectorImpl/implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "putEventOps(Lsun/nio/ch/SelectionKeyImpl;I)V"(thread: JVMThread, args: [sun_nio_ch_SelectionKeyImpl, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectorImpl/putEventOps(Lsun/nio/ch/SelectionKeyImpl;I)V"(thread: JVMThread, args: [sun_nio_ch_SelectionKeyImpl, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "register(Ljava/nio/channels/spi/AbstractSelectableChannel;ILjava/lang/Object;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_spi_AbstractSelectableChannel, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "sun/nio/ch/SelectorImpl/register(Ljava/nio/channels/spi/AbstractSelectableChannel;ILjava/lang/Object;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_spi_AbstractSelectableChannel, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "implRegister(Lsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectorImpl/implRegister(Lsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "processDeregisterQueue()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectorImpl/processDeregisterQueue()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implDereg(Lsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SelectorImpl/implDereg(Lsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "wakeup()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
    public "sun/nio/ch/SelectorImpl/wakeup()Ljava/nio/channels/Selector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Selector) => void): void;
  }
  export class java_nio_channels_spi_AbstractSelectableChannel extends java_nio_channels_SelectableChannel {
    public "java/nio/channels/spi/AbstractSelectableChannel/provider": java_nio_channels_spi_SelectorProvider;
    public "java/nio/channels/spi/AbstractSelectableChannel/keys": JVMArray<java_nio_channels_SelectionKey>;
    public "java/nio/channels/spi/AbstractSelectableChannel/keyCount": number;
    public "java/nio/channels/spi/AbstractSelectableChannel/keyLock": java_lang_Object;
    public "java/nio/channels/spi/AbstractSelectableChannel/regLock": java_lang_Object;
    public "java/nio/channels/spi/AbstractSelectableChannel/blocking": number;
    public static "java/nio/channels/spi/AbstractSelectableChannel/$assertionsDisabled": number;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public "addKey(Ljava/nio/channels/SelectionKey;)V"(thread: JVMThread, args: [java_nio_channels_SelectionKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/addKey(Ljava/nio/channels/SelectionKey;)V"(thread: JVMThread, args: [java_nio_channels_SelectionKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "findKey(Ljava/nio/channels/Selector;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/findKey(Ljava/nio/channels/Selector;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "removeKey(Ljava/nio/channels/SelectionKey;)V"(thread: JVMThread, args: [java_nio_channels_SelectionKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/removeKey(Ljava/nio/channels/SelectionKey;)V"(thread: JVMThread, args: [java_nio_channels_SelectionKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "haveValidKeys()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/haveValidKeys()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRegistered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/isRegistered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keyFor(Ljava/nio/channels/Selector;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/keyFor(Ljava/nio/channels/Selector;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "register(Ljava/nio/channels/Selector;ILjava/lang/Object;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/register(Ljava/nio/channels/Selector;ILjava/lang/Object;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_Selector, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "implCloseChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/implCloseChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isBlocking()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/isBlocking()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "blockingLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/blockingLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "configureBlocking(Z)Ljava/nio/channels/SelectableChannel;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectableChannel) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/configureBlocking(Z)Ljava/nio/channels/SelectableChannel;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectableChannel) => void): void;
    public "implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelectableChannel/implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/channels/spi/AbstractSelectableChannel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_channels_spi_AbstractSelector extends java_nio_channels_Selector {
    public "java/nio/channels/spi/AbstractSelector/selectorOpen": java_util_concurrent_atomic_AtomicBoolean;
    public "java/nio/channels/spi/AbstractSelector/provider": java_nio_channels_spi_SelectorProvider;
    public "java/nio/channels/spi/AbstractSelector/cancelledKeys": java_util_Set;
    public "java/nio/channels/spi/AbstractSelector/interruptor": sun_nio_ch_Interruptible;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelector/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "cancel(Ljava/nio/channels/SelectionKey;)V"(thread: JVMThread, args: [java_nio_channels_SelectionKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelector/cancel(Ljava/nio/channels/SelectionKey;)V"(thread: JVMThread, args: [java_nio_channels_SelectionKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelector/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseSelector()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelector/implCloseSelector()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/spi/AbstractSelector/isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public "java/nio/channels/spi/AbstractSelector/provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public "cancelledKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/nio/channels/spi/AbstractSelector/cancelledKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "register(Ljava/nio/channels/spi/AbstractSelectableChannel;ILjava/lang/Object;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_spi_AbstractSelectableChannel, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "java/nio/channels/spi/AbstractSelector/register(Ljava/nio/channels/spi/AbstractSelectableChannel;ILjava/lang/Object;)Ljava/nio/channels/SelectionKey;"(thread: JVMThread, args: [java_nio_channels_spi_AbstractSelectableChannel, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SelectionKey) => void): void;
    public "deregister(Ljava/nio/channels/spi/AbstractSelectionKey;)V"(thread: JVMThread, args: [java_nio_channels_spi_AbstractSelectionKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelector/deregister(Ljava/nio/channels/spi/AbstractSelectionKey;)V"(thread: JVMThread, args: [java_nio_channels_spi_AbstractSelectionKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "begin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelector/begin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelector/end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_atomic_AtomicBoolean extends java_lang_Object implements java_io_Serializable {
    public static "java/util/concurrent/atomic/AtomicBoolean/serialVersionUID": Long;
    public static "java/util/concurrent/atomic/AtomicBoolean/unsafe": sun_misc_Unsafe;
    public static "java/util/concurrent/atomic/AtomicBoolean/valueOffset": Long;
    public "java/util/concurrent/atomic/AtomicBoolean/value": number;
    public "<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicBoolean/<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicBoolean/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "get()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicBoolean/get()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareAndSet(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicBoolean/compareAndSet(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "weakCompareAndSet(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicBoolean/weakCompareAndSet(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "set(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicBoolean/set(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "lazySet(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicBoolean/lazySet(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAndSet(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicBoolean/getAndSet(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/atomic/AtomicBoolean/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/atomic/AtomicBoolean/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_channels_spi_AbstractSelectionKey extends java_nio_channels_SelectionKey {
    public "java/nio/channels/spi/AbstractSelectionKey/valid": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelectionKey/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/spi/AbstractSelectionKey/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelectionKey/invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "cancel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AbstractSelectionKey/cancel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_SctpChannel extends java_nio_channels_spi_AbstractSelectableChannel {
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/SctpChannel/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public static "com/sun/nio/sctp/SctpChannel/open()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public static "open(Ljava/net/SocketAddress;II)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public static "com/sun/nio/sctp/SctpChannel/open(Ljava/net/SocketAddress;II)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "com/sun/nio/sctp/SctpChannel/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "bind(Ljava/net/SocketAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "com/sun/nio/sctp/SctpChannel/bind(Ljava/net/SocketAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "bindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "com/sun/nio/sctp/SctpChannel/bindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "unbindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "com/sun/nio/sctp/SctpChannel/unbindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "connect(Ljava/net/SocketAddress;)Z"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/SctpChannel/connect(Ljava/net/SocketAddress;)Z"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "connect(Ljava/net/SocketAddress;II)Z"(thread: JVMThread, args: [java_net_SocketAddress, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/SctpChannel/connect(Ljava/net/SocketAddress;II)Z"(thread: JVMThread, args: [java_net_SocketAddress, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConnectionPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/SctpChannel/isConnectionPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finishConnect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/SctpChannel/finishConnect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAllLocalAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "com/sun/nio/sctp/SctpChannel/getAllLocalAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getRemoteAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "com/sun/nio/sctp/SctpChannel/getRemoteAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "shutdown()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "com/sun/nio/sctp/SctpChannel/shutdown()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "getOption(Lcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "com/sun/nio/sctp/SctpChannel/getOption(Lcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setOption(Lcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "com/sun/nio/sctp/SctpChannel/setOption(Lcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "com/sun/nio/sctp/SctpChannel/supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/SctpChannel/validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "receive(Ljava/nio/ByteBuffer;Ljava/lang/Object;Lcom/sun/nio/sctp/NotificationHandler;)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, com_sun_nio_sctp_NotificationHandler], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "com/sun/nio/sctp/SctpChannel/receive(Ljava/nio/ByteBuffer;Ljava/lang/Object;Lcom/sun/nio/sctp/NotificationHandler;)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, com_sun_nio_sctp_NotificationHandler], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "send(Ljava/nio/ByteBuffer;Lcom/sun/nio/sctp/MessageInfo;)I"(thread: JVMThread, args: [java_nio_ByteBuffer, com_sun_nio_sctp_MessageInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/SctpChannel/send(Ljava/nio/ByteBuffer;Lcom/sun/nio/sctp/MessageInfo;)I"(thread: JVMThread, args: [java_nio_ByteBuffer, com_sun_nio_sctp_MessageInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface com_sun_nio_sctp_NotificationHandler extends java_lang_Object {
    "handleNotification(Lcom/sun/nio/sctp/Notification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_Notification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
  }
  export interface com_sun_nio_sctp_Notification extends java_lang_Object {
    "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
  }
  export class com_sun_nio_sctp_HandlerResult extends java_lang_Enum {
    public static "com/sun/nio/sctp/HandlerResult/CONTINUE": com_sun_nio_sctp_HandlerResult;
    public static "com/sun/nio/sctp/HandlerResult/RETURN": com_sun_nio_sctp_HandlerResult;
    public static "com/sun/nio/sctp/HandlerResult/$VALUES": JVMArray<com_sun_nio_sctp_HandlerResult>;
    public static "values()[Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<com_sun_nio_sctp_HandlerResult>) => void): void;
    public static "com/sun/nio/sctp/HandlerResult/values()[Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<com_sun_nio_sctp_HandlerResult>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public static "com/sun/nio/sctp/HandlerResult/valueOf(Ljava/lang/String;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/HandlerResult/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "com/sun/nio/sctp/HandlerResult/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_MessageInfo extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/MessageInfo/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createOutgoing(Ljava/net/SocketAddress;I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public static "com/sun/nio/sctp/MessageInfo/createOutgoing(Ljava/net/SocketAddress;I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public static "createOutgoing(Lcom/sun/nio/sctp/Association;Ljava/net/SocketAddress;I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [com_sun_nio_sctp_Association, java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public static "com/sun/nio/sctp/MessageInfo/createOutgoing(Lcom/sun/nio/sctp/Association;Ljava/net/SocketAddress;I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [com_sun_nio_sctp_Association, java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "com/sun/nio/sctp/MessageInfo/address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "com/sun/nio/sctp/MessageInfo/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "bytes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/MessageInfo/bytes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isComplete()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/MessageInfo/isComplete()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "complete(Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "com/sun/nio/sctp/MessageInfo/complete(Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "isUnordered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/MessageInfo/isUnordered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unordered(Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "com/sun/nio/sctp/MessageInfo/unordered(Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "payloadProtocolID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/MessageInfo/payloadProtocolID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "payloadProtocolID(I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "com/sun/nio/sctp/MessageInfo/payloadProtocolID(I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "streamNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/MessageInfo/streamNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "streamNumber(I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "com/sun/nio/sctp/MessageInfo/streamNumber(I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "timeToLive()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "com/sun/nio/sctp/MessageInfo/timeToLive()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "timeToLive(J)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "com/sun/nio/sctp/MessageInfo/timeToLive(J)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
  }
  export class com_sun_nio_sctp_Association extends java_lang_Object {
    public "com/sun/nio/sctp/Association/associationID": number;
    public "com/sun/nio/sctp/Association/maxInStreams": number;
    public "com/sun/nio/sctp/Association/maxOutStreams": number;
    public "<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/Association/<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "associationID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/Association/associationID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "maxInboundStreams()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/Association/maxInboundStreams()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "maxOutboundStreams()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/Association/maxOutboundStreams()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_channels_spi_SelectorProvider extends java_lang_Object {
    public static "java/nio/channels/spi/SelectorProvider/lock": java_lang_Object;
    public static "java/nio/channels/spi/SelectorProvider/provider": java_nio_channels_spi_SelectorProvider;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/SelectorProvider/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "loadProviderFromProperty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/channels/spi/SelectorProvider/loadProviderFromProperty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "loadProviderAsService()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/channels/spi/SelectorProvider/loadProviderAsService()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public static "java/nio/channels/spi/SelectorProvider/provider()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public "openDatagramChannel()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "java/nio/channels/spi/SelectorProvider/openDatagramChannel()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "openDatagramChannel(Ljava/net/ProtocolFamily;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_ProtocolFamily], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "java/nio/channels/spi/SelectorProvider/openDatagramChannel(Ljava/net/ProtocolFamily;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_ProtocolFamily], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "openPipe()Ljava/nio/channels/Pipe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Pipe) => void): void;
    public "java/nio/channels/spi/SelectorProvider/openPipe()Ljava/nio/channels/Pipe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Pipe) => void): void;
    public "openSelector()Ljava/nio/channels/spi/AbstractSelector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AbstractSelector) => void): void;
    public "java/nio/channels/spi/SelectorProvider/openSelector()Ljava/nio/channels/spi/AbstractSelector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AbstractSelector) => void): void;
    public "openServerSocketChannel()Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "java/nio/channels/spi/SelectorProvider/openServerSocketChannel()Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "openSocketChannel()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "java/nio/channels/spi/SelectorProvider/openSocketChannel()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "inheritedChannel()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public "java/nio/channels/spi/SelectorProvider/inheritedChannel()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public static "access$000()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/channels/spi/SelectorProvider/access$000()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public static "java/nio/channels/spi/SelectorProvider/access$100()Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public static "access$200()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/channels/spi/SelectorProvider/access$200()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$102(Ljava/nio/channels/spi/SelectorProvider;)Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public static "java/nio/channels/spi/SelectorProvider/access$102(Ljava/nio/channels/spi/SelectorProvider;)Ljava/nio/channels/spi/SelectorProvider;"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_SelectorProvider) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/channels/spi/SelectorProvider/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_channels_SocketChannel extends java_nio_channels_spi_AbstractSelectableChannel implements java_nio_channels_ByteChannel, java_nio_channels_ScatteringByteChannel, java_nio_channels_GatheringByteChannel, java_nio_channels_NetworkChannel {
    public static "java/nio/channels/SocketChannel/$assertionsDisabled": number;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/SocketChannel/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public static "java/nio/channels/SocketChannel/open()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public static "open(Ljava/net/SocketAddress;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public static "java/nio/channels/SocketChannel/open(Ljava/net/SocketAddress;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SocketChannel/validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "java/nio/channels/SocketChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "java/nio/channels/SocketChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "shutdownInput()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "java/nio/channels/SocketChannel/shutdownInput()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "shutdownOutput()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "java/nio/channels/SocketChannel/shutdownOutput()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "socket()Ljava/net/Socket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Socket) => void): void;
    public "java/nio/channels/SocketChannel/socket()Ljava/net/Socket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Socket) => void): void;
    public "isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SocketChannel/isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConnectionPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SocketChannel/isConnectionPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "connect(Ljava/net/SocketAddress;)Z"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SocketChannel/connect(Ljava/net/SocketAddress;)Z"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finishConnect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SocketChannel/finishConnect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/nio/channels/SocketChannel/getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SocketChannel/read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/SocketChannel/read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "read([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/SocketChannel/read([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/SocketChannel/write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/SocketChannel/write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/SocketChannel/write([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/nio/channels/SocketChannel/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/SocketChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/SocketChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/channels/SocketChannel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
  }
  export class java_net_Socket extends java_lang_Object implements java_io_Closeable {
    public "java/net/Socket/created": number;
    public "java/net/Socket/bound": number;
    public "java/net/Socket/connected": number;
    public "java/net/Socket/closed": number;
    public "java/net/Socket/closeLock": java_lang_Object;
    public "java/net/Socket/shutIn": number;
    public "java/net/Socket/shutOut": number;
    public "java/net/Socket/impl": java_net_SocketImpl;
    public "java/net/Socket/oldImpl": number;
    public static "java/net/Socket/factory": java_net_SocketImplFactory;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/Proxy;)V"(thread: JVMThread, args: [java_net_Proxy], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>(Ljava/net/Proxy;)V"(thread: JVMThread, args: [java_net_Proxy], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;ILjava/net/InetAddress;I)V"(thread: JVMThread, args: [java_lang_String, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>(Ljava/lang/String;ILjava/net/InetAddress;I)V"(thread: JVMThread, args: [java_lang_String, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/InetAddress;ILjava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>(Ljava/net/InetAddress;ILjava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;IZ)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>(Ljava/lang/String;IZ)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/InetAddress;IZ)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>(Ljava/net/InetAddress;IZ)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/SocketAddress;Ljava/net/SocketAddress;Z)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/<init>(Ljava/net/SocketAddress;Ljava/net/SocketAddress;Z)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "createImpl(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/createImpl(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkOldImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/checkOldImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getImpl()Ljava/net/SocketImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketImpl) => void): void;
    public "java/net/Socket/getImpl()Ljava/net/SocketImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketImpl) => void): void;
    public "connect(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/connect(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/connect(Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/bind(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAddress(Ljava/net/InetAddress;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_InetAddress, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/checkAddress(Ljava/net/InetAddress;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_InetAddress, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "postAccept()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/postAccept()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCreated()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setCreated()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBound()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setBound()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setConnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setConnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInetAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/Socket/getInetAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/Socket/getLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRemoteSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/net/Socket/getRemoteSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "getLocalSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/net/Socket/getLocalSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "getChannel()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "java/net/Socket/getChannel()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/net/Socket/getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "java/net/Socket/getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "setTcpNoDelay(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setTcpNoDelay(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTcpNoDelay()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getTcpNoDelay()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSoLinger(ZI)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setSoLinger(ZI)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSoLinger()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getSoLinger()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/sendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOOBInline(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setOOBInline(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOOBInline()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getOOBInline()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSoTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setSoTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSoTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getSoTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSendBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setSendBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSendBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getSendBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReceiveBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setReceiveBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getReceiveBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getReceiveBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setKeepAlive(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setKeepAlive(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKeepAlive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getKeepAlive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setTrafficClass(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setTrafficClass(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTrafficClass()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getTrafficClass()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReuseAddress(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setReuseAddress(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getReuseAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/getReuseAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownInput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/shutdownInput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownOutput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/shutdownOutput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/Socket/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isClosed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/isClosed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInputShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/isInputShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOutputShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Socket/isOutputShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setSocketImplFactory(Ljava/net/SocketImplFactory;)V"(thread: JVMThread, args: [java_net_SocketImplFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/Socket/setSocketImplFactory(Ljava/net/SocketImplFactory;)V"(thread: JVMThread, args: [java_net_SocketImplFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPerformancePreferences(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Socket/setPerformancePreferences(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/Socket/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_OutputStream extends java_lang_Object implements java_io_Closeable, java_io_Flushable {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStream/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStream/write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStream/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_io_Flushable extends java_lang_Object {
    "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_net_SocketImplFactory extends java_lang_Object {
    "createSocketImpl()Ljava/net/SocketImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketImpl) => void): void;
  }
  export interface java_nio_channels_NetworkChannel extends java_lang_Object, java_nio_channels_Channel {
    "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_channels_GatheringByteChannel extends java_lang_Object, java_nio_channels_WritableByteChannel {
    "write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "write([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_channels_WritableByteChannel extends java_lang_Object, java_nio_channels_Channel {
    "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_channels_ScatteringByteChannel extends java_lang_Object, java_nio_channels_ReadableByteChannel {
    "read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "read([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_channels_ReadableByteChannel extends java_lang_Object, java_nio_channels_Channel {
    "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_channels_ByteChannel extends java_lang_Object, java_nio_channels_ReadableByteChannel, java_nio_channels_WritableByteChannel {
    "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_channels_ServerSocketChannel extends java_nio_channels_spi_AbstractSelectableChannel implements java_nio_channels_NetworkChannel {
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/ServerSocketChannel/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open()Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public static "java/nio/channels/ServerSocketChannel/open()Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/ServerSocketChannel/validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "java/nio/channels/ServerSocketChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;I)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "java/nio/channels/ServerSocketChannel/bind(Ljava/net/SocketAddress;I)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "java/nio/channels/ServerSocketChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "socket()Ljava/net/ServerSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ServerSocket) => void): void;
    public "java/nio/channels/ServerSocketChannel/socket()Ljava/net/ServerSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ServerSocket) => void): void;
    public "accept()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "java/nio/channels/ServerSocketChannel/accept()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/nio/channels/ServerSocketChannel/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/ServerSocketChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/ServerSocketChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
  }
  export class java_net_ServerSocket extends java_lang_Object implements java_io_Closeable {
    public "java/net/ServerSocket/created": number;
    public "java/net/ServerSocket/bound": number;
    public "java/net/ServerSocket/closed": number;
    public "java/net/ServerSocket/closeLock": java_lang_Object;
    public "java/net/ServerSocket/impl": java_net_SocketImpl;
    public "java/net/ServerSocket/oldImpl": number;
    public static "java/net/ServerSocket/factory": java_net_SocketImplFactory;
    public "<init>(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/<init>(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/<init>(IILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "getImpl()Ljava/net/SocketImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketImpl) => void): void;
    public "java/net/ServerSocket/getImpl()Ljava/net/SocketImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketImpl) => void): void;
    public "checkOldImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/checkOldImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/setImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "createImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/createImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/bind(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/bind(Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInetAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/ServerSocket/getInetAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/ServerSocket/getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLocalSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/net/ServerSocket/getLocalSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "accept()Ljava/net/Socket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Socket) => void): void;
    public "java/net/ServerSocket/accept()Ljava/net/Socket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Socket) => void): void;
    public "implAccept(Ljava/net/Socket;)V"(thread: JVMThread, args: [java_net_Socket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/implAccept(Ljava/net/Socket;)V"(thread: JVMThread, args: [java_net_Socket], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChannel()Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "java/net/ServerSocket/getChannel()Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/ServerSocket/isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isClosed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/ServerSocket/isClosed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSoTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/setSoTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSoTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/ServerSocket/getSoTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReuseAddress(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/setReuseAddress(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getReuseAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/ServerSocket/getReuseAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/ServerSocket/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setBound()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/setBound()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCreated()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/setCreated()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setSocketFactory(Ljava/net/SocketImplFactory;)V"(thread: JVMThread, args: [java_net_SocketImplFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/ServerSocket/setSocketFactory(Ljava/net/SocketImplFactory;)V"(thread: JVMThread, args: [java_net_SocketImplFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "setReceiveBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/setReceiveBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getReceiveBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/ServerSocket/getReceiveBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setPerformancePreferences(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ServerSocket/setPerformancePreferences(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/net/ServerSocket;)Ljava/net/SocketImpl;"(thread: JVMThread, args: [java_net_ServerSocket], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketImpl) => void): void;
    public static "java/net/ServerSocket/access$000(Ljava/net/ServerSocket;)Ljava/net/SocketImpl;"(thread: JVMThread, args: [java_net_ServerSocket], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketImpl) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/ServerSocket/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_channels_Pipe extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/Pipe/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "source()Ljava/nio/channels/Pipe$SourceChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Pipe$SourceChannel) => void): void;
    public "java/nio/channels/Pipe/source()Ljava/nio/channels/Pipe$SourceChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Pipe$SourceChannel) => void): void;
    public "sink()Ljava/nio/channels/Pipe$SinkChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Pipe$SinkChannel) => void): void;
    public "java/nio/channels/Pipe/sink()Ljava/nio/channels/Pipe$SinkChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Pipe$SinkChannel) => void): void;
    public static "open()Ljava/nio/channels/Pipe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Pipe) => void): void;
    public static "java/nio/channels/Pipe/open()Ljava/nio/channels/Pipe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Pipe) => void): void;
  }
  export class java_nio_channels_Pipe$SinkChannel extends java_nio_channels_spi_AbstractSelectableChannel implements java_nio_channels_WritableByteChannel, java_nio_channels_GatheringByteChannel {
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/Pipe$SinkChannel/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/Pipe$SinkChannel/validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "write([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class java_nio_channels_Pipe$SourceChannel extends java_nio_channels_spi_AbstractSelectableChannel implements java_nio_channels_ReadableByteChannel, java_nio_channels_ScatteringByteChannel {
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/Pipe$SourceChannel/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/Pipe$SourceChannel/validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "read([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export interface java_net_ProtocolFamily extends java_lang_Object {
    "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_nio_channels_DatagramChannel extends java_nio_channels_spi_AbstractSelectableChannel implements java_nio_channels_ByteChannel, java_nio_channels_ScatteringByteChannel, java_nio_channels_GatheringByteChannel, java_nio_channels_MulticastChannel {
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/DatagramChannel/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public static "java/nio/channels/DatagramChannel/open()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public static "open(Ljava/net/ProtocolFamily;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_ProtocolFamily], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public static "java/nio/channels/DatagramChannel/open(Ljava/net/ProtocolFamily;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_ProtocolFamily], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/DatagramChannel/validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "java/nio/channels/DatagramChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "java/nio/channels/DatagramChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "socket()Ljava/net/DatagramSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_DatagramSocket) => void): void;
    public "java/nio/channels/DatagramChannel/socket()Ljava/net/DatagramSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_DatagramSocket) => void): void;
    public "isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/DatagramChannel/isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "connect(Ljava/net/SocketAddress;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "java/nio/channels/DatagramChannel/connect(Ljava/net/SocketAddress;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "disconnect()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "java/nio/channels/DatagramChannel/disconnect()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/nio/channels/DatagramChannel/getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "receive(Ljava/nio/ByteBuffer;)Ljava/net/SocketAddress;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/nio/channels/DatagramChannel/receive(Ljava/nio/ByteBuffer;)Ljava/net/SocketAddress;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "send(Ljava/nio/ByteBuffer;Ljava/net/SocketAddress;)I"(thread: JVMThread, args: [java_nio_ByteBuffer, java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/DatagramChannel/send(Ljava/nio/ByteBuffer;Ljava/net/SocketAddress;)I"(thread: JVMThread, args: [java_nio_ByteBuffer, java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/DatagramChannel/read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/DatagramChannel/read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "read([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/DatagramChannel/read([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/DatagramChannel/write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/DatagramChannel/write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/DatagramChannel/write([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/nio/channels/DatagramChannel/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/DatagramChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/DatagramChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    "join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
  }
  export class java_nio_channels_MembershipKey extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/MembershipKey/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/MembershipKey/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/MembershipKey/drop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "block(Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "java/nio/channels/MembershipKey/block(Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "unblock(Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "java/nio/channels/MembershipKey/unblock(Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "channel()Ljava/nio/channels/MulticastChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MulticastChannel) => void): void;
    public "java/nio/channels/MembershipKey/channel()Ljava/nio/channels/MulticastChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MulticastChannel) => void): void;
    public "group()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/nio/channels/MembershipKey/group()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "networkInterface()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public "java/nio/channels/MembershipKey/networkInterface()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public "sourceAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/nio/channels/MembershipKey/sourceAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
  }
  export class java_net_DatagramSocket extends java_lang_Object implements java_io_Closeable {
    public "java/net/DatagramSocket/created": number;
    public "java/net/DatagramSocket/bound": number;
    public "java/net/DatagramSocket/closed": number;
    public "java/net/DatagramSocket/closeLock": java_lang_Object;
    public "java/net/DatagramSocket/impl": java_net_DatagramSocketImpl;
    public "java/net/DatagramSocket/oldImpl": number;
    public "java/net/DatagramSocket/explicitFilter": number;
    public "java/net/DatagramSocket/bytesLeftToFilter": number;
    public static "java/net/DatagramSocket/ST_NOT_CONNECTED": number;
    public static "java/net/DatagramSocket/ST_CONNECTED": number;
    public static "java/net/DatagramSocket/ST_CONNECTED_NO_IMPL": number;
    public "java/net/DatagramSocket/connectState": number;
    public "java/net/DatagramSocket/connectedAddress": java_net_InetAddress;
    public "java/net/DatagramSocket/connectedPort": number;
    public static "java/net/DatagramSocket/implClass": java_lang_Class;
    public static "java/net/DatagramSocket/factory": java_net_DatagramSocketImplFactory;
    public "connectInternal(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/connectInternal(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/DatagramSocketImpl;)V"(thread: JVMThread, args: [java_net_DatagramSocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/<init>(Ljava/net/DatagramSocketImpl;)V"(thread: JVMThread, args: [java_net_DatagramSocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/<init>(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/<init>(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkOldImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/checkOldImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "createImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/createImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getImpl()Ljava/net/DatagramSocketImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_DatagramSocketImpl) => void): void;
    public "java/net/DatagramSocket/getImpl()Ljava/net/DatagramSocketImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_DatagramSocketImpl) => void): void;
    public "bind(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/bind(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAddress(Ljava/net/InetAddress;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_InetAddress, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/checkAddress(Ljava/net/InetAddress;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_InetAddress, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/connect(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "disconnect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/disconnect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInetAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/DatagramSocket/getInetAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRemoteSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/net/DatagramSocket/getRemoteSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "getLocalSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/net/DatagramSocket/getLocalSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "send(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/send(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "receive(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/receive(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkFiltering(Ljava/net/DatagramPacket;)Z"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/checkFiltering(Ljava/net/DatagramPacket;)Z"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/DatagramSocket/getLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSoTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/setSoTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSoTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/getSoTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSendBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/setSendBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSendBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/getSendBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReceiveBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/setReceiveBufferSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getReceiveBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/getReceiveBufferSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReuseAddress(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/setReuseAddress(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getReuseAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/getReuseAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setBroadcast(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/setBroadcast(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBroadcast()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/getBroadcast()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setTrafficClass(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/setTrafficClass(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTrafficClass()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/getTrafficClass()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocket/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isClosed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocket/isClosed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getChannel()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "java/net/DatagramSocket/getChannel()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public static "setDatagramSocketImplFactory(Ljava/net/DatagramSocketImplFactory;)V"(thread: JVMThread, args: [java_net_DatagramSocketImplFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/DatagramSocket/setDatagramSocketImplFactory(Ljava/net/DatagramSocketImplFactory;)V"(thread: JVMThread, args: [java_net_DatagramSocketImplFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/DatagramSocket/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_net_DatagramSocketImplFactory extends java_lang_Object {
    "createDatagramSocketImpl()Ljava/net/DatagramSocketImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_DatagramSocketImpl) => void): void;
  }
  export class java_net_DatagramSocketImpl extends java_lang_Object implements java_net_SocketOptions {
    public "java/net/DatagramSocketImpl/localPort": number;
    public "java/net/DatagramSocketImpl/fd": java_io_FileDescriptor;
    public "java/net/DatagramSocketImpl/socket": java_net_DatagramSocket;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dataAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocketImpl/dataAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDatagramSocket(Ljava/net/DatagramSocket;)V"(thread: JVMThread, args: [java_net_DatagramSocket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/setDatagramSocket(Ljava/net/DatagramSocket;)V"(thread: JVMThread, args: [java_net_DatagramSocket], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDatagramSocket()Ljava/net/DatagramSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_DatagramSocket) => void): void;
    public "java/net/DatagramSocketImpl/getDatagramSocket()Ljava/net/DatagramSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_DatagramSocket) => void): void;
    public "create()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/create()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/bind(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "send(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/send(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "disconnect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/disconnect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "peek(Ljava/net/InetAddress;)I"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocketImpl/peek(Ljava/net/InetAddress;)I"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peekData(Ljava/net/DatagramPacket;)I"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocketImpl/peekData(Ljava/net/DatagramPacket;)I"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "receive(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/receive(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTTL(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/setTTL(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTTL()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocketImpl/getTTL()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setTimeToLive(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/setTimeToLive(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTimeToLive()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocketImpl/getTimeToLive()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "join(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/join(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "leave(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/leave(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "joinGroup(Ljava/net/SocketAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/joinGroup(Ljava/net/SocketAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "leaveGroup(Ljava/net/SocketAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/leaveGroup(Ljava/net/SocketAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramSocketImpl/getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramSocketImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/DatagramSocketImpl/getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getFileDescriptor()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "java/net/DatagramSocketImpl/getFileDescriptor()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    "setOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "getOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_net_SocketOptions extends java_lang_Object {
    "setOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "getOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_nio_channels_MulticastChannel extends java_lang_Object, java_nio_channels_NetworkChannel {
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    "join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_HashSet extends java_util_AbstractSet implements java_util_Set, java_lang_Cloneable, java_io_Serializable {
    public static "java/util/HashSet/serialVersionUID": Long;
    public "java/util/HashSet/map": java_util_HashMap;
    public static "java/util/HashSet/PRESENT": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashSet/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashSet/<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashSet/<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashSet/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IFZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashSet/<init>(IFZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/HashSet/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashSet/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashSet/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashSet/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashSet/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashSet/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashSet/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashSet/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashSet/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashSet/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/HashSet/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/HashSet/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_HashMap extends java_util_AbstractMap implements java_util_Map, java_lang_Cloneable, java_io_Serializable {
    public static "java/util/HashMap/serialVersionUID": Long;
    public static "java/util/HashMap/DEFAULT_INITIAL_CAPACITY": number;
    public static "java/util/HashMap/MAXIMUM_CAPACITY": number;
    public static "java/util/HashMap/DEFAULT_LOAD_FACTOR": number;
    public static "java/util/HashMap/TREEIFY_THRESHOLD": number;
    public static "java/util/HashMap/UNTREEIFY_THRESHOLD": number;
    public static "java/util/HashMap/MIN_TREEIFY_CAPACITY": number;
    public "java/util/HashMap/table": JVMArray<java_util_HashMap$Node>;
    public "java/util/HashMap/entrySet": java_util_Set;
    public "java/util/HashMap/size": number;
    public "java/util/HashMap/modCount": number;
    public "java/util/HashMap/threshold": number;
    public "java/util/HashMap/loadFactor": number;
    public static "hash(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/HashMap/hash(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "comparableClassFor(Ljava/lang/Object;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/util/HashMap/comparableClassFor(Ljava/lang/Object;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "compareComparables(Ljava/lang/Class;Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Class, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/HashMap/compareComparables(Ljava/lang/Class;Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Class, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "tableSizeFor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/HashMap/tableSizeFor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "putMapEntries(Ljava/util/Map;Z)V"(thread: JVMThread, args: [java_util_Map, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/putMapEntries(Ljava/util/Map;Z)V"(thread: JVMThread, args: [java_util_Map, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getNode(ILjava/lang/Object;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "java/util/HashMap/getNode(ILjava/lang/Object;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap/containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putVal(ILjava/lang/Object;Ljava/lang/Object;ZZ)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/putVal(ILjava/lang/Object;Ljava/lang/Object;ZZ)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "resize()[Ljava/util/HashMap$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_HashMap$Node>) => void): void;
    public "java/util/HashMap/resize()[Ljava/util/HashMap$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_HashMap$Node>) => void): void;
    public "treeifyBin([Ljava/util/HashMap$Node;I)V"(thread: JVMThread, args: [JVMArray<java_util_HashMap$Node>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/treeifyBin([Ljava/util/HashMap$Node;I)V"(thread: JVMThread, args: [JVMArray<java_util_HashMap$Node>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "removeNode(ILjava/lang/Object;Ljava/lang/Object;ZZ)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "java/util/HashMap/removeNode(ILjava/lang/Object;Ljava/lang/Object;ZZ)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/HashMap/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/HashMap/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/HashMap/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap/remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap/replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "loadFactor()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap/loadFactor()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap/capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "newNode(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "java/util/HashMap/newNode(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "replacementNode(Ljava/util/HashMap$Node;Ljava/util/HashMap$Node;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [java_util_HashMap$Node, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "java/util/HashMap/replacementNode(Ljava/util/HashMap$Node;Ljava/util/HashMap$Node;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [java_util_HashMap$Node, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "newTreeNode(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "java/util/HashMap/newTreeNode(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "replacementTreeNode(Ljava/util/HashMap$Node;Ljava/util/HashMap$Node;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$Node, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "java/util/HashMap/replacementTreeNode(Ljava/util/HashMap$Node;Ljava/util/HashMap$Node;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$Node, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "reinitialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/reinitialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "afterNodeAccess(Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/afterNodeAccess(Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "afterNodeInsertion(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/afterNodeInsertion(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "afterNodeRemoval(Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/afterNodeRemoval(Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "internalWriteEntries(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap/internalWriteEntries(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_HashMap$TreeNode extends java_util_LinkedHashMap$Entry {
    public "java/util/HashMap$TreeNode/parent": java_util_HashMap$TreeNode;
    public "java/util/HashMap$TreeNode/left": java_util_HashMap$TreeNode;
    public "java/util/HashMap$TreeNode/right": java_util_HashMap$TreeNode;
    public "java/util/HashMap$TreeNode/prev": java_util_HashMap$TreeNode;
    public "java/util/HashMap$TreeNode/red": number;
    public static "java/util/HashMap$TreeNode/$assertionsDisabled": number;
    public "<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap$TreeNode/<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "root()Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "java/util/HashMap$TreeNode/root()Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "moveRootToFront([Ljava/util/HashMap$Node;Ljava/util/HashMap$TreeNode;)V"(thread: JVMThread, args: [JVMArray<java_util_HashMap$Node>, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/HashMap$TreeNode/moveRootToFront([Ljava/util/HashMap$Node;Ljava/util/HashMap$TreeNode;)V"(thread: JVMThread, args: [JVMArray<java_util_HashMap$Node>, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable) => void): void;
    public "find(ILjava/lang/Object;Ljava/lang/Class;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "java/util/HashMap$TreeNode/find(ILjava/lang/Object;Ljava/lang/Class;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "getTreeNode(ILjava/lang/Object;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "java/util/HashMap$TreeNode/getTreeNode(ILjava/lang/Object;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "tieBreakOrder(Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/HashMap$TreeNode/tieBreakOrder(Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "treeify([Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [JVMArray<java_util_HashMap$Node>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap$TreeNode/treeify([Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [JVMArray<java_util_HashMap$Node>], cb?: (e?: java_lang_Throwable) => void): void;
    public "untreeify(Ljava/util/HashMap;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [java_util_HashMap], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "java/util/HashMap$TreeNode/untreeify(Ljava/util/HashMap;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [java_util_HashMap], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "putTreeVal(Ljava/util/HashMap;[Ljava/util/HashMap$Node;ILjava/lang/Object;Ljava/lang/Object;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap, JVMArray<java_util_HashMap$Node>, number, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "java/util/HashMap$TreeNode/putTreeVal(Ljava/util/HashMap;[Ljava/util/HashMap$Node;ILjava/lang/Object;Ljava/lang/Object;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap, JVMArray<java_util_HashMap$Node>, number, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "removeTreeNode(Ljava/util/HashMap;[Ljava/util/HashMap$Node;Z)V"(thread: JVMThread, args: [java_util_HashMap, JVMArray<java_util_HashMap$Node>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap$TreeNode/removeTreeNode(Ljava/util/HashMap;[Ljava/util/HashMap$Node;Z)V"(thread: JVMThread, args: [java_util_HashMap, JVMArray<java_util_HashMap$Node>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "split(Ljava/util/HashMap;[Ljava/util/HashMap$Node;II)V"(thread: JVMThread, args: [java_util_HashMap, JVMArray<java_util_HashMap$Node>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap$TreeNode/split(Ljava/util/HashMap;[Ljava/util/HashMap$Node;II)V"(thread: JVMThread, args: [java_util_HashMap, JVMArray<java_util_HashMap$Node>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "rotateLeft(Ljava/util/HashMap$TreeNode;Ljava/util/HashMap$TreeNode;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$TreeNode, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "java/util/HashMap$TreeNode/rotateLeft(Ljava/util/HashMap$TreeNode;Ljava/util/HashMap$TreeNode;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$TreeNode, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "rotateRight(Ljava/util/HashMap$TreeNode;Ljava/util/HashMap$TreeNode;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$TreeNode, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "java/util/HashMap$TreeNode/rotateRight(Ljava/util/HashMap$TreeNode;Ljava/util/HashMap$TreeNode;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$TreeNode, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "balanceInsertion(Ljava/util/HashMap$TreeNode;Ljava/util/HashMap$TreeNode;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$TreeNode, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "java/util/HashMap$TreeNode/balanceInsertion(Ljava/util/HashMap$TreeNode;Ljava/util/HashMap$TreeNode;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$TreeNode, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "balanceDeletion(Ljava/util/HashMap$TreeNode;Ljava/util/HashMap$TreeNode;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$TreeNode, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "java/util/HashMap$TreeNode/balanceDeletion(Ljava/util/HashMap$TreeNode;Ljava/util/HashMap$TreeNode;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$TreeNode, java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public static "checkInvariants(Ljava/util/HashMap$TreeNode;)Z"(thread: JVMThread, args: [java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/HashMap$TreeNode/checkInvariants(Ljava/util/HashMap$TreeNode;)Z"(thread: JVMThread, args: [java_util_HashMap$TreeNode], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/HashMap$TreeNode/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_LinkedHashMap$Entry extends java_util_HashMap$Node {
    public "java/util/LinkedHashMap$Entry/before": java_util_LinkedHashMap$Entry;
    public "java/util/LinkedHashMap$Entry/after": java_util_LinkedHashMap$Entry;
    public "<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap$Entry/<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_HashMap$Node extends java_lang_Object implements java_util_Map$Entry {
    public "java/util/HashMap$Node/hash": number;
    public "java/util/HashMap$Node/key": java_lang_Object;
    public "java/util/HashMap$Node/value": java_lang_Object;
    public "java/util/HashMap$Node/next": java_util_HashMap$Node;
    public "<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/HashMap$Node/<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap$Node/getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap$Node/getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/HashMap$Node/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap$Node/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/HashMap$Node/setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/HashMap$Node/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_Map$Entry extends java_lang_Object {
    "getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_AbstractMap extends java_lang_Object implements java_util_Map {
    public "java/util/AbstractMap/keySet": java_util_Set;
    public "java/util/AbstractMap/values": java_util_Collection;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractMap/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractMap/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractMap/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractMap/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractMap/containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractMap/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractMap/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractMap/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractMap/putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractMap/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/AbstractMap/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/AbstractMap/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/AbstractMap/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractMap/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractMap/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/AbstractMap/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractMap/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "eq(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/AbstractMap/eq(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/AbstractMap/access$000(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_lang_Cloneable extends java_lang_Object {
  }
  export class java_util_AbstractSet extends java_util_AbstractCollection implements java_util_Set {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractSet/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractSet/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractSet/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractSet/removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_AbstractCollection extends java_lang_Object implements java_util_Collection {
    public static "java/util/AbstractCollection/MAX_ARRAY_SIZE": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractCollection/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/AbstractCollection/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractCollection/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractCollection/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractCollection/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/AbstractCollection/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/AbstractCollection/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public static "finishToArray([Ljava/lang/Object;Ljava/util/Iterator;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_util_Iterator], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public static "java/util/AbstractCollection/finishToArray([Ljava/lang/Object;Ljava/util/Iterator;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_util_Iterator], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public static "hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/AbstractCollection/hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractCollection/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractCollection/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractCollection/containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractCollection/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractCollection/removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractCollection/retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractCollection/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/AbstractCollection/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState extends java_lang_Enum {
    public static "sun/nio/ch/sctp/SctpServerChannelImpl$ChannelState/UNINITIALIZED": sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl$ChannelState/INUSE": sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl$ChannelState/KILLPENDING": sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl$ChannelState/KILLED": sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl$ChannelState/$VALUES": JVMArray<sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState>;
    public static "values()[Lsun/nio/ch/sctp/SctpServerChannelImpl$ChannelState;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState>) => void): void;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl$ChannelState/values()[Lsun/nio/ch/sctp/SctpServerChannelImpl$ChannelState;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lsun/nio/ch/sctp/SctpServerChannelImpl$ChannelState;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState) => void): void;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl$ChannelState/valueOf(Ljava/lang/String;)Lsun/nio/ch/sctp/SctpServerChannelImpl$ChannelState;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_SctpServerChannelImpl$ChannelState) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpServerChannelImpl$ChannelState/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpServerChannelImpl$ChannelState/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_nio_ch_SelChImpl extends java_lang_Object, java_nio_channels_Channel {
    "getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    "getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    "validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_SctpServerChannel extends java_nio_channels_spi_AbstractSelectableChannel {
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open()Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public static "com/sun/nio/sctp/SctpServerChannel/open()Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "accept()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/accept()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/bind(Ljava/net/SocketAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;I)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/bind(Ljava/net/SocketAddress;I)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "bindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/bindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "unbindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/unbindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "getAllLocalAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/getAllLocalAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getOption(Lcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/getOption(Lcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setOption(Lcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/setOption(Lcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)Lcom/sun/nio/sctp/SctpServerChannel;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpServerChannel) => void): void;
    public "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/SctpServerChannel/validOps()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_net_SocketAddress extends java_lang_Object implements java_io_Serializable {
    public static "java/net/SocketAddress/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketAddress/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_sctp_SctpNet extends java_lang_Object {
    public static "sun/nio/ch/sctp/SctpNet/osName": java_lang_String;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpNet/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "IPv4MappedAddresses()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/IPv4MappedAddresses()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "throwAlreadyBoundException()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/throwAlreadyBoundException()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "listen(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/listen(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "connect(ILjava/net/InetAddress;I)I"(thread: JVMThread, args: [number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/connect(ILjava/net/InetAddress;I)I"(thread: JVMThread, args: [number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "close(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/close(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "preClose(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/preClose(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "socket(Z)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/socket(Z)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "bindx(I[Ljava/net/InetAddress;IZ)V"(thread: JVMThread, args: [number, JVMArray<java_net_InetAddress>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/bindx(I[Ljava/net/InetAddress;IZ)V"(thread: JVMThread, args: [number, JVMArray<java_net_InetAddress>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getLocalAddresses(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getLocalAddresses(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "getRevealedLocalAddressSet([Ljava/net/SocketAddress;)Ljava/util/Set;"(thread: JVMThread, args: [JVMArray<java_net_SocketAddress>], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getRevealedLocalAddressSet([Ljava/net/SocketAddress;)Ljava/util/Set;"(thread: JVMThread, args: [JVMArray<java_net_SocketAddress>], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "getRevealedLocalAddress(Ljava/net/SocketAddress;Ljava/lang/SecurityManager;)Ljava/net/SocketAddress;"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_SecurityManager], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getRevealedLocalAddress(Ljava/net/SocketAddress;Ljava/lang/SecurityManager;)Ljava/net/SocketAddress;"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_SecurityManager], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public static "getRemoteAddresses(II)Ljava/util/Set;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getRemoteAddresses(II)Ljava/util/Set;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "setSocketOption(ILcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;I)V"(thread: JVMThread, args: [number, com_sun_nio_sctp_SctpSocketOption, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/setSocketOption(ILcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;I)V"(thread: JVMThread, args: [number, com_sun_nio_sctp_SctpSocketOption, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getSocketOption(ILcom/sun/nio/sctp/SctpSocketOption;I)Ljava/lang/Object;"(thread: JVMThread, args: [number, com_sun_nio_sctp_SctpSocketOption, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getSocketOption(ILcom/sun/nio/sctp/SctpSocketOption;I)Ljava/lang/Object;"(thread: JVMThread, args: [number, com_sun_nio_sctp_SctpSocketOption, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "setIntOption(ILcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [number, com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/setIntOption(ILcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [number, com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getIntOption(ILcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [number, com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getIntOption(ILcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [number, com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "shutdown(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/shutdown(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "branch(II)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/branch(II)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "socket0(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/socket0(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "listen0(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/listen0(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "connect0(ILjava/net/InetAddress;I)I"(thread: JVMThread, args: [number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/connect0(ILjava/net/InetAddress;I)I"(thread: JVMThread, args: [number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "close0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/close0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "preClose0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/preClose0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "bindx(I[Ljava/net/InetAddress;IIZZ)V"(thread: JVMThread, args: [number, JVMArray<java_net_InetAddress>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/bindx(I[Ljava/net/InetAddress;IIZZ)V"(thread: JVMThread, args: [number, JVMArray<java_net_InetAddress>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getIntOption0(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getIntOption0(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setIntOption0(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/setIntOption0(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getLocalAddresses0(I)[Ljava/net/SocketAddress;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_SocketAddress>) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getLocalAddresses0(I)[Ljava/net/SocketAddress;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_SocketAddress>) => void): void;
    public static "getRemoteAddresses0(II)[Ljava/net/SocketAddress;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_SocketAddress>) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getRemoteAddresses0(II)[Ljava/net/SocketAddress;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_SocketAddress>) => void): void;
    public static "branch0(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/branch0(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setPrimAddrOption0(IILjava/net/InetAddress;I)V"(thread: JVMThread, args: [number, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/setPrimAddrOption0(IILjava/net/InetAddress;I)V"(thread: JVMThread, args: [number, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setPeerPrimAddrOption0(IILjava/net/InetAddress;IZ)V"(thread: JVMThread, args: [number, number, java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/setPeerPrimAddrOption0(IILjava/net/InetAddress;IZ)V"(thread: JVMThread, args: [number, number, java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getPrimAddrOption0(II)Ljava/net/SocketAddress;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getPrimAddrOption0(II)Ljava/net/SocketAddress;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public static "getInitMsgOption0(I[I)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/getInitMsgOption0(I[I)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setInitMsgOption0(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/setInitMsgOption0(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "shutdown0(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/shutdown0(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpNet/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_sctp_ResultContainer extends java_lang_Object {
    public static "sun/nio/ch/sctp/ResultContainer/NOTHING": number;
    public static "sun/nio/ch/sctp/ResultContainer/MESSAGE": number;
    public static "sun/nio/ch/sctp/ResultContainer/SEND_FAILED": number;
    public static "sun/nio/ch/sctp/ResultContainer/ASSOCIATION_CHANGED": number;
    public static "sun/nio/ch/sctp/ResultContainer/PEER_ADDRESS_CHANGED": number;
    public static "sun/nio/ch/sctp/ResultContainer/SHUTDOWN": number;
    public "sun/nio/ch/sctp/ResultContainer/value": java_lang_Object;
    public "sun/nio/ch/sctp/ResultContainer/type": number;
    public static "sun/nio/ch/sctp/ResultContainer/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "type()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/type()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasSomething()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/hasSomething()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isNotification()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/isNotification()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "notification()Lsun/nio/ch/sctp/SctpNotification;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_SctpNotification) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/notification()Lsun/nio/ch/sctp/SctpNotification;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_SctpNotification) => void): void;
    public "getMessageInfo()Lsun/nio/ch/sctp/MessageInfoImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_MessageInfoImpl) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/getMessageInfo()Lsun/nio/ch/sctp/MessageInfoImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_MessageInfoImpl) => void): void;
    public "getSendFailed()Lsun/nio/ch/sctp/SendFailed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_SendFailed) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/getSendFailed()Lsun/nio/ch/sctp/SendFailed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_SendFailed) => void): void;
    public "getAssociationChanged()Lsun/nio/ch/sctp/AssociationChange;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_AssociationChange) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/getAssociationChanged()Lsun/nio/ch/sctp/AssociationChange;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_AssociationChange) => void): void;
    public "getPeerAddressChanged()Lsun/nio/ch/sctp/PeerAddrChange;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_PeerAddrChange) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/getPeerAddressChanged()Lsun/nio/ch/sctp/PeerAddrChange;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_PeerAddrChange) => void): void;
    public "getShutdown()Lsun/nio/ch/sctp/Shutdown;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_Shutdown) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/getShutdown()Lsun/nio/ch/sctp/Shutdown;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_Shutdown) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/sctp/ResultContainer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/ResultContainer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_sctp_Shutdown extends com_sun_nio_sctp_ShutdownNotification implements sun_nio_ch_sctp_SctpNotification {
    public "sun/nio/ch/sctp/Shutdown/association": com_sun_nio_sctp_Association;
    public "sun/nio/ch/sctp/Shutdown/assocId": number;
    public static "sun/nio/ch/sctp/Shutdown/$assertionsDisabled": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/Shutdown/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "assocId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/Shutdown/assocId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/Shutdown/setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "sun/nio/ch/sctp/Shutdown/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/sctp/Shutdown/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/Shutdown/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_ShutdownNotification extends java_lang_Object implements com_sun_nio_sctp_Notification {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/ShutdownNotification/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "com/sun/nio/sctp/ShutdownNotification/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
  }
  export class sun_nio_ch_sctp_PeerAddrChange extends com_sun_nio_sctp_PeerAddressChangeNotification implements sun_nio_ch_sctp_SctpNotification {
    public static "sun/nio/ch/sctp/PeerAddrChange/SCTP_ADDR_AVAILABLE": number;
    public static "sun/nio/ch/sctp/PeerAddrChange/SCTP_ADDR_UNREACHABLE": number;
    public static "sun/nio/ch/sctp/PeerAddrChange/SCTP_ADDR_REMOVED": number;
    public static "sun/nio/ch/sctp/PeerAddrChange/SCTP_ADDR_ADDED": number;
    public static "sun/nio/ch/sctp/PeerAddrChange/SCTP_ADDR_MADE_PRIM": number;
    public static "sun/nio/ch/sctp/PeerAddrChange/SCTP_ADDR_CONFIRMED": number;
    public "sun/nio/ch/sctp/PeerAddrChange/association": com_sun_nio_sctp_Association;
    public "sun/nio/ch/sctp/PeerAddrChange/assocId": number;
    public "sun/nio/ch/sctp/PeerAddrChange/address": java_net_SocketAddress;
    public "sun/nio/ch/sctp/PeerAddrChange/event": com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent;
    public static "sun/nio/ch/sctp/PeerAddrChange/$assertionsDisabled": number;
    public "<init>(ILjava/net/SocketAddress;I)V"(thread: JVMThread, args: [number, java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/PeerAddrChange/<init>(ILjava/net/SocketAddress;I)V"(thread: JVMThread, args: [number, java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "assocId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/PeerAddrChange/assocId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/PeerAddrChange/setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/sctp/PeerAddrChange/address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "sun/nio/ch/sctp/PeerAddrChange/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "event()Lcom/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent) => void): void;
    public "sun/nio/ch/sctp/PeerAddrChange/event()Lcom/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/sctp/PeerAddrChange/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/PeerAddrChange/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent extends java_lang_Enum {
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/ADDR_AVAILABLE": com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent;
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/ADDR_UNREACHABLE": com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent;
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/ADDR_REMOVED": com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent;
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/ADDR_ADDED": com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent;
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/ADDR_MADE_PRIMARY": com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent;
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/ADDR_CONFIRMED": com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent;
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/$VALUES": JVMArray<com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent>;
    public static "values()[Lcom/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent>) => void): void;
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/values()[Lcom/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lcom/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent) => void): void;
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/valueOf(Ljava/lang/String;)Lcom/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "com/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_PeerAddressChangeNotification extends java_lang_Object implements com_sun_nio_sctp_Notification {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/PeerAddressChangeNotification/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "com/sun/nio/sctp/PeerAddressChangeNotification/address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "com/sun/nio/sctp/PeerAddressChangeNotification/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "event()Lcom/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent) => void): void;
    public "com/sun/nio/sctp/PeerAddressChangeNotification/event()Lcom/sun/nio/sctp/PeerAddressChangeNotification$AddressChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_PeerAddressChangeNotification$AddressChangeEvent) => void): void;
  }
  export class sun_nio_ch_sctp_AssociationChange extends com_sun_nio_sctp_AssociationChangeNotification implements sun_nio_ch_sctp_SctpNotification {
    public static "sun/nio/ch/sctp/AssociationChange/SCTP_COMM_UP": number;
    public static "sun/nio/ch/sctp/AssociationChange/SCTP_COMM_LOST": number;
    public static "sun/nio/ch/sctp/AssociationChange/SCTP_RESTART": number;
    public static "sun/nio/ch/sctp/AssociationChange/SCTP_SHUTDOWN": number;
    public static "sun/nio/ch/sctp/AssociationChange/SCTP_CANT_START": number;
    public "sun/nio/ch/sctp/AssociationChange/association": com_sun_nio_sctp_Association;
    public "sun/nio/ch/sctp/AssociationChange/assocId": number;
    public "sun/nio/ch/sctp/AssociationChange/event": com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent;
    public "sun/nio/ch/sctp/AssociationChange/maxOutStreams": number;
    public "sun/nio/ch/sctp/AssociationChange/maxInStreams": number;
    public static "sun/nio/ch/sctp/AssociationChange/$assertionsDisabled": number;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/AssociationChange/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "assocId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/AssociationChange/assocId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/AssociationChange/setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "sun/nio/ch/sctp/AssociationChange/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "event()Lcom/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent) => void): void;
    public "sun/nio/ch/sctp/AssociationChange/event()Lcom/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent) => void): void;
    public "maxOutStreams()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/AssociationChange/maxOutStreams()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "maxInStreams()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/AssociationChange/maxInStreams()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/sctp/AssociationChange/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/AssociationChange/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent extends java_lang_Enum {
    public static "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/COMM_UP": com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent;
    public static "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/COMM_LOST": com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent;
    public static "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/RESTART": com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent;
    public static "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/SHUTDOWN": com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent;
    public static "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/CANT_START": com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent;
    public static "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/$VALUES": JVMArray<com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent>;
    public static "values()[Lcom/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent>) => void): void;
    public static "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/values()[Lcom/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lcom/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent) => void): void;
    public static "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/valueOf(Ljava/lang/String;)Lcom/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "com/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_AssociationChangeNotification extends java_lang_Object implements com_sun_nio_sctp_Notification {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/AssociationChangeNotification/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "com/sun/nio/sctp/AssociationChangeNotification/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "event()Lcom/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent) => void): void;
    public "com/sun/nio/sctp/AssociationChangeNotification/event()Lcom/sun/nio/sctp/AssociationChangeNotification$AssocChangeEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_AssociationChangeNotification$AssocChangeEvent) => void): void;
  }
  export class sun_nio_ch_sctp_SendFailed extends com_sun_nio_sctp_SendFailedNotification implements sun_nio_ch_sctp_SctpNotification {
    public "sun/nio/ch/sctp/SendFailed/association": com_sun_nio_sctp_Association;
    public "sun/nio/ch/sctp/SendFailed/assocId": number;
    public "sun/nio/ch/sctp/SendFailed/address": java_net_SocketAddress;
    public "sun/nio/ch/sctp/SendFailed/buffer": java_nio_ByteBuffer;
    public "sun/nio/ch/sctp/SendFailed/errorCode": number;
    public "sun/nio/ch/sctp/SendFailed/streamNumber": number;
    public static "sun/nio/ch/sctp/SendFailed/$assertionsDisabled": number;
    public "<init>(ILjava/net/SocketAddress;Ljava/nio/ByteBuffer;II)V"(thread: JVMThread, args: [number, java_net_SocketAddress, java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SendFailed/<init>(ILjava/net/SocketAddress;Ljava/nio/ByteBuffer;II)V"(thread: JVMThread, args: [number, java_net_SocketAddress, java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "assocId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SendFailed/assocId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SendFailed/setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "sun/nio/ch/sctp/SendFailed/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/sctp/SendFailed/address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "buffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/nio/ch/sctp/SendFailed/buffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "errorCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SendFailed/errorCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "streamNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SendFailed/streamNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/sctp/SendFailed/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SendFailed/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_SendFailedNotification extends java_lang_Object implements com_sun_nio_sctp_Notification {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/SendFailedNotification/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "com/sun/nio/sctp/SendFailedNotification/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "com/sun/nio/sctp/SendFailedNotification/address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "buffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "com/sun/nio/sctp/SendFailedNotification/buffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "errorCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/SendFailedNotification/errorCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "streamNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/nio/sctp/SendFailedNotification/streamNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_nio_ch_sctp_MessageInfoImpl extends com_sun_nio_sctp_MessageInfo {
    public "sun/nio/ch/sctp/MessageInfoImpl/address": java_net_SocketAddress;
    public "sun/nio/ch/sctp/MessageInfoImpl/bytes": number;
    public "sun/nio/ch/sctp/MessageInfoImpl/association": com_sun_nio_sctp_Association;
    public "sun/nio/ch/sctp/MessageInfoImpl/assocId": number;
    public "sun/nio/ch/sctp/MessageInfoImpl/streamNumber": number;
    public "sun/nio/ch/sctp/MessageInfoImpl/complete": number;
    public "sun/nio/ch/sctp/MessageInfoImpl/unordered": number;
    public "sun/nio/ch/sctp/MessageInfoImpl/timeToLive": Long;
    public "sun/nio/ch/sctp/MessageInfoImpl/ppid": number;
    public "<init>(Lcom/sun/nio/sctp/Association;Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association, java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/<init>(Lcom/sun/nio/sctp/Association;Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association, java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/net/SocketAddress;IIZZI)V"(thread: JVMThread, args: [number, java_net_SocketAddress, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/<init>(ILjava/net/SocketAddress;IIZZI)V"(thread: JVMThread, args: [number, java_net_SocketAddress, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "associationID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/associationID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "bytes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/bytes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "streamNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/streamNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "streamNumber(I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/streamNumber(I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "payloadProtocolID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/payloadProtocolID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "payloadProtocolID(I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/payloadProtocolID(I)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "isComplete()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/isComplete()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "complete(Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/complete(Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "isUnordered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/isUnordered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unordered(Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/unordered(Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "timeToLive()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/timeToLive()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "timeToLive(J)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/timeToLive(J)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/sctp/MessageInfoImpl/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface sun_nio_ch_sctp_SctpNotification extends java_lang_Object, com_sun_nio_sctp_Notification {
    "assocId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
  }
  export class sun_nio_ch_sctp_SctpChannelImpl extends com_sun_nio_sctp_SctpChannel implements sun_nio_ch_SelChImpl {
    public "sun/nio/ch/sctp/SctpChannelImpl/fd": java_io_FileDescriptor;
    public "sun/nio/ch/sctp/SctpChannelImpl/fdVal": number;
    public "sun/nio/ch/sctp/SctpChannelImpl/receiverThread": Long;
    public "sun/nio/ch/sctp/SctpChannelImpl/senderThread": Long;
    public "sun/nio/ch/sctp/SctpChannelImpl/receiveLock": java_lang_Object;
    public "sun/nio/ch/sctp/SctpChannelImpl/sendLock": java_lang_Object;
    public "sun/nio/ch/sctp/SctpChannelImpl/receiveInvoked": java_lang_ThreadLocal;
    public "sun/nio/ch/sctp/SctpChannelImpl/stateLock": java_lang_Object;
    public "sun/nio/ch/sctp/SctpChannelImpl/state": sun_nio_ch_sctp_SctpChannelImpl$ChannelState;
    public "sun/nio/ch/sctp/SctpChannelImpl/port": number;
    public "sun/nio/ch/sctp/SctpChannelImpl/localAddresses": java_util_HashSet;
    public "sun/nio/ch/sctp/SctpChannelImpl/wildcard": number;
    public "sun/nio/ch/sctp/SctpChannelImpl/readyToConnect": number;
    public "sun/nio/ch/sctp/SctpChannelImpl/isShutdown": number;
    public "sun/nio/ch/sctp/SctpChannelImpl/association": com_sun_nio_sctp_Association;
    public "sun/nio/ch/sctp/SctpChannelImpl/remoteAddresses": java_util_Set;
    public "sun/nio/ch/sctp/SctpChannelImpl/internalNotificationHandler": sun_nio_ch_sctp_SctpChannelImpl$InternalNotificationHandler;
    public static "sun/nio/ch/sctp/SctpChannelImpl/$assertionsDisabled": number;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor, com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor, com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(Ljava/net/SocketAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/bind(Ljava/net/SocketAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "bindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/bindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "unbindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/unbindAddress(Ljava/net/InetAddress;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "bindUnbindAddress(Ljava/net/InetAddress;Z)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/bindUnbindAddress(Ljava/net/InetAddress;Z)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureOpenAndUnconnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/ensureOpenAndUnconnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureReceiveOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/ensureReceiveOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureSendOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/ensureSendOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "receiverCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/receiverCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "senderCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/senderCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/association()Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public "connect(Ljava/net/SocketAddress;)Z"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/connect(Ljava/net/SocketAddress;)Z"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "connect(Ljava/net/SocketAddress;II)Z"(thread: JVMThread, args: [java_net_SocketAddress, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/connect(Ljava/net/SocketAddress;II)Z"(thread: JVMThread, args: [java_net_SocketAddress, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConnectionPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/isConnectionPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finishConnect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/finishConnect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOption(Lcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/setOption(Lcom/sun/nio/sctp/SctpSocketOption;Ljava/lang/Object;)Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "getOption(Lcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/getOption(Lcom/sun/nio/sctp/SctpSocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [com_sun_nio_sctp_SctpSocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "receive(Ljava/nio/ByteBuffer;Ljava/lang/Object;Lcom/sun/nio/sctp/NotificationHandler;)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, com_sun_nio_sctp_NotificationHandler], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/receive(Ljava/nio/ByteBuffer;Ljava/lang/Object;Lcom/sun/nio/sctp/NotificationHandler;)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, com_sun_nio_sctp_NotificationHandler], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "receive(Ljava/nio/ByteBuffer;Ljava/lang/Object;Lcom/sun/nio/sctp/NotificationHandler;Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, com_sun_nio_sctp_NotificationHandler, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/receive(Ljava/nio/ByteBuffer;Ljava/lang/Object;Lcom/sun/nio/sctp/NotificationHandler;Z)Lcom/sun/nio/sctp/MessageInfo;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, com_sun_nio_sctp_NotificationHandler, number], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_MessageInfo) => void): void;
    public "receive(ILjava/nio/ByteBuffer;Lsun/nio/ch/sctp/ResultContainer;Z)I"(thread: JVMThread, args: [number, java_nio_ByteBuffer, sun_nio_ch_sctp_ResultContainer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/receive(ILjava/nio/ByteBuffer;Lsun/nio/ch/sctp/ResultContainer;Z)I"(thread: JVMThread, args: [number, java_nio_ByteBuffer, sun_nio_ch_sctp_ResultContainer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "receiveIntoNativeBuffer(ILsun/nio/ch/sctp/ResultContainer;Ljava/nio/ByteBuffer;IIZ)I"(thread: JVMThread, args: [number, sun_nio_ch_sctp_ResultContainer, java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/receiveIntoNativeBuffer(ILsun/nio/ch/sctp/ResultContainer;Ljava/nio/ByteBuffer;IIZ)I"(thread: JVMThread, args: [number, sun_nio_ch_sctp_ResultContainer, java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "handleNotificationInternal(Lsun/nio/ch/sctp/ResultContainer;)V"(thread: JVMThread, args: [sun_nio_ch_sctp_ResultContainer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/handleNotificationInternal(Lsun/nio/ch/sctp/ResultContainer;)V"(thread: JVMThread, args: [sun_nio_ch_sctp_ResultContainer], cb?: (e?: java_lang_Throwable) => void): void;
    public "invokeNotificationHandler(Lsun/nio/ch/sctp/ResultContainer;Lcom/sun/nio/sctp/NotificationHandler;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [sun_nio_ch_sctp_ResultContainer, com_sun_nio_sctp_NotificationHandler, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/invokeNotificationHandler(Lsun/nio/ch/sctp/ResultContainer;Lcom/sun/nio/sctp/NotificationHandler;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [sun_nio_ch_sctp_ResultContainer, com_sun_nio_sctp_NotificationHandler, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "checkAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/checkAssociation(Lcom/sun/nio/sctp/Association;)V"(thread: JVMThread, args: [com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkStreamNumber(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/checkStreamNumber(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "send(Ljava/nio/ByteBuffer;Lcom/sun/nio/sctp/MessageInfo;)I"(thread: JVMThread, args: [java_nio_ByteBuffer, com_sun_nio_sctp_MessageInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/send(Ljava/nio/ByteBuffer;Lcom/sun/nio/sctp/MessageInfo;)I"(thread: JVMThread, args: [java_nio_ByteBuffer, com_sun_nio_sctp_MessageInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "send(ILjava/nio/ByteBuffer;Lcom/sun/nio/sctp/MessageInfo;)I"(thread: JVMThread, args: [number, java_nio_ByteBuffer, com_sun_nio_sctp_MessageInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/send(ILjava/nio/ByteBuffer;Lcom/sun/nio/sctp/MessageInfo;)I"(thread: JVMThread, args: [number, java_nio_ByteBuffer, com_sun_nio_sctp_MessageInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sendFromNativeBuffer(ILjava/nio/ByteBuffer;Ljava/net/SocketAddress;IZI)I"(thread: JVMThread, args: [number, java_nio_ByteBuffer, java_net_SocketAddress, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/sendFromNativeBuffer(ILjava/nio/ByteBuffer;Ljava/net/SocketAddress;IZI)I"(thread: JVMThread, args: [number, java_nio_ByteBuffer, java_net_SocketAddress, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shutdown()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/shutdown()Lcom/sun/nio/sctp/SctpChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_SctpChannel) => void): void;
    public "getAllLocalAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/getAllLocalAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getRemoteAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl/getRemoteAddresses()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "receive0(ILsun/nio/ch/sctp/ResultContainer;JIZ)I"(thread: JVMThread, args: [number, sun_nio_ch_sctp_ResultContainer, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl/receive0(ILsun/nio/ch/sctp/ResultContainer;JIZ)I"(thread: JVMThread, args: [number, sun_nio_ch_sctp_ResultContainer, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "send0(IJILjava/net/InetAddress;IIIZI)I"(thread: JVMThread, args: [number, Long, any, number, java_net_InetAddress, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl/send0(IJILjava/net/InetAddress;IIIZI)I"(thread: JVMThread, args: [number, Long, any, number, java_net_InetAddress, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkConnect(Ljava/io/FileDescriptor;ZZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl/checkConnect(Ljava/io/FileDescriptor;ZZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Lsun/nio/ch/sctp/SctpChannelImpl;)Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl/access$100(Lsun/nio/ch/sctp/SctpChannelImpl;)Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public static "access$102(Lsun/nio/ch/sctp/SctpChannelImpl;Lcom/sun/nio/sctp/Association;)Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl, com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl/access$102(Lsun/nio/ch/sctp/SctpChannelImpl;Lcom/sun/nio/sctp/Association;)Lcom/sun/nio/sctp/Association;"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl, com_sun_nio_sctp_Association], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_Association) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_sctp_SctpChannelImpl$InternalNotificationHandler extends com_sun_nio_sctp_AbstractNotificationHandler {
    public "sun/nio/ch/sctp/SctpChannelImpl$InternalNotificationHandler/this$0": sun_nio_ch_sctp_SctpChannelImpl;
    public "<init>(Lsun/nio/ch/sctp/SctpChannelImpl;)V"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl$InternalNotificationHandler/<init>(Lsun/nio/ch/sctp/SctpChannelImpl;)V"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "handleNotification(Lcom/sun/nio/sctp/AssociationChangeNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_AssociationChangeNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl$InternalNotificationHandler/handleNotification(Lcom/sun/nio/sctp/AssociationChangeNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_AssociationChangeNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "<init>(Lsun/nio/ch/sctp/SctpChannelImpl;Lsun/nio/ch/sctp/SctpChannelImpl$1;)V"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl, sun_nio_ch_sctp_SctpChannelImpl$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl$InternalNotificationHandler/<init>(Lsun/nio/ch/sctp/SctpChannelImpl;Lsun/nio/ch/sctp/SctpChannelImpl$1;)V"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl, sun_nio_ch_sctp_SctpChannelImpl$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_sctp_SctpChannelImpl$1 extends java_lang_ThreadLocal {
    public "sun/nio/ch/sctp/SctpChannelImpl$1/this$0": sun_nio_ch_sctp_SctpChannelImpl;
    public "<init>(Lsun/nio/ch/sctp/SctpChannelImpl;)V"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl$1/<init>(Lsun/nio/ch/sctp/SctpChannelImpl;)V"(thread: JVMThread, args: [sun_nio_ch_sctp_SctpChannelImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "initialValue()Ljava/lang/Boolean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Boolean) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl$1/initialValue()Ljava/lang/Boolean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Boolean) => void): void;
    public "initialValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl$1/initialValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_lang_Boolean extends java_lang_Object implements java_io_Serializable, java_lang_Comparable {
    public unbox(): number;
    public static box(val: number): java_lang_Boolean;
    public static "java/lang/Boolean/TRUE": java_lang_Boolean;
    public static "java/lang/Boolean/FALSE": java_lang_Boolean;
    public static "java/lang/Boolean/TYPE": java_lang_Class;
    public "java/lang/Boolean/value": number;
    public static "java/lang/Boolean/serialVersionUID": Long;
    public "<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Boolean/<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Boolean/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "parseBoolean(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Boolean/parseBoolean(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "booleanValue()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Boolean/booleanValue()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "valueOf(Z)Ljava/lang/Boolean;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Boolean) => void): void;
    public static "java/lang/Boolean/valueOf(Z)Ljava/lang/Boolean;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Boolean) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/Boolean;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Boolean) => void): void;
    public static "java/lang/Boolean/valueOf(Ljava/lang/String;)Ljava/lang/Boolean;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Boolean) => void): void;
    public static "toString(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Boolean/toString(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Boolean/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Boolean/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hashCode(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Boolean/hashCode(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Boolean/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getBoolean(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Boolean/getBoolean(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Boolean;)I"(thread: JVMThread, args: [java_lang_Boolean], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Boolean/compareTo(Ljava/lang/Boolean;)I"(thread: JVMThread, args: [java_lang_Boolean], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(ZZ)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Boolean/compare(ZZ)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "logicalAnd(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Boolean/logicalAnd(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "logicalOr(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Boolean/logicalOr(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "logicalXor(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Boolean/logicalXor(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Boolean/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Boolean/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class com_sun_nio_sctp_AbstractNotificationHandler extends java_lang_Object implements com_sun_nio_sctp_NotificationHandler {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/nio/sctp/AbstractNotificationHandler/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "handleNotification(Lcom/sun/nio/sctp/Notification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_Notification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "com/sun/nio/sctp/AbstractNotificationHandler/handleNotification(Lcom/sun/nio/sctp/Notification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_Notification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "handleNotification(Lcom/sun/nio/sctp/AssociationChangeNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_AssociationChangeNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "com/sun/nio/sctp/AbstractNotificationHandler/handleNotification(Lcom/sun/nio/sctp/AssociationChangeNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_AssociationChangeNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "handleNotification(Lcom/sun/nio/sctp/PeerAddressChangeNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_PeerAddressChangeNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "com/sun/nio/sctp/AbstractNotificationHandler/handleNotification(Lcom/sun/nio/sctp/PeerAddressChangeNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_PeerAddressChangeNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "handleNotification(Lcom/sun/nio/sctp/SendFailedNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_SendFailedNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "com/sun/nio/sctp/AbstractNotificationHandler/handleNotification(Lcom/sun/nio/sctp/SendFailedNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_SendFailedNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "handleNotification(Lcom/sun/nio/sctp/ShutdownNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_ShutdownNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
    public "com/sun/nio/sctp/AbstractNotificationHandler/handleNotification(Lcom/sun/nio/sctp/ShutdownNotification;Ljava/lang/Object;)Lcom/sun/nio/sctp/HandlerResult;"(thread: JVMThread, args: [com_sun_nio_sctp_ShutdownNotification, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: com_sun_nio_sctp_HandlerResult) => void): void;
  }
  export class sun_nio_ch_sctp_SctpChannelImpl$ChannelState extends java_lang_Enum {
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/UNINITIALIZED": sun_nio_ch_sctp_SctpChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/UNCONNECTED": sun_nio_ch_sctp_SctpChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/PENDING": sun_nio_ch_sctp_SctpChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/CONNECTED": sun_nio_ch_sctp_SctpChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/KILLPENDING": sun_nio_ch_sctp_SctpChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/KILLED": sun_nio_ch_sctp_SctpChannelImpl$ChannelState;
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/$VALUES": JVMArray<sun_nio_ch_sctp_SctpChannelImpl$ChannelState>;
    public static "values()[Lsun/nio/ch/sctp/SctpChannelImpl$ChannelState;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_nio_ch_sctp_SctpChannelImpl$ChannelState>) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/values()[Lsun/nio/ch/sctp/SctpChannelImpl$ChannelState;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_nio_ch_sctp_SctpChannelImpl$ChannelState>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lsun/nio/ch/sctp/SctpChannelImpl$ChannelState;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_SctpChannelImpl$ChannelState) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/valueOf(Ljava/lang/String;)Lsun/nio/ch/sctp/SctpChannelImpl$ChannelState;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_sctp_SctpChannelImpl$ChannelState) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/sctp/SctpChannelImpl$ChannelState/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ThreadLocal extends java_lang_Object {
    public "java/lang/ThreadLocal/threadLocalHashCode": number;
    public static "java/lang/ThreadLocal/nextHashCode": java_util_concurrent_atomic_AtomicInteger;
    public static "java/lang/ThreadLocal/HASH_INCREMENT": number;
    public static "nextHashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ThreadLocal/nextHashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "initialValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ThreadLocal/initialValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "withInitial(Ljava/util/function/Supplier;)Ljava/lang/ThreadLocal;"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal) => void): void;
    public static "java/lang/ThreadLocal/withInitial(Ljava/util/function/Supplier;)Ljava/lang/ThreadLocal;"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ThreadLocal/get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setInitialValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ThreadLocal/setInitialValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal/set(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal/remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMap(Ljava/lang/Thread;)Ljava/lang/ThreadLocal$ThreadLocalMap;"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap) => void): void;
    public "java/lang/ThreadLocal/getMap(Ljava/lang/Thread;)Ljava/lang/ThreadLocal$ThreadLocalMap;"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap) => void): void;
    public "createMap(Ljava/lang/Thread;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Thread, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal/createMap(Ljava/lang/Thread;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Thread, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createInheritedMap(Ljava/lang/ThreadLocal$ThreadLocalMap;)Ljava/lang/ThreadLocal$ThreadLocalMap;"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap) => void): void;
    public static "java/lang/ThreadLocal/createInheritedMap(Ljava/lang/ThreadLocal$ThreadLocalMap;)Ljava/lang/ThreadLocal$ThreadLocalMap;"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap) => void): void;
    public "childValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ThreadLocal/childValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$400(Ljava/lang/ThreadLocal;)I"(thread: JVMThread, args: [java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ThreadLocal/access$400(Ljava/lang/ThreadLocal;)I"(thread: JVMThread, args: [java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ThreadLocal/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ThreadLocal$ThreadLocalMap extends java_lang_Object {
    public static "java/lang/ThreadLocal$ThreadLocalMap/INITIAL_CAPACITY": number;
    public "java/lang/ThreadLocal$ThreadLocalMap/table": JVMArray<java_lang_ThreadLocal$ThreadLocalMap$Entry>;
    public "java/lang/ThreadLocal$ThreadLocalMap/size": number;
    public "java/lang/ThreadLocal$ThreadLocalMap/threshold": number;
    public "setThreshold(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/setThreshold(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "nextIndex(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ThreadLocal$ThreadLocalMap/nextIndex(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "prevIndex(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ThreadLocal$ThreadLocalMap/prevIndex(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/ThreadLocal;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ThreadLocal, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/<init>(Ljava/lang/ThreadLocal;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ThreadLocal, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ThreadLocal$ThreadLocalMap;)V"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/<init>(Ljava/lang/ThreadLocal$ThreadLocalMap;)V"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEntry(Ljava/lang/ThreadLocal;)Ljava/lang/ThreadLocal$ThreadLocalMap$Entry;"(thread: JVMThread, args: [java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap$Entry) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/getEntry(Ljava/lang/ThreadLocal;)Ljava/lang/ThreadLocal$ThreadLocalMap$Entry;"(thread: JVMThread, args: [java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap$Entry) => void): void;
    public "getEntryAfterMiss(Ljava/lang/ThreadLocal;ILjava/lang/ThreadLocal$ThreadLocalMap$Entry;)Ljava/lang/ThreadLocal$ThreadLocalMap$Entry;"(thread: JVMThread, args: [java_lang_ThreadLocal, number, java_lang_ThreadLocal$ThreadLocalMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap$Entry) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/getEntryAfterMiss(Ljava/lang/ThreadLocal;ILjava/lang/ThreadLocal$ThreadLocalMap$Entry;)Ljava/lang/ThreadLocal$ThreadLocalMap$Entry;"(thread: JVMThread, args: [java_lang_ThreadLocal, number, java_lang_ThreadLocal$ThreadLocalMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap$Entry) => void): void;
    public "set(Ljava/lang/ThreadLocal;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ThreadLocal, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/set(Ljava/lang/ThreadLocal;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ThreadLocal, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/ThreadLocal;)V"(thread: JVMThread, args: [java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/remove(Ljava/lang/ThreadLocal;)V"(thread: JVMThread, args: [java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceStaleEntry(Ljava/lang/ThreadLocal;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_ThreadLocal, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/replaceStaleEntry(Ljava/lang/ThreadLocal;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_ThreadLocal, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "expungeStaleEntry(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/expungeStaleEntry(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "cleanSomeSlots(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/cleanSomeSlots(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rehash()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/rehash()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/resize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "expungeStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/expungeStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/lang/ThreadLocal$ThreadLocalMap;Ljava/lang/ThreadLocal;)Ljava/lang/ThreadLocal$ThreadLocalMap$Entry;"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap, java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap$Entry) => void): void;
    public static "java/lang/ThreadLocal$ThreadLocalMap/access$000(Ljava/lang/ThreadLocal$ThreadLocalMap;Ljava/lang/ThreadLocal;)Ljava/lang/ThreadLocal$ThreadLocalMap$Entry;"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap, java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal$ThreadLocalMap$Entry) => void): void;
    public static "access$100(Ljava/lang/ThreadLocal$ThreadLocalMap;Ljava/lang/ThreadLocal;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap, java_lang_ThreadLocal, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ThreadLocal$ThreadLocalMap/access$100(Ljava/lang/ThreadLocal$ThreadLocalMap;Ljava/lang/ThreadLocal;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap, java_lang_ThreadLocal, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$200(Ljava/lang/ThreadLocal$ThreadLocalMap;Ljava/lang/ThreadLocal;)V"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap, java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ThreadLocal$ThreadLocalMap/access$200(Ljava/lang/ThreadLocal$ThreadLocalMap;Ljava/lang/ThreadLocal;)V"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap, java_lang_ThreadLocal], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ThreadLocal$ThreadLocalMap;Ljava/lang/ThreadLocal$1;)V"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap, java_lang_ThreadLocal$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap/<init>(Ljava/lang/ThreadLocal$ThreadLocalMap;Ljava/lang/ThreadLocal$1;)V"(thread: JVMThread, args: [java_lang_ThreadLocal$ThreadLocalMap, java_lang_ThreadLocal$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ThreadLocal$1 extends java_lang_Object {
  }
  export class java_lang_ThreadLocal$ThreadLocalMap$Entry extends java_lang_ref_WeakReference {
    public "java/lang/ThreadLocal$ThreadLocalMap$Entry/value": java_lang_Object;
    public "<init>(Ljava/lang/ThreadLocal;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ThreadLocal, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadLocal$ThreadLocalMap$Entry/<init>(Ljava/lang/ThreadLocal;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ThreadLocal, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ref_WeakReference extends java_lang_ref_Reference {
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/WeakReference/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_ref_ReferenceQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/WeakReference/<init>(Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_ref_ReferenceQueue], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ref_ReferenceQueue extends java_lang_Object {
    public static "java/lang/ref/ReferenceQueue/NULL": java_lang_ref_ReferenceQueue;
    public static "java/lang/ref/ReferenceQueue/ENQUEUED": java_lang_ref_ReferenceQueue;
    public "java/lang/ref/ReferenceQueue/lock": java_lang_ref_ReferenceQueue$Lock;
    public "java/lang/ref/ReferenceQueue/head": java_lang_ref_Reference;
    public "java/lang/ref/ReferenceQueue/queueLength": Long;
    public static "java/lang/ref/ReferenceQueue/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/ReferenceQueue/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "enqueue(Ljava/lang/ref/Reference;)Z"(thread: JVMThread, args: [java_lang_ref_Reference], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ref/ReferenceQueue/enqueue(Ljava/lang/ref/Reference;)Z"(thread: JVMThread, args: [java_lang_ref_Reference], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reallyPoll()Ljava/lang/ref/Reference;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public "java/lang/ref/ReferenceQueue/reallyPoll()Ljava/lang/ref/Reference;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public "poll()Ljava/lang/ref/Reference;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public "java/lang/ref/ReferenceQueue/poll()Ljava/lang/ref/Reference;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public "remove(J)Ljava/lang/ref/Reference;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public "java/lang/ref/ReferenceQueue/remove(J)Ljava/lang/ref/Reference;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public "remove()Ljava/lang/ref/Reference;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public "java/lang/ref/ReferenceQueue/remove()Ljava/lang/ref/Reference;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ref/ReferenceQueue/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ref_ReferenceQueue$Lock extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/ReferenceQueue$Lock/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ref/ReferenceQueue$1;)V"(thread: JVMThread, args: [java_lang_ref_ReferenceQueue$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/ReferenceQueue$Lock/<init>(Ljava/lang/ref/ReferenceQueue$1;)V"(thread: JVMThread, args: [java_lang_ref_ReferenceQueue$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ref_ReferenceQueue$1 extends java_lang_Object {
  }
  export class java_util_concurrent_atomic_AtomicInteger extends java_lang_Number implements java_io_Serializable {
    public static "java/util/concurrent/atomic/AtomicInteger/serialVersionUID": Long;
    public static "java/util/concurrent/atomic/AtomicInteger/unsafe": sun_misc_Unsafe;
    public static "java/util/concurrent/atomic/AtomicInteger/valueOffset": Long;
    public "java/util/concurrent/atomic/AtomicInteger/value": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "get()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/get()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "set(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/set(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "lazySet(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/lazySet(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAndSet(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/getAndSet(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareAndSet(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/compareAndSet(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "weakCompareAndSet(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/weakCompareAndSet(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAndIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/getAndIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAndDecrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/getAndDecrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAndAdd(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/getAndAdd(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "incrementAndGet()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/incrementAndGet()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "decrementAndGet()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/decrementAndGet()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAndGet(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/addAndGet(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAndUpdate(Ljava/util/function/IntUnaryOperator;)I"(thread: JVMThread, args: [java_util_function_IntUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/getAndUpdate(Ljava/util/function/IntUnaryOperator;)I"(thread: JVMThread, args: [java_util_function_IntUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "updateAndGet(Ljava/util/function/IntUnaryOperator;)I"(thread: JVMThread, args: [java_util_function_IntUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/updateAndGet(Ljava/util/function/IntUnaryOperator;)I"(thread: JVMThread, args: [java_util_function_IntUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAndAccumulate(ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/getAndAccumulate(ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "accumulateAndGet(ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/accumulateAndGet(ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicInteger/doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/atomic/AtomicInteger/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_FileChannelImpl extends java_nio_channels_FileChannel {
    public static "sun/nio/ch/FileChannelImpl/allocationGranularity": Long;
    public "sun/nio/ch/FileChannelImpl/nd": sun_nio_ch_FileDispatcher;
    public "sun/nio/ch/FileChannelImpl/fd": java_io_FileDescriptor;
    public "sun/nio/ch/FileChannelImpl/writable": number;
    public "sun/nio/ch/FileChannelImpl/readable": number;
    public "sun/nio/ch/FileChannelImpl/append": number;
    public "sun/nio/ch/FileChannelImpl/parent": java_lang_Object;
    public "sun/nio/ch/FileChannelImpl/path": java_lang_String;
    public "sun/nio/ch/FileChannelImpl/threads": sun_nio_ch_NativeThreadSet;
    public "sun/nio/ch/FileChannelImpl/positionLock": java_lang_Object;
    public static "sun/nio/ch/FileChannelImpl/transferSupported": number;
    public static "sun/nio/ch/FileChannelImpl/pipeSupported": number;
    public static "sun/nio/ch/FileChannelImpl/fileSupported": number;
    public static "sun/nio/ch/FileChannelImpl/MAPPED_TRANSFER_SIZE": Long;
    public static "sun/nio/ch/FileChannelImpl/TRANSFER_SIZE": number;
    public static "sun/nio/ch/FileChannelImpl/MAP_RO": number;
    public static "sun/nio/ch/FileChannelImpl/MAP_RW": number;
    public static "sun/nio/ch/FileChannelImpl/MAP_PV": number;
    public "sun/nio/ch/FileChannelImpl/fileLockTable": sun_nio_ch_FileLockTable;
    public static "sun/nio/ch/FileChannelImpl/isSharedFileLockTable": number;
    public static "sun/nio/ch/FileChannelImpl/propertyChecked": number;
    public static "sun/nio/ch/FileChannelImpl/$assertionsDisabled": number;
    public "<init>(Ljava/io/FileDescriptor;Ljava/lang/String;ZZZLjava/lang/Object;)V"(thread: JVMThread, args: [java_io_FileDescriptor, java_lang_String, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileChannelImpl/<init>(Ljava/io/FileDescriptor;Ljava/lang/String;ZZZLjava/lang/Object;)V"(thread: JVMThread, args: [java_io_FileDescriptor, java_lang_String, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open(Ljava/io/FileDescriptor;Ljava/lang/String;ZZLjava/lang/Object;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_io_FileDescriptor, java_lang_String, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public static "sun/nio/ch/FileChannelImpl/open(Ljava/io/FileDescriptor;Ljava/lang/String;ZZLjava/lang/Object;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_io_FileDescriptor, java_lang_String, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public static "open(Ljava/io/FileDescriptor;Ljava/lang/String;ZZZLjava/lang/Object;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_io_FileDescriptor, java_lang_String, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public static "sun/nio/ch/FileChannelImpl/open(Ljava/io/FileDescriptor;Ljava/lang/String;ZZZLjava/lang/Object;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_io_FileDescriptor, java_lang_String, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileChannelImpl/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileChannelImpl/implCloseChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileChannelImpl/read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileChannelImpl/write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "position()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/position()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "position(J)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "sun/nio/ch/FileChannelImpl/position(J)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "truncate(J)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "sun/nio/ch/FileChannelImpl/truncate(J)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "force(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileChannelImpl/force(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "transferToDirectlyInternal(JILjava/nio/channels/WritableByteChannel;Ljava/io/FileDescriptor;)J"(thread: JVMThread, args: [Long, any, number, java_nio_channels_WritableByteChannel, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/transferToDirectlyInternal(JILjava/nio/channels/WritableByteChannel;Ljava/io/FileDescriptor;)J"(thread: JVMThread, args: [Long, any, number, java_nio_channels_WritableByteChannel, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "transferToDirectly(JILjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, number, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/transferToDirectly(JILjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, number, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "transferToTrustedChannel(JJLjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, Long, any, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/transferToTrustedChannel(JJLjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, Long, any, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "transferToArbitraryChannel(JILjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, number, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/transferToArbitraryChannel(JILjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, number, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "transferTo(JJLjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, Long, any, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/transferTo(JJLjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, Long, any, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "transferFromFileChannel(Lsun/nio/ch/FileChannelImpl;JJ)J"(thread: JVMThread, args: [sun_nio_ch_FileChannelImpl, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/transferFromFileChannel(Lsun/nio/ch/FileChannelImpl;JJ)J"(thread: JVMThread, args: [sun_nio_ch_FileChannelImpl, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "transferFromArbitraryChannel(Ljava/nio/channels/ReadableByteChannel;JJ)J"(thread: JVMThread, args: [java_nio_channels_ReadableByteChannel, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/transferFromArbitraryChannel(Ljava/nio/channels/ReadableByteChannel;JJ)J"(thread: JVMThread, args: [java_nio_channels_ReadableByteChannel, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "transferFrom(Ljava/nio/channels/ReadableByteChannel;JJ)J"(thread: JVMThread, args: [java_nio_channels_ReadableByteChannel, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/transferFrom(Ljava/nio/channels/ReadableByteChannel;JJ)J"(thread: JVMThread, args: [java_nio_channels_ReadableByteChannel, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "read(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileChannelImpl/read(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readInternal(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileChannelImpl/readInternal(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileChannelImpl/write(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeInternal(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileChannelImpl/writeInternal(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "unmap(Ljava/nio/MappedByteBuffer;)V"(thread: JVMThread, args: [java_nio_MappedByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileChannelImpl/unmap(Ljava/nio/MappedByteBuffer;)V"(thread: JVMThread, args: [java_nio_MappedByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "map(Ljava/nio/channels/FileChannel$MapMode;JJ)Ljava/nio/MappedByteBuffer;"(thread: JVMThread, args: [java_nio_channels_FileChannel$MapMode, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_MappedByteBuffer) => void): void;
    public "sun/nio/ch/FileChannelImpl/map(Ljava/nio/channels/FileChannel$MapMode;JJ)Ljava/nio/MappedByteBuffer;"(thread: JVMThread, args: [java_nio_channels_FileChannel$MapMode, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_MappedByteBuffer) => void): void;
    public static "getMappedBufferPool()Lsun/misc/JavaNioAccess$BufferPool;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_JavaNioAccess$BufferPool) => void): void;
    public static "sun/nio/ch/FileChannelImpl/getMappedBufferPool()Lsun/misc/JavaNioAccess$BufferPool;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_JavaNioAccess$BufferPool) => void): void;
    public static "isSharedFileLockTable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileChannelImpl/isSharedFileLockTable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fileLockTable()Lsun/nio/ch/FileLockTable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_FileLockTable) => void): void;
    public "sun/nio/ch/FileChannelImpl/fileLockTable()Lsun/nio/ch/FileLockTable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_FileLockTable) => void): void;
    public "lock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "sun/nio/ch/FileChannelImpl/lock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "tryLock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "sun/nio/ch/FileChannelImpl/tryLock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "release(Lsun/nio/ch/FileLockImpl;)V"(thread: JVMThread, args: [sun_nio_ch_FileLockImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileChannelImpl/release(Lsun/nio/ch/FileLockImpl;)V"(thread: JVMThread, args: [sun_nio_ch_FileLockImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "map0(IJJ)J"(thread: JVMThread, args: [number, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/map0(IJJ)J"(thread: JVMThread, args: [number, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "unmap0(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileChannelImpl/unmap0(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "transferTo0(Ljava/io/FileDescriptor;JJLjava/io/FileDescriptor;)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/transferTo0(Ljava/io/FileDescriptor;JJLjava/io/FileDescriptor;)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "position0(Ljava/io/FileDescriptor;J)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileChannelImpl/position0(Ljava/io/FileDescriptor;J)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "initIDs()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/FileChannelImpl/initIDs()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "truncate(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "sun/nio/ch/FileChannelImpl/truncate(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "position(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "sun/nio/ch/FileChannelImpl/position(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public static "access$000(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileChannelImpl/access$000(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileChannelImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_channels_SeekableByteChannel extends java_lang_Object, java_nio_channels_ByteChannel {
    "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "position()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "position(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    "size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "truncate(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_FileLockImpl extends java_nio_channels_FileLock {
    public "sun/nio/ch/FileLockImpl/valid": number;
    public static "sun/nio/ch/FileLockImpl/$assertionsDisabled": number;
    public "<init>(Ljava/nio/channels/FileChannel;JJZ)V"(thread: JVMThread, args: [java_nio_channels_FileChannel, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileLockImpl/<init>(Ljava/nio/channels/FileChannel;JJZ)V"(thread: JVMThread, args: [java_nio_channels_FileChannel, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/AsynchronousFileChannel;JJZ)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousFileChannel, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileLockImpl/<init>(Ljava/nio/channels/AsynchronousFileChannel;JJZ)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousFileChannel, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileLockImpl/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileLockImpl/invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "release()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileLockImpl/release()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileLockImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_channels_AsynchronousFileChannel extends java_lang_Object implements java_nio_channels_AsynchronousChannel {
    public static "java/nio/channels/AsynchronousFileChannel/NO_ATTRIBUTES": JVMArray<java_nio_file_attribute_FileAttribute>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open(Ljava/nio/file/Path;Ljava/util/Set;Ljava/util/concurrent/ExecutorService;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, java_util_concurrent_ExecutorService, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public static "java/nio/channels/AsynchronousFileChannel/open(Ljava/nio/file/Path;Ljava/util/Set;Ljava/util/concurrent/ExecutorService;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, java_util_concurrent_ExecutorService, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public static "open(Ljava/nio/file/Path;[Ljava/nio/file/OpenOption;)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_OpenOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public static "java/nio/channels/AsynchronousFileChannel/open(Ljava/nio/file/Path;[Ljava/nio/file/OpenOption;)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_OpenOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public "size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "truncate(J)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/truncate(J)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public "force(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/force(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "lock(JJZLjava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [Long, any, Long, any, number, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/lock(JJZLjava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [Long, any, Long, any, number, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "lock(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/lock(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "lock(JJZ)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/lock(JJZ)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "lock()Ljava/util/concurrent/Future;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/lock()Ljava/util/concurrent/Future;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "tryLock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/tryLock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "tryLock()Ljava/nio/channels/FileLock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/tryLock()Ljava/nio/channels/FileLock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "read(Ljava/nio/ByteBuffer;JLjava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/read(Ljava/nio/ByteBuffer;JLjava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/nio/ByteBuffer;J)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/read(Ljava/nio/ByteBuffer;J)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "write(Ljava/nio/ByteBuffer;JLjava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/write(Ljava/nio/ByteBuffer;JLjava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/nio/ByteBuffer;J)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/nio/channels/AsynchronousFileChannel/write(Ljava/nio/ByteBuffer;J)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/channels/AsynchronousFileChannel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_concurrent_Future extends java_lang_Object {
    "cancel(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isCancelled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isDone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "get(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_concurrent_TimeUnit extends java_lang_Enum {
    public static "java/util/concurrent/TimeUnit/NANOSECONDS": java_util_concurrent_TimeUnit;
    public static "java/util/concurrent/TimeUnit/MICROSECONDS": java_util_concurrent_TimeUnit;
    public static "java/util/concurrent/TimeUnit/MILLISECONDS": java_util_concurrent_TimeUnit;
    public static "java/util/concurrent/TimeUnit/SECONDS": java_util_concurrent_TimeUnit;
    public static "java/util/concurrent/TimeUnit/MINUTES": java_util_concurrent_TimeUnit;
    public static "java/util/concurrent/TimeUnit/HOURS": java_util_concurrent_TimeUnit;
    public static "java/util/concurrent/TimeUnit/DAYS": java_util_concurrent_TimeUnit;
    public static "java/util/concurrent/TimeUnit/C0": Long;
    public static "java/util/concurrent/TimeUnit/C1": Long;
    public static "java/util/concurrent/TimeUnit/C2": Long;
    public static "java/util/concurrent/TimeUnit/C3": Long;
    public static "java/util/concurrent/TimeUnit/C4": Long;
    public static "java/util/concurrent/TimeUnit/C5": Long;
    public static "java/util/concurrent/TimeUnit/C6": Long;
    public static "java/util/concurrent/TimeUnit/MAX": Long;
    public static "java/util/concurrent/TimeUnit/$VALUES": JVMArray<java_util_concurrent_TimeUnit>;
    public static "values()[Ljava/util/concurrent/TimeUnit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_concurrent_TimeUnit>) => void): void;
    public static "java/util/concurrent/TimeUnit/values()[Ljava/util/concurrent/TimeUnit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_concurrent_TimeUnit>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/util/concurrent/TimeUnit;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_TimeUnit) => void): void;
    public static "java/util/concurrent/TimeUnit/valueOf(Ljava/lang/String;)Ljava/util/concurrent/TimeUnit;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_TimeUnit) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/TimeUnit/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "x(JJJ)J"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/concurrent/TimeUnit/x(JJJ)J"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "convert(JLjava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit/convert(JLjava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toNanos(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit/toNanos(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toMicros(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit/toMicros(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toMillis(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit/toMillis(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toSeconds(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit/toSeconds(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toMinutes(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit/toMinutes(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toHours(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit/toHours(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toDays(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit/toDays(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "excessNanos(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/TimeUnit/excessNanos(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "timedWait(Ljava/lang/Object;J)V"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/TimeUnit/timedWait(Ljava/lang/Object;J)V"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "timedJoin(Ljava/lang/Thread;J)V"(thread: JVMThread, args: [java_lang_Thread, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/TimeUnit/timedJoin(Ljava/lang/Thread;J)V"(thread: JVMThread, args: [java_lang_Thread, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sleep(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/TimeUnit/sleep(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;ILjava/util/concurrent/TimeUnit$1;)V"(thread: JVMThread, args: [java_lang_String, number, java_util_concurrent_TimeUnit$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/TimeUnit/<init>(Ljava/lang/String;ILjava/util/concurrent/TimeUnit$1;)V"(thread: JVMThread, args: [java_lang_String, number, java_util_concurrent_TimeUnit$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/TimeUnit/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_TimeUnit$1 extends java_util_concurrent_TimeUnit {
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/TimeUnit$1/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toNanos(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit$1/toNanos(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toMicros(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit$1/toMicros(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toMillis(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit$1/toMillis(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toSeconds(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit$1/toSeconds(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toMinutes(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit$1/toMinutes(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toHours(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit$1/toHours(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toDays(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit$1/toDays(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "convert(JLjava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/TimeUnit$1/convert(JLjava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "excessNanos(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/TimeUnit$1/excessNanos(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_nio_channels_CompletionHandler extends java_lang_Object {
    "completed(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "failed(Ljava/lang/Throwable;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Throwable, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_file_OpenOption extends java_lang_Object {
  }
  export interface java_util_concurrent_ExecutorService extends java_lang_Object, java_util_concurrent_Executor {
    "shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "shutdownNow()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "isShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isTerminated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "awaitTermination(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "submit(Ljava/util/concurrent/Callable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_util_concurrent_Callable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    "submit(Ljava/lang/Runnable;Ljava/lang/Object;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    "submit(Ljava/lang/Runnable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    "invokeAll(Ljava/util/Collection;)Ljava/util/List;"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "invokeAll(Ljava/util/Collection;JLjava/util/concurrent/TimeUnit;)Ljava/util/List;"(thread: JVMThread, args: [java_util_Collection, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "invokeAny(Ljava/util/Collection;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "invokeAny(Ljava/util/Collection;JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_concurrent_Callable extends java_lang_Object {
    "call()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_List extends java_lang_Object, java_util_Collection {
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "listIterator()Ljava/util/ListIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    "listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    "subList(II)Ljava/util/List;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
  }
  export interface java_util_ListIterator extends java_lang_Object, java_util_Iterator {
    "hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "next()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "hasPrevious()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "previous()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "nextIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "previousIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "set(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "add(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_concurrent_Executor extends java_lang_Object {
    "execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_file_Path extends java_lang_Object, java_lang_Comparable, java_lang_Iterable, java_nio_file_Watchable {
    "getFileSystem()Ljava/nio/file/FileSystem;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    "isAbsolute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getRoot()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "getFileName()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "getParent()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "getNameCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getName(I)Ljava/nio/file/Path;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "subpath(II)Ljava/nio/file/Path;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "startsWith(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "startsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "endsWith(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "endsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "normalize()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "resolve(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "resolve(Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "resolveSibling(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "resolveSibling(Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "relativize(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "toUri()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    "toAbsolutePath()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "toRealPath([Ljava/nio/file/LinkOption;)Ljava/nio/file/Path;"(thread: JVMThread, args: [JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "toFile()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    "register(Ljava/nio/file/WatchService;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_WatchService, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    "register(Ljava/nio/file/WatchService;[Ljava/nio/file/WatchEvent$Kind;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_WatchService, JVMArray<java_nio_file_WatchEvent$Kind>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "compareTo(Ljava/nio/file/Path;)I"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
  }
  export interface java_nio_file_WatchEvent$Modifier extends java_lang_Object {
    "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_nio_file_WatchEvent$Kind extends java_lang_Object {
    "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "type()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
  }
  export interface java_nio_file_WatchService extends java_lang_Object, java_io_Closeable {
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "poll()Ljava/nio/file/WatchKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    "poll(JLjava/util/concurrent/TimeUnit;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    "take()Ljava/nio/file/WatchKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
  }
  export interface java_nio_file_WatchKey extends java_lang_Object {
    "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "pollEvents()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "reset()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "cancel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "watchable()Ljava/nio/file/Watchable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Watchable) => void): void;
  }
  export class java_nio_file_LinkOption extends java_lang_Enum implements java_nio_file_OpenOption, java_nio_file_CopyOption {
    public static "java/nio/file/LinkOption/NOFOLLOW_LINKS": java_nio_file_LinkOption;
    public static "java/nio/file/LinkOption/$VALUES": JVMArray<java_nio_file_LinkOption>;
    public static "values()[Ljava/nio/file/LinkOption;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_nio_file_LinkOption>) => void): void;
    public static "java/nio/file/LinkOption/values()[Ljava/nio/file/LinkOption;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_nio_file_LinkOption>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/nio/file/LinkOption;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_LinkOption) => void): void;
    public static "java/nio/file/LinkOption/valueOf(Ljava/lang/String;)Ljava/nio/file/LinkOption;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_LinkOption) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/LinkOption/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/file/LinkOption/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_file_CopyOption extends java_lang_Object {
  }
  export class java_net_URI extends java_lang_Object implements java_lang_Comparable, java_io_Serializable {
    public static "java/net/URI/serialVersionUID": Long;
    public "java/net/URI/scheme": java_lang_String;
    public "java/net/URI/fragment": java_lang_String;
    public "java/net/URI/authority": java_lang_String;
    public "java/net/URI/userInfo": java_lang_String;
    public "java/net/URI/host": java_lang_String;
    public "java/net/URI/port": number;
    public "java/net/URI/path": java_lang_String;
    public "java/net/URI/query": java_lang_String;
    public "java/net/URI/schemeSpecificPart": java_lang_String;
    public "java/net/URI/hash": number;
    public "java/net/URI/decodedUserInfo": java_lang_String;
    public "java/net/URI/decodedAuthority": java_lang_String;
    public "java/net/URI/decodedPath": java_lang_String;
    public "java/net/URI/decodedQuery": java_lang_String;
    public "java/net/URI/decodedFragment": java_lang_String;
    public "java/net/URI/decodedSchemeSpecificPart": java_lang_String;
    public "java/net/URI/string": java_lang_String;
    public static "java/net/URI/L_DIGIT": Long;
    public static "java/net/URI/H_DIGIT": Long;
    public static "java/net/URI/L_UPALPHA": Long;
    public static "java/net/URI/H_UPALPHA": Long;
    public static "java/net/URI/L_LOWALPHA": Long;
    public static "java/net/URI/H_LOWALPHA": Long;
    public static "java/net/URI/L_ALPHA": Long;
    public static "java/net/URI/H_ALPHA": Long;
    public static "java/net/URI/L_ALPHANUM": Long;
    public static "java/net/URI/H_ALPHANUM": Long;
    public static "java/net/URI/L_HEX": Long;
    public static "java/net/URI/H_HEX": Long;
    public static "java/net/URI/L_MARK": Long;
    public static "java/net/URI/H_MARK": Long;
    public static "java/net/URI/L_UNRESERVED": Long;
    public static "java/net/URI/H_UNRESERVED": Long;
    public static "java/net/URI/L_RESERVED": Long;
    public static "java/net/URI/H_RESERVED": Long;
    public static "java/net/URI/L_ESCAPED": Long;
    public static "java/net/URI/H_ESCAPED": Long;
    public static "java/net/URI/L_URIC": Long;
    public static "java/net/URI/H_URIC": Long;
    public static "java/net/URI/L_PCHAR": Long;
    public static "java/net/URI/H_PCHAR": Long;
    public static "java/net/URI/L_PATH": Long;
    public static "java/net/URI/H_PATH": Long;
    public static "java/net/URI/L_DASH": Long;
    public static "java/net/URI/H_DASH": Long;
    public static "java/net/URI/L_DOT": Long;
    public static "java/net/URI/H_DOT": Long;
    public static "java/net/URI/L_USERINFO": Long;
    public static "java/net/URI/H_USERINFO": Long;
    public static "java/net/URI/L_REG_NAME": Long;
    public static "java/net/URI/H_REG_NAME": Long;
    public static "java/net/URI/L_SERVER": Long;
    public static "java/net/URI/H_SERVER": Long;
    public static "java/net/URI/L_SERVER_PERCENT": Long;
    public static "java/net/URI/H_SERVER_PERCENT": Long;
    public static "java/net/URI/L_LEFT_BRACKET": Long;
    public static "java/net/URI/H_LEFT_BRACKET": Long;
    public static "java/net/URI/L_SCHEME": Long;
    public static "java/net/URI/H_SCHEME": Long;
    public static "java/net/URI/L_URIC_NO_SLASH": Long;
    public static "java/net/URI/H_URIC_NO_SLASH": Long;
    public static "java/net/URI/hexDigits": JVMArray<number>;
    public static "java/net/URI/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "create(Ljava/lang/String;)Ljava/net/URI;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public static "java/net/URI/create(Ljava/lang/String;)Ljava/net/URI;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "parseServerAuthority()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "java/net/URI/parseServerAuthority()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "normalize()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "java/net/URI/normalize()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "resolve(Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "java/net/URI/resolve(Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "resolve(Ljava/lang/String;)Ljava/net/URI;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "java/net/URI/resolve(Ljava/lang/String;)Ljava/net/URI;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "relativize(Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "java/net/URI/relativize(Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "toURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "java/net/URI/toURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "getScheme()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getScheme()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isAbsolute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URI/isAbsolute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOpaque()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URI/isOpaque()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRawSchemeSpecificPart()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getRawSchemeSpecificPart()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSchemeSpecificPart()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getSchemeSpecificPart()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRawAuthority()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getRawAuthority()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAuthority()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getAuthority()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRawUserInfo()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getRawUserInfo()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getUserInfo()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getUserInfo()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getHost()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getHost()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URI/getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRawPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getRawPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRawQuery()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getRawQuery()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getQuery()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getQuery()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRawFragment()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getRawFragment()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFragment()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/getFragment()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URI/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URI/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/net/URI;)I"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URI/compareTo(Ljava/net/URI;)I"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toASCIIString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/toASCIIString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "toLower(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/toLower(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toUpper(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/toUpper(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "equal(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/equal(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "equalIgnoringCase(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/equalIgnoringCase(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hash(ILjava/lang/String;)I"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/hash(ILjava/lang/String;)I"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "normalizedHash(ILjava/lang/String;)I"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/normalizedHash(ILjava/lang/String;)I"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hashIgnoringCase(ILjava/lang/String;)I"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/hashIgnoringCase(ILjava/lang/String;)I"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(Ljava/lang/String;Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/compare(Ljava/lang/String;Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compareIgnoringCase(Ljava/lang/String;Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/compareIgnoringCase(Ljava/lang/String;Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkPath(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URI/checkPath(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "appendAuthority(Ljava/lang/StringBuffer;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_StringBuffer, java_lang_String, java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/appendAuthority(Ljava/lang/StringBuffer;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_StringBuffer, java_lang_String, java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "appendSchemeSpecificPart(Ljava/lang/StringBuffer;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_StringBuffer, java_lang_String, java_lang_String, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/appendSchemeSpecificPart(Ljava/lang/StringBuffer;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_StringBuffer, java_lang_String, java_lang_String, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "appendFragment(Ljava/lang/StringBuffer;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_StringBuffer, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/appendFragment(Ljava/lang/StringBuffer;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_StringBuffer, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URI/toString(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "defineSchemeSpecificPart()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/defineSchemeSpecificPart()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "defineString()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URI/defineString()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "resolvePath(Ljava/lang/String;Ljava/lang/String;Z)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/resolvePath(Ljava/lang/String;Ljava/lang/String;Z)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "resolve(Ljava/net/URI;Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI, java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public static "java/net/URI/resolve(Ljava/net/URI;Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI, java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public static "normalize(Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public static "java/net/URI/normalize(Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public static "relativize(Ljava/net/URI;Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI, java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public static "java/net/URI/relativize(Ljava/net/URI;Ljava/net/URI;)Ljava/net/URI;"(thread: JVMThread, args: [java_net_URI, java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public static "needsNormalization(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/needsNormalization(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "split([C[I)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URI/split([C[I)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "join([C[I)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/join([C[I)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "removeDots([C[I)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URI/removeDots([C[I)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "maybeAddLeadingDot([C[I)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URI/maybeAddLeadingDot([C[I)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "normalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/normalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "lowMask(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/lowMask(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "highMask(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/highMask(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "lowMask(CC)J"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/lowMask(CC)J"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "highMask(CC)J"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/highMask(CC)J"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "match(CJJ)Z"(thread: JVMThread, args: [number, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/match(CJJ)Z"(thread: JVMThread, args: [number, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "appendEscape(Ljava/lang/StringBuffer;B)V"(thread: JVMThread, args: [java_lang_StringBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URI/appendEscape(Ljava/lang/StringBuffer;B)V"(thread: JVMThread, args: [java_lang_StringBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "appendEncoded(Ljava/lang/StringBuffer;C)V"(thread: JVMThread, args: [java_lang_StringBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URI/appendEncoded(Ljava/lang/StringBuffer;C)V"(thread: JVMThread, args: [java_lang_StringBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "quote(Ljava/lang/String;JJ)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/quote(Ljava/lang/String;JJ)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "encode(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/encode(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "decode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/decode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "decode(CC)B"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/decode(CC)B"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "decode(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/decode(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URI/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$002(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/access$002(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$100()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$100()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$200()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$200()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$300(CJJ)Z"(thread: JVMThread, args: [number, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/access$300(CJJ)Z"(thread: JVMThread, args: [number, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$400()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$400()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$500()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$500()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$600()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$600()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$702(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/access$702(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$800()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$800()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$900()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$900()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1002(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/access$1002(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$1102(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/access$1102(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$1200()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$1200()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1300()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$1300()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1402(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/access$1402(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$1502(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/access$1502(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$1600()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$1600()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1700()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$1700()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1800()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$1800()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1900()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$1900()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$2000()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$2000()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$2100()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$2100()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$2202(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/access$2202(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$2302(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/access$2302(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$2402(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URI/access$2402(Ljava/net/URI;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URI, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$2502(Ljava/net/URI;I)I"(thread: JVMThread, args: [java_net_URI, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URI/access$2502(Ljava/net/URI;I)I"(thread: JVMThread, args: [java_net_URI, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$2600()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$2600()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$2700()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$2700()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$2800()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$2800()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$2900()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$2900()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$3000()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$3000()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$3100()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$3100()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$3200()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$3200()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$3300()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$3300()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$3400()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URI/access$3400()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URI/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_file_FileSystem extends java_lang_Object implements java_io_Closeable {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/FileSystem/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "provider()Ljava/nio/file/spi/FileSystemProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileSystemProvider) => void): void;
    public "java/nio/file/FileSystem/provider()Ljava/nio/file/spi/FileSystemProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileSystemProvider) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/FileSystem/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/FileSystem/isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/FileSystem/isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSeparator()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/FileSystem/getSeparator()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRootDirectories()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "java/nio/file/FileSystem/getRootDirectories()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "getFileStores()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "java/nio/file/FileSystem/getFileStores()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "supportedFileAttributeViews()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/nio/file/FileSystem/supportedFileAttributeViews()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getPath(Ljava/lang/String;[Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "java/nio/file/FileSystem/getPath(Ljava/lang/String;[Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "getPathMatcher(Ljava/lang/String;)Ljava/nio/file/PathMatcher;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_PathMatcher) => void): void;
    public "java/nio/file/FileSystem/getPathMatcher(Ljava/lang/String;)Ljava/nio/file/PathMatcher;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_PathMatcher) => void): void;
    public "getUserPrincipalLookupService()Ljava/nio/file/attribute/UserPrincipalLookupService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_UserPrincipalLookupService) => void): void;
    public "java/nio/file/FileSystem/getUserPrincipalLookupService()Ljava/nio/file/attribute/UserPrincipalLookupService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_UserPrincipalLookupService) => void): void;
    public "newWatchService()Ljava/nio/file/WatchService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchService) => void): void;
    public "java/nio/file/FileSystem/newWatchService()Ljava/nio/file/WatchService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchService) => void): void;
  }
  export class java_nio_file_attribute_UserPrincipalLookupService extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/attribute/UserPrincipalLookupService/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lookupPrincipalByName(Ljava/lang/String;)Ljava/nio/file/attribute/UserPrincipal;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_UserPrincipal) => void): void;
    public "java/nio/file/attribute/UserPrincipalLookupService/lookupPrincipalByName(Ljava/lang/String;)Ljava/nio/file/attribute/UserPrincipal;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_UserPrincipal) => void): void;
    public "lookupPrincipalByGroupName(Ljava/lang/String;)Ljava/nio/file/attribute/GroupPrincipal;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_GroupPrincipal) => void): void;
    public "java/nio/file/attribute/UserPrincipalLookupService/lookupPrincipalByGroupName(Ljava/lang/String;)Ljava/nio/file/attribute/GroupPrincipal;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_GroupPrincipal) => void): void;
  }
  export interface java_nio_file_attribute_GroupPrincipal extends java_lang_Object, java_nio_file_attribute_UserPrincipal {
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "implies(Ljavax/security/auth/Subject;)Z"(thread: JVMThread, args: [javax_security_auth_Subject], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class javax_security_auth_Subject extends java_lang_Object implements java_io_Serializable {
    public static "javax/security/auth/Subject/serialVersionUID": Long;
    public "javax/security/auth/Subject/principals": java_util_Set;
    public "javax/security/auth/Subject/pubCredentials": java_util_Set;
    public "javax/security/auth/Subject/privCredentials": java_util_Set;
    public "javax/security/auth/Subject/readOnly": number;
    public static "javax/security/auth/Subject/PRINCIPAL_SET": number;
    public static "javax/security/auth/Subject/PUB_CREDENTIAL_SET": number;
    public static "javax/security/auth/Subject/PRIV_CREDENTIAL_SET": number;
    public static "javax/security/auth/Subject/NULL_PD_ARRAY": JVMArray<java_security_ProtectionDomain>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/security/auth/Subject/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ZLjava/util/Set;Ljava/util/Set;Ljava/util/Set;)V"(thread: JVMThread, args: [number, java_util_Set, java_util_Set, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/security/auth/Subject/<init>(ZLjava/util/Set;Ljava/util/Set;Ljava/util/Set;)V"(thread: JVMThread, args: [number, java_util_Set, java_util_Set, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "setReadOnly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/security/auth/Subject/setReadOnly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/security/auth/Subject/isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getSubject(Ljava/security/AccessControlContext;)Ljavax/security/auth/Subject;"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: javax_security_auth_Subject) => void): void;
    public static "javax/security/auth/Subject/getSubject(Ljava/security/AccessControlContext;)Ljavax/security/auth/Subject;"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: javax_security_auth_Subject) => void): void;
    public static "doAs(Ljavax/security/auth/Subject;Ljava/security/PrivilegedAction;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_PrivilegedAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "javax/security/auth/Subject/doAs(Ljavax/security/auth/Subject;Ljava/security/PrivilegedAction;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_PrivilegedAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doAs(Ljavax/security/auth/Subject;Ljava/security/PrivilegedExceptionAction;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_PrivilegedExceptionAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "javax/security/auth/Subject/doAs(Ljavax/security/auth/Subject;Ljava/security/PrivilegedExceptionAction;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_PrivilegedExceptionAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doAsPrivileged(Ljavax/security/auth/Subject;Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_PrivilegedAction, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "javax/security/auth/Subject/doAsPrivileged(Ljavax/security/auth/Subject;Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_PrivilegedAction, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doAsPrivileged(Ljavax/security/auth/Subject;Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_PrivilegedExceptionAction, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "javax/security/auth/Subject/doAsPrivileged(Ljavax/security/auth/Subject;Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_PrivilegedExceptionAction, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "createContext(Ljavax/security/auth/Subject;Ljava/security/AccessControlContext;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "javax/security/auth/Subject/createContext(Ljavax/security/auth/Subject;Ljava/security/AccessControlContext;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [javax_security_auth_Subject, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "getPrincipals()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "javax/security/auth/Subject/getPrincipals()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getPrincipals(Ljava/lang/Class;)Ljava/util/Set;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "javax/security/auth/Subject/getPrincipals(Ljava/lang/Class;)Ljava/util/Set;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getPublicCredentials()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "javax/security/auth/Subject/getPublicCredentials()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getPrivateCredentials()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "javax/security/auth/Subject/getPrivateCredentials()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getPublicCredentials(Ljava/lang/Class;)Ljava/util/Set;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "javax/security/auth/Subject/getPublicCredentials(Ljava/lang/Class;)Ljava/util/Set;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getPrivateCredentials(Ljava/lang/Class;)Ljava/util/Set;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "javax/security/auth/Subject/getPrivateCredentials(Ljava/lang/Class;)Ljava/util/Set;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/security/auth/Subject/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/security/auth/Subject/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/security/auth/Subject/toString(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/security/auth/Subject/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCredHashCode(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/security/auth/Subject/getCredHashCode(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/security/auth/Subject/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/security/auth/Subject/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/security/auth/Subject/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_nio_file_attribute_UserPrincipal extends java_lang_Object, java_security_Principal {
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "implies(Ljavax/security/auth/Subject;)Z"(thread: JVMThread, args: [javax_security_auth_Subject], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_security_Principal extends java_lang_Object {
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "implies(Ljavax/security/auth/Subject;)Z"(thread: JVMThread, args: [javax_security_auth_Subject], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_nio_file_PathMatcher extends java_lang_Object {
    "matches(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_nio_file_Watchable extends java_lang_Object {
    "register(Ljava/nio/file/WatchService;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_WatchService, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    "register(Ljava/nio/file/WatchService;[Ljava/nio/file/WatchEvent$Kind;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_WatchService, JVMArray<java_nio_file_WatchEvent$Kind>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
  }
  export interface java_nio_file_attribute_FileAttribute extends java_lang_Object {
    "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "value()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_nio_channels_AsynchronousChannel extends java_lang_Object, java_nio_channels_Channel {
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_channels_FileLock extends java_lang_Object implements java_lang_AutoCloseable {
    public "java/nio/channels/FileLock/channel": java_nio_channels_Channel;
    public "java/nio/channels/FileLock/position": Long;
    public "java/nio/channels/FileLock/size": Long;
    public "java/nio/channels/FileLock/shared": number;
    public "<init>(Ljava/nio/channels/FileChannel;JJZ)V"(thread: JVMThread, args: [java_nio_channels_FileChannel, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/FileLock/<init>(Ljava/nio/channels/FileChannel;JJZ)V"(thread: JVMThread, args: [java_nio_channels_FileChannel, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/AsynchronousFileChannel;JJZ)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousFileChannel, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/FileLock/<init>(Ljava/nio/channels/AsynchronousFileChannel;JJZ)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousFileChannel, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "channel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "java/nio/channels/FileLock/channel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "acquiredBy()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public "java/nio/channels/FileLock/acquiredBy()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public "position()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileLock/position()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileLock/size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "isShared()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/FileLock/isShared()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "overlaps(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/FileLock/overlaps(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/FileLock/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "release()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/FileLock/release()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/FileLock/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/channels/FileLock/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface sun_misc_JavaNioAccess$BufferPool extends java_lang_Object {
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getTotalCapacity()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getMemoryUsed()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class java_nio_channels_FileChannel$MapMode extends java_lang_Object {
    public static "java/nio/channels/FileChannel$MapMode/READ_ONLY": java_nio_channels_FileChannel$MapMode;
    public static "java/nio/channels/FileChannel$MapMode/READ_WRITE": java_nio_channels_FileChannel$MapMode;
    public static "java/nio/channels/FileChannel$MapMode/PRIVATE": java_nio_channels_FileChannel$MapMode;
    public "java/nio/channels/FileChannel$MapMode/name": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/FileChannel$MapMode/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/channels/FileChannel$MapMode/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/channels/FileChannel$MapMode/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_FileLockTable extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileLockTable/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "newSharedFileLockTable(Ljava/nio/channels/Channel;Ljava/io/FileDescriptor;)Lsun/nio/ch/FileLockTable;"(thread: JVMThread, args: [java_nio_channels_Channel, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_FileLockTable) => void): void;
    public static "sun/nio/ch/FileLockTable/newSharedFileLockTable(Ljava/nio/channels/Channel;Ljava/io/FileDescriptor;)Lsun/nio/ch/FileLockTable;"(thread: JVMThread, args: [java_nio_channels_Channel, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_FileLockTable) => void): void;
    public "add(Ljava/nio/channels/FileLock;)V"(thread: JVMThread, args: [java_nio_channels_FileLock], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileLockTable/add(Ljava/nio/channels/FileLock;)V"(thread: JVMThread, args: [java_nio_channels_FileLock], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/nio/channels/FileLock;)V"(thread: JVMThread, args: [java_nio_channels_FileLock], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileLockTable/remove(Ljava/nio/channels/FileLock;)V"(thread: JVMThread, args: [java_nio_channels_FileLock], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeAll()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/nio/ch/FileLockTable/removeAll()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "replace(Ljava/nio/channels/FileLock;Ljava/nio/channels/FileLock;)V"(thread: JVMThread, args: [java_nio_channels_FileLock, java_nio_channels_FileLock], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileLockTable/replace(Ljava/nio/channels/FileLock;Ljava/nio/channels/FileLock;)V"(thread: JVMThread, args: [java_nio_channels_FileLock, java_nio_channels_FileLock], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_NativeThreadSet extends java_lang_Object {
    public "sun/nio/ch/NativeThreadSet/elts": JVMArray<Long>;
    public "sun/nio/ch/NativeThreadSet/used": number;
    public "sun/nio/ch/NativeThreadSet/waitingToEmpty": number;
    public static "sun/nio/ch/NativeThreadSet/$assertionsDisabled": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeThreadSet/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "add()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeThreadSet/add()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeThreadSet/remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "signalAndWait()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeThreadSet/signalAndWait()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/NativeThreadSet/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_FileDispatcher extends sun_nio_ch_NativeDispatcher {
    public static "sun/nio/ch/FileDispatcher/NO_LOCK": number;
    public static "sun/nio/ch/FileDispatcher/LOCKED": number;
    public static "sun/nio/ch/FileDispatcher/RET_EX_LOCK": number;
    public static "sun/nio/ch/FileDispatcher/INTERRUPTED": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileDispatcher/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "force(Ljava/io/FileDescriptor;Z)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcher/force(Ljava/io/FileDescriptor;Z)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "truncate(Ljava/io/FileDescriptor;J)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcher/truncate(Ljava/io/FileDescriptor;J)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size(Ljava/io/FileDescriptor;)J"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileDispatcher/size(Ljava/io/FileDescriptor;)J"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "lock(Ljava/io/FileDescriptor;ZJJZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcher/lock(Ljava/io/FileDescriptor;ZJJZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "release(Ljava/io/FileDescriptor;JJ)V"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileDispatcher/release(Ljava/io/FileDescriptor;JJ)V"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "duplicateForMapping(Ljava/io/FileDescriptor;)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "sun/nio/ch/FileDispatcher/duplicateForMapping(Ljava/io/FileDescriptor;)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "canTransferToDirectly(Ljava/nio/channels/SelectableChannel;)Z"(thread: JVMThread, args: [java_nio_channels_SelectableChannel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcher/canTransferToDirectly(Ljava/nio/channels/SelectableChannel;)Z"(thread: JVMThread, args: [java_nio_channels_SelectableChannel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "transferToDirectlyNeedsPositionLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcher/transferToDirectlyNeedsPositionLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_nio_ch_NativeDispatcher extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeDispatcher/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeDispatcher/read(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "needsPositionLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeDispatcher/needsPositionLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "pread(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeDispatcher/pread(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readv(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/NativeDispatcher/readv(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeDispatcher/write(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "pwrite(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeDispatcher/pwrite(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writev(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/NativeDispatcher/writev(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "close(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeDispatcher/close(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "preClose(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeDispatcher/preClose(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_channels_FileChannel extends java_nio_channels_spi_AbstractInterruptibleChannel implements java_nio_channels_SeekableByteChannel, java_nio_channels_GatheringByteChannel, java_nio_channels_ScatteringByteChannel {
    public static "java/nio/channels/FileChannel/NO_ATTRIBUTES": JVMArray<java_nio_file_attribute_FileAttribute>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/FileChannel/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public static "java/nio/channels/FileChannel/open(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public static "open(Ljava/nio/file/Path;[Ljava/nio/file/OpenOption;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_OpenOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public static "java/nio/channels/FileChannel/open(Ljava/nio/file/Path;[Ljava/nio/file/OpenOption;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_OpenOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/FileChannel/read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileChannel/read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "read([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileChannel/read([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/FileChannel/write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileChannel/write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileChannel/write([Ljava/nio/ByteBuffer;)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "position()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileChannel/position()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "position(J)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "java/nio/channels/FileChannel/position(J)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileChannel/size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "truncate(J)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "java/nio/channels/FileChannel/truncate(J)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "force(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/FileChannel/force(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "transferTo(JJLjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, Long, any, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileChannel/transferTo(JJLjava/nio/channels/WritableByteChannel;)J"(thread: JVMThread, args: [Long, any, Long, any, java_nio_channels_WritableByteChannel], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "transferFrom(Ljava/nio/channels/ReadableByteChannel;JJ)J"(thread: JVMThread, args: [java_nio_channels_ReadableByteChannel, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/channels/FileChannel/transferFrom(Ljava/nio/channels/ReadableByteChannel;JJ)J"(thread: JVMThread, args: [java_nio_channels_ReadableByteChannel, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "read(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/FileChannel/read(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/FileChannel/write(Ljava/nio/ByteBuffer;J)I"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "map(Ljava/nio/channels/FileChannel$MapMode;JJ)Ljava/nio/MappedByteBuffer;"(thread: JVMThread, args: [java_nio_channels_FileChannel$MapMode, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_MappedByteBuffer) => void): void;
    public "java/nio/channels/FileChannel/map(Ljava/nio/channels/FileChannel$MapMode;JJ)Ljava/nio/MappedByteBuffer;"(thread: JVMThread, args: [java_nio_channels_FileChannel$MapMode, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_MappedByteBuffer) => void): void;
    public "lock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "java/nio/channels/FileChannel/lock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "lock()Ljava/nio/channels/FileLock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "java/nio/channels/FileChannel/lock()Ljava/nio/channels/FileLock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "tryLock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "java/nio/channels/FileChannel/tryLock(JJZ)Ljava/nio/channels/FileLock;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "tryLock()Ljava/nio/channels/FileLock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "java/nio/channels/FileChannel/tryLock()Ljava/nio/channels/FileLock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileLock) => void): void;
    public "truncate(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "java/nio/channels/FileChannel/truncate(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "position(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "java/nio/channels/FileChannel/position(J)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/channels/FileChannel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_IOUtil extends java_lang_Object {
    public static "sun/nio/ch/IOUtil/IOV_MAX": number;
    public static "sun/nio/ch/IOUtil/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/IOUtil/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "write(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;JLsun/nio/ch/NativeDispatcher;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, Long, any, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/IOUtil/write(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;JLsun/nio/ch/NativeDispatcher;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, Long, any, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "writeFromNativeBuffer(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;JLsun/nio/ch/NativeDispatcher;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, Long, any, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/IOUtil/writeFromNativeBuffer(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;JLsun/nio/ch/NativeDispatcher;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, Long, any, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "write(Ljava/io/FileDescriptor;[Ljava/nio/ByteBuffer;Lsun/nio/ch/NativeDispatcher;)J"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<java_nio_ByteBuffer>, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/IOUtil/write(Ljava/io/FileDescriptor;[Ljava/nio/ByteBuffer;Lsun/nio/ch/NativeDispatcher;)J"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<java_nio_ByteBuffer>, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "write(Ljava/io/FileDescriptor;[Ljava/nio/ByteBuffer;IILsun/nio/ch/NativeDispatcher;)J"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<java_nio_ByteBuffer>, number, number, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/IOUtil/write(Ljava/io/FileDescriptor;[Ljava/nio/ByteBuffer;IILsun/nio/ch/NativeDispatcher;)J"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<java_nio_ByteBuffer>, number, number, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "read(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;JLsun/nio/ch/NativeDispatcher;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, Long, any, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/IOUtil/read(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;JLsun/nio/ch/NativeDispatcher;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, Long, any, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "readIntoNativeBuffer(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;JLsun/nio/ch/NativeDispatcher;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, Long, any, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/IOUtil/readIntoNativeBuffer(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;JLsun/nio/ch/NativeDispatcher;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, Long, any, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "read(Ljava/io/FileDescriptor;[Ljava/nio/ByteBuffer;Lsun/nio/ch/NativeDispatcher;)J"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<java_nio_ByteBuffer>, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/IOUtil/read(Ljava/io/FileDescriptor;[Ljava/nio/ByteBuffer;Lsun/nio/ch/NativeDispatcher;)J"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<java_nio_ByteBuffer>, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "read(Ljava/io/FileDescriptor;[Ljava/nio/ByteBuffer;IILsun/nio/ch/NativeDispatcher;)J"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<java_nio_ByteBuffer>, number, number, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/IOUtil/read(Ljava/io/FileDescriptor;[Ljava/nio/ByteBuffer;IILsun/nio/ch/NativeDispatcher;)J"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<java_nio_ByteBuffer>, number, number, sun_nio_ch_NativeDispatcher], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "newFD(I)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "sun/nio/ch/IOUtil/newFD(I)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "randomBytes([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/IOUtil/randomBytes([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "makePipe(Z)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/IOUtil/makePipe(Z)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "drain(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/IOUtil/drain(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "configureBlocking(Ljava/io/FileDescriptor;Z)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/IOUtil/configureBlocking(Ljava/io/FileDescriptor;Z)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fdVal(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/IOUtil/fdVal(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setfdVal(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/IOUtil/setfdVal(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fdLimit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/IOUtil/fdLimit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "iovMax()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/IOUtil/iovMax()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/IOUtil/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "load()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/IOUtil/load()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/IOUtil/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_FileDispatcherImpl extends sun_nio_ch_FileDispatcher {
    public "<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/read(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "pread(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/pread(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readv(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/readv(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/write(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "pwrite(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/pwrite(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writev(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/writev(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "force(Ljava/io/FileDescriptor;Z)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/force(Ljava/io/FileDescriptor;Z)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "truncate(Ljava/io/FileDescriptor;J)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/truncate(Ljava/io/FileDescriptor;J)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size(Ljava/io/FileDescriptor;)J"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/size(Ljava/io/FileDescriptor;)J"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "lock(Ljava/io/FileDescriptor;ZJJZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/lock(Ljava/io/FileDescriptor;ZJJZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "release(Ljava/io/FileDescriptor;JJ)V"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/release(Ljava/io/FileDescriptor;JJ)V"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "close(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/close(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "preClose(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/preClose(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "duplicateForMapping(Ljava/io/FileDescriptor;)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/duplicateForMapping(Ljava/io/FileDescriptor;)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "canTransferToDirectly(Ljava/nio/channels/SelectableChannel;)Z"(thread: JVMThread, args: [java_nio_channels_SelectableChannel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/canTransferToDirectly(Ljava/nio/channels/SelectableChannel;)Z"(thread: JVMThread, args: [java_nio_channels_SelectableChannel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "transferToDirectlyNeedsPositionLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileDispatcherImpl/transferToDirectlyNeedsPositionLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "read0(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/read0(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "pread0(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/pread0(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "readv0(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/readv0(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "write0(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/write0(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "pwrite0(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/pwrite0(Ljava/io/FileDescriptor;JIJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "writev0(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/writev0(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "force0(Ljava/io/FileDescriptor;Z)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/force0(Ljava/io/FileDescriptor;Z)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "truncate0(Ljava/io/FileDescriptor;J)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/truncate0(Ljava/io/FileDescriptor;J)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "size0(Ljava/io/FileDescriptor;)J"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/size0(Ljava/io/FileDescriptor;)J"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "lock0(Ljava/io/FileDescriptor;ZJJZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/lock0(Ljava/io/FileDescriptor;ZJJZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "release0(Ljava/io/FileDescriptor;JJ)V"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/release0(Ljava/io/FileDescriptor;JJ)V"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "close0(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/close0(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "preClose0(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/preClose0(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "closeIntFD(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/closeIntFD(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileDispatcherImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_NativeThread extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeThread/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "current()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/NativeThread/current()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "signal(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/NativeThread/signal(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/NativeThread/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/NativeThread/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_UnixAsynchronousSocketChannelImpl extends sun_nio_ch_AsynchronousSocketChannelImpl implements sun_nio_ch_Port$PollableChannel {
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/nd": sun_nio_ch_NativeDispatcher;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/disableSynchronousRead": number;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/port": sun_nio_ch_Port;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/fdVal": number;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/updateLock": java_lang_Object;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/connectPending": number;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/connectHandler": java_nio_channels_CompletionHandler;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/connectAttachment": java_lang_Object;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/connectFuture": sun_nio_ch_PendingFuture;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/pendingRemote": java_net_SocketAddress;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/readPending": number;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/isScatteringRead": number;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/readBuffer": java_nio_ByteBuffer;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/readBuffers": JVMArray<java_nio_ByteBuffer>;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/readHandler": java_nio_channels_CompletionHandler;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/readAttachment": java_lang_Object;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/readFuture": sun_nio_ch_PendingFuture;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/readTimer": java_util_concurrent_Future;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/writePending": number;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/isGatheringWrite": number;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/writeBuffer": java_nio_ByteBuffer;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/writeBuffers": JVMArray<java_nio_ByteBuffer>;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/writeHandler": java_nio_channels_CompletionHandler;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/writeAttachment": java_lang_Object;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/writeFuture": sun_nio_ch_PendingFuture;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/writeTimer": java_util_concurrent_Future;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/readTimeoutTask": java_lang_Runnable;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/writeTimeoutTask": java_lang_Runnable;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/$assertionsDisabled": number;
    public "<init>(Lsun/nio/ch/Port;)V"(thread: JVMThread, args: [sun_nio_ch_Port], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/<init>(Lsun/nio/ch/Port;)V"(thread: JVMThread, args: [sun_nio_ch_Port], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/nio/ch/Port;Ljava/io/FileDescriptor;Ljava/net/InetSocketAddress;)V"(thread: JVMThread, args: [sun_nio_ch_Port, java_io_FileDescriptor, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/<init>(Lsun/nio/ch/Port;Ljava/io/FileDescriptor;Ljava/net/InetSocketAddress;)V"(thread: JVMThread, args: [sun_nio_ch_Port, java_io_FileDescriptor, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "group()Lsun/nio/ch/AsynchronousChannelGroupImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_AsynchronousChannelGroupImpl) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/group()Lsun/nio/ch/AsynchronousChannelGroupImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_AsynchronousChannelGroupImpl) => void): void;
    public "updateEvents()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/updateEvents()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lockAndUpdateEvents()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/lockAndUpdateEvents()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finish(ZZZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/finish(ZZZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "onEvent(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/onEvent(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "onCancel(Lsun/nio/ch/PendingFuture;)V"(thread: JVMThread, args: [sun_nio_ch_PendingFuture], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/onCancel(Lsun/nio/ch/PendingFuture;)V"(thread: JVMThread, args: [sun_nio_ch_PendingFuture], cb?: (e?: java_lang_Throwable) => void): void;
    public "setConnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/setConnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finishConnect(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/finishConnect(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "implConnect(Ljava/net/SocketAddress;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/implConnect(Ljava/net/SocketAddress;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "finishRead(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/finishRead(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "implRead(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/implRead(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "finishWrite(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/finishWrite(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "implWrite(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/UnixAsynchronousSocketChannelImpl/implWrite(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public static "checkConnect(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/checkConnect(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$000(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$100(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Z"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$100(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Z"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$102(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;Z)Z"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$102(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;Z)Z"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/nio/channels/CompletionHandler;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_CompletionHandler) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$200(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/nio/channels/CompletionHandler;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_CompletionHandler) => void): void;
    public static "access$300(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$300(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$400(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Lsun/nio/ch/PendingFuture;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_PendingFuture) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$400(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Lsun/nio/ch/PendingFuture;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_PendingFuture) => void): void;
    public static "access$500(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Z"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$500(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Z"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$502(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;Z)Z"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$502(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;Z)Z"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$600(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/nio/channels/CompletionHandler;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_CompletionHandler) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$600(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/nio/channels/CompletionHandler;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_CompletionHandler) => void): void;
    public static "access$700(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$700(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$800(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Lsun/nio/ch/PendingFuture;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_PendingFuture) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/access$800(Lsun/nio/ch/UnixAsynchronousSocketChannelImpl;)Lsun/nio/ch/PendingFuture;"(thread: JVMThread, args: [sun_nio_ch_UnixAsynchronousSocketChannelImpl], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_PendingFuture) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/UnixAsynchronousSocketChannelImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_AsynchronousChannelGroupImpl extends java_nio_channels_AsynchronousChannelGroup implements java_util_concurrent_Executor {
    public static "sun/nio/ch/AsynchronousChannelGroupImpl/internalThreadCount": number;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/pool": sun_nio_ch_ThreadPool;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/threadCount": java_util_concurrent_atomic_AtomicInteger;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/timeoutExecutor": java_util_concurrent_ScheduledThreadPoolExecutor;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/taskQueue": java_util_Queue;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/shutdown": java_util_concurrent_atomic_AtomicBoolean;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/shutdownNowLock": java_lang_Object;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/terminateInitiated": number;
    public "<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;Lsun/nio/ch/ThreadPool;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider, sun_nio_ch_ThreadPool], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;Lsun/nio/ch/ThreadPool;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider, sun_nio_ch_ThreadPool], cb?: (e?: java_lang_Throwable) => void): void;
    public "executor()Ljava/util/concurrent/ExecutorService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ExecutorService) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/executor()Ljava/util/concurrent/ExecutorService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ExecutorService) => void): void;
    public "isFixedThreadPool()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/isFixedThreadPool()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fixedThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/fixedThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "bindToGroup(Ljava/lang/Runnable;)Ljava/lang/Runnable;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runnable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/bindToGroup(Ljava/lang/Runnable;)Ljava/lang/Runnable;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runnable) => void): void;
    public "startInternalThread(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/startInternalThread(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "startThreads(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/startThreads(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "threadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/threadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "threadExit(Ljava/lang/Runnable;Z)I"(thread: JVMThread, args: [java_lang_Runnable, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/threadExit(Ljava/lang/Runnable;Z)I"(thread: JVMThread, args: [java_lang_Runnable, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "executeOnHandlerTask(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/executeOnHandlerTask(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "executeOnPooledThread(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/executeOnPooledThread(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "offerTask(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/offerTask(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "pollTask()Ljava/lang/Runnable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runnable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/pollTask()Ljava/lang/Runnable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runnable) => void): void;
    public "schedule(Ljava/lang/Runnable;JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/schedule(Ljava/lang/Runnable;JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "isShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/isShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isTerminated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/isTerminated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "attachForeignChannel(Ljava/nio/channels/Channel;Ljava/io/FileDescriptor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_channels_Channel, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/attachForeignChannel(Ljava/nio/channels/Channel;Ljava/io/FileDescriptor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_channels_Channel, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "detachForeignChannel(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/detachForeignChannel(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "closeAllChannels()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/closeAllChannels()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownHandlerTasks()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/shutdownHandlerTasks()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownExecutors()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/shutdownExecutors()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownNow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/shutdownNow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "detachFromThreadPool()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/detachFromThreadPool()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "awaitTermination(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/awaitTermination(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousChannelGroupImpl/execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Lsun/nio/ch/AsynchronousChannelGroupImpl;)Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "sun/nio/ch/AsynchronousChannelGroupImpl/access$000(Lsun/nio/ch/AsynchronousChannelGroupImpl;)Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "access$100(Lsun/nio/ch/AsynchronousChannelGroupImpl;)Ljava/util/concurrent/ScheduledThreadPoolExecutor;"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledThreadPoolExecutor) => void): void;
    public static "sun/nio/ch/AsynchronousChannelGroupImpl/access$100(Lsun/nio/ch/AsynchronousChannelGroupImpl;)Ljava/util/concurrent/ScheduledThreadPoolExecutor;"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledThreadPoolExecutor) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/AsynchronousChannelGroupImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_channels_spi_AsynchronousChannelProvider extends java_lang_Object {
    public static "checkPermission()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "java/nio/channels/spi/AsynchronousChannelProvider/checkPermission()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "<init>(Ljava/lang/Void;)V"(thread: JVMThread, args: [java_lang_Void], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AsynchronousChannelProvider/<init>(Ljava/lang/Void;)V"(thread: JVMThread, args: [java_lang_Void], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/spi/AsynchronousChannelProvider/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "provider()Ljava/nio/channels/spi/AsynchronousChannelProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AsynchronousChannelProvider) => void): void;
    public static "java/nio/channels/spi/AsynchronousChannelProvider/provider()Ljava/nio/channels/spi/AsynchronousChannelProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AsynchronousChannelProvider) => void): void;
    public "openAsynchronousChannelGroup(ILjava/util/concurrent/ThreadFactory;)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public "java/nio/channels/spi/AsynchronousChannelProvider/openAsynchronousChannelGroup(ILjava/util/concurrent/ThreadFactory;)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public "openAsynchronousChannelGroup(Ljava/util/concurrent/ExecutorService;I)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [java_util_concurrent_ExecutorService, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public "java/nio/channels/spi/AsynchronousChannelProvider/openAsynchronousChannelGroup(Ljava/util/concurrent/ExecutorService;I)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [java_util_concurrent_ExecutorService, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public "openAsynchronousServerSocketChannel(Ljava/nio/channels/AsynchronousChannelGroup;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannelGroup], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "java/nio/channels/spi/AsynchronousChannelProvider/openAsynchronousServerSocketChannel(Ljava/nio/channels/AsynchronousChannelGroup;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannelGroup], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "openAsynchronousSocketChannel(Ljava/nio/channels/AsynchronousChannelGroup;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannelGroup], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "java/nio/channels/spi/AsynchronousChannelProvider/openAsynchronousSocketChannel(Ljava/nio/channels/AsynchronousChannelGroup;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannelGroup], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
  }
  export class java_nio_channels_AsynchronousSocketChannel extends java_lang_Object implements java_nio_channels_AsynchronousByteChannel, java_nio_channels_NetworkChannel {
    public "java/nio/channels/AsynchronousSocketChannel/provider": java_nio_channels_spi_AsynchronousChannelProvider;
    public "<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "provider()Ljava/nio/channels/spi/AsynchronousChannelProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AsynchronousChannelProvider) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/provider()Ljava/nio/channels/spi/AsynchronousChannelProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AsynchronousChannelProvider) => void): void;
    public static "open(Ljava/nio/channels/AsynchronousChannelGroup;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannelGroup], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public static "java/nio/channels/AsynchronousSocketChannel/open(Ljava/nio/channels/AsynchronousChannelGroup;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannelGroup], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public static "open()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public static "java/nio/channels/AsynchronousSocketChannel/open()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "shutdownInput()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/shutdownInput()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "shutdownOutput()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/shutdownOutput()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "connect(Ljava/net/SocketAddress;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/connect(Ljava/net/SocketAddress;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/SocketAddress;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/connect(Ljava/net/SocketAddress;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "read(Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/read(Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/nio/ByteBuffer;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/read(Ljava/nio/ByteBuffer;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/read(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "read([Ljava/nio/ByteBuffer;IIJLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/read([Ljava/nio/ByteBuffer;IIJLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/write(Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/nio/ByteBuffer;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/write(Ljava/nio/ByteBuffer;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/write(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "write([Ljava/nio/ByteBuffer;IIJLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/write([Ljava/nio/ByteBuffer;IIJLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/AsynchronousSocketChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
  }
  export interface java_nio_channels_AsynchronousByteChannel extends java_lang_Object, java_nio_channels_AsynchronousChannel {
    "read(Ljava/nio/ByteBuffer;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    "read(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    "write(Ljava/nio/ByteBuffer;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    "write(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_channels_AsynchronousServerSocketChannel extends java_lang_Object implements java_nio_channels_AsynchronousChannel, java_nio_channels_NetworkChannel {
    public "java/nio/channels/AsynchronousServerSocketChannel/provider": java_nio_channels_spi_AsynchronousChannelProvider;
    public "<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "provider()Ljava/nio/channels/spi/AsynchronousChannelProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AsynchronousChannelProvider) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/provider()Ljava/nio/channels/spi/AsynchronousChannelProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AsynchronousChannelProvider) => void): void;
    public static "open(Ljava/nio/channels/AsynchronousChannelGroup;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannelGroup], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public static "java/nio/channels/AsynchronousServerSocketChannel/open(Ljava/nio/channels/AsynchronousChannelGroup;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannelGroup], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public static "open()Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public static "java/nio/channels/AsynchronousServerSocketChannel/open()Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;I)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/bind(Ljava/net/SocketAddress;I)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "accept(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/accept(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "accept()Ljava/util/concurrent/Future;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/accept()Ljava/util/concurrent/Future;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "java/nio/channels/AsynchronousServerSocketChannel/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
  }
  export interface java_util_concurrent_ThreadFactory extends java_lang_Object {
    "newThread(Ljava/lang/Runnable;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
  }
  export class java_lang_Void extends java_lang_Object {
    public unbox(): number;
    public static box(): java_lang_Void;
    public static "java/lang/Void/TYPE": java_lang_Class;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Void/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Void/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_Queue extends java_lang_Object, java_util_Collection {
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ScheduledThreadPoolExecutor extends java_util_concurrent_ThreadPoolExecutor implements java_util_concurrent_ScheduledExecutorService {
    public "java/util/concurrent/ScheduledThreadPoolExecutor/continueExistingPeriodicTasksAfterShutdown": number;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/executeExistingDelayedTasksAfterShutdown": number;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/removeOnCancel": number;
    public static "java/util/concurrent/ScheduledThreadPoolExecutor/sequencer": java_util_concurrent_atomic_AtomicLong;
    public "now()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/now()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "canRunInCurrentRunState(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/canRunInCurrentRunState(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "delayedExecute(Ljava/util/concurrent/RunnableScheduledFuture;)V"(thread: JVMThread, args: [java_util_concurrent_RunnableScheduledFuture], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/delayedExecute(Ljava/util/concurrent/RunnableScheduledFuture;)V"(thread: JVMThread, args: [java_util_concurrent_RunnableScheduledFuture], cb?: (e?: java_lang_Throwable) => void): void;
    public "reExecutePeriodic(Ljava/util/concurrent/RunnableScheduledFuture;)V"(thread: JVMThread, args: [java_util_concurrent_RunnableScheduledFuture], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/reExecutePeriodic(Ljava/util/concurrent/RunnableScheduledFuture;)V"(thread: JVMThread, args: [java_util_concurrent_RunnableScheduledFuture], cb?: (e?: java_lang_Throwable) => void): void;
    public "onShutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/onShutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "decorateTask(Ljava/lang/Runnable;Ljava/util/concurrent/RunnableScheduledFuture;)Ljava/util/concurrent/RunnableScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, java_util_concurrent_RunnableScheduledFuture], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RunnableScheduledFuture) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/decorateTask(Ljava/lang/Runnable;Ljava/util/concurrent/RunnableScheduledFuture;)Ljava/util/concurrent/RunnableScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, java_util_concurrent_RunnableScheduledFuture], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RunnableScheduledFuture) => void): void;
    public "decorateTask(Ljava/util/concurrent/Callable;Ljava/util/concurrent/RunnableScheduledFuture;)Ljava/util/concurrent/RunnableScheduledFuture;"(thread: JVMThread, args: [java_util_concurrent_Callable, java_util_concurrent_RunnableScheduledFuture], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RunnableScheduledFuture) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/decorateTask(Ljava/util/concurrent/Callable;Ljava/util/concurrent/RunnableScheduledFuture;)Ljava/util/concurrent/RunnableScheduledFuture;"(thread: JVMThread, args: [java_util_concurrent_Callable, java_util_concurrent_RunnableScheduledFuture], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RunnableScheduledFuture) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/util/concurrent/ThreadFactory;)V"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/<init>(ILjava/util/concurrent/ThreadFactory;)V"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [number, java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/<init>(ILjava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [number, java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/util/concurrent/ThreadFactory;Ljava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory, java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/<init>(ILjava/util/concurrent/ThreadFactory;Ljava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory, java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "triggerTime(JLjava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/triggerTime(JLjava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "triggerTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/triggerTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "overflowFree(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/overflowFree(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "schedule(Ljava/lang/Runnable;JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/schedule(Ljava/lang/Runnable;JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    public "schedule(Ljava/util/concurrent/Callable;JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_util_concurrent_Callable, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/schedule(Ljava/util/concurrent/Callable;JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_util_concurrent_Callable, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    public "scheduleAtFixedRate(Ljava/lang/Runnable;JJLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/scheduleAtFixedRate(Ljava/lang/Runnable;JJLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    public "scheduleWithFixedDelay(Ljava/lang/Runnable;JJLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/scheduleWithFixedDelay(Ljava/lang/Runnable;JJLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    public "execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "submit(Ljava/lang/Runnable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/submit(Ljava/lang/Runnable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "submit(Ljava/lang/Runnable;Ljava/lang/Object;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/submit(Ljava/lang/Runnable;Ljava/lang/Object;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "submit(Ljava/util/concurrent/Callable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_util_concurrent_Callable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/submit(Ljava/util/concurrent/Callable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_util_concurrent_Callable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "setContinueExistingPeriodicTasksAfterShutdownPolicy(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/setContinueExistingPeriodicTasksAfterShutdownPolicy(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContinueExistingPeriodicTasksAfterShutdownPolicy()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/getContinueExistingPeriodicTasksAfterShutdownPolicy()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setExecuteExistingDelayedTasksAfterShutdownPolicy(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/setExecuteExistingDelayedTasksAfterShutdownPolicy(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getExecuteExistingDelayedTasksAfterShutdownPolicy()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/getExecuteExistingDelayedTasksAfterShutdownPolicy()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setRemoveOnCancelPolicy(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/setRemoveOnCancelPolicy(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRemoveOnCancelPolicy()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/getRemoveOnCancelPolicy()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownNow()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/shutdownNow()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getQueue()Ljava/util/concurrent/BlockingQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_BlockingQueue) => void): void;
    public "java/util/concurrent/ScheduledThreadPoolExecutor/getQueue()Ljava/util/concurrent/BlockingQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_BlockingQueue) => void): void;
    public static "access$000()Ljava/util/concurrent/atomic/AtomicLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicLong) => void): void;
    public static "java/util/concurrent/ScheduledThreadPoolExecutor/access$000()Ljava/util/concurrent/atomic/AtomicLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicLong) => void): void;
    public static "access$100(Ljava/util/concurrent/ScheduledThreadPoolExecutor;)Z"(thread: JVMThread, args: [java_util_concurrent_ScheduledThreadPoolExecutor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ScheduledThreadPoolExecutor/access$100(Ljava/util/concurrent/ScheduledThreadPoolExecutor;)Z"(thread: JVMThread, args: [java_util_concurrent_ScheduledThreadPoolExecutor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/ScheduledThreadPoolExecutor/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_concurrent_BlockingQueue extends java_lang_Object, java_util_Queue {
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "put(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "offer(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "take()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "poll(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remainingCapacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "drainTo(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "drainTo(Ljava/util/Collection;I)I"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_concurrent_ScheduledFuture extends java_lang_Object, java_util_concurrent_Delayed, java_util_concurrent_Future {
    "getDelay(Ljava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "cancel(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isCancelled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isDone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "get(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_concurrent_Delayed extends java_lang_Object, java_lang_Comparable {
    "getDelay(Ljava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_concurrent_RejectedExecutionHandler extends java_lang_Object {
    "rejectedExecution(Ljava/lang/Runnable;Ljava/util/concurrent/ThreadPoolExecutor;)V"(thread: JVMThread, args: [java_lang_Runnable, java_util_concurrent_ThreadPoolExecutor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_concurrent_RunnableScheduledFuture extends java_lang_Object, java_util_concurrent_RunnableFuture, java_util_concurrent_ScheduledFuture {
    "isPeriodic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "cancel(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isCancelled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isDone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "get(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getDelay(Ljava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_concurrent_RunnableFuture extends java_lang_Object, java_lang_Runnable, java_util_concurrent_Future {
    "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "cancel(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isCancelled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isDone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "get(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_concurrent_ScheduledExecutorService extends java_lang_Object, java_util_concurrent_ExecutorService {
    "schedule(Ljava/lang/Runnable;JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    "schedule(Ljava/util/concurrent/Callable;JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_util_concurrent_Callable, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    "scheduleAtFixedRate(Ljava/lang/Runnable;JJLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    "scheduleWithFixedDelay(Ljava/lang/Runnable;JJLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/ScheduledFuture;"(thread: JVMThread, args: [java_lang_Runnable, Long, any, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ScheduledFuture) => void): void;
    "shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "shutdownNow()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "isShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isTerminated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "awaitTermination(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "submit(Ljava/util/concurrent/Callable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_util_concurrent_Callable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    "submit(Ljava/lang/Runnable;Ljava/lang/Object;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    "submit(Ljava/lang/Runnable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    "invokeAll(Ljava/util/Collection;)Ljava/util/List;"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "invokeAll(Ljava/util/Collection;JLjava/util/concurrent/TimeUnit;)Ljava/util/List;"(thread: JVMThread, args: [java_util_Collection, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "invokeAny(Ljava/util/Collection;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "invokeAny(Ljava/util/Collection;JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ThreadPoolExecutor extends java_util_concurrent_AbstractExecutorService {
    public "java/util/concurrent/ThreadPoolExecutor/ctl": java_util_concurrent_atomic_AtomicInteger;
    public static "java/util/concurrent/ThreadPoolExecutor/COUNT_BITS": number;
    public static "java/util/concurrent/ThreadPoolExecutor/CAPACITY": number;
    public static "java/util/concurrent/ThreadPoolExecutor/RUNNING": number;
    public static "java/util/concurrent/ThreadPoolExecutor/SHUTDOWN": number;
    public static "java/util/concurrent/ThreadPoolExecutor/STOP": number;
    public static "java/util/concurrent/ThreadPoolExecutor/TIDYING": number;
    public static "java/util/concurrent/ThreadPoolExecutor/TERMINATED": number;
    public "java/util/concurrent/ThreadPoolExecutor/workQueue": java_util_concurrent_BlockingQueue;
    public "java/util/concurrent/ThreadPoolExecutor/mainLock": java_util_concurrent_locks_ReentrantLock;
    public "java/util/concurrent/ThreadPoolExecutor/workers": java_util_HashSet;
    public "java/util/concurrent/ThreadPoolExecutor/termination": java_util_concurrent_locks_Condition;
    public "java/util/concurrent/ThreadPoolExecutor/largestPoolSize": number;
    public "java/util/concurrent/ThreadPoolExecutor/completedTaskCount": Long;
    public "java/util/concurrent/ThreadPoolExecutor/threadFactory": java_util_concurrent_ThreadFactory;
    public "java/util/concurrent/ThreadPoolExecutor/handler": java_util_concurrent_RejectedExecutionHandler;
    public "java/util/concurrent/ThreadPoolExecutor/keepAliveTime": Long;
    public "java/util/concurrent/ThreadPoolExecutor/allowCoreThreadTimeOut": number;
    public "java/util/concurrent/ThreadPoolExecutor/corePoolSize": number;
    public "java/util/concurrent/ThreadPoolExecutor/maximumPoolSize": number;
    public static "java/util/concurrent/ThreadPoolExecutor/defaultHandler": java_util_concurrent_RejectedExecutionHandler;
    public static "java/util/concurrent/ThreadPoolExecutor/shutdownPerm": java_lang_RuntimePermission;
    public static "java/util/concurrent/ThreadPoolExecutor/ONLY_ONE": number;
    public static "runStateOf(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ThreadPoolExecutor/runStateOf(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "workerCountOf(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ThreadPoolExecutor/workerCountOf(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ctlOf(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ThreadPoolExecutor/ctlOf(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "runStateLessThan(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ThreadPoolExecutor/runStateLessThan(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "runStateAtLeast(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ThreadPoolExecutor/runStateAtLeast(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isRunning(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ThreadPoolExecutor/isRunning(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareAndIncrementWorkerCount(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/compareAndIncrementWorkerCount(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareAndDecrementWorkerCount(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/compareAndDecrementWorkerCount(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "decrementWorkerCount()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/decrementWorkerCount()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "advanceRunState(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/advanceRunState(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "tryTerminate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/tryTerminate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkShutdownAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/checkShutdownAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interruptWorkers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/interruptWorkers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interruptIdleWorkers(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/interruptIdleWorkers(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "interruptIdleWorkers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/interruptIdleWorkers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "reject(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/reject(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "onShutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/onShutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isRunningOrShutdown(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/isRunningOrShutdown(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drainQueue()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/drainQueue()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "addWorker(Ljava/lang/Runnable;Z)Z"(thread: JVMThread, args: [java_lang_Runnable, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/addWorker(Ljava/lang/Runnable;Z)Z"(thread: JVMThread, args: [java_lang_Runnable, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addWorkerFailed(Ljava/util/concurrent/ThreadPoolExecutor$Worker;)V"(thread: JVMThread, args: [java_util_concurrent_ThreadPoolExecutor$Worker], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/addWorkerFailed(Ljava/util/concurrent/ThreadPoolExecutor$Worker;)V"(thread: JVMThread, args: [java_util_concurrent_ThreadPoolExecutor$Worker], cb?: (e?: java_lang_Throwable) => void): void;
    public "processWorkerExit(Ljava/util/concurrent/ThreadPoolExecutor$Worker;Z)V"(thread: JVMThread, args: [java_util_concurrent_ThreadPoolExecutor$Worker, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/processWorkerExit(Ljava/util/concurrent/ThreadPoolExecutor$Worker;Z)V"(thread: JVMThread, args: [java_util_concurrent_ThreadPoolExecutor$Worker, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTask()Ljava/lang/Runnable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runnable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getTask()Ljava/lang/Runnable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runnable) => void): void;
    public "runWorker(Ljava/util/concurrent/ThreadPoolExecutor$Worker;)V"(thread: JVMThread, args: [java_util_concurrent_ThreadPoolExecutor$Worker], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/runWorker(Ljava/util/concurrent/ThreadPoolExecutor$Worker;)V"(thread: JVMThread, args: [java_util_concurrent_ThreadPoolExecutor$Worker], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIJLjava/util/concurrent/TimeUnit;Ljava/util/concurrent/BlockingQueue;)V"(thread: JVMThread, args: [number, number, Long, any, java_util_concurrent_TimeUnit, java_util_concurrent_BlockingQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/<init>(IIJLjava/util/concurrent/TimeUnit;Ljava/util/concurrent/BlockingQueue;)V"(thread: JVMThread, args: [number, number, Long, any, java_util_concurrent_TimeUnit, java_util_concurrent_BlockingQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIJLjava/util/concurrent/TimeUnit;Ljava/util/concurrent/BlockingQueue;Ljava/util/concurrent/ThreadFactory;)V"(thread: JVMThread, args: [number, number, Long, any, java_util_concurrent_TimeUnit, java_util_concurrent_BlockingQueue, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/<init>(IIJLjava/util/concurrent/TimeUnit;Ljava/util/concurrent/BlockingQueue;Ljava/util/concurrent/ThreadFactory;)V"(thread: JVMThread, args: [number, number, Long, any, java_util_concurrent_TimeUnit, java_util_concurrent_BlockingQueue, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIJLjava/util/concurrent/TimeUnit;Ljava/util/concurrent/BlockingQueue;Ljava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [number, number, Long, any, java_util_concurrent_TimeUnit, java_util_concurrent_BlockingQueue, java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/<init>(IIJLjava/util/concurrent/TimeUnit;Ljava/util/concurrent/BlockingQueue;Ljava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [number, number, Long, any, java_util_concurrent_TimeUnit, java_util_concurrent_BlockingQueue, java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIJLjava/util/concurrent/TimeUnit;Ljava/util/concurrent/BlockingQueue;Ljava/util/concurrent/ThreadFactory;Ljava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [number, number, Long, any, java_util_concurrent_TimeUnit, java_util_concurrent_BlockingQueue, java_util_concurrent_ThreadFactory, java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/<init>(IIJLjava/util/concurrent/TimeUnit;Ljava/util/concurrent/BlockingQueue;Ljava/util/concurrent/ThreadFactory;Ljava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [number, number, Long, any, java_util_concurrent_TimeUnit, java_util_concurrent_BlockingQueue, java_util_concurrent_ThreadFactory, java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownNow()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/shutdownNow()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "isShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/isShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isTerminating()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/isTerminating()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isTerminated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/isTerminated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "awaitTermination(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/awaitTermination(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setThreadFactory(Ljava/util/concurrent/ThreadFactory;)V"(thread: JVMThread, args: [java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/setThreadFactory(Ljava/util/concurrent/ThreadFactory;)V"(thread: JVMThread, args: [java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "getThreadFactory()Ljava/util/concurrent/ThreadFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ThreadFactory) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getThreadFactory()Ljava/util/concurrent/ThreadFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ThreadFactory) => void): void;
    public "setRejectedExecutionHandler(Ljava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/setRejectedExecutionHandler(Ljava/util/concurrent/RejectedExecutionHandler;)V"(thread: JVMThread, args: [java_util_concurrent_RejectedExecutionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRejectedExecutionHandler()Ljava/util/concurrent/RejectedExecutionHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RejectedExecutionHandler) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getRejectedExecutionHandler()Ljava/util/concurrent/RejectedExecutionHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RejectedExecutionHandler) => void): void;
    public "setCorePoolSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/setCorePoolSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCorePoolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getCorePoolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "prestartCoreThread()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/prestartCoreThread()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensurePrestart()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/ensurePrestart()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "prestartAllCoreThreads()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/prestartAllCoreThreads()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "allowsCoreThreadTimeOut()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/allowsCoreThreadTimeOut()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "allowCoreThreadTimeOut(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/allowCoreThreadTimeOut(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setMaximumPoolSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/setMaximumPoolSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMaximumPoolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getMaximumPoolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setKeepAliveTime(JLjava/util/concurrent/TimeUnit;)V"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/setKeepAliveTime(JLjava/util/concurrent/TimeUnit;)V"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKeepAliveTime(Ljava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getKeepAliveTime(Ljava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getQueue()Ljava/util/concurrent/BlockingQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_BlockingQueue) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getQueue()Ljava/util/concurrent/BlockingQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_BlockingQueue) => void): void;
    public "remove(Ljava/lang/Runnable;)Z"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/remove(Ljava/lang/Runnable;)Z"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "purge()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/purge()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPoolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getPoolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getActiveCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getActiveCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLargestPoolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getLargestPoolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTaskCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getTaskCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getCompletedTaskCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/getCompletedTaskCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "beforeExecute(Ljava/lang/Thread;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Thread, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/beforeExecute(Ljava/lang/Thread;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Thread, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "afterExecute(Ljava/lang/Runnable;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/afterExecute(Ljava/lang/Runnable;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "terminated()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor/terminated()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/ThreadPoolExecutor/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ThreadPoolExecutor$Worker extends java_util_concurrent_locks_AbstractQueuedSynchronizer implements java_lang_Runnable {
    public static "java/util/concurrent/ThreadPoolExecutor$Worker/serialVersionUID": Long;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/thread": java_lang_Thread;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/firstTask": java_lang_Runnable;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/completedTasks": Long;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/this$0": java_util_concurrent_ThreadPoolExecutor;
    public "<init>(Ljava/util/concurrent/ThreadPoolExecutor;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_util_concurrent_ThreadPoolExecutor, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/<init>(Ljava/util/concurrent/ThreadPoolExecutor;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_util_concurrent_ThreadPoolExecutor, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isHeldExclusively()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/isHeldExclusively()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryAcquire(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/tryAcquire(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryRelease(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/tryRelease(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/lock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "tryLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/tryLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unlock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/unlock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isLocked()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/isLocked()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "interruptIfStarted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ThreadPoolExecutor$Worker/interruptIfStarted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_locks_AbstractQueuedSynchronizer extends java_util_concurrent_locks_AbstractOwnableSynchronizer implements java_io_Serializable {
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/serialVersionUID": Long;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/head": java_util_concurrent_locks_AbstractQueuedSynchronizer$Node;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/tail": java_util_concurrent_locks_AbstractQueuedSynchronizer$Node;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/state": number;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/spinForTimeoutThreshold": Long;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/unsafe": sun_misc_Unsafe;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/stateOffset": Long;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/headOffset": Long;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/tailOffset": Long;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/waitStatusOffset": Long;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/nextOffset": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/getState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setState(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/setState(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "compareAndSetState(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/compareAndSetState(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "enq(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$Node) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/enq(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$Node) => void): void;
    public "addWaiter(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$Node) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/addWaiter(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$Node) => void): void;
    public "setHead(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/setHead(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "unparkSuccessor(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/unparkSuccessor(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "doReleaseShared()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/doReleaseShared()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setHeadAndPropagate(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;I)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/setHeadAndPropagate(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;I)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "cancelAcquire(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/cancelAcquire(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public static "shouldParkAfterFailedAcquire(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/shouldParkAfterFailedAcquire(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "selfInterrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/selfInterrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "parkAndCheckInterrupt()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/parkAndCheckInterrupt()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "acquireQueued(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;I)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/acquireQueued(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;I)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doAcquireInterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/doAcquireInterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "doAcquireNanos(IJ)Z"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/doAcquireNanos(IJ)Z"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doAcquireShared(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/doAcquireShared(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "doAcquireSharedInterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/doAcquireSharedInterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "doAcquireSharedNanos(IJ)Z"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/doAcquireSharedNanos(IJ)Z"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryAcquire(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/tryAcquire(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryRelease(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/tryRelease(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryAcquireShared(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/tryAcquireShared(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryReleaseShared(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/tryReleaseShared(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isHeldExclusively()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/isHeldExclusively()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "acquire(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/acquire(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "acquireInterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/acquireInterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "tryAcquireNanos(IJ)Z"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/tryAcquireNanos(IJ)Z"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "release(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/release(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "acquireShared(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/acquireShared(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "acquireSharedInterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/acquireSharedInterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "tryAcquireSharedNanos(IJ)Z"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/tryAcquireSharedNanos(IJ)Z"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "releaseShared(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/releaseShared(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasQueuedThreads()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/hasQueuedThreads()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasContended()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/hasContended()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFirstQueuedThread()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/getFirstQueuedThread()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public "fullGetFirstQueuedThread()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/fullGetFirstQueuedThread()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public "isQueued(Ljava/lang/Thread;)Z"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/isQueued(Ljava/lang/Thread;)Z"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "apparentlyFirstQueuedIsExclusive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/apparentlyFirstQueuedIsExclusive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasQueuedPredecessors()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/hasQueuedPredecessors()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getQueueLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/getQueueLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/getQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "getExclusiveQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/getExclusiveQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "getSharedQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/getSharedQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isOnSyncQueue(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/isOnSyncQueue(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "findNodeFromTail(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/findNodeFromTail(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "transferForSignal(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/transferForSignal(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "transferAfterCancelledWait(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/transferAfterCancelledWait(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fullyRelease(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)I"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/fullyRelease(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)I"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "owns(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/owns(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasWaiters(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/hasWaiters(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWaitQueueLength(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;)I"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/getWaitQueueLength(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;)I"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWaitingThreads(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;)Ljava/util/Collection;"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/getWaitingThreads(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;)Ljava/util/Collection;"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "compareAndSetHead(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/compareAndSetHead(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareAndSetTail(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer/compareAndSetTail(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compareAndSetWaitStatus(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;II)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/compareAndSetWaitStatus(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;II)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compareAndSetNext(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/compareAndSetNext(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject extends java_lang_Object implements java_util_concurrent_locks_Condition, java_io_Serializable {
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/serialVersionUID": Long;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/firstWaiter": java_util_concurrent_locks_AbstractQueuedSynchronizer$Node;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/lastWaiter": java_util_concurrent_locks_AbstractQueuedSynchronizer$Node;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/REINTERRUPT": number;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/THROW_IE": number;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/this$0": java_util_concurrent_locks_AbstractQueuedSynchronizer;
    public "<init>(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/<init>(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer], cb?: (e?: java_lang_Throwable) => void): void;
    public "addConditionWaiter()Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$Node) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/addConditionWaiter()Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$Node) => void): void;
    public "doSignal(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/doSignal(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "doSignalAll(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/doSignalAll(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "unlinkCancelledWaiters()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/unlinkCancelledWaiters()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "signal()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/signal()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "signalAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/signalAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "awaitUninterruptibly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/awaitUninterruptibly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkInterruptWhileWaiting(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)I"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/checkInterruptWhileWaiting(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)I"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reportInterruptAfterWait(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/reportInterruptAfterWait(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "await()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/await()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "awaitNanos(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/awaitNanos(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "awaitUntil(Ljava/util/Date;)Z"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/awaitUntil(Ljava/util/Date;)Z"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "await(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/await(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOwnedBy(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/isOwnedBy(Ljava/util/concurrent/locks/AbstractQueuedSynchronizer;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_AbstractQueuedSynchronizer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasWaiters()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/hasWaiters()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWaitQueueLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/getWaitQueueLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWaitingThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject/getWaitingThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
  }
  export class java_util_Date extends java_lang_Object implements java_io_Serializable, java_lang_Cloneable, java_lang_Comparable {
    public static "java/util/Date/gcal": sun_util_calendar_BaseCalendar;
    public static "java/util/Date/jcal": sun_util_calendar_BaseCalendar;
    public "java/util/Date/fastTime": Long;
    public "java/util/Date/cdate": sun_util_calendar_BaseCalendar$Date;
    public static "java/util/Date/defaultCenturyStart": number;
    public static "java/util/Date/serialVersionUID": Long;
    public static "java/util/Date/wtb": JVMArray<java_lang_String>;
    public static "java/util/Date/ttb": JVMArray<number>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/<init>(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/<init>(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Date/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "UTC(IIIIII)J"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/Date/UTC(IIIIII)J"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "parse(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/Date/parse(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setYear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/setYear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/getMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMonth(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/setMonth(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDate()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/getDate()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDate(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/setDate(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDay()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/getDay()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHours()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/getHours()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setHours(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/setHours(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMinutes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/getMinutes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMinutes(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/setMinutes(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSeconds()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/getSeconds()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSeconds(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/setSeconds(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/Date/getTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTimeImpl()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/Date/getTimeImpl()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setTime(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/setTime(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "before(Ljava/util/Date;)Z"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/before(Ljava/util/Date;)Z"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "after(Ljava/util/Date;)Z"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/after(Ljava/util/Date;)Z"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getMillisOf(Ljava/util/Date;)J"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/Date/getMillisOf(Ljava/util/Date;)J"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareTo(Ljava/util/Date;)I"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/compareTo(Ljava/util/Date;)I"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Date/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "convertToAbbr(Ljava/lang/StringBuilder;Ljava/lang/String;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_StringBuilder, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public static "java/util/Date/convertToAbbr(Ljava/lang/StringBuilder;Ljava/lang/String;)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_StringBuilder, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "toLocaleString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Date/toLocaleString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toGMTString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Date/toGMTString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTimezoneOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/getTimezoneOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCalendarDate()Lsun/util/calendar/BaseCalendar$Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar$Date) => void): void;
    public "java/util/Date/getCalendarDate()Lsun/util/calendar/BaseCalendar$Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar$Date) => void): void;
    public "normalize()Lsun/util/calendar/BaseCalendar$Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar$Date) => void): void;
    public "java/util/Date/normalize()Lsun/util/calendar/BaseCalendar$Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar$Date) => void): void;
    public "normalize(Lsun/util/calendar/BaseCalendar$Date;)Lsun/util/calendar/BaseCalendar$Date;"(thread: JVMThread, args: [sun_util_calendar_BaseCalendar$Date], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar$Date) => void): void;
    public "java/util/Date/normalize(Lsun/util/calendar/BaseCalendar$Date;)Lsun/util/calendar/BaseCalendar$Date;"(thread: JVMThread, args: [sun_util_calendar_BaseCalendar$Date], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar$Date) => void): void;
    public static "getCalendarSystem(I)Lsun/util/calendar/BaseCalendar;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar) => void): void;
    public static "java/util/Date/getCalendarSystem(I)Lsun/util/calendar/BaseCalendar;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar) => void): void;
    public static "getCalendarSystem(J)Lsun/util/calendar/BaseCalendar;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar) => void): void;
    public static "java/util/Date/getCalendarSystem(J)Lsun/util/calendar/BaseCalendar;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar) => void): void;
    public static "getCalendarSystem(Lsun/util/calendar/BaseCalendar$Date;)Lsun/util/calendar/BaseCalendar;"(thread: JVMThread, args: [sun_util_calendar_BaseCalendar$Date], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar) => void): void;
    public static "java/util/Date/getCalendarSystem(Lsun/util/calendar/BaseCalendar$Date;)Lsun/util/calendar/BaseCalendar;"(thread: JVMThread, args: [sun_util_calendar_BaseCalendar$Date], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar) => void): void;
    public static "getJulianCalendar()Lsun/util/calendar/BaseCalendar;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar) => void): void;
    public static "java/util/Date/getJulianCalendar()Lsun/util/calendar/BaseCalendar;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Date/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "from(Ljava/time/Instant;)Ljava/util/Date;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_util_Date) => void): void;
    public static "java/util/Date/from(Ljava/time/Instant;)Ljava/util/Date;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_util_Date) => void): void;
    public "toInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/util/Date/toInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Date/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Date/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_Instant extends java_lang_Object implements java_time_temporal_Temporal, java_time_temporal_TemporalAdjuster, java_lang_Comparable, java_io_Serializable {
    public static "java/time/Instant/EPOCH": java_time_Instant;
    public static "java/time/Instant/MIN_SECOND": Long;
    public static "java/time/Instant/MAX_SECOND": Long;
    public static "java/time/Instant/MIN": java_time_Instant;
    public static "java/time/Instant/MAX": java_time_Instant;
    public static "java/time/Instant/serialVersionUID": Long;
    public "java/time/Instant/seconds": Long;
    public "java/time/Instant/nanos": number;
    public static "now()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "java/time/Instant/now()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "now(Ljava/time/Clock;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "java/time/Instant/now(Ljava/time/Clock;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "ofEpochSecond(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "java/time/Instant/ofEpochSecond(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "ofEpochSecond(JJ)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "java/time/Instant/ofEpochSecond(JJ)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "ofEpochMilli(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "java/time/Instant/ofEpochMilli(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "java/time/Instant/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "parse(Ljava/lang/CharSequence;)Ljava/time/Instant;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "java/time/Instant/parse(Ljava/lang/CharSequence;)Ljava/time/Instant;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "create(JI)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "java/time/Instant/create(JI)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "<init>(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Instant/<init>(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/Instant/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Instant/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getEpochSecond()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Instant/getEpochSecond()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/with(Ljava/time/temporal/TemporalField;J)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "plusSeconds(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/plusSeconds(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "plusMillis(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/plusMillis(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "plusNanos(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/plusNanos(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "plus(JJ)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/plus(JJ)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "minusSeconds(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/minusSeconds(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "minusMillis(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/minusMillis(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "minusNanos(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Instant/minusNanos(J)Ljava/time/Instant;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/Instant/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Instant/adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Instant/until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "nanosUntil(Ljava/time/Instant;)J"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Instant/nanosUntil(Ljava/time/Instant;)J"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "secondsUntil(Ljava/time/Instant;)J"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Instant/secondsUntil(Ljava/time/Instant;)J"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "atOffset(Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/Instant/atOffset(Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "atZone(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/Instant/atZone(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "toEpochMilli()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Instant/toEpochMilli()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareTo(Ljava/time/Instant;)I"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/compareTo(Ljava/time/Instant;)I"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAfter(Ljava/time/Instant;)Z"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/isAfter(Ljava/time/Instant;)Z"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBefore(Ljava/time/Instant;)Z"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/isBefore(Ljava/time/Instant;)Z"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/Instant/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/Instant/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Instant/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Instant/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/Instant;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public static "java/time/Instant/readExternal(Ljava/io/DataInput;)Ljava/time/Instant;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Instant/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Instant/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Instant/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Instant/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Instant/with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Instant/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Instant/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/Instant/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_io_DataInput extends java_lang_Object {
    "readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    "readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_io_DataOutput extends java_lang_Object {
    "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    "writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    "writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_ZoneId extends java_lang_Object implements java_io_Serializable {
    public static "java/time/ZoneId/SHORT_IDS": java_util_Map;
    public static "java/time/ZoneId/serialVersionUID": Long;
    public static "systemDefault()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "java/time/ZoneId/systemDefault()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "getAvailableZoneIds()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "java/time/ZoneId/getAvailableZoneIds()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "of(Ljava/lang/String;Ljava/util/Map;)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "java/time/ZoneId/of(Ljava/lang/String;Ljava/util/Map;)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "of(Ljava/lang/String;)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "java/time/ZoneId/of(Ljava/lang/String;)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "ofOffset(Ljava/lang/String;Ljava/time/ZoneOffset;)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "java/time/ZoneId/ofOffset(Ljava/lang/String;Ljava/time/ZoneOffset;)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "of(Ljava/lang/String;Z)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "java/time/ZoneId/of(Ljava/lang/String;Z)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "ofWithPrefix(Ljava/lang/String;IZ)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "java/time/ZoneId/ofWithPrefix(Ljava/lang/String;IZ)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "java/time/ZoneId/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/ZoneId;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZoneId/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/ZoneId/getId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/ZoneId/getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toTemporal()Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "java/time/ZoneId/toTemporal()Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "getRules()Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public "java/time/ZoneId/getRules()Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public "normalized()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "java/time/ZoneId/normalized()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZoneId/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZoneId/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZoneId/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/ZoneId/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/ZoneId/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "write(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZoneId/write(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/ZoneId/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_zone_ZoneRules extends java_lang_Object implements java_io_Serializable {
    public static "java/time/zone/ZoneRules/serialVersionUID": Long;
    public static "java/time/zone/ZoneRules/LAST_CACHED_YEAR": number;
    public "java/time/zone/ZoneRules/standardTransitions": JVMArray<Long>;
    public "java/time/zone/ZoneRules/standardOffsets": JVMArray<java_time_ZoneOffset>;
    public "java/time/zone/ZoneRules/savingsInstantTransitions": JVMArray<Long>;
    public "java/time/zone/ZoneRules/savingsLocalTransitions": JVMArray<java_time_LocalDateTime>;
    public "java/time/zone/ZoneRules/wallOffsets": JVMArray<java_time_ZoneOffset>;
    public "java/time/zone/ZoneRules/lastRules": JVMArray<java_time_zone_ZoneOffsetTransitionRule>;
    public "java/time/zone/ZoneRules/lastRulesCache": java_util_concurrent_ConcurrentMap;
    public static "java/time/zone/ZoneRules/EMPTY_LONG_ARRAY": JVMArray<Long>;
    public static "java/time/zone/ZoneRules/EMPTY_LASTRULES": JVMArray<java_time_zone_ZoneOffsetTransitionRule>;
    public static "java/time/zone/ZoneRules/EMPTY_LDT_ARRAY": JVMArray<java_time_LocalDateTime>;
    public static "of(Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;Ljava/util/List;Ljava/util/List;Ljava/util/List;)Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: [java_time_ZoneOffset, java_time_ZoneOffset, java_util_List, java_util_List, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public static "java/time/zone/ZoneRules/of(Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;Ljava/util/List;Ljava/util/List;Ljava/util/List;)Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: [java_time_ZoneOffset, java_time_ZoneOffset, java_util_List, java_util_List, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public static "of(Ljava/time/ZoneOffset;)Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public static "java/time/zone/ZoneRules/of(Ljava/time/ZoneOffset;)Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public "<init>(Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;Ljava/util/List;Ljava/util/List;Ljava/util/List;)V"(thread: JVMThread, args: [java_time_ZoneOffset, java_time_ZoneOffset, java_util_List, java_util_List, java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneRules/<init>(Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;Ljava/util/List;Ljava/util/List;Ljava/util/List;)V"(thread: JVMThread, args: [java_time_ZoneOffset, java_time_ZoneOffset, java_util_List, java_util_List, java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([J[Ljava/time/ZoneOffset;[J[Ljava/time/ZoneOffset;[Ljava/time/zone/ZoneOffsetTransitionRule;)V"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<java_time_ZoneOffset>, JVMArray<Long>, JVMArray<java_time_ZoneOffset>, JVMArray<java_time_zone_ZoneOffsetTransitionRule>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneRules/<init>([J[Ljava/time/ZoneOffset;[J[Ljava/time/ZoneOffset;[Ljava/time/zone/ZoneOffsetTransitionRule;)V"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<java_time_ZoneOffset>, JVMArray<Long>, JVMArray<java_time_ZoneOffset>, JVMArray<java_time_zone_ZoneOffsetTransitionRule>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneRules/<init>(Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneRules/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/zone/ZoneRules/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneRules/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public static "java/time/zone/ZoneRules/readExternal(Ljava/io/DataInput;)Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public "isFixedOffset()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneRules/isFixedOffset()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOffset(Ljava/time/Instant;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/zone/ZoneRules/getOffset(Ljava/time/Instant;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "getOffset(Ljava/time/LocalDateTime;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/zone/ZoneRules/getOffset(Ljava/time/LocalDateTime;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "getValidOffsets(Ljava/time/LocalDateTime;)Ljava/util/List;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/time/zone/ZoneRules/getValidOffsets(Ljava/time/LocalDateTime;)Ljava/util/List;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getTransition(Ljava/time/LocalDateTime;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "java/time/zone/ZoneRules/getTransition(Ljava/time/LocalDateTime;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "getOffsetInfo(Ljava/time/LocalDateTime;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/zone/ZoneRules/getOffsetInfo(Ljava/time/LocalDateTime;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "findOffsetInfo(Ljava/time/LocalDateTime;Ljava/time/zone/ZoneOffsetTransition;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_zone_ZoneOffsetTransition], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/zone/ZoneRules/findOffsetInfo(Ljava/time/LocalDateTime;Ljava/time/zone/ZoneOffsetTransition;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_zone_ZoneOffsetTransition], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "findTransitionArray(I)[Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_zone_ZoneOffsetTransition>) => void): void;
    public "java/time/zone/ZoneRules/findTransitionArray(I)[Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_zone_ZoneOffsetTransition>) => void): void;
    public "getStandardOffset(Ljava/time/Instant;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/zone/ZoneRules/getStandardOffset(Ljava/time/Instant;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "getDaylightSavings(Ljava/time/Instant;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/zone/ZoneRules/getDaylightSavings(Ljava/time/Instant;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "isDaylightSavings(Ljava/time/Instant;)Z"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneRules/isDaylightSavings(Ljava/time/Instant;)Z"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isValidOffset(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;)Z"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneRules/isValidOffset(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;)Z"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextTransition(Ljava/time/Instant;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "java/time/zone/ZoneRules/nextTransition(Ljava/time/Instant;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "previousTransition(Ljava/time/Instant;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "java/time/zone/ZoneRules/previousTransition(Ljava/time/Instant;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "findYear(JLjava/time/ZoneOffset;)I"(thread: JVMThread, args: [Long, any, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneRules/findYear(JLjava/time/ZoneOffset;)I"(thread: JVMThread, args: [Long, any, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransitions()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/time/zone/ZoneRules/getTransitions()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getTransitionRules()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/time/zone/ZoneRules/getTransitionRules()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneRules/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneRules/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/zone/ZoneRules/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/zone/ZoneRules/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_Duration extends java_lang_Object implements java_time_temporal_TemporalAmount, java_lang_Comparable, java_io_Serializable {
    public static "java/time/Duration/ZERO": java_time_Duration;
    public static "java/time/Duration/serialVersionUID": Long;
    public static "java/time/Duration/BI_NANOS_PER_SECOND": java_math_BigInteger;
    public static "java/time/Duration/PATTERN": java_util_regex_Pattern;
    public "java/time/Duration/seconds": Long;
    public "java/time/Duration/nanos": number;
    public static "ofDays(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/ofDays(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "ofHours(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/ofHours(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "ofMinutes(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/ofMinutes(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "ofSeconds(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/ofSeconds(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "ofSeconds(JJ)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/ofSeconds(JJ)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "ofMillis(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/ofMillis(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "ofNanos(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/ofNanos(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "of(JLjava/time/temporal/TemporalUnit;)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/of(JLjava/time/temporal/TemporalUnit;)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "from(Ljava/time/temporal/TemporalAmount;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/from(Ljava/time/temporal/TemporalAmount;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "parse(Ljava/lang/CharSequence;)Ljava/time/Duration;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/parse(Ljava/lang/CharSequence;)Ljava/time/Duration;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "parseNumber(Ljava/lang/CharSequence;Ljava/lang/String;ILjava/lang/String;)J"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/time/Duration/parseNumber(Ljava/lang/CharSequence;Ljava/lang/String;ILjava/lang/String;)J"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "parseFraction(Ljava/lang/CharSequence;Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/Duration/parseFraction(Ljava/lang/CharSequence;Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "create(ZJJJJI)Ljava/time/Duration;"(thread: JVMThread, args: [number, Long, any, Long, any, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/create(ZJJJJI)Ljava/time/Duration;"(thread: JVMThread, args: [number, Long, any, Long, any, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "between(Ljava/time/temporal/Temporal;Ljava/time/temporal/Temporal;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/between(Ljava/time/temporal/Temporal;Ljava/time/temporal/Temporal;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "create(JI)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/create(JI)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "<init>(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Duration/<init>(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Duration/get(Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getUnits()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/time/Duration/getUnits()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "isZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Duration/isZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isNegative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Duration/isNegative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSeconds()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Duration/getSeconds()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Duration/getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "withSeconds(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/withSeconds(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "withNanos(I)Ljava/time/Duration;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/withNanos(I)Ljava/time/Duration;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "plus(Ljava/time/Duration;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/plus(Ljava/time/Duration;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "plusDays(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/plusDays(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "plusHours(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/plusHours(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "plusMinutes(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/plusMinutes(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "plusSeconds(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/plusSeconds(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "plusMillis(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/plusMillis(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "plusNanos(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/plusNanos(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "plus(JJ)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/plus(JJ)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "minus(Ljava/time/Duration;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/minus(Ljava/time/Duration;)Ljava/time/Duration;"(thread: JVMThread, args: [java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "minusDays(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/minusDays(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "minusHours(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/minusHours(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "minusMinutes(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/minusMinutes(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "minusSeconds(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/minusSeconds(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "minusMillis(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/minusMillis(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "minusNanos(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/minusNanos(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "multipliedBy(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/multipliedBy(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "dividedBy(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/dividedBy(J)Ljava/time/Duration;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "toSeconds()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/time/Duration/toSeconds()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "create(Ljava/math/BigDecimal;)Ljava/time/Duration;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/create(Ljava/math/BigDecimal;)Ljava/time/Duration;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "negated()Ljava/time/Duration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/negated()Ljava/time/Duration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "abs()Ljava/time/Duration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/Duration/abs()Ljava/time/Duration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "addTo(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Duration/addTo(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "subtractFrom(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Duration/subtractFrom(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "toDays()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Duration/toDays()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toHours()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Duration/toHours()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toMinutes()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Duration/toMinutes()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toMillis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Duration/toMillis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toNanos()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Duration/toNanos()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareTo(Ljava/time/Duration;)I"(thread: JVMThread, args: [java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Duration/compareTo(Ljava/time/Duration;)I"(thread: JVMThread, args: [java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Duration/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Duration/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/Duration/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/Duration/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Duration/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Duration/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/Duration;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public static "java/time/Duration/readExternal(Ljava/io/DataInput;)Ljava/time/Duration;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Duration/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/Duration/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_math_BigDecimal extends java_lang_Number implements java_lang_Comparable {
    public "java/math/BigDecimal/intVal": java_math_BigInteger;
    public "java/math/BigDecimal/scale": number;
    public "java/math/BigDecimal/precision": number;
    public "java/math/BigDecimal/stringCache": java_lang_String;
    public static "java/math/BigDecimal/INFLATED": Long;
    public static "java/math/BigDecimal/INFLATED_BIGINT": java_math_BigInteger;
    public "java/math/BigDecimal/intCompact": Long;
    public static "java/math/BigDecimal/MAX_COMPACT_DIGITS": number;
    public static "java/math/BigDecimal/serialVersionUID": Long;
    public static "java/math/BigDecimal/threadLocalStringBuilderHelper": java_lang_ThreadLocal;
    public static "java/math/BigDecimal/zeroThroughTen": JVMArray<java_math_BigDecimal>;
    public static "java/math/BigDecimal/ZERO_SCALED_BY": JVMArray<java_math_BigDecimal>;
    public static "java/math/BigDecimal/HALF_LONG_MAX_VALUE": Long;
    public static "java/math/BigDecimal/HALF_LONG_MIN_VALUE": Long;
    public static "java/math/BigDecimal/ZERO": java_math_BigDecimal;
    public static "java/math/BigDecimal/ONE": java_math_BigDecimal;
    public static "java/math/BigDecimal/TEN": java_math_BigDecimal;
    public static "java/math/BigDecimal/ROUND_UP": number;
    public static "java/math/BigDecimal/ROUND_DOWN": number;
    public static "java/math/BigDecimal/ROUND_CEILING": number;
    public static "java/math/BigDecimal/ROUND_FLOOR": number;
    public static "java/math/BigDecimal/ROUND_HALF_UP": number;
    public static "java/math/BigDecimal/ROUND_HALF_DOWN": number;
    public static "java/math/BigDecimal/ROUND_HALF_EVEN": number;
    public static "java/math/BigDecimal/ROUND_UNNECESSARY": number;
    public static "java/math/BigDecimal/double10pow": JVMArray<number>;
    public static "java/math/BigDecimal/float10pow": JVMArray<number>;
    public static "java/math/BigDecimal/LONG_TEN_POWERS_TABLE": JVMArray<Long>;
    public static "java/math/BigDecimal/BIG_TEN_POWERS_TABLE": JVMArray<java_math_BigInteger>;
    public static "java/math/BigDecimal/BIG_TEN_POWERS_TABLE_INITLEN": number;
    public static "java/math/BigDecimal/BIG_TEN_POWERS_TABLE_MAX": number;
    public static "java/math/BigDecimal/THRESHOLDS_TABLE": JVMArray<Long>;
    public static "java/math/BigDecimal/DIV_NUM_BASE": Long;
    public static "java/math/BigDecimal/LONGLONG_TEN_POWERS_TABLE": JVMArray<JVMArray<Long>>;
    public static "java/math/BigDecimal/$assertionsDisabled": number;
    public "<init>(Ljava/math/BigInteger;JII)V"(thread: JVMThread, args: [java_math_BigInteger, Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(Ljava/math/BigInteger;JII)V"(thread: JVMThread, args: [java_math_BigInteger, Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([CIILjava/math/MathContext;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>([CIILjava/math/MathContext;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "adjustScale(IJ)I"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/adjustScale(IJ)I"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "parseExp([CII)J"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/math/BigDecimal/parseExp([CII)J"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "<init>([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([CLjava/math/MathContext;)V"(thread: JVMThread, args: [JVMArray<number>, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>([CLjava/math/MathContext;)V"(thread: JVMThread, args: [JVMArray<number>, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/math/MathContext;)V"(thread: JVMThread, args: [java_lang_String, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(Ljava/lang/String;Ljava/math/MathContext;)V"(thread: JVMThread, args: [java_lang_String, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(DLjava/math/MathContext;)V"(thread: JVMThread, args: [number, any, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(DLjava/math/MathContext;)V"(thread: JVMThread, args: [number, any, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/math/BigInteger;)V"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(Ljava/math/BigInteger;)V"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/math/BigInteger;Ljava/math/MathContext;)V"(thread: JVMThread, args: [java_math_BigInteger, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(Ljava/math/BigInteger;Ljava/math/MathContext;)V"(thread: JVMThread, args: [java_math_BigInteger, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/math/BigInteger;I)V"(thread: JVMThread, args: [java_math_BigInteger, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(Ljava/math/BigInteger;I)V"(thread: JVMThread, args: [java_math_BigInteger, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/math/BigInteger;ILjava/math/MathContext;)V"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(Ljava/math/BigInteger;ILjava/math/MathContext;)V"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/math/MathContext;)V"(thread: JVMThread, args: [number, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(ILjava/math/MathContext;)V"(thread: JVMThread, args: [number, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLjava/math/MathContext;)V"(thread: JVMThread, args: [Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/<init>(JLjava/math/MathContext;)V"(thread: JVMThread, args: [Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable) => void): void;
    public static "valueOf(JI)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/valueOf(JI)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "valueOf(J)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/valueOf(J)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "valueOf(JII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/valueOf(JII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "valueOf(Ljava/math/BigInteger;II)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/valueOf(Ljava/math/BigInteger;II)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "zeroValueOf(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/zeroValueOf(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "valueOf(D)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/valueOf(D)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "add(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/add(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "add(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/add(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "preAlign(Ljava/math/BigDecimal;Ljava/math/BigDecimal;JLjava/math/MathContext;)[Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_BigDecimal, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigDecimal>) => void): void;
    public "java/math/BigDecimal/preAlign(Ljava/math/BigDecimal;Ljava/math/BigDecimal;JLjava/math/MathContext;)[Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_BigDecimal, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigDecimal>) => void): void;
    public "subtract(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/subtract(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "subtract(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/subtract(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "multiply(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/multiply(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "multiply(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/multiply(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "divide(Ljava/math/BigDecimal;II)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/divide(Ljava/math/BigDecimal;II)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "divide(Ljava/math/BigDecimal;ILjava/math/RoundingMode;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, number, java_math_RoundingMode], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/divide(Ljava/math/BigDecimal;ILjava/math/RoundingMode;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, number, java_math_RoundingMode], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "divide(Ljava/math/BigDecimal;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/divide(Ljava/math/BigDecimal;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "divide(Ljava/math/BigDecimal;Ljava/math/RoundingMode;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_RoundingMode], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/divide(Ljava/math/BigDecimal;Ljava/math/RoundingMode;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_RoundingMode], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "divide(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/divide(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "divide(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/divide(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "divideToIntegralValue(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/divideToIntegralValue(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "divideToIntegralValue(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/divideToIntegralValue(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "remainder(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/remainder(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "remainder(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/remainder(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "divideAndRemainder(Ljava/math/BigDecimal;)[Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigDecimal>) => void): void;
    public "java/math/BigDecimal/divideAndRemainder(Ljava/math/BigDecimal;)[Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigDecimal>) => void): void;
    public "divideAndRemainder(Ljava/math/BigDecimal;Ljava/math/MathContext;)[Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigDecimal>) => void): void;
    public "java/math/BigDecimal/divideAndRemainder(Ljava/math/BigDecimal;Ljava/math/MathContext;)[Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_BigDecimal>) => void): void;
    public "pow(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/pow(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "pow(ILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/pow(ILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "abs()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/abs()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "abs(Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/abs(Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "negate()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/negate()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "negate(Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/negate(Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "plus()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/plus()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "plus(Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/plus(Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "signum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/signum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "scale()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/scale()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "precision()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/precision()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unscaledValue()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigDecimal/unscaledValue()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "round(Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/round(Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "setScale(ILjava/math/RoundingMode;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, java_math_RoundingMode], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/setScale(ILjava/math/RoundingMode;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, java_math_RoundingMode], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "setScale(II)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/setScale(II)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "setScale(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/setScale(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "movePointLeft(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/movePointLeft(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "movePointRight(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/movePointRight(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "scaleByPowerOfTen(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/scaleByPowerOfTen(I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "stripTrailingZeros()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/stripTrailingZeros()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "compareTo(Ljava/math/BigDecimal;)I"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/compareTo(Ljava/math/BigDecimal;)I"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareMagnitude(Ljava/math/BigDecimal;)I"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/compareMagnitude(Ljava/math/BigDecimal;)I"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "min(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/min(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "max(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/max(Ljava/math/BigDecimal;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/BigDecimal/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toEngineeringString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/BigDecimal/toEngineeringString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toPlainString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/BigDecimal/toPlainString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getValueString(ILjava/lang/String;I)Ljava/lang/String;"(thread: JVMThread, args: [number, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/BigDecimal/getValueString(ILjava/lang/String;I)Ljava/lang/String;"(thread: JVMThread, args: [number, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toBigInteger()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigDecimal/toBigInteger()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "toBigIntegerExact()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigDecimal/toBigIntegerExact()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/math/BigDecimal/longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "longValueExact()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/math/BigDecimal/longValueExact()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intValueExact()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/intValueExact()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shortValueExact()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/shortValueExact()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "byteValueExact()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/byteValueExact()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ulp()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/ulp()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "layoutChars(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/BigDecimal/layoutChars(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "bigTenToThe(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigDecimal/bigTenToThe(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "expandBigIntegerTenPowers(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigDecimal/expandBigIntegerTenPowers(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "longMultiplyPowerTen(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/math/BigDecimal/longMultiplyPowerTen(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "bigMultiplyPowerTen(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigDecimal/bigMultiplyPowerTen(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "inflated()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/BigDecimal/inflated()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "matchScale([Ljava/math/BigDecimal;)V"(thread: JVMThread, args: [JVMArray<java_math_BigDecimal>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/BigDecimal/matchScale([Ljava/math/BigDecimal;)V"(thread: JVMThread, args: [JVMArray<java_math_BigDecimal>], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/BigDecimal/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "longDigitLength(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/longDigitLength(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "bigDigitLength(Ljava/math/BigInteger;)I"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/bigDigitLength(Ljava/math/BigInteger;)I"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkScale(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/checkScale(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compactValFor(Ljava/math/BigInteger;)J"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/math/BigDecimal/compactValFor(Ljava/math/BigInteger;)J"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "longCompareMagnitude(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/longCompareMagnitude(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "saturateLong(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/saturateLong(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "print(Ljava/lang/String;Ljava/math/BigDecimal;)V"(thread: JVMThread, args: [java_lang_String, java_math_BigDecimal], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/BigDecimal/print(Ljava/lang/String;Ljava/math/BigDecimal;)V"(thread: JVMThread, args: [java_lang_String, java_math_BigDecimal], cb?: (e?: java_lang_Throwable) => void): void;
    public "audit()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/BigDecimal/audit()Ljava/math/BigDecimal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "checkScaleNonZero(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/checkScaleNonZero(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkScale(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/checkScale(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkScale(Ljava/math/BigInteger;J)I"(thread: JVMThread, args: [java_math_BigInteger, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/checkScale(Ljava/math/BigInteger;J)I"(thread: JVMThread, args: [java_math_BigInteger, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "doRound(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/doRound(Ljava/math/BigDecimal;Ljava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigDecimal, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "doRound(JILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/doRound(JILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "doRound(Ljava/math/BigInteger;ILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/doRound(Ljava/math/BigInteger;ILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divideAndRoundByTenPow(Ljava/math/BigInteger;II)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigDecimal/divideAndRoundByTenPow(Ljava/math/BigInteger;II)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "divideAndRound(JJIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divideAndRound(JJIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divideAndRound(JJI)J"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/math/BigDecimal/divideAndRound(JJI)J"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "commonNeedIncrement(IIIZ)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/commonNeedIncrement(IIIZ)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "needIncrement(JIIJJ)Z"(thread: JVMThread, args: [Long, any, number, number, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/needIncrement(JIIJJ)Z"(thread: JVMThread, args: [Long, any, number, number, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "divideAndRound(Ljava/math/BigInteger;JI)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigDecimal/divideAndRound(Ljava/math/BigInteger;JI)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "divideAndRound(Ljava/math/BigInteger;JIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divideAndRound(Ljava/math/BigInteger;JIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "needIncrement(JIILjava/math/MutableBigInteger;J)Z"(thread: JVMThread, args: [Long, any, number, number, java_math_MutableBigInteger, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/needIncrement(JIILjava/math/MutableBigInteger;J)Z"(thread: JVMThread, args: [Long, any, number, number, java_math_MutableBigInteger, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "divideAndRound(Ljava/math/BigInteger;Ljava/math/BigInteger;I)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigDecimal/divideAndRound(Ljava/math/BigInteger;Ljava/math/BigInteger;I)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "divideAndRound(Ljava/math/BigInteger;Ljava/math/BigInteger;III)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divideAndRound(Ljava/math/BigInteger;Ljava/math/BigInteger;III)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "needIncrement(Ljava/math/MutableBigInteger;IILjava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Z"(thread: JVMThread, args: [java_math_MutableBigInteger, number, number, java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/needIncrement(Ljava/math/MutableBigInteger;IILjava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Z"(thread: JVMThread, args: [java_math_MutableBigInteger, number, number, java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "createAndStripZerosToMatchScale(Ljava/math/BigInteger;IJ)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/createAndStripZerosToMatchScale(Ljava/math/BigInteger;IJ)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "createAndStripZerosToMatchScale(JIJ)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/createAndStripZerosToMatchScale(JIJ)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "stripZerosToMatchScale(Ljava/math/BigInteger;JII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/stripZerosToMatchScale(Ljava/math/BigInteger;JII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "add(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/math/BigDecimal/add(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "add(JJI)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/add(JJI)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "add(JIJI)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/add(JIJI)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "add(JILjava/math/BigInteger;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/add(JILjava/math/BigInteger;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "add(Ljava/math/BigInteger;ILjava/math/BigInteger;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/add(Ljava/math/BigInteger;ILjava/math/BigInteger;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "bigMultiplyPowerTen(JI)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigDecimal/bigMultiplyPowerTen(JI)Ljava/math/BigInteger;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "bigMultiplyPowerTen(Ljava/math/BigInteger;I)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigDecimal/bigMultiplyPowerTen(Ljava/math/BigInteger;I)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "divideSmallFastPath(JIJIJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divideSmallFastPath(JIJIJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divide(JIJIJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divide(JIJIJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divide(Ljava/math/BigInteger;IJIJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, Long, any, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divide(Ljava/math/BigInteger;IJIJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, Long, any, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divide(JILjava/math/BigInteger;IJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, java_math_BigInteger, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divide(JILjava/math/BigInteger;IJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, java_math_BigInteger, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divide(Ljava/math/BigInteger;ILjava/math/BigInteger;IJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_BigInteger, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divide(Ljava/math/BigInteger;ILjava/math/BigInteger;IJLjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_BigInteger, number, Long, any, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "multiplyDivideAndRound(JJJIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/multiplyDivideAndRound(JJJIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divideAndRound128(JJJIIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, Long, any, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divideAndRound128(JJJIIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, Long, any, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "roundedTenPower(IIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/roundedTenPower(IIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "scaledTenPow(III)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/scaledTenPow(III)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divRemNegativeLong(JJ)[J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public static "java/math/BigDecimal/divRemNegativeLong(JJ)[J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public static "make64(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/math/BigDecimal/make64(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "mulsub(JJJJJ)J"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/math/BigDecimal/mulsub(JJJJJ)J"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "unsignedLongCompare(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/unsignedLongCompare(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "unsignedLongCompareEq(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/unsignedLongCompareEq(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compareMagnitudeNormalized(JIJI)I"(thread: JVMThread, args: [Long, any, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/compareMagnitudeNormalized(JIJI)I"(thread: JVMThread, args: [Long, any, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compareMagnitudeNormalized(JILjava/math/BigInteger;I)I"(thread: JVMThread, args: [Long, any, number, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/compareMagnitudeNormalized(JILjava/math/BigInteger;I)I"(thread: JVMThread, args: [Long, any, number, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compareMagnitudeNormalized(Ljava/math/BigInteger;ILjava/math/BigInteger;I)I"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/compareMagnitudeNormalized(Ljava/math/BigInteger;ILjava/math/BigInteger;I)I"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "multiply(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/math/BigDecimal/multiply(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "multiply(JJI)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/multiply(JJI)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "multiply(JLjava/math/BigInteger;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/multiply(JLjava/math/BigInteger;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "multiply(Ljava/math/BigInteger;Ljava/math/BigInteger;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/multiply(Ljava/math/BigInteger;Ljava/math/BigInteger;I)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "multiplyAndRound(JJILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/multiplyAndRound(JJILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "multiplyAndRound(JLjava/math/BigInteger;ILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, java_math_BigInteger, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/multiplyAndRound(JLjava/math/BigInteger;ILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, java_math_BigInteger, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "multiplyAndRound(Ljava/math/BigInteger;Ljava/math/BigInteger;ILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/multiplyAndRound(Ljava/math/BigInteger;Ljava/math/BigInteger;ILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, java_math_BigInteger, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "doRound128(JJIILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/doRound128(JJIILjava/math/MathContext;)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, Long, any, number, number, java_math_MathContext], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "precision(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/precision(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "longLongCompareMagnitude(JJJJ)Z"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/BigDecimal/longLongCompareMagnitude(JJJJ)Z"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "divide(JIJIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divide(JIJIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divide(Ljava/math/BigInteger;IJIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divide(Ljava/math/BigInteger;IJIII)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divide(JILjava/math/BigInteger;III)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, java_math_BigInteger, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divide(JILjava/math/BigInteger;III)Ljava/math/BigDecimal;"(thread: JVMThread, args: [Long, any, number, java_math_BigInteger, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "divide(Ljava/math/BigInteger;ILjava/math/BigInteger;III)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_BigInteger, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public static "java/math/BigDecimal/divide(Ljava/math/BigInteger;ILjava/math/BigInteger;III)Ljava/math/BigDecimal;"(thread: JVMThread, args: [java_math_BigInteger, number, java_math_BigInteger, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/BigDecimal/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/math/BigDecimal;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "java/math/BigDecimal/access$000(Ljava/math/BigDecimal;)Ljava/math/BigInteger;"(thread: JVMThread, args: [java_math_BigDecimal], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/BigDecimal/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_math_MutableBigInteger extends java_lang_Object {
    public "java/math/MutableBigInteger/value": JVMArray<number>;
    public "java/math/MutableBigInteger/intLen": number;
    public "java/math/MutableBigInteger/offset": number;
    public static "java/math/MutableBigInteger/ONE": java_math_MutableBigInteger;
    public static "java/math/MutableBigInteger/KNUTH_POW2_THRESH_LEN": number;
    public static "java/math/MutableBigInteger/KNUTH_POW2_THRESH_ZEROS": number;
    public static "java/math/MutableBigInteger/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/<init>([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/math/BigInteger;)V"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/<init>(Ljava/math/BigInteger;)V"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/math/MutableBigInteger;)V"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/<init>(Ljava/math/MutableBigInteger;)V"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "ones(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/ones(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMagnitudeArray()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/math/MutableBigInteger/getMagnitudeArray()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "toLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/math/MutableBigInteger/toLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toBigInteger(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/MutableBigInteger/toBigInteger(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "toBigInteger()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/MutableBigInteger/toBigInteger()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "toBigDecimal(II)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "java/math/MutableBigInteger/toBigDecimal(II)Ljava/math/BigDecimal;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigDecimal) => void): void;
    public "toCompactValue(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/math/MutableBigInteger/toCompactValue(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "compare(Ljava/math/MutableBigInteger;)I"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/compare(Ljava/math/MutableBigInteger;)I"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareShifted(Ljava/math/MutableBigInteger;I)I"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/compareShifted(Ljava/math/MutableBigInteger;I)I"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareHalf(Ljava/math/MutableBigInteger;)I"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/compareHalf(Ljava/math/MutableBigInteger;)I"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLowestSetBit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/getLowestSetBit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/math/MutableBigInteger/getLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "normalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/normalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toIntArray()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/math/MutableBigInteger/toIntArray()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "setInt(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/setInt(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setValue([II)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/setValue([II)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyValue(Ljava/math/MutableBigInteger;)V"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/copyValue(Ljava/math/MutableBigInteger;)V"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyValue([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/copyValue([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "isOne()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/isOne()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/isZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEven()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/isEven()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOdd()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/isOdd()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isNormal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/isNormal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/MutableBigInteger/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "safeRightShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/safeRightShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rightShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/rightShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "safeLeftShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/safeLeftShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "leftShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/leftShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "divadd([I[II)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/divadd([I[II)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mulsub([I[IIII)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/mulsub([I[IIII)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mulsubBorrow([I[IIII)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/mulsubBorrow([I[IIII)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "primitiveRightShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/primitiveRightShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "primitiveLeftShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/primitiveLeftShift(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLower(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/math/MutableBigInteger/getLower(I)Ljava/math/BigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "keepLower(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/keepLower(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/math/MutableBigInteger;)V"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/add(Ljava/math/MutableBigInteger;)V"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "addShifted(Ljava/math/MutableBigInteger;I)V"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/addShifted(Ljava/math/MutableBigInteger;I)V"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addDisjoint(Ljava/math/MutableBigInteger;I)V"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/addDisjoint(Ljava/math/MutableBigInteger;I)V"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addLower(Ljava/math/MutableBigInteger;I)V"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/addLower(Ljava/math/MutableBigInteger;I)V"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "subtract(Ljava/math/MutableBigInteger;)I"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/subtract(Ljava/math/MutableBigInteger;)I"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "difference(Ljava/math/MutableBigInteger;)I"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/difference(Ljava/math/MutableBigInteger;)I"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "multiply(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)V"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/multiply(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)V"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "mul(ILjava/math/MutableBigInteger;)V"(thread: JVMThread, args: [number, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MutableBigInteger/mul(ILjava/math/MutableBigInteger;)V"(thread: JVMThread, args: [number, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "divideOneWord(ILjava/math/MutableBigInteger;)I"(thread: JVMThread, args: [number, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/divideOneWord(ILjava/math/MutableBigInteger;)I"(thread: JVMThread, args: [number, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "divide(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/divide(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "divide(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;Z)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/divide(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;Z)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "divideKnuth(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/divideKnuth(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "divideKnuth(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;Z)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/divideKnuth(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;Z)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "divideAndRemainderBurnikelZiegler(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/divideAndRemainderBurnikelZiegler(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "divide2n1n(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/divide2n1n(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "divide3n2n(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/divide3n2n(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "getBlock(III)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/getBlock(III)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "bitLength()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/math/MutableBigInteger/bitLength()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "divide(JLjava/math/MutableBigInteger;)J"(thread: JVMThread, args: [Long, any, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/math/MutableBigInteger/divide(JLjava/math/MutableBigInteger;)J"(thread: JVMThread, args: [Long, any, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "copyAndShift([III[III)V"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/MutableBigInteger/copyAndShift([III[III)V"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "divideMagnitude(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;Z)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/divideMagnitude(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;Z)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "divideLongMagnitude(JLjava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [Long, any, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/divideLongMagnitude(JLjava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [Long, any, java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "divaddLong(II[II)I"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/divaddLong(II[II)I"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mulsubLong([IIIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/mulsubLong([IIIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unsignedLongCompare(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MutableBigInteger/unsignedLongCompare(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "divWord(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/math/MutableBigInteger/divWord(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "hybridGCD(Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/hybridGCD(Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "binaryGCD(Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/binaryGCD(Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public static "binaryGcd(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/MutableBigInteger/binaryGcd(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mutableModInverse(Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/mutableModInverse(Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "modInverseMP2(I)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/modInverseMP2(I)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public static "inverseMod32(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/math/MutableBigInteger/inverseMod32(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "modInverseBP2(Ljava/math/MutableBigInteger;I)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public static "java/math/MutableBigInteger/modInverseBP2(Ljava/math/MutableBigInteger;I)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "modInverse(Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/modInverse(Ljava/math/MutableBigInteger;)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public static "fixup(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;I)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public static "java/math/MutableBigInteger/fixup(Ljava/math/MutableBigInteger;Ljava/math/MutableBigInteger;I)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [java_math_MutableBigInteger, java_math_MutableBigInteger, number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "euclidModInverse(I)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public "java/math/MutableBigInteger/euclidModInverse(I)Ljava/math/MutableBigInteger;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_MutableBigInteger) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/MutableBigInteger/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_math_RoundingMode extends java_lang_Enum {
    public static "java/math/RoundingMode/UP": java_math_RoundingMode;
    public static "java/math/RoundingMode/DOWN": java_math_RoundingMode;
    public static "java/math/RoundingMode/CEILING": java_math_RoundingMode;
    public static "java/math/RoundingMode/FLOOR": java_math_RoundingMode;
    public static "java/math/RoundingMode/HALF_UP": java_math_RoundingMode;
    public static "java/math/RoundingMode/HALF_DOWN": java_math_RoundingMode;
    public static "java/math/RoundingMode/HALF_EVEN": java_math_RoundingMode;
    public static "java/math/RoundingMode/UNNECESSARY": java_math_RoundingMode;
    public "java/math/RoundingMode/oldMode": number;
    public static "java/math/RoundingMode/$VALUES": JVMArray<java_math_RoundingMode>;
    public static "values()[Ljava/math/RoundingMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_RoundingMode>) => void): void;
    public static "java/math/RoundingMode/values()[Ljava/math/RoundingMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_math_RoundingMode>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/math/RoundingMode;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_math_RoundingMode) => void): void;
    public static "java/math/RoundingMode/valueOf(Ljava/lang/String;)Ljava/math/RoundingMode;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_math_RoundingMode) => void): void;
    public "<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/RoundingMode/<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "valueOf(I)Ljava/math/RoundingMode;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_RoundingMode) => void): void;
    public static "java/math/RoundingMode/valueOf(I)Ljava/math/RoundingMode;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_math_RoundingMode) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/RoundingMode/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_math_MathContext extends java_lang_Object implements java_io_Serializable {
    public static "java/math/MathContext/DEFAULT_DIGITS": number;
    public static "java/math/MathContext/DEFAULT_ROUNDINGMODE": java_math_RoundingMode;
    public static "java/math/MathContext/MIN_DIGITS": number;
    public static "java/math/MathContext/serialVersionUID": Long;
    public static "java/math/MathContext/UNLIMITED": java_math_MathContext;
    public static "java/math/MathContext/DECIMAL32": java_math_MathContext;
    public static "java/math/MathContext/DECIMAL64": java_math_MathContext;
    public static "java/math/MathContext/DECIMAL128": java_math_MathContext;
    public "java/math/MathContext/precision": number;
    public "java/math/MathContext/roundingMode": java_math_RoundingMode;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MathContext/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/math/RoundingMode;)V"(thread: JVMThread, args: [number, java_math_RoundingMode], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MathContext/<init>(ILjava/math/RoundingMode;)V"(thread: JVMThread, args: [number, java_math_RoundingMode], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MathContext/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPrecision()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MathContext/getPrecision()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRoundingMode()Ljava/math/RoundingMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_RoundingMode) => void): void;
    public "java/math/MathContext/getRoundingMode()Ljava/math/RoundingMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_RoundingMode) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MathContext/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/math/MathContext/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/math/MathContext/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/math/MathContext/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/math/MathContext/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_regex_Pattern extends java_lang_Object implements java_io_Serializable {
    public static "java/util/regex/Pattern/UNIX_LINES": number;
    public static "java/util/regex/Pattern/CASE_INSENSITIVE": number;
    public static "java/util/regex/Pattern/COMMENTS": number;
    public static "java/util/regex/Pattern/MULTILINE": number;
    public static "java/util/regex/Pattern/LITERAL": number;
    public static "java/util/regex/Pattern/DOTALL": number;
    public static "java/util/regex/Pattern/UNICODE_CASE": number;
    public static "java/util/regex/Pattern/CANON_EQ": number;
    public static "java/util/regex/Pattern/UNICODE_CHARACTER_CLASS": number;
    public static "java/util/regex/Pattern/serialVersionUID": Long;
    public "java/util/regex/Pattern/pattern": java_lang_String;
    public "java/util/regex/Pattern/flags": number;
    public "java/util/regex/Pattern/compiled": number;
    public "java/util/regex/Pattern/normalizedPattern": java_lang_String;
    public "java/util/regex/Pattern/root": java_util_regex_Pattern$Node;
    public "java/util/regex/Pattern/matchRoot": java_util_regex_Pattern$Node;
    public "java/util/regex/Pattern/buffer": JVMArray<number>;
    public "java/util/regex/Pattern/namedGroups": java_util_Map;
    public "java/util/regex/Pattern/groupNodes": JVMArray<java_util_regex_Pattern$GroupHead>;
    public "java/util/regex/Pattern/temp": JVMArray<number>;
    public "java/util/regex/Pattern/capturingGroupCount": number;
    public "java/util/regex/Pattern/localCount": number;
    public "java/util/regex/Pattern/cursor": number;
    public "java/util/regex/Pattern/patternLength": number;
    public "java/util/regex/Pattern/hasSupplementary": number;
    public static "java/util/regex/Pattern/MAX_REPS": number;
    public static "java/util/regex/Pattern/GREEDY": number;
    public static "java/util/regex/Pattern/LAZY": number;
    public static "java/util/regex/Pattern/POSSESSIVE": number;
    public static "java/util/regex/Pattern/INDEPENDENT": number;
    public static "java/util/regex/Pattern/lookbehindEnd": java_util_regex_Pattern$Node;
    public static "java/util/regex/Pattern/accept": java_util_regex_Pattern$Node;
    public static "java/util/regex/Pattern/lastAccept": java_util_regex_Pattern$Node;
    public static "java/util/regex/Pattern/$assertionsDisabled": number;
    public static "compile(Ljava/lang/String;)Ljava/util/regex/Pattern;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern) => void): void;
    public static "java/util/regex/Pattern/compile(Ljava/lang/String;)Ljava/util/regex/Pattern;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern) => void): void;
    public static "compile(Ljava/lang/String;I)Ljava/util/regex/Pattern;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern) => void): void;
    public static "java/util/regex/Pattern/compile(Ljava/lang/String;I)Ljava/util/regex/Pattern;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern) => void): void;
    public "pattern()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Pattern/pattern()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Pattern/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "matcher(Ljava/lang/CharSequence;)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "java/util/regex/Pattern/matcher(Ljava/lang/CharSequence;)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "flags()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/flags()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "matches(Ljava/lang/String;Ljava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/regex/Pattern/matches(Ljava/lang/String;Ljava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "split(Ljava/lang/CharSequence;I)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/regex/Pattern/split(Ljava/lang/CharSequence;I)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "split(Ljava/lang/CharSequence;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/regex/Pattern/split(Ljava/lang/CharSequence;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "quote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/regex/Pattern/quote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "normalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/normalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "normalizeCharClass(Ljava/lang/StringBuilder;I)I"(thread: JVMThread, args: [java_lang_StringBuilder, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/normalizeCharClass(Ljava/lang/StringBuilder;I)I"(thread: JVMThread, args: [java_lang_StringBuilder, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "produceEquivalentAlternation(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Pattern/produceEquivalentAlternation(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "producePermutations(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/regex/Pattern/producePermutations(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getClass(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/getClass(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "composeOneStep(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Pattern/composeOneStep(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "RemoveQEQuoting()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/RemoveQEQuoting()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "compile()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/compile()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "namedGroups()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/util/regex/Pattern/namedGroups()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "printObjectTree(Ljava/util/regex/Pattern$Node;)V"(thread: JVMThread, args: [java_util_regex_Pattern$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/regex/Pattern/printObjectTree(Ljava/util/regex/Pattern$Node;)V"(thread: JVMThread, args: [java_util_regex_Pattern$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "has(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/has(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "accept(ILjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/accept(ILjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "mark(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/mark(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "peek()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/peek()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readEscaped()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/readEscaped()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "next()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/next()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextEscaped()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/nextEscaped()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peekPastWhitespace(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/peekPastWhitespace(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parsePastWhitespace(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/parsePastWhitespace(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parsePastLine()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/parsePastLine()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peekPastLine()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/peekPastLine()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLineSeparator(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/isLineSeparator(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skip()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/skip()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unread()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/unread()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "error(Ljava/lang/String;)Ljava/util/regex/PatternSyntaxException;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_PatternSyntaxException) => void): void;
    public "java/util/regex/Pattern/error(Ljava/lang/String;)Ljava/util/regex/PatternSyntaxException;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_PatternSyntaxException) => void): void;
    public "findSupplementary(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/findSupplementary(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSupplementary(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/regex/Pattern/isSupplementary(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "expr(Ljava/util/regex/Pattern$Node;)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [java_util_regex_Pattern$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "java/util/regex/Pattern/expr(Ljava/util/regex/Pattern$Node;)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [java_util_regex_Pattern$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "sequence(Ljava/util/regex/Pattern$Node;)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [java_util_regex_Pattern$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "java/util/regex/Pattern/sequence(Ljava/util/regex/Pattern$Node;)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [java_util_regex_Pattern$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "atom()Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "java/util/regex/Pattern/atom()Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "append(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/append(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "ref(I)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "java/util/regex/Pattern/ref(I)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "escape(ZZZ)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/escape(ZZZ)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clazz(Z)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern/clazz(Z)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "bitsOrSingle(Ljava/util/regex/Pattern$BitClass;I)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$BitClass, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern/bitsOrSingle(Ljava/util/regex/Pattern$BitClass;I)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$BitClass, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "range(Ljava/util/regex/Pattern$BitClass;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$BitClass], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern/range(Ljava/util/regex/Pattern$BitClass;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$BitClass], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "family(ZZ)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern/family(ZZ)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "unicodeScriptPropertyFor(Ljava/lang/String;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern/unicodeScriptPropertyFor(Ljava/lang/String;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "unicodeBlockPropertyFor(Ljava/lang/String;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern/unicodeBlockPropertyFor(Ljava/lang/String;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "charPropertyNodeFor(Ljava/lang/String;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern/charPropertyNodeFor(Ljava/lang/String;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "groupname(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Pattern/groupname(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "group0()Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "java/util/regex/Pattern/group0()Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "createGroup(Z)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "java/util/regex/Pattern/createGroup(Z)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "addFlag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/addFlag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "subFlag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/subFlag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "closure(Ljava/util/regex/Pattern$Node;)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [java_util_regex_Pattern$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "java/util/regex/Pattern/closure(Ljava/util/regex/Pattern$Node;)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [java_util_regex_Pattern$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "c()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/c()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "o()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/o()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "x()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/x()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "cursor()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/cursor()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setcursor(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern/setcursor(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "uxxxx()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/uxxxx()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "u()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/u()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "countChars(Ljava/lang/CharSequence;II)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/regex/Pattern/countChars(Ljava/lang/CharSequence;II)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "countCodePoints(Ljava/lang/CharSequence;)I"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/regex/Pattern/countCodePoints(Ljava/lang/CharSequence;)I"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "newSingle(I)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern/newSingle(I)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "newSlice([IIZ)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public "java/util/regex/Pattern/newSlice([IIZ)Ljava/util/regex/Pattern$Node;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$Node) => void): void;
    public static "inRange(III)Z"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/regex/Pattern/inRange(III)Z"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "rangeFor(II)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "java/util/regex/Pattern/rangeFor(II)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "caseInsensitiveRangeFor(II)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern/caseInsensitiveRangeFor(II)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "union(Ljava/util/regex/Pattern$CharProperty;Ljava/util/regex/Pattern$CharProperty;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$CharProperty, java_util_regex_Pattern$CharProperty], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "java/util/regex/Pattern/union(Ljava/util/regex/Pattern$CharProperty;Ljava/util/regex/Pattern$CharProperty;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$CharProperty, java_util_regex_Pattern$CharProperty], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "intersection(Ljava/util/regex/Pattern$CharProperty;Ljava/util/regex/Pattern$CharProperty;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$CharProperty, java_util_regex_Pattern$CharProperty], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "java/util/regex/Pattern/intersection(Ljava/util/regex/Pattern$CharProperty;Ljava/util/regex/Pattern$CharProperty;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$CharProperty, java_util_regex_Pattern$CharProperty], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "setDifference(Ljava/util/regex/Pattern$CharProperty;Ljava/util/regex/Pattern$CharProperty;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$CharProperty, java_util_regex_Pattern$CharProperty], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "java/util/regex/Pattern/setDifference(Ljava/util/regex/Pattern$CharProperty;Ljava/util/regex/Pattern$CharProperty;)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [java_util_regex_Pattern$CharProperty, java_util_regex_Pattern$CharProperty], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "hasBaseCharacter(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/regex/Pattern/hasBaseCharacter(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "asPredicate()Ljava/util/function/Predicate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Predicate) => void): void;
    public "java/util/regex/Pattern/asPredicate()Ljava/util/function/Predicate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_function_Predicate) => void): void;
    public "splitAsStream(Ljava/lang/CharSequence;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    public "java/util/regex/Pattern/splitAsStream(Ljava/lang/CharSequence;)Ljava/util/stream/Stream;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    public "lambda$asPredicate$0(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern/lambda$asPredicate$0(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(III)Z"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/regex/Pattern/access$200(III)Z"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$300(Ljava/lang/CharSequence;II)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/regex/Pattern/access$300(Ljava/lang/CharSequence;II)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$400(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/regex/Pattern/access$400(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$600(II)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "java/util/regex/Pattern/access$600(II)Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/regex/Pattern/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_regex_Pattern$BitClass extends java_util_regex_Pattern$BmpCharProperty {
    public "java/util/regex/Pattern$BitClass/bits": JVMArray<number>;
    public static "java/util/regex/Pattern$BitClass/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$BitClass/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Z)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$BitClass/<init>([Z)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(II)Ljava/util/regex/Pattern$BitClass;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$BitClass) => void): void;
    public "java/util/regex/Pattern$BitClass/add(II)Ljava/util/regex/Pattern$BitClass;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$BitClass) => void): void;
    public "isSatisfiedBy(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$BitClass/isSatisfiedBy(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/regex/Pattern$BitClass/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_regex_Pattern$BmpCharProperty extends java_util_regex_Pattern$CharProperty {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$BmpCharProperty/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "match(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$BmpCharProperty/match(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/util/regex/Pattern$1;)V"(thread: JVMThread, args: [java_util_regex_Pattern$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$BmpCharProperty/<init>(Ljava/util/regex/Pattern$1;)V"(thread: JVMThread, args: [java_util_regex_Pattern$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_regex_Pattern$1 extends java_util_regex_Pattern$CharProperty {
    public "java/util/regex/Pattern$1/val$lower": number;
    public "java/util/regex/Pattern$1/val$upper": number;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$1/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSatisfiedBy(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$1/isSatisfiedBy(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_regex_Pattern$CharProperty extends java_util_regex_Pattern$Node {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$CharProperty/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSatisfiedBy(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$CharProperty/isSatisfiedBy(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "complement()Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "java/util/regex/Pattern$CharProperty/complement()Ljava/util/regex/Pattern$CharProperty;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern$CharProperty) => void): void;
    public "match(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$CharProperty/match(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "study(Ljava/util/regex/Pattern$TreeInfo;)Z"(thread: JVMThread, args: [java_util_regex_Pattern$TreeInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$CharProperty/study(Ljava/util/regex/Pattern$TreeInfo;)Z"(thread: JVMThread, args: [java_util_regex_Pattern$TreeInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/util/regex/Pattern$1;)V"(thread: JVMThread, args: [java_util_regex_Pattern$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$CharProperty/<init>(Ljava/util/regex/Pattern$1;)V"(thread: JVMThread, args: [java_util_regex_Pattern$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_regex_Pattern$TreeInfo extends java_lang_Object {
    public "java/util/regex/Pattern$TreeInfo/minLength": number;
    public "java/util/regex/Pattern$TreeInfo/maxLength": number;
    public "java/util/regex/Pattern$TreeInfo/maxValid": number;
    public "java/util/regex/Pattern$TreeInfo/deterministic": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$TreeInfo/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$TreeInfo/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_regex_PatternSyntaxException extends java_lang_IllegalArgumentException {
    public static "java/util/regex/PatternSyntaxException/serialVersionUID": Long;
    public "java/util/regex/PatternSyntaxException/desc": java_lang_String;
    public "java/util/regex/PatternSyntaxException/pattern": java_lang_String;
    public "java/util/regex/PatternSyntaxException/index": number;
    public static "java/util/regex/PatternSyntaxException/nl": java_lang_String;
    public "<init>(Ljava/lang/String;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/PatternSyntaxException/<init>(Ljava/lang/String;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/PatternSyntaxException/getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/PatternSyntaxException/getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPattern()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/PatternSyntaxException/getPattern()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/PatternSyntaxException/getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/regex/PatternSyntaxException/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_IllegalArgumentException extends java_lang_RuntimeException {
    public static "java/lang/IllegalArgumentException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalArgumentException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalArgumentException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalArgumentException/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalArgumentException/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_RuntimeException extends java_lang_Exception {
    public static "java/lang/RuntimeException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/RuntimeException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/RuntimeException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/RuntimeException/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/RuntimeException/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;ZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/RuntimeException/<init>(Ljava/lang/String;Ljava/lang/Throwable;ZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Exception extends java_lang_Throwable {
    public static "java/lang/Exception/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Exception/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Exception/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Exception/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Exception/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;ZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Exception/<init>(Ljava/lang/String;Ljava/lang/Throwable;ZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_regex_Matcher extends java_lang_Object implements java_util_regex_MatchResult {
    public "java/util/regex/Matcher/parentPattern": java_util_regex_Pattern;
    public "java/util/regex/Matcher/groups": JVMArray<number>;
    public "java/util/regex/Matcher/from": number;
    public "java/util/regex/Matcher/to": number;
    public "java/util/regex/Matcher/lookbehindTo": number;
    public "java/util/regex/Matcher/text": java_lang_CharSequence;
    public static "java/util/regex/Matcher/ENDANCHOR": number;
    public static "java/util/regex/Matcher/NOANCHOR": number;
    public "java/util/regex/Matcher/acceptMode": number;
    public "java/util/regex/Matcher/first": number;
    public "java/util/regex/Matcher/last": number;
    public "java/util/regex/Matcher/oldLast": number;
    public "java/util/regex/Matcher/lastAppendPosition": number;
    public "java/util/regex/Matcher/locals": JVMArray<number>;
    public "java/util/regex/Matcher/hitEnd": number;
    public "java/util/regex/Matcher/requireEnd": number;
    public "java/util/regex/Matcher/transparentBounds": number;
    public "java/util/regex/Matcher/anchoringBounds": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Matcher/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/regex/Pattern;Ljava/lang/CharSequence;)V"(thread: JVMThread, args: [java_util_regex_Pattern, java_lang_CharSequence], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Matcher/<init>(Ljava/util/regex/Pattern;Ljava/lang/CharSequence;)V"(thread: JVMThread, args: [java_util_regex_Pattern, java_lang_CharSequence], cb?: (e?: java_lang_Throwable) => void): void;
    public "pattern()Ljava/util/regex/Pattern;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern) => void): void;
    public "java/util/regex/Matcher/pattern()Ljava/util/regex/Pattern;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern) => void): void;
    public "toMatchResult()Ljava/util/regex/MatchResult;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_MatchResult) => void): void;
    public "java/util/regex/Matcher/toMatchResult()Ljava/util/regex/MatchResult;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_MatchResult) => void): void;
    public "usePattern(Ljava/util/regex/Pattern;)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [java_util_regex_Pattern], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "java/util/regex/Matcher/usePattern(Ljava/util/regex/Pattern;)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [java_util_regex_Pattern], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "reset()Ljava/util/regex/Matcher;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "java/util/regex/Matcher/reset()Ljava/util/regex/Matcher;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "reset(Ljava/lang/CharSequence;)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "java/util/regex/Matcher/reset(Ljava/lang/CharSequence;)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "start()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/start()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "start(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/start(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "start(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/start(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "end()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/end()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "end(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/end(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "end(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/end(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "group()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Matcher/group()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "group(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Matcher/group(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "group(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Matcher/group(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "groupCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/groupCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "matches()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/matches()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "find()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/find()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "find(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/find(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lookingAt()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/lookingAt()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "quoteReplacement(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/regex/Matcher/quoteReplacement(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "appendReplacement(Ljava/lang/StringBuffer;Ljava/lang/String;)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [java_lang_StringBuffer, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "java/util/regex/Matcher/appendReplacement(Ljava/lang/StringBuffer;Ljava/lang/String;)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [java_lang_StringBuffer, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "appendTail(Ljava/lang/StringBuffer;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/util/regex/Matcher/appendTail(Ljava/lang/StringBuffer;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "replaceAll(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Matcher/replaceAll(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "replaceFirst(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Matcher/replaceFirst(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "region(II)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "java/util/regex/Matcher/region(II)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "regionStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/regionStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "regionEnd()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/regionEnd()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasTransparentBounds()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/hasTransparentBounds()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "useTransparentBounds(Z)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "java/util/regex/Matcher/useTransparentBounds(Z)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "hasAnchoringBounds()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/hasAnchoringBounds()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "useAnchoringBounds(Z)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "java/util/regex/Matcher/useAnchoringBounds(Z)Ljava/util/regex/Matcher;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Matcher) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/regex/Matcher/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hitEnd()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/hitEnd()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requireEnd()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/requireEnd()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "search(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/search(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "match(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/match(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTextLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/getTextLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSubSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "java/util/regex/Matcher/getSubSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMatchedGroupIndex(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Matcher/getMatchedGroupIndex(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_regex_MatchResult extends java_lang_Object {
    "start()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "start(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "end()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "end(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "group()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "group(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "groupCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_regex_Pattern$GroupHead extends java_util_regex_Pattern$Node {
    public "java/util/regex/Pattern$GroupHead/localIndex": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$GroupHead/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "match(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$GroupHead/match(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "matchRef(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$GroupHead/matchRef(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_regex_Pattern$Node extends java_lang_Object {
    public "java/util/regex/Pattern$Node/next": java_util_regex_Pattern$Node;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/regex/Pattern$Node/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "match(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$Node/match(Ljava/util/regex/Matcher;ILjava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_util_regex_Matcher, number, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "study(Ljava/util/regex/Pattern$TreeInfo;)Z"(thread: JVMThread, args: [java_util_regex_Pattern$TreeInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/regex/Pattern$Node/study(Ljava/util/regex/Pattern$TreeInfo;)Z"(thread: JVMThread, args: [java_util_regex_Pattern$TreeInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_time_zone_ZoneOffsetTransition extends java_lang_Object implements java_lang_Comparable, java_io_Serializable {
    public static "java/time/zone/ZoneOffsetTransition/serialVersionUID": Long;
    public "java/time/zone/ZoneOffsetTransition/transition": java_time_LocalDateTime;
    public "java/time/zone/ZoneOffsetTransition/offsetBefore": java_time_ZoneOffset;
    public "java/time/zone/ZoneOffsetTransition/offsetAfter": java_time_ZoneOffset;
    public static "of(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public static "java/time/zone/ZoneOffsetTransition/of(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "<init>(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneOffsetTransition/<init>(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLjava/time/ZoneOffset;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [Long, any, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneOffsetTransition/<init>(JLjava/time/ZoneOffset;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [Long, any, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneOffsetTransition/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/zone/ZoneOffsetTransition/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneOffsetTransition/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public static "java/time/zone/ZoneOffsetTransition/readExternal(Ljava/io/DataInput;)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "getInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/zone/ZoneOffsetTransition/getInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "toEpochSecond()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/zone/ZoneOffsetTransition/toEpochSecond()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getDateTimeBefore()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/zone/ZoneOffsetTransition/getDateTimeBefore()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "getDateTimeAfter()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/zone/ZoneOffsetTransition/getDateTimeAfter()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "getOffsetBefore()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/zone/ZoneOffsetTransition/getOffsetBefore()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "getOffsetAfter()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/zone/ZoneOffsetTransition/getOffsetAfter()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "getDuration()Ljava/time/Duration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "java/time/zone/ZoneOffsetTransition/getDuration()Ljava/time/Duration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    public "getDurationSeconds()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransition/getDurationSeconds()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isGap()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransition/isGap()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOverlap()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransition/isOverlap()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isValidOffset(Ljava/time/ZoneOffset;)Z"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransition/isValidOffset(Ljava/time/ZoneOffset;)Z"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getValidOffsets()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/time/zone/ZoneOffsetTransition/getValidOffsets()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "compareTo(Ljava/time/zone/ZoneOffsetTransition;)I"(thread: JVMThread, args: [java_time_zone_ZoneOffsetTransition], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransition/compareTo(Ljava/time/zone/ZoneOffsetTransition;)I"(thread: JVMThread, args: [java_time_zone_ZoneOffsetTransition], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransition/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransition/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/zone/ZoneOffsetTransition/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransition/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_concurrent_ConcurrentMap extends java_lang_Object, java_util_Map {
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "lambda$replaceAll$0(Ljava/util/function/BiFunction;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_function_BiFunction, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_time_zone_ZoneOffsetTransitionRule extends java_lang_Object implements java_io_Serializable {
    public static "java/time/zone/ZoneOffsetTransitionRule/serialVersionUID": Long;
    public "java/time/zone/ZoneOffsetTransitionRule/month": java_time_Month;
    public "java/time/zone/ZoneOffsetTransitionRule/dom": number;
    public "java/time/zone/ZoneOffsetTransitionRule/dow": java_time_DayOfWeek;
    public "java/time/zone/ZoneOffsetTransitionRule/time": java_time_LocalTime;
    public "java/time/zone/ZoneOffsetTransitionRule/timeEndOfDay": number;
    public "java/time/zone/ZoneOffsetTransitionRule/timeDefinition": java_time_zone_ZoneOffsetTransitionRule$TimeDefinition;
    public "java/time/zone/ZoneOffsetTransitionRule/standardOffset": java_time_ZoneOffset;
    public "java/time/zone/ZoneOffsetTransitionRule/offsetBefore": java_time_ZoneOffset;
    public "java/time/zone/ZoneOffsetTransitionRule/offsetAfter": java_time_ZoneOffset;
    public static "of(Ljava/time/Month;ILjava/time/DayOfWeek;Ljava/time/LocalTime;ZLjava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)Ljava/time/zone/ZoneOffsetTransitionRule;"(thread: JVMThread, args: [java_time_Month, number, java_time_DayOfWeek, java_time_LocalTime, number, java_time_zone_ZoneOffsetTransitionRule$TimeDefinition, java_time_ZoneOffset, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransitionRule) => void): void;
    public static "java/time/zone/ZoneOffsetTransitionRule/of(Ljava/time/Month;ILjava/time/DayOfWeek;Ljava/time/LocalTime;ZLjava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)Ljava/time/zone/ZoneOffsetTransitionRule;"(thread: JVMThread, args: [java_time_Month, number, java_time_DayOfWeek, java_time_LocalTime, number, java_time_zone_ZoneOffsetTransitionRule$TimeDefinition, java_time_ZoneOffset, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransitionRule) => void): void;
    public "<init>(Ljava/time/Month;ILjava/time/DayOfWeek;Ljava/time/LocalTime;ZLjava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_Month, number, java_time_DayOfWeek, java_time_LocalTime, number, java_time_zone_ZoneOffsetTransitionRule$TimeDefinition, java_time_ZoneOffset, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/<init>(Ljava/time/Month;ILjava/time/DayOfWeek;Ljava/time/LocalTime;ZLjava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_Month, number, java_time_DayOfWeek, java_time_LocalTime, number, java_time_zone_ZoneOffsetTransitionRule$TimeDefinition, java_time_ZoneOffset, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/zone/ZoneOffsetTransitionRule;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransitionRule) => void): void;
    public static "java/time/zone/ZoneOffsetTransitionRule/readExternal(Ljava/io/DataInput;)Ljava/time/zone/ZoneOffsetTransitionRule;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransitionRule) => void): void;
    public "getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "getDayOfMonthIndicator()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/getDayOfMonthIndicator()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "getLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/getLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "isMidnightEndOfDay()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/isMidnightEndOfDay()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTimeDefinition()Ljava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransitionRule$TimeDefinition) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/getTimeDefinition()Ljava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransitionRule$TimeDefinition) => void): void;
    public "getStandardOffset()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/getStandardOffset()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "getOffsetBefore()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/getOffsetBefore()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "getOffsetAfter()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/getOffsetAfter()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "createTransition(I)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/createTransition(I)Ljava/time/zone/ZoneOffsetTransition;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransition) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_time_zone_ZoneOffsetTransitionRule$TimeDefinition extends java_lang_Enum {
    public static "java/time/zone/ZoneOffsetTransitionRule$TimeDefinition/UTC": java_time_zone_ZoneOffsetTransitionRule$TimeDefinition;
    public static "java/time/zone/ZoneOffsetTransitionRule$TimeDefinition/WALL": java_time_zone_ZoneOffsetTransitionRule$TimeDefinition;
    public static "java/time/zone/ZoneOffsetTransitionRule$TimeDefinition/STANDARD": java_time_zone_ZoneOffsetTransitionRule$TimeDefinition;
    public static "java/time/zone/ZoneOffsetTransitionRule$TimeDefinition/$VALUES": JVMArray<java_time_zone_ZoneOffsetTransitionRule$TimeDefinition>;
    public static "values()[Ljava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_zone_ZoneOffsetTransitionRule$TimeDefinition>) => void): void;
    public static "java/time/zone/ZoneOffsetTransitionRule$TimeDefinition/values()[Ljava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_zone_ZoneOffsetTransitionRule$TimeDefinition>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransitionRule$TimeDefinition) => void): void;
    public static "java/time/zone/ZoneOffsetTransitionRule$TimeDefinition/valueOf(Ljava/lang/String;)Ljava/time/zone/ZoneOffsetTransitionRule$TimeDefinition;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneOffsetTransitionRule$TimeDefinition) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule$TimeDefinition/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "createDateTime(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/zone/ZoneOffsetTransitionRule$TimeDefinition/createDateTime(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneOffset;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/zone/ZoneOffsetTransitionRule$TimeDefinition/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_LocalTime extends java_lang_Object implements java_time_temporal_Temporal, java_time_temporal_TemporalAdjuster, java_lang_Comparable, java_io_Serializable {
    public static "java/time/LocalTime/MIN": java_time_LocalTime;
    public static "java/time/LocalTime/MAX": java_time_LocalTime;
    public static "java/time/LocalTime/MIDNIGHT": java_time_LocalTime;
    public static "java/time/LocalTime/NOON": java_time_LocalTime;
    public static "java/time/LocalTime/HOURS": JVMArray<java_time_LocalTime>;
    public static "java/time/LocalTime/HOURS_PER_DAY": number;
    public static "java/time/LocalTime/MINUTES_PER_HOUR": number;
    public static "java/time/LocalTime/MINUTES_PER_DAY": number;
    public static "java/time/LocalTime/SECONDS_PER_MINUTE": number;
    public static "java/time/LocalTime/SECONDS_PER_HOUR": number;
    public static "java/time/LocalTime/SECONDS_PER_DAY": number;
    public static "java/time/LocalTime/MILLIS_PER_DAY": Long;
    public static "java/time/LocalTime/MICROS_PER_DAY": Long;
    public static "java/time/LocalTime/NANOS_PER_SECOND": Long;
    public static "java/time/LocalTime/NANOS_PER_MINUTE": Long;
    public static "java/time/LocalTime/NANOS_PER_HOUR": Long;
    public static "java/time/LocalTime/NANOS_PER_DAY": Long;
    public static "java/time/LocalTime/serialVersionUID": Long;
    public "java/time/LocalTime/hour": number;
    public "java/time/LocalTime/minute": number;
    public "java/time/LocalTime/second": number;
    public "java/time/LocalTime/nano": number;
    public static "now()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/now()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "now(Ljava/time/ZoneId;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/now(Ljava/time/ZoneId;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "now(Ljava/time/Clock;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/now(Ljava/time/Clock;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "of(II)Ljava/time/LocalTime;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/of(II)Ljava/time/LocalTime;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "of(III)Ljava/time/LocalTime;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/of(III)Ljava/time/LocalTime;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "of(IIII)Ljava/time/LocalTime;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/of(IIII)Ljava/time/LocalTime;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "ofSecondOfDay(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/ofSecondOfDay(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "ofNanoOfDay(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/ofNanoOfDay(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/parse(Ljava/lang/CharSequence;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "create(IIII)Ljava/time/LocalTime;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/create(IIII)Ljava/time/LocalTime;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/LocalTime/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/LocalTime/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalTime/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "get0(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/get0(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "withHour(I)Ljava/time/LocalTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/withHour(I)Ljava/time/LocalTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "withMinute(I)Ljava/time/LocalTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/withMinute(I)Ljava/time/LocalTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "withSecond(I)Ljava/time/LocalTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/withSecond(I)Ljava/time/LocalTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "withNano(I)Ljava/time/LocalTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/withNano(I)Ljava/time/LocalTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "plusHours(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/plusHours(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "plusMinutes(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/plusMinutes(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "plusSeconds(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/plusSeconds(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "plusNanos(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/plusNanos(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "minusHours(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/minusHours(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "minusMinutes(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/minusMinutes(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "minusSeconds(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/minusSeconds(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "minusNanos(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalTime/minusNanos(J)Ljava/time/LocalTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/LocalTime/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalTime/adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalTime/until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/LocalTime/format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "atDate(Ljava/time/LocalDate;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalTime/atDate(Ljava/time/LocalDate;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "atOffset(Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/LocalTime/atOffset(Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "toSecondOfDay()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/toSecondOfDay()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toNanoOfDay()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalTime/toNanoOfDay()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareTo(Ljava/time/LocalTime;)I"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/compareTo(Ljava/time/LocalTime;)I"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAfter(Ljava/time/LocalTime;)Z"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/isAfter(Ljava/time/LocalTime;)Z"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBefore(Ljava/time/LocalTime;)Z"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/isBefore(Ljava/time/LocalTime;)Z"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/LocalTime/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/LocalTime/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/LocalTime/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/LocalTime/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public static "java/time/LocalTime/readExternal(Ljava/io/DataInput;)Ljava/time/LocalTime;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalTime/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/LocalTime/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_time_OffsetTime extends java_lang_Object implements java_time_temporal_Temporal, java_time_temporal_TemporalAdjuster, java_lang_Comparable, java_io_Serializable {
    public static "java/time/OffsetTime/MIN": java_time_OffsetTime;
    public static "java/time/OffsetTime/MAX": java_time_OffsetTime;
    public static "java/time/OffsetTime/serialVersionUID": Long;
    public "java/time/OffsetTime/time": java_time_LocalTime;
    public "java/time/OffsetTime/offset": java_time_ZoneOffset;
    public static "now()Ljava/time/OffsetTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/now()Ljava/time/OffsetTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "now(Ljava/time/ZoneId;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/now(Ljava/time/ZoneId;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "now(Ljava/time/Clock;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/now(Ljava/time/Clock;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "of(Ljava/time/LocalTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_LocalTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/of(Ljava/time/LocalTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_LocalTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "of(IIIILjava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number, number, number, number, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/of(IIIILjava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number, number, number, number, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "ofInstant(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/ofInstant(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/parse(Ljava/lang/CharSequence;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "<init>(Ljava/time/LocalTime;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_LocalTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/OffsetTime/<init>(Ljava/time/LocalTime;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_LocalTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "with(Ljava/time/LocalTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_LocalTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/with(Ljava/time/LocalTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_LocalTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/OffsetTime/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/OffsetTime/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getOffset()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/OffsetTime/getOffset()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "withOffsetSameLocal(Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/withOffsetSameLocal(Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "withOffsetSameInstant(Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/withOffsetSameInstant(Ljava/time/ZoneOffset;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/OffsetTime/toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "withHour(I)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/withHour(I)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "withMinute(I)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/withMinute(I)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "withSecond(I)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/withSecond(I)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "withNano(I)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/withNano(I)Ljava/time/OffsetTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "plusHours(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/plusHours(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "plusMinutes(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/plusMinutes(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "plusSeconds(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/plusSeconds(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "plusNanos(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/plusNanos(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "minusHours(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/minusHours(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "minusMinutes(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/minusMinutes(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "minusSeconds(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/minusSeconds(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "minusNanos(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetTime/minusNanos(J)Ljava/time/OffsetTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/OffsetTime/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetTime/adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/OffsetTime/until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/OffsetTime/format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "atDate(Ljava/time/LocalDate;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetTime/atDate(Ljava/time/LocalDate;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "toEpochNano()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/OffsetTime/toEpochNano()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareTo(Ljava/time/OffsetTime;)I"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/compareTo(Ljava/time/OffsetTime;)I"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAfter(Ljava/time/OffsetTime;)Z"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/isAfter(Ljava/time/OffsetTime;)Z"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBefore(Ljava/time/OffsetTime;)Z"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/isBefore(Ljava/time/OffsetTime;)Z"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEqual(Ljava/time/OffsetTime;)Z"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/isEqual(Ljava/time/OffsetTime;)Z"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/OffsetTime/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/OffsetTime/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/OffsetTime/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/OffsetTime/writeExternal(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/ObjectInput;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public static "java/time/OffsetTime/readExternal(Ljava/io/ObjectInput;)Ljava/time/OffsetTime;"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetTime/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/OffsetTime/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_io_ObjectInput extends java_lang_Object, java_io_DataInput, java_lang_AutoCloseable {
    "readObject()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    "readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_io_ObjectOutput extends java_lang_Object, java_io_DataOutput, java_lang_AutoCloseable {
    "writeObject(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    "writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    "writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_LocalDate extends java_lang_Object implements java_time_temporal_Temporal, java_time_temporal_TemporalAdjuster, java_time_chrono_ChronoLocalDate, java_io_Serializable {
    public static "java/time/LocalDate/MIN": java_time_LocalDate;
    public static "java/time/LocalDate/MAX": java_time_LocalDate;
    public static "java/time/LocalDate/serialVersionUID": Long;
    public static "java/time/LocalDate/DAYS_PER_CYCLE": number;
    public static "java/time/LocalDate/DAYS_0000_TO_1970": Long;
    public "java/time/LocalDate/year": number;
    public "java/time/LocalDate/month": number;
    public "java/time/LocalDate/day": number;
    public static "now()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/now()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "now(Ljava/time/ZoneId;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/now(Ljava/time/ZoneId;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "now(Ljava/time/Clock;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/now(Ljava/time/Clock;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "of(ILjava/time/Month;I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, java_time_Month, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/of(ILjava/time/Month;I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, java_time_Month, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "of(III)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/of(III)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "ofYearDay(II)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/ofYearDay(II)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "ofEpochDay(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/ofEpochDay(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "parse(Ljava/lang/CharSequence;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/parse(Ljava/lang/CharSequence;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "create(III)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/create(III)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "resolvePreviousValid(III)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/resolvePreviousValid(III)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/LocalDate/<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/LocalDate/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalDate/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "get0(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/get0(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getProlepticMonth()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalDate/getProlepticMonth()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getChronology()Ljava/time/chrono/IsoChronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoChronology) => void): void;
    public "java/time/LocalDate/getChronology()Ljava/time/chrono/IsoChronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoChronology) => void): void;
    public "getEra()Ljava/time/chrono/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Era) => void): void;
    public "java/time/LocalDate/getEra()Ljava/time/chrono/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Era) => void): void;
    public "getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonthValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/getMonthValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "java/time/LocalDate/getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/getDayOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "java/time/LocalDate/getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "isLeapYear()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/isLeapYear()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lengthOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/lengthOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lengthOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/lengthOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/with(Ljava/time/temporal/TemporalField;J)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "withYear(I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/withYear(I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "withMonth(I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/withMonth(I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "withDayOfMonth(I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/withDayOfMonth(I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "withDayOfYear(I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/withDayOfYear(I)Ljava/time/LocalDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "plusYears(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/plusYears(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "plusMonths(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/plusMonths(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "plusWeeks(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/plusWeeks(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "plusDays(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/plusDays(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "minusYears(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/minusYears(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "minusMonths(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/minusMonths(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "minusWeeks(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/minusWeeks(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "minusDays(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDate/minusDays(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/LocalDate/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDate/adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalDate/until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "daysUntil(Ljava/time/LocalDate;)J"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalDate/daysUntil(Ljava/time/LocalDate;)J"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "monthsUntil(Ljava/time/LocalDate;)J"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalDate/monthsUntil(Ljava/time/LocalDate;)J"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "until(Ljava/time/chrono/ChronoLocalDate;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/LocalDate/until(Ljava/time/chrono/ChronoLocalDate;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/LocalDate/format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "atTime(Ljava/time/LocalTime;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDate/atTime(Ljava/time/LocalTime;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "atTime(II)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDate/atTime(II)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "atTime(III)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDate/atTime(III)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "atTime(IIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDate/atTime(IIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "atTime(Ljava/time/OffsetTime;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/LocalDate/atTime(Ljava/time/OffsetTime;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_OffsetTime], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "atStartOfDay()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDate/atStartOfDay()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "atStartOfDay(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/LocalDate/atStartOfDay(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "toEpochDay()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalDate/toEpochDay()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareTo(Ljava/time/chrono/ChronoLocalDate;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/compareTo(Ljava/time/chrono/ChronoLocalDate;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo0(Ljava/time/LocalDate;)I"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/compareTo0(Ljava/time/LocalDate;)I"(thread: JVMThread, args: [java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAfter(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/isAfter(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBefore(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/isBefore(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEqual(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/isEqual(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/LocalDate/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/LocalDate/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/LocalDate/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/LocalDate/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public static "java/time/LocalDate/readExternal(Ljava/io/DataInput;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDate/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDate/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDate/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDate/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDate/with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDate/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "atTime(Ljava/time/LocalTime;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "java/time/LocalDate/atTime(Ljava/time/LocalTime;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "until(Ljava/time/chrono/ChronoLocalDate;)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "java/time/LocalDate/until(Ljava/time/chrono/ChronoLocalDate;)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/LocalDate/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/LocalDate/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/LocalDate/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/LocalDate/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/LocalDate/with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/LocalDate/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public "java/time/LocalDate/getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDate/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/LocalDate/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getEra()Ljava/time/chrono/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Era) => void): void;
    "isLeapYear()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lengthOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "atTime(Ljava/time/LocalTime;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "toEpochDay()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "compareTo(Ljava/time/chrono/ChronoLocalDate;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isAfter(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isBefore(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEqual(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_time_chrono_Chronology extends java_lang_Object, java_lang_Comparable {
    "getId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getCalendarType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "date(Ljava/time/chrono/Era;III)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "date(III)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateYearDay(Ljava/time/chrono/Era;II)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateYearDay(II)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateEpochDay(J)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateNow()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateNow(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateNow(Ljava/time/Clock;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "date(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "localDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "zonedDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "zonedDateTime(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "isLeapYear(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "prolepticYear(Ljava/time/chrono/Era;I)I"(thread: JVMThread, args: [java_time_chrono_Era, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "eraOf(I)Ljava/time/chrono/Era;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Era) => void): void;
    "eras()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "range(Ljava/time/temporal/ChronoField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_ChronoField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "resolveDate(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "period(III)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    "compareTo(Ljava/time/chrono/Chronology;)I"(thread: JVMThread, args: [java_time_chrono_Chronology], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_time_format_ResolverStyle extends java_lang_Enum {
    public static "java/time/format/ResolverStyle/STRICT": java_time_format_ResolverStyle;
    public static "java/time/format/ResolverStyle/SMART": java_time_format_ResolverStyle;
    public static "java/time/format/ResolverStyle/LENIENT": java_time_format_ResolverStyle;
    public static "java/time/format/ResolverStyle/$VALUES": JVMArray<java_time_format_ResolverStyle>;
    public static "values()[Ljava/time/format/ResolverStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_format_ResolverStyle>) => void): void;
    public static "java/time/format/ResolverStyle/values()[Ljava/time/format/ResolverStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_format_ResolverStyle>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/time/format/ResolverStyle;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_format_ResolverStyle) => void): void;
    public static "java/time/format/ResolverStyle/valueOf(Ljava/lang/String;)Ljava/time/format/ResolverStyle;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_format_ResolverStyle) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/ResolverStyle/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/format/ResolverStyle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_temporal_ChronoField extends java_lang_Enum implements java_time_temporal_TemporalField {
    public static "java/time/temporal/ChronoField/NANO_OF_SECOND": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/NANO_OF_DAY": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/MICRO_OF_SECOND": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/MICRO_OF_DAY": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/MILLI_OF_SECOND": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/MILLI_OF_DAY": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/SECOND_OF_MINUTE": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/SECOND_OF_DAY": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/MINUTE_OF_HOUR": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/MINUTE_OF_DAY": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/HOUR_OF_AMPM": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/CLOCK_HOUR_OF_AMPM": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/HOUR_OF_DAY": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/CLOCK_HOUR_OF_DAY": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/AMPM_OF_DAY": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/DAY_OF_WEEK": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/ALIGNED_DAY_OF_WEEK_IN_MONTH": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/ALIGNED_DAY_OF_WEEK_IN_YEAR": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/DAY_OF_MONTH": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/DAY_OF_YEAR": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/EPOCH_DAY": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/ALIGNED_WEEK_OF_MONTH": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/ALIGNED_WEEK_OF_YEAR": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/MONTH_OF_YEAR": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/PROLEPTIC_MONTH": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/YEAR_OF_ERA": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/YEAR": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/ERA": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/INSTANT_SECONDS": java_time_temporal_ChronoField;
    public static "java/time/temporal/ChronoField/OFFSET_SECONDS": java_time_temporal_ChronoField;
    public "java/time/temporal/ChronoField/name": java_lang_String;
    public "java/time/temporal/ChronoField/baseUnit": java_time_temporal_TemporalUnit;
    public "java/time/temporal/ChronoField/rangeUnit": java_time_temporal_TemporalUnit;
    public "java/time/temporal/ChronoField/range": java_time_temporal_ValueRange;
    public "java/time/temporal/ChronoField/displayNameKey": java_lang_String;
    public static "java/time/temporal/ChronoField/$VALUES": JVMArray<java_time_temporal_ChronoField>;
    public static "values()[Ljava/time/temporal/ChronoField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_temporal_ChronoField>) => void): void;
    public static "java/time/temporal/ChronoField/values()[Ljava/time/temporal/ChronoField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_temporal_ChronoField>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/time/temporal/ChronoField;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ChronoField) => void): void;
    public static "java/time/temporal/ChronoField/valueOf(Ljava/lang/String;)Ljava/time/temporal/ChronoField;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ChronoField) => void): void;
    public "<init>(Ljava/lang/String;ILjava/lang/String;Ljava/time/temporal/TemporalUnit;Ljava/time/temporal/TemporalUnit;Ljava/time/temporal/ValueRange;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, java_time_temporal_TemporalUnit, java_time_temporal_TemporalUnit, java_time_temporal_ValueRange], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/temporal/ChronoField/<init>(Ljava/lang/String;ILjava/lang/String;Ljava/time/temporal/TemporalUnit;Ljava/time/temporal/TemporalUnit;Ljava/time/temporal/ValueRange;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, java_time_temporal_TemporalUnit, java_time_temporal_TemporalUnit, java_time_temporal_ValueRange], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;ILjava/lang/String;Ljava/time/temporal/TemporalUnit;Ljava/time/temporal/TemporalUnit;Ljava/time/temporal/ValueRange;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, java_time_temporal_TemporalUnit, java_time_temporal_TemporalUnit, java_time_temporal_ValueRange, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/temporal/ChronoField/<init>(Ljava/lang/String;ILjava/lang/String;Ljava/time/temporal/TemporalUnit;Ljava/time/temporal/TemporalUnit;Ljava/time/temporal/ValueRange;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, java_time_temporal_TemporalUnit, java_time_temporal_TemporalUnit, java_time_temporal_ValueRange, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/temporal/ChronoField/getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getBaseUnit()Ljava/time/temporal/TemporalUnit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalUnit) => void): void;
    public "java/time/temporal/ChronoField/getBaseUnit()Ljava/time/temporal/TemporalUnit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalUnit) => void): void;
    public "getRangeUnit()Ljava/time/temporal/TemporalUnit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalUnit) => void): void;
    public "java/time/temporal/ChronoField/getRangeUnit()Ljava/time/temporal/TemporalUnit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalUnit) => void): void;
    public "range()Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/temporal/ChronoField/range()Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "isDateBased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ChronoField/isDateBased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isTimeBased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ChronoField/isTimeBased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkValidValue(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/temporal/ChronoField/checkValidValue(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "checkValidIntValue(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ChronoField/checkValidIntValue(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupportedBy(Ljava/time/temporal/TemporalAccessor;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ChronoField/isSupportedBy(Ljava/time/temporal/TemporalAccessor;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rangeRefinedBy(Ljava/time/temporal/TemporalAccessor;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/temporal/ChronoField/rangeRefinedBy(Ljava/time/temporal/TemporalAccessor;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "getFrom(Ljava/time/temporal/TemporalAccessor;)J"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/temporal/ChronoField/getFrom(Ljava/time/temporal/TemporalAccessor;)J"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/temporal/ChronoField/adjustInto(Ljava/time/temporal/Temporal;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/temporal/ChronoField/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/temporal/ChronoField/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "resolve(Ljava/util/Map;Ljava/time/temporal/TemporalAccessor;Ljava/time/format/ResolverStyle;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_util_Map, java_time_temporal_TemporalAccessor, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    "getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_time_chrono_ChronoZonedDateTime extends java_lang_Object, java_time_temporal_Temporal, java_lang_Comparable {
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "toLocalDate()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    "toLocalDateTime()Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    "getOffset()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    "getZone()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    "withEarlierOffsetAtOverlap()Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "withLaterOffsetAtOverlap()Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "withZoneSameLocal(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "withZoneSameInstant(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "toInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    "toEpochSecond()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "compareTo(Ljava/time/chrono/ChronoZonedDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isBefore(Ljava/time/chrono/ChronoZonedDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isAfter(Ljava/time/chrono/ChronoZonedDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEqual(Ljava/time/chrono/ChronoZonedDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_time_chrono_ChronoPeriod extends java_lang_Object, java_time_temporal_TemporalAmount {
    "get(Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getUnits()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    "isZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isNegative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    "multipliedBy(I)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    "negated()Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    "normalized()Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    "addTo(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "subtractFrom(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_time_chrono_ChronoLocalDateTime extends java_lang_Object, java_time_temporal_Temporal, java_time_temporal_TemporalAdjuster, java_lang_Comparable {
    "getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    "toLocalDate()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "atZone(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "toInstant(Ljava/time/ZoneOffset;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    "toEpochSecond(Ljava/time/ZoneOffset;)J"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "compareTo(Ljava/time/chrono/ChronoLocalDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isAfter(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isBefore(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEqual(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_time_Period extends java_lang_Object implements java_time_chrono_ChronoPeriod, java_io_Serializable {
    public static "java/time/Period/ZERO": java_time_Period;
    public static "java/time/Period/serialVersionUID": Long;
    public static "java/time/Period/PATTERN": java_util_regex_Pattern;
    public static "java/time/Period/SUPPORTED_UNITS": java_util_List;
    public "java/time/Period/years": number;
    public "java/time/Period/months": number;
    public "java/time/Period/days": number;
    public static "ofYears(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/ofYears(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "ofMonths(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/ofMonths(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "ofWeeks(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/ofWeeks(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "ofDays(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/ofDays(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "of(III)Ljava/time/Period;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/of(III)Ljava/time/Period;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "from(Ljava/time/temporal/TemporalAmount;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/from(Ljava/time/temporal/TemporalAmount;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "parse(Ljava/lang/CharSequence;)Ljava/time/Period;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/parse(Ljava/lang/CharSequence;)Ljava/time/Period;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "parseNumber(Ljava/lang/CharSequence;Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/Period/parseNumber(Ljava/lang/CharSequence;Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "between(Ljava/time/LocalDate;Ljava/time/LocalDate;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/between(Ljava/time/LocalDate;Ljava/time/LocalDate;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "create(III)Ljava/time/Period;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/create(III)Ljava/time/Period;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Period/<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Period/get(Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getUnits()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/time/Period/getUnits()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getChronology()Ljava/time/chrono/IsoChronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoChronology) => void): void;
    public "java/time/Period/getChronology()Ljava/time/chrono/IsoChronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoChronology) => void): void;
    public "isZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Period/isZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isNegative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Period/isNegative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getYears()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Period/getYears()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonths()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Period/getMonths()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDays()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Period/getDays()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "withYears(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/withYears(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "withMonths(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/withMonths(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "withDays(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/withDays(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "plusYears(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/plusYears(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "plusMonths(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/plusMonths(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "plusDays(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/plusDays(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "minusYears(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/minusYears(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "minusMonths(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/minusMonths(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "minusDays(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/minusDays(J)Ljava/time/Period;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "multipliedBy(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/multipliedBy(I)Ljava/time/Period;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "negated()Ljava/time/Period;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/negated()Ljava/time/Period;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "normalized()Ljava/time/Period;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/Period/normalized()Ljava/time/Period;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "toTotalMonths()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Period/toTotalMonths()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "addTo(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Period/addTo(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "subtractFrom(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Period/subtractFrom(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "validateChrono(Ljava/time/temporal/TemporalAccessor;)V"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Period/validateChrono(Ljava/time/temporal/TemporalAccessor;)V"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Period/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Period/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/Period/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/Period/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Period/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Period/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/Period;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/Period/readExternal(Ljava/io/DataInput;)Ljava/time/Period;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "normalized()Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "java/time/Period/normalized()Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "negated()Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "java/time/Period/negated()Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "multipliedBy(I)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "java/time/Period/multipliedBy(I)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "java/time/Period/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "java/time/Period/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public "java/time/Period/getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/Period/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isNegative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "negated()Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
  }
  export interface java_time_chrono_Era extends java_lang_Object, java_time_temporal_TemporalAccessor, java_time_temporal_TemporalAdjuster {
    "getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_time_chrono_IsoChronology extends java_time_chrono_AbstractChronology implements java_io_Serializable {
    public static "java/time/chrono/IsoChronology/INSTANCE": java_time_chrono_IsoChronology;
    public static "java/time/chrono/IsoChronology/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/chrono/IsoChronology/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/chrono/IsoChronology/getId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCalendarType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/chrono/IsoChronology/getCalendarType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "date(Ljava/time/chrono/Era;III)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/date(Ljava/time/chrono/Era;III)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "date(III)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/date(III)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "dateYearDay(Ljava/time/chrono/Era;II)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateYearDay(Ljava/time/chrono/Era;II)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "dateYearDay(II)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateYearDay(II)Ljava/time/LocalDate;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "dateEpochDay(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateEpochDay(J)Ljava/time/LocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "date(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/date(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "localDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/chrono/IsoChronology/localDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "zonedDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/chrono/IsoChronology/zonedDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "zonedDateTime(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/chrono/IsoChronology/zonedDateTime(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "dateNow()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateNow()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "dateNow(Ljava/time/ZoneId;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateNow(Ljava/time/ZoneId;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "dateNow(Ljava/time/Clock;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateNow(Ljava/time/Clock;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "isLeapYear(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/chrono/IsoChronology/isLeapYear(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "prolepticYear(Ljava/time/chrono/Era;I)I"(thread: JVMThread, args: [java_time_chrono_Era, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/chrono/IsoChronology/prolepticYear(Ljava/time/chrono/Era;I)I"(thread: JVMThread, args: [java_time_chrono_Era, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "eraOf(I)Ljava/time/chrono/IsoEra;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoEra) => void): void;
    public "java/time/chrono/IsoChronology/eraOf(I)Ljava/time/chrono/IsoEra;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoEra) => void): void;
    public "eras()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/time/chrono/IsoChronology/eras()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "resolveDate(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/resolveDate(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "resolveProlepticMonth(Ljava/util/Map;Ljava/time/format/ResolverStyle;)V"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/chrono/IsoChronology/resolveProlepticMonth(Ljava/util/Map;Ljava/time/format/ResolverStyle;)V"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveYearOfEra(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/resolveYearOfEra(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "resolveYMD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/chrono/IsoChronology/resolveYMD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/LocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "range(Ljava/time/temporal/ChronoField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_ChronoField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/chrono/IsoChronology/range(Ljava/time/temporal/ChronoField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_ChronoField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "period(III)Ljava/time/Period;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "java/time/chrono/IsoChronology/period(III)Ljava/time/Period;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/chrono/IsoChronology/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/chrono/IsoChronology/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveYMD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/resolveYMD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveYearOfEra(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/resolveYearOfEra(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveDate(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/resolveDate(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "period(III)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "java/time/chrono/IsoChronology/period(III)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    public "eraOf(I)Ljava/time/chrono/Era;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Era) => void): void;
    public "java/time/chrono/IsoChronology/eraOf(I)Ljava/time/chrono/Era;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Era) => void): void;
    public "zonedDateTime(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/chrono/IsoChronology/zonedDateTime(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "zonedDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/chrono/IsoChronology/zonedDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "localDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "java/time/chrono/IsoChronology/localDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "date(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/date(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "dateNow(Ljava/time/Clock;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateNow(Ljava/time/Clock;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "dateNow(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateNow(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "dateNow()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateNow()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "dateEpochDay(J)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateEpochDay(J)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "dateYearDay(II)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateYearDay(II)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "dateYearDay(Ljava/time/chrono/Era;II)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/dateYearDay(Ljava/time/chrono/Era;II)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "date(III)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/date(III)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "date(Ljava/time/chrono/Era;III)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/IsoChronology/date(Ljava/time/chrono/Era;III)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/chrono/IsoChronology/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_chrono_IsoEra extends java_lang_Enum implements java_time_chrono_Era {
    public static "java/time/chrono/IsoEra/BCE": java_time_chrono_IsoEra;
    public static "java/time/chrono/IsoEra/CE": java_time_chrono_IsoEra;
    public static "java/time/chrono/IsoEra/$VALUES": JVMArray<java_time_chrono_IsoEra>;
    public static "values()[Ljava/time/chrono/IsoEra;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_chrono_IsoEra>) => void): void;
    public static "java/time/chrono/IsoEra/values()[Ljava/time/chrono/IsoEra;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_chrono_IsoEra>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/time/chrono/IsoEra;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoEra) => void): void;
    public static "java/time/chrono/IsoEra/valueOf(Ljava/lang/String;)Ljava/time/chrono/IsoEra;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoEra) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/chrono/IsoEra/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "of(I)Ljava/time/chrono/IsoEra;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoEra) => void): void;
    public static "java/time/chrono/IsoEra/of(I)Ljava/time/chrono/IsoEra;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_IsoEra) => void): void;
    public "getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/chrono/IsoEra/getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/chrono/IsoEra/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_time_chrono_AbstractChronology extends java_lang_Object implements java_time_chrono_Chronology {
    public static "java/time/chrono/AbstractChronology/DATE_ORDER": java_util_Comparator;
    public static "java/time/chrono/AbstractChronology/DATE_TIME_ORDER": java_util_Comparator;
    public static "java/time/chrono/AbstractChronology/INSTANT_ORDER": java_util_Comparator;
    public static "java/time/chrono/AbstractChronology/CHRONOS_BY_ID": java_util_concurrent_ConcurrentHashMap;
    public static "java/time/chrono/AbstractChronology/CHRONOS_BY_TYPE": java_util_concurrent_ConcurrentHashMap;
    public static "registerChrono(Ljava/time/chrono/Chronology;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_time_chrono_Chronology], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "java/time/chrono/AbstractChronology/registerChrono(Ljava/time/chrono/Chronology;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_time_chrono_Chronology], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "registerChrono(Ljava/time/chrono/Chronology;Ljava/lang/String;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_time_chrono_Chronology, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "java/time/chrono/AbstractChronology/registerChrono(Ljava/time/chrono/Chronology;Ljava/lang/String;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_time_chrono_Chronology, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "initCache()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/chrono/AbstractChronology/initCache()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ofLocale(Ljava/util/Locale;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "java/time/chrono/AbstractChronology/ofLocale(Ljava/util/Locale;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "of(Ljava/lang/String;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "java/time/chrono/AbstractChronology/of(Ljava/lang/String;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "of0(Ljava/lang/String;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "java/time/chrono/AbstractChronology/of0(Ljava/lang/String;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "getAvailableChronologies()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "java/time/chrono/AbstractChronology/getAvailableChronologies()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/chrono/AbstractChronology/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveDate(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/AbstractChronology/resolveDate(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveProlepticMonth(Ljava/util/Map;Ljava/time/format/ResolverStyle;)V"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/chrono/AbstractChronology/resolveProlepticMonth(Ljava/util/Map;Ljava/time/format/ResolverStyle;)V"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveYearOfEra(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/AbstractChronology/resolveYearOfEra(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveYMD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/AbstractChronology/resolveYMD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveYD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/AbstractChronology/resolveYD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveYMAA(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/AbstractChronology/resolveYMAA(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveYMAD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/AbstractChronology/resolveYMAD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveYAA(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/AbstractChronology/resolveYAA(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveYAD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/AbstractChronology/resolveYAD(Ljava/util/Map;Ljava/time/format/ResolverStyle;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_util_Map, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "resolveAligned(Ljava/time/chrono/ChronoLocalDate;JJJ)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/chrono/AbstractChronology/resolveAligned(Ljava/time/chrono/ChronoLocalDate;JJJ)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "addFieldValue(Ljava/util/Map;Ljava/time/temporal/ChronoField;J)V"(thread: JVMThread, args: [java_util_Map, java_time_temporal_ChronoField, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/chrono/AbstractChronology/addFieldValue(Ljava/util/Map;Ljava/time/temporal/ChronoField;J)V"(thread: JVMThread, args: [java_util_Map, java_time_temporal_ChronoField, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "compareTo(Ljava/time/chrono/Chronology;)I"(thread: JVMThread, args: [java_time_chrono_Chronology], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/chrono/AbstractChronology/compareTo(Ljava/time/chrono/Chronology;)I"(thread: JVMThread, args: [java_time_chrono_Chronology], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/chrono/AbstractChronology/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/chrono/AbstractChronology/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/chrono/AbstractChronology/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/chrono/AbstractChronology/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/chrono/AbstractChronology/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/chrono/AbstractChronology/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public static "java/time/chrono/AbstractChronology/readExternal(Ljava/io/DataInput;)Ljava/time/chrono/Chronology;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/chrono/AbstractChronology/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "$deserializeLambda$(Ljava/lang/invoke/SerializedLambda;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_SerializedLambda], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/time/chrono/AbstractChronology/$deserializeLambda$(Ljava/lang/invoke/SerializedLambda;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_SerializedLambda], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "lambda$static$2241c452$1(Ljava/time/chrono/ChronoZonedDateTime;Ljava/time/chrono/ChronoZonedDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime, java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/chrono/AbstractChronology/lambda$static$2241c452$1(Ljava/time/chrono/ChronoZonedDateTime;Ljava/time/chrono/ChronoZonedDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime, java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "lambda$static$b5a61975$1(Ljava/time/chrono/ChronoLocalDateTime;Ljava/time/chrono/ChronoLocalDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime, java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/chrono/AbstractChronology/lambda$static$b5a61975$1(Ljava/time/chrono/ChronoLocalDateTime;Ljava/time/chrono/ChronoLocalDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime, java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "lambda$static$7f2d2d5b$1(Ljava/time/chrono/ChronoLocalDate;Ljava/time/chrono/ChronoLocalDate;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate, java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/chrono/AbstractChronology/lambda$static$7f2d2d5b$1(Ljava/time/chrono/ChronoLocalDate;Ljava/time/chrono/ChronoLocalDate;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate, java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/chrono/AbstractChronology/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getCalendarType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "date(Ljava/time/chrono/Era;III)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "date(III)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateYearDay(Ljava/time/chrono/Era;II)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_chrono_Era, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateYearDay(II)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateEpochDay(J)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateNow()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateNow(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "dateNow(Ljava/time/Clock;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "date(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "localDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "zonedDateTime(Ljava/time/temporal/TemporalAccessor;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "zonedDateTime(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "isLeapYear(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "prolepticYear(Ljava/time/chrono/Era;I)I"(thread: JVMThread, args: [java_time_chrono_Era, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "eraOf(I)Ljava/time/chrono/Era;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Era) => void): void;
    "eras()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "range(Ljava/time/temporal/ChronoField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_ChronoField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "period(III)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_concurrent_ConcurrentHashMap extends java_util_AbstractMap implements java_util_concurrent_ConcurrentMap, java_io_Serializable {
    public static "java/util/concurrent/ConcurrentHashMap/serialVersionUID": Long;
    public static "java/util/concurrent/ConcurrentHashMap/MAXIMUM_CAPACITY": number;
    public static "java/util/concurrent/ConcurrentHashMap/DEFAULT_CAPACITY": number;
    public static "java/util/concurrent/ConcurrentHashMap/MAX_ARRAY_SIZE": number;
    public static "java/util/concurrent/ConcurrentHashMap/DEFAULT_CONCURRENCY_LEVEL": number;
    public static "java/util/concurrent/ConcurrentHashMap/LOAD_FACTOR": number;
    public static "java/util/concurrent/ConcurrentHashMap/TREEIFY_THRESHOLD": number;
    public static "java/util/concurrent/ConcurrentHashMap/UNTREEIFY_THRESHOLD": number;
    public static "java/util/concurrent/ConcurrentHashMap/MIN_TREEIFY_CAPACITY": number;
    public static "java/util/concurrent/ConcurrentHashMap/MIN_TRANSFER_STRIDE": number;
    public static "java/util/concurrent/ConcurrentHashMap/RESIZE_STAMP_BITS": number;
    public static "java/util/concurrent/ConcurrentHashMap/MAX_RESIZERS": number;
    public static "java/util/concurrent/ConcurrentHashMap/RESIZE_STAMP_SHIFT": number;
    public static "java/util/concurrent/ConcurrentHashMap/MOVED": number;
    public static "java/util/concurrent/ConcurrentHashMap/TREEBIN": number;
    public static "java/util/concurrent/ConcurrentHashMap/RESERVED": number;
    public static "java/util/concurrent/ConcurrentHashMap/HASH_BITS": number;
    public static "java/util/concurrent/ConcurrentHashMap/NCPU": number;
    public static "java/util/concurrent/ConcurrentHashMap/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public "java/util/concurrent/ConcurrentHashMap/table": JVMArray<java_util_concurrent_ConcurrentHashMap$Node>;
    public "java/util/concurrent/ConcurrentHashMap/nextTable": JVMArray<java_util_concurrent_ConcurrentHashMap$Node>;
    public "java/util/concurrent/ConcurrentHashMap/baseCount": Long;
    public "java/util/concurrent/ConcurrentHashMap/sizeCtl": number;
    public "java/util/concurrent/ConcurrentHashMap/transferIndex": number;
    public "java/util/concurrent/ConcurrentHashMap/cellsBusy": number;
    public "java/util/concurrent/ConcurrentHashMap/counterCells": JVMArray<java_util_concurrent_ConcurrentHashMap$CounterCell>;
    public "java/util/concurrent/ConcurrentHashMap/keySet": java_util_concurrent_ConcurrentHashMap$KeySetView;
    public "java/util/concurrent/ConcurrentHashMap/values": java_util_concurrent_ConcurrentHashMap$ValuesView;
    public "java/util/concurrent/ConcurrentHashMap/entrySet": java_util_concurrent_ConcurrentHashMap$EntrySetView;
    public static "java/util/concurrent/ConcurrentHashMap/U": sun_misc_Unsafe;
    public static "java/util/concurrent/ConcurrentHashMap/SIZECTL": Long;
    public static "java/util/concurrent/ConcurrentHashMap/TRANSFERINDEX": Long;
    public static "java/util/concurrent/ConcurrentHashMap/BASECOUNT": Long;
    public static "java/util/concurrent/ConcurrentHashMap/CELLSBUSY": Long;
    public static "java/util/concurrent/ConcurrentHashMap/CELLVALUE": Long;
    public static "java/util/concurrent/ConcurrentHashMap/ABASE": Long;
    public static "java/util/concurrent/ConcurrentHashMap/ASHIFT": number;
    public static "spread(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/spread(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "tableSizeFor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/tableSizeFor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "comparableClassFor(Ljava/lang/Object;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/comparableClassFor(Ljava/lang/Object;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "compareComparables(Ljava/lang/Class;Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Class, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/compareComparables(Ljava/lang/Class;Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Class, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "tabAt([Ljava/util/concurrent/ConcurrentHashMap$Node;I)Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, number], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$Node) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/tabAt([Ljava/util/concurrent/ConcurrentHashMap$Node;I)Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, number], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$Node) => void): void;
    public static "casTabAt([Ljava/util/concurrent/ConcurrentHashMap$Node;ILjava/util/concurrent/ConcurrentHashMap$Node;Ljava/util/concurrent/ConcurrentHashMap$Node;)Z"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, number, java_util_concurrent_ConcurrentHashMap$Node, java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/casTabAt([Ljava/util/concurrent/ConcurrentHashMap$Node;ILjava/util/concurrent/ConcurrentHashMap$Node;Ljava/util/concurrent/ConcurrentHashMap$Node;)Z"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, number, java_util_concurrent_ConcurrentHashMap$Node, java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setTabAt([Ljava/util/concurrent/ConcurrentHashMap$Node;ILjava/util/concurrent/ConcurrentHashMap$Node;)V"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, number, java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/setTabAt([Ljava/util/concurrent/ConcurrentHashMap$Node;ILjava/util/concurrent/ConcurrentHashMap$Node;)V"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, number, java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IFI)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/<init>(IFI)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putVal(Ljava/lang/Object;Ljava/lang/Object;Z)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/putVal(Ljava/lang/Object;Ljava/lang/Object;Z)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "replaceNode(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/replaceNode(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "keySet()Ljava/util/concurrent/ConcurrentHashMap$KeySetView;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$KeySetView) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/keySet()Ljava/util/concurrent/ConcurrentHashMap$KeySetView;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$KeySetView) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/keys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "mappingCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/mappingCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "newKeySet()Ljava/util/concurrent/ConcurrentHashMap$KeySetView;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$KeySetView) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/newKeySet()Ljava/util/concurrent/ConcurrentHashMap$KeySetView;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$KeySetView) => void): void;
    public static "newKeySet(I)Ljava/util/concurrent/ConcurrentHashMap$KeySetView;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$KeySetView) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/newKeySet(I)Ljava/util/concurrent/ConcurrentHashMap$KeySetView;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$KeySetView) => void): void;
    public "keySet(Ljava/lang/Object;)Ljava/util/concurrent/ConcurrentHashMap$KeySetView;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$KeySetView) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/keySet(Ljava/lang/Object;)Ljava/util/concurrent/ConcurrentHashMap$KeySetView;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$KeySetView) => void): void;
    public static "resizeStamp(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/resizeStamp(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "initTable()[Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_concurrent_ConcurrentHashMap$Node>) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/initTable()[Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_concurrent_ConcurrentHashMap$Node>) => void): void;
    public "addCount(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/addCount(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "helpTransfer([Ljava/util/concurrent/ConcurrentHashMap$Node;Ljava/util/concurrent/ConcurrentHashMap$Node;)[Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_concurrent_ConcurrentHashMap$Node>) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/helpTransfer([Ljava/util/concurrent/ConcurrentHashMap$Node;Ljava/util/concurrent/ConcurrentHashMap$Node;)[Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_concurrent_ConcurrentHashMap$Node>) => void): void;
    public "tryPresize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/tryPresize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "transfer([Ljava/util/concurrent/ConcurrentHashMap$Node;[Ljava/util/concurrent/ConcurrentHashMap$Node;)V"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, JVMArray<java_util_concurrent_ConcurrentHashMap$Node>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/transfer([Ljava/util/concurrent/ConcurrentHashMap$Node;[Ljava/util/concurrent/ConcurrentHashMap$Node;)V"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, JVMArray<java_util_concurrent_ConcurrentHashMap$Node>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sumCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/sumCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "fullAddCount(JZ)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/fullAddCount(JZ)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "treeifyBin([Ljava/util/concurrent/ConcurrentHashMap$Node;I)V"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/treeifyBin([Ljava/util/concurrent/ConcurrentHashMap$Node;I)V"(thread: JVMThread, args: [JVMArray<java_util_concurrent_ConcurrentHashMap$Node>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "untreeify(Ljava/util/concurrent/ConcurrentHashMap$Node;)Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$Node) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/untreeify(Ljava/util/concurrent/ConcurrentHashMap$Node;)Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$Node) => void): void;
    public "batchFor(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/batchFor(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "forEach(JLjava/util/function/BiConsumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/forEach(JLjava/util/function/BiConsumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "forEach(JLjava/util/function/BiFunction;Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/forEach(JLjava/util/function/BiFunction;Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "search(JLjava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/search(JLjava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduce(JLjava/util/function/BiFunction;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduce(JLjava/util/function/BiFunction;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduceToDouble(JLjava/util/function/ToDoubleBiFunction;DLjava/util/function/DoubleBinaryOperator;)D"(thread: JVMThread, args: [Long, any, java_util_function_ToDoubleBiFunction, number, any, java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceToDouble(JLjava/util/function/ToDoubleBiFunction;DLjava/util/function/DoubleBinaryOperator;)D"(thread: JVMThread, args: [Long, any, java_util_function_ToDoubleBiFunction, number, any, java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reduceToLong(JLjava/util/function/ToLongBiFunction;JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_ToLongBiFunction, Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceToLong(JLjava/util/function/ToLongBiFunction;JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_ToLongBiFunction, Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "reduceToInt(JLjava/util/function/ToIntBiFunction;ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [Long, any, java_util_function_ToIntBiFunction, number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceToInt(JLjava/util/function/ToIntBiFunction;ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [Long, any, java_util_function_ToIntBiFunction, number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "forEachKey(JLjava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/forEachKey(JLjava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "forEachKey(JLjava/util/function/Function;Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/forEachKey(JLjava/util/function/Function;Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "searchKeys(JLjava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/searchKeys(JLjava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduceKeys(JLjava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceKeys(JLjava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduceKeys(JLjava/util/function/Function;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceKeys(JLjava/util/function/Function;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduceKeysToDouble(JLjava/util/function/ToDoubleFunction;DLjava/util/function/DoubleBinaryOperator;)D"(thread: JVMThread, args: [Long, any, java_util_function_ToDoubleFunction, number, any, java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceKeysToDouble(JLjava/util/function/ToDoubleFunction;DLjava/util/function/DoubleBinaryOperator;)D"(thread: JVMThread, args: [Long, any, java_util_function_ToDoubleFunction, number, any, java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reduceKeysToLong(JLjava/util/function/ToLongFunction;JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_ToLongFunction, Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceKeysToLong(JLjava/util/function/ToLongFunction;JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_ToLongFunction, Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "reduceKeysToInt(JLjava/util/function/ToIntFunction;ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [Long, any, java_util_function_ToIntFunction, number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceKeysToInt(JLjava/util/function/ToIntFunction;ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [Long, any, java_util_function_ToIntFunction, number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "forEachValue(JLjava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/forEachValue(JLjava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "forEachValue(JLjava/util/function/Function;Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/forEachValue(JLjava/util/function/Function;Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "searchValues(JLjava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/searchValues(JLjava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduceValues(JLjava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceValues(JLjava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduceValues(JLjava/util/function/Function;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceValues(JLjava/util/function/Function;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduceValuesToDouble(JLjava/util/function/ToDoubleFunction;DLjava/util/function/DoubleBinaryOperator;)D"(thread: JVMThread, args: [Long, any, java_util_function_ToDoubleFunction, number, any, java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceValuesToDouble(JLjava/util/function/ToDoubleFunction;DLjava/util/function/DoubleBinaryOperator;)D"(thread: JVMThread, args: [Long, any, java_util_function_ToDoubleFunction, number, any, java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reduceValuesToLong(JLjava/util/function/ToLongFunction;JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_ToLongFunction, Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceValuesToLong(JLjava/util/function/ToLongFunction;JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_ToLongFunction, Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "reduceValuesToInt(JLjava/util/function/ToIntFunction;ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [Long, any, java_util_function_ToIntFunction, number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceValuesToInt(JLjava/util/function/ToIntFunction;ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [Long, any, java_util_function_ToIntFunction, number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "forEachEntry(JLjava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/forEachEntry(JLjava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "forEachEntry(JLjava/util/function/Function;Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/forEachEntry(JLjava/util/function/Function;Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "searchEntries(JLjava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/searchEntries(JLjava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduceEntries(JLjava/util/function/BiFunction;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceEntries(JLjava/util/function/BiFunction;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [Long, any, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "reduceEntries(JLjava/util/function/Function;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceEntries(JLjava/util/function/Function;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_function_Function, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reduceEntriesToDouble(JLjava/util/function/ToDoubleFunction;DLjava/util/function/DoubleBinaryOperator;)D"(thread: JVMThread, args: [Long, any, java_util_function_ToDoubleFunction, number, any, java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceEntriesToDouble(JLjava/util/function/ToDoubleFunction;DLjava/util/function/DoubleBinaryOperator;)D"(thread: JVMThread, args: [Long, any, java_util_function_ToDoubleFunction, number, any, java_util_function_DoubleBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reduceEntriesToLong(JLjava/util/function/ToLongFunction;JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_ToLongFunction, Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceEntriesToLong(JLjava/util/function/ToLongFunction;JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_ToLongFunction, Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "reduceEntriesToInt(JLjava/util/function/ToIntFunction;ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [Long, any, java_util_function_ToIntFunction, number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/reduceEntriesToInt(JLjava/util/function/ToIntFunction;ILjava/util/function/IntBinaryOperator;)I"(thread: JVMThread, args: [Long, any, java_util_function_ToIntFunction, number, java_util_function_IntBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/concurrent/ConcurrentHashMap/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/ConcurrentHashMap/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "lambda$replaceAll$0(Ljava/util/function/BiFunction;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_function_BiFunction, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_function_ToIntBiFunction extends java_lang_Object {
    "applyAsInt(Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_function_ToLongBiFunction extends java_lang_Object {
    "applyAsLong(Ljava/lang/Object;Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export interface java_util_function_ToDoubleBiFunction extends java_lang_Object {
    "applyAsDouble(Ljava/lang/Object;Ljava/lang/Object;)D"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_util_Enumeration extends java_lang_Object {
    "hasMoreElements()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "nextElement()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_concurrent_ConcurrentHashMap$EntrySetView extends java_util_concurrent_ConcurrentHashMap$CollectionView implements java_util_Set, java_io_Serializable {
    public static "java/util/concurrent/ConcurrentHashMap$EntrySetView/serialVersionUID": Long;
    public "<init>(Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/<init>(Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "add(Ljava/util/Map$Entry;)Z"(thread: JVMThread, args: [java_util_Map$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/add(Ljava/util/Map$Entry;)Z"(thread: JVMThread, args: [java_util_Map$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$EntrySetView/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ConcurrentHashMap$CollectionView extends java_lang_Object implements java_util_Collection, java_io_Serializable {
    public static "java/util/concurrent/ConcurrentHashMap$CollectionView/serialVersionUID": Long;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/map": java_util_concurrent_ConcurrentHashMap;
    public static "java/util/concurrent/ConcurrentHashMap$CollectionView/oomeMsg": java_lang_String;
    public "<init>(Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/<init>(Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMap()Ljava/util/concurrent/ConcurrentHashMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/getMap()Ljava/util/concurrent/ConcurrentHashMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CollectionView/retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ConcurrentHashMap$ValuesView extends java_util_concurrent_ConcurrentHashMap$CollectionView implements java_util_Collection, java_io_Serializable {
    public static "java/util/concurrent/ConcurrentHashMap$ValuesView/serialVersionUID": Long;
    public "<init>(Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$ValuesView/<init>(Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$ValuesView/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$ValuesView/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$ValuesView/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$ValuesView/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$ValuesView/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$ValuesView/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$ValuesView/forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ConcurrentHashMap$KeySetView extends java_util_concurrent_ConcurrentHashMap$CollectionView implements java_util_Set, java_io_Serializable {
    public static "java/util/concurrent/ConcurrentHashMap$KeySetView/serialVersionUID": Long;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/value": java_lang_Object;
    public "<init>(Ljava/util/concurrent/ConcurrentHashMap;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/<init>(Ljava/util/concurrent/ConcurrentHashMap;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMappedValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/getMappedValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMap()Ljava/util/concurrent/ConcurrentHashMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$KeySetView/getMap()Ljava/util/concurrent/ConcurrentHashMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ConcurrentHashMap$CounterCell extends java_lang_Object {
    public "java/util/concurrent/ConcurrentHashMap$CounterCell/value": Long;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$CounterCell/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ConcurrentHashMap$Node extends java_lang_Object implements java_util_Map$Entry {
    public "java/util/concurrent/ConcurrentHashMap$Node/hash": number;
    public "java/util/concurrent/ConcurrentHashMap$Node/key": java_lang_Object;
    public "java/util/concurrent/ConcurrentHashMap$Node/val": java_lang_Object;
    public "java/util/concurrent/ConcurrentHashMap$Node/next": java_util_concurrent_ConcurrentHashMap$Node;
    public "<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/concurrent/ConcurrentHashMap$Node;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$Node/<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/concurrent/ConcurrentHashMap$Node;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_concurrent_ConcurrentHashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$Node/getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$Node/getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$Node/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$Node/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$Node/setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$Node/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "find(ILjava/lang/Object;)Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$Node) => void): void;
    public "java/util/concurrent/ConcurrentHashMap$Node/find(ILjava/lang/Object;)Ljava/util/concurrent/ConcurrentHashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentHashMap$Node) => void): void;
  }
  export interface java_time_chrono_ChronoLocalDate extends java_lang_Object, java_time_temporal_Temporal, java_time_temporal_TemporalAdjuster, java_lang_Comparable {
    "getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    "getEra()Ljava/time/chrono/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Era) => void): void;
    "isLeapYear()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lengthOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "lengthOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "until(Ljava/time/chrono/ChronoLocalDate;)Ljava/time/chrono/ChronoPeriod;"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoPeriod) => void): void;
    "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "atTime(Ljava/time/LocalTime;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "toEpochDay()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "compareTo(Ljava/time/chrono/ChronoLocalDate;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isAfter(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isBefore(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEqual(Ljava/time/chrono/ChronoLocalDate;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_time_format_DateTimeFormatter extends java_lang_Object {
    public "java/time/format/DateTimeFormatter/printerParser": java_time_format_DateTimeFormatterBuilder$CompositePrinterParser;
    public "java/time/format/DateTimeFormatter/locale": java_util_Locale;
    public "java/time/format/DateTimeFormatter/decimalStyle": java_time_format_DecimalStyle;
    public "java/time/format/DateTimeFormatter/resolverStyle": java_time_format_ResolverStyle;
    public "java/time/format/DateTimeFormatter/resolverFields": java_util_Set;
    public "java/time/format/DateTimeFormatter/chrono": java_time_chrono_Chronology;
    public "java/time/format/DateTimeFormatter/zone": java_time_ZoneId;
    public static "java/time/format/DateTimeFormatter/ISO_LOCAL_DATE": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_OFFSET_DATE": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_DATE": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_LOCAL_TIME": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_OFFSET_TIME": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_TIME": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_LOCAL_DATE_TIME": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_OFFSET_DATE_TIME": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_ZONED_DATE_TIME": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_DATE_TIME": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_ORDINAL_DATE": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_WEEK_DATE": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/ISO_INSTANT": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/BASIC_ISO_DATE": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/RFC_1123_DATE_TIME": java_time_format_DateTimeFormatter;
    public static "java/time/format/DateTimeFormatter/PARSED_EXCESS_DAYS": java_time_temporal_TemporalQuery;
    public static "java/time/format/DateTimeFormatter/PARSED_LEAP_SECOND": java_time_temporal_TemporalQuery;
    public static "ofPattern(Ljava/lang/String;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "java/time/format/DateTimeFormatter/ofPattern(Ljava/lang/String;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "ofPattern(Ljava/lang/String;Ljava/util/Locale;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "java/time/format/DateTimeFormatter/ofPattern(Ljava/lang/String;Ljava/util/Locale;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "ofLocalizedDate(Ljava/time/format/FormatStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_FormatStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "java/time/format/DateTimeFormatter/ofLocalizedDate(Ljava/time/format/FormatStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_FormatStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "ofLocalizedTime(Ljava/time/format/FormatStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_FormatStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "java/time/format/DateTimeFormatter/ofLocalizedTime(Ljava/time/format/FormatStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_FormatStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "ofLocalizedDateTime(Ljava/time/format/FormatStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_FormatStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "java/time/format/DateTimeFormatter/ofLocalizedDateTime(Ljava/time/format/FormatStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_FormatStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "ofLocalizedDateTime(Ljava/time/format/FormatStyle;Ljava/time/format/FormatStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_FormatStyle, java_time_format_FormatStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "java/time/format/DateTimeFormatter/ofLocalizedDateTime(Ljava/time/format/FormatStyle;Ljava/time/format/FormatStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_FormatStyle, java_time_format_FormatStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public static "parsedExcessDays()Ljava/time/temporal/TemporalQuery;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalQuery) => void): void;
    public static "java/time/format/DateTimeFormatter/parsedExcessDays()Ljava/time/temporal/TemporalQuery;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalQuery) => void): void;
    public static "parsedLeapSecond()Ljava/time/temporal/TemporalQuery;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalQuery) => void): void;
    public static "java/time/format/DateTimeFormatter/parsedLeapSecond()Ljava/time/temporal/TemporalQuery;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalQuery) => void): void;
    public "<init>(Ljava/time/format/DateTimeFormatterBuilder$CompositePrinterParser;Ljava/util/Locale;Ljava/time/format/DecimalStyle;Ljava/time/format/ResolverStyle;Ljava/util/Set;Ljava/time/chrono/Chronology;Ljava/time/ZoneId;)V"(thread: JVMThread, args: [java_time_format_DateTimeFormatterBuilder$CompositePrinterParser, java_util_Locale, java_time_format_DecimalStyle, java_time_format_ResolverStyle, java_util_Set, java_time_chrono_Chronology, java_time_ZoneId], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeFormatter/<init>(Ljava/time/format/DateTimeFormatterBuilder$CompositePrinterParser;Ljava/util/Locale;Ljava/time/format/DecimalStyle;Ljava/time/format/ResolverStyle;Ljava/util/Set;Ljava/time/chrono/Chronology;Ljava/time/ZoneId;)V"(thread: JVMThread, args: [java_time_format_DateTimeFormatterBuilder$CompositePrinterParser, java_util_Locale, java_time_format_DecimalStyle, java_time_format_ResolverStyle, java_util_Set, java_time_chrono_Chronology, java_time_ZoneId], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/time/format/DateTimeFormatter/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "withLocale(Ljava/util/Locale;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "java/time/format/DateTimeFormatter/withLocale(Ljava/util/Locale;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "getDecimalStyle()Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "java/time/format/DateTimeFormatter/getDecimalStyle()Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "withDecimalStyle(Ljava/time/format/DecimalStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_DecimalStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "java/time/format/DateTimeFormatter/withDecimalStyle(Ljava/time/format/DecimalStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_DecimalStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public "java/time/format/DateTimeFormatter/getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public "withChronology(Ljava/time/chrono/Chronology;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_chrono_Chronology], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "java/time/format/DateTimeFormatter/withChronology(Ljava/time/chrono/Chronology;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_chrono_Chronology], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "getZone()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "java/time/format/DateTimeFormatter/getZone()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "withZone(Ljava/time/ZoneId;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "java/time/format/DateTimeFormatter/withZone(Ljava/time/ZoneId;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "getResolverStyle()Ljava/time/format/ResolverStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_ResolverStyle) => void): void;
    public "java/time/format/DateTimeFormatter/getResolverStyle()Ljava/time/format/ResolverStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_ResolverStyle) => void): void;
    public "withResolverStyle(Ljava/time/format/ResolverStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "java/time/format/DateTimeFormatter/withResolverStyle(Ljava/time/format/ResolverStyle;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "getResolverFields()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/time/format/DateTimeFormatter/getResolverFields()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "withResolverFields([Ljava/time/temporal/TemporalField;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [JVMArray<java_time_temporal_TemporalField>], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "java/time/format/DateTimeFormatter/withResolverFields([Ljava/time/temporal/TemporalField;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [JVMArray<java_time_temporal_TemporalField>], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "withResolverFields(Ljava/util/Set;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "java/time/format/DateTimeFormatter/withResolverFields(Ljava/util/Set;)Ljava/time/format/DateTimeFormatter;"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatter) => void): void;
    public "format(Ljava/time/temporal/TemporalAccessor;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/format/DateTimeFormatter/format(Ljava/time/temporal/TemporalAccessor;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "formatTo(Ljava/time/temporal/TemporalAccessor;Ljava/lang/Appendable;)V"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor, java_lang_Appendable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeFormatter/formatTo(Ljava/time/temporal/TemporalAccessor;Ljava/lang/Appendable;)V"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor, java_lang_Appendable], cb?: (e?: java_lang_Throwable) => void): void;
    public "parse(Ljava/lang/CharSequence;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "java/time/format/DateTimeFormatter/parse(Ljava/lang/CharSequence;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "parse(Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "java/time/format/DateTimeFormatter/parse(Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "parse(Ljava/lang/CharSequence;Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/format/DateTimeFormatter/parse(Ljava/lang/CharSequence;Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "parseBest(Ljava/lang/CharSequence;[Ljava/time/temporal/TemporalQuery;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence, JVMArray<java_time_temporal_TemporalQuery>], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "java/time/format/DateTimeFormatter/parseBest(Ljava/lang/CharSequence;[Ljava/time/temporal/TemporalQuery;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence, JVMArray<java_time_temporal_TemporalQuery>], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "createError(Ljava/lang/CharSequence;Ljava/lang/RuntimeException;)Ljava/time/format/DateTimeParseException;"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_RuntimeException], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeParseException) => void): void;
    public "java/time/format/DateTimeFormatter/createError(Ljava/lang/CharSequence;Ljava/lang/RuntimeException;)Ljava/time/format/DateTimeParseException;"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_RuntimeException], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeParseException) => void): void;
    public "parseResolved0(Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "java/time/format/DateTimeFormatter/parseResolved0(Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "parseUnresolved(Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "java/time/format/DateTimeFormatter/parseUnresolved(Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "parseUnresolved0(Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/format/DateTimeParseContext;"(thread: JVMThread, args: [java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeParseContext) => void): void;
    public "java/time/format/DateTimeFormatter/parseUnresolved0(Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/format/DateTimeParseContext;"(thread: JVMThread, args: [java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeParseContext) => void): void;
    public "toPrinterParser(Z)Ljava/time/format/DateTimeFormatterBuilder$CompositePrinterParser;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatterBuilder$CompositePrinterParser) => void): void;
    public "java/time/format/DateTimeFormatter/toPrinterParser(Z)Ljava/time/format/DateTimeFormatterBuilder$CompositePrinterParser;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatterBuilder$CompositePrinterParser) => void): void;
    public "toFormat()Ljava/text/Format;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_text_Format) => void): void;
    public "java/time/format/DateTimeFormatter/toFormat()Ljava/text/Format;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_text_Format) => void): void;
    public "toFormat(Ljava/time/temporal/TemporalQuery;)Ljava/text/Format;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_text_Format) => void): void;
    public "java/time/format/DateTimeFormatter/toFormat(Ljava/time/temporal/TemporalQuery;)Ljava/text/Format;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_text_Format) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/format/DateTimeFormatter/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "lambda$static$1(Ljava/time/temporal/TemporalAccessor;)Ljava/lang/Boolean;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Boolean) => void): void;
    public static "java/time/format/DateTimeFormatter/lambda$static$1(Ljava/time/temporal/TemporalAccessor;)Ljava/lang/Boolean;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Boolean) => void): void;
    public static "lambda$static$0(Ljava/time/temporal/TemporalAccessor;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "java/time/format/DateTimeFormatter/lambda$static$0(Ljava/time/temporal/TemporalAccessor;)Ljava/time/Period;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_Period) => void): void;
    public static "access$000(Ljava/time/format/DateTimeFormatter;Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter, java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public static "java/time/format/DateTimeFormatter/access$000(Ljava/time/format/DateTimeFormatter;Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter, java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public static "access$100(Ljava/time/format/DateTimeFormatter;Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/format/DateTimeParseContext;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter, java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeParseContext) => void): void;
    public static "java/time/format/DateTimeFormatter/access$100(Ljava/time/format/DateTimeFormatter;Ljava/lang/CharSequence;Ljava/text/ParsePosition;)Ljava/time/format/DateTimeParseContext;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter, java_lang_CharSequence, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeParseContext) => void): void;
    public static "access$200(Ljava/time/format/DateTimeFormatter;)Ljava/time/format/ResolverStyle;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_format_ResolverStyle) => void): void;
    public static "java/time/format/DateTimeFormatter/access$200(Ljava/time/format/DateTimeFormatter;)Ljava/time/format/ResolverStyle;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_format_ResolverStyle) => void): void;
    public static "access$300(Ljava/time/format/DateTimeFormatter;)Ljava/util/Set;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "java/time/format/DateTimeFormatter/access$300(Ljava/time/format/DateTimeFormatter;)Ljava/util/Set;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/format/DateTimeFormatter/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_text_Format extends java_lang_Object implements java_io_Serializable, java_lang_Cloneable {
    public static "java/text/Format/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/Format/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "format(Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/text/Format/format(Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "format(Ljava/lang/Object;Ljava/lang/StringBuffer;Ljava/text/FieldPosition;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_Object, java_lang_StringBuffer, java_text_FieldPosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/text/Format/format(Ljava/lang/Object;Ljava/lang/StringBuffer;Ljava/text/FieldPosition;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_Object, java_lang_StringBuffer, java_text_FieldPosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "formatToCharacterIterator(Ljava/lang/Object;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "java/text/Format/formatToCharacterIterator(Ljava/lang/Object;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "parseObject(Ljava/lang/String;Ljava/text/ParsePosition;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/text/Format/parseObject(Ljava/lang/String;Ljava/text/ParsePosition;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "parseObject(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/text/Format/parseObject(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/text/Format/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "createAttributedCharacterIterator(Ljava/lang/String;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "java/text/Format/createAttributedCharacterIterator(Ljava/lang/String;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "createAttributedCharacterIterator([Ljava/text/AttributedCharacterIterator;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [JVMArray<java_text_AttributedCharacterIterator>], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "java/text/Format/createAttributedCharacterIterator([Ljava/text/AttributedCharacterIterator;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [JVMArray<java_text_AttributedCharacterIterator>], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "createAttributedCharacterIterator(Ljava/lang/String;Ljava/text/AttributedCharacterIterator$Attribute;Ljava/lang/Object;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_lang_String, java_text_AttributedCharacterIterator$Attribute, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "java/text/Format/createAttributedCharacterIterator(Ljava/lang/String;Ljava/text/AttributedCharacterIterator$Attribute;Ljava/lang/Object;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_lang_String, java_text_AttributedCharacterIterator$Attribute, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "createAttributedCharacterIterator(Ljava/text/AttributedCharacterIterator;Ljava/text/AttributedCharacterIterator$Attribute;Ljava/lang/Object;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, java_text_AttributedCharacterIterator$Attribute, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "java/text/Format/createAttributedCharacterIterator(Ljava/text/AttributedCharacterIterator;Ljava/text/AttributedCharacterIterator$Attribute;Ljava/lang/Object;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, java_text_AttributedCharacterIterator$Attribute, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
  }
  export class java_text_AttributedCharacterIterator$Attribute extends java_lang_Object implements java_io_Serializable {
    public "java/text/AttributedCharacterIterator$Attribute/name": java_lang_String;
    public static "java/text/AttributedCharacterIterator$Attribute/instanceMap": java_util_Map;
    public static "java/text/AttributedCharacterIterator$Attribute/LANGUAGE": java_text_AttributedCharacterIterator$Attribute;
    public static "java/text/AttributedCharacterIterator$Attribute/READING": java_text_AttributedCharacterIterator$Attribute;
    public static "java/text/AttributedCharacterIterator$Attribute/INPUT_METHOD_SEGMENT": java_text_AttributedCharacterIterator$Attribute;
    public static "java/text/AttributedCharacterIterator$Attribute/serialVersionUID": Long;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/AttributedCharacterIterator$Attribute/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/AttributedCharacterIterator$Attribute/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/AttributedCharacterIterator$Attribute/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/text/AttributedCharacterIterator$Attribute/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/text/AttributedCharacterIterator$Attribute/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/text/AttributedCharacterIterator$Attribute/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/text/AttributedCharacterIterator$Attribute/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_text_AttributedCharacterIterator extends java_lang_Object, java_text_CharacterIterator {
    "getRunStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getRunStart(Ljava/text/AttributedCharacterIterator$Attribute;)I"(thread: JVMThread, args: [java_text_AttributedCharacterIterator$Attribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getRunStart(Ljava/util/Set;)I"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getRunLimit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getRunLimit(Ljava/text/AttributedCharacterIterator$Attribute;)I"(thread: JVMThread, args: [java_text_AttributedCharacterIterator$Attribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getRunLimit(Ljava/util/Set;)I"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAttributes()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    "getAttribute(Ljava/text/AttributedCharacterIterator$Attribute;)Ljava/lang/Object;"(thread: JVMThread, args: [java_text_AttributedCharacterIterator$Attribute], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getAllAttributeKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "first()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "last()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "current()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "next()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "previous()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setIndex(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getBeginIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getEndIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_text_CharacterIterator extends java_lang_Object, java_lang_Cloneable {
    "first()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "last()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "current()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "next()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "previous()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setIndex(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getBeginIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getEndIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_text_FieldPosition extends java_lang_Object {
    public "java/text/FieldPosition/field": number;
    public "java/text/FieldPosition/endIndex": number;
    public "java/text/FieldPosition/beginIndex": number;
    public "java/text/FieldPosition/attribute": java_text_Format$Field;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/FieldPosition/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/text/Format$Field;)V"(thread: JVMThread, args: [java_text_Format$Field], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/FieldPosition/<init>(Ljava/text/Format$Field;)V"(thread: JVMThread, args: [java_text_Format$Field], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/text/Format$Field;I)V"(thread: JVMThread, args: [java_text_Format$Field, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/FieldPosition/<init>(Ljava/text/Format$Field;I)V"(thread: JVMThread, args: [java_text_Format$Field, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFieldAttribute()Ljava/text/Format$Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_text_Format$Field) => void): void;
    public "java/text/FieldPosition/getFieldAttribute()Ljava/text/Format$Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_text_Format$Field) => void): void;
    public "getField()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/FieldPosition/getField()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBeginIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/FieldPosition/getBeginIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getEndIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/FieldPosition/getEndIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setBeginIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/FieldPosition/setBeginIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setEndIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/FieldPosition/setEndIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFieldDelegate()Ljava/text/Format$FieldDelegate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_text_Format$FieldDelegate) => void): void;
    public "java/text/FieldPosition/getFieldDelegate()Ljava/text/Format$FieldDelegate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_text_Format$FieldDelegate) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/FieldPosition/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/FieldPosition/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/text/FieldPosition/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "matchesField(Ljava/text/Format$Field;)Z"(thread: JVMThread, args: [java_text_Format$Field], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/FieldPosition/matchesField(Ljava/text/Format$Field;)Z"(thread: JVMThread, args: [java_text_Format$Field], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "matchesField(Ljava/text/Format$Field;I)Z"(thread: JVMThread, args: [java_text_Format$Field, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/FieldPosition/matchesField(Ljava/text/Format$Field;I)Z"(thread: JVMThread, args: [java_text_Format$Field, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Ljava/text/FieldPosition;Ljava/text/Format$Field;)Z"(thread: JVMThread, args: [java_text_FieldPosition, java_text_Format$Field], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/text/FieldPosition/access$100(Ljava/text/FieldPosition;Ljava/text/Format$Field;)Z"(thread: JVMThread, args: [java_text_FieldPosition, java_text_Format$Field], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Ljava/text/FieldPosition;Ljava/text/Format$Field;I)Z"(thread: JVMThread, args: [java_text_FieldPosition, java_text_Format$Field, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/text/FieldPosition/access$200(Ljava/text/FieldPosition;Ljava/text/Format$Field;I)Z"(thread: JVMThread, args: [java_text_FieldPosition, java_text_Format$Field, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_text_Format$FieldDelegate extends java_lang_Object {
    "formatted(Ljava/text/Format$Field;Ljava/lang/Object;IILjava/lang/StringBuffer;)V"(thread: JVMThread, args: [java_text_Format$Field, java_lang_Object, number, number, java_lang_StringBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    "formatted(ILjava/text/Format$Field;Ljava/lang/Object;IILjava/lang/StringBuffer;)V"(thread: JVMThread, args: [number, java_text_Format$Field, java_lang_Object, number, number, java_lang_StringBuffer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_text_Format$Field extends java_text_AttributedCharacterIterator$Attribute {
    public static "java/text/Format$Field/serialVersionUID": Long;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/Format$Field/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_format_DateTimeParseContext extends java_lang_Object {
    public "java/time/format/DateTimeParseContext/formatter": java_time_format_DateTimeFormatter;
    public "java/time/format/DateTimeParseContext/caseSensitive": number;
    public "java/time/format/DateTimeParseContext/strict": number;
    public "java/time/format/DateTimeParseContext/parsed": java_util_ArrayList;
    public "java/time/format/DateTimeParseContext/chronoListeners": java_util_ArrayList;
    public "<init>(Ljava/time/format/DateTimeFormatter;)V"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseContext/<init>(Ljava/time/format/DateTimeFormatter;)V"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable) => void): void;
    public "copy()Ljava/time/format/DateTimeParseContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeParseContext) => void): void;
    public "java/time/format/DateTimeParseContext/copy()Ljava/time/format/DateTimeParseContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeParseContext) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/time/format/DateTimeParseContext/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "getDecimalStyle()Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "java/time/format/DateTimeParseContext/getDecimalStyle()Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "getEffectiveChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public "java/time/format/DateTimeParseContext/getEffectiveChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    public "isCaseSensitive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DateTimeParseContext/isCaseSensitive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setCaseSensitive(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseContext/setCaseSensitive(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "subSequenceEquals(Ljava/lang/CharSequence;ILjava/lang/CharSequence;II)Z"(thread: JVMThread, args: [java_lang_CharSequence, number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DateTimeParseContext/subSequenceEquals(Ljava/lang/CharSequence;ILjava/lang/CharSequence;II)Z"(thread: JVMThread, args: [java_lang_CharSequence, number, java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charEquals(CC)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DateTimeParseContext/charEquals(CC)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "charEqualsIgnoreCase(CC)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/format/DateTimeParseContext/charEqualsIgnoreCase(CC)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isStrict()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DateTimeParseContext/isStrict()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setStrict(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseContext/setStrict(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "startOptional()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseContext/startOptional()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "endOptional(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseContext/endOptional(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "currentParsed()Ljava/time/format/Parsed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_Parsed) => void): void;
    public "java/time/format/DateTimeParseContext/currentParsed()Ljava/time/format/Parsed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_Parsed) => void): void;
    public "toUnresolved()Ljava/time/format/Parsed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_Parsed) => void): void;
    public "java/time/format/DateTimeParseContext/toUnresolved()Ljava/time/format/Parsed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_Parsed) => void): void;
    public "toResolved(Ljava/time/format/ResolverStyle;Ljava/util/Set;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_time_format_ResolverStyle, java_util_Set], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "java/time/format/DateTimeParseContext/toResolved(Ljava/time/format/ResolverStyle;Ljava/util/Set;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_time_format_ResolverStyle, java_util_Set], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "getParsed(Ljava/time/temporal/TemporalField;)Ljava/lang/Long;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public "java/time/format/DateTimeParseContext/getParsed(Ljava/time/temporal/TemporalField;)Ljava/lang/Long;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public "setParsedField(Ljava/time/temporal/TemporalField;JII)I"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DateTimeParseContext/setParsedField(Ljava/time/temporal/TemporalField;JII)I"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setParsed(Ljava/time/chrono/Chronology;)V"(thread: JVMThread, args: [java_time_chrono_Chronology], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseContext/setParsed(Ljava/time/chrono/Chronology;)V"(thread: JVMThread, args: [java_time_chrono_Chronology], cb?: (e?: java_lang_Throwable) => void): void;
    public "addChronoChangedListener(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseContext/addChronoChangedListener(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setParsed(Ljava/time/ZoneId;)V"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseContext/setParsed(Ljava/time/ZoneId;)V"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable) => void): void;
    public "setParsedLeapSecond()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseContext/setParsedLeapSecond()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/format/DateTimeParseContext/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_time_format_Parsed extends java_lang_Object implements java_time_temporal_TemporalAccessor {
    public "java/time/format/Parsed/fieldValues": java_util_Map;
    public "java/time/format/Parsed/zone": java_time_ZoneId;
    public "java/time/format/Parsed/chrono": java_time_chrono_Chronology;
    public "java/time/format/Parsed/leapSecond": number;
    public "java/time/format/Parsed/resolverStyle": java_time_format_ResolverStyle;
    public "java/time/format/Parsed/date": java_time_chrono_ChronoLocalDate;
    public "java/time/format/Parsed/time": java_time_LocalTime;
    public "java/time/format/Parsed/excessDays": java_time_Period;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "copy()Ljava/time/format/Parsed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_Parsed) => void): void;
    public "java/time/format/Parsed/copy()Ljava/time/format/Parsed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_Parsed) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/Parsed/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/format/Parsed/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/format/Parsed/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "resolve(Ljava/time/format/ResolverStyle;Ljava/util/Set;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_time_format_ResolverStyle, java_util_Set], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "java/time/format/Parsed/resolve(Ljava/time/format/ResolverStyle;Ljava/util/Set;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_time_format_ResolverStyle, java_util_Set], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "resolveFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolveFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateCheckConflict(Ljava/time/temporal/TemporalField;Ljava/time/temporal/TemporalField;Ljava/lang/Long;)V"(thread: JVMThread, args: [java_time_temporal_TemporalField, java_time_temporal_TemporalField, java_lang_Long], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/updateCheckConflict(Ljava/time/temporal/TemporalField;Ljava/time/temporal/TemporalField;Ljava/lang/Long;)V"(thread: JVMThread, args: [java_time_temporal_TemporalField, java_time_temporal_TemporalField, java_lang_Long], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveInstantFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolveInstantFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveInstantFields0(Ljava/time/ZoneId;)V"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolveInstantFields0(Ljava/time/ZoneId;)V"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveDateFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolveDateFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateCheckConflict(Ljava/time/chrono/ChronoLocalDate;)V"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/updateCheckConflict(Ljava/time/chrono/ChronoLocalDate;)V"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDate], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveTimeFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolveTimeFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveTimeLenient()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolveTimeLenient()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveTime(JJJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolveTime(JJJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolvePeriod()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolvePeriod()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveFractional()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolveFractional()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveInstant()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/resolveInstant()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateCheckConflict(Ljava/time/LocalTime;Ljava/time/Period;)V"(thread: JVMThread, args: [java_time_LocalTime, java_time_Period], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/updateCheckConflict(Ljava/time/LocalTime;Ljava/time/Period;)V"(thread: JVMThread, args: [java_time_LocalTime, java_time_Period], cb?: (e?: java_lang_Throwable) => void): void;
    public "crossCheck()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/crossCheck()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "crossCheck(Ljava/time/temporal/TemporalAccessor;)V"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/Parsed/crossCheck(Ljava/time/temporal/TemporalAccessor;)V"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/format/Parsed/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_ArrayList extends java_util_AbstractList implements java_util_List, java_util_RandomAccess, java_lang_Cloneable, java_io_Serializable {
    public static "java/util/ArrayList/serialVersionUID": Long;
    public static "java/util/ArrayList/DEFAULT_CAPACITY": number;
    public static "java/util/ArrayList/EMPTY_ELEMENTDATA": JVMArray<java_lang_Object>;
    public static "java/util/ArrayList/DEFAULTCAPACITY_EMPTY_ELEMENTDATA": JVMArray<java_lang_Object>;
    public "java/util/ArrayList/elementData": JVMArray<java_lang_Object>;
    public "java/util/ArrayList/size": number;
    public static "java/util/ArrayList/MAX_ARRAY_SIZE": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureCapacityInternal(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/ensureCapacityInternal(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureExplicitCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/ensureExplicitCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/ArrayList/hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ArrayList/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/ArrayList/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/ArrayList/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "elementData(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ArrayList/elementData(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ArrayList/get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ArrayList/set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ArrayList/remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fastRemove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/fastRemove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rangeCheck(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/rangeCheck(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rangeCheckForAdd(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/rangeCheckForAdd(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "outOfBoundsMsg(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ArrayList/outOfBoundsMsg(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "batchRemove(Ljava/util/Collection;Z)Z"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/batchRemove(Ljava/util/Collection;Z)Z"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/ArrayList/listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "listIterator()Ljava/util/ListIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/ArrayList/listIterator()Ljava/util/ListIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/ArrayList/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "subList(II)Ljava/util/List;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/ArrayList/subList(II)Ljava/util/List;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "subListRangeCheck(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/ArrayList/subListRangeCheck(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/ArrayList/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ArrayList/removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    public "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ArrayList/sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$100(Ljava/util/ArrayList;)I"(thread: JVMThread, args: [java_util_ArrayList], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/ArrayList/access$100(Ljava/util/ArrayList;)I"(thread: JVMThread, args: [java_util_ArrayList], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/ArrayList/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_RandomAccess extends java_lang_Object {
  }
  export class java_util_AbstractList extends java_util_AbstractCollection implements java_util_List {
    public "java/util/AbstractList/modCount": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractList/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractList/get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractList/set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractList/add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractList/remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractList/indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractList/lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractList/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractList/addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/AbstractList/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "listIterator()Ljava/util/ListIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/AbstractList/listIterator()Ljava/util/ListIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/AbstractList/listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "subList(II)Ljava/util/List;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/AbstractList/subList(II)Ljava/util/List;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractList/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractList/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractList/removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rangeCheckForAdd(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractList/rangeCheckForAdd(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "outOfBoundsMsg(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/AbstractList/outOfBoundsMsg(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_format_DateTimeParseException extends java_time_DateTimeException {
    public static "java/time/format/DateTimeParseException/serialVersionUID": Long;
    public "java/time/format/DateTimeParseException/parsedString": java_lang_String;
    public "java/time/format/DateTimeParseException/errorIndex": number;
    public "<init>(Ljava/lang/String;Ljava/lang/CharSequence;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseException/<init>(Ljava/lang/String;Ljava/lang/CharSequence;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/CharSequence;ILjava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_CharSequence, number, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeParseException/<init>(Ljava/lang/String;Ljava/lang/CharSequence;ILjava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_CharSequence, number, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "getParsedString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/format/DateTimeParseException/getParsedString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getErrorIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DateTimeParseException/getErrorIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_time_DateTimeException extends java_lang_RuntimeException {
    public static "java/time/DateTimeException/serialVersionUID": Long;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/DateTimeException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/DateTimeException/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_text_ParsePosition extends java_lang_Object {
    public "java/text/ParsePosition/index": number;
    public "java/text/ParsePosition/errorIndex": number;
    public "getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/ParsePosition/getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/ParsePosition/setIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/ParsePosition/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setErrorIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/ParsePosition/setErrorIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getErrorIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/ParsePosition/getErrorIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/ParsePosition/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/ParsePosition/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/text/ParsePosition/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_time_format_FormatStyle extends java_lang_Enum {
    public static "java/time/format/FormatStyle/FULL": java_time_format_FormatStyle;
    public static "java/time/format/FormatStyle/LONG": java_time_format_FormatStyle;
    public static "java/time/format/FormatStyle/MEDIUM": java_time_format_FormatStyle;
    public static "java/time/format/FormatStyle/SHORT": java_time_format_FormatStyle;
    public static "java/time/format/FormatStyle/$VALUES": JVMArray<java_time_format_FormatStyle>;
    public static "values()[Ljava/time/format/FormatStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_format_FormatStyle>) => void): void;
    public static "java/time/format/FormatStyle/values()[Ljava/time/format/FormatStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_format_FormatStyle>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/time/format/FormatStyle;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_format_FormatStyle) => void): void;
    public static "java/time/format/FormatStyle/valueOf(Ljava/lang/String;)Ljava/time/format/FormatStyle;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_format_FormatStyle) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/FormatStyle/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/format/FormatStyle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_format_DecimalStyle extends java_lang_Object {
    public static "java/time/format/DecimalStyle/STANDARD": java_time_format_DecimalStyle;
    public static "java/time/format/DecimalStyle/CACHE": java_util_concurrent_ConcurrentMap;
    public "java/time/format/DecimalStyle/zeroDigit": number;
    public "java/time/format/DecimalStyle/positiveSign": number;
    public "java/time/format/DecimalStyle/negativeSign": number;
    public "java/time/format/DecimalStyle/decimalSeparator": number;
    public static "getAvailableLocales()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "java/time/format/DecimalStyle/getAvailableLocales()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "ofDefaultLocale()Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public static "java/time/format/DecimalStyle/ofDefaultLocale()Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public static "of(Ljava/util/Locale;)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public static "java/time/format/DecimalStyle/of(Ljava/util/Locale;)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public static "create(Ljava/util/Locale;)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public static "java/time/format/DecimalStyle/create(Ljava/util/Locale;)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "<init>(CCCC)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DecimalStyle/<init>(CCCC)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getZeroDigit()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DecimalStyle/getZeroDigit()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "withZeroDigit(C)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "java/time/format/DecimalStyle/withZeroDigit(C)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "getPositiveSign()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DecimalStyle/getPositiveSign()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "withPositiveSign(C)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "java/time/format/DecimalStyle/withPositiveSign(C)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "getNegativeSign()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DecimalStyle/getNegativeSign()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "withNegativeSign(C)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "java/time/format/DecimalStyle/withNegativeSign(C)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "getDecimalSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DecimalStyle/getDecimalSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "withDecimalSeparator(C)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "java/time/format/DecimalStyle/withDecimalSeparator(C)Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "convertToDigit(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DecimalStyle/convertToDigit(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "convertNumberToI18N(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/format/DecimalStyle/convertNumberToI18N(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DecimalStyle/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DecimalStyle/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/format/DecimalStyle/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/format/DecimalStyle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_format_DateTimeFormatterBuilder$CompositePrinterParser extends java_lang_Object implements java_time_format_DateTimeFormatterBuilder$DateTimePrinterParser {
    public "java/time/format/DateTimeFormatterBuilder$CompositePrinterParser/printerParsers": JVMArray<java_time_format_DateTimeFormatterBuilder$DateTimePrinterParser>;
    public "java/time/format/DateTimeFormatterBuilder$CompositePrinterParser/optional": number;
    public "<init>(Ljava/util/List;Z)V"(thread: JVMThread, args: [java_util_List, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeFormatterBuilder$CompositePrinterParser/<init>(Ljava/util/List;Z)V"(thread: JVMThread, args: [java_util_List, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Ljava/time/format/DateTimeFormatterBuilder$DateTimePrinterParser;Z)V"(thread: JVMThread, args: [JVMArray<java_time_format_DateTimeFormatterBuilder$DateTimePrinterParser>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimeFormatterBuilder$CompositePrinterParser/<init>([Ljava/time/format/DateTimeFormatterBuilder$DateTimePrinterParser;Z)V"(thread: JVMThread, args: [JVMArray<java_time_format_DateTimeFormatterBuilder$DateTimePrinterParser>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "withOptional(Z)Ljava/time/format/DateTimeFormatterBuilder$CompositePrinterParser;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatterBuilder$CompositePrinterParser) => void): void;
    public "java/time/format/DateTimeFormatterBuilder$CompositePrinterParser/withOptional(Z)Ljava/time/format/DateTimeFormatterBuilder$CompositePrinterParser;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DateTimeFormatterBuilder$CompositePrinterParser) => void): void;
    public "format(Ljava/time/format/DateTimePrintContext;Ljava/lang/StringBuilder;)Z"(thread: JVMThread, args: [java_time_format_DateTimePrintContext, java_lang_StringBuilder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DateTimeFormatterBuilder$CompositePrinterParser/format(Ljava/time/format/DateTimePrintContext;Ljava/lang/StringBuilder;)Z"(thread: JVMThread, args: [java_time_format_DateTimePrintContext, java_lang_StringBuilder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parse(Ljava/time/format/DateTimeParseContext;Ljava/lang/CharSequence;I)I"(thread: JVMThread, args: [java_time_format_DateTimeParseContext, java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/DateTimeFormatterBuilder$CompositePrinterParser/parse(Ljava/time/format/DateTimeParseContext;Ljava/lang/CharSequence;I)I"(thread: JVMThread, args: [java_time_format_DateTimeParseContext, java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/format/DateTimeFormatterBuilder$CompositePrinterParser/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_time_format_DateTimePrintContext extends java_lang_Object {
    public "java/time/format/DateTimePrintContext/temporal": java_time_temporal_TemporalAccessor;
    public "java/time/format/DateTimePrintContext/formatter": java_time_format_DateTimeFormatter;
    public "java/time/format/DateTimePrintContext/optional": number;
    public "<init>(Ljava/time/temporal/TemporalAccessor;Ljava/time/format/DateTimeFormatter;)V"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimePrintContext/<init>(Ljava/time/temporal/TemporalAccessor;Ljava/time/format/DateTimeFormatter;)V"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable) => void): void;
    public static "adjust(Ljava/time/temporal/TemporalAccessor;Ljava/time/format/DateTimeFormatter;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public static "java/time/format/DateTimePrintContext/adjust(Ljava/time/temporal/TemporalAccessor;Ljava/time/format/DateTimeFormatter;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "getTemporal()Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "java/time/format/DateTimePrintContext/getTemporal()Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/time/format/DateTimePrintContext/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "getDecimalStyle()Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "java/time/format/DateTimePrintContext/getDecimalStyle()Ljava/time/format/DecimalStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_DecimalStyle) => void): void;
    public "startOptional()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimePrintContext/startOptional()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "endOptional()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/DateTimePrintContext/endOptional()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValue(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/format/DateTimePrintContext/getValue(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getValue(Ljava/time/temporal/TemporalField;)Ljava/lang/Long;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public "java/time/format/DateTimePrintContext/getValue(Ljava/time/temporal/TemporalField;)Ljava/lang/Long;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/format/DateTimePrintContext/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_time_format_DateTimeFormatterBuilder$DateTimePrinterParser extends java_lang_Object {
    "format(Ljava/time/format/DateTimePrintContext;Ljava/lang/StringBuilder;)Z"(thread: JVMThread, args: [java_time_format_DateTimePrintContext, java_lang_StringBuilder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "parse(Ljava/time/format/DateTimeParseContext;Ljava/lang/CharSequence;I)I"(thread: JVMThread, args: [java_time_format_DateTimeParseContext, java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_time_DayOfWeek extends java_lang_Enum implements java_time_temporal_TemporalAccessor, java_time_temporal_TemporalAdjuster {
    public static "java/time/DayOfWeek/MONDAY": java_time_DayOfWeek;
    public static "java/time/DayOfWeek/TUESDAY": java_time_DayOfWeek;
    public static "java/time/DayOfWeek/WEDNESDAY": java_time_DayOfWeek;
    public static "java/time/DayOfWeek/THURSDAY": java_time_DayOfWeek;
    public static "java/time/DayOfWeek/FRIDAY": java_time_DayOfWeek;
    public static "java/time/DayOfWeek/SATURDAY": java_time_DayOfWeek;
    public static "java/time/DayOfWeek/SUNDAY": java_time_DayOfWeek;
    public static "java/time/DayOfWeek/ENUMS": JVMArray<java_time_DayOfWeek>;
    public static "java/time/DayOfWeek/$VALUES": JVMArray<java_time_DayOfWeek>;
    public static "values()[Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_DayOfWeek>) => void): void;
    public static "java/time/DayOfWeek/values()[Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_DayOfWeek>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public static "java/time/DayOfWeek/valueOf(Ljava/lang/String;)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/DayOfWeek/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "of(I)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public static "java/time/DayOfWeek/of(I)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public static "java/time/DayOfWeek/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/DayOfWeek/getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/DayOfWeek/getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/DayOfWeek/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/DayOfWeek/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/DayOfWeek/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/DayOfWeek/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "plus(J)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "java/time/DayOfWeek/plus(J)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "minus(J)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "java/time/DayOfWeek/minus(J)Ljava/time/DayOfWeek;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/DayOfWeek/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/DayOfWeek/adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/DayOfWeek/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_time_Month extends java_lang_Enum implements java_time_temporal_TemporalAccessor, java_time_temporal_TemporalAdjuster {
    public static "java/time/Month/JANUARY": java_time_Month;
    public static "java/time/Month/FEBRUARY": java_time_Month;
    public static "java/time/Month/MARCH": java_time_Month;
    public static "java/time/Month/APRIL": java_time_Month;
    public static "java/time/Month/MAY": java_time_Month;
    public static "java/time/Month/JUNE": java_time_Month;
    public static "java/time/Month/JULY": java_time_Month;
    public static "java/time/Month/AUGUST": java_time_Month;
    public static "java/time/Month/SEPTEMBER": java_time_Month;
    public static "java/time/Month/OCTOBER": java_time_Month;
    public static "java/time/Month/NOVEMBER": java_time_Month;
    public static "java/time/Month/DECEMBER": java_time_Month;
    public static "java/time/Month/ENUMS": JVMArray<java_time_Month>;
    public static "java/time/Month/$VALUES": JVMArray<java_time_Month>;
    public static "values()[Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_Month>) => void): void;
    public static "java/time/Month/values()[Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_Month>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/time/Month;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public static "java/time/Month/valueOf(Ljava/lang/String;)Ljava/time/Month;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Month/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "of(I)Ljava/time/Month;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public static "java/time/Month/of(I)Ljava/time/Month;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/Month;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public static "java/time/Month/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/Month;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Month/getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/Month/getDisplayName(Ljava/time/format/TextStyle;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_TextStyle, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Month/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/Month/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Month/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Month/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "plus(J)Ljava/time/Month;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "java/time/Month/plus(J)Ljava/time/Month;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "minus(J)Ljava/time/Month;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "java/time/Month/minus(J)Ljava/time/Month;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "length(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Month/length(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "minLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Month/minLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "maxLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Month/maxLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "firstDayOfYear(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Month/firstDayOfYear(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "firstMonthOfQuarter()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "java/time/Month/firstMonthOfQuarter()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/Month/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/Month/adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/Month/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_time_LocalDateTime extends java_lang_Object implements java_time_temporal_Temporal, java_time_temporal_TemporalAdjuster, java_time_chrono_ChronoLocalDateTime, java_io_Serializable {
    public static "java/time/LocalDateTime/MIN": java_time_LocalDateTime;
    public static "java/time/LocalDateTime/MAX": java_time_LocalDateTime;
    public static "java/time/LocalDateTime/serialVersionUID": Long;
    public "java/time/LocalDateTime/date": java_time_LocalDate;
    public "java/time/LocalDateTime/time": java_time_LocalTime;
    public static "now()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/now()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "now(Ljava/time/ZoneId;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/now(Ljava/time/ZoneId;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "now(Ljava/time/Clock;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/now(Ljava/time/Clock;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "of(ILjava/time/Month;III)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, java_time_Month, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/of(ILjava/time/Month;III)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, java_time_Month, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "of(ILjava/time/Month;IIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, java_time_Month, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/of(ILjava/time/Month;IIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, java_time_Month, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "of(ILjava/time/Month;IIIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, java_time_Month, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/of(ILjava/time/Month;IIIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, java_time_Month, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "of(IIIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/of(IIIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "of(IIIIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/of(IIIIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "of(IIIIIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/of(IIIIIII)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "of(Ljava/time/LocalDate;Ljava/time/LocalTime;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/of(Ljava/time/LocalDate;Ljava/time/LocalTime;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "ofInstant(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/ofInstant(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "ofEpochSecond(JILjava/time/ZoneOffset;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any, number, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/ofEpochSecond(JILjava/time/ZoneOffset;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any, number, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/parse(Ljava/lang/CharSequence;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "<init>(Ljava/time/LocalDate;Ljava/time/LocalTime;)V"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/LocalDateTime/<init>(Ljava/time/LocalDate;Ljava/time/LocalTime;)V"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime], cb?: (e?: java_lang_Throwable) => void): void;
    public "with(Ljava/time/LocalDate;Ljava/time/LocalTime;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/with(Ljava/time/LocalDate;Ljava/time/LocalTime;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/LocalDateTime/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalDateTime/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toLocalDate()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/LocalDateTime/toLocalDate()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonthValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/getMonthValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "java/time/LocalDateTime/getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/getDayOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "java/time/LocalDateTime/getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/LocalDateTime/toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "withYear(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/withYear(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "withMonth(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/withMonth(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "withDayOfMonth(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/withDayOfMonth(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "withDayOfYear(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/withDayOfYear(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "withHour(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/withHour(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "withMinute(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/withMinute(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "withSecond(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/withSecond(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "withNano(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/withNano(I)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plusYears(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plusYears(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plusMonths(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plusMonths(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plusWeeks(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plusWeeks(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plusDays(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plusDays(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plusHours(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plusHours(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plusMinutes(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plusMinutes(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plusSeconds(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plusSeconds(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plusNanos(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plusNanos(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minusYears(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minusYears(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minusMonths(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minusMonths(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minusWeeks(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minusWeeks(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minusDays(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minusDays(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minusHours(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minusHours(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minusMinutes(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minusMinutes(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minusSeconds(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minusSeconds(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minusNanos(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/minusNanos(J)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "plusWithOverflow(Ljava/time/LocalDate;JJJJI)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDate, Long, any, Long, any, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/LocalDateTime/plusWithOverflow(Ljava/time/LocalDate;JJJJI)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_time_LocalDate, Long, any, Long, any, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/LocalDateTime/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDateTime/adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/LocalDateTime/until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/LocalDateTime/format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "atOffset(Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/LocalDateTime/atOffset(Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "atZone(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/LocalDateTime/atZone(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "compareTo(Ljava/time/chrono/ChronoLocalDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/compareTo(Ljava/time/chrono/ChronoLocalDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo0(Ljava/time/LocalDateTime;)I"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/compareTo0(Ljava/time/LocalDateTime;)I"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAfter(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/isAfter(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBefore(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/isBefore(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEqual(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/isEqual(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/LocalDateTime/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/LocalDateTime/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/LocalDateTime/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/LocalDateTime/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public static "java/time/LocalDateTime/readExternal(Ljava/io/DataInput;)Ljava/time/LocalDateTime;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDateTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDateTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDateTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDateTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDateTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/LocalDateTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "atZone(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/LocalDateTime/atZone(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "java/time/LocalDateTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "java/time/LocalDateTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "java/time/LocalDateTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "java/time/LocalDateTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "java/time/LocalDateTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "java/time/LocalDateTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "toLocalDate()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/LocalDateTime/toLocalDate()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/LocalDateTime/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/LocalDateTime/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    "toInstant(Ljava/time/ZoneOffset;)Ljava/time/Instant;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    "toEpochSecond(Ljava/time/ZoneOffset;)J"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "compareTo(Ljava/time/chrono/ChronoLocalDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isAfter(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isBefore(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEqual(Ljava/time/chrono/ChronoLocalDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoLocalDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_Locale extends java_lang_Object implements java_lang_Cloneable, java_io_Serializable {
    public static "java/util/Locale/LOCALECACHE": java_util_Locale$Cache;
    public static "java/util/Locale/ENGLISH": java_util_Locale;
    public static "java/util/Locale/FRENCH": java_util_Locale;
    public static "java/util/Locale/GERMAN": java_util_Locale;
    public static "java/util/Locale/ITALIAN": java_util_Locale;
    public static "java/util/Locale/JAPANESE": java_util_Locale;
    public static "java/util/Locale/KOREAN": java_util_Locale;
    public static "java/util/Locale/CHINESE": java_util_Locale;
    public static "java/util/Locale/SIMPLIFIED_CHINESE": java_util_Locale;
    public static "java/util/Locale/TRADITIONAL_CHINESE": java_util_Locale;
    public static "java/util/Locale/FRANCE": java_util_Locale;
    public static "java/util/Locale/GERMANY": java_util_Locale;
    public static "java/util/Locale/ITALY": java_util_Locale;
    public static "java/util/Locale/JAPAN": java_util_Locale;
    public static "java/util/Locale/KOREA": java_util_Locale;
    public static "java/util/Locale/CHINA": java_util_Locale;
    public static "java/util/Locale/PRC": java_util_Locale;
    public static "java/util/Locale/TAIWAN": java_util_Locale;
    public static "java/util/Locale/UK": java_util_Locale;
    public static "java/util/Locale/US": java_util_Locale;
    public static "java/util/Locale/CANADA": java_util_Locale;
    public static "java/util/Locale/CANADA_FRENCH": java_util_Locale;
    public static "java/util/Locale/ROOT": java_util_Locale;
    public static "java/util/Locale/PRIVATE_USE_EXTENSION": number;
    public static "java/util/Locale/UNICODE_LOCALE_EXTENSION": number;
    public static "java/util/Locale/serialVersionUID": Long;
    public static "java/util/Locale/DISPLAY_LANGUAGE": number;
    public static "java/util/Locale/DISPLAY_COUNTRY": number;
    public static "java/util/Locale/DISPLAY_VARIANT": number;
    public static "java/util/Locale/DISPLAY_SCRIPT": number;
    public "java/util/Locale/baseLocale": sun_util_locale_BaseLocale;
    public "java/util/Locale/localeExtensions": sun_util_locale_LocaleExtensions;
    public "java/util/Locale/hashCodeValue": number;
    public static "java/util/Locale/defaultLocale": java_util_Locale;
    public static "java/util/Locale/defaultDisplayLocale": java_util_Locale;
    public static "java/util/Locale/defaultFormatLocale": java_util_Locale;
    public "java/util/Locale/languageTag": java_lang_String;
    public static "java/util/Locale/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/util/Locale/isoLanguages": JVMArray<java_lang_String>;
    public static "java/util/Locale/isoCountries": JVMArray<java_lang_String>;
    public static "java/util/Locale/$assertionsDisabled": number;
    public "<init>(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;)V"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale/<init>(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;)V"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createConstant(Ljava/lang/String;Ljava/lang/String;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/createConstant(Ljava/lang/String;Ljava/lang/String;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/getInstance(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Lsun/util/locale/LocaleExtensions;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, sun_util_locale_LocaleExtensions], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/getInstance(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Lsun/util/locale/LocaleExtensions;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, sun_util_locale_LocaleExtensions], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "getInstance(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;)Ljava/util/Locale;"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/getInstance(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;)Ljava/util/Locale;"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "getDefault()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/getDefault()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "getDefault(Ljava/util/Locale$Category;)Ljava/util/Locale;"(thread: JVMThread, args: [java_util_Locale$Category], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/getDefault(Ljava/util/Locale$Category;)Ljava/util/Locale;"(thread: JVMThread, args: [java_util_Locale$Category], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "initDefault()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/initDefault()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "initDefault(Ljava/util/Locale$Category;)Ljava/util/Locale;"(thread: JVMThread, args: [java_util_Locale$Category], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/initDefault(Ljava/util/Locale$Category;)Ljava/util/Locale;"(thread: JVMThread, args: [java_util_Locale$Category], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "setDefault(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Locale/setDefault(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setDefault(Ljava/util/Locale$Category;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale$Category, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Locale/setDefault(Ljava/util/Locale$Category;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale$Category, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getAvailableLocales()[Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Locale>) => void): void;
    public static "java/util/Locale/getAvailableLocales()[Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Locale>) => void): void;
    public static "getISOCountries()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/util/Locale/getISOCountries()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "getISOLanguages()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/util/Locale/getISOLanguages()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "getISO2Table(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/util/Locale/getISO2Table(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getLanguage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getLanguage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getScript()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getScript()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCountry()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getCountry()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVariant()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getVariant()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hasExtensions()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Locale/hasExtensions()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "stripExtensions()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/util/Locale/stripExtensions()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "getExtension(C)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getExtension(C)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getExtensionKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/Locale/getExtensionKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getUnicodeLocaleAttributes()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/Locale/getUnicodeLocaleAttributes()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getUnicodeLocaleType(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getUnicodeLocaleType(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getUnicodeLocaleKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/Locale/getUnicodeLocaleKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getBaseLocale()Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public "java/util/Locale/getBaseLocale()Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public "getLocaleExtensions()Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
    public "java/util/Locale/getLocaleExtensions()Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toLanguageTag()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/toLanguageTag()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "forLanguageTag(Ljava/lang/String;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/forLanguageTag(Ljava/lang/String;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "getISO3Language()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getISO3Language()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getISO3Country()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getISO3Country()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getISO3Code(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/Locale/getISO3Code(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayLanguage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayLanguage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayLanguage(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayLanguage(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayScript()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayScript()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayScript(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayScript(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayCountry()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayCountry()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayCountry(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayCountry(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayString(Ljava/lang/String;Ljava/util/Locale;I)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayString(Ljava/lang/String;Ljava/util/Locale;I)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayVariant()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayVariant()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayVariant(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayVariant(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Locale/getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Locale/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Locale/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Locale/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDisplayVariantArray(Ljava/util/Locale;)[Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/Locale/getDisplayVariantArray(Ljava/util/Locale;)[Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "formatList([Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/Locale/formatList([Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "composeList(Ljava/text/MessageFormat;[Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_text_MessageFormat, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/util/Locale/composeList(Ljava/text/MessageFormat;[Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_text_MessageFormat, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "isUnicodeExtensionKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Locale/isUnicodeExtensionKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Locale/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "convertOldISOCodes(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/Locale/convertOldISOCodes(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getCompatibilityExtensions(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
    public static "java/util/Locale/getCompatibilityExtensions(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
    public static "filter(Ljava/util/List;Ljava/util/Collection;Ljava/util/Locale$FilteringMode;)Ljava/util/List;"(thread: JVMThread, args: [java_util_List, java_util_Collection, java_util_Locale$FilteringMode], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/util/Locale/filter(Ljava/util/List;Ljava/util/Collection;Ljava/util/Locale$FilteringMode;)Ljava/util/List;"(thread: JVMThread, args: [java_util_List, java_util_Collection, java_util_Locale$FilteringMode], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "filter(Ljava/util/List;Ljava/util/Collection;)Ljava/util/List;"(thread: JVMThread, args: [java_util_List, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/util/Locale/filter(Ljava/util/List;Ljava/util/Collection;)Ljava/util/List;"(thread: JVMThread, args: [java_util_List, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "filterTags(Ljava/util/List;Ljava/util/Collection;Ljava/util/Locale$FilteringMode;)Ljava/util/List;"(thread: JVMThread, args: [java_util_List, java_util_Collection, java_util_Locale$FilteringMode], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/util/Locale/filterTags(Ljava/util/List;Ljava/util/Collection;Ljava/util/Locale$FilteringMode;)Ljava/util/List;"(thread: JVMThread, args: [java_util_List, java_util_Collection, java_util_Locale$FilteringMode], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "filterTags(Ljava/util/List;Ljava/util/Collection;)Ljava/util/List;"(thread: JVMThread, args: [java_util_List, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/util/Locale/filterTags(Ljava/util/List;Ljava/util/Collection;)Ljava/util/List;"(thread: JVMThread, args: [java_util_List, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "lookup(Ljava/util/List;Ljava/util/Collection;)Ljava/util/Locale;"(thread: JVMThread, args: [java_util_List, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "java/util/Locale/lookup(Ljava/util/List;Ljava/util/Collection;)Ljava/util/Locale;"(thread: JVMThread, args: [java_util_List, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "lookupTag(Ljava/util/List;Ljava/util/Collection;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_List, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/Locale/lookupTag(Ljava/util/List;Ljava/util/Collection;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_List, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "<init>(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;Ljava/util/Locale$1;)V"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions, java_util_Locale$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale/<init>(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;Ljava/util/Locale$1;)V"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions, java_util_Locale$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$600(Ljava/util/Locale;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public static "java/util/Locale/access$600(Ljava/util/Locale;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public static "access$700(Ljava/util/Locale;)Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
    public static "java/util/Locale/access$700(Ljava/util/Locale;)Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
    public static "access$800(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
    public static "java/util/Locale/access$800(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Locale/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_Locale$1 extends java_lang_Object {
    public static "java/util/Locale$1/$SwitchMap$java$util$Locale$Category": JVMArray<number>;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Locale$1/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_Locale$FilteringMode extends java_lang_Enum {
    public static "java/util/Locale$FilteringMode/AUTOSELECT_FILTERING": java_util_Locale$FilteringMode;
    public static "java/util/Locale$FilteringMode/EXTENDED_FILTERING": java_util_Locale$FilteringMode;
    public static "java/util/Locale$FilteringMode/IGNORE_EXTENDED_RANGES": java_util_Locale$FilteringMode;
    public static "java/util/Locale$FilteringMode/MAP_EXTENDED_RANGES": java_util_Locale$FilteringMode;
    public static "java/util/Locale$FilteringMode/REJECT_EXTENDED_RANGES": java_util_Locale$FilteringMode;
    public static "java/util/Locale$FilteringMode/$VALUES": JVMArray<java_util_Locale$FilteringMode>;
    public static "values()[Ljava/util/Locale$FilteringMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Locale$FilteringMode>) => void): void;
    public static "java/util/Locale$FilteringMode/values()[Ljava/util/Locale$FilteringMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Locale$FilteringMode>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/util/Locale$FilteringMode;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale$FilteringMode) => void): void;
    public static "java/util/Locale$FilteringMode/valueOf(Ljava/lang/String;)Ljava/util/Locale$FilteringMode;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale$FilteringMode) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale$FilteringMode/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Locale$FilteringMode/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_text_MessageFormat extends java_text_Format {
    public static "java/text/MessageFormat/serialVersionUID": Long;
    public "java/text/MessageFormat/locale": java_util_Locale;
    public "java/text/MessageFormat/pattern": java_lang_String;
    public static "java/text/MessageFormat/INITIAL_FORMATS": number;
    public "java/text/MessageFormat/formats": JVMArray<java_text_Format>;
    public "java/text/MessageFormat/offsets": JVMArray<number>;
    public "java/text/MessageFormat/argumentNumbers": JVMArray<number>;
    public "java/text/MessageFormat/maxOffset": number;
    public static "java/text/MessageFormat/SEG_RAW": number;
    public static "java/text/MessageFormat/SEG_INDEX": number;
    public static "java/text/MessageFormat/SEG_TYPE": number;
    public static "java/text/MessageFormat/SEG_MODIFIER": number;
    public static "java/text/MessageFormat/TYPE_NULL": number;
    public static "java/text/MessageFormat/TYPE_NUMBER": number;
    public static "java/text/MessageFormat/TYPE_DATE": number;
    public static "java/text/MessageFormat/TYPE_TIME": number;
    public static "java/text/MessageFormat/TYPE_CHOICE": number;
    public static "java/text/MessageFormat/TYPE_KEYWORDS": JVMArray<java_lang_String>;
    public static "java/text/MessageFormat/MODIFIER_DEFAULT": number;
    public static "java/text/MessageFormat/MODIFIER_CURRENCY": number;
    public static "java/text/MessageFormat/MODIFIER_PERCENT": number;
    public static "java/text/MessageFormat/MODIFIER_INTEGER": number;
    public static "java/text/MessageFormat/NUMBER_MODIFIER_KEYWORDS": JVMArray<java_lang_String>;
    public static "java/text/MessageFormat/MODIFIER_SHORT": number;
    public static "java/text/MessageFormat/MODIFIER_MEDIUM": number;
    public static "java/text/MessageFormat/MODIFIER_LONG": number;
    public static "java/text/MessageFormat/MODIFIER_FULL": number;
    public static "java/text/MessageFormat/DATE_TIME_MODIFIER_KEYWORDS": JVMArray<java_lang_String>;
    public static "java/text/MessageFormat/DATE_TIME_MODIFIERS": JVMArray<number>;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/<init>(Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocale(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/setLocale(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/text/MessageFormat/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "applyPattern(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/applyPattern(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "toPattern()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/text/MessageFormat/toPattern()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setFormatsByArgumentIndex([Ljava/text/Format;)V"(thread: JVMThread, args: [JVMArray<java_text_Format>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/setFormatsByArgumentIndex([Ljava/text/Format;)V"(thread: JVMThread, args: [JVMArray<java_text_Format>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFormats([Ljava/text/Format;)V"(thread: JVMThread, args: [JVMArray<java_text_Format>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/setFormats([Ljava/text/Format;)V"(thread: JVMThread, args: [JVMArray<java_text_Format>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFormatByArgumentIndex(ILjava/text/Format;)V"(thread: JVMThread, args: [number, java_text_Format], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/setFormatByArgumentIndex(ILjava/text/Format;)V"(thread: JVMThread, args: [number, java_text_Format], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFormat(ILjava/text/Format;)V"(thread: JVMThread, args: [number, java_text_Format], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/setFormat(ILjava/text/Format;)V"(thread: JVMThread, args: [number, java_text_Format], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFormatsByArgumentIndex()[Ljava/text/Format;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_text_Format>) => void): void;
    public "java/text/MessageFormat/getFormatsByArgumentIndex()[Ljava/text/Format;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_text_Format>) => void): void;
    public "getFormats()[Ljava/text/Format;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_text_Format>) => void): void;
    public "java/text/MessageFormat/getFormats()[Ljava/text/Format;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_text_Format>) => void): void;
    public "format([Ljava/lang/Object;Ljava/lang/StringBuffer;Ljava/text/FieldPosition;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_StringBuffer, java_text_FieldPosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/text/MessageFormat/format([Ljava/lang/Object;Ljava/lang/StringBuffer;Ljava/text/FieldPosition;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_StringBuffer, java_text_FieldPosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public static "format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/text/MessageFormat/format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "format(Ljava/lang/Object;Ljava/lang/StringBuffer;Ljava/text/FieldPosition;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_Object, java_lang_StringBuffer, java_text_FieldPosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/text/MessageFormat/format(Ljava/lang/Object;Ljava/lang/StringBuffer;Ljava/text/FieldPosition;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_Object, java_lang_StringBuffer, java_text_FieldPosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "formatToCharacterIterator(Ljava/lang/Object;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "java/text/MessageFormat/formatToCharacterIterator(Ljava/lang/Object;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "parse(Ljava/lang/String;Ljava/text/ParsePosition;)[Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/text/MessageFormat/parse(Ljava/lang/String;Ljava/text/ParsePosition;)[Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "parse(Ljava/lang/String;)[Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/text/MessageFormat/parse(Ljava/lang/String;)[Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "parseObject(Ljava/lang/String;Ljava/text/ParsePosition;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/text/MessageFormat/parseObject(Ljava/lang/String;Ljava/text/ParsePosition;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_text_ParsePosition], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/text/MessageFormat/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/MessageFormat/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/text/MessageFormat/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "subformat([Ljava/lang/Object;Ljava/lang/StringBuffer;Ljava/text/FieldPosition;Ljava/util/List;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_StringBuffer, java_text_FieldPosition, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "java/text/MessageFormat/subformat([Ljava/lang/Object;Ljava/lang/StringBuffer;Ljava/text/FieldPosition;Ljava/util/List;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_StringBuffer, java_text_FieldPosition, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "append(Ljava/lang/StringBuffer;Ljava/text/CharacterIterator;)V"(thread: JVMThread, args: [java_lang_StringBuffer, java_text_CharacterIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/append(Ljava/lang/StringBuffer;Ljava/text/CharacterIterator;)V"(thread: JVMThread, args: [java_lang_StringBuffer, java_text_CharacterIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "makeFormat(II[Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [number, number, JVMArray<java_lang_StringBuilder>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/makeFormat(II[Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [number, number, JVMArray<java_lang_StringBuilder>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "findKeyword(Ljava/lang/String;[Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/text/MessageFormat/findKeyword(Ljava/lang/String;[Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "copyAndFixQuotes(Ljava/lang/String;IILjava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/text/MessageFormat/copyAndFixQuotes(Ljava/lang/String;IILjava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/MessageFormat/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/text/MessageFormat/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_Locale$Category extends java_lang_Enum {
    public static "java/util/Locale$Category/DISPLAY": java_util_Locale$Category;
    public static "java/util/Locale$Category/FORMAT": java_util_Locale$Category;
    public "java/util/Locale$Category/languageKey": java_lang_String;
    public "java/util/Locale$Category/scriptKey": java_lang_String;
    public "java/util/Locale$Category/countryKey": java_lang_String;
    public "java/util/Locale$Category/variantKey": java_lang_String;
    public static "java/util/Locale$Category/$VALUES": JVMArray<java_util_Locale$Category>;
    public static "values()[Ljava/util/Locale$Category;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Locale$Category>) => void): void;
    public static "java/util/Locale$Category/values()[Ljava/util/Locale$Category;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Locale$Category>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/util/Locale$Category;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale$Category) => void): void;
    public static "java/util/Locale$Category/valueOf(Ljava/lang/String;)Ljava/util/Locale$Category;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale$Category) => void): void;
    public "<init>(Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale$Category/<init>(Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Locale$Category/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_locale_LocaleExtensions extends java_lang_Object {
    public "sun/util/locale/LocaleExtensions/extensionMap": java_util_Map;
    public "sun/util/locale/LocaleExtensions/id": java_lang_String;
    public static "sun/util/locale/LocaleExtensions/CALENDAR_JAPANESE": sun_util_locale_LocaleExtensions;
    public static "sun/util/locale/LocaleExtensions/NUMBER_THAI": sun_util_locale_LocaleExtensions;
    public static "sun/util/locale/LocaleExtensions/$assertionsDisabled": number;
    public "<init>(Ljava/lang/String;Ljava/lang/Character;Lsun/util/locale/Extension;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Character, sun_util_locale_Extension], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/LocaleExtensions/<init>(Ljava/lang/String;Ljava/lang/Character;Lsun/util/locale/Extension;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Character, sun_util_locale_Extension], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Map;Ljava/util/Set;Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map, java_util_Set, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/LocaleExtensions/<init>(Ljava/util/Map;Ljava/util/Set;Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map, java_util_Set, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/util/locale/LocaleExtensions/getKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getExtension(Ljava/lang/Character;)Lsun/util/locale/Extension;"(thread: JVMThread, args: [java_lang_Character], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_Extension) => void): void;
    public "sun/util/locale/LocaleExtensions/getExtension(Ljava/lang/Character;)Lsun/util/locale/Extension;"(thread: JVMThread, args: [java_lang_Character], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_Extension) => void): void;
    public "getExtensionValue(Ljava/lang/Character;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Character], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/LocaleExtensions/getExtensionValue(Ljava/lang/Character;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Character], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getUnicodeLocaleAttributes()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/util/locale/LocaleExtensions/getUnicodeLocaleAttributes()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getUnicodeLocaleKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/util/locale/LocaleExtensions/getUnicodeLocaleKeys()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getUnicodeLocaleType(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/LocaleExtensions/getUnicodeLocaleType(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/locale/LocaleExtensions/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isValidKey(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/util/locale/LocaleExtensions/isValidKey(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isValidUnicodeLocaleKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/util/locale/LocaleExtensions/isValidUnicodeLocaleKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toID(Ljava/util/SortedMap;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_SortedMap], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/util/locale/LocaleExtensions/toID(Ljava/util/SortedMap;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_SortedMap], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/LocaleExtensions/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getID()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/LocaleExtensions/getID()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/locale/LocaleExtensions/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/locale/LocaleExtensions/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/locale/LocaleExtensions/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_SortedMap extends java_lang_Object, java_util_Map {
    "comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "subMap(Ljava/lang/Object;Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    "headMap(Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    "tailMap(Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    "firstKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "lastKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_util_locale_Extension extends java_lang_Object {
    public "sun/util/locale/Extension/key": number;
    public "sun/util/locale/Extension/value": java_lang_String;
    public "sun/util/locale/Extension/id": java_lang_String;
    public "<init>(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/Extension/<init>(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(CLjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/Extension/<init>(CLjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setValue(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/Extension/setValue(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKey()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/locale/Extension/getKey()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getValue()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/Extension/getValue()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getID()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/Extension/getID()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/Extension/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_lang_Character extends java_lang_Object implements java_io_Serializable, java_lang_Comparable {
    public unbox(): number;
    public static box(val: number): java_lang_Character;
    public static "java/lang/Character/MIN_RADIX": number;
    public static "java/lang/Character/MAX_RADIX": number;
    public static "java/lang/Character/MIN_VALUE": number;
    public static "java/lang/Character/MAX_VALUE": number;
    public static "java/lang/Character/TYPE": java_lang_Class;
    public static "java/lang/Character/UNASSIGNED": number;
    public static "java/lang/Character/UPPERCASE_LETTER": number;
    public static "java/lang/Character/LOWERCASE_LETTER": number;
    public static "java/lang/Character/TITLECASE_LETTER": number;
    public static "java/lang/Character/MODIFIER_LETTER": number;
    public static "java/lang/Character/OTHER_LETTER": number;
    public static "java/lang/Character/NON_SPACING_MARK": number;
    public static "java/lang/Character/ENCLOSING_MARK": number;
    public static "java/lang/Character/COMBINING_SPACING_MARK": number;
    public static "java/lang/Character/DECIMAL_DIGIT_NUMBER": number;
    public static "java/lang/Character/LETTER_NUMBER": number;
    public static "java/lang/Character/OTHER_NUMBER": number;
    public static "java/lang/Character/SPACE_SEPARATOR": number;
    public static "java/lang/Character/LINE_SEPARATOR": number;
    public static "java/lang/Character/PARAGRAPH_SEPARATOR": number;
    public static "java/lang/Character/CONTROL": number;
    public static "java/lang/Character/FORMAT": number;
    public static "java/lang/Character/PRIVATE_USE": number;
    public static "java/lang/Character/SURROGATE": number;
    public static "java/lang/Character/DASH_PUNCTUATION": number;
    public static "java/lang/Character/START_PUNCTUATION": number;
    public static "java/lang/Character/END_PUNCTUATION": number;
    public static "java/lang/Character/CONNECTOR_PUNCTUATION": number;
    public static "java/lang/Character/OTHER_PUNCTUATION": number;
    public static "java/lang/Character/MATH_SYMBOL": number;
    public static "java/lang/Character/CURRENCY_SYMBOL": number;
    public static "java/lang/Character/MODIFIER_SYMBOL": number;
    public static "java/lang/Character/OTHER_SYMBOL": number;
    public static "java/lang/Character/INITIAL_QUOTE_PUNCTUATION": number;
    public static "java/lang/Character/FINAL_QUOTE_PUNCTUATION": number;
    public static "java/lang/Character/ERROR": number;
    public static "java/lang/Character/DIRECTIONALITY_UNDEFINED": number;
    public static "java/lang/Character/DIRECTIONALITY_LEFT_TO_RIGHT": number;
    public static "java/lang/Character/DIRECTIONALITY_RIGHT_TO_LEFT": number;
    public static "java/lang/Character/DIRECTIONALITY_RIGHT_TO_LEFT_ARABIC": number;
    public static "java/lang/Character/DIRECTIONALITY_EUROPEAN_NUMBER": number;
    public static "java/lang/Character/DIRECTIONALITY_EUROPEAN_NUMBER_SEPARATOR": number;
    public static "java/lang/Character/DIRECTIONALITY_EUROPEAN_NUMBER_TERMINATOR": number;
    public static "java/lang/Character/DIRECTIONALITY_ARABIC_NUMBER": number;
    public static "java/lang/Character/DIRECTIONALITY_COMMON_NUMBER_SEPARATOR": number;
    public static "java/lang/Character/DIRECTIONALITY_NONSPACING_MARK": number;
    public static "java/lang/Character/DIRECTIONALITY_BOUNDARY_NEUTRAL": number;
    public static "java/lang/Character/DIRECTIONALITY_PARAGRAPH_SEPARATOR": number;
    public static "java/lang/Character/DIRECTIONALITY_SEGMENT_SEPARATOR": number;
    public static "java/lang/Character/DIRECTIONALITY_WHITESPACE": number;
    public static "java/lang/Character/DIRECTIONALITY_OTHER_NEUTRALS": number;
    public static "java/lang/Character/DIRECTIONALITY_LEFT_TO_RIGHT_EMBEDDING": number;
    public static "java/lang/Character/DIRECTIONALITY_LEFT_TO_RIGHT_OVERRIDE": number;
    public static "java/lang/Character/DIRECTIONALITY_RIGHT_TO_LEFT_EMBEDDING": number;
    public static "java/lang/Character/DIRECTIONALITY_RIGHT_TO_LEFT_OVERRIDE": number;
    public static "java/lang/Character/DIRECTIONALITY_POP_DIRECTIONAL_FORMAT": number;
    public static "java/lang/Character/MIN_HIGH_SURROGATE": number;
    public static "java/lang/Character/MAX_HIGH_SURROGATE": number;
    public static "java/lang/Character/MIN_LOW_SURROGATE": number;
    public static "java/lang/Character/MAX_LOW_SURROGATE": number;
    public static "java/lang/Character/MIN_SURROGATE": number;
    public static "java/lang/Character/MAX_SURROGATE": number;
    public static "java/lang/Character/MIN_SUPPLEMENTARY_CODE_POINT": number;
    public static "java/lang/Character/MIN_CODE_POINT": number;
    public static "java/lang/Character/MAX_CODE_POINT": number;
    public "java/lang/Character/value": number;
    public static "java/lang/Character/serialVersionUID": Long;
    public static "java/lang/Character/SIZE": number;
    public static "java/lang/Character/BYTES": number;
    public static "java/lang/Character/$assertionsDisabled": number;
    public "<init>(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Character/<init>(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "valueOf(C)Ljava/lang/Character;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Character) => void): void;
    public static "java/lang/Character/valueOf(C)Ljava/lang/Character;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Character) => void): void;
    public "charValue()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Character/charValue()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Character/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hashCode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/hashCode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Character/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Character/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toString(C)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Character/toString(C)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "isValidCodePoint(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isValidCodePoint(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isBmpCodePoint(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isBmpCodePoint(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSupplementaryCodePoint(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isSupplementaryCodePoint(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isHighSurrogate(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isHighSurrogate(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLowSurrogate(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isLowSurrogate(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSurrogate(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isSurrogate(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSurrogatePair(CC)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isSurrogatePair(CC)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "charCount(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/charCount(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toCodePoint(CC)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/toCodePoint(CC)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointAt(Ljava/lang/CharSequence;I)I"(thread: JVMThread, args: [java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointAt(Ljava/lang/CharSequence;I)I"(thread: JVMThread, args: [java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointAt([CI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointAt([CI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointAt([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointAt([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointAtImpl([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointAtImpl([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointBefore(Ljava/lang/CharSequence;I)I"(thread: JVMThread, args: [java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointBefore(Ljava/lang/CharSequence;I)I"(thread: JVMThread, args: [java_lang_CharSequence, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointBefore([CI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointBefore([CI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointBefore([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointBefore([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointBeforeImpl([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointBeforeImpl([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "highSurrogate(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/highSurrogate(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "lowSurrogate(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/lowSurrogate(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toChars(I[CI)I"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/toChars(I[CI)I"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toChars(I)[C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/lang/Character/toChars(I)[C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "toSurrogates(I[CI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Character/toSurrogates(I[CI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "codePointCount(Ljava/lang/CharSequence;II)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointCount(Ljava/lang/CharSequence;II)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointCount([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointCount([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "codePointCountImpl([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/codePointCountImpl([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "offsetByCodePoints(Ljava/lang/CharSequence;II)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/offsetByCodePoints(Ljava/lang/CharSequence;II)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "offsetByCodePoints([CIIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/offsetByCodePoints([CIIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "offsetByCodePointsImpl([CIIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/offsetByCodePointsImpl([CIIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLowerCase(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isLowerCase(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLowerCase(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isLowerCase(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isUpperCase(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isUpperCase(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isUpperCase(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isUpperCase(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isTitleCase(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isTitleCase(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isTitleCase(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isTitleCase(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isDigit(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isDigit(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isDefined(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isDefined(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isDefined(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isDefined(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLetter(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isLetter(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLetter(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isLetter(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLetterOrDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isLetterOrDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLetterOrDigit(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isLetterOrDigit(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isJavaLetter(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isJavaLetter(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isJavaLetterOrDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isJavaLetterOrDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isAlphabetic(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isAlphabetic(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isIdeographic(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isIdeographic(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isJavaIdentifierStart(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isJavaIdentifierStart(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isJavaIdentifierStart(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isJavaIdentifierStart(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isJavaIdentifierPart(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isJavaIdentifierPart(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isJavaIdentifierPart(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isJavaIdentifierPart(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isUnicodeIdentifierStart(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isUnicodeIdentifierStart(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isUnicodeIdentifierStart(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isUnicodeIdentifierStart(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isUnicodeIdentifierPart(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isUnicodeIdentifierPart(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isUnicodeIdentifierPart(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isUnicodeIdentifierPart(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isIdentifierIgnorable(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isIdentifierIgnorable(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isIdentifierIgnorable(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isIdentifierIgnorable(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toLowerCase(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/toLowerCase(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toLowerCase(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/toLowerCase(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toUpperCase(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/toUpperCase(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toUpperCase(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/toUpperCase(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toTitleCase(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/toTitleCase(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toTitleCase(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/toTitleCase(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "digit(CI)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/digit(CI)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "digit(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/digit(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getNumericValue(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/getNumericValue(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getNumericValue(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/getNumericValue(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSpace(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isSpace(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSpaceChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isSpaceChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSpaceChar(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isSpaceChar(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isWhitespace(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isWhitespace(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isWhitespace(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isWhitespace(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isISOControl(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isISOControl(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isISOControl(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isISOControl(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getType(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/getType(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getType(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/getType(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "forDigit(II)C"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/forDigit(II)C"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDirectionality(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/getDirectionality(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDirectionality(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/getDirectionality(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isMirrored(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isMirrored(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isMirrored(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/isMirrored(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Character;)I"(thread: JVMThread, args: [java_lang_Character], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Character/compareTo(Ljava/lang/Character;)I"(thread: JVMThread, args: [java_lang_Character], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(CC)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/compare(CC)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toUpperCaseEx(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/toUpperCaseEx(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toUpperCaseCharArray(I)[C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/lang/Character/toUpperCaseCharArray(I)[C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "reverseBytes(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Character/reverseBytes(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getName(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Character/getName(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Character/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Character/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_locale_BaseLocale extends java_lang_Object {
    public static "sun/util/locale/BaseLocale/SEP": java_lang_String;
    public static "sun/util/locale/BaseLocale/CACHE": sun_util_locale_BaseLocale$Cache;
    public "sun/util/locale/BaseLocale/language": java_lang_String;
    public "sun/util/locale/BaseLocale/script": java_lang_String;
    public "sun/util/locale/BaseLocale/region": java_lang_String;
    public "sun/util/locale/BaseLocale/variant": java_lang_String;
    public "sun/util/locale/BaseLocale/hash": number;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/BaseLocale/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/BaseLocale/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createInstance(Ljava/lang/String;Ljava/lang/String;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public static "sun/util/locale/BaseLocale/createInstance(Ljava/lang/String;Ljava/lang/String;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public static "sun/util/locale/BaseLocale/getInstance(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public "getLanguage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/BaseLocale/getLanguage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getScript()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/BaseLocale/getScript()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRegion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/BaseLocale/getRegion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVariant()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/BaseLocale/getVariant()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/locale/BaseLocale/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/locale/BaseLocale/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/locale/BaseLocale/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Lsun/util/locale/BaseLocale$1;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, sun_util_locale_BaseLocale$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/BaseLocale/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Lsun/util/locale/BaseLocale$1;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, sun_util_locale_BaseLocale$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/locale/BaseLocale/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_locale_BaseLocale$1 extends java_lang_Object {
  }
  export class sun_util_locale_BaseLocale$Cache extends sun_util_locale_LocaleObjectCache {
    public static "sun/util/locale/BaseLocale$Cache/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/BaseLocale$Cache/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "normalizeKey(Lsun/util/locale/BaseLocale$Key;)Lsun/util/locale/BaseLocale$Key;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale$Key) => void): void;
    public "sun/util/locale/BaseLocale$Cache/normalizeKey(Lsun/util/locale/BaseLocale$Key;)Lsun/util/locale/BaseLocale$Key;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale$Key) => void): void;
    public "createObject(Lsun/util/locale/BaseLocale$Key;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public "sun/util/locale/BaseLocale$Cache/createObject(Lsun/util/locale/BaseLocale$Key;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public "normalizeKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/util/locale/BaseLocale$Cache/normalizeKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "createObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/util/locale/BaseLocale$Cache/createObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/locale/BaseLocale$Cache/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_locale_BaseLocale$Key extends java_lang_Object {
    public "sun/util/locale/BaseLocale$Key/lang": java_lang_ref_SoftReference;
    public "sun/util/locale/BaseLocale$Key/scrt": java_lang_ref_SoftReference;
    public "sun/util/locale/BaseLocale$Key/regn": java_lang_ref_SoftReference;
    public "sun/util/locale/BaseLocale$Key/vart": java_lang_ref_SoftReference;
    public "sun/util/locale/BaseLocale$Key/normalized": number;
    public "sun/util/locale/BaseLocale$Key/hash": number;
    public static "sun/util/locale/BaseLocale$Key/$assertionsDisabled": number;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/BaseLocale$Key/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/BaseLocale$Key/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/BaseLocale$Key/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/locale/BaseLocale$Key/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/locale/BaseLocale$Key/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "normalize(Lsun/util/locale/BaseLocale$Key;)Lsun/util/locale/BaseLocale$Key;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale$Key) => void): void;
    public static "sun/util/locale/BaseLocale$Key/normalize(Lsun/util/locale/BaseLocale$Key;)Lsun/util/locale/BaseLocale$Key;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale$Key) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Lsun/util/locale/BaseLocale$1;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, sun_util_locale_BaseLocale$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/BaseLocale$Key/<init>(Ljava/lang/String;Ljava/lang/String;Lsun/util/locale/BaseLocale$1;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, sun_util_locale_BaseLocale$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$100(Lsun/util/locale/BaseLocale$Key;)Ljava/lang/ref/SoftReference;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_SoftReference) => void): void;
    public static "sun/util/locale/BaseLocale$Key/access$100(Lsun/util/locale/BaseLocale$Key;)Ljava/lang/ref/SoftReference;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_SoftReference) => void): void;
    public static "access$200(Lsun/util/locale/BaseLocale$Key;)Ljava/lang/ref/SoftReference;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_SoftReference) => void): void;
    public static "sun/util/locale/BaseLocale$Key/access$200(Lsun/util/locale/BaseLocale$Key;)Ljava/lang/ref/SoftReference;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_SoftReference) => void): void;
    public static "access$300(Lsun/util/locale/BaseLocale$Key;)Ljava/lang/ref/SoftReference;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_SoftReference) => void): void;
    public static "sun/util/locale/BaseLocale$Key/access$300(Lsun/util/locale/BaseLocale$Key;)Ljava/lang/ref/SoftReference;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_SoftReference) => void): void;
    public static "access$400(Lsun/util/locale/BaseLocale$Key;)Ljava/lang/ref/SoftReference;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_SoftReference) => void): void;
    public static "sun/util/locale/BaseLocale$Key/access$400(Lsun/util/locale/BaseLocale$Key;)Ljava/lang/ref/SoftReference;"(thread: JVMThread, args: [sun_util_locale_BaseLocale$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_SoftReference) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/locale/BaseLocale$Key/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ref_SoftReference extends java_lang_ref_Reference {
    public static "java/lang/ref/SoftReference/clock": Long;
    public "java/lang/ref/SoftReference/timestamp": Long;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/SoftReference/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_ref_ReferenceQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/SoftReference/<init>(Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_ref_ReferenceQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ref/SoftReference/get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_util_locale_LocaleObjectCache extends java_lang_Object {
    public "sun/util/locale/LocaleObjectCache/map": java_util_concurrent_ConcurrentMap;
    public "sun/util/locale/LocaleObjectCache/queue": java_lang_ref_ReferenceQueue;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/LocaleObjectCache/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IFI)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/LocaleObjectCache/<init>(IFI)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/util/locale/LocaleObjectCache/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/util/locale/LocaleObjectCache/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "cleanStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/locale/LocaleObjectCache/cleanStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "createObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/util/locale/LocaleObjectCache/createObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "normalizeKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/util/locale/LocaleObjectCache/normalizeKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_Locale$Cache extends sun_util_locale_LocaleObjectCache {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale$Cache/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "createObject(Ljava/util/Locale$LocaleKey;)Ljava/util/Locale;"(thread: JVMThread, args: [java_util_Locale$LocaleKey], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/util/Locale$Cache/createObject(Ljava/util/Locale$LocaleKey;)Ljava/util/Locale;"(thread: JVMThread, args: [java_util_Locale$LocaleKey], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "createObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Locale$Cache/createObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "<init>(Ljava/util/Locale$1;)V"(thread: JVMThread, args: [java_util_Locale$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale$Cache/<init>(Ljava/util/Locale$1;)V"(thread: JVMThread, args: [java_util_Locale$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_Locale$LocaleKey extends java_lang_Object {
    public "java/util/Locale$LocaleKey/base": sun_util_locale_BaseLocale;
    public "java/util/Locale$LocaleKey/exts": sun_util_locale_LocaleExtensions;
    public "java/util/Locale$LocaleKey/hash": number;
    public "<init>(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;)V"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale$LocaleKey/<init>(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;)V"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Locale$LocaleKey/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Locale$LocaleKey/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;Ljava/util/Locale$1;)V"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions, java_util_Locale$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Locale$LocaleKey/<init>(Lsun/util/locale/BaseLocale;Lsun/util/locale/LocaleExtensions;Ljava/util/Locale$1;)V"(thread: JVMThread, args: [sun_util_locale_BaseLocale, sun_util_locale_LocaleExtensions, java_util_Locale$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$200(Ljava/util/Locale$LocaleKey;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [java_util_Locale$LocaleKey], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public static "java/util/Locale$LocaleKey/access$200(Ljava/util/Locale$LocaleKey;)Lsun/util/locale/BaseLocale;"(thread: JVMThread, args: [java_util_Locale$LocaleKey], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_BaseLocale) => void): void;
    public static "access$300(Ljava/util/Locale$LocaleKey;)Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: [java_util_Locale$LocaleKey], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
    public static "java/util/Locale$LocaleKey/access$300(Ljava/util/Locale$LocaleKey;)Lsun/util/locale/LocaleExtensions;"(thread: JVMThread, args: [java_util_Locale$LocaleKey], cb?: (e?: java_lang_Throwable, rv?: sun_util_locale_LocaleExtensions) => void): void;
  }
  export class java_time_format_TextStyle extends java_lang_Enum {
    public static "java/time/format/TextStyle/FULL": java_time_format_TextStyle;
    public static "java/time/format/TextStyle/FULL_STANDALONE": java_time_format_TextStyle;
    public static "java/time/format/TextStyle/SHORT": java_time_format_TextStyle;
    public static "java/time/format/TextStyle/SHORT_STANDALONE": java_time_format_TextStyle;
    public static "java/time/format/TextStyle/NARROW": java_time_format_TextStyle;
    public static "java/time/format/TextStyle/NARROW_STANDALONE": java_time_format_TextStyle;
    public "java/time/format/TextStyle/calendarStyle": number;
    public "java/time/format/TextStyle/zoneNameStyleIndex": number;
    public static "java/time/format/TextStyle/$VALUES": JVMArray<java_time_format_TextStyle>;
    public static "values()[Ljava/time/format/TextStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_format_TextStyle>) => void): void;
    public static "java/time/format/TextStyle/values()[Ljava/time/format/TextStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_time_format_TextStyle>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/time/format/TextStyle;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_format_TextStyle) => void): void;
    public static "java/time/format/TextStyle/valueOf(Ljava/lang/String;)Ljava/time/format/TextStyle;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_format_TextStyle) => void): void;
    public "<init>(Ljava/lang/String;III)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/format/TextStyle/<init>(Ljava/lang/String;III)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isStandalone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/TextStyle/isStandalone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "asStandalone()Ljava/time/format/TextStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_TextStyle) => void): void;
    public "java/time/format/TextStyle/asStandalone()Ljava/time/format/TextStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_TextStyle) => void): void;
    public "asNormal()Ljava/time/format/TextStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_TextStyle) => void): void;
    public "java/time/format/TextStyle/asNormal()Ljava/time/format/TextStyle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_format_TextStyle) => void): void;
    public "toCalendarStyle()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/TextStyle/toCalendarStyle()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "zoneNameStyleIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/format/TextStyle/zoneNameStyleIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/format/TextStyle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_time_ZonedDateTime extends java_lang_Object implements java_time_temporal_Temporal, java_time_chrono_ChronoZonedDateTime, java_io_Serializable {
    public static "java/time/ZonedDateTime/serialVersionUID": Long;
    public "java/time/ZonedDateTime/dateTime": java_time_LocalDateTime;
    public "java/time/ZonedDateTime/offset": java_time_ZoneOffset;
    public "java/time/ZonedDateTime/zone": java_time_ZoneId;
    public static "now()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/now()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "now(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/now(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "now(Ljava/time/Clock;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/now(Ljava/time/Clock;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "of(Ljava/time/LocalDate;Ljava/time/LocalTime;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/of(Ljava/time/LocalDate;Ljava/time/LocalTime;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "of(Ljava/time/LocalDateTime;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/of(Ljava/time/LocalDateTime;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "of(IIIIIIILjava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number, number, number, number, number, number, number, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/of(IIIIIIILjava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number, number, number, number, number, number, number, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "ofLocal(Ljava/time/LocalDateTime;Ljava/time/ZoneId;Ljava/time/ZoneOffset;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneId, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/ofLocal(Ljava/time/LocalDateTime;Ljava/time/ZoneId;Ljava/time/ZoneOffset;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneId, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "ofInstant(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/ofInstant(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "ofInstant(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/ofInstant(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "create(JILjava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any, number, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/create(JILjava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any, number, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "ofStrict(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/ofStrict(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "ofLenient(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/ofLenient(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/parse(Ljava/lang/CharSequence;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "<init>(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneId;)V"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneId], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZonedDateTime/<init>(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;Ljava/time/ZoneId;)V"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset, java_time_ZoneId], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveLocal(Ljava/time/LocalDateTime;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/resolveLocal(Ljava/time/LocalDateTime;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "resolveInstant(Ljava/time/LocalDateTime;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/resolveInstant(Ljava/time/LocalDateTime;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "resolveOffset(Ljava/time/ZoneOffset;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/resolveOffset(Ljava/time/ZoneOffset;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/ZonedDateTime/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/ZonedDateTime/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getOffset()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/ZonedDateTime/getOffset()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "withEarlierOffsetAtOverlap()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withEarlierOffsetAtOverlap()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withLaterOffsetAtOverlap()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withLaterOffsetAtOverlap()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "getZone()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "java/time/ZonedDateTime/getZone()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "withZoneSameLocal(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withZoneSameLocal(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withZoneSameInstant(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withZoneSameInstant(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withFixedOffsetZone()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withFixedOffsetZone()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "toLocalDateTime()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/ZonedDateTime/toLocalDateTime()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "toLocalDate()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/ZonedDateTime/toLocalDate()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonthValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/getMonthValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "java/time/ZonedDateTime/getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/getDayOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "java/time/ZonedDateTime/getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/ZonedDateTime/toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withYear(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withYear(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withMonth(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withMonth(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withDayOfMonth(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withDayOfMonth(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withDayOfYear(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withDayOfYear(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withHour(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withHour(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withMinute(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withMinute(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withSecond(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withSecond(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "withNano(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withNano(I)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plusYears(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plusYears(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plusMonths(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plusMonths(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plusWeeks(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plusWeeks(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plusDays(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plusDays(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plusHours(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plusHours(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plusMinutes(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plusMinutes(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plusSeconds(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plusSeconds(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "plusNanos(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plusNanos(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minusYears(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minusYears(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minusMonths(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minusMonths(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minusWeeks(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minusWeeks(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minusDays(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minusDays(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minusHours(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minusHours(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minusMinutes(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minusMinutes(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minusSeconds(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minusSeconds(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minusNanos(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minusNanos(J)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/ZonedDateTime/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/ZonedDateTime/until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/ZonedDateTime/format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toOffsetDateTime()Ljava/time/OffsetDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/ZonedDateTime/toOffsetDateTime()Ljava/time/OffsetDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZonedDateTime/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/ZonedDateTime/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/ZonedDateTime/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZonedDateTime/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZonedDateTime/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/ObjectInput;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public static "java/time/ZonedDateTime/readExternal(Ljava/io/ObjectInput;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/ZonedDateTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/ZonedDateTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/ZonedDateTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/ZonedDateTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/ZonedDateTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/ZonedDateTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "withZoneSameInstant(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withZoneSameInstant(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "withZoneSameLocal(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withZoneSameLocal(Ljava/time/ZoneId;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "withLaterOffsetAtOverlap()Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withLaterOffsetAtOverlap()Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "withEarlierOffsetAtOverlap()Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "java/time/ZonedDateTime/withEarlierOffsetAtOverlap()Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    public "toLocalDateTime()Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "java/time/ZonedDateTime/toLocalDateTime()Ljava/time/chrono/ChronoLocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDateTime) => void): void;
    public "toLocalDate()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    public "java/time/ZonedDateTime/toLocalDate()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "getChronology()Ljava/time/chrono/Chronology;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_Chronology) => void): void;
    "toInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    "toEpochSecond()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "compareTo(Ljava/time/chrono/ChronoZonedDateTime;)I"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isBefore(Ljava/time/chrono/ChronoZonedDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isAfter(Ljava/time/chrono/ChronoZonedDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEqual(Ljava/time/chrono/ChronoZonedDateTime;)Z"(thread: JVMThread, args: [java_time_chrono_ChronoZonedDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "toLocalDate()Ljava/time/chrono/ChronoLocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoLocalDate) => void): void;
    "toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/chrono/ChronoZonedDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_chrono_ChronoZonedDateTime) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
  }
  export class java_time_ZoneOffset extends java_time_ZoneId implements java_time_temporal_TemporalAccessor, java_time_temporal_TemporalAdjuster, java_lang_Comparable, java_io_Serializable {
    public static "java/time/ZoneOffset/SECONDS_CACHE": java_util_concurrent_ConcurrentMap;
    public static "java/time/ZoneOffset/ID_CACHE": java_util_concurrent_ConcurrentMap;
    public static "java/time/ZoneOffset/MAX_SECONDS": number;
    public static "java/time/ZoneOffset/serialVersionUID": Long;
    public static "java/time/ZoneOffset/UTC": java_time_ZoneOffset;
    public static "java/time/ZoneOffset/MIN": java_time_ZoneOffset;
    public static "java/time/ZoneOffset/MAX": java_time_ZoneOffset;
    public "java/time/ZoneOffset/totalSeconds": number;
    public "java/time/ZoneOffset/id": java_lang_String;
    public static "of(Ljava/lang/String;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "java/time/ZoneOffset/of(Ljava/lang/String;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "parseNumber(Ljava/lang/CharSequence;IZ)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/ZoneOffset/parseNumber(Ljava/lang/CharSequence;IZ)I"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ofHours(I)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "java/time/ZoneOffset/ofHours(I)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "ofHoursMinutes(II)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "java/time/ZoneOffset/ofHoursMinutes(II)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "ofHoursMinutesSeconds(III)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "java/time/ZoneOffset/ofHoursMinutesSeconds(III)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "java/time/ZoneOffset/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "validate(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/ZoneOffset/validate(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "totalSeconds(III)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/ZoneOffset/totalSeconds(III)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ofTotalSeconds(I)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "java/time/ZoneOffset/ofTotalSeconds(I)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZoneOffset/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "buildId(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/time/ZoneOffset/buildId(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTotalSeconds()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZoneOffset/getTotalSeconds()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/ZoneOffset/getId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRules()Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public "java/time/ZoneOffset/getRules()Ljava/time/zone/ZoneRules;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_zone_ZoneRules) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZoneOffset/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/ZoneOffset/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZoneOffset/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/ZoneOffset/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/ZoneOffset/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/ZoneOffset/adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "compareTo(Ljava/time/ZoneOffset;)I"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZoneOffset/compareTo(Ljava/time/ZoneOffset;)I"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZoneOffset/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZoneOffset/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/ZoneOffset/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/ZoneOffset/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZoneOffset/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZoneOffset/write(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/ZoneOffset/writeExternal(Ljava/io/DataOutput;)V"(thread: JVMThread, args: [java_io_DataOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/DataInput;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public static "java/time/ZoneOffset/readExternal(Ljava/io/DataInput;)Ljava/time/ZoneOffset;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/ZoneOffset/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/ZoneOffset/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_time_OffsetDateTime extends java_lang_Object implements java_time_temporal_Temporal, java_time_temporal_TemporalAdjuster, java_lang_Comparable, java_io_Serializable {
    public static "java/time/OffsetDateTime/MIN": java_time_OffsetDateTime;
    public static "java/time/OffsetDateTime/MAX": java_time_OffsetDateTime;
    public static "java/time/OffsetDateTime/serialVersionUID": Long;
    public "java/time/OffsetDateTime/dateTime": java_time_LocalDateTime;
    public "java/time/OffsetDateTime/offset": java_time_ZoneOffset;
    public static "timeLineOrder()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public static "java/time/OffsetDateTime/timeLineOrder()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public static "compareInstant(Ljava/time/OffsetDateTime;Ljava/time/OffsetDateTime;)I"(thread: JVMThread, args: [java_time_OffsetDateTime, java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/time/OffsetDateTime/compareInstant(Ljava/time/OffsetDateTime;Ljava/time/OffsetDateTime;)I"(thread: JVMThread, args: [java_time_OffsetDateTime, java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "now()Ljava/time/OffsetDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/now()Ljava/time/OffsetDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "now(Ljava/time/ZoneId;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/now(Ljava/time/ZoneId;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "now(Ljava/time/Clock;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/now(Ljava/time/Clock;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_Clock], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "of(Ljava/time/LocalDate;Ljava/time/LocalTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/of(Ljava/time/LocalDate;Ljava/time/LocalTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_LocalDate, java_time_LocalTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "of(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/of(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "of(IIIIIIILjava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number, number, number, number, number, number, number, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/of(IIIIIIILjava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number, number, number, number, number, number, number, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "ofInstant(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/ofInstant(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/from(Ljava/time/temporal/TemporalAccessor;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/parse(Ljava/lang/CharSequence;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/parse(Ljava/lang/CharSequence;Ljava/time/format/DateTimeFormatter;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_lang_CharSequence, java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "<init>(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/OffsetDateTime/<init>(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;)V"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable) => void): void;
    public "with(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/with(Ljava/time/LocalDateTime;Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_LocalDateTime, java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "java/time/OffsetDateTime/range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/OffsetDateTime/getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getOffset()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "java/time/OffsetDateTime/getOffset()Ljava/time/ZoneOffset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneOffset) => void): void;
    public "withOffsetSameLocal(Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withOffsetSameLocal(Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "withOffsetSameInstant(Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withOffsetSameInstant(Ljava/time/ZoneOffset;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_ZoneOffset], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "toLocalDateTime()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "java/time/OffsetDateTime/toLocalDateTime()Ljava/time/LocalDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDateTime) => void): void;
    public "toLocalDate()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "java/time/OffsetDateTime/toLocalDate()Ljava/time/LocalDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalDate) => void): void;
    public "getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonthValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/getMonthValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "java/time/OffsetDateTime/getMonth()Ljava/time/Month;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Month) => void): void;
    public "getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/getDayOfYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "java/time/OffsetDateTime/getDayOfWeek()Ljava/time/DayOfWeek;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_DayOfWeek) => void): void;
    public "toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "java/time/OffsetDateTime/toLocalTime()Ljava/time/LocalTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_LocalTime) => void): void;
    public "getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/getHour()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/getMinute()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/getSecond()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/getNano()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "withYear(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withYear(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "withMonth(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withMonth(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "withDayOfMonth(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withDayOfMonth(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "withDayOfYear(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withDayOfYear(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "withHour(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withHour(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "withMinute(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withMinute(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "withSecond(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withSecond(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "withNano(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/withNano(I)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/truncatedTo(Ljava/time/temporal/TemporalUnit;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plusYears(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plusYears(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plusMonths(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plusMonths(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plusWeeks(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plusWeeks(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plusDays(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plusDays(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plusHours(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plusHours(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plusMinutes(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plusMinutes(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plusSeconds(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plusSeconds(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "plusNanos(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/plusNanos(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minusYears(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minusYears(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minusMonths(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minusMonths(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minusWeeks(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minusWeeks(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minusDays(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minusDays(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minusHours(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minusHours(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minusMinutes(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minusMinutes(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minusSeconds(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minusSeconds(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minusNanos(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "java/time/OffsetDateTime/minusNanos(J)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/OffsetDateTime/query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetDateTime/adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/OffsetDateTime/until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/OffsetDateTime/format(Ljava/time/format/DateTimeFormatter;)Ljava/lang/String;"(thread: JVMThread, args: [java_time_format_DateTimeFormatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "atZoneSameInstant(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/OffsetDateTime/atZoneSameInstant(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "atZoneSimilarLocal(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/OffsetDateTime/atZoneSimilarLocal(Ljava/time/ZoneId;)Ljava/time/ZonedDateTime;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "toOffsetTime()Ljava/time/OffsetTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "java/time/OffsetDateTime/toOffsetTime()Ljava/time/OffsetTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetTime) => void): void;
    public "toZonedDateTime()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "java/time/OffsetDateTime/toZonedDateTime()Ljava/time/ZonedDateTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZonedDateTime) => void): void;
    public "toInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/OffsetDateTime/toInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "toEpochSecond()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/OffsetDateTime/toEpochSecond()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareTo(Ljava/time/OffsetDateTime;)I"(thread: JVMThread, args: [java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/compareTo(Ljava/time/OffsetDateTime;)I"(thread: JVMThread, args: [java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAfter(Ljava/time/OffsetDateTime;)Z"(thread: JVMThread, args: [java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/isAfter(Ljava/time/OffsetDateTime;)Z"(thread: JVMThread, args: [java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBefore(Ljava/time/OffsetDateTime;)Z"(thread: JVMThread, args: [java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/isBefore(Ljava/time/OffsetDateTime;)Z"(thread: JVMThread, args: [java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEqual(Ljava/time/OffsetDateTime;)Z"(thread: JVMThread, args: [java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/isEqual(Ljava/time/OffsetDateTime;)Z"(thread: JVMThread, args: [java_time_OffsetDateTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/OffsetDateTime/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/time/OffsetDateTime/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/OffsetDateTime/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternal(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/OffsetDateTime/writeExternal(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readExternal(Ljava/io/ObjectInput;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public static "java/time/OffsetDateTime/readExternal(Ljava/io/ObjectInput;)Ljava/time/OffsetDateTime;"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable, rv?: java_time_OffsetDateTime) => void): void;
    public "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetDateTime/minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetDateTime/minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetDateTime/plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetDateTime/plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetDateTime/with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "java/time/OffsetDateTime/with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/OffsetDateTime/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/time/OffsetDateTime/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_time_temporal_TemporalQuery extends java_lang_Object {
    "queryFrom(Ljava/time/temporal/TemporalAccessor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_time_temporal_TemporalAmount extends java_lang_Object {
    "get(Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getUnits()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "addTo(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "subtractFrom(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
  }
  export class java_time_temporal_ValueRange extends java_lang_Object implements java_io_Serializable {
    public static "java/time/temporal/ValueRange/serialVersionUID": Long;
    public "java/time/temporal/ValueRange/minSmallest": Long;
    public "java/time/temporal/ValueRange/minLargest": Long;
    public "java/time/temporal/ValueRange/maxSmallest": Long;
    public "java/time/temporal/ValueRange/maxLargest": Long;
    public static "of(JJ)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public static "java/time/temporal/ValueRange/of(JJ)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public static "of(JJJ)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public static "java/time/temporal/ValueRange/of(JJJ)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public static "of(JJJJ)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public static "java/time/temporal/ValueRange/of(JJJJ)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    public "<init>(JJJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/temporal/ValueRange/<init>(JJJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "isFixed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ValueRange/isFixed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinimum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/temporal/ValueRange/getMinimum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLargestMinimum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/temporal/ValueRange/getLargestMinimum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getSmallestMaximum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/temporal/ValueRange/getSmallestMaximum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMaximum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/temporal/ValueRange/getMaximum()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "isIntValue()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ValueRange/isIntValue()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isValidValue(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ValueRange/isValidValue(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isValidIntValue(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ValueRange/isValidIntValue(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkValidValue(JLjava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/temporal/ValueRange/checkValidValue(JLjava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "checkValidIntValue(JLjava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ValueRange/checkValidIntValue(JLjava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "genInvalidFieldMessage(Ljava/time/temporal/TemporalField;J)Ljava/lang/String;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/temporal/ValueRange/genInvalidFieldMessage(Ljava/time/temporal/TemporalField;J)Ljava/lang/String;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/temporal/ValueRange/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ValueRange/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/temporal/ValueRange/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/time/temporal/ValueRange/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_time_temporal_TemporalUnit extends java_lang_Object {
    "getDuration()Ljava/time/Duration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Duration) => void): void;
    "isDurationEstimated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isDateBased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isTimeBased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSupportedBy(Ljava/time/temporal/Temporal;)Z"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addTo(Ljava/time/temporal/Temporal;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "between(Ljava/time/temporal/Temporal;Ljava/time/temporal/Temporal;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_time_temporal_TemporalField extends java_lang_Object {
    "getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getBaseUnit()Ljava/time/temporal/TemporalUnit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalUnit) => void): void;
    "getRangeUnit()Ljava/time/temporal/TemporalUnit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalUnit) => void): void;
    "range()Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "isDateBased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isTimeBased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSupportedBy(Ljava/time/temporal/TemporalAccessor;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "rangeRefinedBy(Ljava/time/temporal/TemporalAccessor;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "getFrom(Ljava/time/temporal/TemporalAccessor;)J"(thread: JVMThread, args: [java_time_temporal_TemporalAccessor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "adjustInto(Ljava/time/temporal/Temporal;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "resolve(Ljava/util/Map;Ljava/time/temporal/TemporalAccessor;Ljava/time/format/ResolverStyle;)Ljava/time/temporal/TemporalAccessor;"(thread: JVMThread, args: [java_util_Map, java_time_temporal_TemporalAccessor, java_time_format_ResolverStyle], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_TemporalAccessor) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_time_temporal_TemporalAccessor extends java_lang_Object {
    "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_time_Clock extends java_lang_Object {
    public static "systemUTC()Ljava/time/Clock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "java/time/Clock/systemUTC()Ljava/time/Clock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "systemDefaultZone()Ljava/time/Clock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "java/time/Clock/systemDefaultZone()Ljava/time/Clock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "system(Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "java/time/Clock/system(Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "tickSeconds(Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "java/time/Clock/tickSeconds(Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "tickMinutes(Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "java/time/Clock/tickMinutes(Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "tick(Ljava/time/Clock;Ljava/time/Duration;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_Clock, java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "java/time/Clock/tick(Ljava/time/Clock;Ljava/time/Duration;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_Clock, java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "fixed(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "java/time/Clock/fixed(Ljava/time/Instant;Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_Instant, java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "offset(Ljava/time/Clock;Ljava/time/Duration;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_Clock, java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public static "java/time/Clock/offset(Ljava/time/Clock;Ljava/time/Duration;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_Clock, java_time_Duration], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/time/Clock/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getZone()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "java/time/Clock/getZone()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "withZone(Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public "java/time/Clock/withZone(Ljava/time/ZoneId;)Ljava/time/Clock;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_time_Clock) => void): void;
    public "millis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/time/Clock/millis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "instant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/time/Clock/instant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Clock/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/time/Clock/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_time_temporal_TemporalAdjuster extends java_lang_Object {
    "adjustInto(Ljava/time/temporal/Temporal;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_Temporal], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
  }
  export interface java_time_temporal_Temporal extends java_lang_Object, java_time_temporal_TemporalAccessor {
    "isSupported(Ljava/time/temporal/TemporalUnit;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "with(Ljava/time/temporal/TemporalAdjuster;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAdjuster], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "with(Ljava/time/temporal/TemporalField;J)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalField, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "plus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(Ljava/time/temporal/TemporalAmount;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [java_time_temporal_TemporalAmount], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "minus(JLjava/time/temporal/TemporalUnit;)Ljava/time/temporal/Temporal;"(thread: JVMThread, args: [Long, any, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_Temporal) => void): void;
    "until(Ljava/time/temporal/Temporal;Ljava/time/temporal/TemporalUnit;)J"(thread: JVMThread, args: [java_time_temporal_Temporal, java_time_temporal_TemporalUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "isSupported(Ljava/time/temporal/TemporalField;)Z"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "range(Ljava/time/temporal/TemporalField;)Ljava/time/temporal/ValueRange;"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: java_time_temporal_ValueRange) => void): void;
    "get(Ljava/time/temporal/TemporalField;)I"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLong(Ljava/time/temporal/TemporalField;)J"(thread: JVMThread, args: [java_time_temporal_TemporalField], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "query(Ljava/time/temporal/TemporalQuery;)Ljava/lang/Object;"(thread: JVMThread, args: [java_time_temporal_TemporalQuery], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_util_calendar_BaseCalendar$Date extends sun_util_calendar_CalendarDate {
    public "sun/util/calendar/BaseCalendar$Date/cachedYear": number;
    public "sun/util/calendar/BaseCalendar$Date/cachedFixedDateJan1": Long;
    public "sun/util/calendar/BaseCalendar$Date/cachedFixedDateNextJan1": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/TimeZone;)V"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/<init>(Ljava/util/TimeZone;)V"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable) => void): void;
    public "setNormalizedDate(III)Lsun/util/calendar/BaseCalendar$Date;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar$Date) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/setNormalizedDate(III)Lsun/util/calendar/BaseCalendar$Date;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_BaseCalendar$Date) => void): void;
    public "getNormalizedYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/getNormalizedYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setNormalizedYear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/setNormalizedYear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "hit(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/hit(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hit(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/hit(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCachedYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/getCachedYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCachedJan1()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/getCachedJan1()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setCache(IJI)V"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/BaseCalendar$Date/setCache(IJI)V"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_calendar_CalendarDate extends java_lang_Object implements java_lang_Cloneable {
    public static "sun/util/calendar/CalendarDate/FIELD_UNDEFINED": number;
    public static "sun/util/calendar/CalendarDate/TIME_UNDEFINED": Long;
    public "sun/util/calendar/CalendarDate/era": sun_util_calendar_Era;
    public "sun/util/calendar/CalendarDate/year": number;
    public "sun/util/calendar/CalendarDate/month": number;
    public "sun/util/calendar/CalendarDate/dayOfMonth": number;
    public "sun/util/calendar/CalendarDate/dayOfWeek": number;
    public "sun/util/calendar/CalendarDate/leapYear": number;
    public "sun/util/calendar/CalendarDate/hours": number;
    public "sun/util/calendar/CalendarDate/minutes": number;
    public "sun/util/calendar/CalendarDate/seconds": number;
    public "sun/util/calendar/CalendarDate/millis": number;
    public "sun/util/calendar/CalendarDate/fraction": Long;
    public "sun/util/calendar/CalendarDate/normalized": number;
    public "sun/util/calendar/CalendarDate/zoneinfo": java_util_TimeZone;
    public "sun/util/calendar/CalendarDate/zoneOffset": number;
    public "sun/util/calendar/CalendarDate/daylightSaving": number;
    public "sun/util/calendar/CalendarDate/forceStandardTime": number;
    public "sun/util/calendar/CalendarDate/locale": java_util_Locale;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/TimeZone;)V"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/<init>(Ljava/util/TimeZone;)V"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEra()Lsun/util/calendar/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Era) => void): void;
    public "sun/util/calendar/CalendarDate/getEra()Lsun/util/calendar/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Era) => void): void;
    public "setEra(Lsun/util/calendar/Era;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [sun_util_calendar_Era], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setEra(Lsun/util/calendar/Era;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [sun_util_calendar_Era], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setYear(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setYear(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "addYear(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/addYear(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "isLeapYear()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/isLeapYear()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLeapYear(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/setLeapYear(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMonth(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setMonth(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "addMonth(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/addMonth(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getDayOfMonth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDayOfMonth(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setDayOfMonth(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "addDayOfMonth(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/addDayOfMonth(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getDayOfWeek()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getDayOfWeek()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHours()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getHours()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setHours(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setHours(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "addHours(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/addHours(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getMinutes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getMinutes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMinutes(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setMinutes(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "addMinutes(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/addMinutes(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getSeconds()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getSeconds()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSeconds(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setSeconds(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "addSeconds(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/addSeconds(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getMillis()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getMillis()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMillis(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setMillis(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "addMillis(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/addMillis(I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getTimeOfDay()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/CalendarDate/getTimeOfDay()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setDate(III)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setDate(III)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "addDate(III)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/addDate(III)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "setTimeOfDay(IIII)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setTimeOfDay(IIII)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "addTimeOfDay(IIII)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/addTimeOfDay(IIII)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "setTimeOfDay(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/setTimeOfDay(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "isNormalized()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/isNormalized()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isStandardTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/isStandardTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setStandardTime(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/setStandardTime(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isDaylightTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/isDaylightTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLocale(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/setLocale(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "getZone()Ljava/util/TimeZone;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public "sun/util/calendar/CalendarDate/getZone()Ljava/util/TimeZone;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public "setZone(Ljava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarDate/setZone(Ljava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "isSameDate(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/isSameDate(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/util/calendar/CalendarDate/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/calendar/CalendarDate/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setDayOfWeek(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/setDayOfWeek(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setNormalized(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/setNormalized(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getZoneOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getZoneOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setZoneOffset(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/setZoneOffset(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDaylightSaving()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarDate/getDaylightSaving()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDaylightSaving(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarDate/setDaylightSaving(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_calendar_Era extends java_lang_Object {
    public "sun/util/calendar/Era/name": java_lang_String;
    public "sun/util/calendar/Era/abbr": java_lang_String;
    public "sun/util/calendar/Era/since": Long;
    public "sun/util/calendar/Era/sinceDate": sun_util_calendar_CalendarDate;
    public "sun/util/calendar/Era/localTime": number;
    public "sun/util/calendar/Era/hash": number;
    public "<init>(Ljava/lang/String;Ljava/lang/String;JZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/Era/<init>(Ljava/lang/String;Ljava/lang/String;JZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/calendar/Era/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/calendar/Era/getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAbbreviation()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/calendar/Era/getAbbreviation()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDiaplayAbbreviation(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/calendar/Era/getDiaplayAbbreviation(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSince(Ljava/util/TimeZone;)J"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/Era/getSince(Ljava/util/TimeZone;)J"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getSinceDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/Era/getSinceDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "isLocalTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/Era/isLocalTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/Era/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/Era/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/calendar/Era/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_util_calendar_BaseCalendar extends sun_util_calendar_AbstractCalendar {
    public static "sun/util/calendar/BaseCalendar/JANUARY": number;
    public static "sun/util/calendar/BaseCalendar/FEBRUARY": number;
    public static "sun/util/calendar/BaseCalendar/MARCH": number;
    public static "sun/util/calendar/BaseCalendar/APRIL": number;
    public static "sun/util/calendar/BaseCalendar/MAY": number;
    public static "sun/util/calendar/BaseCalendar/JUNE": number;
    public static "sun/util/calendar/BaseCalendar/JULY": number;
    public static "sun/util/calendar/BaseCalendar/AUGUST": number;
    public static "sun/util/calendar/BaseCalendar/SEPTEMBER": number;
    public static "sun/util/calendar/BaseCalendar/OCTOBER": number;
    public static "sun/util/calendar/BaseCalendar/NOVEMBER": number;
    public static "sun/util/calendar/BaseCalendar/DECEMBER": number;
    public static "sun/util/calendar/BaseCalendar/SUNDAY": number;
    public static "sun/util/calendar/BaseCalendar/MONDAY": number;
    public static "sun/util/calendar/BaseCalendar/TUESDAY": number;
    public static "sun/util/calendar/BaseCalendar/WEDNESDAY": number;
    public static "sun/util/calendar/BaseCalendar/THURSDAY": number;
    public static "sun/util/calendar/BaseCalendar/FRIDAY": number;
    public static "sun/util/calendar/BaseCalendar/SATURDAY": number;
    public static "sun/util/calendar/BaseCalendar/BASE_YEAR": number;
    public static "sun/util/calendar/BaseCalendar/FIXED_DATES": JVMArray<number>;
    public static "sun/util/calendar/BaseCalendar/DAYS_IN_MONTH": JVMArray<number>;
    public static "sun/util/calendar/BaseCalendar/ACCUMULATED_DAYS_IN_MONTH": JVMArray<number>;
    public static "sun/util/calendar/BaseCalendar/ACCUMULATED_DAYS_IN_MONTH_LEAP": JVMArray<number>;
    public static "sun/util/calendar/BaseCalendar/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/BaseCalendar/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "validate(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/validate(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "normalize(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/normalize(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "normalizeMonth(Lsun/util/calendar/CalendarDate;)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/BaseCalendar/normalizeMonth(Lsun/util/calendar/CalendarDate;)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable) => void): void;
    public "getYearLength(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/getYearLength(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getYearLengthInMonths(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/getYearLengthInMonths(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonthLength(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/getMonthLength(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonthLength(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/getMonthLength(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDayOfYear(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/BaseCalendar/getDayOfYear(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getDayOfYear(III)J"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/BaseCalendar/getDayOfYear(III)J"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFixedDate(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/BaseCalendar/getFixedDate(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFixedDate(IIILsun/util/calendar/BaseCalendar$Date;)J"(thread: JVMThread, args: [number, number, number, sun_util_calendar_BaseCalendar$Date], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/BaseCalendar/getFixedDate(IIILsun/util/calendar/BaseCalendar$Date;)J"(thread: JVMThread, args: [number, number, number, sun_util_calendar_BaseCalendar$Date], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getCalendarDateFromFixedDate(Lsun/util/calendar/CalendarDate;J)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/BaseCalendar/getCalendarDateFromFixedDate(Lsun/util/calendar/CalendarDate;J)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDayOfWeek(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/getDayOfWeek(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDayOfWeekFromFixedDate(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/util/calendar/BaseCalendar/getDayOfWeekFromFixedDate(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getYearFromFixedDate(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/getYearFromFixedDate(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGregorianYearFromFixedDate(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/getGregorianYearFromFixedDate(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLeapYear(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/isLeapYear(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLeapYear(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/BaseCalendar/isLeapYear(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/calendar/BaseCalendar/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_calendar_AbstractCalendar extends sun_util_calendar_CalendarSystem {
    public static "sun/util/calendar/AbstractCalendar/SECOND_IN_MILLIS": number;
    public static "sun/util/calendar/AbstractCalendar/MINUTE_IN_MILLIS": number;
    public static "sun/util/calendar/AbstractCalendar/HOUR_IN_MILLIS": number;
    public static "sun/util/calendar/AbstractCalendar/DAY_IN_MILLIS": number;
    public static "sun/util/calendar/AbstractCalendar/EPOCH_OFFSET": number;
    public "sun/util/calendar/AbstractCalendar/eras": JVMArray<sun_util_calendar_Era>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/AbstractCalendar/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEra(Ljava/lang/String;)Lsun/util/calendar/Era;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Era) => void): void;
    public "sun/util/calendar/AbstractCalendar/getEra(Ljava/lang/String;)Lsun/util/calendar/Era;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Era) => void): void;
    public "getEras()[Lsun/util/calendar/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_util_calendar_Era>) => void): void;
    public "sun/util/calendar/AbstractCalendar/getEras()[Lsun/util/calendar/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_util_calendar_Era>) => void): void;
    public "setEra(Lsun/util/calendar/CalendarDate;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/AbstractCalendar/setEra(Lsun/util/calendar/CalendarDate;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setEras([Lsun/util/calendar/Era;)V"(thread: JVMThread, args: [JVMArray<sun_util_calendar_Era>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/AbstractCalendar/setEras([Lsun/util/calendar/Era;)V"(thread: JVMThread, args: [JVMArray<sun_util_calendar_Era>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/AbstractCalendar/getCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getCalendarDate(J)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/AbstractCalendar/getCalendarDate(J)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getCalendarDate(JLjava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/AbstractCalendar/getCalendarDate(JLjava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getCalendarDate(JLsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/AbstractCalendar/getCalendarDate(JLsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getTime(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/AbstractCalendar/getTime(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTimeOfDay(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/AbstractCalendar/getTimeOfDay(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTimeOfDayValue(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/AbstractCalendar/getTimeOfDayValue(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setTimeOfDay(Lsun/util/calendar/CalendarDate;I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/AbstractCalendar/setTimeOfDay(Lsun/util/calendar/CalendarDate;I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getWeekLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/AbstractCalendar/getWeekLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLeapYear(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/AbstractCalendar/isLeapYear(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNthDayOfWeek(IILsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/AbstractCalendar/getNthDayOfWeek(IILsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public static "getDayOfWeekDateBefore(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/util/calendar/AbstractCalendar/getDayOfWeekDateBefore(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getDayOfWeekDateAfter(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/util/calendar/AbstractCalendar/getDayOfWeekDateAfter(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getDayOfWeekDateOnOrBefore(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/util/calendar/AbstractCalendar/getDayOfWeekDateOnOrBefore(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFixedDate(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/AbstractCalendar/getFixedDate(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getCalendarDateFromFixedDate(Lsun/util/calendar/CalendarDate;J)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/AbstractCalendar/getCalendarDateFromFixedDate(Lsun/util/calendar/CalendarDate;J)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "validateTime(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/AbstractCalendar/validateTime(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "normalizeTime(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/AbstractCalendar/normalizeTime(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_util_calendar_CalendarSystem extends java_lang_Object {
    public static "sun/util/calendar/CalendarSystem/initialized": number;
    public static "sun/util/calendar/CalendarSystem/names": java_util_concurrent_ConcurrentMap;
    public static "sun/util/calendar/CalendarSystem/calendars": java_util_concurrent_ConcurrentMap;
    public static "sun/util/calendar/CalendarSystem/PACKAGE_NAME": java_lang_String;
    public static "sun/util/calendar/CalendarSystem/namePairs": JVMArray<java_lang_String>;
    public static "sun/util/calendar/CalendarSystem/GREGORIAN_INSTANCE": sun_util_calendar_Gregorian;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarSystem/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initNames()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/calendar/CalendarSystem/initNames()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getGregorianCalendar()Lsun/util/calendar/Gregorian;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian) => void): void;
    public static "sun/util/calendar/CalendarSystem/getGregorianCalendar()Lsun/util/calendar/Gregorian;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian) => void): void;
    public static "forName(Ljava/lang/String;)Lsun/util/calendar/CalendarSystem;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarSystem) => void): void;
    public static "sun/util/calendar/CalendarSystem/forName(Ljava/lang/String;)Lsun/util/calendar/CalendarSystem;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarSystem) => void): void;
    public static "getCalendarProperties()Ljava/util/Properties;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "sun/util/calendar/CalendarSystem/getCalendarProperties()Ljava/util/Properties;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/calendar/CalendarSystem/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarSystem/getCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getCalendarDate(J)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarSystem/getCalendarDate(J)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getCalendarDate(JLsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarSystem/getCalendarDate(JLsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getCalendarDate(JLjava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarSystem/getCalendarDate(JLjava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "newCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarSystem/newCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "newCalendarDate(Ljava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarSystem/newCalendarDate(Ljava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getTime(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/util/calendar/CalendarSystem/getTime(Lsun/util/calendar/CalendarDate;)J"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getYearLength(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarSystem/getYearLength(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getYearLengthInMonths(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarSystem/getYearLengthInMonths(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMonthLength(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarSystem/getMonthLength(Lsun/util/calendar/CalendarDate;)I"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWeekLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarSystem/getWeekLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getEra(Ljava/lang/String;)Lsun/util/calendar/Era;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Era) => void): void;
    public "sun/util/calendar/CalendarSystem/getEra(Ljava/lang/String;)Lsun/util/calendar/Era;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Era) => void): void;
    public "getEras()[Lsun/util/calendar/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_util_calendar_Era>) => void): void;
    public "sun/util/calendar/CalendarSystem/getEras()[Lsun/util/calendar/Era;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_util_calendar_Era>) => void): void;
    public "setEra(Lsun/util/calendar/CalendarDate;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/CalendarSystem/setEra(Lsun/util/calendar/CalendarDate;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNthDayOfWeek(IILsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarSystem/getNthDayOfWeek(IILsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [number, number, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "setTimeOfDay(Lsun/util/calendar/CalendarDate;I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/CalendarSystem/setTimeOfDay(Lsun/util/calendar/CalendarDate;I)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [sun_util_calendar_CalendarDate, number], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "validate(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarSystem/validate(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "normalize(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/CalendarSystem/normalize(Lsun/util/calendar/CalendarDate;)Z"(thread: JVMThread, args: [sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/calendar/CalendarSystem/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_calendar_Gregorian extends sun_util_calendar_BaseCalendar {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/Gregorian/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/calendar/Gregorian/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCalendarDate()Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "sun/util/calendar/Gregorian/getCalendarDate()Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "getCalendarDate(J)Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "sun/util/calendar/Gregorian/getCalendarDate(J)Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "getCalendarDate(JLsun/util/calendar/CalendarDate;)Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: [Long, any, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "sun/util/calendar/Gregorian/getCalendarDate(JLsun/util/calendar/CalendarDate;)Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: [Long, any, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "getCalendarDate(JLjava/util/TimeZone;)Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: [Long, any, java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "sun/util/calendar/Gregorian/getCalendarDate(JLjava/util/TimeZone;)Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: [Long, any, java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "newCalendarDate()Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "sun/util/calendar/Gregorian/newCalendarDate()Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "newCalendarDate(Ljava/util/TimeZone;)Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "sun/util/calendar/Gregorian/newCalendarDate(Ljava/util/TimeZone;)Lsun/util/calendar/Gregorian$Date;"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_Gregorian$Date) => void): void;
    public "getCalendarDate(JLsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/Gregorian/getCalendarDate(JLsun/util/calendar/CalendarDate;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, sun_util_calendar_CalendarDate], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getCalendarDate(JLjava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/Gregorian/getCalendarDate(JLjava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any, java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getCalendarDate(J)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/Gregorian/getCalendarDate(J)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "getCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/Gregorian/getCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "newCalendarDate(Ljava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/Gregorian/newCalendarDate(Ljava/util/TimeZone;)Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "newCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
    public "sun/util/calendar/Gregorian/newCalendarDate()Lsun/util/calendar/CalendarDate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_calendar_CalendarDate) => void): void;
  }
  export class sun_util_calendar_Gregorian$Date extends sun_util_calendar_BaseCalendar$Date {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/Gregorian$Date/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/TimeZone;)V"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/Gregorian$Date/<init>(Ljava/util/TimeZone;)V"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNormalizedYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/calendar/Gregorian$Date/getNormalizedYear()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setNormalizedYear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/calendar/Gregorian$Date/setNormalizedYear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_locks_AbstractQueuedSynchronizer$Node extends java_lang_Object {
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/SHARED": java_util_concurrent_locks_AbstractQueuedSynchronizer$Node;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/EXCLUSIVE": java_util_concurrent_locks_AbstractQueuedSynchronizer$Node;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/CANCELLED": number;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/SIGNAL": number;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/CONDITION": number;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/PROPAGATE": number;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/waitStatus": number;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/prev": java_util_concurrent_locks_AbstractQueuedSynchronizer$Node;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/next": java_util_concurrent_locks_AbstractQueuedSynchronizer$Node;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/thread": java_lang_Thread;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/nextWaiter": java_util_concurrent_locks_AbstractQueuedSynchronizer$Node;
    public "isShared()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/isShared()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "predecessor()Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$Node) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/predecessor()Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$Node) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Thread;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_lang_Thread, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/<init>(Ljava/lang/Thread;Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$Node;)V"(thread: JVMThread, args: [java_lang_Thread, java_util_concurrent_locks_AbstractQueuedSynchronizer$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Thread;I)V"(thread: JVMThread, args: [java_lang_Thread, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/<init>(Ljava/lang/Thread;I)V"(thread: JVMThread, args: [java_lang_Thread, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/locks/AbstractQueuedSynchronizer$Node/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_locks_AbstractOwnableSynchronizer extends java_lang_Object implements java_io_Serializable {
    public static "java/util/concurrent/locks/AbstractOwnableSynchronizer/serialVersionUID": Long;
    public "java/util/concurrent/locks/AbstractOwnableSynchronizer/exclusiveOwnerThread": java_lang_Thread;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractOwnableSynchronizer/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setExclusiveOwnerThread(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/AbstractOwnableSynchronizer/setExclusiveOwnerThread(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "getExclusiveOwnerThread()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public "java/util/concurrent/locks/AbstractOwnableSynchronizer/getExclusiveOwnerThread()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
  }
  export class java_lang_RuntimePermission extends java_security_BasicPermission {
    public static "java/lang/RuntimePermission/serialVersionUID": Long;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/RuntimePermission/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/RuntimePermission/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_BasicPermission extends java_security_Permission implements java_io_Serializable {
    public static "java/security/BasicPermission/serialVersionUID": Long;
    public "java/security/BasicPermission/wildcard": number;
    public "java/security/BasicPermission/path": java_lang_String;
    public "java/security/BasicPermission/exitVM": number;
    public "init(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/BasicPermission/init(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/BasicPermission/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/BasicPermission/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/BasicPermission/implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/BasicPermission/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/BasicPermission/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getActions()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/BasicPermission/getActions()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "newPermissionCollection()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public "java/security/BasicPermission/newPermissionCollection()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/BasicPermission/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCanonicalName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/BasicPermission/getCanonicalName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_security_PermissionCollection extends java_lang_Object implements java_io_Serializable {
    public static "java/security/PermissionCollection/serialVersionUID": Long;
    public "java/security/PermissionCollection/readOnly": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/PermissionCollection/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/PermissionCollection/add(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
    public "implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/PermissionCollection/implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/security/PermissionCollection/elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "setReadOnly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/PermissionCollection/setReadOnly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/PermissionCollection/isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/PermissionCollection/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_security_Permission extends java_lang_Object implements java_security_Guard, java_io_Serializable {
    public static "java/security/Permission/serialVersionUID": Long;
    public "java/security/Permission/name": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Permission/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkGuard(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Permission/checkGuard(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Permission/implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Permission/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Permission/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Permission/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getActions()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Permission/getActions()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "newPermissionCollection()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public "java/security/Permission/newPermissionCollection()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Permission/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_security_Guard extends java_lang_Object {
    "checkGuard(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_concurrent_locks_Condition extends java_lang_Object {
    "await()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "awaitUninterruptibly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "awaitNanos(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "await(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "awaitUntil(Ljava/util/Date;)Z"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "signal()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "signalAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_locks_ReentrantLock extends java_lang_Object implements java_util_concurrent_locks_Lock, java_io_Serializable {
    public static "java/util/concurrent/locks/ReentrantLock/serialVersionUID": Long;
    public "java/util/concurrent/locks/ReentrantLock/sync": java_util_concurrent_locks_ReentrantLock$Sync;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "lock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/lock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lockInterruptibly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/lockInterruptibly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "tryLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/tryLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryLock(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/tryLock(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unlock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/unlock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "newCondition()Ljava/util/concurrent/locks/Condition;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_Condition) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/newCondition()Ljava/util/concurrent/locks/Condition;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_Condition) => void): void;
    public "getHoldCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/getHoldCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isHeldByCurrentThread()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/isHeldByCurrentThread()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLocked()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/isLocked()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFair()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/isFair()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOwner()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/getOwner()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public "hasQueuedThreads()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/hasQueuedThreads()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasQueuedThread(Ljava/lang/Thread;)Z"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/hasQueuedThread(Ljava/lang/Thread;)Z"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getQueueLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/getQueueLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/getQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "hasWaiters(Ljava/util/concurrent/locks/Condition;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_Condition], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/hasWaiters(Ljava/util/concurrent/locks/Condition;)Z"(thread: JVMThread, args: [java_util_concurrent_locks_Condition], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWaitQueueLength(Ljava/util/concurrent/locks/Condition;)I"(thread: JVMThread, args: [java_util_concurrent_locks_Condition], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/getWaitQueueLength(Ljava/util/concurrent/locks/Condition;)I"(thread: JVMThread, args: [java_util_concurrent_locks_Condition], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWaitingThreads(Ljava/util/concurrent/locks/Condition;)Ljava/util/Collection;"(thread: JVMThread, args: [java_util_concurrent_locks_Condition], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/getWaitingThreads(Ljava/util/concurrent/locks/Condition;)Ljava/util/Collection;"(thread: JVMThread, args: [java_util_concurrent_locks_Condition], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/locks/ReentrantLock/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_util_concurrent_locks_ReentrantLock$Sync extends java_util_concurrent_locks_AbstractQueuedSynchronizer {
    public static "java/util/concurrent/locks/ReentrantLock$Sync/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/lock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "nonfairTryAcquire(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/nonfairTryAcquire(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryRelease(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/tryRelease(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isHeldExclusively()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/isHeldExclusively()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "newCondition()Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/newCondition()Ljava/util/concurrent/locks/AbstractQueuedSynchronizer$ConditionObject;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_AbstractQueuedSynchronizer$ConditionObject) => void): void;
    public "getOwner()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/getOwner()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public "getHoldCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/getHoldCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLocked()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/isLocked()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/locks/ReentrantLock$Sync/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_concurrent_locks_Lock extends java_lang_Object {
    "lock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "lockInterruptibly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "tryLock()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "tryLock(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "unlock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "newCondition()Ljava/util/concurrent/locks/Condition;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_Condition) => void): void;
  }
  export class java_util_concurrent_AbstractExecutorService extends java_lang_Object implements java_util_concurrent_ExecutorService {
    public static "java/util/concurrent/AbstractExecutorService/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/AbstractExecutorService/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "newTaskFor(Ljava/lang/Runnable;Ljava/lang/Object;)Ljava/util/concurrent/RunnableFuture;"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RunnableFuture) => void): void;
    public "java/util/concurrent/AbstractExecutorService/newTaskFor(Ljava/lang/Runnable;Ljava/lang/Object;)Ljava/util/concurrent/RunnableFuture;"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RunnableFuture) => void): void;
    public "newTaskFor(Ljava/util/concurrent/Callable;)Ljava/util/concurrent/RunnableFuture;"(thread: JVMThread, args: [java_util_concurrent_Callable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RunnableFuture) => void): void;
    public "java/util/concurrent/AbstractExecutorService/newTaskFor(Ljava/util/concurrent/Callable;)Ljava/util/concurrent/RunnableFuture;"(thread: JVMThread, args: [java_util_concurrent_Callable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_RunnableFuture) => void): void;
    public "submit(Ljava/lang/Runnable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/util/concurrent/AbstractExecutorService/submit(Ljava/lang/Runnable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "submit(Ljava/lang/Runnable;Ljava/lang/Object;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/util/concurrent/AbstractExecutorService/submit(Ljava/lang/Runnable;Ljava/lang/Object;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Runnable, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "submit(Ljava/util/concurrent/Callable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_util_concurrent_Callable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "java/util/concurrent/AbstractExecutorService/submit(Ljava/util/concurrent/Callable;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_util_concurrent_Callable], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "doInvokeAny(Ljava/util/Collection;ZJ)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/AbstractExecutorService/doInvokeAny(Ljava/util/Collection;ZJ)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invokeAny(Ljava/util/Collection;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/AbstractExecutorService/invokeAny(Ljava/util/Collection;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invokeAny(Ljava/util/Collection;JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/AbstractExecutorService/invokeAny(Ljava/util/Collection;JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_Collection, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invokeAll(Ljava/util/Collection;)Ljava/util/List;"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/concurrent/AbstractExecutorService/invokeAll(Ljava/util/Collection;)Ljava/util/List;"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "invokeAll(Ljava/util/Collection;JLjava/util/concurrent/TimeUnit;)Ljava/util/List;"(thread: JVMThread, args: [java_util_Collection, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/concurrent/AbstractExecutorService/invokeAll(Ljava/util/Collection;JLjava/util/concurrent/TimeUnit;)Ljava/util/List;"(thread: JVMThread, args: [java_util_Collection, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/AbstractExecutorService/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "shutdownNow()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "isShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isTerminated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "awaitTermination(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "execute(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_ThreadPool extends java_lang_Object {
    public static "sun/nio/ch/ThreadPool/DEFAULT_THREAD_POOL_THREAD_FACTORY": java_lang_String;
    public static "sun/nio/ch/ThreadPool/DEFAULT_THREAD_POOL_INITIAL_SIZE": java_lang_String;
    public "sun/nio/ch/ThreadPool/executor": java_util_concurrent_ExecutorService;
    public "sun/nio/ch/ThreadPool/isFixed": number;
    public "sun/nio/ch/ThreadPool/poolSize": number;
    public "<init>(Ljava/util/concurrent/ExecutorService;ZI)V"(thread: JVMThread, args: [java_util_concurrent_ExecutorService, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/ThreadPool/<init>(Ljava/util/concurrent/ExecutorService;ZI)V"(thread: JVMThread, args: [java_util_concurrent_ExecutorService, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "executor()Ljava/util/concurrent/ExecutorService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ExecutorService) => void): void;
    public "sun/nio/ch/ThreadPool/executor()Ljava/util/concurrent/ExecutorService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ExecutorService) => void): void;
    public "isFixedThreadPool()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ThreadPool/isFixedThreadPool()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "poolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ThreadPool/poolSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "defaultThreadFactory()Ljava/util/concurrent/ThreadFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ThreadFactory) => void): void;
    public static "sun/nio/ch/ThreadPool/defaultThreadFactory()Ljava/util/concurrent/ThreadFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ThreadFactory) => void): void;
    public static "getDefault()Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "sun/nio/ch/ThreadPool/getDefault()Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "createDefault()Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "sun/nio/ch/ThreadPool/createDefault()Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "create(ILjava/util/concurrent/ThreadFactory;)Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "sun/nio/ch/ThreadPool/create(ILjava/util/concurrent/ThreadFactory;)Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "wrap(Ljava/util/concurrent/ExecutorService;I)Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: [java_util_concurrent_ExecutorService, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "sun/nio/ch/ThreadPool/wrap(Ljava/util/concurrent/ExecutorService;I)Lsun/nio/ch/ThreadPool;"(thread: JVMThread, args: [java_util_concurrent_ExecutorService, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_ThreadPool) => void): void;
    public static "getDefaultThreadPoolInitialSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/ThreadPool/getDefaultThreadPoolInitialSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDefaultThreadPoolThreadFactory()Ljava/util/concurrent/ThreadFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ThreadFactory) => void): void;
    public static "sun/nio/ch/ThreadPool/getDefaultThreadPoolThreadFactory()Ljava/util/concurrent/ThreadFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ThreadFactory) => void): void;
    public static "lambda$defaultThreadFactory$2(Ljava/lang/Runnable;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "sun/nio/ch/ThreadPool/lambda$defaultThreadFactory$2(Ljava/lang/Runnable;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "lambda$null$1(Ljava/lang/Runnable;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "sun/nio/ch/ThreadPool/lambda$null$1(Ljava/lang/Runnable;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "lambda$defaultThreadFactory$0(Ljava/lang/Runnable;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "sun/nio/ch/ThreadPool/lambda$defaultThreadFactory$0(Ljava/lang/Runnable;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
  }
  export class java_nio_channels_AsynchronousChannelGroup extends java_lang_Object {
    public "java/nio/channels/AsynchronousChannelGroup/provider": java_nio_channels_spi_AsynchronousChannelProvider;
    public "<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousChannelGroup/<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "provider()Ljava/nio/channels/spi/AsynchronousChannelProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AsynchronousChannelProvider) => void): void;
    public "java/nio/channels/AsynchronousChannelGroup/provider()Ljava/nio/channels/spi/AsynchronousChannelProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_spi_AsynchronousChannelProvider) => void): void;
    public static "withFixedThreadPool(ILjava/util/concurrent/ThreadFactory;)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public static "java/nio/channels/AsynchronousChannelGroup/withFixedThreadPool(ILjava/util/concurrent/ThreadFactory;)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [number, java_util_concurrent_ThreadFactory], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public static "withCachedThreadPool(Ljava/util/concurrent/ExecutorService;I)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [java_util_concurrent_ExecutorService, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public static "java/nio/channels/AsynchronousChannelGroup/withCachedThreadPool(Ljava/util/concurrent/ExecutorService;I)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [java_util_concurrent_ExecutorService, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public static "withThreadPool(Ljava/util/concurrent/ExecutorService;)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [java_util_concurrent_ExecutorService], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public static "java/nio/channels/AsynchronousChannelGroup/withThreadPool(Ljava/util/concurrent/ExecutorService;)Ljava/nio/channels/AsynchronousChannelGroup;"(thread: JVMThread, args: [java_util_concurrent_ExecutorService], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannelGroup) => void): void;
    public "isShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/AsynchronousChannelGroup/isShutdown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isTerminated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/AsynchronousChannelGroup/isTerminated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousChannelGroup/shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownNow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/channels/AsynchronousChannelGroup/shutdownNow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "awaitTermination(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/channels/AsynchronousChannelGroup/awaitTermination(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_nio_ch_PendingFuture extends java_lang_Object implements java_util_concurrent_Future {
    public static "sun/nio/ch/PendingFuture/CANCELLED": java_util_concurrent_CancellationException;
    public "sun/nio/ch/PendingFuture/channel": java_nio_channels_AsynchronousChannel;
    public "sun/nio/ch/PendingFuture/handler": java_nio_channels_CompletionHandler;
    public "sun/nio/ch/PendingFuture/attachment": java_lang_Object;
    public "sun/nio/ch/PendingFuture/haveResult": number;
    public "sun/nio/ch/PendingFuture/result": java_lang_Object;
    public "sun/nio/ch/PendingFuture/exc": java_lang_Throwable;
    public "sun/nio/ch/PendingFuture/latch": java_util_concurrent_CountDownLatch;
    public "sun/nio/ch/PendingFuture/timeoutTask": java_util_concurrent_Future;
    public "sun/nio/ch/PendingFuture/context": java_lang_Object;
    public "<init>(Ljava/nio/channels/AsynchronousChannel;Ljava/nio/channels/CompletionHandler;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannel, java_nio_channels_CompletionHandler, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/<init>(Ljava/nio/channels/AsynchronousChannel;Ljava/nio/channels/CompletionHandler;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannel, java_nio_channels_CompletionHandler, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/AsynchronousChannel;Ljava/nio/channels/CompletionHandler;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannel, java_nio_channels_CompletionHandler, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/<init>(Ljava/nio/channels/AsynchronousChannel;Ljava/nio/channels/CompletionHandler;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannel, java_nio_channels_CompletionHandler, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/AsynchronousChannel;)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannel], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/<init>(Ljava/nio/channels/AsynchronousChannel;)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannel], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/AsynchronousChannel;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannel, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/<init>(Ljava/nio/channels/AsynchronousChannel;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_nio_channels_AsynchronousChannel, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "channel()Ljava/nio/channels/AsynchronousChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannel) => void): void;
    public "sun/nio/ch/PendingFuture/channel()Ljava/nio/channels/AsynchronousChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousChannel) => void): void;
    public "handler()Ljava/nio/channels/CompletionHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_CompletionHandler) => void): void;
    public "sun/nio/ch/PendingFuture/handler()Ljava/nio/channels/CompletionHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_CompletionHandler) => void): void;
    public "attachment()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/PendingFuture/attachment()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setContext(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/setContext(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContext()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/PendingFuture/getContext()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setTimeoutTask(Ljava/util/concurrent/Future;)V"(thread: JVMThread, args: [java_util_concurrent_Future], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/setTimeoutTask(Ljava/util/concurrent/Future;)V"(thread: JVMThread, args: [java_util_concurrent_Future], cb?: (e?: java_lang_Throwable) => void): void;
    public "prepareForWait()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/PendingFuture/prepareForWait()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setResult(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/setResult(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFailure(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/setFailure(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "setResult(Ljava/lang/Object;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/setResult(Ljava/lang/Object;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/PendingFuture/get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "get(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/PendingFuture/get(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "exception()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PendingFuture/exception()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "value()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/PendingFuture/value()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isCancelled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/PendingFuture/isCancelled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/PendingFuture/isDone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "cancel(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/PendingFuture/cancel(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/PendingFuture/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_CountDownLatch extends java_lang_Object {
    public "java/util/concurrent/CountDownLatch/sync": java_util_concurrent_CountDownLatch$Sync;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CountDownLatch/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "await()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CountDownLatch/await()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "await(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CountDownLatch/await(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countDown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CountDownLatch/countDown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/CountDownLatch/getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/CountDownLatch/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_util_concurrent_CountDownLatch$Sync extends java_util_concurrent_locks_AbstractQueuedSynchronizer {
    public static "java/util/concurrent/CountDownLatch$Sync/serialVersionUID": Long;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CountDownLatch$Sync/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CountDownLatch$Sync/getCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryAcquireShared(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CountDownLatch$Sync/tryAcquireShared(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryReleaseShared(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CountDownLatch$Sync/tryReleaseShared(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_concurrent_CancellationException extends java_lang_IllegalStateException {
    public static "java/util/concurrent/CancellationException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CancellationException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CancellationException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_IllegalStateException extends java_lang_RuntimeException {
    public static "java/lang/IllegalStateException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalStateException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalStateException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalStateException/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalStateException/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_Port extends sun_nio_ch_AsynchronousChannelGroupImpl {
    public "sun/nio/ch/Port/fdToChannelLock": java_util_concurrent_locks_ReadWriteLock;
    public "sun/nio/ch/Port/fdToChannel": java_util_Map;
    public "<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;Lsun/nio/ch/ThreadPool;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider, sun_nio_ch_ThreadPool], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/Port/<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;Lsun/nio/ch/ThreadPool;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider, sun_nio_ch_ThreadPool], cb?: (e?: java_lang_Throwable) => void): void;
    public "register(ILsun/nio/ch/Port$PollableChannel;)V"(thread: JVMThread, args: [number, sun_nio_ch_Port$PollableChannel], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/Port/register(ILsun/nio/ch/Port$PollableChannel;)V"(thread: JVMThread, args: [number, sun_nio_ch_Port$PollableChannel], cb?: (e?: java_lang_Throwable) => void): void;
    public "preUnregister(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/Port/preUnregister(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "unregister(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/Port/unregister(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "startPoll(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/Port/startPoll(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/Port/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "attachForeignChannel(Ljava/nio/channels/Channel;Ljava/io/FileDescriptor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_channels_Channel, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/Port/attachForeignChannel(Ljava/nio/channels/Channel;Ljava/io/FileDescriptor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_channels_Channel, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "detachForeignChannel(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/Port/detachForeignChannel(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "closeAllChannels()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/Port/closeAllChannels()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_concurrent_locks_ReadWriteLock extends java_lang_Object {
    "readLock()Ljava/util/concurrent/locks/Lock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_Lock) => void): void;
    "writeLock()Ljava/util/concurrent/locks/Lock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_Lock) => void): void;
  }
  export interface sun_nio_ch_Port$PollableChannel extends java_lang_Object, java_io_Closeable {
    "onEvent(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_AsynchronousSocketChannelImpl extends java_nio_channels_AsynchronousSocketChannel implements sun_nio_ch_Cancellable, sun_nio_ch_Groupable {
    public "sun/nio/ch/AsynchronousSocketChannelImpl/fd": java_io_FileDescriptor;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/stateLock": java_lang_Object;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/localAddress": java_net_InetSocketAddress;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/remoteAddress": java_net_InetSocketAddress;
    public static "sun/nio/ch/AsynchronousSocketChannelImpl/ST_UNINITIALIZED": number;
    public static "sun/nio/ch/AsynchronousSocketChannelImpl/ST_UNCONNECTED": number;
    public static "sun/nio/ch/AsynchronousSocketChannelImpl/ST_PENDING": number;
    public static "sun/nio/ch/AsynchronousSocketChannelImpl/ST_CONNECTED": number;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/state": number;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/readLock": java_lang_Object;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/reading": number;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/readShutdown": number;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/readKilled": number;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/writeLock": java_lang_Object;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/writing": number;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/writeShutdown": number;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/writeKilled": number;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/closeLock": java_util_concurrent_locks_ReadWriteLock;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/open": number;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/isReuseAddress": number;
    public "<init>(Lsun/nio/ch/AsynchronousChannelGroupImpl;)V"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/<init>(Lsun/nio/ch/AsynchronousChannelGroupImpl;)V"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/nio/ch/AsynchronousChannelGroupImpl;Ljava/io/FileDescriptor;Ljava/net/InetSocketAddress;)V"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl, java_io_FileDescriptor, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/<init>(Lsun/nio/ch/AsynchronousChannelGroupImpl;Ljava/io/FileDescriptor;Ljava/net/InetSocketAddress;)V"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl, java_io_FileDescriptor, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "begin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/begin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "enableReading(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/enableReading(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "enableReading()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/enableReading()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "enableWriting(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/enableWriting(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "enableWriting()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/enableWriting()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "killReading()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/killReading()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "killWriting()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/killWriting()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "killConnect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/killConnect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implConnect(Ljava/net/SocketAddress;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/implConnect(Ljava/net/SocketAddress;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "connect(Ljava/net/SocketAddress;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/connect(Ljava/net/SocketAddress;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "connect(Ljava/net/SocketAddress;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/connect(Ljava/net/SocketAddress;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "implRead(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/implRead(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "read(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/read(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "read(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/read(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "read(Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/read(Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "read([Ljava/nio/ByteBuffer;IIJLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/read([Ljava/nio/ByteBuffer;IIJLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "implWrite(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/implWrite(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "write(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/write(ZLjava/nio/ByteBuffer;[Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [number, java_nio_ByteBuffer, JVMArray<java_nio_ByteBuffer>, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "write(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/write(Ljava/nio/ByteBuffer;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "write(Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/write(Ljava/nio/ByteBuffer;JLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_nio_ByteBuffer, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([Ljava/nio/ByteBuffer;IIJLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/write([Ljava/nio/ByteBuffer;IIJLjava/util/concurrent/TimeUnit;Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number, Long, any, java_util_concurrent_TimeUnit, java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "shutdownInput()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/shutdownInput()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "shutdownOutput()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/shutdownOutput()Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "sun/nio/ch/AsynchronousSocketChannelImpl/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "onCancel(Lsun/nio/ch/PendingFuture;)V"(thread: JVMThread, args: [sun_nio_ch_PendingFuture], cb?: (e?: java_lang_Throwable) => void): void;
    "group()Lsun/nio/ch/AsynchronousChannelGroupImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_AsynchronousChannelGroupImpl) => void): void;
  }
  export interface sun_nio_ch_Groupable extends java_lang_Object {
    "group()Lsun/nio/ch/AsynchronousChannelGroupImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_AsynchronousChannelGroupImpl) => void): void;
  }
  export interface sun_nio_ch_Cancellable extends java_lang_Object {
    "onCancel(Lsun/nio/ch/PendingFuture;)V"(thread: JVMThread, args: [sun_nio_ch_PendingFuture], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_UnixAsynchronousServerSocketChannelImpl extends sun_nio_ch_AsynchronousServerSocketChannelImpl implements sun_nio_ch_Port$PollableChannel {
    public static "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/nd": sun_nio_ch_NativeDispatcher;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/port": sun_nio_ch_Port;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/fdVal": number;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/accepting": java_util_concurrent_atomic_AtomicBoolean;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/updateLock": java_lang_Object;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/acceptPending": number;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/acceptHandler": java_nio_channels_CompletionHandler;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/acceptAttachment": java_lang_Object;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/acceptFuture": sun_nio_ch_PendingFuture;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/acceptAcc": java_security_AccessControlContext;
    public "enableAccept()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/enableAccept()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/nio/ch/Port;)V"(thread: JVMThread, args: [sun_nio_ch_Port], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/<init>(Lsun/nio/ch/Port;)V"(thread: JVMThread, args: [sun_nio_ch_Port], cb?: (e?: java_lang_Throwable) => void): void;
    public "implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "group()Lsun/nio/ch/AsynchronousChannelGroupImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_AsynchronousChannelGroupImpl) => void): void;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/group()Lsun/nio/ch/AsynchronousChannelGroupImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_AsynchronousChannelGroupImpl) => void): void;
    public "onEvent(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/onEvent(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "finishAccept(Ljava/io/FileDescriptor;Ljava/net/InetSocketAddress;Ljava/security/AccessControlContext;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_InetSocketAddress, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/finishAccept(Ljava/io/FileDescriptor;Ljava/net/InetSocketAddress;Ljava/security/AccessControlContext;)Ljava/nio/channels/AsynchronousSocketChannel;"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_InetSocketAddress, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousSocketChannel) => void): void;
    public "implAccept(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/implAccept(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "accept(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/accept(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "accept0(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/accept0(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/UnixAsynchronousServerSocketChannelImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_AsynchronousServerSocketChannelImpl extends java_nio_channels_AsynchronousServerSocketChannel implements sun_nio_ch_Cancellable, sun_nio_ch_Groupable {
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/fd": java_io_FileDescriptor;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/localAddress": java_net_InetSocketAddress;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/stateLock": java_lang_Object;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/closeLock": java_util_concurrent_locks_ReadWriteLock;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/open": number;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/acceptKilled": number;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/isReuseAddress": number;
    public "<init>(Lsun/nio/ch/AsynchronousChannelGroupImpl;)V"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/<init>(Lsun/nio/ch/AsynchronousChannelGroupImpl;)V"(thread: JVMThread, args: [sun_nio_ch_AsynchronousChannelGroupImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "begin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/begin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implAccept(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/implAccept(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)Ljava/util/concurrent/Future;"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "accept()Ljava/util/concurrent/Future;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/accept()Ljava/util/concurrent/Future;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Future) => void): void;
    public "accept(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/accept(Ljava/lang/Object;Ljava/nio/channels/CompletionHandler;)V"(thread: JVMThread, args: [java_lang_Object, java_nio_channels_CompletionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAcceptKilled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/isAcceptKilled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "onCancel(Lsun/nio/ch/PendingFuture;)V"(thread: JVMThread, args: [sun_nio_ch_PendingFuture], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/onCancel(Lsun/nio/ch/PendingFuture;)V"(thread: JVMThread, args: [sun_nio_ch_PendingFuture], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(Ljava/net/SocketAddress;I)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/bind(Ljava/net/SocketAddress;I)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/AsynchronousServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousServerSocketChannel) => void): void;
    public "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "sun/nio/ch/AsynchronousServerSocketChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    "group()Lsun/nio/ch/AsynchronousChannelGroupImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_AsynchronousChannelGroupImpl) => void): void;
  }
  export class sun_nio_ch_SocketChannelImpl extends java_nio_channels_SocketChannel implements sun_nio_ch_SelChImpl {
    public static "sun/nio/ch/SocketChannelImpl/nd": sun_nio_ch_NativeDispatcher;
    public "sun/nio/ch/SocketChannelImpl/fd": java_io_FileDescriptor;
    public "sun/nio/ch/SocketChannelImpl/fdVal": number;
    public "sun/nio/ch/SocketChannelImpl/readerThread": Long;
    public "sun/nio/ch/SocketChannelImpl/writerThread": Long;
    public "sun/nio/ch/SocketChannelImpl/readLock": java_lang_Object;
    public "sun/nio/ch/SocketChannelImpl/writeLock": java_lang_Object;
    public "sun/nio/ch/SocketChannelImpl/stateLock": java_lang_Object;
    public "sun/nio/ch/SocketChannelImpl/isReuseAddress": number;
    public static "sun/nio/ch/SocketChannelImpl/ST_UNINITIALIZED": number;
    public static "sun/nio/ch/SocketChannelImpl/ST_UNCONNECTED": number;
    public static "sun/nio/ch/SocketChannelImpl/ST_PENDING": number;
    public static "sun/nio/ch/SocketChannelImpl/ST_CONNECTED": number;
    public static "sun/nio/ch/SocketChannelImpl/ST_KILLPENDING": number;
    public static "sun/nio/ch/SocketChannelImpl/ST_KILLED": number;
    public "sun/nio/ch/SocketChannelImpl/state": number;
    public "sun/nio/ch/SocketChannelImpl/localAddress": java_net_InetSocketAddress;
    public "sun/nio/ch/SocketChannelImpl/remoteAddress": java_net_InetSocketAddress;
    public "sun/nio/ch/SocketChannelImpl/isInputOpen": number;
    public "sun/nio/ch/SocketChannelImpl/isOutputOpen": number;
    public "sun/nio/ch/SocketChannelImpl/readyToConnect": number;
    public "sun/nio/ch/SocketChannelImpl/socket": java_net_Socket;
    public static "sun/nio/ch/SocketChannelImpl/$assertionsDisabled": number;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;Z)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;Z)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;Ljava/net/InetSocketAddress;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;Ljava/net/InetSocketAddress;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "socket()Ljava/net/Socket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Socket) => void): void;
    public "sun/nio/ch/SocketChannelImpl/socket()Ljava/net/Socket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Socket) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/SocketChannelImpl/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/SocketChannelImpl/getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "sun/nio/ch/SocketChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/SocketChannelImpl/getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/SocketChannelImpl/supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "ensureReadOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/ensureReadOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureWriteOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/ensureWriteOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readerCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/readerCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writerCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/writerCleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/SocketChannelImpl/read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/SocketChannelImpl/write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sendOutOfBandData(B)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/sendOutOfBandData(B)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "localAddress()Ljava/net/InetSocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public "sun/nio/ch/SocketChannelImpl/localAddress()Ljava/net/InetSocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public "remoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/SocketChannelImpl/remoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "sun/nio/ch/SocketChannelImpl/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConnectionPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/isConnectionPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureOpenAndUnconnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/ensureOpenAndUnconnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/SocketAddress;)Z"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/connect(Ljava/net/SocketAddress;)Z"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finishConnect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/finishConnect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shutdownInput()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "sun/nio/ch/SocketChannelImpl/shutdownInput()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "shutdownOutput()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "sun/nio/ch/SocketChannelImpl/shutdownOutput()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "isInputOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/isInputOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOutputOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/isOutputOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "poll(IJ)I"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/poll(IJ)I"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/SocketChannelImpl/translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "sun/nio/ch/SocketChannelImpl/getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/SocketChannelImpl/getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/SocketChannelImpl/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "checkConnect(Ljava/io/FileDescriptor;ZZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/SocketChannelImpl/checkConnect(Ljava/io/FileDescriptor;ZZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sendOutOfBandData(Ljava/io/FileDescriptor;B)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/SocketChannelImpl/sendOutOfBandData(Ljava/io/FileDescriptor;B)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "sun/nio/ch/SocketChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "sun/nio/ch/SocketChannelImpl/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/SocketChannelImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_InetSocketAddress extends java_net_SocketAddress {
    public "java/net/InetSocketAddress/holder": java_net_InetSocketAddress$InetSocketAddressHolder;
    public static "java/net/InetSocketAddress/serialVersionUID": Long;
    public static "java/net/InetSocketAddress/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/net/InetSocketAddress/FIELDS_OFFSET": Long;
    public static "java/net/InetSocketAddress/UNSAFE": sun_misc_Unsafe;
    public static "checkPort(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/InetSocketAddress/checkPort(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkHost(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/InetSocketAddress/checkHost(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetSocketAddress/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetSocketAddress/<init>(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetSocketAddress/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetSocketAddress/<init>(ILjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createUnresolved(Ljava/lang/String;I)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "java/net/InetSocketAddress/createUnresolved(Ljava/lang/String;I)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetSocketAddress/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetSocketAddress/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObjectNoData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetSocketAddress/readObjectNoData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetSocketAddress/getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/InetSocketAddress/getAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetSocketAddress/getHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getHostString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetSocketAddress/getHostString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isUnresolved()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetSocketAddress/isUnresolved()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetSocketAddress/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetSocketAddress/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetSocketAddress/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/InetSocketAddress/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_InetSocketAddress$InetSocketAddressHolder extends java_lang_Object {
    public "java/net/InetSocketAddress$InetSocketAddressHolder/hostname": java_lang_String;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/addr": java_net_InetAddress;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/port": number;
    public "<init>(Ljava/lang/String;Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/<init>(Ljava/lang/String;Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/getAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/getHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getHostString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/getHostString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isUnresolved()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/isUnresolved()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/String;Ljava/net/InetAddress;ILjava/net/InetSocketAddress$1;)V"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress, number, java_net_InetSocketAddress$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetSocketAddress$InetSocketAddressHolder/<init>(Ljava/lang/String;Ljava/net/InetAddress;ILjava/net/InetSocketAddress$1;)V"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress, number, java_net_InetSocketAddress$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$100(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/InetSocketAddress$InetSocketAddressHolder/access$100(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$200(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "java/net/InetSocketAddress$InetSocketAddressHolder/access$200(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "access$300(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)I"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/InetSocketAddress$InetSocketAddressHolder/access$300(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)I"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$400(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)I"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/InetSocketAddress$InetSocketAddressHolder/access$400(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)I"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "java/net/InetSocketAddress$InetSocketAddressHolder/access$500(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "access$600(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/InetSocketAddress$InetSocketAddressHolder/access$600(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$700(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/InetSocketAddress$InetSocketAddressHolder/access$700(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$800(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Z"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/InetSocketAddress$InetSocketAddressHolder/access$800(Ljava/net/InetSocketAddress$InetSocketAddressHolder;)Z"(thread: JVMThread, args: [java_net_InetSocketAddress$InetSocketAddressHolder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_net_InetSocketAddress$1 extends java_lang_Object {
  }
  export class sun_nio_ch_ServerSocketChannelImpl extends java_nio_channels_ServerSocketChannel implements sun_nio_ch_SelChImpl {
    public static "sun/nio/ch/ServerSocketChannelImpl/nd": sun_nio_ch_NativeDispatcher;
    public "sun/nio/ch/ServerSocketChannelImpl/fd": java_io_FileDescriptor;
    public "sun/nio/ch/ServerSocketChannelImpl/fdVal": number;
    public "sun/nio/ch/ServerSocketChannelImpl/thread": Long;
    public "sun/nio/ch/ServerSocketChannelImpl/lock": java_lang_Object;
    public "sun/nio/ch/ServerSocketChannelImpl/stateLock": java_lang_Object;
    public static "sun/nio/ch/ServerSocketChannelImpl/ST_UNINITIALIZED": number;
    public static "sun/nio/ch/ServerSocketChannelImpl/ST_INUSE": number;
    public static "sun/nio/ch/ServerSocketChannelImpl/ST_KILLED": number;
    public "sun/nio/ch/ServerSocketChannelImpl/state": number;
    public "sun/nio/ch/ServerSocketChannelImpl/localAddress": java_net_InetSocketAddress;
    public "sun/nio/ch/ServerSocketChannelImpl/isReuseAddress": number;
    public "sun/nio/ch/ServerSocketChannelImpl/socket": java_net_ServerSocket;
    public static "sun/nio/ch/ServerSocketChannelImpl/$assertionsDisabled": number;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;Z)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;Z)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socket()Ljava/net/ServerSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ServerSocket) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/socket()Ljava/net/ServerSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ServerSocket) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/isBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "localAddress()Ljava/net/InetSocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/localAddress()Ljava/net/InetSocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public "bind(Ljava/net/SocketAddress;I)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/bind(Ljava/net/SocketAddress;I)Ljava/nio/channels/ServerSocketChannel;"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_ServerSocketChannel) => void): void;
    public "accept()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/accept()Ljava/nio/channels/SocketChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SocketChannel) => void): void;
    public "implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "poll(IJ)I"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/poll(IJ)I"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "accept(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/accept(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "accept0(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/accept0(Ljava/io/FileDescriptor;Ljava/io/FileDescriptor;[Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_io_FileDescriptor, JVMArray<java_net_InetSocketAddress>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/ServerSocketChannelImpl/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "sun/nio/ch/ServerSocketChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/ServerSocketChannelImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_PollArrayWrapper extends sun_nio_ch_AbstractPollArrayWrapper {
    public "sun/nio/ch/PollArrayWrapper/interruptFD": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PollArrayWrapper/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "initInterrupt(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PollArrayWrapper/initInterrupt(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "release(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PollArrayWrapper/release(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "free()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PollArrayWrapper/free()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addEntry(Lsun/nio/ch/SelChImpl;)V"(thread: JVMThread, args: [sun_nio_ch_SelChImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PollArrayWrapper/addEntry(Lsun/nio/ch/SelChImpl;)V"(thread: JVMThread, args: [sun_nio_ch_SelChImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public static "replaceEntry(Lsun/nio/ch/PollArrayWrapper;ILsun/nio/ch/PollArrayWrapper;I)V"(thread: JVMThread, args: [sun_nio_ch_PollArrayWrapper, number, sun_nio_ch_PollArrayWrapper, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/PollArrayWrapper/replaceEntry(Lsun/nio/ch/PollArrayWrapper;ILsun/nio/ch/PollArrayWrapper;I)V"(thread: JVMThread, args: [sun_nio_ch_PollArrayWrapper, number, sun_nio_ch_PollArrayWrapper, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PollArrayWrapper/grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "poll(IIJ)I"(thread: JVMThread, args: [number, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/PollArrayWrapper/poll(IIJ)I"(thread: JVMThread, args: [number, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "interrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/PollArrayWrapper/interrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "poll0(JIJ)I"(thread: JVMThread, args: [Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/PollArrayWrapper/poll0(JIJ)I"(thread: JVMThread, args: [Long, any, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "interrupt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/PollArrayWrapper/interrupt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/PollArrayWrapper/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_AbstractPollArrayWrapper extends java_lang_Object {
    public static "sun/nio/ch/AbstractPollArrayWrapper/SIZE_POLLFD": number;
    public static "sun/nio/ch/AbstractPollArrayWrapper/FD_OFFSET": number;
    public static "sun/nio/ch/AbstractPollArrayWrapper/EVENT_OFFSET": number;
    public static "sun/nio/ch/AbstractPollArrayWrapper/REVENT_OFFSET": number;
    public "sun/nio/ch/AbstractPollArrayWrapper/pollArray": sun_nio_ch_AllocatedNativeObject;
    public "sun/nio/ch/AbstractPollArrayWrapper/totalChannels": number;
    public "sun/nio/ch/AbstractPollArrayWrapper/pollArrayAddress": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AbstractPollArrayWrapper/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEventOps(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AbstractPollArrayWrapper/getEventOps(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getReventOps(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AbstractPollArrayWrapper/getReventOps(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescriptor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/AbstractPollArrayWrapper/getDescriptor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putEventOps(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AbstractPollArrayWrapper/putEventOps(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "putReventOps(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AbstractPollArrayWrapper/putReventOps(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "putDescriptor(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AbstractPollArrayWrapper/putDescriptor(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_AllocatedNativeObject extends sun_nio_ch_NativeObject {
    public "<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AllocatedNativeObject/<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "free()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/AllocatedNativeObject/free()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_NativeObject extends java_lang_Object {
    public static "sun/nio/ch/NativeObject/unsafe": sun_misc_Unsafe;
    public "sun/nio/ch/NativeObject/allocationAddress": Long;
    public "sun/nio/ch/NativeObject/address": Long;
    public static "sun/nio/ch/NativeObject/byteOrder": java_nio_ByteOrder;
    public static "sun/nio/ch/NativeObject/pageSize": number;
    public static "sun/nio/ch/NativeObject/$assertionsDisabled": number;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/<init>(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "address()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/NativeObject/address()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "allocationAddress()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/NativeObject/allocationAddress()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "subObject(I)Lsun/nio/ch/NativeObject;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_NativeObject) => void): void;
    public "sun/nio/ch/NativeObject/subObject(I)Lsun/nio/ch/NativeObject;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_NativeObject) => void): void;
    public "getObject(I)Lsun/nio/ch/NativeObject;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_NativeObject) => void): void;
    public "sun/nio/ch/NativeObject/getObject(I)Lsun/nio/ch/NativeObject;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_NativeObject) => void): void;
    public "putObject(ILsun/nio/ch/NativeObject;)V"(thread: JVMThread, args: [number, sun_nio_ch_NativeObject], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/putObject(ILsun/nio/ch/NativeObject;)V"(thread: JVMThread, args: [number, sun_nio_ch_NativeObject], cb?: (e?: java_lang_Throwable) => void): void;
    public "getByte(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeObject/getByte(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putByte(IB)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/putByte(IB)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getShort(I)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeObject/getShort(I)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putShort(IS)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/putShort(IS)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChar(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeObject/getChar(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putChar(IC)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/putChar(IC)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeObject/getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putInt(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/putInt(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/NativeObject/getLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putLong(IJ)V"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/putLong(IJ)V"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeObject/getFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloat(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/putFloat(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDouble(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/NativeObject/getDouble(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDouble(ID)V"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/NativeObject/putDouble(ID)V"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "addressSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/NativeObject/addressSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "byteOrder()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public static "sun/nio/ch/NativeObject/byteOrder()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public static "pageSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/NativeObject/pageSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/NativeObject/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_ByteOrder extends java_lang_Object {
    public "java/nio/ByteOrder/name": java_lang_String;
    public static "java/nio/ByteOrder/BIG_ENDIAN": java_nio_ByteOrder;
    public static "java/nio/ByteOrder/LITTLE_ENDIAN": java_nio_ByteOrder;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/ByteOrder/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "nativeOrder()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public static "java/nio/ByteOrder/nativeOrder()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/ByteOrder/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/ByteOrder/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_Net extends java_lang_Object {
    public static "sun/nio/ch/Net/UNSPEC": java_net_ProtocolFamily;
    public static "sun/nio/ch/Net/exclusiveBind": number;
    public static "sun/nio/ch/Net/fastLoopback": number;
    public static "sun/nio/ch/Net/checkedIPv6": number;
    public static "sun/nio/ch/Net/isIPv6Available": number;
    public static "sun/nio/ch/Net/SHUT_RD": number;
    public static "sun/nio/ch/Net/SHUT_WR": number;
    public static "sun/nio/ch/Net/SHUT_RDWR": number;
    public static "sun/nio/ch/Net/POLLIN": number;
    public static "sun/nio/ch/Net/POLLOUT": number;
    public static "sun/nio/ch/Net/POLLERR": number;
    public static "sun/nio/ch/Net/POLLHUP": number;
    public static "sun/nio/ch/Net/POLLNVAL": number;
    public static "sun/nio/ch/Net/POLLCONN": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/Net/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isIPv6Available()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/isIPv6Available()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "useExclusiveBind()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/useExclusiveBind()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "canIPv6SocketJoinIPv4Group()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/canIPv6SocketJoinIPv4Group()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "canJoin6WithIPv4Group()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/canJoin6WithIPv4Group()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkAddress(Ljava/net/SocketAddress;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "sun/nio/ch/Net/checkAddress(Ljava/net/SocketAddress;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "asInetSocketAddress(Ljava/net/SocketAddress;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "sun/nio/ch/Net/asInetSocketAddress(Ljava/net/SocketAddress;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "translateToSocketException(Ljava/lang/Exception;)V"(thread: JVMThread, args: [java_lang_Exception], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/translateToSocketException(Ljava/lang/Exception;)V"(thread: JVMThread, args: [java_lang_Exception], cb?: (e?: java_lang_Throwable) => void): void;
    public static "translateException(Ljava/lang/Exception;Z)V"(thread: JVMThread, args: [java_lang_Exception, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/translateException(Ljava/lang/Exception;Z)V"(thread: JVMThread, args: [java_lang_Exception, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "translateException(Ljava/lang/Exception;)V"(thread: JVMThread, args: [java_lang_Exception], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/translateException(Ljava/lang/Exception;)V"(thread: JVMThread, args: [java_lang_Exception], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getRevealedLocalAddress(Ljava/net/InetSocketAddress;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "sun/nio/ch/Net/getRevealedLocalAddress(Ljava/net/InetSocketAddress;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "getRevealedLocalAddressAsString(Ljava/net/InetSocketAddress;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/nio/ch/Net/getRevealedLocalAddressAsString(Ljava/net/InetSocketAddress;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getLoopbackAddress(I)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "sun/nio/ch/Net/getLoopbackAddress(I)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "anyInet4Address(Ljava/net/NetworkInterface;)Ljava/net/Inet4Address;"(thread: JVMThread, args: [java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: java_net_Inet4Address) => void): void;
    public static "sun/nio/ch/Net/anyInet4Address(Ljava/net/NetworkInterface;)Ljava/net/Inet4Address;"(thread: JVMThread, args: [java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: java_net_Inet4Address) => void): void;
    public static "inet4AsInt(Ljava/net/InetAddress;)I"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/inet4AsInt(Ljava/net/InetAddress;)I"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "inet4FromInt(I)Ljava/net/InetAddress;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "sun/nio/ch/Net/inet4FromInt(I)Ljava/net/InetAddress;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "inet6AsByteArray(Ljava/net/InetAddress;)[B"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/ch/Net/inet6AsByteArray(Ljava/net/InetAddress;)[B"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "setSocketOption(Ljava/io/FileDescriptor;Ljava/net/ProtocolFamily;Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_ProtocolFamily, java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/setSocketOption(Ljava/io/FileDescriptor;Ljava/net/ProtocolFamily;Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_ProtocolFamily, java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getSocketOption(Ljava/io/FileDescriptor;Ljava/net/ProtocolFamily;Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_ProtocolFamily, java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/nio/ch/Net/getSocketOption(Ljava/io/FileDescriptor;Ljava/net/ProtocolFamily;Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_ProtocolFamily, java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "isFastTcpLoopbackRequested()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/isFastTcpLoopbackRequested()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isIPv6Available0()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/isIPv6Available0()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isExclusiveBindAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/isExclusiveBindAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "canIPv6SocketJoinIPv4Group0()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/canIPv6SocketJoinIPv4Group0()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "canJoin6WithIPv4Group0()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/canJoin6WithIPv4Group0()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "socket(Z)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "sun/nio/ch/Net/socket(Z)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "socket(Ljava/net/ProtocolFamily;Z)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [java_net_ProtocolFamily, number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "sun/nio/ch/Net/socket(Ljava/net/ProtocolFamily;Z)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [java_net_ProtocolFamily, number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "serverSocket(Z)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "sun/nio/ch/Net/serverSocket(Z)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "socket0(ZZZZ)I"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/socket0(ZZZZ)I"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "bind(Ljava/io/FileDescriptor;Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/bind(Ljava/io/FileDescriptor;Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "bind(Ljava/net/ProtocolFamily;Ljava/io/FileDescriptor;Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_ProtocolFamily, java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/bind(Ljava/net/ProtocolFamily;Ljava/io/FileDescriptor;Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_ProtocolFamily, java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "bind0(Ljava/io/FileDescriptor;ZZLjava/net/InetAddress;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/bind0(Ljava/io/FileDescriptor;ZZLjava/net/InetAddress;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "listen(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/listen(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "connect(Ljava/io/FileDescriptor;Ljava/net/InetAddress;I)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/connect(Ljava/io/FileDescriptor;Ljava/net/InetAddress;I)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "connect(Ljava/net/ProtocolFamily;Ljava/io/FileDescriptor;Ljava/net/InetAddress;I)I"(thread: JVMThread, args: [java_net_ProtocolFamily, java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/connect(Ljava/net/ProtocolFamily;Ljava/io/FileDescriptor;Ljava/net/InetAddress;I)I"(thread: JVMThread, args: [java_net_ProtocolFamily, java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "connect0(ZLjava/io/FileDescriptor;Ljava/net/InetAddress;I)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/connect0(ZLjava/io/FileDescriptor;Ljava/net/InetAddress;I)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "shutdown(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/shutdown(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "localPort(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/localPort(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "localInetAddress(Ljava/io/FileDescriptor;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "sun/nio/ch/Net/localInetAddress(Ljava/io/FileDescriptor;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "localAddress(Ljava/io/FileDescriptor;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "sun/nio/ch/Net/localAddress(Ljava/io/FileDescriptor;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "remotePort(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/remotePort(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "remoteInetAddress(Ljava/io/FileDescriptor;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "sun/nio/ch/Net/remoteInetAddress(Ljava/io/FileDescriptor;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "remoteAddress(Ljava/io/FileDescriptor;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "sun/nio/ch/Net/remoteAddress(Ljava/io/FileDescriptor;)Ljava/net/InetSocketAddress;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: java_net_InetSocketAddress) => void): void;
    public static "getIntOption0(Ljava/io/FileDescriptor;ZII)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/getIntOption0(Ljava/io/FileDescriptor;ZII)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setIntOption0(Ljava/io/FileDescriptor;ZIIIZ)V"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/setIntOption0(Ljava/io/FileDescriptor;ZIIIZ)V"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "poll(Ljava/io/FileDescriptor;IJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/poll(Ljava/io/FileDescriptor;IJ)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "join4(Ljava/io/FileDescriptor;III)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/join4(Ljava/io/FileDescriptor;III)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "drop4(Ljava/io/FileDescriptor;III)V"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/drop4(Ljava/io/FileDescriptor;III)V"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "joinOrDrop4(ZLjava/io/FileDescriptor;III)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/joinOrDrop4(ZLjava/io/FileDescriptor;III)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "block4(Ljava/io/FileDescriptor;III)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/block4(Ljava/io/FileDescriptor;III)I"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "unblock4(Ljava/io/FileDescriptor;III)V"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/unblock4(Ljava/io/FileDescriptor;III)V"(thread: JVMThread, args: [java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "blockOrUnblock4(ZLjava/io/FileDescriptor;III)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/blockOrUnblock4(ZLjava/io/FileDescriptor;III)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "join6(Ljava/io/FileDescriptor;[BI[B)I"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/join6(Ljava/io/FileDescriptor;[BI[B)I"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "drop6(Ljava/io/FileDescriptor;[BI[B)V"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/drop6(Ljava/io/FileDescriptor;[BI[B)V"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "joinOrDrop6(ZLjava/io/FileDescriptor;[BI[B)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/joinOrDrop6(ZLjava/io/FileDescriptor;[BI[B)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "block6(Ljava/io/FileDescriptor;[BI[B)I"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/block6(Ljava/io/FileDescriptor;[BI[B)I"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "unblock6(Ljava/io/FileDescriptor;[BI[B)V"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/unblock6(Ljava/io/FileDescriptor;[BI[B)V"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "blockOrUnblock6(ZLjava/io/FileDescriptor;[BI[B)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/blockOrUnblock6(ZLjava/io/FileDescriptor;[BI[B)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setInterface4(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/setInterface4(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInterface4(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/getInterface4(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setInterface6(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/setInterface6(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInterface6(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/getInterface6(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "pollinValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/pollinValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "polloutValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/polloutValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "pollerrValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/pollerrValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "pollhupValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/pollhupValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "pollnvalValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/pollnvalValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "pollconnValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/Net/pollconnValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/Net/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_InheritedChannel extends java_lang_Object {
    public static "sun/nio/ch/InheritedChannel/UNKNOWN": number;
    public static "sun/nio/ch/InheritedChannel/SOCK_STREAM": number;
    public static "sun/nio/ch/InheritedChannel/SOCK_DGRAM": number;
    public static "sun/nio/ch/InheritedChannel/O_RDONLY": number;
    public static "sun/nio/ch/InheritedChannel/O_WRONLY": number;
    public static "sun/nio/ch/InheritedChannel/O_RDWR": number;
    public static "sun/nio/ch/InheritedChannel/devnull": number;
    public static "sun/nio/ch/InheritedChannel/haveChannel": number;
    public static "sun/nio/ch/InheritedChannel/channel": java_nio_channels_Channel;
    public static "sun/nio/ch/InheritedChannel/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/InheritedChannel/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "detachIOStreams()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/InheritedChannel/detachIOStreams()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkAccess(Ljava/nio/channels/Channel;)V"(thread: JVMThread, args: [java_nio_channels_Channel], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/InheritedChannel/checkAccess(Ljava/nio/channels/Channel;)V"(thread: JVMThread, args: [java_nio_channels_Channel], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createChannel()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public static "sun/nio/ch/InheritedChannel/createChannel()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public static "getChannel()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public static "sun/nio/ch/InheritedChannel/getChannel()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public static "dup(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/InheritedChannel/dup(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "dup2(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/InheritedChannel/dup2(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open0(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/InheritedChannel/open0(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "close0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/InheritedChannel/close0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "soType0(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/InheritedChannel/soType0(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "peerAddress0(I)Ljava/net/InetAddress;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "sun/nio/ch/InheritedChannel/peerAddress0(I)Ljava/net/InetAddress;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "peerPort0(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/InheritedChannel/peerPort0(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/InheritedChannel/access$000()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/InheritedChannel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_FileKey extends java_lang_Object {
    public "sun/nio/ch/FileKey/st_dev": Long;
    public "sun/nio/ch/FileKey/st_ino": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileKey/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "create(Ljava/io/FileDescriptor;)Lsun/nio/ch/FileKey;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_FileKey) => void): void;
    public static "sun/nio/ch/FileKey/create(Ljava/io/FileDescriptor;)Lsun/nio/ch/FileKey;"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_FileKey) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileKey/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/FileKey/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "init(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/FileKey/init(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileKey/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/FileKey/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_EPollPort extends sun_nio_ch_Port {
    public static "sun/nio/ch/EPollPort/MAX_EPOLL_EVENTS": number;
    public static "sun/nio/ch/EPollPort/ENOENT": number;
    public "sun/nio/ch/EPollPort/epfd": number;
    public "sun/nio/ch/EPollPort/closed": number;
    public "sun/nio/ch/EPollPort/sp": JVMArray<number>;
    public "sun/nio/ch/EPollPort/wakeupCount": java_util_concurrent_atomic_AtomicInteger;
    public "sun/nio/ch/EPollPort/address": Long;
    public "sun/nio/ch/EPollPort/queue": java_util_concurrent_ArrayBlockingQueue;
    public "sun/nio/ch/EPollPort/NEED_TO_POLL": sun_nio_ch_EPollPort$Event;
    public "sun/nio/ch/EPollPort/EXECUTE_TASK_OR_SHUTDOWN": sun_nio_ch_EPollPort$Event;
    public "<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;Lsun/nio/ch/ThreadPool;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider, sun_nio_ch_ThreadPool], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollPort/<init>(Ljava/nio/channels/spi/AsynchronousChannelProvider;Lsun/nio/ch/ThreadPool;)V"(thread: JVMThread, args: [java_nio_channels_spi_AsynchronousChannelProvider, sun_nio_ch_ThreadPool], cb?: (e?: java_lang_Throwable) => void): void;
    public "start()Lsun/nio/ch/EPollPort;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_EPollPort) => void): void;
    public "sun/nio/ch/EPollPort/start()Lsun/nio/ch/EPollPort;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_EPollPort) => void): void;
    public "implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollPort/implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "wakeup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollPort/wakeup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "executeOnHandlerTask(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollPort/executeOnHandlerTask(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownHandlerTasks()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollPort/shutdownHandlerTasks()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "startPoll(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollPort/startPoll(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "socketpair([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollPort/socketpair([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "interrupt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollPort/interrupt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "drain1(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollPort/drain1(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "close0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollPort/close0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$100(Lsun/nio/ch/EPollPort;)I"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPollPort/access$100(Lsun/nio/ch/EPollPort;)I"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Lsun/nio/ch/EPollPort;)J"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/EPollPort/access$200(Lsun/nio/ch/EPollPort;)J"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$300(Lsun/nio/ch/EPollPort;)[I"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/ch/EPollPort/access$300(Lsun/nio/ch/EPollPort;)[I"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$400(Lsun/nio/ch/EPollPort;)Ljava/util/concurrent/atomic/AtomicInteger;"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicInteger) => void): void;
    public static "sun/nio/ch/EPollPort/access$400(Lsun/nio/ch/EPollPort;)Ljava/util/concurrent/atomic/AtomicInteger;"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicInteger) => void): void;
    public static "access$500(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollPort/access$500(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$600(Lsun/nio/ch/EPollPort;)Lsun/nio/ch/EPollPort$Event;"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_EPollPort$Event) => void): void;
    public static "sun/nio/ch/EPollPort/access$600(Lsun/nio/ch/EPollPort;)Lsun/nio/ch/EPollPort$Event;"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_EPollPort$Event) => void): void;
    public static "access$700(Lsun/nio/ch/EPollPort;)Ljava/util/concurrent/ArrayBlockingQueue;"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ArrayBlockingQueue) => void): void;
    public static "sun/nio/ch/EPollPort/access$700(Lsun/nio/ch/EPollPort;)Ljava/util/concurrent/ArrayBlockingQueue;"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ArrayBlockingQueue) => void): void;
    public static "access$800(Lsun/nio/ch/EPollPort;)Lsun/nio/ch/EPollPort$Event;"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_EPollPort$Event) => void): void;
    public static "sun/nio/ch/EPollPort/access$800(Lsun/nio/ch/EPollPort;)Lsun/nio/ch/EPollPort$Event;"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_EPollPort$Event) => void): void;
    public static "access$900(Lsun/nio/ch/EPollPort;)V"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollPort/access$900(Lsun/nio/ch/EPollPort;)V"(thread: JVMThread, args: [sun_nio_ch_EPollPort], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollPort/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_EPollPort$Event extends java_lang_Object {
    public "sun/nio/ch/EPollPort$Event/channel": sun_nio_ch_Port$PollableChannel;
    public "sun/nio/ch/EPollPort$Event/events": number;
    public "<init>(Lsun/nio/ch/Port$PollableChannel;I)V"(thread: JVMThread, args: [sun_nio_ch_Port$PollableChannel, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollPort$Event/<init>(Lsun/nio/ch/Port$PollableChannel;I)V"(thread: JVMThread, args: [sun_nio_ch_Port$PollableChannel, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "channel()Lsun/nio/ch/Port$PollableChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_Port$PollableChannel) => void): void;
    public "sun/nio/ch/EPollPort$Event/channel()Lsun/nio/ch/Port$PollableChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_ch_Port$PollableChannel) => void): void;
    public "events()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollPort$Event/events()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_concurrent_ArrayBlockingQueue extends java_util_AbstractQueue implements java_util_concurrent_BlockingQueue, java_io_Serializable {
    public static "java/util/concurrent/ArrayBlockingQueue/serialVersionUID": Long;
    public "java/util/concurrent/ArrayBlockingQueue/items": JVMArray<java_lang_Object>;
    public "java/util/concurrent/ArrayBlockingQueue/takeIndex": number;
    public "java/util/concurrent/ArrayBlockingQueue/putIndex": number;
    public "java/util/concurrent/ArrayBlockingQueue/count": number;
    public "java/util/concurrent/ArrayBlockingQueue/lock": java_util_concurrent_locks_ReentrantLock;
    public "java/util/concurrent/ArrayBlockingQueue/notEmpty": java_util_concurrent_locks_Condition;
    public "java/util/concurrent/ArrayBlockingQueue/notFull": java_util_concurrent_locks_Condition;
    public "java/util/concurrent/ArrayBlockingQueue/itrs": java_util_concurrent_ArrayBlockingQueue$Itrs;
    public "dec(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/dec(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "itemAt(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/itemAt(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "checkNotNull(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/ArrayBlockingQueue/checkNotNull(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "enqueue(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/enqueue(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "dequeue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/dequeue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "removeAt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/removeAt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IZLjava/util/Collection;)V"(thread: JVMThread, args: [number, number, java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/<init>(IZLjava/util/Collection;)V"(thread: JVMThread, args: [number, number, java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/put(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "offer(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/offer(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "take()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/take()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "poll(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/poll(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remainingCapacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/remainingCapacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "drainTo(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/drainTo(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drainTo(Ljava/util/Collection;I)I"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/drainTo(Ljava/util/Collection;I)I"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ArrayBlockingQueue$Itrs extends java_lang_Object {
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/cycles": number;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/head": java_util_concurrent_ArrayBlockingQueue$Itrs$Node;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/sweeper": java_util_concurrent_ArrayBlockingQueue$Itrs$Node;
    public static "java/util/concurrent/ArrayBlockingQueue$Itrs/SHORT_SWEEP_PROBES": number;
    public static "java/util/concurrent/ArrayBlockingQueue$Itrs/LONG_SWEEP_PROBES": number;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/this$0": java_util_concurrent_ArrayBlockingQueue;
    public "<init>(Ljava/util/concurrent/ArrayBlockingQueue;Ljava/util/concurrent/ArrayBlockingQueue$Itr;)V"(thread: JVMThread, args: [java_util_concurrent_ArrayBlockingQueue, java_util_concurrent_ArrayBlockingQueue$Itr], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/<init>(Ljava/util/concurrent/ArrayBlockingQueue;Ljava/util/concurrent/ArrayBlockingQueue$Itr;)V"(thread: JVMThread, args: [java_util_concurrent_ArrayBlockingQueue, java_util_concurrent_ArrayBlockingQueue$Itr], cb?: (e?: java_lang_Throwable) => void): void;
    public "doSomeSweeping(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/doSomeSweeping(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "register(Ljava/util/concurrent/ArrayBlockingQueue$Itr;)V"(thread: JVMThread, args: [java_util_concurrent_ArrayBlockingQueue$Itr], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/register(Ljava/util/concurrent/ArrayBlockingQueue$Itr;)V"(thread: JVMThread, args: [java_util_concurrent_ArrayBlockingQueue$Itr], cb?: (e?: java_lang_Throwable) => void): void;
    public "takeIndexWrapped()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/takeIndexWrapped()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removedAt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/removedAt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "queueIsEmpty()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/queueIsEmpty()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "elementDequeued()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs/elementDequeued()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ArrayBlockingQueue$Itr extends java_lang_Object implements java_util_Iterator {
    public "java/util/concurrent/ArrayBlockingQueue$Itr/cursor": number;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/nextItem": java_lang_Object;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/nextIndex": number;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/lastItem": java_lang_Object;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/lastRet": number;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/prevTakeIndex": number;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/prevCycles": number;
    public static "java/util/concurrent/ArrayBlockingQueue$Itr/NONE": number;
    public static "java/util/concurrent/ArrayBlockingQueue$Itr/REMOVED": number;
    public static "java/util/concurrent/ArrayBlockingQueue$Itr/DETACHED": number;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/this$0": java_util_concurrent_ArrayBlockingQueue;
    public "<init>(Ljava/util/concurrent/ArrayBlockingQueue;)V"(thread: JVMThread, args: [java_util_concurrent_ArrayBlockingQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/<init>(Ljava/util/concurrent/ArrayBlockingQueue;)V"(thread: JVMThread, args: [java_util_concurrent_ArrayBlockingQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "isDetached()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/isDetached()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "incCursor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/incCursor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "invalidated(IIJI)Z"(thread: JVMThread, args: [number, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/invalidated(IIJI)Z"(thread: JVMThread, args: [number, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "incorporateDequeues()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/incorporateDequeues()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "detach()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/detach()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "noNext()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/noNext()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "next()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/next()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "distance(III)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/distance(III)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removedAt(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/removedAt(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "takeIndexWrapped()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itr/takeIndexWrapped()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEachRemaining(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "remove()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_ArrayBlockingQueue$Itrs$Node extends java_lang_ref_WeakReference {
    public "java/util/concurrent/ArrayBlockingQueue$Itrs$Node/next": java_util_concurrent_ArrayBlockingQueue$Itrs$Node;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs$Node/this$1": java_util_concurrent_ArrayBlockingQueue$Itrs;
    public "<init>(Ljava/util/concurrent/ArrayBlockingQueue$Itrs;Ljava/util/concurrent/ArrayBlockingQueue$Itr;Ljava/util/concurrent/ArrayBlockingQueue$Itrs$Node;)V"(thread: JVMThread, args: [java_util_concurrent_ArrayBlockingQueue$Itrs, java_util_concurrent_ArrayBlockingQueue$Itr, java_util_concurrent_ArrayBlockingQueue$Itrs$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/ArrayBlockingQueue$Itrs$Node/<init>(Ljava/util/concurrent/ArrayBlockingQueue$Itrs;Ljava/util/concurrent/ArrayBlockingQueue$Itr;Ljava/util/concurrent/ArrayBlockingQueue$Itrs$Node;)V"(thread: JVMThread, args: [java_util_concurrent_ArrayBlockingQueue$Itrs, java_util_concurrent_ArrayBlockingQueue$Itr, java_util_concurrent_ArrayBlockingQueue$Itrs$Node], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_AbstractQueue extends java_util_AbstractCollection implements java_util_Queue {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractQueue/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractQueue/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractQueue/remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractQueue/element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractQueue/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractQueue/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_EPollArrayWrapper extends java_lang_Object {
    public static "sun/nio/ch/EPollArrayWrapper/EPOLLIN": number;
    public static "sun/nio/ch/EPollArrayWrapper/EPOLL_CTL_ADD": number;
    public static "sun/nio/ch/EPollArrayWrapper/EPOLL_CTL_DEL": number;
    public static "sun/nio/ch/EPollArrayWrapper/EPOLL_CTL_MOD": number;
    public static "sun/nio/ch/EPollArrayWrapper/SIZE_EPOLLEVENT": number;
    public static "sun/nio/ch/EPollArrayWrapper/EVENT_OFFSET": number;
    public static "sun/nio/ch/EPollArrayWrapper/DATA_OFFSET": number;
    public static "sun/nio/ch/EPollArrayWrapper/FD_OFFSET": number;
    public static "sun/nio/ch/EPollArrayWrapper/OPEN_MAX": number;
    public static "sun/nio/ch/EPollArrayWrapper/NUM_EPOLLEVENTS": number;
    public static "sun/nio/ch/EPollArrayWrapper/KILLED": number;
    public static "sun/nio/ch/EPollArrayWrapper/INITIAL_PENDING_UPDATE_SIZE": number;
    public static "sun/nio/ch/EPollArrayWrapper/MAX_UPDATE_ARRAY_SIZE": number;
    public "sun/nio/ch/EPollArrayWrapper/epfd": number;
    public "sun/nio/ch/EPollArrayWrapper/pollArray": sun_nio_ch_AllocatedNativeObject;
    public "sun/nio/ch/EPollArrayWrapper/pollArrayAddress": Long;
    public "sun/nio/ch/EPollArrayWrapper/outgoingInterruptFD": number;
    public "sun/nio/ch/EPollArrayWrapper/incomingInterruptFD": number;
    public "sun/nio/ch/EPollArrayWrapper/interruptedIndex": number;
    public "sun/nio/ch/EPollArrayWrapper/updated": number;
    public "sun/nio/ch/EPollArrayWrapper/updateLock": java_lang_Object;
    public "sun/nio/ch/EPollArrayWrapper/updateCount": number;
    public "sun/nio/ch/EPollArrayWrapper/updateDescriptors": JVMArray<number>;
    public "sun/nio/ch/EPollArrayWrapper/eventsLow": JVMArray<number>;
    public "sun/nio/ch/EPollArrayWrapper/eventsHigh": java_util_Map;
    public "sun/nio/ch/EPollArrayWrapper/registered": java_util_BitSet;
    public "sun/nio/ch/EPollArrayWrapper/interrupted": number;
    public static "sun/nio/ch/EPollArrayWrapper/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "initInterrupt(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/initInterrupt(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "putEventOps(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/putEventOps(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "putDescriptor(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/putDescriptor(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEventOps(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/getEventOps(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescriptor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/getDescriptor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEventsHighKilled(Ljava/lang/Integer;)Z"(thread: JVMThread, args: [java_lang_Integer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/isEventsHighKilled(Ljava/lang/Integer;)Z"(thread: JVMThread, args: [java_lang_Integer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setUpdateEvents(IBZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/setUpdateEvents(IBZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUpdateEvents(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/getUpdateEvents(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setInterest(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/setInterest(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/add(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "closeEPollFD()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/closeEPollFD()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "poll(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/poll(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "updateRegistrations()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/updateRegistrations()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/interrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interruptedIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/interruptedIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "interrupted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/interrupted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clearInterrupted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/clearInterrupted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "epollCreate()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/epollCreate()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "epollCtl(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/epollCtl(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "epollWait(JIJI)I"(thread: JVMThread, args: [Long, any, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/EPollArrayWrapper/epollWait(JIJI)I"(thread: JVMThread, args: [Long, any, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sizeofEPollEvent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPollArrayWrapper/sizeofEPollEvent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "offsetofData()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPollArrayWrapper/offsetofData()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "interrupt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollArrayWrapper/interrupt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollArrayWrapper/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPollArrayWrapper/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_BitSet extends java_lang_Object implements java_lang_Cloneable, java_io_Serializable {
    public static "java/util/BitSet/ADDRESS_BITS_PER_WORD": number;
    public static "java/util/BitSet/BITS_PER_WORD": number;
    public static "java/util/BitSet/BIT_INDEX_MASK": number;
    public static "java/util/BitSet/WORD_MASK": Long;
    public static "java/util/BitSet/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public "java/util/BitSet/words": JVMArray<Long>;
    public "java/util/BitSet/wordsInUse": number;
    public "java/util/BitSet/sizeIsSticky": number;
    public static "java/util/BitSet/serialVersionUID": Long;
    public static "java/util/BitSet/$assertionsDisabled": number;
    public static "wordIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/BitSet/wordIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkInvariants()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/checkInvariants()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "recalculateWordsInUse()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/recalculateWordsInUse()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "initWords(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/initWords(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([J)V"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/<init>([J)V"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "valueOf([J)Ljava/util/BitSet;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public static "java/util/BitSet/valueOf([J)Ljava/util/BitSet;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public static "valueOf(Ljava/nio/LongBuffer;)Ljava/util/BitSet;"(thread: JVMThread, args: [java_nio_LongBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public static "java/util/BitSet/valueOf(Ljava/nio/LongBuffer;)Ljava/util/BitSet;"(thread: JVMThread, args: [java_nio_LongBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public static "valueOf([B)Ljava/util/BitSet;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public static "java/util/BitSet/valueOf([B)Ljava/util/BitSet;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public static "valueOf(Ljava/nio/ByteBuffer;)Ljava/util/BitSet;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public static "java/util/BitSet/valueOf(Ljava/nio/ByteBuffer;)Ljava/util/BitSet;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public "toByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/util/BitSet/toByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "toLongArray()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "java/util/BitSet/toLongArray()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "expandTo(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/expandTo(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/BitSet/checkRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flip(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/flip(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flip(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/flip(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/set(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/set(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/set(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(IIZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/set(IIZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/clear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/clear(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/get(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(II)Ljava/util/BitSet;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public "java/util/BitSet/get(II)Ljava/util/BitSet;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_BitSet) => void): void;
    public "nextSetBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/nextSetBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextClearBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/nextClearBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "previousSetBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/previousSetBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "previousClearBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/previousClearBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(Ljava/util/BitSet;)Z"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/intersects(Ljava/util/BitSet;)Z"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "cardinality()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/cardinality()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "and(Ljava/util/BitSet;)V"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/and(Ljava/util/BitSet;)V"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "or(Ljava/util/BitSet;)V"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/or(Ljava/util/BitSet;)V"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "xor(Ljava/util/BitSet;)V"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/xor(Ljava/util/BitSet;)V"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "andNot(Ljava/util/BitSet;)V"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/andNot(Ljava/util/BitSet;)V"(thread: JVMThread, args: [java_util_BitSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/BitSet/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/BitSet/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/BitSet/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/BitSet/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "stream()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "java/util/BitSet/stream()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "lambda$stream$0()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    public "java/util/BitSet/lambda$stream$0()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/BitSet/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_LongBuffer extends java_nio_Buffer implements java_lang_Comparable {
    public "java/nio/LongBuffer/hb": JVMArray<Long>;
    public "java/nio/LongBuffer/offset": number;
    public "java/nio/LongBuffer/isReadOnly": number;
    public "<init>(IIII[JI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/LongBuffer/<init>(IIII[JI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/LongBuffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "allocate(I)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public static "java/nio/LongBuffer/allocate(I)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public static "wrap([JII)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public static "java/nio/LongBuffer/wrap([JII)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public static "wrap([J)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public static "java/nio/LongBuffer/wrap([J)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "slice()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/slice()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "duplicate()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/duplicate()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "asReadOnlyBuffer()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/asReadOnlyBuffer()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "get()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/LongBuffer/get()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "put(J)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/put(J)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "get(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/LongBuffer/get(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "put(IJ)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/put(IJ)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "get([JII)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/get([JII)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "get([J)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/get([J)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "put(Ljava/nio/LongBuffer;)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [java_nio_LongBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/put(Ljava/nio/LongBuffer;)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [java_nio_LongBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "put([JII)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/put([JII)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "put([J)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/put([J)Ljava/nio/LongBuffer;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/LongBuffer/hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "array()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "java/nio/LongBuffer/array()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/LongBuffer/arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compact()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/LongBuffer/compact()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/LongBuffer/isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/LongBuffer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/LongBuffer/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/LongBuffer/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "equals(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/LongBuffer/equals(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/nio/LongBuffer;)I"(thread: JVMThread, args: [java_nio_LongBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/LongBuffer/compareTo(Ljava/nio/LongBuffer;)I"(thread: JVMThread, args: [java_nio_LongBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/LongBuffer/compare(JJ)I"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "java/nio/LongBuffer/order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/LongBuffer/array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/LongBuffer/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_Buffer extends java_lang_Object {
    public static "java/nio/Buffer/SPLITERATOR_CHARACTERISTICS": number;
    public "java/nio/Buffer/mark": number;
    public "java/nio/Buffer/position": number;
    public "java/nio/Buffer/limit": number;
    public "java/nio/Buffer/capacity": number;
    public "java/nio/Buffer/address": Long;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/Buffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "position()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/position()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "position(I)Ljava/nio/Buffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "java/nio/Buffer/position(I)Ljava/nio/Buffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "limit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/limit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "limit(I)Ljava/nio/Buffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "java/nio/Buffer/limit(I)Ljava/nio/Buffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "mark()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "java/nio/Buffer/mark()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "reset()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "java/nio/Buffer/reset()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "clear()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "java/nio/Buffer/clear()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "flip()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "java/nio/Buffer/flip()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "rewind()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "java/nio/Buffer/rewind()Ljava/nio/Buffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_Buffer) => void): void;
    public "remaining()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/remaining()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasRemaining()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/hasRemaining()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/Buffer/array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextGetIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/nextGetIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextGetIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/nextGetIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextPutIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/nextPutIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextPutIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/nextPutIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/checkIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkIndex(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/checkIndex(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "markValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/Buffer/markValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "truncate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/Buffer/truncate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "discardMark()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/Buffer/discardMark()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkBounds(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Buffer/checkBounds(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_EPoll extends java_lang_Object {
    public static "sun/nio/ch/EPoll/unsafe": sun_misc_Unsafe;
    public static "sun/nio/ch/EPoll/SIZEOF_EPOLLEVENT": number;
    public static "sun/nio/ch/EPoll/OFFSETOF_EVENTS": number;
    public static "sun/nio/ch/EPoll/OFFSETOF_FD": number;
    public static "sun/nio/ch/EPoll/EPOLL_CTL_ADD": number;
    public static "sun/nio/ch/EPoll/EPOLL_CTL_DEL": number;
    public static "sun/nio/ch/EPoll/EPOLL_CTL_MOD": number;
    public static "sun/nio/ch/EPoll/EPOLLONESHOT": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/EPoll/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "allocatePollArray(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/EPoll/allocatePollArray(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "freePollArray(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPoll/freePollArray(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getEvent(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/EPoll/getEvent(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getDescriptor(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPoll/getDescriptor(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getEvents(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPoll/getEvents(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "eventSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPoll/eventSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "eventsOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPoll/eventsOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "dataOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPoll/dataOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "epollCreate()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPoll/epollCreate()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "epollCtl(IIII)I"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPoll/epollCtl(IIII)I"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "epollWait(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/EPoll/epollWait(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/EPoll/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_DatagramDispatcher extends sun_nio_ch_NativeDispatcher {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramDispatcher/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramDispatcher/read(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readv(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/DatagramDispatcher/readv(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramDispatcher/write(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writev(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/DatagramDispatcher/writev(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "close(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramDispatcher/close(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "preClose(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramDispatcher/preClose(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "read0(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/DatagramDispatcher/read0(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "readv0(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/DatagramDispatcher/readv0(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "write0(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/ch/DatagramDispatcher/write0(Ljava/io/FileDescriptor;JI)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "writev0(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/ch/DatagramDispatcher/writev0(Ljava/io/FileDescriptor;JI)J"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/DatagramDispatcher/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_DatagramChannelImpl extends java_nio_channels_DatagramChannel implements sun_nio_ch_SelChImpl {
    public static "sun/nio/ch/DatagramChannelImpl/nd": sun_nio_ch_NativeDispatcher;
    public "sun/nio/ch/DatagramChannelImpl/fd": java_io_FileDescriptor;
    public "sun/nio/ch/DatagramChannelImpl/fdVal": number;
    public "sun/nio/ch/DatagramChannelImpl/family": java_net_ProtocolFamily;
    public "sun/nio/ch/DatagramChannelImpl/readerThread": Long;
    public "sun/nio/ch/DatagramChannelImpl/writerThread": Long;
    public "sun/nio/ch/DatagramChannelImpl/cachedSenderInetAddress": java_net_InetAddress;
    public "sun/nio/ch/DatagramChannelImpl/cachedSenderPort": number;
    public "sun/nio/ch/DatagramChannelImpl/readLock": java_lang_Object;
    public "sun/nio/ch/DatagramChannelImpl/writeLock": java_lang_Object;
    public "sun/nio/ch/DatagramChannelImpl/stateLock": java_lang_Object;
    public static "sun/nio/ch/DatagramChannelImpl/ST_UNINITIALIZED": number;
    public static "sun/nio/ch/DatagramChannelImpl/ST_UNCONNECTED": number;
    public static "sun/nio/ch/DatagramChannelImpl/ST_CONNECTED": number;
    public static "sun/nio/ch/DatagramChannelImpl/ST_KILLED": number;
    public "sun/nio/ch/DatagramChannelImpl/state": number;
    public "sun/nio/ch/DatagramChannelImpl/localAddress": java_net_InetSocketAddress;
    public "sun/nio/ch/DatagramChannelImpl/remoteAddress": java_net_InetSocketAddress;
    public "sun/nio/ch/DatagramChannelImpl/socket": java_net_DatagramSocket;
    public "sun/nio/ch/DatagramChannelImpl/registry": sun_nio_ch_MembershipRegistry;
    public "sun/nio/ch/DatagramChannelImpl/reuseAddressEmulated": number;
    public "sun/nio/ch/DatagramChannelImpl/isReuseAddress": number;
    public "sun/nio/ch/DatagramChannelImpl/sender": java_net_SocketAddress;
    public static "sun/nio/ch/DatagramChannelImpl/$assertionsDisabled": number;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/net/ProtocolFamily;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_net_ProtocolFamily], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/net/ProtocolFamily;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_net_ProtocolFamily], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/<init>(Ljava/nio/channels/spi/SelectorProvider;Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_nio_channels_spi_SelectorProvider, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "socket()Ljava/net/DatagramSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_DatagramSocket) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/socket()Ljava/net/DatagramSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_DatagramSocket) => void): void;
    public "getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/getLocalAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/getRemoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/supportedOptions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "receive(Ljava/nio/ByteBuffer;)Ljava/net/SocketAddress;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/receive(Ljava/nio/ByteBuffer;)Ljava/net/SocketAddress;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "receive(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/receive(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "receiveIntoNativeBuffer(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;II)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/receiveIntoNativeBuffer(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;II)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "send(Ljava/nio/ByteBuffer;Ljava/net/SocketAddress;)I"(thread: JVMThread, args: [java_nio_ByteBuffer, java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/send(Ljava/nio/ByteBuffer;Ljava/net/SocketAddress;)I"(thread: JVMThread, args: [java_nio_ByteBuffer, java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "send(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/send(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sendFromNativeBuffer(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/sendFromNativeBuffer(Ljava/io/FileDescriptor;Ljava/nio/ByteBuffer;Ljava/net/InetSocketAddress;)I"(thread: JVMThread, args: [java_io_FileDescriptor, java_nio_ByteBuffer, java_net_InetSocketAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/read(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/read([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/write(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/write([Ljava/nio/ByteBuffer;II)J"(thread: JVMThread, args: [JVMArray<java_nio_ByteBuffer>, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/implConfigureBlocking(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "localAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/localAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "remoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/remoteAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/isConnected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureOpenAndUnconnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/ensureOpenAndUnconnected()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/SocketAddress;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/connect(Ljava/net/SocketAddress;)Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "disconnect()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/disconnect()Ljava/nio/channels/DatagramChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_DatagramChannel) => void): void;
    public "innerJoin(Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/innerJoin(Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "drop(Lsun/nio/ch/MembershipKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/drop(Lsun/nio/ch/MembershipKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "block(Lsun/nio/ch/MembershipKeyImpl;Ljava/net/InetAddress;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/block(Lsun/nio/ch/MembershipKeyImpl;Ljava/net/InetAddress;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "unblock(Lsun/nio/ch/MembershipKeyImpl;Ljava/net/InetAddress;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/unblock(Lsun/nio/ch/MembershipKeyImpl;Ljava/net/InetAddress;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/implCloseSelectableChannel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/kill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/translateReadyOps(IILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/translateAndUpdateReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/translateAndSetReadyOps(ILsun/nio/ch/SelectionKeyImpl;)Z"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "poll(IJ)I"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/poll(IJ)I"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/translateAndSetInterestOps(ILsun/nio/ch/SelectionKeyImpl;)V"(thread: JVMThread, args: [number, sun_nio_ch_SelectionKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/getFDVal()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/DatagramChannelImpl/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "disconnect0(Ljava/io/FileDescriptor;Z)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/DatagramChannelImpl/disconnect0(Ljava/io/FileDescriptor;Z)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "receive0(Ljava/io/FileDescriptor;JIZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/receive0(Ljava/io/FileDescriptor;JIZ)I"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "send0(ZLjava/io/FileDescriptor;JILjava/net/InetAddress;I)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, Long, any, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/send0(ZLjava/io/FileDescriptor;JILjava/net/InetAddress;I)I"(thread: JVMThread, args: [number, java_io_FileDescriptor, Long, any, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public "sun/nio/ch/DatagramChannelImpl/bind(Ljava/net/SocketAddress;)Ljava/nio/channels/NetworkChannel;"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_NetworkChannel) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/ch/DatagramChannelImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_MembershipKeyImpl extends java_nio_channels_MembershipKey {
    public "sun/nio/ch/MembershipKeyImpl/ch": java_nio_channels_MulticastChannel;
    public "sun/nio/ch/MembershipKeyImpl/group": java_net_InetAddress;
    public "sun/nio/ch/MembershipKeyImpl/interf": java_net_NetworkInterface;
    public "sun/nio/ch/MembershipKeyImpl/source": java_net_InetAddress;
    public "sun/nio/ch/MembershipKeyImpl/valid": number;
    public "sun/nio/ch/MembershipKeyImpl/stateLock": java_lang_Object;
    public "sun/nio/ch/MembershipKeyImpl/blockedSet": java_util_HashSet;
    public "<init>(Ljava/nio/channels/MulticastChannel;Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_nio_channels_MulticastChannel, java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/<init>(Ljava/nio/channels/MulticastChannel;Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_nio_channels_MulticastChannel, java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "drop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/drop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "channel()Ljava/nio/channels/MulticastChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MulticastChannel) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/channel()Ljava/nio/channels/MulticastChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MulticastChannel) => void): void;
    public "group()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/group()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "networkInterface()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/networkInterface()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public "sourceAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/sourceAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "block(Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/block(Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "unblock(Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/unblock(Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "<init>(Ljava/nio/channels/MulticastChannel;Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;Lsun/nio/ch/MembershipKeyImpl$1;)V"(thread: JVMThread, args: [java_nio_channels_MulticastChannel, java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress, sun_nio_ch_MembershipKeyImpl$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/MembershipKeyImpl/<init>(Ljava/nio/channels/MulticastChannel;Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;Lsun/nio/ch/MembershipKeyImpl$1;)V"(thread: JVMThread, args: [java_nio_channels_MulticastChannel, java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress, sun_nio_ch_MembershipKeyImpl$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_ch_MembershipKeyImpl$1 extends java_lang_Object {
  }
  export class sun_nio_ch_MembershipRegistry extends java_lang_Object {
    public "sun/nio/ch/MembershipRegistry/groups": java_util_Map;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/MembershipRegistry/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkMembership(Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "sun/nio/ch/MembershipRegistry/checkMembership(Ljava/net/InetAddress;Ljava/net/NetworkInterface;Ljava/net/InetAddress;)Ljava/nio/channels/MembershipKey;"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_MembershipKey) => void): void;
    public "add(Lsun/nio/ch/MembershipKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/MembershipRegistry/add(Lsun/nio/ch/MembershipKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Lsun/nio/ch/MembershipKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/MembershipRegistry/remove(Lsun/nio/ch/MembershipKeyImpl;)V"(thread: JVMThread, args: [sun_nio_ch_MembershipKeyImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidateAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/ch/MembershipRegistry/invalidateAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_UnixFileStoreAttributes extends java_lang_Object {
    public "sun/nio/fs/UnixFileStoreAttributes/f_frsize": Long;
    public "sun/nio/fs/UnixFileStoreAttributes/f_blocks": Long;
    public "sun/nio/fs/UnixFileStoreAttributes/f_bfree": Long;
    public "sun/nio/fs/UnixFileStoreAttributes/f_bavail": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileStoreAttributes/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "get(Lsun/nio/fs/UnixPath;)Lsun/nio/fs/UnixFileStoreAttributes;"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileStoreAttributes) => void): void;
    public static "sun/nio/fs/UnixFileStoreAttributes/get(Lsun/nio/fs/UnixPath;)Lsun/nio/fs/UnixFileStoreAttributes;"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileStoreAttributes) => void): void;
    public "blockSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/UnixFileStoreAttributes/blockSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "totalBlocks()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/UnixFileStoreAttributes/totalBlocks()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "freeBlocks()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/UnixFileStoreAttributes/freeBlocks()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "availableBlocks()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/UnixFileStoreAttributes/availableBlocks()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class sun_nio_fs_UnixPath extends sun_nio_fs_AbstractPath {
    public static "sun/nio/fs/UnixPath/encoder": java_lang_ThreadLocal;
    public "sun/nio/fs/UnixPath/fs": sun_nio_fs_UnixFileSystem;
    public "sun/nio/fs/UnixPath/path": JVMArray<number>;
    public "sun/nio/fs/UnixPath/stringValue": java_lang_String;
    public "sun/nio/fs/UnixPath/hash": number;
    public "sun/nio/fs/UnixPath/offsets": JVMArray<number>;
    public static "sun/nio/fs/UnixPath/$assertionsDisabled": number;
    public "<init>(Lsun/nio/fs/UnixFileSystem;[B)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixPath/<init>(Lsun/nio/fs/UnixFileSystem;[B)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/nio/fs/UnixFileSystem;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixPath/<init>(Lsun/nio/fs/UnixFileSystem;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "normalizeAndCheck(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/nio/fs/UnixPath/normalizeAndCheck(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "checkNotNul(Ljava/lang/String;C)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixPath/checkNotNul(Ljava/lang/String;C)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "normalize(Ljava/lang/String;II)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/nio/fs/UnixPath/normalize(Ljava/lang/String;II)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "encode(Lsun/nio/fs/UnixFileSystem;Ljava/lang/String;)[B"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixPath/encode(Lsun/nio/fs/UnixFileSystem;Ljava/lang/String;)[B"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "asByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/nio/fs/UnixPath/asByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getByteArrayForSysCalls()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/nio/fs/UnixPath/getByteArrayForSysCalls()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPathForExceptionMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixPath/getPathForExceptionMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPathForPermissionCheck()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixPath/getPathForPermissionCheck()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toUnixPath(Ljava/nio/file/Path;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public static "sun/nio/fs/UnixPath/toUnixPath(Ljava/nio/file/Path;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "initOffsets()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixPath/initOffsets()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "emptyPath()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/emptyPath()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "getFileSystem()Lsun/nio/fs/UnixFileSystem;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileSystem) => void): void;
    public "sun/nio/fs/UnixPath/getFileSystem()Lsun/nio/fs/UnixFileSystem;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileSystem) => void): void;
    public "getRoot()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/getRoot()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "getFileName()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/getFileName()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "getParent()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/getParent()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "getNameCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/getNameCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName(I)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/getName(I)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "subpath(II)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/subpath(II)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "isAbsolute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/isAbsolute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "resolve([B[B)[B"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixPath/resolve([B[B)[B"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "resolve(Ljava/nio/file/Path;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/resolve(Ljava/nio/file/Path;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "resolve([B)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/resolve([B)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "relativize(Ljava/nio/file/Path;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/relativize(Ljava/nio/file/Path;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "normalize()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/normalize()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "startsWith(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/startsWith(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "endsWith(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/endsWith(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/nio/file/Path;)I"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/compareTo(Ljava/nio/file/Path;)I"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixPath/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "openForAttributeAccess(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/openForAttributeAccess(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkRead()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixPath/checkRead()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkWrite()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixPath/checkWrite()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkDelete()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixPath/checkDelete()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toAbsolutePath()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixPath/toAbsolutePath()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "toRealPath([Ljava/nio/file/LinkOption;)Ljava/nio/file/Path;"(thread: JVMThread, args: [JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/toRealPath([Ljava/nio/file/LinkOption;)Ljava/nio/file/Path;"(thread: JVMThread, args: [JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "toUri()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "sun/nio/fs/UnixPath/toUri()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "register(Ljava/nio/file/WatchService;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_WatchService, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "sun/nio/fs/UnixPath/register(Ljava/nio/file/WatchService;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_WatchService, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "toAbsolutePath()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/toAbsolutePath()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "relativize(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/relativize(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "resolve(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/resolve(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "subpath(II)Ljava/nio/file/Path;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/subpath(II)Ljava/nio/file/Path;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "getName(I)Ljava/nio/file/Path;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/getName(I)Ljava/nio/file/Path;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "getParent()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/getParent()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "getFileName()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/getFileName()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "getRoot()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixPath/getRoot()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "getFileSystem()Ljava/nio/file/FileSystem;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "sun/nio/fs/UnixPath/getFileSystem()Ljava/nio/file/FileSystem;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixPath/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixPath/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_UnixFileSystem extends java_nio_file_FileSystem {
    public "sun/nio/fs/UnixFileSystem/provider": sun_nio_fs_UnixFileSystemProvider;
    public "sun/nio/fs/UnixFileSystem/defaultDirectory": JVMArray<number>;
    public "sun/nio/fs/UnixFileSystem/needToResolveAgainstDefaultDirectory": number;
    public "sun/nio/fs/UnixFileSystem/rootDirectory": sun_nio_fs_UnixPath;
    public static "sun/nio/fs/UnixFileSystem/GLOB_SYNTAX": java_lang_String;
    public static "sun/nio/fs/UnixFileSystem/REGEX_SYNTAX": java_lang_String;
    public "<init>(Lsun/nio/fs/UnixFileSystemProvider;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystemProvider, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystem/<init>(Lsun/nio/fs/UnixFileSystemProvider;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystemProvider, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "defaultDirectory()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/nio/fs/UnixFileSystem/defaultDirectory()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "needToResolveAgainstDefaultDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileSystem/needToResolveAgainstDefaultDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rootDirectory()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixFileSystem/rootDirectory()Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "isSolaris()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileSystem/isSolaris()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "standardFileAttributeViews()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "sun/nio/fs/UnixFileSystem/standardFileAttributeViews()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "provider()Ljava/nio/file/spi/FileSystemProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileSystemProvider) => void): void;
    public "sun/nio/fs/UnixFileSystem/provider()Ljava/nio/file/spi/FileSystemProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileSystemProvider) => void): void;
    public "getSeparator()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixFileSystem/getSeparator()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileSystem/isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileSystem/isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystem/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyNonPosixAttributes(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystem/copyNonPosixAttributes(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRootDirectories()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "sun/nio/fs/UnixFileSystem/getRootDirectories()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "getMountEntries()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "sun/nio/fs/UnixFileSystem/getMountEntries()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "getFileStore(Lsun/nio/fs/UnixMountEntry;)Ljava/nio/file/FileStore;"(thread: JVMThread, args: [sun_nio_fs_UnixMountEntry], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileStore) => void): void;
    public "sun/nio/fs/UnixFileSystem/getFileStore(Lsun/nio/fs/UnixMountEntry;)Ljava/nio/file/FileStore;"(thread: JVMThread, args: [sun_nio_fs_UnixMountEntry], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileStore) => void): void;
    public "getFileStores()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "sun/nio/fs/UnixFileSystem/getFileStores()Ljava/lang/Iterable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Iterable) => void): void;
    public "getPath(Ljava/lang/String;[Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixFileSystem/getPath(Ljava/lang/String;[Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "getPathMatcher(Ljava/lang/String;)Ljava/nio/file/PathMatcher;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_PathMatcher) => void): void;
    public "sun/nio/fs/UnixFileSystem/getPathMatcher(Ljava/lang/String;)Ljava/nio/file/PathMatcher;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_PathMatcher) => void): void;
    public "getUserPrincipalLookupService()Ljava/nio/file/attribute/UserPrincipalLookupService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_UserPrincipalLookupService) => void): void;
    public "sun/nio/fs/UnixFileSystem/getUserPrincipalLookupService()Ljava/nio/file/attribute/UserPrincipalLookupService;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_UserPrincipalLookupService) => void): void;
    public "compilePathMatchPattern(Ljava/lang/String;)Ljava/util/regex/Pattern;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern) => void): void;
    public "sun/nio/fs/UnixFileSystem/compilePathMatchPattern(Ljava/lang/String;)Ljava/util/regex/Pattern;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_regex_Pattern) => void): void;
    public "normalizeNativePath([C)[C"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/nio/fs/UnixFileSystem/normalizeNativePath([C)[C"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "normalizeJavaPath(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixFileSystem/normalizeJavaPath(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000(Lsun/nio/fs/UnixFileSystem;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public static "sun/nio/fs/UnixFileSystem/access$000(Lsun/nio/fs/UnixFileSystem;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
  }
  export class java_nio_file_FileStore extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/FileStore/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/FileStore/name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "type()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/FileStore/type()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/FileStore/isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTotalSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/file/FileStore/getTotalSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getUsableSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/file/FileStore/getUsableSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getUnallocatedSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/file/FileStore/getUnallocatedSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "supportsFileAttributeView(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/FileStore/supportsFileAttributeView(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "supportsFileAttributeView(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/FileStore/supportsFileAttributeView(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFileStoreAttributeView(Ljava/lang/Class;)Ljava/nio/file/attribute/FileStoreAttributeView;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileStoreAttributeView) => void): void;
    public "java/nio/file/FileStore/getFileStoreAttributeView(Ljava/lang/Class;)Ljava/nio/file/attribute/FileStoreAttributeView;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileStoreAttributeView) => void): void;
    public "getAttribute(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/file/FileStore/getAttribute(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_nio_file_attribute_FileStoreAttributeView extends java_lang_Object, java_nio_file_attribute_AttributeView {
    "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_nio_file_attribute_AttributeView extends java_lang_Object {
    "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_nio_fs_UnixFileSystemProvider extends sun_nio_fs_AbstractFileSystemProvider {
    public static "sun/nio/fs/UnixFileSystemProvider/USER_DIR": java_lang_String;
    public "sun/nio/fs/UnixFileSystemProvider/theFileSystem": sun_nio_fs_UnixFileSystem;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "newFileSystem(Ljava/lang/String;)Lsun/nio/fs/UnixFileSystem;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileSystem) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/newFileSystem(Ljava/lang/String;)Lsun/nio/fs/UnixFileSystem;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileSystem) => void): void;
    public "getScheme()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/getScheme()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "checkUri(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/checkUri(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
    public "newFileSystem(Ljava/net/URI;Ljava/util/Map;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_net_URI, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/newFileSystem(Ljava/net/URI;Ljava/util/Map;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_net_URI, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "getFileSystem(Ljava/net/URI;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/getFileSystem(Ljava/net/URI;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "getPath(Ljava/net/URI;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/getPath(Ljava/net/URI;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "checkPath(Ljava/nio/file/Path;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/checkPath(Ljava/nio/file/Path;)Lsun/nio/fs/UnixPath;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixPath) => void): void;
    public "getFileAttributeView(Ljava/nio/file/Path;Ljava/lang/Class;[Ljava/nio/file/LinkOption;)Ljava/nio/file/attribute/FileAttributeView;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_Class, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileAttributeView) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/getFileAttributeView(Ljava/nio/file/Path;Ljava/lang/Class;[Ljava/nio/file/LinkOption;)Ljava/nio/file/attribute/FileAttributeView;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_Class, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileAttributeView) => void): void;
    public "readAttributes(Ljava/nio/file/Path;Ljava/lang/Class;[Ljava/nio/file/LinkOption;)Ljava/nio/file/attribute/BasicFileAttributes;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_Class, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_BasicFileAttributes) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/readAttributes(Ljava/nio/file/Path;Ljava/lang/Class;[Ljava/nio/file/LinkOption;)Ljava/nio/file/attribute/BasicFileAttributes;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_Class, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_BasicFileAttributes) => void): void;
    public "getFileAttributeView(Ljava/nio/file/Path;Ljava/lang/String;[Ljava/nio/file/LinkOption;)Lsun/nio/fs/DynamicFileAttributeView;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_DynamicFileAttributeView) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/getFileAttributeView(Ljava/nio/file/Path;Ljava/lang/String;[Ljava/nio/file/LinkOption;)Lsun/nio/fs/DynamicFileAttributeView;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_DynamicFileAttributeView) => void): void;
    public "newFileChannel(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/newFileChannel(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "newAsynchronousFileChannel(Ljava/nio/file/Path;Ljava/util/Set;Ljava/util/concurrent/ExecutorService;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, java_util_concurrent_ExecutorService, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/newAsynchronousFileChannel(Ljava/nio/file/Path;Ljava/util/Set;Ljava/util/concurrent/ExecutorService;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, java_util_concurrent_ExecutorService, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public "newByteChannel(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/newByteChannel(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "implDelete(Ljava/nio/file/Path;Z)Z"(thread: JVMThread, args: [java_nio_file_Path, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/implDelete(Ljava/nio/file/Path;Z)Z"(thread: JVMThread, args: [java_nio_file_Path, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "copy(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/copy(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "move(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/move(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAccess(Ljava/nio/file/Path;[Ljava/nio/file/AccessMode;)V"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_AccessMode>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/checkAccess(Ljava/nio/file/Path;[Ljava/nio/file/AccessMode;)V"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_AccessMode>], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSameFile(Ljava/nio/file/Path;Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/isSameFile(Ljava/nio/file/Path;Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isHidden(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/isHidden(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFileStore(Lsun/nio/fs/UnixPath;)Ljava/nio/file/FileStore;"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileStore) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/getFileStore(Lsun/nio/fs/UnixPath;)Ljava/nio/file/FileStore;"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileStore) => void): void;
    public "getFileStore(Ljava/nio/file/Path;)Ljava/nio/file/FileStore;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileStore) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/getFileStore(Ljava/nio/file/Path;)Ljava/nio/file/FileStore;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileStore) => void): void;
    public "createDirectory(Ljava/nio/file/Path;[Ljava/nio/file/attribute/FileAttribute;)V"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/createDirectory(Ljava/nio/file/Path;[Ljava/nio/file/attribute/FileAttribute;)V"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable) => void): void;
    public "newDirectoryStream(Ljava/nio/file/Path;Ljava/nio/file/DirectoryStream$Filter;)Ljava/nio/file/DirectoryStream;"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_DirectoryStream$Filter], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_DirectoryStream) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/newDirectoryStream(Ljava/nio/file/Path;Ljava/nio/file/DirectoryStream$Filter;)Ljava/nio/file/DirectoryStream;"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_DirectoryStream$Filter], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_DirectoryStream) => void): void;
    public "createSymbolicLink(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/attribute/FileAttribute;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/createSymbolicLink(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/attribute/FileAttribute;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable) => void): void;
    public "createLink(Ljava/nio/file/Path;Ljava/nio/file/Path;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/createLink(Ljava/nio/file/Path;Ljava/nio/file/Path;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path], cb?: (e?: java_lang_Throwable) => void): void;
    public "readSymbolicLink(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/readSymbolicLink(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "getFileTypeDetector()Ljava/nio/file/spi/FileTypeDetector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileTypeDetector) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/getFileTypeDetector()Ljava/nio/file/spi/FileTypeDetector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileTypeDetector) => void): void;
    public "chain([Lsun/nio/fs/AbstractFileTypeDetector;)Ljava/nio/file/spi/FileTypeDetector;"(thread: JVMThread, args: [JVMArray<sun_nio_fs_AbstractFileTypeDetector>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileTypeDetector) => void): void;
    public "sun/nio/fs/UnixFileSystemProvider/chain([Lsun/nio/fs/AbstractFileTypeDetector;)Ljava/nio/file/spi/FileTypeDetector;"(thread: JVMThread, args: [JVMArray<sun_nio_fs_AbstractFileTypeDetector>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileTypeDetector) => void): void;
  }
  export class sun_nio_fs_AbstractFileTypeDetector extends java_nio_file_spi_FileTypeDetector {
    public static "sun/nio/fs/AbstractFileTypeDetector/TSPECIALS": java_lang_String;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractFileTypeDetector/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "probeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/AbstractFileTypeDetector/probeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "implProbeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/AbstractFileTypeDetector/implProbeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "parse(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/nio/fs/AbstractFileTypeDetector/parse(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "isTokenChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/AbstractFileTypeDetector/isTokenChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isValidToken(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/AbstractFileTypeDetector/isValidToken(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_file_spi_FileTypeDetector extends java_lang_Object {
    public static "checkPermission()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "java/nio/file/spi/FileTypeDetector/checkPermission()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "<init>(Ljava/lang/Void;)V"(thread: JVMThread, args: [java_lang_Void], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileTypeDetector/<init>(Ljava/lang/Void;)V"(thread: JVMThread, args: [java_lang_Void], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileTypeDetector/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "probeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/spi/FileTypeDetector/probeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_nio_file_DirectoryStream$Filter extends java_lang_Object {
    "accept(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_nio_file_DirectoryStream extends java_lang_Object, java_io_Closeable, java_lang_Iterable {
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
  }
  export class java_nio_file_AccessMode extends java_lang_Enum {
    public static "java/nio/file/AccessMode/READ": java_nio_file_AccessMode;
    public static "java/nio/file/AccessMode/WRITE": java_nio_file_AccessMode;
    public static "java/nio/file/AccessMode/EXECUTE": java_nio_file_AccessMode;
    public static "java/nio/file/AccessMode/$VALUES": JVMArray<java_nio_file_AccessMode>;
    public static "values()[Ljava/nio/file/AccessMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_nio_file_AccessMode>) => void): void;
    public static "java/nio/file/AccessMode/values()[Ljava/nio/file/AccessMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_nio_file_AccessMode>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/nio/file/AccessMode;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_AccessMode) => void): void;
    public static "java/nio/file/AccessMode/valueOf(Ljava/lang/String;)Ljava/nio/file/AccessMode;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_AccessMode) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/AccessMode/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/file/AccessMode/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_nio_fs_DynamicFileAttributeView extends java_lang_Object {
    "setAttribute(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "readAttributes([Ljava/lang/String;)Ljava/util/Map;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
  }
  export interface java_nio_file_attribute_BasicFileAttributes extends java_lang_Object {
    "lastModifiedTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    "lastAccessTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    "creationTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    "isRegularFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSymbolicLink()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isOther()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "fileKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_nio_file_attribute_FileTime extends java_lang_Object implements java_lang_Comparable {
    public "java/nio/file/attribute/FileTime/unit": java_util_concurrent_TimeUnit;
    public "java/nio/file/attribute/FileTime/value": Long;
    public "java/nio/file/attribute/FileTime/instant": java_time_Instant;
    public "java/nio/file/attribute/FileTime/valueAsString": java_lang_String;
    public static "java/nio/file/attribute/FileTime/HOURS_PER_DAY": Long;
    public static "java/nio/file/attribute/FileTime/MINUTES_PER_HOUR": Long;
    public static "java/nio/file/attribute/FileTime/SECONDS_PER_MINUTE": Long;
    public static "java/nio/file/attribute/FileTime/SECONDS_PER_HOUR": Long;
    public static "java/nio/file/attribute/FileTime/SECONDS_PER_DAY": Long;
    public static "java/nio/file/attribute/FileTime/MILLIS_PER_SECOND": Long;
    public static "java/nio/file/attribute/FileTime/MICROS_PER_SECOND": Long;
    public static "java/nio/file/attribute/FileTime/NANOS_PER_SECOND": Long;
    public static "java/nio/file/attribute/FileTime/NANOS_PER_MILLI": number;
    public static "java/nio/file/attribute/FileTime/NANOS_PER_MICRO": number;
    public static "java/nio/file/attribute/FileTime/MIN_SECOND": Long;
    public static "java/nio/file/attribute/FileTime/MAX_SECOND": Long;
    public static "java/nio/file/attribute/FileTime/DAYS_PER_10000_YEARS": Long;
    public static "java/nio/file/attribute/FileTime/SECONDS_PER_10000_YEARS": Long;
    public static "java/nio/file/attribute/FileTime/SECONDS_0000_TO_1970": Long;
    public "<init>(JLjava/util/concurrent/TimeUnit;Ljava/time/Instant;)V"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit, java_time_Instant], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/attribute/FileTime/<init>(JLjava/util/concurrent/TimeUnit;Ljava/time/Instant;)V"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit, java_time_Instant], cb?: (e?: java_lang_Throwable) => void): void;
    public static "from(JLjava/util/concurrent/TimeUnit;)Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public static "java/nio/file/attribute/FileTime/from(JLjava/util/concurrent/TimeUnit;)Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public static "fromMillis(J)Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public static "java/nio/file/attribute/FileTime/fromMillis(J)Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public static "from(Ljava/time/Instant;)Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public static "java/nio/file/attribute/FileTime/from(Ljava/time/Instant;)Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: [java_time_Instant], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "to(Ljava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/file/attribute/FileTime/to(Ljava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toMillis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/file/attribute/FileTime/toMillis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "scale(JJJ)J"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/nio/file/attribute/FileTime/scale(JJJ)J"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "java/nio/file/attribute/FileTime/toInstant()Ljava/time/Instant;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_Instant) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/attribute/FileTime/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/attribute/FileTime/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toDays()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/file/attribute/FileTime/toDays()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toExcessNanos(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/file/attribute/FileTime/toExcessNanos(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareTo(Ljava/nio/file/attribute/FileTime;)I"(thread: JVMThread, args: [java_nio_file_attribute_FileTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/attribute/FileTime/compareTo(Ljava/nio/file/attribute/FileTime;)I"(thread: JVMThread, args: [java_nio_file_attribute_FileTime], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "append(Ljava/lang/StringBuilder;II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_StringBuilder, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "java/nio/file/attribute/FileTime/append(Ljava/lang/StringBuilder;II)Ljava/lang/StringBuilder;"(thread: JVMThread, args: [java_lang_StringBuilder, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/attribute/FileTime/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/attribute/FileTime/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_nio_file_attribute_FileAttributeView extends java_lang_Object, java_nio_file_attribute_AttributeView {
    "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_nio_fs_AbstractFileSystemProvider extends java_nio_file_spi_FileSystemProvider {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractFileSystemProvider/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "split(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "sun/nio/fs/AbstractFileSystemProvider/split(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getFileAttributeView(Ljava/nio/file/Path;Ljava/lang/String;[Ljava/nio/file/LinkOption;)Lsun/nio/fs/DynamicFileAttributeView;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_DynamicFileAttributeView) => void): void;
    public "sun/nio/fs/AbstractFileSystemProvider/getFileAttributeView(Ljava/nio/file/Path;Ljava/lang/String;[Ljava/nio/file/LinkOption;)Lsun/nio/fs/DynamicFileAttributeView;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_DynamicFileAttributeView) => void): void;
    public "setAttribute(Ljava/nio/file/Path;Ljava/lang/String;Ljava/lang/Object;[Ljava/nio/file/LinkOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, java_lang_Object, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractFileSystemProvider/setAttribute(Ljava/nio/file/Path;Ljava/lang/String;Ljava/lang/Object;[Ljava/nio/file/LinkOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, java_lang_Object, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "readAttributes(Ljava/nio/file/Path;Ljava/lang/String;[Ljava/nio/file/LinkOption;)Ljava/util/Map;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "sun/nio/fs/AbstractFileSystemProvider/readAttributes(Ljava/nio/file/Path;Ljava/lang/String;[Ljava/nio/file/LinkOption;)Ljava/util/Map;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "implDelete(Ljava/nio/file/Path;Z)Z"(thread: JVMThread, args: [java_nio_file_Path, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/AbstractFileSystemProvider/implDelete(Ljava/nio/file/Path;Z)Z"(thread: JVMThread, args: [java_nio_file_Path, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "delete(Ljava/nio/file/Path;)V"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractFileSystemProvider/delete(Ljava/nio/file/Path;)V"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable) => void): void;
    public "deleteIfExists(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/AbstractFileSystemProvider/deleteIfExists(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_nio_fs_AbstractPath extends java_lang_Object implements java_nio_file_Path {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractPath/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "startsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/AbstractPath/startsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "endsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/AbstractPath/endsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "resolve(Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/AbstractPath/resolve(Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "resolveSibling(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/AbstractPath/resolveSibling(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "resolveSibling(Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "sun/nio/fs/AbstractPath/resolveSibling(Ljava/lang/String;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "sun/nio/fs/AbstractPath/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "toFile()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "sun/nio/fs/AbstractPath/toFile()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "register(Ljava/nio/file/WatchService;[Ljava/nio/file/WatchEvent$Kind;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_WatchService, JVMArray<java_nio_file_WatchEvent$Kind>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "sun/nio/fs/AbstractPath/register(Ljava/nio/file/WatchService;[Ljava/nio/file/WatchEvent$Kind;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_WatchService, JVMArray<java_nio_file_WatchEvent$Kind>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    "getFileSystem()Ljava/nio/file/FileSystem;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    "isAbsolute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getRoot()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "getFileName()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "getParent()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "getNameCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getName(I)Ljava/nio/file/Path;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "subpath(II)Ljava/nio/file/Path;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "startsWith(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "endsWith(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "normalize()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "resolve(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "relativize(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "toUri()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    "toAbsolutePath()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "toRealPath([Ljava/nio/file/LinkOption;)Ljava/nio/file/Path;"(thread: JVMThread, args: [JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    "register(Ljava/nio/file/WatchService;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_WatchService, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    "compareTo(Ljava/nio/file/Path;)I"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
  }
  export class sun_nio_fs_UnixFileAttributes extends java_lang_Object implements java_nio_file_attribute_PosixFileAttributes {
    public "sun/nio/fs/UnixFileAttributes/st_mode": number;
    public "sun/nio/fs/UnixFileAttributes/st_ino": Long;
    public "sun/nio/fs/UnixFileAttributes/st_dev": Long;
    public "sun/nio/fs/UnixFileAttributes/st_rdev": Long;
    public "sun/nio/fs/UnixFileAttributes/st_nlink": number;
    public "sun/nio/fs/UnixFileAttributes/st_uid": number;
    public "sun/nio/fs/UnixFileAttributes/st_gid": number;
    public "sun/nio/fs/UnixFileAttributes/st_size": Long;
    public "sun/nio/fs/UnixFileAttributes/st_atime_sec": Long;
    public "sun/nio/fs/UnixFileAttributes/st_atime_nsec": Long;
    public "sun/nio/fs/UnixFileAttributes/st_mtime_sec": Long;
    public "sun/nio/fs/UnixFileAttributes/st_mtime_nsec": Long;
    public "sun/nio/fs/UnixFileAttributes/st_ctime_sec": Long;
    public "sun/nio/fs/UnixFileAttributes/st_ctime_nsec": Long;
    public "sun/nio/fs/UnixFileAttributes/st_birthtime_sec": Long;
    public "sun/nio/fs/UnixFileAttributes/owner": java_nio_file_attribute_UserPrincipal;
    public "sun/nio/fs/UnixFileAttributes/group": java_nio_file_attribute_GroupPrincipal;
    public "sun/nio/fs/UnixFileAttributes/key": sun_nio_fs_UnixFileKey;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileAttributes/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "get(Lsun/nio/fs/UnixPath;Z)Lsun/nio/fs/UnixFileAttributes;"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileAttributes) => void): void;
    public static "sun/nio/fs/UnixFileAttributes/get(Lsun/nio/fs/UnixPath;Z)Lsun/nio/fs/UnixFileAttributes;"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileAttributes) => void): void;
    public static "get(I)Lsun/nio/fs/UnixFileAttributes;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileAttributes) => void): void;
    public static "sun/nio/fs/UnixFileAttributes/get(I)Lsun/nio/fs/UnixFileAttributes;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileAttributes) => void): void;
    public static "get(ILsun/nio/fs/UnixPath;Z)Lsun/nio/fs/UnixFileAttributes;"(thread: JVMThread, args: [number, sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileAttributes) => void): void;
    public static "sun/nio/fs/UnixFileAttributes/get(ILsun/nio/fs/UnixPath;Z)Lsun/nio/fs/UnixFileAttributes;"(thread: JVMThread, args: [number, sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileAttributes) => void): void;
    public "isSameFile(Lsun/nio/fs/UnixFileAttributes;)Z"(thread: JVMThread, args: [sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/isSameFile(Lsun/nio/fs/UnixFileAttributes;)Z"(thread: JVMThread, args: [sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/mode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ino()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/UnixFileAttributes/ino()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "dev()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/UnixFileAttributes/dev()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "rdev()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/UnixFileAttributes/rdev()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "nlink()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/nlink()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "uid()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/uid()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "gid()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/gid()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toFileTime(JJ)Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public static "sun/nio/fs/UnixFileAttributes/toFileTime(JJ)Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "ctime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "sun/nio/fs/UnixFileAttributes/ctime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "isDevice()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/isDevice()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastModifiedTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "sun/nio/fs/UnixFileAttributes/lastModifiedTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "lastAccessTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "sun/nio/fs/UnixFileAttributes/lastAccessTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "creationTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "sun/nio/fs/UnixFileAttributes/creationTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "isRegularFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/isRegularFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/isDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSymbolicLink()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/isSymbolicLink()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOther()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileAttributes/isOther()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/UnixFileAttributes/size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "fileKey()Lsun/nio/fs/UnixFileKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileKey) => void): void;
    public "sun/nio/fs/UnixFileAttributes/fileKey()Lsun/nio/fs/UnixFileKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileKey) => void): void;
    public "owner()Ljava/nio/file/attribute/UserPrincipal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_UserPrincipal) => void): void;
    public "sun/nio/fs/UnixFileAttributes/owner()Ljava/nio/file/attribute/UserPrincipal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_UserPrincipal) => void): void;
    public "group()Ljava/nio/file/attribute/GroupPrincipal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_GroupPrincipal) => void): void;
    public "sun/nio/fs/UnixFileAttributes/group()Ljava/nio/file/attribute/GroupPrincipal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_GroupPrincipal) => void): void;
    public "permissions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "sun/nio/fs/UnixFileAttributes/permissions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "asBasicFileAttributes()Ljava/nio/file/attribute/BasicFileAttributes;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_BasicFileAttributes) => void): void;
    public "sun/nio/fs/UnixFileAttributes/asBasicFileAttributes()Ljava/nio/file/attribute/BasicFileAttributes;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_BasicFileAttributes) => void): void;
    public static "toUnixFileAttributes(Ljava/nio/file/attribute/BasicFileAttributes;)Lsun/nio/fs/UnixFileAttributes;"(thread: JVMThread, args: [java_nio_file_attribute_BasicFileAttributes], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileAttributes) => void): void;
    public static "sun/nio/fs/UnixFileAttributes/toUnixFileAttributes(Ljava/nio/file/attribute/BasicFileAttributes;)Lsun/nio/fs/UnixFileAttributes;"(thread: JVMThread, args: [java_nio_file_attribute_BasicFileAttributes], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixFileAttributes) => void): void;
    public "fileKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/fs/UnixFileAttributes/fileKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_nio_fs_UnixFileKey extends java_lang_Object {
    public "sun/nio/fs/UnixFileKey/st_dev": Long;
    public "sun/nio/fs/UnixFileKey/st_ino": Long;
    public "<init>(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixFileKey/<init>(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileKey/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixFileKey/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixFileKey/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_nio_file_attribute_PosixFileAttributes extends java_lang_Object, java_nio_file_attribute_BasicFileAttributes {
    "owner()Ljava/nio/file/attribute/UserPrincipal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_UserPrincipal) => void): void;
    "group()Ljava/nio/file/attribute/GroupPrincipal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_GroupPrincipal) => void): void;
    "permissions()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "lastModifiedTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    "lastAccessTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    "creationTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    "isRegularFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSymbolicLink()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isOther()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "size()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "fileKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_nio_fs_UnixNativeDispatcher extends java_lang_Object {
    public static "sun/nio/fs/UnixNativeDispatcher/SUPPORTS_OPENAT": number;
    public static "sun/nio/fs/UnixNativeDispatcher/SUPPORTS_FUTIMES": number;
    public static "sun/nio/fs/UnixNativeDispatcher/SUPPORTS_BIRTHTIME": number;
    public static "sun/nio/fs/UnixNativeDispatcher/capabilities": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixNativeDispatcher/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyToNativeBuffer(Lsun/nio/fs/UnixPath;)Lsun/nio/fs/NativeBuffer;"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_NativeBuffer) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/copyToNativeBuffer(Lsun/nio/fs/UnixPath;)Lsun/nio/fs/NativeBuffer;"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_NativeBuffer) => void): void;
    public static "getcwd()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/getcwd()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "dup(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/dup(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "open(Lsun/nio/fs/UnixPath;II)I"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/open(Lsun/nio/fs/UnixPath;II)I"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "open0(JII)I"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/open0(JII)I"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "openat(I[BII)I"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/openat(I[BII)I"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "openat0(IJII)I"(thread: JVMThread, args: [number, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/openat0(IJII)I"(thread: JVMThread, args: [number, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "close(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/close(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fopen(Lsun/nio/fs/UnixPath;Ljava/lang/String;)J"(thread: JVMThread, args: [sun_nio_fs_UnixPath, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fopen(Lsun/nio/fs/UnixPath;Ljava/lang/String;)J"(thread: JVMThread, args: [sun_nio_fs_UnixPath, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "fopen0(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fopen0(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "fclose(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fclose(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "link(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/link(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "link0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/link0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "unlink(Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/unlink(Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "unlink0(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/unlink0(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "unlinkat(I[BI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/unlinkat(I[BI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "unlinkat0(IJI)V"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/unlinkat0(IJI)V"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "mknod(Lsun/nio/fs/UnixPath;IJ)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/mknod(Lsun/nio/fs/UnixPath;IJ)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "mknod0(JIJ)V"(thread: JVMThread, args: [Long, any, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/mknod0(JIJ)V"(thread: JVMThread, args: [Long, any, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "rename(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/rename(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "rename0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/rename0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "renameat(I[BI[B)V"(thread: JVMThread, args: [number, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/renameat(I[BI[B)V"(thread: JVMThread, args: [number, JVMArray<number>, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "renameat0(IJIJ)V"(thread: JVMThread, args: [number, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/renameat0(IJIJ)V"(thread: JVMThread, args: [number, Long, any, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "mkdir(Lsun/nio/fs/UnixPath;I)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/mkdir(Lsun/nio/fs/UnixPath;I)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "mkdir0(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/mkdir0(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "rmdir(Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/rmdir(Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "rmdir0(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/rmdir0(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readlink(Lsun/nio/fs/UnixPath;)[B"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/readlink(Lsun/nio/fs/UnixPath;)[B"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "readlink0(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/readlink0(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "realpath(Lsun/nio/fs/UnixPath;)[B"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/realpath(Lsun/nio/fs/UnixPath;)[B"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "realpath0(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/realpath0(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "symlink([BLsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [JVMArray<number>, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/symlink([BLsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [JVMArray<number>, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "symlink0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/symlink0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "stat(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/stat(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "stat0(JLsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [Long, any, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/stat0(JLsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [Long, any, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "lstat(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/lstat(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "lstat0(JLsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [Long, any, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/lstat0(JLsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [Long, any, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fstat(ILsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [number, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fstat(ILsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [number, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fstatat(I[BILsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [number, JVMArray<number>, number, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fstatat(I[BILsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [number, JVMArray<number>, number, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fstatat0(IJILsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [number, Long, any, number, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fstatat0(IJILsun/nio/fs/UnixFileAttributes;)V"(thread: JVMThread, args: [number, Long, any, number, sun_nio_fs_UnixFileAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "chown(Lsun/nio/fs/UnixPath;II)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/chown(Lsun/nio/fs/UnixPath;II)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "chown0(JII)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/chown0(JII)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "lchown(Lsun/nio/fs/UnixPath;II)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/lchown(Lsun/nio/fs/UnixPath;II)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "lchown0(JII)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/lchown0(JII)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fchown(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fchown(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "chmod(Lsun/nio/fs/UnixPath;I)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/chmod(Lsun/nio/fs/UnixPath;I)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "chmod0(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/chmod0(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fchmod(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fchmod(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "utimes(Lsun/nio/fs/UnixPath;JJ)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/utimes(Lsun/nio/fs/UnixPath;JJ)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "utimes0(JJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/utimes0(JJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "futimes(IJJ)V"(thread: JVMThread, args: [number, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/futimes(IJJ)V"(thread: JVMThread, args: [number, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "opendir(Lsun/nio/fs/UnixPath;)J"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/opendir(Lsun/nio/fs/UnixPath;)J"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "opendir0(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/opendir0(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "fdopendir(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fdopendir(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "closedir(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/closedir(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readdir(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/readdir(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "read(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/read(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "write(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/write(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access(Lsun/nio/fs/UnixPath;I)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/access(Lsun/nio/fs/UnixPath;I)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access0(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/access0(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getpwuid(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/getpwuid(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getgrgid(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/getgrgid(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getpwnam(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/getpwnam(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getpwnam0(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/getpwnam0(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getgrnam(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/getgrnam(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getgrnam0(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/getgrnam0(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "statvfs(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileStoreAttributes;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileStoreAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/statvfs(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileStoreAttributes;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileStoreAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "statvfs0(JLsun/nio/fs/UnixFileStoreAttributes;)V"(thread: JVMThread, args: [Long, any, sun_nio_fs_UnixFileStoreAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/statvfs0(JLsun/nio/fs/UnixFileStoreAttributes;)V"(thread: JVMThread, args: [Long, any, sun_nio_fs_UnixFileStoreAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public static "pathconf(Lsun/nio/fs/UnixPath;I)J"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/pathconf(Lsun/nio/fs/UnixPath;I)J"(thread: JVMThread, args: [sun_nio_fs_UnixPath, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "pathconf0(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/pathconf0(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "fpathconf(II)J"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/fpathconf(II)J"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "strerror(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/strerror(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "openatSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/openatSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "futimesSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/futimesSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "birthtimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/birthtimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "init()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/init()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixNativeDispatcher/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_NativeBuffer extends java_lang_Object {
    public static "sun/nio/fs/NativeBuffer/unsafe": sun_misc_Unsafe;
    public "sun/nio/fs/NativeBuffer/address": Long;
    public "sun/nio/fs/NativeBuffer/size": number;
    public "sun/nio/fs/NativeBuffer/cleaner": sun_misc_Cleaner;
    public "sun/nio/fs/NativeBuffer/owner": java_lang_Object;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/NativeBuffer/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "release()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/NativeBuffer/release()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "address()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/NativeBuffer/address()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/NativeBuffer/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "cleaner()Lsun/misc/Cleaner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Cleaner) => void): void;
    public "sun/nio/fs/NativeBuffer/cleaner()Lsun/misc/Cleaner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Cleaner) => void): void;
    public "setOwner(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/NativeBuffer/setOwner(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "owner()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/fs/NativeBuffer/owner()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$000()Lsun/misc/Unsafe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Unsafe) => void): void;
    public static "sun/nio/fs/NativeBuffer/access$000()Lsun/misc/Unsafe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Unsafe) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/NativeBuffer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_Cleaner extends java_lang_ref_PhantomReference {
    public static "sun/misc/Cleaner/dummyQueue": java_lang_ref_ReferenceQueue;
    public static "sun/misc/Cleaner/first": sun_misc_Cleaner;
    public "sun/misc/Cleaner/next": sun_misc_Cleaner;
    public "sun/misc/Cleaner/prev": sun_misc_Cleaner;
    public "sun/misc/Cleaner/thunk": java_lang_Runnable;
    public static "add(Lsun/misc/Cleaner;)Lsun/misc/Cleaner;"(thread: JVMThread, args: [sun_misc_Cleaner], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Cleaner) => void): void;
    public static "sun/misc/Cleaner/add(Lsun/misc/Cleaner;)Lsun/misc/Cleaner;"(thread: JVMThread, args: [sun_misc_Cleaner], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Cleaner) => void): void;
    public static "remove(Lsun/misc/Cleaner;)Z"(thread: JVMThread, args: [sun_misc_Cleaner], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Cleaner/remove(Lsun/misc/Cleaner;)Z"(thread: JVMThread, args: [sun_misc_Cleaner], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/Object;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Cleaner/<init>(Ljava/lang/Object;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "create(Ljava/lang/Object;Ljava/lang/Runnable;)Lsun/misc/Cleaner;"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Cleaner) => void): void;
    public static "sun/misc/Cleaner/create(Ljava/lang/Object;Ljava/lang/Runnable;)Lsun/misc/Cleaner;"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Cleaner) => void): void;
    public "clean()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Cleaner/clean()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Cleaner/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ref_PhantomReference extends java_lang_ref_Reference {
    public "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ref/PhantomReference/get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "<init>(Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_ref_ReferenceQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/PhantomReference/<init>(Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_ref_ReferenceQueue], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_UnixCopyFile extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixCopyFile/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyDirectory(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixCopyFile/copyDirectory(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyFile(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;J)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixCopyFile/copyFile(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;J)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyLink(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixCopyFile/copyLink(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copySpecial(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixCopyFile/copySpecial(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags], cb?: (e?: java_lang_Throwable) => void): void;
    public static "move(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixCopyFile/move(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copy(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixCopyFile/copy(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "transfer(IIJ)V"(thread: JVMThread, args: [number, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixCopyFile/transfer(IIJ)V"(thread: JVMThread, args: [number, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;J)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixCopyFile/access$000(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixFileAttributes;Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixCopyFile$Flags;J)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixFileAttributes, sun_nio_fs_UnixPath, sun_nio_fs_UnixCopyFile$Flags, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/UnixCopyFile/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_UnixCopyFile$Flags extends java_lang_Object {
    public "sun/nio/fs/UnixCopyFile$Flags/replaceExisting": number;
    public "sun/nio/fs/UnixCopyFile$Flags/atomicMove": number;
    public "sun/nio/fs/UnixCopyFile$Flags/followLinks": number;
    public "sun/nio/fs/UnixCopyFile$Flags/interruptible": number;
    public "sun/nio/fs/UnixCopyFile$Flags/copyBasicAttributes": number;
    public "sun/nio/fs/UnixCopyFile$Flags/copyPosixAttributes": number;
    public "sun/nio/fs/UnixCopyFile$Flags/copyNonPosixAttributes": number;
    public "sun/nio/fs/UnixCopyFile$Flags/failIfUnableToCopyBasic": number;
    public "sun/nio/fs/UnixCopyFile$Flags/failIfUnableToCopyPosix": number;
    public "sun/nio/fs/UnixCopyFile$Flags/failIfUnableToCopyNonPosix": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixCopyFile$Flags/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fromCopyOptions([Ljava/nio/file/CopyOption;)Lsun/nio/fs/UnixCopyFile$Flags;"(thread: JVMThread, args: [JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixCopyFile$Flags) => void): void;
    public static "sun/nio/fs/UnixCopyFile$Flags/fromCopyOptions([Ljava/nio/file/CopyOption;)Lsun/nio/fs/UnixCopyFile$Flags;"(thread: JVMThread, args: [JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixCopyFile$Flags) => void): void;
    public static "fromMoveOptions([Ljava/nio/file/CopyOption;)Lsun/nio/fs/UnixCopyFile$Flags;"(thread: JVMThread, args: [JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixCopyFile$Flags) => void): void;
    public static "sun/nio/fs/UnixCopyFile$Flags/fromMoveOptions([Ljava/nio/file/CopyOption;)Lsun/nio/fs/UnixCopyFile$Flags;"(thread: JVMThread, args: [JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_UnixCopyFile$Flags) => void): void;
  }
  export class sun_nio_fs_MagicFileTypeDetector extends sun_nio_fs_AbstractFileTypeDetector {
    public static "sun/nio/fs/MagicFileTypeDetector/UNKNOW_MIME_TYPE": java_lang_String;
    public "sun/nio/fs/MagicFileTypeDetector/libmagicAvailable": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/MagicFileTypeDetector/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implProbeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/MagicFileTypeDetector/implProbeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initialize0()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/MagicFileTypeDetector/initialize0()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "probe0(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/MagicFileTypeDetector/probe0(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/MagicFileTypeDetector/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_LinuxWatchService extends sun_nio_fs_AbstractWatchService {
    public static "sun/nio/fs/LinuxWatchService/unsafe": sun_misc_Unsafe;
    public "sun/nio/fs/LinuxWatchService/poller": sun_nio_fs_LinuxWatchService$Poller;
    public "<init>(Lsun/nio/fs/UnixFileSystem;)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/LinuxWatchService/<init>(Lsun/nio/fs/UnixFileSystem;)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem], cb?: (e?: java_lang_Throwable) => void): void;
    public "register(Ljava/nio/file/Path;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "sun/nio/fs/LinuxWatchService/register(Ljava/nio/file/Path;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/LinuxWatchService/implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "eventSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxWatchService/eventSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "eventOffsets()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/LinuxWatchService/eventOffsets()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "inotifyInit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxWatchService/inotifyInit()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "inotifyAddWatch(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxWatchService/inotifyAddWatch(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "inotifyRmWatch(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxWatchService/inotifyRmWatch(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "configureBlocking(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxWatchService/configureBlocking(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "socketpair([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxWatchService/socketpair([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "poll(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxWatchService/poll(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxWatchService/access$000(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$100(Lsun/nio/fs/LinuxWatchService;)Lsun/nio/fs/LinuxWatchService$Poller;"(thread: JVMThread, args: [sun_nio_fs_LinuxWatchService], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_LinuxWatchService$Poller) => void): void;
    public static "sun/nio/fs/LinuxWatchService/access$100(Lsun/nio/fs/LinuxWatchService;)Lsun/nio/fs/LinuxWatchService$Poller;"(thread: JVMThread, args: [sun_nio_fs_LinuxWatchService], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_LinuxWatchService$Poller) => void): void;
    public static "access$200()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxWatchService/access$200()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$300()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/LinuxWatchService/access$300()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$400()Lsun/misc/Unsafe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Unsafe) => void): void;
    public static "sun/nio/fs/LinuxWatchService/access$400()Lsun/misc/Unsafe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Unsafe) => void): void;
    public static "access$500(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxWatchService/access$500(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$600(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxWatchService/access$600(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxWatchService/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_LinuxWatchService$Poller extends sun_nio_fs_AbstractPoller {
    public static "sun/nio/fs/LinuxWatchService$Poller/SIZEOF_INOTIFY_EVENT": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/offsets": JVMArray<number>;
    public static "sun/nio/fs/LinuxWatchService$Poller/OFFSETOF_WD": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/OFFSETOF_MASK": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/OFFSETOF_LEN": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/OFFSETOF_NAME": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/IN_MODIFY": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/IN_ATTRIB": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/IN_MOVED_FROM": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/IN_MOVED_TO": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/IN_CREATE": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/IN_DELETE": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/IN_UNMOUNT": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/IN_Q_OVERFLOW": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/IN_IGNORED": number;
    public static "sun/nio/fs/LinuxWatchService$Poller/BUFFER_SIZE": number;
    public "sun/nio/fs/LinuxWatchService$Poller/fs": sun_nio_fs_UnixFileSystem;
    public "sun/nio/fs/LinuxWatchService$Poller/watcher": sun_nio_fs_LinuxWatchService;
    public "sun/nio/fs/LinuxWatchService$Poller/ifd": number;
    public "sun/nio/fs/LinuxWatchService$Poller/socketpair": JVMArray<number>;
    public "sun/nio/fs/LinuxWatchService$Poller/wdToKey": java_util_Map;
    public "sun/nio/fs/LinuxWatchService$Poller/address": Long;
    public "<init>(Lsun/nio/fs/UnixFileSystem;Lsun/nio/fs/LinuxWatchService;I[I)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem, sun_nio_fs_LinuxWatchService, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/LinuxWatchService$Poller/<init>(Lsun/nio/fs/UnixFileSystem;Lsun/nio/fs/LinuxWatchService;I[I)V"(thread: JVMThread, args: [sun_nio_fs_UnixFileSystem, sun_nio_fs_LinuxWatchService, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "wakeup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/LinuxWatchService$Poller/wakeup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implRegister(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/fs/LinuxWatchService$Poller/implRegister(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implCancelKey(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/LinuxWatchService$Poller/implCancelKey(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/LinuxWatchService$Poller/implCloseAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/LinuxWatchService$Poller/run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "maskToEventKind(I)Ljava/nio/file/WatchEvent$Kind;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchEvent$Kind) => void): void;
    public "sun/nio/fs/LinuxWatchService$Poller/maskToEventKind(I)Ljava/nio/file/WatchEvent$Kind;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchEvent$Kind) => void): void;
    public "processEvent(IILsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [number, number, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/LinuxWatchService$Poller/processEvent(IILsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [number, number, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxWatchService$Poller/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_AbstractPoller extends java_lang_Object implements java_lang_Runnable {
    public "sun/nio/fs/AbstractPoller/requestList": java_util_LinkedList;
    public "sun/nio/fs/AbstractPoller/shutdown": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractPoller/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "start()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractPoller/start()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "wakeup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractPoller/wakeup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implRegister(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/fs/AbstractPoller/implRegister(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implCancelKey(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractPoller/implCancelKey(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "implCloseAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractPoller/implCloseAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "register(Ljava/nio/file/Path;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "sun/nio/fs/AbstractPoller/register(Ljava/nio/file/Path;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "cancel(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractPoller/cancel(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractPoller/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "invoke(Lsun/nio/fs/AbstractPoller$RequestType;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_nio_fs_AbstractPoller$RequestType, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/fs/AbstractPoller/invoke(Lsun/nio/fs/AbstractPoller$RequestType;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_nio_fs_AbstractPoller$RequestType, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "processRequests()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/AbstractPoller/processRequests()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_AbstractPoller$RequestType extends java_lang_Enum {
    public static "sun/nio/fs/AbstractPoller$RequestType/REGISTER": sun_nio_fs_AbstractPoller$RequestType;
    public static "sun/nio/fs/AbstractPoller$RequestType/CANCEL": sun_nio_fs_AbstractPoller$RequestType;
    public static "sun/nio/fs/AbstractPoller$RequestType/CLOSE": sun_nio_fs_AbstractPoller$RequestType;
    public static "sun/nio/fs/AbstractPoller$RequestType/$VALUES": JVMArray<sun_nio_fs_AbstractPoller$RequestType>;
    public static "values()[Lsun/nio/fs/AbstractPoller$RequestType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_nio_fs_AbstractPoller$RequestType>) => void): void;
    public static "sun/nio/fs/AbstractPoller$RequestType/values()[Lsun/nio/fs/AbstractPoller$RequestType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_nio_fs_AbstractPoller$RequestType>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lsun/nio/fs/AbstractPoller$RequestType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_AbstractPoller$RequestType) => void): void;
    public static "sun/nio/fs/AbstractPoller$RequestType/valueOf(Ljava/lang/String;)Lsun/nio/fs/AbstractPoller$RequestType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_fs_AbstractPoller$RequestType) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractPoller$RequestType/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/AbstractPoller$RequestType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_LinkedList extends java_util_AbstractSequentialList implements java_util_List, java_util_Deque, java_lang_Cloneable, java_io_Serializable {
    public "java/util/LinkedList/size": number;
    public "java/util/LinkedList/first": java_util_LinkedList$Node;
    public "java/util/LinkedList/last": java_util_LinkedList$Node;
    public static "java/util/LinkedList/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "linkFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/linkFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "linkLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/linkLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "linkBefore(Ljava/lang/Object;Ljava/util/LinkedList$Node;)V"(thread: JVMThread, args: [java_lang_Object, java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/linkBefore(Ljava/lang/Object;Ljava/util/LinkedList$Node;)V"(thread: JVMThread, args: [java_lang_Object, java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "unlinkFirst(Ljava/util/LinkedList$Node;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/unlinkFirst(Ljava/util/LinkedList$Node;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "unlinkLast(Ljava/util/LinkedList$Node;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/unlinkLast(Ljava/util/LinkedList$Node;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "unlink(Ljava/util/LinkedList$Node;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/unlink(Ljava/util/LinkedList$Node;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/getFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/getLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "removeFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/removeFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "removeLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/removeLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "addFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/addFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "addLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/addLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isElementIndex(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/isElementIndex(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPositionIndex(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/isPositionIndex(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "outOfBoundsMsg(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/LinkedList/outOfBoundsMsg(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "checkElementIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/checkElementIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkPositionIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/checkPositionIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "node(I)Ljava/util/LinkedList$Node;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList$Node) => void): void;
    public "java/util/LinkedList/node(I)Ljava/util/LinkedList$Node;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList$Node) => void): void;
    public "indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offerFirst(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/offerFirst(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offerLast(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/offerLast(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peekFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/peekFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "peekLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/peekLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "pollFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/pollFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "pollLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/pollLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "push(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/push(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "pop()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/pop()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "removeFirstOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/removeFirstOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeLastOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedList/removeLastOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/LinkedList/listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "descendingIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/LinkedList/descendingIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "superClone()Ljava/util/LinkedList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList) => void): void;
    public "java/util/LinkedList/superClone()Ljava/util/LinkedList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedList/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/LinkedList/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/LinkedList/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/LinkedList/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_LinkedList$Node extends java_lang_Object {
    public "java/util/LinkedList$Node/item": java_lang_Object;
    public "java/util/LinkedList$Node/next": java_util_LinkedList$Node;
    public "java/util/LinkedList$Node/prev": java_util_LinkedList$Node;
    public "<init>(Ljava/util/LinkedList$Node;Ljava/lang/Object;Ljava/util/LinkedList$Node;)V"(thread: JVMThread, args: [java_util_LinkedList$Node, java_lang_Object, java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedList$Node/<init>(Ljava/util/LinkedList$Node;Ljava/lang/Object;Ljava/util/LinkedList$Node;)V"(thread: JVMThread, args: [java_util_LinkedList$Node, java_lang_Object, java_util_LinkedList$Node], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_Deque extends java_lang_Object, java_util_Queue {
    "addFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "addLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "offerFirst(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "offerLast(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "removeLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "pollFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "pollLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "peekFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "peekLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "removeFirstOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeLastOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "push(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "pop()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "descendingIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_AbstractSequentialList extends java_util_AbstractList {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractSequentialList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractSequentialList/get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractSequentialList/set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/AbstractSequentialList/add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/AbstractSequentialList/remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/AbstractSequentialList/addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/AbstractSequentialList/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/AbstractSequentialList/listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
  }
  export class sun_nio_fs_AbstractWatchService extends java_lang_Object implements java_nio_file_WatchService {
    public "sun/nio/fs/AbstractWatchService/pendingKeys": java_util_concurrent_LinkedBlockingDeque;
    public "sun/nio/fs/AbstractWatchService/CLOSE_KEY": java_nio_file_WatchKey;
    public "sun/nio/fs/AbstractWatchService/closed": number;
    public "sun/nio/fs/AbstractWatchService/closeLock": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractWatchService/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "register(Ljava/nio/file/Path;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "sun/nio/fs/AbstractWatchService/register(Ljava/nio/file/Path;[Ljava/nio/file/WatchEvent$Kind;[Ljava/nio/file/WatchEvent$Modifier;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_WatchEvent$Kind>, JVMArray<java_nio_file_WatchEvent$Modifier>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "enqueueKey(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractWatchService/enqueueKey(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractWatchService/checkOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkKey(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractWatchService/checkKey(Ljava/nio/file/WatchKey;)V"(thread: JVMThread, args: [java_nio_file_WatchKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "poll()Ljava/nio/file/WatchKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "sun/nio/fs/AbstractWatchService/poll()Ljava/nio/file/WatchKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "poll(JLjava/util/concurrent/TimeUnit;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "sun/nio/fs/AbstractWatchService/poll(JLjava/util/concurrent/TimeUnit;)Ljava/nio/file/WatchKey;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "take()Ljava/nio/file/WatchKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "sun/nio/fs/AbstractWatchService/take()Ljava/nio/file/WatchKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_WatchKey) => void): void;
    public "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/AbstractWatchService/isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "closeLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/nio/fs/AbstractWatchService/closeLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractWatchService/implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/AbstractWatchService/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_LinkedBlockingDeque extends java_util_AbstractQueue implements java_util_concurrent_BlockingDeque, java_io_Serializable {
    public static "java/util/concurrent/LinkedBlockingDeque/serialVersionUID": Long;
    public "java/util/concurrent/LinkedBlockingDeque/first": java_util_concurrent_LinkedBlockingDeque$Node;
    public "java/util/concurrent/LinkedBlockingDeque/last": java_util_concurrent_LinkedBlockingDeque$Node;
    public "java/util/concurrent/LinkedBlockingDeque/count": number;
    public "java/util/concurrent/LinkedBlockingDeque/capacity": number;
    public "java/util/concurrent/LinkedBlockingDeque/lock": java_util_concurrent_locks_ReentrantLock;
    public "java/util/concurrent/LinkedBlockingDeque/notEmpty": java_util_concurrent_locks_Condition;
    public "java/util/concurrent/LinkedBlockingDeque/notFull": java_util_concurrent_locks_Condition;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "linkFirst(Ljava/util/concurrent/LinkedBlockingDeque$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_LinkedBlockingDeque$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/linkFirst(Ljava/util/concurrent/LinkedBlockingDeque$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_LinkedBlockingDeque$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "linkLast(Ljava/util/concurrent/LinkedBlockingDeque$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_LinkedBlockingDeque$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/linkLast(Ljava/util/concurrent/LinkedBlockingDeque$Node;)Z"(thread: JVMThread, args: [java_util_concurrent_LinkedBlockingDeque$Node], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unlinkFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/unlinkFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "unlinkLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/unlinkLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "unlink(Ljava/util/concurrent/LinkedBlockingDeque$Node;)V"(thread: JVMThread, args: [java_util_concurrent_LinkedBlockingDeque$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/unlink(Ljava/util/concurrent/LinkedBlockingDeque$Node;)V"(thread: JVMThread, args: [java_util_concurrent_LinkedBlockingDeque$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "addFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/addFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "addLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/addLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "offerFirst(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/offerFirst(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offerLast(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/offerLast(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/putFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "putLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/putLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "offerFirst(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/offerFirst(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offerLast(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/offerLast(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/removeFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "removeLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/removeLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "pollFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/pollFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "pollLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/pollLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "takeFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/takeFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "takeLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/takeLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "pollFirst(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/pollFirst(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "pollLast(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/pollLast(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/getFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/getLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "peekFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/peekFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "peekLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/peekLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "removeFirstOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/removeFirstOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeLastOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/removeLastOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/put(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "offer(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/offer(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "take()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/take()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "poll(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/poll(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remainingCapacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/remainingCapacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drainTo(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/drainTo(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drainTo(Ljava/util/Collection;I)I"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/drainTo(Ljava/util/Collection;I)I"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "push(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/push(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "pop()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/pop()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "descendingIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/descendingIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_LinkedBlockingDeque$Node extends java_lang_Object {
    public "java/util/concurrent/LinkedBlockingDeque$Node/item": java_lang_Object;
    public "java/util/concurrent/LinkedBlockingDeque$Node/prev": java_util_concurrent_LinkedBlockingDeque$Node;
    public "java/util/concurrent/LinkedBlockingDeque$Node/next": java_util_concurrent_LinkedBlockingDeque$Node;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/LinkedBlockingDeque$Node/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_concurrent_BlockingDeque extends java_lang_Object, java_util_concurrent_BlockingQueue, java_util_Deque {
    "addFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "addLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "offerFirst(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "offerLast(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "putFirst(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "putLast(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "offerFirst(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "offerLast(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "takeFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "takeLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "pollFirst(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "pollLast(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "removeFirstOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeLastOccurrence(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "put(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "offer(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "take()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "poll(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "element()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "push(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "remainingCapacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "drainTo(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "drainTo(Ljava/util/Collection;I)I"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "removeFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "removeLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "pollFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "pollLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "peekFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "peekLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "pop()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "descendingIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_UnixMountEntry extends java_lang_Object {
    public "sun/nio/fs/UnixMountEntry/name": JVMArray<number>;
    public "sun/nio/fs/UnixMountEntry/dir": JVMArray<number>;
    public "sun/nio/fs/UnixMountEntry/fstype": JVMArray<number>;
    public "sun/nio/fs/UnixMountEntry/opts": JVMArray<number>;
    public "sun/nio/fs/UnixMountEntry/dev": Long;
    public "sun/nio/fs/UnixMountEntry/fstypeAsString": java_lang_String;
    public "sun/nio/fs/UnixMountEntry/optionsAsString": java_lang_String;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixMountEntry/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixMountEntry/name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "fstype()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixMountEntry/fstype()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "dir()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/nio/fs/UnixMountEntry/dir()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "dev()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/nio/fs/UnixMountEntry/dev()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "hasOption(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixMountEntry/hasOption(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isIgnored()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixMountEntry/isIgnored()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixMountEntry/isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_nio_fs_LinuxNativeDispatcher extends sun_nio_fs_UnixNativeDispatcher {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/LinuxNativeDispatcher/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setmntent([B[B)J"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/setmntent([B[B)J"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "setmntent0(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/setmntent0(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getmntent(JLsun/nio/fs/UnixMountEntry;)I"(thread: JVMThread, args: [Long, any, sun_nio_fs_UnixMountEntry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/getmntent(JLsun/nio/fs/UnixMountEntry;)I"(thread: JVMThread, args: [Long, any, sun_nio_fs_UnixMountEntry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "endmntent(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/endmntent(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fgetxattr(I[BJI)I"(thread: JVMThread, args: [number, JVMArray<number>, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/fgetxattr(I[BJI)I"(thread: JVMThread, args: [number, JVMArray<number>, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "fgetxattr0(IJJI)I"(thread: JVMThread, args: [number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/fgetxattr0(IJJI)I"(thread: JVMThread, args: [number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "fsetxattr(I[BJI)V"(thread: JVMThread, args: [number, JVMArray<number>, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/fsetxattr(I[BJI)V"(thread: JVMThread, args: [number, JVMArray<number>, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fsetxattr0(IJJI)V"(thread: JVMThread, args: [number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/fsetxattr0(IJJI)V"(thread: JVMThread, args: [number, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fremovexattr(I[B)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/fremovexattr(I[B)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fremovexattr0(IJ)V"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/fremovexattr0(IJ)V"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "flistxattr(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/flistxattr(IJI)I"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/LinuxNativeDispatcher/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_GnomeFileTypeDetector extends sun_nio_fs_AbstractFileTypeDetector {
    public static "sun/nio/fs/GnomeFileTypeDetector/GNOME_VFS_MIME_TYPE_UNKNOWN": java_lang_String;
    public "sun/nio/fs/GnomeFileTypeDetector/gioAvailable": number;
    public "sun/nio/fs/GnomeFileTypeDetector/gnomeVfsAvailable": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/GnomeFileTypeDetector/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implProbeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/GnomeFileTypeDetector/implProbeContentType(Ljava/nio/file/Path;)Ljava/lang/String;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initializeGio()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/GnomeFileTypeDetector/initializeGio()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "probeUsingGio(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/GnomeFileTypeDetector/probeUsingGio(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "initializeGnomeVfs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/nio/fs/GnomeFileTypeDetector/initializeGnomeVfs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "probeUsingGnomeVfs(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/nio/fs/GnomeFileTypeDetector/probeUsingGnomeVfs(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/fs/GnomeFileTypeDetector/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_file_spi_FileSystemProvider extends java_lang_Object {
    public static "java/nio/file/spi/FileSystemProvider/lock": java_lang_Object;
    public static "java/nio/file/spi/FileSystemProvider/installedProviders": java_util_List;
    public static "java/nio/file/spi/FileSystemProvider/loadingProviders": number;
    public static "checkPermission()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "java/nio/file/spi/FileSystemProvider/checkPermission()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "<init>(Ljava/lang/Void;)V"(thread: JVMThread, args: [java_lang_Void], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/<init>(Ljava/lang/Void;)V"(thread: JVMThread, args: [java_lang_Void], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "loadInstalledProviders()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/nio/file/spi/FileSystemProvider/loadInstalledProviders()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "installedProviders()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/nio/file/spi/FileSystemProvider/installedProviders()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getScheme()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/spi/FileSystemProvider/getScheme()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "newFileSystem(Ljava/net/URI;Ljava/util/Map;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_net_URI, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "java/nio/file/spi/FileSystemProvider/newFileSystem(Ljava/net/URI;Ljava/util/Map;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_net_URI, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "getFileSystem(Ljava/net/URI;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "java/nio/file/spi/FileSystemProvider/getFileSystem(Ljava/net/URI;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "getPath(Ljava/net/URI;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "java/nio/file/spi/FileSystemProvider/getPath(Ljava/net/URI;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "newFileSystem(Ljava/nio/file/Path;Ljava/util/Map;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "java/nio/file/spi/FileSystemProvider/newFileSystem(Ljava/nio/file/Path;Ljava/util/Map;)Ljava/nio/file/FileSystem;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileSystem) => void): void;
    public "newInputStream(Ljava/nio/file/Path;[Ljava/nio/file/OpenOption;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_OpenOption>], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/nio/file/spi/FileSystemProvider/newInputStream(Ljava/nio/file/Path;[Ljava/nio/file/OpenOption;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_OpenOption>], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "newOutputStream(Ljava/nio/file/Path;[Ljava/nio/file/OpenOption;)Ljava/io/OutputStream;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_OpenOption>], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "java/nio/file/spi/FileSystemProvider/newOutputStream(Ljava/nio/file/Path;[Ljava/nio/file/OpenOption;)Ljava/io/OutputStream;"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_OpenOption>], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "newFileChannel(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "java/nio/file/spi/FileSystemProvider/newFileChannel(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "newAsynchronousFileChannel(Ljava/nio/file/Path;Ljava/util/Set;Ljava/util/concurrent/ExecutorService;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, java_util_concurrent_ExecutorService, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public "java/nio/file/spi/FileSystemProvider/newAsynchronousFileChannel(Ljava/nio/file/Path;Ljava/util/Set;Ljava/util/concurrent/ExecutorService;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/AsynchronousFileChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, java_util_concurrent_ExecutorService, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_AsynchronousFileChannel) => void): void;
    public "newByteChannel(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "java/nio/file/spi/FileSystemProvider/newByteChannel(Ljava/nio/file/Path;Ljava/util/Set;[Ljava/nio/file/attribute/FileAttribute;)Ljava/nio/channels/SeekableByteChannel;"(thread: JVMThread, args: [java_nio_file_Path, java_util_Set, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_SeekableByteChannel) => void): void;
    public "newDirectoryStream(Ljava/nio/file/Path;Ljava/nio/file/DirectoryStream$Filter;)Ljava/nio/file/DirectoryStream;"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_DirectoryStream$Filter], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_DirectoryStream) => void): void;
    public "java/nio/file/spi/FileSystemProvider/newDirectoryStream(Ljava/nio/file/Path;Ljava/nio/file/DirectoryStream$Filter;)Ljava/nio/file/DirectoryStream;"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_DirectoryStream$Filter], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_DirectoryStream) => void): void;
    public "createDirectory(Ljava/nio/file/Path;[Ljava/nio/file/attribute/FileAttribute;)V"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/createDirectory(Ljava/nio/file/Path;[Ljava/nio/file/attribute/FileAttribute;)V"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable) => void): void;
    public "createSymbolicLink(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/attribute/FileAttribute;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/createSymbolicLink(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/attribute/FileAttribute;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_attribute_FileAttribute>], cb?: (e?: java_lang_Throwable) => void): void;
    public "createLink(Ljava/nio/file/Path;Ljava/nio/file/Path;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/createLink(Ljava/nio/file/Path;Ljava/nio/file/Path;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path], cb?: (e?: java_lang_Throwable) => void): void;
    public "delete(Ljava/nio/file/Path;)V"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/delete(Ljava/nio/file/Path;)V"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable) => void): void;
    public "deleteIfExists(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/spi/FileSystemProvider/deleteIfExists(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readSymbolicLink(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "java/nio/file/spi/FileSystemProvider/readSymbolicLink(Ljava/nio/file/Path;)Ljava/nio/file/Path;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "copy(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/copy(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "move(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/move(Ljava/nio/file/Path;Ljava/nio/file/Path;[Ljava/nio/file/CopyOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path, JVMArray<java_nio_file_CopyOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSameFile(Ljava/nio/file/Path;Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/spi/FileSystemProvider/isSameFile(Ljava/nio/file/Path;Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path, java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isHidden(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/file/spi/FileSystemProvider/isHidden(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFileStore(Ljava/nio/file/Path;)Ljava/nio/file/FileStore;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileStore) => void): void;
    public "java/nio/file/spi/FileSystemProvider/getFileStore(Ljava/nio/file/Path;)Ljava/nio/file/FileStore;"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_FileStore) => void): void;
    public "checkAccess(Ljava/nio/file/Path;[Ljava/nio/file/AccessMode;)V"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_AccessMode>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/checkAccess(Ljava/nio/file/Path;[Ljava/nio/file/AccessMode;)V"(thread: JVMThread, args: [java_nio_file_Path, JVMArray<java_nio_file_AccessMode>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFileAttributeView(Ljava/nio/file/Path;Ljava/lang/Class;[Ljava/nio/file/LinkOption;)Ljava/nio/file/attribute/FileAttributeView;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_Class, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileAttributeView) => void): void;
    public "java/nio/file/spi/FileSystemProvider/getFileAttributeView(Ljava/nio/file/Path;Ljava/lang/Class;[Ljava/nio/file/LinkOption;)Ljava/nio/file/attribute/FileAttributeView;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_Class, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileAttributeView) => void): void;
    public "readAttributes(Ljava/nio/file/Path;Ljava/lang/Class;[Ljava/nio/file/LinkOption;)Ljava/nio/file/attribute/BasicFileAttributes;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_Class, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_BasicFileAttributes) => void): void;
    public "java/nio/file/spi/FileSystemProvider/readAttributes(Ljava/nio/file/Path;Ljava/lang/Class;[Ljava/nio/file/LinkOption;)Ljava/nio/file/attribute/BasicFileAttributes;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_Class, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_BasicFileAttributes) => void): void;
    public "readAttributes(Ljava/nio/file/Path;Ljava/lang/String;[Ljava/nio/file/LinkOption;)Ljava/util/Map;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/nio/file/spi/FileSystemProvider/readAttributes(Ljava/nio/file/Path;Ljava/lang/String;[Ljava/nio/file/LinkOption;)Ljava/util/Map;"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "setAttribute(Ljava/nio/file/Path;Ljava/lang/String;Ljava/lang/Object;[Ljava/nio/file/LinkOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, java_lang_Object, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/spi/FileSystemProvider/setAttribute(Ljava/nio/file/Path;Ljava/lang/String;Ljava/lang/Object;[Ljava/nio/file/LinkOption;)V"(thread: JVMThread, args: [java_nio_file_Path, java_lang_String, java_lang_Object, JVMArray<java_nio_file_LinkOption>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/nio/file/spi/FileSystemProvider/access$000()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/file/spi/FileSystemProvider/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_net_dns_ResolverConfigurationImpl extends sun_net_dns_ResolverConfiguration {
    public static "sun/net/dns/ResolverConfigurationImpl/lock": java_lang_Object;
    public static "sun/net/dns/ResolverConfigurationImpl/lastRefresh": Long;
    public static "sun/net/dns/ResolverConfigurationImpl/TIMEOUT": number;
    public "sun/net/dns/ResolverConfigurationImpl/opts": sun_net_dns_ResolverConfiguration$Options;
    public "sun/net/dns/ResolverConfigurationImpl/searchlist": java_util_LinkedList;
    public "sun/net/dns/ResolverConfigurationImpl/nameservers": java_util_LinkedList;
    public static "sun/net/dns/ResolverConfigurationImpl/$assertionsDisabled": number;
    public "resolvconf(Ljava/lang/String;II)Ljava/util/LinkedList;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList) => void): void;
    public "sun/net/dns/ResolverConfigurationImpl/resolvconf(Ljava/lang/String;II)Ljava/util/LinkedList;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList) => void): void;
    public "loadConfig()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/dns/ResolverConfigurationImpl/loadConfig()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSearchList()Ljava/util/LinkedList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList) => void): void;
    public "sun/net/dns/ResolverConfigurationImpl/getSearchList()Ljava/util/LinkedList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/dns/ResolverConfigurationImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "searchlist()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/net/dns/ResolverConfigurationImpl/searchlist()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "nameservers()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/net/dns/ResolverConfigurationImpl/nameservers()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "options()Lsun/net/dns/ResolverConfiguration$Options;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_net_dns_ResolverConfiguration$Options) => void): void;
    public "sun/net/dns/ResolverConfigurationImpl/options()Lsun/net/dns/ResolverConfiguration$Options;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_net_dns_ResolverConfiguration$Options) => void): void;
    public static "localDomain0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/net/dns/ResolverConfigurationImpl/localDomain0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "fallbackDomain0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/net/dns/ResolverConfigurationImpl/fallbackDomain0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000(Lsun/net/dns/ResolverConfigurationImpl;Ljava/lang/String;II)Ljava/util/LinkedList;"(thread: JVMThread, args: [sun_net_dns_ResolverConfigurationImpl, java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList) => void): void;
    public static "sun/net/dns/ResolverConfigurationImpl/access$000(Lsun/net/dns/ResolverConfigurationImpl;Ljava/lang/String;II)Ljava/util/LinkedList;"(thread: JVMThread, args: [sun_net_dns_ResolverConfigurationImpl, java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_LinkedList) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/dns/ResolverConfigurationImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_net_dns_ResolverConfiguration$Options extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/dns/ResolverConfiguration$Options/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "attempts()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/net/dns/ResolverConfiguration$Options/attempts()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "retrans()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/net/dns/ResolverConfiguration$Options/retrans()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_net_dns_ResolverConfiguration extends java_lang_Object {
    public static "sun/net/dns/ResolverConfiguration/lock": java_lang_Object;
    public static "sun/net/dns/ResolverConfiguration/provider": sun_net_dns_ResolverConfiguration;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/dns/ResolverConfiguration/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open()Lsun/net/dns/ResolverConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_net_dns_ResolverConfiguration) => void): void;
    public static "sun/net/dns/ResolverConfiguration/open()Lsun/net/dns/ResolverConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_net_dns_ResolverConfiguration) => void): void;
    public "searchlist()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/net/dns/ResolverConfiguration/searchlist()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "nameservers()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/net/dns/ResolverConfiguration/nameservers()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "options()Lsun/net/dns/ResolverConfiguration$Options;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_net_dns_ResolverConfiguration$Options) => void): void;
    public "sun/net/dns/ResolverConfiguration/options()Lsun/net/dns/ResolverConfiguration$Options;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_net_dns_ResolverConfiguration$Options) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/dns/ResolverConfiguration/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_net_sdp_SdpSupport extends java_lang_Object {
    public static "sun/net/sdp/SdpSupport/os": java_lang_String;
    public static "sun/net/sdp/SdpSupport/isSupported": number;
    public static "sun/net/sdp/SdpSupport/fdAccess": sun_misc_JavaIOFileDescriptorAccess;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/sdp/SdpSupport/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createSocket()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "sun/net/sdp/SdpSupport/createSocket()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "convertSocket(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/sdp/SdpSupport/convertSocket(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "create0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/sdp/SdpSupport/create0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "convert0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/sdp/SdpSupport/convert0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/sdp/SdpSupport/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_misc_JavaIOFileDescriptorAccess extends java_lang_Object {
    "set(Ljava/io/FileDescriptor;I)V"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable) => void): void;
    "get(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setHandle(Ljava/io/FileDescriptor;J)V"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    "getHandle(Ljava/io/FileDescriptor;)J"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class java_net_Proxy extends java_lang_Object {
    public "java/net/Proxy/type": java_net_Proxy$Type;
    public "java/net/Proxy/sa": java_net_SocketAddress;
    public static "java/net/Proxy/NO_PROXY": java_net_Proxy;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Proxy/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/Proxy$Type;Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_Proxy$Type, java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Proxy/<init>(Ljava/net/Proxy$Type;Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_Proxy$Type, java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "type()Ljava/net/Proxy$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Proxy$Type) => void): void;
    public "java/net/Proxy/type()Ljava/net/Proxy$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Proxy$Type) => void): void;
    public "address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/net/Proxy/address()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/Proxy/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Proxy/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Proxy/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/Proxy/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_Proxy$Type extends java_lang_Enum {
    public static "java/net/Proxy$Type/DIRECT": java_net_Proxy$Type;
    public static "java/net/Proxy$Type/HTTP": java_net_Proxy$Type;
    public static "java/net/Proxy$Type/SOCKS": java_net_Proxy$Type;
    public static "java/net/Proxy$Type/$VALUES": JVMArray<java_net_Proxy$Type>;
    public static "values()[Ljava/net/Proxy$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_Proxy$Type>) => void): void;
    public static "java/net/Proxy$Type/values()[Ljava/net/Proxy$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_Proxy$Type>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/net/Proxy$Type;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_Proxy$Type) => void): void;
    public static "java/net/Proxy$Type/valueOf(Ljava/lang/String;)Ljava/net/Proxy$Type;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_Proxy$Type) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Proxy$Type/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/Proxy$Type/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_net_spi_DefaultProxySelector extends java_net_ProxySelector {
    public static "sun/net/spi/DefaultProxySelector/props": JVMArray<JVMArray<java_lang_String>>;
    public static "sun/net/spi/DefaultProxySelector/SOCKS_PROXY_VERSION": java_lang_String;
    public static "sun/net/spi/DefaultProxySelector/hasSystemProxies": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/spi/DefaultProxySelector/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "select(Ljava/net/URI;)Ljava/util/List;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/net/spi/DefaultProxySelector/select(Ljava/net/URI;)Ljava/util/List;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "connectFailed(Ljava/net/URI;Ljava/net/SocketAddress;Ljava/io/IOException;)V"(thread: JVMThread, args: [java_net_URI, java_net_SocketAddress, java_io_IOException], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/spi/DefaultProxySelector/connectFailed(Ljava/net/URI;Ljava/net/SocketAddress;Ljava/io/IOException;)V"(thread: JVMThread, args: [java_net_URI, java_net_SocketAddress, java_io_IOException], cb?: (e?: java_lang_Throwable) => void): void;
    public "defaultPort(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/net/spi/DefaultProxySelector/defaultPort(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "init()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/spi/DefaultProxySelector/init()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSystemProxy(Ljava/lang/String;Ljava/lang/String;)Ljava/net/Proxy;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_Proxy) => void): void;
    public "sun/net/spi/DefaultProxySelector/getSystemProxy(Ljava/lang/String;Ljava/lang/String;)Ljava/net/Proxy;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_Proxy) => void): void;
    public static "access$000()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/spi/DefaultProxySelector/access$000()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Lsun/net/spi/DefaultProxySelector;Ljava/lang/String;Ljava/lang/String;)Ljava/net/Proxy;"(thread: JVMThread, args: [sun_net_spi_DefaultProxySelector, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_Proxy) => void): void;
    public static "sun/net/spi/DefaultProxySelector/access$100(Lsun/net/spi/DefaultProxySelector;Ljava/lang/String;Ljava/lang/String;)Ljava/net/Proxy;"(thread: JVMThread, args: [sun_net_spi_DefaultProxySelector, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_Proxy) => void): void;
    public static "access$200(Lsun/net/spi/DefaultProxySelector;Ljava/lang/String;)I"(thread: JVMThread, args: [sun_net_spi_DefaultProxySelector, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/spi/DefaultProxySelector/access$200(Lsun/net/spi/DefaultProxySelector;Ljava/lang/String;)I"(thread: JVMThread, args: [sun_net_spi_DefaultProxySelector, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/spi/DefaultProxySelector/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_IOException extends java_lang_Exception {
    public static "java/io/IOException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/IOException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/IOException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/IOException/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/IOException/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_ProxySelector extends java_lang_Object {
    public static "java/net/ProxySelector/theProxySelector": java_net_ProxySelector;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ProxySelector/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getDefault()Ljava/net/ProxySelector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ProxySelector) => void): void;
    public static "java/net/ProxySelector/getDefault()Ljava/net/ProxySelector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ProxySelector) => void): void;
    public static "setDefault(Ljava/net/ProxySelector;)V"(thread: JVMThread, args: [java_net_ProxySelector], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/ProxySelector/setDefault(Ljava/net/ProxySelector;)V"(thread: JVMThread, args: [java_net_ProxySelector], cb?: (e?: java_lang_Throwable) => void): void;
    public "select(Ljava/net/URI;)Ljava/util/List;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/net/ProxySelector/select(Ljava/net/URI;)Ljava/util/List;"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "connectFailed(Ljava/net/URI;Ljava/net/SocketAddress;Ljava/io/IOException;)V"(thread: JVMThread, args: [java_net_URI, java_net_SocketAddress, java_io_IOException], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ProxySelector/connectFailed(Ljava/net/URI;Ljava/net/SocketAddress;Ljava/io/IOException;)V"(thread: JVMThread, args: [java_net_URI, java_net_SocketAddress, java_io_IOException], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/ProxySelector/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_net_PortConfig extends java_lang_Object {
    public static "sun/net/PortConfig/defaultUpper": number;
    public static "sun/net/PortConfig/defaultLower": number;
    public static "sun/net/PortConfig/upper": number;
    public static "sun/net/PortConfig/lower": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/PortConfig/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getLower0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/PortConfig/getLower0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getUpper0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/PortConfig/getUpper0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getLower()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/PortConfig/getLower()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getUpper()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/PortConfig/getUpper()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$002(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/PortConfig/access$002(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$102(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/PortConfig/access$102(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/PortConfig/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class jdk_net_SocketFlow extends java_lang_Object {
    public static "jdk/net/SocketFlow/UNSET": number;
    public static "jdk/net/SocketFlow/NORMAL_PRIORITY": number;
    public static "jdk/net/SocketFlow/HIGH_PRIORITY": number;
    public "jdk/net/SocketFlow/priority": number;
    public "jdk/net/SocketFlow/bandwidth": Long;
    public "jdk/net/SocketFlow/status": jdk_net_SocketFlow$Status;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "jdk/net/SocketFlow/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "create()Ljdk/net/SocketFlow;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow) => void): void;
    public static "jdk/net/SocketFlow/create()Ljdk/net/SocketFlow;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow) => void): void;
    public "priority(I)Ljdk/net/SocketFlow;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow) => void): void;
    public "jdk/net/SocketFlow/priority(I)Ljdk/net/SocketFlow;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow) => void): void;
    public "bandwidth(J)Ljdk/net/SocketFlow;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow) => void): void;
    public "jdk/net/SocketFlow/bandwidth(J)Ljdk/net/SocketFlow;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow) => void): void;
    public "priority()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "jdk/net/SocketFlow/priority()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "bandwidth()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "jdk/net/SocketFlow/bandwidth()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "status()Ljdk/net/SocketFlow$Status;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow$Status) => void): void;
    public "jdk/net/SocketFlow/status()Ljdk/net/SocketFlow$Status;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow$Status) => void): void;
  }
  export class jdk_net_SocketFlow$Status extends java_lang_Enum {
    public static "jdk/net/SocketFlow$Status/NO_STATUS": jdk_net_SocketFlow$Status;
    public static "jdk/net/SocketFlow$Status/OK": jdk_net_SocketFlow$Status;
    public static "jdk/net/SocketFlow$Status/NO_PERMISSION": jdk_net_SocketFlow$Status;
    public static "jdk/net/SocketFlow$Status/NOT_CONNECTED": jdk_net_SocketFlow$Status;
    public static "jdk/net/SocketFlow$Status/NOT_SUPPORTED": jdk_net_SocketFlow$Status;
    public static "jdk/net/SocketFlow$Status/ALREADY_CREATED": jdk_net_SocketFlow$Status;
    public static "jdk/net/SocketFlow$Status/IN_PROGRESS": jdk_net_SocketFlow$Status;
    public static "jdk/net/SocketFlow$Status/OTHER": jdk_net_SocketFlow$Status;
    public static "jdk/net/SocketFlow$Status/$VALUES": JVMArray<jdk_net_SocketFlow$Status>;
    public static "values()[Ljdk/net/SocketFlow$Status;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<jdk_net_SocketFlow$Status>) => void): void;
    public static "jdk/net/SocketFlow$Status/values()[Ljdk/net/SocketFlow$Status;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<jdk_net_SocketFlow$Status>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljdk/net/SocketFlow$Status;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow$Status) => void): void;
    public static "jdk/net/SocketFlow$Status/valueOf(Ljava/lang/String;)Ljdk/net/SocketFlow$Status;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: jdk_net_SocketFlow$Status) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "jdk/net/SocketFlow$Status/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "jdk/net/SocketFlow$Status/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_net_ExtendedOptionsImpl extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/ExtendedOptionsImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkSetOptionPermission(Ljava/net/SocketOption;)V"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/ExtendedOptionsImpl/checkSetOptionPermission(Ljava/net/SocketOption;)V"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkGetOptionPermission(Ljava/net/SocketOption;)V"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/ExtendedOptionsImpl/checkGetOptionPermission(Ljava/net/SocketOption;)V"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkValueType(Ljava/lang/Object;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/ExtendedOptionsImpl/checkValueType(Ljava/lang/Object;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/ExtendedOptionsImpl/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setFlowOption(Ljava/io/FileDescriptor;Ljdk/net/SocketFlow;)V"(thread: JVMThread, args: [java_io_FileDescriptor, jdk_net_SocketFlow], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/ExtendedOptionsImpl/setFlowOption(Ljava/io/FileDescriptor;Ljdk/net/SocketFlow;)V"(thread: JVMThread, args: [java_io_FileDescriptor, jdk_net_SocketFlow], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getFlowOption(Ljava/io/FileDescriptor;Ljdk/net/SocketFlow;)V"(thread: JVMThread, args: [java_io_FileDescriptor, jdk_net_SocketFlow], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/ExtendedOptionsImpl/getFlowOption(Ljava/io/FileDescriptor;Ljdk/net/SocketFlow;)V"(thread: JVMThread, args: [java_io_FileDescriptor, jdk_net_SocketFlow], cb?: (e?: java_lang_Throwable) => void): void;
    public static "flowSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/ExtendedOptionsImpl/flowSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "lambda$static$0()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "sun/net/ExtendedOptionsImpl/lambda$static$0()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/net/ExtendedOptionsImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_Unsafe extends java_lang_Object {
    public static "sun/misc/Unsafe/theUnsafe": sun_misc_Unsafe;
    public static "sun/misc/Unsafe/INVALID_FIELD_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_BOOLEAN_BASE_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_BYTE_BASE_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_SHORT_BASE_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_CHAR_BASE_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_INT_BASE_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_LONG_BASE_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_FLOAT_BASE_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_DOUBLE_BASE_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_OBJECT_BASE_OFFSET": number;
    public static "sun/misc/Unsafe/ARRAY_BOOLEAN_INDEX_SCALE": number;
    public static "sun/misc/Unsafe/ARRAY_BYTE_INDEX_SCALE": number;
    public static "sun/misc/Unsafe/ARRAY_SHORT_INDEX_SCALE": number;
    public static "sun/misc/Unsafe/ARRAY_CHAR_INDEX_SCALE": number;
    public static "sun/misc/Unsafe/ARRAY_INT_INDEX_SCALE": number;
    public static "sun/misc/Unsafe/ARRAY_LONG_INDEX_SCALE": number;
    public static "sun/misc/Unsafe/ARRAY_FLOAT_INDEX_SCALE": number;
    public static "sun/misc/Unsafe/ARRAY_DOUBLE_INDEX_SCALE": number;
    public static "sun/misc/Unsafe/ARRAY_OBJECT_INDEX_SCALE": number;
    public static "sun/misc/Unsafe/ADDRESS_SIZE": number;
    public static "registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Unsafe/registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getUnsafe()Lsun/misc/Unsafe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Unsafe) => void): void;
    public static "sun/misc/Unsafe/getUnsafe()Lsun/misc/Unsafe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Unsafe) => void): void;
    public "getInt(Ljava/lang/Object;J)I"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getInt(Ljava/lang/Object;J)I"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putInt(Ljava/lang/Object;JI)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putInt(Ljava/lang/Object;JI)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObject(Ljava/lang/Object;J)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/misc/Unsafe/getObject(Ljava/lang/Object;J)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putObject(Ljava/lang/Object;JLjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putObject(Ljava/lang/Object;JLjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBoolean(Ljava/lang/Object;J)Z"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getBoolean(Ljava/lang/Object;J)Z"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putBoolean(Ljava/lang/Object;JZ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putBoolean(Ljava/lang/Object;JZ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getByte(Ljava/lang/Object;J)B"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getByte(Ljava/lang/Object;J)B"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putByte(Ljava/lang/Object;JB)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putByte(Ljava/lang/Object;JB)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getShort(Ljava/lang/Object;J)S"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getShort(Ljava/lang/Object;J)S"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putShort(Ljava/lang/Object;JS)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putShort(Ljava/lang/Object;JS)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChar(Ljava/lang/Object;J)C"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getChar(Ljava/lang/Object;J)C"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putChar(Ljava/lang/Object;JC)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putChar(Ljava/lang/Object;JC)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLong(Ljava/lang/Object;J)J"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/getLong(Ljava/lang/Object;J)J"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putLong(Ljava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putLong(Ljava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFloat(Ljava/lang/Object;J)F"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getFloat(Ljava/lang/Object;J)F"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloat(Ljava/lang/Object;JF)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putFloat(Ljava/lang/Object;JF)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDouble(Ljava/lang/Object;J)D"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getDouble(Ljava/lang/Object;J)D"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDouble(Ljava/lang/Object;JD)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putDouble(Ljava/lang/Object;JD)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInt(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getInt(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putInt(Ljava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putInt(Ljava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObject(Ljava/lang/Object;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/misc/Unsafe/getObject(Ljava/lang/Object;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putObject(Ljava/lang/Object;ILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putObject(Ljava/lang/Object;ILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBoolean(Ljava/lang/Object;I)Z"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getBoolean(Ljava/lang/Object;I)Z"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putBoolean(Ljava/lang/Object;IZ)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putBoolean(Ljava/lang/Object;IZ)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getByte(Ljava/lang/Object;I)B"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getByte(Ljava/lang/Object;I)B"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putByte(Ljava/lang/Object;IB)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putByte(Ljava/lang/Object;IB)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getShort(Ljava/lang/Object;I)S"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getShort(Ljava/lang/Object;I)S"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putShort(Ljava/lang/Object;IS)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putShort(Ljava/lang/Object;IS)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChar(Ljava/lang/Object;I)C"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getChar(Ljava/lang/Object;I)C"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putChar(Ljava/lang/Object;IC)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putChar(Ljava/lang/Object;IC)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLong(Ljava/lang/Object;I)J"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/getLong(Ljava/lang/Object;I)J"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putLong(Ljava/lang/Object;IJ)V"(thread: JVMThread, args: [java_lang_Object, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putLong(Ljava/lang/Object;IJ)V"(thread: JVMThread, args: [java_lang_Object, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFloat(Ljava/lang/Object;I)F"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getFloat(Ljava/lang/Object;I)F"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloat(Ljava/lang/Object;IF)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putFloat(Ljava/lang/Object;IF)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDouble(Ljava/lang/Object;I)D"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getDouble(Ljava/lang/Object;I)D"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDouble(Ljava/lang/Object;ID)V"(thread: JVMThread, args: [java_lang_Object, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putDouble(Ljava/lang/Object;ID)V"(thread: JVMThread, args: [java_lang_Object, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getByte(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getByte(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putByte(JB)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putByte(JB)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getShort(J)S"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getShort(J)S"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putShort(JS)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putShort(JS)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChar(J)C"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getChar(J)C"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putChar(JC)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putChar(JC)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInt(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getInt(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putInt(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putInt(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLong(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/getLong(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putLong(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putLong(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFloat(J)F"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getFloat(J)F"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloat(JF)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putFloat(JF)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDouble(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getDouble(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDouble(JD)V"(thread: JVMThread, args: [Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putDouble(JD)V"(thread: JVMThread, args: [Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAddress(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/getAddress(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putAddress(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putAddress(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "allocateMemory(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/allocateMemory(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "reallocateMemory(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/reallocateMemory(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setMemory(Ljava/lang/Object;JJB)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/setMemory(Ljava/lang/Object;JJB)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setMemory(JJB)V"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/setMemory(JJB)V"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyMemory(Ljava/lang/Object;JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/copyMemory(Ljava/lang/Object;JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyMemory(JJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/copyMemory(JJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "freeMemory(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/freeMemory(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "fieldOffset(Ljava/lang/reflect/Field;)I"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/fieldOffset(Ljava/lang/reflect/Field;)I"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "staticFieldBase(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/misc/Unsafe/staticFieldBase(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "staticFieldOffset(Ljava/lang/reflect/Field;)J"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/staticFieldOffset(Ljava/lang/reflect/Field;)J"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "objectFieldOffset(Ljava/lang/reflect/Field;)J"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/objectFieldOffset(Ljava/lang/reflect/Field;)J"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "staticFieldBase(Ljava/lang/reflect/Field;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/misc/Unsafe/staticFieldBase(Ljava/lang/reflect/Field;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "shouldBeInitialized(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/shouldBeInitialized(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureClassInitialized(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/ensureClassInitialized(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "arrayBaseOffset(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/arrayBaseOffset(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "arrayIndexScale(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/arrayIndexScale(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addressSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/addressSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "pageSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/pageSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "defineClass(Ljava/lang/String;[BIILjava/lang/ClassLoader;Ljava/security/ProtectionDomain;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number, java_lang_ClassLoader, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/misc/Unsafe/defineClass(Ljava/lang/String;[BIILjava/lang/ClassLoader;Ljava/security/ProtectionDomain;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number, java_lang_ClassLoader, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "defineAnonymousClass(Ljava/lang/Class;[B[Ljava/lang/Object;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class, JVMArray<number>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/misc/Unsafe/defineAnonymousClass(Ljava/lang/Class;[B[Ljava/lang/Object;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class, JVMArray<number>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "allocateInstance(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/misc/Unsafe/allocateInstance(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "monitorEnter(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/monitorEnter(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "monitorExit(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/monitorExit(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "tryMonitorEnter(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/tryMonitorEnter(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "throwException(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/throwException(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "compareAndSwapObject(Ljava/lang/Object;JLjava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/compareAndSwapObject(Ljava/lang/Object;JLjava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareAndSwapInt(Ljava/lang/Object;JII)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/compareAndSwapInt(Ljava/lang/Object;JII)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareAndSwapLong(Ljava/lang/Object;JJJ)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/compareAndSwapLong(Ljava/lang/Object;JJJ)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getObjectVolatile(Ljava/lang/Object;J)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/misc/Unsafe/getObjectVolatile(Ljava/lang/Object;J)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putObjectVolatile(Ljava/lang/Object;JLjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putObjectVolatile(Ljava/lang/Object;JLjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getIntVolatile(Ljava/lang/Object;J)I"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getIntVolatile(Ljava/lang/Object;J)I"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putIntVolatile(Ljava/lang/Object;JI)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putIntVolatile(Ljava/lang/Object;JI)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBooleanVolatile(Ljava/lang/Object;J)Z"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getBooleanVolatile(Ljava/lang/Object;J)Z"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putBooleanVolatile(Ljava/lang/Object;JZ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putBooleanVolatile(Ljava/lang/Object;JZ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getByteVolatile(Ljava/lang/Object;J)B"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getByteVolatile(Ljava/lang/Object;J)B"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putByteVolatile(Ljava/lang/Object;JB)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putByteVolatile(Ljava/lang/Object;JB)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getShortVolatile(Ljava/lang/Object;J)S"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getShortVolatile(Ljava/lang/Object;J)S"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putShortVolatile(Ljava/lang/Object;JS)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putShortVolatile(Ljava/lang/Object;JS)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCharVolatile(Ljava/lang/Object;J)C"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getCharVolatile(Ljava/lang/Object;J)C"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putCharVolatile(Ljava/lang/Object;JC)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putCharVolatile(Ljava/lang/Object;JC)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLongVolatile(Ljava/lang/Object;J)J"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/getLongVolatile(Ljava/lang/Object;J)J"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putLongVolatile(Ljava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putLongVolatile(Ljava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFloatVolatile(Ljava/lang/Object;J)F"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getFloatVolatile(Ljava/lang/Object;J)F"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloatVolatile(Ljava/lang/Object;JF)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putFloatVolatile(Ljava/lang/Object;JF)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDoubleVolatile(Ljava/lang/Object;J)D"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getDoubleVolatile(Ljava/lang/Object;J)D"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDoubleVolatile(Ljava/lang/Object;JD)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putDoubleVolatile(Ljava/lang/Object;JD)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "putOrderedObject(Ljava/lang/Object;JLjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putOrderedObject(Ljava/lang/Object;JLjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "putOrderedInt(Ljava/lang/Object;JI)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putOrderedInt(Ljava/lang/Object;JI)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "putOrderedLong(Ljava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/putOrderedLong(Ljava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "unpark(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/unpark(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "park(ZJ)V"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/park(ZJ)V"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLoadAverage([DI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getLoadAverage([DI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAndAddInt(Ljava/lang/Object;JI)I"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getAndAddInt(Ljava/lang/Object;JI)I"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAndAddLong(Ljava/lang/Object;JJ)J"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/getAndAddLong(Ljava/lang/Object;JJ)J"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getAndSetInt(Ljava/lang/Object;JI)I"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Unsafe/getAndSetInt(Ljava/lang/Object;JI)I"(thread: JVMThread, args: [java_lang_Object, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAndSetLong(Ljava/lang/Object;JJ)J"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Unsafe/getAndSetLong(Ljava/lang/Object;JJ)J"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getAndSetObject(Ljava/lang/Object;JLjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/misc/Unsafe/getAndSetObject(Ljava/lang/Object;JLjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, Long, any, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "loadFence()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/loadFence()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "storeFence()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/storeFence()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "fullFence()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Unsafe/fullFence()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "throwIllegalAccessError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Unsafe/throwIllegalAccessError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Unsafe/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_VM extends java_lang_Object {
    public static "sun/misc/VM/suspended": number;
    public static "sun/misc/VM/STATE_GREEN": number;
    public static "sun/misc/VM/STATE_YELLOW": number;
    public static "sun/misc/VM/STATE_RED": number;
    public static "sun/misc/VM/booted": number;
    public static "sun/misc/VM/lock": java_lang_Object;
    public static "sun/misc/VM/directMemory": Long;
    public static "sun/misc/VM/pageAlignDirectMemory": number;
    public static "sun/misc/VM/defaultAllowArraySyntax": number;
    public static "sun/misc/VM/allowArraySyntax": number;
    public static "sun/misc/VM/savedProps": java_util_Properties;
    public static "sun/misc/VM/finalRefCount": number;
    public static "sun/misc/VM/peakFinalRefCount": number;
    public static "sun/misc/VM/JVMTI_THREAD_STATE_ALIVE": number;
    public static "sun/misc/VM/JVMTI_THREAD_STATE_TERMINATED": number;
    public static "sun/misc/VM/JVMTI_THREAD_STATE_RUNNABLE": number;
    public static "sun/misc/VM/JVMTI_THREAD_STATE_BLOCKED_ON_MONITOR_ENTER": number;
    public static "sun/misc/VM/JVMTI_THREAD_STATE_WAITING_INDEFINITELY": number;
    public static "sun/misc/VM/JVMTI_THREAD_STATE_WAITING_WITH_TIMEOUT": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/VM/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "threadsSuspended()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/threadsSuspended()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "allowThreadSuspension(Ljava/lang/ThreadGroup;Z)Z"(thread: JVMThread, args: [java_lang_ThreadGroup, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/allowThreadSuspension(Ljava/lang/ThreadGroup;Z)Z"(thread: JVMThread, args: [java_lang_ThreadGroup, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "suspendThreads()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/suspendThreads()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "unsuspendThreads()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/unsuspendThreads()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "unsuspendSomeThreads()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/unsuspendSomeThreads()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/getState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "registerVMNotification(Lsun/misc/VMNotification;)V"(thread: JVMThread, args: [sun_misc_VMNotification], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/registerVMNotification(Lsun/misc/VMNotification;)V"(thread: JVMThread, args: [sun_misc_VMNotification], cb?: (e?: java_lang_Throwable) => void): void;
    public static "asChange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/asChange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "asChange_otherthread(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/asChange_otherthread(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "booted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/booted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isBooted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/isBooted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "awaitBooted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/awaitBooted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "maxDirectMemory()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/misc/VM/maxDirectMemory()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "isDirectMemoryPageAligned()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/isDirectMemoryPageAligned()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "allowArraySyntax()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/allowArraySyntax()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSystemDomainLoader(Ljava/lang/ClassLoader;)Z"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/isSystemDomainLoader(Ljava/lang/ClassLoader;)Z"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getSavedProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/VM/getSavedProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "saveAndRemoveProperties(Ljava/util/Properties;)V"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/saveAndRemoveProperties(Ljava/util/Properties;)V"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initializeOSEnvironment()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/initializeOSEnvironment()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getFinalRefCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/getFinalRefCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getPeakFinalRefCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VM/getPeakFinalRefCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "addFinalRefCount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/addFinalRefCount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "toThreadState(I)Ljava/lang/Thread$State;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$State) => void): void;
    public static "sun/misc/VM/toThreadState(I)Ljava/lang/Thread$State;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$State) => void): void;
    public static "latestUserDefinedLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "sun/misc/VM/latestUserDefinedLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "initialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/initialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VM/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Thread$State extends java_lang_Enum {
    public static "java/lang/Thread$State/NEW": java_lang_Thread$State;
    public static "java/lang/Thread$State/RUNNABLE": java_lang_Thread$State;
    public static "java/lang/Thread$State/BLOCKED": java_lang_Thread$State;
    public static "java/lang/Thread$State/WAITING": java_lang_Thread$State;
    public static "java/lang/Thread$State/TIMED_WAITING": java_lang_Thread$State;
    public static "java/lang/Thread$State/TERMINATED": java_lang_Thread$State;
    public static "java/lang/Thread$State/$VALUES": JVMArray<java_lang_Thread$State>;
    public static "values()[Ljava/lang/Thread$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread$State>) => void): void;
    public static "java/lang/Thread$State/values()[Ljava/lang/Thread$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread$State>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/Thread$State;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$State) => void): void;
    public static "java/lang/Thread$State/valueOf(Ljava/lang/String;)Ljava/lang/Thread$State;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$State) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread$State/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Thread$State/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_misc_VMNotification extends java_lang_Object {
    "newAllocState(IIZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ThreadGroup extends java_lang_Object implements java_lang_Thread$UncaughtExceptionHandler {
    public "java/lang/ThreadGroup/parent": java_lang_ThreadGroup;
    public "java/lang/ThreadGroup/name": java_lang_String;
    public "java/lang/ThreadGroup/maxPriority": number;
    public "java/lang/ThreadGroup/destroyed": number;
    public "java/lang/ThreadGroup/daemon": number;
    public "java/lang/ThreadGroup/vmAllowSuspension": number;
    public "java/lang/ThreadGroup/nUnstartedThreads": number;
    public "java/lang/ThreadGroup/nthreads": number;
    public "java/lang/ThreadGroup/threads": JVMArray<java_lang_Thread>;
    public "java/lang/ThreadGroup/ngroups": number;
    public "java/lang/ThreadGroup/groups": JVMArray<java_lang_ThreadGroup>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ThreadGroup;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/<init>(Ljava/lang/ThreadGroup;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Void;Ljava/lang/ThreadGroup;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Void, java_lang_ThreadGroup, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/<init>(Ljava/lang/Void;Ljava/lang/ThreadGroup;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Void, java_lang_ThreadGroup, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkParentAccess(Ljava/lang/ThreadGroup;)Ljava/lang/Void;"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "java/lang/ThreadGroup/checkParentAccess(Ljava/lang/ThreadGroup;)Ljava/lang/Void;"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/ThreadGroup/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getParent()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public "java/lang/ThreadGroup/getParent()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public "getMaxPriority()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/getMaxPriority()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDaemon()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/isDaemon()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDestroyed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/isDestroyed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDaemon(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/setDaemon(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setMaxPriority(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/setMaxPriority(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "parentOf(Ljava/lang/ThreadGroup;)Z"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/parentOf(Ljava/lang/ThreadGroup;)Z"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/checkAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "activeCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/activeCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "enumerate([Ljava/lang/Thread;)I"(thread: JVMThread, args: [JVMArray<java_lang_Thread>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/enumerate([Ljava/lang/Thread;)I"(thread: JVMThread, args: [JVMArray<java_lang_Thread>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "enumerate([Ljava/lang/Thread;Z)I"(thread: JVMThread, args: [JVMArray<java_lang_Thread>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/enumerate([Ljava/lang/Thread;Z)I"(thread: JVMThread, args: [JVMArray<java_lang_Thread>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "enumerate([Ljava/lang/Thread;IZ)I"(thread: JVMThread, args: [JVMArray<java_lang_Thread>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/enumerate([Ljava/lang/Thread;IZ)I"(thread: JVMThread, args: [JVMArray<java_lang_Thread>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "activeGroupCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/activeGroupCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "enumerate([Ljava/lang/ThreadGroup;)I"(thread: JVMThread, args: [JVMArray<java_lang_ThreadGroup>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/enumerate([Ljava/lang/ThreadGroup;)I"(thread: JVMThread, args: [JVMArray<java_lang_ThreadGroup>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "enumerate([Ljava/lang/ThreadGroup;Z)I"(thread: JVMThread, args: [JVMArray<java_lang_ThreadGroup>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/enumerate([Ljava/lang/ThreadGroup;Z)I"(thread: JVMThread, args: [JVMArray<java_lang_ThreadGroup>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "enumerate([Ljava/lang/ThreadGroup;IZ)I"(thread: JVMThread, args: [JVMArray<java_lang_ThreadGroup>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/enumerate([Ljava/lang/ThreadGroup;IZ)I"(thread: JVMThread, args: [JVMArray<java_lang_ThreadGroup>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "stop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/stop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/interrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "suspend()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/suspend()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "stopOrSuspend(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/stopOrSuspend(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "resume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/resume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "destroy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/destroy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/lang/ThreadGroup;)V"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/add(Ljava/lang/ThreadGroup;)V"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/ThreadGroup;)V"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/remove(Ljava/lang/ThreadGroup;)V"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "addUnstarted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/addUnstarted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/add(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "threadStartFailed(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/threadStartFailed(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "threadTerminated(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/threadTerminated(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/remove(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "list()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/list()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "list(Ljava/io/PrintStream;I)V"(thread: JVMThread, args: [java_io_PrintStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/list(Ljava/io/PrintStream;I)V"(thread: JVMThread, args: [java_io_PrintStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "uncaughtException(Ljava/lang/Thread;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Thread, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ThreadGroup/uncaughtException(Ljava/lang/Thread;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Thread, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "allowThreadSuspension(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ThreadGroup/allowThreadSuspension(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/ThreadGroup/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_lang_Thread$UncaughtExceptionHandler extends java_lang_Object {
    "uncaughtException(Ljava/lang/Thread;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Thread, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_Version extends java_lang_Object {
    public static "sun/misc/Version/launcher_name": java_lang_String;
    public static "sun/misc/Version/java_version": java_lang_String;
    public static "sun/misc/Version/java_runtime_name": java_lang_String;
    public static "sun/misc/Version/java_profile_name": java_lang_String;
    public static "sun/misc/Version/java_runtime_version": java_lang_String;
    public static "sun/misc/Version/versionsInitialized": number;
    public static "sun/misc/Version/jvm_major_version": number;
    public static "sun/misc/Version/jvm_minor_version": number;
    public static "sun/misc/Version/jvm_micro_version": number;
    public static "sun/misc/Version/jvm_update_version": number;
    public static "sun/misc/Version/jvm_build_number": number;
    public static "sun/misc/Version/jvm_special_version": java_lang_String;
    public static "sun/misc/Version/jdk_major_version": number;
    public static "sun/misc/Version/jdk_minor_version": number;
    public static "sun/misc/Version/jdk_micro_version": number;
    public static "sun/misc/Version/jdk_update_version": number;
    public static "sun/misc/Version/jdk_build_number": number;
    public static "sun/misc/Version/jdk_special_version": java_lang_String;
    public static "sun/misc/Version/jvmVersionInfoAvailable": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Version/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Version/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "print()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Version/print()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "println()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Version/println()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "print(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Version/print(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "jvmMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jvmMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jvmMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jvmMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jvmMicroVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jvmMicroVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jvmUpdateVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jvmUpdateVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jvmSpecialVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/Version/jvmSpecialVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getJvmSpecialVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/Version/getJvmSpecialVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "jvmBuildNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jvmBuildNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jdkMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jdkMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jdkMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jdkMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jdkMicroVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jdkMicroVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jdkUpdateVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jdkUpdateVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "jdkSpecialVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/Version/jdkSpecialVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getJdkSpecialVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/Version/getJdkSpecialVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "jdkBuildNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/jdkBuildNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initVersions()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Version/initVersions()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getJvmVersionInfo()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Version/getJvmVersionInfo()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getJdkVersionInfo()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Version/getJdkVersionInfo()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Version/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_Signal extends java_lang_Object {
    public static "sun/misc/Signal/handlers": java_util_Hashtable;
    public static "sun/misc/Signal/signals": java_util_Hashtable;
    public "sun/misc/Signal/number": number;
    public "sun/misc/Signal/name": java_lang_String;
    public "getNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Signal/getNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/misc/Signal/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Signal/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Signal/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/misc/Signal/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Signal/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "handle(Lsun/misc/Signal;Lsun/misc/SignalHandler;)Lsun/misc/SignalHandler;"(thread: JVMThread, args: [sun_misc_Signal, sun_misc_SignalHandler], cb?: (e?: java_lang_Throwable, rv?: sun_misc_SignalHandler) => void): void;
    public static "sun/misc/Signal/handle(Lsun/misc/Signal;Lsun/misc/SignalHandler;)Lsun/misc/SignalHandler;"(thread: JVMThread, args: [sun_misc_Signal, sun_misc_SignalHandler], cb?: (e?: java_lang_Throwable, rv?: sun_misc_SignalHandler) => void): void;
    public static "raise(Lsun/misc/Signal;)V"(thread: JVMThread, args: [sun_misc_Signal], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Signal/raise(Lsun/misc/Signal;)V"(thread: JVMThread, args: [sun_misc_Signal], cb?: (e?: java_lang_Throwable) => void): void;
    public static "dispatch(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Signal/dispatch(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "findSignal(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/Signal/findSignal(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "handle0(IJ)J"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/misc/Signal/handle0(IJ)J"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "raise0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Signal/raise0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Signal/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_misc_SignalHandler extends java_lang_Object {
    "handle(Lsun/misc/Signal;)V"(thread: JVMThread, args: [sun_misc_Signal], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_Hashtable extends java_util_Dictionary implements java_util_Map, java_lang_Cloneable, java_io_Serializable {
    public "java/util/Hashtable/table": JVMArray<java_util_Hashtable$Entry>;
    public "java/util/Hashtable/count": number;
    public "java/util/Hashtable/threshold": number;
    public "java/util/Hashtable/loadFactor": number;
    public "java/util/Hashtable/modCount": number;
    public static "java/util/Hashtable/serialVersionUID": Long;
    public static "java/util/Hashtable/MAX_ARRAY_SIZE": number;
    public "java/util/Hashtable/keySet": java_util_Set;
    public "java/util/Hashtable/entrySet": java_util_Set;
    public "java/util/Hashtable/values": java_util_Collection;
    public static "java/util/Hashtable/KEYS": number;
    public static "java/util/Hashtable/VALUES": number;
    public static "java/util/Hashtable/ENTRIES": number;
    public "<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/Hashtable/keys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/Hashtable/elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable/containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "rehash()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/rehash()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addEntry(ILjava/lang/Object;Ljava/lang/Object;I)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/addEntry(ILjava/lang/Object;Ljava/lang/Object;I)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Hashtable/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getEnumeration(I)Ljava/util/Enumeration;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/Hashtable/getEnumeration(I)Ljava/util/Enumeration;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getIterator(I)Ljava/util/Iterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/Hashtable/getIterator(I)Ljava/util/Iterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/Hashtable/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/Hashtable/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/Hashtable/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable/remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable/replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable/merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "reconstitutionPut([Ljava/util/Hashtable$Entry;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_util_Hashtable$Entry>, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable/reconstitutionPut([Ljava/util/Hashtable$Entry;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_util_Hashtable$Entry>, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$100(Ljava/util/Hashtable;I)Ljava/util/Iterator;"(thread: JVMThread, args: [java_util_Hashtable, number], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public static "java/util/Hashtable/access$100(Ljava/util/Hashtable;I)Ljava/util/Iterator;"(thread: JVMThread, args: [java_util_Hashtable, number], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public static "access$200(Ljava/util/Hashtable;)I"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Hashtable/access$200(Ljava/util/Hashtable;)I"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$400(Ljava/util/Hashtable;)[Ljava/util/Hashtable$Entry;"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Hashtable$Entry>) => void): void;
    public static "java/util/Hashtable/access$400(Ljava/util/Hashtable;)[Ljava/util/Hashtable$Entry;"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Hashtable$Entry>) => void): void;
    public static "access$508(Ljava/util/Hashtable;)I"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Hashtable/access$508(Ljava/util/Hashtable;)I"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$210(Ljava/util/Hashtable;)I"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Hashtable/access$210(Ljava/util/Hashtable;)I"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500(Ljava/util/Hashtable;)I"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Hashtable/access$500(Ljava/util/Hashtable;)I"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_Hashtable$Entry extends java_lang_Object implements java_util_Map$Entry {
    public "java/util/Hashtable$Entry/hash": number;
    public "java/util/Hashtable$Entry/key": java_lang_Object;
    public "java/util/Hashtable$Entry/value": java_lang_Object;
    public "java/util/Hashtable$Entry/next": java_util_Hashtable$Entry;
    public "<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/Hashtable$Entry;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_Hashtable$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Hashtable$Entry/<init>(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/Hashtable$Entry;)V"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_Hashtable$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable$Entry/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable$Entry/getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable$Entry/getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Hashtable$Entry/setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable$Entry/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Hashtable$Entry/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Hashtable$Entry/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_util_Dictionary extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Dictionary/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Dictionary/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Dictionary/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/Dictionary/keys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/Dictionary/elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Dictionary/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Dictionary/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Dictionary/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_misc_NativeSignalHandler extends java_lang_Object implements sun_misc_SignalHandler {
    public "sun/misc/NativeSignalHandler/handler": Long;
    public "getHandler()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/NativeSignalHandler/getHandler()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/NativeSignalHandler/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "handle(Lsun/misc/Signal;)V"(thread: JVMThread, args: [sun_misc_Signal], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/NativeSignalHandler/handle(Lsun/misc/Signal;)V"(thread: JVMThread, args: [sun_misc_Signal], cb?: (e?: java_lang_Throwable) => void): void;
    public static "handle0(IJ)V"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/NativeSignalHandler/handle0(IJ)V"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_URL extends java_lang_Object implements java_io_Serializable {
    public static "java/net/URL/serialVersionUID": Long;
    public static "java/net/URL/protocolPathProp": java_lang_String;
    public "java/net/URL/protocol": java_lang_String;
    public "java/net/URL/host": java_lang_String;
    public "java/net/URL/port": number;
    public "java/net/URL/file": java_lang_String;
    public "java/net/URL/query": java_lang_String;
    public "java/net/URL/authority": java_lang_String;
    public "java/net/URL/path": java_lang_String;
    public "java/net/URL/userInfo": java_lang_String;
    public "java/net/URL/ref": java_lang_String;
    public "java/net/URL/hostAddress": java_net_InetAddress;
    public "java/net/URL/handler": java_net_URLStreamHandler;
    public "java/net/URL/hashCode": number;
    public static "java/net/URL/factory": java_net_URLStreamHandlerFactory;
    public static "java/net/URL/handlers": java_util_Hashtable;
    public static "java/net/URL/streamHandlerLock": java_lang_Object;
    public "<init>(Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/<init>(Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/net/URLStreamHandler;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_lang_String, java_net_URLStreamHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/<init>(Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/net/URLStreamHandler;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_lang_String, java_net_URLStreamHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/URL;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_URL, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/<init>(Ljava/net/URL;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_URL, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/URL;Ljava/lang/String;Ljava/net/URLStreamHandler;)V"(thread: JVMThread, args: [java_net_URL, java_lang_String, java_net_URLStreamHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/<init>(Ljava/net/URL;Ljava/lang/String;Ljava/net/URLStreamHandler;)V"(thread: JVMThread, args: [java_net_URL, java_lang_String, java_net_URLStreamHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "isValidProtocol(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URL/isValidProtocol(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkSpecifyHandler(Ljava/lang/SecurityManager;)V"(thread: JVMThread, args: [java_lang_SecurityManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/checkSpecifyHandler(Ljava/lang/SecurityManager;)V"(thread: JVMThread, args: [java_lang_SecurityManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/set(Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/set(Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getQuery()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/getQuery()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/getPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getUserInfo()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/getUserInfo()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAuthority()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/getAuthority()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URL/getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDefaultPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URL/getDefaultPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getProtocol()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/getProtocol()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getHost()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/getHost()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/getFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRef()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/getRef()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URL/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URL/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sameFile(Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URL/sameFile(Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toExternalForm()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URL/toExternalForm()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toURI()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "java/net/URL/toURI()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "openConnection()Ljava/net/URLConnection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URLConnection) => void): void;
    public "java/net/URL/openConnection()Ljava/net/URLConnection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URLConnection) => void): void;
    public "openConnection(Ljava/net/Proxy;)Ljava/net/URLConnection;"(thread: JVMThread, args: [java_net_Proxy], cb?: (e?: java_lang_Throwable, rv?: java_net_URLConnection) => void): void;
    public "java/net/URL/openConnection(Ljava/net/Proxy;)Ljava/net/URLConnection;"(thread: JVMThread, args: [java_net_Proxy], cb?: (e?: java_lang_Throwable, rv?: java_net_URLConnection) => void): void;
    public "openStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/net/URL/openStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getContent()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/URL/getContent()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getContent([Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/URL/getContent([Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "setURLStreamHandlerFactory(Ljava/net/URLStreamHandlerFactory;)V"(thread: JVMThread, args: [java_net_URLStreamHandlerFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URL/setURLStreamHandlerFactory(Ljava/net/URLStreamHandlerFactory;)V"(thread: JVMThread, args: [java_net_URLStreamHandlerFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getURLStreamHandler(Ljava/lang/String;)Ljava/net/URLStreamHandler;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URLStreamHandler) => void): void;
    public static "java/net/URL/getURLStreamHandler(Ljava/lang/String;)Ljava/net/URLStreamHandler;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URLStreamHandler) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URL/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URL/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_URLConnection extends java_lang_Object {
    public "java/net/URLConnection/url": java_net_URL;
    public "java/net/URLConnection/doInput": number;
    public "java/net/URLConnection/doOutput": number;
    public static "java/net/URLConnection/defaultAllowUserInteraction": number;
    public "java/net/URLConnection/allowUserInteraction": number;
    public static "java/net/URLConnection/defaultUseCaches": number;
    public "java/net/URLConnection/useCaches": number;
    public "java/net/URLConnection/ifModifiedSince": Long;
    public "java/net/URLConnection/connected": number;
    public "java/net/URLConnection/connectTimeout": number;
    public "java/net/URLConnection/readTimeout": number;
    public "java/net/URLConnection/requests": sun_net_www_MessageHeader;
    public static "java/net/URLConnection/fileNameMap": java_net_FileNameMap;
    public static "java/net/URLConnection/fileNameMapLoaded": number;
    public static "java/net/URLConnection/factory": java_net_ContentHandlerFactory;
    public static "java/net/URLConnection/handlers": java_util_Hashtable;
    public static "java/net/URLConnection/contentClassPrefix": java_lang_String;
    public static "java/net/URLConnection/contentPathProp": java_lang_String;
    public static "getFileNameMap()Ljava/net/FileNameMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_FileNameMap) => void): void;
    public static "java/net/URLConnection/getFileNameMap()Ljava/net/FileNameMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_FileNameMap) => void): void;
    public static "setFileNameMap(Ljava/net/FileNameMap;)V"(thread: JVMThread, args: [java_net_FileNameMap], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URLConnection/setFileNameMap(Ljava/net/FileNameMap;)V"(thread: JVMThread, args: [java_net_FileNameMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/connect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setConnectTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/setConnectTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getConnectTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLConnection/getConnectTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReadTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/setReadTimeout(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getReadTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLConnection/getReadTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/net/URL;)V"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/<init>(Ljava/net/URL;)V"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable) => void): void;
    public "getURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "java/net/URLConnection/getURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "getContentLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLConnection/getContentLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getContentLengthLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/net/URLConnection/getContentLengthLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getContentType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/getContentType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getContentEncoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/getContentEncoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getExpiration()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/net/URLConnection/getExpiration()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getDate()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/net/URLConnection/getDate()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLastModified()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/net/URLConnection/getLastModified()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getHeaderField(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/getHeaderField(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getHeaderFields()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/net/URLConnection/getHeaderFields()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getHeaderFieldInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLConnection/getHeaderFieldInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeaderFieldLong(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/net/URLConnection/getHeaderFieldLong(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getHeaderFieldDate(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/net/URLConnection/getHeaderFieldDate(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getHeaderFieldKey(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/getHeaderFieldKey(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getHeaderField(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/getHeaderField(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getContent()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/URLConnection/getContent()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getContent([Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/URLConnection/getContent([Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getPermission()Ljava/security/Permission;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Permission) => void): void;
    public "java/net/URLConnection/getPermission()Ljava/security/Permission;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Permission) => void): void;
    public "getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/net/URLConnection/getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "java/net/URLConnection/getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setDoInput(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/setDoInput(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDoInput()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLConnection/getDoInput()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDoOutput(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/setDoOutput(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDoOutput()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLConnection/getDoOutput()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setAllowUserInteraction(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/setAllowUserInteraction(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAllowUserInteraction()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLConnection/getAllowUserInteraction()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setDefaultAllowUserInteraction(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URLConnection/setDefaultAllowUserInteraction(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getDefaultAllowUserInteraction()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URLConnection/getDefaultAllowUserInteraction()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setUseCaches(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/setUseCaches(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUseCaches()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLConnection/getUseCaches()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setIfModifiedSince(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/setIfModifiedSince(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getIfModifiedSince()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/net/URLConnection/getIfModifiedSince()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getDefaultUseCaches()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLConnection/getDefaultUseCaches()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDefaultUseCaches(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/setDefaultUseCaches(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRequestProperty(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/setRequestProperty(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "addRequestProperty(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLConnection/addRequestProperty(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRequestProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/getRequestProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRequestProperties()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/net/URLConnection/getRequestProperties()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "setDefaultRequestProperty(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URLConnection/setDefaultRequestProperty(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getDefaultRequestProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URLConnection/getDefaultRequestProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "setContentHandlerFactory(Ljava/net/ContentHandlerFactory;)V"(thread: JVMThread, args: [java_net_ContentHandlerFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URLConnection/setContentHandlerFactory(Ljava/net/ContentHandlerFactory;)V"(thread: JVMThread, args: [java_net_ContentHandlerFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContentHandler()Ljava/net/ContentHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ContentHandler) => void): void;
    public "java/net/URLConnection/getContentHandler()Ljava/net/ContentHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ContentHandler) => void): void;
    public "stripOffParameters(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/stripOffParameters(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "lookupContentHandlerClassFor(Ljava/lang/String;)Ljava/net/ContentHandler;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_ContentHandler) => void): void;
    public "java/net/URLConnection/lookupContentHandlerClassFor(Ljava/lang/String;)Ljava/net/ContentHandler;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_ContentHandler) => void): void;
    public "typeToPackageName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/typeToPackageName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getContentHandlerPkgPrefixes()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLConnection/getContentHandlerPkgPrefixes()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "guessContentTypeFromName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URLConnection/guessContentTypeFromName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "guessContentTypeFromStream(Ljava/io/InputStream;)Ljava/lang/String;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/URLConnection/guessContentTypeFromStream(Ljava/io/InputStream;)Ljava/lang/String;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "checkfpx(Ljava/io/InputStream;)Z"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URLConnection/checkfpx(Ljava/io/InputStream;)Z"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "readBytes([IILjava/io/InputStream;)I"(thread: JVMThread, args: [JVMArray<number>, number, java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/URLConnection/readBytes([IILjava/io/InputStream;)I"(thread: JVMThread, args: [JVMArray<number>, number, java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "skipForward(Ljava/io/InputStream;J)J"(thread: JVMThread, args: [java_io_InputStream, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/net/URLConnection/skipForward(Ljava/io/InputStream;J)J"(thread: JVMThread, args: [java_io_InputStream, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$000()Ljava/net/FileNameMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_FileNameMap) => void): void;
    public static "java/net/URLConnection/access$000()Ljava/net/FileNameMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_FileNameMap) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/URLConnection/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_ContentHandler extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/ContentHandler/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContent(Ljava/net/URLConnection;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_URLConnection], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/ContentHandler/getContent(Ljava/net/URLConnection;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_URLConnection], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getContent(Ljava/net/URLConnection;[Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_URLConnection, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/ContentHandler/getContent(Ljava/net/URLConnection;[Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_URLConnection, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_net_ContentHandlerFactory extends java_lang_Object {
    "createContentHandler(Ljava/lang/String;)Ljava/net/ContentHandler;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_ContentHandler) => void): void;
  }
  export interface java_net_FileNameMap extends java_lang_Object {
    "getContentTypeFor(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_net_www_MessageHeader extends java_lang_Object {
    public "sun/net/www/MessageHeader/keys": JVMArray<java_lang_String>;
    public "sun/net/www/MessageHeader/values": JVMArray<java_lang_String>;
    public "sun/net/www/MessageHeader/nkeys": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getHeaderNamesInList()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/net/www/MessageHeader/getHeaderNamesInList()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "findValue(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/net/www/MessageHeader/findValue(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getKey(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/net/www/MessageHeader/getKey(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getKey(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/net/www/MessageHeader/getKey(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getValue(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/net/www/MessageHeader/getValue(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "findNextValue(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/net/www/MessageHeader/findNextValue(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "filterNTLMResponses(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/net/www/MessageHeader/filterNTLMResponses(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "multiValueIterator(Ljava/lang/String;)Ljava/util/Iterator;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "sun/net/www/MessageHeader/multiValueIterator(Ljava/lang/String;)Ljava/util/Iterator;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "getHeaders()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "sun/net/www/MessageHeader/getHeaders()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getHeaders([Ljava/lang/String;)Ljava/util/Map;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "sun/net/www/MessageHeader/getHeaders([Ljava/lang/String;)Ljava/util/Map;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "filterAndAddHeaders([Ljava/lang/String;Ljava/util/Map;)Ljava/util/Map;"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "sun/net/www/MessageHeader/filterAndAddHeaders([Ljava/lang/String;Ljava/util/Map;)Ljava/util/Map;"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "print(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/print(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/add(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "prepend(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/prepend(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(ILjava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/set(ILjava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "grow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/grow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/set(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setIfNotSet(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/setIfNotSet(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "canonicalID(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/net/www/MessageHeader/canonicalID(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "parseHeader(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/parseHeader(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "mergeHeader(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/net/www/MessageHeader/mergeHeader(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/net/www/MessageHeader/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000(Lsun/net/www/MessageHeader;)I"(thread: JVMThread, args: [sun_net_www_MessageHeader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/net/www/MessageHeader/access$000(Lsun/net/www/MessageHeader;)I"(thread: JVMThread, args: [sun_net_www_MessageHeader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Lsun/net/www/MessageHeader;)[Ljava/lang/String;"(thread: JVMThread, args: [sun_net_www_MessageHeader], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "sun/net/www/MessageHeader/access$100(Lsun/net/www/MessageHeader;)[Ljava/lang/String;"(thread: JVMThread, args: [sun_net_www_MessageHeader], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "access$200(Lsun/net/www/MessageHeader;)[Ljava/lang/String;"(thread: JVMThread, args: [sun_net_www_MessageHeader], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "sun/net/www/MessageHeader/access$200(Lsun/net/www/MessageHeader;)[Ljava/lang/String;"(thread: JVMThread, args: [sun_net_www_MessageHeader], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
  }
  export interface java_net_URLStreamHandlerFactory extends java_lang_Object {
    "createURLStreamHandler(Ljava/lang/String;)Ljava/net/URLStreamHandler;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URLStreamHandler) => void): void;
  }
  export class java_net_URLStreamHandler extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLStreamHandler/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "openConnection(Ljava/net/URL;)Ljava/net/URLConnection;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_net_URLConnection) => void): void;
    public "java/net/URLStreamHandler/openConnection(Ljava/net/URL;)Ljava/net/URLConnection;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_net_URLConnection) => void): void;
    public "openConnection(Ljava/net/URL;Ljava/net/Proxy;)Ljava/net/URLConnection;"(thread: JVMThread, args: [java_net_URL, java_net_Proxy], cb?: (e?: java_lang_Throwable, rv?: java_net_URLConnection) => void): void;
    public "java/net/URLStreamHandler/openConnection(Ljava/net/URL;Ljava/net/Proxy;)Ljava/net/URLConnection;"(thread: JVMThread, args: [java_net_URL, java_net_Proxy], cb?: (e?: java_lang_Throwable, rv?: java_net_URLConnection) => void): void;
    public "parseURL(Ljava/net/URL;Ljava/lang/String;II)V"(thread: JVMThread, args: [java_net_URL, java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLStreamHandler/parseURL(Ljava/net/URL;Ljava/lang/String;II)V"(thread: JVMThread, args: [java_net_URL, java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDefaultPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLStreamHandler/getDefaultPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/net/URL;Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLStreamHandler/equals(Ljava/net/URL;Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode(Ljava/net/URL;)I"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLStreamHandler/hashCode(Ljava/net/URL;)I"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sameFile(Ljava/net/URL;Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLStreamHandler/sameFile(Ljava/net/URL;Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHostAddress(Ljava/net/URL;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/URLStreamHandler/getHostAddress(Ljava/net/URL;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "hostsEqual(Ljava/net/URL;Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/URLStreamHandler/hostsEqual(Ljava/net/URL;Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toExternalForm(Ljava/net/URL;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/URLStreamHandler/toExternalForm(Ljava/net/URL;)Ljava/lang/String;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setURL(Ljava/net/URL;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_URL, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLStreamHandler/setURL(Ljava/net/URL;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_URL, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setURL(Ljava/net/URL;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_URL, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/URLStreamHandler/setURL(Ljava/net/URL;Ljava/lang/String;Ljava/lang/String;ILjava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_net_URL, java_lang_String, java_lang_String, number, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_URLClassPath extends java_lang_Object {
    public static "sun/misc/URLClassPath/USER_AGENT_JAVA_VERSION": java_lang_String;
    public static "sun/misc/URLClassPath/JAVA_VERSION": java_lang_String;
    public static "sun/misc/URLClassPath/DEBUG": number;
    public static "sun/misc/URLClassPath/DEBUG_LOOKUP_CACHE": number;
    public static "sun/misc/URLClassPath/DISABLE_JAR_CHECKING": number;
    public "sun/misc/URLClassPath/path": java_util_ArrayList;
    public "sun/misc/URLClassPath/urls": java_util_Stack;
    public "sun/misc/URLClassPath/loaders": java_util_ArrayList;
    public "sun/misc/URLClassPath/lmap": java_util_HashMap;
    public "sun/misc/URLClassPath/jarHandler": java_net_URLStreamHandler;
    public "sun/misc/URLClassPath/closed": number;
    public static "sun/misc/URLClassPath/lookupCacheEnabled": number;
    public "sun/misc/URLClassPath/lookupCacheURLs": JVMArray<java_net_URL>;
    public "sun/misc/URLClassPath/lookupCacheLoader": java_lang_ClassLoader;
    public "<init>([Ljava/net/URL;Ljava/net/URLStreamHandlerFactory;)V"(thread: JVMThread, args: [JVMArray<java_net_URL>, java_net_URLStreamHandlerFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/URLClassPath/<init>([Ljava/net/URL;Ljava/net/URLStreamHandlerFactory;)V"(thread: JVMThread, args: [JVMArray<java_net_URL>, java_net_URLStreamHandlerFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Ljava/net/URL;)V"(thread: JVMThread, args: [JVMArray<java_net_URL>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/URLClassPath/<init>([Ljava/net/URL;)V"(thread: JVMThread, args: [JVMArray<java_net_URL>], cb?: (e?: java_lang_Throwable) => void): void;
    public "closeLoaders()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/misc/URLClassPath/closeLoaders()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "addURL(Ljava/net/URL;)V"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/URLClassPath/addURL(Ljava/net/URL;)V"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable) => void): void;
    public "getURLs()[Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_URL>) => void): void;
    public "sun/misc/URLClassPath/getURLs()[Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_URL>) => void): void;
    public "findResource(Ljava/lang/String;Z)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "sun/misc/URLClassPath/findResource(Ljava/lang/String;Z)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "getResource(Ljava/lang/String;Z)Lsun/misc/Resource;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Resource) => void): void;
    public "sun/misc/URLClassPath/getResource(Ljava/lang/String;Z)Lsun/misc/Resource;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Resource) => void): void;
    public "findResources(Ljava/lang/String;Z)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "sun/misc/URLClassPath/findResources(Ljava/lang/String;Z)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getResource(Ljava/lang/String;)Lsun/misc/Resource;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Resource) => void): void;
    public "sun/misc/URLClassPath/getResource(Ljava/lang/String;)Lsun/misc/Resource;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Resource) => void): void;
    public "getResources(Ljava/lang/String;Z)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "sun/misc/URLClassPath/getResources(Ljava/lang/String;Z)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "sun/misc/URLClassPath/getResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "initLookupCache(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/URLClassPath/initLookupCache(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public static "disableAllLookupCaches()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/URLClassPath/disableAllLookupCaches()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getLookupCacheURLs(Ljava/lang/ClassLoader;)[Ljava/net/URL;"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_URL>) => void): void;
    public static "sun/misc/URLClassPath/getLookupCacheURLs(Ljava/lang/ClassLoader;)[Ljava/net/URL;"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_URL>) => void): void;
    public static "getLookupCacheForClassLoader(Ljava/lang/ClassLoader;Ljava/lang/String;)[I"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/misc/URLClassPath/getLookupCacheForClassLoader(Ljava/lang/ClassLoader;Ljava/lang/String;)[I"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "knownToNotExist0(Ljava/lang/ClassLoader;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/URLClassPath/knownToNotExist0(Ljava/lang/ClassLoader;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "knownToNotExist(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/URLClassPath/knownToNotExist(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLookupCache(Ljava/lang/String;)[I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/misc/URLClassPath/getLookupCache(Ljava/lang/String;)[I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "ensureLoaderOpened(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/URLClassPath/ensureLoaderOpened(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "validateLookupCache(ILjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/URLClassPath/validateLookupCache(ILjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNextLoader([II)Lsun/misc/URLClassPath$Loader;"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath$Loader) => void): void;
    public "sun/misc/URLClassPath/getNextLoader([II)Lsun/misc/URLClassPath$Loader;"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath$Loader) => void): void;
    public "getLoader(I)Lsun/misc/URLClassPath$Loader;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath$Loader) => void): void;
    public "sun/misc/URLClassPath/getLoader(I)Lsun/misc/URLClassPath$Loader;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath$Loader) => void): void;
    public "getLoader(Ljava/net/URL;)Lsun/misc/URLClassPath$Loader;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath$Loader) => void): void;
    public "sun/misc/URLClassPath/getLoader(Ljava/net/URL;)Lsun/misc/URLClassPath$Loader;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath$Loader) => void): void;
    public "push([Ljava/net/URL;)V"(thread: JVMThread, args: [JVMArray<java_net_URL>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/URLClassPath/push([Ljava/net/URL;)V"(thread: JVMThread, args: [JVMArray<java_net_URL>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "pathToURLs(Ljava/lang/String;)[Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_URL>) => void): void;
    public static "sun/misc/URLClassPath/pathToURLs(Ljava/lang/String;)[Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_URL>) => void): void;
    public "checkURL(Ljava/net/URL;)Ljava/net/URL;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "sun/misc/URLClassPath/checkURL(Ljava/net/URL;)Ljava/net/URL;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public static "check(Ljava/net/URL;)V"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/URLClassPath/check(Ljava/net/URL;)V"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Lsun/misc/URLClassPath;Ljava/lang/String;)[I"(thread: JVMThread, args: [sun_misc_URLClassPath, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/misc/URLClassPath/access$000(Lsun/misc/URLClassPath;Ljava/lang/String;)[I"(thread: JVMThread, args: [sun_misc_URLClassPath, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$100(Lsun/misc/URLClassPath;[II)Lsun/misc/URLClassPath$Loader;"(thread: JVMThread, args: [sun_misc_URLClassPath, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath$Loader) => void): void;
    public static "sun/misc/URLClassPath/access$100(Lsun/misc/URLClassPath;[II)Lsun/misc/URLClassPath$Loader;"(thread: JVMThread, args: [sun_misc_URLClassPath, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath$Loader) => void): void;
    public static "access$200(Lsun/misc/URLClassPath;)Ljava/net/URLStreamHandler;"(thread: JVMThread, args: [sun_misc_URLClassPath], cb?: (e?: java_lang_Throwable, rv?: java_net_URLStreamHandler) => void): void;
    public static "sun/misc/URLClassPath/access$200(Lsun/misc/URLClassPath;)Ljava/net/URLStreamHandler;"(thread: JVMThread, args: [sun_misc_URLClassPath], cb?: (e?: java_lang_Throwable, rv?: java_net_URLStreamHandler) => void): void;
    public static "access$400()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/URLClassPath/access$400()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1100()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/URLClassPath/access$1100()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/URLClassPath/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_URLClassPath$Loader extends java_lang_Object implements java_io_Closeable {
    public "sun/misc/URLClassPath$Loader/base": java_net_URL;
    public "sun/misc/URLClassPath$Loader/jarfile": java_util_jar_JarFile;
    public "<init>(Ljava/net/URL;)V"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/URLClassPath$Loader/<init>(Ljava/net/URL;)V"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBaseURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "sun/misc/URLClassPath$Loader/getBaseURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "findResource(Ljava/lang/String;Z)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "sun/misc/URLClassPath$Loader/findResource(Ljava/lang/String;Z)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "getResource(Ljava/lang/String;Z)Lsun/misc/Resource;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Resource) => void): void;
    public "sun/misc/URLClassPath$Loader/getResource(Ljava/lang/String;Z)Lsun/misc/Resource;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Resource) => void): void;
    public "getResource(Ljava/lang/String;)Lsun/misc/Resource;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Resource) => void): void;
    public "sun/misc/URLClassPath$Loader/getResource(Ljava/lang/String;)Lsun/misc/Resource;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Resource) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/URLClassPath$Loader/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getClassPath()[Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_URL>) => void): void;
    public "sun/misc/URLClassPath$Loader/getClassPath()[Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_URL>) => void): void;
    public static "access$300(Lsun/misc/URLClassPath$Loader;)Ljava/net/URL;"(thread: JVMThread, args: [sun_misc_URLClassPath$Loader], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public static "sun/misc/URLClassPath$Loader/access$300(Lsun/misc/URLClassPath$Loader;)Ljava/net/URL;"(thread: JVMThread, args: [sun_misc_URLClassPath$Loader], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
  }
  export class sun_misc_Resource extends java_lang_Object {
    public "sun/misc/Resource/cis": java_io_InputStream;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Resource/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/misc/Resource/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "sun/misc/Resource/getURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "getCodeSourceURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "sun/misc/Resource/getCodeSourceURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "sun/misc/Resource/getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getContentLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/Resource/getContentLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "cachedInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "sun/misc/Resource/cachedInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/misc/Resource/getBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getByteBuffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/misc/Resource/getByteBuffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "getManifest()Ljava/util/jar/Manifest;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public "sun/misc/Resource/getManifest()Ljava/util/jar/Manifest;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public "getCertificates()[Ljava/security/cert/Certificate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "sun/misc/Resource/getCertificates()[Ljava/security/cert/Certificate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "getCodeSigners()[Ljava/security/CodeSigner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "sun/misc/Resource/getCodeSigners()[Ljava/security/CodeSigner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
  }
  export class java_security_CodeSigner extends java_lang_Object implements java_io_Serializable {
    public static "java/security/CodeSigner/serialVersionUID": Long;
    public "java/security/CodeSigner/signerCertPath": java_security_cert_CertPath;
    public "java/security/CodeSigner/timestamp": java_security_Timestamp;
    public "java/security/CodeSigner/myhash": number;
    public "<init>(Ljava/security/cert/CertPath;Ljava/security/Timestamp;)V"(thread: JVMThread, args: [java_security_cert_CertPath, java_security_Timestamp], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/CodeSigner/<init>(Ljava/security/cert/CertPath;Ljava/security/Timestamp;)V"(thread: JVMThread, args: [java_security_cert_CertPath, java_security_Timestamp], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSignerCertPath()Ljava/security/cert/CertPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "java/security/CodeSigner/getSignerCertPath()Ljava/security/cert/CertPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "getTimestamp()Ljava/security/Timestamp;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Timestamp) => void): void;
    public "java/security/CodeSigner/getTimestamp()Ljava/security/Timestamp;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Timestamp) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/CodeSigner/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/CodeSigner/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/CodeSigner/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/CodeSigner/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_Timestamp extends java_lang_Object implements java_io_Serializable {
    public static "java/security/Timestamp/serialVersionUID": Long;
    public "java/security/Timestamp/timestamp": java_util_Date;
    public "java/security/Timestamp/signerCertPath": java_security_cert_CertPath;
    public "java/security/Timestamp/myhash": number;
    public "<init>(Ljava/util/Date;Ljava/security/cert/CertPath;)V"(thread: JVMThread, args: [java_util_Date, java_security_cert_CertPath], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Timestamp/<init>(Ljava/util/Date;Ljava/security/cert/CertPath;)V"(thread: JVMThread, args: [java_util_Date, java_security_cert_CertPath], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTimestamp()Ljava/util/Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Date) => void): void;
    public "java/security/Timestamp/getTimestamp()Ljava/util/Date;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Date) => void): void;
    public "getSignerCertPath()Ljava/security/cert/CertPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "java/security/Timestamp/getSignerCertPath()Ljava/security/cert/CertPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Timestamp/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Timestamp/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Timestamp/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Timestamp/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_cert_CertPath extends java_lang_Object implements java_io_Serializable {
    public static "java/security/cert/CertPath/serialVersionUID": Long;
    public "java/security/cert/CertPath/type": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/cert/CertPath/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/cert/CertPath/getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getEncodings()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/security/cert/CertPath/getEncodings()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/cert/CertPath/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/cert/CertPath/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/cert/CertPath/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getEncoded()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/security/cert/CertPath/getEncoded()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getEncoded(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/security/cert/CertPath/getEncoded(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getCertificates()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/security/cert/CertPath/getCertificates()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/cert/CertPath/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_security_cert_Certificate extends java_lang_Object implements java_io_Serializable {
    public static "java/security/cert/Certificate/serialVersionUID": Long;
    public "java/security/cert/Certificate/type": java_lang_String;
    public "java/security/cert/Certificate/hash": number;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/cert/Certificate/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/cert/Certificate/getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/cert/Certificate/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/cert/Certificate/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getEncoded()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/security/cert/Certificate/getEncoded()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "verify(Ljava/security/PublicKey;)V"(thread: JVMThread, args: [java_security_PublicKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/cert/Certificate/verify(Ljava/security/PublicKey;)V"(thread: JVMThread, args: [java_security_PublicKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "verify(Ljava/security/PublicKey;Ljava/lang/String;)V"(thread: JVMThread, args: [java_security_PublicKey, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/cert/Certificate/verify(Ljava/security/PublicKey;Ljava/lang/String;)V"(thread: JVMThread, args: [java_security_PublicKey, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "verify(Ljava/security/PublicKey;Ljava/security/Provider;)V"(thread: JVMThread, args: [java_security_PublicKey, java_security_Provider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/cert/Certificate/verify(Ljava/security/PublicKey;Ljava/security/Provider;)V"(thread: JVMThread, args: [java_security_PublicKey, java_security_Provider], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/cert/Certificate/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPublicKey()Ljava/security/PublicKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PublicKey) => void): void;
    public "java/security/cert/Certificate/getPublicKey()Ljava/security/PublicKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PublicKey) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/cert/Certificate/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_security_Provider extends java_util_Properties {
    public static "java/security/Provider/serialVersionUID": Long;
    public static "java/security/Provider/debug": sun_security_util_Debug;
    public "java/security/Provider/name": java_lang_String;
    public "java/security/Provider/info": java_lang_String;
    public "java/security/Provider/version": number;
    public "java/security/Provider/entrySet": java_util_Set;
    public "java/security/Provider/entrySetCallCount": number;
    public "java/security/Provider/initialized": number;
    public "java/security/Provider/legacyChanged": number;
    public "java/security/Provider/servicesChanged": number;
    public "java/security/Provider/legacyStrings": java_util_Map;
    public "java/security/Provider/serviceMap": java_util_Map;
    public "java/security/Provider/legacyMap": java_util_Map;
    public "java/security/Provider/serviceSet": java_util_Set;
    public static "java/security/Provider/ALIAS_PREFIX": java_lang_String;
    public static "java/security/Provider/ALIAS_PREFIX_LOWER": java_lang_String;
    public static "java/security/Provider/ALIAS_LENGTH": number;
    public static "java/security/Provider/previousKey": java_security_Provider$ServiceKey;
    public static "java/security/Provider/knownEngines": java_util_Map;
    public "<init>(Ljava/lang/String;DLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, any, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/<init>(Ljava/lang/String;DLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, any, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Provider/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVersion()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider/getVersion()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInfo()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Provider/getInfo()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Provider/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "load(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/load(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/security/Provider/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/security/Provider/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/security/Provider/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider/remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider/replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "keys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/security/Provider/keys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/security/Provider/elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Provider/getProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "checkInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/checkInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "check(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/check(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "putId()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/putId()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkLegacy(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider/checkLegacy(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "implPutAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/implPutAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "implRemove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/implRemove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implRemove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider/implRemove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "implReplace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider/implReplace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "implReplace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/implReplace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implReplaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/implReplaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "implMerge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/implMerge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implCompute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/implCompute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implComputeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/implComputeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implComputeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/implComputeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implPut(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/implPut(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implPutIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider/implPutIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "implClear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/implClear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureLegacyParsed()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/ensureLegacyParsed()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeInvalidServices(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/removeInvalidServices(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTypeAndAlgorithm(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/security/Provider/getTypeAndAlgorithm(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "parseLegacyPut(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/parseLegacyPut(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getService(Ljava/lang/String;Ljava/lang/String;)Ljava/security/Provider$Service;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider$Service) => void): void;
    public "java/security/Provider/getService(Ljava/lang/String;Ljava/lang/String;)Ljava/security/Provider$Service;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider$Service) => void): void;
    public "getServices()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/security/Provider/getServices()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "putService(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/putService(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public "putPropertyStrings(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/putPropertyStrings(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePropertyStrings(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/removePropertyStrings(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeService(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/removeService(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public "implRemoveService(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider/implRemoveService(Ljava/security/Provider$Service;)V"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable) => void): void;
    public static "addEngine(Ljava/lang/String;ZLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/security/Provider/addEngine(Ljava/lang/String;ZLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getEngineName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/security/Provider/getEngineName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$900(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/security/Provider/access$900(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$1000()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/security/Provider/access$1000()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/security/Provider/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_Provider$Service extends java_lang_Object {
    public "java/security/Provider$Service/type": java_lang_String;
    public "java/security/Provider$Service/algorithm": java_lang_String;
    public "java/security/Provider$Service/className": java_lang_String;
    public "java/security/Provider$Service/provider": java_security_Provider;
    public "java/security/Provider$Service/aliases": java_util_List;
    public "java/security/Provider$Service/attributes": java_util_Map;
    public "java/security/Provider$Service/classRef": java_lang_ref_Reference;
    public "java/security/Provider$Service/hasKeyAttributes": java_lang_Boolean;
    public "java/security/Provider$Service/supportedFormats": JVMArray<java_lang_String>;
    public "java/security/Provider$Service/supportedClasses": JVMArray<java_lang_Class>;
    public "java/security/Provider$Service/registered": number;
    public static "java/security/Provider$Service/CLASS0": JVMArray<java_lang_Class>;
    public "<init>(Ljava/security/Provider;)V"(thread: JVMThread, args: [java_security_Provider], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider$Service/<init>(Ljava/security/Provider;)V"(thread: JVMThread, args: [java_security_Provider], cb?: (e?: java_lang_Throwable) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider$Service/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAlias(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider$Service/addAlias(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAttribute(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider$Service/addAttribute(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/security/Provider;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/util/List;Ljava/util/Map;)V"(thread: JVMThread, args: [java_security_Provider, java_lang_String, java_lang_String, java_lang_String, java_util_List, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider$Service/<init>(Ljava/security/Provider;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/util/List;Ljava/util/Map;)V"(thread: JVMThread, args: [java_security_Provider, java_lang_String, java_lang_String, java_lang_String, java_util_List, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Provider$Service/getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAlgorithm()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Provider$Service/getAlgorithm()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getProvider()Ljava/security/Provider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public "java/security/Provider$Service/getProvider()Ljava/security/Provider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public "getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Provider$Service/getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAliases()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/security/Provider$Service/getAliases()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getAttribute(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Provider$Service/getAttribute(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "newInstance(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider$Service/newInstance(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getImplClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/security/Provider$Service/getImplClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "newInstanceGeneric(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/Provider$Service/newInstanceGeneric(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "supportsParameter(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider$Service/supportsParameter(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasKeyAttributes()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider$Service/hasKeyAttributes()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getKeyClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/security/Provider$Service/getKeyClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "supportsKeyFormat(Ljava/security/Key;)Z"(thread: JVMThread, args: [java_security_Key], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider$Service/supportsKeyFormat(Ljava/security/Key;)Z"(thread: JVMThread, args: [java_security_Key], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "supportsKeyClass(Ljava/security/Key;)Z"(thread: JVMThread, args: [java_security_Key], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider$Service/supportsKeyClass(Ljava/security/Key;)Z"(thread: JVMThread, args: [java_security_Key], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/Provider$Service/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000(Ljava/security/Provider$Service;)Z"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/security/Provider$Service/access$000(Ljava/security/Provider$Service;)Z"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/security/Provider;Ljava/security/Provider$1;)V"(thread: JVMThread, args: [java_security_Provider, java_security_Provider$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider$Service/<init>(Ljava/security/Provider;Ljava/security/Provider$1;)V"(thread: JVMThread, args: [java_security_Provider, java_security_Provider$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$302(Ljava/security/Provider$Service;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_Provider$Service, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/security/Provider$Service/access$302(Ljava/security/Provider$Service;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_Provider$Service, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$402(Ljava/security/Provider$Service;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_Provider$Service, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/security/Provider$Service/access$402(Ljava/security/Provider$Service;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_Provider$Service, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$500(Ljava/security/Provider$Service;Ljava/lang/String;)V"(thread: JVMThread, args: [java_security_Provider$Service, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/security/Provider$Service/access$500(Ljava/security/Provider$Service;Ljava/lang/String;)V"(thread: JVMThread, args: [java_security_Provider$Service, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$602(Ljava/security/Provider$Service;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_Provider$Service, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/security/Provider$Service/access$602(Ljava/security/Provider$Service;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_Provider$Service, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$700(Ljava/security/Provider$Service;)Ljava/util/List;"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/security/Provider$Service/access$700(Ljava/security/Provider$Service;)Ljava/util/List;"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "access$800(Ljava/security/Provider$Service;)Ljava/util/Map;"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/security/Provider$Service/access$800(Ljava/security/Provider$Service;)Ljava/util/Map;"(thread: JVMThread, args: [java_security_Provider$Service], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/security/Provider$Service/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_Provider$1 extends java_lang_Object {
  }
  export interface java_security_Key extends java_lang_Object, java_io_Serializable {
    "getAlgorithm()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getFormat()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getEncoded()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export class java_security_Provider$ServiceKey extends java_lang_Object {
    public "java/security/Provider$ServiceKey/type": java_lang_String;
    public "java/security/Provider$ServiceKey/algorithm": java_lang_String;
    public "java/security/Provider$ServiceKey/originalAlgorithm": java_lang_String;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider$ServiceKey/<init>(Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider$ServiceKey/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider$ServiceKey/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "matches(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/Provider$ServiceKey/matches(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;ZLjava/security/Provider$1;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_security_Provider$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/Provider$ServiceKey/<init>(Ljava/lang/String;Ljava/lang/String;ZLjava/security/Provider$1;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, java_security_Provider$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_security_util_Debug extends java_lang_Object {
    public "sun/security/util/Debug/prefix": java_lang_String;
    public static "sun/security/util/Debug/args": java_lang_String;
    public static "sun/security/util/Debug/hexDigits": JVMArray<number>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/Debug/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "Help()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/security/util/Debug/Help()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance(Ljava/lang/String;)Lsun/security/util/Debug;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_Debug) => void): void;
    public static "sun/security/util/Debug/getInstance(Ljava/lang/String;)Lsun/security/util/Debug;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_Debug) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/lang/String;)Lsun/security/util/Debug;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_Debug) => void): void;
    public static "sun/security/util/Debug/getInstance(Ljava/lang/String;Ljava/lang/String;)Lsun/security/util/Debug;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_Debug) => void): void;
    public static "isOn(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/security/util/Debug/isOn(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "println(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/Debug/println(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "println()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/Debug/println()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "println(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/security/util/Debug/println(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "toHexString(Ljava/math/BigInteger;)Ljava/lang/String;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/security/util/Debug/toHexString(Ljava/math/BigInteger;)Ljava/lang/String;"(thread: JVMThread, args: [java_math_BigInteger], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "marshal(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/security/util/Debug/marshal(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toString([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/security/util/Debug/toString([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/security/util/Debug/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_security_PublicKey extends java_lang_Object, java_security_Key {
    "getAlgorithm()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getFormat()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getEncoded()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export class java_util_jar_Manifest extends java_lang_Object implements java_lang_Cloneable {
    public "java/util/jar/Manifest/attr": java_util_jar_Attributes;
    public "java/util/jar/Manifest/entries": java_util_Map;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest/<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/jar/Manifest;)V"(thread: JVMThread, args: [java_util_jar_Manifest], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest/<init>(Ljava/util/jar/Manifest;)V"(thread: JVMThread, args: [java_util_jar_Manifest], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMainAttributes()Ljava/util/jar/Attributes;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Attributes) => void): void;
    public "java/util/jar/Manifest/getMainAttributes()Ljava/util/jar/Attributes;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Attributes) => void): void;
    public "getEntries()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/util/jar/Manifest/getEntries()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getAttributes(Ljava/lang/String;)Ljava/util/jar/Attributes;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Attributes) => void): void;
    public "java/util/jar/Manifest/getAttributes(Ljava/lang/String;)Ljava/util/jar/Attributes;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Attributes) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest/write(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make72Safe(Ljava/lang/StringBuffer;)V"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/jar/Manifest/make72Safe(Ljava/lang/StringBuffer;)V"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest/read(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "parseName([BI)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/jar/Manifest/parseName([BI)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toLower(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Manifest/toLower(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Manifest/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Manifest/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/jar/Manifest/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_jar_Attributes extends java_lang_Object implements java_util_Map, java_lang_Cloneable {
    public "java/util/jar/Attributes/map": java_util_Map;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Attributes/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Attributes/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/jar/Attributes;)V"(thread: JVMThread, args: [java_util_jar_Attributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Attributes/<init>(Ljava/util/jar/Attributes;)V"(thread: JVMThread, args: [java_util_jar_Attributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/jar/Attributes/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getValue(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/jar/Attributes/getValue(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getValue(Ljava/util/jar/Attributes$Name;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_jar_Attributes$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/jar/Attributes/getValue(Ljava/util/jar/Attributes$Name;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_jar_Attributes$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/jar/Attributes/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putValue(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/jar/Attributes/putValue(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/jar/Attributes/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Attributes/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Attributes/containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Attributes/putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Attributes/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Attributes/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Attributes/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/jar/Attributes/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/jar/Attributes/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/jar/Attributes/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Attributes/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Attributes/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/jar/Attributes/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "write(Ljava/io/DataOutputStream;)V"(thread: JVMThread, args: [java_io_DataOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Attributes/write(Ljava/io/DataOutputStream;)V"(thread: JVMThread, args: [java_io_DataOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeMain(Ljava/io/DataOutputStream;)V"(thread: JVMThread, args: [java_io_DataOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Attributes/writeMain(Ljava/io/DataOutputStream;)V"(thread: JVMThread, args: [java_io_DataOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/util/jar/Manifest$FastInputStream;[B)V"(thread: JVMThread, args: [java_util_jar_Manifest$FastInputStream, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Attributes/read(Ljava/util/jar/Manifest$FastInputStream;[B)V"(thread: JVMThread, args: [java_util_jar_Manifest$FastInputStream, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_jar_Manifest$FastInputStream extends java_io_FilterInputStream {
    public "java/util/jar/Manifest$FastInputStream/buf": JVMArray<number>;
    public "java/util/jar/Manifest$FastInputStream/count": number;
    public "java/util/jar/Manifest$FastInputStream/pos": number;
    public "<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest$FastInputStream/<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/InputStream;I)V"(thread: JVMThread, args: [java_io_InputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest$FastInputStream/<init>(Ljava/io/InputStream;I)V"(thread: JVMThread, args: [java_io_InputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Manifest$FastInputStream/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Manifest$FastInputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readLine([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Manifest$FastInputStream/readLine([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peek()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Manifest$FastInputStream/peek()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readLine([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Manifest$FastInputStream/readLine([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/jar/Manifest$FastInputStream/skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Manifest$FastInputStream/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest$FastInputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "fill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Manifest$FastInputStream/fill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_FilterInputStream extends java_io_InputStream {
    public "java/io/FilterInputStream/in": java_io_InputStream;
    public "<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterInputStream/<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FilterInputStream/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FilterInputStream/read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FilterInputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/FilterInputStream/skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FilterInputStream/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterInputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mark(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterInputStream/mark(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterInputStream/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "markSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FilterInputStream/markSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_io_DataOutputStream extends java_io_FilterOutputStream implements java_io_DataOutput {
    public "java/io/DataOutputStream/written": number;
    public "java/io/DataOutputStream/bytearr": JVMArray<number>;
    public "java/io/DataOutputStream/writeBuffer": JVMArray<number>;
    public "<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "incCount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/incCount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataOutputStream/writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "writeUTF(Ljava/lang/String;Ljava/io/DataOutput;)I"(thread: JVMThread, args: [java_lang_String, java_io_DataOutput], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/DataOutputStream/writeUTF(Ljava/lang/String;Ljava/io/DataOutput;)I"(thread: JVMThread, args: [java_lang_String, java_io_DataOutput], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataOutputStream/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_io_FilterOutputStream extends java_io_OutputStream {
    public "java/io/FilterOutputStream/out": java_io_OutputStream;
    public "<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterOutputStream/<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterOutputStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterOutputStream/write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterOutputStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterOutputStream/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FilterOutputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_jar_Attributes$Name extends java_lang_Object {
    public "java/util/jar/Attributes$Name/name": java_lang_String;
    public "java/util/jar/Attributes$Name/hashCode": number;
    public static "java/util/jar/Attributes$Name/MANIFEST_VERSION": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/SIGNATURE_VERSION": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/CONTENT_TYPE": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/CLASS_PATH": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/MAIN_CLASS": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/SEALED": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/EXTENSION_LIST": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/EXTENSION_NAME": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/EXTENSION_INSTALLATION": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/IMPLEMENTATION_TITLE": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/IMPLEMENTATION_VERSION": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/IMPLEMENTATION_VENDOR": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/IMPLEMENTATION_VENDOR_ID": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/IMPLEMENTATION_URL": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/SPECIFICATION_TITLE": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/SPECIFICATION_VERSION": java_util_jar_Attributes$Name;
    public static "java/util/jar/Attributes$Name/SPECIFICATION_VENDOR": java_util_jar_Attributes$Name;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/Attributes$Name/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isValid(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/jar/Attributes$Name/isValid(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isValid(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/jar/Attributes$Name/isValid(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isAlpha(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/jar/Attributes$Name/isAlpha(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/jar/Attributes$Name/isDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Attributes$Name/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/Attributes$Name/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/jar/Attributes$Name/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/jar/Attributes$Name/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_Stack extends java_util_Vector {
    public static "java/util/Stack/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Stack/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "push(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Stack/push(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "pop()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Stack/pop()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Stack/peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "empty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Stack/empty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "search(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Stack/search(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_Vector extends java_util_AbstractList implements java_util_List, java_util_RandomAccess, java_lang_Cloneable, java_io_Serializable {
    public "java/util/Vector/elementData": JVMArray<java_lang_Object>;
    public "java/util/Vector/elementCount": number;
    public "java/util/Vector/capacityIncrement": number;
    public static "java/util/Vector/serialVersionUID": Long;
    public static "java/util/Vector/MAX_ARRAY_SIZE": number;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyInto([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/copyInto([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureCapacityHelper(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/ensureCapacityHelper(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Vector/hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/setSize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/capacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/Vector/elements()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/indexOf(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/lastIndexOf(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "elementAt(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Vector/elementAt(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "firstElement()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Vector/firstElement()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "lastElement()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Vector/lastElement()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setElementAt(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/setElementAt(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeElementAt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/removeElementAt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "insertElementAt(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/insertElementAt(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addElement(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/addElement(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeElement(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/removeElement(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeAllElements()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/removeAllElements()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Vector/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/Vector/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/Vector/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "elementData(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Vector/elementData(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Vector/get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Vector/set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Vector/remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Vector/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "subList(II)Ljava/util/List;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/Vector/subList(II)Ljava/util/List;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/Vector/listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "listIterator()Ljava/util/ListIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/Vector/listIterator()Ljava/util/ListIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/Vector/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Vector/removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    public "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Vector/sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/Vector/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_Perf extends java_lang_Object {
    public static "sun/misc/Perf/instance": sun_misc_Perf;
    public static "sun/misc/Perf/PERF_MODE_RO": number;
    public static "sun/misc/Perf/PERF_MODE_RW": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Perf/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getPerf()Lsun/misc/Perf;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Perf) => void): void;
    public static "sun/misc/Perf/getPerf()Lsun/misc/Perf;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Perf) => void): void;
    public "attach(ILjava/lang/String;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/misc/Perf/attach(ILjava/lang/String;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "attach(Ljava/lang/String;ILjava/lang/String;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/misc/Perf/attach(Ljava/lang/String;ILjava/lang/String;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "attachImpl(Ljava/lang/String;II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/misc/Perf/attachImpl(Ljava/lang/String;II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "attach(Ljava/lang/String;II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/misc/Perf/attach(Ljava/lang/String;II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "detach(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/Perf/detach(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "createLong(Ljava/lang/String;IIJ)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/misc/Perf/createLong(Ljava/lang/String;IIJ)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "createString(Ljava/lang/String;IILjava/lang/String;I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/misc/Perf/createString(Ljava/lang/String;IILjava/lang/String;I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "createString(Ljava/lang/String;IILjava/lang/String;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/misc/Perf/createString(Ljava/lang/String;IILjava/lang/String;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "createByteArray(Ljava/lang/String;II[BI)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/misc/Perf/createByteArray(Ljava/lang/String;II[BI)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public static "getBytes(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/misc/Perf/getBytes(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "highResCounter()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Perf/highResCounter()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "highResFrequency()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/misc/Perf/highResFrequency()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Perf/registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000()Lsun/misc/Perf;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Perf) => void): void;
    public static "sun/misc/Perf/access$000()Lsun/misc/Perf;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Perf) => void): void;
    public static "access$100(Lsun/misc/Perf;Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [sun_misc_Perf, java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Perf/access$100(Lsun/misc/Perf;Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [sun_misc_Perf, java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/Perf/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_VMSupport extends java_lang_Object {
    public static "sun/misc/VMSupport/agentProps": java_util_Properties;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/VMSupport/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getAgentProperties()Ljava/util/Properties;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "sun/misc/VMSupport/getAgentProperties()Ljava/util/Properties;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "initAgentProperties(Ljava/util/Properties;)Ljava/util/Properties;"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "sun/misc/VMSupport/initAgentProperties(Ljava/util/Properties;)Ljava/util/Properties;"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "serializePropertiesToByteArray(Ljava/util/Properties;)[B"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/misc/VMSupport/serializePropertiesToByteArray(Ljava/util/Properties;)[B"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "serializePropertiesToByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/misc/VMSupport/serializePropertiesToByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "serializeAgentPropertiesToByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/misc/VMSupport/serializeAgentPropertiesToByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "isClassPathAttributePresent(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/misc/VMSupport/isClassPathAttributePresent(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getVMTemporaryDirectory()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/VMSupport/getVMTemporaryDirectory()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/VMSupport/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_MessageUtils extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/MessageUtils/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "subst(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/MessageUtils/subst(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "subst(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/MessageUtils/subst(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "subst(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/MessageUtils/subst(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "subst(Ljava/lang/String;[Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/MessageUtils/subst(Ljava/lang/String;[Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "substProp(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/MessageUtils/substProp(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "substProp(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/MessageUtils/substProp(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "substProp(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/misc/MessageUtils/substProp(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toStderr(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/MessageUtils/toStderr(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "toStdout(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/MessageUtils/toStdout(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "err(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/MessageUtils/err(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "out(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/MessageUtils/out(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "where()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/MessageUtils/where()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_GC extends java_lang_Object {
    public static "sun/misc/GC/NO_TARGET": Long;
    public static "sun/misc/GC/latencyTarget": Long;
    public static "sun/misc/GC/daemon": java_lang_Thread;
    public static "sun/misc/GC/lock": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/GC/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "maxObjectInspectionAge()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/misc/GC/maxObjectInspectionAge()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "setLatencyTarget(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/GC/setLatencyTarget(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "requestLatency(J)Lsun/misc/GC$LatencyRequest;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_misc_GC$LatencyRequest) => void): void;
    public static "sun/misc/GC/requestLatency(J)Lsun/misc/GC$LatencyRequest;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_misc_GC$LatencyRequest) => void): void;
    public static "currentLatencyTarget()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/misc/GC/currentLatencyTarget()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$100()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/misc/GC/access$100()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$200()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/misc/GC/access$200()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$302(Ljava/lang/Thread;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "sun/misc/GC/access$302(Ljava/lang/Thread;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "access$500(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/GC/access$500(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/GC/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_GC$LatencyRequest extends java_lang_Object implements java_lang_Comparable {
    public static "sun/misc/GC$LatencyRequest/counter": Long;
    public static "sun/misc/GC$LatencyRequest/requests": java_util_SortedSet;
    public "sun/misc/GC$LatencyRequest/latency": Long;
    public "sun/misc/GC$LatencyRequest/id": Long;
    public static "adjustLatencyIfNeeded()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/GC$LatencyRequest/adjustLatencyIfNeeded()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/GC$LatencyRequest/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "cancel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/GC$LatencyRequest/cancel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "compareTo(Lsun/misc/GC$LatencyRequest;)I"(thread: JVMThread, args: [sun_misc_GC$LatencyRequest], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/GC$LatencyRequest/compareTo(Lsun/misc/GC$LatencyRequest;)I"(thread: JVMThread, args: [sun_misc_GC$LatencyRequest], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/misc/GC$LatencyRequest/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/misc/GC$LatencyRequest/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(JLsun/misc/GC$1;)V"(thread: JVMThread, args: [Long, any, sun_misc_GC$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/misc/GC$LatencyRequest/<init>(JLsun/misc/GC$1;)V"(thread: JVMThread, args: [Long, any, sun_misc_GC$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/misc/GC$LatencyRequest/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_misc_GC$1 extends java_lang_Object {
  }
  export interface java_util_SortedSet extends java_lang_Object, java_util_Set {
    "comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "subSet(Ljava/lang/Object;Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    "headSet(Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    "tailSet(Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    "first()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "last()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
  }
  export class sun_management_VMManagementImpl extends java_lang_Object implements sun_management_VMManagement {
    public static "sun/management/VMManagementImpl/version": java_lang_String;
    public static "sun/management/VMManagementImpl/compTimeMonitoringSupport": number;
    public static "sun/management/VMManagementImpl/threadContentionMonitoringSupport": number;
    public static "sun/management/VMManagementImpl/currentThreadCpuTimeSupport": number;
    public static "sun/management/VMManagementImpl/otherThreadCpuTimeSupport": number;
    public static "sun/management/VMManagementImpl/bootClassPathSupport": number;
    public static "sun/management/VMManagementImpl/objectMonitorUsageSupport": number;
    public static "sun/management/VMManagementImpl/synchronizerUsageSupport": number;
    public static "sun/management/VMManagementImpl/threadAllocatedMemorySupport": number;
    public static "sun/management/VMManagementImpl/gcNotificationSupport": number;
    public static "sun/management/VMManagementImpl/remoteDiagnosticCommandsSupport": number;
    public "sun/management/VMManagementImpl/vmArgs": java_util_List;
    public "sun/management/VMManagementImpl/perfInstr": sun_management_counter_perf_PerfInstrumentation;
    public "sun/management/VMManagementImpl/noPerfData": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/VMManagementImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getVersion0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/management/VMManagementImpl/getVersion0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initOptionalSupportFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/VMManagementImpl/initOptionalSupportFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isCompilationTimeMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isCompilationTimeMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadContentionMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isThreadContentionMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isCurrentThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isCurrentThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOtherThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isOtherThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBootClassPathSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isBootClassPathSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isObjectMonitorUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isObjectMonitorUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSynchronizerUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isSynchronizerUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadAllocatedMemorySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isThreadAllocatedMemorySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isGcNotificationSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isGcNotificationSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRemoteDiagnosticCommandsSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isRemoteDiagnosticCommandsSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadContentionMonitoringEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isThreadContentionMonitoringEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadCpuTimeEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isThreadCpuTimeEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadAllocatedMemoryEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/isThreadAllocatedMemoryEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLoadedClassCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/getLoadedClassCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTotalClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getTotalClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getUnloadedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getUnloadedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getVerboseClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/getVerboseClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getVerboseGC()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/getVerboseGC()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getManagementVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getManagementVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVmId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getVmId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getProcessId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/getProcessId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getVmName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getVmName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVmVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getVmVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVmVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getVmVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVmSpecName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getVmSpecName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVmSpecVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getVmSpecVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVmSpecVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getVmSpecVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getClassPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getClassPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getLibraryPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getLibraryPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getBootClassPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getBootClassPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getUptime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getUptime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getVmArguments()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/management/VMManagementImpl/getVmArguments()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getVmArguments0()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/management/VMManagementImpl/getVmArguments0()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getStartupTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getStartupTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getUptime0()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getUptime0()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getAvailableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/getAvailableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCompilerName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getCompilerName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTotalCompileTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getTotalCompileTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTotalThreadCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getTotalThreadCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLiveThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/getLiveThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPeakThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/getPeakThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDaemonThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/VMManagementImpl/getDaemonThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOsName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getOsName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getOsArch()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getOsArch()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getOsVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/VMManagementImpl/getOsVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSafepointCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getSafepointCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTotalSafepointTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getTotalSafepointTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getSafepointSyncTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getSafepointSyncTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTotalApplicationNonStoppedTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getTotalApplicationNonStoppedTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLoadedClassSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getLoadedClassSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getUnloadedClassSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getUnloadedClassSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getClassLoadingTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getClassLoadingTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMethodDataSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getMethodDataSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getInitializedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getInitializedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getClassInitializationTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getClassInitializationTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getClassVerificationTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/VMManagementImpl/getClassVerificationTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getPerfInstrumentation()Lsun/management/counter/perf/PerfInstrumentation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_perf_PerfInstrumentation) => void): void;
    public "sun/management/VMManagementImpl/getPerfInstrumentation()Lsun/management/counter/perf/PerfInstrumentation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_perf_PerfInstrumentation) => void): void;
    public "getInternalCounters(Ljava/lang/String;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/management/VMManagementImpl/getInternalCounters(Ljava/lang/String;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/VMManagementImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_counter_perf_PerfInstrumentation extends java_lang_Object {
    public "sun/management/counter/perf/PerfInstrumentation/buffer": java_nio_ByteBuffer;
    public "sun/management/counter/perf/PerfInstrumentation/prologue": sun_management_counter_perf_Prologue;
    public "sun/management/counter/perf/PerfInstrumentation/lastModificationTime": Long;
    public "sun/management/counter/perf/PerfInstrumentation/lastUsed": Long;
    public "sun/management/counter/perf/PerfInstrumentation/nextEntry": number;
    public "sun/management/counter/perf/PerfInstrumentation/map": java_util_SortedMap;
    public static "sun/management/counter/perf/PerfInstrumentation/$assertionsDisabled": number;
    public "<init>(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/counter/perf/PerfInstrumentation/<init>(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/PerfInstrumentation/getMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/PerfInstrumentation/getMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getModificationTimeStamp()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/counter/perf/PerfInstrumentation/getModificationTimeStamp()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "rewind()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/counter/perf/PerfInstrumentation/rewind()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/PerfInstrumentation/hasNext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNextCounter()Lsun/management/counter/Counter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_Counter) => void): void;
    public "sun/management/counter/perf/PerfInstrumentation/getNextCounter()Lsun/management/counter/Counter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_Counter) => void): void;
    public "getAllCounters()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/management/counter/perf/PerfInstrumentation/getAllCounters()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "findByPattern(Ljava/lang/String;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/management/counter/perf/PerfInstrumentation/findByPattern(Ljava/lang/String;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/counter/perf/PerfInstrumentation/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_management_counter_Counter extends java_lang_Object, java_io_Serializable {
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getUnits()Lsun/management/counter/Units;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_Units) => void): void;
    "getVariability()Lsun/management/counter/Variability;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_Variability) => void): void;
    "isVector()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getVectorLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "isInternal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getFlags()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_management_counter_Variability extends java_lang_Object implements java_io_Serializable {
    public static "sun/management/counter/Variability/NATTRIBUTES": number;
    public static "sun/management/counter/Variability/map": JVMArray<sun_management_counter_Variability>;
    public "sun/management/counter/Variability/name": java_lang_String;
    public "sun/management/counter/Variability/value": number;
    public static "sun/management/counter/Variability/INVALID": sun_management_counter_Variability;
    public static "sun/management/counter/Variability/CONSTANT": sun_management_counter_Variability;
    public static "sun/management/counter/Variability/MONOTONIC": sun_management_counter_Variability;
    public static "sun/management/counter/Variability/VARIABLE": sun_management_counter_Variability;
    public static "sun/management/counter/Variability/serialVersionUID": Long;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/counter/Variability/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/Variability/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toVariability(I)Lsun/management/counter/Variability;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_Variability) => void): void;
    public static "sun/management/counter/Variability/toVariability(I)Lsun/management/counter/Variability;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_Variability) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/counter/Variability/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/counter/Variability/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_counter_Units extends java_lang_Object implements java_io_Serializable {
    public static "sun/management/counter/Units/NUNITS": number;
    public static "sun/management/counter/Units/map": JVMArray<sun_management_counter_Units>;
    public "sun/management/counter/Units/name": java_lang_String;
    public "sun/management/counter/Units/value": number;
    public static "sun/management/counter/Units/INVALID": sun_management_counter_Units;
    public static "sun/management/counter/Units/NONE": sun_management_counter_Units;
    public static "sun/management/counter/Units/BYTES": sun_management_counter_Units;
    public static "sun/management/counter/Units/TICKS": sun_management_counter_Units;
    public static "sun/management/counter/Units/EVENTS": sun_management_counter_Units;
    public static "sun/management/counter/Units/STRING": sun_management_counter_Units;
    public static "sun/management/counter/Units/HERTZ": sun_management_counter_Units;
    public static "sun/management/counter/Units/serialVersionUID": Long;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/counter/Units/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/Units/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toUnits(I)Lsun/management/counter/Units;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_Units) => void): void;
    public static "sun/management/counter/Units/toUnits(I)Lsun/management/counter/Units;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_management_counter_Units) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/counter/Units/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/counter/Units/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_counter_perf_Prologue extends java_lang_Object {
    public static "sun/management/counter/perf/Prologue/PERFDATA_BIG_ENDIAN": number;
    public static "sun/management/counter/perf/Prologue/PERFDATA_LITTLE_ENDIAN": number;
    public static "sun/management/counter/perf/Prologue/PERFDATA_MAGIC": number;
    public "sun/management/counter/perf/Prologue/header": java_nio_ByteBuffer;
    public "sun/management/counter/perf/Prologue/magic": number;
    public "<init>(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/counter/perf/Prologue/<init>(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMagic()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/Prologue/getMagic()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/Prologue/getMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/Prologue/getMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getByteOrder()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "sun/management/counter/perf/Prologue/getByteOrder()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "getEntryOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/Prologue/getEntryOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUsed()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/Prologue/getUsed()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOverflow()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/Prologue/getOverflow()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getModificationTimeStamp()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/counter/perf/Prologue/getModificationTimeStamp()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getNumEntries()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/Prologue/getNumEntries()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAccessible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/counter/perf/Prologue/isAccessible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface sun_management_VMManagement extends java_lang_Object {
    "isCompilationTimeMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadContentionMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadContentionMonitoringEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isCurrentThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isOtherThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadCpuTimeEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isBootClassPathSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isObjectMonitorUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSynchronizerUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadAllocatedMemorySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadAllocatedMemoryEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isGcNotificationSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isRemoteDiagnosticCommandsSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTotalClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getLoadedClassCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getUnloadedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getVerboseClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getVerboseGC()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getManagementVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVmId()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVmName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVmVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVmVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVmSpecName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVmSpecVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVmSpecVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getClassPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getLibraryPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getBootClassPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVmArguments()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "getStartupTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getUptime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getAvailableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getCompilerName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getTotalCompileTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getTotalThreadCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getLiveThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getPeakThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getDaemonThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getOsName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getOsArch()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getOsVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getSafepointCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getTotalSafepointTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getSafepointSyncTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getTotalApplicationNonStoppedTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getLoadedClassSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getUnloadedClassSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getClassLoadingTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getMethodDataSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getInitializedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getClassInitializationTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getClassVerificationTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getInternalCounters(Ljava/lang/String;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
  }
  export class java_lang_management_ThreadInfo extends java_lang_Object {
    public "java/lang/management/ThreadInfo/threadName": java_lang_String;
    public "java/lang/management/ThreadInfo/threadId": Long;
    public "java/lang/management/ThreadInfo/blockedTime": Long;
    public "java/lang/management/ThreadInfo/blockedCount": Long;
    public "java/lang/management/ThreadInfo/waitedTime": Long;
    public "java/lang/management/ThreadInfo/waitedCount": Long;
    public "java/lang/management/ThreadInfo/lock": java_lang_management_LockInfo;
    public "java/lang/management/ThreadInfo/lockName": java_lang_String;
    public "java/lang/management/ThreadInfo/lockOwnerId": Long;
    public "java/lang/management/ThreadInfo/lockOwnerName": java_lang_String;
    public "java/lang/management/ThreadInfo/inNative": number;
    public "java/lang/management/ThreadInfo/suspended": number;
    public "java/lang/management/ThreadInfo/threadState": java_lang_Thread$State;
    public "java/lang/management/ThreadInfo/stackTrace": JVMArray<java_lang_StackTraceElement>;
    public "java/lang/management/ThreadInfo/lockedMonitors": JVMArray<java_lang_management_MonitorInfo>;
    public "java/lang/management/ThreadInfo/lockedSynchronizers": JVMArray<java_lang_management_LockInfo>;
    public static "java/lang/management/ThreadInfo/EMPTY_MONITORS": JVMArray<java_lang_management_MonitorInfo>;
    public static "java/lang/management/ThreadInfo/EMPTY_SYNCS": JVMArray<java_lang_management_LockInfo>;
    public static "java/lang/management/ThreadInfo/MAX_FRAMES": number;
    public static "java/lang/management/ThreadInfo/NO_STACK_TRACE": JVMArray<java_lang_StackTraceElement>;
    public static "java/lang/management/ThreadInfo/$assertionsDisabled": number;
    public "<init>(Ljava/lang/Thread;ILjava/lang/Object;Ljava/lang/Thread;JJJJ[Ljava/lang/StackTraceElement;)V"(thread: JVMThread, args: [java_lang_Thread, number, java_lang_Object, java_lang_Thread, Long, any, Long, any, Long, any, Long, any, JVMArray<java_lang_StackTraceElement>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/ThreadInfo/<init>(Ljava/lang/Thread;ILjava/lang/Object;Ljava/lang/Thread;JJJJ[Ljava/lang/StackTraceElement;)V"(thread: JVMThread, args: [java_lang_Thread, number, java_lang_Object, java_lang_Thread, Long, any, Long, any, Long, any, Long, any, JVMArray<java_lang_StackTraceElement>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Thread;ILjava/lang/Object;Ljava/lang/Thread;JJJJ[Ljava/lang/StackTraceElement;[Ljava/lang/Object;[I[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Thread, number, java_lang_Object, java_lang_Thread, Long, any, Long, any, Long, any, Long, any, JVMArray<java_lang_StackTraceElement>, JVMArray<java_lang_Object>, JVMArray<number>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/ThreadInfo/<init>(Ljava/lang/Thread;ILjava/lang/Object;Ljava/lang/Thread;JJJJ[Ljava/lang/StackTraceElement;[Ljava/lang/Object;[I[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Thread, number, java_lang_Object, java_lang_Thread, Long, any, Long, any, Long, any, Long, any, JVMArray<java_lang_StackTraceElement>, JVMArray<java_lang_Object>, JVMArray<number>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "initialize(Ljava/lang/Thread;ILjava/lang/Object;Ljava/lang/Thread;JJJJ[Ljava/lang/StackTraceElement;[Ljava/lang/management/MonitorInfo;[Ljava/lang/management/LockInfo;)V"(thread: JVMThread, args: [java_lang_Thread, number, java_lang_Object, java_lang_Thread, Long, any, Long, any, Long, any, Long, any, JVMArray<java_lang_StackTraceElement>, JVMArray<java_lang_management_MonitorInfo>, JVMArray<java_lang_management_LockInfo>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/ThreadInfo/initialize(Ljava/lang/Thread;ILjava/lang/Object;Ljava/lang/Thread;JJJJ[Ljava/lang/StackTraceElement;[Ljava/lang/management/MonitorInfo;[Ljava/lang/management/LockInfo;)V"(thread: JVMThread, args: [java_lang_Thread, number, java_lang_Object, java_lang_Thread, Long, any, Long, any, Long, any, Long, any, JVMArray<java_lang_StackTraceElement>, JVMArray<java_lang_management_MonitorInfo>, JVMArray<java_lang_management_LockInfo>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljavax/management/openmbean/CompositeData;)V"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/ThreadInfo/<init>(Ljavax/management/openmbean/CompositeData;)V"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable) => void): void;
    public "getThreadId()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/ThreadInfo/getThreadId()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getThreadName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/management/ThreadInfo/getThreadName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getThreadState()Ljava/lang/Thread$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$State) => void): void;
    public "java/lang/management/ThreadInfo/getThreadState()Ljava/lang/Thread$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$State) => void): void;
    public "getBlockedTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/ThreadInfo/getBlockedTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getBlockedCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/ThreadInfo/getBlockedCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getWaitedTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/ThreadInfo/getWaitedTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getWaitedCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/ThreadInfo/getWaitedCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLockInfo()Ljava/lang/management/LockInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_LockInfo) => void): void;
    public "java/lang/management/ThreadInfo/getLockInfo()Ljava/lang/management/LockInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_LockInfo) => void): void;
    public "getLockName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/management/ThreadInfo/getLockName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getLockOwnerId()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/ThreadInfo/getLockOwnerId()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLockOwnerName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/management/ThreadInfo/getLockOwnerName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getStackTrace()[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_StackTraceElement>) => void): void;
    public "java/lang/management/ThreadInfo/getStackTrace()[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_StackTraceElement>) => void): void;
    public "isSuspended()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/management/ThreadInfo/isSuspended()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInNative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/management/ThreadInfo/isInNative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/management/ThreadInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "from(Ljavax/management/openmbean/CompositeData;)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    public static "java/lang/management/ThreadInfo/from(Ljavax/management/openmbean/CompositeData;)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    public "getLockedMonitors()[Ljava/lang/management/MonitorInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MonitorInfo>) => void): void;
    public "java/lang/management/ThreadInfo/getLockedMonitors()[Ljava/lang/management/MonitorInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MonitorInfo>) => void): void;
    public "getLockedSynchronizers()[Ljava/lang/management/LockInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_LockInfo>) => void): void;
    public "java/lang/management/ThreadInfo/getLockedSynchronizers()[Ljava/lang/management/LockInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_LockInfo>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/management/ThreadInfo/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface javax_management_openmbean_CompositeData extends java_lang_Object {
    "getCompositeType()Ljavax/management/openmbean/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_CompositeType) => void): void;
    "get(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getAll([Ljava/lang/String;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "containsKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class javax_management_openmbean_CompositeType extends javax_management_openmbean_OpenType {
    public static "javax/management/openmbean/CompositeType/serialVersionUID": Long;
    public "javax/management/openmbean/CompositeType/nameToDescription": java_util_TreeMap;
    public "javax/management/openmbean/CompositeType/nameToType": java_util_TreeMap;
    public "javax/management/openmbean/CompositeType/myHashCode": java_lang_Integer;
    public "javax/management/openmbean/CompositeType/myToString": java_lang_String;
    public "javax/management/openmbean/CompositeType/myNamesSet": java_util_Set;
    public "<init>(Ljava/lang/String;Ljava/lang/String;[Ljava/lang/String;[Ljava/lang/String;[Ljavax/management/openmbean/OpenType;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<java_lang_String>, JVMArray<java_lang_String>, JVMArray<javax_management_openmbean_OpenType>], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/openmbean/CompositeType/<init>(Ljava/lang/String;Ljava/lang/String;[Ljava/lang/String;[Ljava/lang/String;[Ljavax/management/openmbean/OpenType;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<java_lang_String>, JVMArray<java_lang_String>, JVMArray<javax_management_openmbean_OpenType>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkForNullElement([Ljava/lang/Object;Ljava/lang/String;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/openmbean/CompositeType/checkForNullElement([Ljava/lang/Object;Ljava/lang/String;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkForEmptyString([Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/openmbean/CompositeType/checkForEmptyString([Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "containsKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/CompositeType/containsKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescription(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/openmbean/CompositeType/getDescription(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getType(Ljava/lang/String;)Ljavax/management/openmbean/OpenType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_OpenType) => void): void;
    public "javax/management/openmbean/CompositeType/getType(Ljava/lang/String;)Ljavax/management/openmbean/OpenType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_OpenType) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "javax/management/openmbean/CompositeType/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "isValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/CompositeType/isValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAssignableFrom(Ljavax/management/openmbean/OpenType;)Z"(thread: JVMThread, args: [javax_management_openmbean_OpenType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/CompositeType/isAssignableFrom(Ljavax/management/openmbean/OpenType;)Z"(thread: JVMThread, args: [javax_management_openmbean_OpenType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/CompositeType/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/CompositeType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/openmbean/CompositeType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_util_TreeMap extends java_util_AbstractMap implements java_util_NavigableMap, java_lang_Cloneable, java_io_Serializable {
    public "java/util/TreeMap/comparator": java_util_Comparator;
    public "java/util/TreeMap/root": java_util_TreeMap$Entry;
    public "java/util/TreeMap/size": number;
    public "java/util/TreeMap/modCount": number;
    public "java/util/TreeMap/entrySet": java_util_TreeMap$EntrySet;
    public "java/util/TreeMap/navigableKeySet": java_util_TreeMap$KeySet;
    public "java/util/TreeMap/descendingMap": java_util_NavigableMap;
    public static "java/util/TreeMap/UNBOUNDED": java_lang_Object;
    public static "java/util/TreeMap/RED": number;
    public static "java/util/TreeMap/BLACK": number;
    public static "java/util/TreeMap/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/<init>(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/SortedMap;)V"(thread: JVMThread, args: [java_util_SortedMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/<init>(Ljava/util/SortedMap;)V"(thread: JVMThread, args: [java_util_SortedMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap/containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public "java/util/TreeMap/comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public "firstKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/firstKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "lastKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/lastKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "java/util/TreeMap/getEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "getEntryUsingComparator(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "java/util/TreeMap/getEntryUsingComparator(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "getCeilingEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "java/util/TreeMap/getCeilingEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "getFloorEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "java/util/TreeMap/getFloorEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "getHigherEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "java/util/TreeMap/getHigherEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "getLowerEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "java/util/TreeMap/getLowerEntry(Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "firstEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "java/util/TreeMap/firstEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "lastEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "java/util/TreeMap/lastEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "pollFirstEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "java/util/TreeMap/pollFirstEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "pollLastEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "java/util/TreeMap/pollLastEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "lowerEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "java/util/TreeMap/lowerEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "lowerKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/lowerKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "floorEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "java/util/TreeMap/floorEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "floorKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/floorKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "ceilingEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "java/util/TreeMap/ceilingEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "ceilingKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/ceilingKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "higherEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "java/util/TreeMap/higherEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public "higherKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/higherKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/TreeMap/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "navigableKeySet()Ljava/util/NavigableSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "java/util/TreeMap/navigableKeySet()Ljava/util/NavigableSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "descendingKeySet()Ljava/util/NavigableSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "java/util/TreeMap/descendingKeySet()Ljava/util/NavigableSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/TreeMap/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/TreeMap/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "descendingMap()Ljava/util/NavigableMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    public "java/util/TreeMap/descendingMap()Ljava/util/NavigableMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    public "subMap(Ljava/lang/Object;ZLjava/lang/Object;Z)Ljava/util/NavigableMap;"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    public "java/util/TreeMap/subMap(Ljava/lang/Object;ZLjava/lang/Object;Z)Ljava/util/NavigableMap;"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    public "headMap(Ljava/lang/Object;Z)Ljava/util/NavigableMap;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    public "java/util/TreeMap/headMap(Ljava/lang/Object;Z)Ljava/util/NavigableMap;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    public "tailMap(Ljava/lang/Object;Z)Ljava/util/NavigableMap;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    public "java/util/TreeMap/tailMap(Ljava/lang/Object;Z)Ljava/util/NavigableMap;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    public "subMap(Ljava/lang/Object;Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public "java/util/TreeMap/subMap(Ljava/lang/Object;Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public "headMap(Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public "java/util/TreeMap/headMap(Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public "tailMap(Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public "java/util/TreeMap/tailMap(Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap/replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap/replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "keyIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/TreeMap/keyIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "descendingKeyIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/TreeMap/descendingKeyIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "compare(Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap/compare(Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "valEquals(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/TreeMap/valEquals(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "exportEntry(Ljava/util/TreeMap$Entry;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public static "java/util/TreeMap/exportEntry(Ljava/util/TreeMap$Entry;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    public static "keyOrNull(Ljava/util/TreeMap$Entry;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/util/TreeMap/keyOrNull(Ljava/util/TreeMap$Entry;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "key(Ljava/util/TreeMap$Entry;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/util/TreeMap/key(Ljava/util/TreeMap$Entry;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getFirstEntry()Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "java/util/TreeMap/getFirstEntry()Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "getLastEntry()Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "java/util/TreeMap/getLastEntry()Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "successor(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "java/util/TreeMap/successor(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "predecessor(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "java/util/TreeMap/predecessor(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "colorOf(Ljava/util/TreeMap$Entry;)Z"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/TreeMap/colorOf(Ljava/util/TreeMap$Entry;)Z"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "parentOf(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "java/util/TreeMap/parentOf(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "setColor(Ljava/util/TreeMap$Entry;Z)V"(thread: JVMThread, args: [java_util_TreeMap$Entry, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/TreeMap/setColor(Ljava/util/TreeMap$Entry;Z)V"(thread: JVMThread, args: [java_util_TreeMap$Entry, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "leftOf(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "java/util/TreeMap/leftOf(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "rightOf(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "java/util/TreeMap/rightOf(Ljava/util/TreeMap$Entry;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "rotateLeft(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/rotateLeft(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotateRight(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/rotateRight(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "fixAfterInsertion(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/fixAfterInsertion(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "deleteEntry(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/deleteEntry(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "fixAfterDeletion(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/fixAfterDeletion(Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readTreeSet(ILjava/io/ObjectInputStream;Ljava/lang/Object;)V"(thread: JVMThread, args: [number, java_io_ObjectInputStream, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/readTreeSet(ILjava/io/ObjectInputStream;Ljava/lang/Object;)V"(thread: JVMThread, args: [number, java_io_ObjectInputStream, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAllForTreeSet(Ljava/util/SortedSet;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_SortedSet, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/addAllForTreeSet(Ljava/util/SortedSet;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_SortedSet, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "buildFromSorted(ILjava/util/Iterator;Ljava/io/ObjectInputStream;Ljava/lang/Object;)V"(thread: JVMThread, args: [number, java_util_Iterator, java_io_ObjectInputStream, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap/buildFromSorted(ILjava/util/Iterator;Ljava/io/ObjectInputStream;Ljava/lang/Object;)V"(thread: JVMThread, args: [number, java_util_Iterator, java_io_ObjectInputStream, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "buildFromSorted(IIIILjava/util/Iterator;Ljava/io/ObjectInputStream;Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [number, number, number, number, java_util_Iterator, java_io_ObjectInputStream, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public "java/util/TreeMap/buildFromSorted(IIIILjava/util/Iterator;Ljava/io/ObjectInputStream;Ljava/lang/Object;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [number, number, number, number, java_util_Iterator, java_io_ObjectInputStream, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "computeRedLevel(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/TreeMap/computeRedLevel(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "keySpliteratorFor(Ljava/util/NavigableMap;)Ljava/util/Spliterator;"(thread: JVMThread, args: [java_util_NavigableMap], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public static "java/util/TreeMap/keySpliteratorFor(Ljava/util/NavigableMap;)Ljava/util/Spliterator;"(thread: JVMThread, args: [java_util_NavigableMap], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "keySpliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/TreeMap/keySpliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "descendingKeySpliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/TreeMap/descendingKeySpliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public static "access$000(Ljava/util/TreeMap;Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap, java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/TreeMap/access$000(Ljava/util/TreeMap;Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_util_TreeMap, java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$100(Ljava/util/TreeMap;)I"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/TreeMap/access$100(Ljava/util/TreeMap;)I"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/util/TreeMap/access$200()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$300(Ljava/util/TreeMap;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public static "java/util/TreeMap/access$300(Ljava/util/TreeMap;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public static "access$400(Ljava/util/TreeMap;)I"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/TreeMap/access$400(Ljava/util/TreeMap;)I"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500(Ljava/util/TreeMap;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "java/util/TreeMap/access$500(Ljava/util/TreeMap;)Ljava/util/TreeMap$Entry;"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap$Entry) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/TreeMap/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_NavigableSet extends java_lang_Object, java_util_SortedSet {
    "lower(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "floor(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "ceiling(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "higher(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "pollFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "pollLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "descendingSet()Ljava/util/NavigableSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    "descendingIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    "subSet(Ljava/lang/Object;ZLjava/lang/Object;Z)Ljava/util/NavigableSet;"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    "headSet(Ljava/lang/Object;Z)Ljava/util/NavigableSet;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    "tailSet(Ljava/lang/Object;Z)Ljava/util/NavigableSet;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    "subSet(Ljava/lang/Object;Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    "headSet(Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    "tailSet(Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    "comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "first()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "last()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_TreeMap$KeySet extends java_util_AbstractSet implements java_util_NavigableSet {
    public "java/util/TreeMap$KeySet/m": java_util_NavigableMap;
    public "<init>(Ljava/util/NavigableMap;)V"(thread: JVMThread, args: [java_util_NavigableMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap$KeySet/<init>(Ljava/util/NavigableMap;)V"(thread: JVMThread, args: [java_util_NavigableMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/TreeMap$KeySet/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "descendingIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/TreeMap$KeySet/descendingIterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap$KeySet/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap$KeySet/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap$KeySet/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap$KeySet/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lower(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$KeySet/lower(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "floor(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$KeySet/floor(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "ceiling(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$KeySet/ceiling(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "higher(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$KeySet/higher(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "first()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$KeySet/first()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "last()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$KeySet/last()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public "java/util/TreeMap$KeySet/comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public "pollFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$KeySet/pollFirst()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "pollLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$KeySet/pollLast()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap$KeySet/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "subSet(Ljava/lang/Object;ZLjava/lang/Object;Z)Ljava/util/NavigableSet;"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "java/util/TreeMap$KeySet/subSet(Ljava/lang/Object;ZLjava/lang/Object;Z)Ljava/util/NavigableSet;"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "headSet(Ljava/lang/Object;Z)Ljava/util/NavigableSet;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "java/util/TreeMap$KeySet/headSet(Ljava/lang/Object;Z)Ljava/util/NavigableSet;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "tailSet(Ljava/lang/Object;Z)Ljava/util/NavigableSet;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "java/util/TreeMap$KeySet/tailSet(Ljava/lang/Object;Z)Ljava/util/NavigableSet;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "subSet(Ljava/lang/Object;Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    public "java/util/TreeMap$KeySet/subSet(Ljava/lang/Object;Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    public "headSet(Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    public "java/util/TreeMap$KeySet/headSet(Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    public "tailSet(Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    public "java/util/TreeMap$KeySet/tailSet(Ljava/lang/Object;)Ljava/util/SortedSet;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedSet) => void): void;
    public "descendingSet()Ljava/util/NavigableSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "java/util/TreeMap$KeySet/descendingSet()Ljava/util/NavigableSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/TreeMap$KeySet/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_TreeMap$EntrySet extends java_util_AbstractSet {
    public "java/util/TreeMap$EntrySet/this$0": java_util_TreeMap;
    public "<init>(Ljava/util/TreeMap;)V"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap$EntrySet/<init>(Ljava/util/TreeMap;)V"(thread: JVMThread, args: [java_util_TreeMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/TreeMap$EntrySet/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap$EntrySet/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap$EntrySet/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap$EntrySet/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap$EntrySet/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/TreeMap$EntrySet/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
  }
  export class java_util_TreeMap$Entry extends java_lang_Object implements java_util_Map$Entry {
    public "java/util/TreeMap$Entry/key": java_lang_Object;
    public "java/util/TreeMap$Entry/value": java_lang_Object;
    public "java/util/TreeMap$Entry/left": java_util_TreeMap$Entry;
    public "java/util/TreeMap$Entry/right": java_util_TreeMap$Entry;
    public "java/util/TreeMap$Entry/parent": java_util_TreeMap$Entry;
    public "java/util/TreeMap$Entry/color": number;
    public "<init>(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TreeMap$Entry/<init>(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/TreeMap$Entry;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_TreeMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$Entry/getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$Entry/getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TreeMap$Entry/setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap$Entry/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TreeMap$Entry/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/TreeMap$Entry/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_util_NavigableMap extends java_lang_Object, java_util_SortedMap {
    "lowerEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    "lowerKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "floorEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    "floorKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "ceilingEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    "ceilingKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "higherEntry(Ljava/lang/Object;)Ljava/util/Map$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    "higherKey(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "firstEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    "lastEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    "pollFirstEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    "pollLastEntry()Ljava/util/Map$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map$Entry) => void): void;
    "descendingMap()Ljava/util/NavigableMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    "navigableKeySet()Ljava/util/NavigableSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    "descendingKeySet()Ljava/util/NavigableSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableSet) => void): void;
    "subMap(Ljava/lang/Object;ZLjava/lang/Object;Z)Ljava/util/NavigableMap;"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    "headMap(Ljava/lang/Object;Z)Ljava/util/NavigableMap;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    "tailMap(Ljava/lang/Object;Z)Ljava/util/NavigableMap;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_util_NavigableMap) => void): void;
    "subMap(Ljava/lang/Object;Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    "headMap(Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    "tailMap(Ljava/lang/Object;)Ljava/util/SortedMap;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    "comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "firstKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "lastKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class javax_management_openmbean_OpenType extends java_lang_Object implements java_io_Serializable {
    public static "javax/management/openmbean/OpenType/serialVersionUID": Long;
    public static "javax/management/openmbean/OpenType/ALLOWED_CLASSNAMES_LIST": java_util_List;
    public static "javax/management/openmbean/OpenType/ALLOWED_CLASSNAMES": JVMArray<java_lang_String>;
    public "javax/management/openmbean/OpenType/className": java_lang_String;
    public "javax/management/openmbean/OpenType/description": java_lang_String;
    public "javax/management/openmbean/OpenType/typeName": java_lang_String;
    public "javax/management/openmbean/OpenType/isArray": number;
    public "javax/management/openmbean/OpenType/descriptor": javax_management_Descriptor;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/openmbean/OpenType/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/openmbean/OpenType/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkClassNameOverride()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/openmbean/OpenType/checkClassNameOverride()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "overridesGetClassName(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "javax/management/openmbean/OpenType/overridesGetClassName(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "validClassName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "javax/management/openmbean/OpenType/validClassName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valid(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "javax/management/openmbean/OpenType/valid(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDescriptor()Ljavax/management/Descriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_Descriptor) => void): void;
    public "javax/management/openmbean/OpenType/getDescriptor()Ljavax/management/Descriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_Descriptor) => void): void;
    public "getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/openmbean/OpenType/getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "safeGetClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/openmbean/OpenType/safeGetClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTypeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/openmbean/OpenType/getTypeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/openmbean/OpenType/getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/OpenType/isArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/OpenType/isValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAssignableFrom(Ljavax/management/openmbean/OpenType;)Z"(thread: JVMThread, args: [javax_management_openmbean_OpenType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/OpenType/isAssignableFrom(Ljavax/management/openmbean/OpenType;)Z"(thread: JVMThread, args: [javax_management_openmbean_OpenType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/OpenType/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/openmbean/OpenType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/openmbean/OpenType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/openmbean/OpenType/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/openmbean/OpenType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface javax_management_Descriptor extends java_lang_Object, java_io_Serializable, java_lang_Cloneable {
    "getFieldValue(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "setField(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "getFields()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "getFieldNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "getFieldValues([Ljava/lang/String;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "removeField(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setFields([Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_lang_management_MonitorInfo extends java_lang_management_LockInfo {
    public "java/lang/management/MonitorInfo/stackDepth": number;
    public "java/lang/management/MonitorInfo/stackFrame": java_lang_StackTraceElement;
    public "<init>(Ljava/lang/String;IILjava/lang/StackTraceElement;)V"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_StackTraceElement], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/MonitorInfo/<init>(Ljava/lang/String;IILjava/lang/StackTraceElement;)V"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_StackTraceElement], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLockedStackDepth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/management/MonitorInfo/getLockedStackDepth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLockedStackFrame()Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_StackTraceElement) => void): void;
    public "java/lang/management/MonitorInfo/getLockedStackFrame()Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_StackTraceElement) => void): void;
    public static "from(Ljavax/management/openmbean/CompositeData;)Ljava/lang/management/MonitorInfo;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MonitorInfo) => void): void;
    public static "java/lang/management/MonitorInfo/from(Ljavax/management/openmbean/CompositeData;)Ljava/lang/management/MonitorInfo;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MonitorInfo) => void): void;
  }
  export class java_lang_management_LockInfo extends java_lang_Object {
    public "java/lang/management/LockInfo/className": java_lang_String;
    public "java/lang/management/LockInfo/identityHashCode": number;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/LockInfo/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/LockInfo/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/management/LockInfo/getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getIdentityHashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/management/LockInfo/getIdentityHashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "from(Ljavax/management/openmbean/CompositeData;)Ljava/lang/management/LockInfo;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_LockInfo) => void): void;
    public static "java/lang/management/LockInfo/from(Ljavax/management/openmbean/CompositeData;)Ljava/lang/management/LockInfo;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_LockInfo) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/management/LockInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_management_ThreadImpl extends java_lang_Object implements com_sun_management_ThreadMXBean {
    public "sun/management/ThreadImpl/jvm": sun_management_VMManagement;
    public "sun/management/ThreadImpl/contentionMonitoringEnabled": number;
    public "sun/management/ThreadImpl/cpuTimeEnabled": number;
    public "sun/management/ThreadImpl/allocatedMemoryEnabled": number;
    public "<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/ThreadImpl/<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "getThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/getThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPeakThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/getPeakThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTotalStartedThreadCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/ThreadImpl/getTotalStartedThreadCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getDaemonThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/getDaemonThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadContentionMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/isThreadContentionMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadContentionMonitoringEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/isThreadContentionMonitoringEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/isThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isCurrentThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/isCurrentThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadAllocatedMemorySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/isThreadAllocatedMemorySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadCpuTimeEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/isThreadCpuTimeEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isThreadAllocatedMemoryEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/isThreadAllocatedMemoryEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAllThreadIds()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "sun/management/ThreadImpl/getAllThreadIds()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "getThreadInfo(J)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    public "sun/management/ThreadImpl/getThreadInfo(J)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    public "getThreadInfo(JI)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    public "sun/management/ThreadImpl/getThreadInfo(JI)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    public "getThreadInfo([J)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public "sun/management/ThreadImpl/getThreadInfo([J)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public "verifyThreadIds([J)V"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/ThreadImpl/verifyThreadIds([J)V"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getThreadInfo([JI)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public "sun/management/ThreadImpl/getThreadInfo([JI)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public "setThreadContentionMonitoringEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/ThreadImpl/setThreadContentionMonitoringEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "verifyCurrentThreadCpuTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/verifyCurrentThreadCpuTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCurrentThreadCpuTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/ThreadImpl/getCurrentThreadCpuTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getThreadCpuTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/ThreadImpl/getThreadCpuTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "verifyThreadCpuTime([J)Z"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/verifyThreadCpuTime([J)Z"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getThreadCpuTime([J)[J"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "sun/management/ThreadImpl/getThreadCpuTime([J)[J"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "getCurrentThreadUserTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/ThreadImpl/getCurrentThreadUserTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getThreadUserTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/ThreadImpl/getThreadUserTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getThreadUserTime([J)[J"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "sun/management/ThreadImpl/getThreadUserTime([J)[J"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "setThreadCpuTimeEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/ThreadImpl/setThreadCpuTimeEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getThreadAllocatedBytes(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/ThreadImpl/getThreadAllocatedBytes(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "verifyThreadAllocatedMemory([J)Z"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/verifyThreadAllocatedMemory([J)Z"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getThreadAllocatedBytes([J)[J"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "sun/management/ThreadImpl/getThreadAllocatedBytes([J)[J"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "setThreadAllocatedMemoryEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/ThreadImpl/setThreadAllocatedMemoryEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "findMonitorDeadlockedThreads()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "sun/management/ThreadImpl/findMonitorDeadlockedThreads()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "findDeadlockedThreads()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "sun/management/ThreadImpl/findDeadlockedThreads()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "resetPeakThreadCount()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/ThreadImpl/resetPeakThreadCount()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isObjectMonitorUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/isObjectMonitorUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSynchronizerUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ThreadImpl/isSynchronizerUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "verifyDumpThreads(ZZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/ThreadImpl/verifyDumpThreads(ZZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getThreadInfo([JZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public "sun/management/ThreadImpl/getThreadInfo([JZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public "dumpAllThreads(ZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public "sun/management/ThreadImpl/dumpAllThreads(ZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public static "getThreads()[Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread>) => void): void;
    public static "sun/management/ThreadImpl/getThreads()[Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread>) => void): void;
    public static "getThreadInfo1([JI[Ljava/lang/management/ThreadInfo;)V"(thread: JVMThread, args: [JVMArray<Long>, number, JVMArray<java_lang_management_ThreadInfo>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ThreadImpl/getThreadInfo1([JI[Ljava/lang/management/ThreadInfo;)V"(thread: JVMThread, args: [JVMArray<Long>, number, JVMArray<java_lang_management_ThreadInfo>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getThreadTotalCpuTime0(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/management/ThreadImpl/getThreadTotalCpuTime0(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getThreadTotalCpuTime1([J[J)V"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ThreadImpl/getThreadTotalCpuTime1([J[J)V"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getThreadUserCpuTime0(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/management/ThreadImpl/getThreadUserCpuTime0(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getThreadUserCpuTime1([J[J)V"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ThreadImpl/getThreadUserCpuTime1([J[J)V"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getThreadAllocatedMemory1([J[J)V"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ThreadImpl/getThreadAllocatedMemory1([J[J)V"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setThreadCpuTimeEnabled0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ThreadImpl/setThreadCpuTimeEnabled0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setThreadAllocatedMemoryEnabled0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ThreadImpl/setThreadAllocatedMemoryEnabled0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setThreadContentionMonitoringEnabled0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ThreadImpl/setThreadContentionMonitoringEnabled0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "findMonitorDeadlockedThreads0()[Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread>) => void): void;
    public static "sun/management/ThreadImpl/findMonitorDeadlockedThreads0()[Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread>) => void): void;
    public static "findDeadlockedThreads0()[Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread>) => void): void;
    public static "sun/management/ThreadImpl/findDeadlockedThreads0()[Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread>) => void): void;
    public static "resetPeakThreadCount0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ThreadImpl/resetPeakThreadCount0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "dumpThreads0([JZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public static "sun/management/ThreadImpl/dumpThreads0([JZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    public static "resetContentionTimes0(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ThreadImpl/resetContentionTimes0(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "sun/management/ThreadImpl/getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class javax_management_ObjectName extends java_lang_Object implements java_lang_Comparable, javax_management_QueryExp {
    public static "javax/management/ObjectName/oldSerialVersionUID": Long;
    public static "javax/management/ObjectName/newSerialVersionUID": Long;
    public static "javax/management/ObjectName/oldSerialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "javax/management/ObjectName/newSerialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "javax/management/ObjectName/serialVersionUID": Long;
    public static "javax/management/ObjectName/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "javax/management/ObjectName/compat": number;
    public static "javax/management/ObjectName/_Empty_property_array": JVMArray<javax_management_ObjectName$Property>;
    public "javax/management/ObjectName/_canonicalName": java_lang_String;
    public "javax/management/ObjectName/_kp_array": JVMArray<javax_management_ObjectName$Property>;
    public "javax/management/ObjectName/_ca_array": JVMArray<javax_management_ObjectName$Property>;
    public "javax/management/ObjectName/_domain_length": number;
    public "javax/management/ObjectName/_propertyList": java_util_Map;
    public "javax/management/ObjectName/_domain_pattern": number;
    public "javax/management/ObjectName/_property_list_pattern": number;
    public "javax/management/ObjectName/_property_value_pattern": number;
    public static "javax/management/ObjectName/WILDCARD": javax_management_ObjectName;
    public "construct(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/construct(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "construct(Ljava/lang/String;Ljava/util/Map;)V"(thread: JVMThread, args: [java_lang_String, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/construct(Ljava/lang/String;Ljava/util/Map;)V"(thread: JVMThread, args: [java_lang_String, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "addProperty(Ljavax/management/ObjectName$Property;ILjava/util/Map;Ljava/lang/String;)V"(thread: JVMThread, args: [javax_management_ObjectName$Property, number, java_util_Map, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/addProperty(Ljavax/management/ObjectName$Property;ILjava/util/Map;Ljava/lang/String;)V"(thread: JVMThread, args: [javax_management_ObjectName$Property, number, java_util_Map, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCanonicalName([C[C[Ljava/lang/String;Ljava/util/Map;II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, JVMArray<java_lang_String>, java_util_Map, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/setCanonicalName([C[C[Ljava/lang/String;Ljava/util/Map;II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, JVMArray<java_lang_String>, java_util_Map, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "parseKey([CI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "javax/management/ObjectName/parseKey([CI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "parseValue([CI)[I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "javax/management/ObjectName/parseValue([CI)[I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "checkValue(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "javax/management/ObjectName/checkValue(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkKey(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/ObjectName/checkKey(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isDomain(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/isDomain(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance(Ljava/lang/String;)Ljavax/management/ObjectName;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "javax/management/ObjectName/getInstance(Ljava/lang/String;)Ljavax/management/ObjectName;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljavax/management/ObjectName;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "javax/management/ObjectName/getInstance(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljavax/management/ObjectName;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/util/Hashtable;)Ljavax/management/ObjectName;"(thread: JVMThread, args: [java_lang_String, java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "javax/management/ObjectName/getInstance(Ljava/lang/String;Ljava/util/Hashtable;)Ljavax/management/ObjectName;"(thread: JVMThread, args: [java_lang_String, java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "getInstance(Ljavax/management/ObjectName;)Ljavax/management/ObjectName;"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "javax/management/ObjectName/getInstance(Ljavax/management/ObjectName;)Ljavax/management/ObjectName;"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/util/Hashtable;)V"(thread: JVMThread, args: [java_lang_String, java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/<init>(Ljava/lang/String;Ljava/util/Hashtable;)V"(thread: JVMThread, args: [java_lang_String, java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public "isPattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/isPattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDomainPattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/isDomainPattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPropertyPattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/isPropertyPattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPropertyListPattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/isPropertyListPattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPropertyValuePattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/isPropertyValuePattern()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPropertyValuePattern(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/isPropertyValuePattern(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCanonicalName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectName/getCanonicalName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDomain()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectName/getDomain()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getKeyProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectName/getKeyProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "_getKeyPropertyList()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "javax/management/ObjectName/_getKeyPropertyList()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getKeyPropertyList()Ljava/util/Hashtable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Hashtable) => void): void;
    public "javax/management/ObjectName/getKeyPropertyList()Ljava/util/Hashtable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Hashtable) => void): void;
    public "getKeyPropertyListString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectName/getKeyPropertyListString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSerializedNameString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectName/getSerializedNameString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeKeyPropertyListString([C[CI)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/writeKeyPropertyListString([C[CI)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCanonicalKeyPropertyListString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectName/getCanonicalKeyPropertyListString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectName/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "quote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "javax/management/ObjectName/quote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "unquote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "javax/management/ObjectName/unquote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "apply(Ljavax/management/ObjectName;)Z"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/apply(Ljavax/management/ObjectName;)Z"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "matchDomains(Ljavax/management/ObjectName;)Z"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/matchDomains(Ljavax/management/ObjectName;)Z"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "matchKeys(Ljavax/management/ObjectName;)Z"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/matchKeys(Ljavax/management/ObjectName;)Z"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMBeanServer(Ljavax/management/MBeanServer;)V"(thread: JVMThread, args: [javax_management_MBeanServer], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName/setMBeanServer(Ljavax/management/MBeanServer;)V"(thread: JVMThread, args: [javax_management_MBeanServer], cb?: (e?: java_lang_Throwable) => void): void;
    public "compareTo(Ljavax/management/ObjectName;)I"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/compareTo(Ljavax/management/ObjectName;)I"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectName/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/ObjectName/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface javax_management_MBeanServer extends java_lang_Object, javax_management_MBeanServerConnection {
    "createMBean(Ljava/lang/String;Ljavax/management/ObjectName;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "createMBean(Ljava/lang/String;Ljavax/management/ObjectName;Ljavax/management/ObjectName;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName, javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "createMBean(Ljava/lang/String;Ljavax/management/ObjectName;[Ljava/lang/Object;[Ljava/lang/String;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "createMBean(Ljava/lang/String;Ljavax/management/ObjectName;Ljavax/management/ObjectName;[Ljava/lang/Object;[Ljava/lang/String;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName, javax_management_ObjectName, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "registerMBean(Ljava/lang/Object;Ljavax/management/ObjectName;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [java_lang_Object, javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "unregisterMBean(Ljavax/management/ObjectName;)V"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable) => void): void;
    "getObjectInstance(Ljavax/management/ObjectName;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "queryMBeans(Ljavax/management/ObjectName;Ljavax/management/QueryExp;)Ljava/util/Set;"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_QueryExp], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "queryNames(Ljavax/management/ObjectName;Ljavax/management/QueryExp;)Ljava/util/Set;"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_QueryExp], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "isRegistered(Ljavax/management/ObjectName;)Z"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMBeanCount()Ljava/lang/Integer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    "getAttribute(Ljavax/management/ObjectName;Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_management_ObjectName, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getAttributes(Ljavax/management/ObjectName;[Ljava/lang/String;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [javax_management_ObjectName, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    "setAttribute(Ljavax/management/ObjectName;Ljavax/management/Attribute;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    "setAttributes(Ljavax/management/ObjectName;Ljavax/management/AttributeList;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    "invoke(Ljavax/management/ObjectName;Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_management_ObjectName, java_lang_String, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getDefaultDomain()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getDomains()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "addNotificationListener(Ljavax/management/ObjectName;Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "addNotificationListener(Ljavax/management/ObjectName;Ljavax/management/ObjectName;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_ObjectName, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/ObjectName;Ljavax/management/ObjectName;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_ObjectName], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/ObjectName;Ljavax/management/ObjectName;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_ObjectName, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/ObjectName;Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/ObjectName;Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "getMBeanInfo(Ljavax/management/ObjectName;)Ljavax/management/MBeanInfo;"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_MBeanInfo) => void): void;
    "isInstanceOf(Ljavax/management/ObjectName;Ljava/lang/String;)Z"(thread: JVMThread, args: [javax_management_ObjectName, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "instantiate(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "instantiate(Ljava/lang/String;Ljavax/management/ObjectName;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "instantiate(Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "instantiate(Ljava/lang/String;Ljavax/management/ObjectName;[Ljava/lang/Object;[Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "deserialize(Ljavax/management/ObjectName;[B)Ljava/io/ObjectInputStream;"(thread: JVMThread, args: [javax_management_ObjectName, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream) => void): void;
    "deserialize(Ljava/lang/String;[B)Ljava/io/ObjectInputStream;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream) => void): void;
    "deserialize(Ljava/lang/String;Ljavax/management/ObjectName;[B)Ljava/io/ObjectInputStream;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream) => void): void;
    "getClassLoaderFor(Ljavax/management/ObjectName;)Ljava/lang/ClassLoader;"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    "getClassLoader(Ljavax/management/ObjectName;)Ljava/lang/ClassLoader;"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    "getClassLoaderRepository()Ljavax/management/loading/ClassLoaderRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_loading_ClassLoaderRepository) => void): void;
  }
  export interface javax_management_loading_ClassLoaderRepository extends java_lang_Object {
    "loadClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    "loadClassWithout(Ljava/lang/ClassLoader;Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    "loadClassBefore(Ljava/lang/ClassLoader;Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
  }
  export class javax_management_MBeanInfo extends java_lang_Object implements java_lang_Cloneable, java_io_Serializable, javax_management_DescriptorRead {
    public static "javax/management/MBeanInfo/serialVersionUID": Long;
    public "javax/management/MBeanInfo/descriptor": javax_management_Descriptor;
    public "javax/management/MBeanInfo/description": java_lang_String;
    public "javax/management/MBeanInfo/className": java_lang_String;
    public "javax/management/MBeanInfo/attributes": JVMArray<javax_management_MBeanAttributeInfo>;
    public "javax/management/MBeanInfo/operations": JVMArray<javax_management_MBeanOperationInfo>;
    public "javax/management/MBeanInfo/constructors": JVMArray<javax_management_MBeanConstructorInfo>;
    public "javax/management/MBeanInfo/notifications": JVMArray<javax_management_MBeanNotificationInfo>;
    public "javax/management/MBeanInfo/hashCode": number;
    public "javax/management/MBeanInfo/arrayGettersSafe": number;
    public static "javax/management/MBeanInfo/arrayGettersSafeMap": java_util_Map;
    public "<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanAttributeInfo;[Ljavax/management/MBeanConstructorInfo;[Ljavax/management/MBeanOperationInfo;[Ljavax/management/MBeanNotificationInfo;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanAttributeInfo>, JVMArray<javax_management_MBeanConstructorInfo>, JVMArray<javax_management_MBeanOperationInfo>, JVMArray<javax_management_MBeanNotificationInfo>], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanInfo/<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanAttributeInfo;[Ljavax/management/MBeanConstructorInfo;[Ljavax/management/MBeanOperationInfo;[Ljavax/management/MBeanNotificationInfo;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanAttributeInfo>, JVMArray<javax_management_MBeanConstructorInfo>, JVMArray<javax_management_MBeanOperationInfo>, JVMArray<javax_management_MBeanNotificationInfo>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanAttributeInfo;[Ljavax/management/MBeanConstructorInfo;[Ljavax/management/MBeanOperationInfo;[Ljavax/management/MBeanNotificationInfo;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanAttributeInfo>, JVMArray<javax_management_MBeanConstructorInfo>, JVMArray<javax_management_MBeanOperationInfo>, JVMArray<javax_management_MBeanNotificationInfo>, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanInfo/<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanAttributeInfo;[Ljavax/management/MBeanConstructorInfo;[Ljavax/management/MBeanOperationInfo;[Ljavax/management/MBeanNotificationInfo;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanAttributeInfo>, JVMArray<javax_management_MBeanConstructorInfo>, JVMArray<javax_management_MBeanOperationInfo>, JVMArray<javax_management_MBeanNotificationInfo>, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/management/MBeanInfo/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanInfo/getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanInfo/getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAttributes()[Ljavax/management/MBeanAttributeInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanAttributeInfo>) => void): void;
    public "javax/management/MBeanInfo/getAttributes()[Ljavax/management/MBeanAttributeInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanAttributeInfo>) => void): void;
    public "fastGetAttributes()[Ljavax/management/MBeanAttributeInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanAttributeInfo>) => void): void;
    public "javax/management/MBeanInfo/fastGetAttributes()[Ljavax/management/MBeanAttributeInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanAttributeInfo>) => void): void;
    public "nonNullAttributes()[Ljavax/management/MBeanAttributeInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanAttributeInfo>) => void): void;
    public "javax/management/MBeanInfo/nonNullAttributes()[Ljavax/management/MBeanAttributeInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanAttributeInfo>) => void): void;
    public "getOperations()[Ljavax/management/MBeanOperationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanOperationInfo>) => void): void;
    public "javax/management/MBeanInfo/getOperations()[Ljavax/management/MBeanOperationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanOperationInfo>) => void): void;
    public "fastGetOperations()[Ljavax/management/MBeanOperationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanOperationInfo>) => void): void;
    public "javax/management/MBeanInfo/fastGetOperations()[Ljavax/management/MBeanOperationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanOperationInfo>) => void): void;
    public "nonNullOperations()[Ljavax/management/MBeanOperationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanOperationInfo>) => void): void;
    public "javax/management/MBeanInfo/nonNullOperations()[Ljavax/management/MBeanOperationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanOperationInfo>) => void): void;
    public "getConstructors()[Ljavax/management/MBeanConstructorInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanConstructorInfo>) => void): void;
    public "javax/management/MBeanInfo/getConstructors()[Ljavax/management/MBeanConstructorInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanConstructorInfo>) => void): void;
    public "fastGetConstructors()[Ljavax/management/MBeanConstructorInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanConstructorInfo>) => void): void;
    public "javax/management/MBeanInfo/fastGetConstructors()[Ljavax/management/MBeanConstructorInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanConstructorInfo>) => void): void;
    public "nonNullConstructors()[Ljavax/management/MBeanConstructorInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanConstructorInfo>) => void): void;
    public "javax/management/MBeanInfo/nonNullConstructors()[Ljavax/management/MBeanConstructorInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanConstructorInfo>) => void): void;
    public "getNotifications()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "javax/management/MBeanInfo/getNotifications()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "fastGetNotifications()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "javax/management/MBeanInfo/fastGetNotifications()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "nonNullNotifications()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "javax/management/MBeanInfo/nonNullNotifications()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "getDescriptor()Ljavax/management/Descriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_Descriptor) => void): void;
    public "javax/management/MBeanInfo/getDescriptor()Ljavax/management/Descriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_Descriptor) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanInfo/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanInfo/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "arrayGettersSafe(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "javax/management/MBeanInfo/arrayGettersSafe(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isEqual(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "javax/management/MBeanInfo/isEqual(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanInfo/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanInfo/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/MBeanInfo/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_management_MBeanNotificationInfo extends javax_management_MBeanFeatureInfo implements java_lang_Cloneable {
    public static "javax/management/MBeanNotificationInfo/serialVersionUID": Long;
    public static "javax/management/MBeanNotificationInfo/NO_TYPES": JVMArray<java_lang_String>;
    public static "javax/management/MBeanNotificationInfo/NO_NOTIFICATIONS": JVMArray<javax_management_MBeanNotificationInfo>;
    public "javax/management/MBeanNotificationInfo/types": JVMArray<java_lang_String>;
    public "javax/management/MBeanNotificationInfo/arrayGettersSafe": number;
    public "<init>([Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanNotificationInfo/<init>([Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String, java_lang_String, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanNotificationInfo/<init>([Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String, java_lang_String, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/management/MBeanNotificationInfo/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getNotifTypes()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "javax/management/MBeanNotificationInfo/getNotifTypes()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "fastGetNotifTypes()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "javax/management/MBeanNotificationInfo/fastGetNotifTypes()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanNotificationInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanNotificationInfo/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanNotificationInfo/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanNotificationInfo/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/MBeanNotificationInfo/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_management_MBeanFeatureInfo extends java_lang_Object implements java_io_Serializable, javax_management_DescriptorRead {
    public static "javax/management/MBeanFeatureInfo/serialVersionUID": Long;
    public "javax/management/MBeanFeatureInfo/name": java_lang_String;
    public "javax/management/MBeanFeatureInfo/description": java_lang_String;
    public "javax/management/MBeanFeatureInfo/descriptor": javax_management_Descriptor;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanFeatureInfo/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanFeatureInfo/<init>(Ljava/lang/String;Ljava/lang/String;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanFeatureInfo/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanFeatureInfo/getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDescriptor()Ljavax/management/Descriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_Descriptor) => void): void;
    public "javax/management/MBeanFeatureInfo/getDescriptor()Ljavax/management/Descriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_Descriptor) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanFeatureInfo/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanFeatureInfo/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanFeatureInfo/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanFeatureInfo/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_management_MBeanConstructorInfo extends javax_management_MBeanFeatureInfo implements java_lang_Cloneable {
    public static "javax/management/MBeanConstructorInfo/serialVersionUID": Long;
    public static "javax/management/MBeanConstructorInfo/NO_CONSTRUCTORS": JVMArray<javax_management_MBeanConstructorInfo>;
    public "javax/management/MBeanConstructorInfo/arrayGettersSafe": number;
    public "javax/management/MBeanConstructorInfo/signature": JVMArray<javax_management_MBeanParameterInfo>;
    public "<init>(Ljava/lang/String;Ljava/lang/reflect/Constructor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanConstructorInfo/<init>(Ljava/lang/String;Ljava/lang/reflect/Constructor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanParameterInfo;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanParameterInfo>], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanConstructorInfo/<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanParameterInfo;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanParameterInfo>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanParameterInfo;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanParameterInfo>, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanConstructorInfo/<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanParameterInfo;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanParameterInfo>, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/management/MBeanConstructorInfo/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getSignature()[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public "javax/management/MBeanConstructorInfo/getSignature()[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public "fastGetSignature()[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public "javax/management/MBeanConstructorInfo/fastGetSignature()[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanConstructorInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanConstructorInfo/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanConstructorInfo/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "constructorSignature(Ljava/lang/reflect/Constructor;)[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public static "javax/management/MBeanConstructorInfo/constructorSignature(Ljava/lang/reflect/Constructor;)[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/MBeanConstructorInfo/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_management_MBeanParameterInfo extends javax_management_MBeanFeatureInfo implements java_lang_Cloneable {
    public static "javax/management/MBeanParameterInfo/serialVersionUID": Long;
    public static "javax/management/MBeanParameterInfo/NO_PARAMS": JVMArray<javax_management_MBeanParameterInfo>;
    public "javax/management/MBeanParameterInfo/type": java_lang_String;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanParameterInfo/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanParameterInfo/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/management/MBeanParameterInfo/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanParameterInfo/getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanParameterInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanParameterInfo/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanParameterInfo/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/MBeanParameterInfo/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_management_MBeanOperationInfo extends javax_management_MBeanFeatureInfo implements java_lang_Cloneable {
    public static "javax/management/MBeanOperationInfo/serialVersionUID": Long;
    public static "javax/management/MBeanOperationInfo/NO_OPERATIONS": JVMArray<javax_management_MBeanOperationInfo>;
    public static "javax/management/MBeanOperationInfo/INFO": number;
    public static "javax/management/MBeanOperationInfo/ACTION": number;
    public static "javax/management/MBeanOperationInfo/ACTION_INFO": number;
    public static "javax/management/MBeanOperationInfo/UNKNOWN": number;
    public "javax/management/MBeanOperationInfo/type": java_lang_String;
    public "javax/management/MBeanOperationInfo/signature": JVMArray<javax_management_MBeanParameterInfo>;
    public "javax/management/MBeanOperationInfo/impact": number;
    public "javax/management/MBeanOperationInfo/arrayGettersSafe": number;
    public static "javax/management/MBeanOperationInfo/$assertionsDisabled": number;
    public "<init>(Ljava/lang/String;Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_String, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanOperationInfo/<init>(Ljava/lang/String;Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_String, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanParameterInfo;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanParameterInfo>, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanOperationInfo/<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanParameterInfo;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanParameterInfo>, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanParameterInfo;Ljava/lang/String;ILjavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanParameterInfo>, java_lang_String, number, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanOperationInfo/<init>(Ljava/lang/String;Ljava/lang/String;[Ljavax/management/MBeanParameterInfo;Ljava/lang/String;ILjavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<javax_management_MBeanParameterInfo>, java_lang_String, number, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/management/MBeanOperationInfo/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getReturnType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanOperationInfo/getReturnType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSignature()[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public "javax/management/MBeanOperationInfo/getSignature()[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public "fastGetSignature()[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public "javax/management/MBeanOperationInfo/fastGetSignature()[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public "getImpact()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanOperationInfo/getImpact()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanOperationInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanOperationInfo/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanOperationInfo/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "methodSignature(Ljava/lang/reflect/Method;)[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public static "javax/management/MBeanOperationInfo/methodSignature(Ljava/lang/reflect/Method;)[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public static "parameters([Ljava/lang/Class;[[Ljava/lang/annotation/Annotation;)[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, JVMArray<JVMArray<java_lang_annotation_Annotation>>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public static "javax/management/MBeanOperationInfo/parameters([Ljava/lang/Class;[[Ljava/lang/annotation/Annotation;)[Ljavax/management/MBeanParameterInfo;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, JVMArray<JVMArray<java_lang_annotation_Annotation>>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanParameterInfo>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/MBeanOperationInfo/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_lang_annotation_Annotation extends java_lang_Object {
    "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "annotationType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
  }
  export class javax_management_MBeanAttributeInfo extends javax_management_MBeanFeatureInfo implements java_lang_Cloneable {
    public static "javax/management/MBeanAttributeInfo/serialVersionUID": Long;
    public static "javax/management/MBeanAttributeInfo/NO_ATTRIBUTES": JVMArray<javax_management_MBeanAttributeInfo>;
    public "javax/management/MBeanAttributeInfo/attributeType": java_lang_String;
    public "javax/management/MBeanAttributeInfo/isWrite": number;
    public "javax/management/MBeanAttributeInfo/isRead": number;
    public "javax/management/MBeanAttributeInfo/is": number;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ZZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanAttributeInfo/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ZZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ZZZLjavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number, number, number, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanAttributeInfo/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ZZZLjavax/management/Descriptor;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number, number, number, javax_management_Descriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/reflect/Method;Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_reflect_Method, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/MBeanAttributeInfo/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/reflect/Method;Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_reflect_Method, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/management/MBeanAttributeInfo/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanAttributeInfo/getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isReadable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanAttributeInfo/isReadable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isWritable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanAttributeInfo/isWritable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isIs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanAttributeInfo/isIs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/MBeanAttributeInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanAttributeInfo/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/MBeanAttributeInfo/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isIs(Ljava/lang/reflect/Method;)Z"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "javax/management/MBeanAttributeInfo/isIs(Ljava/lang/reflect/Method;)Z"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "attributeType(Ljava/lang/reflect/Method;Ljava/lang/reflect/Method;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "javax/management/MBeanAttributeInfo/attributeType(Ljava/lang/reflect/Method;Ljava/lang/reflect/Method;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/MBeanAttributeInfo/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface javax_management_DescriptorRead extends java_lang_Object {
    "getDescriptor()Ljavax/management/Descriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_Descriptor) => void): void;
  }
  export interface javax_management_NotificationFilter extends java_lang_Object, java_io_Serializable {
    "isNotificationEnabled(Ljavax/management/Notification;)Z"(thread: JVMThread, args: [javax_management_Notification], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class javax_management_Notification extends java_util_EventObject {
    public static "javax/management/Notification/oldSerialVersionUID": Long;
    public static "javax/management/Notification/newSerialVersionUID": Long;
    public static "javax/management/Notification/oldSerialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "javax/management/Notification/newSerialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "javax/management/Notification/serialVersionUID": Long;
    public static "javax/management/Notification/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "javax/management/Notification/compat": number;
    public "javax/management/Notification/type": java_lang_String;
    public "javax/management/Notification/sequenceNumber": Long;
    public "javax/management/Notification/timeStamp": Long;
    public "javax/management/Notification/userData": java_lang_Object;
    public "javax/management/Notification/message": java_lang_String;
    public "javax/management/Notification/source": java_lang_Object;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;J)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/<init>(Ljava/lang/String;Ljava/lang/Object;J)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;JLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, Long, any, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/<init>(Ljava/lang/String;Ljava/lang/Object;JLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, Long, any, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/<init>(Ljava/lang/String;Ljava/lang/Object;JJ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;JJLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, Long, any, Long, any, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/<init>(Ljava/lang/String;Ljava/lang/Object;JJLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, Long, any, Long, any, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSource(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/setSource(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSequenceNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "javax/management/Notification/getSequenceNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setSequenceNumber(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/setSequenceNumber(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/Notification/getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTimeStamp()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "javax/management/Notification/getTimeStamp()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setTimeStamp(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/setTimeStamp(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/Notification/getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getUserData()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/management/Notification/getUserData()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setUserData(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/setUserData(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/Notification/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Notification/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/management/Notification/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_EventObject extends java_lang_Object implements java_io_Serializable {
    public static "java/util/EventObject/serialVersionUID": Long;
    public "java/util/EventObject/source": java_lang_Object;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/EventObject/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSource()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/EventObject/getSource()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/EventObject/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface javax_management_NotificationListener extends java_lang_Object, java_util_EventListener {
    "handleNotification(Ljavax/management/Notification;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_Notification, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_EventListener extends java_lang_Object {
  }
  export class javax_management_Attribute extends java_lang_Object implements java_io_Serializable {
    public static "javax/management/Attribute/serialVersionUID": Long;
    public "javax/management/Attribute/name": java_lang_String;
    public "javax/management/Attribute/value": java_lang_Object;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/Attribute/<init>(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/Attribute/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/management/Attribute/getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/Attribute/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/Attribute/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/Attribute/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class javax_management_AttributeList extends java_util_ArrayList {
    public "javax/management/AttributeList/typeSafe": number;
    public "javax/management/AttributeList/tainted": number;
    public static "javax/management/AttributeList/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljavax/management/AttributeList;)V"(thread: JVMThread, args: [javax_management_AttributeList], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/<init>(Ljavax/management/AttributeList;)V"(thread: JVMThread, args: [javax_management_AttributeList], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/<init>(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "asList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "javax/management/AttributeList/asList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "add(Ljavax/management/Attribute;)V"(thread: JVMThread, args: [javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/add(Ljavax/management/Attribute;)V"(thread: JVMThread, args: [javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(ILjavax/management/Attribute;)V"(thread: JVMThread, args: [number, javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/add(ILjavax/management/Attribute;)V"(thread: JVMThread, args: [number, javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(ILjavax/management/Attribute;)V"(thread: JVMThread, args: [number, javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/set(ILjavax/management/Attribute;)V"(thread: JVMThread, args: [number, javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAll(Ljavax/management/AttributeList;)Z"(thread: JVMThread, args: [javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/AttributeList/addAll(Ljavax/management/AttributeList;)Z"(thread: JVMThread, args: [javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(ILjavax/management/AttributeList;)Z"(thread: JVMThread, args: [number, javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/AttributeList/addAll(ILjavax/management/AttributeList;)Z"(thread: JVMThread, args: [number, javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/AttributeList/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/AttributeList/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/AttributeList/addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/management/AttributeList/set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "adding(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/adding(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "adding(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/AttributeList/adding(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_management_ObjectInstance extends java_lang_Object implements java_io_Serializable {
    public static "javax/management/ObjectInstance/serialVersionUID": Long;
    public "javax/management/ObjectInstance/name": javax_management_ObjectName;
    public "javax/management/ObjectInstance/className": java_lang_String;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectInstance/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljavax/management/ObjectName;Ljava/lang/String;)V"(thread: JVMThread, args: [javax_management_ObjectName, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectInstance/<init>(Ljavax/management/ObjectName;Ljava/lang/String;)V"(thread: JVMThread, args: [javax_management_ObjectName, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectInstance/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/management/ObjectInstance/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "javax/management/ObjectInstance/getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectInstance/getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectInstance/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface javax_management_MBeanServerConnection extends java_lang_Object {
    "createMBean(Ljava/lang/String;Ljavax/management/ObjectName;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "createMBean(Ljava/lang/String;Ljavax/management/ObjectName;Ljavax/management/ObjectName;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName, javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "createMBean(Ljava/lang/String;Ljavax/management/ObjectName;[Ljava/lang/Object;[Ljava/lang/String;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "createMBean(Ljava/lang/String;Ljavax/management/ObjectName;Ljavax/management/ObjectName;[Ljava/lang/Object;[Ljava/lang/String;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [java_lang_String, javax_management_ObjectName, javax_management_ObjectName, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "unregisterMBean(Ljavax/management/ObjectName;)V"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable) => void): void;
    "getObjectInstance(Ljavax/management/ObjectName;)Ljavax/management/ObjectInstance;"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectInstance) => void): void;
    "queryMBeans(Ljavax/management/ObjectName;Ljavax/management/QueryExp;)Ljava/util/Set;"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_QueryExp], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "queryNames(Ljavax/management/ObjectName;Ljavax/management/QueryExp;)Ljava/util/Set;"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_QueryExp], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    "isRegistered(Ljavax/management/ObjectName;)Z"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMBeanCount()Ljava/lang/Integer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Integer) => void): void;
    "getAttribute(Ljavax/management/ObjectName;Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_management_ObjectName, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getAttributes(Ljavax/management/ObjectName;[Ljava/lang/String;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [javax_management_ObjectName, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    "setAttribute(Ljavax/management/ObjectName;Ljavax/management/Attribute;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    "setAttributes(Ljavax/management/ObjectName;Ljavax/management/AttributeList;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    "invoke(Ljavax/management/ObjectName;Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [javax_management_ObjectName, java_lang_String, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getDefaultDomain()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getDomains()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "addNotificationListener(Ljavax/management/ObjectName;Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "addNotificationListener(Ljavax/management/ObjectName;Ljavax/management/ObjectName;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_ObjectName, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/ObjectName;Ljavax/management/ObjectName;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_ObjectName], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/ObjectName;Ljavax/management/ObjectName;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_ObjectName, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/ObjectName;Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/ObjectName;Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_ObjectName, javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "getMBeanInfo(Ljavax/management/ObjectName;)Ljavax/management/MBeanInfo;"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: javax_management_MBeanInfo) => void): void;
    "isInstanceOf(Ljavax/management/ObjectName;Ljava/lang/String;)Z"(thread: JVMThread, args: [javax_management_ObjectName, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class javax_management_ObjectName$Property extends java_lang_Object {
    public "javax/management/ObjectName$Property/_key_index": number;
    public "javax/management/ObjectName$Property/_key_length": number;
    public "javax/management/ObjectName$Property/_value_length": number;
    public "<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName$Property/<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setKeyIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/management/ObjectName$Property/setKeyIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKeyString(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectName$Property/getKeyString(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getValueString(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/management/ObjectName$Property/getValueString(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface javax_management_QueryExp extends java_lang_Object, java_io_Serializable {
    "apply(Ljavax/management/ObjectName;)Z"(thread: JVMThread, args: [javax_management_ObjectName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setMBeanServer(Ljavax/management/MBeanServer;)V"(thread: JVMThread, args: [javax_management_MBeanServer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface com_sun_management_ThreadMXBean extends java_lang_Object, java_lang_management_ThreadMXBean {
    "getThreadCpuTime([J)[J"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "getThreadUserTime([J)[J"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "getThreadAllocatedBytes(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getThreadAllocatedBytes([J)[J"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "isThreadAllocatedMemorySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadAllocatedMemoryEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setThreadAllocatedMemoryEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getPeakThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTotalStartedThreadCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getDaemonThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAllThreadIds()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "getThreadInfo(J)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    "getThreadInfo([J)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    "getThreadInfo(JI)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    "getThreadInfo([JI)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    "isThreadContentionMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadContentionMonitoringEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setThreadContentionMonitoringEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getCurrentThreadCpuTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getCurrentThreadUserTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getThreadCpuTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getThreadUserTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "isThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isCurrentThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadCpuTimeEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setThreadCpuTimeEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "findMonitorDeadlockedThreads()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "resetPeakThreadCount()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "findDeadlockedThreads()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "isObjectMonitorUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSynchronizerUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getThreadInfo([JZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    "dumpAllThreads(ZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export interface java_lang_management_ThreadMXBean extends java_lang_Object, java_lang_management_PlatformManagedObject {
    "getThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getPeakThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTotalStartedThreadCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getDaemonThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAllThreadIds()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "getThreadInfo(J)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    "getThreadInfo([J)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    "getThreadInfo(JI)Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_ThreadInfo) => void): void;
    "getThreadInfo([JI)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    "isThreadContentionMonitoringSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadContentionMonitoringEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setThreadContentionMonitoringEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getCurrentThreadCpuTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getCurrentThreadUserTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getThreadCpuTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getThreadUserTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "isThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isCurrentThreadCpuTimeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isThreadCpuTimeEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setThreadCpuTimeEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "findMonitorDeadlockedThreads()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "resetPeakThreadCount()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "findDeadlockedThreads()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    "isObjectMonitorUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isSynchronizerUsageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getThreadInfo([JZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    "dumpAllThreads(ZZ)[Ljava/lang/management/ThreadInfo;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_ThreadInfo>) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export interface java_lang_management_PlatformManagedObject extends java_lang_Object {
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class sun_management_OperatingSystemImpl extends sun_management_BaseOperatingSystemImpl implements com_sun_management_UnixOperatingSystemMXBean {
    public "<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/OperatingSystemImpl/<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCommittedVirtualMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/OperatingSystemImpl/getCommittedVirtualMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTotalSwapSpaceSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/OperatingSystemImpl/getTotalSwapSpaceSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFreeSwapSpaceSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/OperatingSystemImpl/getFreeSwapSpaceSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getProcessCpuTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/OperatingSystemImpl/getProcessCpuTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFreePhysicalMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/OperatingSystemImpl/getFreePhysicalMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTotalPhysicalMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/OperatingSystemImpl/getTotalPhysicalMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getOpenFileDescriptorCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/OperatingSystemImpl/getOpenFileDescriptorCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMaxFileDescriptorCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/OperatingSystemImpl/getMaxFileDescriptorCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getSystemCpuLoad()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/OperatingSystemImpl/getSystemCpuLoad()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getProcessCpuLoad()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/OperatingSystemImpl/getProcessCpuLoad()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/OperatingSystemImpl/initialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/OperatingSystemImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface com_sun_management_UnixOperatingSystemMXBean extends java_lang_Object, com_sun_management_OperatingSystemMXBean {
    "getOpenFileDescriptorCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getMaxFileDescriptorCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getCommittedVirtualMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getTotalSwapSpaceSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getFreeSwapSpaceSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getProcessCpuTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getFreePhysicalMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getTotalPhysicalMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getSystemCpuLoad()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getProcessCpuLoad()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getArch()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getAvailableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSystemLoadAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export interface com_sun_management_OperatingSystemMXBean extends java_lang_Object, java_lang_management_OperatingSystemMXBean {
    "getCommittedVirtualMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getTotalSwapSpaceSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getFreeSwapSpaceSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getProcessCpuTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getFreePhysicalMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getTotalPhysicalMemorySize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getSystemCpuLoad()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getProcessCpuLoad()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getArch()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getAvailableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSystemLoadAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export interface java_lang_management_OperatingSystemMXBean extends java_lang_Object, java_lang_management_PlatformManagedObject {
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getArch()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getAvailableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSystemLoadAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class sun_management_BaseOperatingSystemImpl extends java_lang_Object implements java_lang_management_OperatingSystemMXBean {
    public "sun/management/BaseOperatingSystemImpl/jvm": sun_management_VMManagement;
    public static "sun/management/BaseOperatingSystemImpl/unsafe": sun_misc_Unsafe;
    public "sun/management/BaseOperatingSystemImpl/loadavg": JVMArray<number>;
    public "<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/BaseOperatingSystemImpl/<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/BaseOperatingSystemImpl/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getArch()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/BaseOperatingSystemImpl/getArch()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/BaseOperatingSystemImpl/getVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAvailableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/BaseOperatingSystemImpl/getAvailableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSystemLoadAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/BaseOperatingSystemImpl/getSystemLoadAverage()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "sun/management/BaseOperatingSystemImpl/getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/BaseOperatingSystemImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_Sensor extends java_lang_Object {
    public "sun/management/Sensor/lock": java_lang_Object;
    public "sun/management/Sensor/name": java_lang_String;
    public "sun/management/Sensor/count": Long;
    public "sun/management/Sensor/on": number;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Sensor/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/Sensor/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/Sensor/getCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "isOn()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/Sensor/isOn()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "trigger()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Sensor/trigger()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "trigger(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Sensor/trigger(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "trigger(ILjava/lang/management/MemoryUsage;)V"(thread: JVMThread, args: [number, java_lang_management_MemoryUsage], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Sensor/trigger(ILjava/lang/management/MemoryUsage;)V"(thread: JVMThread, args: [number, java_lang_management_MemoryUsage], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Sensor/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Sensor/clear(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/Sensor/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "triggerAction()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Sensor/triggerAction()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "triggerAction(Ljava/lang/management/MemoryUsage;)V"(thread: JVMThread, args: [java_lang_management_MemoryUsage], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Sensor/triggerAction(Ljava/lang/management/MemoryUsage;)V"(thread: JVMThread, args: [java_lang_management_MemoryUsage], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearAction()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Sensor/clearAction()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_MemoryPoolImpl extends java_lang_Object implements java_lang_management_MemoryPoolMXBean {
    public "sun/management/MemoryPoolImpl/name": java_lang_String;
    public "sun/management/MemoryPoolImpl/isHeap": number;
    public "sun/management/MemoryPoolImpl/isValid": number;
    public "sun/management/MemoryPoolImpl/collectionThresholdSupported": number;
    public "sun/management/MemoryPoolImpl/usageThresholdSupported": number;
    public "sun/management/MemoryPoolImpl/managers": JVMArray<java_lang_management_MemoryManagerMXBean>;
    public "sun/management/MemoryPoolImpl/usageThreshold": Long;
    public "sun/management/MemoryPoolImpl/collectionThreshold": Long;
    public "sun/management/MemoryPoolImpl/usageSensorRegistered": number;
    public "sun/management/MemoryPoolImpl/gcSensorRegistered": number;
    public "sun/management/MemoryPoolImpl/usageSensor": sun_management_Sensor;
    public "sun/management/MemoryPoolImpl/gcSensor": sun_management_Sensor;
    public "<init>(Ljava/lang/String;ZJJ)V"(thread: JVMThread, args: [java_lang_String, number, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryPoolImpl/<init>(Ljava/lang/String;ZJJ)V"(thread: JVMThread, args: [java_lang_String, number, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/MemoryPoolImpl/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/MemoryPoolImpl/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getType()Ljava/lang/management/MemoryType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryType) => void): void;
    public "sun/management/MemoryPoolImpl/getType()Ljava/lang/management/MemoryType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryType) => void): void;
    public "getUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "sun/management/MemoryPoolImpl/getUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "getPeakUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "sun/management/MemoryPoolImpl/getPeakUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "getUsageThreshold()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/MemoryPoolImpl/getUsageThreshold()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setUsageThreshold(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryPoolImpl/setUsageThreshold(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMemoryManagers()[Ljava/lang/management/MemoryManagerMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryManagerMXBean>) => void): void;
    public "sun/management/MemoryPoolImpl/getMemoryManagers()[Ljava/lang/management/MemoryManagerMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryManagerMXBean>) => void): void;
    public "getMemoryManagerNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/management/MemoryPoolImpl/getMemoryManagerNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "resetPeakUsage()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryPoolImpl/resetPeakUsage()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isUsageThresholdExceeded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/MemoryPoolImpl/isUsageThresholdExceeded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUsageThresholdCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/MemoryPoolImpl/getUsageThresholdCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "isUsageThresholdSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/MemoryPoolImpl/isUsageThresholdSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCollectionUsageThreshold()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/MemoryPoolImpl/getCollectionUsageThreshold()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setCollectionUsageThreshold(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryPoolImpl/setCollectionUsageThreshold(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "isCollectionUsageThresholdExceeded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/MemoryPoolImpl/isCollectionUsageThresholdExceeded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCollectionUsageThresholdCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/MemoryPoolImpl/getCollectionUsageThresholdCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getCollectionUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "sun/management/MemoryPoolImpl/getCollectionUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "isCollectionUsageThresholdSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/MemoryPoolImpl/isCollectionUsageThresholdSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUsage0()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "sun/management/MemoryPoolImpl/getUsage0()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "getPeakUsage0()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "sun/management/MemoryPoolImpl/getPeakUsage0()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "getCollectionUsage0()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "sun/management/MemoryPoolImpl/getCollectionUsage0()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "setUsageThreshold0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryPoolImpl/setUsageThreshold0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCollectionThreshold0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryPoolImpl/setCollectionThreshold0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "resetPeakUsage0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryPoolImpl/resetPeakUsage0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMemoryManagers0()[Ljava/lang/management/MemoryManagerMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryManagerMXBean>) => void): void;
    public "sun/management/MemoryPoolImpl/getMemoryManagers0()[Ljava/lang/management/MemoryManagerMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryManagerMXBean>) => void): void;
    public "setPoolUsageSensor(Lsun/management/Sensor;)V"(thread: JVMThread, args: [sun_management_Sensor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryPoolImpl/setPoolUsageSensor(Lsun/management/Sensor;)V"(thread: JVMThread, args: [sun_management_Sensor], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPoolCollectionSensor(Lsun/management/Sensor;)V"(thread: JVMThread, args: [sun_management_Sensor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryPoolImpl/setPoolCollectionSensor(Lsun/management/Sensor;)V"(thread: JVMThread, args: [sun_management_Sensor], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "sun/management/MemoryPoolImpl/getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "access$000(Lsun/management/MemoryPoolImpl;)Lsun/management/Sensor;"(thread: JVMThread, args: [sun_management_MemoryPoolImpl], cb?: (e?: java_lang_Throwable, rv?: sun_management_Sensor) => void): void;
    public static "sun/management/MemoryPoolImpl/access$000(Lsun/management/MemoryPoolImpl;)Lsun/management/Sensor;"(thread: JVMThread, args: [sun_management_MemoryPoolImpl], cb?: (e?: java_lang_Throwable, rv?: sun_management_Sensor) => void): void;
  }
  export class java_lang_management_MemoryType extends java_lang_Enum {
    public static "java/lang/management/MemoryType/HEAP": java_lang_management_MemoryType;
    public static "java/lang/management/MemoryType/NON_HEAP": java_lang_management_MemoryType;
    public "java/lang/management/MemoryType/description": java_lang_String;
    public static "java/lang/management/MemoryType/serialVersionUID": Long;
    public static "java/lang/management/MemoryType/$VALUES": JVMArray<java_lang_management_MemoryType>;
    public static "values()[Ljava/lang/management/MemoryType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryType>) => void): void;
    public static "java/lang/management/MemoryType/values()[Ljava/lang/management/MemoryType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryType>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/management/MemoryType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryType) => void): void;
    public static "java/lang/management/MemoryType/valueOf(Ljava/lang/String;)Ljava/lang/management/MemoryType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryType) => void): void;
    public "<init>(Ljava/lang/String;ILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/MemoryType/<init>(Ljava/lang/String;ILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/management/MemoryType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/management/MemoryType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_MemoryManagerImpl extends sun_management_NotificationEmitterSupport implements java_lang_management_MemoryManagerMXBean {
    public "sun/management/MemoryManagerImpl/name": java_lang_String;
    public "sun/management/MemoryManagerImpl/isValid": number;
    public "sun/management/MemoryManagerImpl/pools": JVMArray<java_lang_management_MemoryPoolMXBean>;
    public "sun/management/MemoryManagerImpl/notifInfo": JVMArray<javax_management_MBeanNotificationInfo>;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryManagerImpl/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/MemoryManagerImpl/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/MemoryManagerImpl/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMemoryPoolNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/management/MemoryManagerImpl/getMemoryPoolNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getMemoryPools()[Ljava/lang/management/MemoryPoolMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryPoolMXBean>) => void): void;
    public "sun/management/MemoryManagerImpl/getMemoryPools()[Ljava/lang/management/MemoryPoolMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryPoolMXBean>) => void): void;
    public "getMemoryPools0()[Ljava/lang/management/MemoryPoolMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryPoolMXBean>) => void): void;
    public "sun/management/MemoryManagerImpl/getMemoryPools0()[Ljava/lang/management/MemoryPoolMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryPoolMXBean>) => void): void;
    public "getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "sun/management/MemoryManagerImpl/getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "sun/management/MemoryManagerImpl/getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class sun_management_NotificationEmitterSupport extends java_lang_Object implements javax_management_NotificationEmitter {
    public "sun/management/NotificationEmitterSupport/listenerLock": java_lang_Object;
    public "sun/management/NotificationEmitterSupport/listenerList": java_util_List;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/NotificationEmitterSupport/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/NotificationEmitterSupport/addNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotificationListener(Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/NotificationEmitterSupport/removeNotificationListener(Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/NotificationEmitterSupport/removeNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sendNotification(Ljavax/management/Notification;)V"(thread: JVMThread, args: [javax_management_Notification], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/NotificationEmitterSupport/sendNotification(Ljavax/management/Notification;)V"(thread: JVMThread, args: [javax_management_Notification], cb?: (e?: java_lang_Throwable) => void): void;
    public "hasListeners()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/NotificationEmitterSupport/hasListeners()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "sun/management/NotificationEmitterSupport/getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
  }
  export interface javax_management_NotificationEmitter extends java_lang_Object, javax_management_NotificationBroadcaster {
    "removeNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "addNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    "getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
  }
  export interface javax_management_NotificationBroadcaster extends java_lang_Object {
    "addNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "removeNotificationListener(Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    "getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
  }
  export interface java_lang_management_MemoryManagerMXBean extends java_lang_Object, java_lang_management_PlatformManagedObject {
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMemoryPoolNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export interface java_lang_management_MemoryPoolMXBean extends java_lang_Object, java_lang_management_PlatformManagedObject {
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getType()Ljava/lang/management/MemoryType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryType) => void): void;
    "getUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    "getPeakUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    "resetPeakUsage()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMemoryManagerNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "getUsageThreshold()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "setUsageThreshold(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    "isUsageThresholdExceeded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getUsageThresholdCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "isUsageThresholdSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getCollectionUsageThreshold()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "setCollectionUsageThreshold(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    "isCollectionUsageThresholdExceeded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getCollectionUsageThresholdCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getCollectionUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    "isCollectionUsageThresholdSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class sun_management_MemoryImpl extends sun_management_NotificationEmitterSupport implements java_lang_management_MemoryMXBean {
    public "sun/management/MemoryImpl/jvm": sun_management_VMManagement;
    public static "sun/management/MemoryImpl/pools": JVMArray<java_lang_management_MemoryPoolMXBean>;
    public static "sun/management/MemoryImpl/mgrs": JVMArray<java_lang_management_MemoryManagerMXBean>;
    public static "sun/management/MemoryImpl/notifName": java_lang_String;
    public static "sun/management/MemoryImpl/notifTypes": JVMArray<java_lang_String>;
    public static "sun/management/MemoryImpl/notifMsgs": JVMArray<java_lang_String>;
    public "sun/management/MemoryImpl/notifInfo": JVMArray<javax_management_MBeanNotificationInfo>;
    public static "sun/management/MemoryImpl/seqNumber": Long;
    public "<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryImpl/<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjectPendingFinalizationCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/MemoryImpl/getObjectPendingFinalizationCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "gc()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryImpl/gc()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getHeapMemoryUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "sun/management/MemoryImpl/getHeapMemoryUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "getNonHeapMemoryUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "sun/management/MemoryImpl/getNonHeapMemoryUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "isVerbose()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/MemoryImpl/isVerbose()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setVerbose(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryImpl/setVerbose(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getMemoryPools()[Ljava/lang/management/MemoryPoolMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryPoolMXBean>) => void): void;
    public static "sun/management/MemoryImpl/getMemoryPools()[Ljava/lang/management/MemoryPoolMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryPoolMXBean>) => void): void;
    public static "getMemoryManagers()[Ljava/lang/management/MemoryManagerMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryManagerMXBean>) => void): void;
    public static "sun/management/MemoryImpl/getMemoryManagers()[Ljava/lang/management/MemoryManagerMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryManagerMXBean>) => void): void;
    public static "getMemoryPools0()[Ljava/lang/management/MemoryPoolMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryPoolMXBean>) => void): void;
    public static "sun/management/MemoryImpl/getMemoryPools0()[Ljava/lang/management/MemoryPoolMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryPoolMXBean>) => void): void;
    public static "getMemoryManagers0()[Ljava/lang/management/MemoryManagerMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryManagerMXBean>) => void): void;
    public static "sun/management/MemoryImpl/getMemoryManagers0()[Ljava/lang/management/MemoryManagerMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_management_MemoryManagerMXBean>) => void): void;
    public "getMemoryUsage0(Z)Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "sun/management/MemoryImpl/getMemoryUsage0(Z)Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public "setVerboseGC(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/MemoryImpl/setVerboseGC(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "sun/management/MemoryImpl/getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public static "getNotifMsg(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/management/MemoryImpl/getNotifMsg(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getNextSeqNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/management/MemoryImpl/getNextSeqNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "createNotification(Ljava/lang/String;Ljava/lang/String;Ljava/lang/management/MemoryUsage;J)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_management_MemoryUsage, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/MemoryImpl/createNotification(Ljava/lang/String;Ljava/lang/String;Ljava/lang/management/MemoryUsage;J)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_management_MemoryUsage, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "sun/management/MemoryImpl/getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/MemoryImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_lang_management_MemoryMXBean extends java_lang_Object, java_lang_management_PlatformManagedObject {
    "getObjectPendingFinalizationCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getHeapMemoryUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    "getNonHeapMemoryUsage()Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    "isVerbose()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVerbose(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "gc()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class sun_management_HotspotThread extends java_lang_Object implements sun_management_HotspotThreadMBean {
    public "sun/management/HotspotThread/jvm": sun_management_VMManagement;
    public static "sun/management/HotspotThread/JAVA_THREADS": java_lang_String;
    public static "sun/management/HotspotThread/COM_SUN_THREADS": java_lang_String;
    public static "sun/management/HotspotThread/SUN_THREADS": java_lang_String;
    public static "sun/management/HotspotThread/THREADS_COUNTER_NAME_PATTERN": java_lang_String;
    public "<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/HotspotThread/<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInternalThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/HotspotThread/getInternalThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInternalThreadCpuTimes()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "sun/management/HotspotThread/getInternalThreadCpuTimes()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getInternalThreadTimes0([Ljava/lang/String;[J)I"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/HotspotThread/getInternalThreadTimes0([Ljava/lang/String;[J)I"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInternalThreadingCounters()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/management/HotspotThread/getInternalThreadingCounters()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
  }
  export interface sun_management_HotspotThreadMBean extends java_lang_Object {
    "getInternalThreadCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getInternalThreadCpuTimes()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    "getInternalThreadingCounters()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
  }
  export class sun_management_HotSpotDiagnostic extends java_lang_Object implements com_sun_management_HotSpotDiagnosticMXBean {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/HotSpotDiagnostic/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dumpHeap(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/HotSpotDiagnostic/dumpHeap(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "dumpHeap0(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/HotSpotDiagnostic/dumpHeap0(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDiagnosticOptions()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/management/HotSpotDiagnostic/getDiagnosticOptions()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getVMOption(Ljava/lang/String;)Lcom/sun/management/VMOption;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption) => void): void;
    public "sun/management/HotSpotDiagnostic/getVMOption(Ljava/lang/String;)Lcom/sun/management/VMOption;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption) => void): void;
    public "setVMOption(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/HotSpotDiagnostic/setVMOption(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "sun/management/HotSpotDiagnostic/getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class com_sun_management_VMOption extends java_lang_Object {
    public "com/sun/management/VMOption/name": java_lang_String;
    public "com/sun/management/VMOption/value": java_lang_String;
    public "com/sun/management/VMOption/writeable": number;
    public "com/sun/management/VMOption/origin": com_sun_management_VMOption$Origin;
    public "<init>(Ljava/lang/String;Ljava/lang/String;ZLcom/sun/management/VMOption$Origin;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, com_sun_management_VMOption$Origin], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/management/VMOption/<init>(Ljava/lang/String;Ljava/lang/String;ZLcom/sun/management/VMOption$Origin;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number, com_sun_management_VMOption$Origin], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljavax/management/openmbean/CompositeData;)V"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/management/VMOption/<init>(Ljavax/management/openmbean/CompositeData;)V"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "com/sun/management/VMOption/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getValue()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "com/sun/management/VMOption/getValue()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getOrigin()Lcom/sun/management/VMOption$Origin;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption$Origin) => void): void;
    public "com/sun/management/VMOption/getOrigin()Lcom/sun/management/VMOption$Origin;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption$Origin) => void): void;
    public "isWriteable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/management/VMOption/isWriteable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "com/sun/management/VMOption/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "from(Ljavax/management/openmbean/CompositeData;)Lcom/sun/management/VMOption;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption) => void): void;
    public static "com/sun/management/VMOption/from(Ljavax/management/openmbean/CompositeData;)Lcom/sun/management/VMOption;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption) => void): void;
  }
  export class com_sun_management_VMOption$Origin extends java_lang_Enum {
    public static "com/sun/management/VMOption$Origin/DEFAULT": com_sun_management_VMOption$Origin;
    public static "com/sun/management/VMOption$Origin/VM_CREATION": com_sun_management_VMOption$Origin;
    public static "com/sun/management/VMOption$Origin/ENVIRON_VAR": com_sun_management_VMOption$Origin;
    public static "com/sun/management/VMOption$Origin/CONFIG_FILE": com_sun_management_VMOption$Origin;
    public static "com/sun/management/VMOption$Origin/MANAGEMENT": com_sun_management_VMOption$Origin;
    public static "com/sun/management/VMOption$Origin/ERGONOMIC": com_sun_management_VMOption$Origin;
    public static "com/sun/management/VMOption$Origin/OTHER": com_sun_management_VMOption$Origin;
    public static "com/sun/management/VMOption$Origin/$VALUES": JVMArray<com_sun_management_VMOption$Origin>;
    public static "values()[Lcom/sun/management/VMOption$Origin;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<com_sun_management_VMOption$Origin>) => void): void;
    public static "com/sun/management/VMOption$Origin/values()[Lcom/sun/management/VMOption$Origin;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<com_sun_management_VMOption$Origin>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lcom/sun/management/VMOption$Origin;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption$Origin) => void): void;
    public static "com/sun/management/VMOption$Origin/valueOf(Ljava/lang/String;)Lcom/sun/management/VMOption$Origin;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption$Origin) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/management/VMOption$Origin/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "com/sun/management/VMOption$Origin/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface com_sun_management_HotSpotDiagnosticMXBean extends java_lang_Object, java_lang_management_PlatformManagedObject {
    "dumpHeap(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    "getDiagnosticOptions()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "getVMOption(Ljava/lang/String;)Lcom/sun/management/VMOption;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption) => void): void;
    "setVMOption(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class java_lang_management_MemoryUsage extends java_lang_Object {
    public "java/lang/management/MemoryUsage/init": Long;
    public "java/lang/management/MemoryUsage/used": Long;
    public "java/lang/management/MemoryUsage/committed": Long;
    public "java/lang/management/MemoryUsage/max": Long;
    public "<init>(JJJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/MemoryUsage/<init>(JJJJ)V"(thread: JVMThread, args: [Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljavax/management/openmbean/CompositeData;)V"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/management/MemoryUsage/<init>(Ljavax/management/openmbean/CompositeData;)V"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInit()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/MemoryUsage/getInit()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getUsed()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/MemoryUsage/getUsed()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getCommitted()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/MemoryUsage/getCommitted()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMax()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/management/MemoryUsage/getMax()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/management/MemoryUsage/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "from(Ljavax/management/openmbean/CompositeData;)Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
    public static "java/lang/management/MemoryUsage/from(Ljavax/management/openmbean/CompositeData;)Ljava/lang/management/MemoryUsage;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: java_lang_management_MemoryUsage) => void): void;
  }
  export class com_sun_management_GcInfo extends java_lang_Object implements javax_management_openmbean_CompositeData, javax_management_openmbean_CompositeDataView {
    public "com/sun/management/GcInfo/index": Long;
    public "com/sun/management/GcInfo/startTime": Long;
    public "com/sun/management/GcInfo/endTime": Long;
    public "com/sun/management/GcInfo/usageBeforeGc": java_util_Map;
    public "com/sun/management/GcInfo/usageAfterGc": java_util_Map;
    public "com/sun/management/GcInfo/extAttributes": JVMArray<java_lang_Object>;
    public "com/sun/management/GcInfo/cdata": javax_management_openmbean_CompositeData;
    public "com/sun/management/GcInfo/builder": sun_management_GcInfoBuilder;
    public "<init>(Lsun/management/GcInfoBuilder;JJJ[Ljava/lang/management/MemoryUsage;[Ljava/lang/management/MemoryUsage;[Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_management_GcInfoBuilder, Long, any, Long, any, Long, any, JVMArray<java_lang_management_MemoryUsage>, JVMArray<java_lang_management_MemoryUsage>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/management/GcInfo/<init>(Lsun/management/GcInfoBuilder;JJJ[Ljava/lang/management/MemoryUsage;[Ljava/lang/management/MemoryUsage;[Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_management_GcInfoBuilder, Long, any, Long, any, Long, any, JVMArray<java_lang_management_MemoryUsage>, JVMArray<java_lang_management_MemoryUsage>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljavax/management/openmbean/CompositeData;)V"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable) => void): void;
    public "com/sun/management/GcInfo/<init>(Ljavax/management/openmbean/CompositeData;)V"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable) => void): void;
    public "getId()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "com/sun/management/GcInfo/getId()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getStartTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "com/sun/management/GcInfo/getStartTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getEndTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "com/sun/management/GcInfo/getEndTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getDuration()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "com/sun/management/GcInfo/getDuration()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMemoryUsageBeforeGc()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "com/sun/management/GcInfo/getMemoryUsageBeforeGc()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getMemoryUsageAfterGc()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "com/sun/management/GcInfo/getMemoryUsageAfterGc()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "from(Ljavax/management/openmbean/CompositeData;)Lcom/sun/management/GcInfo;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_GcInfo) => void): void;
    public static "com/sun/management/GcInfo/from(Ljavax/management/openmbean/CompositeData;)Lcom/sun/management/GcInfo;"(thread: JVMThread, args: [javax_management_openmbean_CompositeData], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_GcInfo) => void): void;
    public "containsKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/management/GcInfo/containsKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/management/GcInfo/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/management/GcInfo/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "com/sun/management/GcInfo/get(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getAll([Ljava/lang/String;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "com/sun/management/GcInfo/getAll([Ljava/lang/String;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "getCompositeType()Ljavax/management/openmbean/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_CompositeType) => void): void;
    public "com/sun/management/GcInfo/getCompositeType()Ljavax/management/openmbean/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_CompositeType) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "com/sun/management/GcInfo/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "com/sun/management/GcInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "com/sun/management/GcInfo/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "toCompositeData(Ljavax/management/openmbean/CompositeType;)Ljavax/management/openmbean/CompositeData;"(thread: JVMThread, args: [javax_management_openmbean_CompositeType], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_CompositeData) => void): void;
    public "com/sun/management/GcInfo/toCompositeData(Ljavax/management/openmbean/CompositeType;)Ljavax/management/openmbean/CompositeData;"(thread: JVMThread, args: [javax_management_openmbean_CompositeType], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_CompositeData) => void): void;
  }
  export interface javax_management_openmbean_CompositeDataView extends java_lang_Object {
    "toCompositeData(Ljavax/management/openmbean/CompositeType;)Ljavax/management/openmbean/CompositeData;"(thread: JVMThread, args: [javax_management_openmbean_CompositeType], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_CompositeData) => void): void;
  }
  export interface java_lang_management_GarbageCollectorMXBean extends java_lang_Object, java_lang_management_MemoryManagerMXBean {
    "getCollectionCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getCollectionTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMemoryPoolNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class sun_management_GcInfoBuilder extends java_lang_Object {
    public "sun/management/GcInfoBuilder/gc": java_lang_management_GarbageCollectorMXBean;
    public "sun/management/GcInfoBuilder/poolNames": JVMArray<java_lang_String>;
    public "sun/management/GcInfoBuilder/allItemNames": JVMArray<java_lang_String>;
    public "sun/management/GcInfoBuilder/gcInfoCompositeType": javax_management_openmbean_CompositeType;
    public "sun/management/GcInfoBuilder/gcExtItemCount": number;
    public "sun/management/GcInfoBuilder/gcExtItemNames": JVMArray<java_lang_String>;
    public "sun/management/GcInfoBuilder/gcExtItemDescs": JVMArray<java_lang_String>;
    public "sun/management/GcInfoBuilder/gcExtItemTypes": JVMArray<number>;
    public "<init>(Ljava/lang/management/GarbageCollectorMXBean;[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_management_GarbageCollectorMXBean, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/GcInfoBuilder/<init>(Ljava/lang/management/GarbageCollectorMXBean;[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_management_GarbageCollectorMXBean, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLastGcInfo()Lcom/sun/management/GcInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_GcInfo) => void): void;
    public "sun/management/GcInfoBuilder/getLastGcInfo()Lcom/sun/management/GcInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_GcInfo) => void): void;
    public "getPoolNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/management/GcInfoBuilder/getPoolNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getGcExtItemCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/GcInfoBuilder/getGcExtItemCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGcInfoCompositeType()Ljavax/management/openmbean/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_CompositeType) => void): void;
    public "sun/management/GcInfoBuilder/getGcInfoCompositeType()Ljavax/management/openmbean/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_openmbean_CompositeType) => void): void;
    public "getItemNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/management/GcInfoBuilder/getItemNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getNumGcExtAttributes(Ljava/lang/management/GarbageCollectorMXBean;)I"(thread: JVMThread, args: [java_lang_management_GarbageCollectorMXBean], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/GcInfoBuilder/getNumGcExtAttributes(Ljava/lang/management/GarbageCollectorMXBean;)I"(thread: JVMThread, args: [java_lang_management_GarbageCollectorMXBean], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fillGcAttributeInfo(Ljava/lang/management/GarbageCollectorMXBean;I[Ljava/lang/String;[C[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_management_GarbageCollectorMXBean, number, JVMArray<java_lang_String>, JVMArray<number>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/GcInfoBuilder/fillGcAttributeInfo(Ljava/lang/management/GarbageCollectorMXBean;I[Ljava/lang/String;[C[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_management_GarbageCollectorMXBean, number, JVMArray<java_lang_String>, JVMArray<number>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLastGcInfo0(Ljava/lang/management/GarbageCollectorMXBean;I[Ljava/lang/Object;[C[Ljava/lang/management/MemoryUsage;[Ljava/lang/management/MemoryUsage;)Lcom/sun/management/GcInfo;"(thread: JVMThread, args: [java_lang_management_GarbageCollectorMXBean, number, JVMArray<java_lang_Object>, JVMArray<number>, JVMArray<java_lang_management_MemoryUsage>, JVMArray<java_lang_management_MemoryUsage>], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_GcInfo) => void): void;
    public "sun/management/GcInfoBuilder/getLastGcInfo0(Ljava/lang/management/GarbageCollectorMXBean;I[Ljava/lang/Object;[C[Ljava/lang/management/MemoryUsage;[Ljava/lang/management/MemoryUsage;)Lcom/sun/management/GcInfo;"(thread: JVMThread, args: [java_lang_management_GarbageCollectorMXBean, number, JVMArray<java_lang_Object>, JVMArray<number>, JVMArray<java_lang_management_MemoryUsage>, JVMArray<java_lang_management_MemoryUsage>], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_GcInfo) => void): void;
  }
  export interface com_sun_management_GarbageCollectorMXBean extends java_lang_Object, java_lang_management_GarbageCollectorMXBean {
    "getLastGcInfo()Lcom/sun/management/GcInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_GcInfo) => void): void;
    "getCollectionCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getCollectionTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMemoryPoolNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class sun_management_GarbageCollectorImpl extends sun_management_MemoryManagerImpl implements com_sun_management_GarbageCollectorMXBean {
    public "sun/management/GarbageCollectorImpl/poolNames": JVMArray<java_lang_String>;
    public "sun/management/GarbageCollectorImpl/gcInfoBuilder": sun_management_GcInfoBuilder;
    public static "sun/management/GarbageCollectorImpl/notifName": java_lang_String;
    public static "sun/management/GarbageCollectorImpl/gcNotifTypes": JVMArray<java_lang_String>;
    public "sun/management/GarbageCollectorImpl/notifInfo": JVMArray<javax_management_MBeanNotificationInfo>;
    public static "sun/management/GarbageCollectorImpl/seqNumber": Long;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/GarbageCollectorImpl/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCollectionCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/GarbageCollectorImpl/getCollectionCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getCollectionTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/GarbageCollectorImpl/getCollectionTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getAllPoolNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/management/GarbageCollectorImpl/getAllPoolNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getGcInfoBuilder()Lsun/management/GcInfoBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_GcInfoBuilder) => void): void;
    public "sun/management/GarbageCollectorImpl/getGcInfoBuilder()Lsun/management/GcInfoBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_GcInfoBuilder) => void): void;
    public "getLastGcInfo()Lcom/sun/management/GcInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_GcInfo) => void): void;
    public "sun/management/GarbageCollectorImpl/getLastGcInfo()Lcom/sun/management/GcInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_GcInfo) => void): void;
    public "getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "sun/management/GarbageCollectorImpl/getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public static "getNextSeqNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/management/GarbageCollectorImpl/getNextSeqNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "createGCNotification(JLjava/lang/String;Ljava/lang/String;Ljava/lang/String;Lcom/sun/management/GcInfo;)V"(thread: JVMThread, args: [Long, any, java_lang_String, java_lang_String, java_lang_String, com_sun_management_GcInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/GarbageCollectorImpl/createGCNotification(JLjava/lang/String;Ljava/lang/String;Ljava/lang/String;Lcom/sun/management/GcInfo;)V"(thread: JVMThread, args: [Long, any, java_lang_String, java_lang_String, java_lang_String, com_sun_management_GcInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "addNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/GarbageCollectorImpl/addNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotificationListener(Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/GarbageCollectorImpl/removeNotificationListener(Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/GarbageCollectorImpl/removeNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "sun/management/GarbageCollectorImpl/getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "setNotificationEnabled(Lcom/sun/management/GarbageCollectorMXBean;Z)V"(thread: JVMThread, args: [com_sun_management_GarbageCollectorMXBean, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/GarbageCollectorImpl/setNotificationEnabled(Lcom/sun/management/GarbageCollectorMXBean;Z)V"(thread: JVMThread, args: [com_sun_management_GarbageCollectorMXBean, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/GarbageCollectorImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_Flag extends java_lang_Object {
    public "sun/management/Flag/name": java_lang_String;
    public "sun/management/Flag/value": java_lang_Object;
    public "sun/management/Flag/origin": com_sun_management_VMOption$Origin;
    public "sun/management/Flag/writeable": number;
    public "sun/management/Flag/external": number;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;ZZLcom/sun/management/VMOption$Origin;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, number, number, com_sun_management_VMOption$Origin], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/Flag/<init>(Ljava/lang/String;Ljava/lang/Object;ZZLcom/sun/management/VMOption$Origin;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, number, number, com_sun_management_VMOption$Origin], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/management/Flag/getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isWriteable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/Flag/isWriteable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isExternal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/Flag/isExternal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getVMOption()Lcom/sun/management/VMOption;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption) => void): void;
    public "sun/management/Flag/getVMOption()Lcom/sun/management/VMOption;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: com_sun_management_VMOption) => void): void;
    public static "getFlag(Ljava/lang/String;)Lsun/management/Flag;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_management_Flag) => void): void;
    public static "sun/management/Flag/getFlag(Ljava/lang/String;)Lsun/management/Flag;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_management_Flag) => void): void;
    public static "getAllFlags()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "sun/management/Flag/getAllFlags()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "getFlags([Ljava/lang/String;I)Ljava/util/List;"(thread: JVMThread, args: [JVMArray<java_lang_String>, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "sun/management/Flag/getFlags([Ljava/lang/String;I)Ljava/util/List;"(thread: JVMThread, args: [JVMArray<java_lang_String>, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "getAllFlagNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "sun/management/Flag/getAllFlagNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "getFlags([Ljava/lang/String;[Lsun/management/Flag;I)I"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<sun_management_Flag>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/management/Flag/getFlags([Ljava/lang/String;[Lsun/management/Flag;I)I"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<sun_management_Flag>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getInternalFlagCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/management/Flag/getInternalFlagCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setLongValue(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/Flag/setLongValue(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setBooleanValue(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/Flag/setBooleanValue(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setStringValue(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/Flag/setStringValue(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/Flag/initialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/Flag/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_FileSystemImpl extends sun_management_FileSystem {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/FileSystemImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "supportsFileSecurity(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/FileSystemImpl/supportsFileSecurity(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAccessUserOnly(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/FileSystemImpl/isAccessUserOnly(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isAccessUserOnly0(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/management/FileSystemImpl/isAccessUserOnly0(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/FileSystemImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_FileSystem extends java_lang_Object {
    public static "sun/management/FileSystem/lock": java_lang_Object;
    public static "sun/management/FileSystem/fs": sun_management_FileSystem;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/FileSystem/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "open()Lsun/management/FileSystem;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_FileSystem) => void): void;
    public static "sun/management/FileSystem/open()Lsun/management/FileSystem;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_management_FileSystem) => void): void;
    public "supportsFileSecurity(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/FileSystem/supportsFileSecurity(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAccessUserOnly(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/FileSystem/isAccessUserOnly(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/FileSystem/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_DiagnosticCommandInfo extends java_lang_Object {
    public "sun/management/DiagnosticCommandInfo/name": java_lang_String;
    public "sun/management/DiagnosticCommandInfo/description": java_lang_String;
    public "sun/management/DiagnosticCommandInfo/impact": java_lang_String;
    public "sun/management/DiagnosticCommandInfo/permissionClass": java_lang_String;
    public "sun/management/DiagnosticCommandInfo/permissionName": java_lang_String;
    public "sun/management/DiagnosticCommandInfo/permissionAction": java_lang_String;
    public "sun/management/DiagnosticCommandInfo/enabled": number;
    public "sun/management/DiagnosticCommandInfo/arguments": java_util_List;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/DiagnosticCommandInfo/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/DiagnosticCommandInfo/getDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getImpact()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/DiagnosticCommandInfo/getImpact()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPermissionClass()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/DiagnosticCommandInfo/getPermissionClass()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPermissionName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/DiagnosticCommandInfo/getPermissionName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPermissionAction()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/DiagnosticCommandInfo/getPermissionAction()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/DiagnosticCommandInfo/isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getArgumentsInfo()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/management/DiagnosticCommandInfo/getArgumentsInfo()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ZLjava/util/List;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, number, java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/DiagnosticCommandInfo/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;ZLjava/util/List;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, number, java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_DiagnosticCommandImpl extends sun_management_NotificationEmitterSupport implements com_sun_management_DiagnosticCommandMBean {
    public "sun/management/DiagnosticCommandImpl/jvm": sun_management_VMManagement;
    public "sun/management/DiagnosticCommandImpl/wrappers": java_util_Map;
    public static "sun/management/DiagnosticCommandImpl/strClassName": java_lang_String;
    public static "sun/management/DiagnosticCommandImpl/strArrayClassName": java_lang_String;
    public "sun/management/DiagnosticCommandImpl/isSupported": number;
    public static "sun/management/DiagnosticCommandImpl/notifName": java_lang_String;
    public static "sun/management/DiagnosticCommandImpl/diagFramNotifTypes": JVMArray<java_lang_String>;
    public "sun/management/DiagnosticCommandImpl/notifInfo": JVMArray<javax_management_MBeanNotificationInfo>;
    public static "sun/management/DiagnosticCommandImpl/seqNumber": Long;
    public "getAttribute(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/management/DiagnosticCommandImpl/getAttribute(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setAttribute(Ljavax/management/Attribute;)V"(thread: JVMThread, args: [javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/DiagnosticCommandImpl/setAttribute(Ljavax/management/Attribute;)V"(thread: JVMThread, args: [javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAttributes([Ljava/lang/String;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    public "sun/management/DiagnosticCommandImpl/getAttributes([Ljava/lang/String;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    public "setAttributes(Ljavax/management/AttributeList;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    public "sun/management/DiagnosticCommandImpl/setAttributes(Ljavax/management/AttributeList;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    public "<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/DiagnosticCommandImpl/<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMBeanInfo()Ljavax/management/MBeanInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_MBeanInfo) => void): void;
    public "sun/management/DiagnosticCommandImpl/getMBeanInfo()Ljavax/management/MBeanInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_MBeanInfo) => void): void;
    public "invoke(Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/management/DiagnosticCommandImpl/invoke(Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "transform(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/management/DiagnosticCommandImpl/transform(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "commandDescriptor(Lsun/management/DiagnosticCommandImpl$Wrapper;)Ljavax/management/Descriptor;"(thread: JVMThread, args: [sun_management_DiagnosticCommandImpl$Wrapper], cb?: (e?: java_lang_Throwable, rv?: javax_management_Descriptor) => void): void;
    public "sun/management/DiagnosticCommandImpl/commandDescriptor(Lsun/management/DiagnosticCommandImpl$Wrapper;)Ljavax/management/Descriptor;"(thread: JVMThread, args: [sun_management_DiagnosticCommandImpl$Wrapper], cb?: (e?: java_lang_Throwable, rv?: javax_management_Descriptor) => void): void;
    public "getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public "sun/management/DiagnosticCommandImpl/getNotificationInfo()[Ljavax/management/MBeanNotificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_management_MBeanNotificationInfo>) => void): void;
    public static "getNextSeqNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/management/DiagnosticCommandImpl/getNextSeqNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "createDiagnosticFrameworkNotification()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/DiagnosticCommandImpl/createDiagnosticFrameworkNotification()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/DiagnosticCommandImpl/addNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotificationListener(Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/DiagnosticCommandImpl/removeNotificationListener(Ljavax/management/NotificationListener;)V"(thread: JVMThread, args: [javax_management_NotificationListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/DiagnosticCommandImpl/removeNotificationListener(Ljavax/management/NotificationListener;Ljavax/management/NotificationFilter;Ljava/lang/Object;)V"(thread: JVMThread, args: [javax_management_NotificationListener, javax_management_NotificationFilter, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "setNotificationEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/DiagnosticCommandImpl/setNotificationEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDiagnosticCommands()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/management/DiagnosticCommandImpl/getDiagnosticCommands()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getDiagnosticCommandInfo([Ljava/lang/String;)[Lsun/management/DiagnosticCommandInfo;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_management_DiagnosticCommandInfo>) => void): void;
    public "sun/management/DiagnosticCommandImpl/getDiagnosticCommandInfo([Ljava/lang/String;)[Lsun/management/DiagnosticCommandInfo;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_management_DiagnosticCommandInfo>) => void): void;
    public "executeDiagnosticCommand(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/DiagnosticCommandImpl/executeDiagnosticCommand(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000(Lsun/management/DiagnosticCommandImpl;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [sun_management_DiagnosticCommandImpl, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/management/DiagnosticCommandImpl/access$000(Lsun/management/DiagnosticCommandImpl;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [sun_management_DiagnosticCommandImpl, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/DiagnosticCommandImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_management_DiagnosticCommandImpl$Wrapper extends java_lang_Object {
    public "sun/management/DiagnosticCommandImpl$Wrapper/name": java_lang_String;
    public "sun/management/DiagnosticCommandImpl$Wrapper/cmd": java_lang_String;
    public "sun/management/DiagnosticCommandImpl$Wrapper/info": sun_management_DiagnosticCommandInfo;
    public "sun/management/DiagnosticCommandImpl$Wrapper/permission": java_security_Permission;
    public "sun/management/DiagnosticCommandImpl$Wrapper/this$0": sun_management_DiagnosticCommandImpl;
    public "<init>(Lsun/management/DiagnosticCommandImpl;Ljava/lang/String;Ljava/lang/String;Lsun/management/DiagnosticCommandInfo;)V"(thread: JVMThread, args: [sun_management_DiagnosticCommandImpl, java_lang_String, java_lang_String, sun_management_DiagnosticCommandInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/DiagnosticCommandImpl$Wrapper/<init>(Lsun/management/DiagnosticCommandImpl;Ljava/lang/String;Ljava/lang/String;Lsun/management/DiagnosticCommandInfo;)V"(thread: JVMThread, args: [sun_management_DiagnosticCommandImpl, java_lang_String, java_lang_String, sun_management_DiagnosticCommandInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "execute([Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/management/DiagnosticCommandImpl$Wrapper/execute([Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface com_sun_management_DiagnosticCommandMBean extends java_lang_Object, javax_management_DynamicMBean {
    "getAttribute(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "setAttribute(Ljavax/management/Attribute;)V"(thread: JVMThread, args: [javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    "getAttributes([Ljava/lang/String;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    "setAttributes(Ljavax/management/AttributeList;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    "invoke(Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getMBeanInfo()Ljavax/management/MBeanInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_MBeanInfo) => void): void;
  }
  export interface javax_management_DynamicMBean extends java_lang_Object {
    "getAttribute(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "setAttribute(Ljavax/management/Attribute;)V"(thread: JVMThread, args: [javax_management_Attribute], cb?: (e?: java_lang_Throwable) => void): void;
    "getAttributes([Ljava/lang/String;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    "setAttributes(Ljavax/management/AttributeList;)Ljavax/management/AttributeList;"(thread: JVMThread, args: [javax_management_AttributeList], cb?: (e?: java_lang_Throwable, rv?: javax_management_AttributeList) => void): void;
    "invoke(Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getMBeanInfo()Ljavax/management/MBeanInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_MBeanInfo) => void): void;
  }
  export class sun_management_ClassLoadingImpl extends java_lang_Object implements java_lang_management_ClassLoadingMXBean {
    public "sun/management/ClassLoadingImpl/jvm": sun_management_VMManagement;
    public "<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/ClassLoadingImpl/<init>(Lsun/management/VMManagement;)V"(thread: JVMThread, args: [sun_management_VMManagement], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTotalLoadedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/ClassLoadingImpl/getTotalLoadedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLoadedClassCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ClassLoadingImpl/getLoadedClassCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUnloadedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/management/ClassLoadingImpl/getUnloadedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "isVerbose()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/management/ClassLoadingImpl/isVerbose()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setVerbose(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/management/ClassLoadingImpl/setVerbose(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setVerboseClass(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/management/ClassLoadingImpl/setVerboseClass(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
    public "sun/management/ClassLoadingImpl/getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export interface java_lang_management_ClassLoadingMXBean extends java_lang_Object, java_lang_management_PlatformManagedObject {
    "getTotalLoadedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getLoadedClassCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getUnloadedClassCount()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "isVerbose()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVerbose(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getObjectName()Ljavax/management/ObjectName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_management_ObjectName) => void): void;
  }
  export class sun_font_GlyphList extends java_lang_Object {
    public static "sun/font/GlyphList/MINGRAYLENGTH": number;
    public static "sun/font/GlyphList/MAXGRAYLENGTH": number;
    public static "sun/font/GlyphList/DEFAULT_LENGTH": number;
    public "sun/font/GlyphList/glyphindex": number;
    public "sun/font/GlyphList/metrics": JVMArray<number>;
    public "sun/font/GlyphList/graybits": JVMArray<number>;
    public "sun/font/GlyphList/strikelist": java_lang_Object;
    public "sun/font/GlyphList/len": number;
    public "sun/font/GlyphList/maxLen": number;
    public "sun/font/GlyphList/maxPosLen": number;
    public "sun/font/GlyphList/glyphData": JVMArray<number>;
    public "sun/font/GlyphList/chData": JVMArray<number>;
    public "sun/font/GlyphList/images": JVMArray<Long>;
    public "sun/font/GlyphList/positions": JVMArray<number>;
    public "sun/font/GlyphList/x": number;
    public "sun/font/GlyphList/y": number;
    public "sun/font/GlyphList/gposx": number;
    public "sun/font/GlyphList/gposy": number;
    public "sun/font/GlyphList/usePositions": number;
    public "sun/font/GlyphList/lcdRGBOrder": number;
    public "sun/font/GlyphList/lcdSubPixPos": number;
    public static "sun/font/GlyphList/reusableGL": sun_font_GlyphList;
    public static "sun/font/GlyphList/inUse": number;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphList/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance()Lsun/font/GlyphList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphList) => void): void;
    public static "sun/font/GlyphList/getInstance()Lsun/font/GlyphList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphList) => void): void;
    public "setFromString(Lsun/java2d/loops/FontInfo;Ljava/lang/String;FF)Z"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphList/setFromString(Lsun/java2d/loops/FontInfo;Ljava/lang/String;FF)Z"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFromChars(Lsun/java2d/loops/FontInfo;[CIIFF)Z"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphList/setFromChars(Lsun/java2d/loops/FontInfo;[CIIFF)Z"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mapChars(Lsun/java2d/loops/FontInfo;I)Z"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphList/mapChars(Lsun/java2d/loops/FontInfo;I)Z"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFromGlyphVector(Lsun/java2d/loops/FontInfo;Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphList/setFromGlyphVector(Lsun/java2d/loops/FontInfo;Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBounds()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/GlyphList/getBounds()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "setGlyphIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphList/setGlyphIndex(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMetrics()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/GlyphList/getMetrics()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getGrayBits()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/GlyphList/getGrayBits()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getImages()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "sun/font/GlyphList/getImages()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "usePositions()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphList/usePositions()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPositions()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/GlyphList/getPositions()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getX()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphList/getX()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphList/getY()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getStrike()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/GlyphList/getStrike()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isSubPixPos()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphList/isSubPixPos()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRGBOrder()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphList/isRGBOrder()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphList/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphList/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fillBounds([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphList/fillBounds([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/GlyphList/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_font_GlyphVector extends java_lang_Object implements java_lang_Cloneable {
    public static "java/awt/font/GlyphVector/FLAG_HAS_TRANSFORMS": number;
    public static "java/awt/font/GlyphVector/FLAG_HAS_POSITION_ADJUSTMENTS": number;
    public static "java/awt/font/GlyphVector/FLAG_RUN_RTL": number;
    public static "java/awt/font/GlyphVector/FLAG_COMPLEX_GLYPHS": number;
    public static "java/awt/font/GlyphVector/FLAG_MASK": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/GlyphVector/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/font/GlyphVector/getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "java/awt/font/GlyphVector/getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "performDefaultLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/GlyphVector/performDefaultLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphVector/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphVector/getGlyphCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphCodes(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/font/GlyphVector/getGlyphCodes(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getGlyphCharIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphVector/getGlyphCharIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphCharIndices(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/font/GlyphVector/getGlyphCharIndices(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getLogicalBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/font/GlyphVector/getLogicalBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getVisualBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/font/GlyphVector/getVisualBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getPixelBounds(Ljava/awt/font/FontRenderContext;FF)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/font/GlyphVector/getPixelBounds(Ljava/awt/font/FontRenderContext;FF)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getOutline()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/font/GlyphVector/getOutline()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getOutline(FF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/font/GlyphVector/getOutline(FF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphOutline(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/font/GlyphVector/getGlyphOutline(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphOutline(IFF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/font/GlyphVector/getGlyphOutline(IFF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphPosition(I)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/font/GlyphVector/getGlyphPosition(I)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "setGlyphPosition(ILjava/awt/geom/Point2D;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/GlyphVector/setGlyphPosition(ILjava/awt/geom/Point2D;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphTransform(I)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "java/awt/font/GlyphVector/getGlyphTransform(I)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "setGlyphTransform(ILjava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [number, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/GlyphVector/setGlyphTransform(ILjava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [number, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLayoutFlags()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphVector/getLayoutFlags()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphPositions(II[F)[F"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/font/GlyphVector/getGlyphPositions(II[F)[F"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getGlyphLogicalBounds(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/font/GlyphVector/getGlyphLogicalBounds(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphVisualBounds(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/font/GlyphVector/getGlyphVisualBounds(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphPixelBounds(ILjava/awt/font/FontRenderContext;FF)Ljava/awt/Rectangle;"(thread: JVMThread, args: [number, java_awt_font_FontRenderContext, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/font/GlyphVector/getGlyphPixelBounds(ILjava/awt/font/FontRenderContext;FF)Ljava/awt/Rectangle;"(thread: JVMThread, args: [number, java_awt_font_FontRenderContext, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getGlyphMetrics(I)Ljava/awt/font/GlyphMetrics;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphMetrics) => void): void;
    public "java/awt/font/GlyphVector/getGlyphMetrics(I)Ljava/awt/font/GlyphMetrics;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphMetrics) => void): void;
    public "getGlyphJustificationInfo(I)Ljava/awt/font/GlyphJustificationInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphJustificationInfo) => void): void;
    public "java/awt/font/GlyphVector/getGlyphJustificationInfo(I)Ljava/awt/font/GlyphJustificationInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphJustificationInfo) => void): void;
    public "equals(Ljava/awt/font/GlyphVector;)Z"(thread: JVMThread, args: [java_awt_font_GlyphVector], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphVector/equals(Ljava/awt/font/GlyphVector;)Z"(thread: JVMThread, args: [java_awt_font_GlyphVector], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_font_GlyphJustificationInfo extends java_lang_Object {
    public static "java/awt/font/GlyphJustificationInfo/PRIORITY_KASHIDA": number;
    public static "java/awt/font/GlyphJustificationInfo/PRIORITY_WHITESPACE": number;
    public static "java/awt/font/GlyphJustificationInfo/PRIORITY_INTERCHAR": number;
    public static "java/awt/font/GlyphJustificationInfo/PRIORITY_NONE": number;
    public "java/awt/font/GlyphJustificationInfo/weight": number;
    public "java/awt/font/GlyphJustificationInfo/growPriority": number;
    public "java/awt/font/GlyphJustificationInfo/growAbsorb": number;
    public "java/awt/font/GlyphJustificationInfo/growLeftLimit": number;
    public "java/awt/font/GlyphJustificationInfo/growRightLimit": number;
    public "java/awt/font/GlyphJustificationInfo/shrinkPriority": number;
    public "java/awt/font/GlyphJustificationInfo/shrinkAbsorb": number;
    public "java/awt/font/GlyphJustificationInfo/shrinkLeftLimit": number;
    public "java/awt/font/GlyphJustificationInfo/shrinkRightLimit": number;
    public "<init>(FZIFFZIFF)V"(thread: JVMThread, args: [number, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/GlyphJustificationInfo/<init>(FZIFFZIFF)V"(thread: JVMThread, args: [number, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "priorityIsValid(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/GlyphJustificationInfo/priorityIsValid(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_font_GlyphMetrics extends java_lang_Object {
    public "java/awt/font/GlyphMetrics/horizontal": number;
    public "java/awt/font/GlyphMetrics/advanceX": number;
    public "java/awt/font/GlyphMetrics/advanceY": number;
    public "java/awt/font/GlyphMetrics/bounds": java_awt_geom_Rectangle2D$Float;
    public "java/awt/font/GlyphMetrics/glyphType": number;
    public static "java/awt/font/GlyphMetrics/STANDARD": number;
    public static "java/awt/font/GlyphMetrics/LIGATURE": number;
    public static "java/awt/font/GlyphMetrics/COMBINING": number;
    public static "java/awt/font/GlyphMetrics/COMPONENT": number;
    public static "java/awt/font/GlyphMetrics/WHITESPACE": number;
    public "<init>(FLjava/awt/geom/Rectangle2D;B)V"(thread: JVMThread, args: [number, java_awt_geom_Rectangle2D, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/GlyphMetrics/<init>(FLjava/awt/geom/Rectangle2D;B)V"(thread: JVMThread, args: [number, java_awt_geom_Rectangle2D, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ZFFLjava/awt/geom/Rectangle2D;B)V"(thread: JVMThread, args: [number, number, number, java_awt_geom_Rectangle2D, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/GlyphMetrics/<init>(ZFFLjava/awt/geom/Rectangle2D;B)V"(thread: JVMThread, args: [number, number, number, java_awt_geom_Rectangle2D, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAdvance()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/getAdvance()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAdvanceX()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/getAdvanceX()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAdvanceY()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/getAdvanceY()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/font/GlyphMetrics/getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getLSB()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/getLSB()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRSB()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/getRSB()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isStandard()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/isStandard()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLigature()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/isLigature()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isCombining()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/isCombining()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isComponent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/isComponent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isWhitespace()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GlyphMetrics/isWhitespace()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_geom_AffineTransform extends java_lang_Object implements java_lang_Cloneable, java_io_Serializable {
    public static "java/awt/geom/AffineTransform/TYPE_UNKNOWN": number;
    public static "java/awt/geom/AffineTransform/TYPE_IDENTITY": number;
    public static "java/awt/geom/AffineTransform/TYPE_TRANSLATION": number;
    public static "java/awt/geom/AffineTransform/TYPE_UNIFORM_SCALE": number;
    public static "java/awt/geom/AffineTransform/TYPE_GENERAL_SCALE": number;
    public static "java/awt/geom/AffineTransform/TYPE_MASK_SCALE": number;
    public static "java/awt/geom/AffineTransform/TYPE_FLIP": number;
    public static "java/awt/geom/AffineTransform/TYPE_QUADRANT_ROTATION": number;
    public static "java/awt/geom/AffineTransform/TYPE_GENERAL_ROTATION": number;
    public static "java/awt/geom/AffineTransform/TYPE_MASK_ROTATION": number;
    public static "java/awt/geom/AffineTransform/TYPE_GENERAL_TRANSFORM": number;
    public static "java/awt/geom/AffineTransform/APPLY_IDENTITY": number;
    public static "java/awt/geom/AffineTransform/APPLY_TRANSLATE": number;
    public static "java/awt/geom/AffineTransform/APPLY_SCALE": number;
    public static "java/awt/geom/AffineTransform/APPLY_SHEAR": number;
    public static "java/awt/geom/AffineTransform/HI_SHIFT": number;
    public static "java/awt/geom/AffineTransform/HI_IDENTITY": number;
    public static "java/awt/geom/AffineTransform/HI_TRANSLATE": number;
    public static "java/awt/geom/AffineTransform/HI_SCALE": number;
    public static "java/awt/geom/AffineTransform/HI_SHEAR": number;
    public "java/awt/geom/AffineTransform/m00": number;
    public "java/awt/geom/AffineTransform/m10": number;
    public "java/awt/geom/AffineTransform/m01": number;
    public "java/awt/geom/AffineTransform/m11": number;
    public "java/awt/geom/AffineTransform/m02": number;
    public "java/awt/geom/AffineTransform/m12": number;
    public "java/awt/geom/AffineTransform/state": number;
    public "java/awt/geom/AffineTransform/type": number;
    public static "java/awt/geom/AffineTransform/rot90conversion": JVMArray<number>;
    public static "java/awt/geom/AffineTransform/serialVersionUID": Long;
    public "<init>(DDDDDDI)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/<init>(DDDDDDI)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/<init>(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(FFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/<init>(FFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([F)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/<init>([F)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(DDDDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/<init>(DDDDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([D)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/<init>([D)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getTranslateInstance(DD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "java/awt/geom/AffineTransform/getTranslateInstance(DD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getRotateInstance(D)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "java/awt/geom/AffineTransform/getRotateInstance(D)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getRotateInstance(DDD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "java/awt/geom/AffineTransform/getRotateInstance(DDD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getRotateInstance(DD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "java/awt/geom/AffineTransform/getRotateInstance(DD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getRotateInstance(DDDD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "java/awt/geom/AffineTransform/getRotateInstance(DDDD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getQuadrantRotateInstance(I)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "java/awt/geom/AffineTransform/getQuadrantRotateInstance(I)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getQuadrantRotateInstance(IDD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "java/awt/geom/AffineTransform/getQuadrantRotateInstance(IDD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getScaleInstance(DD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "java/awt/geom/AffineTransform/getScaleInstance(DD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getShearInstance(DD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "java/awt/geom/AffineTransform/getShearInstance(DD)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "calculateType()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/calculateType()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDeterminant()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/getDeterminant()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "updateState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/updateState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "stateError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/stateError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMatrix([D)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/getMatrix([D)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getScaleX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/getScaleX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getScaleY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/getScaleY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getShearX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/getShearX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getShearY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/getShearY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTranslateX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/getTranslateX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTranslateY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/getTranslateY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "translate(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/translate(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate90()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/rotate90()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate180()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/rotate180()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate270()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/rotate270()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/rotate(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate(DDD)V"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/rotate(DDD)V"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/rotate(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/rotate(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "quadrantRotate(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/quadrantRotate(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "quadrantRotate(IDD)V"(thread: JVMThread, args: [number, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/quadrantRotate(IDD)V"(thread: JVMThread, args: [number, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "scale(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/scale(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "shear(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/shear(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToIdentity()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToIdentity()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToTranslation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToTranslation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToRotation(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToRotation(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToRotation(DDD)V"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToRotation(DDD)V"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToRotation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToRotation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToRotation(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToRotation(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToQuadrantRotation(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToQuadrantRotation(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToQuadrantRotation(IDD)V"(thread: JVMThread, args: [number, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToQuadrantRotation(IDD)V"(thread: JVMThread, args: [number, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToScale(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToScale(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setToShear(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setToShear(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTransform(DDDDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/setTransform(DDDDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "concatenate(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/concatenate(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "preConcatenate(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/preConcatenate(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "createInverse()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "java/awt/geom/AffineTransform/createInverse()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "invert()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/invert()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "transform(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/geom/AffineTransform/transform(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "transform([Ljava/awt/geom/Point2D;I[Ljava/awt/geom/Point2D;II)V"(thread: JVMThread, args: [JVMArray<java_awt_geom_Point2D>, number, JVMArray<java_awt_geom_Point2D>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/transform([Ljava/awt/geom/Point2D;I[Ljava/awt/geom/Point2D;II)V"(thread: JVMThread, args: [JVMArray<java_awt_geom_Point2D>, number, JVMArray<java_awt_geom_Point2D>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "transform([FI[FII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/transform([FI[FII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "transform([DI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/transform([DI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "transform([FI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/transform([FI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "transform([DI[FII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/transform([DI[FII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "inverseTransform(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/geom/AffineTransform/inverseTransform(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "inverseTransform([DI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/inverseTransform([DI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "deltaTransform(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/geom/AffineTransform/deltaTransform(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "deltaTransform([DI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/deltaTransform([DI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "createTransformedShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/geom/AffineTransform/createTransformedShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public static "_matround(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/AffineTransform/_matround(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/geom/AffineTransform/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isIdentity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/isIdentity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/geom/AffineTransform/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/AffineTransform/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/AffineTransform/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/geom/AffineTransform/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_geom_Point2D extends java_lang_Object implements java_lang_Cloneable {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D/getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D/getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLocation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D/setLocation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocation(Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D/setLocation(Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public static "distanceSq(DDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Point2D/distanceSq(DDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "distance(DDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Point2D/distance(DDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "distanceSq(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D/distanceSq(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "distanceSq(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D/distanceSq(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "distance(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D/distance(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "distance(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D/distance(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/geom/Point2D/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_Shape extends java_lang_Object {
    "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    "contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "intersects(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    "getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
  }
  export interface java_awt_geom_PathIterator extends java_lang_Object {
    "getWindingRule()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isDone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "next()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "currentSegment([F)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "currentSegment([D)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Rectangle extends java_awt_geom_Rectangle2D implements java_awt_Shape, java_io_Serializable {
    public "java/awt/Rectangle/x": number;
    public "java/awt/Rectangle/y": number;
    public "java/awt/Rectangle/width": number;
    public "java/awt/Rectangle/height": number;
    public static "java/awt/Rectangle/serialVersionUID": Long;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Rectangle/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/<init>(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Point;Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Point, java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/<init>(Ljava/awt/Point;Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Point, java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/<init>(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/<init>(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Rectangle/getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/Rectangle/getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "setBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/setBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBounds(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/setBounds(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRect(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/setRect(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "clip(DZ)I"(thread: JVMThread, args: [number, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Rectangle/clip(DZ)I"(thread: JVMThread, args: [number, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reshape(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/reshape(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Rectangle/getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "setLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/setLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocation(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/setLocation(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "move(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/move(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Rectangle/getSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "setSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/setSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/setSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "resize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/resize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "contains(Ljava/awt/Point;)Z"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/contains(Ljava/awt/Point;)Z"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/contains(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/awt/Rectangle;)Z"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/contains(Ljava/awt/Rectangle;)Z"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/contains(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inside(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/inside(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(Ljava/awt/Rectangle;)Z"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/intersects(Ljava/awt/Rectangle;)Z"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersection(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Rectangle/intersection(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "union(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Rectangle/union(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "add(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/add(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/add(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/add(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "grow(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Rectangle/grow(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "outcode(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/outcode(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createIntersection(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/Rectangle/createIntersection(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "createUnion(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/Rectangle/createUnion(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Rectangle/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Rectangle/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Rectangle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Dimension extends java_awt_geom_Dimension2D implements java_io_Serializable {
    public "java/awt/Dimension/width": number;
    public "java/awt/Dimension/height": number;
    public static "java/awt/Dimension/serialVersionUID": Long;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Dimension/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dimension/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dimension/<init>(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dimension/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dimension/getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dimension/getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSize(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dimension/setSize(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Dimension/getSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "setSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dimension/setSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dimension/setSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dimension/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dimension/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Dimension/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Dimension/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_geom_Dimension2D extends java_lang_Object implements java_lang_Cloneable {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Dimension2D/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Dimension2D/getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Dimension2D/getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSize(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Dimension2D/setSize(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSize(Ljava/awt/geom/Dimension2D;)V"(thread: JVMThread, args: [java_awt_geom_Dimension2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Dimension2D/setSize(Ljava/awt/geom/Dimension2D;)V"(thread: JVMThread, args: [java_awt_geom_Dimension2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/geom/Dimension2D/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_awt_Point extends java_awt_geom_Point2D implements java_io_Serializable {
    public "java/awt/Point/x": number;
    public "java/awt/Point/y": number;
    public static "java/awt/Point/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Point/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Point/<init>(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Point/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Point/getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Point/getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Point/getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "setLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Point/setLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocation(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Point/setLocation(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Point/setLocation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "move(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Point/move(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Point/translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Point/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Point/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_geom_Rectangle2D extends java_awt_geom_RectangularShape {
    public static "java/awt/geom/Rectangle2D/OUT_LEFT": number;
    public static "java/awt/geom/Rectangle2D/OUT_TOP": number;
    public static "java/awt/geom/Rectangle2D/OUT_RIGHT": number;
    public static "java/awt/geom/Rectangle2D/OUT_BOTTOM": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRect(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D/setRect(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRect(Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D/setRect(Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "intersectsLine(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D/intersectsLine(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersectsLine(Ljava/awt/geom/Line2D;)Z"(thread: JVMThread, args: [java_awt_geom_Line2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D/intersectsLine(Ljava/awt/geom/Line2D;)Z"(thread: JVMThread, args: [java_awt_geom_Line2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "outcode(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D/outcode(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "outcode(Ljava/awt/geom/Point2D;)I"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D/outcode(Ljava/awt/geom/Point2D;)I"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFrame(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D/setFrame(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/geom/Rectangle2D/getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D/contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D/intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D/contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createIntersection(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/geom/Rectangle2D/createIntersection(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public static "intersect(Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D, java_awt_geom_Rectangle2D, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/geom/Rectangle2D/intersect(Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D, java_awt_geom_Rectangle2D, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "createUnion(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/geom/Rectangle2D/createUnion(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public static "union(Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D, java_awt_geom_Rectangle2D, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/geom/Rectangle2D/union(Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D, java_awt_geom_Rectangle2D, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D/add(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D/add(Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D/add(Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "java/awt/geom/Rectangle2D/getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "java/awt/geom/Rectangle2D/getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_geom_Line2D extends java_lang_Object implements java_awt_Shape, java_lang_Cloneable {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Line2D/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX1()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/getX1()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY1()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/getY1()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getP1()Ljava/awt/geom/Point2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/geom/Line2D/getP1()Ljava/awt/geom/Point2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "getX2()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/getX2()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY2()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/getY2()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getP2()Ljava/awt/geom/Point2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/geom/Line2D/getP2()Ljava/awt/geom/Point2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "setLine(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Line2D/setLine(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLine(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Line2D/setLine(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLine(Ljava/awt/geom/Line2D;)V"(thread: JVMThread, args: [java_awt_geom_Line2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Line2D/setLine(Ljava/awt/geom/Line2D;)V"(thread: JVMThread, args: [java_awt_geom_Line2D], cb?: (e?: java_lang_Throwable) => void): void;
    public static "relativeCCW(DDDDDD)I"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Line2D/relativeCCW(DDDDDD)I"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "relativeCCW(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/relativeCCW(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "relativeCCW(Ljava/awt/geom/Point2D;)I"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/relativeCCW(Ljava/awt/geom/Point2D;)I"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "linesIntersect(DDDDDDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Line2D/linesIntersect(DDDDDDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersectsLine(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/intersectsLine(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersectsLine(Ljava/awt/geom/Line2D;)Z"(thread: JVMThread, args: [java_awt_geom_Line2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/intersectsLine(Ljava/awt/geom/Line2D;)Z"(thread: JVMThread, args: [java_awt_geom_Line2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ptSegDistSq(DDDDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Line2D/ptSegDistSq(DDDDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ptSegDist(DDDDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Line2D/ptSegDist(DDDDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ptSegDistSq(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/ptSegDistSq(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ptSegDistSq(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/ptSegDistSq(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ptSegDist(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/ptSegDist(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ptSegDist(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/ptSegDist(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ptLineDistSq(DDDDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Line2D/ptLineDistSq(DDDDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ptLineDist(DDDDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Line2D/ptLineDist(DDDDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ptLineDistSq(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/ptLineDistSq(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ptLineDistSq(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/ptLineDistSq(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ptLineDist(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/ptLineDist(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ptLineDist(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/ptLineDist(Ljava/awt/geom/Point2D;)D"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/contains(Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/intersects(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Line2D/contains(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/geom/Line2D/getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "java/awt/geom/Line2D/getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "java/awt/geom/Line2D/getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/geom/Line2D/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
  }
  export class java_awt_geom_RectangularShape extends java_lang_Object implements java_awt_Shape, java_lang_Cloneable {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/RectangularShape/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getMinX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getMinY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaxX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getMaxX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaxY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getMaxY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCenterX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getCenterX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCenterY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/getCenterY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFrame()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/geom/RectangularShape/getFrame()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFrame(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/RectangularShape/setFrame(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFrame(Ljava/awt/geom/Point2D;Ljava/awt/geom/Dimension2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Dimension2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/RectangularShape/setFrame(Ljava/awt/geom/Point2D;Ljava/awt/geom/Dimension2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Dimension2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFrame(Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/RectangularShape/setFrame(Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFrameFromDiagonal(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/RectangularShape/setFrameFromDiagonal(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFrameFromDiagonal(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/RectangularShape/setFrameFromDiagonal(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFrameFromCenter(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/RectangularShape/setFrameFromCenter(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFrameFromCenter(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/RectangularShape/setFrameFromCenter(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)V"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "contains(Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/contains(Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/intersects(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/RectangularShape/contains(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/geom/RectangularShape/getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "java/awt/geom/RectangularShape/getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/geom/RectangularShape/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    "contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
  }
  export class java_awt_font_FontRenderContext extends java_lang_Object {
    public "java/awt/font/FontRenderContext/tx": java_awt_geom_AffineTransform;
    public "java/awt/font/FontRenderContext/aaHintValue": java_lang_Object;
    public "java/awt/font/FontRenderContext/fmHintValue": java_lang_Object;
    public "java/awt/font/FontRenderContext/defaulting": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/FontRenderContext/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/geom/AffineTransform;ZZ)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/FontRenderContext/<init>(Ljava/awt/geom/AffineTransform;ZZ)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/geom/AffineTransform;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/FontRenderContext/<init>(Ljava/awt/geom/AffineTransform;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "isTransformed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/FontRenderContext/isTransformed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransformType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/FontRenderContext/getTransformType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "java/awt/font/FontRenderContext/getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "isAntiAliased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/FontRenderContext/isAntiAliased()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "usesFractionalMetrics()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/FontRenderContext/usesFractionalMetrics()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAntiAliasingHint()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/font/FontRenderContext/getAntiAliasingHint()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getFractionalMetricsHint()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/font/FontRenderContext/getFractionalMetricsHint()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/FontRenderContext/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/awt/font/FontRenderContext;)Z"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/FontRenderContext/equals(Ljava/awt/font/FontRenderContext;)Z"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/FontRenderContext/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Font extends java_lang_Object implements java_io_Serializable {
    public "java/awt/Font/fRequestedAttributes": java_util_Hashtable;
    public static "java/awt/Font/DIALOG": java_lang_String;
    public static "java/awt/Font/DIALOG_INPUT": java_lang_String;
    public static "java/awt/Font/SANS_SERIF": java_lang_String;
    public static "java/awt/Font/SERIF": java_lang_String;
    public static "java/awt/Font/MONOSPACED": java_lang_String;
    public static "java/awt/Font/PLAIN": number;
    public static "java/awt/Font/BOLD": number;
    public static "java/awt/Font/ITALIC": number;
    public static "java/awt/Font/ROMAN_BASELINE": number;
    public static "java/awt/Font/CENTER_BASELINE": number;
    public static "java/awt/Font/HANGING_BASELINE": number;
    public static "java/awt/Font/TRUETYPE_FONT": number;
    public static "java/awt/Font/TYPE1_FONT": number;
    public "java/awt/Font/name": java_lang_String;
    public "java/awt/Font/style": number;
    public "java/awt/Font/size": number;
    public "java/awt/Font/pointSize": number;
    public "java/awt/Font/peer": java_awt_peer_FontPeer;
    public "java/awt/Font/pData": Long;
    public "java/awt/Font/font2DHandle": sun_font_Font2DHandle;
    public "java/awt/Font/values": sun_font_AttributeValues;
    public "java/awt/Font/hasLayoutAttributes": number;
    public "java/awt/Font/createdFont": number;
    public "java/awt/Font/nonIdentityTx": number;
    public static "java/awt/Font/identityTx": java_awt_geom_AffineTransform;
    public static "java/awt/Font/serialVersionUID": Long;
    public static "java/awt/Font/RECOGNIZED_MASK": number;
    public static "java/awt/Font/PRIMARY_MASK": number;
    public static "java/awt/Font/SECONDARY_MASK": number;
    public static "java/awt/Font/LAYOUT_MASK": number;
    public static "java/awt/Font/EXTRA_MASK": number;
    public static "java/awt/Font/ssinfo": JVMArray<number>;
    public "java/awt/Font/hash": number;
    public "java/awt/Font/fontSerializedDataVersion": number;
    public "java/awt/Font/flmref": java_lang_ref_SoftReference;
    public static "java/awt/Font/LAYOUT_LEFT_TO_RIGHT": number;
    public static "java/awt/Font/LAYOUT_RIGHT_TO_LEFT": number;
    public static "java/awt/Font/LAYOUT_NO_START_CONTEXT": number;
    public static "java/awt/Font/LAYOUT_NO_LIMIT_CONTEXT": number;
    public "getPeer()Ljava/awt/peer/FontPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FontPeer) => void): void;
    public "java/awt/Font/getPeer()Ljava/awt/peer/FontPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FontPeer) => void): void;
    public "getPeer_NoClientCode()Ljava/awt/peer/FontPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FontPeer) => void): void;
    public "java/awt/Font/getPeer_NoClientCode()Ljava/awt/peer/FontPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FontPeer) => void): void;
    public "getAttributeValues()Lsun/font/AttributeValues;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "java/awt/Font/getAttributeValues()Lsun/font/AttributeValues;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "getFont2D()Lsun/font/Font2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "java/awt/Font/getFont2D()Lsun/font/Font2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;IF)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/<init>(Ljava/lang/String;IF)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;IFZLsun/font/Font2DHandle;)V"(thread: JVMThread, args: [java_lang_String, number, number, number, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/<init>(Ljava/lang/String;IFZLsun/font/Font2DHandle;)V"(thread: JVMThread, args: [java_lang_String, number, number, number, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;IZLsun/font/CreatedFontTracker;)V"(thread: JVMThread, args: [java_io_File, number, number, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/<init>(Ljava/io/File;IZLsun/font/CreatedFontTracker;)V"(thread: JVMThread, args: [java_io_File, number, number, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/AttributeValues;Ljava/lang/String;IZLsun/font/Font2DHandle;)V"(thread: JVMThread, args: [sun_font_AttributeValues, java_lang_String, number, number, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/<init>(Lsun/font/AttributeValues;Ljava/lang/String;IZLsun/font/Font2DHandle;)V"(thread: JVMThread, args: [sun_font_AttributeValues, java_lang_String, number, number, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/<init>(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "initFromValues(Lsun/font/AttributeValues;)V"(thread: JVMThread, args: [sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/initFromValues(Lsun/font/AttributeValues;)V"(thread: JVMThread, args: [sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getFont(Ljava/util/Map;)Ljava/awt/Font;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "java/awt/Font/getFont(Ljava/util/Map;)Ljava/awt/Font;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "hasTempPermission()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Font/hasTempPermission()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "createFont(ILjava/io/InputStream;)Ljava/awt/Font;"(thread: JVMThread, args: [number, java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "java/awt/Font/createFont(ILjava/io/InputStream;)Ljava/awt/Font;"(thread: JVMThread, args: [number, java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "createFont0(ILjava/io/InputStream;Lsun/font/CreatedFontTracker;)Ljava/awt/Font;"(thread: JVMThread, args: [number, java_io_InputStream, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "java/awt/Font/createFont0(ILjava/io/InputStream;Lsun/font/CreatedFontTracker;)Ljava/awt/Font;"(thread: JVMThread, args: [number, java_io_InputStream, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "createFont(ILjava/io/File;)Ljava/awt/Font;"(thread: JVMThread, args: [number, java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "java/awt/Font/createFont(ILjava/io/File;)Ljava/awt/Font;"(thread: JVMThread, args: [number, java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "java/awt/Font/getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "getFamily()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Font/getFamily()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFamily_NoClientCode()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Font/getFamily_NoClientCode()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFamily(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Font/getFamily(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPSName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Font/getPSName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Font/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFontName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Font/getFontName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFontName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Font/getFontName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getStyle()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/getStyle()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/getSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSize2D()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/getSize2D()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPlain()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/isPlain()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBold()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/isBold()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isItalic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/isItalic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isTransformed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/isTransformed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasLayoutAttributes()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/hasLayoutAttributes()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFont(Ljava/lang/String;)Ljava/awt/Font;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "java/awt/Font/getFont(Ljava/lang/String;)Ljava/awt/Font;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "decode(Ljava/lang/String;)Ljava/awt/Font;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "java/awt/Font/decode(Ljava/lang/String;)Ljava/awt/Font;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "getFont(Ljava/lang/String;Ljava/awt/Font;)Ljava/awt/Font;"(thread: JVMThread, args: [java_lang_String, java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "java/awt/Font/getFont(Ljava/lang/String;Ljava/awt/Font;)Ljava/awt/Font;"(thread: JVMThread, args: [java_lang_String, java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Font/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Font/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBaselineFor(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/getBaselineFor(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAttributes()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/awt/Font/getAttributes()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getAvailableAttributes()[Ljava/text/AttributedCharacterIterator$Attribute;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_text_AttributedCharacterIterator$Attribute>) => void): void;
    public "java/awt/Font/getAvailableAttributes()[Ljava/text/AttributedCharacterIterator$Attribute;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_text_AttributedCharacterIterator$Attribute>) => void): void;
    public "deriveFont(IF)Ljava/awt/Font;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/Font/deriveFont(IF)Ljava/awt/Font;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "deriveFont(ILjava/awt/geom/AffineTransform;)Ljava/awt/Font;"(thread: JVMThread, args: [number, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/Font/deriveFont(ILjava/awt/geom/AffineTransform;)Ljava/awt/Font;"(thread: JVMThread, args: [number, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "deriveFont(F)Ljava/awt/Font;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/Font/deriveFont(F)Ljava/awt/Font;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "deriveFont(Ljava/awt/geom/AffineTransform;)Ljava/awt/Font;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/Font/deriveFont(Ljava/awt/geom/AffineTransform;)Ljava/awt/Font;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "deriveFont(I)Ljava/awt/Font;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/Font/deriveFont(I)Ljava/awt/Font;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "deriveFont(Ljava/util/Map;)Ljava/awt/Font;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/Font/deriveFont(Ljava/util/Map;)Ljava/awt/Font;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "canDisplay(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/canDisplay(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDisplay(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/canDisplay(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDisplayUpTo(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/canDisplayUpTo(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDisplayUpTo([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/canDisplayUpTo([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDisplayUpTo(Ljava/text/CharacterIterator;II)I"(thread: JVMThread, args: [java_text_CharacterIterator, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/canDisplayUpTo(Ljava/text/CharacterIterator;II)I"(thread: JVMThread, args: [java_text_CharacterIterator, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getItalicAngle()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/getItalicAngle()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getItalicAngle(Ljava/awt/font/FontRenderContext;)F"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/getItalicAngle(Ljava/awt/font/FontRenderContext;)F"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasUniformLineMetrics()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Font/hasUniformLineMetrics()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "defaultLineMetrics(Ljava/awt/font/FontRenderContext;)Lsun/font/FontLineMetrics;"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontLineMetrics) => void): void;
    public "java/awt/Font/defaultLineMetrics(Ljava/awt/font/FontRenderContext;)Lsun/font/FontLineMetrics;"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontLineMetrics) => void): void;
    public "getLineMetrics(Ljava/lang/String;Ljava/awt/font/FontRenderContext;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_lang_String, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "java/awt/Font/getLineMetrics(Ljava/lang/String;Ljava/awt/font/FontRenderContext;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_lang_String, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "getLineMetrics(Ljava/lang/String;IILjava/awt/font/FontRenderContext;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_lang_String, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "java/awt/Font/getLineMetrics(Ljava/lang/String;IILjava/awt/font/FontRenderContext;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_lang_String, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "getLineMetrics([CIILjava/awt/font/FontRenderContext;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "java/awt/Font/getLineMetrics([CIILjava/awt/font/FontRenderContext;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "getLineMetrics(Ljava/text/CharacterIterator;IILjava/awt/font/FontRenderContext;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_text_CharacterIterator, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "java/awt/Font/getLineMetrics(Ljava/text/CharacterIterator;IILjava/awt/font/FontRenderContext;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_text_CharacterIterator, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "getStringBounds(Ljava/lang/String;Ljava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_lang_String, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/Font/getStringBounds(Ljava/lang/String;Ljava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_lang_String, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getStringBounds(Ljava/lang/String;IILjava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_lang_String, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/Font/getStringBounds(Ljava/lang/String;IILjava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_lang_String, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getStringBounds([CIILjava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/Font/getStringBounds([CIILjava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getStringBounds(Ljava/text/CharacterIterator;IILjava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_text_CharacterIterator, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/Font/getStringBounds(Ljava/text/CharacterIterator;IILjava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_text_CharacterIterator, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getMaxCharBounds(Ljava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/Font/getMaxCharBounds(Ljava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "createGlyphVector(Ljava/awt/font/FontRenderContext;Ljava/lang/String;)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public "java/awt/Font/createGlyphVector(Ljava/awt/font/FontRenderContext;Ljava/lang/String;)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public "createGlyphVector(Ljava/awt/font/FontRenderContext;[C)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public "java/awt/Font/createGlyphVector(Ljava/awt/font/FontRenderContext;[C)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public "createGlyphVector(Ljava/awt/font/FontRenderContext;Ljava/text/CharacterIterator;)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, java_text_CharacterIterator], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public "java/awt/Font/createGlyphVector(Ljava/awt/font/FontRenderContext;Ljava/text/CharacterIterator;)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, java_text_CharacterIterator], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public "createGlyphVector(Ljava/awt/font/FontRenderContext;[I)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public "java/awt/Font/createGlyphVector(Ljava/awt/font/FontRenderContext;[I)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public "layoutGlyphVector(Ljava/awt/font/FontRenderContext;[CIII)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public "java/awt/Font/layoutGlyphVector(Ljava/awt/font/FontRenderContext;[CIII)Ljava/awt/font/GlyphVector;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphVector) => void): void;
    public static "applyTransform(Ljava/awt/geom/AffineTransform;Lsun/font/AttributeValues;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Font/applyTransform(Ljava/awt/geom/AffineTransform;Lsun/font/AttributeValues;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public static "applyStyle(ILsun/font/AttributeValues;)V"(thread: JVMThread, args: [number, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Font/applyStyle(ILsun/font/AttributeValues;)V"(thread: JVMThread, args: [number, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Font/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/awt/Font;)Lsun/font/Font2D;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public static "java/awt/Font/access$000(Ljava/awt/Font;)Lsun/font/Font2D;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public static "access$102(Ljava/awt/Font;Lsun/font/Font2DHandle;)Lsun/font/Font2DHandle;"(thread: JVMThread, args: [java_awt_Font, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2DHandle) => void): void;
    public static "java/awt/Font/access$102(Ljava/awt/Font;Lsun/font/Font2DHandle;)Lsun/font/Font2DHandle;"(thread: JVMThread, args: [java_awt_Font, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2DHandle) => void): void;
    public static "access$202(Ljava/awt/Font;Z)Z"(thread: JVMThread, args: [java_awt_Font, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Font/access$202(Ljava/awt/Font;Z)Z"(thread: JVMThread, args: [java_awt_Font, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Ljava/awt/Font;)Z"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Font/access$200(Ljava/awt/Font;)Z"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Font/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_font_LineMetrics extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/LineMetrics/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumChars()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getNumChars()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAscent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getAscent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getDescent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLeading()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getLeading()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getHeight()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBaselineIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getBaselineIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBaselineOffsets()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/font/LineMetrics/getBaselineOffsets()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getStrikethroughOffset()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getStrikethroughOffset()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getStrikethroughThickness()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getStrikethroughThickness()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUnderlineOffset()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getUnderlineOffset()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUnderlineThickness()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/LineMetrics/getUnderlineThickness()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_font_FontLineMetrics extends java_awt_font_LineMetrics implements java_lang_Cloneable {
    public "sun/font/FontLineMetrics/numchars": number;
    public "sun/font/FontLineMetrics/cm": sun_font_CoreMetrics;
    public "sun/font/FontLineMetrics/frc": java_awt_font_FontRenderContext;
    public "<init>(ILsun/font/CoreMetrics;Ljava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [number, sun_font_CoreMetrics, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontLineMetrics/<init>(ILsun/font/CoreMetrics;Ljava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [number, sun_font_CoreMetrics, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumChars()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getNumChars()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAscent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getAscent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getDescent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLeading()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getLeading()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getHeight()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBaselineIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getBaselineIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBaselineOffsets()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/FontLineMetrics/getBaselineOffsets()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getStrikethroughOffset()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getStrikethroughOffset()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getStrikethroughThickness()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getStrikethroughThickness()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUnderlineOffset()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getUnderlineOffset()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUnderlineThickness()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/getUnderlineThickness()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontLineMetrics/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/FontLineMetrics/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_font_CoreMetrics extends java_lang_Object {
    public "sun/font/CoreMetrics/ascent": number;
    public "sun/font/CoreMetrics/descent": number;
    public "sun/font/CoreMetrics/leading": number;
    public "sun/font/CoreMetrics/height": number;
    public "sun/font/CoreMetrics/baselineIndex": number;
    public "sun/font/CoreMetrics/baselineOffsets": JVMArray<number>;
    public "sun/font/CoreMetrics/strikethroughOffset": number;
    public "sun/font/CoreMetrics/strikethroughThickness": number;
    public "sun/font/CoreMetrics/underlineOffset": number;
    public "sun/font/CoreMetrics/underlineThickness": number;
    public "sun/font/CoreMetrics/ssOffset": number;
    public "sun/font/CoreMetrics/italicAngle": number;
    public "<init>(FFFFI[FFFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CoreMetrics/<init>(FFFFI[FFFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "get(Ljava/awt/font/LineMetrics;)Lsun/font/CoreMetrics;"(thread: JVMThread, args: [java_awt_font_LineMetrics], cb?: (e?: java_lang_Throwable, rv?: sun_font_CoreMetrics) => void): void;
    public static "sun/font/CoreMetrics/get(Ljava/awt/font/LineMetrics;)Lsun/font/CoreMetrics;"(thread: JVMThread, args: [java_awt_font_LineMetrics], cb?: (e?: java_lang_Throwable, rv?: sun_font_CoreMetrics) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CoreMetrics/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CoreMetrics/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Lsun/font/CoreMetrics;)Z"(thread: JVMThread, args: [sun_font_CoreMetrics], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CoreMetrics/equals(Lsun/font/CoreMetrics;)Z"(thread: JVMThread, args: [sun_font_CoreMetrics], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "effectiveBaselineOffset([F)F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CoreMetrics/effectiveBaselineOffset([F)F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_font_CreatedFontTracker extends java_lang_Object {
    public static "sun/font/CreatedFontTracker/MAX_FILE_SIZE": number;
    public static "sun/font/CreatedFontTracker/MAX_TOTAL_BYTES": number;
    public static "sun/font/CreatedFontTracker/tracker": sun_font_CreatedFontTracker;
    public "sun/font/CreatedFontTracker/numBytes": number;
    public static "getTracker()Lsun/font/CreatedFontTracker;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CreatedFontTracker) => void): void;
    public static "sun/font/CreatedFontTracker/getTracker()Lsun/font/CreatedFontTracker;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CreatedFontTracker) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CreatedFontTracker/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumBytes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CreatedFontTracker/getNumBytes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addBytes(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CreatedFontTracker/addBytes(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "subBytes(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CreatedFontTracker/subBytes(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getCS()Ljava/util/concurrent/Semaphore;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Semaphore) => void): void;
    public static "sun/font/CreatedFontTracker/getCS()Ljava/util/concurrent/Semaphore;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Semaphore) => void): void;
    public "acquirePermit()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CreatedFontTracker/acquirePermit()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "releasePermit()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CreatedFontTracker/releasePermit()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CreatedFontTracker/add(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(Ljava/io/File;Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_File, java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CreatedFontTracker/set(Ljava/io/File;Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_File, java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CreatedFontTracker/remove(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_Semaphore extends java_lang_Object implements java_io_Serializable {
    public static "java/util/concurrent/Semaphore/serialVersionUID": Long;
    public "java/util/concurrent/Semaphore/sync": java_util_concurrent_Semaphore$Sync;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore/<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "acquire()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore/acquire()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "acquireUninterruptibly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore/acquireUninterruptibly()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "tryAcquire()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore/tryAcquire()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryAcquire(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore/tryAcquire(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "release()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore/release()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "acquire(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore/acquire(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "acquireUninterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore/acquireUninterruptibly(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "tryAcquire(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore/tryAcquire(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryAcquire(IJLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [number, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore/tryAcquire(IJLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [number, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "release(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore/release(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "availablePermits()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore/availablePermits()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drainPermits()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore/drainPermits()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reducePermits(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore/reducePermits(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isFair()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore/isFair()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasQueuedThreads()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore/hasQueuedThreads()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getQueueLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore/getQueueLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/concurrent/Semaphore/getQueuedThreads()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/Semaphore/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_util_concurrent_Semaphore$Sync extends java_util_concurrent_locks_AbstractQueuedSynchronizer {
    public static "java/util/concurrent/Semaphore$Sync/serialVersionUID": Long;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore$Sync/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPermits()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore$Sync/getPermits()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nonfairTryAcquireShared(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore$Sync/nonfairTryAcquireShared(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "tryReleaseShared(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore$Sync/tryReleaseShared(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reducePermits(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/Semaphore$Sync/reducePermits(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drainPermits()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/Semaphore$Sync/drainPermits()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_font_AttributeValues extends java_lang_Object implements java_lang_Cloneable {
    public "sun/font/AttributeValues/defined": number;
    public "sun/font/AttributeValues/nondefault": number;
    public "sun/font/AttributeValues/family": java_lang_String;
    public "sun/font/AttributeValues/weight": number;
    public "sun/font/AttributeValues/width": number;
    public "sun/font/AttributeValues/posture": number;
    public "sun/font/AttributeValues/size": number;
    public "sun/font/AttributeValues/tracking": number;
    public "sun/font/AttributeValues/numericShaping": java_awt_font_NumericShaper;
    public "sun/font/AttributeValues/transform": java_awt_geom_AffineTransform;
    public "sun/font/AttributeValues/charReplacement": java_awt_font_GraphicAttribute;
    public "sun/font/AttributeValues/foreground": java_awt_Paint;
    public "sun/font/AttributeValues/background": java_awt_Paint;
    public "sun/font/AttributeValues/justification": number;
    public "sun/font/AttributeValues/imHighlight": java_lang_Object;
    public "sun/font/AttributeValues/font": java_awt_Font;
    public "sun/font/AttributeValues/imUnderline": number;
    public "sun/font/AttributeValues/superscript": number;
    public "sun/font/AttributeValues/underline": number;
    public "sun/font/AttributeValues/runDirection": number;
    public "sun/font/AttributeValues/bidiEmbedding": number;
    public "sun/font/AttributeValues/kerning": number;
    public "sun/font/AttributeValues/ligatures": number;
    public "sun/font/AttributeValues/strikethrough": number;
    public "sun/font/AttributeValues/swapColors": number;
    public "sun/font/AttributeValues/baselineTransform": java_awt_geom_AffineTransform;
    public "sun/font/AttributeValues/charTransform": java_awt_geom_AffineTransform;
    public static "sun/font/AttributeValues/DEFAULT": sun_font_AttributeValues;
    public static "sun/font/AttributeValues/MASK_ALL": number;
    public static "sun/font/AttributeValues/DEFINED_KEY": java_lang_String;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFamily()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/AttributeValues/getFamily()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setFamily(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setFamily(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWeight()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getWeight()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setWeight(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setWeight(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWidth()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getWidth()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setWidth(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setWidth(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPosture()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getPosture()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setPosture(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setPosture(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSize()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getSize()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSize(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setSize(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "sun/font/AttributeValues/getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTransform(Ljava/awt/font/TransformAttribute;)V"(thread: JVMThread, args: [java_awt_font_TransformAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setTransform(Ljava/awt/font/TransformAttribute;)V"(thread: JVMThread, args: [java_awt_font_TransformAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSuperscript()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getSuperscript()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSuperscript(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setSuperscript(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "sun/font/AttributeValues/getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCharReplacement()Ljava/awt/font/GraphicAttribute;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GraphicAttribute) => void): void;
    public "sun/font/AttributeValues/getCharReplacement()Ljava/awt/font/GraphicAttribute;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GraphicAttribute) => void): void;
    public "setCharReplacement(Ljava/awt/font/GraphicAttribute;)V"(thread: JVMThread, args: [java_awt_font_GraphicAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setCharReplacement(Ljava/awt/font/GraphicAttribute;)V"(thread: JVMThread, args: [java_awt_font_GraphicAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "getForeground()Ljava/awt/Paint;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Paint) => void): void;
    public "sun/font/AttributeValues/getForeground()Ljava/awt/Paint;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Paint) => void): void;
    public "setForeground(Ljava/awt/Paint;)V"(thread: JVMThread, args: [java_awt_Paint], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setForeground(Ljava/awt/Paint;)V"(thread: JVMThread, args: [java_awt_Paint], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBackground()Ljava/awt/Paint;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Paint) => void): void;
    public "sun/font/AttributeValues/getBackground()Ljava/awt/Paint;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Paint) => void): void;
    public "setBackground(Ljava/awt/Paint;)V"(thread: JVMThread, args: [java_awt_Paint], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setBackground(Ljava/awt/Paint;)V"(thread: JVMThread, args: [java_awt_Paint], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUnderline()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getUnderline()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setUnderline(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setUnderline(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getStrikethrough()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getStrikethrough()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setStrikethrough(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setStrikethrough(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRunDirection()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getRunDirection()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setRunDirection(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setRunDirection(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBidiEmbedding()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getBidiEmbedding()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setBidiEmbedding(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setBidiEmbedding(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getJustification()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getJustification()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setJustification(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setJustification(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInputMethodHighlight()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/AttributeValues/getInputMethodHighlight()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setInputMethodHighlight(Ljava/text/Annotation;)V"(thread: JVMThread, args: [java_text_Annotation], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setInputMethodHighlight(Ljava/text/Annotation;)V"(thread: JVMThread, args: [java_text_Annotation], cb?: (e?: java_lang_Throwable) => void): void;
    public "setInputMethodHighlight(Ljava/awt/im/InputMethodHighlight;)V"(thread: JVMThread, args: [java_awt_im_InputMethodHighlight], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setInputMethodHighlight(Ljava/awt/im/InputMethodHighlight;)V"(thread: JVMThread, args: [java_awt_im_InputMethodHighlight], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInputMethodUnderline()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getInputMethodUnderline()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setInputMethodUnderline(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setInputMethodUnderline(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSwapColors()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getSwapColors()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSwapColors(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setSwapColors(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumericShaping()Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public "sun/font/AttributeValues/getNumericShaping()Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public "setNumericShaping(Ljava/awt/font/NumericShaper;)V"(thread: JVMThread, args: [java_awt_font_NumericShaper], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setNumericShaping(Ljava/awt/font/NumericShaper;)V"(thread: JVMThread, args: [java_awt_font_NumericShaper], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKerning()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getKerning()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setKerning(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setKerning(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTracking()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getTracking()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setTracking(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setTracking(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLigatures()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/getLigatures()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLigatures(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setLigatures(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBaselineTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "sun/font/AttributeValues/getBaselineTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "getCharTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "sun/font/AttributeValues/getCharTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getMask(Lsun/font/EAttribute;)I"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/AttributeValues/getMask(Lsun/font/EAttribute;)I"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getMask([Lsun/font/EAttribute;)I"(thread: JVMThread, args: [JVMArray<sun_font_EAttribute>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/AttributeValues/getMask([Lsun/font/EAttribute;)I"(thread: JVMThread, args: [JVMArray<sun_font_EAttribute>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unsetDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/unsetDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "defineAll(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/defineAll(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "allDefined(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/allDefined(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "anyDefined(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/anyDefined(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "anyNonDefault(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/anyNonDefault(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDefined(Lsun/font/EAttribute;)Z"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/isDefined(Lsun/font/EAttribute;)Z"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isNonDefault(Lsun/font/EAttribute;)Z"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/isNonDefault(Lsun/font/EAttribute;)Z"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDefault(Lsun/font/EAttribute;)V"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/setDefault(Lsun/font/EAttribute;)V"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "unset(Lsun/font/EAttribute;)V"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/unset(Lsun/font/EAttribute;)V"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(Lsun/font/EAttribute;Lsun/font/AttributeValues;)V"(thread: JVMThread, args: [sun_font_EAttribute, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/set(Lsun/font/EAttribute;Lsun/font/AttributeValues;)V"(thread: JVMThread, args: [sun_font_EAttribute, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(Lsun/font/EAttribute;Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_font_EAttribute, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/set(Lsun/font/EAttribute;Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_font_EAttribute, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Lsun/font/EAttribute;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/AttributeValues/get(Lsun/font/EAttribute;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "merge(Ljava/util/Map;)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "sun/font/AttributeValues/merge(Ljava/util/Map;)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "merge(Ljava/util/Map;I)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Map, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "sun/font/AttributeValues/merge(Ljava/util/Map;I)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Map, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "merge(Lsun/font/AttributeValues;)Lsun/font/AttributeValues;"(thread: JVMThread, args: [sun_font_AttributeValues], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "sun/font/AttributeValues/merge(Lsun/font/AttributeValues;)Lsun/font/AttributeValues;"(thread: JVMThread, args: [sun_font_AttributeValues], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "merge(Lsun/font/AttributeValues;I)Lsun/font/AttributeValues;"(thread: JVMThread, args: [sun_font_AttributeValues, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "sun/font/AttributeValues/merge(Lsun/font/AttributeValues;I)Lsun/font/AttributeValues;"(thread: JVMThread, args: [sun_font_AttributeValues, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public static "fromMap(Ljava/util/Map;)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public static "sun/font/AttributeValues/fromMap(Ljava/util/Map;)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public static "fromMap(Ljava/util/Map;I)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Map, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public static "sun/font/AttributeValues/fromMap(Ljava/util/Map;I)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Map, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "toMap(Ljava/util/Map;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "sun/font/AttributeValues/toMap(Ljava/util/Map;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "is16Hashtable(Ljava/util/Hashtable;)Z"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/AttributeValues/is16Hashtable(Ljava/util/Hashtable;)Z"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "fromSerializableHashtable(Ljava/util/Hashtable;)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public static "sun/font/AttributeValues/fromSerializableHashtable(Ljava/util/Hashtable;)Lsun/font/AttributeValues;"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "toSerializableHashtable()Ljava/util/Hashtable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Hashtable) => void): void;
    public "sun/font/AttributeValues/toSerializableHashtable()Ljava/util/Hashtable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Hashtable) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Lsun/font/AttributeValues;)Z"(thread: JVMThread, args: [sun_font_AttributeValues], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/equals(Lsun/font/AttributeValues;)Z"(thread: JVMThread, args: [sun_font_AttributeValues], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Lsun/font/AttributeValues;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "sun/font/AttributeValues/clone()Lsun/font/AttributeValues;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/AttributeValues/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "equals(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/AttributeValues/equals(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "update(Lsun/font/EAttribute;)V"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/update(Lsun/font/EAttribute;)V"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "i_set(Lsun/font/EAttribute;Lsun/font/AttributeValues;)V"(thread: JVMThread, args: [sun_font_EAttribute, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/i_set(Lsun/font/EAttribute;Lsun/font/AttributeValues;)V"(thread: JVMThread, args: [sun_font_EAttribute, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable) => void): void;
    public "i_equals(Lsun/font/EAttribute;Lsun/font/AttributeValues;)Z"(thread: JVMThread, args: [sun_font_EAttribute, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/i_equals(Lsun/font/EAttribute;Lsun/font/AttributeValues;)Z"(thread: JVMThread, args: [sun_font_EAttribute, sun_font_AttributeValues], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "i_set(Lsun/font/EAttribute;Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_font_EAttribute, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/i_set(Lsun/font/EAttribute;Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_font_EAttribute, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "i_get(Lsun/font/EAttribute;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/AttributeValues/i_get(Lsun/font/EAttribute;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "i_validate(Lsun/font/EAttribute;)Z"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/AttributeValues/i_validate(Lsun/font/EAttribute;)Z"(thread: JVMThread, args: [sun_font_EAttribute], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getJustification(Ljava/util/Map;)F"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/AttributeValues/getJustification(Ljava/util/Map;)F"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getNumericShaping(Ljava/util/Map;)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "sun/font/AttributeValues/getNumericShaping(Ljava/util/Map;)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public "applyIMHighlight()Lsun/font/AttributeValues;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public "sun/font/AttributeValues/applyIMHighlight()Lsun/font/AttributeValues;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_AttributeValues) => void): void;
    public static "getBaselineTransform(Ljava/util/Map;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "sun/font/AttributeValues/getBaselineTransform(Ljava/util/Map;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "getCharTransform(Ljava/util/Map;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "sun/font/AttributeValues/getCharTransform(Ljava/util/Map;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "updateDerivedTransforms()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/AttributeValues/updateDerivedTransforms()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "extractXRotation(Ljava/awt/geom/AffineTransform;Z)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "sun/font/AttributeValues/extractXRotation(Ljava/awt/geom/AffineTransform;Z)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "extractYRotation(Ljava/awt/geom/AffineTransform;Z)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "sun/font/AttributeValues/extractYRotation(Ljava/awt/geom/AffineTransform;Z)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "extractRotation(Ljava/awt/geom/Point2D$Double;Ljava/awt/geom/AffineTransform;Z)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_awt_geom_Point2D$Double, java_awt_geom_AffineTransform, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "sun/font/AttributeValues/extractRotation(Ljava/awt/geom/Point2D$Double;Ljava/awt/geom/AffineTransform;Z)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_awt_geom_Point2D$Double, java_awt_geom_AffineTransform, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/AttributeValues/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/AttributeValues/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_geom_Point2D$Double extends java_awt_geom_Point2D implements java_io_Serializable {
    public "java/awt/geom/Point2D$Double/x": number;
    public "java/awt/geom/Point2D$Double/y": number;
    public static "java/awt/geom/Point2D$Double/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D$Double/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D$Double/<init>(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D$Double/getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D$Double/getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLocation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D$Double/setLocation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/geom/Point2D$Double/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_font_EAttribute extends java_lang_Enum {
    public static "sun/font/EAttribute/EFAMILY": sun_font_EAttribute;
    public static "sun/font/EAttribute/EWEIGHT": sun_font_EAttribute;
    public static "sun/font/EAttribute/EWIDTH": sun_font_EAttribute;
    public static "sun/font/EAttribute/EPOSTURE": sun_font_EAttribute;
    public static "sun/font/EAttribute/ESIZE": sun_font_EAttribute;
    public static "sun/font/EAttribute/ETRANSFORM": sun_font_EAttribute;
    public static "sun/font/EAttribute/ESUPERSCRIPT": sun_font_EAttribute;
    public static "sun/font/EAttribute/EFONT": sun_font_EAttribute;
    public static "sun/font/EAttribute/ECHAR_REPLACEMENT": sun_font_EAttribute;
    public static "sun/font/EAttribute/EFOREGROUND": sun_font_EAttribute;
    public static "sun/font/EAttribute/EBACKGROUND": sun_font_EAttribute;
    public static "sun/font/EAttribute/EUNDERLINE": sun_font_EAttribute;
    public static "sun/font/EAttribute/ESTRIKETHROUGH": sun_font_EAttribute;
    public static "sun/font/EAttribute/ERUN_DIRECTION": sun_font_EAttribute;
    public static "sun/font/EAttribute/EBIDI_EMBEDDING": sun_font_EAttribute;
    public static "sun/font/EAttribute/EJUSTIFICATION": sun_font_EAttribute;
    public static "sun/font/EAttribute/EINPUT_METHOD_HIGHLIGHT": sun_font_EAttribute;
    public static "sun/font/EAttribute/EINPUT_METHOD_UNDERLINE": sun_font_EAttribute;
    public static "sun/font/EAttribute/ESWAP_COLORS": sun_font_EAttribute;
    public static "sun/font/EAttribute/ENUMERIC_SHAPING": sun_font_EAttribute;
    public static "sun/font/EAttribute/EKERNING": sun_font_EAttribute;
    public static "sun/font/EAttribute/ELIGATURES": sun_font_EAttribute;
    public static "sun/font/EAttribute/ETRACKING": sun_font_EAttribute;
    public static "sun/font/EAttribute/EBASELINE_TRANSFORM": sun_font_EAttribute;
    public "sun/font/EAttribute/mask": number;
    public "sun/font/EAttribute/att": java_awt_font_TextAttribute;
    public static "sun/font/EAttribute/atts": JVMArray<sun_font_EAttribute>;
    public static "sun/font/EAttribute/$VALUES": JVMArray<sun_font_EAttribute>;
    public static "values()[Lsun/font/EAttribute;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_EAttribute>) => void): void;
    public static "sun/font/EAttribute/values()[Lsun/font/EAttribute;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_EAttribute>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lsun/font/EAttribute;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_EAttribute) => void): void;
    public static "sun/font/EAttribute/valueOf(Ljava/lang/String;)Lsun/font/EAttribute;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_EAttribute) => void): void;
    public "<init>(Ljava/lang/String;ILjava/awt/font/TextAttribute;)V"(thread: JVMThread, args: [java_lang_String, number, java_awt_font_TextAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/EAttribute/<init>(Ljava/lang/String;ILjava/awt/font/TextAttribute;)V"(thread: JVMThread, args: [java_lang_String, number, java_awt_font_TextAttribute], cb?: (e?: java_lang_Throwable) => void): void;
    public static "forAttribute(Ljava/text/AttributedCharacterIterator$Attribute;)Lsun/font/EAttribute;"(thread: JVMThread, args: [java_text_AttributedCharacterIterator$Attribute], cb?: (e?: java_lang_Throwable, rv?: sun_font_EAttribute) => void): void;
    public static "sun/font/EAttribute/forAttribute(Ljava/text/AttributedCharacterIterator$Attribute;)Lsun/font/EAttribute;"(thread: JVMThread, args: [java_text_AttributedCharacterIterator$Attribute], cb?: (e?: java_lang_Throwable, rv?: sun_font_EAttribute) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/EAttribute/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/EAttribute/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_font_TextAttribute extends java_text_AttributedCharacterIterator$Attribute {
    public static "java/awt/font/TextAttribute/instanceMap": java_util_Map;
    public static "java/awt/font/TextAttribute/serialVersionUID": Long;
    public static "java/awt/font/TextAttribute/FAMILY": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/WEIGHT": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/WEIGHT_EXTRA_LIGHT": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_LIGHT": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_DEMILIGHT": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_REGULAR": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_SEMIBOLD": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_MEDIUM": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_DEMIBOLD": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_BOLD": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_HEAVY": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_EXTRABOLD": java_lang_Float;
    public static "java/awt/font/TextAttribute/WEIGHT_ULTRABOLD": java_lang_Float;
    public static "java/awt/font/TextAttribute/WIDTH": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/WIDTH_CONDENSED": java_lang_Float;
    public static "java/awt/font/TextAttribute/WIDTH_SEMI_CONDENSED": java_lang_Float;
    public static "java/awt/font/TextAttribute/WIDTH_REGULAR": java_lang_Float;
    public static "java/awt/font/TextAttribute/WIDTH_SEMI_EXTENDED": java_lang_Float;
    public static "java/awt/font/TextAttribute/WIDTH_EXTENDED": java_lang_Float;
    public static "java/awt/font/TextAttribute/POSTURE": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/POSTURE_REGULAR": java_lang_Float;
    public static "java/awt/font/TextAttribute/POSTURE_OBLIQUE": java_lang_Float;
    public static "java/awt/font/TextAttribute/SIZE": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/TRANSFORM": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/SUPERSCRIPT": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/SUPERSCRIPT_SUPER": java_lang_Integer;
    public static "java/awt/font/TextAttribute/SUPERSCRIPT_SUB": java_lang_Integer;
    public static "java/awt/font/TextAttribute/FONT": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/CHAR_REPLACEMENT": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/FOREGROUND": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/BACKGROUND": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/UNDERLINE": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/UNDERLINE_ON": java_lang_Integer;
    public static "java/awt/font/TextAttribute/STRIKETHROUGH": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/STRIKETHROUGH_ON": java_lang_Boolean;
    public static "java/awt/font/TextAttribute/RUN_DIRECTION": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/RUN_DIRECTION_LTR": java_lang_Boolean;
    public static "java/awt/font/TextAttribute/RUN_DIRECTION_RTL": java_lang_Boolean;
    public static "java/awt/font/TextAttribute/BIDI_EMBEDDING": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/JUSTIFICATION": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/JUSTIFICATION_FULL": java_lang_Float;
    public static "java/awt/font/TextAttribute/JUSTIFICATION_NONE": java_lang_Float;
    public static "java/awt/font/TextAttribute/INPUT_METHOD_HIGHLIGHT": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/INPUT_METHOD_UNDERLINE": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/UNDERLINE_LOW_ONE_PIXEL": java_lang_Integer;
    public static "java/awt/font/TextAttribute/UNDERLINE_LOW_TWO_PIXEL": java_lang_Integer;
    public static "java/awt/font/TextAttribute/UNDERLINE_LOW_DOTTED": java_lang_Integer;
    public static "java/awt/font/TextAttribute/UNDERLINE_LOW_GRAY": java_lang_Integer;
    public static "java/awt/font/TextAttribute/UNDERLINE_LOW_DASHED": java_lang_Integer;
    public static "java/awt/font/TextAttribute/SWAP_COLORS": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/SWAP_COLORS_ON": java_lang_Boolean;
    public static "java/awt/font/TextAttribute/NUMERIC_SHAPING": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/KERNING": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/KERNING_ON": java_lang_Integer;
    public static "java/awt/font/TextAttribute/LIGATURES": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/LIGATURES_ON": java_lang_Integer;
    public static "java/awt/font/TextAttribute/TRACKING": java_awt_font_TextAttribute;
    public static "java/awt/font/TextAttribute/TRACKING_TIGHT": java_lang_Float;
    public static "java/awt/font/TextAttribute/TRACKING_LOOSE": java_lang_Float;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/TextAttribute/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/font/TextAttribute/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/font/TextAttribute/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_im_InputMethodHighlight extends java_lang_Object {
    public static "java/awt/im/InputMethodHighlight/RAW_TEXT": number;
    public static "java/awt/im/InputMethodHighlight/CONVERTED_TEXT": number;
    public static "java/awt/im/InputMethodHighlight/UNSELECTED_RAW_TEXT_HIGHLIGHT": java_awt_im_InputMethodHighlight;
    public static "java/awt/im/InputMethodHighlight/SELECTED_RAW_TEXT_HIGHLIGHT": java_awt_im_InputMethodHighlight;
    public static "java/awt/im/InputMethodHighlight/UNSELECTED_CONVERTED_TEXT_HIGHLIGHT": java_awt_im_InputMethodHighlight;
    public static "java/awt/im/InputMethodHighlight/SELECTED_CONVERTED_TEXT_HIGHLIGHT": java_awt_im_InputMethodHighlight;
    public "java/awt/im/InputMethodHighlight/selected": number;
    public "java/awt/im/InputMethodHighlight/state": number;
    public "java/awt/im/InputMethodHighlight/variation": number;
    public "java/awt/im/InputMethodHighlight/style": java_util_Map;
    public "<init>(ZI)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputMethodHighlight/<init>(ZI)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ZII)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputMethodHighlight/<init>(ZII)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ZIILjava/util/Map;)V"(thread: JVMThread, args: [number, number, number, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputMethodHighlight/<init>(ZIILjava/util/Map;)V"(thread: JVMThread, args: [number, number, number, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSelected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/im/InputMethodHighlight/isSelected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/im/InputMethodHighlight/getState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getVariation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/im/InputMethodHighlight/getVariation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getStyle()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/awt/im/InputMethodHighlight/getStyle()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/im/InputMethodHighlight/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_text_Annotation extends java_lang_Object {
    public "java/text/Annotation/value": java_lang_Object;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/text/Annotation/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/text/Annotation/getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/text/Annotation/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_font_TransformAttribute extends java_lang_Object implements java_io_Serializable {
    public "java/awt/font/TransformAttribute/transform": java_awt_geom_AffineTransform;
    public static "java/awt/font/TransformAttribute/IDENTITY": java_awt_font_TransformAttribute;
    public static "java/awt/font/TransformAttribute/serialVersionUID": Long;
    public "<init>(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/TransformAttribute/<init>(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "java/awt/font/TransformAttribute/getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "isIdentity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/TransformAttribute/isIdentity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/TransformAttribute/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/font/TransformAttribute/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/TransformAttribute/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/TransformAttribute/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/font/TransformAttribute/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_Paint extends java_lang_Object, java_awt_Transparency {
    "createContext(Ljava/awt/image/ColorModel;Ljava/awt/Rectangle;Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/AffineTransform;Ljava/awt/RenderingHints;)Ljava/awt/PaintContext;"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_Rectangle, java_awt_geom_Rectangle2D, java_awt_geom_AffineTransform, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable, rv?: java_awt_PaintContext) => void): void;
    "getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_RenderingHints extends java_lang_Object implements java_util_Map, java_lang_Cloneable {
    public "java/awt/RenderingHints/hintmap": java_util_HashMap;
    public static "java/awt/RenderingHints/KEY_ANTIALIASING": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/VALUE_ANTIALIAS_ON": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_ANTIALIAS_OFF": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_ANTIALIAS_DEFAULT": java_lang_Object;
    public static "java/awt/RenderingHints/KEY_RENDERING": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/VALUE_RENDER_SPEED": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_RENDER_QUALITY": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_RENDER_DEFAULT": java_lang_Object;
    public static "java/awt/RenderingHints/KEY_DITHERING": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/VALUE_DITHER_DISABLE": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_DITHER_ENABLE": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_DITHER_DEFAULT": java_lang_Object;
    public static "java/awt/RenderingHints/KEY_TEXT_ANTIALIASING": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/VALUE_TEXT_ANTIALIAS_ON": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_TEXT_ANTIALIAS_OFF": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_TEXT_ANTIALIAS_DEFAULT": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_TEXT_ANTIALIAS_GASP": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_TEXT_ANTIALIAS_LCD_HRGB": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_TEXT_ANTIALIAS_LCD_HBGR": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_TEXT_ANTIALIAS_LCD_VRGB": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_TEXT_ANTIALIAS_LCD_VBGR": java_lang_Object;
    public static "java/awt/RenderingHints/KEY_TEXT_LCD_CONTRAST": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/KEY_FRACTIONALMETRICS": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/VALUE_FRACTIONALMETRICS_OFF": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_FRACTIONALMETRICS_ON": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_FRACTIONALMETRICS_DEFAULT": java_lang_Object;
    public static "java/awt/RenderingHints/KEY_INTERPOLATION": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/VALUE_INTERPOLATION_NEAREST_NEIGHBOR": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_INTERPOLATION_BILINEAR": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_INTERPOLATION_BICUBIC": java_lang_Object;
    public static "java/awt/RenderingHints/KEY_ALPHA_INTERPOLATION": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/VALUE_ALPHA_INTERPOLATION_SPEED": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_ALPHA_INTERPOLATION_QUALITY": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_ALPHA_INTERPOLATION_DEFAULT": java_lang_Object;
    public static "java/awt/RenderingHints/KEY_COLOR_RENDERING": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/VALUE_COLOR_RENDER_SPEED": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_COLOR_RENDER_QUALITY": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_COLOR_RENDER_DEFAULT": java_lang_Object;
    public static "java/awt/RenderingHints/KEY_STROKE_CONTROL": java_awt_RenderingHints$Key;
    public static "java/awt/RenderingHints/VALUE_STROKE_DEFAULT": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_STROKE_NORMALIZE": java_lang_Object;
    public static "java/awt/RenderingHints/VALUE_STROKE_PURE": java_lang_Object;
    public "<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/RenderingHints/<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/RenderingHints$Key;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_RenderingHints$Key, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/RenderingHints/<init>(Ljava/awt/RenderingHints$Key;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_RenderingHints$Key, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints/containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/RenderingHints/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/RenderingHints/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "add(Ljava/awt/RenderingHints;)V"(thread: JVMThread, args: [java_awt_RenderingHints], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/RenderingHints/add(Ljava/awt/RenderingHints;)V"(thread: JVMThread, args: [java_awt_RenderingHints], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/RenderingHints/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/RenderingHints/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/RenderingHints/putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/awt/RenderingHints/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/awt/RenderingHints/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/awt/RenderingHints/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/RenderingHints/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/RenderingHints/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/RenderingHints/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_awt_RenderingHints$Key extends java_lang_Object {
    public static "java/awt/RenderingHints$Key/identitymap": java_util_HashMap;
    public "java/awt/RenderingHints$Key/privatekey": number;
    public "getIdentity()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/RenderingHints$Key/getIdentity()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "recordIdentity(Ljava/awt/RenderingHints$Key;)V"(thread: JVMThread, args: [java_awt_RenderingHints$Key], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/RenderingHints$Key/recordIdentity(Ljava/awt/RenderingHints$Key;)V"(thread: JVMThread, args: [java_awt_RenderingHints$Key], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/RenderingHints$Key/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isCompatibleValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints$Key/isCompatibleValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intKey()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints$Key/intKey()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints$Key/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/RenderingHints$Key/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/RenderingHints$Key/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_image_ColorModel extends java_lang_Object implements java_awt_Transparency {
    public "java/awt/image/ColorModel/pData": Long;
    public "java/awt/image/ColorModel/pixel_bits": number;
    public "java/awt/image/ColorModel/nBits": JVMArray<number>;
    public "java/awt/image/ColorModel/transparency": number;
    public "java/awt/image/ColorModel/supportsAlpha": number;
    public "java/awt/image/ColorModel/isAlphaPremultiplied": number;
    public "java/awt/image/ColorModel/numComponents": number;
    public "java/awt/image/ColorModel/numColorComponents": number;
    public "java/awt/image/ColorModel/colorSpace": java_awt_color_ColorSpace;
    public "java/awt/image/ColorModel/colorSpaceType": number;
    public "java/awt/image/ColorModel/maxBits": number;
    public "java/awt/image/ColorModel/is_sRGB": number;
    public "java/awt/image/ColorModel/transferType": number;
    public static "java/awt/image/ColorModel/loaded": number;
    public static "java/awt/image/ColorModel/RGBdefault": java_awt_image_ColorModel;
    public static "java/awt/image/ColorModel/l8Tos8": JVMArray<number>;
    public static "java/awt/image/ColorModel/s8Tol8": JVMArray<number>;
    public static "java/awt/image/ColorModel/l16Tos8": JVMArray<number>;
    public static "java/awt/image/ColorModel/s8Tol16": JVMArray<number>;
    public static "java/awt/image/ColorModel/g8Tos8Map": java_util_Map;
    public static "java/awt/image/ColorModel/lg16Toog8Map": java_util_Map;
    public static "java/awt/image/ColorModel/g16Tos8Map": java_util_Map;
    public static "java/awt/image/ColorModel/lg16Toog16Map": java_util_Map;
    public static "loadLibraries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/ColorModel/loadLibraries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/ColorModel/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getRGBdefault()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public static "java/awt/image/ColorModel/getRGBdefault()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/ColorModel/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I[ILjava/awt/color/ColorSpace;ZZII)V"(thread: JVMThread, args: [number, JVMArray<number>, java_awt_color_ColorSpace, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/ColorModel/<init>(I[ILjava/awt/color/ColorSpace;ZZII)V"(thread: JVMThread, args: [number, JVMArray<number>, java_awt_color_ColorSpace, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "hasAlpha()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/hasAlpha()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAlphaPremultiplied()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/isAlphaPremultiplied()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransferType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getTransferType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPixelSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getPixelSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getComponentSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getComponentSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getComponentSize()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/ColorModel/getComponentSize()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getNumComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumColorComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getNumColorComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRed(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getRed(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGreen(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getGreen(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBlue(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getBlue(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAlpha(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getAlpha(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRGB(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getRGB(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRed(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getRed(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGreen(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getGreen(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBlue(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getBlue(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAlpha(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getAlpha(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRGB(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getRGB(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDataElements(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/ColorModel/getDataElements(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getComponents(I[II)[I"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/ColorModel/getComponents(I[II)[I"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getComponents(Ljava/lang/Object;[II)[I"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/ColorModel/getComponents(Ljava/lang/Object;[II)[I"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getUnnormalizedComponents([FI[II)[I"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/ColorModel/getUnnormalizedComponents([FI[II)[I"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getNormalizedComponents([II[FI)[F"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/ColorModel/getNormalizedComponents([II[FI)[F"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getDataElement([II)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getDataElement([II)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDataElements([IILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<number>, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/ColorModel/getDataElements([IILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<number>, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getDataElement([FI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/getDataElement([FI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDataElements([FILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<number>, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/ColorModel/getDataElements([FILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<number>, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getNormalizedComponents(Ljava/lang/Object;[FI)[F"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/ColorModel/getNormalizedComponents(Ljava/lang/Object;[FI)[F"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getColorSpace()Ljava/awt/color/ColorSpace;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ColorSpace) => void): void;
    public "java/awt/image/ColorModel/getColorSpace()Ljava/awt/color/ColorSpace;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ColorSpace) => void): void;
    public "coerceData(Ljava/awt/image/WritableRaster;Z)Ljava/awt/image/ColorModel;"(thread: JVMThread, args: [java_awt_image_WritableRaster, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "java/awt/image/ColorModel/coerceData(Ljava/awt/image/WritableRaster;Z)Ljava/awt/image/ColorModel;"(thread: JVMThread, args: [java_awt_image_WritableRaster, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "isCompatibleRaster(Ljava/awt/image/Raster;)Z"(thread: JVMThread, args: [java_awt_image_Raster], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/isCompatibleRaster(Ljava/awt/image/Raster;)Z"(thread: JVMThread, args: [java_awt_image_Raster], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createCompatibleWritableRaster(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/ColorModel/createCompatibleWritableRaster(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "createCompatibleSampleModel(II)Ljava/awt/image/SampleModel;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "java/awt/image/ColorModel/createCompatibleSampleModel(II)Ljava/awt/image/SampleModel;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "isCompatibleSampleModel(Ljava/awt/image/SampleModel;)Z"(thread: JVMThread, args: [java_awt_image_SampleModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/ColorModel/isCompatibleSampleModel(Ljava/awt/image/SampleModel;)Z"(thread: JVMThread, args: [java_awt_image_SampleModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/ColorModel/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAlphaRaster(Ljava/awt/image/WritableRaster;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/ColorModel/getAlphaRaster(Ljava/awt/image/WritableRaster;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/image/ColorModel/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getDefaultTransferType(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/image/ColorModel/getDefaultTransferType(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLinearRGBspace(Ljava/awt/color/ColorSpace;)Z"(thread: JVMThread, args: [java_awt_color_ColorSpace], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/image/ColorModel/isLinearRGBspace(Ljava/awt/color/ColorSpace;)Z"(thread: JVMThread, args: [java_awt_color_ColorSpace], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLinearGRAYspace(Ljava/awt/color/ColorSpace;)Z"(thread: JVMThread, args: [java_awt_color_ColorSpace], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/image/ColorModel/isLinearGRAYspace(Ljava/awt/color/ColorSpace;)Z"(thread: JVMThread, args: [java_awt_color_ColorSpace], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getLinearRGB8TosRGB8LUT()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/image/ColorModel/getLinearRGB8TosRGB8LUT()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getsRGB8ToLinearRGB8LUT()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/image/ColorModel/getsRGB8ToLinearRGB8LUT()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getLinearRGB16TosRGB8LUT()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/image/ColorModel/getLinearRGB16TosRGB8LUT()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getsRGB8ToLinearRGB16LUT()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/image/ColorModel/getsRGB8ToLinearRGB16LUT()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getGray8TosRGB8LUT(Ljava/awt/color/ICC_ColorSpace;)[B"(thread: JVMThread, args: [java_awt_color_ICC__ColorSpace], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/image/ColorModel/getGray8TosRGB8LUT(Ljava/awt/color/ICC_ColorSpace;)[B"(thread: JVMThread, args: [java_awt_color_ICC__ColorSpace], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getLinearGray16ToOtherGray8LUT(Ljava/awt/color/ICC_ColorSpace;)[B"(thread: JVMThread, args: [java_awt_color_ICC__ColorSpace], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/image/ColorModel/getLinearGray16ToOtherGray8LUT(Ljava/awt/color/ICC_ColorSpace;)[B"(thread: JVMThread, args: [java_awt_color_ICC__ColorSpace], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getGray16TosRGB8LUT(Ljava/awt/color/ICC_ColorSpace;)[B"(thread: JVMThread, args: [java_awt_color_ICC__ColorSpace], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/image/ColorModel/getGray16TosRGB8LUT(Ljava/awt/color/ICC_ColorSpace;)[B"(thread: JVMThread, args: [java_awt_color_ICC__ColorSpace], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getLinearGray16ToOtherGray16LUT(Ljava/awt/color/ICC_ColorSpace;)[S"(thread: JVMThread, args: [java_awt_color_ICC__ColorSpace], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/image/ColorModel/getLinearGray16ToOtherGray16LUT(Ljava/awt/color/ICC_ColorSpace;)[S"(thread: JVMThread, args: [java_awt_color_ICC__ColorSpace], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/ColorModel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_color_ICC__ColorSpace extends java_awt_color_ColorSpace {
    public static "java/awt/color/ICC_ColorSpace/serialVersionUID": Long;
    public "java/awt/color/ICC_ColorSpace/thisProfile": java_awt_color_ICC__Profile;
    public "java/awt/color/ICC_ColorSpace/minVal": JVMArray<number>;
    public "java/awt/color/ICC_ColorSpace/maxVal": JVMArray<number>;
    public "java/awt/color/ICC_ColorSpace/diffMinMax": JVMArray<number>;
    public "java/awt/color/ICC_ColorSpace/invDiffMinMax": JVMArray<number>;
    public "java/awt/color/ICC_ColorSpace/needScaleInit": number;
    public "java/awt/color/ICC_ColorSpace/this2srgb": sun_java2d_cmm_ColorTransform;
    public "java/awt/color/ICC_ColorSpace/srgb2this": sun_java2d_cmm_ColorTransform;
    public "java/awt/color/ICC_ColorSpace/this2xyz": sun_java2d_cmm_ColorTransform;
    public "java/awt/color/ICC_ColorSpace/xyz2this": sun_java2d_cmm_ColorTransform;
    public "<init>(Ljava/awt/color/ICC_Profile;)V"(thread: JVMThread, args: [java_awt_color_ICC__Profile], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_ColorSpace/<init>(Ljava/awt/color/ICC_Profile;)V"(thread: JVMThread, args: [java_awt_color_ICC__Profile], cb?: (e?: java_lang_Throwable) => void): void;
    public "getProfile()Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public "java/awt/color/ICC_ColorSpace/getProfile()Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public "toRGB([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ICC_ColorSpace/toRGB([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "fromRGB([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ICC_ColorSpace/fromRGB([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "toCIEXYZ([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ICC_ColorSpace/toCIEXYZ([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "fromCIEXYZ([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ICC_ColorSpace/fromCIEXYZ([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getMinValue(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_ColorSpace/getMinValue(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaxValue(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_ColorSpace/getMaxValue(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMinMax()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_ColorSpace/setMinMax()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setComponentScaling()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_ColorSpace/setComponentScaling()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_cmm_ColorTransform extends java_lang_Object {
    "getNumInComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getNumOutComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "colorConvert(Ljava/awt/image/BufferedImage;Ljava/awt/image/BufferedImage;)V"(thread: JVMThread, args: [java_awt_image_BufferedImage, java_awt_image_BufferedImage], cb?: (e?: java_lang_Throwable) => void): void;
    "colorConvert(Ljava/awt/image/Raster;Ljava/awt/image/WritableRaster;[F[F[F[F)V"(thread: JVMThread, args: [java_awt_image_Raster, java_awt_image_WritableRaster, JVMArray<number>, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    "colorConvert(Ljava/awt/image/Raster;Ljava/awt/image/WritableRaster;)V"(thread: JVMThread, args: [java_awt_image_Raster, java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable) => void): void;
    "colorConvert([S[S)[S"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    "colorConvert([B[B)[B"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export class java_awt_image_BufferedImage extends java_awt_Image implements java_awt_image_WritableRenderedImage, java_awt_Transparency {
    public "java/awt/image/BufferedImage/imageType": number;
    public "java/awt/image/BufferedImage/colorModel": java_awt_image_ColorModel;
    public "java/awt/image/BufferedImage/raster": java_awt_image_WritableRaster;
    public "java/awt/image/BufferedImage/osis": sun_awt_image_OffScreenImageSource;
    public "java/awt/image/BufferedImage/properties": java_util_Hashtable;
    public static "java/awt/image/BufferedImage/TYPE_CUSTOM": number;
    public static "java/awt/image/BufferedImage/TYPE_INT_RGB": number;
    public static "java/awt/image/BufferedImage/TYPE_INT_ARGB": number;
    public static "java/awt/image/BufferedImage/TYPE_INT_ARGB_PRE": number;
    public static "java/awt/image/BufferedImage/TYPE_INT_BGR": number;
    public static "java/awt/image/BufferedImage/TYPE_3BYTE_BGR": number;
    public static "java/awt/image/BufferedImage/TYPE_4BYTE_ABGR": number;
    public static "java/awt/image/BufferedImage/TYPE_4BYTE_ABGR_PRE": number;
    public static "java/awt/image/BufferedImage/TYPE_USHORT_565_RGB": number;
    public static "java/awt/image/BufferedImage/TYPE_USHORT_555_RGB": number;
    public static "java/awt/image/BufferedImage/TYPE_BYTE_GRAY": number;
    public static "java/awt/image/BufferedImage/TYPE_USHORT_GRAY": number;
    public static "java/awt/image/BufferedImage/TYPE_BYTE_BINARY": number;
    public static "java/awt/image/BufferedImage/TYPE_BYTE_INDEXED": number;
    public static "java/awt/image/BufferedImage/DCM_RED_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_GREEN_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_BLUE_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_ALPHA_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_565_RED_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_565_GRN_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_565_BLU_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_555_RED_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_555_GRN_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_555_BLU_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_BGR_RED_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_BGR_GRN_MASK": number;
    public static "java/awt/image/BufferedImage/DCM_BGR_BLU_MASK": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/BufferedImage/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIILjava/awt/image/IndexColorModel;)V"(thread: JVMThread, args: [number, number, number, java_awt_image_IndexColorModel], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/<init>(IIILjava/awt/image/IndexColorModel;)V"(thread: JVMThread, args: [number, number, number, java_awt_image_IndexColorModel], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/image/ColorModel;Ljava/awt/image/WritableRaster;ZLjava/util/Hashtable;)V"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_image_WritableRaster, number, java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/<init>(Ljava/awt/image/ColorModel;Ljava/awt/image/WritableRaster;ZLjava/util/Hashtable;)V"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_image_WritableRaster, number, java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isStandard(Ljava/awt/image/ColorModel;Ljava/awt/image/WritableRaster;)Z"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/image/BufferedImage/isStandard(Ljava/awt/image/ColorModel;Ljava/awt/image/WritableRaster;)Z"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "java/awt/image/BufferedImage/getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "getRaster()Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/BufferedImage/getRaster()Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "getAlphaRaster()Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/BufferedImage/getAlphaRaster()Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "getRGB(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getRGB(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRGB(IIII[III)[I"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/BufferedImage/getRGB(IIII[III)[I"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "setRGB(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/setRGB(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRGB(IIII[III)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/setRGB(IIII[III)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWidth(Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getWidth(Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight(Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getHeight(Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSource()Ljava/awt/image/ImageProducer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ImageProducer) => void): void;
    public "java/awt/image/BufferedImage/getSource()Ljava/awt/image/ImageProducer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ImageProducer) => void): void;
    public "getProperty(Ljava/lang/String;Ljava/awt/image/ImageObserver;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/BufferedImage/getProperty(Ljava/lang/String;Ljava/awt/image/ImageObserver;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getProperty(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/BufferedImage/getProperty(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "java/awt/image/BufferedImage/getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "createGraphics()Ljava/awt/Graphics2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics2D) => void): void;
    public "java/awt/image/BufferedImage/createGraphics()Ljava/awt/Graphics2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics2D) => void): void;
    public "getSubimage(IIII)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "java/awt/image/BufferedImage/getSubimage(IIII)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "isAlphaPremultiplied()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/isAlphaPremultiplied()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "coerceData(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/coerceData(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/image/BufferedImage/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSources()Ljava/util/Vector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    public "java/awt/image/BufferedImage/getSources()Ljava/util/Vector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    public "getPropertyNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/awt/image/BufferedImage/getPropertyNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getMinX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getMinX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getMinY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSampleModel()Ljava/awt/image/SampleModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "java/awt/image/BufferedImage/getSampleModel()Ljava/awt/image/SampleModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "getNumXTiles()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getNumXTiles()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumYTiles()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getNumYTiles()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinTileX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getMinTileX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinTileY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getMinTileY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTileWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getTileWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTileHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getTileHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTileGridXOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getTileGridXOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTileGridYOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getTileGridYOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTile(II)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "java/awt/image/BufferedImage/getTile(II)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "getData()Ljava/awt/image/Raster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "java/awt/image/BufferedImage/getData()Ljava/awt/image/Raster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "getData(Ljava/awt/Rectangle;)Ljava/awt/image/Raster;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "java/awt/image/BufferedImage/getData(Ljava/awt/Rectangle;)Ljava/awt/image/Raster;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "copyData(Ljava/awt/image/WritableRaster;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/BufferedImage/copyData(Ljava/awt/image/WritableRaster;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "setData(Ljava/awt/image/Raster;)V"(thread: JVMThread, args: [java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/setData(Ljava/awt/image/Raster;)V"(thread: JVMThread, args: [java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "addTileObserver(Ljava/awt/image/TileObserver;)V"(thread: JVMThread, args: [java_awt_image_TileObserver], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/addTileObserver(Ljava/awt/image/TileObserver;)V"(thread: JVMThread, args: [java_awt_image_TileObserver], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeTileObserver(Ljava/awt/image/TileObserver;)V"(thread: JVMThread, args: [java_awt_image_TileObserver], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/removeTileObserver(Ljava/awt/image/TileObserver;)V"(thread: JVMThread, args: [java_awt_image_TileObserver], cb?: (e?: java_lang_Throwable) => void): void;
    public "isTileWritable(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/isTileWritable(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWritableTileIndices()[Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Point>) => void): void;
    public "java/awt/image/BufferedImage/getWritableTileIndices()[Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Point>) => void): void;
    public "hasTileWriters()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/hasTileWriters()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWritableTile(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/BufferedImage/getWritableTile(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "releaseWritableTile(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferedImage/releaseWritableTile(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferedImage/getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/BufferedImage/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_image_TileObserver extends java_lang_Object {
    "tileUpdate(Ljava/awt/image/WritableRenderedImage;IIZ)V"(thread: JVMThread, args: [java_awt_image_WritableRenderedImage, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Graphics2D extends java_awt_Graphics {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "draw3DRect(IIIIZ)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/draw3DRect(IIIIZ)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fill3DRect(IIIIZ)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/fill3DRect(IIIIZ)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "draw(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/draw(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawImage(Ljava/awt/Image;Ljava/awt/geom/AffineTransform;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, java_awt_geom_AffineTransform, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Graphics2D/drawImage(Ljava/awt/Image;Ljava/awt/geom/AffineTransform;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, java_awt_geom_AffineTransform, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/image/BufferedImage;Ljava/awt/image/BufferedImageOp;II)V"(thread: JVMThread, args: [java_awt_image_BufferedImage, java_awt_image_BufferedImageOp, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/drawImage(Ljava/awt/image/BufferedImage;Ljava/awt/image/BufferedImageOp;II)V"(thread: JVMThread, args: [java_awt_image_BufferedImage, java_awt_image_BufferedImageOp, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRenderedImage(Ljava/awt/image/RenderedImage;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_image_RenderedImage, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/drawRenderedImage(Ljava/awt/image/RenderedImage;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_image_RenderedImage, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRenderableImage(Ljava/awt/image/renderable/RenderableImage;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_image_renderable_RenderableImage, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/drawRenderableImage(Ljava/awt/image/renderable/RenderableImage;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_image_renderable_RenderableImage, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/drawString(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Ljava/lang/String;FF)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/drawString(Ljava/lang/String;FF)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Ljava/text/AttributedCharacterIterator;II)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/drawString(Ljava/text/AttributedCharacterIterator;II)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Ljava/text/AttributedCharacterIterator;FF)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/drawString(Ljava/text/AttributedCharacterIterator;FF)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawGlyphVector(Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/drawGlyphVector(Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fill(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/fill(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "hit(Ljava/awt/Rectangle;Ljava/awt/Shape;Z)Z"(thread: JVMThread, args: [java_awt_Rectangle, java_awt_Shape, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Graphics2D/hit(Ljava/awt/Rectangle;Ljava/awt/Shape;Z)Z"(thread: JVMThread, args: [java_awt_Rectangle, java_awt_Shape, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDeviceConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "java/awt/Graphics2D/getDeviceConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "setComposite(Ljava/awt/Composite;)V"(thread: JVMThread, args: [java_awt_Composite], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/setComposite(Ljava/awt/Composite;)V"(thread: JVMThread, args: [java_awt_Composite], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPaint(Ljava/awt/Paint;)V"(thread: JVMThread, args: [java_awt_Paint], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/setPaint(Ljava/awt/Paint;)V"(thread: JVMThread, args: [java_awt_Paint], cb?: (e?: java_lang_Throwable) => void): void;
    public "setStroke(Ljava/awt/Stroke;)V"(thread: JVMThread, args: [java_awt_Stroke], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/setStroke(Ljava/awt/Stroke;)V"(thread: JVMThread, args: [java_awt_Stroke], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRenderingHint(Ljava/awt/RenderingHints$Key;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_RenderingHints$Key, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/setRenderingHint(Ljava/awt/RenderingHints$Key;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_RenderingHints$Key, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRenderingHint(Ljava/awt/RenderingHints$Key;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_RenderingHints$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/Graphics2D/getRenderingHint(Ljava/awt/RenderingHints$Key;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_RenderingHints$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setRenderingHints(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/setRenderingHints(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "addRenderingHints(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/addRenderingHints(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRenderingHints()Ljava/awt/RenderingHints;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_RenderingHints) => void): void;
    public "java/awt/Graphics2D/getRenderingHints()Ljava/awt/RenderingHints;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_RenderingHints) => void): void;
    public "translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "translate(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/translate(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/rotate(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate(DDD)V"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/rotate(DDD)V"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "scale(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/scale(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "shear(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/shear(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "transform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/transform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "java/awt/Graphics2D/getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "getPaint()Ljava/awt/Paint;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Paint) => void): void;
    public "java/awt/Graphics2D/getPaint()Ljava/awt/Paint;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Paint) => void): void;
    public "getComposite()Ljava/awt/Composite;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Composite) => void): void;
    public "java/awt/Graphics2D/getComposite()Ljava/awt/Composite;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Composite) => void): void;
    public "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "java/awt/Graphics2D/getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "getStroke()Ljava/awt/Stroke;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Stroke) => void): void;
    public "java/awt/Graphics2D/getStroke()Ljava/awt/Stroke;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Stroke) => void): void;
    public "clip(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics2D/clip(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "java/awt/Graphics2D/getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
  }
  export class java_awt_Color extends java_lang_Object implements java_awt_Paint, java_io_Serializable {
    public static "java/awt/Color/white": java_awt_Color;
    public static "java/awt/Color/WHITE": java_awt_Color;
    public static "java/awt/Color/lightGray": java_awt_Color;
    public static "java/awt/Color/LIGHT_GRAY": java_awt_Color;
    public static "java/awt/Color/gray": java_awt_Color;
    public static "java/awt/Color/GRAY": java_awt_Color;
    public static "java/awt/Color/darkGray": java_awt_Color;
    public static "java/awt/Color/DARK_GRAY": java_awt_Color;
    public static "java/awt/Color/black": java_awt_Color;
    public static "java/awt/Color/BLACK": java_awt_Color;
    public static "java/awt/Color/red": java_awt_Color;
    public static "java/awt/Color/RED": java_awt_Color;
    public static "java/awt/Color/pink": java_awt_Color;
    public static "java/awt/Color/PINK": java_awt_Color;
    public static "java/awt/Color/orange": java_awt_Color;
    public static "java/awt/Color/ORANGE": java_awt_Color;
    public static "java/awt/Color/yellow": java_awt_Color;
    public static "java/awt/Color/YELLOW": java_awt_Color;
    public static "java/awt/Color/green": java_awt_Color;
    public static "java/awt/Color/GREEN": java_awt_Color;
    public static "java/awt/Color/magenta": java_awt_Color;
    public static "java/awt/Color/MAGENTA": java_awt_Color;
    public static "java/awt/Color/cyan": java_awt_Color;
    public static "java/awt/Color/CYAN": java_awt_Color;
    public static "java/awt/Color/blue": java_awt_Color;
    public static "java/awt/Color/BLUE": java_awt_Color;
    public "java/awt/Color/value": number;
    public "java/awt/Color/frgbvalue": JVMArray<number>;
    public "java/awt/Color/fvalue": JVMArray<number>;
    public "java/awt/Color/falpha": number;
    public "java/awt/Color/cs": java_awt_color_ColorSpace;
    public static "java/awt/Color/serialVersionUID": Long;
    public static "java/awt/Color/FACTOR": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Color/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "testColorValueRange(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Color/testColorValueRange(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "testColorValueRange(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Color/testColorValueRange(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Color/<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Color/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Color/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Color/<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(FFF)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Color/<init>(FFF)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Color/<init>(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/color/ColorSpace;[FF)V"(thread: JVMThread, args: [java_awt_color_ColorSpace, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Color/<init>(Ljava/awt/color/ColorSpace;[FF)V"(thread: JVMThread, args: [java_awt_color_ColorSpace, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRed()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Color/getRed()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGreen()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Color/getGreen()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBlue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Color/getBlue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAlpha()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Color/getAlpha()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRGB()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Color/getRGB()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "brighter()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "java/awt/Color/brighter()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "darker()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "java/awt/Color/darker()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Color/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Color/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Color/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "decode(Ljava/lang/String;)Ljava/awt/Color;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public static "java/awt/Color/decode(Ljava/lang/String;)Ljava/awt/Color;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public static "getColor(Ljava/lang/String;)Ljava/awt/Color;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public static "java/awt/Color/getColor(Ljava/lang/String;)Ljava/awt/Color;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public static "getColor(Ljava/lang/String;Ljava/awt/Color;)Ljava/awt/Color;"(thread: JVMThread, args: [java_lang_String, java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public static "java/awt/Color/getColor(Ljava/lang/String;Ljava/awt/Color;)Ljava/awt/Color;"(thread: JVMThread, args: [java_lang_String, java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public static "getColor(Ljava/lang/String;I)Ljava/awt/Color;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public static "java/awt/Color/getColor(Ljava/lang/String;I)Ljava/awt/Color;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public static "HSBtoRGB(FFF)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Color/HSBtoRGB(FFF)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "RGBtoHSB(III[F)[F"(thread: JVMThread, args: [number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/Color/RGBtoHSB(III[F)[F"(thread: JVMThread, args: [number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getHSBColor(FFF)Ljava/awt/Color;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public static "java/awt/Color/getHSBColor(FFF)Ljava/awt/Color;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "getRGBComponents([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/Color/getRGBComponents([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getRGBColorComponents([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/Color/getRGBColorComponents([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getComponents([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/Color/getComponents([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getColorComponents([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/Color/getColorComponents([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getComponents(Ljava/awt/color/ColorSpace;[F)[F"(thread: JVMThread, args: [java_awt_color_ColorSpace, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/Color/getComponents(Ljava/awt/color/ColorSpace;[F)[F"(thread: JVMThread, args: [java_awt_color_ColorSpace, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getColorComponents(Ljava/awt/color/ColorSpace;[F)[F"(thread: JVMThread, args: [java_awt_color_ColorSpace, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/Color/getColorComponents(Ljava/awt/color/ColorSpace;[F)[F"(thread: JVMThread, args: [java_awt_color_ColorSpace, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getColorSpace()Ljava/awt/color/ColorSpace;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ColorSpace) => void): void;
    public "java/awt/Color/getColorSpace()Ljava/awt/color/ColorSpace;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ColorSpace) => void): void;
    public "createContext(Ljava/awt/image/ColorModel;Ljava/awt/Rectangle;Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/AffineTransform;Ljava/awt/RenderingHints;)Ljava/awt/PaintContext;"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_Rectangle, java_awt_geom_Rectangle2D, java_awt_geom_AffineTransform, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable, rv?: java_awt_PaintContext) => void): void;
    public "java/awt/Color/createContext(Ljava/awt/image/ColorModel;Ljava/awt/Rectangle;Ljava/awt/geom/Rectangle2D;Ljava/awt/geom/AffineTransform;Ljava/awt/RenderingHints;)Ljava/awt/PaintContext;"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_Rectangle, java_awt_geom_Rectangle2D, java_awt_geom_AffineTransform, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable, rv?: java_awt_PaintContext) => void): void;
    public "getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Color/getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Color/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_Stroke extends java_lang_Object {
    "createStrokedShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
  }
  export interface java_awt_Composite extends java_lang_Object {
    "createContext(Ljava/awt/image/ColorModel;Ljava/awt/image/ColorModel;Ljava/awt/RenderingHints;)Ljava/awt/CompositeContext;"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_image_ColorModel, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable, rv?: java_awt_CompositeContext) => void): void;
  }
  export interface java_awt_CompositeContext extends java_lang_Object {
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "compose(Ljava/awt/image/Raster;Ljava/awt/image/Raster;Ljava/awt/image/WritableRaster;)V"(thread: JVMThread, args: [java_awt_image_Raster, java_awt_image_Raster, java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_GraphicsConfiguration extends java_lang_Object {
    public static "java/awt/GraphicsConfiguration/defaultBufferCaps": java_awt_BufferCapabilities;
    public static "java/awt/GraphicsConfiguration/defaultImageCaps": java_awt_ImageCapabilities;
    public static "java/awt/GraphicsConfiguration/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/GraphicsConfiguration/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDevice()Ljava/awt/GraphicsDevice;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsDevice) => void): void;
    public "java/awt/GraphicsConfiguration/getDevice()Ljava/awt/GraphicsDevice;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsDevice) => void): void;
    public "createCompatibleImage(II)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "java/awt/GraphicsConfiguration/createCompatibleImage(II)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "createCompatibleImage(III)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "java/awt/GraphicsConfiguration/createCompatibleImage(III)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "createCompatibleVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "java/awt/GraphicsConfiguration/createCompatibleVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "createCompatibleVolatileImage(III)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "java/awt/GraphicsConfiguration/createCompatibleVolatileImage(III)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "createCompatibleVolatileImage(IILjava/awt/ImageCapabilities;)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number, java_awt_ImageCapabilities], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "java/awt/GraphicsConfiguration/createCompatibleVolatileImage(IILjava/awt/ImageCapabilities;)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number, java_awt_ImageCapabilities], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "createCompatibleVolatileImage(IILjava/awt/ImageCapabilities;I)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number, java_awt_ImageCapabilities, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "java/awt/GraphicsConfiguration/createCompatibleVolatileImage(IILjava/awt/ImageCapabilities;I)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number, java_awt_ImageCapabilities, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "java/awt/GraphicsConfiguration/getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "getColorModel(I)Ljava/awt/image/ColorModel;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "java/awt/GraphicsConfiguration/getColorModel(I)Ljava/awt/image/ColorModel;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "getDefaultTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "java/awt/GraphicsConfiguration/getDefaultTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "getNormalizingTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "java/awt/GraphicsConfiguration/getNormalizingTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/GraphicsConfiguration/getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getBufferCapabilities()Ljava/awt/BufferCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_BufferCapabilities) => void): void;
    public "java/awt/GraphicsConfiguration/getBufferCapabilities()Ljava/awt/BufferCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_BufferCapabilities) => void): void;
    public "getImageCapabilities()Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "java/awt/GraphicsConfiguration/getImageCapabilities()Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "isTranslucencyCapable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/GraphicsConfiguration/isTranslucencyCapable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/GraphicsConfiguration/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_image_VolatileImage extends java_awt_Image implements java_awt_Transparency {
    public static "java/awt/image/VolatileImage/IMAGE_OK": number;
    public static "java/awt/image/VolatileImage/IMAGE_RESTORED": number;
    public static "java/awt/image/VolatileImage/IMAGE_INCOMPATIBLE": number;
    public "java/awt/image/VolatileImage/transparency": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/VolatileImage/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSnapshot()Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "java/awt/image/VolatileImage/getSnapshot()Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/VolatileImage/getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/VolatileImage/getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSource()Ljava/awt/image/ImageProducer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ImageProducer) => void): void;
    public "java/awt/image/VolatileImage/getSource()Ljava/awt/image/ImageProducer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ImageProducer) => void): void;
    public "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "java/awt/image/VolatileImage/getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "createGraphics()Ljava/awt/Graphics2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics2D) => void): void;
    public "java/awt/image/VolatileImage/createGraphics()Ljava/awt/Graphics2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics2D) => void): void;
    public "validate(Ljava/awt/GraphicsConfiguration;)I"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/VolatileImage/validate(Ljava/awt/GraphicsConfiguration;)I"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contentsLost()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/VolatileImage/contentsLost()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCapabilities()Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "java/awt/image/VolatileImage/getCapabilities()Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/VolatileImage/getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_GraphicsDevice extends java_lang_Object {
    public "java/awt/GraphicsDevice/fullScreenWindow": java_awt_Window;
    public "java/awt/GraphicsDevice/fullScreenAppContext": sun_awt_AppContext;
    public "java/awt/GraphicsDevice/fsAppContextLock": java_lang_Object;
    public "java/awt/GraphicsDevice/windowedModeBounds": java_awt_Rectangle;
    public static "java/awt/GraphicsDevice/TYPE_RASTER_SCREEN": number;
    public static "java/awt/GraphicsDevice/TYPE_PRINTER": number;
    public static "java/awt/GraphicsDevice/TYPE_IMAGE_BUFFER": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/GraphicsDevice/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/GraphicsDevice/getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getIDstring()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/GraphicsDevice/getIDstring()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getConfigurations()[Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_GraphicsConfiguration>) => void): void;
    public "java/awt/GraphicsDevice/getConfigurations()[Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_GraphicsConfiguration>) => void): void;
    public "getDefaultConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "java/awt/GraphicsDevice/getDefaultConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "getBestConfiguration(Ljava/awt/GraphicsConfigTemplate;)Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: [java_awt_GraphicsConfigTemplate], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "java/awt/GraphicsDevice/getBestConfiguration(Ljava/awt/GraphicsConfigTemplate;)Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: [java_awt_GraphicsConfigTemplate], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "isFullScreenSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/GraphicsDevice/isFullScreenSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFullScreenWindow(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/GraphicsDevice/setFullScreenWindow(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFullScreenWindow()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "java/awt/GraphicsDevice/getFullScreenWindow()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "isDisplayChangeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/GraphicsDevice/isDisplayChangeSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDisplayMode(Ljava/awt/DisplayMode;)V"(thread: JVMThread, args: [java_awt_DisplayMode], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/GraphicsDevice/setDisplayMode(Ljava/awt/DisplayMode;)V"(thread: JVMThread, args: [java_awt_DisplayMode], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDisplayMode()Ljava/awt/DisplayMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_DisplayMode) => void): void;
    public "java/awt/GraphicsDevice/getDisplayMode()Ljava/awt/DisplayMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_DisplayMode) => void): void;
    public "getDisplayModes()[Ljava/awt/DisplayMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_DisplayMode>) => void): void;
    public "java/awt/GraphicsDevice/getDisplayModes()[Ljava/awt/DisplayMode;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_DisplayMode>) => void): void;
    public "getAvailableAcceleratedMemory()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/GraphicsDevice/getAvailableAcceleratedMemory()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isWindowTranslucencySupported(Ljava/awt/GraphicsDevice$WindowTranslucency;)Z"(thread: JVMThread, args: [java_awt_GraphicsDevice$WindowTranslucency], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/GraphicsDevice/isWindowTranslucencySupported(Ljava/awt/GraphicsDevice$WindowTranslucency;)Z"(thread: JVMThread, args: [java_awt_GraphicsDevice$WindowTranslucency], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isWindowShapingSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/GraphicsDevice/isWindowShapingSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isWindowOpacitySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/GraphicsDevice/isWindowOpacitySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isWindowPerpixelTranslucencySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/GraphicsDevice/isWindowPerpixelTranslucencySupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTranslucencyCapableGC()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "java/awt/GraphicsDevice/getTranslucencyCapableGC()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
  }
  export class java_awt_GraphicsDevice$WindowTranslucency extends java_lang_Enum {
    public static "java/awt/GraphicsDevice$WindowTranslucency/PERPIXEL_TRANSPARENT": java_awt_GraphicsDevice$WindowTranslucency;
    public static "java/awt/GraphicsDevice$WindowTranslucency/TRANSLUCENT": java_awt_GraphicsDevice$WindowTranslucency;
    public static "java/awt/GraphicsDevice$WindowTranslucency/PERPIXEL_TRANSLUCENT": java_awt_GraphicsDevice$WindowTranslucency;
    public static "java/awt/GraphicsDevice$WindowTranslucency/$VALUES": JVMArray<java_awt_GraphicsDevice$WindowTranslucency>;
    public static "values()[Ljava/awt/GraphicsDevice$WindowTranslucency;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_GraphicsDevice$WindowTranslucency>) => void): void;
    public static "java/awt/GraphicsDevice$WindowTranslucency/values()[Ljava/awt/GraphicsDevice$WindowTranslucency;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_GraphicsDevice$WindowTranslucency>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/awt/GraphicsDevice$WindowTranslucency;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsDevice$WindowTranslucency) => void): void;
    public static "java/awt/GraphicsDevice$WindowTranslucency/valueOf(Ljava/lang/String;)Ljava/awt/GraphicsDevice$WindowTranslucency;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsDevice$WindowTranslucency) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/GraphicsDevice$WindowTranslucency/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/GraphicsDevice$WindowTranslucency/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_DisplayMode extends java_lang_Object {
    public "java/awt/DisplayMode/size": java_awt_Dimension;
    public "java/awt/DisplayMode/bitDepth": number;
    public "java/awt/DisplayMode/refreshRate": number;
    public static "java/awt/DisplayMode/BIT_DEPTH_MULTI": number;
    public static "java/awt/DisplayMode/REFRESH_RATE_UNKNOWN": number;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/DisplayMode/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/DisplayMode/getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/DisplayMode/getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBitDepth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/DisplayMode/getBitDepth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRefreshRate()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/DisplayMode/getRefreshRate()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/awt/DisplayMode;)Z"(thread: JVMThread, args: [java_awt_DisplayMode], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/DisplayMode/equals(Ljava/awt/DisplayMode;)Z"(thread: JVMThread, args: [java_awt_DisplayMode], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/DisplayMode/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/DisplayMode/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_GraphicsConfigTemplate extends java_lang_Object implements java_io_Serializable {
    public static "java/awt/GraphicsConfigTemplate/serialVersionUID": Long;
    public static "java/awt/GraphicsConfigTemplate/REQUIRED": number;
    public static "java/awt/GraphicsConfigTemplate/PREFERRED": number;
    public static "java/awt/GraphicsConfigTemplate/UNNECESSARY": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/GraphicsConfigTemplate/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBestConfiguration([Ljava/awt/GraphicsConfiguration;)Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: [JVMArray<java_awt_GraphicsConfiguration>], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "java/awt/GraphicsConfigTemplate/getBestConfiguration([Ljava/awt/GraphicsConfiguration;)Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: [JVMArray<java_awt_GraphicsConfiguration>], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "isGraphicsConfigSupported(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/GraphicsConfigTemplate/isGraphicsConfigSupported(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_awt_AppContext extends java_lang_Object {
    public static "sun/awt/AppContext/log": sun_util_logging_PlatformLogger;
    public static "sun/awt/AppContext/EVENT_QUEUE_KEY": java_lang_Object;
    public static "sun/awt/AppContext/EVENT_QUEUE_LOCK_KEY": java_lang_Object;
    public static "sun/awt/AppContext/EVENT_QUEUE_COND_KEY": java_lang_Object;
    public static "sun/awt/AppContext/threadGroup2appContext": java_util_Map;
    public static "sun/awt/AppContext/mainAppContext": sun_awt_AppContext;
    public static "sun/awt/AppContext/getAppContextLock": java_lang_Object;
    public "sun/awt/AppContext/table": java_util_Map;
    public "sun/awt/AppContext/threadGroup": java_lang_ThreadGroup;
    public "sun/awt/AppContext/changeSupport": java_beans_PropertyChangeSupport;
    public static "sun/awt/AppContext/DISPOSED_PROPERTY_NAME": java_lang_String;
    public static "sun/awt/AppContext/GUI_DISPOSED": java_lang_String;
    public "sun/awt/AppContext/state": sun_awt_AppContext$State;
    public static "sun/awt/AppContext/numAppContexts": java_util_concurrent_atomic_AtomicInteger;
    public "sun/awt/AppContext/contextClassLoader": java_lang_ClassLoader;
    public static "sun/awt/AppContext/threadAppContext": java_lang_ThreadLocal;
    public "sun/awt/AppContext/DISPOSAL_TIMEOUT": Long;
    public "sun/awt/AppContext/THREAD_INTERRUPT_TIMEOUT": Long;
    public "sun/awt/AppContext/mostRecentKeyValue": sun_awt_MostRecentKeyValue;
    public "sun/awt/AppContext/shadowMostRecentKeyValue": sun_awt_MostRecentKeyValue;
    public static "getAppContexts()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "sun/awt/AppContext/getAppContexts()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "isDisposed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/AppContext/isDisposed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/ThreadGroup;)V"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/AppContext/<init>(Ljava/lang/ThreadGroup;)V"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initMainAppContext()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/AppContext/initMainAppContext()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getAppContext()Lsun/awt/AppContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "sun/awt/AppContext/getAppContext()Lsun/awt/AppContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "isMainContext(Lsun/awt/AppContext;)Z"(thread: JVMThread, args: [sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/AppContext/isMainContext(Lsun/awt/AppContext;)Z"(thread: JVMThread, args: [sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getExecutionAppContext()Lsun/awt/AppContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "sun/awt/AppContext/getExecutionAppContext()Lsun/awt/AppContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/AppContext/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "stopEventDispatchThreads()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/AppContext/stopEventDispatchThreads()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/awt/AppContext/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/awt/AppContext/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/awt/AppContext/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getThreadGroup()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public "sun/awt/AppContext/getThreadGroup()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public "getContextClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "sun/awt/AppContext/getContextClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/AppContext/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPropertyChangeListeners()[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "sun/awt/AppContext/getPropertyChangeListeners()[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/AppContext/addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/AppContext/removePropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPropertyChangeListeners(Ljava/lang/String;)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "sun/awt/AppContext/getPropertyChangeListeners(Ljava/lang/String;)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public static "getSoftReferenceValue(Ljava/lang/Object;Ljava/util/function/Supplier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/awt/AppContext/getSoftReferenceValue(Ljava/lang/Object;Ljava/util/function/Supplier;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Supplier], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$102(Lsun/awt/AppContext;)Lsun/awt/AppContext;"(thread: JVMThread, args: [sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "sun/awt/AppContext/access$102(Lsun/awt/AppContext;)Lsun/awt/AppContext;"(thread: JVMThread, args: [sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "access$200()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/awt/AppContext/access$200()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$300()Ljava/util/concurrent/atomic/AtomicInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicInteger) => void): void;
    public static "sun/awt/AppContext/access$300()Ljava/util/concurrent/atomic/AtomicInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicInteger) => void): void;
    public static "access$400()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/AppContext/access$400()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$500()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "sun/awt/AppContext/access$500()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "access$600()Ljava/lang/ThreadLocal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal) => void): void;
    public static "sun/awt/AppContext/access$600()Ljava/lang/ThreadLocal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadLocal) => void): void;
    public static "access$700()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "sun/awt/AppContext/access$700()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "access$800(Lsun/awt/AppContext;)Ljava/lang/ThreadGroup;"(thread: JVMThread, args: [sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public static "sun/awt/AppContext/access$800(Lsun/awt/AppContext;)Ljava/lang/ThreadGroup;"(thread: JVMThread, args: [sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public static "access$900()Lsun/awt/AppContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "sun/awt/AppContext/access$900()Lsun/awt/AppContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "access$100()Lsun/awt/AppContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "sun/awt/AppContext/access$100()Lsun/awt/AppContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/AppContext/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_beans_PropertyChangeListener extends java_lang_Object, java_util_EventListener {
    "propertyChange(Ljava/beans/PropertyChangeEvent;)V"(thread: JVMThread, args: [java_beans_PropertyChangeEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_beans_PropertyChangeEvent extends java_util_EventObject {
    public static "java/beans/PropertyChangeEvent/serialVersionUID": Long;
    public "java/beans/PropertyChangeEvent/propertyName": java_lang_String;
    public "java/beans/PropertyChangeEvent/newValue": java_lang_Object;
    public "java/beans/PropertyChangeEvent/oldValue": java_lang_Object;
    public "java/beans/PropertyChangeEvent/propagationId": java_lang_Object;
    public "<init>(Ljava/lang/Object;Ljava/lang/String;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_String, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeEvent/<init>(Ljava/lang/Object;Ljava/lang/String;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_String, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPropertyName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/beans/PropertyChangeEvent/getPropertyName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getNewValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/beans/PropertyChangeEvent/getNewValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getOldValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/beans/PropertyChangeEvent/getOldValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setPropagationId(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeEvent/setPropagationId(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPropagationId()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/beans/PropertyChangeEvent/getPropagationId()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/beans/PropertyChangeEvent/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "appendTo(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeEvent/appendTo(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_MostRecentKeyValue extends java_lang_Object {
    public "sun/awt/MostRecentKeyValue/key": java_lang_Object;
    public "sun/awt/MostRecentKeyValue/value": java_lang_Object;
    public "<init>(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/MostRecentKeyValue/<init>(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPair(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/MostRecentKeyValue/setPair(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_AppContext$State extends java_lang_Enum {
    public static "sun/awt/AppContext$State/VALID": sun_awt_AppContext$State;
    public static "sun/awt/AppContext$State/BEING_DISPOSED": sun_awt_AppContext$State;
    public static "sun/awt/AppContext$State/DISPOSED": sun_awt_AppContext$State;
    public static "sun/awt/AppContext$State/$VALUES": JVMArray<sun_awt_AppContext$State>;
    public static "values()[Lsun/awt/AppContext$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_AppContext$State>) => void): void;
    public static "sun/awt/AppContext$State/values()[Lsun/awt/AppContext$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_AppContext$State>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lsun/awt/AppContext$State;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext$State) => void): void;
    public static "sun/awt/AppContext$State/valueOf(Ljava/lang/String;)Lsun/awt/AppContext$State;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext$State) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/AppContext$State/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/AppContext$State/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_beans_PropertyChangeSupport extends java_lang_Object implements java_io_Serializable {
    public "java/beans/PropertyChangeSupport/map": java_beans_PropertyChangeSupport$PropertyChangeListenerMap;
    public "java/beans/PropertyChangeSupport/source": java_lang_Object;
    public static "java/beans/PropertyChangeSupport/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/beans/PropertyChangeSupport/serialVersionUID": Long;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/removePropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPropertyChangeListeners()[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "java/beans/PropertyChangeSupport/getPropertyChangeListeners()[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/removePropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPropertyChangeListeners(Ljava/lang/String;)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "java/beans/PropertyChangeSupport/getPropertyChangeListeners(Ljava/lang/String;)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "firePropertyChange(Ljava/lang/String;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/firePropertyChange(Ljava/lang/String;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/firePropertyChange(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;ZZ)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/firePropertyChange(Ljava/lang/String;ZZ)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/beans/PropertyChangeEvent;)V"(thread: JVMThread, args: [java_beans_PropertyChangeEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/firePropertyChange(Ljava/beans/PropertyChangeEvent;)V"(thread: JVMThread, args: [java_beans_PropertyChangeEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fire([Ljava/beans/PropertyChangeListener;Ljava/beans/PropertyChangeEvent;)V"(thread: JVMThread, args: [JVMArray<java_beans_PropertyChangeListener>, java_beans_PropertyChangeEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/beans/PropertyChangeSupport/fire([Ljava/beans/PropertyChangeListener;Ljava/beans/PropertyChangeEvent;)V"(thread: JVMThread, args: [JVMArray<java_beans_PropertyChangeListener>, java_beans_PropertyChangeEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "fireIndexedPropertyChange(Ljava/lang/String;ILjava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/fireIndexedPropertyChange(Ljava/lang/String;ILjava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "fireIndexedPropertyChange(Ljava/lang/String;III)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/fireIndexedPropertyChange(Ljava/lang/String;III)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fireIndexedPropertyChange(Ljava/lang/String;IZZ)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/fireIndexedPropertyChange(Ljava/lang/String;IZZ)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "hasListeners(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/beans/PropertyChangeSupport/hasListeners(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/beans/PropertyChangeSupport/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_beans_PropertyChangeSupport$PropertyChangeListenerMap extends java_beans_ChangeListenerMap {
    public static "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/EMPTY": JVMArray<java_beans_PropertyChangeListener>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "newArray(I)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/newArray(I)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "newProxy(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable, rv?: java_beans_PropertyChangeListener) => void): void;
    public "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/newProxy(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable, rv?: java_beans_PropertyChangeListener) => void): void;
    public "extract(Ljava/beans/PropertyChangeListener;)Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable, rv?: java_beans_PropertyChangeListener) => void): void;
    public "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/extract(Ljava/beans/PropertyChangeListener;)Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable, rv?: java_beans_PropertyChangeListener) => void): void;
    public "extract(Ljava/util/EventListener;)Ljava/util/EventListener;"(thread: JVMThread, args: [java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: java_util_EventListener) => void): void;
    public "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/extract(Ljava/util/EventListener;)Ljava/util/EventListener;"(thread: JVMThread, args: [java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: java_util_EventListener) => void): void;
    public "newProxy(Ljava/lang/String;Ljava/util/EventListener;)Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_String, java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: java_util_EventListener) => void): void;
    public "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/newProxy(Ljava/lang/String;Ljava/util/EventListener;)Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_String, java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: java_util_EventListener) => void): void;
    public "newArray(I)[Ljava/util/EventListener;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/newArray(I)[Ljava/util/EventListener;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "<init>(Ljava/beans/PropertyChangeSupport$1;)V"(thread: JVMThread, args: [java_beans_PropertyChangeSupport$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/<init>(Ljava/beans/PropertyChangeSupport$1;)V"(thread: JVMThread, args: [java_beans_PropertyChangeSupport$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/beans/PropertyChangeSupport$PropertyChangeListenerMap/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_beans_PropertyChangeSupport$1 extends java_lang_Object {
  }
  export class java_beans_ChangeListenerMap extends java_lang_Object {
    public "java/beans/ChangeListenerMap/map": java_util_Map;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/ChangeListenerMap/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "newArray(I)[Ljava/util/EventListener;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/beans/ChangeListenerMap/newArray(I)[Ljava/util/EventListener;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "newProxy(Ljava/lang/String;Ljava/util/EventListener;)Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_String, java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: java_util_EventListener) => void): void;
    public "java/beans/ChangeListenerMap/newProxy(Ljava/lang/String;Ljava/util/EventListener;)Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_String, java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: java_util_EventListener) => void): void;
    public "add(Ljava/lang/String;Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_String, java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/ChangeListenerMap/add(Ljava/lang/String;Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_String, java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/String;Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_String, java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/ChangeListenerMap/remove(Ljava/lang/String;Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_String, java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/String;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/beans/ChangeListenerMap/get(Ljava/lang/String;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "set(Ljava/lang/String;[Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_util_EventListener>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/beans/ChangeListenerMap/set(Ljava/lang/String;[Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_util_EventListener>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getListeners()[Ljava/util/EventListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/beans/ChangeListenerMap/getListeners()[Ljava/util/EventListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "getListeners(Ljava/lang/String;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/beans/ChangeListenerMap/getListeners(Ljava/lang/String;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "hasListeners(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/beans/ChangeListenerMap/hasListeners(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getEntries()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/beans/ChangeListenerMap/getEntries()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "extract(Ljava/util/EventListener;)Ljava/util/EventListener;"(thread: JVMThread, args: [java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: java_util_EventListener) => void): void;
    public "java/beans/ChangeListenerMap/extract(Ljava/util/EventListener;)Ljava/util/EventListener;"(thread: JVMThread, args: [java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: java_util_EventListener) => void): void;
  }
  export class sun_util_logging_PlatformLogger extends java_lang_Object {
    public static "sun/util/logging/PlatformLogger/OFF": number;
    public static "sun/util/logging/PlatformLogger/SEVERE": number;
    public static "sun/util/logging/PlatformLogger/WARNING": number;
    public static "sun/util/logging/PlatformLogger/INFO": number;
    public static "sun/util/logging/PlatformLogger/CONFIG": number;
    public static "sun/util/logging/PlatformLogger/FINE": number;
    public static "sun/util/logging/PlatformLogger/FINER": number;
    public static "sun/util/logging/PlatformLogger/FINEST": number;
    public static "sun/util/logging/PlatformLogger/ALL": number;
    public static "sun/util/logging/PlatformLogger/DEFAULT_LEVEL": sun_util_logging_PlatformLogger$Level;
    public static "sun/util/logging/PlatformLogger/loggingEnabled": number;
    public static "sun/util/logging/PlatformLogger/loggers": java_util_Map;
    public "sun/util/logging/PlatformLogger/loggerProxy": sun_util_logging_PlatformLogger$LoggerProxy;
    public "sun/util/logging/PlatformLogger/javaLoggerProxy": sun_util_logging_PlatformLogger$JavaLoggerProxy;
    public static "getLogger(Ljava/lang/String;)Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "sun/util/logging/PlatformLogger/getLogger(Ljava/lang/String;)Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "redirectPlatformLoggers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/logging/PlatformLogger/redirectPlatformLoggers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "redirectToJavaLoggerProxy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/redirectToJavaLoggerProxy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/logging/PlatformLogger/isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/util/logging/PlatformLogger/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isLoggable(Lsun/util/logging/PlatformLogger$Level;)Z"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/logging/PlatformLogger/isLoggable(Lsun/util/logging/PlatformLogger$Level;)Z"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "level()Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public "sun/util/logging/PlatformLogger/level()Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public "setLevel(Lsun/util/logging/PlatformLogger$Level;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/setLevel(Lsun/util/logging/PlatformLogger$Level;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "severe(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/severe(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "severe(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/severe(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "severe(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/severe(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "warning(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/warning(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "warning(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/warning(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "warning(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/warning(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "info(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/info(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "info(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/info(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "info(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/info(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "config(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/config(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "config(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/config(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "config(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/config(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "fine(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/fine(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "fine(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/fine(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "fine(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/fine(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "finer(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/finer(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "finer(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/finer(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "finer(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/finer(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "finest(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/finest(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "finest(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/finest(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "finest(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger/finest(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000()Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public static "sun/util/logging/PlatformLogger/access$000()Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/logging/PlatformLogger/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_logging_PlatformLogger$JavaLoggerProxy extends sun_util_logging_PlatformLogger$LoggerProxy {
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/javaLogger": java_lang_Object;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Lsun/util/logging/PlatformLogger$Level;)V"(thread: JVMThread, args: [java_lang_String, sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/<init>(Ljava/lang/String;Lsun/util/logging/PlatformLogger$Level;)V"(thread: JVMThread, args: [java_lang_String, sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLevel()Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/getLevel()Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public "setLevel(Lsun/util/logging/PlatformLogger$Level;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/setLevel(Lsun/util/logging/PlatformLogger$Level;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "isLoggable(Lsun/util/logging/PlatformLogger$Level;)Z"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/logging/PlatformLogger$JavaLoggerProxy/isLoggable(Lsun/util/logging/PlatformLogger$Level;)Z"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/logging/PlatformLogger$JavaLoggerProxy/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_util_logging_PlatformLogger$LoggerProxy extends java_lang_Object {
    public "sun/util/logging/PlatformLogger$LoggerProxy/name": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$LoggerProxy/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/logging/PlatformLogger$LoggerProxy/isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLevel()Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public "sun/util/logging/PlatformLogger$LoggerProxy/getLevel()Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public "setLevel(Lsun/util/logging/PlatformLogger$Level;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$LoggerProxy/setLevel(Lsun/util/logging/PlatformLogger$Level;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$LoggerProxy/doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$LoggerProxy/doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$LoggerProxy/doLog(Lsun/util/logging/PlatformLogger$Level;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "isLoggable(Lsun/util/logging/PlatformLogger$Level;)Z"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/logging/PlatformLogger$LoggerProxy/isLoggable(Lsun/util/logging/PlatformLogger$Level;)Z"(thread: JVMThread, args: [sun_util_logging_PlatformLogger$Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_util_logging_PlatformLogger$Level extends java_lang_Enum {
    public static "sun/util/logging/PlatformLogger$Level/ALL": sun_util_logging_PlatformLogger$Level;
    public static "sun/util/logging/PlatformLogger$Level/FINEST": sun_util_logging_PlatformLogger$Level;
    public static "sun/util/logging/PlatformLogger$Level/FINER": sun_util_logging_PlatformLogger$Level;
    public static "sun/util/logging/PlatformLogger$Level/FINE": sun_util_logging_PlatformLogger$Level;
    public static "sun/util/logging/PlatformLogger$Level/CONFIG": sun_util_logging_PlatformLogger$Level;
    public static "sun/util/logging/PlatformLogger$Level/INFO": sun_util_logging_PlatformLogger$Level;
    public static "sun/util/logging/PlatformLogger$Level/WARNING": sun_util_logging_PlatformLogger$Level;
    public static "sun/util/logging/PlatformLogger$Level/SEVERE": sun_util_logging_PlatformLogger$Level;
    public static "sun/util/logging/PlatformLogger$Level/OFF": sun_util_logging_PlatformLogger$Level;
    public "sun/util/logging/PlatformLogger$Level/javaLevel": java_lang_Object;
    public static "sun/util/logging/PlatformLogger$Level/LEVEL_VALUES": JVMArray<number>;
    public static "sun/util/logging/PlatformLogger$Level/$VALUES": JVMArray<sun_util_logging_PlatformLogger$Level>;
    public static "values()[Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_util_logging_PlatformLogger$Level>) => void): void;
    public static "sun/util/logging/PlatformLogger$Level/values()[Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_util_logging_PlatformLogger$Level>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public static "sun/util/logging/PlatformLogger$Level/valueOf(Ljava/lang/String;)Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/util/logging/PlatformLogger$Level/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/util/logging/PlatformLogger$Level/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "valueOf(I)Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public static "sun/util/logging/PlatformLogger$Level/valueOf(I)Lsun/util/logging/PlatformLogger$Level;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger$Level) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/util/logging/PlatformLogger$Level/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Window extends java_awt_Container implements javax_accessibility_Accessible {
    public "java/awt/Window/warningString": java_lang_String;
    public "java/awt/Window/icons": java_util_List;
    public "java/awt/Window/temporaryLostComponent": java_awt_Component;
    public static "java/awt/Window/systemSyncLWRequests": number;
    public "java/awt/Window/syncLWRequests": number;
    public "java/awt/Window/beforeFirstShow": number;
    public "java/awt/Window/disposing": number;
    public "java/awt/Window/disposerRecord": java_awt_Window$WindowDisposerRecord;
    public static "java/awt/Window/OPENED": number;
    public "java/awt/Window/state": number;
    public "java/awt/Window/alwaysOnTop": number;
    public static "java/awt/Window/allWindows": sun_awt_util_IdentityArrayList;
    public "java/awt/Window/ownedWindowList": java_util_Vector;
    public "java/awt/Window/weakThis": java_lang_ref_WeakReference;
    public "java/awt/Window/showWithParent": number;
    public "java/awt/Window/modalBlocker": java_awt_Dialog;
    public "java/awt/Window/modalExclusionType": java_awt_Dialog$ModalExclusionType;
    public "java/awt/Window/windowListener": java_awt_event_WindowListener;
    public "java/awt/Window/windowStateListener": java_awt_event_WindowStateListener;
    public "java/awt/Window/windowFocusListener": java_awt_event_WindowFocusListener;
    public "java/awt/Window/inputContext": java_awt_im_InputContext;
    public "java/awt/Window/inputContextLock": java_lang_Object;
    public "java/awt/Window/focusMgr": java_awt_FocusManager;
    public "java/awt/Window/focusableWindowState": number;
    public "java/awt/Window/autoRequestFocus": number;
    public "java/awt/Window/isInShow": number;
    public "java/awt/Window/opacity": number;
    public "java/awt/Window/shape": java_awt_Shape;
    public static "java/awt/Window/base": java_lang_String;
    public static "java/awt/Window/nameCounter": number;
    public static "java/awt/Window/serialVersionUID": Long;
    public static "java/awt/Window/log": sun_util_logging_PlatformLogger;
    public static "java/awt/Window/locationByPlatformProp": number;
    public "java/awt/Window/isTrayIconWindow": number;
    public "java/awt/Window/securityWarningWidth": number;
    public "java/awt/Window/securityWarningHeight": number;
    public "java/awt/Window/securityWarningPointX": number;
    public "java/awt/Window/securityWarningPointY": number;
    public "java/awt/Window/securityWarningAlignmentX": number;
    public "java/awt/Window/securityWarningAlignmentY": number;
    public "java/awt/Window/anchor": java_lang_Object;
    public static "java/awt/Window/beforeFirstWindowShown": java_util_concurrent_atomic_AtomicBoolean;
    public "java/awt/Window/type": java_awt_Window$Type;
    public "java/awt/Window/windowSerializedDataVersion": number;
    public "java/awt/Window/locationByPlatform": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Window/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/<init>(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "initGC(Ljava/awt/GraphicsConfiguration;)Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "java/awt/Window/initGC(Ljava/awt/GraphicsConfiguration;)Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "init(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/init(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Frame;)V"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/<init>(Ljava/awt/Frame;)V"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/<init>(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_Window, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/<init>(Ljava/awt/Window;Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_Window, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "ownedInit(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/ownedInit(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Window/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getIconImages()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/awt/Window/getIconImages()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "setIconImages(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setIconImages(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "setIconImage(Ljava/awt/Image;)V"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setIconImage(Ljava/awt/Image;)V"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "pack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/pack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setMinimumSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setMinimumSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocation(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setLocation(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "reshape(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/reshape(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setClientSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setClientSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "closeSplashScreen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/closeSplashScreen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "show()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/show()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "updateChildFocusableWindowState(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Window/updateChildFocusableWindowState(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "postWindowEvent(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/postWindowEvent(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "hide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/hide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearMostRecentFocusOwnerOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/clearMostRecentFocusOwnerOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "disposeImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/disposeImpl()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "doDispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/doDispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "adjustListeningChildrenOnParent(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/adjustListeningChildrenOnParent(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "adjustDecendantsOnParent(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/adjustDecendantsOnParent(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toFront()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/toFront()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toFront_NoClientCode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/toFront_NoClientCode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toBack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/toBack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toBack_NoClientCode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/toBack_NoClientCode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getToolkit()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public "java/awt/Window/getToolkit()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public "getWarningString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Window/getWarningString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setWarningString()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setWarningString()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/awt/Window/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "getInputContext()Ljava/awt/im/InputContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputContext) => void): void;
    public "java/awt/Window/getInputContext()Ljava/awt/im/InputContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputContext) => void): void;
    public "setCursor(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setCursor(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOwner()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "java/awt/Window/getOwner()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "getOwner_NoClientCode()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "java/awt/Window/getOwner_NoClientCode()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "getOwnedWindows()[Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public "java/awt/Window/getOwnedWindows()[Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public "getOwnedWindows_NoClientCode()[Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public "java/awt/Window/getOwnedWindows_NoClientCode()[Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public "isModalBlocked()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isModalBlocked()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setModalBlocked(Ljava/awt/Dialog;ZZ)V"(thread: JVMThread, args: [java_awt_Dialog, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setModalBlocked(Ljava/awt/Dialog;ZZ)V"(thread: JVMThread, args: [java_awt_Dialog, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getModalBlocker()Ljava/awt/Dialog;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog) => void): void;
    public "java/awt/Window/getModalBlocker()Ljava/awt/Dialog;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog) => void): void;
    public static "getAllWindows()Lsun/awt/util/IdentityArrayList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_util_IdentityArrayList) => void): void;
    public static "java/awt/Window/getAllWindows()Lsun/awt/util/IdentityArrayList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_util_IdentityArrayList) => void): void;
    public static "getAllUnblockedWindows()Lsun/awt/util/IdentityArrayList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_util_IdentityArrayList) => void): void;
    public static "java/awt/Window/getAllUnblockedWindows()Lsun/awt/util/IdentityArrayList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_util_IdentityArrayList) => void): void;
    public static "getWindows(Lsun/awt/AppContext;)[Ljava/awt/Window;"(thread: JVMThread, args: [sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public static "java/awt/Window/getWindows(Lsun/awt/AppContext;)[Ljava/awt/Window;"(thread: JVMThread, args: [sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public static "getWindows()[Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public static "java/awt/Window/getWindows()[Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public static "getOwnerlessWindows()[Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public static "java/awt/Window/getOwnerlessWindows()[Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window>) => void): void;
    public "getDocumentRoot()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "java/awt/Window/getDocumentRoot()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "setModalExclusionType(Ljava/awt/Dialog$ModalExclusionType;)V"(thread: JVMThread, args: [java_awt_Dialog$ModalExclusionType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setModalExclusionType(Ljava/awt/Dialog$ModalExclusionType;)V"(thread: JVMThread, args: [java_awt_Dialog$ModalExclusionType], cb?: (e?: java_lang_Throwable) => void): void;
    public "getModalExclusionType()Ljava/awt/Dialog$ModalExclusionType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog$ModalExclusionType) => void): void;
    public "java/awt/Window/getModalExclusionType()Ljava/awt/Dialog$ModalExclusionType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog$ModalExclusionType) => void): void;
    public "isModalExcluded(Ljava/awt/Dialog$ModalExclusionType;)Z"(thread: JVMThread, args: [java_awt_Dialog$ModalExclusionType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isModalExcluded(Ljava/awt/Dialog$ModalExclusionType;)Z"(thread: JVMThread, args: [java_awt_Dialog$ModalExclusionType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "updateChildrenBlocking()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/updateChildrenBlocking()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addWindowListener(Ljava/awt/event/WindowListener;)V"(thread: JVMThread, args: [java_awt_event_WindowListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/addWindowListener(Ljava/awt/event/WindowListener;)V"(thread: JVMThread, args: [java_awt_event_WindowListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "addWindowStateListener(Ljava/awt/event/WindowStateListener;)V"(thread: JVMThread, args: [java_awt_event_WindowStateListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/addWindowStateListener(Ljava/awt/event/WindowStateListener;)V"(thread: JVMThread, args: [java_awt_event_WindowStateListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "addWindowFocusListener(Ljava/awt/event/WindowFocusListener;)V"(thread: JVMThread, args: [java_awt_event_WindowFocusListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/addWindowFocusListener(Ljava/awt/event/WindowFocusListener;)V"(thread: JVMThread, args: [java_awt_event_WindowFocusListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeWindowListener(Ljava/awt/event/WindowListener;)V"(thread: JVMThread, args: [java_awt_event_WindowListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/removeWindowListener(Ljava/awt/event/WindowListener;)V"(thread: JVMThread, args: [java_awt_event_WindowListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeWindowStateListener(Ljava/awt/event/WindowStateListener;)V"(thread: JVMThread, args: [java_awt_event_WindowStateListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/removeWindowStateListener(Ljava/awt/event/WindowStateListener;)V"(thread: JVMThread, args: [java_awt_event_WindowStateListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeWindowFocusListener(Ljava/awt/event/WindowFocusListener;)V"(thread: JVMThread, args: [java_awt_event_WindowFocusListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/removeWindowFocusListener(Ljava/awt/event/WindowFocusListener;)V"(thread: JVMThread, args: [java_awt_event_WindowFocusListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWindowListeners()[Ljava/awt/event/WindowListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_WindowListener>) => void): void;
    public "java/awt/Window/getWindowListeners()[Ljava/awt/event/WindowListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_WindowListener>) => void): void;
    public "getWindowFocusListeners()[Ljava/awt/event/WindowFocusListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_WindowFocusListener>) => void): void;
    public "java/awt/Window/getWindowFocusListeners()[Ljava/awt/event/WindowFocusListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_WindowFocusListener>) => void): void;
    public "getWindowStateListeners()[Ljava/awt/event/WindowStateListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_WindowStateListener>) => void): void;
    public "java/awt/Window/getWindowStateListeners()[Ljava/awt/event/WindowStateListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_WindowStateListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/Window/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processWindowEvent(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/processWindowEvent(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processWindowFocusEvent(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/processWindowFocusEvent(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processWindowStateEvent(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/processWindowStateEvent(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "preProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/preProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "postProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/postProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "setAlwaysOnTop(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setAlwaysOnTop(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOwnedWindowsAlwaysOnTop(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setOwnedWindowsAlwaysOnTop(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAlwaysOnTopSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isAlwaysOnTopSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAlwaysOnTop()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isAlwaysOnTop()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFocusOwner()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Window/getFocusOwner()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getMostRecentFocusOwner()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Window/getMostRecentFocusOwner()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "isActive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isActive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFocused()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isFocused()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFocusTraversalKeys(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/awt/Window/getFocusTraversalKeys(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "setFocusCycleRoot(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setFocusCycleRoot(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isFocusCycleRoot()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isFocusCycleRoot()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFocusCycleRootAncestor()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Window/getFocusCycleRootAncestor()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "isFocusableWindow()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isFocusableWindow()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFocusableWindowState()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/getFocusableWindowState()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFocusableWindowState(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setFocusableWindowState(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setAutoRequestFocus(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setAutoRequestFocus(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAutoRequestFocus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isAutoRequestFocus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "isValidateRoot()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isValidateRoot()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "dispatchEventImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/dispatchEventImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "postEvent(Ljava/awt/Event;)Z"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/postEvent(Ljava/awt/Event;)Z"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isShowing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isShowing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDisposing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isDisposing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "applyResourceBundle(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/applyResourceBundle(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "applyResourceBundle(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/applyResourceBundle(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "addOwnedWindow(Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/addOwnedWindow(Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeOwnedWindow(Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/removeOwnedWindow(Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
    public "connectOwnedWindow(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/connectOwnedWindow(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "addToWindowList()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/addToWindowList()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "removeFromWindowList(Lsun/awt/AppContext;Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [sun_awt_AppContext, java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Window/removeFromWindowList(Lsun/awt/AppContext;Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [sun_awt_AppContext, java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeFromWindowList()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/removeFromWindowList()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setType(Ljava/awt/Window$Type;)V"(thread: JVMThread, args: [java_awt_Window$Type], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setType(Ljava/awt/Window$Type;)V"(thread: JVMThread, args: [java_awt_Window$Type], cb?: (e?: java_lang_Throwable) => void): void;
    public "getType()Ljava/awt/Window$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window$Type) => void): void;
    public "java/awt/Window/getType()Ljava/awt/Window$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window$Type) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "initDeserializedWindow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/initDeserializedWindow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "deserializeResources(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/deserializeResources(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Window/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "setGraphicsConfiguration(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setGraphicsConfiguration(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocationRelativeTo(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setLocationRelativeTo(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "deliverMouseWheelToAncestor(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/deliverMouseWheelToAncestor(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchMouseWheelToAncestor(Ljava/awt/event/MouseWheelEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/dispatchMouseWheelToAncestor(Ljava/awt/event/MouseWheelEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createBufferStrategy(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/createBufferStrategy(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "createBufferStrategy(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/createBufferStrategy(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBufferStrategy()Ljava/awt/image/BufferStrategy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferStrategy) => void): void;
    public "java/awt/Window/getBufferStrategy()Ljava/awt/image/BufferStrategy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferStrategy) => void): void;
    public "getTemporaryLostComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Window/getTemporaryLostComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "setTemporaryLostComponent(Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Window/setTemporaryLostComponent(Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "canContainFocusOwner(Ljava/awt/Component;)Z"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/canContainFocusOwner(Ljava/awt/Component;)Z"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLocationByPlatform(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setLocationByPlatform(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isLocationByPlatform()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isLocationByPlatform()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setBounds(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setBounds(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "isRecursivelyVisible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isRecursivelyVisible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOpacity()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/getOpacity()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getShape()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/Window/getShape()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "setShape(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setShape(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "java/awt/Window/getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "isOpaque()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Window/isOpaque()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "updateWindow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/updateWindow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setLayersOpaque(Ljava/awt/Component;Z)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Window/setLayersOpaque(Ljava/awt/Component;Z)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Window/getContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "applyCompoundShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/applyCompoundShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "applyCurrentShape()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/applyCurrentShape()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnReshaping()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/mixOnReshaping()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocationOnWindow()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Window/getLocationOnWindow()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public static "limit(DDD)D"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/limit(DDD)D"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "calculateSecurityWarningPosition(DDDD)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/Window/calculateSecurityWarningPosition(DDDD)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "updateZOrder()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window/updateZOrder()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/awt/Window;)Ljava/lang/ref/WeakReference;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_WeakReference) => void): void;
    public static "java/awt/Window/access$000(Ljava/awt/Window;)Ljava/lang/ref/WeakReference;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_WeakReference) => void): void;
    public static "access$100(Lsun/awt/AppContext;Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [sun_awt_AppContext, java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Window/access$100(Lsun/awt/AppContext;Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [sun_awt_AppContext, java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$202(Ljava/awt/Window;Z)Z"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$202(Ljava/awt/Window;Z)Z"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$300(Ljava/awt/Window;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/awt/Window/access$300(Ljava/awt/Window;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$400(Ljava/awt/Window;)F"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$400(Ljava/awt/Window;)F"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Window/access$500(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$600(Ljava/awt/Window;)I"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$600(Ljava/awt/Window;)I"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$700(Ljava/awt/Window;)I"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$700(Ljava/awt/Window;)I"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$602(Ljava/awt/Window;I)I"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$602(Ljava/awt/Window;I)I"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$702(Ljava/awt/Window;I)I"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$702(Ljava/awt/Window;I)I"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$802(Ljava/awt/Window;D)D"(thread: JVMThread, args: [java_awt_Window, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$802(Ljava/awt/Window;D)D"(thread: JVMThread, args: [java_awt_Window, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$902(Ljava/awt/Window;D)D"(thread: JVMThread, args: [java_awt_Window, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$902(Ljava/awt/Window;D)D"(thread: JVMThread, args: [java_awt_Window, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1002(Ljava/awt/Window;F)F"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$1002(Ljava/awt/Window;F)F"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1102(Ljava/awt/Window;F)F"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$1102(Ljava/awt/Window;F)F"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1200(Ljava/awt/Window;DDDD)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [java_awt_Window, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public static "java/awt/Window/access$1200(Ljava/awt/Window;DDDD)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [java_awt_Window, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public static "access$1300(Ljava/awt/Window;)Z"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Window/access$1300(Ljava/awt/Window;)Z"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Window/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_image_BufferStrategy extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferStrategy/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCapabilities()Ljava/awt/BufferCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_BufferCapabilities) => void): void;
    public "java/awt/image/BufferStrategy/getCapabilities()Ljava/awt/BufferCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_BufferCapabilities) => void): void;
    public "getDrawGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "java/awt/image/BufferStrategy/getDrawGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "contentsLost()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferStrategy/contentsLost()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contentsRestored()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/BufferStrategy/contentsRestored()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "show()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferStrategy/show()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/BufferStrategy/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_MouseWheelEvent extends java_awt_event_MouseEvent {
    public static "java/awt/event/MouseWheelEvent/WHEEL_UNIT_SCROLL": number;
    public static "java/awt/event/MouseWheelEvent/WHEEL_BLOCK_SCROLL": number;
    public "java/awt/event/MouseWheelEvent/scrollType": number;
    public "java/awt/event/MouseWheelEvent/scrollAmount": number;
    public "java/awt/event/MouseWheelEvent/wheelRotation": number;
    public "java/awt/event/MouseWheelEvent/preciseWheelRotation": number;
    public static "java/awt/event/MouseWheelEvent/serialVersionUID": Long;
    public "<init>(Ljava/awt/Component;IJIIIIZIII)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseWheelEvent/<init>(Ljava/awt/Component;IJIIIIZIII)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;IJIIIIIIZIII)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseWheelEvent/<init>(Ljava/awt/Component;IJIIIIIIZIII)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;IJIIIIIIZIIID)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number, number, number, number, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseWheelEvent/<init>(Ljava/awt/Component;IJIIIIIIZIIID)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number, number, number, number, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getScrollType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseWheelEvent/getScrollType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getScrollAmount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseWheelEvent/getScrollAmount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWheelRotation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseWheelEvent/getWheelRotation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPreciseWheelRotation()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseWheelEvent/getPreciseWheelRotation()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUnitsToScroll()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseWheelEvent/getUnitsToScroll()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/MouseWheelEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_event_MouseEvent extends java_awt_event_InputEvent {
    public static "java/awt/event/MouseEvent/MOUSE_FIRST": number;
    public static "java/awt/event/MouseEvent/MOUSE_LAST": number;
    public static "java/awt/event/MouseEvent/MOUSE_CLICKED": number;
    public static "java/awt/event/MouseEvent/MOUSE_PRESSED": number;
    public static "java/awt/event/MouseEvent/MOUSE_RELEASED": number;
    public static "java/awt/event/MouseEvent/MOUSE_MOVED": number;
    public static "java/awt/event/MouseEvent/MOUSE_ENTERED": number;
    public static "java/awt/event/MouseEvent/MOUSE_EXITED": number;
    public static "java/awt/event/MouseEvent/MOUSE_DRAGGED": number;
    public static "java/awt/event/MouseEvent/MOUSE_WHEEL": number;
    public static "java/awt/event/MouseEvent/NOBUTTON": number;
    public static "java/awt/event/MouseEvent/BUTTON1": number;
    public static "java/awt/event/MouseEvent/BUTTON2": number;
    public static "java/awt/event/MouseEvent/BUTTON3": number;
    public "java/awt/event/MouseEvent/x": number;
    public "java/awt/event/MouseEvent/y": number;
    public "java/awt/event/MouseEvent/xAbs": number;
    public "java/awt/event/MouseEvent/yAbs": number;
    public "java/awt/event/MouseEvent/clickCount": number;
    public "java/awt/event/MouseEvent/button": number;
    public "java/awt/event/MouseEvent/popupTrigger": number;
    public static "java/awt/event/MouseEvent/serialVersionUID": Long;
    public static "java/awt/event/MouseEvent/cachedNumberOfButtons": number;
    public "java/awt/event/MouseEvent/shouldExcludeButtonFromExtModifiers": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/event/MouseEvent/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/event/MouseEvent/getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "getXOnScreen()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseEvent/getXOnScreen()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getYOnScreen()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseEvent/getYOnScreen()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/awt/Component;IJIIIIZI)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseEvent/<init>(Ljava/awt/Component;IJIIIIZI)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;IJIIIIZ)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseEvent/<init>(Ljava/awt/Component;IJIIIIZ)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getModifiersEx()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseEvent/getModifiersEx()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/awt/Component;IJIIIIIIZI)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseEvent/<init>(Ljava/awt/Component;IJIIIIIIZI)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseEvent/getX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseEvent/getY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPoint()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/event/MouseEvent/getPoint()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "translatePoint(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseEvent/translatePoint(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getClickCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseEvent/getClickCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getButton()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseEvent/getButton()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPopupTrigger()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/MouseEvent/isPopupTrigger()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getMouseModifiersText(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/event/MouseEvent/getMouseModifiersText(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/MouseEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setNewModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseEvent/setNewModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOldModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseEvent/setOldModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/MouseEvent/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/event/MouseEvent/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_InputEvent extends java_awt_event_ComponentEvent {
    public static "java/awt/event/InputEvent/logger": sun_util_logging_PlatformLogger;
    public static "java/awt/event/InputEvent/SHIFT_MASK": number;
    public static "java/awt/event/InputEvent/CTRL_MASK": number;
    public static "java/awt/event/InputEvent/META_MASK": number;
    public static "java/awt/event/InputEvent/ALT_MASK": number;
    public static "java/awt/event/InputEvent/ALT_GRAPH_MASK": number;
    public static "java/awt/event/InputEvent/BUTTON1_MASK": number;
    public static "java/awt/event/InputEvent/BUTTON2_MASK": number;
    public static "java/awt/event/InputEvent/BUTTON3_MASK": number;
    public static "java/awt/event/InputEvent/SHIFT_DOWN_MASK": number;
    public static "java/awt/event/InputEvent/CTRL_DOWN_MASK": number;
    public static "java/awt/event/InputEvent/META_DOWN_MASK": number;
    public static "java/awt/event/InputEvent/ALT_DOWN_MASK": number;
    public static "java/awt/event/InputEvent/BUTTON1_DOWN_MASK": number;
    public static "java/awt/event/InputEvent/BUTTON2_DOWN_MASK": number;
    public static "java/awt/event/InputEvent/BUTTON3_DOWN_MASK": number;
    public static "java/awt/event/InputEvent/ALT_GRAPH_DOWN_MASK": number;
    public static "java/awt/event/InputEvent/BUTTON_DOWN_MASK": JVMArray<number>;
    public static "java/awt/event/InputEvent/FIRST_HIGH_BIT": number;
    public static "java/awt/event/InputEvent/JDK_1_3_MODIFIERS": number;
    public static "java/awt/event/InputEvent/HIGH_MODIFIERS": number;
    public "java/awt/event/InputEvent/when": Long;
    public "java/awt/event/InputEvent/modifiers": number;
    public "java/awt/event/InputEvent/canAccessSystemClipboard": number;
    public static "java/awt/event/InputEvent/serialVersionUID": Long;
    public static "getButtonDownMasks()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/event/InputEvent/getButtonDownMasks()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getMaskForButton(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/event/InputEvent/getMaskForButton(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/event/InputEvent/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;IJI)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InputEvent/<init>(Ljava/awt/Component;IJI)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "canAccessSystemClipboard()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputEvent/canAccessSystemClipboard()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isShiftDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputEvent/isShiftDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isControlDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputEvent/isControlDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMetaDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputEvent/isMetaDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAltDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputEvent/isAltDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAltGraphDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputEvent/isAltGraphDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWhen()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/awt/event/InputEvent/getWhen()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputEvent/getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getModifiersEx()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputEvent/getModifiersEx()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InputEvent/consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isConsumed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputEvent/isConsumed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getModifiersExText(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/event/InputEvent/getModifiersExText(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/event/InputEvent/access$000()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/event/InputEvent/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_ComponentEvent extends java_awt_AWTEvent {
    public static "java/awt/event/ComponentEvent/COMPONENT_FIRST": number;
    public static "java/awt/event/ComponentEvent/COMPONENT_LAST": number;
    public static "java/awt/event/ComponentEvent/COMPONENT_MOVED": number;
    public static "java/awt/event/ComponentEvent/COMPONENT_RESIZED": number;
    public static "java/awt/event/ComponentEvent/COMPONENT_SHOWN": number;
    public static "java/awt/event/ComponentEvent/COMPONENT_HIDDEN": number;
    public static "java/awt/event/ComponentEvent/serialVersionUID": Long;
    public "<init>(Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/ComponentEvent/<init>(Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/event/ComponentEvent/getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/ComponentEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class javax_accessibility_AccessibleContext extends java_lang_Object {
    public "javax/accessibility/AccessibleContext/targetAppContext": sun_awt_AppContext;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_NAME_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_DESCRIPTION_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_STATE_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_VALUE_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_SELECTION_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_CARET_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_VISIBLE_DATA_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_CHILD_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_ACTIVE_DESCENDANT_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_TABLE_CAPTION_CHANGED": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_TABLE_SUMMARY_CHANGED": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_TABLE_MODEL_CHANGED": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_TABLE_ROW_HEADER_CHANGED": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_TABLE_ROW_DESCRIPTION_CHANGED": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_TABLE_COLUMN_HEADER_CHANGED": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_TABLE_COLUMN_DESCRIPTION_CHANGED": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_ACTION_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_HYPERTEXT_OFFSET": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_TEXT_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_INVALIDATE_CHILDREN": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_TEXT_ATTRIBUTES_CHANGED": java_lang_String;
    public static "javax/accessibility/AccessibleContext/ACCESSIBLE_COMPONENT_BOUNDS_CHANGED": java_lang_String;
    public "javax/accessibility/AccessibleContext/accessibleParent": javax_accessibility_Accessible;
    public "javax/accessibility/AccessibleContext/accessibleName": java_lang_String;
    public "javax/accessibility/AccessibleContext/accessibleDescription": java_lang_String;
    public "javax/accessibility/AccessibleContext/accessibleChangeSupport": java_beans_PropertyChangeSupport;
    public "javax/accessibility/AccessibleContext/relationSet": javax_accessibility_AccessibleRelationSet;
    public "javax/accessibility/AccessibleContext/nativeAXResource": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleContext/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setAccessibleName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleContext/setAccessibleName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setAccessibleDescription(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleContext/setAccessibleDescription(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleRole()Ljavax/accessibility/AccessibleRole;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleRole) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleRole()Ljavax/accessibility/AccessibleRole;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleRole) => void): void;
    public "getAccessibleStateSet()Ljavax/accessibility/AccessibleStateSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleStateSet) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleStateSet()Ljavax/accessibility/AccessibleStateSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleStateSet) => void): void;
    public "getAccessibleParent()Ljavax/accessibility/Accessible;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleParent()Ljavax/accessibility/Accessible;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    public "setAccessibleParent(Ljavax/accessibility/Accessible;)V"(thread: JVMThread, args: [javax_accessibility_Accessible], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleContext/setAccessibleParent(Ljavax/accessibility/Accessible;)V"(thread: JVMThread, args: [javax_accessibility_Accessible], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleIndexInParent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleIndexInParent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAccessibleChildrenCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleChildrenCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAccessibleChild(I)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleChild(I)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "javax/accessibility/AccessibleContext/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleContext/addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleContext/removePropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleAction()Ljavax/accessibility/AccessibleAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleAction) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleAction()Ljavax/accessibility/AccessibleAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleAction) => void): void;
    public "getAccessibleComponent()Ljavax/accessibility/AccessibleComponent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleComponent) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleComponent()Ljavax/accessibility/AccessibleComponent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleComponent) => void): void;
    public "getAccessibleSelection()Ljavax/accessibility/AccessibleSelection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleSelection) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleSelection()Ljavax/accessibility/AccessibleSelection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleSelection) => void): void;
    public "getAccessibleText()Ljavax/accessibility/AccessibleText;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleText) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleText()Ljavax/accessibility/AccessibleText;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleText) => void): void;
    public "getAccessibleEditableText()Ljavax/accessibility/AccessibleEditableText;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleEditableText) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleEditableText()Ljavax/accessibility/AccessibleEditableText;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleEditableText) => void): void;
    public "getAccessibleValue()Ljavax/accessibility/AccessibleValue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleValue) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleValue()Ljavax/accessibility/AccessibleValue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleValue) => void): void;
    public "getAccessibleIcon()[Ljavax/accessibility/AccessibleIcon;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_accessibility_AccessibleIcon>) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleIcon()[Ljavax/accessibility/AccessibleIcon;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_accessibility_AccessibleIcon>) => void): void;
    public "getAccessibleRelationSet()Ljavax/accessibility/AccessibleRelationSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleRelationSet) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleRelationSet()Ljavax/accessibility/AccessibleRelationSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleRelationSet) => void): void;
    public "getAccessibleTable()Ljavax/accessibility/AccessibleTable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleTable) => void): void;
    public "javax/accessibility/AccessibleContext/getAccessibleTable()Ljavax/accessibility/AccessibleTable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleTable) => void): void;
    public "firePropertyChange(Ljava/lang/String;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleContext/firePropertyChange(Ljava/lang/String;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$002(Ljavax/accessibility/AccessibleContext;Lsun/awt/AppContext;)Lsun/awt/AppContext;"(thread: JVMThread, args: [javax_accessibility_AccessibleContext, sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "javax/accessibility/AccessibleContext/access$002(Ljavax/accessibility/AccessibleContext;Lsun/awt/AppContext;)Lsun/awt/AppContext;"(thread: JVMThread, args: [javax_accessibility_AccessibleContext, sun_awt_AppContext], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "access$000(Ljavax/accessibility/AccessibleContext;)Lsun/awt/AppContext;"(thread: JVMThread, args: [javax_accessibility_AccessibleContext], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "javax/accessibility/AccessibleContext/access$000(Ljavax/accessibility/AccessibleContext;)Lsun/awt/AppContext;"(thread: JVMThread, args: [javax_accessibility_AccessibleContext], cb?: (e?: java_lang_Throwable, rv?: sun_awt_AppContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/accessibility/AccessibleContext/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface javax_accessibility_AccessibleTable extends java_lang_Object {
    "getAccessibleCaption()Ljavax/accessibility/Accessible;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    "setAccessibleCaption(Ljavax/accessibility/Accessible;)V"(thread: JVMThread, args: [javax_accessibility_Accessible], cb?: (e?: java_lang_Throwable) => void): void;
    "getAccessibleSummary()Ljavax/accessibility/Accessible;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    "setAccessibleSummary(Ljavax/accessibility/Accessible;)V"(thread: JVMThread, args: [javax_accessibility_Accessible], cb?: (e?: java_lang_Throwable) => void): void;
    "getAccessibleRowCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAccessibleColumnCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAccessibleAt(II)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    "getAccessibleRowExtentAt(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAccessibleColumnExtentAt(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAccessibleRowHeader()Ljavax/accessibility/AccessibleTable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleTable) => void): void;
    "setAccessibleRowHeader(Ljavax/accessibility/AccessibleTable;)V"(thread: JVMThread, args: [javax_accessibility_AccessibleTable], cb?: (e?: java_lang_Throwable) => void): void;
    "getAccessibleColumnHeader()Ljavax/accessibility/AccessibleTable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleTable) => void): void;
    "setAccessibleColumnHeader(Ljavax/accessibility/AccessibleTable;)V"(thread: JVMThread, args: [javax_accessibility_AccessibleTable], cb?: (e?: java_lang_Throwable) => void): void;
    "getAccessibleRowDescription(I)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    "setAccessibleRowDescription(ILjavax/accessibility/Accessible;)V"(thread: JVMThread, args: [number, javax_accessibility_Accessible], cb?: (e?: java_lang_Throwable) => void): void;
    "getAccessibleColumnDescription(I)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    "setAccessibleColumnDescription(ILjavax/accessibility/Accessible;)V"(thread: JVMThread, args: [number, javax_accessibility_Accessible], cb?: (e?: java_lang_Throwable) => void): void;
    "isAccessibleSelected(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isAccessibleRowSelected(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isAccessibleColumnSelected(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSelectedAccessibleRows()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    "getSelectedAccessibleColumns()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export interface javax_accessibility_AccessibleIcon extends java_lang_Object {
    "getAccessibleIconDescription()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "setAccessibleIconDescription(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "getAccessibleIconWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAccessibleIconHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface javax_accessibility_AccessibleValue extends java_lang_Object {
    "getCurrentAccessibleValue()Ljava/lang/Number;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Number) => void): void;
    "setCurrentAccessibleValue(Ljava/lang/Number;)Z"(thread: JVMThread, args: [java_lang_Number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinimumAccessibleValue()Ljava/lang/Number;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Number) => void): void;
    "getMaximumAccessibleValue()Ljava/lang/Number;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Number) => void): void;
  }
  export interface javax_accessibility_AccessibleEditableText extends java_lang_Object, javax_accessibility_AccessibleText {
    "setTextContents(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "insertTextAtIndex(ILjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "getTextRange(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "delete(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "cut(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "paste(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceText(IILjava/lang/String;)V"(thread: JVMThread, args: [number, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "selectText(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setAttributes(IILjavax/swing/text/AttributeSet;)V"(thread: JVMThread, args: [number, number, javax_swing_text_AttributeSet], cb?: (e?: java_lang_Throwable) => void): void;
    "getIndexAtPoint(Ljava/awt/Point;)I"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getCharacterBounds(I)Ljava/awt/Rectangle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    "getCharCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getCaretPosition()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAtIndex(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getAfterIndex(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getBeforeIndex(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getCharacterAttribute(I)Ljavax/swing/text/AttributeSet;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: javax_swing_text_AttributeSet) => void): void;
    "getSelectionStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSelectionEnd()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSelectedText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface javax_swing_text_AttributeSet extends java_lang_Object {
    "getAttributeCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isDefined(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isEqual(Ljavax/swing/text/AttributeSet;)Z"(thread: JVMThread, args: [javax_swing_text_AttributeSet], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "copyAttributes()Ljavax/swing/text/AttributeSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_swing_text_AttributeSet) => void): void;
    "getAttribute(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getAttributeNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    "containsAttribute(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "containsAttributes(Ljavax/swing/text/AttributeSet;)Z"(thread: JVMThread, args: [javax_swing_text_AttributeSet], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getResolveParent()Ljavax/swing/text/AttributeSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_swing_text_AttributeSet) => void): void;
  }
  export interface javax_accessibility_AccessibleText extends java_lang_Object {
    "getIndexAtPoint(Ljava/awt/Point;)I"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getCharacterBounds(I)Ljava/awt/Rectangle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    "getCharCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getCaretPosition()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAtIndex(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getAfterIndex(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getBeforeIndex(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getCharacterAttribute(I)Ljavax/swing/text/AttributeSet;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: javax_swing_text_AttributeSet) => void): void;
    "getSelectionStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSelectionEnd()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSelectedText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface javax_accessibility_AccessibleSelection extends java_lang_Object {
    "getAccessibleSelectionCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAccessibleSelection(I)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    "isAccessibleChildSelected(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAccessibleSelection(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "removeAccessibleSelection(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "clearAccessibleSelection()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "selectAllAccessibleSelection()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface javax_accessibility_AccessibleComponent extends java_lang_Object {
    "getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "getForeground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "getCursor()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    "setCursor(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "isVisible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "isShowing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "contains(Ljava/awt/Point;)Z"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "setLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    "setBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    "getSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "setSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    "getAccessibleAt(Ljava/awt/Point;)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    "isFocusTraversable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "requestFocus()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "addFocusListener(Ljava/awt/event/FocusListener;)V"(thread: JVMThread, args: [java_awt_event_FocusListener], cb?: (e?: java_lang_Throwable) => void): void;
    "removeFocusListener(Ljava/awt/event/FocusListener;)V"(thread: JVMThread, args: [java_awt_event_FocusListener], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_FocusListener extends java_lang_Object, java_util_EventListener {
    "focusGained(Ljava/awt/event/FocusEvent;)V"(thread: JVMThread, args: [java_awt_event_FocusEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "focusLost(Ljava/awt/event/FocusEvent;)V"(thread: JVMThread, args: [java_awt_event_FocusEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_FocusEvent extends java_awt_event_ComponentEvent {
    public static "java/awt/event/FocusEvent/FOCUS_FIRST": number;
    public static "java/awt/event/FocusEvent/FOCUS_LAST": number;
    public static "java/awt/event/FocusEvent/FOCUS_GAINED": number;
    public static "java/awt/event/FocusEvent/FOCUS_LOST": number;
    public "java/awt/event/FocusEvent/temporary": number;
    public "java/awt/event/FocusEvent/opposite": java_awt_Component;
    public static "java/awt/event/FocusEvent/serialVersionUID": Long;
    public "<init>(Ljava/awt/Component;IZLjava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component, number, number, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/FocusEvent/<init>(Ljava/awt/Component;IZLjava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component, number, number, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;IZ)V"(thread: JVMThread, args: [java_awt_Component, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/FocusEvent/<init>(Ljava/awt/Component;IZ)V"(thread: JVMThread, args: [java_awt_Component, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/FocusEvent/<init>(Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isTemporary()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/FocusEvent/isTemporary()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOppositeComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/event/FocusEvent/getOppositeComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/FocusEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_FontMetrics extends java_lang_Object implements java_io_Serializable {
    public static "java/awt/FontMetrics/DEFAULT_FRC": java_awt_font_FontRenderContext;
    public "java/awt/FontMetrics/font": java_awt_Font;
    public static "java/awt/FontMetrics/serialVersionUID": Long;
    public "<init>(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FontMetrics/<init>(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/FontMetrics/getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "java/awt/FontMetrics/getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "getLeading()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/getLeading()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAscent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/getAscent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/getDescent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaxAscent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/getMaxAscent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaxDescent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/getMaxDescent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaxDecent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/getMaxDecent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaxAdvance()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/getMaxAdvance()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charWidth(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/charWidth(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charWidth(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/charWidth(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "stringWidth(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/stringWidth(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charsWidth([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/charsWidth([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "bytesWidth([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/bytesWidth([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWidths()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/FontMetrics/getWidths()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "hasUniformLineMetrics()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FontMetrics/hasUniformLineMetrics()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLineMetrics(Ljava/lang/String;Ljava/awt/Graphics;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_lang_String, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "java/awt/FontMetrics/getLineMetrics(Ljava/lang/String;Ljava/awt/Graphics;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_lang_String, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "getLineMetrics(Ljava/lang/String;IILjava/awt/Graphics;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_lang_String, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "java/awt/FontMetrics/getLineMetrics(Ljava/lang/String;IILjava/awt/Graphics;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_lang_String, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "getLineMetrics([CIILjava/awt/Graphics;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "java/awt/FontMetrics/getLineMetrics([CIILjava/awt/Graphics;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "getLineMetrics(Ljava/text/CharacterIterator;IILjava/awt/Graphics;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_text_CharacterIterator, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "java/awt/FontMetrics/getLineMetrics(Ljava/text/CharacterIterator;IILjava/awt/Graphics;)Ljava/awt/font/LineMetrics;"(thread: JVMThread, args: [java_text_CharacterIterator, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_LineMetrics) => void): void;
    public "getStringBounds(Ljava/lang/String;Ljava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_lang_String, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/FontMetrics/getStringBounds(Ljava/lang/String;Ljava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_lang_String, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getStringBounds(Ljava/lang/String;IILjava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_lang_String, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/FontMetrics/getStringBounds(Ljava/lang/String;IILjava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_lang_String, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getStringBounds([CIILjava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/FontMetrics/getStringBounds([CIILjava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getStringBounds(Ljava/text/CharacterIterator;IILjava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_text_CharacterIterator, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/FontMetrics/getStringBounds(Ljava/text/CharacterIterator;IILjava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_text_CharacterIterator, number, number, java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getMaxCharBounds(Ljava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/FontMetrics/getMaxCharBounds(Ljava/awt/Graphics;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "myFRC(Ljava/awt/Graphics;)Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "java/awt/FontMetrics/myFRC(Ljava/awt/Graphics;)Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/FontMetrics/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/FontMetrics/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/FontMetrics/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface javax_accessibility_AccessibleAction extends java_lang_Object {
    "getAccessibleActionCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAccessibleActionDescription(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "doAccessibleAction(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class javax_accessibility_AccessibleStateSet extends java_lang_Object {
    public "javax/accessibility/AccessibleStateSet/states": java_util_Vector;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleStateSet/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Ljavax/accessibility/AccessibleState;)V"(thread: JVMThread, args: [JVMArray<javax_accessibility_AccessibleState>], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleStateSet/<init>([Ljavax/accessibility/AccessibleState;)V"(thread: JVMThread, args: [JVMArray<javax_accessibility_AccessibleState>], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljavax/accessibility/AccessibleState;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleState], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/accessibility/AccessibleStateSet/add(Ljavax/accessibility/AccessibleState;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleState], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll([Ljavax/accessibility/AccessibleState;)V"(thread: JVMThread, args: [JVMArray<javax_accessibility_AccessibleState>], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleStateSet/addAll([Ljavax/accessibility/AccessibleState;)V"(thread: JVMThread, args: [JVMArray<javax_accessibility_AccessibleState>], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljavax/accessibility/AccessibleState;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleState], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/accessibility/AccessibleStateSet/remove(Ljavax/accessibility/AccessibleState;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleState], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleStateSet/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "contains(Ljavax/accessibility/AccessibleState;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleState], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/accessibility/AccessibleStateSet/contains(Ljavax/accessibility/AccessibleState;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleState], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toArray()[Ljavax/accessibility/AccessibleState;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_accessibility_AccessibleState>) => void): void;
    public "javax/accessibility/AccessibleStateSet/toArray()[Ljavax/accessibility/AccessibleState;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_accessibility_AccessibleState>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/accessibility/AccessibleStateSet/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class javax_accessibility_AccessibleState extends javax_accessibility_AccessibleBundle {
    public static "javax/accessibility/AccessibleState/ACTIVE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/PRESSED": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/ARMED": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/BUSY": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/CHECKED": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/EDITABLE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/EXPANDABLE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/COLLAPSED": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/EXPANDED": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/ENABLED": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/FOCUSABLE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/FOCUSED": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/ICONIFIED": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/MODAL": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/OPAQUE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/RESIZABLE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/MULTISELECTABLE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/SELECTABLE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/SELECTED": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/SHOWING": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/VISIBLE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/VERTICAL": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/HORIZONTAL": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/SINGLE_LINE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/MULTI_LINE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/TRANSIENT": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/MANAGES_DESCENDANTS": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/INDETERMINATE": javax_accessibility_AccessibleState;
    public static "javax/accessibility/AccessibleState/TRUNCATED": javax_accessibility_AccessibleState;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleState/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/accessibility/AccessibleState/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_accessibility_AccessibleBundle extends java_lang_Object {
    public static "javax/accessibility/AccessibleBundle/table": java_util_Hashtable;
    public "javax/accessibility/AccessibleBundle/defaultResourceBundleName": java_lang_String;
    public "javax/accessibility/AccessibleBundle/key": java_lang_String;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleBundle/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toDisplayString(Ljava/lang/String;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/accessibility/AccessibleBundle/toDisplayString(Ljava/lang/String;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toDisplayString(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/accessibility/AccessibleBundle/toDisplayString(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toDisplayString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/accessibility/AccessibleBundle/toDisplayString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/accessibility/AccessibleBundle/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "loadResourceBundle(Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleBundle/loadResourceBundle(Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/accessibility/AccessibleBundle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_accessibility_AccessibleRole extends javax_accessibility_AccessibleBundle {
    public static "javax/accessibility/AccessibleRole/ALERT": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/COLUMN_HEADER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/CANVAS": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/COMBO_BOX": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/DESKTOP_ICON": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/HTML_CONTAINER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/INTERNAL_FRAME": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/DESKTOP_PANE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/OPTION_PANE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/WINDOW": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/FRAME": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/DIALOG": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/COLOR_CHOOSER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/DIRECTORY_PANE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/FILE_CHOOSER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/FILLER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/HYPERLINK": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/ICON": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/LABEL": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/ROOT_PANE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/GLASS_PANE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/LAYERED_PANE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/LIST": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/LIST_ITEM": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/MENU_BAR": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/POPUP_MENU": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/MENU": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/MENU_ITEM": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/SEPARATOR": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/PAGE_TAB_LIST": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/PAGE_TAB": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/PANEL": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/PROGRESS_BAR": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/PASSWORD_TEXT": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/PUSH_BUTTON": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/TOGGLE_BUTTON": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/CHECK_BOX": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/RADIO_BUTTON": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/ROW_HEADER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/SCROLL_PANE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/SCROLL_BAR": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/VIEWPORT": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/SLIDER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/SPLIT_PANE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/TABLE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/TEXT": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/TREE": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/TOOL_BAR": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/TOOL_TIP": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/AWT_COMPONENT": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/SWING_COMPONENT": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/UNKNOWN": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/STATUS_BAR": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/DATE_EDITOR": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/SPIN_BOX": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/FONT_CHOOSER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/GROUP_BOX": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/HEADER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/FOOTER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/PARAGRAPH": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/RULER": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/EDITBAR": javax_accessibility_AccessibleRole;
    public static "javax/accessibility/AccessibleRole/PROGRESS_MONITOR": javax_accessibility_AccessibleRole;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRole/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/accessibility/AccessibleRole/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_accessibility_AccessibleRelationSet extends java_lang_Object {
    public "javax/accessibility/AccessibleRelationSet/relations": java_util_Vector;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRelationSet/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Ljavax/accessibility/AccessibleRelation;)V"(thread: JVMThread, args: [JVMArray<javax_accessibility_AccessibleRelation>], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRelationSet/<init>([Ljavax/accessibility/AccessibleRelation;)V"(thread: JVMThread, args: [JVMArray<javax_accessibility_AccessibleRelation>], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljavax/accessibility/AccessibleRelation;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleRelation], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/accessibility/AccessibleRelationSet/add(Ljavax/accessibility/AccessibleRelation;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleRelation], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll([Ljavax/accessibility/AccessibleRelation;)V"(thread: JVMThread, args: [JVMArray<javax_accessibility_AccessibleRelation>], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRelationSet/addAll([Ljavax/accessibility/AccessibleRelation;)V"(thread: JVMThread, args: [JVMArray<javax_accessibility_AccessibleRelation>], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljavax/accessibility/AccessibleRelation;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleRelation], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/accessibility/AccessibleRelationSet/remove(Ljavax/accessibility/AccessibleRelation;)Z"(thread: JVMThread, args: [javax_accessibility_AccessibleRelation], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRelationSet/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/accessibility/AccessibleRelationSet/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/accessibility/AccessibleRelationSet/contains(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;)Ljavax/accessibility/AccessibleRelation;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleRelation) => void): void;
    public "javax/accessibility/AccessibleRelationSet/get(Ljava/lang/String;)Ljavax/accessibility/AccessibleRelation;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleRelation) => void): void;
    public "toArray()[Ljavax/accessibility/AccessibleRelation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_accessibility_AccessibleRelation>) => void): void;
    public "javax/accessibility/AccessibleRelationSet/toArray()[Ljavax/accessibility/AccessibleRelation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<javax_accessibility_AccessibleRelation>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/accessibility/AccessibleRelationSet/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class javax_accessibility_AccessibleRelation extends javax_accessibility_AccessibleBundle {
    public "javax/accessibility/AccessibleRelation/target": JVMArray<java_lang_Object>;
    public static "javax/accessibility/AccessibleRelation/LABEL_FOR": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/LABELED_BY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/MEMBER_OF": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/CONTROLLER_FOR": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/CONTROLLED_BY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/FLOWS_TO": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/FLOWS_FROM": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/SUBWINDOW_OF": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/PARENT_WINDOW_OF": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/EMBEDS": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/EMBEDDED_BY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/CHILD_NODE_OF": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/LABEL_FOR_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/LABELED_BY_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/MEMBER_OF_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/CONTROLLER_FOR_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/CONTROLLED_BY_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/FLOWS_TO_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/FLOWS_FROM_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/SUBWINDOW_OF_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/PARENT_WINDOW_OF_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/EMBEDS_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/EMBEDDED_BY_PROPERTY": java_lang_String;
    public static "javax/accessibility/AccessibleRelation/CHILD_NODE_OF_PROPERTY": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRelation/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRelation/<init>(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRelation/<init>(Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKey()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/accessibility/AccessibleRelation/getKey()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTarget()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "javax/accessibility/AccessibleRelation/getTarget()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "setTarget(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRelation/setTarget(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTarget([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/accessibility/AccessibleRelation/setTarget([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/accessibility/AccessibleRelation/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_ResourceBundle extends java_lang_Object {
    public static "java/util/ResourceBundle/INITIAL_CACHE_SIZE": number;
    public static "java/util/ResourceBundle/NONEXISTENT_BUNDLE": java_util_ResourceBundle;
    public static "java/util/ResourceBundle/cacheList": java_util_concurrent_ConcurrentMap;
    public static "java/util/ResourceBundle/referenceQueue": java_lang_ref_ReferenceQueue;
    public "java/util/ResourceBundle/parent": java_util_ResourceBundle;
    public "java/util/ResourceBundle/locale": java_util_Locale;
    public "java/util/ResourceBundle/name": java_lang_String;
    public "java/util/ResourceBundle/expired": number;
    public "java/util/ResourceBundle/cacheKey": java_util_ResourceBundle$CacheKey;
    public "java/util/ResourceBundle/keySet": java_util_Set;
    public static "java/util/ResourceBundle/providers": java_util_List;
    public static "java/util/ResourceBundle/$assertionsDisabled": number;
    public "getBaseBundleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ResourceBundle/getBaseBundleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getString(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ResourceBundle/getString(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getStringArray(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/ResourceBundle/getStringArray(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getObject(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ResourceBundle/getObject(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/util/ResourceBundle/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "getLoader(Ljava/lang/Class;)Ljava/lang/ClassLoader;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "java/util/ResourceBundle/getLoader(Ljava/lang/Class;)Ljava/lang/ClassLoader;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "setParent(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle/setParent(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getBundle(Ljava/lang/String;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/getBundle(Ljava/lang/String;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "getBundle(Ljava/lang/String;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/getBundle(Ljava/lang/String;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "getBundle(Ljava/lang/String;Ljava/util/Locale;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/getBundle(Ljava/lang/String;Ljava/util/Locale;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "getBundle(Ljava/lang/String;Ljava/util/Locale;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/getBundle(Ljava/lang/String;Ljava/util/Locale;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "getBundle(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/ClassLoader;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/getBundle(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/ClassLoader;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "getBundle(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/ClassLoader;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_ClassLoader, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/getBundle(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/ClassLoader;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_ClassLoader, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "getDefaultControl(Ljava/lang/String;)Ljava/util/ResourceBundle$Control;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$Control) => void): void;
    public static "java/util/ResourceBundle/getDefaultControl(Ljava/lang/String;)Ljava/util/ResourceBundle$Control;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$Control) => void): void;
    public static "getBundleImpl(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/ClassLoader;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_ClassLoader, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/getBundleImpl(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/ClassLoader;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_ClassLoader, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "checkList(Ljava/util/List;)Z"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/ResourceBundle/checkList(Ljava/util/List;)Z"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "findBundle(Ljava/util/ResourceBundle$CacheKey;Ljava/util/List;Ljava/util/List;ILjava/util/ResourceBundle$Control;Ljava/util/ResourceBundle;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_List, java_util_List, number, java_util_ResourceBundle$Control, java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/findBundle(Ljava/util/ResourceBundle$CacheKey;Ljava/util/List;Ljava/util/List;ILjava/util/ResourceBundle$Control;Ljava/util/ResourceBundle;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_List, java_util_List, number, java_util_ResourceBundle$Control, java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "loadBundle(Ljava/util/ResourceBundle$CacheKey;Ljava/util/List;Ljava/util/ResourceBundle$Control;Z)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_List, java_util_ResourceBundle$Control, number], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/loadBundle(Ljava/util/ResourceBundle$CacheKey;Ljava/util/List;Ljava/util/ResourceBundle$Control;Z)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_List, java_util_ResourceBundle$Control, number], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "isValidBundle(Ljava/util/ResourceBundle;)Z"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/ResourceBundle/isValidBundle(Ljava/util/ResourceBundle;)Z"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hasValidParentChain(Ljava/util/ResourceBundle;)Z"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/ResourceBundle/hasValidParentChain(Ljava/util/ResourceBundle;)Z"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "throwMissingResourceException(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/ResourceBundle/throwMissingResourceException(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "findBundleInCache(Ljava/util/ResourceBundle$CacheKey;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/findBundleInCache(Ljava/util/ResourceBundle$CacheKey;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "putBundleInCache(Ljava/util/ResourceBundle$CacheKey;Ljava/util/ResourceBundle;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_ResourceBundle, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/ResourceBundle/putBundleInCache(Ljava/util/ResourceBundle$CacheKey;Ljava/util/ResourceBundle;Ljava/util/ResourceBundle$Control;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_ResourceBundle, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "setExpirationTime(Ljava/util/ResourceBundle$CacheKey;Ljava/util/ResourceBundle$Control;)V"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/ResourceBundle/setExpirationTime(Ljava/util/ResourceBundle$CacheKey;Ljava/util/ResourceBundle$Control;)V"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_util_ResourceBundle$Control], cb?: (e?: java_lang_Throwable) => void): void;
    public static "clearCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/ResourceBundle/clearCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "clearCache(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/ResourceBundle/clearCache(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "handleGetObject(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ResourceBundle/handleGetObject(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getKeys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/ResourceBundle/getKeys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "containsKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ResourceBundle/containsKey(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/ResourceBundle/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "handleKeySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/ResourceBundle/handleKeySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "access$200()Ljava/lang/ref/ReferenceQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_ReferenceQueue) => void): void;
    public static "java/util/ResourceBundle/access$200()Ljava/lang/ref/ReferenceQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_ReferenceQueue) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/ResourceBundle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_ResourceBundle$Control extends java_lang_Object {
    public static "java/util/ResourceBundle$Control/FORMAT_DEFAULT": java_util_List;
    public static "java/util/ResourceBundle$Control/FORMAT_CLASS": java_util_List;
    public static "java/util/ResourceBundle$Control/FORMAT_PROPERTIES": java_util_List;
    public static "java/util/ResourceBundle$Control/TTL_DONT_CACHE": Long;
    public static "java/util/ResourceBundle$Control/TTL_NO_EXPIRATION_CONTROL": Long;
    public static "java/util/ResourceBundle$Control/INSTANCE": java_util_ResourceBundle$Control;
    public static "java/util/ResourceBundle$Control/CANDIDATES_CACHE": java_util_ResourceBundle$Control$CandidateListCache;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$Control/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getControl(Ljava/util/List;)Ljava/util/ResourceBundle$Control;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$Control) => void): void;
    public static "java/util/ResourceBundle$Control/getControl(Ljava/util/List;)Ljava/util/ResourceBundle$Control;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$Control) => void): void;
    public static "getNoFallbackControl(Ljava/util/List;)Ljava/util/ResourceBundle$Control;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$Control) => void): void;
    public static "java/util/ResourceBundle$Control/getNoFallbackControl(Ljava/util/List;)Ljava/util/ResourceBundle$Control;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$Control) => void): void;
    public "getFormats(Ljava/lang/String;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/ResourceBundle$Control/getFormats(Ljava/lang/String;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getCandidateLocales(Ljava/lang/String;Ljava/util/Locale;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/ResourceBundle$Control/getCandidateLocales(Ljava/lang/String;Ljava/util/Locale;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getFallbackLocale(Ljava/lang/String;Ljava/util/Locale;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/util/ResourceBundle$Control/getFallbackLocale(Ljava/lang/String;Ljava/util/Locale;)Ljava/util/Locale;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "newBundle(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/String;Ljava/lang/ClassLoader;Z)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_String, java_lang_ClassLoader, number], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "java/util/ResourceBundle$Control/newBundle(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/String;Ljava/lang/ClassLoader;Z)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_String, java_lang_ClassLoader, number], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "getTimeToLive(Ljava/lang/String;Ljava/util/Locale;)J"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/ResourceBundle$Control/getTimeToLive(Ljava/lang/String;Ljava/util/Locale;)J"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "needsReload(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/String;Ljava/lang/ClassLoader;Ljava/util/ResourceBundle;J)Z"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_String, java_lang_ClassLoader, java_util_ResourceBundle, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ResourceBundle$Control/needsReload(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/String;Ljava/lang/ClassLoader;Ljava/util/ResourceBundle;J)Z"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_String, java_lang_ClassLoader, java_util_ResourceBundle, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toBundleName(Ljava/lang/String;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ResourceBundle$Control/toBundleName(Ljava/lang/String;Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toResourceName(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ResourceBundle$Control/toResourceName(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toResourceName0(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ResourceBundle$Control/toResourceName0(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$300()Ljava/util/ResourceBundle$Control;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$Control) => void): void;
    public static "java/util/ResourceBundle$Control/access$300()Ljava/util/ResourceBundle$Control;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$Control) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/ResourceBundle$Control/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_ResourceBundle$Control$CandidateListCache extends sun_util_locale_LocaleObjectCache {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$Control$CandidateListCache/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "createObject(Lsun/util/locale/BaseLocale;)Ljava/util/List;"(thread: JVMThread, args: [sun_util_locale_BaseLocale], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/ResourceBundle$Control$CandidateListCache/createObject(Lsun/util/locale/BaseLocale;)Ljava/util/List;"(thread: JVMThread, args: [sun_util_locale_BaseLocale], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "getDefaultList(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/util/ResourceBundle$Control$CandidateListCache/getDefaultList(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "createObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ResourceBundle$Control$CandidateListCache/createObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "<init>(Ljava/util/ResourceBundle$1;)V"(thread: JVMThread, args: [java_util_ResourceBundle$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$Control$CandidateListCache/<init>(Ljava/util/ResourceBundle$1;)V"(thread: JVMThread, args: [java_util_ResourceBundle$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_ResourceBundle$1 extends java_util_ResourceBundle {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$1/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKeys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/ResourceBundle$1/getKeys()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "handleGetObject(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ResourceBundle$1/handleGetObject(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ResourceBundle$1/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_util_ResourceBundle$CacheKey extends java_lang_Object implements java_lang_Cloneable {
    public "java/util/ResourceBundle$CacheKey/name": java_lang_String;
    public "java/util/ResourceBundle$CacheKey/locale": java_util_Locale;
    public "java/util/ResourceBundle$CacheKey/loaderRef": java_util_ResourceBundle$LoaderReference;
    public "java/util/ResourceBundle$CacheKey/format": java_lang_String;
    public "java/util/ResourceBundle$CacheKey/loadTime": Long;
    public "java/util/ResourceBundle$CacheKey/expirationTime": Long;
    public "java/util/ResourceBundle$CacheKey/cause": java_lang_Throwable;
    public "java/util/ResourceBundle$CacheKey/hashCodeCache": number;
    public "<init>(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$CacheKey/<init>(Ljava/lang/String;Ljava/util/Locale;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_util_Locale, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ResourceBundle$CacheKey/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setName(Ljava/lang/String;)Ljava/util/ResourceBundle$CacheKey;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$CacheKey) => void): void;
    public "java/util/ResourceBundle$CacheKey/setName(Ljava/lang/String;)Ljava/util/ResourceBundle$CacheKey;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$CacheKey) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/util/ResourceBundle$CacheKey/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "setLocale(Ljava/util/Locale;)Ljava/util/ResourceBundle$CacheKey;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$CacheKey) => void): void;
    public "java/util/ResourceBundle$CacheKey/setLocale(Ljava/util/Locale;)Ljava/util/ResourceBundle$CacheKey;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$CacheKey) => void): void;
    public "getLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/util/ResourceBundle$CacheKey/getLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ResourceBundle$CacheKey/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/ResourceBundle$CacheKey/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "calculateHashCode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$CacheKey/calculateHashCode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/ResourceBundle$CacheKey/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getFormat()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ResourceBundle$CacheKey/getFormat()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setFormat(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$CacheKey/setFormat(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCause(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$CacheKey/setCause(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$CacheKey/getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/ResourceBundle$CacheKey/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$400(Ljava/util/ResourceBundle$CacheKey;)Ljava/lang/Throwable;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public static "java/util/ResourceBundle$CacheKey/access$400(Ljava/util/ResourceBundle$CacheKey;)Ljava/lang/Throwable;"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public static "access$500(Ljava/util/ResourceBundle$CacheKey;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/ResourceBundle$CacheKey/access$500(Ljava/util/ResourceBundle$CacheKey;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$600(Ljava/util/ResourceBundle$CacheKey;)J"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/ResourceBundle$CacheKey/access$600(Ljava/util/ResourceBundle$CacheKey;)J"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$700(Ljava/util/ResourceBundle$CacheKey;)J"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/ResourceBundle$CacheKey/access$700(Ljava/util/ResourceBundle$CacheKey;)J"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$702(Ljava/util/ResourceBundle$CacheKey;J)J"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/ResourceBundle$CacheKey/access$702(Ljava/util/ResourceBundle$CacheKey;J)J"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$602(Ljava/util/ResourceBundle$CacheKey;J)J"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/ResourceBundle$CacheKey/access$602(Ljava/util/ResourceBundle$CacheKey;J)J"(thread: JVMThread, args: [java_util_ResourceBundle$CacheKey, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class java_util_ResourceBundle$LoaderReference extends java_lang_ref_WeakReference implements java_util_ResourceBundle$CacheKeyReference {
    public "java/util/ResourceBundle$LoaderReference/cacheKey": java_util_ResourceBundle$CacheKey;
    public "<init>(Ljava/lang/ClassLoader;Ljava/lang/ref/ReferenceQueue;Ljava/util/ResourceBundle$CacheKey;)V"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_ref_ReferenceQueue, java_util_ResourceBundle$CacheKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/ResourceBundle$LoaderReference/<init>(Ljava/lang/ClassLoader;Ljava/lang/ref/ReferenceQueue;Ljava/util/ResourceBundle$CacheKey;)V"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_ref_ReferenceQueue, java_util_ResourceBundle$CacheKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCacheKey()Ljava/util/ResourceBundle$CacheKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$CacheKey) => void): void;
    public "java/util/ResourceBundle$LoaderReference/getCacheKey()Ljava/util/ResourceBundle$CacheKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$CacheKey) => void): void;
  }
  export interface java_util_ResourceBundle$CacheKeyReference extends java_lang_Object {
    "getCacheKey()Ljava/util/ResourceBundle$CacheKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle$CacheKey) => void): void;
  }
  export class java_awt_Event extends java_lang_Object implements java_io_Serializable {
    public "java/awt/Event/data": Long;
    public static "java/awt/Event/SHIFT_MASK": number;
    public static "java/awt/Event/CTRL_MASK": number;
    public static "java/awt/Event/META_MASK": number;
    public static "java/awt/Event/ALT_MASK": number;
    public static "java/awt/Event/HOME": number;
    public static "java/awt/Event/END": number;
    public static "java/awt/Event/PGUP": number;
    public static "java/awt/Event/PGDN": number;
    public static "java/awt/Event/UP": number;
    public static "java/awt/Event/DOWN": number;
    public static "java/awt/Event/LEFT": number;
    public static "java/awt/Event/RIGHT": number;
    public static "java/awt/Event/F1": number;
    public static "java/awt/Event/F2": number;
    public static "java/awt/Event/F3": number;
    public static "java/awt/Event/F4": number;
    public static "java/awt/Event/F5": number;
    public static "java/awt/Event/F6": number;
    public static "java/awt/Event/F7": number;
    public static "java/awt/Event/F8": number;
    public static "java/awt/Event/F9": number;
    public static "java/awt/Event/F10": number;
    public static "java/awt/Event/F11": number;
    public static "java/awt/Event/F12": number;
    public static "java/awt/Event/PRINT_SCREEN": number;
    public static "java/awt/Event/SCROLL_LOCK": number;
    public static "java/awt/Event/CAPS_LOCK": number;
    public static "java/awt/Event/NUM_LOCK": number;
    public static "java/awt/Event/PAUSE": number;
    public static "java/awt/Event/INSERT": number;
    public static "java/awt/Event/ENTER": number;
    public static "java/awt/Event/BACK_SPACE": number;
    public static "java/awt/Event/TAB": number;
    public static "java/awt/Event/ESCAPE": number;
    public static "java/awt/Event/DELETE": number;
    public static "java/awt/Event/WINDOW_EVENT": number;
    public static "java/awt/Event/WINDOW_DESTROY": number;
    public static "java/awt/Event/WINDOW_EXPOSE": number;
    public static "java/awt/Event/WINDOW_ICONIFY": number;
    public static "java/awt/Event/WINDOW_DEICONIFY": number;
    public static "java/awt/Event/WINDOW_MOVED": number;
    public static "java/awt/Event/KEY_EVENT": number;
    public static "java/awt/Event/KEY_PRESS": number;
    public static "java/awt/Event/KEY_RELEASE": number;
    public static "java/awt/Event/KEY_ACTION": number;
    public static "java/awt/Event/KEY_ACTION_RELEASE": number;
    public static "java/awt/Event/MOUSE_EVENT": number;
    public static "java/awt/Event/MOUSE_DOWN": number;
    public static "java/awt/Event/MOUSE_UP": number;
    public static "java/awt/Event/MOUSE_MOVE": number;
    public static "java/awt/Event/MOUSE_ENTER": number;
    public static "java/awt/Event/MOUSE_EXIT": number;
    public static "java/awt/Event/MOUSE_DRAG": number;
    public static "java/awt/Event/SCROLL_EVENT": number;
    public static "java/awt/Event/SCROLL_LINE_UP": number;
    public static "java/awt/Event/SCROLL_LINE_DOWN": number;
    public static "java/awt/Event/SCROLL_PAGE_UP": number;
    public static "java/awt/Event/SCROLL_PAGE_DOWN": number;
    public static "java/awt/Event/SCROLL_ABSOLUTE": number;
    public static "java/awt/Event/SCROLL_BEGIN": number;
    public static "java/awt/Event/SCROLL_END": number;
    public static "java/awt/Event/LIST_EVENT": number;
    public static "java/awt/Event/LIST_SELECT": number;
    public static "java/awt/Event/LIST_DESELECT": number;
    public static "java/awt/Event/MISC_EVENT": number;
    public static "java/awt/Event/ACTION_EVENT": number;
    public static "java/awt/Event/LOAD_FILE": number;
    public static "java/awt/Event/SAVE_FILE": number;
    public static "java/awt/Event/GOT_FOCUS": number;
    public static "java/awt/Event/LOST_FOCUS": number;
    public "java/awt/Event/target": java_lang_Object;
    public "java/awt/Event/when": Long;
    public "java/awt/Event/id": number;
    public "java/awt/Event/x": number;
    public "java/awt/Event/y": number;
    public "java/awt/Event/key": number;
    public "java/awt/Event/modifiers": number;
    public "java/awt/Event/clickCount": number;
    public "java/awt/Event/arg": java_lang_Object;
    public "java/awt/Event/evt": java_awt_Event;
    public static "java/awt/Event/actionKeyCodes": JVMArray<JVMArray<number>>;
    public "java/awt/Event/consumed": number;
    public static "java/awt/Event/serialVersionUID": Long;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Event/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;JIIIIILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number, number, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Event/<init>(Ljava/lang/Object;JIIIIILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number, number, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;JIIIII)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Event/<init>(Ljava/lang/Object;JIIIII)V"(thread: JVMThread, args: [java_lang_Object, Long, any, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;ILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Event/<init>(Ljava/lang/Object;ILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Event/translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "shiftDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Event/shiftDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "controlDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Event/controlDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "metaDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Event/metaDown()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Event/consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isConsumed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Event/isConsumed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getOldEventKey(Ljava/awt/event/KeyEvent;)I"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Event/getOldEventKey(Ljava/awt/event/KeyEvent;)I"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getKeyEventChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Event/getKeyEventChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Event/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Event/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Event/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_KeyEvent extends java_awt_event_InputEvent {
    public "java/awt/event/KeyEvent/isProxyActive": number;
    public static "java/awt/event/KeyEvent/KEY_FIRST": number;
    public static "java/awt/event/KeyEvent/KEY_LAST": number;
    public static "java/awt/event/KeyEvent/KEY_TYPED": number;
    public static "java/awt/event/KeyEvent/KEY_PRESSED": number;
    public static "java/awt/event/KeyEvent/KEY_RELEASED": number;
    public static "java/awt/event/KeyEvent/VK_ENTER": number;
    public static "java/awt/event/KeyEvent/VK_BACK_SPACE": number;
    public static "java/awt/event/KeyEvent/VK_TAB": number;
    public static "java/awt/event/KeyEvent/VK_CANCEL": number;
    public static "java/awt/event/KeyEvent/VK_CLEAR": number;
    public static "java/awt/event/KeyEvent/VK_SHIFT": number;
    public static "java/awt/event/KeyEvent/VK_CONTROL": number;
    public static "java/awt/event/KeyEvent/VK_ALT": number;
    public static "java/awt/event/KeyEvent/VK_PAUSE": number;
    public static "java/awt/event/KeyEvent/VK_CAPS_LOCK": number;
    public static "java/awt/event/KeyEvent/VK_ESCAPE": number;
    public static "java/awt/event/KeyEvent/VK_SPACE": number;
    public static "java/awt/event/KeyEvent/VK_PAGE_UP": number;
    public static "java/awt/event/KeyEvent/VK_PAGE_DOWN": number;
    public static "java/awt/event/KeyEvent/VK_END": number;
    public static "java/awt/event/KeyEvent/VK_HOME": number;
    public static "java/awt/event/KeyEvent/VK_LEFT": number;
    public static "java/awt/event/KeyEvent/VK_UP": number;
    public static "java/awt/event/KeyEvent/VK_RIGHT": number;
    public static "java/awt/event/KeyEvent/VK_DOWN": number;
    public static "java/awt/event/KeyEvent/VK_COMMA": number;
    public static "java/awt/event/KeyEvent/VK_MINUS": number;
    public static "java/awt/event/KeyEvent/VK_PERIOD": number;
    public static "java/awt/event/KeyEvent/VK_SLASH": number;
    public static "java/awt/event/KeyEvent/VK_0": number;
    public static "java/awt/event/KeyEvent/VK_1": number;
    public static "java/awt/event/KeyEvent/VK_2": number;
    public static "java/awt/event/KeyEvent/VK_3": number;
    public static "java/awt/event/KeyEvent/VK_4": number;
    public static "java/awt/event/KeyEvent/VK_5": number;
    public static "java/awt/event/KeyEvent/VK_6": number;
    public static "java/awt/event/KeyEvent/VK_7": number;
    public static "java/awt/event/KeyEvent/VK_8": number;
    public static "java/awt/event/KeyEvent/VK_9": number;
    public static "java/awt/event/KeyEvent/VK_SEMICOLON": number;
    public static "java/awt/event/KeyEvent/VK_EQUALS": number;
    public static "java/awt/event/KeyEvent/VK_A": number;
    public static "java/awt/event/KeyEvent/VK_B": number;
    public static "java/awt/event/KeyEvent/VK_C": number;
    public static "java/awt/event/KeyEvent/VK_D": number;
    public static "java/awt/event/KeyEvent/VK_E": number;
    public static "java/awt/event/KeyEvent/VK_F": number;
    public static "java/awt/event/KeyEvent/VK_G": number;
    public static "java/awt/event/KeyEvent/VK_H": number;
    public static "java/awt/event/KeyEvent/VK_I": number;
    public static "java/awt/event/KeyEvent/VK_J": number;
    public static "java/awt/event/KeyEvent/VK_K": number;
    public static "java/awt/event/KeyEvent/VK_L": number;
    public static "java/awt/event/KeyEvent/VK_M": number;
    public static "java/awt/event/KeyEvent/VK_N": number;
    public static "java/awt/event/KeyEvent/VK_O": number;
    public static "java/awt/event/KeyEvent/VK_P": number;
    public static "java/awt/event/KeyEvent/VK_Q": number;
    public static "java/awt/event/KeyEvent/VK_R": number;
    public static "java/awt/event/KeyEvent/VK_S": number;
    public static "java/awt/event/KeyEvent/VK_T": number;
    public static "java/awt/event/KeyEvent/VK_U": number;
    public static "java/awt/event/KeyEvent/VK_V": number;
    public static "java/awt/event/KeyEvent/VK_W": number;
    public static "java/awt/event/KeyEvent/VK_X": number;
    public static "java/awt/event/KeyEvent/VK_Y": number;
    public static "java/awt/event/KeyEvent/VK_Z": number;
    public static "java/awt/event/KeyEvent/VK_OPEN_BRACKET": number;
    public static "java/awt/event/KeyEvent/VK_BACK_SLASH": number;
    public static "java/awt/event/KeyEvent/VK_CLOSE_BRACKET": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD0": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD1": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD2": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD3": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD4": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD5": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD6": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD7": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD8": number;
    public static "java/awt/event/KeyEvent/VK_NUMPAD9": number;
    public static "java/awt/event/KeyEvent/VK_MULTIPLY": number;
    public static "java/awt/event/KeyEvent/VK_ADD": number;
    public static "java/awt/event/KeyEvent/VK_SEPARATER": number;
    public static "java/awt/event/KeyEvent/VK_SEPARATOR": number;
    public static "java/awt/event/KeyEvent/VK_SUBTRACT": number;
    public static "java/awt/event/KeyEvent/VK_DECIMAL": number;
    public static "java/awt/event/KeyEvent/VK_DIVIDE": number;
    public static "java/awt/event/KeyEvent/VK_DELETE": number;
    public static "java/awt/event/KeyEvent/VK_NUM_LOCK": number;
    public static "java/awt/event/KeyEvent/VK_SCROLL_LOCK": number;
    public static "java/awt/event/KeyEvent/VK_F1": number;
    public static "java/awt/event/KeyEvent/VK_F2": number;
    public static "java/awt/event/KeyEvent/VK_F3": number;
    public static "java/awt/event/KeyEvent/VK_F4": number;
    public static "java/awt/event/KeyEvent/VK_F5": number;
    public static "java/awt/event/KeyEvent/VK_F6": number;
    public static "java/awt/event/KeyEvent/VK_F7": number;
    public static "java/awt/event/KeyEvent/VK_F8": number;
    public static "java/awt/event/KeyEvent/VK_F9": number;
    public static "java/awt/event/KeyEvent/VK_F10": number;
    public static "java/awt/event/KeyEvent/VK_F11": number;
    public static "java/awt/event/KeyEvent/VK_F12": number;
    public static "java/awt/event/KeyEvent/VK_F13": number;
    public static "java/awt/event/KeyEvent/VK_F14": number;
    public static "java/awt/event/KeyEvent/VK_F15": number;
    public static "java/awt/event/KeyEvent/VK_F16": number;
    public static "java/awt/event/KeyEvent/VK_F17": number;
    public static "java/awt/event/KeyEvent/VK_F18": number;
    public static "java/awt/event/KeyEvent/VK_F19": number;
    public static "java/awt/event/KeyEvent/VK_F20": number;
    public static "java/awt/event/KeyEvent/VK_F21": number;
    public static "java/awt/event/KeyEvent/VK_F22": number;
    public static "java/awt/event/KeyEvent/VK_F23": number;
    public static "java/awt/event/KeyEvent/VK_F24": number;
    public static "java/awt/event/KeyEvent/VK_PRINTSCREEN": number;
    public static "java/awt/event/KeyEvent/VK_INSERT": number;
    public static "java/awt/event/KeyEvent/VK_HELP": number;
    public static "java/awt/event/KeyEvent/VK_META": number;
    public static "java/awt/event/KeyEvent/VK_BACK_QUOTE": number;
    public static "java/awt/event/KeyEvent/VK_QUOTE": number;
    public static "java/awt/event/KeyEvent/VK_KP_UP": number;
    public static "java/awt/event/KeyEvent/VK_KP_DOWN": number;
    public static "java/awt/event/KeyEvent/VK_KP_LEFT": number;
    public static "java/awt/event/KeyEvent/VK_KP_RIGHT": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_GRAVE": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_ACUTE": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_CIRCUMFLEX": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_TILDE": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_MACRON": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_BREVE": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_ABOVEDOT": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_DIAERESIS": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_ABOVERING": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_DOUBLEACUTE": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_CARON": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_CEDILLA": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_OGONEK": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_IOTA": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_VOICED_SOUND": number;
    public static "java/awt/event/KeyEvent/VK_DEAD_SEMIVOICED_SOUND": number;
    public static "java/awt/event/KeyEvent/VK_AMPERSAND": number;
    public static "java/awt/event/KeyEvent/VK_ASTERISK": number;
    public static "java/awt/event/KeyEvent/VK_QUOTEDBL": number;
    public static "java/awt/event/KeyEvent/VK_LESS": number;
    public static "java/awt/event/KeyEvent/VK_GREATER": number;
    public static "java/awt/event/KeyEvent/VK_BRACELEFT": number;
    public static "java/awt/event/KeyEvent/VK_BRACERIGHT": number;
    public static "java/awt/event/KeyEvent/VK_AT": number;
    public static "java/awt/event/KeyEvent/VK_COLON": number;
    public static "java/awt/event/KeyEvent/VK_CIRCUMFLEX": number;
    public static "java/awt/event/KeyEvent/VK_DOLLAR": number;
    public static "java/awt/event/KeyEvent/VK_EURO_SIGN": number;
    public static "java/awt/event/KeyEvent/VK_EXCLAMATION_MARK": number;
    public static "java/awt/event/KeyEvent/VK_INVERTED_EXCLAMATION_MARK": number;
    public static "java/awt/event/KeyEvent/VK_LEFT_PARENTHESIS": number;
    public static "java/awt/event/KeyEvent/VK_NUMBER_SIGN": number;
    public static "java/awt/event/KeyEvent/VK_PLUS": number;
    public static "java/awt/event/KeyEvent/VK_RIGHT_PARENTHESIS": number;
    public static "java/awt/event/KeyEvent/VK_UNDERSCORE": number;
    public static "java/awt/event/KeyEvent/VK_WINDOWS": number;
    public static "java/awt/event/KeyEvent/VK_CONTEXT_MENU": number;
    public static "java/awt/event/KeyEvent/VK_FINAL": number;
    public static "java/awt/event/KeyEvent/VK_CONVERT": number;
    public static "java/awt/event/KeyEvent/VK_NONCONVERT": number;
    public static "java/awt/event/KeyEvent/VK_ACCEPT": number;
    public static "java/awt/event/KeyEvent/VK_MODECHANGE": number;
    public static "java/awt/event/KeyEvent/VK_KANA": number;
    public static "java/awt/event/KeyEvent/VK_KANJI": number;
    public static "java/awt/event/KeyEvent/VK_ALPHANUMERIC": number;
    public static "java/awt/event/KeyEvent/VK_KATAKANA": number;
    public static "java/awt/event/KeyEvent/VK_HIRAGANA": number;
    public static "java/awt/event/KeyEvent/VK_FULL_WIDTH": number;
    public static "java/awt/event/KeyEvent/VK_HALF_WIDTH": number;
    public static "java/awt/event/KeyEvent/VK_ROMAN_CHARACTERS": number;
    public static "java/awt/event/KeyEvent/VK_ALL_CANDIDATES": number;
    public static "java/awt/event/KeyEvent/VK_PREVIOUS_CANDIDATE": number;
    public static "java/awt/event/KeyEvent/VK_CODE_INPUT": number;
    public static "java/awt/event/KeyEvent/VK_JAPANESE_KATAKANA": number;
    public static "java/awt/event/KeyEvent/VK_JAPANESE_HIRAGANA": number;
    public static "java/awt/event/KeyEvent/VK_JAPANESE_ROMAN": number;
    public static "java/awt/event/KeyEvent/VK_KANA_LOCK": number;
    public static "java/awt/event/KeyEvent/VK_INPUT_METHOD_ON_OFF": number;
    public static "java/awt/event/KeyEvent/VK_CUT": number;
    public static "java/awt/event/KeyEvent/VK_COPY": number;
    public static "java/awt/event/KeyEvent/VK_PASTE": number;
    public static "java/awt/event/KeyEvent/VK_UNDO": number;
    public static "java/awt/event/KeyEvent/VK_AGAIN": number;
    public static "java/awt/event/KeyEvent/VK_FIND": number;
    public static "java/awt/event/KeyEvent/VK_PROPS": number;
    public static "java/awt/event/KeyEvent/VK_STOP": number;
    public static "java/awt/event/KeyEvent/VK_COMPOSE": number;
    public static "java/awt/event/KeyEvent/VK_ALT_GRAPH": number;
    public static "java/awt/event/KeyEvent/VK_BEGIN": number;
    public static "java/awt/event/KeyEvent/VK_UNDEFINED": number;
    public static "java/awt/event/KeyEvent/CHAR_UNDEFINED": number;
    public static "java/awt/event/KeyEvent/KEY_LOCATION_UNKNOWN": number;
    public static "java/awt/event/KeyEvent/KEY_LOCATION_STANDARD": number;
    public static "java/awt/event/KeyEvent/KEY_LOCATION_LEFT": number;
    public static "java/awt/event/KeyEvent/KEY_LOCATION_RIGHT": number;
    public static "java/awt/event/KeyEvent/KEY_LOCATION_NUMPAD": number;
    public "java/awt/event/KeyEvent/keyCode": number;
    public "java/awt/event/KeyEvent/keyChar": number;
    public "java/awt/event/KeyEvent/keyLocation": number;
    public "java/awt/event/KeyEvent/rawCode": Long;
    public "java/awt/event/KeyEvent/primaryLevelUnicode": Long;
    public "java/awt/event/KeyEvent/scancode": Long;
    public "java/awt/event/KeyEvent/extendedKeyCode": Long;
    public static "java/awt/event/KeyEvent/serialVersionUID": Long;
    public "java/awt/event/KeyEvent/originalSource": java_awt_Component;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/event/KeyEvent/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;IJIICIZ)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/<init>(Ljava/awt/Component;IJIICIZ)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;IJIICI)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/<init>(Ljava/awt/Component;IJIICI)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;IJIIC)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/<init>(Ljava/awt/Component;IJIIC)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;IJII)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/<init>(Ljava/awt/Component;IJII)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKeyCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/KeyEvent/getKeyCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setKeyCode(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/setKeyCode(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKeyChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/KeyEvent/getKeyChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setKeyChar(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/setKeyChar(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setModifiers(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/setModifiers(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKeyLocation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/KeyEvent/getKeyLocation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getKeyText(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/event/KeyEvent/getKeyText(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getKeyModifiersText(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/event/KeyEvent/getKeyModifiersText(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isActionKey()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/KeyEvent/isActionKey()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/KeyEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getExtendedKeyCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/KeyEvent/getExtendedKeyCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getExtendedKeyCodeForChar(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/event/KeyEvent/getExtendedKeyCodeForChar(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setNewModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/setNewModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOldModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/setOldModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/KeyEvent/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$002(Ljava/awt/event/KeyEvent;J)J"(thread: JVMThread, args: [java_awt_event_KeyEvent, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/awt/event/KeyEvent/access$002(Ljava/awt/event/KeyEvent;J)J"(thread: JVMThread, args: [java_awt_event_KeyEvent, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$102(Ljava/awt/event/KeyEvent;J)J"(thread: JVMThread, args: [java_awt_event_KeyEvent, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/awt/event/KeyEvent/access$102(Ljava/awt/event/KeyEvent;J)J"(thread: JVMThread, args: [java_awt_event_KeyEvent, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$202(Ljava/awt/event/KeyEvent;J)J"(thread: JVMThread, args: [java_awt_event_KeyEvent, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/awt/event/KeyEvent/access$202(Ljava/awt/event/KeyEvent;J)J"(thread: JVMThread, args: [java_awt_event_KeyEvent, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$300(Ljava/awt/event/KeyEvent;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public static "java/awt/event/KeyEvent/access$300(Ljava/awt/event/KeyEvent;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/event/KeyEvent/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_WindowEvent extends java_awt_event_ComponentEvent {
    public static "java/awt/event/WindowEvent/WINDOW_FIRST": number;
    public static "java/awt/event/WindowEvent/WINDOW_OPENED": number;
    public static "java/awt/event/WindowEvent/WINDOW_CLOSING": number;
    public static "java/awt/event/WindowEvent/WINDOW_CLOSED": number;
    public static "java/awt/event/WindowEvent/WINDOW_ICONIFIED": number;
    public static "java/awt/event/WindowEvent/WINDOW_DEICONIFIED": number;
    public static "java/awt/event/WindowEvent/WINDOW_ACTIVATED": number;
    public static "java/awt/event/WindowEvent/WINDOW_DEACTIVATED": number;
    public static "java/awt/event/WindowEvent/WINDOW_GAINED_FOCUS": number;
    public static "java/awt/event/WindowEvent/WINDOW_LOST_FOCUS": number;
    public static "java/awt/event/WindowEvent/WINDOW_STATE_CHANGED": number;
    public static "java/awt/event/WindowEvent/WINDOW_LAST": number;
    public "java/awt/event/WindowEvent/opposite": java_awt_Window;
    public "java/awt/event/WindowEvent/oldState": number;
    public "java/awt/event/WindowEvent/newState": number;
    public static "java/awt/event/WindowEvent/serialVersionUID": Long;
    public "<init>(Ljava/awt/Window;ILjava/awt/Window;II)V"(thread: JVMThread, args: [java_awt_Window, number, java_awt_Window, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/WindowEvent/<init>(Ljava/awt/Window;ILjava/awt/Window;II)V"(thread: JVMThread, args: [java_awt_Window, number, java_awt_Window, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;ILjava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window, number, java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/WindowEvent/<init>(Ljava/awt/Window;ILjava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window, number, java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;III)V"(thread: JVMThread, args: [java_awt_Window, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/WindowEvent/<init>(Ljava/awt/Window;III)V"(thread: JVMThread, args: [java_awt_Window, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;I)V"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/WindowEvent/<init>(Ljava/awt/Window;I)V"(thread: JVMThread, args: [java_awt_Window, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWindow()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "java/awt/event/WindowEvent/getWindow()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "getOppositeWindow()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "java/awt/event/WindowEvent/getOppositeWindow()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "getOldState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/WindowEvent/getOldState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNewState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/WindowEvent/getNewState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/WindowEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_AWTEvent extends java_util_EventObject {
    public static "java/awt/AWTEvent/log": sun_util_logging_PlatformLogger;
    public "java/awt/AWTEvent/bdata": JVMArray<number>;
    public "java/awt/AWTEvent/id": number;
    public "java/awt/AWTEvent/consumed": number;
    public "java/awt/AWTEvent/acc": java_security_AccessControlContext;
    public "java/awt/AWTEvent/focusManagerIsDispatching": number;
    public "java/awt/AWTEvent/isPosted": number;
    public "java/awt/AWTEvent/isSystemGenerated": number;
    public static "java/awt/AWTEvent/COMPONENT_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/CONTAINER_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/FOCUS_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/KEY_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/MOUSE_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/MOUSE_MOTION_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/WINDOW_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/ACTION_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/ADJUSTMENT_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/ITEM_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/TEXT_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/INPUT_METHOD_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/INPUT_METHODS_ENABLED_MASK": Long;
    public static "java/awt/AWTEvent/PAINT_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/INVOCATION_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/HIERARCHY_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/HIERARCHY_BOUNDS_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/MOUSE_WHEEL_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/WINDOW_STATE_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/WINDOW_FOCUS_EVENT_MASK": Long;
    public static "java/awt/AWTEvent/RESERVED_ID_MAX": number;
    public static "java/awt/AWTEvent/inputEvent_CanAccessSystemClipboard_Field": java_lang_reflect_Field;
    public static "java/awt/AWTEvent/serialVersionUID": Long;
    public "getAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "java/awt/AWTEvent/getAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "get_InputEvent_CanAccessSystemClipboard()Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public static "java/awt/AWTEvent/get_InputEvent_CanAccessSystemClipboard()Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/AWTEvent/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Event;)V"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AWTEvent/<init>(Ljava/awt/Event;)V"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AWTEvent/<init>(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSource(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AWTEvent/setSource(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "nativeSetSource(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AWTEvent/nativeSetSource(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    public "getID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/AWTEvent/getID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/AWTEvent/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/AWTEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AWTEvent/consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isConsumed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/AWTEvent/isConsumed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "convertToOld()Ljava/awt/Event;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Event) => void): void;
    public "java/awt/AWTEvent/convertToOld()Ljava/awt/Event;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Event) => void): void;
    public "copyPrivateDataInto(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AWTEvent/copyPrivateDataInto(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatched()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AWTEvent/dispatched()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$002(Ljava/awt/AWTEvent;Z)Z"(thread: JVMThread, args: [java_awt_AWTEvent, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/AWTEvent/access$002(Ljava/awt/AWTEvent;Z)Z"(thread: JVMThread, args: [java_awt_AWTEvent, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/AWTEvent/access$000(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Ljava/awt/AWTEvent;)[B"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/AWTEvent/access$100(Ljava/awt/AWTEvent;)[B"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$102(Ljava/awt/AWTEvent;[B)[B"(thread: JVMThread, args: [java_awt_AWTEvent, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/AWTEvent/access$102(Ljava/awt/AWTEvent;[B)[B"(thread: JVMThread, args: [java_awt_AWTEvent, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$200()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "java/awt/AWTEvent/access$200()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/AWTEvent/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_ComponentPeer extends java_lang_Object {
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_peer_ContainerPeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "getInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    "beginValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "beginLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Insets extends java_lang_Object implements java_lang_Cloneable, java_io_Serializable {
    public "java/awt/Insets/top": number;
    public "java/awt/Insets/left": number;
    public "java/awt/Insets/bottom": number;
    public "java/awt/Insets/right": number;
    public static "java/awt/Insets/serialVersionUID": Long;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Insets/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "set(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Insets/set(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Insets/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Insets/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Insets/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/Insets/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Insets/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Insets/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_BufferCapabilities$FlipContents extends java_awt_AttributeValue {
    public static "java/awt/BufferCapabilities$FlipContents/I_UNDEFINED": number;
    public static "java/awt/BufferCapabilities$FlipContents/I_BACKGROUND": number;
    public static "java/awt/BufferCapabilities$FlipContents/I_PRIOR": number;
    public static "java/awt/BufferCapabilities$FlipContents/I_COPIED": number;
    public static "java/awt/BufferCapabilities$FlipContents/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/BufferCapabilities$FlipContents/UNDEFINED": java_awt_BufferCapabilities$FlipContents;
    public static "java/awt/BufferCapabilities$FlipContents/BACKGROUND": java_awt_BufferCapabilities$FlipContents;
    public static "java/awt/BufferCapabilities$FlipContents/PRIOR": java_awt_BufferCapabilities$FlipContents;
    public static "java/awt/BufferCapabilities$FlipContents/COPIED": java_awt_BufferCapabilities$FlipContents;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/BufferCapabilities$FlipContents/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/BufferCapabilities$FlipContents/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BufferCapabilities$FlipContents/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/BufferCapabilities$FlipContents/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_AttributeValue extends java_lang_Object {
    public static "java/awt/AttributeValue/log": sun_util_logging_PlatformLogger;
    public "java/awt/AttributeValue/value": number;
    public "java/awt/AttributeValue/names": JVMArray<java_lang_String>;
    public "<init>(I[Ljava/lang/String;)V"(thread: JVMThread, args: [number, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AttributeValue/<init>(I[Ljava/lang/String;)V"(thread: JVMThread, args: [number, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/AttributeValue/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/AttributeValue/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/AttributeValue/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_CausedFocusEvent$Cause extends java_lang_Enum {
    public static "sun/awt/CausedFocusEvent$Cause/UNKNOWN": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/MOUSE_EVENT": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/TRAVERSAL": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/TRAVERSAL_UP": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/TRAVERSAL_DOWN": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/TRAVERSAL_FORWARD": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/TRAVERSAL_BACKWARD": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/MANUAL_REQUEST": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/AUTOMATIC_TRAVERSE": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/ROLLBACK": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/NATIVE_SYSTEM": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/ACTIVATION": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/CLEAR_GLOBAL_FOCUS_OWNER": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/RETARGETED": sun_awt_CausedFocusEvent$Cause;
    public static "sun/awt/CausedFocusEvent$Cause/$VALUES": JVMArray<sun_awt_CausedFocusEvent$Cause>;
    public static "values()[Lsun/awt/CausedFocusEvent$Cause;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_CausedFocusEvent$Cause>) => void): void;
    public static "sun/awt/CausedFocusEvent$Cause/values()[Lsun/awt/CausedFocusEvent$Cause;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_CausedFocusEvent$Cause>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lsun/awt/CausedFocusEvent$Cause;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_awt_CausedFocusEvent$Cause) => void): void;
    public static "sun/awt/CausedFocusEvent$Cause/valueOf(Ljava/lang/String;)Lsun/awt/CausedFocusEvent$Cause;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_awt_CausedFocusEvent$Cause) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/CausedFocusEvent$Cause/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/CausedFocusEvent$Cause/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_PaintEvent extends java_awt_event_ComponentEvent {
    public static "java/awt/event/PaintEvent/PAINT_FIRST": number;
    public static "java/awt/event/PaintEvent/PAINT_LAST": number;
    public static "java/awt/event/PaintEvent/PAINT": number;
    public static "java/awt/event/PaintEvent/UPDATE": number;
    public "java/awt/event/PaintEvent/updateRect": java_awt_Rectangle;
    public static "java/awt/event/PaintEvent/serialVersionUID": Long;
    public "<init>(Ljava/awt/Component;ILjava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/PaintEvent/<init>(Ljava/awt/Component;ILjava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUpdateRect()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/event/PaintEvent/getUpdateRect()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "setUpdateRect(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/PaintEvent/setUpdateRect(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/PaintEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_Cursor extends java_lang_Object implements java_io_Serializable {
    public static "java/awt/Cursor/DEFAULT_CURSOR": number;
    public static "java/awt/Cursor/CROSSHAIR_CURSOR": number;
    public static "java/awt/Cursor/TEXT_CURSOR": number;
    public static "java/awt/Cursor/WAIT_CURSOR": number;
    public static "java/awt/Cursor/SW_RESIZE_CURSOR": number;
    public static "java/awt/Cursor/SE_RESIZE_CURSOR": number;
    public static "java/awt/Cursor/NW_RESIZE_CURSOR": number;
    public static "java/awt/Cursor/NE_RESIZE_CURSOR": number;
    public static "java/awt/Cursor/N_RESIZE_CURSOR": number;
    public static "java/awt/Cursor/S_RESIZE_CURSOR": number;
    public static "java/awt/Cursor/W_RESIZE_CURSOR": number;
    public static "java/awt/Cursor/E_RESIZE_CURSOR": number;
    public static "java/awt/Cursor/HAND_CURSOR": number;
    public static "java/awt/Cursor/MOVE_CURSOR": number;
    public static "java/awt/Cursor/predefined": JVMArray<java_awt_Cursor>;
    public static "java/awt/Cursor/predefinedPrivate": JVMArray<java_awt_Cursor>;
    public static "java/awt/Cursor/cursorProperties": JVMArray<JVMArray<java_lang_String>>;
    public "java/awt/Cursor/type": number;
    public static "java/awt/Cursor/CUSTOM_CURSOR": number;
    public static "java/awt/Cursor/systemCustomCursors": java_util_Hashtable;
    public static "java/awt/Cursor/systemCustomCursorDirPrefix": java_lang_String;
    public static "java/awt/Cursor/systemCustomCursorPropertiesFile": java_lang_String;
    public static "java/awt/Cursor/systemCustomCursorProperties": java_util_Properties;
    public static "java/awt/Cursor/CursorDotPrefix": java_lang_String;
    public static "java/awt/Cursor/DotFileSuffix": java_lang_String;
    public static "java/awt/Cursor/DotHotspotSuffix": java_lang_String;
    public static "java/awt/Cursor/DotNameSuffix": java_lang_String;
    public static "java/awt/Cursor/serialVersionUID": Long;
    public static "java/awt/Cursor/log": sun_util_logging_PlatformLogger;
    public "java/awt/Cursor/pData": Long;
    public "java/awt/Cursor/anchor": java_lang_Object;
    public "java/awt/Cursor/disposer": java_awt_Cursor$CursorDisposer;
    public "java/awt/Cursor/name": java_lang_String;
    public static "initCursorDir()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/Cursor/initCursorDir()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Cursor/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPData(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Cursor/setPData(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getPredefinedCursor(I)Ljava/awt/Cursor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public static "java/awt/Cursor/getPredefinedCursor(I)Ljava/awt/Cursor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public static "getSystemCustomCursor(Ljava/lang/String;)Ljava/awt/Cursor;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public static "java/awt/Cursor/getSystemCustomCursor(Ljava/lang/String;)Ljava/awt/Cursor;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public static "getDefaultCursor()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public static "java/awt/Cursor/getDefaultCursor()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Cursor/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Cursor/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Cursor/getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Cursor/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Cursor/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "loadSystemCustomCursorProperties()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Cursor/loadSystemCustomCursorProperties()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "finalizeImpl(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Cursor/finalizeImpl(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/awt/Cursor;)J"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/awt/Cursor/access$000(Ljava/awt/Cursor;)J"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$002(Ljava/awt/Cursor;J)J"(thread: JVMThread, args: [java_awt_Cursor, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/awt/Cursor/access$002(Ljava/awt/Cursor;J)J"(thread: JVMThread, args: [java_awt_Cursor, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$100(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Cursor/access$100(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$200()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/Cursor/access$200()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$300()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/Cursor/access$300()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$400()Ljava/util/Properties;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "java/awt/Cursor/access$400()Ljava/util/Properties;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Cursor/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Cursor$CursorDisposer extends java_lang_Object implements sun_java2d_DisposerRecord {
    public "java/awt/Cursor$CursorDisposer/pData": Long;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Cursor$CursorDisposer/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Cursor$CursorDisposer/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_DisposerRecord extends java_lang_Object {
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Toolkit extends java_lang_Object {
    public static "java/awt/Toolkit/lightweightMarker": java_awt_peer_LightweightPeer;
    public static "java/awt/Toolkit/toolkit": java_awt_Toolkit;
    public static "java/awt/Toolkit/atNames": java_lang_String;
    public static "java/awt/Toolkit/resources": java_util_ResourceBundle;
    public static "java/awt/Toolkit/platformResources": java_util_ResourceBundle;
    public static "java/awt/Toolkit/loaded": number;
    public "java/awt/Toolkit/desktopProperties": java_util_Map;
    public "java/awt/Toolkit/desktopPropsSupport": java_beans_PropertyChangeSupport;
    public static "java/awt/Toolkit/LONG_BITS": number;
    public "java/awt/Toolkit/calls": JVMArray<number>;
    public static "java/awt/Toolkit/enabledOnToolkitMask": Long;
    public "java/awt/Toolkit/eventListener": java_awt_event_AWTEventListener;
    public "java/awt/Toolkit/listener2SelectiveListener": java_util_WeakHashMap;
    public static "java/awt/Toolkit/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "createDesktopPeer(Ljava/awt/Desktop;)Ljava/awt/peer/DesktopPeer;"(thread: JVMThread, args: [java_awt_Desktop], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_DesktopPeer) => void): void;
    public "java/awt/Toolkit/createDesktopPeer(Ljava/awt/Desktop;)Ljava/awt/peer/DesktopPeer;"(thread: JVMThread, args: [java_awt_Desktop], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_DesktopPeer) => void): void;
    public "createButton(Ljava/awt/Button;)Ljava/awt/peer/ButtonPeer;"(thread: JVMThread, args: [java_awt_Button], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ButtonPeer) => void): void;
    public "java/awt/Toolkit/createButton(Ljava/awt/Button;)Ljava/awt/peer/ButtonPeer;"(thread: JVMThread, args: [java_awt_Button], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ButtonPeer) => void): void;
    public "createTextField(Ljava/awt/TextField;)Ljava/awt/peer/TextFieldPeer;"(thread: JVMThread, args: [java_awt_TextField], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_TextFieldPeer) => void): void;
    public "java/awt/Toolkit/createTextField(Ljava/awt/TextField;)Ljava/awt/peer/TextFieldPeer;"(thread: JVMThread, args: [java_awt_TextField], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_TextFieldPeer) => void): void;
    public "createLabel(Ljava/awt/Label;)Ljava/awt/peer/LabelPeer;"(thread: JVMThread, args: [java_awt_Label], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_LabelPeer) => void): void;
    public "java/awt/Toolkit/createLabel(Ljava/awt/Label;)Ljava/awt/peer/LabelPeer;"(thread: JVMThread, args: [java_awt_Label], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_LabelPeer) => void): void;
    public "createList(Ljava/awt/List;)Ljava/awt/peer/ListPeer;"(thread: JVMThread, args: [java_awt_List], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ListPeer) => void): void;
    public "java/awt/Toolkit/createList(Ljava/awt/List;)Ljava/awt/peer/ListPeer;"(thread: JVMThread, args: [java_awt_List], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ListPeer) => void): void;
    public "createCheckbox(Ljava/awt/Checkbox;)Ljava/awt/peer/CheckboxPeer;"(thread: JVMThread, args: [java_awt_Checkbox], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_CheckboxPeer) => void): void;
    public "java/awt/Toolkit/createCheckbox(Ljava/awt/Checkbox;)Ljava/awt/peer/CheckboxPeer;"(thread: JVMThread, args: [java_awt_Checkbox], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_CheckboxPeer) => void): void;
    public "createScrollbar(Ljava/awt/Scrollbar;)Ljava/awt/peer/ScrollbarPeer;"(thread: JVMThread, args: [java_awt_Scrollbar], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ScrollbarPeer) => void): void;
    public "java/awt/Toolkit/createScrollbar(Ljava/awt/Scrollbar;)Ljava/awt/peer/ScrollbarPeer;"(thread: JVMThread, args: [java_awt_Scrollbar], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ScrollbarPeer) => void): void;
    public "createScrollPane(Ljava/awt/ScrollPane;)Ljava/awt/peer/ScrollPanePeer;"(thread: JVMThread, args: [java_awt_ScrollPane], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ScrollPanePeer) => void): void;
    public "java/awt/Toolkit/createScrollPane(Ljava/awt/ScrollPane;)Ljava/awt/peer/ScrollPanePeer;"(thread: JVMThread, args: [java_awt_ScrollPane], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ScrollPanePeer) => void): void;
    public "createTextArea(Ljava/awt/TextArea;)Ljava/awt/peer/TextAreaPeer;"(thread: JVMThread, args: [java_awt_TextArea], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_TextAreaPeer) => void): void;
    public "java/awt/Toolkit/createTextArea(Ljava/awt/TextArea;)Ljava/awt/peer/TextAreaPeer;"(thread: JVMThread, args: [java_awt_TextArea], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_TextAreaPeer) => void): void;
    public "createChoice(Ljava/awt/Choice;)Ljava/awt/peer/ChoicePeer;"(thread: JVMThread, args: [java_awt_Choice], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ChoicePeer) => void): void;
    public "java/awt/Toolkit/createChoice(Ljava/awt/Choice;)Ljava/awt/peer/ChoicePeer;"(thread: JVMThread, args: [java_awt_Choice], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ChoicePeer) => void): void;
    public "createFrame(Ljava/awt/Frame;)Ljava/awt/peer/FramePeer;"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FramePeer) => void): void;
    public "java/awt/Toolkit/createFrame(Ljava/awt/Frame;)Ljava/awt/peer/FramePeer;"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FramePeer) => void): void;
    public "createCanvas(Ljava/awt/Canvas;)Ljava/awt/peer/CanvasPeer;"(thread: JVMThread, args: [java_awt_Canvas], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_CanvasPeer) => void): void;
    public "java/awt/Toolkit/createCanvas(Ljava/awt/Canvas;)Ljava/awt/peer/CanvasPeer;"(thread: JVMThread, args: [java_awt_Canvas], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_CanvasPeer) => void): void;
    public "createPanel(Ljava/awt/Panel;)Ljava/awt/peer/PanelPeer;"(thread: JVMThread, args: [java_awt_Panel], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_PanelPeer) => void): void;
    public "java/awt/Toolkit/createPanel(Ljava/awt/Panel;)Ljava/awt/peer/PanelPeer;"(thread: JVMThread, args: [java_awt_Panel], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_PanelPeer) => void): void;
    public "createWindow(Ljava/awt/Window;)Ljava/awt/peer/WindowPeer;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_WindowPeer) => void): void;
    public "java/awt/Toolkit/createWindow(Ljava/awt/Window;)Ljava/awt/peer/WindowPeer;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_WindowPeer) => void): void;
    public "createDialog(Ljava/awt/Dialog;)Ljava/awt/peer/DialogPeer;"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_DialogPeer) => void): void;
    public "java/awt/Toolkit/createDialog(Ljava/awt/Dialog;)Ljava/awt/peer/DialogPeer;"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_DialogPeer) => void): void;
    public "createMenuBar(Ljava/awt/MenuBar;)Ljava/awt/peer/MenuBarPeer;"(thread: JVMThread, args: [java_awt_MenuBar], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MenuBarPeer) => void): void;
    public "java/awt/Toolkit/createMenuBar(Ljava/awt/MenuBar;)Ljava/awt/peer/MenuBarPeer;"(thread: JVMThread, args: [java_awt_MenuBar], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MenuBarPeer) => void): void;
    public "createMenu(Ljava/awt/Menu;)Ljava/awt/peer/MenuPeer;"(thread: JVMThread, args: [java_awt_Menu], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MenuPeer) => void): void;
    public "java/awt/Toolkit/createMenu(Ljava/awt/Menu;)Ljava/awt/peer/MenuPeer;"(thread: JVMThread, args: [java_awt_Menu], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MenuPeer) => void): void;
    public "createPopupMenu(Ljava/awt/PopupMenu;)Ljava/awt/peer/PopupMenuPeer;"(thread: JVMThread, args: [java_awt_PopupMenu], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_PopupMenuPeer) => void): void;
    public "java/awt/Toolkit/createPopupMenu(Ljava/awt/PopupMenu;)Ljava/awt/peer/PopupMenuPeer;"(thread: JVMThread, args: [java_awt_PopupMenu], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_PopupMenuPeer) => void): void;
    public "createMenuItem(Ljava/awt/MenuItem;)Ljava/awt/peer/MenuItemPeer;"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MenuItemPeer) => void): void;
    public "java/awt/Toolkit/createMenuItem(Ljava/awt/MenuItem;)Ljava/awt/peer/MenuItemPeer;"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MenuItemPeer) => void): void;
    public "createFileDialog(Ljava/awt/FileDialog;)Ljava/awt/peer/FileDialogPeer;"(thread: JVMThread, args: [java_awt_FileDialog], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FileDialogPeer) => void): void;
    public "java/awt/Toolkit/createFileDialog(Ljava/awt/FileDialog;)Ljava/awt/peer/FileDialogPeer;"(thread: JVMThread, args: [java_awt_FileDialog], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FileDialogPeer) => void): void;
    public "createCheckboxMenuItem(Ljava/awt/CheckboxMenuItem;)Ljava/awt/peer/CheckboxMenuItemPeer;"(thread: JVMThread, args: [java_awt_CheckboxMenuItem], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_CheckboxMenuItemPeer) => void): void;
    public "java/awt/Toolkit/createCheckboxMenuItem(Ljava/awt/CheckboxMenuItem;)Ljava/awt/peer/CheckboxMenuItemPeer;"(thread: JVMThread, args: [java_awt_CheckboxMenuItem], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_CheckboxMenuItemPeer) => void): void;
    public "getMouseInfoPeer()Ljava/awt/peer/MouseInfoPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MouseInfoPeer) => void): void;
    public "java/awt/Toolkit/getMouseInfoPeer()Ljava/awt/peer/MouseInfoPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MouseInfoPeer) => void): void;
    public "createComponent(Ljava/awt/Component;)Ljava/awt/peer/LightweightPeer;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_LightweightPeer) => void): void;
    public "java/awt/Toolkit/createComponent(Ljava/awt/Component;)Ljava/awt/peer/LightweightPeer;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_LightweightPeer) => void): void;
    public "getFontPeer(Ljava/lang/String;I)Ljava/awt/peer/FontPeer;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FontPeer) => void): void;
    public "java/awt/Toolkit/getFontPeer(Ljava/lang/String;I)Ljava/awt/peer/FontPeer;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_FontPeer) => void): void;
    public "loadSystemColors([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/loadSystemColors([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDynamicLayout(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/setDynamicLayout(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isDynamicLayoutSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/isDynamicLayoutSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDynamicLayoutActive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/isDynamicLayoutActive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getScreenSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Toolkit/getScreenSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getScreenResolution()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/getScreenResolution()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getScreenInsets(Ljava/awt/GraphicsConfiguration;)Ljava/awt/Insets;"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public "java/awt/Toolkit/getScreenInsets(Ljava/awt/GraphicsConfiguration;)Ljava/awt/Insets;"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "java/awt/Toolkit/getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "getFontList()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/awt/Toolkit/getFontList()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "java/awt/Toolkit/getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initAssistiveTechnologies()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Toolkit/initAssistiveTechnologies()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "loadAssistiveTechnologies()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Toolkit/loadAssistiveTechnologies()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getDefaultToolkit()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public static "java/awt/Toolkit/getDefaultToolkit()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public "getImage(Ljava/lang/String;)Ljava/awt/Image;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Toolkit/getImage(Ljava/lang/String;)Ljava/awt/Image;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "getImage(Ljava/net/URL;)Ljava/awt/Image;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Toolkit/getImage(Ljava/net/URL;)Ljava/awt/Image;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "createImage(Ljava/lang/String;)Ljava/awt/Image;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Toolkit/createImage(Ljava/lang/String;)Ljava/awt/Image;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "createImage(Ljava/net/URL;)Ljava/awt/Image;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Toolkit/createImage(Ljava/net/URL;)Ljava/awt/Image;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Toolkit/createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "createImage([B)Ljava/awt/Image;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Toolkit/createImage([B)Ljava/awt/Image;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "createImage([BII)Ljava/awt/Image;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Toolkit/createImage([BII)Ljava/awt/Image;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "getPrintJob(Ljava/awt/Frame;Ljava/lang/String;Ljava/util/Properties;)Ljava/awt/PrintJob;"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, java_util_Properties], cb?: (e?: java_lang_Throwable, rv?: java_awt_PrintJob) => void): void;
    public "java/awt/Toolkit/getPrintJob(Ljava/awt/Frame;Ljava/lang/String;Ljava/util/Properties;)Ljava/awt/PrintJob;"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, java_util_Properties], cb?: (e?: java_lang_Throwable, rv?: java_awt_PrintJob) => void): void;
    public "getPrintJob(Ljava/awt/Frame;Ljava/lang/String;Ljava/awt/JobAttributes;Ljava/awt/PageAttributes;)Ljava/awt/PrintJob;"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, java_awt_JobAttributes, java_awt_PageAttributes], cb?: (e?: java_lang_Throwable, rv?: java_awt_PrintJob) => void): void;
    public "java/awt/Toolkit/getPrintJob(Ljava/awt/Frame;Ljava/lang/String;Ljava/awt/JobAttributes;Ljava/awt/PageAttributes;)Ljava/awt/PrintJob;"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, java_awt_JobAttributes, java_awt_PageAttributes], cb?: (e?: java_lang_Throwable, rv?: java_awt_PrintJob) => void): void;
    public "beep()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/beep()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSystemClipboard()Ljava/awt/datatransfer/Clipboard;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Clipboard) => void): void;
    public "java/awt/Toolkit/getSystemClipboard()Ljava/awt/datatransfer/Clipboard;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Clipboard) => void): void;
    public "getSystemSelection()Ljava/awt/datatransfer/Clipboard;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Clipboard) => void): void;
    public "java/awt/Toolkit/getSystemSelection()Ljava/awt/datatransfer/Clipboard;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Clipboard) => void): void;
    public "getMenuShortcutKeyMask()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/getMenuShortcutKeyMask()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLockingKeyState(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/getLockingKeyState(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLockingKeyState(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/setLockingKeyState(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getNativeContainer(Ljava/awt/Component;)Ljava/awt/Container;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public static "java/awt/Toolkit/getNativeContainer(Ljava/awt/Component;)Ljava/awt/Container;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "createCustomCursor(Ljava/awt/Image;Ljava/awt/Point;Ljava/lang/String;)Ljava/awt/Cursor;"(thread: JVMThread, args: [java_awt_Image, java_awt_Point, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public "java/awt/Toolkit/createCustomCursor(Ljava/awt/Image;Ljava/awt/Point;Ljava/lang/String;)Ljava/awt/Cursor;"(thread: JVMThread, args: [java_awt_Image, java_awt_Point, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public "getBestCursorSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Toolkit/getBestCursorSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getMaximumCursorColors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/getMaximumCursorColors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFrameStateSupported(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/isFrameStateSupported(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setPlatformResources(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Toolkit/setPlatformResources(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Toolkit/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "loadLibraries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Toolkit/loadLibraries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/Toolkit/getProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSystemEventQueue()Ljava/awt/EventQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventQueue) => void): void;
    public "java/awt/Toolkit/getSystemEventQueue()Ljava/awt/EventQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventQueue) => void): void;
    public "getSystemEventQueueImpl()Ljava/awt/EventQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventQueue) => void): void;
    public "java/awt/Toolkit/getSystemEventQueueImpl()Ljava/awt/EventQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventQueue) => void): void;
    public static "getEventQueue()Ljava/awt/EventQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventQueue) => void): void;
    public static "java/awt/Toolkit/getEventQueue()Ljava/awt/EventQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventQueue) => void): void;
    public "createDragSourceContextPeer(Ljava/awt/dnd/DragGestureEvent;)Ljava/awt/dnd/peer/DragSourceContextPeer;"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_peer_DragSourceContextPeer) => void): void;
    public "java/awt/Toolkit/createDragSourceContextPeer(Ljava/awt/dnd/DragGestureEvent;)Ljava/awt/dnd/peer/DragSourceContextPeer;"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_peer_DragSourceContextPeer) => void): void;
    public "createDragGestureRecognizer(Ljava/lang/Class;Ljava/awt/dnd/DragSource;Ljava/awt/Component;ILjava/awt/dnd/DragGestureListener;)Ljava/awt/dnd/DragGestureRecognizer;"(thread: JVMThread, args: [java_lang_Class, java_awt_dnd_DragSource, java_awt_Component, number, java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureRecognizer) => void): void;
    public "java/awt/Toolkit/createDragGestureRecognizer(Ljava/lang/Class;Ljava/awt/dnd/DragSource;Ljava/awt/Component;ILjava/awt/dnd/DragGestureListener;)Ljava/awt/dnd/DragGestureRecognizer;"(thread: JVMThread, args: [java_lang_Class, java_awt_dnd_DragSource, java_awt_Component, number, java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureRecognizer) => void): void;
    public "getDesktopProperty(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/Toolkit/getDesktopProperty(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setDesktopProperty(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/setDesktopProperty(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "lazilyLoadDesktopProperty(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/Toolkit/lazilyLoadDesktopProperty(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "initializeDesktopProperties()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/initializeDesktopProperties()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/removePropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPropertyChangeListeners()[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "java/awt/Toolkit/getPropertyChangeListeners()[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "getPropertyChangeListeners(Ljava/lang/String;)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "java/awt/Toolkit/getPropertyChangeListeners(Ljava/lang/String;)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "isAlwaysOnTopSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/isAlwaysOnTopSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isModalityTypeSupported(Ljava/awt/Dialog$ModalityType;)Z"(thread: JVMThread, args: [java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/isModalityTypeSupported(Ljava/awt/Dialog$ModalityType;)Z"(thread: JVMThread, args: [java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isModalExclusionTypeSupported(Ljava/awt/Dialog$ModalExclusionType;)Z"(thread: JVMThread, args: [java_awt_Dialog$ModalExclusionType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/isModalExclusionTypeSupported(Ljava/awt/Dialog$ModalExclusionType;)Z"(thread: JVMThread, args: [java_awt_Dialog$ModalExclusionType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "deProxyAWTEventListener(Ljava/awt/event/AWTEventListener;)Ljava/awt/event/AWTEventListener;"(thread: JVMThread, args: [java_awt_event_AWTEventListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_event_AWTEventListener) => void): void;
    public static "java/awt/Toolkit/deProxyAWTEventListener(Ljava/awt/event/AWTEventListener;)Ljava/awt/event/AWTEventListener;"(thread: JVMThread, args: [java_awt_event_AWTEventListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_event_AWTEventListener) => void): void;
    public "addAWTEventListener(Ljava/awt/event/AWTEventListener;J)V"(thread: JVMThread, args: [java_awt_event_AWTEventListener, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/addAWTEventListener(Ljava/awt/event/AWTEventListener;J)V"(thread: JVMThread, args: [java_awt_event_AWTEventListener, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeAWTEventListener(Ljava/awt/event/AWTEventListener;)V"(thread: JVMThread, args: [java_awt_event_AWTEventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/removeAWTEventListener(Ljava/awt/event/AWTEventListener;)V"(thread: JVMThread, args: [java_awt_event_AWTEventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public static "enabledOnToolkit(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Toolkit/enabledOnToolkit(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countAWTEventListeners(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/countAWTEventListeners(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAWTEventListeners()[Ljava/awt/event/AWTEventListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_AWTEventListener>) => void): void;
    public "java/awt/Toolkit/getAWTEventListeners()[Ljava/awt/event/AWTEventListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_AWTEventListener>) => void): void;
    public "getAWTEventListeners(J)[Ljava/awt/event/AWTEventListener;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_AWTEventListener>) => void): void;
    public "java/awt/Toolkit/getAWTEventListeners(J)[Ljava/awt/event/AWTEventListener;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_AWTEventListener>) => void): void;
    public "notifyAWTEventListeners(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Toolkit/notifyAWTEventListeners(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "mapInputMethodHighlight(Ljava/awt/im/InputMethodHighlight;)Ljava/util/Map;"(thread: JVMThread, args: [java_awt_im_InputMethodHighlight], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/awt/Toolkit/mapInputMethodHighlight(Ljava/awt/im/InputMethodHighlight;)Ljava/util/Map;"(thread: JVMThread, args: [java_awt_im_InputMethodHighlight], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "createPropertyChangeSupport(Ljava/awt/Toolkit;)Ljava/beans/PropertyChangeSupport;"(thread: JVMThread, args: [java_awt_Toolkit], cb?: (e?: java_lang_Throwable, rv?: java_beans_PropertyChangeSupport) => void): void;
    public static "java/awt/Toolkit/createPropertyChangeSupport(Ljava/awt/Toolkit;)Ljava/beans/PropertyChangeSupport;"(thread: JVMThread, args: [java_awt_Toolkit], cb?: (e?: java_lang_Throwable, rv?: java_beans_PropertyChangeSupport) => void): void;
    public "areExtraMouseButtonsEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Toolkit/areExtraMouseButtonsEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$002(Ljava/awt/Toolkit;)Ljava/awt/Toolkit;"(thread: JVMThread, args: [java_awt_Toolkit], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public static "java/awt/Toolkit/access$002(Ljava/awt/Toolkit;)Ljava/awt/Toolkit;"(thread: JVMThread, args: [java_awt_Toolkit], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public static "access$000()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public static "java/awt/Toolkit/access$000()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public static "access$100(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Toolkit/access$100(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$202(Ljava/util/ResourceBundle;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/awt/Toolkit/access$202(Ljava/util/ResourceBundle;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Toolkit/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Dialog$ModalityType extends java_lang_Enum {
    public static "java/awt/Dialog$ModalityType/MODELESS": java_awt_Dialog$ModalityType;
    public static "java/awt/Dialog$ModalityType/DOCUMENT_MODAL": java_awt_Dialog$ModalityType;
    public static "java/awt/Dialog$ModalityType/APPLICATION_MODAL": java_awt_Dialog$ModalityType;
    public static "java/awt/Dialog$ModalityType/TOOLKIT_MODAL": java_awt_Dialog$ModalityType;
    public static "java/awt/Dialog$ModalityType/$VALUES": JVMArray<java_awt_Dialog$ModalityType>;
    public static "values()[Ljava/awt/Dialog$ModalityType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Dialog$ModalityType>) => void): void;
    public static "java/awt/Dialog$ModalityType/values()[Ljava/awt/Dialog$ModalityType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Dialog$ModalityType>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/awt/Dialog$ModalityType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog$ModalityType) => void): void;
    public static "java/awt/Dialog$ModalityType/valueOf(Ljava/lang/String;)Ljava/awt/Dialog$ModalityType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog$ModalityType) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog$ModalityType/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Dialog$ModalityType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_dnd_DragGestureListener extends java_lang_Object, java_util_EventListener {
    "dragGestureRecognized(Ljava/awt/dnd/DragGestureEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_dnd_DragSource extends java_lang_Object implements java_io_Serializable {
    public static "java/awt/dnd/DragSource/serialVersionUID": Long;
    public static "java/awt/dnd/DragSource/DefaultCopyDrop": java_awt_Cursor;
    public static "java/awt/dnd/DragSource/DefaultMoveDrop": java_awt_Cursor;
    public static "java/awt/dnd/DragSource/DefaultLinkDrop": java_awt_Cursor;
    public static "java/awt/dnd/DragSource/DefaultCopyNoDrop": java_awt_Cursor;
    public static "java/awt/dnd/DragSource/DefaultMoveNoDrop": java_awt_Cursor;
    public static "java/awt/dnd/DragSource/DefaultLinkNoDrop": java_awt_Cursor;
    public static "java/awt/dnd/DragSource/dflt": java_awt_dnd_DragSource;
    public static "java/awt/dnd/DragSource/dragSourceListenerK": java_lang_String;
    public static "java/awt/dnd/DragSource/dragSourceMotionListenerK": java_lang_String;
    public "java/awt/dnd/DragSource/flavorMap": java_awt_datatransfer_FlavorMap;
    public "java/awt/dnd/DragSource/listener": java_awt_dnd_DragSourceListener;
    public "java/awt/dnd/DragSource/motionListener": java_awt_dnd_DragSourceMotionListener;
    public static "load(Ljava/lang/String;)Ljava/awt/Cursor;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public static "java/awt/dnd/DragSource/load(Ljava/lang/String;)Ljava/awt/Cursor;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public static "getDefaultDragSource()Ljava/awt/dnd/DragSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSource) => void): void;
    public static "java/awt/dnd/DragSource/getDefaultDragSource()Ljava/awt/dnd/DragSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSource) => void): void;
    public static "isDragImageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/dnd/DragSource/isDragImageSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "startDrag(Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;Ljava/awt/datatransfer/FlavorMap;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener, java_awt_datatransfer_FlavorMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/startDrag(Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;Ljava/awt/datatransfer/FlavorMap;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener, java_awt_datatransfer_FlavorMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "startDrag(Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;Ljava/awt/datatransfer/FlavorMap;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener, java_awt_datatransfer_FlavorMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/startDrag(Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;Ljava/awt/datatransfer/FlavorMap;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener, java_awt_datatransfer_FlavorMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "startDrag(Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/startDrag(Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "startDrag(Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/startDrag(Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "createDragSourceContext(Ljava/awt/dnd/peer/DragSourceContextPeer;Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)Ljava/awt/dnd/DragSourceContext;"(thread: JVMThread, args: [java_awt_dnd_peer_DragSourceContextPeer, java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSourceContext) => void): void;
    public "java/awt/dnd/DragSource/createDragSourceContext(Ljava/awt/dnd/peer/DragSourceContextPeer;Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)Ljava/awt/dnd/DragSourceContext;"(thread: JVMThread, args: [java_awt_dnd_peer_DragSourceContextPeer, java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSourceContext) => void): void;
    public "getFlavorMap()Ljava/awt/datatransfer/FlavorMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_FlavorMap) => void): void;
    public "java/awt/dnd/DragSource/getFlavorMap()Ljava/awt/datatransfer/FlavorMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_FlavorMap) => void): void;
    public "createDragGestureRecognizer(Ljava/lang/Class;Ljava/awt/Component;ILjava/awt/dnd/DragGestureListener;)Ljava/awt/dnd/DragGestureRecognizer;"(thread: JVMThread, args: [java_lang_Class, java_awt_Component, number, java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureRecognizer) => void): void;
    public "java/awt/dnd/DragSource/createDragGestureRecognizer(Ljava/lang/Class;Ljava/awt/Component;ILjava/awt/dnd/DragGestureListener;)Ljava/awt/dnd/DragGestureRecognizer;"(thread: JVMThread, args: [java_lang_Class, java_awt_Component, number, java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureRecognizer) => void): void;
    public "createDefaultDragGestureRecognizer(Ljava/awt/Component;ILjava/awt/dnd/DragGestureListener;)Ljava/awt/dnd/DragGestureRecognizer;"(thread: JVMThread, args: [java_awt_Component, number, java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureRecognizer) => void): void;
    public "java/awt/dnd/DragSource/createDefaultDragGestureRecognizer(Ljava/awt/Component;ILjava/awt/dnd/DragGestureListener;)Ljava/awt/dnd/DragGestureRecognizer;"(thread: JVMThread, args: [java_awt_Component, number, java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureRecognizer) => void): void;
    public "addDragSourceListener(Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/addDragSourceListener(Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeDragSourceListener(Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/removeDragSourceListener(Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDragSourceListeners()[Ljava/awt/dnd/DragSourceListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_dnd_DragSourceListener>) => void): void;
    public "java/awt/dnd/DragSource/getDragSourceListeners()[Ljava/awt/dnd/DragSourceListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_dnd_DragSourceListener>) => void): void;
    public "addDragSourceMotionListener(Ljava/awt/dnd/DragSourceMotionListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceMotionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/addDragSourceMotionListener(Ljava/awt/dnd/DragSourceMotionListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceMotionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeDragSourceMotionListener(Ljava/awt/dnd/DragSourceMotionListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceMotionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/removeDragSourceMotionListener(Ljava/awt/dnd/DragSourceMotionListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceMotionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDragSourceMotionListeners()[Ljava/awt/dnd/DragSourceMotionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_dnd_DragSourceMotionListener>) => void): void;
    public "java/awt/dnd/DragSource/getDragSourceMotionListeners()[Ljava/awt/dnd/DragSourceMotionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_dnd_DragSourceMotionListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/dnd/DragSource/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "processDragEnter(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/processDragEnter(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processDragOver(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/processDragOver(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processDropActionChanged(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/processDropActionChanged(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processDragExit(Ljava/awt/dnd/DragSourceEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/processDragExit(Ljava/awt/dnd/DragSourceEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processDragDropEnd(Ljava/awt/dnd/DragSourceDropEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDropEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/processDragDropEnd(Ljava/awt/dnd/DragSourceDropEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDropEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processDragMouseMoved(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/processDragMouseMoved(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSource/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getDragThreshold()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/dnd/DragSource/getDragThreshold()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/dnd/DragSource/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_dnd_DragSourceDropEvent extends java_awt_dnd_DragSourceEvent {
    public static "java/awt/dnd/DragSourceDropEvent/serialVersionUID": Long;
    public "java/awt/dnd/DragSourceDropEvent/dropSuccess": number;
    public "java/awt/dnd/DragSourceDropEvent/dropAction": number;
    public "<init>(Ljava/awt/dnd/DragSourceContext;IZ)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceDropEvent/<init>(Ljava/awt/dnd/DragSourceContext;IZ)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/dnd/DragSourceContext;IZII)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceDropEvent/<init>(Ljava/awt/dnd/DragSourceContext;IZII)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/dnd/DragSourceContext;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceDropEvent/<init>(Ljava/awt/dnd/DragSourceContext;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDropSuccess()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceDropEvent/getDropSuccess()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDropAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceDropEvent/getDropAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_dnd_DragSourceEvent extends java_util_EventObject {
    public static "java/awt/dnd/DragSourceEvent/serialVersionUID": Long;
    public "java/awt/dnd/DragSourceEvent/locationSpecified": number;
    public "java/awt/dnd/DragSourceEvent/x": number;
    public "java/awt/dnd/DragSourceEvent/y": number;
    public "<init>(Ljava/awt/dnd/DragSourceContext;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceEvent/<init>(Ljava/awt/dnd/DragSourceContext;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/dnd/DragSourceContext;II)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceEvent/<init>(Ljava/awt/dnd/DragSourceContext;II)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDragSourceContext()Ljava/awt/dnd/DragSourceContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSourceContext) => void): void;
    public "java/awt/dnd/DragSourceEvent/getDragSourceContext()Ljava/awt/dnd/DragSourceContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSourceContext) => void): void;
    public "getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/dnd/DragSourceEvent/getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "getX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceEvent/getX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceEvent/getY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_dnd_DragSourceDragEvent extends java_awt_dnd_DragSourceEvent {
    public static "java/awt/dnd/DragSourceDragEvent/serialVersionUID": Long;
    public static "java/awt/dnd/DragSourceDragEvent/JDK_1_3_MODIFIERS": number;
    public static "java/awt/dnd/DragSourceDragEvent/JDK_1_4_MODIFIERS": number;
    public "java/awt/dnd/DragSourceDragEvent/targetActions": number;
    public "java/awt/dnd/DragSourceDragEvent/dropAction": number;
    public "java/awt/dnd/DragSourceDragEvent/gestureModifiers": number;
    public "java/awt/dnd/DragSourceDragEvent/invalidModifiers": number;
    public "<init>(Ljava/awt/dnd/DragSourceContext;III)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceDragEvent/<init>(Ljava/awt/dnd/DragSourceContext;III)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/dnd/DragSourceContext;IIIII)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceDragEvent/<init>(Ljava/awt/dnd/DragSourceContext;IIIII)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTargetActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceDragEvent/getTargetActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGestureModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceDragEvent/getGestureModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGestureModifiersEx()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceDragEvent/getGestureModifiersEx()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUserAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceDragEvent/getUserAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDropAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceDragEvent/getDropAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setNewModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceDragEvent/setNewModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOldModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceDragEvent/setOldModifiers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_dnd_DragSourceContext extends java_lang_Object implements java_awt_dnd_DragSourceListener, java_awt_dnd_DragSourceMotionListener, java_io_Serializable {
    public static "java/awt/dnd/DragSourceContext/serialVersionUID": Long;
    public static "java/awt/dnd/DragSourceContext/DEFAULT": number;
    public static "java/awt/dnd/DragSourceContext/ENTER": number;
    public static "java/awt/dnd/DragSourceContext/OVER": number;
    public static "java/awt/dnd/DragSourceContext/CHANGED": number;
    public static "java/awt/dnd/DragSourceContext/emptyTransferable": java_awt_datatransfer_Transferable;
    public "java/awt/dnd/DragSourceContext/peer": java_awt_dnd_peer_DragSourceContextPeer;
    public "java/awt/dnd/DragSourceContext/trigger": java_awt_dnd_DragGestureEvent;
    public "java/awt/dnd/DragSourceContext/cursor": java_awt_Cursor;
    public "java/awt/dnd/DragSourceContext/transferable": java_awt_datatransfer_Transferable;
    public "java/awt/dnd/DragSourceContext/listener": java_awt_dnd_DragSourceListener;
    public "java/awt/dnd/DragSourceContext/useCustomCursor": number;
    public "java/awt/dnd/DragSourceContext/sourceActions": number;
    public "<init>(Ljava/awt/dnd/peer/DragSourceContextPeer;Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_peer_DragSourceContextPeer, java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/<init>(Ljava/awt/dnd/peer/DragSourceContextPeer;Ljava/awt/dnd/DragGestureEvent;Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_peer_DragSourceContextPeer, java_awt_dnd_DragGestureEvent, java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDragSource()Ljava/awt/dnd/DragSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSource) => void): void;
    public "java/awt/dnd/DragSourceContext/getDragSource()Ljava/awt/dnd/DragSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSource) => void): void;
    public "getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/dnd/DragSourceContext/getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getTrigger()Ljava/awt/dnd/DragGestureEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureEvent) => void): void;
    public "java/awt/dnd/DragSourceContext/getTrigger()Ljava/awt/dnd/DragGestureEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureEvent) => void): void;
    public "getSourceActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragSourceContext/getSourceActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setCursor(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/setCursor(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCursor()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public "java/awt/dnd/DragSourceContext/getCursor()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public "addDragSourceListener(Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/addDragSourceListener(Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeDragSourceListener(Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/removeDragSourceListener(Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "transferablesFlavorsChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/transferablesFlavorsChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dragEnter(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/dragEnter(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dragOver(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/dragOver(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dragExit(Ljava/awt/dnd/DragSourceEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/dragExit(Ljava/awt/dnd/DragSourceEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dropActionChanged(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/dropActionChanged(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dragDropEnd(Ljava/awt/dnd/DragSourceDropEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDropEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/dragDropEnd(Ljava/awt/dnd/DragSourceDropEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDropEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dragMouseMoved(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/dragMouseMoved(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "java/awt/dnd/DragSourceContext/getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "updateCurrentCursor(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/updateCurrentCursor(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCursorImpl(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/setCursorImpl(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragSourceContext/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_datatransfer_Transferable extends java_lang_Object {
    "getTransferDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    "isDataFlavorSupported(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTransferData(Ljava/awt/datatransfer/DataFlavor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_awt_datatransfer_DataFlavor extends java_lang_Object implements java_io_Externalizable, java_lang_Cloneable {
    public static "java/awt/datatransfer/DataFlavor/serialVersionUID": Long;
    public static "java/awt/datatransfer/DataFlavor/ioInputStreamClass": java_lang_Class;
    public static "java/awt/datatransfer/DataFlavor/stringFlavor": java_awt_datatransfer_DataFlavor;
    public static "java/awt/datatransfer/DataFlavor/imageFlavor": java_awt_datatransfer_DataFlavor;
    public static "java/awt/datatransfer/DataFlavor/plainTextFlavor": java_awt_datatransfer_DataFlavor;
    public static "java/awt/datatransfer/DataFlavor/javaSerializedObjectMimeType": java_lang_String;
    public static "java/awt/datatransfer/DataFlavor/javaFileListFlavor": java_awt_datatransfer_DataFlavor;
    public static "java/awt/datatransfer/DataFlavor/javaJVMLocalObjectMimeType": java_lang_String;
    public static "java/awt/datatransfer/DataFlavor/javaRemoteObjectMimeType": java_lang_String;
    public static "java/awt/datatransfer/DataFlavor/selectionHtmlFlavor": java_awt_datatransfer_DataFlavor;
    public static "java/awt/datatransfer/DataFlavor/fragmentHtmlFlavor": java_awt_datatransfer_DataFlavor;
    public static "java/awt/datatransfer/DataFlavor/allHtmlFlavor": java_awt_datatransfer_DataFlavor;
    public static "java/awt/datatransfer/DataFlavor/textFlavorComparator": java_util_Comparator;
    public "java/awt/datatransfer/DataFlavor/atom": number;
    public "java/awt/datatransfer/DataFlavor/mimeType": java_awt_datatransfer_MimeType;
    public "java/awt/datatransfer/DataFlavor/humanPresentableName": java_lang_String;
    public "java/awt/datatransfer/DataFlavor/representationClass": java_lang_Class;
    public static "tryToLoadClass(Ljava/lang/String;Ljava/lang/ClassLoader;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/awt/datatransfer/DataFlavor/tryToLoadClass(Ljava/lang/String;Ljava/lang/ClassLoader;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "createConstant(Ljava/lang/Class;Ljava/lang/String;)Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public static "java/awt/datatransfer/DataFlavor/createConstant(Ljava/lang/Class;Ljava/lang/String;)Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public static "createConstant(Ljava/lang/String;Ljava/lang/String;)Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public static "java/awt/datatransfer/DataFlavor/createConstant(Ljava/lang/String;Ljava/lang/String;)Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public static "initHtmlDataFlavor(Ljava/lang/String;)Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public static "java/awt/datatransfer/DataFlavor/initHtmlDataFlavor(Ljava/lang/String;)Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/awt/datatransfer/MimeTypeParameterList;Ljava/lang/Class;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_awt_datatransfer_MimeTypeParameterList, java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/awt/datatransfer/MimeTypeParameterList;Ljava/lang/Class;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_awt_datatransfer_MimeTypeParameterList, java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Class;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/<init>(Ljava/lang/Class;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "initialize(Ljava/lang/String;Ljava/lang/String;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/initialize(Ljava/lang/String;Ljava/lang/String;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getTextPlainUnicodeFlavor()Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public static "java/awt/datatransfer/DataFlavor/getTextPlainUnicodeFlavor()Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public static "selectBestTextFlavor([Ljava/awt/datatransfer/DataFlavor;)Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: [JVMArray<java_awt_datatransfer_DataFlavor>], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public static "java/awt/datatransfer/DataFlavor/selectBestTextFlavor([Ljava/awt/datatransfer/DataFlavor;)Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: [JVMArray<java_awt_datatransfer_DataFlavor>], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_DataFlavor) => void): void;
    public "getReaderForText(Ljava/awt/datatransfer/Transferable;)Ljava/io/Reader;"(thread: JVMThread, args: [java_awt_datatransfer_Transferable], cb?: (e?: java_lang_Throwable, rv?: java_io_Reader) => void): void;
    public "java/awt/datatransfer/DataFlavor/getReaderForText(Ljava/awt/datatransfer/Transferable;)Ljava/io/Reader;"(thread: JVMThread, args: [java_awt_datatransfer_Transferable], cb?: (e?: java_lang_Throwable, rv?: java_io_Reader) => void): void;
    public "getMimeType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/getMimeType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRepresentationClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/awt/datatransfer/DataFlavor/getRepresentationClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getHumanPresentableName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/getHumanPresentableName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPrimaryType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/getPrimaryType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSubType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/getSubType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getParameter(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/getParameter(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setHumanPresentableName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/setHumanPresentableName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/equals(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/equals(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "match(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/match(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMimeTypeEqual(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isMimeTypeEqual(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMimeTypeEqual(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isMimeTypeEqual(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMimeTypeEqual(Ljava/awt/datatransfer/MimeType;)Z"(thread: JVMThread, args: [java_awt_datatransfer_MimeType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isMimeTypeEqual(Ljava/awt/datatransfer/MimeType;)Z"(thread: JVMThread, args: [java_awt_datatransfer_MimeType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isStandardTextRepresentationClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isStandardTextRepresentationClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMimeTypeSerializedObject()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isMimeTypeSerializedObject()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDefaultRepresentationClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/awt/datatransfer/DataFlavor/getDefaultRepresentationClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getDefaultRepresentationClassAsString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/getDefaultRepresentationClassAsString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isRepresentationClassInputStream()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isRepresentationClassInputStream()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRepresentationClassReader()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isRepresentationClassReader()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRepresentationClassCharBuffer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isRepresentationClassCharBuffer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRepresentationClassByteBuffer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isRepresentationClassByteBuffer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRepresentationClassSerializable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isRepresentationClassSerializable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRepresentationClassRemote()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isRepresentationClassRemote()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFlavorSerializedObjectType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isFlavorSerializedObjectType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFlavorRemoteObjectType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isFlavorRemoteObjectType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFlavorJavaFileListType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isFlavorJavaFileListType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFlavorTextType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/DataFlavor/isFlavorTextType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeExternal(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/writeExternal(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "readExternal(Ljava/io/ObjectInput;)V"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/DataFlavor/readExternal(Ljava/io/ObjectInput;)V"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/datatransfer/DataFlavor/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "normalizeMimeTypeParameter(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/normalizeMimeTypeParameter(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "normalizeMimeType(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/DataFlavor/normalizeMimeType(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/datatransfer/DataFlavor/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_Reader extends java_lang_Object implements java_lang_Readable, java_io_Closeable {
    public "java/io/Reader/lock": java_lang_Object;
    public static "java/io/Reader/maxSkipBufferSize": number;
    public "java/io/Reader/skipBuffer": JVMArray<number>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Reader/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Reader/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "read(Ljava/nio/CharBuffer;)I"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/Reader/read(Ljava/nio/CharBuffer;)I"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/Reader/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([C)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/Reader/read([C)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/Reader/read([CII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/Reader/skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "ready()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/Reader/ready()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "markSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/Reader/markSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mark(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Reader/mark(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Reader/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Reader/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_CharBuffer extends java_nio_Buffer implements java_lang_Comparable, java_lang_Appendable, java_lang_CharSequence, java_lang_Readable {
    public "java/nio/CharBuffer/hb": JVMArray<number>;
    public "java/nio/CharBuffer/offset": number;
    public "java/nio/CharBuffer/isReadOnly": number;
    public "<init>(IIII[CI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/CharBuffer/<init>(IIII[CI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/CharBuffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "allocate(I)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public static "java/nio/CharBuffer/allocate(I)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public static "wrap([CII)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public static "java/nio/CharBuffer/wrap([CII)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public static "wrap([C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public static "java/nio/CharBuffer/wrap([C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "read(Ljava/nio/CharBuffer;)I"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/read(Ljava/nio/CharBuffer;)I"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "wrap(Ljava/lang/CharSequence;II)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public static "java/nio/CharBuffer/wrap(Ljava/lang/CharSequence;II)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public static "wrap(Ljava/lang/CharSequence;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public static "java/nio/CharBuffer/wrap(Ljava/lang/CharSequence;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "slice()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/slice()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "duplicate()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/duplicate()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "asReadOnlyBuffer()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/asReadOnlyBuffer()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "get()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/get()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/put(C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "get(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/get(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUnchecked(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/getUnchecked(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(IC)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/put(IC)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "get([CII)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/get([CII)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "get([C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/get([C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "put(Ljava/nio/CharBuffer;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/put(Ljava/nio/CharBuffer;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "put([CII)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/put([CII)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "put([C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/put([C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "put(Ljava/lang/String;II)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/put(Ljava/lang/String;II)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "put(Ljava/lang/String;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/put(Ljava/lang/String;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "array()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/nio/CharBuffer/array()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compact()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/compact()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "equals(CC)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/CharBuffer/equals(CC)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/nio/CharBuffer;)I"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/compareTo(Ljava/nio/CharBuffer;)I"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(CC)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/CharBuffer/compare(CC)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/CharBuffer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/CharBuffer/toString(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "subSequence(II)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/subSequence(II)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/append(Ljava/lang/CharSequence;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/append(Ljava/lang/CharSequence;II)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "append(C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/CharBuffer/append(C)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "java/nio/CharBuffer/order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "chars()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "java/nio/CharBuffer/chars()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    public "array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/CharBuffer/array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/CharBuffer/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/nio/CharBuffer/append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/nio/CharBuffer/append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/nio/CharBuffer/append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "java/nio/CharBuffer/subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "lambda$chars$0()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    public "java/nio/CharBuffer/lambda$chars$0()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    "codePoints()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "lambda$codePoints$1()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    "chars()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
  }
  export interface java_lang_Readable extends java_lang_Object {
    "read(Ljava/nio/CharBuffer;)I"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_datatransfer_MimeTypeParameterList extends java_lang_Object implements java_lang_Cloneable {
    public "java/awt/datatransfer/MimeTypeParameterList/parameters": java_util_Hashtable;
    public static "java/awt/datatransfer/MimeTypeParameterList/TSPECIALS": java_lang_String;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parse(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/parse(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/get(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "set(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/set(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/getNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/datatransfer/MimeTypeParameterList/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "isTokenChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/datatransfer/MimeTypeParameterList/isTokenChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "skipWhiteSpace(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/datatransfer/MimeTypeParameterList/skipWhiteSpace(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "quote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/datatransfer/MimeTypeParameterList/quote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "unquote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/datatransfer/MimeTypeParameterList/unquote(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_datatransfer_MimeType extends java_lang_Object implements java_io_Externalizable, java_lang_Cloneable {
    public static "java/awt/datatransfer/MimeType/serialVersionUID": Long;
    public "java/awt/datatransfer/MimeType/primaryType": java_lang_String;
    public "java/awt/datatransfer/MimeType/subType": java_lang_String;
    public "java/awt/datatransfer/MimeType/parameters": java_awt_datatransfer_MimeTypeParameterList;
    public static "java/awt/datatransfer/MimeType/TSPECIALS": java_lang_String;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeType/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeType/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeType/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/awt/datatransfer/MimeTypeParameterList;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_awt_datatransfer_MimeTypeParameterList], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeType/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/awt/datatransfer/MimeTypeParameterList;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_awt_datatransfer_MimeTypeParameterList], cb?: (e?: java_lang_Throwable) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/MimeType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/MimeType/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parse(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeType/parse(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPrimaryType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/MimeType/getPrimaryType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSubType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/MimeType/getSubType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getParameters()Ljava/awt/datatransfer/MimeTypeParameterList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_MimeTypeParameterList) => void): void;
    public "java/awt/datatransfer/MimeType/getParameters()Ljava/awt/datatransfer/MimeTypeParameterList;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_MimeTypeParameterList) => void): void;
    public "getParameter(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/MimeType/getParameter(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setParameter(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeType/setParameter(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeParameter(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeType/removeParameter(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/MimeType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getBaseType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/MimeType/getBaseType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "match(Ljava/awt/datatransfer/MimeType;)Z"(thread: JVMThread, args: [java_awt_datatransfer_MimeType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/MimeType/match(Ljava/awt/datatransfer/MimeType;)Z"(thread: JVMThread, args: [java_awt_datatransfer_MimeType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "match(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/MimeType/match(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeExternal(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeType/writeExternal(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "readExternal(Ljava/io/ObjectInput;)V"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/MimeType/readExternal(Ljava/io/ObjectInput;)V"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/datatransfer/MimeType/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "isTokenChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/datatransfer/MimeType/isTokenChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isValidToken(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/MimeType/isValidToken(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_io_Externalizable extends java_lang_Object, java_io_Serializable {
    "writeExternal(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    "readExternal(Ljava/io/ObjectInput;)V"(thread: JVMThread, args: [java_io_ObjectInput], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_dnd_DragSourceMotionListener extends java_lang_Object, java_util_EventListener {
    "dragMouseMoved(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_dnd_DragSourceListener extends java_lang_Object, java_util_EventListener {
    "dragEnter(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "dragOver(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "dropActionChanged(Ljava/awt/dnd/DragSourceDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "dragExit(Ljava/awt/dnd/DragSourceEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "dragDropEnd(Ljava/awt/dnd/DragSourceDropEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceDropEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_datatransfer_FlavorMap extends java_lang_Object {
    "getNativesForFlavors([Ljava/awt/datatransfer/DataFlavor;)Ljava/util/Map;"(thread: JVMThread, args: [JVMArray<java_awt_datatransfer_DataFlavor>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    "getFlavorsForNatives([Ljava/lang/String;)Ljava/util/Map;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
  }
  export class java_awt_dnd_DragGestureRecognizer extends java_lang_Object implements java_io_Serializable {
    public static "java/awt/dnd/DragGestureRecognizer/serialVersionUID": Long;
    public "java/awt/dnd/DragGestureRecognizer/dragSource": java_awt_dnd_DragSource;
    public "java/awt/dnd/DragGestureRecognizer/component": java_awt_Component;
    public "java/awt/dnd/DragGestureRecognizer/dragGestureListener": java_awt_dnd_DragGestureListener;
    public "java/awt/dnd/DragGestureRecognizer/sourceActions": number;
    public "java/awt/dnd/DragGestureRecognizer/events": java_util_ArrayList;
    public "<init>(Ljava/awt/dnd/DragSource;Ljava/awt/Component;ILjava/awt/dnd/DragGestureListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSource, java_awt_Component, number, java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/<init>(Ljava/awt/dnd/DragSource;Ljava/awt/Component;ILjava/awt/dnd/DragGestureListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragSource, java_awt_Component, number, java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/dnd/DragSource;Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_dnd_DragSource, java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/<init>(Ljava/awt/dnd/DragSource;Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_dnd_DragSource, java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/dnd/DragSource;Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_dnd_DragSource, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/<init>(Ljava/awt/dnd/DragSource;Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_dnd_DragSource, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/dnd/DragSource;)V"(thread: JVMThread, args: [java_awt_dnd_DragSource], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/<init>(Ljava/awt/dnd/DragSource;)V"(thread: JVMThread, args: [java_awt_dnd_DragSource], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerListeners()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/registerListeners()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "unregisterListeners()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/unregisterListeners()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDragSource()Ljava/awt/dnd/DragSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSource) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/getDragSource()Ljava/awt/dnd/DragSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSource) => void): void;
    public "getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "setComponent(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/setComponent(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSourceActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/getSourceActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSourceActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/setSourceActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTriggerEvent()Ljava/awt/event/InputEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_event_InputEvent) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/getTriggerEvent()Ljava/awt/event/InputEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_event_InputEvent) => void): void;
    public "resetRecognizer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/resetRecognizer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addDragGestureListener(Ljava/awt/dnd/DragGestureListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/addDragGestureListener(Ljava/awt/dnd/DragGestureListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeDragGestureListener(Ljava/awt/dnd/DragGestureListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/removeDragGestureListener(Ljava/awt/dnd/DragGestureListener;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "fireDragGestureRecognized(ILjava/awt/Point;)V"(thread: JVMThread, args: [number, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/fireDragGestureRecognized(ILjava/awt/Point;)V"(thread: JVMThread, args: [number, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "appendEvent(Ljava/awt/event/InputEvent;)V"(thread: JVMThread, args: [java_awt_event_InputEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/appendEvent(Ljava/awt/event/InputEvent;)V"(thread: JVMThread, args: [java_awt_event_InputEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureRecognizer/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_dnd_DragGestureEvent extends java_util_EventObject {
    public static "java/awt/dnd/DragGestureEvent/serialVersionUID": Long;
    public "java/awt/dnd/DragGestureEvent/events": java_util_List;
    public "java/awt/dnd/DragGestureEvent/dragSource": java_awt_dnd_DragSource;
    public "java/awt/dnd/DragGestureEvent/component": java_awt_Component;
    public "java/awt/dnd/DragGestureEvent/origin": java_awt_Point;
    public "java/awt/dnd/DragGestureEvent/action": number;
    public "<init>(Ljava/awt/dnd/DragGestureRecognizer;ILjava/awt/Point;Ljava/util/List;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureRecognizer, number, java_awt_Point, java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureEvent/<init>(Ljava/awt/dnd/DragGestureRecognizer;ILjava/awt/Point;Ljava/util/List;)V"(thread: JVMThread, args: [java_awt_dnd_DragGestureRecognizer, number, java_awt_Point, java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSourceAsDragGestureRecognizer()Ljava/awt/dnd/DragGestureRecognizer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureRecognizer) => void): void;
    public "java/awt/dnd/DragGestureEvent/getSourceAsDragGestureRecognizer()Ljava/awt/dnd/DragGestureRecognizer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragGestureRecognizer) => void): void;
    public "getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/dnd/DragGestureEvent/getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getDragSource()Ljava/awt/dnd/DragSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSource) => void): void;
    public "java/awt/dnd/DragGestureEvent/getDragSource()Ljava/awt/dnd/DragSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DragSource) => void): void;
    public "getDragOrigin()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/dnd/DragGestureEvent/getDragOrigin()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/awt/dnd/DragGestureEvent/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/awt/dnd/DragGestureEvent/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/awt/dnd/DragGestureEvent/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "getDragAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DragGestureEvent/getDragAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTriggerEvent()Ljava/awt/event/InputEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_event_InputEvent) => void): void;
    public "java/awt/dnd/DragGestureEvent/getTriggerEvent()Ljava/awt/event/InputEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_event_InputEvent) => void): void;
    public "startDrag(Ljava/awt/Cursor;Ljava/awt/datatransfer/Transferable;)V"(thread: JVMThread, args: [java_awt_Cursor, java_awt_datatransfer_Transferable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureEvent/startDrag(Ljava/awt/Cursor;Ljava/awt/datatransfer/Transferable;)V"(thread: JVMThread, args: [java_awt_Cursor, java_awt_datatransfer_Transferable], cb?: (e?: java_lang_Throwable) => void): void;
    public "startDrag(Ljava/awt/Cursor;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_Cursor, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureEvent/startDrag(Ljava/awt/Cursor;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_Cursor, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "startDrag(Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureEvent/startDrag(Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;Ljava/awt/datatransfer/Transferable;Ljava/awt/dnd/DragSourceListener;)V"(thread: JVMThread, args: [java_awt_Cursor, java_awt_Image, java_awt_Point, java_awt_datatransfer_Transferable, java_awt_dnd_DragSourceListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureEvent/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DragGestureEvent/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_dnd_peer_DragSourceContextPeer extends java_lang_Object {
    "startDrag(Ljava/awt/dnd/DragSourceContext;Ljava/awt/Cursor;Ljava/awt/Image;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_dnd_DragSourceContext, java_awt_Cursor, java_awt_Image, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    "getCursor()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    "setCursor(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    "transferablesFlavorsChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_EventQueue extends java_lang_Object {
    public static "java/awt/EventQueue/threadInitNumber": java_util_concurrent_atomic_AtomicInteger;
    public static "java/awt/EventQueue/LOW_PRIORITY": number;
    public static "java/awt/EventQueue/NORM_PRIORITY": number;
    public static "java/awt/EventQueue/HIGH_PRIORITY": number;
    public static "java/awt/EventQueue/ULTIMATE_PRIORITY": number;
    public static "java/awt/EventQueue/NUM_PRIORITIES": number;
    public "java/awt/EventQueue/queues": JVMArray<java_awt_Queue>;
    public "java/awt/EventQueue/nextQueue": java_awt_EventQueue;
    public "java/awt/EventQueue/previousQueue": java_awt_EventQueue;
    public "java/awt/EventQueue/pushPopLock": java_util_concurrent_locks_Lock;
    public "java/awt/EventQueue/pushPopCond": java_util_concurrent_locks_Condition;
    public static "java/awt/EventQueue/dummyRunnable": java_lang_Runnable;
    public "java/awt/EventQueue/dispatchThread": java_awt_EventDispatchThread;
    public "java/awt/EventQueue/threadGroup": java_lang_ThreadGroup;
    public "java/awt/EventQueue/classLoader": java_lang_ClassLoader;
    public "java/awt/EventQueue/mostRecentEventTime": Long;
    public "java/awt/EventQueue/mostRecentKeyEventTime": Long;
    public "java/awt/EventQueue/currentEvent": java_lang_ref_WeakReference;
    public "java/awt/EventQueue/waitForID": number;
    public "java/awt/EventQueue/appContext": sun_awt_AppContext;
    public "java/awt/EventQueue/name": java_lang_String;
    public "java/awt/EventQueue/fwDispatcher": sun_awt_FwDispatcher;
    public static "java/awt/EventQueue/eventLog": sun_util_logging_PlatformLogger;
    public static "java/awt/EventQueue/PAINT": number;
    public static "java/awt/EventQueue/UPDATE": number;
    public static "java/awt/EventQueue/MOVE": number;
    public static "java/awt/EventQueue/DRAG": number;
    public static "java/awt/EventQueue/PEER": number;
    public static "java/awt/EventQueue/CACHE_LENGTH": number;
    public static "java/awt/EventQueue/javaSecurityAccess": sun_misc_JavaSecurityAccess;
    public static "getEventLog()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "java/awt/EventQueue/getEventLog()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "postEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/postEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "postEventPrivate(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/postEventPrivate(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getPriority(Ljava/awt/AWTEvent;)I"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/EventQueue/getPriority(Ljava/awt/AWTEvent;)I"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "postEvent(Ljava/awt/AWTEvent;I)V"(thread: JVMThread, args: [java_awt_AWTEvent, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/postEvent(Ljava/awt/AWTEvent;I)V"(thread: JVMThread, args: [java_awt_AWTEvent, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "coalescePaintEvent(Ljava/awt/event/PaintEvent;)Z"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/EventQueue/coalescePaintEvent(Ljava/awt/event/PaintEvent;)Z"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mergePaintEvents(Ljava/awt/event/PaintEvent;Ljava/awt/event/PaintEvent;)Ljava/awt/event/PaintEvent;"(thread: JVMThread, args: [java_awt_event_PaintEvent, java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_event_PaintEvent) => void): void;
    public "java/awt/EventQueue/mergePaintEvents(Ljava/awt/event/PaintEvent;Ljava/awt/event/PaintEvent;)Ljava/awt/event/PaintEvent;"(thread: JVMThread, args: [java_awt_event_PaintEvent, java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_event_PaintEvent) => void): void;
    public "coalesceMouseEvent(Ljava/awt/event/MouseEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/EventQueue/coalesceMouseEvent(Ljava/awt/event/MouseEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "coalescePeerEvent(Lsun/awt/PeerEvent;)Z"(thread: JVMThread, args: [sun_awt_PeerEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/EventQueue/coalescePeerEvent(Lsun/awt/PeerEvent;)Z"(thread: JVMThread, args: [sun_awt_PeerEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "coalesceOtherEvent(Ljava/awt/AWTEvent;I)Z"(thread: JVMThread, args: [java_awt_AWTEvent, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/EventQueue/coalesceOtherEvent(Ljava/awt/AWTEvent;I)Z"(thread: JVMThread, args: [java_awt_AWTEvent, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "coalesceEvent(Ljava/awt/AWTEvent;I)Z"(thread: JVMThread, args: [java_awt_AWTEvent, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/EventQueue/coalesceEvent(Ljava/awt/AWTEvent;I)Z"(thread: JVMThread, args: [java_awt_AWTEvent, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "cacheEQItem(Lsun/awt/EventQueueItem;)V"(thread: JVMThread, args: [sun_awt_EventQueueItem], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/cacheEQItem(Lsun/awt/EventQueueItem;)V"(thread: JVMThread, args: [sun_awt_EventQueueItem], cb?: (e?: java_lang_Throwable) => void): void;
    public "uncacheEQItem(Lsun/awt/EventQueueItem;)V"(thread: JVMThread, args: [sun_awt_EventQueueItem], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/uncacheEQItem(Lsun/awt/EventQueueItem;)V"(thread: JVMThread, args: [sun_awt_EventQueueItem], cb?: (e?: java_lang_Throwable) => void): void;
    public static "eventToCacheIndex(Ljava/awt/AWTEvent;)I"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/EventQueue/eventToCacheIndex(Ljava/awt/AWTEvent;)I"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "noEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/EventQueue/noEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNextEvent()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "java/awt/EventQueue/getNextEvent()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "getNextEventPrivate()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "java/awt/EventQueue/getNextEventPrivate()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "getNextEvent(I)Ljava/awt/AWTEvent;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "java/awt/EventQueue/getNextEvent(I)Ljava/awt/AWTEvent;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "peekEvent()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "java/awt/EventQueue/peekEvent()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "peekEvent(I)Ljava/awt/AWTEvent;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "java/awt/EventQueue/peekEvent(I)Ljava/awt/AWTEvent;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "dispatchEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/dispatchEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getAccessControlContextFrom(Ljava/lang/Object;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "java/awt/EventQueue/getAccessControlContextFrom(Ljava/lang/Object;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "dispatchEventImpl(Ljava/awt/AWTEvent;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_AWTEvent, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/dispatchEventImpl(Ljava/awt/AWTEvent;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_AWTEvent, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getMostRecentEventTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/awt/EventQueue/getMostRecentEventTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMostRecentEventTimeImpl()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/awt/EventQueue/getMostRecentEventTimeImpl()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMostRecentEventTimeEx()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/awt/EventQueue/getMostRecentEventTimeEx()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getCurrentEvent()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public static "java/awt/EventQueue/getCurrentEvent()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "getCurrentEventImpl()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "java/awt/EventQueue/getCurrentEventImpl()Ljava/awt/AWTEvent;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "push(Ljava/awt/EventQueue;)V"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/push(Ljava/awt/EventQueue;)V"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "pop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/pop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "createSecondaryLoop()Ljava/awt/SecondaryLoop;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_SecondaryLoop) => void): void;
    public "java/awt/EventQueue/createSecondaryLoop()Ljava/awt/SecondaryLoop;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_SecondaryLoop) => void): void;
    public "createSecondaryLoop(Ljava/awt/Conditional;Ljava/awt/EventFilter;J)Ljava/awt/SecondaryLoop;"(thread: JVMThread, args: [java_awt_Conditional, java_awt_EventFilter, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_SecondaryLoop) => void): void;
    public "java/awt/EventQueue/createSecondaryLoop(Ljava/awt/Conditional;Ljava/awt/EventFilter;J)Ljava/awt/SecondaryLoop;"(thread: JVMThread, args: [java_awt_Conditional, java_awt_EventFilter, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_SecondaryLoop) => void): void;
    public static "isDispatchThread()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/EventQueue/isDispatchThread()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDispatchThreadImpl()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/EventQueue/isDispatchThreadImpl()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "initDispatchThread()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/initDispatchThread()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "detachDispatchThread(Ljava/awt/EventDispatchThread;)V"(thread: JVMThread, args: [java_awt_EventDispatchThread], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/detachDispatchThread(Ljava/awt/EventDispatchThread;)V"(thread: JVMThread, args: [java_awt_EventDispatchThread], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDispatchThread()Ljava/awt/EventDispatchThread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventDispatchThread) => void): void;
    public "java/awt/EventQueue/getDispatchThread()Ljava/awt/EventDispatchThread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventDispatchThread) => void): void;
    public "removeSourceEvents(Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/removeSourceEvents(Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMostRecentKeyEventTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/awt/EventQueue/getMostRecentKeyEventTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "setCurrentEventAndMostRecentTime(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventQueue/setCurrentEventAndMostRecentTime(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCurrentEventAndMostRecentTimeImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/setCurrentEventAndMostRecentTimeImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public static "invokeLater(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventQueue/invokeLater(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "invokeAndWait(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventQueue/invokeAndWait(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "invokeAndWait(Ljava/lang/Object;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventQueue/invokeAndWait(Ljava/lang/Object;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "wakeup(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/wakeup(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFwDispatcher(Lsun/awt/FwDispatcher;)V"(thread: JVMThread, args: [sun_awt_FwDispatcher], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventQueue/setFwDispatcher(Lsun/awt/FwDispatcher;)V"(thread: JVMThread, args: [sun_awt_FwDispatcher], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/awt/EventQueue;)Z"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/EventQueue/access$000(Ljava/awt/EventQueue;)Z"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Ljava/awt/EventQueue;Z)V"(thread: JVMThread, args: [java_awt_EventQueue, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventQueue/access$100(Ljava/awt/EventQueue;Z)V"(thread: JVMThread, args: [java_awt_EventQueue, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$200(Ljava/awt/EventQueue;Lsun/awt/FwDispatcher;)V"(thread: JVMThread, args: [java_awt_EventQueue, sun_awt_FwDispatcher], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventQueue/access$200(Ljava/awt/EventQueue;Lsun/awt/FwDispatcher;)V"(thread: JVMThread, args: [java_awt_EventQueue, sun_awt_FwDispatcher], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$300(Ljava/awt/EventQueue;)J"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/awt/EventQueue/access$300(Ljava/awt/EventQueue;)J"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$400(Ljava/awt/EventQueue;)Lsun/awt/FwDispatcher;"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FwDispatcher) => void): void;
    public static "java/awt/EventQueue/access$400(Ljava/awt/EventQueue;)Lsun/awt/FwDispatcher;"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FwDispatcher) => void): void;
    public static "access$500(Ljava/awt/EventQueue;Ljava/awt/AWTEvent;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_EventQueue, java_awt_AWTEvent, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventQueue/access$500(Ljava/awt/EventQueue;Ljava/awt/AWTEvent;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_EventQueue, java_awt_AWTEvent, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$600()Lsun/misc/JavaSecurityAccess;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_JavaSecurityAccess) => void): void;
    public static "java/awt/EventQueue/access$600()Lsun/misc/JavaSecurityAccess;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_JavaSecurityAccess) => void): void;
    public static "access$700(Ljava/awt/EventQueue;)Ljava/lang/ThreadGroup;"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public static "java/awt/EventQueue/access$700(Ljava/awt/EventQueue;)Ljava/lang/ThreadGroup;"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public static "access$800(Ljava/awt/EventQueue;)Ljava/lang/String;"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/awt/EventQueue/access$800(Ljava/awt/EventQueue;)Ljava/lang/String;"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$900(Ljava/awt/EventQueue;)Ljava/lang/ClassLoader;"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "java/awt/EventQueue/access$900(Ljava/awt/EventQueue;)Ljava/lang/ClassLoader;"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventQueue/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_EventFilter extends java_lang_Object {
    "acceptEvent(Ljava/awt/AWTEvent;)Ljava/awt/EventFilter$FilterAction;"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventFilter$FilterAction) => void): void;
  }
  export class java_awt_EventFilter$FilterAction extends java_lang_Enum {
    public static "java/awt/EventFilter$FilterAction/ACCEPT": java_awt_EventFilter$FilterAction;
    public static "java/awt/EventFilter$FilterAction/REJECT": java_awt_EventFilter$FilterAction;
    public static "java/awt/EventFilter$FilterAction/ACCEPT_IMMEDIATELY": java_awt_EventFilter$FilterAction;
    public static "java/awt/EventFilter$FilterAction/$VALUES": JVMArray<java_awt_EventFilter$FilterAction>;
    public static "values()[Ljava/awt/EventFilter$FilterAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_EventFilter$FilterAction>) => void): void;
    public static "java/awt/EventFilter$FilterAction/values()[Ljava/awt/EventFilter$FilterAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_EventFilter$FilterAction>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/awt/EventFilter$FilterAction;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventFilter$FilterAction) => void): void;
    public static "java/awt/EventFilter$FilterAction/valueOf(Ljava/lang/String;)Ljava/awt/EventFilter$FilterAction;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventFilter$FilterAction) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventFilter$FilterAction/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventFilter$FilterAction/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_Conditional extends java_lang_Object {
    "evaluate()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_SecondaryLoop extends java_lang_Object {
    "enter()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "exit()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_awt_EventQueueItem extends java_lang_Object {
    public "sun/awt/EventQueueItem/event": java_awt_AWTEvent;
    public "sun/awt/EventQueueItem/next": sun_awt_EventQueueItem;
    public "<init>(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/EventQueueItem/<init>(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_PeerEvent extends java_awt_event_InvocationEvent {
    public static "sun/awt/PeerEvent/PRIORITY_EVENT": Long;
    public static "sun/awt/PeerEvent/ULTIMATE_PRIORITY_EVENT": Long;
    public static "sun/awt/PeerEvent/LOW_PRIORITY_EVENT": Long;
    public "sun/awt/PeerEvent/flags": Long;
    public "<init>(Ljava/lang/Object;Ljava/lang/Runnable;J)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/PeerEvent/<init>(Ljava/lang/Object;Ljava/lang/Runnable;J)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;Ljava/lang/Runnable;Ljava/lang/Object;ZJ)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable, java_lang_Object, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/PeerEvent/<init>(Ljava/lang/Object;Ljava/lang/Runnable;Ljava/lang/Object;ZJ)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable, java_lang_Object, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFlags()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/awt/PeerEvent/getFlags()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "coalesceEvents(Lsun/awt/PeerEvent;)Lsun/awt/PeerEvent;"(thread: JVMThread, args: [sun_awt_PeerEvent], cb?: (e?: java_lang_Throwable, rv?: sun_awt_PeerEvent) => void): void;
    public "sun/awt/PeerEvent/coalesceEvents(Lsun/awt/PeerEvent;)Lsun/awt/PeerEvent;"(thread: JVMThread, args: [sun_awt_PeerEvent], cb?: (e?: java_lang_Throwable, rv?: sun_awt_PeerEvent) => void): void;
  }
  export class java_awt_event_InvocationEvent extends java_awt_AWTEvent implements java_awt_ActiveEvent {
    public static "java/awt/event/InvocationEvent/INVOCATION_FIRST": number;
    public static "java/awt/event/InvocationEvent/INVOCATION_DEFAULT": number;
    public static "java/awt/event/InvocationEvent/INVOCATION_LAST": number;
    public "java/awt/event/InvocationEvent/runnable": java_lang_Runnable;
    public "java/awt/event/InvocationEvent/notifier": java_lang_Object;
    public "java/awt/event/InvocationEvent/listener": java_lang_Runnable;
    public "java/awt/event/InvocationEvent/dispatched": number;
    public "java/awt/event/InvocationEvent/catchExceptions": number;
    public "java/awt/event/InvocationEvent/exception": java_lang_Exception;
    public "java/awt/event/InvocationEvent/throwable": java_lang_Throwable;
    public "java/awt/event/InvocationEvent/when": Long;
    public static "java/awt/event/InvocationEvent/serialVersionUID": Long;
    public "<init>(Ljava/lang/Object;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InvocationEvent/<init>(Ljava/lang/Object;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;Ljava/lang/Runnable;Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InvocationEvent/<init>(Ljava/lang/Object;Ljava/lang/Runnable;Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;Ljava/lang/Runnable;Ljava/lang/Runnable;Z)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable, java_lang_Runnable, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InvocationEvent/<init>(Ljava/lang/Object;Ljava/lang/Runnable;Ljava/lang/Runnable;Z)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Runnable, java_lang_Runnable, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;ILjava/lang/Runnable;Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Runnable, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InvocationEvent/<init>(Ljava/lang/Object;ILjava/lang/Runnable;Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Runnable, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;ILjava/lang/Runnable;Ljava/lang/Object;Ljava/lang/Runnable;Z)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Runnable, java_lang_Object, java_lang_Runnable, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InvocationEvent/<init>(Ljava/lang/Object;ILjava/lang/Runnable;Ljava/lang/Object;Ljava/lang/Runnable;Z)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Runnable, java_lang_Object, java_lang_Runnable, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatch()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InvocationEvent/dispatch()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getException()Ljava/lang/Exception;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Exception) => void): void;
    public "java/awt/event/InvocationEvent/getException()Ljava/lang/Exception;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Exception) => void): void;
    public "getThrowable()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/awt/event/InvocationEvent/getThrowable()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "getWhen()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/awt/event/InvocationEvent/getWhen()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "isDispatched()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InvocationEvent/isDispatched()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finishedDispatching(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InvocationEvent/finishedDispatching(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/InvocationEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000(Ljava/awt/event/InvocationEvent;Z)V"(thread: JVMThread, args: [java_awt_event_InvocationEvent, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/event/InvocationEvent/access$000(Ljava/awt/event/InvocationEvent;Z)V"(thread: JVMThread, args: [java_awt_event_InvocationEvent, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/event/InvocationEvent/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_ActiveEvent extends java_lang_Object {
    "dispatch()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_misc_JavaSecurityAccess extends java_lang_Object {
    "doIntersectionPrivilege(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction, java_security_AccessControlContext, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "doIntersectionPrivilege(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface sun_awt_FwDispatcher extends java_lang_Object {
    "isDispatchThread()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "scheduleDispatch(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    "createSecondaryLoop()Ljava/awt/SecondaryLoop;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_SecondaryLoop) => void): void;
  }
  export class java_awt_EventDispatchThread extends java_lang_Thread {
    public static "java/awt/EventDispatchThread/eventLog": sun_util_logging_PlatformLogger;
    public "java/awt/EventDispatchThread/theQueue": java_awt_EventQueue;
    public "java/awt/EventDispatchThread/doDispatch": number;
    public static "java/awt/EventDispatchThread/ANY_EVENT": number;
    public "java/awt/EventDispatchThread/eventFilters": java_util_ArrayList;
    public "<init>(Ljava/lang/ThreadGroup;Ljava/lang/String;Ljava/awt/EventQueue;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_String, java_awt_EventQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/<init>(Ljava/lang/ThreadGroup;Ljava/lang/String;Ljava/awt/EventQueue;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_String, java_awt_EventQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "stopDispatching()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/stopDispatching()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "pumpEvents(Ljava/awt/Conditional;)V"(thread: JVMThread, args: [java_awt_Conditional], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/pumpEvents(Ljava/awt/Conditional;)V"(thread: JVMThread, args: [java_awt_Conditional], cb?: (e?: java_lang_Throwable) => void): void;
    public "pumpEventsForHierarchy(Ljava/awt/Conditional;Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Conditional, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/pumpEventsForHierarchy(Ljava/awt/Conditional;Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Conditional, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "pumpEvents(ILjava/awt/Conditional;)V"(thread: JVMThread, args: [number, java_awt_Conditional], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/pumpEvents(ILjava/awt/Conditional;)V"(thread: JVMThread, args: [number, java_awt_Conditional], cb?: (e?: java_lang_Throwable) => void): void;
    public "pumpEventsForHierarchy(ILjava/awt/Conditional;Ljava/awt/Component;)V"(thread: JVMThread, args: [number, java_awt_Conditional, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/pumpEventsForHierarchy(ILjava/awt/Conditional;Ljava/awt/Component;)V"(thread: JVMThread, args: [number, java_awt_Conditional, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "pumpEventsForFilter(Ljava/awt/Conditional;Ljava/awt/EventFilter;)V"(thread: JVMThread, args: [java_awt_Conditional, java_awt_EventFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/pumpEventsForFilter(Ljava/awt/Conditional;Ljava/awt/EventFilter;)V"(thread: JVMThread, args: [java_awt_Conditional, java_awt_EventFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "pumpEventsForFilter(ILjava/awt/Conditional;Ljava/awt/EventFilter;)V"(thread: JVMThread, args: [number, java_awt_Conditional, java_awt_EventFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/pumpEventsForFilter(ILjava/awt/Conditional;Ljava/awt/EventFilter;)V"(thread: JVMThread, args: [number, java_awt_Conditional, java_awt_EventFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "addEventFilter(Ljava/awt/EventFilter;)V"(thread: JVMThread, args: [java_awt_EventFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/addEventFilter(Ljava/awt/EventFilter;)V"(thread: JVMThread, args: [java_awt_EventFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeEventFilter(Ljava/awt/EventFilter;)V"(thread: JVMThread, args: [java_awt_EventFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/removeEventFilter(Ljava/awt/EventFilter;)V"(thread: JVMThread, args: [java_awt_EventFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "pumpOneEventForFilters(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/pumpOneEventForFilters(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "processException(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/processException(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEventQueue()Ljava/awt/EventQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventQueue) => void): void;
    public "java/awt/EventDispatchThread/getEventQueue()Ljava/awt/EventQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventQueue) => void): void;
    public "setEventQueue(Ljava/awt/EventQueue;)V"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/EventDispatchThread/setEventQueue(Ljava/awt/EventQueue;)V"(thread: JVMThread, args: [java_awt_EventQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/EventDispatchThread/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Queue extends java_lang_Object {
    public "java/awt/Queue/head": sun_awt_EventQueueItem;
    public "java/awt/Queue/tail": sun_awt_EventQueueItem;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Queue/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_datatransfer_Clipboard extends java_lang_Object {
    public "java/awt/datatransfer/Clipboard/name": java_lang_String;
    public "java/awt/datatransfer/Clipboard/owner": java_awt_datatransfer_ClipboardOwner;
    public "java/awt/datatransfer/Clipboard/contents": java_awt_datatransfer_Transferable;
    public "java/awt/datatransfer/Clipboard/flavorListeners": sun_awt_EventListenerAggregate;
    public "java/awt/datatransfer/Clipboard/currentDataFlavors": java_util_Set;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/Clipboard/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/datatransfer/Clipboard/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setContents(Ljava/awt/datatransfer/Transferable;Ljava/awt/datatransfer/ClipboardOwner;)V"(thread: JVMThread, args: [java_awt_datatransfer_Transferable, java_awt_datatransfer_ClipboardOwner], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/Clipboard/setContents(Ljava/awt/datatransfer/Transferable;Ljava/awt/datatransfer/ClipboardOwner;)V"(thread: JVMThread, args: [java_awt_datatransfer_Transferable, java_awt_datatransfer_ClipboardOwner], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContents(Ljava/lang/Object;)Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "java/awt/datatransfer/Clipboard/getContents(Ljava/lang/Object;)Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "getAvailableDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "java/awt/datatransfer/Clipboard/getAvailableDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "isDataFlavorAvailable(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/datatransfer/Clipboard/isDataFlavorAvailable(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getData(Ljava/awt/datatransfer/DataFlavor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/datatransfer/Clipboard/getData(Ljava/awt/datatransfer/DataFlavor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "addFlavorListener(Ljava/awt/datatransfer/FlavorListener;)V"(thread: JVMThread, args: [java_awt_datatransfer_FlavorListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/Clipboard/addFlavorListener(Ljava/awt/datatransfer/FlavorListener;)V"(thread: JVMThread, args: [java_awt_datatransfer_FlavorListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeFlavorListener(Ljava/awt/datatransfer/FlavorListener;)V"(thread: JVMThread, args: [java_awt_datatransfer_FlavorListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/Clipboard/removeFlavorListener(Ljava/awt/datatransfer/FlavorListener;)V"(thread: JVMThread, args: [java_awt_datatransfer_FlavorListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFlavorListeners()[Ljava/awt/datatransfer/FlavorListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_FlavorListener>) => void): void;
    public "java/awt/datatransfer/Clipboard/getFlavorListeners()[Ljava/awt/datatransfer/FlavorListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_FlavorListener>) => void): void;
    public "fireFlavorsChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/Clipboard/fireFlavorsChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAvailableDataFlavorSet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/awt/datatransfer/Clipboard/getAvailableDataFlavorSet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
  }
  export interface java_awt_datatransfer_FlavorListener extends java_lang_Object, java_util_EventListener {
    "flavorsChanged(Ljava/awt/datatransfer/FlavorEvent;)V"(thread: JVMThread, args: [java_awt_datatransfer_FlavorEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_datatransfer_FlavorEvent extends java_util_EventObject {
    public "<init>(Ljava/awt/datatransfer/Clipboard;)V"(thread: JVMThread, args: [java_awt_datatransfer_Clipboard], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/datatransfer/FlavorEvent/<init>(Ljava/awt/datatransfer/Clipboard;)V"(thread: JVMThread, args: [java_awt_datatransfer_Clipboard], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_EventListenerAggregate extends java_lang_Object {
    public "sun/awt/EventListenerAggregate/listenerList": JVMArray<java_util_EventListener>;
    public "<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/EventListenerAggregate/<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "getListenerClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/awt/EventListenerAggregate/getListenerClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "add(Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/EventListenerAggregate/add(Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/util/EventListener;)Z"(thread: JVMThread, args: [java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/EventListenerAggregate/remove(Ljava/util/EventListener;)Z"(thread: JVMThread, args: [java_util_EventListener], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getListenersInternal()[Ljava/util/EventListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "sun/awt/EventListenerAggregate/getListenersInternal()[Ljava/util/EventListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "getListenersCopy()[Ljava/util/EventListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "sun/awt/EventListenerAggregate/getListenersCopy()[Ljava/util/EventListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/EventListenerAggregate/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/EventListenerAggregate/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_datatransfer_ClipboardOwner extends java_lang_Object {
    "lostOwnership(Ljava/awt/datatransfer/Clipboard;Ljava/awt/datatransfer/Transferable;)V"(thread: JVMThread, args: [java_awt_datatransfer_Clipboard, java_awt_datatransfer_Transferable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_PageAttributes extends java_lang_Object implements java_lang_Cloneable {
    public "java/awt/PageAttributes/color": java_awt_PageAttributes$ColorType;
    public "java/awt/PageAttributes/media": java_awt_PageAttributes$MediaType;
    public "java/awt/PageAttributes/orientationRequested": java_awt_PageAttributes$OrientationRequestedType;
    public "java/awt/PageAttributes/origin": java_awt_PageAttributes$OriginType;
    public "java/awt/PageAttributes/printQuality": java_awt_PageAttributes$PrintQualityType;
    public "java/awt/PageAttributes/printerResolution": JVMArray<number>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/PageAttributes;)V"(thread: JVMThread, args: [java_awt_PageAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/<init>(Ljava/awt/PageAttributes;)V"(thread: JVMThread, args: [java_awt_PageAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/PageAttributes$ColorType;Ljava/awt/PageAttributes$MediaType;Ljava/awt/PageAttributes$OrientationRequestedType;Ljava/awt/PageAttributes$OriginType;Ljava/awt/PageAttributes$PrintQualityType;[I)V"(thread: JVMThread, args: [java_awt_PageAttributes$ColorType, java_awt_PageAttributes$MediaType, java_awt_PageAttributes$OrientationRequestedType, java_awt_PageAttributes$OriginType, java_awt_PageAttributes$PrintQualityType, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/<init>(Ljava/awt/PageAttributes$ColorType;Ljava/awt/PageAttributes$MediaType;Ljava/awt/PageAttributes$OrientationRequestedType;Ljava/awt/PageAttributes$OriginType;Ljava/awt/PageAttributes$PrintQualityType;[I)V"(thread: JVMThread, args: [java_awt_PageAttributes$ColorType, java_awt_PageAttributes$MediaType, java_awt_PageAttributes$OrientationRequestedType, java_awt_PageAttributes$OriginType, java_awt_PageAttributes$PrintQualityType, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/PageAttributes/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(Ljava/awt/PageAttributes;)V"(thread: JVMThread, args: [java_awt_PageAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/set(Ljava/awt/PageAttributes;)V"(thread: JVMThread, args: [java_awt_PageAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "getColor()Ljava/awt/PageAttributes$ColorType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$ColorType) => void): void;
    public "java/awt/PageAttributes/getColor()Ljava/awt/PageAttributes$ColorType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$ColorType) => void): void;
    public "setColor(Ljava/awt/PageAttributes$ColorType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$ColorType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setColor(Ljava/awt/PageAttributes$ColorType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$ColorType], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMedia()Ljava/awt/PageAttributes$MediaType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$MediaType) => void): void;
    public "java/awt/PageAttributes/getMedia()Ljava/awt/PageAttributes$MediaType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$MediaType) => void): void;
    public "setMedia(Ljava/awt/PageAttributes$MediaType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$MediaType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setMedia(Ljava/awt/PageAttributes$MediaType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$MediaType], cb?: (e?: java_lang_Throwable) => void): void;
    public "setMediaToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setMediaToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOrientationRequested()Ljava/awt/PageAttributes$OrientationRequestedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$OrientationRequestedType) => void): void;
    public "java/awt/PageAttributes/getOrientationRequested()Ljava/awt/PageAttributes$OrientationRequestedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$OrientationRequestedType) => void): void;
    public "setOrientationRequested(Ljava/awt/PageAttributes$OrientationRequestedType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$OrientationRequestedType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setOrientationRequested(Ljava/awt/PageAttributes$OrientationRequestedType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$OrientationRequestedType], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOrientationRequested(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setOrientationRequested(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOrientationRequestedToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setOrientationRequestedToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOrigin()Ljava/awt/PageAttributes$OriginType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$OriginType) => void): void;
    public "java/awt/PageAttributes/getOrigin()Ljava/awt/PageAttributes$OriginType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$OriginType) => void): void;
    public "setOrigin(Ljava/awt/PageAttributes$OriginType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$OriginType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setOrigin(Ljava/awt/PageAttributes$OriginType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$OriginType], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPrintQuality()Ljava/awt/PageAttributes$PrintQualityType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$PrintQualityType) => void): void;
    public "java/awt/PageAttributes/getPrintQuality()Ljava/awt/PageAttributes$PrintQualityType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_PageAttributes$PrintQualityType) => void): void;
    public "setPrintQuality(Ljava/awt/PageAttributes$PrintQualityType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$PrintQualityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setPrintQuality(Ljava/awt/PageAttributes$PrintQualityType;)V"(thread: JVMThread, args: [java_awt_PageAttributes$PrintQualityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPrintQuality(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setPrintQuality(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPrintQualityToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setPrintQualityToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPrinterResolution()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/PageAttributes/getPrinterResolution()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "setPrinterResolution([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setPrinterResolution([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPrinterResolution(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setPrinterResolution(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPrinterResolutionToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes/setPrinterResolutionToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/PageAttributes/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/PageAttributes/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/PageAttributes/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_PageAttributes$PrintQualityType extends java_awt_AttributeValue {
    public static "java/awt/PageAttributes$PrintQualityType/I_HIGH": number;
    public static "java/awt/PageAttributes$PrintQualityType/I_NORMAL": number;
    public static "java/awt/PageAttributes$PrintQualityType/I_DRAFT": number;
    public static "java/awt/PageAttributes$PrintQualityType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/PageAttributes$PrintQualityType/HIGH": java_awt_PageAttributes$PrintQualityType;
    public static "java/awt/PageAttributes$PrintQualityType/NORMAL": java_awt_PageAttributes$PrintQualityType;
    public static "java/awt/PageAttributes$PrintQualityType/DRAFT": java_awt_PageAttributes$PrintQualityType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes$PrintQualityType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/PageAttributes$PrintQualityType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/PageAttributes$PrintQualityType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/PageAttributes$PrintQualityType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_PageAttributes$OriginType extends java_awt_AttributeValue {
    public static "java/awt/PageAttributes$OriginType/I_PHYSICAL": number;
    public static "java/awt/PageAttributes$OriginType/I_PRINTABLE": number;
    public static "java/awt/PageAttributes$OriginType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/PageAttributes$OriginType/PHYSICAL": java_awt_PageAttributes$OriginType;
    public static "java/awt/PageAttributes$OriginType/PRINTABLE": java_awt_PageAttributes$OriginType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes$OriginType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/PageAttributes$OriginType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/PageAttributes$OriginType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/PageAttributes$OriginType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_PageAttributes$OrientationRequestedType extends java_awt_AttributeValue {
    public static "java/awt/PageAttributes$OrientationRequestedType/I_PORTRAIT": number;
    public static "java/awt/PageAttributes$OrientationRequestedType/I_LANDSCAPE": number;
    public static "java/awt/PageAttributes$OrientationRequestedType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/PageAttributes$OrientationRequestedType/PORTRAIT": java_awt_PageAttributes$OrientationRequestedType;
    public static "java/awt/PageAttributes$OrientationRequestedType/LANDSCAPE": java_awt_PageAttributes$OrientationRequestedType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes$OrientationRequestedType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/PageAttributes$OrientationRequestedType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/PageAttributes$OrientationRequestedType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/PageAttributes$OrientationRequestedType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_PageAttributes$MediaType extends java_awt_AttributeValue {
    public static "java/awt/PageAttributes$MediaType/I_ISO_4A0": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_2A0": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A0": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A1": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A2": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A3": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A4": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A5": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A6": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A7": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A8": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A9": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_A10": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B0": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B1": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B2": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B3": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B4": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B5": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B6": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B7": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B8": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B9": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_B10": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B0": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B1": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B2": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B3": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B4": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B5": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B6": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B7": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B8": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B9": number;
    public static "java/awt/PageAttributes$MediaType/I_JIS_B10": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C0": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C1": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C2": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C3": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C4": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C5": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C6": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C7": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C8": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C9": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_C10": number;
    public static "java/awt/PageAttributes$MediaType/I_ISO_DESIGNATED_LONG": number;
    public static "java/awt/PageAttributes$MediaType/I_EXECUTIVE": number;
    public static "java/awt/PageAttributes$MediaType/I_FOLIO": number;
    public static "java/awt/PageAttributes$MediaType/I_INVOICE": number;
    public static "java/awt/PageAttributes$MediaType/I_LEDGER": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_LETTER": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_LEGAL": number;
    public static "java/awt/PageAttributes$MediaType/I_QUARTO": number;
    public static "java/awt/PageAttributes$MediaType/I_A": number;
    public static "java/awt/PageAttributes$MediaType/I_B": number;
    public static "java/awt/PageAttributes$MediaType/I_C": number;
    public static "java/awt/PageAttributes$MediaType/I_D": number;
    public static "java/awt/PageAttributes$MediaType/I_E": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_10X15_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_10X14_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_10X13_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_9X12_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_9X11_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_7X9_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_6X9_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_NUMBER_9_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_NUMBER_10_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_NUMBER_11_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_NUMBER_12_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_NA_NUMBER_14_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_INVITE_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_ITALY_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_MONARCH_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/I_PERSONAL_ENVELOPE": number;
    public static "java/awt/PageAttributes$MediaType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/PageAttributes$MediaType/ISO_4A0": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_2A0": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A0": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A1": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A2": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A3": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A4": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A5": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A6": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A7": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A8": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_A10": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B0": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B1": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B2": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B3": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B4": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B5": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B6": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B7": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B8": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B10": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B0": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B1": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B2": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B3": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B4": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B5": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B6": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B7": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B8": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/JIS_B10": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C0": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C1": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C2": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C3": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C4": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C5": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C6": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C7": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C8": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C10": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_DESIGNATED_LONG": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/EXECUTIVE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/FOLIO": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/INVOICE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/LEDGER": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_LETTER": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_LEGAL": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/QUARTO": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/D": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/E": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_10X15_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_10X14_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_10X13_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_9X12_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_9X11_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_7X9_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_6X9_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_NUMBER_9_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_NUMBER_10_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_NUMBER_11_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_NUMBER_12_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NA_NUMBER_14_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/INVITE_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ITALY_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/MONARCH_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/PERSONAL_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A0": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A1": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A2": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A3": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A4": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A5": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A6": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A7": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A8": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/A10": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B0": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B1": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B2": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B3": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B4": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B4_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B5": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_B5_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B6": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B7": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B8": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/B10": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C0": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C0_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C1": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C1_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C2": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C2_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C3": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C3_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C4": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C4_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C5": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C5_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C6": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C6_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C7": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C7_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C8": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C8_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C9_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/C10": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_C10_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ISO_DESIGNATED_LONG_ENVELOPE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/STATEMENT": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/TABLOID": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/LETTER": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/NOTE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/LEGAL": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_10X15": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_10X14": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_10X13": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_9X12": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_9X11": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_7X9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_6X9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_9": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_10": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_11": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_12": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_14": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_INVITE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_ITALY": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_MONARCH": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ENV_PERSONAL": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/INVITE": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/ITALY": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/MONARCH": java_awt_PageAttributes$MediaType;
    public static "java/awt/PageAttributes$MediaType/PERSONAL": java_awt_PageAttributes$MediaType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes$MediaType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/PageAttributes$MediaType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/PageAttributes$MediaType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/PageAttributes$MediaType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_PageAttributes$ColorType extends java_awt_AttributeValue {
    public static "java/awt/PageAttributes$ColorType/I_COLOR": number;
    public static "java/awt/PageAttributes$ColorType/I_MONOCHROME": number;
    public static "java/awt/PageAttributes$ColorType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/PageAttributes$ColorType/COLOR": java_awt_PageAttributes$ColorType;
    public static "java/awt/PageAttributes$ColorType/MONOCHROME": java_awt_PageAttributes$ColorType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PageAttributes$ColorType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/PageAttributes$ColorType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/PageAttributes$ColorType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/PageAttributes$ColorType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_JobAttributes extends java_lang_Object implements java_lang_Cloneable {
    public "java/awt/JobAttributes/copies": number;
    public "java/awt/JobAttributes/defaultSelection": java_awt_JobAttributes$DefaultSelectionType;
    public "java/awt/JobAttributes/destination": java_awt_JobAttributes$DestinationType;
    public "java/awt/JobAttributes/dialog": java_awt_JobAttributes$DialogType;
    public "java/awt/JobAttributes/fileName": java_lang_String;
    public "java/awt/JobAttributes/fromPage": number;
    public "java/awt/JobAttributes/maxPage": number;
    public "java/awt/JobAttributes/minPage": number;
    public "java/awt/JobAttributes/multipleDocumentHandling": java_awt_JobAttributes$MultipleDocumentHandlingType;
    public "java/awt/JobAttributes/pageRanges": JVMArray<JVMArray<number>>;
    public "java/awt/JobAttributes/prFirst": number;
    public "java/awt/JobAttributes/prLast": number;
    public "java/awt/JobAttributes/printer": java_lang_String;
    public "java/awt/JobAttributes/sides": java_awt_JobAttributes$SidesType;
    public "java/awt/JobAttributes/toPage": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/JobAttributes;)V"(thread: JVMThread, args: [java_awt_JobAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/<init>(Ljava/awt/JobAttributes;)V"(thread: JVMThread, args: [java_awt_JobAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/awt/JobAttributes$DefaultSelectionType;Ljava/awt/JobAttributes$DestinationType;Ljava/awt/JobAttributes$DialogType;Ljava/lang/String;IILjava/awt/JobAttributes$MultipleDocumentHandlingType;[[ILjava/lang/String;Ljava/awt/JobAttributes$SidesType;)V"(thread: JVMThread, args: [number, java_awt_JobAttributes$DefaultSelectionType, java_awt_JobAttributes$DestinationType, java_awt_JobAttributes$DialogType, java_lang_String, number, number, java_awt_JobAttributes$MultipleDocumentHandlingType, JVMArray<JVMArray<number>>, java_lang_String, java_awt_JobAttributes$SidesType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/<init>(ILjava/awt/JobAttributes$DefaultSelectionType;Ljava/awt/JobAttributes$DestinationType;Ljava/awt/JobAttributes$DialogType;Ljava/lang/String;IILjava/awt/JobAttributes$MultipleDocumentHandlingType;[[ILjava/lang/String;Ljava/awt/JobAttributes$SidesType;)V"(thread: JVMThread, args: [number, java_awt_JobAttributes$DefaultSelectionType, java_awt_JobAttributes$DestinationType, java_awt_JobAttributes$DialogType, java_lang_String, number, number, java_awt_JobAttributes$MultipleDocumentHandlingType, JVMArray<JVMArray<number>>, java_lang_String, java_awt_JobAttributes$SidesType], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/JobAttributes/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(Ljava/awt/JobAttributes;)V"(thread: JVMThread, args: [java_awt_JobAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/set(Ljava/awt/JobAttributes;)V"(thread: JVMThread, args: [java_awt_JobAttributes], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCopies()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes/getCopies()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setCopies(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setCopies(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCopiesToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setCopiesToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDefaultSelection()Ljava/awt/JobAttributes$DefaultSelectionType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$DefaultSelectionType) => void): void;
    public "java/awt/JobAttributes/getDefaultSelection()Ljava/awt/JobAttributes$DefaultSelectionType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$DefaultSelectionType) => void): void;
    public "setDefaultSelection(Ljava/awt/JobAttributes$DefaultSelectionType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$DefaultSelectionType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setDefaultSelection(Ljava/awt/JobAttributes$DefaultSelectionType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$DefaultSelectionType], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDestination()Ljava/awt/JobAttributes$DestinationType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$DestinationType) => void): void;
    public "java/awt/JobAttributes/getDestination()Ljava/awt/JobAttributes$DestinationType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$DestinationType) => void): void;
    public "setDestination(Ljava/awt/JobAttributes$DestinationType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$DestinationType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setDestination(Ljava/awt/JobAttributes$DestinationType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$DestinationType], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDialog()Ljava/awt/JobAttributes$DialogType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$DialogType) => void): void;
    public "java/awt/JobAttributes/getDialog()Ljava/awt/JobAttributes$DialogType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$DialogType) => void): void;
    public "setDialog(Ljava/awt/JobAttributes$DialogType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$DialogType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setDialog(Ljava/awt/JobAttributes$DialogType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$DialogType], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFileName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/JobAttributes/getFileName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setFileName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setFileName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFromPage()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes/getFromPage()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFromPage(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setFromPage(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMaxPage()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes/getMaxPage()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMaxPage(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setMaxPage(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMinPage()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes/getMinPage()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMinPage(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setMinPage(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMultipleDocumentHandling()Ljava/awt/JobAttributes$MultipleDocumentHandlingType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$MultipleDocumentHandlingType) => void): void;
    public "java/awt/JobAttributes/getMultipleDocumentHandling()Ljava/awt/JobAttributes$MultipleDocumentHandlingType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$MultipleDocumentHandlingType) => void): void;
    public "setMultipleDocumentHandling(Ljava/awt/JobAttributes$MultipleDocumentHandlingType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$MultipleDocumentHandlingType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setMultipleDocumentHandling(Ljava/awt/JobAttributes$MultipleDocumentHandlingType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$MultipleDocumentHandlingType], cb?: (e?: java_lang_Throwable) => void): void;
    public "setMultipleDocumentHandlingToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setMultipleDocumentHandlingToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPageRanges()[[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<number>>) => void): void;
    public "java/awt/JobAttributes/getPageRanges()[[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<number>>) => void): void;
    public "setPageRanges([[I)V"(thread: JVMThread, args: [JVMArray<JVMArray<number>>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setPageRanges([[I)V"(thread: JVMThread, args: [JVMArray<JVMArray<number>>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPrinter()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/JobAttributes/getPrinter()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setPrinter(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setPrinter(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSides()Ljava/awt/JobAttributes$SidesType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$SidesType) => void): void;
    public "java/awt/JobAttributes/getSides()Ljava/awt/JobAttributes$SidesType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_JobAttributes$SidesType) => void): void;
    public "setSides(Ljava/awt/JobAttributes$SidesType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$SidesType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setSides(Ljava/awt/JobAttributes$SidesType;)V"(thread: JVMThread, args: [java_awt_JobAttributes$SidesType], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSidesToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setSidesToDefault()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getToPage()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes/getToPage()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setToPage(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes/setToPage(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/JobAttributes/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_JobAttributes$SidesType extends java_awt_AttributeValue {
    public static "java/awt/JobAttributes$SidesType/I_ONE_SIDED": number;
    public static "java/awt/JobAttributes$SidesType/I_TWO_SIDED_LONG_EDGE": number;
    public static "java/awt/JobAttributes$SidesType/I_TWO_SIDED_SHORT_EDGE": number;
    public static "java/awt/JobAttributes$SidesType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/JobAttributes$SidesType/ONE_SIDED": java_awt_JobAttributes$SidesType;
    public static "java/awt/JobAttributes$SidesType/TWO_SIDED_LONG_EDGE": java_awt_JobAttributes$SidesType;
    public static "java/awt/JobAttributes$SidesType/TWO_SIDED_SHORT_EDGE": java_awt_JobAttributes$SidesType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes$SidesType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/JobAttributes$SidesType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes$SidesType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/JobAttributes$SidesType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_JobAttributes$MultipleDocumentHandlingType extends java_awt_AttributeValue {
    public static "java/awt/JobAttributes$MultipleDocumentHandlingType/I_SEPARATE_DOCUMENTS_COLLATED_COPIES": number;
    public static "java/awt/JobAttributes$MultipleDocumentHandlingType/I_SEPARATE_DOCUMENTS_UNCOLLATED_COPIES": number;
    public static "java/awt/JobAttributes$MultipleDocumentHandlingType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/JobAttributes$MultipleDocumentHandlingType/SEPARATE_DOCUMENTS_COLLATED_COPIES": java_awt_JobAttributes$MultipleDocumentHandlingType;
    public static "java/awt/JobAttributes$MultipleDocumentHandlingType/SEPARATE_DOCUMENTS_UNCOLLATED_COPIES": java_awt_JobAttributes$MultipleDocumentHandlingType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes$MultipleDocumentHandlingType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/JobAttributes$MultipleDocumentHandlingType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes$MultipleDocumentHandlingType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/JobAttributes$MultipleDocumentHandlingType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_JobAttributes$DialogType extends java_awt_AttributeValue {
    public static "java/awt/JobAttributes$DialogType/I_COMMON": number;
    public static "java/awt/JobAttributes$DialogType/I_NATIVE": number;
    public static "java/awt/JobAttributes$DialogType/I_NONE": number;
    public static "java/awt/JobAttributes$DialogType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/JobAttributes$DialogType/COMMON": java_awt_JobAttributes$DialogType;
    public static "java/awt/JobAttributes$DialogType/NATIVE": java_awt_JobAttributes$DialogType;
    public static "java/awt/JobAttributes$DialogType/NONE": java_awt_JobAttributes$DialogType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes$DialogType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/JobAttributes$DialogType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes$DialogType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/JobAttributes$DialogType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_JobAttributes$DestinationType extends java_awt_AttributeValue {
    public static "java/awt/JobAttributes$DestinationType/I_FILE": number;
    public static "java/awt/JobAttributes$DestinationType/I_PRINTER": number;
    public static "java/awt/JobAttributes$DestinationType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/JobAttributes$DestinationType/FILE": java_awt_JobAttributes$DestinationType;
    public static "java/awt/JobAttributes$DestinationType/PRINTER": java_awt_JobAttributes$DestinationType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes$DestinationType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/JobAttributes$DestinationType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes$DestinationType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/JobAttributes$DestinationType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_JobAttributes$DefaultSelectionType extends java_awt_AttributeValue {
    public static "java/awt/JobAttributes$DefaultSelectionType/I_ALL": number;
    public static "java/awt/JobAttributes$DefaultSelectionType/I_RANGE": number;
    public static "java/awt/JobAttributes$DefaultSelectionType/I_SELECTION": number;
    public static "java/awt/JobAttributes$DefaultSelectionType/NAMES": JVMArray<java_lang_String>;
    public static "java/awt/JobAttributes$DefaultSelectionType/ALL": java_awt_JobAttributes$DefaultSelectionType;
    public static "java/awt/JobAttributes$DefaultSelectionType/RANGE": java_awt_JobAttributes$DefaultSelectionType;
    public static "java/awt/JobAttributes$DefaultSelectionType/SELECTION": java_awt_JobAttributes$DefaultSelectionType;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/JobAttributes$DefaultSelectionType/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/JobAttributes$DefaultSelectionType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/JobAttributes$DefaultSelectionType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/JobAttributes$DefaultSelectionType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_PrintJob extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PrintJob/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "java/awt/PrintJob/getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "getPageDimension()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/PrintJob/getPageDimension()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getPageResolution()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/PrintJob/getPageResolution()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastPageFirst()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/PrintJob/lastPageFirst()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PrintJob/end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PrintJob/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_MouseInfoPeer extends java_lang_Object {
    "fillPointWithCoords(Ljava/awt/Point;)I"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isWindowUnderMouse(Ljava/awt/Window;)Z"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_CheckboxMenuItem extends java_awt_MenuItem implements java_awt_ItemSelectable, javax_accessibility_Accessible {
    public "java/awt/CheckboxMenuItem/state": number;
    public "java/awt/CheckboxMenuItem/itemListener": java_awt_event_ItemListener;
    public static "java/awt/CheckboxMenuItem/base": java_lang_String;
    public static "java/awt/CheckboxMenuItem/nameCounter": number;
    public static "java/awt/CheckboxMenuItem/serialVersionUID": Long;
    public "java/awt/CheckboxMenuItem/checkboxMenuItemSerializedDataVersion": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/CheckboxMenuItem/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getState()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/CheckboxMenuItem/getState()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setState(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/setState(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSelectedObjects()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/awt/CheckboxMenuItem/getSelectedObjects()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "addItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/addItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/removeItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getItemListeners()[Ljava/awt/event/ItemListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ItemListener>) => void): void;
    public "java/awt/CheckboxMenuItem/getItemListeners()[Ljava/awt/event/ItemListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ItemListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/CheckboxMenuItem/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/CheckboxMenuItem/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processItemEvent(Ljava/awt/event/ItemEvent;)V"(thread: JVMThread, args: [java_awt_event_ItemEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/processItemEvent(Ljava/awt/event/ItemEvent;)V"(thread: JVMThread, args: [java_awt_event_ItemEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "doMenuEvent(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/doMenuEvent(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/CheckboxMenuItem/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxMenuItem/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/CheckboxMenuItem/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/CheckboxMenuItem/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/CheckboxMenuItem/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_ItemEvent extends java_awt_AWTEvent {
    public static "java/awt/event/ItemEvent/ITEM_FIRST": number;
    public static "java/awt/event/ItemEvent/ITEM_LAST": number;
    public static "java/awt/event/ItemEvent/ITEM_STATE_CHANGED": number;
    public static "java/awt/event/ItemEvent/SELECTED": number;
    public static "java/awt/event/ItemEvent/DESELECTED": number;
    public "java/awt/event/ItemEvent/item": java_lang_Object;
    public "java/awt/event/ItemEvent/stateChange": number;
    public static "java/awt/event/ItemEvent/serialVersionUID": Long;
    public "<init>(Ljava/awt/ItemSelectable;ILjava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_ItemSelectable, number, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/ItemEvent/<init>(Ljava/awt/ItemSelectable;ILjava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_ItemSelectable, number, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getItemSelectable()Ljava/awt/ItemSelectable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ItemSelectable) => void): void;
    public "java/awt/event/ItemEvent/getItemSelectable()Ljava/awt/ItemSelectable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ItemSelectable) => void): void;
    public "getItem()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/event/ItemEvent/getItem()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getStateChange()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/ItemEvent/getStateChange()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/ItemEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_awt_event_ItemListener extends java_lang_Object, java_util_EventListener {
    "itemStateChanged(Ljava/awt/event/ItemEvent;)V"(thread: JVMThread, args: [java_awt_event_ItemEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_ItemSelectable extends java_lang_Object {
    "getSelectedObjects()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    "addItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    "removeItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_CheckboxMenuItemPeer extends java_lang_Object, java_awt_peer_MenuItemPeer {
    "setState(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_FileDialog extends java_awt_Dialog {
    public static "java/awt/FileDialog/LOAD": number;
    public static "java/awt/FileDialog/SAVE": number;
    public "java/awt/FileDialog/mode": number;
    public "java/awt/FileDialog/dir": java_lang_String;
    public "java/awt/FileDialog/file": java_lang_String;
    public "java/awt/FileDialog/files": JVMArray<java_io_File>;
    public "java/awt/FileDialog/multipleMode": number;
    public "java/awt/FileDialog/filter": java_io_FilenameFilter;
    public static "java/awt/FileDialog/base": java_lang_String;
    public static "java/awt/FileDialog/nameCounter": number;
    public static "java/awt/FileDialog/serialVersionUID": Long;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/FileDialog/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Frame;)V"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/<init>(Ljava/awt/Frame;)V"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Frame;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/<init>(Ljava/awt/Frame;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Frame;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/<init>(Ljava/awt/Frame;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Dialog;)V"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/<init>(Ljava/awt/Dialog;)V"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Dialog;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/<init>(Ljava/awt/Dialog;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Dialog;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/<init>(Ljava/awt/Dialog;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/FileDialog/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FileDialog/getMode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMode(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/setMode(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDirectory()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/FileDialog/getDirectory()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setDirectory(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/setDirectory(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/FileDialog/getFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFiles()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "java/awt/FileDialog/getFiles()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "setFiles([Ljava/io/File;)V"(thread: JVMThread, args: [JVMArray<java_io_File>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/setFiles([Ljava/io/File;)V"(thread: JVMThread, args: [JVMArray<java_io_File>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFile(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/setFile(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setMultipleMode(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/setMultipleMode(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isMultipleMode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FileDialog/isMultipleMode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFilenameFilter()Ljava/io/FilenameFilter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FilenameFilter) => void): void;
    public "java/awt/FileDialog/getFilenameFilter()Ljava/io/FilenameFilter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FilenameFilter) => void): void;
    public "setFilenameFilter(Ljava/io/FilenameFilter;)V"(thread: JVMThread, args: [java_io_FilenameFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/setFilenameFilter(Ljava/io/FilenameFilter;)V"(thread: JVMThread, args: [java_io_FilenameFilter], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FileDialog/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/FileDialog/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "postsOldMouseEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/FileDialog/postsOldMouseEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/awt/FileDialog;[Ljava/io/File;)V"(thread: JVMThread, args: [java_awt_FileDialog, JVMArray<java_io_File>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/FileDialog/access$000(Ljava/awt/FileDialog;[Ljava/io/File;)V"(thread: JVMThread, args: [java_awt_FileDialog, JVMArray<java_io_File>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$100(Ljava/awt/FileDialog;)Z"(thread: JVMThread, args: [java_awt_FileDialog], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/FileDialog/access$100(Ljava/awt/FileDialog;)Z"(thread: JVMThread, args: [java_awt_FileDialog], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/FileDialog/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_io_FilenameFilter extends java_lang_Object {
    "accept(Ljava/io/File;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_peer_FileDialogPeer extends java_lang_Object, java_awt_peer_DialogPeer {
    "setFile(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setDirectory(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setFilenameFilter(Ljava/io/FilenameFilter;)V"(thread: JVMThread, args: [java_io_FilenameFilter], cb?: (e?: java_lang_Throwable) => void): void;
    "setTitle(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setResizable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "blockWindows(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    "toFront()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "toBack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateAlwaysOnTopState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateFocusableWindowState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setModalBlocked(Ljava/awt/Dialog;Z)V"(thread: JVMThread, args: [java_awt_Dialog, number], cb?: (e?: java_lang_Throwable) => void): void;
    "updateMinimumSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateIconImages()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setOpaque(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "updateWindow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "repositionSecurityWarning()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    "beginValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "beginLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_MenuItem extends java_awt_MenuComponent implements javax_accessibility_Accessible {
    public "java/awt/MenuItem/enabled": number;
    public "java/awt/MenuItem/label": java_lang_String;
    public "java/awt/MenuItem/actionCommand": java_lang_String;
    public "java/awt/MenuItem/eventMask": Long;
    public "java/awt/MenuItem/actionListener": java_awt_event_ActionListener;
    public "java/awt/MenuItem/shortcut": java_awt_MenuShortcut;
    public static "java/awt/MenuItem/base": java_lang_String;
    public static "java/awt/MenuItem/nameCounter": number;
    public static "java/awt/MenuItem/serialVersionUID": Long;
    public "java/awt/MenuItem/menuItemSerializedDataVersion": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_lang_String, java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/<init>(Ljava/lang/String;Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_lang_String, java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuItem/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLabel()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuItem/getLabel()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuItem/isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "enable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/enable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "enable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/enable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "disable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/disable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getShortcut()Ljava/awt/MenuShortcut;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuShortcut) => void): void;
    public "java/awt/MenuItem/getShortcut()Ljava/awt/MenuShortcut;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuShortcut) => void): void;
    public "setShortcut(Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/setShortcut(Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "deleteShortcut()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/deleteShortcut()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "deleteShortcut(Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/deleteShortcut(Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "doMenuEvent(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/doMenuEvent(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isItemEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuItem/isItemEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "handleShortcut(Ljava/awt/event/KeyEvent;)Z"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuItem/handleShortcut(Ljava/awt/event/KeyEvent;)Z"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getShortcutMenuItem(Ljava/awt/MenuShortcut;)Ljava/awt/MenuItem;"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "java/awt/MenuItem/getShortcutMenuItem(Ljava/awt/MenuShortcut;)Ljava/awt/MenuItem;"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "enableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/enableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "disableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/disableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setActionCommand(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/setActionCommand(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getActionCommand()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuItem/getActionCommand()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getActionCommandImpl()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuItem/getActionCommandImpl()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "java/awt/MenuItem/getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/MenuItem/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuItem/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processActionEvent(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/processActionEvent(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuItem/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuItem/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/MenuItem/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/MenuItem/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "access$000(Ljava/awt/MenuItem;)Ljava/awt/MenuShortcut;"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuShortcut) => void): void;
    public static "java/awt/MenuItem/access$000(Ljava/awt/MenuItem;)Ljava/awt/MenuShortcut;"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuShortcut) => void): void;
    public static "access$100(Ljava/awt/MenuItem;)Z"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/MenuItem/access$100(Ljava/awt/MenuItem;)Z"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/MenuItem/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_ActionEvent extends java_awt_AWTEvent {
    public static "java/awt/event/ActionEvent/SHIFT_MASK": number;
    public static "java/awt/event/ActionEvent/CTRL_MASK": number;
    public static "java/awt/event/ActionEvent/META_MASK": number;
    public static "java/awt/event/ActionEvent/ALT_MASK": number;
    public static "java/awt/event/ActionEvent/ACTION_FIRST": number;
    public static "java/awt/event/ActionEvent/ACTION_LAST": number;
    public static "java/awt/event/ActionEvent/ACTION_PERFORMED": number;
    public "java/awt/event/ActionEvent/actionCommand": java_lang_String;
    public "java/awt/event/ActionEvent/when": Long;
    public "java/awt/event/ActionEvent/modifiers": number;
    public static "java/awt/event/ActionEvent/serialVersionUID": Long;
    public "<init>(Ljava/lang/Object;ILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/ActionEvent/<init>(Ljava/lang/Object;ILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;ILjava/lang/String;I)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/ActionEvent/<init>(Ljava/lang/Object;ILjava/lang/String;I)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;ILjava/lang/String;JI)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_String, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/ActionEvent/<init>(Ljava/lang/Object;ILjava/lang/String;JI)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_String, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getActionCommand()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/ActionEvent/getActionCommand()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getWhen()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/awt/event/ActionEvent/getWhen()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/ActionEvent/getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/ActionEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_MenuShortcut extends java_lang_Object implements java_io_Serializable {
    public "java/awt/MenuShortcut/key": number;
    public "java/awt/MenuShortcut/usesShift": number;
    public static "java/awt/MenuShortcut/serialVersionUID": Long;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuShortcut/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuShortcut/<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKey()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuShortcut/getKey()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "usesShiftModifier()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuShortcut/usesShiftModifier()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/awt/MenuShortcut;)Z"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuShortcut/equals(Ljava/awt/MenuShortcut;)Z"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuShortcut/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuShortcut/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuShortcut/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuShortcut/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_awt_event_ActionListener extends java_lang_Object, java_util_EventListener {
    "actionPerformed(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_MenuComponent extends java_lang_Object implements java_io_Serializable {
    public "java/awt/MenuComponent/peer": java_awt_peer_MenuComponentPeer;
    public "java/awt/MenuComponent/parent": java_awt_MenuContainer;
    public "java/awt/MenuComponent/appContext": sun_awt_AppContext;
    public "java/awt/MenuComponent/font": java_awt_Font;
    public "java/awt/MenuComponent/name": java_lang_String;
    public "java/awt/MenuComponent/nameExplicitlySet": number;
    public "java/awt/MenuComponent/newEventsOnly": number;
    public "java/awt/MenuComponent/acc": java_security_AccessControlContext;
    public static "java/awt/MenuComponent/actionListenerK": java_lang_String;
    public static "java/awt/MenuComponent/itemListenerK": java_lang_String;
    public static "java/awt/MenuComponent/serialVersionUID": Long;
    public "java/awt/MenuComponent/accessibleContext": javax_accessibility_AccessibleContext;
    public "getAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "java/awt/MenuComponent/getAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuComponent/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuComponent/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuComponent/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuComponent/setName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getParent()Ljava/awt/MenuContainer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuContainer) => void): void;
    public "java/awt/MenuComponent/getParent()Ljava/awt/MenuContainer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuContainer) => void): void;
    public "getParent_NoClientCode()Ljava/awt/MenuContainer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuContainer) => void): void;
    public "java/awt/MenuComponent/getParent_NoClientCode()Ljava/awt/MenuContainer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuContainer) => void): void;
    public "getPeer()Ljava/awt/peer/MenuComponentPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MenuComponentPeer) => void): void;
    public "java/awt/MenuComponent/getPeer()Ljava/awt/peer/MenuComponentPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_MenuComponentPeer) => void): void;
    public "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/MenuComponent/getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "getFont_NoClientCode()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/MenuComponent/getFont_NoClientCode()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuComponent/setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuComponent/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "postEvent(Ljava/awt/Event;)Z"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuComponent/postEvent(Ljava/awt/Event;)Z"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "dispatchEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuComponent/dispatchEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchEventImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuComponent/dispatchEventImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuComponent/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuComponent/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuComponent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuComponent/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTreeLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/MenuComponent/getTreeLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuComponent/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/MenuComponent/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/MenuComponent/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "getAccessibleIndexInParent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuComponent/getAccessibleIndexInParent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAccessibleChildIndex(Ljava/awt/MenuComponent;)I"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuComponent/getAccessibleChildIndex(Ljava/awt/MenuComponent;)I"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAccessibleStateSet()Ljavax/accessibility/AccessibleStateSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleStateSet) => void): void;
    public "java/awt/MenuComponent/getAccessibleStateSet()Ljavax/accessibility/AccessibleStateSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleStateSet) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/MenuComponent/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_MenuContainer extends java_lang_Object {
    "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    "remove(Ljava/awt/MenuComponent;)V"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable) => void): void;
    "postEvent(Ljava/awt/Event;)Z"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_peer_MenuComponentPeer extends java_lang_Object {
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_MenuItemPeer extends java_lang_Object, java_awt_peer_MenuComponentPeer {
    "setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_PopupMenu extends java_awt_Menu {
    public static "java/awt/PopupMenu/base": java_lang_String;
    public static "java/awt/PopupMenu/nameCounter": number;
    public "java/awt/PopupMenu/isTrayIconPopup": number;
    public static "java/awt/PopupMenu/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PopupMenu/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PopupMenu/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getParent()Ljava/awt/MenuContainer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuContainer) => void): void;
    public "java/awt/PopupMenu/getParent()Ljava/awt/MenuContainer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuContainer) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/PopupMenu/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PopupMenu/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "show(Ljava/awt/Component;II)V"(thread: JVMThread, args: [java_awt_Component, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PopupMenu/show(Ljava/awt/Component;II)V"(thread: JVMThread, args: [java_awt_Component, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/PopupMenu/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/PopupMenu/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_PopupMenuPeer extends java_lang_Object, java_awt_peer_MenuPeer {
    "show(Ljava/awt/Event;)V"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable) => void): void;
    "addSeparator()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "addItem(Ljava/awt/MenuItem;)V"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable) => void): void;
    "delItem(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Menu extends java_awt_MenuItem implements java_awt_MenuContainer, javax_accessibility_Accessible {
    public "java/awt/Menu/items": java_util_Vector;
    public "java/awt/Menu/tearOff": number;
    public "java/awt/Menu/isHelpMenu": number;
    public static "java/awt/Menu/base": java_lang_String;
    public static "java/awt/Menu/nameCounter": number;
    public static "java/awt/Menu/serialVersionUID": Long;
    public "java/awt/Menu/menuSerializedDataVersion": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Menu/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isTearOff()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Menu/isTearOff()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getItemCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Menu/getItemCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countItems()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Menu/countItems()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countItemsImpl()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Menu/countItemsImpl()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getItem(I)Ljava/awt/MenuItem;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "java/awt/Menu/getItem(I)Ljava/awt/MenuItem;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "getItemImpl(I)Ljava/awt/MenuItem;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "java/awt/Menu/getItemImpl(I)Ljava/awt/MenuItem;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "add(Ljava/awt/MenuItem;)Ljava/awt/MenuItem;"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "java/awt/Menu/add(Ljava/awt/MenuItem;)Ljava/awt/MenuItem;"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "add(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/add(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "insert(Ljava/awt/MenuItem;I)V"(thread: JVMThread, args: [java_awt_MenuItem, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/insert(Ljava/awt/MenuItem;I)V"(thread: JVMThread, args: [java_awt_MenuItem, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "insert(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/insert(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addSeparator()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/addSeparator()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "insertSeparator(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/insertSeparator(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/awt/MenuComponent;)V"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/remove(Ljava/awt/MenuComponent;)V"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "handleShortcut(Ljava/awt/event/KeyEvent;)Z"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Menu/handleShortcut(Ljava/awt/event/KeyEvent;)Z"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getShortcutMenuItem(Ljava/awt/MenuShortcut;)Ljava/awt/MenuItem;"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "java/awt/Menu/getShortcutMenuItem(Ljava/awt/MenuShortcut;)Ljava/awt/MenuItem;"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "shortcuts()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/awt/Menu/shortcuts()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "deleteShortcut(Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/deleteShortcut(Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Menu/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Menu/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Menu/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Menu/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "getAccessibleChildIndex(Ljava/awt/MenuComponent;)I"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Menu/getAccessibleChildIndex(Ljava/awt/MenuComponent;)I"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Menu/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_MenuPeer extends java_lang_Object, java_awt_peer_MenuItemPeer {
    "addSeparator()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "addItem(Ljava/awt/MenuItem;)V"(thread: JVMThread, args: [java_awt_MenuItem], cb?: (e?: java_lang_Throwable) => void): void;
    "delItem(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_MenuBar extends java_awt_MenuComponent implements java_awt_MenuContainer, javax_accessibility_Accessible {
    public "java/awt/MenuBar/menus": java_util_Vector;
    public "java/awt/MenuBar/helpMenu": java_awt_Menu;
    public static "java/awt/MenuBar/base": java_lang_String;
    public static "java/awt/MenuBar/nameCounter": number;
    public static "java/awt/MenuBar/serialVersionUID": Long;
    public "java/awt/MenuBar/menuBarSerializedDataVersion": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuBar/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/MenuBar/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuBar/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuBar/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getHelpMenu()Ljava/awt/Menu;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Menu) => void): void;
    public "java/awt/MenuBar/getHelpMenu()Ljava/awt/Menu;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Menu) => void): void;
    public "setHelpMenu(Ljava/awt/Menu;)V"(thread: JVMThread, args: [java_awt_Menu], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuBar/setHelpMenu(Ljava/awt/Menu;)V"(thread: JVMThread, args: [java_awt_Menu], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/awt/Menu;)Ljava/awt/Menu;"(thread: JVMThread, args: [java_awt_Menu], cb?: (e?: java_lang_Throwable, rv?: java_awt_Menu) => void): void;
    public "java/awt/MenuBar/add(Ljava/awt/Menu;)Ljava/awt/Menu;"(thread: JVMThread, args: [java_awt_Menu], cb?: (e?: java_lang_Throwable, rv?: java_awt_Menu) => void): void;
    public "remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuBar/remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/awt/MenuComponent;)V"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuBar/remove(Ljava/awt/MenuComponent;)V"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMenuCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuBar/getMenuCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countMenus()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuBar/countMenus()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMenuCountImpl()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuBar/getMenuCountImpl()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMenu(I)Ljava/awt/Menu;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Menu) => void): void;
    public "java/awt/MenuBar/getMenu(I)Ljava/awt/Menu;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Menu) => void): void;
    public "getMenuImpl(I)Ljava/awt/Menu;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Menu) => void): void;
    public "java/awt/MenuBar/getMenuImpl(I)Ljava/awt/Menu;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Menu) => void): void;
    public "shortcuts()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/awt/MenuBar/shortcuts()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getShortcutMenuItem(Ljava/awt/MenuShortcut;)Ljava/awt/MenuItem;"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "java/awt/MenuBar/getShortcutMenuItem(Ljava/awt/MenuShortcut;)Ljava/awt/MenuItem;"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuItem) => void): void;
    public "handleShortcut(Ljava/awt/event/KeyEvent;)Z"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuBar/handleShortcut(Ljava/awt/event/KeyEvent;)Z"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "deleteShortcut(Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuBar/deleteShortcut(Ljava/awt/MenuShortcut;)V"(thread: JVMThread, args: [java_awt_MenuShortcut], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuBar/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/MenuBar/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/MenuBar/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/MenuBar/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "getAccessibleChildIndex(Ljava/awt/MenuComponent;)I"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/MenuBar/getAccessibleChildIndex(Ljava/awt/MenuComponent;)I"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/MenuBar/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_MenuBarPeer extends java_lang_Object, java_awt_peer_MenuComponentPeer {
    "addMenu(Ljava/awt/Menu;)V"(thread: JVMThread, args: [java_awt_Menu], cb?: (e?: java_lang_Throwable) => void): void;
    "delMenu(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "addHelpMenu(Ljava/awt/Menu;)V"(thread: JVMThread, args: [java_awt_Menu], cb?: (e?: java_lang_Throwable) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_DialogPeer extends java_lang_Object, java_awt_peer_WindowPeer {
    "setTitle(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setResizable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "blockWindows(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    "toFront()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "toBack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateAlwaysOnTopState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateFocusableWindowState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setModalBlocked(Ljava/awt/Dialog;Z)V"(thread: JVMThread, args: [java_awt_Dialog, number], cb?: (e?: java_lang_Throwable) => void): void;
    "updateMinimumSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateIconImages()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setOpaque(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "updateWindow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "repositionSecurityWarning()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    "beginValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "beginLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_peer_WindowPeer extends java_lang_Object, java_awt_peer_ContainerPeer {
    "toFront()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "toBack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateAlwaysOnTopState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateFocusableWindowState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setModalBlocked(Ljava/awt/Dialog;Z)V"(thread: JVMThread, args: [java_awt_Dialog, number], cb?: (e?: java_lang_Throwable) => void): void;
    "updateMinimumSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateIconImages()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setOpaque(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "updateWindow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "repositionSecurityWarning()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    "beginValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "beginLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Panel extends java_awt_Container implements javax_accessibility_Accessible {
    public static "java/awt/Panel/base": java_lang_String;
    public static "java/awt/Panel/nameCounter": number;
    public static "java/awt/Panel/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Panel/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/LayoutManager;)V"(thread: JVMThread, args: [java_awt_LayoutManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Panel/<init>(Ljava/awt/LayoutManager;)V"(thread: JVMThread, args: [java_awt_LayoutManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Panel/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Panel/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Panel/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Panel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_LayoutManager extends java_lang_Object {
    "addLayoutComponent(Ljava/lang/String;Ljava/awt/Component;)V"(thread: JVMThread, args: [java_lang_String, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    "removeLayoutComponent(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    "preferredLayoutSize(Ljava/awt/Container;)Ljava/awt/Dimension;"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "minimumLayoutSize(Ljava/awt/Container;)Ljava/awt/Dimension;"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "layoutContainer(Ljava/awt/Container;)V"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_PanelPeer extends java_lang_Object, java_awt_peer_ContainerPeer {
    "getInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    "beginValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "beginLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Canvas extends java_awt_Component implements javax_accessibility_Accessible {
    public static "java/awt/Canvas/base": java_lang_String;
    public static "java/awt/Canvas/nameCounter": number;
    public static "java/awt/Canvas/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Canvas/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Canvas/<init>(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "setGraphicsConfiguration(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Canvas/setGraphicsConfiguration(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Canvas/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Canvas/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Canvas/paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Canvas/update(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "postsOldMouseEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Canvas/postsOldMouseEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createBufferStrategy(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Canvas/createBufferStrategy(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "createBufferStrategy(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Canvas/createBufferStrategy(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBufferStrategy()Ljava/awt/image/BufferStrategy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferStrategy) => void): void;
    public "java/awt/Canvas/getBufferStrategy()Ljava/awt/image/BufferStrategy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferStrategy) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Canvas/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Canvas/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_CanvasPeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "getAppropriateGraphicsConfiguration(Ljava/awt/GraphicsConfiguration;)Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_peer_FramePeer extends java_lang_Object, java_awt_peer_WindowPeer {
    "setTitle(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setMenuBar(Ljava/awt/MenuBar;)V"(thread: JVMThread, args: [java_awt_MenuBar], cb?: (e?: java_lang_Throwable) => void): void;
    "setResizable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setState(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setMaximizedBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    "setBoundsPrivate(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "getBoundsPrivate()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    "emulateActivation(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "toFront()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "toBack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateAlwaysOnTopState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateFocusableWindowState()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setModalBlocked(Ljava/awt/Dialog;Z)V"(thread: JVMThread, args: [java_awt_Dialog, number], cb?: (e?: java_lang_Throwable) => void): void;
    "updateMinimumSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "updateIconImages()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setOpaque(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "updateWindow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "repositionSecurityWarning()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    "beginValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "beginLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Choice extends java_awt_Component implements java_awt_ItemSelectable, javax_accessibility_Accessible {
    public "java/awt/Choice/pItems": java_util_Vector;
    public "java/awt/Choice/selectedIndex": number;
    public "java/awt/Choice/itemListener": java_awt_event_ItemListener;
    public static "java/awt/Choice/base": java_lang_String;
    public static "java/awt/Choice/nameCounter": number;
    public static "java/awt/Choice/serialVersionUID": Long;
    public "java/awt/Choice/choiceSerializedDataVersion": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Choice/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getItemCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Choice/getItemCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countItems()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Choice/countItems()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getItem(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Choice/getItem(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getItemImpl(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Choice/getItemImpl(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "add(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/add(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "addItem(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/addItem(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "insertNoInvalidate(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/insertNoInvalidate(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "insert(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/insert(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNoInvalidate(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/removeNoInvalidate(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSelectedItem()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Choice/getSelectedItem()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSelectedObjects()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/awt/Choice/getSelectedObjects()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "getSelectedIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Choice/getSelectedIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "select(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/select(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "select(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/select(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "addItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/addItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/removeItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getItemListeners()[Ljava/awt/event/ItemListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ItemListener>) => void): void;
    public "java/awt/Choice/getItemListeners()[Ljava/awt/event/ItemListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ItemListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/Choice/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Choice/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processItemEvent(Ljava/awt/event/ItemEvent;)V"(thread: JVMThread, args: [java_awt_event_ItemEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/processItemEvent(Ljava/awt/event/ItemEvent;)V"(thread: JVMThread, args: [java_awt_event_ItemEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Choice/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Choice/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Choice/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Choice/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Choice/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_ChoicePeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "add(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    "remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "select(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_TextArea extends java_awt_TextComponent {
    public "java/awt/TextArea/rows": number;
    public "java/awt/TextArea/columns": number;
    public static "java/awt/TextArea/base": java_lang_String;
    public static "java/awt/TextArea/nameCounter": number;
    public static "java/awt/TextArea/SCROLLBARS_BOTH": number;
    public static "java/awt/TextArea/SCROLLBARS_VERTICAL_ONLY": number;
    public static "java/awt/TextArea/SCROLLBARS_HORIZONTAL_ONLY": number;
    public static "java/awt/TextArea/SCROLLBARS_NONE": number;
    public "java/awt/TextArea/scrollbarVisibility": number;
    public static "java/awt/TextArea/forwardTraversalKeys": java_util_Set;
    public static "java/awt/TextArea/backwardTraversalKeys": java_util_Set;
    public static "java/awt/TextArea/serialVersionUID": Long;
    public "java/awt/TextArea/textAreaSerializedDataVersion": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/TextArea/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;III)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/<init>(Ljava/lang/String;III)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/TextArea/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "insert(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/insert(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "insertText(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/insertText(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/append(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "appendText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/appendText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceRange(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/replaceRange(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceText(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/replaceText(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRows()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextArea/getRows()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setRows(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/setRows(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getColumns()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextArea/getColumns()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setColumns(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/setColumns(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getScrollbarVisibility()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextArea/getScrollbarVisibility()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPreferredSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextArea/getPreferredSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "preferredSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextArea/preferredSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextArea/getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextArea/preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getMinimumSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextArea/getMinimumSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "minimumSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextArea/minimumSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextArea/getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextArea/minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/TextArea/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextArea/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/TextArea/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/TextArea/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_TextComponent extends java_awt_Component implements javax_accessibility_Accessible {
    public "java/awt/TextComponent/text": java_lang_String;
    public "java/awt/TextComponent/editable": number;
    public "java/awt/TextComponent/selectionStart": number;
    public "java/awt/TextComponent/selectionEnd": number;
    public "java/awt/TextComponent/backgroundSetByClientCode": number;
    public "java/awt/TextComponent/textListener": java_awt_event_TextListener;
    public static "java/awt/TextComponent/serialVersionUID": Long;
    public "java/awt/TextComponent/textComponentSerializedDataVersion": number;
    public "java/awt/TextComponent/checkForEnableIM": number;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "enableInputMethodsIfNecessary()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/enableInputMethodsIfNecessary()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "enableInputMethods(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/enableInputMethods(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "areInputMethodsEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextComponent/areInputMethodsEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInputMethodRequests()Ljava/awt/im/InputMethodRequests;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputMethodRequests) => void): void;
    public "java/awt/TextComponent/getInputMethodRequests()Ljava/awt/im/InputMethodRequests;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputMethodRequests) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/TextComponent/getText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSelectedText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/TextComponent/getSelectedText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isEditable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextComponent/isEditable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setEditable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/setEditable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "java/awt/TextComponent/getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSelectionStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextComponent/getSelectionStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSelectionStart(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/setSelectionStart(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSelectionEnd()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextComponent/getSelectionEnd()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSelectionEnd(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/setSelectionEnd(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "select(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/select(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "selectAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/selectAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCaretPosition(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/setCaretPosition(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCaretPosition()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextComponent/getCaretPosition()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addTextListener(Ljava/awt/event/TextListener;)V"(thread: JVMThread, args: [java_awt_event_TextListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/addTextListener(Ljava/awt/event/TextListener;)V"(thread: JVMThread, args: [java_awt_event_TextListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeTextListener(Ljava/awt/event/TextListener;)V"(thread: JVMThread, args: [java_awt_event_TextListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/removeTextListener(Ljava/awt/event/TextListener;)V"(thread: JVMThread, args: [java_awt_event_TextListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTextListeners()[Ljava/awt/event/TextListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_TextListener>) => void): void;
    public "java/awt/TextComponent/getTextListeners()[Ljava/awt/event/TextListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_TextListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/TextComponent/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextComponent/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processTextEvent(Ljava/awt/event/TextEvent;)V"(thread: JVMThread, args: [java_awt_event_TextEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/processTextEvent(Ljava/awt/event/TextEvent;)V"(thread: JVMThread, args: [java_awt_event_TextEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/TextComponent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "canAccessClipboard()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextComponent/canAccessClipboard()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextComponent/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/TextComponent/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
  }
  export class java_awt_event_TextEvent extends java_awt_AWTEvent {
    public static "java/awt/event/TextEvent/TEXT_FIRST": number;
    public static "java/awt/event/TextEvent/TEXT_LAST": number;
    public static "java/awt/event/TextEvent/TEXT_VALUE_CHANGED": number;
    public static "java/awt/event/TextEvent/serialVersionUID": Long;
    public "<init>(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/TextEvent/<init>(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/TextEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_awt_im_InputMethodRequests extends java_lang_Object {
    "getTextLocation(Ljava/awt/font/TextHitInfo;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_font_TextHitInfo], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    "getLocationOffset(II)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    "getInsertPositionOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getCommittedText(II[Ljava/text/AttributedCharacterIterator$Attribute;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [number, number, JVMArray<java_text_AttributedCharacterIterator$Attribute>], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    "getCommittedTextLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "cancelLatestCommittedText([Ljava/text/AttributedCharacterIterator$Attribute;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [JVMArray<java_text_AttributedCharacterIterator$Attribute>], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    "getSelectedText([Ljava/text/AttributedCharacterIterator$Attribute;)Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: [JVMArray<java_text_AttributedCharacterIterator$Attribute>], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
  }
  export class java_awt_font_TextHitInfo extends java_lang_Object {
    public "java/awt/font/TextHitInfo/charIndex": number;
    public "java/awt/font/TextHitInfo/isLeadingEdge": number;
    public "<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/TextHitInfo/<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCharIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/TextHitInfo/getCharIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLeadingEdge()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/TextHitInfo/isLeadingEdge()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInsertionIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/TextHitInfo/getInsertionIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/TextHitInfo/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/TextHitInfo/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/awt/font/TextHitInfo;)Z"(thread: JVMThread, args: [java_awt_font_TextHitInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/TextHitInfo/equals(Ljava/awt/font/TextHitInfo;)Z"(thread: JVMThread, args: [java_awt_font_TextHitInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/font/TextHitInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "leading(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public static "java/awt/font/TextHitInfo/leading(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public static "trailing(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public static "java/awt/font/TextHitInfo/trailing(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public static "beforeOffset(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public static "java/awt/font/TextHitInfo/beforeOffset(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public static "afterOffset(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public static "java/awt/font/TextHitInfo/afterOffset(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public "getOtherHit()Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public "java/awt/font/TextHitInfo/getOtherHit()Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public "getOffsetHit(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public "java/awt/font/TextHitInfo/getOffsetHit(I)Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
  }
  export interface java_awt_event_TextListener extends java_lang_Object, java_util_EventListener {
    "textValueChanged(Ljava/awt/event/TextEvent;)V"(thread: JVMThread, args: [java_awt_event_TextEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_TextAreaPeer extends java_lang_Object, java_awt_peer_TextComponentPeer {
    "insert(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceRange(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "getPreferredSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize(II)Ljava/awt/Dimension;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "setEditable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "getSelectionStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSelectionEnd()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "select(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setCaretPosition(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getCaretPosition()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getInputMethodRequests()Ljava/awt/im/InputMethodRequests;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputMethodRequests) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_peer_TextComponentPeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "setEditable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "getSelectionStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSelectionEnd()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "select(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setCaretPosition(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getCaretPosition()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getInputMethodRequests()Ljava/awt/im/InputMethodRequests;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputMethodRequests) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_ScrollPane extends java_awt_Container implements javax_accessibility_Accessible {
    public static "java/awt/ScrollPane/SCROLLBARS_AS_NEEDED": number;
    public static "java/awt/ScrollPane/SCROLLBARS_ALWAYS": number;
    public static "java/awt/ScrollPane/SCROLLBARS_NEVER": number;
    public "java/awt/ScrollPane/scrollbarDisplayPolicy": number;
    public "java/awt/ScrollPane/vAdjustable": java_awt_ScrollPaneAdjustable;
    public "java/awt/ScrollPane/hAdjustable": java_awt_ScrollPaneAdjustable;
    public static "java/awt/ScrollPane/base": java_lang_String;
    public static "java/awt/ScrollPane/nameCounter": number;
    public static "java/awt/ScrollPane/defaultWheelScroll": number;
    public "java/awt/ScrollPane/wheelScrollingEnabled": number;
    public static "java/awt/ScrollPane/serialVersionUID": Long;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/ScrollPane/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/ScrollPane/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addToPanel(Ljava/awt/Component;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/addToPanel(Ljava/awt/Component;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addImpl(Ljava/awt/Component;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/addImpl(Ljava/awt/Component;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getScrollbarDisplayPolicy()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPane/getScrollbarDisplayPolicy()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getViewportSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/ScrollPane/getViewportSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getHScrollbarHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPane/getHScrollbarHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getVScrollbarWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPane/getVScrollbarWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getVAdjustable()Ljava/awt/Adjustable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Adjustable) => void): void;
    public "java/awt/ScrollPane/getVAdjustable()Ljava/awt/Adjustable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Adjustable) => void): void;
    public "getHAdjustable()Ljava/awt/Adjustable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Adjustable) => void): void;
    public "java/awt/ScrollPane/getHAdjustable()Ljava/awt/Adjustable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Adjustable) => void): void;
    public "setScrollPosition(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/setScrollPosition(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setScrollPosition(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/setScrollPosition(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "getScrollPosition()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/ScrollPane/getScrollPosition()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "setLayout(Ljava/awt/LayoutManager;)V"(thread: JVMThread, args: [java_awt_LayoutManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/setLayout(Ljava/awt/LayoutManager;)V"(thread: JVMThread, args: [java_awt_LayoutManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/doLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "calculateChildSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/ScrollPane/calculateChildSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "printComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/printComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/ScrollPane/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "autoProcessMouseWheel(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/autoProcessMouseWheel(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processMouseWheelEvent(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/processMouseWheelEvent(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "eventTypeEnabled(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPane/eventTypeEnabled(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setWheelScrollingEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/setWheelScrollingEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isWheelScrollingEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPane/isWheelScrollingEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPane/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/ScrollPane/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/ScrollPane/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_Adjustable extends java_lang_Object {
    "getOrientation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setMinimum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getMinimum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setMaximum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getMaximum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setUnitIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getUnitIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setBlockIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getBlockIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisibleAmount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getVisibleAmount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setValue(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "addAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
    "removeAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_AdjustmentListener extends java_lang_Object, java_util_EventListener {
    "adjustmentValueChanged(Ljava/awt/event/AdjustmentEvent;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_event_AdjustmentEvent extends java_awt_AWTEvent {
    public static "java/awt/event/AdjustmentEvent/ADJUSTMENT_FIRST": number;
    public static "java/awt/event/AdjustmentEvent/ADJUSTMENT_LAST": number;
    public static "java/awt/event/AdjustmentEvent/ADJUSTMENT_VALUE_CHANGED": number;
    public static "java/awt/event/AdjustmentEvent/UNIT_INCREMENT": number;
    public static "java/awt/event/AdjustmentEvent/UNIT_DECREMENT": number;
    public static "java/awt/event/AdjustmentEvent/BLOCK_DECREMENT": number;
    public static "java/awt/event/AdjustmentEvent/BLOCK_INCREMENT": number;
    public static "java/awt/event/AdjustmentEvent/TRACK": number;
    public "java/awt/event/AdjustmentEvent/adjustable": java_awt_Adjustable;
    public "java/awt/event/AdjustmentEvent/value": number;
    public "java/awt/event/AdjustmentEvent/adjustmentType": number;
    public "java/awt/event/AdjustmentEvent/isAdjusting": number;
    public static "java/awt/event/AdjustmentEvent/serialVersionUID": Long;
    public "<init>(Ljava/awt/Adjustable;III)V"(thread: JVMThread, args: [java_awt_Adjustable, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/AdjustmentEvent/<init>(Ljava/awt/Adjustable;III)V"(thread: JVMThread, args: [java_awt_Adjustable, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Adjustable;IIIZ)V"(thread: JVMThread, args: [java_awt_Adjustable, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/AdjustmentEvent/<init>(Ljava/awt/Adjustable;IIIZ)V"(thread: JVMThread, args: [java_awt_Adjustable, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAdjustable()Ljava/awt/Adjustable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Adjustable) => void): void;
    public "java/awt/event/AdjustmentEvent/getAdjustable()Ljava/awt/Adjustable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Adjustable) => void): void;
    public "getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/AdjustmentEvent/getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAdjustmentType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/AdjustmentEvent/getAdjustmentType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getValueIsAdjusting()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/AdjustmentEvent/getValueIsAdjusting()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/AdjustmentEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_ScrollPaneAdjustable extends java_lang_Object implements java_awt_Adjustable, java_io_Serializable {
    public "java/awt/ScrollPaneAdjustable/sp": java_awt_ScrollPane;
    public "java/awt/ScrollPaneAdjustable/orientation": number;
    public "java/awt/ScrollPaneAdjustable/value": number;
    public "java/awt/ScrollPaneAdjustable/minimum": number;
    public "java/awt/ScrollPaneAdjustable/maximum": number;
    public "java/awt/ScrollPaneAdjustable/visibleAmount": number;
    public "java/awt/ScrollPaneAdjustable/isAdjusting": number;
    public "java/awt/ScrollPaneAdjustable/unitIncrement": number;
    public "java/awt/ScrollPaneAdjustable/blockIncrement": number;
    public "java/awt/ScrollPaneAdjustable/adjustmentListener": java_awt_event_AdjustmentListener;
    public static "java/awt/ScrollPaneAdjustable/SCROLLPANE_ONLY": java_lang_String;
    public static "java/awt/ScrollPaneAdjustable/serialVersionUID": Long;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/ScrollPaneAdjustable/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/ScrollPane;Ljava/awt/event/AdjustmentListener;I)V"(thread: JVMThread, args: [java_awt_ScrollPane, java_awt_event_AdjustmentListener, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/<init>(Ljava/awt/ScrollPane;Ljava/awt/event/AdjustmentListener;I)V"(thread: JVMThread, args: [java_awt_ScrollPane, java_awt_event_AdjustmentListener, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSpan(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/setSpan(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOrientation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPaneAdjustable/getOrientation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMinimum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/setMinimum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMinimum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPaneAdjustable/getMinimum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMaximum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/setMaximum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMaximum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPaneAdjustable/getMaximum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setUnitIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/setUnitIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUnitIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPaneAdjustable/getUnitIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setBlockIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/setBlockIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBlockIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPaneAdjustable/getBlockIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setVisibleAmount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/setVisibleAmount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getVisibleAmount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPaneAdjustable/getVisibleAmount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setValueIsAdjusting(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/setValueIsAdjusting(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValueIsAdjusting()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPaneAdjustable/getValueIsAdjusting()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setValue(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/setValue(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTypedValue(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/setTypedValue(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ScrollPaneAdjustable/getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/addAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ScrollPaneAdjustable/removeAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAdjustmentListeners()[Ljava/awt/event/AdjustmentListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_AdjustmentListener>) => void): void;
    public "java/awt/ScrollPaneAdjustable/getAdjustmentListeners()[Ljava/awt/event/AdjustmentListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_AdjustmentListener>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/ScrollPaneAdjustable/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/ScrollPaneAdjustable/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000(Ljava/awt/ScrollPaneAdjustable;II)V"(thread: JVMThread, args: [java_awt_ScrollPaneAdjustable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/ScrollPaneAdjustable/access$000(Ljava/awt/ScrollPaneAdjustable;II)V"(thread: JVMThread, args: [java_awt_ScrollPaneAdjustable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/ScrollPaneAdjustable/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_ScrollPanePeer extends java_lang_Object, java_awt_peer_ContainerPeer {
    "getHScrollbarHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getVScrollbarWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setScrollPosition(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "childResized(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setUnitIncrement(Ljava/awt/Adjustable;I)V"(thread: JVMThread, args: [java_awt_Adjustable, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setValue(Ljava/awt/Adjustable;I)V"(thread: JVMThread, args: [java_awt_Adjustable, number], cb?: (e?: java_lang_Throwable) => void): void;
    "getInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    "beginValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endValidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "beginLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "endLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Scrollbar extends java_awt_Component implements java_awt_Adjustable, javax_accessibility_Accessible {
    public static "java/awt/Scrollbar/HORIZONTAL": number;
    public static "java/awt/Scrollbar/VERTICAL": number;
    public "java/awt/Scrollbar/value": number;
    public "java/awt/Scrollbar/maximum": number;
    public "java/awt/Scrollbar/minimum": number;
    public "java/awt/Scrollbar/visibleAmount": number;
    public "java/awt/Scrollbar/orientation": number;
    public "java/awt/Scrollbar/lineIncrement": number;
    public "java/awt/Scrollbar/pageIncrement": number;
    public "java/awt/Scrollbar/isAdjusting": number;
    public "java/awt/Scrollbar/adjustmentListener": java_awt_event_AdjustmentListener;
    public static "java/awt/Scrollbar/base": java_lang_String;
    public static "java/awt/Scrollbar/nameCounter": number;
    public static "java/awt/Scrollbar/serialVersionUID": Long;
    public "java/awt/Scrollbar/scrollbarSerializedDataVersion": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Scrollbar/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/<init>(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Scrollbar/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOrientation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getOrientation()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setOrientation(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setOrientation(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setValue(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setValue(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMinimum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getMinimum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMinimum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setMinimum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMaximum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getMaximum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMaximum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setMaximum(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getVisibleAmount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getVisibleAmount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getVisible()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getVisible()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setVisibleAmount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setVisibleAmount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setUnitIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setUnitIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLineIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setLineIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUnitIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getUnitIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLineIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getLineIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setBlockIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setBlockIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPageIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setPageIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBlockIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getBlockIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPageIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getPageIncrement()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setValues(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setValues(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValueIsAdjusting()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/getValueIsAdjusting()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setValueIsAdjusting(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/setValueIsAdjusting(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/addAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/removeAdjustmentListener(Ljava/awt/event/AdjustmentListener;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAdjustmentListeners()[Ljava/awt/event/AdjustmentListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_AdjustmentListener>) => void): void;
    public "java/awt/Scrollbar/getAdjustmentListeners()[Ljava/awt/event/AdjustmentListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_AdjustmentListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/Scrollbar/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Scrollbar/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processAdjustmentEvent(Ljava/awt/event/AdjustmentEvent;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/processAdjustmentEvent(Ljava/awt/event/AdjustmentEvent;)V"(thread: JVMThread, args: [java_awt_event_AdjustmentEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Scrollbar/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Scrollbar/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Scrollbar/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Scrollbar/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_ScrollbarPeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "setValues(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setLineIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setPageIncrement(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Checkbox extends java_awt_Component implements java_awt_ItemSelectable, javax_accessibility_Accessible {
    public "java/awt/Checkbox/label": java_lang_String;
    public "java/awt/Checkbox/state": number;
    public "java/awt/Checkbox/group": java_awt_CheckboxGroup;
    public "java/awt/Checkbox/itemListener": java_awt_event_ItemListener;
    public static "java/awt/Checkbox/base": java_lang_String;
    public static "java/awt/Checkbox/nameCounter": number;
    public static "java/awt/Checkbox/serialVersionUID": Long;
    public "java/awt/Checkbox/checkboxSerializedDataVersion": number;
    public "setStateInternal(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/setStateInternal(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;ZLjava/awt/CheckboxGroup;)V"(thread: JVMThread, args: [java_lang_String, number, java_awt_CheckboxGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/<init>(Ljava/lang/String;ZLjava/awt/CheckboxGroup;)V"(thread: JVMThread, args: [java_lang_String, number, java_awt_CheckboxGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/awt/CheckboxGroup;Z)V"(thread: JVMThread, args: [java_lang_String, java_awt_CheckboxGroup, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/<init>(Ljava/lang/String;Ljava/awt/CheckboxGroup;Z)V"(thread: JVMThread, args: [java_lang_String, java_awt_CheckboxGroup, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Checkbox/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLabel()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Checkbox/getLabel()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getState()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Checkbox/getState()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setState(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/setState(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSelectedObjects()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/awt/Checkbox/getSelectedObjects()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "getCheckboxGroup()Ljava/awt/CheckboxGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_CheckboxGroup) => void): void;
    public "java/awt/Checkbox/getCheckboxGroup()Ljava/awt/CheckboxGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_CheckboxGroup) => void): void;
    public "setCheckboxGroup(Ljava/awt/CheckboxGroup;)V"(thread: JVMThread, args: [java_awt_CheckboxGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/setCheckboxGroup(Ljava/awt/CheckboxGroup;)V"(thread: JVMThread, args: [java_awt_CheckboxGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "addItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/addItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/removeItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getItemListeners()[Ljava/awt/event/ItemListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ItemListener>) => void): void;
    public "java/awt/Checkbox/getItemListeners()[Ljava/awt/event/ItemListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ItemListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/Checkbox/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Checkbox/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processItemEvent(Ljava/awt/event/ItemEvent;)V"(thread: JVMThread, args: [java_awt_event_ItemEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/processItemEvent(Ljava/awt/event/ItemEvent;)V"(thread: JVMThread, args: [java_awt_event_ItemEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Checkbox/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Checkbox/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Checkbox/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Checkbox/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Checkbox/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_CheckboxGroup extends java_lang_Object implements java_io_Serializable {
    public "java/awt/CheckboxGroup/selectedCheckbox": java_awt_Checkbox;
    public static "java/awt/CheckboxGroup/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxGroup/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSelectedCheckbox()Ljava/awt/Checkbox;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Checkbox) => void): void;
    public "java/awt/CheckboxGroup/getSelectedCheckbox()Ljava/awt/Checkbox;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Checkbox) => void): void;
    public "getCurrent()Ljava/awt/Checkbox;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Checkbox) => void): void;
    public "java/awt/CheckboxGroup/getCurrent()Ljava/awt/Checkbox;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Checkbox) => void): void;
    public "setSelectedCheckbox(Ljava/awt/Checkbox;)V"(thread: JVMThread, args: [java_awt_Checkbox], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxGroup/setSelectedCheckbox(Ljava/awt/Checkbox;)V"(thread: JVMThread, args: [java_awt_Checkbox], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCurrent(Ljava/awt/Checkbox;)V"(thread: JVMThread, args: [java_awt_Checkbox], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/CheckboxGroup/setCurrent(Ljava/awt/Checkbox;)V"(thread: JVMThread, args: [java_awt_Checkbox], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/CheckboxGroup/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_awt_peer_CheckboxPeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "setState(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setCheckboxGroup(Ljava/awt/CheckboxGroup;)V"(thread: JVMThread, args: [java_awt_CheckboxGroup], cb?: (e?: java_lang_Throwable) => void): void;
    "setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_List extends java_awt_Component implements java_awt_ItemSelectable, javax_accessibility_Accessible {
    public "java/awt/List/items": java_util_Vector;
    public "java/awt/List/rows": number;
    public "java/awt/List/multipleMode": number;
    public "java/awt/List/selected": JVMArray<number>;
    public "java/awt/List/visibleIndex": number;
    public "java/awt/List/actionListener": java_awt_event_ActionListener;
    public "java/awt/List/itemListener": java_awt_event_ItemListener;
    public static "java/awt/List/base": java_lang_String;
    public static "java/awt/List/nameCounter": number;
    public static "java/awt/List/serialVersionUID": Long;
    public static "java/awt/List/DEFAULT_VISIBLE_ROWS": number;
    public "java/awt/List/listSerializedDataVersion": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/List/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getItemCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/getItemCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countItems()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/countItems()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getItem(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/List/getItem(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getItemImpl(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/List/getItemImpl(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getItems()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/awt/List/getItems()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "add(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/add(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "addItem(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/addItem(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/add(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addItem(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/addItem(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceItem(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/replaceItem(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "delItem(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/delItem(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSelectedIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/getSelectedIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSelectedIndexes()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/List/getSelectedIndexes()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getSelectedItem()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/List/getSelectedItem()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSelectedItems()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/awt/List/getSelectedItems()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getSelectedObjects()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/awt/List/getSelectedObjects()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "select(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/select(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "deselect(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/deselect(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isIndexSelected(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/isIndexSelected(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSelected(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/isSelected(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRows()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/getRows()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMultipleMode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/isMultipleMode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "allowsMultipleSelections()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/allowsMultipleSelections()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMultipleMode(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/setMultipleMode(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setMultipleSelections(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/setMultipleSelections(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getVisibleIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/getVisibleIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "makeVisible(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/makeVisible(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPreferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/List/getPreferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "preferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/List/preferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/List/getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/List/preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getMinimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/List/getMinimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "minimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/List/minimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/List/getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/List/minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "addItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/addItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/removeItemListener(Ljava/awt/event/ItemListener;)V"(thread: JVMThread, args: [java_awt_event_ItemListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getItemListeners()[Ljava/awt/event/ItemListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ItemListener>) => void): void;
    public "java/awt/List/getItemListeners()[Ljava/awt/event/ItemListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ItemListener>) => void): void;
    public "addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "java/awt/List/getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/List/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/List/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processItemEvent(Ljava/awt/event/ItemEvent;)V"(thread: JVMThread, args: [java_awt_event_ItemEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/processItemEvent(Ljava/awt/event/ItemEvent;)V"(thread: JVMThread, args: [java_awt_event_ItemEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processActionEvent(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/processActionEvent(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/List/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "delItems(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/delItems(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/List/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/List/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/List/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_ListPeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "getSelectedIndexes()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    "add(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    "delItems(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "select(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "deselect(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "makeVisible(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setMultipleMode(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getPreferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Label extends java_awt_Component implements javax_accessibility_Accessible {
    public static "java/awt/Label/LEFT": number;
    public static "java/awt/Label/CENTER": number;
    public static "java/awt/Label/RIGHT": number;
    public "java/awt/Label/text": java_lang_String;
    public "java/awt/Label/alignment": number;
    public static "java/awt/Label/base": java_lang_String;
    public static "java/awt/Label/nameCounter": number;
    public static "java/awt/Label/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Label/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Label/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Label/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Label/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Label/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Label/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAlignment()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Label/getAlignment()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setAlignment(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Label/setAlignment(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Label/getText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Label/setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Label/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Label/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Label/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Label/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_LabelPeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "setAlignment(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_TextField extends java_awt_TextComponent {
    public "java/awt/TextField/columns": number;
    public "java/awt/TextField/echoChar": number;
    public "java/awt/TextField/actionListener": java_awt_event_ActionListener;
    public static "java/awt/TextField/base": java_lang_String;
    public static "java/awt/TextField/nameCounter": number;
    public static "java/awt/TextField/serialVersionUID": Long;
    public "java/awt/TextField/textFieldSerializedDataVersion": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/TextField/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/TextField/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEchoChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextField/getEchoChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setEchoChar(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/setEchoChar(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setEchoCharacter(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/setEchoCharacter(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "echoCharIsSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextField/echoCharIsSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getColumns()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextField/getColumns()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setColumns(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/setColumns(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPreferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextField/getPreferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "preferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextField/preferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextField/getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextField/preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getMinimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextField/getMinimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "minimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextField/minimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextField/getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/TextField/minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "java/awt/TextField/getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/TextField/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/TextField/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processActionEvent(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/processActionEvent(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/TextField/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/TextField/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/TextField/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/TextField/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_TextFieldPeer extends java_lang_Object, java_awt_peer_TextComponentPeer {
    "setEchoChar(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getPreferredSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize(I)Ljava/awt/Dimension;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "setEditable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getText()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "setText(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "getSelectionStart()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getSelectionEnd()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "select(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setCaretPosition(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getCaretPosition()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getInputMethodRequests()Ljava/awt/im/InputMethodRequests;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputMethodRequests) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Button extends java_awt_Component implements javax_accessibility_Accessible {
    public "java/awt/Button/label": java_lang_String;
    public "java/awt/Button/actionCommand": java_lang_String;
    public "java/awt/Button/actionListener": java_awt_event_ActionListener;
    public static "java/awt/Button/base": java_lang_String;
    public static "java/awt/Button/nameCounter": number;
    public static "java/awt/Button/serialVersionUID": Long;
    public "java/awt/Button/buttonSerializedDataVersion": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Button/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Button/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLabel()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Button/getLabel()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setActionCommand(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/setActionCommand(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getActionCommand()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Button/getActionCommand()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "java/awt/Button/getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/Button/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Button/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processActionEvent(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/processActionEvent(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Button/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Button/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Button/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Button/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_ButtonPeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "setLabel(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Desktop extends java_lang_Object {
    public "java/awt/Desktop/peer": java_awt_peer_DesktopPeer;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getDesktop()Ljava/awt/Desktop;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Desktop) => void): void;
    public static "java/awt/Desktop/getDesktop()Ljava/awt/Desktop;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Desktop) => void): void;
    public static "isDesktopSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Desktop/isDesktopSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupported(Ljava/awt/Desktop$Action;)Z"(thread: JVMThread, args: [java_awt_Desktop$Action], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Desktop/isSupported(Ljava/awt/Desktop$Action;)Z"(thread: JVMThread, args: [java_awt_Desktop$Action], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkFileValidation(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Desktop/checkFileValidation(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkActionSupport(Ljava/awt/Desktop$Action;)V"(thread: JVMThread, args: [java_awt_Desktop$Action], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/checkActionSupport(Ljava/awt/Desktop$Action;)V"(thread: JVMThread, args: [java_awt_Desktop$Action], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAWTPermission()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/checkAWTPermission()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "open(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/open(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "edit(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/edit(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/print(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "browse(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/browse(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
    public "mail()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/mail()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mail(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/mail(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkExec()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop/checkExec()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Desktop$Action extends java_lang_Enum {
    public static "java/awt/Desktop$Action/OPEN": java_awt_Desktop$Action;
    public static "java/awt/Desktop$Action/EDIT": java_awt_Desktop$Action;
    public static "java/awt/Desktop$Action/PRINT": java_awt_Desktop$Action;
    public static "java/awt/Desktop$Action/MAIL": java_awt_Desktop$Action;
    public static "java/awt/Desktop$Action/BROWSE": java_awt_Desktop$Action;
    public static "java/awt/Desktop$Action/$VALUES": JVMArray<java_awt_Desktop$Action>;
    public static "values()[Ljava/awt/Desktop$Action;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Desktop$Action>) => void): void;
    public static "java/awt/Desktop$Action/values()[Ljava/awt/Desktop$Action;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Desktop$Action>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/awt/Desktop$Action;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Desktop$Action) => void): void;
    public static "java/awt/Desktop$Action/valueOf(Ljava/lang/String;)Ljava/awt/Desktop$Action;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Desktop$Action) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Desktop$Action/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Desktop$Action/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_DesktopPeer extends java_lang_Object {
    "isSupported(Ljava/awt/Desktop$Action;)Z"(thread: JVMThread, args: [java_awt_Desktop$Action], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "open(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    "edit(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    "mail(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
    "browse(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_WeakHashMap extends java_util_AbstractMap implements java_util_Map {
    public static "java/util/WeakHashMap/DEFAULT_INITIAL_CAPACITY": number;
    public static "java/util/WeakHashMap/MAXIMUM_CAPACITY": number;
    public static "java/util/WeakHashMap/DEFAULT_LOAD_FACTOR": number;
    public "java/util/WeakHashMap/table": JVMArray<java_util_WeakHashMap$Entry>;
    public "java/util/WeakHashMap/size": number;
    public "java/util/WeakHashMap/threshold": number;
    public "java/util/WeakHashMap/loadFactor": number;
    public "java/util/WeakHashMap/queue": java_lang_ref_ReferenceQueue;
    public "java/util/WeakHashMap/modCount": number;
    public static "java/util/WeakHashMap/NULL_KEY": java_lang_Object;
    public "java/util/WeakHashMap/entrySet": java_util_Set;
    public "newTable(I)[Ljava/util/WeakHashMap$Entry;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_WeakHashMap$Entry>) => void): void;
    public "java/util/WeakHashMap/newTable(I)[Ljava/util/WeakHashMap$Entry;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_WeakHashMap$Entry>) => void): void;
    public "<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public static "maskNull(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/util/WeakHashMap/maskNull(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "unmaskNull(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/util/WeakHashMap/unmaskNull(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "eq(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/WeakHashMap/eq(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hash(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/WeakHashMap/hash(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "indexFor(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/WeakHashMap/indexFor(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "expungeStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/expungeStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTable()[Ljava/util/WeakHashMap$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_WeakHashMap$Entry>) => void): void;
    public "java/util/WeakHashMap/getTable()[Ljava/util/WeakHashMap$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_WeakHashMap$Entry>) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/WeakHashMap/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/WeakHashMap/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/WeakHashMap/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/WeakHashMap/containsKey(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getEntry(Ljava/lang/Object;)Ljava/util/WeakHashMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_WeakHashMap$Entry) => void): void;
    public "java/util/WeakHashMap/getEntry(Ljava/lang/Object;)Ljava/util/WeakHashMap$Entry;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_util_WeakHashMap$Entry) => void): void;
    public "put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/WeakHashMap/put(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "resize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/resize(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "transfer([Ljava/util/WeakHashMap$Entry;[Ljava/util/WeakHashMap$Entry;)V"(thread: JVMThread, args: [JVMArray<java_util_WeakHashMap$Entry>, JVMArray<java_util_WeakHashMap$Entry>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/transfer([Ljava/util/WeakHashMap$Entry;[Ljava/util/WeakHashMap$Entry;)V"(thread: JVMThread, args: [JVMArray<java_util_WeakHashMap$Entry>, JVMArray<java_util_WeakHashMap$Entry>], cb?: (e?: java_lang_Throwable) => void): void;
    public "putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/putAll(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/WeakHashMap/remove(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "removeMapping(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/WeakHashMap/removeMapping(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/WeakHashMap/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsNullValue()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/WeakHashMap/containsNullValue()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/WeakHashMap/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/WeakHashMap/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/WeakHashMap/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap/replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/WeakHashMap/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_WeakHashMap$Entry extends java_lang_ref_WeakReference implements java_util_Map$Entry {
    public "java/util/WeakHashMap$Entry/value": java_lang_Object;
    public "java/util/WeakHashMap$Entry/hash": number;
    public "java/util/WeakHashMap$Entry/next": java_util_WeakHashMap$Entry;
    public "<init>(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;ILjava/util/WeakHashMap$Entry;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_ref_ReferenceQueue, number, java_util_WeakHashMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/WeakHashMap$Entry/<init>(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;ILjava/util/WeakHashMap$Entry;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_ref_ReferenceQueue, number, java_util_WeakHashMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/WeakHashMap$Entry/getKey()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/WeakHashMap$Entry/getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/WeakHashMap$Entry/setValue(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/WeakHashMap$Entry/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/WeakHashMap$Entry/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/WeakHashMap$Entry/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_awt_event_AWTEventListener extends java_lang_Object, java_util_EventListener {
    "eventDispatched(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_LightweightPeer extends java_lang_Object, java_awt_peer_ComponentPeer {
    "isObscured()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "canDetermineObscurity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    "setBounds(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "handleEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "coalescePaintEvent(Ljava/awt/event/PaintEvent;)V"(thread: JVMThread, args: [java_awt_event_PaintEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "requestFocus(Ljava/awt/Component;ZZJLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, number, number, Long, any, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    "handlesWheelScrolling()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createBuffers(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "flip(IIIILjava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    "destroyBuffers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "reparent(Ljava/awt/peer/ContainerPeer;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "isReparentSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "applyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    "setZOrder(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_Frame extends java_awt_Window implements java_awt_MenuContainer {
    public static "java/awt/Frame/DEFAULT_CURSOR": number;
    public static "java/awt/Frame/CROSSHAIR_CURSOR": number;
    public static "java/awt/Frame/TEXT_CURSOR": number;
    public static "java/awt/Frame/WAIT_CURSOR": number;
    public static "java/awt/Frame/SW_RESIZE_CURSOR": number;
    public static "java/awt/Frame/SE_RESIZE_CURSOR": number;
    public static "java/awt/Frame/NW_RESIZE_CURSOR": number;
    public static "java/awt/Frame/NE_RESIZE_CURSOR": number;
    public static "java/awt/Frame/N_RESIZE_CURSOR": number;
    public static "java/awt/Frame/S_RESIZE_CURSOR": number;
    public static "java/awt/Frame/W_RESIZE_CURSOR": number;
    public static "java/awt/Frame/E_RESIZE_CURSOR": number;
    public static "java/awt/Frame/HAND_CURSOR": number;
    public static "java/awt/Frame/MOVE_CURSOR": number;
    public static "java/awt/Frame/NORMAL": number;
    public static "java/awt/Frame/ICONIFIED": number;
    public static "java/awt/Frame/MAXIMIZED_HORIZ": number;
    public static "java/awt/Frame/MAXIMIZED_VERT": number;
    public static "java/awt/Frame/MAXIMIZED_BOTH": number;
    public "java/awt/Frame/maximizedBounds": java_awt_Rectangle;
    public "java/awt/Frame/title": java_lang_String;
    public "java/awt/Frame/menuBar": java_awt_MenuBar;
    public "java/awt/Frame/resizable": number;
    public "java/awt/Frame/undecorated": number;
    public "java/awt/Frame/mbManagement": number;
    public "java/awt/Frame/state": number;
    public "java/awt/Frame/ownedWindows": java_util_Vector;
    public static "java/awt/Frame/base": java_lang_String;
    public static "java/awt/Frame/nameCounter": number;
    public static "java/awt/Frame/serialVersionUID": Long;
    public "java/awt/Frame/frameSerializedDataVersion": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/<init>(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_lang_String, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/<init>(Ljava/lang/String;Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_lang_String, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "init(Ljava/lang/String;Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_lang_String, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/init(Ljava/lang/String;Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_lang_String, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Frame/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTitle()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Frame/getTitle()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setTitle(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setTitle(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getIconImage()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Frame/getIconImage()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "setIconImage(Ljava/awt/Image;)V"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setIconImage(Ljava/awt/Image;)V"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMenuBar()Ljava/awt/MenuBar;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuBar) => void): void;
    public "java/awt/Frame/getMenuBar()Ljava/awt/MenuBar;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_MenuBar) => void): void;
    public "setMenuBar(Ljava/awt/MenuBar;)V"(thread: JVMThread, args: [java_awt_MenuBar], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setMenuBar(Ljava/awt/MenuBar;)V"(thread: JVMThread, args: [java_awt_MenuBar], cb?: (e?: java_lang_Throwable) => void): void;
    public "isResizable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Frame/isResizable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setResizable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setResizable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setState(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setState(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setExtendedState(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setExtendedState(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isFrameStateSupported(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Frame/isFrameStateSupported(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Frame/getState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getExtendedState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Frame/getExtendedState()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setMaximizedBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setMaximizedBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMaximizedBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Frame/getMaximizedBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "setUndecorated(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setUndecorated(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isUndecorated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Frame/isUndecorated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setShape(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setShape(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/awt/MenuComponent;)V"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/remove(Ljava/awt/MenuComponent;)V"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "postProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/postProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Frame/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setCursor(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/setCursor(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCursorType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Frame/getCursorType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFrames()[Ljava/awt/Frame;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Frame>) => void): void;
    public static "java/awt/Frame/getFrames()[Ljava/awt/Frame;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Frame>) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Frame/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Frame/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Frame/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "access$002(Ljava/awt/Frame;I)I"(thread: JVMThread, args: [java_awt_Frame, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Frame/access$002(Ljava/awt/Frame;I)I"(thread: JVMThread, args: [java_awt_Frame, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/awt/Frame;)I"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Frame/access$000(Ljava/awt/Frame;)I"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Frame/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Window$Type extends java_lang_Enum {
    public static "java/awt/Window$Type/NORMAL": java_awt_Window$Type;
    public static "java/awt/Window$Type/UTILITY": java_awt_Window$Type;
    public static "java/awt/Window$Type/POPUP": java_awt_Window$Type;
    public static "java/awt/Window$Type/$VALUES": JVMArray<java_awt_Window$Type>;
    public static "values()[Ljava/awt/Window$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window$Type>) => void): void;
    public static "java/awt/Window$Type/values()[Ljava/awt/Window$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Window$Type>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/awt/Window$Type;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window$Type) => void): void;
    public static "java/awt/Window$Type/valueOf(Ljava/lang/String;)Ljava/awt/Window$Type;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window$Type) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window$Type/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Window$Type/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_FocusManager extends java_lang_Object implements java_io_Serializable {
    public "java/awt/FocusManager/focusRoot": java_awt_Container;
    public "java/awt/FocusManager/focusOwner": java_awt_Component;
    public static "java/awt/FocusManager/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FocusManager/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_im_InputContext extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputContext/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance()Ljava/awt/im/InputContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputContext) => void): void;
    public static "java/awt/im/InputContext/getInstance()Ljava/awt/im/InputContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputContext) => void): void;
    public "selectInputMethod(Ljava/util/Locale;)Z"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/im/InputContext/selectInputMethod(Ljava/util/Locale;)Z"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/awt/im/InputContext/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "setCharacterSubsets([Ljava/lang/Character$Subset;)V"(thread: JVMThread, args: [JVMArray<java_lang_Character$Subset>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputContext/setCharacterSubsets([Ljava/lang/Character$Subset;)V"(thread: JVMThread, args: [JVMArray<java_lang_Character$Subset>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCompositionEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputContext/setCompositionEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isCompositionEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/im/InputContext/isCompositionEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reconvert()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputContext/reconvert()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputContext/dispatchEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputContext/removeNotify(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "endComposition()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputContext/endComposition()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/im/InputContext/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInputMethodControlObject()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/im/InputContext/getInputMethodControlObject()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_lang_Character$Subset extends java_lang_Object {
    public "java/lang/Character$Subset/name": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Character$Subset/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Character$Subset/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Character$Subset/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Character$Subset/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_awt_event_WindowFocusListener extends java_lang_Object, java_util_EventListener {
    "windowGainedFocus(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "windowLostFocus(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_WindowStateListener extends java_lang_Object, java_util_EventListener {
    "windowStateChanged(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_WindowListener extends java_lang_Object, java_util_EventListener {
    "windowOpened(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "windowClosing(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "windowClosed(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "windowIconified(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "windowDeiconified(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "windowActivated(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "windowDeactivated(Ljava/awt/event/WindowEvent;)V"(thread: JVMThread, args: [java_awt_event_WindowEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Dialog$ModalExclusionType extends java_lang_Enum {
    public static "java/awt/Dialog$ModalExclusionType/NO_EXCLUDE": java_awt_Dialog$ModalExclusionType;
    public static "java/awt/Dialog$ModalExclusionType/APPLICATION_EXCLUDE": java_awt_Dialog$ModalExclusionType;
    public static "java/awt/Dialog$ModalExclusionType/TOOLKIT_EXCLUDE": java_awt_Dialog$ModalExclusionType;
    public static "java/awt/Dialog$ModalExclusionType/$VALUES": JVMArray<java_awt_Dialog$ModalExclusionType>;
    public static "values()[Ljava/awt/Dialog$ModalExclusionType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Dialog$ModalExclusionType>) => void): void;
    public static "java/awt/Dialog$ModalExclusionType/values()[Ljava/awt/Dialog$ModalExclusionType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Dialog$ModalExclusionType>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/awt/Dialog$ModalExclusionType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog$ModalExclusionType) => void): void;
    public static "java/awt/Dialog$ModalExclusionType/valueOf(Ljava/lang/String;)Ljava/awt/Dialog$ModalExclusionType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog$ModalExclusionType) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog$ModalExclusionType/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Dialog$ModalExclusionType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Dialog extends java_awt_Window {
    public "java/awt/Dialog/resizable": number;
    public "java/awt/Dialog/undecorated": number;
    public "java/awt/Dialog/initialized": number;
    public static "java/awt/Dialog/DEFAULT_MODALITY_TYPE": java_awt_Dialog$ModalityType;
    public "java/awt/Dialog/modal": number;
    public "java/awt/Dialog/modalityType": java_awt_Dialog$ModalityType;
    public static "java/awt/Dialog/modalDialogs": sun_awt_util_IdentityArrayList;
    public "java/awt/Dialog/blockedWindows": sun_awt_util_IdentityArrayList;
    public "java/awt/Dialog/title": java_lang_String;
    public "java/awt/Dialog/modalFilter": java_awt_ModalEventFilter;
    public "java/awt/Dialog/secondaryLoop": java_awt_SecondaryLoop;
    public "java/awt/Dialog/isInHide": number;
    public "java/awt/Dialog/isInDispose": number;
    public static "java/awt/Dialog/base": java_lang_String;
    public static "java/awt/Dialog/nameCounter": number;
    public static "java/awt/Dialog/serialVersionUID": Long;
    public "<init>(Ljava/awt/Frame;)V"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Frame;)V"(thread: JVMThread, args: [java_awt_Frame], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Frame;Z)V"(thread: JVMThread, args: [java_awt_Frame, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Frame;Z)V"(thread: JVMThread, args: [java_awt_Frame, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Frame;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Frame;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Frame;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Frame;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Frame;Ljava/lang/String;ZLjava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, number, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Frame;Ljava/lang/String;ZLjava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_Frame, java_lang_String, number, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Dialog;)V"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Dialog;)V"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Dialog;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Dialog;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Dialog;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Dialog;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Dialog;Ljava/lang/String;ZLjava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String, number, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Dialog;Ljava/lang/String;ZLjava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_Dialog, java_lang_String, number, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Window, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Window;Ljava/lang/String;)V"(thread: JVMThread, args: [java_awt_Window, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;Ljava/awt/Dialog$ModalityType;)V"(thread: JVMThread, args: [java_awt_Window, java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Window;Ljava/awt/Dialog$ModalityType;)V"(thread: JVMThread, args: [java_awt_Window, java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;Ljava/lang/String;Ljava/awt/Dialog$ModalityType;)V"(thread: JVMThread, args: [java_awt_Window, java_lang_String, java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Window;Ljava/lang/String;Ljava/awt/Dialog$ModalityType;)V"(thread: JVMThread, args: [java_awt_Window, java_lang_String, java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Window;Ljava/lang/String;Ljava/awt/Dialog$ModalityType;Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_Window, java_lang_String, java_awt_Dialog$ModalityType, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/<init>(Ljava/awt/Window;Ljava/lang/String;Ljava/awt/Dialog$ModalityType;Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_Window, java_lang_String, java_awt_Dialog$ModalityType, java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Dialog/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isModal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dialog/isModal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isModal_NoClientCode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dialog/isModal_NoClientCode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setModal(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/setModal(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getModalityType()Ljava/awt/Dialog$ModalityType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog$ModalityType) => void): void;
    public "java/awt/Dialog/getModalityType()Ljava/awt/Dialog$ModalityType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog$ModalityType) => void): void;
    public "setModalityType(Ljava/awt/Dialog$ModalityType;)V"(thread: JVMThread, args: [java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/setModalityType(Ljava/awt/Dialog$ModalityType;)V"(thread: JVMThread, args: [java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTitle()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Dialog/getTitle()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setTitle(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/setTitle(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "conditionalShow(Ljava/awt/Component;Ljava/util/concurrent/atomic/AtomicLong;)Z"(thread: JVMThread, args: [java_awt_Component, java_util_concurrent_atomic_AtomicLong], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dialog/conditionalShow(Ljava/awt/Component;Ljava/util/concurrent/atomic/AtomicLong;)Z"(thread: JVMThread, args: [java_awt_Component, java_util_concurrent_atomic_AtomicLong], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "show()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/show()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "modalityPushed()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/modalityPushed()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "modalityPopped()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/modalityPopped()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interruptBlocking()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/interruptBlocking()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "hideAndDisposePreHandler()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/hideAndDisposePreHandler()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "hideAndDisposeHandler()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/hideAndDisposeHandler()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "hide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/hide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "doDispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/doDispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toBack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/toBack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isResizable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dialog/isResizable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setResizable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/setResizable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setUndecorated(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/setUndecorated(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isUndecorated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dialog/isUndecorated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/setOpacity(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setShape(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/setShape(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Dialog/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Dialog/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "modalShow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/modalShow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "modalHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/modalHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shouldBlock(Ljava/awt/Window;)Z"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Dialog/shouldBlock(Ljava/awt/Window;)Z"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "blockWindow(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/blockWindow(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "blockWindows(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/blockWindows(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "unblockWindow(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/unblockWindow(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkShouldBeBlocked(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Dialog/checkShouldBeBlocked(Ljava/awt/Window;)V"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkModalityPermission(Ljava/awt/Dialog$ModalityType;)V"(thread: JVMThread, args: [java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/checkModalityPermission(Ljava/awt/Dialog$ModalityType;)V"(thread: JVMThread, args: [java_awt_Dialog$ModalityType], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Dialog/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Dialog/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Dialog/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_ModalEventFilter extends java_lang_Object implements java_awt_EventFilter {
    public "java/awt/ModalEventFilter/modalDialog": java_awt_Dialog;
    public "java/awt/ModalEventFilter/disabled": number;
    public "<init>(Ljava/awt/Dialog;)V"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ModalEventFilter/<init>(Ljava/awt/Dialog;)V"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable) => void): void;
    public "getModalDialog()Ljava/awt/Dialog;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog) => void): void;
    public "java/awt/ModalEventFilter/getModalDialog()Ljava/awt/Dialog;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dialog) => void): void;
    public "acceptEvent(Ljava/awt/AWTEvent;)Ljava/awt/EventFilter$FilterAction;"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventFilter$FilterAction) => void): void;
    public "java/awt/ModalEventFilter/acceptEvent(Ljava/awt/AWTEvent;)Ljava/awt/EventFilter$FilterAction;"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventFilter$FilterAction) => void): void;
    public "acceptWindow(Ljava/awt/Window;)Ljava/awt/EventFilter$FilterAction;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventFilter$FilterAction) => void): void;
    public "java/awt/ModalEventFilter/acceptWindow(Ljava/awt/Window;)Ljava/awt/EventFilter$FilterAction;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_awt_EventFilter$FilterAction) => void): void;
    public "disable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ModalEventFilter/disable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "compareTo(Ljava/awt/ModalEventFilter;)I"(thread: JVMThread, args: [java_awt_ModalEventFilter], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ModalEventFilter/compareTo(Ljava/awt/ModalEventFilter;)I"(thread: JVMThread, args: [java_awt_ModalEventFilter], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "createFilterForDialog(Ljava/awt/Dialog;)Ljava/awt/ModalEventFilter;"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable, rv?: java_awt_ModalEventFilter) => void): void;
    public static "java/awt/ModalEventFilter/createFilterForDialog(Ljava/awt/Dialog;)Ljava/awt/ModalEventFilter;"(thread: JVMThread, args: [java_awt_Dialog], cb?: (e?: java_lang_Throwable, rv?: java_awt_ModalEventFilter) => void): void;
  }
  export class sun_awt_util_IdentityArrayList extends java_util_AbstractList implements java_util_List, java_util_RandomAccess {
    public "sun/awt/util/IdentityArrayList/elementData": JVMArray<java_lang_Object>;
    public "sun/awt/util/IdentityArrayList/size": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/trimToSize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/util/IdentityArrayList/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/util/IdentityArrayList/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/util/IdentityArrayList/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/util/IdentityArrayList/indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/util/IdentityArrayList/lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "sun/awt/util/IdentityArrayList/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "sun/awt/util/IdentityArrayList/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/awt/util/IdentityArrayList/get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/awt/util/IdentityArrayList/set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/util/IdentityArrayList/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/awt/util/IdentityArrayList/remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/util/IdentityArrayList/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fastRemove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/fastRemove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/util/IdentityArrayList/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/util/IdentityArrayList/addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rangeCheck(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/rangeCheck(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rangeCheckForAdd(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/util/IdentityArrayList/rangeCheckForAdd(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "outOfBoundsMsg(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/util/IdentityArrayList/outOfBoundsMsg(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Window$WindowDisposerRecord extends java_lang_Object implements sun_java2d_DisposerRecord {
    public "java/awt/Window$WindowDisposerRecord/owner": java_lang_ref_WeakReference;
    public "java/awt/Window$WindowDisposerRecord/weakThis": java_lang_ref_WeakReference;
    public "java/awt/Window$WindowDisposerRecord/context": java_lang_ref_WeakReference;
    public "<init>(Lsun/awt/AppContext;Ljava/awt/Window;)V"(thread: JVMThread, args: [sun_awt_AppContext, java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window$WindowDisposerRecord/<init>(Lsun/awt/AppContext;Ljava/awt/Window;)V"(thread: JVMThread, args: [sun_awt_AppContext, java_awt_Window], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateOwner()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window$WindowDisposerRecord/updateOwner()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Window$WindowDisposerRecord/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_Component extends java_lang_Object implements java_awt_image_ImageObserver, java_awt_MenuContainer, java_io_Serializable {
    public static "java/awt/Component/log": sun_util_logging_PlatformLogger;
    public static "java/awt/Component/eventLog": sun_util_logging_PlatformLogger;
    public static "java/awt/Component/focusLog": sun_util_logging_PlatformLogger;
    public static "java/awt/Component/mixingLog": sun_util_logging_PlatformLogger;
    public "java/awt/Component/peer": java_awt_peer_ComponentPeer;
    public "java/awt/Component/parent": java_awt_Container;
    public "java/awt/Component/appContext": sun_awt_AppContext;
    public "java/awt/Component/x": number;
    public "java/awt/Component/y": number;
    public "java/awt/Component/width": number;
    public "java/awt/Component/height": number;
    public "java/awt/Component/foreground": java_awt_Color;
    public "java/awt/Component/background": java_awt_Color;
    public "java/awt/Component/font": java_awt_Font;
    public "java/awt/Component/peerFont": java_awt_Font;
    public "java/awt/Component/cursor": java_awt_Cursor;
    public "java/awt/Component/locale": java_util_Locale;
    public "java/awt/Component/graphicsConfig": java_awt_GraphicsConfiguration;
    public "java/awt/Component/bufferStrategy": java_awt_image_BufferStrategy;
    public "java/awt/Component/ignoreRepaint": number;
    public "java/awt/Component/visible": number;
    public "java/awt/Component/enabled": number;
    public "java/awt/Component/valid": number;
    public "java/awt/Component/dropTarget": java_awt_dnd_DropTarget;
    public "java/awt/Component/popups": java_util_Vector;
    public "java/awt/Component/name": java_lang_String;
    public "java/awt/Component/nameExplicitlySet": number;
    public "java/awt/Component/focusable": number;
    public static "java/awt/Component/FOCUS_TRAVERSABLE_UNKNOWN": number;
    public static "java/awt/Component/FOCUS_TRAVERSABLE_DEFAULT": number;
    public static "java/awt/Component/FOCUS_TRAVERSABLE_SET": number;
    public "java/awt/Component/isFocusTraversableOverridden": number;
    public "java/awt/Component/focusTraversalKeys": JVMArray<java_util_Set>;
    public static "java/awt/Component/focusTraversalKeyPropertyNames": JVMArray<java_lang_String>;
    public "java/awt/Component/focusTraversalKeysEnabled": number;
    public static "java/awt/Component/LOCK": java_lang_Object;
    public "java/awt/Component/acc": java_security_AccessControlContext;
    public "java/awt/Component/minSize": java_awt_Dimension;
    public "java/awt/Component/minSizeSet": number;
    public "java/awt/Component/prefSize": java_awt_Dimension;
    public "java/awt/Component/prefSizeSet": number;
    public "java/awt/Component/maxSize": java_awt_Dimension;
    public "java/awt/Component/maxSizeSet": number;
    public "java/awt/Component/componentOrientation": java_awt_ComponentOrientation;
    public "java/awt/Component/newEventsOnly": number;
    public "java/awt/Component/componentListener": java_awt_event_ComponentListener;
    public "java/awt/Component/focusListener": java_awt_event_FocusListener;
    public "java/awt/Component/hierarchyListener": java_awt_event_HierarchyListener;
    public "java/awt/Component/hierarchyBoundsListener": java_awt_event_HierarchyBoundsListener;
    public "java/awt/Component/keyListener": java_awt_event_KeyListener;
    public "java/awt/Component/mouseListener": java_awt_event_MouseListener;
    public "java/awt/Component/mouseMotionListener": java_awt_event_MouseMotionListener;
    public "java/awt/Component/mouseWheelListener": java_awt_event_MouseWheelListener;
    public "java/awt/Component/inputMethodListener": java_awt_event_InputMethodListener;
    public "java/awt/Component/windowClosingException": java_lang_RuntimeException;
    public static "java/awt/Component/actionListenerK": java_lang_String;
    public static "java/awt/Component/adjustmentListenerK": java_lang_String;
    public static "java/awt/Component/componentListenerK": java_lang_String;
    public static "java/awt/Component/containerListenerK": java_lang_String;
    public static "java/awt/Component/focusListenerK": java_lang_String;
    public static "java/awt/Component/itemListenerK": java_lang_String;
    public static "java/awt/Component/keyListenerK": java_lang_String;
    public static "java/awt/Component/mouseListenerK": java_lang_String;
    public static "java/awt/Component/mouseMotionListenerK": java_lang_String;
    public static "java/awt/Component/mouseWheelListenerK": java_lang_String;
    public static "java/awt/Component/textListenerK": java_lang_String;
    public static "java/awt/Component/ownedWindowK": java_lang_String;
    public static "java/awt/Component/windowListenerK": java_lang_String;
    public static "java/awt/Component/inputMethodListenerK": java_lang_String;
    public static "java/awt/Component/hierarchyListenerK": java_lang_String;
    public static "java/awt/Component/hierarchyBoundsListenerK": java_lang_String;
    public static "java/awt/Component/windowStateListenerK": java_lang_String;
    public static "java/awt/Component/windowFocusListenerK": java_lang_String;
    public "java/awt/Component/eventMask": Long;
    public static "java/awt/Component/isInc": number;
    public static "java/awt/Component/incRate": number;
    public static "java/awt/Component/TOP_ALIGNMENT": number;
    public static "java/awt/Component/CENTER_ALIGNMENT": number;
    public static "java/awt/Component/BOTTOM_ALIGNMENT": number;
    public static "java/awt/Component/LEFT_ALIGNMENT": number;
    public static "java/awt/Component/RIGHT_ALIGNMENT": number;
    public static "java/awt/Component/serialVersionUID": Long;
    public "java/awt/Component/changeSupport": java_beans_PropertyChangeSupport;
    public "java/awt/Component/objectLock": java_lang_Object;
    public "java/awt/Component/isPacked": number;
    public "java/awt/Component/boundsOp": number;
    public "java/awt/Component/compoundShape": sun_java2d_pipe_Region;
    public "java/awt/Component/mixingCutoutRegion": sun_java2d_pipe_Region;
    public "java/awt/Component/isAddNotifyComplete": number;
    public "java/awt/Component/backgroundEraseDisabled": number;
    public "java/awt/Component/eventCache": JVMArray<sun_awt_EventQueueItem>;
    public "java/awt/Component/coalescingEnabled": number;
    public static "java/awt/Component/coalesceMap": java_util_Map;
    public static "java/awt/Component/coalesceEventsParams": JVMArray<java_lang_Class>;
    public static "java/awt/Component/requestFocusController": sun_awt_RequestFocusController;
    public "java/awt/Component/autoFocusTransferOnDisposal": number;
    public "java/awt/Component/componentSerializedDataVersion": number;
    public "java/awt/Component/accessibleContext": javax_accessibility_AccessibleContext;
    public static "java/awt/Component/$assertionsDisabled": number;
    public "getObjectLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/Component/getObjectLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "java/awt/Component/getAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "getBoundsOp()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getBoundsOp()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setBoundsOp(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setBoundsOp(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "initializeFocusTraversalKeys()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/initializeFocusTraversalKeys()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Component/constructComponentName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Component/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getParent()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Component/getParent()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "getParent_NoClientCode()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Component/getParent_NoClientCode()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "getContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Component/getContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "getPeer()Ljava/awt/peer/ComponentPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ComponentPeer) => void): void;
    public "java/awt/Component/getPeer()Ljava/awt/peer/ComponentPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ComponentPeer) => void): void;
    public "setDropTarget(Ljava/awt/dnd/DropTarget;)V"(thread: JVMThread, args: [java_awt_dnd_DropTarget], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setDropTarget(Ljava/awt/dnd/DropTarget;)V"(thread: JVMThread, args: [java_awt_dnd_DropTarget], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDropTarget()Ljava/awt/dnd/DropTarget;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTarget) => void): void;
    public "java/awt/Component/getDropTarget()Ljava/awt/dnd/DropTarget;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTarget) => void): void;
    public "getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "java/awt/Component/getGraphicsConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "getGraphicsConfiguration_NoClientCode()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "java/awt/Component/getGraphicsConfiguration_NoClientCode()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "setGraphicsConfiguration(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setGraphicsConfiguration(Ljava/awt/GraphicsConfiguration;)V"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkGD(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/checkGD(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTreeLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/Component/getTreeLock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "checkTreeLock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/checkTreeLock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getToolkit()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public "java/awt/Component/getToolkit()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public "getToolkitImpl()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public "java/awt/Component/getToolkitImpl()Ljava/awt/Toolkit;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Toolkit) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDisplayable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isDisplayable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isVisible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isVisible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isVisible_NoClientCode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isVisible_NoClientCode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRecursivelyVisible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isRecursivelyVisible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRecursivelyVisibleBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Component/getRecursivelyVisibleBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "pointRelativeToComponent(Ljava/awt/Point;)Ljava/awt/Point;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Component/pointRelativeToComponent(Ljava/awt/Point;)Ljava/awt/Point;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "findUnderMouseInWindow(Ljava/awt/PointerInfo;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_PointerInfo], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Component/findUnderMouseInWindow(Ljava/awt/PointerInfo;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_PointerInfo], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getMousePosition()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Component/getMousePosition()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "isSameOrAncestorOf(Ljava/awt/Component;Z)Z"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isSameOrAncestorOf(Ljava/awt/Component;Z)Z"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isShowing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isShowing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEnabledImpl()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isEnabledImpl()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "enable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/enable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "enable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/enable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "disable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/disable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isDoubleBuffered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isDoubleBuffered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "enableInputMethods(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/enableInputMethods(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setVisible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "show()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/show()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "show(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/show(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "containsFocus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/containsFocus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clearMostRecentFocusOwnerOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/clearMostRecentFocusOwnerOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearCurrentFocusCycleRootOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/clearCurrentFocusCycleRootOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "hide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/hide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getForeground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "java/awt/Component/getForeground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setForeground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "isForegroundSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isForegroundSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "java/awt/Component/getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "isBackgroundSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isBackgroundSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/Component/getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "getFont_NoClientCode()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/Component/getFont_NoClientCode()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "isFontSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isFontSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/awt/Component/getLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "setLocale(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setLocale(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "java/awt/Component/getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Component/getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Component/getLocationOnScreen()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "getLocationOnScreen_NoTreeLock()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Component/getLocationOnScreen_NoTreeLock()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "location()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Component/location()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "location_NoClientCode()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Component/location_NoClientCode()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "setLocation(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setLocation(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "move(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/move(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Component/getSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "size()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Component/size()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "setSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setSize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "resize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/resize(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "resize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/resize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Component/getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "bounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Component/bounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "setBounds(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setBounds(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "reshape(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/reshape(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "repaintParentIfNeeded(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/repaintParentIfNeeded(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "reshapeNativePeer(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/reshapeNativePeer(IIIII)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "notifyNewBounds(ZZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/notifyNewBounds(ZZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setBounds(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Component/getBounds(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getSize(Ljava/awt/Dimension;)Ljava/awt/Dimension;"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Component/getSize(Ljava/awt/Dimension;)Ljava/awt/Dimension;"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getLocation(Ljava/awt/Point;)Ljava/awt/Point;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Component/getLocation(Ljava/awt/Point;)Ljava/awt/Point;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "isOpaque()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isOpaque()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLightweight()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isLightweight()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setPreferredSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setPreferredSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "isPreferredSizeSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isPreferredSizeSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Component/getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Component/preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "setMinimumSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setMinimumSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "isMinimumSizeSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isMinimumSizeSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Component/getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Component/minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "setMaximumSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setMaximumSize(Ljava/awt/Dimension;)V"(thread: JVMThread, args: [java_awt_Dimension], cb?: (e?: java_lang_Throwable) => void): void;
    public "isMaximumSizeSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isMaximumSizeSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaximumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Component/getMaximumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getAlignmentX()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getAlignmentX()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAlignmentY()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getAlignmentY()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBaseline(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getBaseline(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBaselineResizeBehavior()Ljava/awt/Component$BaselineResizeBehavior;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component$BaselineResizeBehavior) => void): void;
    public "java/awt/Component/getBaselineResizeBehavior()Ljava/awt/Component$BaselineResizeBehavior;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component$BaselineResizeBehavior) => void): void;
    public "doLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/doLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "validate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/validate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidateParent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/invalidateParent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidateIfValid()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/invalidateIfValid()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "revalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/revalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "revalidateSynchronously()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/revalidateSynchronously()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "java/awt/Component/getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "getGraphics_NoClientCode()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "java/awt/Component/getGraphics_NoClientCode()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "java/awt/Component/getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "setCursor(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setCursor(Ljava/awt/Cursor;)V"(thread: JVMThread, args: [java_awt_Cursor], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/updateCursorImmediately()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCursor()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public "java/awt/Component/getCursor()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public "getCursor_NoClientCode()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public "java/awt/Component/getCursor_NoClientCode()Ljava/awt/Cursor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Cursor) => void): void;
    public "isCursorSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isCursorSet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/update(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "paintAll(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/paintAll(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "lightweightPaint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/lightweightPaint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "paintHeavyweightComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/paintHeavyweightComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "repaint()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/repaint()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "repaint(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/repaint(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "repaint(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/repaint(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "repaint(JIIII)V"(thread: JVMThread, args: [Long, any, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/repaint(JIIII)V"(thread: JVMThread, args: [Long, any, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "printAll(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/printAll(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "lightweightPrint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/lightweightPrint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "printHeavyweightComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/printHeavyweightComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInsets_NoClientCode()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public "java/awt/Component/getInsets_NoClientCode()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public "imageUpdate(Ljava/awt/Image;IIIII)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/imageUpdate(Ljava/awt/Image;IIIII)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Component/createImage(Ljava/awt/image/ImageProducer;)Ljava/awt/Image;"(thread: JVMThread, args: [java_awt_image_ImageProducer], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Component/createImage(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "java/awt/Component/createVolatileImage(II)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "createVolatileImage(IILjava/awt/ImageCapabilities;)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number, java_awt_ImageCapabilities], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "java/awt/Component/createVolatileImage(IILjava/awt/ImageCapabilities;)Ljava/awt/image/VolatileImage;"(thread: JVMThread, args: [number, number, java_awt_ImageCapabilities], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_VolatileImage) => void): void;
    public "prepareImage(Ljava/awt/Image;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/prepareImage(Ljava/awt/Image;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/prepareImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkImage(Ljava/awt/Image;Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/checkImage(Ljava/awt/Image;Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/checkImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createBufferStrategy(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/createBufferStrategy(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "createBufferStrategy(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/createBufferStrategy(ILjava/awt/BufferCapabilities;)V"(thread: JVMThread, args: [number, java_awt_BufferCapabilities], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBufferStrategy()Ljava/awt/image/BufferStrategy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferStrategy) => void): void;
    public "java/awt/Component/getBufferStrategy()Ljava/awt/image/BufferStrategy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferStrategy) => void): void;
    public "getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Component/getBackBuffer()Ljava/awt/Image;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "setIgnoreRepaint(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setIgnoreRepaint(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getIgnoreRepaint()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getIgnoreRepaint()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/contains(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inside(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/inside(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/awt/Point;)Z"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/contains(Ljava/awt/Point;)Z"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getComponentAt(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Component/getComponentAt(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "locate(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Component/locate(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getComponentAt(Ljava/awt/Point;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Component/getComponentAt(Ljava/awt/Point;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "deliverEvent(Ljava/awt/Event;)V"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/deliverEvent(Ljava/awt/Event;)V"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/dispatchEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchEventImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/dispatchEventImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "autoProcessMouseWheel(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/autoProcessMouseWheel(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchMouseWheelToAncestor(Ljava/awt/event/MouseWheelEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/dispatchMouseWheelToAncestor(Ljava/awt/event/MouseWheelEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkWindowClosingException()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/checkWindowClosingException()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "areInputMethodsEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/areInputMethodsEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "eventTypeEnabled(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/eventTypeEnabled(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "postEvent(Ljava/awt/Event;)Z"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/postEvent(Ljava/awt/Event;)Z"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addComponentListener(Ljava/awt/event/ComponentListener;)V"(thread: JVMThread, args: [java_awt_event_ComponentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addComponentListener(Ljava/awt/event/ComponentListener;)V"(thread: JVMThread, args: [java_awt_event_ComponentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeComponentListener(Ljava/awt/event/ComponentListener;)V"(thread: JVMThread, args: [java_awt_event_ComponentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeComponentListener(Ljava/awt/event/ComponentListener;)V"(thread: JVMThread, args: [java_awt_event_ComponentListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComponentListeners()[Ljava/awt/event/ComponentListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ComponentListener>) => void): void;
    public "java/awt/Component/getComponentListeners()[Ljava/awt/event/ComponentListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ComponentListener>) => void): void;
    public "addFocusListener(Ljava/awt/event/FocusListener;)V"(thread: JVMThread, args: [java_awt_event_FocusListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addFocusListener(Ljava/awt/event/FocusListener;)V"(thread: JVMThread, args: [java_awt_event_FocusListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeFocusListener(Ljava/awt/event/FocusListener;)V"(thread: JVMThread, args: [java_awt_event_FocusListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeFocusListener(Ljava/awt/event/FocusListener;)V"(thread: JVMThread, args: [java_awt_event_FocusListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFocusListeners()[Ljava/awt/event/FocusListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_FocusListener>) => void): void;
    public "java/awt/Component/getFocusListeners()[Ljava/awt/event/FocusListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_FocusListener>) => void): void;
    public "addHierarchyListener(Ljava/awt/event/HierarchyListener;)V"(thread: JVMThread, args: [java_awt_event_HierarchyListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addHierarchyListener(Ljava/awt/event/HierarchyListener;)V"(thread: JVMThread, args: [java_awt_event_HierarchyListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeHierarchyListener(Ljava/awt/event/HierarchyListener;)V"(thread: JVMThread, args: [java_awt_event_HierarchyListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeHierarchyListener(Ljava/awt/event/HierarchyListener;)V"(thread: JVMThread, args: [java_awt_event_HierarchyListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getHierarchyListeners()[Ljava/awt/event/HierarchyListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_HierarchyListener>) => void): void;
    public "java/awt/Component/getHierarchyListeners()[Ljava/awt/event/HierarchyListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_HierarchyListener>) => void): void;
    public "addHierarchyBoundsListener(Ljava/awt/event/HierarchyBoundsListener;)V"(thread: JVMThread, args: [java_awt_event_HierarchyBoundsListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addHierarchyBoundsListener(Ljava/awt/event/HierarchyBoundsListener;)V"(thread: JVMThread, args: [java_awt_event_HierarchyBoundsListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeHierarchyBoundsListener(Ljava/awt/event/HierarchyBoundsListener;)V"(thread: JVMThread, args: [java_awt_event_HierarchyBoundsListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeHierarchyBoundsListener(Ljava/awt/event/HierarchyBoundsListener;)V"(thread: JVMThread, args: [java_awt_event_HierarchyBoundsListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "numListening(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/numListening(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countHierarchyMembers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/countHierarchyMembers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createHierarchyEvents(ILjava/awt/Component;Ljava/awt/Container;JZ)I"(thread: JVMThread, args: [number, java_awt_Component, java_awt_Container, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/createHierarchyEvents(ILjava/awt/Component;Ljava/awt/Container;JZ)I"(thread: JVMThread, args: [number, java_awt_Component, java_awt_Container, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHierarchyBoundsListeners()[Ljava/awt/event/HierarchyBoundsListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_HierarchyBoundsListener>) => void): void;
    public "java/awt/Component/getHierarchyBoundsListeners()[Ljava/awt/event/HierarchyBoundsListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_HierarchyBoundsListener>) => void): void;
    public "adjustListeningChildrenOnParent(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/adjustListeningChildrenOnParent(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addKeyListener(Ljava/awt/event/KeyListener;)V"(thread: JVMThread, args: [java_awt_event_KeyListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addKeyListener(Ljava/awt/event/KeyListener;)V"(thread: JVMThread, args: [java_awt_event_KeyListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeKeyListener(Ljava/awt/event/KeyListener;)V"(thread: JVMThread, args: [java_awt_event_KeyListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeKeyListener(Ljava/awt/event/KeyListener;)V"(thread: JVMThread, args: [java_awt_event_KeyListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getKeyListeners()[Ljava/awt/event/KeyListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_KeyListener>) => void): void;
    public "java/awt/Component/getKeyListeners()[Ljava/awt/event/KeyListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_KeyListener>) => void): void;
    public "addMouseListener(Ljava/awt/event/MouseListener;)V"(thread: JVMThread, args: [java_awt_event_MouseListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addMouseListener(Ljava/awt/event/MouseListener;)V"(thread: JVMThread, args: [java_awt_event_MouseListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeMouseListener(Ljava/awt/event/MouseListener;)V"(thread: JVMThread, args: [java_awt_event_MouseListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeMouseListener(Ljava/awt/event/MouseListener;)V"(thread: JVMThread, args: [java_awt_event_MouseListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMouseListeners()[Ljava/awt/event/MouseListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_MouseListener>) => void): void;
    public "java/awt/Component/getMouseListeners()[Ljava/awt/event/MouseListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_MouseListener>) => void): void;
    public "addMouseMotionListener(Ljava/awt/event/MouseMotionListener;)V"(thread: JVMThread, args: [java_awt_event_MouseMotionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addMouseMotionListener(Ljava/awt/event/MouseMotionListener;)V"(thread: JVMThread, args: [java_awt_event_MouseMotionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeMouseMotionListener(Ljava/awt/event/MouseMotionListener;)V"(thread: JVMThread, args: [java_awt_event_MouseMotionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeMouseMotionListener(Ljava/awt/event/MouseMotionListener;)V"(thread: JVMThread, args: [java_awt_event_MouseMotionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMouseMotionListeners()[Ljava/awt/event/MouseMotionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_MouseMotionListener>) => void): void;
    public "java/awt/Component/getMouseMotionListeners()[Ljava/awt/event/MouseMotionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_MouseMotionListener>) => void): void;
    public "addMouseWheelListener(Ljava/awt/event/MouseWheelListener;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addMouseWheelListener(Ljava/awt/event/MouseWheelListener;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeMouseWheelListener(Ljava/awt/event/MouseWheelListener;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeMouseWheelListener(Ljava/awt/event/MouseWheelListener;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMouseWheelListeners()[Ljava/awt/event/MouseWheelListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_MouseWheelListener>) => void): void;
    public "java/awt/Component/getMouseWheelListeners()[Ljava/awt/event/MouseWheelListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_MouseWheelListener>) => void): void;
    public "addInputMethodListener(Ljava/awt/event/InputMethodListener;)V"(thread: JVMThread, args: [java_awt_event_InputMethodListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addInputMethodListener(Ljava/awt/event/InputMethodListener;)V"(thread: JVMThread, args: [java_awt_event_InputMethodListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeInputMethodListener(Ljava/awt/event/InputMethodListener;)V"(thread: JVMThread, args: [java_awt_event_InputMethodListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeInputMethodListener(Ljava/awt/event/InputMethodListener;)V"(thread: JVMThread, args: [java_awt_event_InputMethodListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInputMethodListeners()[Ljava/awt/event/InputMethodListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_InputMethodListener>) => void): void;
    public "java/awt/Component/getInputMethodListeners()[Ljava/awt/event/InputMethodListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_InputMethodListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/Component/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "getInputMethodRequests()Ljava/awt/im/InputMethodRequests;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputMethodRequests) => void): void;
    public "java/awt/Component/getInputMethodRequests()Ljava/awt/im/InputMethodRequests;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputMethodRequests) => void): void;
    public "getInputContext()Ljava/awt/im/InputContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputContext) => void): void;
    public "java/awt/Component/getInputContext()Ljava/awt/im/InputContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_im_InputContext) => void): void;
    public "enableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/enableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "disableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/disableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkCoalescing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/checkCoalescing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isCoalesceEventsOverriden(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Component/isCoalesceEventsOverriden(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isCoalescingEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isCoalescingEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "coalesceEvents(Ljava/awt/AWTEvent;Ljava/awt/AWTEvent;)Ljava/awt/AWTEvent;"(thread: JVMThread, args: [java_awt_AWTEvent, java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "java/awt/Component/coalesceEvents(Ljava/awt/AWTEvent;Ljava/awt/AWTEvent;)Ljava/awt/AWTEvent;"(thread: JVMThread, args: [java_awt_AWTEvent, java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: java_awt_AWTEvent) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processComponentEvent(Ljava/awt/event/ComponentEvent;)V"(thread: JVMThread, args: [java_awt_event_ComponentEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processComponentEvent(Ljava/awt/event/ComponentEvent;)V"(thread: JVMThread, args: [java_awt_event_ComponentEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processFocusEvent(Ljava/awt/event/FocusEvent;)V"(thread: JVMThread, args: [java_awt_event_FocusEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processFocusEvent(Ljava/awt/event/FocusEvent;)V"(thread: JVMThread, args: [java_awt_event_FocusEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processMouseEvent(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processMouseEvent(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processMouseMotionEvent(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processMouseMotionEvent(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processMouseWheelEvent(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processMouseWheelEvent(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "postsOldMouseEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/postsOldMouseEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processInputMethodEvent(Ljava/awt/event/InputMethodEvent;)V"(thread: JVMThread, args: [java_awt_event_InputMethodEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processInputMethodEvent(Ljava/awt/event/InputMethodEvent;)V"(thread: JVMThread, args: [java_awt_event_InputMethodEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processHierarchyEvent(Ljava/awt/event/HierarchyEvent;)V"(thread: JVMThread, args: [java_awt_event_HierarchyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processHierarchyEvent(Ljava/awt/event/HierarchyEvent;)V"(thread: JVMThread, args: [java_awt_event_HierarchyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processHierarchyBoundsEvent(Ljava/awt/event/HierarchyEvent;)V"(thread: JVMThread, args: [java_awt_event_HierarchyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/processHierarchyBoundsEvent(Ljava/awt/event/HierarchyEvent;)V"(thread: JVMThread, args: [java_awt_event_HierarchyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "handleEvent(Ljava/awt/Event;)Z"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/handleEvent(Ljava/awt/Event;)Z"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mouseDown(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/mouseDown(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mouseDrag(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/mouseDrag(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mouseUp(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/mouseUp(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mouseMove(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/mouseMove(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mouseEnter(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/mouseEnter(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mouseExit(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/mouseExit(Ljava/awt/Event;II)Z"(thread: JVMThread, args: [java_awt_Event, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keyDown(Ljava/awt/Event;I)Z"(thread: JVMThread, args: [java_awt_Event, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/keyDown(Ljava/awt/Event;I)Z"(thread: JVMThread, args: [java_awt_Event, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keyUp(Ljava/awt/Event;I)Z"(thread: JVMThread, args: [java_awt_Event, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/keyUp(Ljava/awt/Event;I)Z"(thread: JVMThread, args: [java_awt_Event, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "action(Ljava/awt/Event;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_awt_Event, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/action(Ljava/awt/Event;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_awt_Event, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "gotFocus(Ljava/awt/Event;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_awt_Event, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/gotFocus(Ljava/awt/Event;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_awt_Event, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lostFocus(Ljava/awt/Event;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_awt_Event, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/lostFocus(Ljava/awt/Event;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_awt_Event, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFocusTraversable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isFocusTraversable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isFocusable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFocusable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setFocusable(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isFocusTraversableOverridden()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isFocusTraversableOverridden()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFocusTraversalKeys(ILjava/util/Set;)V"(thread: JVMThread, args: [number, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setFocusTraversalKeys(ILjava/util/Set;)V"(thread: JVMThread, args: [number, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFocusTraversalKeys(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/awt/Component/getFocusTraversalKeys(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "setFocusTraversalKeys_NoIDCheck(ILjava/util/Set;)V"(thread: JVMThread, args: [number, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setFocusTraversalKeys_NoIDCheck(ILjava/util/Set;)V"(thread: JVMThread, args: [number, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFocusTraversalKeys_NoIDCheck(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/awt/Component/getFocusTraversalKeys_NoIDCheck(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "areFocusTraversalKeysSet(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/areFocusTraversalKeysSet(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFocusTraversalKeysEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setFocusTraversalKeysEnabled(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFocusTraversalKeysEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getFocusTraversalKeysEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requestFocus()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/requestFocus()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "requestFocus(Lsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/requestFocus(Lsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requestFocus(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/requestFocus(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requestFocus(ZLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [number, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/requestFocus(ZLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [number, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requestFocusInWindow()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/requestFocusInWindow()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requestFocusInWindow(Lsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/requestFocusInWindow(Lsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requestFocusInWindow(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/requestFocusInWindow(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requestFocusInWindow(ZLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [number, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/requestFocusInWindow(ZLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [number, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requestFocusHelper(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/requestFocusHelper(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "requestFocusHelper(ZZLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [number, number, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/requestFocusHelper(ZZLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [number, number, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRequestFocusAccepted(ZZLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [number, number, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isRequestFocusAccepted(ZZLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [number, number, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setRequestFocusController(Lsun/awt/RequestFocusController;)V"(thread: JVMThread, args: [sun_awt_RequestFocusController], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Component/setRequestFocusController(Lsun/awt/RequestFocusController;)V"(thread: JVMThread, args: [sun_awt_RequestFocusController], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFocusCycleRootAncestor()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Component/getFocusCycleRootAncestor()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "isFocusCycleRoot(Ljava/awt/Container;)Z"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isFocusCycleRoot(Ljava/awt/Container;)Z"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTraversalRoot()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Component/getTraversalRoot()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "transferFocus()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/transferFocus()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "nextFocus()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/nextFocus()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "transferFocus(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/transferFocus(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNextFocusCandidate()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Component/getNextFocusCandidate()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "transferFocusBackward()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/transferFocusBackward()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "transferFocusBackward(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/transferFocusBackward(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "transferFocusUpCycle()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/transferFocusUpCycle()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "hasFocus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/hasFocus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFocusOwner()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isFocusOwner()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setAutoFocusTransferOnDisposal(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setAutoFocusTransferOnDisposal(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAutoFocusTransferOnDisposal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isAutoFocusTransferOnDisposal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/awt/PopupMenu;)V"(thread: JVMThread, args: [java_awt_PopupMenu], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/add(Ljava/awt/PopupMenu;)V"(thread: JVMThread, args: [java_awt_PopupMenu], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/awt/MenuComponent;)V"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/remove(Ljava/awt/MenuComponent;)V"(thread: JVMThread, args: [java_awt_MenuComponent], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Component/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Component/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "list()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/list()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "list(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/list(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "list(Ljava/io/PrintStream;I)V"(thread: JVMThread, args: [java_io_PrintStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/list(Ljava/io/PrintStream;I)V"(thread: JVMThread, args: [java_io_PrintStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "list(Ljava/io/PrintWriter;)V"(thread: JVMThread, args: [java_io_PrintWriter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/list(Ljava/io/PrintWriter;)V"(thread: JVMThread, args: [java_io_PrintWriter], cb?: (e?: java_lang_Throwable) => void): void;
    public "list(Ljava/io/PrintWriter;I)V"(thread: JVMThread, args: [java_io_PrintWriter, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/list(Ljava/io/PrintWriter;I)V"(thread: JVMThread, args: [java_io_PrintWriter, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNativeContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Component/getNativeContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removePropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPropertyChangeListeners()[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "java/awt/Component/getPropertyChangeListeners()[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/removePropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPropertyChangeListeners(Ljava/lang/String;)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "java/awt/Component/getPropertyChangeListeners(Ljava/lang/String;)[Ljava/beans/PropertyChangeListener;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_beans_PropertyChangeListener>) => void): void;
    public "firePropertyChange(Ljava/lang/String;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/firePropertyChange(Ljava/lang/String;Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;ZZ)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/firePropertyChange(Ljava/lang/String;ZZ)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/firePropertyChange(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;BB)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/firePropertyChange(Ljava/lang/String;BB)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;CC)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/firePropertyChange(Ljava/lang/String;CC)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;SS)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/firePropertyChange(Ljava/lang/String;SS)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;JJ)V"(thread: JVMThread, args: [java_lang_String, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/firePropertyChange(Ljava/lang/String;JJ)V"(thread: JVMThread, args: [java_lang_String, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;FF)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/firePropertyChange(Ljava/lang/String;FF)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "firePropertyChange(Ljava/lang/String;DD)V"(thread: JVMThread, args: [java_lang_String, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/firePropertyChange(Ljava/lang/String;DD)V"(thread: JVMThread, args: [java_lang_String, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "doSwingSerialization()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/doSwingSerialization()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "setComponentOrientation(Ljava/awt/ComponentOrientation;)V"(thread: JVMThread, args: [java_awt_ComponentOrientation], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/setComponentOrientation(Ljava/awt/ComponentOrientation;)V"(thread: JVMThread, args: [java_awt_ComponentOrientation], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComponentOrientation()Ljava/awt/ComponentOrientation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ComponentOrientation) => void): void;
    public "java/awt/Component/getComponentOrientation()Ljava/awt/ComponentOrientation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ComponentOrientation) => void): void;
    public "applyComponentOrientation(Ljava/awt/ComponentOrientation;)V"(thread: JVMThread, args: [java_awt_ComponentOrientation], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/applyComponentOrientation(Ljava/awt/ComponentOrientation;)V"(thread: JVMThread, args: [java_awt_ComponentOrientation], cb?: (e?: java_lang_Throwable) => void): void;
    public "canBeFocusOwner()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/canBeFocusOwner()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canBeFocusOwnerRecursively()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/canBeFocusOwnerRecursively()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "relocateComponent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/relocateComponent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContainingWindow()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public "java/awt/Component/getContainingWindow()Ljava/awt/Window;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Window) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Component/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "java/awt/Component/getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
    public "getAccessibleIndexInParent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getAccessibleIndexInParent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAccessibleStateSet()Ljavax/accessibility/AccessibleStateSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleStateSet) => void): void;
    public "java/awt/Component/getAccessibleStateSet()Ljavax/accessibility/AccessibleStateSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleStateSet) => void): void;
    public static "isInstanceOf(Ljava/lang/Object;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Component/isInstanceOf(Ljava/lang/Object;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "areBoundsValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/areBoundsValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "applyCompoundShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/applyCompoundShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAppliedShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "java/awt/Component/getAppliedShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getLocationOnWindow()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Component/getLocationOnWindow()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "getNormalShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "java/awt/Component/getNormalShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getOpaqueShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "java/awt/Component/getOpaqueShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getSiblingIndexAbove()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getSiblingIndexAbove()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHWPeerAboveMe()Ljava/awt/peer/ComponentPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ComponentPeer) => void): void;
    public "java/awt/Component/getHWPeerAboveMe()Ljava/awt/peer/ComponentPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_peer_ComponentPeer) => void): void;
    public "getSiblingIndexBelow()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/getSiblingIndexBelow()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isNonOpaqueForMixing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isNonOpaqueForMixing()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "calculateCurrentShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "java/awt/Component/calculateCurrentShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "applyCurrentShape()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/applyCurrentShape()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "subtractAndApplyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/subtractAndApplyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "applyCurrentShapeBelowMe()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/applyCurrentShapeBelowMe()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "subtractAndApplyShapeBelowMe()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/subtractAndApplyShapeBelowMe()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnShowing()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/mixOnShowing()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnHiding(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/mixOnHiding(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnReshaping()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/mixOnReshaping()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnZOrderChanging(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/mixOnZOrderChanging(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnValidating()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/mixOnValidating()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isMixingNeeded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Component/isMixingNeeded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "updateZOrder()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component/updateZOrder()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$002(Ljava/awt/Component;Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Component, sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "java/awt/Component/access$002(Ljava/awt/Component;Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Component, sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "access$100(Ljava/awt/Component;)Ljava/awt/Point;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public static "java/awt/Component/access$100(Ljava/awt/Component;)Ljava/awt/Point;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public static "access$400(Ljava/awt/Component;)Ljava/awt/Insets;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public static "java/awt/Component/access$400(Ljava/awt/Component;)Ljava/awt/Insets;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public static "access$500(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Component/access$500(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Component/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_PrintWriter extends java_io_Writer {
    public "java/io/PrintWriter/out": java_io_Writer;
    public "java/io/PrintWriter/autoFlush": number;
    public "java/io/PrintWriter/trouble": number;
    public "java/io/PrintWriter/formatter": java_util_Formatter;
    public "java/io/PrintWriter/psOut": java_io_PrintStream;
    public "java/io/PrintWriter/lineSeparator": java_lang_String;
    public static "toCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/io/PrintWriter/toCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "<init>(Ljava/io/Writer;)V"(thread: JVMThread, args: [java_io_Writer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/<init>(Ljava/io/Writer;)V"(thread: JVMThread, args: [java_io_Writer], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/Writer;Z)V"(thread: JVMThread, args: [java_io_Writer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/<init>(Ljava/io/Writer;Z)V"(thread: JVMThread, args: [java_io_Writer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;Z)V"(thread: JVMThread, args: [java_io_OutputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/<init>(Ljava/io/OutputStream;Z)V"(thread: JVMThread, args: [java_io_OutputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/charset/Charset;Ljava/io/File;)V"(thread: JVMThread, args: [java_nio_charset_Charset, java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/<init>(Ljava/nio/charset/Charset;Ljava/io/File;)V"(thread: JVMThread, args: [java_nio_charset_Charset, java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkError()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/PrintWriter/checkError()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/setError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/clearError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/write([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/write(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "newLine()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/newLine()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/print(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/print(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/print(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/print(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/print(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/print(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "print([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/print([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/print(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/print(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "println()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "println([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintWriter/println(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "printf(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "java/io/PrintWriter/printf(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "printf(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "java/io/PrintWriter/printf(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "java/io/PrintWriter/format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "format(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "java/io/PrintWriter/format(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "java/io/PrintWriter/append(Ljava/lang/CharSequence;)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "java/io/PrintWriter/append(Ljava/lang/CharSequence;II)Ljava/io/PrintWriter;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "append(C)Ljava/io/PrintWriter;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "java/io/PrintWriter/append(C)Ljava/io/PrintWriter;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "append(C)Ljava/io/Writer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "java/io/PrintWriter/append(C)Ljava/io/Writer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/io/Writer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "java/io/PrintWriter/append(Ljava/lang/CharSequence;II)Ljava/io/Writer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/io/Writer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "java/io/PrintWriter/append(Ljava/lang/CharSequence;)Ljava/io/Writer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/io/PrintWriter/append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/io/PrintWriter/append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/io/PrintWriter/append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
  }
  export class java_nio_charset_Charset extends java_lang_Object implements java_lang_Comparable {
    public static "java/nio/charset/Charset/bugLevel": java_lang_String;
    public static "java/nio/charset/Charset/standardProvider": java_nio_charset_spi_CharsetProvider;
    public static "java/nio/charset/Charset/cache1": JVMArray<java_lang_Object>;
    public static "java/nio/charset/Charset/cache2": JVMArray<java_lang_Object>;
    public static "java/nio/charset/Charset/gate": java_lang_ThreadLocal;
    public static "java/nio/charset/Charset/defaultCharset": java_nio_charset_Charset;
    public "java/nio/charset/Charset/name": java_lang_String;
    public "java/nio/charset/Charset/aliases": JVMArray<java_lang_String>;
    public "java/nio/charset/Charset/aliasSet": java_util_Set;
    public static "atBugLevel(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/charset/Charset/atBugLevel(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/charset/Charset/checkName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "cache(Ljava/lang/String;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_lang_String, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/charset/Charset/cache(Ljava/lang/String;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_lang_String, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public static "providers()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public static "java/nio/charset/Charset/providers()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public static "lookupViaProviders(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/nio/charset/Charset/lookupViaProviders(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "lookupExtendedCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/nio/charset/Charset/lookupExtendedCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "lookup(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/nio/charset/Charset/lookup(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "lookup2(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/nio/charset/Charset/lookup2(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "isSupported(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/charset/Charset/isSupported(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "forName(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/nio/charset/Charset/forName(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "put(Ljava/util/Iterator;Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Iterator, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/charset/Charset/put(Ljava/util/Iterator;Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Iterator, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public static "availableCharsets()Ljava/util/SortedMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public static "java/nio/charset/Charset/availableCharsets()Ljava/util/SortedMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public static "defaultCharset()Ljava/nio/charset/Charset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/nio/charset/Charset/defaultCharset()Ljava/nio/charset/Charset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "<init>(Ljava/lang/String;[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/Charset/<init>(Ljava/lang/String;[Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/charset/Charset/name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "aliases()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/nio/charset/Charset/aliases()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "displayName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/charset/Charset/displayName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isRegistered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/Charset/isRegistered()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "displayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/charset/Charset/displayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "contains(Ljava/nio/charset/Charset;)Z"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/Charset/contains(Ljava/nio/charset/Charset;)Z"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "newDecoder()Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "java/nio/charset/Charset/newDecoder()Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "newEncoder()Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "java/nio/charset/Charset/newEncoder()Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "canEncode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/Charset/canEncode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "decode(Ljava/nio/ByteBuffer;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/charset/Charset/decode(Ljava/nio/ByteBuffer;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "encode(Ljava/nio/CharBuffer;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/charset/Charset/encode(Ljava/nio/CharBuffer;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "encode(Ljava/lang/String;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/charset/Charset/encode(Ljava/lang/String;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "compareTo(Ljava/nio/charset/Charset;)I"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/Charset/compareTo(Ljava/nio/charset/Charset;)I"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/Charset/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/Charset/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/charset/Charset/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/Charset/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public static "java/nio/charset/Charset/access$000()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public static "access$100()Ljava/nio/charset/spi/CharsetProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_spi_CharsetProvider) => void): void;
    public static "java/nio/charset/Charset/access$100()Ljava/nio/charset/spi/CharsetProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_spi_CharsetProvider) => void): void;
    public static "access$200(Ljava/util/Iterator;Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Iterator, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/charset/Charset/access$200(Ljava/util/Iterator;Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Iterator, java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/charset/Charset/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_charset_CharsetEncoder extends java_lang_Object {
    public "java/nio/charset/CharsetEncoder/charset": java_nio_charset_Charset;
    public "java/nio/charset/CharsetEncoder/averageBytesPerChar": number;
    public "java/nio/charset/CharsetEncoder/maxBytesPerChar": number;
    public "java/nio/charset/CharsetEncoder/replacement": JVMArray<number>;
    public "java/nio/charset/CharsetEncoder/malformedInputAction": java_nio_charset_CodingErrorAction;
    public "java/nio/charset/CharsetEncoder/unmappableCharacterAction": java_nio_charset_CodingErrorAction;
    public static "java/nio/charset/CharsetEncoder/ST_RESET": number;
    public static "java/nio/charset/CharsetEncoder/ST_CODING": number;
    public static "java/nio/charset/CharsetEncoder/ST_END": number;
    public static "java/nio/charset/CharsetEncoder/ST_FLUSHED": number;
    public "java/nio/charset/CharsetEncoder/state": number;
    public static "java/nio/charset/CharsetEncoder/stateNames": JVMArray<java_lang_String>;
    public "java/nio/charset/CharsetEncoder/cachedDecoder": java_lang_ref_WeakReference;
    public static "java/nio/charset/CharsetEncoder/$assertionsDisabled": number;
    public "<init>(Ljava/nio/charset/Charset;FF[B)V"(thread: JVMThread, args: [java_nio_charset_Charset, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetEncoder/<init>(Ljava/nio/charset/Charset;FF[B)V"(thread: JVMThread, args: [java_nio_charset_Charset, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/charset/Charset;FF)V"(thread: JVMThread, args: [java_nio_charset_Charset, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetEncoder/<init>(Ljava/nio/charset/Charset;FF)V"(thread: JVMThread, args: [java_nio_charset_Charset, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "charset()Ljava/nio/charset/Charset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "java/nio/charset/CharsetEncoder/charset()Ljava/nio/charset/Charset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "replacement()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/nio/charset/CharsetEncoder/replacement()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "replaceWith([B)Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "java/nio/charset/CharsetEncoder/replaceWith([B)Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "implReplaceWith([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetEncoder/implReplaceWith([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "isLegalReplacement([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetEncoder/isLegalReplacement([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "malformedInputAction()Ljava/nio/charset/CodingErrorAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CodingErrorAction) => void): void;
    public "java/nio/charset/CharsetEncoder/malformedInputAction()Ljava/nio/charset/CodingErrorAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CodingErrorAction) => void): void;
    public "onMalformedInput(Ljava/nio/charset/CodingErrorAction;)Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "java/nio/charset/CharsetEncoder/onMalformedInput(Ljava/nio/charset/CodingErrorAction;)Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "implOnMalformedInput(Ljava/nio/charset/CodingErrorAction;)V"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetEncoder/implOnMalformedInput(Ljava/nio/charset/CodingErrorAction;)V"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable) => void): void;
    public "unmappableCharacterAction()Ljava/nio/charset/CodingErrorAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CodingErrorAction) => void): void;
    public "java/nio/charset/CharsetEncoder/unmappableCharacterAction()Ljava/nio/charset/CodingErrorAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CodingErrorAction) => void): void;
    public "onUnmappableCharacter(Ljava/nio/charset/CodingErrorAction;)Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "java/nio/charset/CharsetEncoder/onUnmappableCharacter(Ljava/nio/charset/CodingErrorAction;)Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "implOnUnmappableCharacter(Ljava/nio/charset/CodingErrorAction;)V"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetEncoder/implOnUnmappableCharacter(Ljava/nio/charset/CodingErrorAction;)V"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable) => void): void;
    public "averageBytesPerChar()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetEncoder/averageBytesPerChar()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "maxBytesPerChar()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetEncoder/maxBytesPerChar()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "encode(Ljava/nio/CharBuffer;Ljava/nio/ByteBuffer;Z)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_CharBuffer, java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CharsetEncoder/encode(Ljava/nio/CharBuffer;Ljava/nio/ByteBuffer;Z)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_CharBuffer, java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "flush(Ljava/nio/ByteBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CharsetEncoder/flush(Ljava/nio/ByteBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "implFlush(Ljava/nio/ByteBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CharsetEncoder/implFlush(Ljava/nio/ByteBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "reset()Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "java/nio/charset/CharsetEncoder/reset()Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "implReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetEncoder/implReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "encodeLoop(Ljava/nio/CharBuffer;Ljava/nio/ByteBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_CharBuffer, java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CharsetEncoder/encodeLoop(Ljava/nio/CharBuffer;Ljava/nio/ByteBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_CharBuffer, java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "encode(Ljava/nio/CharBuffer;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/charset/CharsetEncoder/encode(Ljava/nio/CharBuffer;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "canEncode(Ljava/nio/CharBuffer;)Z"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetEncoder/canEncode(Ljava/nio/CharBuffer;)Z"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canEncode(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetEncoder/canEncode(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canEncode(Ljava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetEncoder/canEncode(Ljava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "throwIllegalStateException(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetEncoder/throwIllegalStateException(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/charset/CharsetEncoder/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_charset_CoderResult extends java_lang_Object {
    public static "java/nio/charset/CoderResult/CR_UNDERFLOW": number;
    public static "java/nio/charset/CoderResult/CR_OVERFLOW": number;
    public static "java/nio/charset/CoderResult/CR_ERROR_MIN": number;
    public static "java/nio/charset/CoderResult/CR_MALFORMED": number;
    public static "java/nio/charset/CoderResult/CR_UNMAPPABLE": number;
    public static "java/nio/charset/CoderResult/names": JVMArray<java_lang_String>;
    public "java/nio/charset/CoderResult/type": number;
    public "java/nio/charset/CoderResult/length": number;
    public static "java/nio/charset/CoderResult/UNDERFLOW": java_nio_charset_CoderResult;
    public static "java/nio/charset/CoderResult/OVERFLOW": java_nio_charset_CoderResult;
    public static "java/nio/charset/CoderResult/malformedCache": java_nio_charset_CoderResult$Cache;
    public static "java/nio/charset/CoderResult/unmappableCache": java_nio_charset_CoderResult$Cache;
    public static "java/nio/charset/CoderResult/$assertionsDisabled": number;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CoderResult/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/charset/CoderResult/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isUnderflow()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CoderResult/isUnderflow()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOverflow()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CoderResult/isOverflow()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isError()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CoderResult/isError()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMalformed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CoderResult/isMalformed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isUnmappable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CoderResult/isUnmappable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CoderResult/length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "malformedForLength(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public static "java/nio/charset/CoderResult/malformedForLength(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public static "unmappableForLength(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public static "java/nio/charset/CoderResult/unmappableForLength(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "throwException()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CoderResult/throwException()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IILjava/nio/charset/CoderResult$1;)V"(thread: JVMThread, args: [number, number, java_nio_charset_CoderResult$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CoderResult/<init>(IILjava/nio/charset/CoderResult$1;)V"(thread: JVMThread, args: [number, number, java_nio_charset_CoderResult$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/charset/CoderResult/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_charset_CoderResult$1 extends java_nio_charset_CoderResult$Cache {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CoderResult$1/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "create(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CoderResult$1/create(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
  }
  export class java_nio_charset_CoderResult$Cache extends java_lang_Object {
    public "java/nio/charset/CoderResult$Cache/cache": java_util_Map;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CoderResult$Cache/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "create(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CoderResult$Cache/create(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "get(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CoderResult$Cache/get(I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "<init>(Ljava/nio/charset/CoderResult$1;)V"(thread: JVMThread, args: [java_nio_charset_CoderResult$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CoderResult$Cache/<init>(Ljava/nio/charset/CoderResult$1;)V"(thread: JVMThread, args: [java_nio_charset_CoderResult$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$200(Ljava/nio/charset/CoderResult$Cache;I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_charset_CoderResult$Cache, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public static "java/nio/charset/CoderResult$Cache/access$200(Ljava/nio/charset/CoderResult$Cache;I)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_charset_CoderResult$Cache, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
  }
  export class java_nio_charset_CodingErrorAction extends java_lang_Object {
    public "java/nio/charset/CodingErrorAction/name": java_lang_String;
    public static "java/nio/charset/CodingErrorAction/IGNORE": java_nio_charset_CodingErrorAction;
    public static "java/nio/charset/CodingErrorAction/REPLACE": java_nio_charset_CodingErrorAction;
    public static "java/nio/charset/CodingErrorAction/REPORT": java_nio_charset_CodingErrorAction;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CodingErrorAction/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/charset/CodingErrorAction/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/charset/CodingErrorAction/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_charset_CharsetDecoder extends java_lang_Object {
    public "java/nio/charset/CharsetDecoder/charset": java_nio_charset_Charset;
    public "java/nio/charset/CharsetDecoder/averageCharsPerByte": number;
    public "java/nio/charset/CharsetDecoder/maxCharsPerByte": number;
    public "java/nio/charset/CharsetDecoder/replacement": java_lang_String;
    public "java/nio/charset/CharsetDecoder/malformedInputAction": java_nio_charset_CodingErrorAction;
    public "java/nio/charset/CharsetDecoder/unmappableCharacterAction": java_nio_charset_CodingErrorAction;
    public static "java/nio/charset/CharsetDecoder/ST_RESET": number;
    public static "java/nio/charset/CharsetDecoder/ST_CODING": number;
    public static "java/nio/charset/CharsetDecoder/ST_END": number;
    public static "java/nio/charset/CharsetDecoder/ST_FLUSHED": number;
    public "java/nio/charset/CharsetDecoder/state": number;
    public static "java/nio/charset/CharsetDecoder/stateNames": JVMArray<java_lang_String>;
    public static "java/nio/charset/CharsetDecoder/$assertionsDisabled": number;
    public "<init>(Ljava/nio/charset/Charset;FFLjava/lang/String;)V"(thread: JVMThread, args: [java_nio_charset_Charset, number, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetDecoder/<init>(Ljava/nio/charset/Charset;FFLjava/lang/String;)V"(thread: JVMThread, args: [java_nio_charset_Charset, number, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/charset/Charset;FF)V"(thread: JVMThread, args: [java_nio_charset_Charset, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetDecoder/<init>(Ljava/nio/charset/Charset;FF)V"(thread: JVMThread, args: [java_nio_charset_Charset, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "charset()Ljava/nio/charset/Charset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "java/nio/charset/CharsetDecoder/charset()Ljava/nio/charset/Charset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "replacement()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/charset/CharsetDecoder/replacement()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "replaceWith(Ljava/lang/String;)Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "java/nio/charset/CharsetDecoder/replaceWith(Ljava/lang/String;)Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "implReplaceWith(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetDecoder/implReplaceWith(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "malformedInputAction()Ljava/nio/charset/CodingErrorAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CodingErrorAction) => void): void;
    public "java/nio/charset/CharsetDecoder/malformedInputAction()Ljava/nio/charset/CodingErrorAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CodingErrorAction) => void): void;
    public "onMalformedInput(Ljava/nio/charset/CodingErrorAction;)Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "java/nio/charset/CharsetDecoder/onMalformedInput(Ljava/nio/charset/CodingErrorAction;)Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "implOnMalformedInput(Ljava/nio/charset/CodingErrorAction;)V"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetDecoder/implOnMalformedInput(Ljava/nio/charset/CodingErrorAction;)V"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable) => void): void;
    public "unmappableCharacterAction()Ljava/nio/charset/CodingErrorAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CodingErrorAction) => void): void;
    public "java/nio/charset/CharsetDecoder/unmappableCharacterAction()Ljava/nio/charset/CodingErrorAction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CodingErrorAction) => void): void;
    public "onUnmappableCharacter(Ljava/nio/charset/CodingErrorAction;)Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "java/nio/charset/CharsetDecoder/onUnmappableCharacter(Ljava/nio/charset/CodingErrorAction;)Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "implOnUnmappableCharacter(Ljava/nio/charset/CodingErrorAction;)V"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetDecoder/implOnUnmappableCharacter(Ljava/nio/charset/CodingErrorAction;)V"(thread: JVMThread, args: [java_nio_charset_CodingErrorAction], cb?: (e?: java_lang_Throwable) => void): void;
    public "averageCharsPerByte()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetDecoder/averageCharsPerByte()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "maxCharsPerByte()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetDecoder/maxCharsPerByte()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "decode(Ljava/nio/ByteBuffer;Ljava/nio/CharBuffer;Z)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_nio_CharBuffer, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CharsetDecoder/decode(Ljava/nio/ByteBuffer;Ljava/nio/CharBuffer;Z)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_nio_CharBuffer, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "flush(Ljava/nio/CharBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CharsetDecoder/flush(Ljava/nio/CharBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "implFlush(Ljava/nio/CharBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CharsetDecoder/implFlush(Ljava/nio/CharBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "reset()Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "java/nio/charset/CharsetDecoder/reset()Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "implReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetDecoder/implReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "decodeLoop(Ljava/nio/ByteBuffer;Ljava/nio/CharBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "java/nio/charset/CharsetDecoder/decodeLoop(Ljava/nio/ByteBuffer;Ljava/nio/CharBuffer;)Ljava/nio/charset/CoderResult;"(thread: JVMThread, args: [java_nio_ByteBuffer, java_nio_CharBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CoderResult) => void): void;
    public "decode(Ljava/nio/ByteBuffer;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/charset/CharsetDecoder/decode(Ljava/nio/ByteBuffer;)Ljava/nio/CharBuffer;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "isAutoDetecting()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetDecoder/isAutoDetecting()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isCharsetDetected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/charset/CharsetDecoder/isCharsetDetected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "detectedCharset()Ljava/nio/charset/Charset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "java/nio/charset/CharsetDecoder/detectedCharset()Ljava/nio/charset/Charset;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "throwIllegalStateException(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/CharsetDecoder/throwIllegalStateException(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/charset/CharsetDecoder/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_charset_spi_CharsetProvider extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/spi/CharsetProvider/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "charsets()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/nio/charset/spi/CharsetProvider/charsets()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "charsetForName(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "java/nio/charset/spi/CharsetProvider/charsetForName(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
  }
  export class java_util_Formatter extends java_lang_Object implements java_io_Closeable, java_io_Flushable {
    public "java/util/Formatter/a": java_lang_Appendable;
    public "java/util/Formatter/l": java_util_Locale;
    public "java/util/Formatter/lastException": java_io_IOException;
    public "java/util/Formatter/zero": number;
    public static "java/util/Formatter/scaleUp": number;
    public static "java/util/Formatter/MAX_FD_CHARS": number;
    public static "java/util/Formatter/formatSpecifier": java_lang_String;
    public static "java/util/Formatter/fsPattern": java_util_regex_Pattern;
    public static "toCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/util/Formatter/toCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "nonNullAppendable(Ljava/lang/Appendable;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_Appendable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public static "java/util/Formatter/nonNullAppendable(Ljava/lang/Appendable;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_Appendable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "<init>(Ljava/util/Locale;Ljava/lang/Appendable;)V"(thread: JVMThread, args: [java_util_Locale, java_lang_Appendable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/util/Locale;Ljava/lang/Appendable;)V"(thread: JVMThread, args: [java_util_Locale, java_lang_Appendable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/charset/Charset;Ljava/util/Locale;Ljava/io/File;)V"(thread: JVMThread, args: [java_nio_charset_Charset, java_util_Locale, java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/nio/charset/Charset;Ljava/util/Locale;Ljava/io/File;)V"(thread: JVMThread, args: [java_nio_charset_Charset, java_util_Locale, java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Appendable;)V"(thread: JVMThread, args: [java_lang_Appendable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/lang/Appendable;)V"(thread: JVMThread, args: [java_lang_Appendable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Appendable;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_lang_Appendable, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/lang/Appendable;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_lang_Appendable, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_io_File, java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/io/File;Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_io_File, java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/<init>(Ljava/io/OutputStream;Ljava/lang/String;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getZero(Ljava/util/Locale;)C"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Formatter/getZero(Ljava/util/Locale;)C"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "locale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "java/util/Formatter/locale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "out()Ljava/lang/Appendable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/util/Formatter/out()Ljava/lang/Appendable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Formatter/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Formatter/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ioException()Ljava/io/IOException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_IOException) => void): void;
    public "java/util/Formatter/ioException()Ljava/io/IOException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_IOException) => void): void;
    public "format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/util/Formatter;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_util_Formatter) => void): void;
    public "java/util/Formatter/format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/util/Formatter;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_util_Formatter) => void): void;
    public "format(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/util/Formatter;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_util_Formatter) => void): void;
    public "java/util/Formatter/format(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/util/Formatter;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_util_Formatter) => void): void;
    public "parse(Ljava/lang/String;)[Ljava/util/Formatter$FormatString;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Formatter$FormatString>) => void): void;
    public "java/util/Formatter/parse(Ljava/lang/String;)[Ljava/util/Formatter$FormatString;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_Formatter$FormatString>) => void): void;
    public static "checkText(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Formatter/checkText(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/util/Formatter;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_util_Formatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public static "java/util/Formatter/access$000(Ljava/util/Formatter;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_util_Formatter], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public static "access$202(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Formatter/access$202(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Formatter/access$200()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$300(Ljava/util/Formatter;)C"(thread: JVMThread, args: [java_util_Formatter], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Formatter/access$300(Ljava/util/Formatter;)C"(thread: JVMThread, args: [java_util_Formatter], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Formatter/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_Formatter$FormatString extends java_lang_Object {
    "index()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "print(Ljava/lang/Object;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_lang_Object, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_io_Writer extends java_lang_Object implements java_lang_Appendable, java_io_Closeable, java_io_Flushable {
    public "java/io/Writer/writeBuffer": JVMArray<number>;
    public static "java/io/Writer/WRITE_BUFFER_SIZE": number;
    public "java/io/Writer/lock": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Writer/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Writer/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Writer/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Writer/write([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Writer/write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Writer/write(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Writer/write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/io/Writer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "java/io/Writer/append(Ljava/lang/CharSequence;)Ljava/io/Writer;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/io/Writer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "java/io/Writer/append(Ljava/lang/CharSequence;II)Ljava/io/Writer;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "append(C)Ljava/io/Writer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "java/io/Writer/append(C)Ljava/io/Writer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_Writer) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Writer/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Writer/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/io/Writer/append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/io/Writer/append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/io/Writer/append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
  }
  export class java_awt_event_HierarchyEvent extends java_awt_AWTEvent {
    public static "java/awt/event/HierarchyEvent/serialVersionUID": Long;
    public static "java/awt/event/HierarchyEvent/HIERARCHY_FIRST": number;
    public static "java/awt/event/HierarchyEvent/HIERARCHY_CHANGED": number;
    public static "java/awt/event/HierarchyEvent/ANCESTOR_MOVED": number;
    public static "java/awt/event/HierarchyEvent/ANCESTOR_RESIZED": number;
    public static "java/awt/event/HierarchyEvent/HIERARCHY_LAST": number;
    public static "java/awt/event/HierarchyEvent/PARENT_CHANGED": number;
    public static "java/awt/event/HierarchyEvent/DISPLAYABILITY_CHANGED": number;
    public static "java/awt/event/HierarchyEvent/SHOWING_CHANGED": number;
    public "java/awt/event/HierarchyEvent/changed": java_awt_Component;
    public "java/awt/event/HierarchyEvent/changedParent": java_awt_Container;
    public "java/awt/event/HierarchyEvent/changeFlags": Long;
    public "<init>(Ljava/awt/Component;ILjava/awt/Component;Ljava/awt/Container;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_Component, java_awt_Container], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/HierarchyEvent/<init>(Ljava/awt/Component;ILjava/awt/Component;Ljava/awt/Container;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_Component, java_awt_Container], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;ILjava/awt/Component;Ljava/awt/Container;J)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_Component, java_awt_Container, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/HierarchyEvent/<init>(Ljava/awt/Component;ILjava/awt/Component;Ljava/awt/Container;J)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_Component, java_awt_Container, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/event/HierarchyEvent/getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getChanged()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/event/HierarchyEvent/getChanged()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getChangedParent()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/event/HierarchyEvent/getChangedParent()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "getChangeFlags()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/awt/event/HierarchyEvent/getChangeFlags()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/HierarchyEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_event_InputMethodEvent extends java_awt_AWTEvent {
    public static "java/awt/event/InputMethodEvent/serialVersionUID": Long;
    public static "java/awt/event/InputMethodEvent/INPUT_METHOD_FIRST": number;
    public static "java/awt/event/InputMethodEvent/INPUT_METHOD_TEXT_CHANGED": number;
    public static "java/awt/event/InputMethodEvent/CARET_POSITION_CHANGED": number;
    public static "java/awt/event/InputMethodEvent/INPUT_METHOD_LAST": number;
    public "java/awt/event/InputMethodEvent/when": Long;
    public "java/awt/event/InputMethodEvent/text": java_text_AttributedCharacterIterator;
    public "java/awt/event/InputMethodEvent/committedCharacterCount": number;
    public "java/awt/event/InputMethodEvent/caret": java_awt_font_TextHitInfo;
    public "java/awt/event/InputMethodEvent/visiblePosition": java_awt_font_TextHitInfo;
    public "<init>(Ljava/awt/Component;IJLjava/text/AttributedCharacterIterator;ILjava/awt/font/TextHitInfo;Ljava/awt/font/TextHitInfo;)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, java_text_AttributedCharacterIterator, number, java_awt_font_TextHitInfo, java_awt_font_TextHitInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InputMethodEvent/<init>(Ljava/awt/Component;IJLjava/text/AttributedCharacterIterator;ILjava/awt/font/TextHitInfo;Ljava/awt/font/TextHitInfo;)V"(thread: JVMThread, args: [java_awt_Component, number, Long, any, java_text_AttributedCharacterIterator, number, java_awt_font_TextHitInfo, java_awt_font_TextHitInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;ILjava/text/AttributedCharacterIterator;ILjava/awt/font/TextHitInfo;Ljava/awt/font/TextHitInfo;)V"(thread: JVMThread, args: [java_awt_Component, number, java_text_AttributedCharacterIterator, number, java_awt_font_TextHitInfo, java_awt_font_TextHitInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InputMethodEvent/<init>(Ljava/awt/Component;ILjava/text/AttributedCharacterIterator;ILjava/awt/font/TextHitInfo;Ljava/awt/font/TextHitInfo;)V"(thread: JVMThread, args: [java_awt_Component, number, java_text_AttributedCharacterIterator, number, java_awt_font_TextHitInfo, java_awt_font_TextHitInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;ILjava/awt/font/TextHitInfo;Ljava/awt/font/TextHitInfo;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_font_TextHitInfo, java_awt_font_TextHitInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InputMethodEvent/<init>(Ljava/awt/Component;ILjava/awt/font/TextHitInfo;Ljava/awt/font/TextHitInfo;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_font_TextHitInfo, java_awt_font_TextHitInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "getText()Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "java/awt/event/InputMethodEvent/getText()Ljava/text/AttributedCharacterIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_text_AttributedCharacterIterator) => void): void;
    public "getCommittedCharacterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputMethodEvent/getCommittedCharacterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCaret()Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public "java/awt/event/InputMethodEvent/getCaret()Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public "getVisiblePosition()Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public "java/awt/event/InputMethodEvent/getVisiblePosition()Ljava/awt/font/TextHitInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_TextHitInfo) => void): void;
    public "consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InputMethodEvent/consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isConsumed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/event/InputMethodEvent/isConsumed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWhen()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/awt/event/InputMethodEvent/getWhen()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/InputMethodEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/InputMethodEvent/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getMostRecentEventTimeForSource(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/awt/event/InputMethodEvent/getMostRecentEventTimeForSource(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class java_awt_Component$BaselineResizeBehavior extends java_lang_Enum {
    public static "java/awt/Component$BaselineResizeBehavior/CONSTANT_ASCENT": java_awt_Component$BaselineResizeBehavior;
    public static "java/awt/Component$BaselineResizeBehavior/CONSTANT_DESCENT": java_awt_Component$BaselineResizeBehavior;
    public static "java/awt/Component$BaselineResizeBehavior/CENTER_OFFSET": java_awt_Component$BaselineResizeBehavior;
    public static "java/awt/Component$BaselineResizeBehavior/OTHER": java_awt_Component$BaselineResizeBehavior;
    public static "java/awt/Component$BaselineResizeBehavior/$VALUES": JVMArray<java_awt_Component$BaselineResizeBehavior>;
    public static "values()[Ljava/awt/Component$BaselineResizeBehavior;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Component$BaselineResizeBehavior>) => void): void;
    public static "java/awt/Component$BaselineResizeBehavior/values()[Ljava/awt/Component$BaselineResizeBehavior;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Component$BaselineResizeBehavior>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/awt/Component$BaselineResizeBehavior;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component$BaselineResizeBehavior) => void): void;
    public static "java/awt/Component$BaselineResizeBehavior/valueOf(Ljava/lang/String;)Ljava/awt/Component$BaselineResizeBehavior;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component$BaselineResizeBehavior) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Component$BaselineResizeBehavior/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Component$BaselineResizeBehavior/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_PointerInfo extends java_lang_Object {
    public "java/awt/PointerInfo/device": java_awt_GraphicsDevice;
    public "java/awt/PointerInfo/location": java_awt_Point;
    public "<init>(Ljava/awt/GraphicsDevice;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_GraphicsDevice, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/PointerInfo/<init>(Ljava/awt/GraphicsDevice;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_GraphicsDevice, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDevice()Ljava/awt/GraphicsDevice;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsDevice) => void): void;
    public "java/awt/PointerInfo/getDevice()Ljava/awt/GraphicsDevice;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsDevice) => void): void;
    public "getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/PointerInfo/getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
  }
  export interface sun_awt_RequestFocusController extends java_lang_Object {
    "acceptRequestFocus(Ljava/awt/Component;Ljava/awt/Component;ZZLsun/awt/CausedFocusEvent$Cause;)Z"(thread: JVMThread, args: [java_awt_Component, java_awt_Component, number, number, sun_awt_CausedFocusEvent$Cause], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_event_InputMethodListener extends java_lang_Object, java_util_EventListener {
    "inputMethodTextChanged(Ljava/awt/event/InputMethodEvent;)V"(thread: JVMThread, args: [java_awt_event_InputMethodEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "caretPositionChanged(Ljava/awt/event/InputMethodEvent;)V"(thread: JVMThread, args: [java_awt_event_InputMethodEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_MouseWheelListener extends java_lang_Object, java_util_EventListener {
    "mouseWheelMoved(Ljava/awt/event/MouseWheelEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseWheelEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_MouseMotionListener extends java_lang_Object, java_util_EventListener {
    "mouseDragged(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "mouseMoved(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_MouseListener extends java_lang_Object, java_util_EventListener {
    "mouseClicked(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "mousePressed(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "mouseReleased(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "mouseEntered(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "mouseExited(Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_KeyListener extends java_lang_Object, java_util_EventListener {
    "keyTyped(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "keyPressed(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "keyReleased(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_HierarchyBoundsListener extends java_lang_Object, java_util_EventListener {
    "ancestorMoved(Ljava/awt/event/HierarchyEvent;)V"(thread: JVMThread, args: [java_awt_event_HierarchyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "ancestorResized(Ljava/awt/event/HierarchyEvent;)V"(thread: JVMThread, args: [java_awt_event_HierarchyEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_HierarchyListener extends java_lang_Object, java_util_EventListener {
    "hierarchyChanged(Ljava/awt/event/HierarchyEvent;)V"(thread: JVMThread, args: [java_awt_event_HierarchyEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_event_ComponentListener extends java_lang_Object, java_util_EventListener {
    "componentResized(Ljava/awt/event/ComponentEvent;)V"(thread: JVMThread, args: [java_awt_event_ComponentEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "componentMoved(Ljava/awt/event/ComponentEvent;)V"(thread: JVMThread, args: [java_awt_event_ComponentEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "componentShown(Ljava/awt/event/ComponentEvent;)V"(thread: JVMThread, args: [java_awt_event_ComponentEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "componentHidden(Ljava/awt/event/ComponentEvent;)V"(thread: JVMThread, args: [java_awt_event_ComponentEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_ComponentOrientation extends java_lang_Object implements java_io_Serializable {
    public static "java/awt/ComponentOrientation/serialVersionUID": Long;
    public static "java/awt/ComponentOrientation/UNK_BIT": number;
    public static "java/awt/ComponentOrientation/HORIZ_BIT": number;
    public static "java/awt/ComponentOrientation/LTR_BIT": number;
    public static "java/awt/ComponentOrientation/LEFT_TO_RIGHT": java_awt_ComponentOrientation;
    public static "java/awt/ComponentOrientation/RIGHT_TO_LEFT": java_awt_ComponentOrientation;
    public static "java/awt/ComponentOrientation/UNKNOWN": java_awt_ComponentOrientation;
    public "java/awt/ComponentOrientation/orientation": number;
    public "isHorizontal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ComponentOrientation/isHorizontal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLeftToRight()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ComponentOrientation/isLeftToRight()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getOrientation(Ljava/util/Locale;)Ljava/awt/ComponentOrientation;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_awt_ComponentOrientation) => void): void;
    public static "java/awt/ComponentOrientation/getOrientation(Ljava/util/Locale;)Ljava/awt/ComponentOrientation;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_awt_ComponentOrientation) => void): void;
    public static "getOrientation(Ljava/util/ResourceBundle;)Ljava/awt/ComponentOrientation;"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: java_awt_ComponentOrientation) => void): void;
    public static "java/awt/ComponentOrientation/getOrientation(Ljava/util/ResourceBundle;)Ljava/awt/ComponentOrientation;"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: java_awt_ComponentOrientation) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ComponentOrientation/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/ComponentOrientation/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_dnd_DropTarget extends java_lang_Object implements java_awt_dnd_DropTargetListener, java_io_Serializable {
    public static "java/awt/dnd/DropTarget/serialVersionUID": Long;
    public "java/awt/dnd/DropTarget/dropTargetContext": java_awt_dnd_DropTargetContext;
    public "java/awt/dnd/DropTarget/component": java_awt_Component;
    public "java/awt/dnd/DropTarget/componentPeer": java_awt_peer_ComponentPeer;
    public "java/awt/dnd/DropTarget/nativePeer": java_awt_peer_ComponentPeer;
    public "java/awt/dnd/DropTarget/actions": number;
    public "java/awt/dnd/DropTarget/active": number;
    public "java/awt/dnd/DropTarget/autoScroller": java_awt_dnd_DropTarget$DropTargetAutoScroller;
    public "java/awt/dnd/DropTarget/dtListener": java_awt_dnd_DropTargetListener;
    public "java/awt/dnd/DropTarget/flavorMap": java_awt_datatransfer_FlavorMap;
    public "java/awt/dnd/DropTarget/isDraggingInside": number;
    public "<init>(Ljava/awt/Component;ILjava/awt/dnd/DropTargetListener;ZLjava/awt/datatransfer/FlavorMap;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_dnd_DropTargetListener, number, java_awt_datatransfer_FlavorMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/<init>(Ljava/awt/Component;ILjava/awt/dnd/DropTargetListener;ZLjava/awt/datatransfer/FlavorMap;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_dnd_DropTargetListener, number, java_awt_datatransfer_FlavorMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;ILjava/awt/dnd/DropTargetListener;Z)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_dnd_DropTargetListener, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/<init>(Ljava/awt/Component;ILjava/awt/dnd/DropTargetListener;Z)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_dnd_DropTargetListener, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;Ljava/awt/dnd/DropTargetListener;)V"(thread: JVMThread, args: [java_awt_Component, java_awt_dnd_DropTargetListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/<init>(Ljava/awt/Component;Ljava/awt/dnd/DropTargetListener;)V"(thread: JVMThread, args: [java_awt_Component, java_awt_dnd_DropTargetListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Component;ILjava/awt/dnd/DropTargetListener;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_dnd_DropTargetListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/<init>(Ljava/awt/Component;ILjava/awt/dnd/DropTargetListener;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_dnd_DropTargetListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "setComponent(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/setComponent(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/dnd/DropTarget/getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "setDefaultActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/setDefaultActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "doSetDefaultActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/doSetDefaultActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDefaultActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTarget/getDefaultActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setActive(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/setActive(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isActive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTarget/isActive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addDropTargetListener(Ljava/awt/dnd/DropTargetListener;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/addDropTargetListener(Ljava/awt/dnd/DropTargetListener;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeDropTargetListener(Ljava/awt/dnd/DropTargetListener;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/removeDropTargetListener(Ljava/awt/dnd/DropTargetListener;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "dragEnter(Ljava/awt/dnd/DropTargetDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/dragEnter(Ljava/awt/dnd/DropTargetDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dragOver(Ljava/awt/dnd/DropTargetDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/dragOver(Ljava/awt/dnd/DropTargetDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dropActionChanged(Ljava/awt/dnd/DropTargetDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/dropActionChanged(Ljava/awt/dnd/DropTargetDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dragExit(Ljava/awt/dnd/DropTargetEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/dragExit(Ljava/awt/dnd/DropTargetEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "drop(Ljava/awt/dnd/DropTargetDropEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDropEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/drop(Ljava/awt/dnd/DropTargetDropEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDropEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFlavorMap()Ljava/awt/datatransfer/FlavorMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_FlavorMap) => void): void;
    public "java/awt/dnd/DropTarget/getFlavorMap()Ljava/awt/datatransfer/FlavorMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_FlavorMap) => void): void;
    public "setFlavorMap(Ljava/awt/datatransfer/FlavorMap;)V"(thread: JVMThread, args: [java_awt_datatransfer_FlavorMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/setFlavorMap(Ljava/awt/datatransfer/FlavorMap;)V"(thread: JVMThread, args: [java_awt_datatransfer_FlavorMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "addNotify(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/addNotify(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/removeNotify(Ljava/awt/peer/ComponentPeer;)V"(thread: JVMThread, args: [java_awt_peer_ComponentPeer], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDropTargetContext()Ljava/awt/dnd/DropTargetContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTargetContext) => void): void;
    public "java/awt/dnd/DropTarget/getDropTargetContext()Ljava/awt/dnd/DropTargetContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTargetContext) => void): void;
    public "createDropTargetContext()Ljava/awt/dnd/DropTargetContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTargetContext) => void): void;
    public "java/awt/dnd/DropTarget/createDropTargetContext()Ljava/awt/dnd/DropTargetContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTargetContext) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "createDropTargetAutoScroller(Ljava/awt/Component;Ljava/awt/Point;)Ljava/awt/dnd/DropTarget$DropTargetAutoScroller;"(thread: JVMThread, args: [java_awt_Component, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTarget$DropTargetAutoScroller) => void): void;
    public "java/awt/dnd/DropTarget/createDropTargetAutoScroller(Ljava/awt/Component;Ljava/awt/Point;)Ljava/awt/dnd/DropTarget$DropTargetAutoScroller;"(thread: JVMThread, args: [java_awt_Component, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTarget$DropTargetAutoScroller) => void): void;
    public "initializeAutoscrolling(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/initializeAutoscrolling(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateAutoscroll(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/updateAutoscroll(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearAutoscroll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget/clearAutoscroll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_dnd_DropTargetDropEvent extends java_awt_dnd_DropTargetEvent {
    public static "java/awt/dnd/DropTargetDropEvent/serialVersionUID": Long;
    public static "java/awt/dnd/DropTargetDropEvent/zero": java_awt_Point;
    public "java/awt/dnd/DropTargetDropEvent/location": java_awt_Point;
    public "java/awt/dnd/DropTargetDropEvent/actions": number;
    public "java/awt/dnd/DropTargetDropEvent/dropAction": number;
    public "java/awt/dnd/DropTargetDropEvent/isLocalTx": number;
    public "<init>(Ljava/awt/dnd/DropTargetContext;Ljava/awt/Point;II)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetContext, java_awt_Point, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/<init>(Ljava/awt/dnd/DropTargetContext;Ljava/awt/Point;II)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetContext, java_awt_Point, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/dnd/DropTargetContext;Ljava/awt/Point;IIZ)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetContext, java_awt_Point, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/<init>(Ljava/awt/dnd/DropTargetContext;Ljava/awt/Point;IIZ)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetContext, java_awt_Point, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "getCurrentDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/getCurrentDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "getCurrentDataFlavorsAsList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/getCurrentDataFlavorsAsList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "isDataFlavorSupported(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/isDataFlavorSupported(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSourceActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/getSourceActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDropAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/getDropAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "acceptDrop(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/acceptDrop(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rejectDrop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/rejectDrop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dropComplete(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/dropComplete(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isLocalTransfer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTargetDropEvent/isLocalTransfer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/dnd/DropTargetDropEvent/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_dnd_DropTargetEvent extends java_util_EventObject {
    public static "java/awt/dnd/DropTargetEvent/serialVersionUID": Long;
    public "java/awt/dnd/DropTargetEvent/context": java_awt_dnd_DropTargetContext;
    public "<init>(Ljava/awt/dnd/DropTargetContext;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetEvent/<init>(Ljava/awt/dnd/DropTargetContext;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDropTargetContext()Ljava/awt/dnd/DropTargetContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTargetContext) => void): void;
    public "java/awt/dnd/DropTargetEvent/getDropTargetContext()Ljava/awt/dnd/DropTargetContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTargetContext) => void): void;
  }
  export class java_awt_dnd_DropTargetDragEvent extends java_awt_dnd_DropTargetEvent {
    public static "java/awt/dnd/DropTargetDragEvent/serialVersionUID": Long;
    public "java/awt/dnd/DropTargetDragEvent/location": java_awt_Point;
    public "java/awt/dnd/DropTargetDragEvent/actions": number;
    public "java/awt/dnd/DropTargetDragEvent/dropAction": number;
    public "<init>(Ljava/awt/dnd/DropTargetContext;Ljava/awt/Point;II)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetContext, java_awt_Point, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/<init>(Ljava/awt/dnd/DropTargetContext;Ljava/awt/Point;II)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetContext, java_awt_Point, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/getLocation()Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "getCurrentDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/getCurrentDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "getCurrentDataFlavorsAsList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/getCurrentDataFlavorsAsList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "isDataFlavorSupported(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/isDataFlavorSupported(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSourceActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/getSourceActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDropAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/getDropAction()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "acceptDrag(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/acceptDrag(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rejectDrag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetDragEvent/rejectDrag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_dnd_DropTarget$DropTargetAutoScroller extends java_lang_Object implements java_awt_event_ActionListener {
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/component": java_awt_Component;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/autoScroll": java_awt_dnd_Autoscroll;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/timer": javax_swing_Timer;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/locn": java_awt_Point;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/prev": java_awt_Point;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/outer": java_awt_Rectangle;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/inner": java_awt_Rectangle;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/hysteresis": number;
    public "<init>(Ljava/awt/Component;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Component, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/<init>(Ljava/awt/Component;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Component, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateRegion()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/updateRegion()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/updateLocation(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "stop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/stop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "actionPerformed(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTarget$DropTargetAutoScroller/actionPerformed(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_swing_Timer extends java_lang_Object implements java_io_Serializable {
    public "javax/swing/Timer/listenerList": javax_swing_event_EventListenerList;
    public "javax/swing/Timer/notify": java_util_concurrent_atomic_AtomicBoolean;
    public "javax/swing/Timer/initialDelay": number;
    public "javax/swing/Timer/delay": number;
    public "javax/swing/Timer/repeats": number;
    public "javax/swing/Timer/coalesce": number;
    public "javax/swing/Timer/doPostEvent": java_lang_Runnable;
    public static "javax/swing/Timer/logTimers": number;
    public "javax/swing/Timer/lock": java_util_concurrent_locks_Lock;
    public "javax/swing/Timer/delayedTimer": javax_swing_TimerQueue$DelayedTimer;
    public "javax/swing/Timer/actionCommand": java_lang_String;
    public "javax/swing/Timer/acc": java_security_AccessControlContext;
    public "<init>(ILjava/awt/event/ActionListener;)V"(thread: JVMThread, args: [number, java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/<init>(ILjava/awt/event/ActionListener;)V"(thread: JVMThread, args: [number, java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "javax/swing/Timer/getAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/addActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/removeActionListener(Ljava/awt/event/ActionListener;)V"(thread: JVMThread, args: [java_awt_event_ActionListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "javax/swing/Timer/getActionListeners()[Ljava/awt/event/ActionListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ActionListener>) => void): void;
    public "fireActionPerformed(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/fireActionPerformed(Ljava/awt/event/ActionEvent;)V"(thread: JVMThread, args: [java_awt_event_ActionEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "javax/swing/Timer/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "timerQueue()Ljavax/swing/TimerQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_swing_TimerQueue) => void): void;
    public "javax/swing/Timer/timerQueue()Ljavax/swing/TimerQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_swing_TimerQueue) => void): void;
    public static "setLogTimers(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/swing/Timer/setLogTimers(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getLogTimers()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "javax/swing/Timer/getLogTimers()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setDelay(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/setDelay(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDelay()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/Timer/getDelay()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setInitialDelay(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/setInitialDelay(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInitialDelay()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/Timer/getInitialDelay()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setRepeats(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/setRepeats(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isRepeats()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/Timer/isRepeats()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setCoalesce(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/setCoalesce(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isCoalesce()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/Timer/isCoalesce()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setActionCommand(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/setActionCommand(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getActionCommand()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/swing/Timer/getActionCommand()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "start()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/start()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isRunning()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/Timer/isRunning()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "stop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/stop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "restart()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/restart()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "cancelEvent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/cancelEvent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "post()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/post()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLock()Ljava/util/concurrent/locks/Lock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_Lock) => void): void;
    public "javax/swing/Timer/getLock()Ljava/util/concurrent/locks/Lock;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_locks_Lock) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/Timer/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "javax/swing/Timer/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$000()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "javax/swing/Timer/access$000()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Ljavax/swing/Timer;)Ljava/util/concurrent/atomic/AtomicBoolean;"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicBoolean) => void): void;
    public static "javax/swing/Timer/access$100(Ljavax/swing/Timer;)Ljava/util/concurrent/atomic/AtomicBoolean;"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicBoolean) => void): void;
    public static "access$200(Ljavax/swing/Timer;)Z"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "javax/swing/Timer/access$200(Ljavax/swing/Timer;)Z"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$300(Ljavax/swing/Timer;)Ljava/lang/Runnable;"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runnable) => void): void;
    public static "javax/swing/Timer/access$300(Ljavax/swing/Timer;)Ljava/lang/Runnable;"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runnable) => void): void;
  }
  export class javax_swing_TimerQueue extends java_lang_Object implements java_lang_Runnable {
    public static "javax/swing/TimerQueue/sharedInstanceKey": java_lang_Object;
    public static "javax/swing/TimerQueue/expiredTimersKey": java_lang_Object;
    public "javax/swing/TimerQueue/queue": java_util_concurrent_DelayQueue;
    public "javax/swing/TimerQueue/running": number;
    public "javax/swing/TimerQueue/runningLock": java_util_concurrent_locks_Lock;
    public static "javax/swing/TimerQueue/classLock": java_lang_Object;
    public static "javax/swing/TimerQueue/NANO_ORIGIN": Long;
    public static "javax/swing/TimerQueue/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/TimerQueue/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sharedInstance()Ljavax/swing/TimerQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_swing_TimerQueue) => void): void;
    public static "javax/swing/TimerQueue/sharedInstance()Ljavax/swing/TimerQueue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_swing_TimerQueue) => void): void;
    public "startIfNeeded()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/TimerQueue/startIfNeeded()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addTimer(Ljavax/swing/Timer;J)V"(thread: JVMThread, args: [javax_swing_Timer, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/TimerQueue/addTimer(Ljavax/swing/Timer;J)V"(thread: JVMThread, args: [javax_swing_Timer, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "addTimer(Ljavax/swing/TimerQueue$DelayedTimer;)V"(thread: JVMThread, args: [javax_swing_TimerQueue$DelayedTimer], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/TimerQueue/addTimer(Ljavax/swing/TimerQueue$DelayedTimer;)V"(thread: JVMThread, args: [javax_swing_TimerQueue$DelayedTimer], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeTimer(Ljavax/swing/Timer;)V"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/TimerQueue/removeTimer(Ljavax/swing/Timer;)V"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable) => void): void;
    public "containsTimer(Ljavax/swing/Timer;)Z"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/TimerQueue/containsTimer(Ljavax/swing/Timer;)Z"(thread: JVMThread, args: [javax_swing_Timer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/TimerQueue/run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/swing/TimerQueue/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "now()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "javax/swing/TimerQueue/now()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$000()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "javax/swing/TimerQueue/access$000()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/swing/TimerQueue/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_DelayQueue extends java_util_AbstractQueue implements java_util_concurrent_BlockingQueue {
    public "java/util/concurrent/DelayQueue/lock": java_util_concurrent_locks_ReentrantLock;
    public "java/util/concurrent/DelayQueue/q": java_util_PriorityQueue;
    public "java/util/concurrent/DelayQueue/leader": java_lang_Thread;
    public "java/util/concurrent/DelayQueue/available": java_util_concurrent_locks_Condition;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/DelayQueue/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/DelayQueue/<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/util/concurrent/Delayed;)Z"(thread: JVMThread, args: [java_util_concurrent_Delayed], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/add(Ljava/util/concurrent/Delayed;)Z"(thread: JVMThread, args: [java_util_concurrent_Delayed], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offer(Ljava/util/concurrent/Delayed;)Z"(thread: JVMThread, args: [java_util_concurrent_Delayed], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/offer(Ljava/util/concurrent/Delayed;)Z"(thread: JVMThread, args: [java_util_concurrent_Delayed], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(Ljava/util/concurrent/Delayed;)V"(thread: JVMThread, args: [java_util_concurrent_Delayed], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/DelayQueue/put(Ljava/util/concurrent/Delayed;)V"(thread: JVMThread, args: [java_util_concurrent_Delayed], cb?: (e?: java_lang_Throwable) => void): void;
    public "offer(Ljava/util/concurrent/Delayed;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_util_concurrent_Delayed, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/offer(Ljava/util/concurrent/Delayed;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_util_concurrent_Delayed, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "poll()Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "java/util/concurrent/DelayQueue/poll()Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "take()Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "java/util/concurrent/DelayQueue/take()Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "poll(JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "java/util/concurrent/DelayQueue/poll(JLjava/util/concurrent/TimeUnit;)Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "peek()Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "java/util/concurrent/DelayQueue/peek()Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peekExpired()Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "java/util/concurrent/DelayQueue/peekExpired()Ljava/util/concurrent/Delayed;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Delayed) => void): void;
    public "drainTo(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/drainTo(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drainTo(Ljava/util/Collection;I)I"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/drainTo(Ljava/util/Collection;I)I"(thread: JVMThread, args: [java_util_Collection, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/DelayQueue/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "remainingCapacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/remainingCapacity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/DelayQueue/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/DelayQueue/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeEQ(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/DelayQueue/removeEQ(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/concurrent/DelayQueue/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/DelayQueue/peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/DelayQueue/poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "poll(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/DelayQueue/poll(JLjava/util/concurrent/TimeUnit;)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "take()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/DelayQueue/take()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "offer(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/DelayQueue/offer(Ljava/lang/Object;JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [java_lang_Object, Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/DelayQueue/put(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_PriorityQueue extends java_util_AbstractQueue implements java_io_Serializable {
    public static "java/util/PriorityQueue/serialVersionUID": Long;
    public static "java/util/PriorityQueue/DEFAULT_INITIAL_CAPACITY": number;
    public "java/util/PriorityQueue/queue": JVMArray<java_lang_Object>;
    public "java/util/PriorityQueue/size": number;
    public "java/util/PriorityQueue/comparator": java_util_Comparator;
    public "java/util/PriorityQueue/modCount": number;
    public static "java/util/PriorityQueue/MAX_ARRAY_SIZE": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/<init>(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/util/Comparator;)V"(thread: JVMThread, args: [number, java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/<init>(ILjava/util/Comparator;)V"(thread: JVMThread, args: [number, java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/PriorityQueue;)V"(thread: JVMThread, args: [java_util_PriorityQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/<init>(Ljava/util/PriorityQueue;)V"(thread: JVMThread, args: [java_util_PriorityQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/SortedSet;)V"(thread: JVMThread, args: [java_util_SortedSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/<init>(Ljava/util/SortedSet;)V"(thread: JVMThread, args: [java_util_SortedSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "initFromPriorityQueue(Ljava/util/PriorityQueue;)V"(thread: JVMThread, args: [java_util_PriorityQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/initFromPriorityQueue(Ljava/util/PriorityQueue;)V"(thread: JVMThread, args: [java_util_PriorityQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "initElementsFromCollection(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/initElementsFromCollection(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "initFromCollection(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/initFromCollection(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/PriorityQueue/hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/PriorityQueue/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/PriorityQueue/offer(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/PriorityQueue/peek()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/PriorityQueue/indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/PriorityQueue/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeEq(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/PriorityQueue/removeEq(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/PriorityQueue/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/PriorityQueue/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/PriorityQueue/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/PriorityQueue/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/PriorityQueue/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/PriorityQueue/poll()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "removeAt(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/PriorityQueue/removeAt(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "siftUp(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/siftUp(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "siftUpComparable(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/siftUpComparable(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "siftUpUsingComparator(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/siftUpUsingComparator(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "siftDown(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/siftDown(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "siftDownComparable(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/siftDownComparable(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "siftDownUsingComparator(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/siftDownUsingComparator(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "heapify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/heapify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public "java/util/PriorityQueue/comparator()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/PriorityQueue/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/PriorityQueue/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public static "access$100(Ljava/util/PriorityQueue;)I"(thread: JVMThread, args: [java_util_PriorityQueue], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/PriorityQueue/access$100(Ljava/util/PriorityQueue;)I"(thread: JVMThread, args: [java_util_PriorityQueue], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Ljava/util/PriorityQueue;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_PriorityQueue, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/util/PriorityQueue/access$200(Ljava/util/PriorityQueue;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_PriorityQueue, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class javax_swing_TimerQueue$DelayedTimer extends java_lang_Object implements java_util_concurrent_Delayed {
    public static "javax/swing/TimerQueue$DelayedTimer/sequencer": java_util_concurrent_atomic_AtomicLong;
    public "javax/swing/TimerQueue$DelayedTimer/sequenceNumber": Long;
    public "javax/swing/TimerQueue$DelayedTimer/time": Long;
    public "javax/swing/TimerQueue$DelayedTimer/timer": javax_swing_Timer;
    public "<init>(Ljavax/swing/Timer;J)V"(thread: JVMThread, args: [javax_swing_Timer, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/TimerQueue$DelayedTimer/<init>(Ljavax/swing/Timer;J)V"(thread: JVMThread, args: [javax_swing_Timer, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDelay(Ljava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "javax/swing/TimerQueue$DelayedTimer/getDelay(Ljava/util/concurrent/TimeUnit;)J"(thread: JVMThread, args: [java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setTime(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/TimerQueue$DelayedTimer/setTime(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTimer()Ljavax/swing/Timer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_swing_Timer) => void): void;
    public "javax/swing/TimerQueue$DelayedTimer/getTimer()Ljavax/swing/Timer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_swing_Timer) => void): void;
    public "compareTo(Ljava/util/concurrent/Delayed;)I"(thread: JVMThread, args: [java_util_concurrent_Delayed], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/TimerQueue$DelayedTimer/compareTo(Ljava/util/concurrent/Delayed;)I"(thread: JVMThread, args: [java_util_concurrent_Delayed], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/TimerQueue$DelayedTimer/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/swing/TimerQueue$DelayedTimer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_swing_event_EventListenerList extends java_lang_Object implements java_io_Serializable {
    public static "javax/swing/event/EventListenerList/NULL_ARRAY": JVMArray<java_lang_Object>;
    public "javax/swing/event/EventListenerList/listenerList": JVMArray<java_lang_Object>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/event/EventListenerList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getListenerList()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "javax/swing/event/EventListenerList/getListenerList()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "javax/swing/event/EventListenerList/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "getListenerCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/event/EventListenerList/getListenerCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getListenerCount(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/event/EventListenerList/getListenerCount(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getListenerCount([Ljava/lang/Object;Ljava/lang/Class;)I"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "javax/swing/event/EventListenerList/getListenerCount([Ljava/lang/Object;Ljava/lang/Class;)I"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/lang/Class;Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_Class, java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/event/EventListenerList/add(Ljava/lang/Class;Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_Class, java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/lang/Class;Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_Class, java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/event/EventListenerList/remove(Ljava/lang/Class;Ljava/util/EventListener;)V"(thread: JVMThread, args: [java_lang_Class, java_util_EventListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/event/EventListenerList/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/event/EventListenerList/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "javax/swing/event/EventListenerList/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "javax/swing/event/EventListenerList/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_dnd_Autoscroll extends java_lang_Object {
    "getAutoscrollInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    "autoscroll(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_dnd_DropTargetContext extends java_lang_Object implements java_io_Serializable {
    public static "java/awt/dnd/DropTargetContext/serialVersionUID": Long;
    public "java/awt/dnd/DropTargetContext/dropTarget": java_awt_dnd_DropTarget;
    public "java/awt/dnd/DropTargetContext/dropTargetContextPeer": java_awt_dnd_peer_DropTargetContextPeer;
    public "java/awt/dnd/DropTargetContext/transferable": java_awt_datatransfer_Transferable;
    public "<init>(Ljava/awt/dnd/DropTarget;)V"(thread: JVMThread, args: [java_awt_dnd_DropTarget], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetContext/<init>(Ljava/awt/dnd/DropTarget;)V"(thread: JVMThread, args: [java_awt_dnd_DropTarget], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDropTarget()Ljava/awt/dnd/DropTarget;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTarget) => void): void;
    public "java/awt/dnd/DropTargetContext/getDropTarget()Ljava/awt/dnd/DropTarget;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTarget) => void): void;
    public "getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/dnd/DropTargetContext/getComponent()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "addNotify(Ljava/awt/dnd/peer/DropTargetContextPeer;)V"(thread: JVMThread, args: [java_awt_dnd_peer_DropTargetContextPeer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetContext/addNotify(Ljava/awt/dnd/peer/DropTargetContextPeer;)V"(thread: JVMThread, args: [java_awt_dnd_peer_DropTargetContextPeer], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetContext/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTargetActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetContext/setTargetActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTargetActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTargetContext/getTargetActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "dropComplete(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetContext/dropComplete(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "acceptDrag(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetContext/acceptDrag(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rejectDrag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetContext/rejectDrag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "acceptDrop(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetContext/acceptDrop(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rejectDrop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/dnd/DropTargetContext/rejectDrop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCurrentDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "java/awt/dnd/DropTargetContext/getCurrentDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "getCurrentDataFlavorsAsList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/awt/dnd/DropTargetContext/getCurrentDataFlavorsAsList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "isDataFlavorSupported(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/dnd/DropTargetContext/isDataFlavorSupported(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "java/awt/dnd/DropTargetContext/getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "getDropTargetContextPeer()Ljava/awt/dnd/peer/DropTargetContextPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_peer_DropTargetContextPeer) => void): void;
    public "java/awt/dnd/DropTargetContext/getDropTargetContextPeer()Ljava/awt/dnd/peer/DropTargetContextPeer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_peer_DropTargetContextPeer) => void): void;
    public "createTransferableProxy(Ljava/awt/datatransfer/Transferable;Z)Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: [java_awt_datatransfer_Transferable, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "java/awt/dnd/DropTargetContext/createTransferableProxy(Ljava/awt/datatransfer/Transferable;Z)Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: [java_awt_datatransfer_Transferable, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
  }
  export interface java_awt_dnd_peer_DropTargetContextPeer extends java_lang_Object {
    "setTargetActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getTargetActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getDropTarget()Ljava/awt/dnd/DropTarget;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTarget) => void): void;
    "getTransferDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    "getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    "isTransferableJVMLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "acceptDrag(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "rejectDrag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "acceptDrop(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "rejectDrop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "dropComplete(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_dnd_DropTargetListener extends java_lang_Object, java_util_EventListener {
    "dragEnter(Ljava/awt/dnd/DropTargetDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "dragOver(Ljava/awt/dnd/DropTargetDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "dropActionChanged(Ljava/awt/dnd/DropTargetDragEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDragEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "dragExit(Ljava/awt/dnd/DropTargetEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "drop(Ljava/awt/dnd/DropTargetDropEvent;)V"(thread: JVMThread, args: [java_awt_dnd_DropTargetDropEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface javax_accessibility_Accessible extends java_lang_Object {
    "getAccessibleContext()Ljavax/accessibility/AccessibleContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_AccessibleContext) => void): void;
  }
  export class java_awt_Container extends java_awt_Component {
    public static "java/awt/Container/log": sun_util_logging_PlatformLogger;
    public static "java/awt/Container/eventLog": sun_util_logging_PlatformLogger;
    public static "java/awt/Container/EMPTY_ARRAY": JVMArray<java_awt_Component>;
    public "java/awt/Container/component": java_util_List;
    public "java/awt/Container/layoutMgr": java_awt_LayoutManager;
    public "java/awt/Container/dispatcher": java_awt_LightweightDispatcher;
    public "java/awt/Container/focusTraversalPolicy": java_awt_FocusTraversalPolicy;
    public "java/awt/Container/focusCycleRoot": number;
    public "java/awt/Container/focusTraversalPolicyProvider": number;
    public "java/awt/Container/printingThreads": java_util_Set;
    public "java/awt/Container/printing": number;
    public "java/awt/Container/containerListener": java_awt_event_ContainerListener;
    public "java/awt/Container/listeningChildren": number;
    public "java/awt/Container/listeningBoundsChildren": number;
    public "java/awt/Container/descendantsCount": number;
    public "java/awt/Container/preserveBackgroundColor": java_awt_Color;
    public static "java/awt/Container/serialVersionUID": Long;
    public static "java/awt/Container/INCLUDE_SELF": number;
    public static "java/awt/Container/SEARCH_HEAVYWEIGHTS": number;
    public "java/awt/Container/numOfHWComponents": number;
    public "java/awt/Container/numOfLWComponents": number;
    public static "java/awt/Container/mixingLog": sun_util_logging_PlatformLogger;
    public static "java/awt/Container/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/awt/Container/isJavaAwtSmartInvalidate": number;
    public static "java/awt/Container/descendUnconditionallyWhenValidating": number;
    public "java/awt/Container/modalComp": java_awt_Component;
    public "java/awt/Container/modalAppContext": sun_awt_AppContext;
    public "java/awt/Container/containerSerializedDataVersion": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Container/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "initializeFocusTraversalKeys()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/initializeFocusTraversalKeys()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComponentCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/getComponentCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/countComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getComponent(I)Ljava/awt/Component;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/getComponent(I)Ljava/awt/Component;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getComponents()[Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Component>) => void): void;
    public "java/awt/Container/getComponents()[Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Component>) => void): void;
    public "getComponents_NoClientCode()[Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Component>) => void): void;
    public "java/awt/Container/getComponents_NoClientCode()[Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Component>) => void): void;
    public "getComponentsSync()[Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Component>) => void): void;
    public "java/awt/Container/getComponentsSync()[Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Component>) => void): void;
    public "getInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public "java/awt/Container/getInsets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public "insets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public "java/awt/Container/insets()Ljava/awt/Insets;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Insets) => void): void;
    public "add(Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/add(Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "add(Ljava/lang/String;Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_lang_String, java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/add(Ljava/lang/String;Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_lang_String, java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "add(Ljava/awt/Component;I)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/add(Ljava/awt/Component;I)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "checkAddToSelf(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/checkAddToSelf(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkNotAWindow(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/checkNotAWindow(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAdding(Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/checkAdding(Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeDelicately(Ljava/awt/Component;Ljava/awt/Container;I)Z"(thread: JVMThread, args: [java_awt_Component, java_awt_Container, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/removeDelicately(Ljava/awt/Component;Ljava/awt/Container;I)Z"(thread: JVMThread, args: [java_awt_Component, java_awt_Container, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canContainFocusOwner(Ljava/awt/Component;)Z"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/canContainFocusOwner(Ljava/awt/Component;)Z"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasHeavyweightDescendants()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/hasHeavyweightDescendants()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasLightweightDescendants()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/hasLightweightDescendants()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeavyweightContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Container/getHeavyweightContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public static "isRemoveNotifyNeeded(Ljava/awt/Component;Ljava/awt/Container;Ljava/awt/Container;)Z"(thread: JVMThread, args: [java_awt_Component, java_awt_Container, java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/Container/isRemoveNotifyNeeded(Ljava/awt/Component;Ljava/awt/Container;Ljava/awt/Container;)Z"(thread: JVMThread, args: [java_awt_Component, java_awt_Container, java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setComponentZOrder(Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/setComponentZOrder(Ljava/awt/Component;I)V"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "reparentTraverse(Ljava/awt/peer/ContainerPeer;Ljava/awt/Container;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer, java_awt_Container], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/reparentTraverse(Ljava/awt/peer/ContainerPeer;Ljava/awt/Container;)V"(thread: JVMThread, args: [java_awt_peer_ContainerPeer, java_awt_Container], cb?: (e?: java_lang_Throwable) => void): void;
    public "reparentChild(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/reparentChild(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "addDelicately(Ljava/awt/Component;Ljava/awt/Container;I)V"(thread: JVMThread, args: [java_awt_Component, java_awt_Container, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/addDelicately(Ljava/awt/Component;Ljava/awt/Container;I)V"(thread: JVMThread, args: [java_awt_Component, java_awt_Container, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComponentZOrder(Ljava/awt/Component;)I"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/getComponentZOrder(Ljava/awt/Component;)I"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/awt/Component;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/add(Ljava/awt/Component;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(Ljava/awt/Component;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/add(Ljava/awt/Component;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addImpl(Ljava/awt/Component;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/addImpl(Ljava/awt/Component;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_awt_Component, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/updateGraphicsData(Ljava/awt/GraphicsConfiguration;)Z"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkGD(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/checkGD(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/remove(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/removeAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "numListening(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/numListening(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "adjustListeningChildren(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/adjustListeningChildren(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "adjustDescendants(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/adjustDescendants(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "adjustDecendantsOnParent(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/adjustDecendantsOnParent(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "countHierarchyMembers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/countHierarchyMembers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getListenersCount(IZ)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/getListenersCount(IZ)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createHierarchyEvents(ILjava/awt/Component;Ljava/awt/Container;JZ)I"(thread: JVMThread, args: [number, java_awt_Component, java_awt_Container, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/createHierarchyEvents(ILjava/awt/Component;Ljava/awt/Container;JZ)I"(thread: JVMThread, args: [number, java_awt_Component, java_awt_Container, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createChildHierarchyEvents(IJZ)V"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/createChildHierarchyEvents(IJZ)V"(thread: JVMThread, args: [number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLayout()Ljava/awt/LayoutManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_LayoutManager) => void): void;
    public "java/awt/Container/getLayout()Ljava/awt/LayoutManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_LayoutManager) => void): void;
    public "setLayout(Ljava/awt/LayoutManager;)V"(thread: JVMThread, args: [java_awt_LayoutManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/setLayout(Ljava/awt/LayoutManager;)V"(thread: JVMThread, args: [java_awt_LayoutManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/doLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/layout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isValidateRoot()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/isValidateRoot()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "invalidateParent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/invalidateParent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "validate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/validate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "validateUnconditionally()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/validateUnconditionally()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "validateTree()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/validateTree()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidateTree()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/invalidateTree()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Container/getPreferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Container/preferredSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Container/getMinimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Container/minimumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getMaximumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "java/awt/Container/getMaximumSize()Ljava/awt/Dimension;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Dimension) => void): void;
    public "getAlignmentX()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/getAlignmentX()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAlignmentY()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/getAlignmentY()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/paint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/update(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/print(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "paintComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/paintComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "lightweightPaint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/lightweightPaint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "paintHeavyweightComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/paintHeavyweightComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "printComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/printComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "lightweightPrint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/lightweightPrint(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "printHeavyweightComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/printHeavyweightComponents(Ljava/awt/Graphics;)V"(thread: JVMThread, args: [java_awt_Graphics], cb?: (e?: java_lang_Throwable) => void): void;
    public "addContainerListener(Ljava/awt/event/ContainerListener;)V"(thread: JVMThread, args: [java_awt_event_ContainerListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/addContainerListener(Ljava/awt/event/ContainerListener;)V"(thread: JVMThread, args: [java_awt_event_ContainerListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeContainerListener(Ljava/awt/event/ContainerListener;)V"(thread: JVMThread, args: [java_awt_event_ContainerListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/removeContainerListener(Ljava/awt/event/ContainerListener;)V"(thread: JVMThread, args: [java_awt_event_ContainerListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContainerListeners()[Ljava/awt/event/ContainerListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ContainerListener>) => void): void;
    public "java/awt/Container/getContainerListeners()[Ljava/awt/event/ContainerListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_event_ContainerListener>) => void): void;
    public "getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "java/awt/Container/getListeners(Ljava/lang/Class;)[Ljava/util/EventListener;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_EventListener>) => void): void;
    public "eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/eventEnabled(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/processEvent(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "processContainerEvent(Ljava/awt/event/ContainerEvent;)V"(thread: JVMThread, args: [java_awt_event_ContainerEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/processContainerEvent(Ljava/awt/event/ContainerEvent;)V"(thread: JVMThread, args: [java_awt_event_ContainerEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchEventImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/dispatchEventImpl(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchEventToSelf(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/dispatchEventToSelf(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMouseEventTarget(IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/getMouseEventTarget(IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getDropTargetEventTarget(IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/getDropTargetEventTarget(IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getMouseEventTarget(IIZLjava/awt/Container$EventTargetFilter;Z)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number, java_awt_Container$EventTargetFilter, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/getMouseEventTarget(IIZLjava/awt/Container$EventTargetFilter;Z)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number, java_awt_Container$EventTargetFilter, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getMouseEventTargetImpl(IIZLjava/awt/Container$EventTargetFilter;ZZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number, java_awt_Container$EventTargetFilter, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/getMouseEventTargetImpl(IIZLjava/awt/Container$EventTargetFilter;ZZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number, java_awt_Container$EventTargetFilter, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "proxyEnableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/proxyEnableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "deliverEvent(Ljava/awt/Event;)V"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/deliverEvent(Ljava/awt/Event;)V"(thread: JVMThread, args: [java_awt_Event], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComponentAt(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/getComponentAt(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "locate(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/locate(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getComponentAt(Ljava/awt/Point;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/getComponentAt(Ljava/awt/Point;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getMousePosition(Z)Ljava/awt/Point;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "java/awt/Container/getMousePosition(Z)Ljava/awt/Point;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Point) => void): void;
    public "isSameOrAncestorOf(Ljava/awt/Component;Z)Z"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/isSameOrAncestorOf(Ljava/awt/Component;Z)Z"(thread: JVMThread, args: [java_awt_Component, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "findComponentAt(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/findComponentAt(II)Ljava/awt/Component;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "findComponentAt(IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/findComponentAt(IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "findComponentAtImpl(IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/findComponentAtImpl(IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public static "getChildAt(Ljava/awt/Component;IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public static "java/awt/Container/getChildAt(Ljava/awt/Component;IIZ)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "findComponentAt(Ljava/awt/Point;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/Container/findComponentAt(Ljava/awt/Point;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/addNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/removeNotify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAncestorOf(Ljava/awt/Component;)Z"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/isAncestorOf(Ljava/awt/Component;)Z"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "startLWModal()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/startLWModal()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "stopLWModal()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/stopLWModal()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Container/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "list(Ljava/io/PrintStream;I)V"(thread: JVMThread, args: [java_io_PrintStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/list(Ljava/io/PrintStream;I)V"(thread: JVMThread, args: [java_io_PrintStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "list(Ljava/io/PrintWriter;I)V"(thread: JVMThread, args: [java_io_PrintWriter, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/list(Ljava/io/PrintWriter;I)V"(thread: JVMThread, args: [java_io_PrintWriter, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFocusTraversalKeys(ILjava/util/Set;)V"(thread: JVMThread, args: [number, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/setFocusTraversalKeys(ILjava/util/Set;)V"(thread: JVMThread, args: [number, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFocusTraversalKeys(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/awt/Container/getFocusTraversalKeys(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "areFocusTraversalKeysSet(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/areFocusTraversalKeysSet(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFocusCycleRoot(Ljava/awt/Container;)Z"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/isFocusCycleRoot(Ljava/awt/Container;)Z"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "findTraversalRoot()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Container/findTraversalRoot()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "containsFocus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/containsFocus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isParentOf(Ljava/awt/Component;)Z"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/isParentOf(Ljava/awt/Component;)Z"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clearMostRecentFocusOwnerOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/clearMostRecentFocusOwnerOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearCurrentFocusCycleRootOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/clearCurrentFocusCycleRootOnHide()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTraversalRoot()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/Container/getTraversalRoot()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "setFocusTraversalPolicy(Ljava/awt/FocusTraversalPolicy;)V"(thread: JVMThread, args: [java_awt_FocusTraversalPolicy], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/setFocusTraversalPolicy(Ljava/awt/FocusTraversalPolicy;)V"(thread: JVMThread, args: [java_awt_FocusTraversalPolicy], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFocusTraversalPolicy()Ljava/awt/FocusTraversalPolicy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_FocusTraversalPolicy) => void): void;
    public "java/awt/Container/getFocusTraversalPolicy()Ljava/awt/FocusTraversalPolicy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_FocusTraversalPolicy) => void): void;
    public "isFocusTraversalPolicySet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/isFocusTraversalPolicySet()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFocusCycleRoot(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/setFocusCycleRoot(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isFocusCycleRoot()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/isFocusCycleRoot()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFocusTraversalPolicyProvider(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/setFocusTraversalPolicyProvider(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isFocusTraversalPolicyProvider()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/isFocusTraversalPolicyProvider()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "transferFocusDownCycle()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/transferFocusDownCycle()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "preProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/preProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "postProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/postProcessKeyEvent(Ljava/awt/event/KeyEvent;)V"(thread: JVMThread, args: [java_awt_event_KeyEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "postsOldMouseEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/postsOldMouseEvents()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "applyComponentOrientation(Ljava/awt/ComponentOrientation;)V"(thread: JVMThread, args: [java_awt_ComponentOrientation], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/applyComponentOrientation(Ljava/awt/ComponentOrientation;)V"(thread: JVMThread, args: [java_awt_ComponentOrientation], cb?: (e?: java_lang_Throwable) => void): void;
    public "addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/addPropertyChangeListener(Ljava/lang/String;Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_lang_String, java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccessibleAt(Ljava/awt/Point;)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    public "java/awt/Container/getAccessibleAt(Ljava/awt/Point;)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    public "getAccessibleChildrenCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/getAccessibleChildrenCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAccessibleChild(I)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    public "java/awt/Container/getAccessibleChild(I)Ljavax/accessibility/Accessible;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: javax_accessibility_Accessible) => void): void;
    public "increaseComponentCount(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/increaseComponentCount(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "decreaseComponentCount(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/decreaseComponentCount(Ljava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTopmostComponentIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/getTopmostComponentIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBottommostComponentIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/getBottommostComponentIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOpaqueShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "java/awt/Container/getOpaqueShape()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "recursiveSubtractAndApplyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/recursiveSubtractAndApplyShape(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "recursiveSubtractAndApplyShape(Lsun/java2d/pipe/Region;I)V"(thread: JVMThread, args: [sun_java2d_pipe_Region, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/recursiveSubtractAndApplyShape(Lsun/java2d/pipe/Region;I)V"(thread: JVMThread, args: [sun_java2d_pipe_Region, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "recursiveSubtractAndApplyShape(Lsun/java2d/pipe/Region;II)V"(thread: JVMThread, args: [sun_java2d_pipe_Region, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/recursiveSubtractAndApplyShape(Lsun/java2d/pipe/Region;II)V"(thread: JVMThread, args: [sun_java2d_pipe_Region, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "recursiveApplyCurrentShape()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/recursiveApplyCurrentShape()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "recursiveApplyCurrentShape(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/recursiveApplyCurrentShape(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "recursiveApplyCurrentShape(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/recursiveApplyCurrentShape(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "recursiveShowHeavyweightChildren()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/recursiveShowHeavyweightChildren()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "recursiveHideHeavyweightChildren()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/recursiveHideHeavyweightChildren()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "recursiveRelocateHeavyweightChildren(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/recursiveRelocateHeavyweightChildren(Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "isRecursivelyVisibleUpToHeavyweightContainer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Container/isRecursivelyVisibleUpToHeavyweightContainer()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mixOnShowing()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/mixOnShowing()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnHiding(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/mixOnHiding(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnReshaping()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/mixOnReshaping()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnZOrderChanging(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/mixOnZOrderChanging(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "mixOnValidating()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Container/mixOnValidating()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Container/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_Container$EventTargetFilter extends java_lang_Object {
    "accept(Ljava/awt/Component;)Z"(thread: JVMThread, args: [java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_event_ContainerEvent extends java_awt_event_ComponentEvent {
    public static "java/awt/event/ContainerEvent/CONTAINER_FIRST": number;
    public static "java/awt/event/ContainerEvent/CONTAINER_LAST": number;
    public static "java/awt/event/ContainerEvent/COMPONENT_ADDED": number;
    public static "java/awt/event/ContainerEvent/COMPONENT_REMOVED": number;
    public "java/awt/event/ContainerEvent/child": java_awt_Component;
    public static "java/awt/event/ContainerEvent/serialVersionUID": Long;
    public "<init>(Ljava/awt/Component;ILjava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/event/ContainerEvent/<init>(Ljava/awt/Component;ILjava/awt/Component;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_Component], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "java/awt/event/ContainerEvent/getContainer()Ljava/awt/Container;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public "getChild()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/event/ContainerEvent/getChild()Ljava/awt/Component;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/event/ContainerEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_awt_event_ContainerListener extends java_lang_Object, java_util_EventListener {
    "componentAdded(Ljava/awt/event/ContainerEvent;)V"(thread: JVMThread, args: [java_awt_event_ContainerEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "componentRemoved(Ljava/awt/event/ContainerEvent;)V"(thread: JVMThread, args: [java_awt_event_ContainerEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_FocusTraversalPolicy extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/FocusTraversalPolicy/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComponentAfter(Ljava/awt/Container;Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container, java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/FocusTraversalPolicy/getComponentAfter(Ljava/awt/Container;Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container, java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getComponentBefore(Ljava/awt/Container;Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container, java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/FocusTraversalPolicy/getComponentBefore(Ljava/awt/Container;Ljava/awt/Component;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container, java_awt_Component], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getFirstComponent(Ljava/awt/Container;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/FocusTraversalPolicy/getFirstComponent(Ljava/awt/Container;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getLastComponent(Ljava/awt/Container;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/FocusTraversalPolicy/getLastComponent(Ljava/awt/Container;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getDefaultComponent(Ljava/awt/Container;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/FocusTraversalPolicy/getDefaultComponent(Ljava/awt/Container;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "getInitialComponent(Ljava/awt/Window;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/FocusTraversalPolicy/getInitialComponent(Ljava/awt/Window;)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Window], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
  }
  export class java_awt_LightweightDispatcher extends java_lang_Object implements java_io_Serializable, java_awt_event_AWTEventListener {
    public static "java/awt/LightweightDispatcher/serialVersionUID": Long;
    public static "java/awt/LightweightDispatcher/LWD_MOUSE_DRAGGED_OVER": number;
    public static "java/awt/LightweightDispatcher/eventLog": sun_util_logging_PlatformLogger;
    public static "java/awt/LightweightDispatcher/BUTTONS_DOWN_MASK": number;
    public "java/awt/LightweightDispatcher/nativeContainer": java_awt_Container;
    public "java/awt/LightweightDispatcher/focus": java_awt_Component;
    public "java/awt/LightweightDispatcher/mouseEventTarget": java_lang_ref_WeakReference;
    public "java/awt/LightweightDispatcher/targetLastEntered": java_lang_ref_WeakReference;
    public "java/awt/LightweightDispatcher/targetLastEnteredDT": java_lang_ref_WeakReference;
    public "java/awt/LightweightDispatcher/isMouseInNativeContainer": number;
    public "java/awt/LightweightDispatcher/isMouseDTInNativeContainer": number;
    public "java/awt/LightweightDispatcher/nativeCursor": java_awt_Cursor;
    public "java/awt/LightweightDispatcher/eventMask": Long;
    public static "java/awt/LightweightDispatcher/PROXY_EVENT_MASK": Long;
    public static "java/awt/LightweightDispatcher/MOUSE_MASK": Long;
    public static "java/awt/LightweightDispatcher/$assertionsDisabled": number;
    public "<init>(Ljava/awt/Container;)V"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/LightweightDispatcher/<init>(Ljava/awt/Container;)V"(thread: JVMThread, args: [java_awt_Container], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/LightweightDispatcher/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "enableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/LightweightDispatcher/enableEvents(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchEvent(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/LightweightDispatcher/dispatchEvent(Ljava/awt/AWTEvent;)Z"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMouseGrab(Ljava/awt/event/MouseEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/LightweightDispatcher/isMouseGrab(Ljava/awt/event/MouseEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processMouseEvent(Ljava/awt/event/MouseEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/LightweightDispatcher/processMouseEvent(Ljava/awt/event/MouseEvent;)Z"(thread: JVMThread, args: [java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processDropTargetEvent(Lsun/awt/dnd/SunDropTargetEvent;)Z"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/LightweightDispatcher/processDropTargetEvent(Lsun/awt/dnd/SunDropTargetEvent;)Z"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "trackDropTargetEnterExit(Ljava/awt/Component;Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_Component, java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/LightweightDispatcher/trackDropTargetEnterExit(Ljava/awt/Component;Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_Component, java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "trackMouseEnterExit(Ljava/awt/Component;Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_Component, java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/LightweightDispatcher/trackMouseEnterExit(Ljava/awt/Component;Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_Component, java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "retargetMouseEnterExit(Ljava/awt/Component;Ljava/awt/event/MouseEvent;Ljava/awt/Component;Z)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component, java_awt_event_MouseEvent, java_awt_Component, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "java/awt/LightweightDispatcher/retargetMouseEnterExit(Ljava/awt/Component;Ljava/awt/event/MouseEvent;Ljava/awt/Component;Z)Ljava/awt/Component;"(thread: JVMThread, args: [java_awt_Component, java_awt_event_MouseEvent, java_awt_Component, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Component) => void): void;
    public "startListeningForOtherDrags()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/LightweightDispatcher/startListeningForOtherDrags()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "stopListeningForOtherDrags()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/LightweightDispatcher/stopListeningForOtherDrags()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "eventDispatched(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/LightweightDispatcher/eventDispatched(Ljava/awt/AWTEvent;)V"(thread: JVMThread, args: [java_awt_AWTEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "retargetMouseEvent(Ljava/awt/Component;ILjava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/LightweightDispatcher/retargetMouseEvent(Ljava/awt/Component;ILjava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_Component, number, java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/awt/LightweightDispatcher;)Ljava/awt/Container;"(thread: JVMThread, args: [java_awt_LightweightDispatcher], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public static "java/awt/LightweightDispatcher/access$000(Ljava/awt/LightweightDispatcher;)Ljava/awt/Container;"(thread: JVMThread, args: [java_awt_LightweightDispatcher], cb?: (e?: java_lang_Throwable, rv?: java_awt_Container) => void): void;
    public static "access$100(Ljava/awt/LightweightDispatcher;Ljava/awt/Component;Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_LightweightDispatcher, java_awt_Component, java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/LightweightDispatcher/access$100(Ljava/awt/LightweightDispatcher;Ljava/awt/Component;Ljava/awt/event/MouseEvent;)V"(thread: JVMThread, args: [java_awt_LightweightDispatcher, java_awt_Component, java_awt_event_MouseEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/LightweightDispatcher/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_dnd_SunDropTargetEvent extends java_awt_event_MouseEvent {
    public static "sun/awt/dnd/SunDropTargetEvent/MOUSE_DROPPED": number;
    public "sun/awt/dnd/SunDropTargetEvent/dispatcher": sun_awt_dnd_SunDropTargetContextPeer$EventDispatcher;
    public "<init>(Ljava/awt/Component;IIILsun/awt/dnd/SunDropTargetContextPeer$EventDispatcher;)V"(thread: JVMThread, args: [java_awt_Component, number, number, number, sun_awt_dnd_SunDropTargetContextPeer$EventDispatcher], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetEvent/<init>(Ljava/awt/Component;IIILsun/awt/dnd/SunDropTargetContextPeer$EventDispatcher;)V"(thread: JVMThread, args: [java_awt_Component, number, number, number, sun_awt_dnd_SunDropTargetContextPeer$EventDispatcher], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatch()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetEvent/dispatch()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetEvent/consume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDispatcher()Lsun/awt/dnd/SunDropTargetContextPeer$EventDispatcher;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_dnd_SunDropTargetContextPeer$EventDispatcher) => void): void;
    public "sun/awt/dnd/SunDropTargetEvent/getDispatcher()Lsun/awt/dnd/SunDropTargetContextPeer$EventDispatcher;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_dnd_SunDropTargetContextPeer$EventDispatcher) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/dnd/SunDropTargetEvent/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_awt_dnd_SunDropTargetContextPeer$EventDispatcher extends java_lang_Object {
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/peer": sun_awt_dnd_SunDropTargetContextPeer;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/dropAction": number;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/actions": number;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/formats": JVMArray<Long>;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/nativeCtxt": Long;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/dispatchType": number;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/dispatcherDone": number;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/returnValue": number;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/eventSet": java_util_HashSet;
    public static "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/handler": sun_awt_datatransfer_ToolkitThreadBlockedHandler;
    public "<init>(Lsun/awt/dnd/SunDropTargetContextPeer;II[JJZ)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, number, number, JVMArray<Long>, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/<init>(Lsun/awt/dnd/SunDropTargetContextPeer;II[JJZ)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, number, number, JVMArray<Long>, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/dispatchEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchEnterEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/dispatchEnterEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchMotionEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/dispatchMotionEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchExitEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/dispatchExitEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchDropEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/dispatchDropEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "setReturnValue(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/setReturnValue(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getReturnValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/getReturnValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/isDone()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "registerEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/registerEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "unregisterEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/unregisterEvent(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "unregisterAllEvents()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/unregisterAllEvents()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer$EventDispatcher/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_awt_datatransfer_ToolkitThreadBlockedHandler extends java_lang_Object {
    "lock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "unlock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "enter()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "exit()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_dnd_SunDropTargetContextPeer extends java_lang_Object implements java_awt_dnd_peer_DropTargetContextPeer, java_awt_datatransfer_Transferable {
    public static "sun/awt/dnd/SunDropTargetContextPeer/DISPATCH_SYNC": number;
    public "sun/awt/dnd/SunDropTargetContextPeer/currentDT": java_awt_dnd_DropTarget;
    public "sun/awt/dnd/SunDropTargetContextPeer/currentDTC": java_awt_dnd_DropTargetContext;
    public "sun/awt/dnd/SunDropTargetContextPeer/currentT": JVMArray<Long>;
    public "sun/awt/dnd/SunDropTargetContextPeer/currentA": number;
    public "sun/awt/dnd/SunDropTargetContextPeer/currentSA": number;
    public "sun/awt/dnd/SunDropTargetContextPeer/currentDA": number;
    public "sun/awt/dnd/SunDropTargetContextPeer/previousDA": number;
    public "sun/awt/dnd/SunDropTargetContextPeer/nativeDragContext": Long;
    public "sun/awt/dnd/SunDropTargetContextPeer/local": java_awt_datatransfer_Transferable;
    public "sun/awt/dnd/SunDropTargetContextPeer/dragRejected": number;
    public "sun/awt/dnd/SunDropTargetContextPeer/dropStatus": number;
    public "sun/awt/dnd/SunDropTargetContextPeer/dropComplete": number;
    public "sun/awt/dnd/SunDropTargetContextPeer/dropInProcess": number;
    public static "sun/awt/dnd/SunDropTargetContextPeer/_globalLock": java_lang_Object;
    public static "sun/awt/dnd/SunDropTargetContextPeer/dndLog": sun_util_logging_PlatformLogger;
    public static "sun/awt/dnd/SunDropTargetContextPeer/currentJVMLocalSourceTransferable": java_awt_datatransfer_Transferable;
    public static "sun/awt/dnd/SunDropTargetContextPeer/STATUS_NONE": number;
    public static "sun/awt/dnd/SunDropTargetContextPeer/STATUS_WAIT": number;
    public static "sun/awt/dnd/SunDropTargetContextPeer/STATUS_ACCEPT": number;
    public static "sun/awt/dnd/SunDropTargetContextPeer/STATUS_REJECT": number;
    public static "setCurrentJVMLocalSourceTransferable(Ljava/awt/datatransfer/Transferable;)V"(thread: JVMThread, args: [java_awt_datatransfer_Transferable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/setCurrentJVMLocalSourceTransferable(Ljava/awt/datatransfer/Transferable;)V"(thread: JVMThread, args: [java_awt_datatransfer_Transferable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getJVMLocalSourceTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/getJVMLocalSourceTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDropTarget()Ljava/awt/dnd/DropTarget;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTarget) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/getDropTarget()Ljava/awt/dnd/DropTarget;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_dnd_DropTarget) => void): void;
    public "setTargetActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/setTargetActions(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTargetActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/getTargetActions()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/getTransferable()Ljava/awt/datatransfer/Transferable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_datatransfer_Transferable) => void): void;
    public "getTransferDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/getTransferDataFlavors()[Ljava/awt/datatransfer/DataFlavor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_datatransfer_DataFlavor>) => void): void;
    public "isDataFlavorSupported(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/isDataFlavorSupported(Ljava/awt/datatransfer/DataFlavor;)Z"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransferData(Ljava/awt/datatransfer/DataFlavor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/getTransferData(Ljava/awt/datatransfer/DataFlavor;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_datatransfer_DataFlavor], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getNativeData(J)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/getNativeData(J)Ljava/lang/Object;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isTransferableJVMLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/isTransferableJVMLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "handleEnterMessage(Ljava/awt/Component;IIII[JJ)I"(thread: JVMThread, args: [java_awt_Component, number, number, number, number, JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/handleEnterMessage(Ljava/awt/Component;IIII[JJ)I"(thread: JVMThread, args: [java_awt_Component, number, number, number, number, JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processEnterMessage(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/processEnterMessage(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "handleExitMessage(Ljava/awt/Component;J)V"(thread: JVMThread, args: [java_awt_Component, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/handleExitMessage(Ljava/awt/Component;J)V"(thread: JVMThread, args: [java_awt_Component, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "processExitMessage(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/processExitMessage(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "handleMotionMessage(Ljava/awt/Component;IIII[JJ)I"(thread: JVMThread, args: [java_awt_Component, number, number, number, number, JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/handleMotionMessage(Ljava/awt/Component;IIII[JJ)I"(thread: JVMThread, args: [java_awt_Component, number, number, number, number, JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "processMotionMessage(Lsun/awt/dnd/SunDropTargetEvent;Z)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/processMotionMessage(Lsun/awt/dnd/SunDropTargetEvent;Z)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "handleDropMessage(Ljava/awt/Component;IIII[JJ)V"(thread: JVMThread, args: [java_awt_Component, number, number, number, number, JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/handleDropMessage(Ljava/awt/Component;IIII[JJ)V"(thread: JVMThread, args: [java_awt_Component, number, number, number, number, JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "processDropMessage(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/processDropMessage(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "postDropTargetEvent(Ljava/awt/Component;IIII[JJIZ)I"(thread: JVMThread, args: [java_awt_Component, number, number, number, number, JVMArray<Long>, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/postDropTargetEvent(Ljava/awt/Component;IIII[JJIZ)I"(thread: JVMThread, args: [java_awt_Component, number, number, number, number, JVMArray<Long>, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "acceptDrag(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/acceptDrag(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rejectDrag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/rejectDrag()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "acceptDrop(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/acceptDrop(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rejectDrop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/rejectDrop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mapOperation(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/mapOperation(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "dropComplete(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/dropComplete(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "doDropDone(ZIZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/doDropDone(ZIZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNativeDragContext()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/getNativeDragContext()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "eventPosted(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/eventPosted(Lsun/awt/dnd/SunDropTargetEvent;)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent], cb?: (e?: java_lang_Throwable) => void): void;
    public "eventProcessed(Lsun/awt/dnd/SunDropTargetEvent;IZ)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/dnd/SunDropTargetContextPeer/eventProcessed(Lsun/awt/dnd/SunDropTargetEvent;IZ)V"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetEvent, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$002(Lsun/awt/dnd/SunDropTargetContextPeer;I)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/access$002(Lsun/awt/dnd/SunDropTargetContextPeer;I)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$102(Lsun/awt/dnd/SunDropTargetContextPeer;J)J"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/access$102(Lsun/awt/dnd/SunDropTargetContextPeer;J)J"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$202(Lsun/awt/dnd/SunDropTargetContextPeer;[J)[J"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/access$202(Lsun/awt/dnd/SunDropTargetContextPeer;[J)[J"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, JVMArray<Long>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public static "access$302(Lsun/awt/dnd/SunDropTargetContextPeer;I)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/access$302(Lsun/awt/dnd/SunDropTargetContextPeer;I)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$402(Lsun/awt/dnd/SunDropTargetContextPeer;I)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/access$402(Lsun/awt/dnd/SunDropTargetContextPeer;I)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$400(Lsun/awt/dnd/SunDropTargetContextPeer;)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/access$400(Lsun/awt/dnd/SunDropTargetContextPeer;)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Lsun/awt/dnd/SunDropTargetContextPeer;)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/access$000(Lsun/awt/dnd/SunDropTargetContextPeer;)I"(thread: JVMThread, args: [sun_awt_dnd_SunDropTargetContextPeer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/access$500()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/dnd/SunDropTargetContextPeer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_ImageCapabilities extends java_lang_Object implements java_lang_Cloneable {
    public "java/awt/ImageCapabilities/accelerated": number;
    public "<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/ImageCapabilities/<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAccelerated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ImageCapabilities/isAccelerated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isTrueVolatile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/ImageCapabilities/isTrueVolatile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/ImageCapabilities/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_awt_BufferCapabilities extends java_lang_Object implements java_lang_Cloneable {
    public "java/awt/BufferCapabilities/frontCaps": java_awt_ImageCapabilities;
    public "java/awt/BufferCapabilities/backCaps": java_awt_ImageCapabilities;
    public "java/awt/BufferCapabilities/flipContents": java_awt_BufferCapabilities$FlipContents;
    public "<init>(Ljava/awt/ImageCapabilities;Ljava/awt/ImageCapabilities;Ljava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [java_awt_ImageCapabilities, java_awt_ImageCapabilities, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/BufferCapabilities/<init>(Ljava/awt/ImageCapabilities;Ljava/awt/ImageCapabilities;Ljava/awt/BufferCapabilities$FlipContents;)V"(thread: JVMThread, args: [java_awt_ImageCapabilities, java_awt_ImageCapabilities, java_awt_BufferCapabilities$FlipContents], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFrontBufferCapabilities()Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "java/awt/BufferCapabilities/getFrontBufferCapabilities()Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "getBackBufferCapabilities()Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "java/awt/BufferCapabilities/getBackBufferCapabilities()Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "isPageFlipping()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BufferCapabilities/isPageFlipping()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFlipContents()Ljava/awt/BufferCapabilities$FlipContents;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_BufferCapabilities$FlipContents) => void): void;
    public "java/awt/BufferCapabilities/getFlipContents()Ljava/awt/BufferCapabilities$FlipContents;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_BufferCapabilities$FlipContents) => void): void;
    public "isFullScreenRequired()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BufferCapabilities/isFullScreenRequired()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMultiBufferAvailable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BufferCapabilities/isMultiBufferAvailable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/BufferCapabilities/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_awt_image_renderable_RenderableImage extends java_lang_Object {
    "getSources()Ljava/util/Vector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    "getProperty(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getPropertyNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "isDynamic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getWidth()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getHeight()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinX()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinY()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "createScaledRendering(IILjava/awt/RenderingHints;)Ljava/awt/image/RenderedImage;"(thread: JVMThread, args: [number, number, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_RenderedImage) => void): void;
    "createDefaultRendering()Ljava/awt/image/RenderedImage;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_RenderedImage) => void): void;
    "createRendering(Ljava/awt/image/renderable/RenderContext;)Ljava/awt/image/RenderedImage;"(thread: JVMThread, args: [java_awt_image_renderable_RenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_RenderedImage) => void): void;
  }
  export class java_awt_image_renderable_RenderContext extends java_lang_Object implements java_lang_Cloneable {
    public "java/awt/image/renderable/RenderContext/hints": java_awt_RenderingHints;
    public "java/awt/image/renderable/RenderContext/usr2dev": java_awt_geom_AffineTransform;
    public "java/awt/image/renderable/RenderContext/aoi": java_awt_Shape;
    public "<init>(Ljava/awt/geom/AffineTransform;Ljava/awt/Shape;Ljava/awt/RenderingHints;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_Shape, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/<init>(Ljava/awt/geom/AffineTransform;Ljava/awt/Shape;Ljava/awt/RenderingHints;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_Shape, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/<init>(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/geom/AffineTransform;Ljava/awt/RenderingHints;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/<init>(Ljava/awt/geom/AffineTransform;Ljava/awt/RenderingHints;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/geom/AffineTransform;Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/<init>(Ljava/awt/geom/AffineTransform;Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRenderingHints()Ljava/awt/RenderingHints;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_RenderingHints) => void): void;
    public "java/awt/image/renderable/RenderContext/getRenderingHints()Ljava/awt/RenderingHints;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_RenderingHints) => void): void;
    public "setRenderingHints(Ljava/awt/RenderingHints;)V"(thread: JVMThread, args: [java_awt_RenderingHints], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/setRenderingHints(Ljava/awt/RenderingHints;)V"(thread: JVMThread, args: [java_awt_RenderingHints], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "preConcatenateTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/preConcatenateTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "preConcetenateTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/preConcetenateTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "concatenateTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/concatenateTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "concetenateTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/concetenateTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "java/awt/image/renderable/RenderContext/getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "setAreaOfInterest(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/renderable/RenderContext/setAreaOfInterest(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAreaOfInterest()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/image/renderable/RenderContext/getAreaOfInterest()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/renderable/RenderContext/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_awt_image_RenderedImage extends java_lang_Object {
    "getSources()Ljava/util/Vector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    "getProperty(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getPropertyNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getSampleModel()Ljava/awt/image/SampleModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getNumXTiles()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getNumYTiles()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinTileX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinTileY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTileWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTileHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTileGridXOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTileGridYOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTile(II)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    "getData()Ljava/awt/image/Raster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    "getData(Ljava/awt/Rectangle;)Ljava/awt/image/Raster;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    "copyData(Ljava/awt/image/WritableRaster;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
  }
  export interface java_awt_image_BufferedImageOp extends java_lang_Object {
    "filter(Ljava/awt/image/BufferedImage;Ljava/awt/image/BufferedImage;)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [java_awt_image_BufferedImage, java_awt_image_BufferedImage], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    "getBounds2D(Ljava/awt/image/BufferedImage;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_image_BufferedImage], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    "createCompatibleDestImage(Ljava/awt/image/BufferedImage;Ljava/awt/image/ColorModel;)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [java_awt_image_BufferedImage, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    "getPoint2D(Ljava/awt/geom/Point2D;Ljava/awt/geom/Point2D;)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [java_awt_geom_Point2D, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    "getRenderingHints()Ljava/awt/RenderingHints;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_RenderingHints) => void): void;
  }
  export class java_awt_Graphics extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "create()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "java/awt/Graphics/create()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "create(IIII)Ljava/awt/Graphics;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "java/awt/Graphics/create(IIII)Ljava/awt/Graphics;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getColor()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "java/awt/Graphics/getColor()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "setColor(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/setColor(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPaintMode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/setPaintMode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setXORMode(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/setXORMode(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "java/awt/Graphics/getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontMetrics()Ljava/awt/FontMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "java/awt/Graphics/getFontMetrics()Ljava/awt/FontMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "java/awt/Graphics/getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "getClipBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Graphics/getClipBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "clipRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/clipRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setClip(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/setClip(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getClip()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/Graphics/getClip()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "setClip(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/setClip(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyArea(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/copyArea(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawLine(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawLine(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/fillRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/clearRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRoundRect(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawRoundRect(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRoundRect(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/fillRoundRect(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "draw3DRect(IIIIZ)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/draw3DRect(IIIIZ)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fill3DRect(IIIIZ)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/fill3DRect(IIIIZ)V"(thread: JVMThread, args: [number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawOval(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawOval(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillOval(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/fillOval(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawArc(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawArc(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillArc(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/fillArc(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolyline([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawPolyline([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolygon([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawPolygon([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolygon(Ljava/awt/Polygon;)V"(thread: JVMThread, args: [java_awt_Polygon], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawPolygon(Ljava/awt/Polygon;)V"(thread: JVMThread, args: [java_awt_Polygon], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillPolygon([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/fillPolygon([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillPolygon(Ljava/awt/Polygon;)V"(thread: JVMThread, args: [java_awt_Polygon], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/fillPolygon(Ljava/awt/Polygon;)V"(thread: JVMThread, args: [java_awt_Polygon], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawString(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Ljava/text/AttributedCharacterIterator;II)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawString(Ljava/text/AttributedCharacterIterator;II)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawChars([CIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawChars([CIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawBytes([BIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/drawBytes([BIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Graphics/drawImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IIIILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Graphics/drawImage(Ljava/awt/Image;IIIILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Graphics/drawImage(Ljava/awt/Image;IILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Graphics/drawImage(Ljava/awt/Image;IIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IIIIIIIILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Graphics/drawImage(Ljava/awt/Image;IIIIIIIILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IIIIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Graphics/drawImage(Ljava/awt/Image;IIIIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Graphics/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/Graphics/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getClipRect()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Graphics/getClipRect()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "hitClip(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Graphics/hitClip(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getClipBounds(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Graphics/getClipBounds(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
  }
  export class java_awt_Polygon extends java_lang_Object implements java_awt_Shape, java_io_Serializable {
    public "java/awt/Polygon/npoints": number;
    public "java/awt/Polygon/xpoints": JVMArray<number>;
    public "java/awt/Polygon/ypoints": JVMArray<number>;
    public "java/awt/Polygon/bounds": java_awt_Rectangle;
    public static "java/awt/Polygon/serialVersionUID": Long;
    public static "java/awt/Polygon/MIN_LENGTH": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Polygon/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Polygon/<init>([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Polygon/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Polygon/invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Polygon/translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "calculateBounds([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Polygon/calculateBounds([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateBounds(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Polygon/updateBounds(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addPoint(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Polygon/addPoint(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Polygon/getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getBoundingBox()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/Polygon/getBoundingBox()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "contains(Ljava/awt/Point;)Z"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Polygon/contains(Ljava/awt/Point;)Z"(thread: JVMThread, args: [java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Polygon/contains(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inside(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Polygon/inside(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/Polygon/getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Polygon/contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCrossings(DDDD)Lsun/awt/geom/Crossings;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: sun_awt_geom_Crossings) => void): void;
    public "java/awt/Polygon/getCrossings(DDDD)Lsun/awt/geom/Crossings;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: sun_awt_geom_Crossings) => void): void;
    public "contains(Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Polygon/contains(Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Polygon/intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Polygon/intersects(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Polygon/contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Polygon/contains(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "java/awt/Polygon/getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "java/awt/Polygon/getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
  }
  export class sun_awt_geom_Crossings extends java_lang_Object {
    public static "sun/awt/geom/Crossings/debug": number;
    public "sun/awt/geom/Crossings/limit": number;
    public "sun/awt/geom/Crossings/yranges": JVMArray<number>;
    public "sun/awt/geom/Crossings/xlo": number;
    public "sun/awt/geom/Crossings/ylo": number;
    public "sun/awt/geom/Crossings/xhi": number;
    public "sun/awt/geom/Crossings/yhi": number;
    public "sun/awt/geom/Crossings/tmp": java_util_Vector;
    public "<init>(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/geom/Crossings/<init>(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getXLo()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/getXLo()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getYLo()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/getYLo()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getXHi()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/getXHi()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getYHi()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/getYHi()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "record(DDI)V"(thread: JVMThread, args: [number, any, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/geom/Crossings/record(DDI)V"(thread: JVMThread, args: [number, any, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/geom/Crossings/print()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "covers(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/covers(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "findCrossings(Ljava/util/Vector;DDDD)Lsun/awt/geom/Crossings;"(thread: JVMThread, args: [java_util_Vector, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: sun_awt_geom_Crossings) => void): void;
    public static "sun/awt/geom/Crossings/findCrossings(Ljava/util/Vector;DDDD)Lsun/awt/geom/Crossings;"(thread: JVMThread, args: [java_util_Vector, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: sun_awt_geom_Crossings) => void): void;
    public static "findCrossings(Ljava/awt/geom/PathIterator;DDDD)Lsun/awt/geom/Crossings;"(thread: JVMThread, args: [java_awt_geom_PathIterator, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: sun_awt_geom_Crossings) => void): void;
    public static "sun/awt/geom/Crossings/findCrossings(Ljava/awt/geom/PathIterator;DDDD)Lsun/awt/geom/Crossings;"(thread: JVMThread, args: [java_awt_geom_PathIterator, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: sun_awt_geom_Crossings) => void): void;
    public "accumulateLine(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/accumulateLine(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "accumulateLine(DDDDI)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/accumulateLine(DDDDI)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "accumulateQuad(DD[D)Z"(thread: JVMThread, args: [number, any, number, any, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/accumulateQuad(DD[D)Z"(thread: JVMThread, args: [number, any, number, any, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "accumulateCubic(DD[D)Z"(thread: JVMThread, args: [number, any, number, any, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/geom/Crossings/accumulateCubic(DD[D)Z"(thread: JVMThread, args: [number, any, number, any, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_image_ImageProducer extends java_lang_Object {
    "addConsumer(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "isConsumer(Ljava/awt/image/ImageConsumer;)Z"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "removeConsumer(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "startProduction(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "requestTopDownLeftRightResend(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_image_ImageConsumer extends java_lang_Object {
    "setDimensions(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setProperties(Ljava/util/Hashtable;)V"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    "setColorModel(Ljava/awt/image/ColorModel;)V"(thread: JVMThread, args: [java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable) => void): void;
    "setHints(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "setPixels(IIIILjava/awt/image/ColorModel;[BII)V"(thread: JVMThread, args: [number, number, number, number, java_awt_image_ColorModel, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setPixels(IIIILjava/awt/image/ColorModel;[III)V"(thread: JVMThread, args: [number, number, number, number, java_awt_image_ColorModel, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "imageComplete(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_image_ImageObserver extends java_lang_Object {
    "imageUpdate(Ljava/awt/Image;IIIII)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_image_IndexColorModel extends java_awt_image_ColorModel {
    public "java/awt/image/IndexColorModel/rgb": JVMArray<number>;
    public "java/awt/image/IndexColorModel/map_size": number;
    public "java/awt/image/IndexColorModel/pixel_mask": number;
    public "java/awt/image/IndexColorModel/transparent_index": number;
    public "java/awt/image/IndexColorModel/allgrayopaque": number;
    public "java/awt/image/IndexColorModel/validBits": java_math_BigInteger;
    public "java/awt/image/IndexColorModel/colorData": sun_awt_image_BufImgSurfaceData$ICMColorData;
    public static "java/awt/image/IndexColorModel/opaqueBits": JVMArray<number>;
    public static "java/awt/image/IndexColorModel/alphaBits": JVMArray<number>;
    public static "java/awt/image/IndexColorModel/CACHESIZE": number;
    public "java/awt/image/IndexColorModel/lookupcache": JVMArray<number>;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/IndexColorModel/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II[B[B[B)V"(thread: JVMThread, args: [number, number, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/<init>(II[B[B[B)V"(thread: JVMThread, args: [number, number, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II[B[B[BI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/<init>(II[B[B[BI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II[B[B[B[B)V"(thread: JVMThread, args: [number, number, JVMArray<number>, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/<init>(II[B[B[B[B)V"(thread: JVMThread, args: [number, number, JVMArray<number>, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II[BIZ)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/<init>(II[BIZ)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II[BIZI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/<init>(II[BIZI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II[IIZII)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/<init>(II[IIZII)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II[IIILjava/math/BigInteger;)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number, java_math_BigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/<init>(II[IIILjava/math/BigInteger;)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number, number, java_math_BigInteger], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRGBs(I[B[B[B[B)V"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/setRGBs(I[B[B[B[B)V"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRGBs(I[IIZ)V"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/setRGBs(I[IIZ)V"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "calcRealMapSize(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/calcRealMapSize(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAllValid()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/awt/image/IndexColorModel/getAllValid()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getComponentSize()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/IndexColorModel/getComponentSize()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getMapSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/getMapSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransparentPixel()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/getTransparentPixel()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getReds([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/getReds([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGreens([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/getGreens([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBlues([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/getBlues([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAlphas([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/getAlphas([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRGBs([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/getRGBs([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTransparentPixel(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/setTransparentPixel(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTransparency(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/setTransparency(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "calculatePixelMask()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/calculatePixelMask()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRed(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/getRed(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGreen(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/getGreen(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBlue(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/getBlue(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAlpha(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/getAlpha(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRGB(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/getRGB(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDataElements(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/IndexColorModel/getDataElements(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "installpixel(Ljava/lang/Object;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/IndexColorModel/installpixel(Ljava/lang/Object;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getComponents(I[II)[I"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/IndexColorModel/getComponents(I[II)[I"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getComponents(Ljava/lang/Object;[II)[I"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/IndexColorModel/getComponents(Ljava/lang/Object;[II)[I"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getDataElement([II)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/getDataElement([II)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDataElements([IILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<number>, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/IndexColorModel/getDataElements([IILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<number>, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "createCompatibleWritableRaster(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/IndexColorModel/createCompatibleWritableRaster(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "isCompatibleRaster(Ljava/awt/image/Raster;)Z"(thread: JVMThread, args: [java_awt_image_Raster], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/isCompatibleRaster(Ljava/awt/image/Raster;)Z"(thread: JVMThread, args: [java_awt_image_Raster], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createCompatibleSampleModel(II)Ljava/awt/image/SampleModel;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "java/awt/image/IndexColorModel/createCompatibleSampleModel(II)Ljava/awt/image/SampleModel;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "isCompatibleSampleModel(Ljava/awt/image/SampleModel;)Z"(thread: JVMThread, args: [java_awt_image_SampleModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/isCompatibleSampleModel(Ljava/awt/image/SampleModel;)Z"(thread: JVMThread, args: [java_awt_image_SampleModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "convertToIntDiscrete(Ljava/awt/image/Raster;Z)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [java_awt_image_Raster, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "java/awt/image/IndexColorModel/convertToIntDiscrete(Ljava/awt/image/Raster;Z)Ljava/awt/image/BufferedImage;"(thread: JVMThread, args: [java_awt_image_Raster, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_BufferedImage) => void): void;
    public "isValid(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/isValid(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/IndexColorModel/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getValidPixels()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "java/awt/image/IndexColorModel/getValidPixels()Ljava/math/BigInteger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_math_BigInteger) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/IndexColorModel/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/image/IndexColorModel/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/IndexColorModel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_image_BufImgSurfaceData$ICMColorData extends java_lang_Object {
    public "sun/awt/image/BufImgSurfaceData$ICMColorData/pData": Long;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/BufImgSurfaceData$ICMColorData/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/BufImgSurfaceData$ICMColorData/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_image_OffScreenImageSource extends java_lang_Object implements java_awt_image_ImageProducer {
    public "sun/awt/image/OffScreenImageSource/image": java_awt_image_BufferedImage;
    public "sun/awt/image/OffScreenImageSource/width": number;
    public "sun/awt/image/OffScreenImageSource/height": number;
    public "sun/awt/image/OffScreenImageSource/properties": java_util_Hashtable;
    public "sun/awt/image/OffScreenImageSource/theConsumer": java_awt_image_ImageConsumer;
    public "<init>(Ljava/awt/image/BufferedImage;Ljava/util/Hashtable;)V"(thread: JVMThread, args: [java_awt_image_BufferedImage, java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/OffScreenImageSource/<init>(Ljava/awt/image/BufferedImage;Ljava/util/Hashtable;)V"(thread: JVMThread, args: [java_awt_image_BufferedImage, java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/image/BufferedImage;)V"(thread: JVMThread, args: [java_awt_image_BufferedImage], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/OffScreenImageSource/<init>(Ljava/awt/image/BufferedImage;)V"(thread: JVMThread, args: [java_awt_image_BufferedImage], cb?: (e?: java_lang_Throwable) => void): void;
    public "addConsumer(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/OffScreenImageSource/addConsumer(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "isConsumer(Ljava/awt/image/ImageConsumer;)Z"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/image/OffScreenImageSource/isConsumer(Ljava/awt/image/ImageConsumer;)Z"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeConsumer(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/OffScreenImageSource/removeConsumer(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "startProduction(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/OffScreenImageSource/startProduction(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "requestTopDownLeftRightResend(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/OffScreenImageSource/requestTopDownLeftRightResend(Ljava/awt/image/ImageConsumer;)V"(thread: JVMThread, args: [java_awt_image_ImageConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sendPixels()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/OffScreenImageSource/sendPixels()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "produce()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/OffScreenImageSource/produce()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_image_WritableRenderedImage extends java_lang_Object, java_awt_image_RenderedImage {
    "addTileObserver(Ljava/awt/image/TileObserver;)V"(thread: JVMThread, args: [java_awt_image_TileObserver], cb?: (e?: java_lang_Throwable) => void): void;
    "removeTileObserver(Ljava/awt/image/TileObserver;)V"(thread: JVMThread, args: [java_awt_image_TileObserver], cb?: (e?: java_lang_Throwable) => void): void;
    "getWritableTile(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    "releaseWritableTile(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "isTileWritable(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getWritableTileIndices()[Ljava/awt/Point;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Point>) => void): void;
    "hasTileWriters()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "setData(Ljava/awt/image/Raster;)V"(thread: JVMThread, args: [java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    "getSources()Ljava/util/Vector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    "getProperty(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getPropertyNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getSampleModel()Ljava/awt/image/SampleModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getNumXTiles()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getNumYTiles()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinTileX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getMinTileY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTileWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTileHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTileGridXOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTileGridYOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTile(II)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    "getData()Ljava/awt/image/Raster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    "getData(Ljava/awt/Rectangle;)Ljava/awt/image/Raster;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    "copyData(Ljava/awt/image/WritableRaster;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
  }
  export class java_awt_Image extends java_lang_Object {
    public static "java/awt/Image/defaultImageCaps": java_awt_ImageCapabilities;
    public "java/awt/Image/accelerationPriority": number;
    public static "java/awt/Image/UndefinedProperty": java_lang_Object;
    public static "java/awt/Image/SCALE_DEFAULT": number;
    public static "java/awt/Image/SCALE_FAST": number;
    public static "java/awt/Image/SCALE_SMOOTH": number;
    public static "java/awt/Image/SCALE_REPLICATE": number;
    public static "java/awt/Image/SCALE_AREA_AVERAGING": number;
    public "java/awt/Image/surfaceManager": sun_awt_image_SurfaceManager;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Image/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWidth(Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Image/getWidth(Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight(Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Image/getHeight(Ljava/awt/image/ImageObserver;)I"(thread: JVMThread, args: [java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSource()Ljava/awt/image/ImageProducer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ImageProducer) => void): void;
    public "java/awt/Image/getSource()Ljava/awt/image/ImageProducer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ImageProducer) => void): void;
    public "getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "java/awt/Image/getGraphics()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "getProperty(Ljava/lang/String;Ljava/awt/image/ImageObserver;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/Image/getProperty(Ljava/lang/String;Ljava/awt/image/ImageObserver;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getScaledInstance(III)Ljava/awt/Image;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "java/awt/Image/getScaledInstance(III)Ljava/awt/Image;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Image/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCapabilities(Ljava/awt/GraphicsConfiguration;)Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "java/awt/Image/getCapabilities(Ljava/awt/GraphicsConfiguration;)Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "setAccelerationPriority(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/Image/setAccelerationPriority(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAccelerationPriority()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/Image/getAccelerationPriority()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/Image/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_image_SurfaceManager extends java_lang_Object {
    public static "sun/awt/image/SurfaceManager/imgaccessor": sun_awt_image_SurfaceManager$ImageAccessor;
    public "sun/awt/image/SurfaceManager/cacheMap": java_util_concurrent_ConcurrentHashMap;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/SurfaceManager/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setImageAccessor(Lsun/awt/image/SurfaceManager$ImageAccessor;)V"(thread: JVMThread, args: [sun_awt_image_SurfaceManager$ImageAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/image/SurfaceManager/setImageAccessor(Lsun/awt/image/SurfaceManager$ImageAccessor;)V"(thread: JVMThread, args: [sun_awt_image_SurfaceManager$ImageAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getManager(Ljava/awt/Image;)Lsun/awt/image/SurfaceManager;"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: sun_awt_image_SurfaceManager) => void): void;
    public static "sun/awt/image/SurfaceManager/getManager(Ljava/awt/Image;)Lsun/awt/image/SurfaceManager;"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: sun_awt_image_SurfaceManager) => void): void;
    public static "setManager(Ljava/awt/Image;Lsun/awt/image/SurfaceManager;)V"(thread: JVMThread, args: [java_awt_Image, sun_awt_image_SurfaceManager], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/image/SurfaceManager/setManager(Ljava/awt/Image;Lsun/awt/image/SurfaceManager;)V"(thread: JVMThread, args: [java_awt_Image, sun_awt_image_SurfaceManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCacheData(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/awt/image/SurfaceManager/getCacheData(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setCacheData(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/SurfaceManager/setCacheData(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPrimarySurfaceData()Lsun/java2d/SurfaceData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "sun/awt/image/SurfaceManager/getPrimarySurfaceData()Lsun/java2d/SurfaceData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "restoreContents()Lsun/java2d/SurfaceData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "sun/awt/image/SurfaceManager/restoreContents()Lsun/java2d/SurfaceData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "acceleratedSurfaceLost()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/SurfaceManager/acceleratedSurfaceLost()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCapabilities(Ljava/awt/GraphicsConfiguration;)Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "sun/awt/image/SurfaceManager/getCapabilities(Ljava/awt/GraphicsConfiguration;)Ljava/awt/ImageCapabilities;"(thread: JVMThread, args: [java_awt_GraphicsConfiguration], cb?: (e?: java_lang_Throwable, rv?: java_awt_ImageCapabilities) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/SurfaceManager/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/SurfaceManager/flush(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setAccelerationPriority(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/SurfaceManager/setAccelerationPriority(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getImageScale(Ljava/awt/Image;)I"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/image/SurfaceManager/getImageScale(Ljava/awt/Image;)I"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_java2d_SurfaceData extends java_lang_Object implements java_awt_Transparency, sun_java2d_DisposerTarget, sun_java2d_StateTrackable, sun_java2d_Surface {
    public "sun/java2d/SurfaceData/pData": Long;
    public "sun/java2d/SurfaceData/valid": number;
    public "sun/java2d/SurfaceData/surfaceLost": number;
    public "sun/java2d/SurfaceData/surfaceType": sun_java2d_loops_SurfaceType;
    public "sun/java2d/SurfaceData/colorModel": java_awt_image_ColorModel;
    public "sun/java2d/SurfaceData/disposerReferent": java_lang_Object;
    public "sun/java2d/SurfaceData/blitProxyKey": java_lang_Object;
    public "sun/java2d/SurfaceData/stateDelegate": sun_java2d_StateTrackableDelegate;
    public static "sun/java2d/SurfaceData/colorPrimitives": sun_java2d_pipe_LoopPipe;
    public static "sun/java2d/SurfaceData/outlineTextRenderer": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/solidTextRenderer": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/aaTextRenderer": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/lcdTextRenderer": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/colorPipe": sun_java2d_pipe_AlphaColorPipe;
    public static "sun/java2d/SurfaceData/colorViaShape": sun_java2d_pipe_PixelToShapeConverter;
    public static "sun/java2d/SurfaceData/colorViaPgram": sun_java2d_pipe_PixelToParallelogramConverter;
    public static "sun/java2d/SurfaceData/colorText": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/clipColorPipe": sun_java2d_pipe_CompositePipe;
    public static "sun/java2d/SurfaceData/clipColorText": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/AAColorShape": sun_java2d_pipe_AAShapePipe;
    public static "sun/java2d/SurfaceData/AAColorViaShape": sun_java2d_pipe_PixelToParallelogramConverter;
    public static "sun/java2d/SurfaceData/AAColorViaPgram": sun_java2d_pipe_PixelToParallelogramConverter;
    public static "sun/java2d/SurfaceData/AAClipColorShape": sun_java2d_pipe_AAShapePipe;
    public static "sun/java2d/SurfaceData/AAClipColorViaShape": sun_java2d_pipe_PixelToParallelogramConverter;
    public static "sun/java2d/SurfaceData/paintPipe": sun_java2d_pipe_CompositePipe;
    public static "sun/java2d/SurfaceData/paintShape": sun_java2d_pipe_SpanShapeRenderer;
    public static "sun/java2d/SurfaceData/paintViaShape": sun_java2d_pipe_PixelToShapeConverter;
    public static "sun/java2d/SurfaceData/paintText": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/clipPaintPipe": sun_java2d_pipe_CompositePipe;
    public static "sun/java2d/SurfaceData/clipPaintText": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/AAPaintShape": sun_java2d_pipe_AAShapePipe;
    public static "sun/java2d/SurfaceData/AAPaintViaShape": sun_java2d_pipe_PixelToParallelogramConverter;
    public static "sun/java2d/SurfaceData/AAClipPaintShape": sun_java2d_pipe_AAShapePipe;
    public static "sun/java2d/SurfaceData/AAClipPaintViaShape": sun_java2d_pipe_PixelToParallelogramConverter;
    public static "sun/java2d/SurfaceData/compPipe": sun_java2d_pipe_CompositePipe;
    public static "sun/java2d/SurfaceData/compShape": sun_java2d_pipe_SpanShapeRenderer;
    public static "sun/java2d/SurfaceData/compViaShape": sun_java2d_pipe_PixelToShapeConverter;
    public static "sun/java2d/SurfaceData/compText": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/clipCompPipe": sun_java2d_pipe_CompositePipe;
    public static "sun/java2d/SurfaceData/clipCompText": sun_java2d_pipe_TextPipe;
    public static "sun/java2d/SurfaceData/AACompShape": sun_java2d_pipe_AAShapePipe;
    public static "sun/java2d/SurfaceData/AACompViaShape": sun_java2d_pipe_PixelToParallelogramConverter;
    public static "sun/java2d/SurfaceData/AAClipCompShape": sun_java2d_pipe_AAShapePipe;
    public static "sun/java2d/SurfaceData/AAClipCompViaShape": sun_java2d_pipe_PixelToParallelogramConverter;
    public static "sun/java2d/SurfaceData/imagepipe": sun_java2d_pipe_DrawImagePipe;
    public static "sun/java2d/SurfaceData/LOOP_UNKNOWN": number;
    public static "sun/java2d/SurfaceData/LOOP_FOUND": number;
    public static "sun/java2d/SurfaceData/LOOP_NOTFOUND": number;
    public "sun/java2d/SurfaceData/haveLCDLoop": number;
    public "sun/java2d/SurfaceData/havePgramXORLoop": number;
    public "sun/java2d/SurfaceData/havePgramSolidLoop": number;
    public static "sun/java2d/SurfaceData/loopcache": sun_java2d_loops_RenderCache;
    public static "sun/java2d/SurfaceData/compPermission": java_security_Permission;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/SurfaceData/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Ljava/awt/image/ColorModel;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/<init>(Lsun/java2d/loops/SurfaceType;Ljava/awt/image/ColorModel;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/StateTrackable$State;Lsun/java2d/loops/SurfaceType;Ljava/awt/image/ColorModel;)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, sun_java2d_loops_SurfaceType, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/<init>(Lsun/java2d/StateTrackable$State;Lsun/java2d/loops/SurfaceType;Ljava/awt/image/ColorModel;)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, sun_java2d_loops_SurfaceType, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/StateTrackableDelegate;Lsun/java2d/loops/SurfaceType;Ljava/awt/image/ColorModel;)V"(thread: JVMThread, args: [sun_java2d_StateTrackableDelegate, sun_java2d_loops_SurfaceType, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/<init>(Lsun/java2d/StateTrackableDelegate;Lsun/java2d/loops/SurfaceType;Ljava/awt/image/ColorModel;)V"(thread: JVMThread, args: [sun_java2d_StateTrackableDelegate, sun_java2d_loops_SurfaceType, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/StateTrackable$State;)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/<init>(Lsun/java2d/StateTrackable$State;)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBlitProxyKey(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/setBlitProxyKey(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSourceSurfaceData(Ljava/awt/Image;ILsun/java2d/loops/CompositeType;Ljava/awt/Color;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [java_awt_Image, number, sun_java2d_loops_CompositeType, java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "sun/java2d/SurfaceData/getSourceSurfaceData(Ljava/awt/Image;ILsun/java2d/loops/CompositeType;Ljava/awt/Color;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [java_awt_Image, number, sun_java2d_loops_CompositeType, java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "makeProxyFor(Lsun/java2d/SurfaceData;)Lsun/java2d/SurfaceDataProxy;"(thread: JVMThread, args: [sun_java2d_SurfaceData], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceDataProxy) => void): void;
    public "sun/java2d/SurfaceData/makeProxyFor(Lsun/java2d/SurfaceData;)Lsun/java2d/SurfaceDataProxy;"(thread: JVMThread, args: [sun_java2d_SurfaceData], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceDataProxy) => void): void;
    public static "getPrimarySurfaceData(Ljava/awt/Image;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public static "sun/java2d/SurfaceData/getPrimarySurfaceData(Ljava/awt/Image;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public static "restoreContents(Ljava/awt/Image;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public static "sun/java2d/SurfaceData/restoreContents(Ljava/awt/Image;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "getState()Lsun/java2d/StateTrackable$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTrackable$State) => void): void;
    public "sun/java2d/SurfaceData/getState()Lsun/java2d/StateTrackable$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTrackable$State) => void): void;
    public "getStateTracker()Lsun/java2d/StateTracker;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTracker) => void): void;
    public "sun/java2d/SurfaceData/getStateTracker()Lsun/java2d/StateTracker;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTracker) => void): void;
    public "markDirty()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/markDirty()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSurfaceLost(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/setSurfaceLost(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSurfaceLost()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/isSurfaceLost()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDisposerReferent()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/SurfaceData/getDisposerReferent()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getNativeOps()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/java2d/SurfaceData/getNativeOps()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getReplacement()Lsun/java2d/SurfaceData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "sun/java2d/SurfaceData/getReplacement()Lsun/java2d/SurfaceData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public static "makeConverter(Lsun/java2d/pipe/AAShapePipe;Lsun/java2d/pipe/ParallelogramPipe;)Lsun/java2d/pipe/PixelToParallelogramConverter;"(thread: JVMThread, args: [sun_java2d_pipe_AAShapePipe, sun_java2d_pipe_ParallelogramPipe], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_PixelToParallelogramConverter) => void): void;
    public static "sun/java2d/SurfaceData/makeConverter(Lsun/java2d/pipe/AAShapePipe;Lsun/java2d/pipe/ParallelogramPipe;)Lsun/java2d/pipe/PixelToParallelogramConverter;"(thread: JVMThread, args: [sun_java2d_pipe_AAShapePipe, sun_java2d_pipe_ParallelogramPipe], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_PixelToParallelogramConverter) => void): void;
    public static "makeConverter(Lsun/java2d/pipe/AAShapePipe;)Lsun/java2d/pipe/PixelToParallelogramConverter;"(thread: JVMThread, args: [sun_java2d_pipe_AAShapePipe], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_PixelToParallelogramConverter) => void): void;
    public static "sun/java2d/SurfaceData/makeConverter(Lsun/java2d/pipe/AAShapePipe;)Lsun/java2d/pipe/PixelToParallelogramConverter;"(thread: JVMThread, args: [sun_java2d_pipe_AAShapePipe], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_PixelToParallelogramConverter) => void): void;
    public "canRenderLCDText(Lsun/java2d/SunGraphics2D;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/canRenderLCDText(Lsun/java2d/SunGraphics2D;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canRenderParallelograms(Lsun/java2d/SunGraphics2D;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/canRenderParallelograms(Lsun/java2d/SunGraphics2D;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "validatePipe(Lsun/java2d/SunGraphics2D;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/validatePipe(Lsun/java2d/SunGraphics2D;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTextPipe(Lsun/java2d/SunGraphics2D;Z)Lsun/java2d/pipe/TextPipe;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_TextPipe) => void): void;
    public "sun/java2d/SurfaceData/getTextPipe(Lsun/java2d/SunGraphics2D;Z)Lsun/java2d/pipe/TextPipe;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_TextPipe) => void): void;
    public static "getPaintSurfaceType(Lsun/java2d/SunGraphics2D;)Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public static "sun/java2d/SurfaceData/getPaintSurfaceType(Lsun/java2d/SunGraphics2D;)Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public static "getFillCompositeType(Lsun/java2d/SunGraphics2D;)Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public static "sun/java2d/SurfaceData/getFillCompositeType(Lsun/java2d/SunGraphics2D;)Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public "getMaskFill(Lsun/java2d/SunGraphics2D;)Lsun/java2d/loops/MaskFill;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_MaskFill) => void): void;
    public "sun/java2d/SurfaceData/getMaskFill(Lsun/java2d/SunGraphics2D;)Lsun/java2d/loops/MaskFill;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_MaskFill) => void): void;
    public "getRenderLoops(Lsun/java2d/SunGraphics2D;)Lsun/java2d/loops/RenderLoops;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_RenderLoops) => void): void;
    public "sun/java2d/SurfaceData/getRenderLoops(Lsun/java2d/SunGraphics2D;)Lsun/java2d/loops/RenderLoops;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_RenderLoops) => void): void;
    public static "makeRenderLoops(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/RenderLoops;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_RenderLoops) => void): void;
    public static "sun/java2d/SurfaceData/makeRenderLoops(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/RenderLoops;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_RenderLoops) => void): void;
    public "getDeviceConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "sun/java2d/SurfaceData/getDeviceConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "getSurfaceType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "sun/java2d/SurfaceData/getSurfaceType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "sun/java2d/SurfaceData/getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRaster(IIII)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "sun/java2d/SurfaceData/getRaster(IIII)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "useTightBBoxes()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/useTightBBoxes()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "pixelFor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/pixelFor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "pixelFor(Ljava/awt/Color;)I"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/pixelFor(Ljava/awt/Color;)I"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rgbFor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/rgbFor(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "sun/java2d/SurfaceData/getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "checkCustomComposite()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/checkCustomComposite()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isOpaqueGray(Ljava/awt/image/IndexColorModel;)Z"(thread: JVMThread, args: [java_awt_image_IndexColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/SurfaceData/isOpaqueGray(Ljava/awt/image/IndexColorModel;)Z"(thread: JVMThread, args: [java_awt_image_IndexColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isNull(Lsun/java2d/SurfaceData;)Z"(thread: JVMThread, args: [sun_java2d_SurfaceData], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/SurfaceData/isNull(Lsun/java2d/SurfaceData;)Z"(thread: JVMThread, args: [sun_java2d_SurfaceData], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "copyArea(Lsun/java2d/SunGraphics2D;IIIIII)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/copyArea(Lsun/java2d/SunGraphics2D;IIIIII)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceData/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDestination()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/SurfaceData/getDestination()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getDefaultScale()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceData/getDefaultScale()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/SurfaceData/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_RenderLoops extends java_lang_Object {
    public static "sun/java2d/loops/RenderLoops/primTypeID": number;
    public "sun/java2d/loops/RenderLoops/drawLineLoop": sun_java2d_loops_DrawLine;
    public "sun/java2d/loops/RenderLoops/fillRectLoop": sun_java2d_loops_FillRect;
    public "sun/java2d/loops/RenderLoops/drawRectLoop": sun_java2d_loops_DrawRect;
    public "sun/java2d/loops/RenderLoops/drawPolygonsLoop": sun_java2d_loops_DrawPolygons;
    public "sun/java2d/loops/RenderLoops/drawPathLoop": sun_java2d_loops_DrawPath;
    public "sun/java2d/loops/RenderLoops/fillPathLoop": sun_java2d_loops_FillPath;
    public "sun/java2d/loops/RenderLoops/fillSpansLoop": sun_java2d_loops_FillSpans;
    public "sun/java2d/loops/RenderLoops/fillParallelogramLoop": sun_java2d_loops_FillParallelogram;
    public "sun/java2d/loops/RenderLoops/drawParallelogramLoop": sun_java2d_loops_DrawParallelogram;
    public "sun/java2d/loops/RenderLoops/drawGlyphListLoop": sun_java2d_loops_DrawGlyphList;
    public "sun/java2d/loops/RenderLoops/drawGlyphListAALoop": sun_java2d_loops_DrawGlyphListAA;
    public "sun/java2d/loops/RenderLoops/drawGlyphListLCDLoop": sun_java2d_loops_DrawGlyphListLCD;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/RenderLoops/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/RenderLoops/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_DrawGlyphListLCD extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/DrawGlyphListLCD/methodSignature": java_lang_String;
    public static "sun/java2d/loops/DrawGlyphListLCD/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawGlyphListLCD;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawGlyphListLCD) => void): void;
    public static "sun/java2d/loops/DrawGlyphListLCD/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawGlyphListLCD;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawGlyphListLCD) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawGlyphListLCD/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawGlyphListLCD/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "DrawGlyphListLCD(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawGlyphListLCD/DrawGlyphListLCD(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawGlyphListLCD/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawGlyphListLCD/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/DrawGlyphListLCD/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_GraphicsPrimitive extends java_lang_Object {
    public "sun/java2d/loops/GraphicsPrimitive/methodSignature": java_lang_String;
    public "sun/java2d/loops/GraphicsPrimitive/uniqueID": number;
    public static "sun/java2d/loops/GraphicsPrimitive/unusedPrimID": number;
    public "sun/java2d/loops/GraphicsPrimitive/sourceType": sun_java2d_loops_SurfaceType;
    public "sun/java2d/loops/GraphicsPrimitive/compositeType": sun_java2d_loops_CompositeType;
    public "sun/java2d/loops/GraphicsPrimitive/destType": sun_java2d_loops_SurfaceType;
    public "sun/java2d/loops/GraphicsPrimitive/pNativePrim": Long;
    public static "sun/java2d/loops/GraphicsPrimitive/traceMap": java_util_HashMap;
    public static "sun/java2d/loops/GraphicsPrimitive/traceflags": number;
    public static "sun/java2d/loops/GraphicsPrimitive/tracefile": java_lang_String;
    public static "sun/java2d/loops/GraphicsPrimitive/traceout": java_io_PrintStream;
    public static "sun/java2d/loops/GraphicsPrimitive/TRACELOG": number;
    public static "sun/java2d/loops/GraphicsPrimitive/TRACETIMESTAMP": number;
    public static "sun/java2d/loops/GraphicsPrimitive/TRACECOUNTS": number;
    public "sun/java2d/loops/GraphicsPrimitive/cachedname": java_lang_String;
    public static "makePrimTypeID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/makePrimTypeID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "makeUniqueID(ILsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)I"(thread: JVMThread, args: [number, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/makeUniqueID(ILsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)I"(thread: JVMThread, args: [number, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/String;ILsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [java_lang_String, number, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/<init>(Ljava/lang/String;ILsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [java_lang_String, number, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLjava/lang/String;ILsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, java_lang_String, number, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/<init>(JLjava/lang/String;ILsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, java_lang_String, number, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUniqueID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/getUniqueID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPrimTypeID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/getPrimTypeID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNativePrim()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/getNativePrim()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getSourceType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/getSourceType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "getCompositeType()Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/getCompositeType()Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public "getDestType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/getDestType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "satisfies(Ljava/lang/String;Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Z"(thread: JVMThread, args: [java_lang_String, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/satisfies(Ljava/lang/String;Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Z"(thread: JVMThread, args: [java_lang_String, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "satisfiesSameAs(Lsun/java2d/loops/GraphicsPrimitive;)Z"(thread: JVMThread, args: [sun_java2d_loops_GraphicsPrimitive], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/satisfiesSameAs(Lsun/java2d/loops/GraphicsPrimitive;)Z"(thread: JVMThread, args: [sun_java2d_loops_GraphicsPrimitive], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "tracingEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/tracingEnabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getTraceOutputFile()Ljava/io/PrintStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/getTraceOutputFile()Ljava/io/PrintStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public static "tracePrimitive(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/tracePrimitive(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "setupGeneralBinaryOp(Lsun/java2d/loops/GraphicsPrimitive$GeneralBinaryOp;)V"(thread: JVMThread, args: [sun_java2d_loops_GraphicsPrimitive$GeneralBinaryOp], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/setupGeneralBinaryOp(Lsun/java2d/loops/GraphicsPrimitive$GeneralBinaryOp;)V"(thread: JVMThread, args: [sun_java2d_loops_GraphicsPrimitive$GeneralBinaryOp], cb?: (e?: java_lang_Throwable) => void): void;
    public "setupGeneralUnaryOp(Lsun/java2d/loops/GraphicsPrimitive$GeneralUnaryOp;)V"(thread: JVMThread, args: [sun_java2d_loops_GraphicsPrimitive$GeneralUnaryOp], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/setupGeneralUnaryOp(Lsun/java2d/loops/GraphicsPrimitive$GeneralUnaryOp;)V"(thread: JVMThread, args: [sun_java2d_loops_GraphicsPrimitive$GeneralUnaryOp], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createConverter(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/Blit;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_Blit) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/createConverter(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/Blit;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_Blit) => void): void;
    public static "convertFrom(Lsun/java2d/loops/Blit;Lsun/java2d/SurfaceData;IIIILsun/java2d/SurfaceData;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [sun_java2d_loops_Blit, sun_java2d_SurfaceData, number, number, number, number, sun_java2d_SurfaceData], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/convertFrom(Lsun/java2d/loops/Blit;Lsun/java2d/SurfaceData;IIIILsun/java2d/SurfaceData;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [sun_java2d_loops_Blit, sun_java2d_SurfaceData, number, number, number, number, sun_java2d_SurfaceData], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public static "convertFrom(Lsun/java2d/loops/Blit;Lsun/java2d/SurfaceData;IIIILsun/java2d/SurfaceData;I)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [sun_java2d_loops_Blit, sun_java2d_SurfaceData, number, number, number, number, sun_java2d_SurfaceData, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/convertFrom(Lsun/java2d/loops/Blit;Lsun/java2d/SurfaceData;IIIILsun/java2d/SurfaceData;I)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [sun_java2d_loops_Blit, sun_java2d_SurfaceData, number, number, number, number, sun_java2d_SurfaceData, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public static "convertTo(Lsun/java2d/loops/Blit;Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;Lsun/java2d/pipe/Region;IIII)V"(thread: JVMThread, args: [sun_java2d_loops_Blit, sun_java2d_SurfaceData, sun_java2d_SurfaceData, sun_java2d_pipe_Region, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/convertTo(Lsun/java2d/loops/Blit;Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;Lsun/java2d/pipe/Region;IIII)V"(thread: JVMThread, args: [sun_java2d_loops_Blit, sun_java2d_SurfaceData, sun_java2d_SurfaceData, sun_java2d_pipe_Region, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getGeneralOp(ILsun/java2d/loops/CompositeType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [number, sun_java2d_loops_CompositeType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/getGeneralOp(ILsun/java2d/loops/CompositeType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [number, sun_java2d_loops_CompositeType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "simplename([Ljava/lang/reflect/Field;Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Field>, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/simplename([Ljava/lang/reflect/Field;Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Field>, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "simplename(Lsun/java2d/loops/SurfaceType;)Ljava/lang/String;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/simplename(Lsun/java2d/loops/SurfaceType;)Ljava/lang/String;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "simplename(Lsun/java2d/loops/CompositeType;)Ljava/lang/String;"(thread: JVMThread, args: [sun_java2d_loops_CompositeType], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/simplename(Lsun/java2d/loops/CompositeType;)Ljava/lang/String;"(thread: JVMThread, args: [sun_java2d_loops_CompositeType], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/java2d/loops/GraphicsPrimitive/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000()Ljava/io/PrintStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/access$000()Ljava/io/PrintStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/GraphicsPrimitive/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_Blit extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/Blit/methodSignature": java_lang_String;
    public static "sun/java2d/loops/Blit/primTypeID": number;
    public static "sun/java2d/loops/Blit/blitcache": sun_java2d_loops_RenderCache;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/Blit;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_Blit) => void): void;
    public static "sun/java2d/loops/Blit/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/Blit;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_Blit) => void): void;
    public static "getFromCache(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/Blit;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_Blit) => void): void;
    public static "sun/java2d/loops/Blit/getFromCache(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/Blit;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_Blit) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/Blit/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/Blit/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "Blit(Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;Ljava/awt/Composite;Lsun/java2d/pipe/Region;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SurfaceData, sun_java2d_SurfaceData, java_awt_Composite, sun_java2d_pipe_Region, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/Blit/Blit(Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;Ljava/awt/Composite;Lsun/java2d/pipe/Region;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SurfaceData, sun_java2d_SurfaceData, java_awt_Composite, sun_java2d_pipe_Region, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/Blit/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/Blit/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/Blit/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_loops_GraphicsPrimitive$GeneralUnaryOp extends java_lang_Object {
    "setPrimitives(Lsun/java2d/loops/Blit;Lsun/java2d/loops/GraphicsPrimitive;Lsun/java2d/loops/Blit;)V"(thread: JVMThread, args: [sun_java2d_loops_Blit, sun_java2d_loops_GraphicsPrimitive, sun_java2d_loops_Blit], cb?: (e?: java_lang_Throwable) => void): void;
    "getCompositeType()Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    "getDestType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    "getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getPrimTypeID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface sun_java2d_loops_GraphicsPrimitive$GeneralBinaryOp extends java_lang_Object {
    "setPrimitives(Lsun/java2d/loops/Blit;Lsun/java2d/loops/Blit;Lsun/java2d/loops/GraphicsPrimitive;Lsun/java2d/loops/Blit;)V"(thread: JVMThread, args: [sun_java2d_loops_Blit, sun_java2d_loops_Blit, sun_java2d_loops_GraphicsPrimitive, sun_java2d_loops_Blit], cb?: (e?: java_lang_Throwable) => void): void;
    "getSourceType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    "getCompositeType()Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    "getDestType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    "getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getPrimTypeID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_java2d_loops_DrawGlyphListAA extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/DrawGlyphListAA/methodSignature": java_lang_String;
    public static "sun/java2d/loops/DrawGlyphListAA/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawGlyphListAA;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawGlyphListAA) => void): void;
    public static "sun/java2d/loops/DrawGlyphListAA/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawGlyphListAA;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawGlyphListAA) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawGlyphListAA/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawGlyphListAA/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "DrawGlyphListAA(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawGlyphListAA/DrawGlyphListAA(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawGlyphListAA/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawGlyphListAA/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/DrawGlyphListAA/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_DrawGlyphList extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/DrawGlyphList/methodSignature": java_lang_String;
    public static "sun/java2d/loops/DrawGlyphList/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawGlyphList;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawGlyphList) => void): void;
    public static "sun/java2d/loops/DrawGlyphList/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawGlyphList;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawGlyphList) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawGlyphList/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawGlyphList/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "DrawGlyphList(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawGlyphList/DrawGlyphList(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawGlyphList/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawGlyphList/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/DrawGlyphList/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_DrawParallelogram extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/DrawParallelogram/methodSignature": java_lang_String;
    public static "sun/java2d/loops/DrawParallelogram/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawParallelogram;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawParallelogram) => void): void;
    public static "sun/java2d/loops/DrawParallelogram/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawParallelogram;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawParallelogram) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawParallelogram/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawParallelogram/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "DrawParallelogram(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;DDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawParallelogram/DrawParallelogram(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;DDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawParallelogram/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawParallelogram/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/DrawParallelogram/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_FillParallelogram extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/FillParallelogram/methodSignature": java_lang_String;
    public static "sun/java2d/loops/FillParallelogram/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/FillParallelogram;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FillParallelogram) => void): void;
    public static "sun/java2d/loops/FillParallelogram/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/FillParallelogram;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FillParallelogram) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillParallelogram/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillParallelogram/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "FillParallelogram(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;DDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillParallelogram/FillParallelogram(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;DDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/FillParallelogram/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/FillParallelogram/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/FillParallelogram/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_FillSpans extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/FillSpans/methodSignature": java_lang_String;
    public static "sun/java2d/loops/FillSpans/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/FillSpans;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FillSpans) => void): void;
    public static "sun/java2d/loops/FillSpans/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/FillSpans;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FillSpans) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillSpans/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillSpans/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "FillSpans(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IJLsun/java2d/pipe/SpanIterator;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, Long, any, sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillSpans/FillSpans(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IJLsun/java2d/pipe/SpanIterator;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, Long, any, sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "FillSpans(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Lsun/java2d/pipe/SpanIterator;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillSpans/FillSpans(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Lsun/java2d/pipe/SpanIterator;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/FillSpans/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/FillSpans/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/FillSpans/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_pipe_SpanIterator extends java_lang_Object {
    "getPathBox([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    "intersectClipBox(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "nextSpan([I)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "skipDownTo(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "getNativeIterator()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class sun_java2d_loops_FillPath extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/FillPath/methodSignature": java_lang_String;
    public static "sun/java2d/loops/FillPath/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/FillPath;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FillPath) => void): void;
    public static "sun/java2d/loops/FillPath/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/FillPath;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FillPath) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillPath/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillPath/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "FillPath(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IILjava/awt/geom/Path2D$Float;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, java_awt_geom_Path2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillPath/FillPath(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IILjava/awt/geom/Path2D$Float;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, java_awt_geom_Path2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/FillPath/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/FillPath/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/FillPath/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_geom_Path2D$Float extends java_awt_geom_Path2D implements java_io_Serializable {
    public "java/awt/geom/Path2D$Float/floatCoords": JVMArray<number>;
    public static "java/awt/geom/Path2D$Float/serialVersionUID": Long;
    public static "java/awt/geom/Path2D$Float/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/<init>(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_Shape, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/<init>(Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_Shape, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "cloneCoordsFloat(Ljava/awt/geom/AffineTransform;)[F"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/geom/Path2D$Float/cloneCoordsFloat(Ljava/awt/geom/AffineTransform;)[F"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "cloneCoordsDouble(Ljava/awt/geom/AffineTransform;)[D"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/geom/Path2D$Float/cloneCoordsDouble(Ljava/awt/geom/AffineTransform;)[D"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "append(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/append(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/append(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPoint(I)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/geom/Path2D$Float/getPoint(I)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "needRoom(ZI)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/needRoom(ZI)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "expandCoords([FI)[F"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/geom/Path2D$Float/expandCoords([FI)[F"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "moveTo(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/moveTo(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "moveTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/moveTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "lineTo(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/lineTo(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "lineTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/lineTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "quadTo(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/quadTo(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "quadTo(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/quadTo(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "curveTo(DDDDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/curveTo(DDDDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "curveTo(FFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/curveTo(FFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "pointCrossings(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D$Float/pointCrossings(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rectCrossings(DDDD)I"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D$Float/rectCrossings(DDDD)I"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "append(Ljava/awt/geom/PathIterator;Z)V"(thread: JVMThread, args: [java_awt_geom_PathIterator, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/append(Ljava/awt/geom/PathIterator;Z)V"(thread: JVMThread, args: [java_awt_geom_PathIterator, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "transform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/transform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/geom/Path2D$Float/getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "java/awt/geom/Path2D$Float/getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/geom/Path2D$Float/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D$Float/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/geom/Path2D$Float/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_geom_Path2D extends java_lang_Object implements java_awt_Shape, java_lang_Cloneable {
    public static "java/awt/geom/Path2D/WIND_EVEN_ODD": number;
    public static "java/awt/geom/Path2D/WIND_NON_ZERO": number;
    public static "java/awt/geom/Path2D/SEG_MOVETO": number;
    public static "java/awt/geom/Path2D/SEG_LINETO": number;
    public static "java/awt/geom/Path2D/SEG_QUADTO": number;
    public static "java/awt/geom/Path2D/SEG_CUBICTO": number;
    public static "java/awt/geom/Path2D/SEG_CLOSE": number;
    public "java/awt/geom/Path2D/pointTypes": JVMArray<number>;
    public "java/awt/geom/Path2D/numTypes": number;
    public "java/awt/geom/Path2D/numCoords": number;
    public "java/awt/geom/Path2D/windingRule": number;
    public static "java/awt/geom/Path2D/INIT_SIZE": number;
    public static "java/awt/geom/Path2D/EXPAND_MAX": number;
    public static "java/awt/geom/Path2D/EXPAND_MAX_COORDS": number;
    public static "java/awt/geom/Path2D/EXPAND_MIN": number;
    public static "java/awt/geom/Path2D/SERIAL_STORAGE_FLT_ARRAY": number;
    public static "java/awt/geom/Path2D/SERIAL_STORAGE_DBL_ARRAY": number;
    public static "java/awt/geom/Path2D/SERIAL_SEG_FLT_MOVETO": number;
    public static "java/awt/geom/Path2D/SERIAL_SEG_FLT_LINETO": number;
    public static "java/awt/geom/Path2D/SERIAL_SEG_FLT_QUADTO": number;
    public static "java/awt/geom/Path2D/SERIAL_SEG_FLT_CUBICTO": number;
    public static "java/awt/geom/Path2D/SERIAL_SEG_DBL_MOVETO": number;
    public static "java/awt/geom/Path2D/SERIAL_SEG_DBL_LINETO": number;
    public static "java/awt/geom/Path2D/SERIAL_SEG_DBL_QUADTO": number;
    public static "java/awt/geom/Path2D/SERIAL_SEG_DBL_CUBICTO": number;
    public static "java/awt/geom/Path2D/SERIAL_SEG_CLOSE": number;
    public static "java/awt/geom/Path2D/SERIAL_PATH_END": number;
    public static "java/awt/geom/Path2D/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "cloneCoordsFloat(Ljava/awt/geom/AffineTransform;)[F"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/geom/Path2D/cloneCoordsFloat(Ljava/awt/geom/AffineTransform;)[F"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "cloneCoordsDouble(Ljava/awt/geom/AffineTransform;)[D"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/geom/Path2D/cloneCoordsDouble(Ljava/awt/geom/AffineTransform;)[D"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "append(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/append(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/append(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPoint(I)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/geom/Path2D/getPoint(I)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "needRoom(ZI)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/needRoom(ZI)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "pointCrossings(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D/pointCrossings(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rectCrossings(DDDD)I"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D/rectCrossings(DDDD)I"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "expandPointTypes([BI)[B"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/geom/Path2D/expandPointTypes([BI)[B"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "moveTo(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/moveTo(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "lineTo(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/lineTo(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "quadTo(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/quadTo(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "curveTo(DDDDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/curveTo(DDDDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "closePath()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/closePath()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(Ljava/awt/Shape;Z)V"(thread: JVMThread, args: [java_awt_Shape, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/append(Ljava/awt/Shape;Z)V"(thread: JVMThread, args: [java_awt_Shape, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "append(Ljava/awt/geom/PathIterator;Z)V"(thread: JVMThread, args: [java_awt_geom_PathIterator, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/append(Ljava/awt/geom/PathIterator;Z)V"(thread: JVMThread, args: [java_awt_geom_PathIterator, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWindingRule()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D/getWindingRule()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setWindingRule(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/setWindingRule(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCurrentPoint()Ljava/awt/geom/Point2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "java/awt/geom/Path2D/getCurrentPoint()Ljava/awt/geom/Point2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "transform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/transform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "createTransformedShape(Ljava/awt/geom/AffineTransform;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/geom/Path2D/createTransformedShape(Ljava/awt/geom/AffineTransform;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/geom/Path2D/getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public static "contains(Ljava/awt/geom/PathIterator;DD)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Path2D/contains(Ljava/awt/geom/PathIterator;DD)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "contains(Ljava/awt/geom/PathIterator;Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Path2D/contains(Ljava/awt/geom/PathIterator;Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D/contains(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D/contains(Ljava/awt/geom/Point2D;)Z"(thread: JVMThread, args: [java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "contains(Ljava/awt/geom/PathIterator;DDDD)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Path2D/contains(Ljava/awt/geom/PathIterator;DDDD)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "contains(Ljava/awt/geom/PathIterator;Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Path2D/contains(Ljava/awt/geom/PathIterator;Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D/contains(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D/contains(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "intersects(Ljava/awt/geom/PathIterator;DDDD)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Path2D/intersects(Ljava/awt/geom/PathIterator;DDDD)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "intersects(Ljava/awt/geom/PathIterator;Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/geom/Path2D/intersects(Ljava/awt/geom/PathIterator;Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_PathIterator, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D/intersects(DDDD)Z"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersects(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Path2D/intersects(Ljava/awt/geom/Rectangle2D;)Z"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "java/awt/geom/Path2D/getPathIterator(Ljava/awt/geom/AffineTransform;D)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/geom/Path2D/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;Z)V"(thread: JVMThread, args: [java_io_ObjectOutputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/writeObject(Ljava/io/ObjectOutputStream;Z)V"(thread: JVMThread, args: [java_io_ObjectOutputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;Z)V"(thread: JVMThread, args: [java_io_ObjectInputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Path2D/readObject(Ljava/io/ObjectInputStream;Z)V"(thread: JVMThread, args: [java_io_ObjectInputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/geom/Path2D/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    "getPathIterator(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/PathIterator;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_PathIterator) => void): void;
  }
  export class sun_java2d_loops_DrawPath extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/DrawPath/methodSignature": java_lang_String;
    public static "sun/java2d/loops/DrawPath/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawPath;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawPath) => void): void;
    public static "sun/java2d/loops/DrawPath/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawPath;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawPath) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawPath/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawPath/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "DrawPath(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IILjava/awt/geom/Path2D$Float;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, java_awt_geom_Path2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawPath/DrawPath(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IILjava/awt/geom/Path2D$Float;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, java_awt_geom_Path2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawPath/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawPath/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/DrawPath/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_DrawPolygons extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/DrawPolygons/methodSignature": java_lang_String;
    public static "sun/java2d/loops/DrawPolygons/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawPolygons;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawPolygons) => void): void;
    public static "sun/java2d/loops/DrawPolygons/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawPolygons;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawPolygons) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawPolygons/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawPolygons/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "DrawPolygons(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;[I[I[IIIIZ)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, JVMArray<number>, JVMArray<number>, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawPolygons/DrawPolygons(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;[I[I[IIIIZ)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, JVMArray<number>, JVMArray<number>, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawPolygons/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawPolygons/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/DrawPolygons/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_DrawRect extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/DrawRect/methodSignature": java_lang_String;
    public static "sun/java2d/loops/DrawRect/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawRect;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawRect) => void): void;
    public static "sun/java2d/loops/DrawRect/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawRect;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawRect) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawRect/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawRect/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "DrawRect(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawRect/DrawRect(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawRect/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawRect/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/DrawRect/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_FillRect extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/FillRect/methodSignature": java_lang_String;
    public static "sun/java2d/loops/FillRect/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/FillRect;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FillRect) => void): void;
    public static "sun/java2d/loops/FillRect/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/FillRect;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FillRect) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillRect/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillRect/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "FillRect(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FillRect/FillRect(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/FillRect/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/FillRect/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/FillRect/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_DrawLine extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/DrawLine/methodSignature": java_lang_String;
    public static "sun/java2d/loops/DrawLine/primTypeID": number;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawLine;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawLine) => void): void;
    public static "sun/java2d/loops/DrawLine/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/DrawLine;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_DrawLine) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawLine/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawLine/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "DrawLine(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/DrawLine/DrawLine(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawLine/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/DrawLine/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/DrawLine/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_MaskFill extends sun_java2d_loops_GraphicsPrimitive {
    public static "sun/java2d/loops/MaskFill/methodSignature": java_lang_String;
    public static "sun/java2d/loops/MaskFill/fillPgramSignature": java_lang_String;
    public static "sun/java2d/loops/MaskFill/drawPgramSignature": java_lang_String;
    public static "sun/java2d/loops/MaskFill/primTypeID": number;
    public static "sun/java2d/loops/MaskFill/fillcache": sun_java2d_loops_RenderCache;
    public static "locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/MaskFill;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_MaskFill) => void): void;
    public static "sun/java2d/loops/MaskFill/locate(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/MaskFill;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_MaskFill) => void): void;
    public static "locatePrim(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/MaskFill;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_MaskFill) => void): void;
    public static "sun/java2d/loops/MaskFill/locatePrim(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/MaskFill;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_MaskFill) => void): void;
    public static "getFromCache(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/MaskFill;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_MaskFill) => void): void;
    public static "sun/java2d/loops/MaskFill/getFromCache(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/MaskFill;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_MaskFill) => void): void;
    public "<init>(Ljava/lang/String;Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [java_lang_String, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/MaskFill/<init>(Ljava/lang/String;Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [java_lang_String, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/MaskFill/<init>(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/MaskFill/<init>(JLsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)V"(thread: JVMThread, args: [Long, any, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable) => void): void;
    public "MaskFill(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Ljava/awt/Composite;IIII[BII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, java_awt_Composite, number, number, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/MaskFill/MaskFill(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Ljava/awt/Composite;IIII[BII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, java_awt_Composite, number, number, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "FillAAPgram(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Ljava/awt/Composite;DDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, java_awt_Composite, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/MaskFill/FillAAPgram(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Ljava/awt/Composite;DDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, java_awt_Composite, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "DrawAAPgram(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Ljava/awt/Composite;DDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, java_awt_Composite, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/MaskFill/DrawAAPgram(Lsun/java2d/SunGraphics2D;Lsun/java2d/SurfaceData;Ljava/awt/Composite;DDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_SurfaceData, java_awt_Composite, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "canDoParallelograms()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/MaskFill/canDoParallelograms()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/MaskFill/makePrimitive(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public "sun/java2d/loops/MaskFill/traceWrap()Lsun/java2d/loops/GraphicsPrimitive;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_GraphicsPrimitive) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/MaskFill/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_SunGraphics2D extends java_awt_Graphics2D implements sun_awt_ConstrainableGraphics, java_lang_Cloneable, sun_java2d_DestSurfaceProvider {
    public static "sun/java2d/SunGraphics2D/PAINT_CUSTOM": number;
    public static "sun/java2d/SunGraphics2D/PAINT_TEXTURE": number;
    public static "sun/java2d/SunGraphics2D/PAINT_RAD_GRADIENT": number;
    public static "sun/java2d/SunGraphics2D/PAINT_LIN_GRADIENT": number;
    public static "sun/java2d/SunGraphics2D/PAINT_GRADIENT": number;
    public static "sun/java2d/SunGraphics2D/PAINT_ALPHACOLOR": number;
    public static "sun/java2d/SunGraphics2D/PAINT_OPAQUECOLOR": number;
    public static "sun/java2d/SunGraphics2D/COMP_CUSTOM": number;
    public static "sun/java2d/SunGraphics2D/COMP_XOR": number;
    public static "sun/java2d/SunGraphics2D/COMP_ALPHA": number;
    public static "sun/java2d/SunGraphics2D/COMP_ISCOPY": number;
    public static "sun/java2d/SunGraphics2D/STROKE_CUSTOM": number;
    public static "sun/java2d/SunGraphics2D/STROKE_WIDE": number;
    public static "sun/java2d/SunGraphics2D/STROKE_THINDASHED": number;
    public static "sun/java2d/SunGraphics2D/STROKE_THIN": number;
    public static "sun/java2d/SunGraphics2D/TRANSFORM_GENERIC": number;
    public static "sun/java2d/SunGraphics2D/TRANSFORM_TRANSLATESCALE": number;
    public static "sun/java2d/SunGraphics2D/TRANSFORM_ANY_TRANSLATE": number;
    public static "sun/java2d/SunGraphics2D/TRANSFORM_INT_TRANSLATE": number;
    public static "sun/java2d/SunGraphics2D/TRANSFORM_ISIDENT": number;
    public static "sun/java2d/SunGraphics2D/CLIP_SHAPE": number;
    public static "sun/java2d/SunGraphics2D/CLIP_RECTANGULAR": number;
    public static "sun/java2d/SunGraphics2D/CLIP_DEVICE": number;
    public "sun/java2d/SunGraphics2D/eargb": number;
    public "sun/java2d/SunGraphics2D/pixel": number;
    public "sun/java2d/SunGraphics2D/surfaceData": sun_java2d_SurfaceData;
    public "sun/java2d/SunGraphics2D/drawpipe": sun_java2d_pipe_PixelDrawPipe;
    public "sun/java2d/SunGraphics2D/fillpipe": sun_java2d_pipe_PixelFillPipe;
    public "sun/java2d/SunGraphics2D/imagepipe": sun_java2d_pipe_DrawImagePipe;
    public "sun/java2d/SunGraphics2D/shapepipe": sun_java2d_pipe_ShapeDrawPipe;
    public "sun/java2d/SunGraphics2D/textpipe": sun_java2d_pipe_TextPipe;
    public "sun/java2d/SunGraphics2D/alphafill": sun_java2d_loops_MaskFill;
    public "sun/java2d/SunGraphics2D/loops": sun_java2d_loops_RenderLoops;
    public "sun/java2d/SunGraphics2D/imageComp": sun_java2d_loops_CompositeType;
    public "sun/java2d/SunGraphics2D/paintState": number;
    public "sun/java2d/SunGraphics2D/compositeState": number;
    public "sun/java2d/SunGraphics2D/strokeState": number;
    public "sun/java2d/SunGraphics2D/transformState": number;
    public "sun/java2d/SunGraphics2D/clipState": number;
    public "sun/java2d/SunGraphics2D/foregroundColor": java_awt_Color;
    public "sun/java2d/SunGraphics2D/backgroundColor": java_awt_Color;
    public "sun/java2d/SunGraphics2D/transform": java_awt_geom_AffineTransform;
    public "sun/java2d/SunGraphics2D/transX": number;
    public "sun/java2d/SunGraphics2D/transY": number;
    public static "sun/java2d/SunGraphics2D/defaultStroke": java_awt_Stroke;
    public static "sun/java2d/SunGraphics2D/defaultComposite": java_awt_Composite;
    public static "sun/java2d/SunGraphics2D/defaultFont": java_awt_Font;
    public "sun/java2d/SunGraphics2D/paint": java_awt_Paint;
    public "sun/java2d/SunGraphics2D/stroke": java_awt_Stroke;
    public "sun/java2d/SunGraphics2D/composite": java_awt_Composite;
    public "sun/java2d/SunGraphics2D/font": java_awt_Font;
    public "sun/java2d/SunGraphics2D/fontMetrics": java_awt_FontMetrics;
    public "sun/java2d/SunGraphics2D/renderHint": number;
    public "sun/java2d/SunGraphics2D/antialiasHint": number;
    public "sun/java2d/SunGraphics2D/textAntialiasHint": number;
    public "sun/java2d/SunGraphics2D/fractionalMetricsHint": number;
    public "sun/java2d/SunGraphics2D/lcdTextContrast": number;
    public static "sun/java2d/SunGraphics2D/lcdTextContrastDefaultValue": number;
    public "sun/java2d/SunGraphics2D/interpolationHint": number;
    public "sun/java2d/SunGraphics2D/strokeHint": number;
    public "sun/java2d/SunGraphics2D/interpolationType": number;
    public "sun/java2d/SunGraphics2D/hints": java_awt_RenderingHints;
    public "sun/java2d/SunGraphics2D/constrainClip": sun_java2d_pipe_Region;
    public "sun/java2d/SunGraphics2D/constrainX": number;
    public "sun/java2d/SunGraphics2D/constrainY": number;
    public "sun/java2d/SunGraphics2D/clipRegion": sun_java2d_pipe_Region;
    public "sun/java2d/SunGraphics2D/usrClip": java_awt_Shape;
    public "sun/java2d/SunGraphics2D/devClip": sun_java2d_pipe_Region;
    public "sun/java2d/SunGraphics2D/devScale": number;
    public "sun/java2d/SunGraphics2D/resolutionVariantHint": number;
    public "sun/java2d/SunGraphics2D/validFontInfo": number;
    public "sun/java2d/SunGraphics2D/fontInfo": sun_java2d_loops_FontInfo;
    public "sun/java2d/SunGraphics2D/glyphVectorFontInfo": sun_java2d_loops_FontInfo;
    public "sun/java2d/SunGraphics2D/glyphVectorFRC": java_awt_font_FontRenderContext;
    public static "sun/java2d/SunGraphics2D/slowTextTransformMask": number;
    public static "sun/java2d/SunGraphics2D/invalidpipe": sun_java2d_pipe_ValidatePipe;
    public static "sun/java2d/SunGraphics2D/IDENT_MATRIX": JVMArray<number>;
    public static "sun/java2d/SunGraphics2D/IDENT_ATX": java_awt_geom_AffineTransform;
    public static "sun/java2d/SunGraphics2D/MINALLOCATED": number;
    public static "sun/java2d/SunGraphics2D/TEXTARRSIZE": number;
    public static "sun/java2d/SunGraphics2D/textTxArr": JVMArray<JVMArray<number>>;
    public static "sun/java2d/SunGraphics2D/textAtArr": JVMArray<java_awt_geom_AffineTransform>;
    public static "sun/java2d/SunGraphics2D/NON_UNIFORM_SCALE_MASK": number;
    public static "sun/java2d/SunGraphics2D/MinPenSizeAA": number;
    public static "sun/java2d/SunGraphics2D/MinPenSizeAASquared": number;
    public static "sun/java2d/SunGraphics2D/MinPenSizeSquared": number;
    public static "sun/java2d/SunGraphics2D/NON_RECTILINEAR_TRANSFORM_MASK": number;
    public "sun/java2d/SunGraphics2D/lastCAblit": sun_java2d_loops_Blit;
    public "sun/java2d/SunGraphics2D/lastCAcomp": java_awt_Composite;
    public "sun/java2d/SunGraphics2D/cachedFRC": java_awt_font_FontRenderContext;
    public "<init>(Lsun/java2d/SurfaceData;Ljava/awt/Color;Ljava/awt/Color;Ljava/awt/Font;)V"(thread: JVMThread, args: [sun_java2d_SurfaceData, java_awt_Color, java_awt_Color, java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/<init>(Lsun/java2d/SurfaceData;Ljava/awt/Color;Ljava/awt/Color;Ljava/awt/Font;)V"(thread: JVMThread, args: [sun_java2d_SurfaceData, java_awt_Color, java_awt_Color, java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/SunGraphics2D/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "create()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "sun/java2d/SunGraphics2D/create()Ljava/awt/Graphics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Graphics) => void): void;
    public "setDevClip(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setDevClip(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDevClip(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setDevClip(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "constrain(IIIILsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [number, number, number, number, sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/constrain(IIIILsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [number, number, number, number, sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "constrain(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/constrain(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidatePipe()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/invalidatePipe()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "validatePipe()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/validatePipe()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "intersectShapes(Ljava/awt/Shape;Ljava/awt/Shape;ZZ)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape, java_awt_Shape, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/java2d/SunGraphics2D/intersectShapes(Ljava/awt/Shape;Ljava/awt/Shape;ZZ)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape, java_awt_Shape, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "intersectRectShape(Ljava/awt/geom/Rectangle2D;Ljava/awt/Shape;ZZ)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D, java_awt_Shape, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/java2d/SunGraphics2D/intersectRectShape(Ljava/awt/geom/Rectangle2D;Ljava/awt/Shape;ZZ)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D, java_awt_Shape, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public static "cloneShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public static "sun/java2d/SunGraphics2D/cloneShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "intersectByArea(Ljava/awt/Shape;Ljava/awt/Shape;ZZ)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape, java_awt_Shape, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/java2d/SunGraphics2D/intersectByArea(Ljava/awt/Shape;Ljava/awt/Shape;ZZ)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape, java_awt_Shape, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getCompClip()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/SunGraphics2D/getCompClip()Lsun/java2d/pipe/Region;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "sun/java2d/SunGraphics2D/getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "checkFontInfo(Lsun/java2d/loops/FontInfo;Ljava/awt/Font;Ljava/awt/font/FontRenderContext;)Lsun/java2d/loops/FontInfo;"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, java_awt_Font, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FontInfo) => void): void;
    public "sun/java2d/SunGraphics2D/checkFontInfo(Lsun/java2d/loops/FontInfo;Ljava/awt/Font;Ljava/awt/font/FontRenderContext;)Lsun/java2d/loops/FontInfo;"(thread: JVMThread, args: [sun_java2d_loops_FontInfo, java_awt_Font, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FontInfo) => void): void;
    public static "isRotated([D)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/SunGraphics2D/isRotated([D)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setFont(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontInfo()Lsun/java2d/loops/FontInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FontInfo) => void): void;
    public "sun/java2d/SunGraphics2D/getFontInfo()Lsun/java2d/loops/FontInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FontInfo) => void): void;
    public "getGVFontInfo(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;)Lsun/java2d/loops/FontInfo;"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FontInfo) => void): void;
    public "sun/java2d/SunGraphics2D/getGVFontInfo(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;)Lsun/java2d/loops/FontInfo;"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_FontInfo) => void): void;
    public "getFontMetrics()Ljava/awt/FontMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "sun/java2d/SunGraphics2D/getFontMetrics()Ljava/awt/FontMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "sun/java2d/SunGraphics2D/getFontMetrics(Ljava/awt/Font;)Ljava/awt/FontMetrics;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: java_awt_FontMetrics) => void): void;
    public "hit(Ljava/awt/Rectangle;Ljava/awt/Shape;Z)Z"(thread: JVMThread, args: [java_awt_Rectangle, java_awt_Shape, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/hit(Ljava/awt/Rectangle;Ljava/awt/Shape;Z)Z"(thread: JVMThread, args: [java_awt_Rectangle, java_awt_Shape, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDeviceColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "sun/java2d/SunGraphics2D/getDeviceColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    public "getDeviceConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "sun/java2d/SunGraphics2D/getDeviceConfiguration()Ljava/awt/GraphicsConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_GraphicsConfiguration) => void): void;
    public "getSurfaceData()Lsun/java2d/SurfaceData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "sun/java2d/SunGraphics2D/getSurfaceData()Lsun/java2d/SurfaceData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "setComposite(Ljava/awt/Composite;)V"(thread: JVMThread, args: [java_awt_Composite], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setComposite(Ljava/awt/Composite;)V"(thread: JVMThread, args: [java_awt_Composite], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPaint(Ljava/awt/Paint;)V"(thread: JVMThread, args: [java_awt_Paint], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setPaint(Ljava/awt/Paint;)V"(thread: JVMThread, args: [java_awt_Paint], cb?: (e?: java_lang_Throwable) => void): void;
    public "validateBasicStroke(Ljava/awt/BasicStroke;)V"(thread: JVMThread, args: [java_awt_BasicStroke], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/validateBasicStroke(Ljava/awt/BasicStroke;)V"(thread: JVMThread, args: [java_awt_BasicStroke], cb?: (e?: java_lang_Throwable) => void): void;
    public "setStroke(Ljava/awt/Stroke;)V"(thread: JVMThread, args: [java_awt_Stroke], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setStroke(Ljava/awt/Stroke;)V"(thread: JVMThread, args: [java_awt_Stroke], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRenderingHint(Ljava/awt/RenderingHints$Key;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_RenderingHints$Key, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setRenderingHint(Ljava/awt/RenderingHints$Key;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_awt_RenderingHints$Key, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRenderingHint(Ljava/awt/RenderingHints$Key;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_RenderingHints$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/SunGraphics2D/getRenderingHint(Ljava/awt/RenderingHints$Key;)Ljava/lang/Object;"(thread: JVMThread, args: [java_awt_RenderingHints$Key], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setRenderingHints(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setRenderingHints(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "addRenderingHints(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/addRenderingHints(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRenderingHints()Ljava/awt/RenderingHints;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_RenderingHints) => void): void;
    public "sun/java2d/SunGraphics2D/getRenderingHints()Ljava/awt/RenderingHints;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_RenderingHints) => void): void;
    public "makeHints(Ljava/util/Map;)Ljava/awt/RenderingHints;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_RenderingHints) => void): void;
    public "sun/java2d/SunGraphics2D/makeHints(Ljava/util/Map;)Ljava/awt/RenderingHints;"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_awt_RenderingHints) => void): void;
    public "translate(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/translate(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/rotate(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "rotate(DDD)V"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/rotate(DDD)V"(thread: JVMThread, args: [number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "scale(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/scale(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "shear(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/shear(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "transform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/transform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/translate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setTransform(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidateTransform()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/invalidateTransform()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "sun/java2d/SunGraphics2D/getTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "cloneTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "sun/java2d/SunGraphics2D/cloneTransform()Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "getPaint()Ljava/awt/Paint;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Paint) => void): void;
    public "sun/java2d/SunGraphics2D/getPaint()Ljava/awt/Paint;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Paint) => void): void;
    public "getComposite()Ljava/awt/Composite;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Composite) => void): void;
    public "sun/java2d/SunGraphics2D/getComposite()Ljava/awt/Composite;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Composite) => void): void;
    public "getColor()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "sun/java2d/SunGraphics2D/getColor()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "validateColor()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/validateColor()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setColor(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setColor(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setBackground(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "sun/java2d/SunGraphics2D/getBackground()Ljava/awt/Color;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Color) => void): void;
    public "getStroke()Ljava/awt/Stroke;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Stroke) => void): void;
    public "sun/java2d/SunGraphics2D/getStroke()Ljava/awt/Stroke;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Stroke) => void): void;
    public "getClipBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "sun/java2d/SunGraphics2D/getClipBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getClipBounds(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "sun/java2d/SunGraphics2D/getClipBounds(Ljava/awt/Rectangle;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "hitClip(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/hitClip(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "validateCompClip()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/validateCompClip()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "transformShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/java2d/SunGraphics2D/transformShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "untransformShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/java2d/SunGraphics2D/untransformShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public static "transformShape(IILjava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number, java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public static "sun/java2d/SunGraphics2D/transformShape(IILjava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number, java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public static "transformShape(Ljava/awt/geom/AffineTransform;Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public static "sun/java2d/SunGraphics2D/transformShape(Ljava/awt/geom/AffineTransform;Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public static "fixRectangleOrientation([DLjava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [JVMArray<number>, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/SunGraphics2D/fixRectangleOrientation([DLjava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [JVMArray<number>, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "clipRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/clipRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setClip(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setClip(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getClip()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/java2d/SunGraphics2D/getClip()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "setClip(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setClip(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "clip(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/clip(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPaintMode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setPaintMode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setXORMode(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/setXORMode(Ljava/awt/Color;)V"(thread: JVMThread, args: [java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyArea(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/copyArea(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "doCopyArea(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/doCopyArea(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawLine(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawLine(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRoundRect(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawRoundRect(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRoundRect(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/fillRoundRect(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawOval(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawOval(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillOval(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/fillOval(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawArc(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawArc(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillArc(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/fillArc(IIIIII)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolyline([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawPolyline([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolygon([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawPolygon([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillPolygon([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/fillPolygon([I[II)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/fillRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "revalidateAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/revalidateAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/clearRect(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "draw(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/draw(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "fill(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/fill(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isIntegerTranslation(Ljava/awt/geom/AffineTransform;)Z"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/SunGraphics2D/isIntegerTranslation(Ljava/awt/geom/AffineTransform;)Z"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getTileIndex(III)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/SunGraphics2D/getTileIndex(III)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getImageRegion(Ljava/awt/image/RenderedImage;Lsun/java2d/pipe/Region;Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;II)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_image_RenderedImage, sun_java2d_pipe_Region, java_awt_geom_AffineTransform, java_awt_geom_AffineTransform, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public static "sun/java2d/SunGraphics2D/getImageRegion(Ljava/awt/image/RenderedImage;Lsun/java2d/pipe/Region;Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;II)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_image_RenderedImage, sun_java2d_pipe_Region, java_awt_geom_AffineTransform, java_awt_geom_AffineTransform, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "drawRenderedImage(Ljava/awt/image/RenderedImage;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_image_RenderedImage, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawRenderedImage(Ljava/awt/image/RenderedImage;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_image_RenderedImage, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "clipTo(Ljava/awt/Rectangle;Ljava/awt/Rectangle;)Z"(thread: JVMThread, args: [java_awt_Rectangle, java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/clipTo(Ljava/awt/Rectangle;Ljava/awt/Rectangle;)Z"(thread: JVMThread, args: [java_awt_Rectangle, java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawTranslatedRenderedImage(Ljava/awt/image/RenderedImage;Ljava/awt/Rectangle;II)V"(thread: JVMThread, args: [java_awt_image_RenderedImage, java_awt_Rectangle, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawTranslatedRenderedImage(Ljava/awt/image/RenderedImage;Ljava/awt/Rectangle;II)V"(thread: JVMThread, args: [java_awt_image_RenderedImage, java_awt_Rectangle, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRenderableImage(Ljava/awt/image/renderable/RenderableImage;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_image_renderable_RenderableImage, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawRenderableImage(Ljava/awt/image/renderable/RenderableImage;Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_image_renderable_RenderableImage, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "transformBounds(Ljava/awt/Rectangle;Ljava/awt/geom/AffineTransform;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "sun/java2d/SunGraphics2D/transformBounds(Ljava/awt/Rectangle;Ljava/awt/geom/AffineTransform;)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_Rectangle, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "drawString(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawString(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Ljava/lang/String;FF)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawString(Ljava/lang/String;FF)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Ljava/text/AttributedCharacterIterator;II)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawString(Ljava/text/AttributedCharacterIterator;II)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Ljava/text/AttributedCharacterIterator;FF)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawString(Ljava/text/AttributedCharacterIterator;FF)V"(thread: JVMThread, args: [java_text_AttributedCharacterIterator, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawGlyphVector(Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawGlyphVector(Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawChars([CIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawChars([CIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawBytes([BIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawBytes([BIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isHiDPIImage(Ljava/awt/Image;)Z"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/isHiDPIImage(Ljava/awt/Image;)Z"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawHiDPIImage(Ljava/awt/Image;IIIIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/drawHiDPIImage(Ljava/awt/Image;IIIIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getResolutionVariant(Lsun/awt/image/MultiResolutionImage;IIIIIIIIII)Ljava/awt/Image;"(thread: JVMThread, args: [sun_awt_image_MultiResolutionImage, number, number, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "sun/java2d/SunGraphics2D/getResolutionVariant(Lsun/awt/image/MultiResolutionImage;IIIIIIIIII)Ljava/awt/Image;"(thread: JVMThread, args: [sun_awt_image_MultiResolutionImage, number, number, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    public "drawImage(Ljava/awt/Image;IIIILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/drawImage(Ljava/awt/Image;IIIILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "copyImage(Ljava/awt/Image;IIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/copyImage(Ljava/awt/Image;IIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/drawImage(Ljava/awt/Image;IIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/drawImage(Ljava/awt/Image;IILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/drawImage(Ljava/awt/Image;IILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IIIIIIIILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/drawImage(Ljava/awt/Image;IIIIIIIILjava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;IIIIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/drawImage(Ljava/awt/Image;IIIIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, number, number, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/Image;Ljava/awt/geom/AffineTransform;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, java_awt_geom_AffineTransform, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SunGraphics2D/drawImage(Ljava/awt/Image;Ljava/awt/geom/AffineTransform;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [java_awt_Image, java_awt_geom_AffineTransform, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawImage(Ljava/awt/image/BufferedImage;Ljava/awt/image/BufferedImageOp;II)V"(thread: JVMThread, args: [java_awt_image_BufferedImage, java_awt_image_BufferedImageOp, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/drawImage(Ljava/awt/image/BufferedImage;Ljava/awt/image/BufferedImageOp;II)V"(thread: JVMThread, args: [java_awt_image_BufferedImage, java_awt_image_BufferedImageOp, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "sun/java2d/SunGraphics2D/getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SunGraphics2D/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDestination()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/SunGraphics2D/getDestination()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getDestSurface()Lsun/java2d/Surface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_Surface) => void): void;
    public "sun/java2d/SunGraphics2D/getDestSurface()Lsun/java2d/Surface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_Surface) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/SunGraphics2D/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_awt_image_MultiResolutionImage extends java_lang_Object {
    "getResolutionVariant(II)Ljava/awt/Image;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Image) => void): void;
    "getResolutionVariants()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
  }
  export class java_awt_BasicStroke extends java_lang_Object implements java_awt_Stroke {
    public static "java/awt/BasicStroke/JOIN_MITER": number;
    public static "java/awt/BasicStroke/JOIN_ROUND": number;
    public static "java/awt/BasicStroke/JOIN_BEVEL": number;
    public static "java/awt/BasicStroke/CAP_BUTT": number;
    public static "java/awt/BasicStroke/CAP_ROUND": number;
    public static "java/awt/BasicStroke/CAP_SQUARE": number;
    public "java/awt/BasicStroke/width": number;
    public "java/awt/BasicStroke/join": number;
    public "java/awt/BasicStroke/cap": number;
    public "java/awt/BasicStroke/miterlimit": number;
    public "java/awt/BasicStroke/dash": JVMArray<number>;
    public "java/awt/BasicStroke/dash_phase": number;
    public "<init>(FIIF[FF)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/BasicStroke/<init>(FIIF[FF)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(FIIF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/BasicStroke/<init>(FIIF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(FII)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/BasicStroke/<init>(FII)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/BasicStroke/<init>(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/BasicStroke/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "createStrokedShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/BasicStroke/createStrokedShape(Ljava/awt/Shape;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getLineWidth()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BasicStroke/getLineWidth()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getEndCap()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BasicStroke/getEndCap()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLineJoin()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BasicStroke/getLineJoin()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMiterLimit()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BasicStroke/getMiterLimit()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDashArray()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/BasicStroke/getDashArray()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getDashPhase()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BasicStroke/getDashPhase()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BasicStroke/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/BasicStroke/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_java2d_pipe_ValidatePipe extends java_lang_Object implements sun_java2d_pipe_PixelDrawPipe, sun_java2d_pipe_PixelFillPipe, sun_java2d_pipe_ShapeDrawPipe, sun_java2d_pipe_TextPipe, sun_java2d_pipe_DrawImagePipe {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "validate(Lsun/java2d/SunGraphics2D;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/ValidatePipe/validate(Lsun/java2d/SunGraphics2D;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawLine(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawLine(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/fillRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/fillRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/fillOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/fillArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolyline(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawPolyline(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/fillPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Lsun/java2d/SunGraphics2D;Ljava/lang/String;DD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_lang_String, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawString(Lsun/java2d/SunGraphics2D;Ljava/lang/String;DD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_lang_String, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawGlyphVector(Lsun/java2d/SunGraphics2D;Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawGlyphVector(Lsun/java2d/SunGraphics2D;Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawChars(Lsun/java2d/SunGraphics2D;[CIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/drawChars(Lsun/java2d/SunGraphics2D;[CIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/ValidatePipe/copyImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "copyImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/ValidatePipe/copyImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "scaleImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/ValidatePipe/scaleImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "scaleImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IIIIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/ValidatePipe/scaleImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IIIIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "transformImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;Ljava/awt/geom/AffineTransform;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, java_awt_geom_AffineTransform, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/ValidatePipe/transformImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;Ljava/awt/geom/AffineTransform;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, java_awt_geom_AffineTransform, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "transformImage(Lsun/java2d/SunGraphics2D;Ljava/awt/image/BufferedImage;Ljava/awt/image/BufferedImageOp;II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_image_BufferedImage, java_awt_image_BufferedImageOp, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ValidatePipe/transformImage(Lsun/java2d/SunGraphics2D;Ljava/awt/image/BufferedImage;Ljava/awt/image/BufferedImageOp;II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_image_BufferedImage, java_awt_image_BufferedImageOp, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_pipe_ShapeDrawPipe extends java_lang_Object {
    "draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    "fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_pipe_PixelFillPipe extends java_lang_Object {
    "fillRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "fillRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "fillOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "fillArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "fillPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_pipe_PixelDrawPipe extends java_lang_Object {
    "drawLine(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "drawRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "drawRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "drawOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "drawArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "drawPolyline(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    "drawPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_DestSurfaceProvider extends java_lang_Object {
    "getDestSurface()Lsun/java2d/Surface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_Surface) => void): void;
  }
  export interface sun_awt_ConstrainableGraphics extends java_lang_Object {
    "constrain(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_pipe_ParallelogramPipe extends java_lang_Object {
    "fillParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    "drawParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_StateTracker extends java_lang_Object {
    "isCurrent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_java2d_SurfaceDataProxy extends java_lang_Object implements sun_awt_DisplayChangedListener, sun_awt_image_SurfaceManager$FlushableCacheData {
    public static "sun/java2d/SurfaceDataProxy/cachingAllowed": number;
    public static "sun/java2d/SurfaceDataProxy/defaultThreshold": number;
    public static "sun/java2d/SurfaceDataProxy/UNCACHED": sun_java2d_SurfaceDataProxy;
    public "sun/java2d/SurfaceDataProxy/threshold": number;
    public "sun/java2d/SurfaceDataProxy/srcTracker": sun_java2d_StateTracker;
    public "sun/java2d/SurfaceDataProxy/numtries": number;
    public "sun/java2d/SurfaceDataProxy/cachedSD": sun_java2d_SurfaceData;
    public "sun/java2d/SurfaceDataProxy/cacheTracker": sun_java2d_StateTracker;
    public "sun/java2d/SurfaceDataProxy/valid": number;
    public static "isCachingAllowed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/SurfaceDataProxy/isCachingAllowed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSupportedOperation(Lsun/java2d/SurfaceData;ILsun/java2d/loops/CompositeType;Ljava/awt/Color;)Z"(thread: JVMThread, args: [sun_java2d_SurfaceData, number, sun_java2d_loops_CompositeType, java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceDataProxy/isSupportedOperation(Lsun/java2d/SurfaceData;ILsun/java2d/loops/CompositeType;Ljava/awt/Color;)Z"(thread: JVMThread, args: [sun_java2d_SurfaceData, number, sun_java2d_loops_CompositeType, java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "validateSurfaceData(Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;II)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [sun_java2d_SurfaceData, sun_java2d_SurfaceData, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "sun/java2d/SurfaceDataProxy/validateSurfaceData(Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;II)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [sun_java2d_SurfaceData, sun_java2d_SurfaceData, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "getRetryTracker(Lsun/java2d/SurfaceData;)Lsun/java2d/StateTracker;"(thread: JVMThread, args: [sun_java2d_SurfaceData], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTracker) => void): void;
    public "sun/java2d/SurfaceDataProxy/getRetryTracker(Lsun/java2d/SurfaceData;)Lsun/java2d/StateTracker;"(thread: JVMThread, args: [sun_java2d_SurfaceData], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTracker) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceDataProxy/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceDataProxy/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceDataProxy/isValid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceDataProxy/invalidate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceDataProxy/flush(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceDataProxy/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAccelerated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/SurfaceDataProxy/isAccelerated()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "activateDisplayListener()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceDataProxy/activateDisplayListener()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "displayChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceDataProxy/displayChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "paletteChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceDataProxy/paletteChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceData(Lsun/java2d/SurfaceData;ILsun/java2d/loops/CompositeType;Ljava/awt/Color;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [sun_java2d_SurfaceData, number, sun_java2d_loops_CompositeType, java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "sun/java2d/SurfaceDataProxy/replaceData(Lsun/java2d/SurfaceData;ILsun/java2d/loops/CompositeType;Ljava/awt/Color;)Lsun/java2d/SurfaceData;"(thread: JVMThread, args: [sun_java2d_SurfaceData, number, sun_java2d_loops_CompositeType, java_awt_Color], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_SurfaceData) => void): void;
    public "updateSurfaceData(Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;II)V"(thread: JVMThread, args: [sun_java2d_SurfaceData, sun_java2d_SurfaceData, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceDataProxy/updateSurfaceData(Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;II)V"(thread: JVMThread, args: [sun_java2d_SurfaceData, sun_java2d_SurfaceData, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateSurfaceDataBg(Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;IILjava/awt/Color;)V"(thread: JVMThread, args: [sun_java2d_SurfaceData, sun_java2d_SurfaceData, number, number, java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/SurfaceDataProxy/updateSurfaceDataBg(Lsun/java2d/SurfaceData;Lsun/java2d/SurfaceData;IILjava/awt/Color;)V"(thread: JVMThread, args: [sun_java2d_SurfaceData, sun_java2d_SurfaceData, number, number, java_awt_Color], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/SurfaceDataProxy/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_awt_image_SurfaceManager$FlushableCacheData extends java_lang_Object {
    "flush(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface sun_awt_DisplayChangedListener extends java_lang_Object, java_util_EventListener {
    "displayChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "paletteChanged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_CompositeType extends java_lang_Object {
    public static "sun/java2d/loops/CompositeType/unusedUID": number;
    public static "sun/java2d/loops/CompositeType/compositeUIDMap": java_util_HashMap;
    public static "sun/java2d/loops/CompositeType/DESC_ANY": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_XOR": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_CLEAR": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_SRC": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_DST": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_SRC_OVER": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_DST_OVER": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_SRC_IN": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_DST_IN": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_SRC_OUT": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_DST_OUT": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_SRC_ATOP": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_DST_ATOP": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_ALPHA_XOR": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_SRC_NO_EA": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_SRC_OVER_NO_EA": java_lang_String;
    public static "sun/java2d/loops/CompositeType/DESC_ANY_ALPHA": java_lang_String;
    public static "sun/java2d/loops/CompositeType/Any": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/General": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/AnyAlpha": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/Xor": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/Clear": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/Src": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/Dst": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/SrcOver": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/DstOver": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/SrcIn": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/DstIn": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/SrcOut": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/DstOut": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/SrcAtop": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/DstAtop": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/AlphaXor": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/SrcNoEa": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/SrcOverNoEa": sun_java2d_loops_CompositeType;
    public static "sun/java2d/loops/CompositeType/OpaqueSrcOverNoEa": sun_java2d_loops_CompositeType;
    public "sun/java2d/loops/CompositeType/uniqueID": number;
    public "sun/java2d/loops/CompositeType/desc": java_lang_String;
    public "sun/java2d/loops/CompositeType/next": sun_java2d_loops_CompositeType;
    public "deriveSubType(Ljava/lang/String;)Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public "sun/java2d/loops/CompositeType/deriveSubType(Ljava/lang/String;)Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public static "forAlphaComposite(Ljava/awt/AlphaComposite;)Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: [java_awt_AlphaComposite], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public static "sun/java2d/loops/CompositeType/forAlphaComposite(Ljava/awt/AlphaComposite;)Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: [java_awt_AlphaComposite], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public "<init>(Lsun/java2d/loops/CompositeType;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_java2d_loops_CompositeType, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/CompositeType/<init>(Lsun/java2d/loops/CompositeType;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_java2d_loops_CompositeType, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "makeUniqueID(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/loops/CompositeType/makeUniqueID(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUniqueID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/CompositeType/getUniqueID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescriptor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/java2d/loops/CompositeType/getDescriptor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSuperType()Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public "sun/java2d/loops/CompositeType/getSuperType()Lsun/java2d/loops/CompositeType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_CompositeType) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/CompositeType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDerivedFrom(Lsun/java2d/loops/CompositeType;)Z"(thread: JVMThread, args: [sun_java2d_loops_CompositeType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/CompositeType/isDerivedFrom(Lsun/java2d/loops/CompositeType;)Z"(thread: JVMThread, args: [sun_java2d_loops_CompositeType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/CompositeType/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/java2d/loops/CompositeType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/CompositeType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_AlphaComposite extends java_lang_Object implements java_awt_Composite {
    public static "java/awt/AlphaComposite/CLEAR": number;
    public static "java/awt/AlphaComposite/SRC": number;
    public static "java/awt/AlphaComposite/DST": number;
    public static "java/awt/AlphaComposite/SRC_OVER": number;
    public static "java/awt/AlphaComposite/DST_OVER": number;
    public static "java/awt/AlphaComposite/SRC_IN": number;
    public static "java/awt/AlphaComposite/DST_IN": number;
    public static "java/awt/AlphaComposite/SRC_OUT": number;
    public static "java/awt/AlphaComposite/DST_OUT": number;
    public static "java/awt/AlphaComposite/SRC_ATOP": number;
    public static "java/awt/AlphaComposite/DST_ATOP": number;
    public static "java/awt/AlphaComposite/XOR": number;
    public static "java/awt/AlphaComposite/Clear": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/Src": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/Dst": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/SrcOver": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/DstOver": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/SrcIn": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/DstIn": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/SrcOut": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/DstOut": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/SrcAtop": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/DstAtop": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/Xor": java_awt_AlphaComposite;
    public static "java/awt/AlphaComposite/MIN_RULE": number;
    public static "java/awt/AlphaComposite/MAX_RULE": number;
    public "java/awt/AlphaComposite/extraAlpha": number;
    public "java/awt/AlphaComposite/rule": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AlphaComposite/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/AlphaComposite/<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance(I)Ljava/awt/AlphaComposite;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AlphaComposite) => void): void;
    public static "java/awt/AlphaComposite/getInstance(I)Ljava/awt/AlphaComposite;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AlphaComposite) => void): void;
    public static "getInstance(IF)Ljava/awt/AlphaComposite;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AlphaComposite) => void): void;
    public static "java/awt/AlphaComposite/getInstance(IF)Ljava/awt/AlphaComposite;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AlphaComposite) => void): void;
    public "createContext(Ljava/awt/image/ColorModel;Ljava/awt/image/ColorModel;Ljava/awt/RenderingHints;)Ljava/awt/CompositeContext;"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_image_ColorModel, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable, rv?: java_awt_CompositeContext) => void): void;
    public "java/awt/AlphaComposite/createContext(Ljava/awt/image/ColorModel;Ljava/awt/image/ColorModel;Ljava/awt/RenderingHints;)Ljava/awt/CompositeContext;"(thread: JVMThread, args: [java_awt_image_ColorModel, java_awt_image_ColorModel, java_awt_RenderingHints], cb?: (e?: java_lang_Throwable, rv?: java_awt_CompositeContext) => void): void;
    public "getAlpha()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/AlphaComposite/getAlpha()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRule()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/AlphaComposite/getRule()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "derive(I)Ljava/awt/AlphaComposite;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AlphaComposite) => void): void;
    public "java/awt/AlphaComposite/derive(I)Ljava/awt/AlphaComposite;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AlphaComposite) => void): void;
    public "derive(F)Ljava/awt/AlphaComposite;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AlphaComposite) => void): void;
    public "java/awt/AlphaComposite/derive(F)Ljava/awt/AlphaComposite;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_AlphaComposite) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/AlphaComposite/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/AlphaComposite/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/AlphaComposite/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_StateTrackable$State extends java_lang_Enum {
    public static "sun/java2d/StateTrackable$State/IMMUTABLE": sun_java2d_StateTrackable$State;
    public static "sun/java2d/StateTrackable$State/STABLE": sun_java2d_StateTrackable$State;
    public static "sun/java2d/StateTrackable$State/DYNAMIC": sun_java2d_StateTrackable$State;
    public static "sun/java2d/StateTrackable$State/UNTRACKABLE": sun_java2d_StateTrackable$State;
    public static "sun/java2d/StateTrackable$State/$VALUES": JVMArray<sun_java2d_StateTrackable$State>;
    public static "values()[Lsun/java2d/StateTrackable$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_java2d_StateTrackable$State>) => void): void;
    public static "sun/java2d/StateTrackable$State/values()[Lsun/java2d/StateTrackable$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_java2d_StateTrackable$State>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lsun/java2d/StateTrackable$State;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTrackable$State) => void): void;
    public static "sun/java2d/StateTrackable$State/valueOf(Ljava/lang/String;)Lsun/java2d/StateTrackable$State;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTrackable$State) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/StateTrackable$State/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/StateTrackable$State/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_RenderCache extends java_lang_Object {
    public "sun/java2d/loops/RenderCache/entries": JVMArray<sun_java2d_loops_RenderCache$Entry>;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/RenderCache/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/loops/RenderCache/get(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Ljava/lang/Object;"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "put(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/RenderCache/put(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_RenderCache$Entry extends java_lang_Object {
    public "sun/java2d/loops/RenderCache$Entry/src": sun_java2d_loops_SurfaceType;
    public "sun/java2d/loops/RenderCache$Entry/comp": sun_java2d_loops_CompositeType;
    public "sun/java2d/loops/RenderCache$Entry/dst": sun_java2d_loops_SurfaceType;
    public "sun/java2d/loops/RenderCache$Entry/value": java_lang_Object;
    public "sun/java2d/loops/RenderCache$Entry/this$0": sun_java2d_loops_RenderCache;
    public "<init>(Lsun/java2d/loops/RenderCache;Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_java2d_loops_RenderCache, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/RenderCache$Entry/<init>(Lsun/java2d/loops/RenderCache;Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;Ljava/lang/Object;)V"(thread: JVMThread, args: [sun_java2d_loops_RenderCache, sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "matches(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Z"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/RenderCache$Entry/matches(Lsun/java2d/loops/SurfaceType;Lsun/java2d/loops/CompositeType;Lsun/java2d/loops/SurfaceType;)Z"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, sun_java2d_loops_CompositeType, sun_java2d_loops_SurfaceType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/loops/RenderCache$Entry/getValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface sun_java2d_pipe_DrawImagePipe extends java_lang_Object {
    "copyImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "copyImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "scaleImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "scaleImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;IIIIIIIILjava/awt/Color;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, number, number, number, number, number, number, number, number, java_awt_Color, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "transformImage(Lsun/java2d/SunGraphics2D;Ljava/awt/Image;Ljava/awt/geom/AffineTransform;Ljava/awt/image/ImageObserver;)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Image, java_awt_geom_AffineTransform, java_awt_image_ImageObserver], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "transformImage(Lsun/java2d/SunGraphics2D;Ljava/awt/image/BufferedImage;Ljava/awt/image/BufferedImageOp;II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_image_BufferedImage, java_awt_image_BufferedImageOp, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_pipe_SpanShapeRenderer extends java_lang_Object implements sun_java2d_pipe_ShapeDrawPipe {
    public static "sun/java2d/pipe/SpanShapeRenderer/RenderEngine": sun_java2d_pipe_RenderingEngine;
    public static "sun/java2d/pipe/SpanShapeRenderer/NON_RECTILINEAR_TRANSFORM_MASK": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/SpanShapeRenderer/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/SpanShapeRenderer/draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/SpanShapeRenderer/fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "startSequence(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;Ljava/awt/Rectangle;[I)Ljava/lang/Object;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape, java_awt_Rectangle, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/pipe/SpanShapeRenderer/startSequence(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;Ljava/awt/Rectangle;[I)Ljava/lang/Object;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape, java_awt_Rectangle, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "renderBox(Ljava/lang/Object;IIII)V"(thread: JVMThread, args: [java_lang_Object, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/SpanShapeRenderer/renderBox(Ljava/lang/Object;IIII)V"(thread: JVMThread, args: [java_lang_Object, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "endSequence(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/SpanShapeRenderer/endSequence(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "renderRect(Lsun/java2d/SunGraphics2D;Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/SpanShapeRenderer/renderRect(Lsun/java2d/SunGraphics2D;Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "renderSpans(Lsun/java2d/SunGraphics2D;Lsun/java2d/pipe/Region;Ljava/awt/Shape;Lsun/java2d/pipe/ShapeSpanIterator;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_pipe_Region, java_awt_Shape, sun_java2d_pipe_ShapeSpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/SpanShapeRenderer/renderSpans(Lsun/java2d/SunGraphics2D;Lsun/java2d/pipe/Region;Ljava/awt/Shape;Lsun/java2d/pipe/ShapeSpanIterator;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_pipe_Region, java_awt_Shape, sun_java2d_pipe_ShapeSpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "spanClipLoop(Ljava/lang/Object;Lsun/java2d/pipe/SpanIterator;Lsun/java2d/pipe/Region;[I)V"(thread: JVMThread, args: [java_lang_Object, sun_java2d_pipe_SpanIterator, sun_java2d_pipe_Region, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/SpanShapeRenderer/spanClipLoop(Ljava/lang/Object;Lsun/java2d/pipe/SpanIterator;Lsun/java2d/pipe/Region;[I)V"(thread: JVMThread, args: [java_lang_Object, sun_java2d_pipe_SpanIterator, sun_java2d_pipe_Region, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/SpanShapeRenderer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_pipe_ShapeSpanIterator extends java_lang_Object implements sun_java2d_pipe_SpanIterator, sun_awt_geom_PathConsumer2D {
    public "sun/java2d/pipe/ShapeSpanIterator/pData": Long;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/ShapeSpanIterator/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "appendPath(Ljava/awt/geom/PathIterator;)V"(thread: JVMThread, args: [java_awt_geom_PathIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/appendPath(Ljava/awt/geom/PathIterator;)V"(thread: JVMThread, args: [java_awt_geom_PathIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "appendPoly([I[IIII)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/appendPoly([I[IIII)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setNormalize(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/setNormalize(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOutputAreaXYWH(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/setOutputAreaXYWH(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOutputAreaXYXY(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/setOutputAreaXYXY(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOutputArea(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/setOutputArea(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOutputArea(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/setOutputArea(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRule(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/setRule(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addSegment(I[F)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/addSegment(I[F)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPathBox([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/getPathBox([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "intersectClipBox(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/intersectClipBox(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "nextSpan([I)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/nextSpan([I)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skipDownTo(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/skipDownTo(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNativeIterator()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/getNativeIterator()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "moveTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/moveTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "lineTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/lineTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "quadTo(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/quadTo(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "curveTo(FFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/curveTo(FFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "closePath()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/closePath()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "pathDone()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/pathDone()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNativeConsumer()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/java2d/pipe/ShapeSpanIterator/getNativeConsumer()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/ShapeSpanIterator/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_awt_geom_PathConsumer2D extends java_lang_Object {
    "moveTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "lineTo(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "quadTo(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "curveTo(FFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "closePath()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "pathDone()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getNativeConsumer()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class sun_java2d_pipe_RenderingEngine extends java_lang_Object {
    public static "sun/java2d/pipe/RenderingEngine/reImpl": sun_java2d_pipe_RenderingEngine;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/RenderingEngine/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance()Lsun/java2d/pipe/RenderingEngine;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_RenderingEngine) => void): void;
    public static "sun/java2d/pipe/RenderingEngine/getInstance()Lsun/java2d/pipe/RenderingEngine;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_RenderingEngine) => void): void;
    public "createStrokedShape(Ljava/awt/Shape;FIIF[FF)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape, number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/java2d/pipe/RenderingEngine/createStrokedShape(Ljava/awt/Shape;FIIF[FF)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_Shape, number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "strokeTo(Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;Ljava/awt/BasicStroke;ZZZLsun/awt/geom/PathConsumer2D;)V"(thread: JVMThread, args: [java_awt_Shape, java_awt_geom_AffineTransform, java_awt_BasicStroke, number, number, number, sun_awt_geom_PathConsumer2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/RenderingEngine/strokeTo(Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;Ljava/awt/BasicStroke;ZZZLsun/awt/geom/PathConsumer2D;)V"(thread: JVMThread, args: [java_awt_Shape, java_awt_geom_AffineTransform, java_awt_BasicStroke, number, number, number, sun_awt_geom_PathConsumer2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAATileGenerator(Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;Lsun/java2d/pipe/Region;Ljava/awt/BasicStroke;ZZ[I)Lsun/java2d/pipe/AATileGenerator;"(thread: JVMThread, args: [java_awt_Shape, java_awt_geom_AffineTransform, sun_java2d_pipe_Region, java_awt_BasicStroke, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_AATileGenerator) => void): void;
    public "sun/java2d/pipe/RenderingEngine/getAATileGenerator(Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;Lsun/java2d/pipe/Region;Ljava/awt/BasicStroke;ZZ[I)Lsun/java2d/pipe/AATileGenerator;"(thread: JVMThread, args: [java_awt_Shape, java_awt_geom_AffineTransform, sun_java2d_pipe_Region, java_awt_BasicStroke, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_AATileGenerator) => void): void;
    public "getAATileGenerator(DDDDDDDDLsun/java2d/pipe/Region;[I)Lsun/java2d/pipe/AATileGenerator;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, sun_java2d_pipe_Region, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_AATileGenerator) => void): void;
    public "sun/java2d/pipe/RenderingEngine/getAATileGenerator(DDDDDDDDLsun/java2d/pipe/Region;[I)Lsun/java2d/pipe/AATileGenerator;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, sun_java2d_pipe_Region, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_AATileGenerator) => void): void;
    public "getMinimumAAPenSize()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/RenderingEngine/getMinimumAAPenSize()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "feedConsumer(Ljava/awt/geom/PathIterator;Lsun/awt/geom/PathConsumer2D;)V"(thread: JVMThread, args: [java_awt_geom_PathIterator, sun_awt_geom_PathConsumer2D], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/RenderingEngine/feedConsumer(Ljava/awt/geom/PathIterator;Lsun/awt/geom/PathConsumer2D;)V"(thread: JVMThread, args: [java_awt_geom_PathIterator, sun_awt_geom_PathConsumer2D], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_pipe_AATileGenerator extends java_lang_Object {
    "getTileWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTileHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getTypicalAlpha()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "nextTile()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getAlpha([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_pipe_AAShapePipe extends java_lang_Object implements sun_java2d_pipe_ShapeDrawPipe, sun_java2d_pipe_ParallelogramPipe {
    public static "sun/java2d/pipe/AAShapePipe/renderengine": sun_java2d_pipe_RenderingEngine;
    public "sun/java2d/pipe/AAShapePipe/outpipe": sun_java2d_pipe_CompositePipe;
    public static "sun/java2d/pipe/AAShapePipe/theTile": JVMArray<number>;
    public "<init>(Lsun/java2d/pipe/CompositePipe;)V"(thread: JVMThread, args: [sun_java2d_pipe_CompositePipe], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AAShapePipe/<init>(Lsun/java2d/pipe/CompositePipe;)V"(thread: JVMThread, args: [sun_java2d_pipe_CompositePipe], cb?: (e?: java_lang_Throwable) => void): void;
    public "draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AAShapePipe/draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AAShapePipe/fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public static "computeBBox(DDDD)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public static "sun/java2d/pipe/AAShapePipe/computeBBox(DDDD)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "fillParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AAShapePipe/fillParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AAShapePipe/drawParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getAlphaTile(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/java2d/pipe/AAShapePipe/getAlphaTile(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "dropAlphaTile([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/AAShapePipe/dropAlphaTile([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "renderPath(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;Ljava/awt/BasicStroke;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape, java_awt_BasicStroke], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AAShapePipe/renderPath(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;Ljava/awt/BasicStroke;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape, java_awt_BasicStroke], cb?: (e?: java_lang_Throwable) => void): void;
    public "renderTiles(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;Lsun/java2d/pipe/AATileGenerator;[I)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape, sun_java2d_pipe_AATileGenerator, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AAShapePipe/renderTiles(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;Lsun/java2d/pipe/AATileGenerator;[I)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape, sun_java2d_pipe_AATileGenerator, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/AAShapePipe/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_pipe_CompositePipe extends java_lang_Object {
    "startSequence(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;Ljava/awt/Rectangle;[I)Ljava/lang/Object;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape, java_awt_Rectangle, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "needTile(Ljava/lang/Object;IIII)Z"(thread: JVMThread, args: [java_lang_Object, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "renderPathTile(Ljava/lang/Object;[BIIIIII)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "skipTile(Ljava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "endSequence(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_pipe_PixelToParallelogramConverter extends sun_java2d_pipe_PixelToShapeConverter implements sun_java2d_pipe_ShapeDrawPipe {
    public "sun/java2d/pipe/PixelToParallelogramConverter/outrenderer": sun_java2d_pipe_ParallelogramPipe;
    public "sun/java2d/pipe/PixelToParallelogramConverter/minPenSize": number;
    public "sun/java2d/pipe/PixelToParallelogramConverter/normPosition": number;
    public "sun/java2d/pipe/PixelToParallelogramConverter/normRoundingBias": number;
    public "sun/java2d/pipe/PixelToParallelogramConverter/adjustfill": number;
    public "<init>(Lsun/java2d/pipe/ShapeDrawPipe;Lsun/java2d/pipe/ParallelogramPipe;DDZ)V"(thread: JVMThread, args: [sun_java2d_pipe_ShapeDrawPipe, sun_java2d_pipe_ParallelogramPipe, number, any, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/<init>(Lsun/java2d/pipe/ShapeDrawPipe;Lsun/java2d/pipe/ParallelogramPipe;DDZ)V"(thread: JVMThread, args: [sun_java2d_pipe_ShapeDrawPipe, sun_java2d_pipe_ParallelogramPipe, number, any, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawLine(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/drawLine(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/drawRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/fillRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public static "len(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/pipe/PixelToParallelogramConverter/len(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "normalize(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/normalize(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "drawGeneralLine(Lsun/java2d/SunGraphics2D;DDDD)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/drawGeneralLine(Lsun/java2d/SunGraphics2D;DDDD)Z"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fillRectangle(Lsun/java2d/SunGraphics2D;DDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/fillRectangle(Lsun/java2d/SunGraphics2D;DDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRectangle(Lsun/java2d/SunGraphics2D;DDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/drawRectangle(Lsun/java2d/SunGraphics2D;DDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillOuterParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToParallelogramConverter/fillOuterParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_pipe_PixelToShapeConverter extends java_lang_Object implements sun_java2d_pipe_PixelDrawPipe, sun_java2d_pipe_PixelFillPipe {
    public "sun/java2d/pipe/PixelToShapeConverter/outpipe": sun_java2d_pipe_ShapeDrawPipe;
    public "<init>(Lsun/java2d/pipe/ShapeDrawPipe;)V"(thread: JVMThread, args: [sun_java2d_pipe_ShapeDrawPipe], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/<init>(Lsun/java2d/pipe/ShapeDrawPipe;)V"(thread: JVMThread, args: [sun_java2d_pipe_ShapeDrawPipe], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawLine(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/drawLine(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/drawRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/fillRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/drawRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/fillRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/drawOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/fillOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/drawArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/fillArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "makePoly([I[IIZ)Ljava/awt/Shape;"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/makePoly([I[IIZ)Ljava/awt/Shape;"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "drawPolyline(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/drawPolyline(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/drawPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/PixelToShapeConverter/fillPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_pipe_AlphaColorPipe extends java_lang_Object implements sun_java2d_pipe_CompositePipe, sun_java2d_pipe_ParallelogramPipe {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AlphaColorPipe/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "startSequence(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;Ljava/awt/Rectangle;[I)Ljava/lang/Object;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape, java_awt_Rectangle, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/pipe/AlphaColorPipe/startSequence(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;Ljava/awt/Rectangle;[I)Ljava/lang/Object;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape, java_awt_Rectangle, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "needTile(Ljava/lang/Object;IIII)Z"(thread: JVMThread, args: [java_lang_Object, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/AlphaColorPipe/needTile(Ljava/lang/Object;IIII)Z"(thread: JVMThread, args: [java_lang_Object, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "renderPathTile(Ljava/lang/Object;[BIIIIII)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AlphaColorPipe/renderPathTile(Ljava/lang/Object;[BIIIIII)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "skipTile(Ljava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AlphaColorPipe/skipTile(Ljava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "endSequence(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AlphaColorPipe/endSequence(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AlphaColorPipe/fillParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/AlphaColorPipe/drawParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_pipe_TextPipe extends java_lang_Object {
    "drawString(Lsun/java2d/SunGraphics2D;Ljava/lang/String;DD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_lang_String, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    "drawGlyphVector(Lsun/java2d/SunGraphics2D;Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "drawChars(Lsun/java2d/SunGraphics2D;[CIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_pipe_LoopPipe extends java_lang_Object implements sun_java2d_pipe_PixelDrawPipe, sun_java2d_pipe_PixelFillPipe, sun_java2d_pipe_ParallelogramPipe, sun_java2d_pipe_ShapeDrawPipe, sun_java2d_pipe_LoopBasedPipe {
    public static "sun/java2d/pipe/LoopPipe/RenderEngine": sun_java2d_pipe_RenderingEngine;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawLine(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/drawLine(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/drawRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/drawRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/drawOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/drawArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolyline(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/drawPolyline(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/drawPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/fillRect(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/fillRoundRect(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/fillOval(Lsun/java2d/SunGraphics2D;IIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/fillArc(Lsun/java2d/SunGraphics2D;IIIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/fillPolygon(Lsun/java2d/SunGraphics2D;[I[II)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/draw(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getFillSSI(Lsun/java2d/SunGraphics2D;)Lsun/java2d/pipe/ShapeSpanIterator;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_ShapeSpanIterator) => void): void;
    public static "sun/java2d/pipe/LoopPipe/getFillSSI(Lsun/java2d/SunGraphics2D;)Lsun/java2d/pipe/ShapeSpanIterator;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_ShapeSpanIterator) => void): void;
    public static "getStrokeSpans(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)Lsun/java2d/pipe/ShapeSpanIterator;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_ShapeSpanIterator) => void): void;
    public static "sun/java2d/pipe/LoopPipe/getStrokeSpans(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)Lsun/java2d/pipe/ShapeSpanIterator;"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_ShapeSpanIterator) => void): void;
    public "fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/fill(Lsun/java2d/SunGraphics2D;Ljava/awt/Shape;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public static "fillSpans(Lsun/java2d/SunGraphics2D;Lsun/java2d/pipe/SpanIterator;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/LoopPipe/fillSpans(Lsun/java2d/SunGraphics2D;Lsun/java2d/pipe/SpanIterator;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/fillParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/LoopPipe/drawParallelogram(Lsun/java2d/SunGraphics2D;DDDDDDDDDDDD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/LoopPipe/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_pipe_LoopBasedPipe extends java_lang_Object {
  }
  export class sun_java2d_StateTrackableDelegate extends java_lang_Object implements sun_java2d_StateTrackable {
    public static "sun/java2d/StateTrackableDelegate/UNTRACKABLE_DELEGATE": sun_java2d_StateTrackableDelegate;
    public static "sun/java2d/StateTrackableDelegate/IMMUTABLE_DELEGATE": sun_java2d_StateTrackableDelegate;
    public "sun/java2d/StateTrackableDelegate/theState": sun_java2d_StateTrackable$State;
    public "sun/java2d/StateTrackableDelegate/theTracker": sun_java2d_StateTracker;
    public "sun/java2d/StateTrackableDelegate/numDynamicAgents": number;
    public static "createInstance(Lsun/java2d/StateTrackable$State;)Lsun/java2d/StateTrackableDelegate;"(thread: JVMThread, args: [sun_java2d_StateTrackable$State], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTrackableDelegate) => void): void;
    public static "sun/java2d/StateTrackableDelegate/createInstance(Lsun/java2d/StateTrackable$State;)Lsun/java2d/StateTrackableDelegate;"(thread: JVMThread, args: [sun_java2d_StateTrackable$State], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTrackableDelegate) => void): void;
    public "<init>(Lsun/java2d/StateTrackable$State;)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/StateTrackableDelegate/<init>(Lsun/java2d/StateTrackable$State;)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State], cb?: (e?: java_lang_Throwable) => void): void;
    public "getState()Lsun/java2d/StateTrackable$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTrackable$State) => void): void;
    public "sun/java2d/StateTrackableDelegate/getState()Lsun/java2d/StateTrackable$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTrackable$State) => void): void;
    public "getStateTracker()Lsun/java2d/StateTracker;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTracker) => void): void;
    public "sun/java2d/StateTrackableDelegate/getStateTracker()Lsun/java2d/StateTracker;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTracker) => void): void;
    public "setImmutable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/StateTrackableDelegate/setImmutable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setUntrackable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/StateTrackableDelegate/setUntrackable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addDynamicAgent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/StateTrackableDelegate/addDynamicAgent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeDynamicAgent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/StateTrackableDelegate/removeDynamicAgent()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "markDirty()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/StateTrackableDelegate/markDirty()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/StateTrackableDelegate/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_loops_SurfaceType extends java_lang_Object {
    public static "sun/java2d/loops/SurfaceType/unusedUID": number;
    public static "sun/java2d/loops/SurfaceType/surfaceUIDMap": java_util_HashMap;
    public static "sun/java2d/loops/SurfaceType/DESC_ANY": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_INT_RGB": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_INT_ARGB": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_INT_ARGB_PRE": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_INT_BGR": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_3BYTE_BGR": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_4BYTE_ABGR": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_4BYTE_ABGR_PRE": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_USHORT_565_RGB": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_USHORT_555_RGB": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_USHORT_555_RGBx": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_USHORT_4444_ARGB": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_BYTE_GRAY": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_USHORT_INDEXED": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_USHORT_GRAY": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_BYTE_BINARY": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_BYTE_INDEXED": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_ANY_INT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_ANY_SHORT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_ANY_BYTE": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_ANY_3BYTE": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_ANY_4BYTE": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_ANY_INT_DCM": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_INT_RGBx": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_INT_BGRx": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_3BYTE_RGB": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_INT_ARGB_BM": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_BYTE_INDEXED_BM": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_BYTE_INDEXED_OPAQUE": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_INDEX8_GRAY": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_INDEX12_GRAY": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_BYTE_BINARY_1BIT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_BYTE_BINARY_2BIT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_BYTE_BINARY_4BIT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_ANY_PAINT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_ANY_COLOR": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_OPAQUE_COLOR": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_GRADIENT_PAINT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_OPAQUE_GRADIENT_PAINT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_TEXTURE_PAINT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_OPAQUE_TEXTURE_PAINT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_LINEAR_GRADIENT_PAINT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_OPAQUE_LINEAR_GRADIENT_PAINT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_RADIAL_GRADIENT_PAINT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/DESC_OPAQUE_RADIAL_GRADIENT_PAINT": java_lang_String;
    public static "sun/java2d/loops/SurfaceType/Any": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/AnyInt": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/AnyShort": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/AnyByte": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/AnyByteBinary": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/Any3Byte": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/Any4Byte": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/AnyDcm": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/Custom": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/IntRgb": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/IntArgb": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/IntArgbPre": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/IntBgr": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/ThreeByteBgr": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/FourByteAbgr": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/FourByteAbgrPre": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/Ushort565Rgb": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/Ushort555Rgb": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/Ushort555Rgbx": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/Ushort4444Argb": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/UshortIndexed": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/ByteGray": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/UshortGray": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/ByteBinary1Bit": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/ByteBinary2Bit": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/ByteBinary4Bit": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/ByteIndexed": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/IntRgbx": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/IntBgrx": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/ThreeByteRgb": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/IntArgbBm": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/ByteIndexedBm": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/ByteIndexedOpaque": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/Index8Gray": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/Index12Gray": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/AnyPaint": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/AnyColor": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/OpaqueColor": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/GradientPaint": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/OpaqueGradientPaint": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/LinearGradientPaint": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/OpaqueLinearGradientPaint": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/RadialGradientPaint": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/OpaqueRadialGradientPaint": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/TexturePaint": sun_java2d_loops_SurfaceType;
    public static "sun/java2d/loops/SurfaceType/OpaqueTexturePaint": sun_java2d_loops_SurfaceType;
    public "sun/java2d/loops/SurfaceType/uniqueID": number;
    public "sun/java2d/loops/SurfaceType/desc": java_lang_String;
    public "sun/java2d/loops/SurfaceType/next": sun_java2d_loops_SurfaceType;
    public "sun/java2d/loops/SurfaceType/pixelConverter": sun_awt_image_PixelConverter;
    public "deriveSubType(Ljava/lang/String;)Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "sun/java2d/loops/SurfaceType/deriveSubType(Ljava/lang/String;)Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "deriveSubType(Ljava/lang/String;Lsun/awt/image/PixelConverter;)Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: [java_lang_String, sun_awt_image_PixelConverter], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "sun/java2d/loops/SurfaceType/deriveSubType(Ljava/lang/String;Lsun/awt/image/PixelConverter;)Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: [java_lang_String, sun_awt_image_PixelConverter], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Ljava/lang/String;Lsun/awt/image/PixelConverter;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, java_lang_String, sun_awt_image_PixelConverter], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/SurfaceType/<init>(Lsun/java2d/loops/SurfaceType;Ljava/lang/String;Lsun/awt/image/PixelConverter;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, java_lang_String, sun_awt_image_PixelConverter], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/loops/SurfaceType;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/SurfaceType/<init>(Lsun/java2d/loops/SurfaceType;Ljava/lang/String;)V"(thread: JVMThread, args: [sun_java2d_loops_SurfaceType, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "makeUniqueID(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/loops/SurfaceType/makeUniqueID(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUniqueID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/SurfaceType/getUniqueID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescriptor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/java2d/loops/SurfaceType/getDescriptor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSuperType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "sun/java2d/loops/SurfaceType/getSuperType()Lsun/java2d/loops/SurfaceType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_loops_SurfaceType) => void): void;
    public "getPixelConverter()Lsun/awt/image/PixelConverter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_image_PixelConverter) => void): void;
    public "sun/java2d/loops/SurfaceType/getPixelConverter()Lsun/awt/image/PixelConverter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_image_PixelConverter) => void): void;
    public "pixelFor(ILjava/awt/image/ColorModel;)I"(thread: JVMThread, args: [number, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/SurfaceType/pixelFor(ILjava/awt/image/ColorModel;)I"(thread: JVMThread, args: [number, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rgbFor(ILjava/awt/image/ColorModel;)I"(thread: JVMThread, args: [number, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/SurfaceType/rgbFor(ILjava/awt/image/ColorModel;)I"(thread: JVMThread, args: [number, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAlphaMask()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/SurfaceType/getAlphaMask()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/SurfaceType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/loops/SurfaceType/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/java2d/loops/SurfaceType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/loops/SurfaceType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_awt_image_PixelConverter extends java_lang_Object {
    public static "sun/awt/image/PixelConverter/instance": sun_awt_image_PixelConverter;
    public "sun/awt/image/PixelConverter/alphaMask": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/PixelConverter/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "rgbToPixel(ILjava/awt/image/ColorModel;)I"(thread: JVMThread, args: [number, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/image/PixelConverter/rgbToPixel(ILjava/awt/image/ColorModel;)I"(thread: JVMThread, args: [number, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "pixelToRgb(ILjava/awt/image/ColorModel;)I"(thread: JVMThread, args: [number, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/image/PixelConverter/pixelToRgb(ILjava/awt/image/ColorModel;)I"(thread: JVMThread, args: [number, java_awt_image_ColorModel], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAlphaMask()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/image/PixelConverter/getAlphaMask()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/image/PixelConverter/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_Surface extends java_lang_Object {
  }
  export interface sun_java2d_StateTrackable extends java_lang_Object {
    "getState()Lsun/java2d/StateTrackable$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTrackable$State) => void): void;
    "getStateTracker()Lsun/java2d/StateTracker;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_StateTracker) => void): void;
  }
  export interface sun_java2d_DisposerTarget extends java_lang_Object {
    "getDisposerReferent()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_awt_image_SurfaceManager$ImageAccessor extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/SurfaceManager$ImageAccessor/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSurfaceManager(Ljava/awt/Image;)Lsun/awt/image/SurfaceManager;"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: sun_awt_image_SurfaceManager) => void): void;
    public "sun/awt/image/SurfaceManager$ImageAccessor/getSurfaceManager(Ljava/awt/Image;)Lsun/awt/image/SurfaceManager;"(thread: JVMThread, args: [java_awt_Image], cb?: (e?: java_lang_Throwable, rv?: sun_awt_image_SurfaceManager) => void): void;
    public "setSurfaceManager(Ljava/awt/Image;Lsun/awt/image/SurfaceManager;)V"(thread: JVMThread, args: [java_awt_Image, sun_awt_image_SurfaceManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/image/SurfaceManager$ImageAccessor/setSurfaceManager(Ljava/awt/Image;Lsun/awt/image/SurfaceManager;)V"(thread: JVMThread, args: [java_awt_Image, sun_awt_image_SurfaceManager], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_color_ICC__Profile extends java_lang_Object implements java_io_Serializable {
    public static "java/awt/color/ICC_Profile/serialVersionUID": Long;
    public "java/awt/color/ICC_Profile/cmmProfile": sun_java2d_cmm_Profile;
    public "java/awt/color/ICC_Profile/deferralInfo": sun_java2d_cmm_ProfileDeferralInfo;
    public "java/awt/color/ICC_Profile/profileActivator": sun_java2d_cmm_ProfileActivator;
    public static "java/awt/color/ICC_Profile/sRGBprofile": java_awt_color_ICC__Profile;
    public static "java/awt/color/ICC_Profile/XYZprofile": java_awt_color_ICC__Profile;
    public static "java/awt/color/ICC_Profile/PYCCprofile": java_awt_color_ICC__Profile;
    public static "java/awt/color/ICC_Profile/GRAYprofile": java_awt_color_ICC__Profile;
    public static "java/awt/color/ICC_Profile/LINEAR_RGBprofile": java_awt_color_ICC__Profile;
    public static "java/awt/color/ICC_Profile/CLASS_INPUT": number;
    public static "java/awt/color/ICC_Profile/CLASS_DISPLAY": number;
    public static "java/awt/color/ICC_Profile/CLASS_OUTPUT": number;
    public static "java/awt/color/ICC_Profile/CLASS_DEVICELINK": number;
    public static "java/awt/color/ICC_Profile/CLASS_COLORSPACECONVERSION": number;
    public static "java/awt/color/ICC_Profile/CLASS_ABSTRACT": number;
    public static "java/awt/color/ICC_Profile/CLASS_NAMEDCOLOR": number;
    public static "java/awt/color/ICC_Profile/icSigXYZData": number;
    public static "java/awt/color/ICC_Profile/icSigLabData": number;
    public static "java/awt/color/ICC_Profile/icSigLuvData": number;
    public static "java/awt/color/ICC_Profile/icSigYCbCrData": number;
    public static "java/awt/color/ICC_Profile/icSigYxyData": number;
    public static "java/awt/color/ICC_Profile/icSigRgbData": number;
    public static "java/awt/color/ICC_Profile/icSigGrayData": number;
    public static "java/awt/color/ICC_Profile/icSigHsvData": number;
    public static "java/awt/color/ICC_Profile/icSigHlsData": number;
    public static "java/awt/color/ICC_Profile/icSigCmykData": number;
    public static "java/awt/color/ICC_Profile/icSigCmyData": number;
    public static "java/awt/color/ICC_Profile/icSigSpace2CLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpace3CLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpace4CLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpace5CLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpace6CLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpace7CLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpace8CLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpace9CLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpaceACLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpaceBCLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpaceCCLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpaceDCLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpaceECLR": number;
    public static "java/awt/color/ICC_Profile/icSigSpaceFCLR": number;
    public static "java/awt/color/ICC_Profile/icSigInputClass": number;
    public static "java/awt/color/ICC_Profile/icSigDisplayClass": number;
    public static "java/awt/color/ICC_Profile/icSigOutputClass": number;
    public static "java/awt/color/ICC_Profile/icSigLinkClass": number;
    public static "java/awt/color/ICC_Profile/icSigAbstractClass": number;
    public static "java/awt/color/ICC_Profile/icSigColorSpaceClass": number;
    public static "java/awt/color/ICC_Profile/icSigNamedColorClass": number;
    public static "java/awt/color/ICC_Profile/icPerceptual": number;
    public static "java/awt/color/ICC_Profile/icRelativeColorimetric": number;
    public static "java/awt/color/ICC_Profile/icMediaRelativeColorimetric": number;
    public static "java/awt/color/ICC_Profile/icSaturation": number;
    public static "java/awt/color/ICC_Profile/icAbsoluteColorimetric": number;
    public static "java/awt/color/ICC_Profile/icICCAbsoluteColorimetric": number;
    public static "java/awt/color/ICC_Profile/icSigHead": number;
    public static "java/awt/color/ICC_Profile/icSigAToB0Tag": number;
    public static "java/awt/color/ICC_Profile/icSigAToB1Tag": number;
    public static "java/awt/color/ICC_Profile/icSigAToB2Tag": number;
    public static "java/awt/color/ICC_Profile/icSigBlueColorantTag": number;
    public static "java/awt/color/ICC_Profile/icSigBlueMatrixColumnTag": number;
    public static "java/awt/color/ICC_Profile/icSigBlueTRCTag": number;
    public static "java/awt/color/ICC_Profile/icSigBToA0Tag": number;
    public static "java/awt/color/ICC_Profile/icSigBToA1Tag": number;
    public static "java/awt/color/ICC_Profile/icSigBToA2Tag": number;
    public static "java/awt/color/ICC_Profile/icSigCalibrationDateTimeTag": number;
    public static "java/awt/color/ICC_Profile/icSigCharTargetTag": number;
    public static "java/awt/color/ICC_Profile/icSigCopyrightTag": number;
    public static "java/awt/color/ICC_Profile/icSigCrdInfoTag": number;
    public static "java/awt/color/ICC_Profile/icSigDeviceMfgDescTag": number;
    public static "java/awt/color/ICC_Profile/icSigDeviceModelDescTag": number;
    public static "java/awt/color/ICC_Profile/icSigDeviceSettingsTag": number;
    public static "java/awt/color/ICC_Profile/icSigGamutTag": number;
    public static "java/awt/color/ICC_Profile/icSigGrayTRCTag": number;
    public static "java/awt/color/ICC_Profile/icSigGreenColorantTag": number;
    public static "java/awt/color/ICC_Profile/icSigGreenMatrixColumnTag": number;
    public static "java/awt/color/ICC_Profile/icSigGreenTRCTag": number;
    public static "java/awt/color/ICC_Profile/icSigLuminanceTag": number;
    public static "java/awt/color/ICC_Profile/icSigMeasurementTag": number;
    public static "java/awt/color/ICC_Profile/icSigMediaBlackPointTag": number;
    public static "java/awt/color/ICC_Profile/icSigMediaWhitePointTag": number;
    public static "java/awt/color/ICC_Profile/icSigNamedColor2Tag": number;
    public static "java/awt/color/ICC_Profile/icSigOutputResponseTag": number;
    public static "java/awt/color/ICC_Profile/icSigPreview0Tag": number;
    public static "java/awt/color/ICC_Profile/icSigPreview1Tag": number;
    public static "java/awt/color/ICC_Profile/icSigPreview2Tag": number;
    public static "java/awt/color/ICC_Profile/icSigProfileDescriptionTag": number;
    public static "java/awt/color/ICC_Profile/icSigProfileSequenceDescTag": number;
    public static "java/awt/color/ICC_Profile/icSigPs2CRD0Tag": number;
    public static "java/awt/color/ICC_Profile/icSigPs2CRD1Tag": number;
    public static "java/awt/color/ICC_Profile/icSigPs2CRD2Tag": number;
    public static "java/awt/color/ICC_Profile/icSigPs2CRD3Tag": number;
    public static "java/awt/color/ICC_Profile/icSigPs2CSATag": number;
    public static "java/awt/color/ICC_Profile/icSigPs2RenderingIntentTag": number;
    public static "java/awt/color/ICC_Profile/icSigRedColorantTag": number;
    public static "java/awt/color/ICC_Profile/icSigRedMatrixColumnTag": number;
    public static "java/awt/color/ICC_Profile/icSigRedTRCTag": number;
    public static "java/awt/color/ICC_Profile/icSigScreeningDescTag": number;
    public static "java/awt/color/ICC_Profile/icSigScreeningTag": number;
    public static "java/awt/color/ICC_Profile/icSigTechnologyTag": number;
    public static "java/awt/color/ICC_Profile/icSigUcrBgTag": number;
    public static "java/awt/color/ICC_Profile/icSigViewingCondDescTag": number;
    public static "java/awt/color/ICC_Profile/icSigViewingConditionsTag": number;
    public static "java/awt/color/ICC_Profile/icSigChromaticityTag": number;
    public static "java/awt/color/ICC_Profile/icSigChromaticAdaptationTag": number;
    public static "java/awt/color/ICC_Profile/icSigColorantOrderTag": number;
    public static "java/awt/color/ICC_Profile/icSigColorantTableTag": number;
    public static "java/awt/color/ICC_Profile/icHdrSize": number;
    public static "java/awt/color/ICC_Profile/icHdrCmmId": number;
    public static "java/awt/color/ICC_Profile/icHdrVersion": number;
    public static "java/awt/color/ICC_Profile/icHdrDeviceClass": number;
    public static "java/awt/color/ICC_Profile/icHdrColorSpace": number;
    public static "java/awt/color/ICC_Profile/icHdrPcs": number;
    public static "java/awt/color/ICC_Profile/icHdrDate": number;
    public static "java/awt/color/ICC_Profile/icHdrMagic": number;
    public static "java/awt/color/ICC_Profile/icHdrPlatform": number;
    public static "java/awt/color/ICC_Profile/icHdrFlags": number;
    public static "java/awt/color/ICC_Profile/icHdrManufacturer": number;
    public static "java/awt/color/ICC_Profile/icHdrModel": number;
    public static "java/awt/color/ICC_Profile/icHdrAttributes": number;
    public static "java/awt/color/ICC_Profile/icHdrRenderingIntent": number;
    public static "java/awt/color/ICC_Profile/icHdrIlluminant": number;
    public static "java/awt/color/ICC_Profile/icHdrCreator": number;
    public static "java/awt/color/ICC_Profile/icHdrProfileID": number;
    public static "java/awt/color/ICC_Profile/icTagType": number;
    public static "java/awt/color/ICC_Profile/icTagReserved": number;
    public static "java/awt/color/ICC_Profile/icCurveCount": number;
    public static "java/awt/color/ICC_Profile/icCurveData": number;
    public static "java/awt/color/ICC_Profile/icXYZNumberX": number;
    public "java/awt/color/ICC_Profile/iccProfileSerializedDataVersion": number;
    public "java/awt/color/ICC_Profile/resolvedDeserializedProfile": java_awt_color_ICC__Profile;
    public "<init>(Lsun/java2d/cmm/Profile;)V"(thread: JVMThread, args: [sun_java2d_cmm_Profile], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/<init>(Lsun/java2d/cmm/Profile;)V"(thread: JVMThread, args: [sun_java2d_cmm_Profile], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/cmm/ProfileDeferralInfo;)V"(thread: JVMThread, args: [sun_java2d_cmm_ProfileDeferralInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/<init>(Lsun/java2d/cmm/ProfileDeferralInfo;)V"(thread: JVMThread, args: [sun_java2d_cmm_ProfileDeferralInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance([B)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "java/awt/color/ICC_Profile/getInstance([B)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "getInstance(I)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "java/awt/color/ICC_Profile/getInstance(I)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "getStandardProfile(Ljava/lang/String;)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "java/awt/color/ICC_Profile/getStandardProfile(Ljava/lang/String;)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "getInstance(Ljava/lang/String;)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "java/awt/color/ICC_Profile/getInstance(Ljava/lang/String;)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "getInstance(Ljava/io/InputStream;)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "java/awt/color/ICC_Profile/getInstance(Ljava/io/InputStream;)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "getProfileDataFromStream(Ljava/io/InputStream;)[B"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/color/ICC_Profile/getProfileDataFromStream(Ljava/io/InputStream;)[B"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getDeferredInstance(Lsun/java2d/cmm/ProfileDeferralInfo;)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [sun_java2d_cmm_ProfileDeferralInfo], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public static "java/awt/color/ICC_Profile/getDeferredInstance(Lsun/java2d/cmm/ProfileDeferralInfo;)Ljava/awt/color/ICC_Profile;"(thread: JVMThread, args: [sun_java2d_cmm_ProfileDeferralInfo], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ICC__Profile) => void): void;
    public "activateDeferredProfile()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/activateDeferredProfile()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_Profile/getMajorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_Profile/getMinorVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getProfileClass()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_Profile/getProfileClass()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getColorSpaceType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_Profile/getColorSpaceType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getColorSpaceType(Lsun/java2d/cmm/Profile;)I"(thread: JVMThread, args: [sun_java2d_cmm_Profile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/color/ICC_Profile/getColorSpaceType(Lsun/java2d/cmm/Profile;)I"(thread: JVMThread, args: [sun_java2d_cmm_Profile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPCSType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_Profile/getPCSType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getPCSType(Lsun/java2d/cmm/Profile;)I"(thread: JVMThread, args: [sun_java2d_cmm_Profile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/color/ICC_Profile/getPCSType(Lsun/java2d/cmm/Profile;)I"(thread: JVMThread, args: [sun_java2d_cmm_Profile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/write(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/write(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getData()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ICC_Profile/getData()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getData(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ICC_Profile/getData(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getData(Lsun/java2d/cmm/Profile;I)[B"(thread: JVMThread, args: [sun_java2d_cmm_Profile, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/color/ICC_Profile/getData(Lsun/java2d/cmm/Profile;I)[B"(thread: JVMThread, args: [sun_java2d_cmm_Profile, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "setData(I[B)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/setData(I[B)V"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRenderingIntent(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/setRenderingIntent(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRenderingIntent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_Profile/getRenderingIntent()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_Profile/getNumComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMediaWhitePoint()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ICC_Profile/getMediaWhitePoint()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getXYZTag(I)[F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ICC_Profile/getXYZTag(I)[F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getGamma(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ICC_Profile/getGamma(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTRC(I)[S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ICC_Profile/getTRC(I)[S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "iccCStoJCS(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/color/ICC_Profile/iccCStoJCS(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "intFromBigEndian([BI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/color/ICC_Profile/intFromBigEndian([BI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "intToBigEndian(I[BI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/color/ICC_Profile/intToBigEndian(I[BI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "shortFromBigEndian([BI)S"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/color/ICC_Profile/shortFromBigEndian([BI)S"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "shortToBigEndian(S[BI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/color/ICC_Profile/shortToBigEndian(S[BI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getProfileFile(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/awt/color/ICC_Profile/getProfileFile(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "getStandardProfileFile(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/awt/color/ICC_Profile/getStandardProfileFile(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "isChildOf(Ljava/io/File;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/color/ICC_Profile/isChildOf(Ljava/io/File;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "standardProfileExists(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/color/ICC_Profile/standardProfileExists(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ICC_Profile/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/color/ICC_Profile/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$000(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/awt/color/ICC_Profile/access$000(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
  }
  export interface sun_java2d_cmm_ProfileActivator extends java_lang_Object {
    "activate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_cmm_ProfileDeferralInfo extends java_io_InputStream {
    public "sun/java2d/cmm/ProfileDeferralInfo/colorSpaceType": number;
    public "sun/java2d/cmm/ProfileDeferralInfo/numComponents": number;
    public "sun/java2d/cmm/ProfileDeferralInfo/profileClass": number;
    public "sun/java2d/cmm/ProfileDeferralInfo/filename": java_lang_String;
    public "<init>(Ljava/lang/String;III)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/cmm/ProfileDeferralInfo/<init>(Ljava/lang/String;III)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/cmm/ProfileDeferralInfo/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_java2d_cmm_Profile extends java_lang_Object {
    public "sun/java2d/cmm/Profile/nativePtr": Long;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/cmm/Profile/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNativePtr()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/java2d/cmm/Profile/getNativePtr()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class java_awt_image_SampleModel extends java_lang_Object {
    public "java/awt/image/SampleModel/width": number;
    public "java/awt/image/SampleModel/height": number;
    public "java/awt/image/SampleModel/numBands": number;
    public "java/awt/image/SampleModel/dataType": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/SampleModel/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumBands()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getNumBands()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumDataElements()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getNumDataElements()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDataType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getDataType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransferType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getTransferType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPixel(II[ILjava/awt/image/DataBuffer;)[I"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getPixel(II[ILjava/awt/image/DataBuffer;)[I"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getDataElements(IILjava/lang/Object;Ljava/awt/image/DataBuffer;)Ljava/lang/Object;"(thread: JVMThread, args: [number, number, java_lang_Object, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/SampleModel/getDataElements(IILjava/lang/Object;Ljava/awt/image/DataBuffer;)Ljava/lang/Object;"(thread: JVMThread, args: [number, number, java_lang_Object, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getDataElements(IIIILjava/lang/Object;Ljava/awt/image/DataBuffer;)Ljava/lang/Object;"(thread: JVMThread, args: [number, number, number, number, java_lang_Object, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/SampleModel/getDataElements(IIIILjava/lang/Object;Ljava/awt/image/DataBuffer;)Ljava/lang/Object;"(thread: JVMThread, args: [number, number, number, number, java_lang_Object, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setDataElements(IILjava/lang/Object;Ljava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, java_lang_Object, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setDataElements(IILjava/lang/Object;Ljava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, java_lang_Object, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDataElements(IIIILjava/lang/Object;Ljava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, java_lang_Object, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setDataElements(IIIILjava/lang/Object;Ljava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, java_lang_Object, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPixel(II[FLjava/awt/image/DataBuffer;)[F"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getPixel(II[FLjava/awt/image/DataBuffer;)[F"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPixel(II[DLjava/awt/image/DataBuffer;)[D"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getPixel(II[DLjava/awt/image/DataBuffer;)[D"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPixels(IIII[ILjava/awt/image/DataBuffer;)[I"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getPixels(IIII[ILjava/awt/image/DataBuffer;)[I"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPixels(IIII[FLjava/awt/image/DataBuffer;)[F"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getPixels(IIII[FLjava/awt/image/DataBuffer;)[F"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPixels(IIII[DLjava/awt/image/DataBuffer;)[D"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getPixels(IIII[DLjava/awt/image/DataBuffer;)[D"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getSample(IIILjava/awt/image/DataBuffer;)I"(thread: JVMThread, args: [number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getSample(IIILjava/awt/image/DataBuffer;)I"(thread: JVMThread, args: [number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSampleFloat(IIILjava/awt/image/DataBuffer;)F"(thread: JVMThread, args: [number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getSampleFloat(IIILjava/awt/image/DataBuffer;)F"(thread: JVMThread, args: [number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSampleDouble(IIILjava/awt/image/DataBuffer;)D"(thread: JVMThread, args: [number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getSampleDouble(IIILjava/awt/image/DataBuffer;)D"(thread: JVMThread, args: [number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSamples(IIIII[ILjava/awt/image/DataBuffer;)[I"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getSamples(IIIII[ILjava/awt/image/DataBuffer;)[I"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getSamples(IIIII[FLjava/awt/image/DataBuffer;)[F"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getSamples(IIIII[FLjava/awt/image/DataBuffer;)[F"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getSamples(IIIII[DLjava/awt/image/DataBuffer;)[D"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getSamples(IIIII[DLjava/awt/image/DataBuffer;)[D"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "setPixel(II[ILjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setPixel(II[ILjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixel(II[FLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setPixel(II[FLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixel(II[DLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setPixel(II[DLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixels(IIII[ILjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setPixels(IIII[ILjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixels(IIII[FLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setPixels(IIII[FLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixels(IIII[DLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setPixels(IIII[DLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSample(IIIILjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setSample(IIIILjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSample(IIIFLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setSample(IIIFLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSample(IIIDLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, any, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setSample(IIIDLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, any, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSamples(IIIII[ILjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setSamples(IIIII[ILjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSamples(IIIII[FLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setSamples(IIIII[FLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSamples(IIIII[DLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/SampleModel/setSamples(IIIII[DLjava/awt/image/DataBuffer;)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_image_DataBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "createCompatibleSampleModel(II)Ljava/awt/image/SampleModel;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "java/awt/image/SampleModel/createCompatibleSampleModel(II)Ljava/awt/image/SampleModel;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "createSubsetSampleModel([I)Ljava/awt/image/SampleModel;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "java/awt/image/SampleModel/createSubsetSampleModel([I)Ljava/awt/image/SampleModel;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "createDataBuffer()Ljava/awt/image/DataBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_DataBuffer) => void): void;
    public "java/awt/image/SampleModel/createDataBuffer()Ljava/awt/image/DataBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_DataBuffer) => void): void;
    public "getSampleSize()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/SampleModel/getSampleSize()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getSampleSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/SampleModel/getSampleSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/SampleModel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_image_DataBuffer extends java_lang_Object {
    public static "java/awt/image/DataBuffer/TYPE_BYTE": number;
    public static "java/awt/image/DataBuffer/TYPE_USHORT": number;
    public static "java/awt/image/DataBuffer/TYPE_SHORT": number;
    public static "java/awt/image/DataBuffer/TYPE_INT": number;
    public static "java/awt/image/DataBuffer/TYPE_FLOAT": number;
    public static "java/awt/image/DataBuffer/TYPE_DOUBLE": number;
    public static "java/awt/image/DataBuffer/TYPE_UNDEFINED": number;
    public "java/awt/image/DataBuffer/dataType": number;
    public "java/awt/image/DataBuffer/banks": number;
    public "java/awt/image/DataBuffer/offset": number;
    public "java/awt/image/DataBuffer/size": number;
    public "java/awt/image/DataBuffer/offsets": JVMArray<number>;
    public "java/awt/image/DataBuffer/theTrackable": sun_java2d_StateTrackableDelegate;
    public static "java/awt/image/DataBuffer/dataTypeSize": JVMArray<number>;
    public static "getDataTypeSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/image/DataBuffer/getDataTypeSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/StateTrackable$State;II)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/<init>(Lsun/java2d/StateTrackable$State;II)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/<init>(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/StateTrackable$State;III)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/<init>(Lsun/java2d/StateTrackable$State;III)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/StateTrackable$State;IIII)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/<init>(Lsun/java2d/StateTrackable$State;IIII)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(III[I)V"(thread: JVMThread, args: [number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/<init>(III[I)V"(thread: JVMThread, args: [number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/java2d/StateTrackable$State;III[I)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/<init>(Lsun/java2d/StateTrackable$State;III[I)V"(thread: JVMThread, args: [sun_java2d_StateTrackable$State, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDataType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getDataType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOffsets()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/DataBuffer/getOffsets()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getNumBanks()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getNumBanks()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getElem(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getElem(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getElem(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getElem(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setElem(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/setElem(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setElem(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/setElem(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getElemFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getElemFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getElemFloat(II)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getElemFloat(II)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setElemFloat(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/setElemFloat(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setElemFloat(IIF)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/setElemFloat(IIF)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getElemDouble(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getElemDouble(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getElemDouble(II)D"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/DataBuffer/getElemDouble(II)D"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setElemDouble(ID)V"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/setElemDouble(ID)V"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setElemDouble(IID)V"(thread: JVMThread, args: [number, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/DataBuffer/setElemDouble(IID)V"(thread: JVMThread, args: [number, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "toIntArray(Ljava/lang/Object;)[I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/awt/image/DataBuffer/toIntArray(Ljava/lang/Object;)[I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/DataBuffer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_image_Raster extends java_lang_Object {
    public "java/awt/image/Raster/sampleModel": java_awt_image_SampleModel;
    public "java/awt/image/Raster/dataBuffer": java_awt_image_DataBuffer;
    public "java/awt/image/Raster/minX": number;
    public "java/awt/image/Raster/minY": number;
    public "java/awt/image/Raster/width": number;
    public "java/awt/image/Raster/height": number;
    public "java/awt/image/Raster/sampleModelTranslateX": number;
    public "java/awt/image/Raster/sampleModelTranslateY": number;
    public "java/awt/image/Raster/numBands": number;
    public "java/awt/image/Raster/numDataElements": number;
    public "java/awt/image/Raster/parent": java_awt_image_Raster;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/Raster/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createInterleavedRaster(IIIILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createInterleavedRaster(IIIILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createInterleavedRaster(IIIII[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createInterleavedRaster(IIIII[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createBandedRaster(IIIILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createBandedRaster(IIIILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createBandedRaster(IIII[I[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createBandedRaster(IIII[I[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createPackedRaster(III[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createPackedRaster(III[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createPackedRaster(IIIIILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, number, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createPackedRaster(IIIIILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, number, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createInterleavedRaster(Ljava/awt/image/DataBuffer;IIII[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_DataBuffer, number, number, number, number, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createInterleavedRaster(Ljava/awt/image/DataBuffer;IIII[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_DataBuffer, number, number, number, number, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createBandedRaster(Ljava/awt/image/DataBuffer;III[I[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_DataBuffer, number, number, number, JVMArray<number>, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createBandedRaster(Ljava/awt/image/DataBuffer;III[I[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_DataBuffer, number, number, number, JVMArray<number>, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createPackedRaster(Ljava/awt/image/DataBuffer;III[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_DataBuffer, number, number, number, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createPackedRaster(Ljava/awt/image/DataBuffer;III[ILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_DataBuffer, number, number, number, JVMArray<number>, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createPackedRaster(Ljava/awt/image/DataBuffer;IIILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_DataBuffer, number, number, number, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createPackedRaster(Ljava/awt/image/DataBuffer;IIILjava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_DataBuffer, number, number, number, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createRaster(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Point;)Ljava/awt/image/Raster;"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public static "java/awt/image/Raster/createRaster(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Point;)Ljava/awt/image/Raster;"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public static "createWritableRaster(Ljava/awt/image/SampleModel;Ljava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createWritableRaster(Ljava/awt/image/SampleModel;Ljava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "createWritableRaster(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public static "java/awt/image/Raster/createWritableRaster(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Point;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Point], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "<init>(Ljava/awt/image/SampleModel;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/Raster/<init>(Ljava/awt/image/SampleModel;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/Raster/<init>(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Rectangle;Ljava/awt/Point;Ljava/awt/image/Raster;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Rectangle, java_awt_Point, java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/Raster/<init>(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Rectangle;Ljava/awt/Point;Ljava/awt/image/Raster;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Rectangle, java_awt_Point, java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "getParent()Ljava/awt/image/Raster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "java/awt/image/Raster/getParent()Ljava/awt/image/Raster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "getSampleModelTranslateX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getSampleModelTranslateX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSampleModelTranslateY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getSampleModelTranslateY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createCompatibleWritableRaster()Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/Raster/createCompatibleWritableRaster()Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "createCompatibleWritableRaster(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/Raster/createCompatibleWritableRaster(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "createCompatibleWritableRaster(Ljava/awt/Rectangle;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/Raster/createCompatibleWritableRaster(Ljava/awt/Rectangle;)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "createCompatibleWritableRaster(IIII)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/Raster/createCompatibleWritableRaster(IIII)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "createTranslatedChild(II)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "java/awt/image/Raster/createTranslatedChild(II)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "createChild(IIIIII[I)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "java/awt/image/Raster/createChild(IIIIII[I)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
    public "getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "java/awt/image/Raster/getBounds()Ljava/awt/Rectangle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getMinX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getMinX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMinY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getMinY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumBands()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getNumBands()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumDataElements()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getNumDataElements()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTransferType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getTransferType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDataBuffer()Ljava/awt/image/DataBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_DataBuffer) => void): void;
    public "java/awt/image/Raster/getDataBuffer()Ljava/awt/image/DataBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_DataBuffer) => void): void;
    public "getSampleModel()Ljava/awt/image/SampleModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "java/awt/image/Raster/getSampleModel()Ljava/awt/image/SampleModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_SampleModel) => void): void;
    public "getDataElements(IILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/Raster/getDataElements(IILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getDataElements(IIIILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/awt/image/Raster/getDataElements(IIIILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getPixel(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/Raster/getPixel(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPixel(II[F)[F"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/Raster/getPixel(II[F)[F"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPixel(II[D)[D"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/Raster/getPixel(II[D)[D"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPixels(IIII[I)[I"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/Raster/getPixels(IIII[I)[I"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPixels(IIII[F)[F"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/Raster/getPixels(IIII[F)[F"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getPixels(IIII[D)[D"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/Raster/getPixels(IIII[D)[D"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getSample(III)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getSample(III)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSampleFloat(III)F"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getSampleFloat(III)F"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSampleDouble(III)D"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/image/Raster/getSampleDouble(III)D"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSamples(IIIII[I)[I"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/Raster/getSamples(IIIII[I)[I"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getSamples(IIIII[F)[F"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/Raster/getSamples(IIIII[F)[F"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getSamples(IIIII[D)[D"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/image/Raster/getSamples(IIIII[D)[D"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/image/Raster/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_image_WritableRaster extends java_awt_image_Raster {
    public "<init>(Ljava/awt/image/SampleModel;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/<init>(Ljava/awt/image/SampleModel;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/<init>(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Rectangle;Ljava/awt/Point;Ljava/awt/image/WritableRaster;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Rectangle, java_awt_Point, java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/<init>(Ljava/awt/image/SampleModel;Ljava/awt/image/DataBuffer;Ljava/awt/Rectangle;Ljava/awt/Point;Ljava/awt/image/WritableRaster;)V"(thread: JVMThread, args: [java_awt_image_SampleModel, java_awt_image_DataBuffer, java_awt_Rectangle, java_awt_Point, java_awt_image_WritableRaster], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWritableParent()Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/WritableRaster/getWritableParent()Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "createWritableTranslatedChild(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/WritableRaster/createWritableTranslatedChild(II)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "createWritableChild(IIIIII[I)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "java/awt/image/WritableRaster/createWritableChild(IIIIII[I)Ljava/awt/image/WritableRaster;"(thread: JVMThread, args: [number, number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_WritableRaster) => void): void;
    public "setDataElements(IILjava/lang/Object;)V"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setDataElements(IILjava/lang/Object;)V"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDataElements(IILjava/awt/image/Raster;)V"(thread: JVMThread, args: [number, number, java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setDataElements(IILjava/awt/image/Raster;)V"(thread: JVMThread, args: [number, number, java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDataElements(IIIILjava/lang/Object;)V"(thread: JVMThread, args: [number, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setDataElements(IIIILjava/lang/Object;)V"(thread: JVMThread, args: [number, number, number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRect(Ljava/awt/image/Raster;)V"(thread: JVMThread, args: [java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setRect(Ljava/awt/image/Raster;)V"(thread: JVMThread, args: [java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRect(IILjava/awt/image/Raster;)V"(thread: JVMThread, args: [number, number, java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setRect(IILjava/awt/image/Raster;)V"(thread: JVMThread, args: [number, number, java_awt_image_Raster], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixel(II[I)V"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setPixel(II[I)V"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixel(II[F)V"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setPixel(II[F)V"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixel(II[D)V"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setPixel(II[D)V"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixels(IIII[I)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setPixels(IIII[I)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixels(IIII[F)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setPixels(IIII[F)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPixels(IIII[D)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setPixels(IIII[D)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSample(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setSample(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSample(IIIF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setSample(IIIF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSample(IIID)V"(thread: JVMThread, args: [number, number, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setSample(IIID)V"(thread: JVMThread, args: [number, number, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSamples(IIIII[I)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setSamples(IIIII[I)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSamples(IIIII[F)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setSamples(IIIII[F)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSamples(IIIII[D)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/image/WritableRaster/setSamples(IIIII[D)V"(thread: JVMThread, args: [number, number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_color_ColorSpace extends java_lang_Object implements java_io_Serializable {
    public static "java/awt/color/ColorSpace/serialVersionUID": Long;
    public "java/awt/color/ColorSpace/type": number;
    public "java/awt/color/ColorSpace/numComponents": number;
    public "java/awt/color/ColorSpace/compName": JVMArray<java_lang_String>;
    public static "java/awt/color/ColorSpace/sRGBspace": java_awt_color_ColorSpace;
    public static "java/awt/color/ColorSpace/XYZspace": java_awt_color_ColorSpace;
    public static "java/awt/color/ColorSpace/PYCCspace": java_awt_color_ColorSpace;
    public static "java/awt/color/ColorSpace/GRAYspace": java_awt_color_ColorSpace;
    public static "java/awt/color/ColorSpace/LINEAR_RGBspace": java_awt_color_ColorSpace;
    public static "java/awt/color/ColorSpace/TYPE_XYZ": number;
    public static "java/awt/color/ColorSpace/TYPE_Lab": number;
    public static "java/awt/color/ColorSpace/TYPE_Luv": number;
    public static "java/awt/color/ColorSpace/TYPE_YCbCr": number;
    public static "java/awt/color/ColorSpace/TYPE_Yxy": number;
    public static "java/awt/color/ColorSpace/TYPE_RGB": number;
    public static "java/awt/color/ColorSpace/TYPE_GRAY": number;
    public static "java/awt/color/ColorSpace/TYPE_HSV": number;
    public static "java/awt/color/ColorSpace/TYPE_HLS": number;
    public static "java/awt/color/ColorSpace/TYPE_CMYK": number;
    public static "java/awt/color/ColorSpace/TYPE_CMY": number;
    public static "java/awt/color/ColorSpace/TYPE_2CLR": number;
    public static "java/awt/color/ColorSpace/TYPE_3CLR": number;
    public static "java/awt/color/ColorSpace/TYPE_4CLR": number;
    public static "java/awt/color/ColorSpace/TYPE_5CLR": number;
    public static "java/awt/color/ColorSpace/TYPE_6CLR": number;
    public static "java/awt/color/ColorSpace/TYPE_7CLR": number;
    public static "java/awt/color/ColorSpace/TYPE_8CLR": number;
    public static "java/awt/color/ColorSpace/TYPE_9CLR": number;
    public static "java/awt/color/ColorSpace/TYPE_ACLR": number;
    public static "java/awt/color/ColorSpace/TYPE_BCLR": number;
    public static "java/awt/color/ColorSpace/TYPE_CCLR": number;
    public static "java/awt/color/ColorSpace/TYPE_DCLR": number;
    public static "java/awt/color/ColorSpace/TYPE_ECLR": number;
    public static "java/awt/color/ColorSpace/TYPE_FCLR": number;
    public static "java/awt/color/ColorSpace/CS_sRGB": number;
    public static "java/awt/color/ColorSpace/CS_LINEAR_RGB": number;
    public static "java/awt/color/ColorSpace/CS_CIEXYZ": number;
    public static "java/awt/color/ColorSpace/CS_PYCC": number;
    public static "java/awt/color/ColorSpace/CS_GRAY": number;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/color/ColorSpace/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance(I)Ljava/awt/color/ColorSpace;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ColorSpace) => void): void;
    public static "java/awt/color/ColorSpace/getInstance(I)Ljava/awt/color/ColorSpace;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_color_ColorSpace) => void): void;
    public "isCS_sRGB()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ColorSpace/isCS_sRGB()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toRGB([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ColorSpace/toRGB([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "fromRGB([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ColorSpace/fromRGB([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "toCIEXYZ([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ColorSpace/toCIEXYZ([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "fromCIEXYZ([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/awt/color/ColorSpace/fromCIEXYZ([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ColorSpace/getType()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ColorSpace/getNumComponents()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/color/ColorSpace/getName(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getMinValue(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ColorSpace/getMinValue(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaxValue(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/color/ColorSpace/getMaxValue(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isCS_CIEXYZ(Ljava/awt/color/ColorSpace;)Z"(thread: JVMThread, args: [java_awt_color_ColorSpace], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/color/ColorSpace/isCS_CIEXYZ(Ljava/awt/color/ColorSpace;)Z"(thread: JVMThread, args: [java_awt_color_ColorSpace], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_awt_PaintContext extends java_lang_Object {
    "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "getColorModel()Ljava/awt/image/ColorModel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_ColorModel) => void): void;
    "getRaster(IIII)Ljava/awt/image/Raster;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_image_Raster) => void): void;
  }
  export interface java_awt_Transparency extends java_lang_Object {
    "getTransparency()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_awt_font_GraphicAttribute extends java_lang_Object {
    public "java/awt/font/GraphicAttribute/fAlignment": number;
    public static "java/awt/font/GraphicAttribute/TOP_ALIGNMENT": number;
    public static "java/awt/font/GraphicAttribute/BOTTOM_ALIGNMENT": number;
    public static "java/awt/font/GraphicAttribute/ROMAN_BASELINE": number;
    public static "java/awt/font/GraphicAttribute/CENTER_BASELINE": number;
    public static "java/awt/font/GraphicAttribute/HANGING_BASELINE": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/GraphicAttribute/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAscent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GraphicAttribute/getAscent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GraphicAttribute/getDescent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAdvance()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GraphicAttribute/getAdvance()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/font/GraphicAttribute/getBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getOutline(Ljava/awt/geom/AffineTransform;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "java/awt/font/GraphicAttribute/getOutline(Ljava/awt/geom/AffineTransform;)Ljava/awt/Shape;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "draw(Ljava/awt/Graphics2D;FF)V"(thread: JVMThread, args: [java_awt_Graphics2D, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/GraphicAttribute/draw(Ljava/awt/Graphics2D;FF)V"(thread: JVMThread, args: [java_awt_Graphics2D, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAlignment()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/GraphicAttribute/getAlignment()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getJustificationInfo()Ljava/awt/font/GlyphJustificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphJustificationInfo) => void): void;
    public "java/awt/font/GraphicAttribute/getJustificationInfo()Ljava/awt/font/GlyphJustificationInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphJustificationInfo) => void): void;
  }
  export class java_awt_font_NumericShaper extends java_lang_Object implements java_io_Serializable {
    public "java/awt/font/NumericShaper/key": number;
    public "java/awt/font/NumericShaper/mask": number;
    public "java/awt/font/NumericShaper/shapingRange": java_awt_font_NumericShaper$Range;
    public "java/awt/font/NumericShaper/rangeSet": java_util_Set;
    public "java/awt/font/NumericShaper/rangeArray": JVMArray<java_awt_font_NumericShaper$Range>;
    public static "java/awt/font/NumericShaper/BSEARCH_THRESHOLD": number;
    public static "java/awt/font/NumericShaper/serialVersionUID": Long;
    public static "java/awt/font/NumericShaper/EUROPEAN": number;
    public static "java/awt/font/NumericShaper/ARABIC": number;
    public static "java/awt/font/NumericShaper/EASTERN_ARABIC": number;
    public static "java/awt/font/NumericShaper/DEVANAGARI": number;
    public static "java/awt/font/NumericShaper/BENGALI": number;
    public static "java/awt/font/NumericShaper/GURMUKHI": number;
    public static "java/awt/font/NumericShaper/GUJARATI": number;
    public static "java/awt/font/NumericShaper/ORIYA": number;
    public static "java/awt/font/NumericShaper/TAMIL": number;
    public static "java/awt/font/NumericShaper/TELUGU": number;
    public static "java/awt/font/NumericShaper/KANNADA": number;
    public static "java/awt/font/NumericShaper/MALAYALAM": number;
    public static "java/awt/font/NumericShaper/THAI": number;
    public static "java/awt/font/NumericShaper/LAO": number;
    public static "java/awt/font/NumericShaper/TIBETAN": number;
    public static "java/awt/font/NumericShaper/MYANMAR": number;
    public static "java/awt/font/NumericShaper/ETHIOPIC": number;
    public static "java/awt/font/NumericShaper/KHMER": number;
    public static "java/awt/font/NumericShaper/MONGOLIAN": number;
    public static "java/awt/font/NumericShaper/ALL_RANGES": number;
    public static "java/awt/font/NumericShaper/EUROPEAN_KEY": number;
    public static "java/awt/font/NumericShaper/ARABIC_KEY": number;
    public static "java/awt/font/NumericShaper/EASTERN_ARABIC_KEY": number;
    public static "java/awt/font/NumericShaper/DEVANAGARI_KEY": number;
    public static "java/awt/font/NumericShaper/BENGALI_KEY": number;
    public static "java/awt/font/NumericShaper/GURMUKHI_KEY": number;
    public static "java/awt/font/NumericShaper/GUJARATI_KEY": number;
    public static "java/awt/font/NumericShaper/ORIYA_KEY": number;
    public static "java/awt/font/NumericShaper/TAMIL_KEY": number;
    public static "java/awt/font/NumericShaper/TELUGU_KEY": number;
    public static "java/awt/font/NumericShaper/KANNADA_KEY": number;
    public static "java/awt/font/NumericShaper/MALAYALAM_KEY": number;
    public static "java/awt/font/NumericShaper/THAI_KEY": number;
    public static "java/awt/font/NumericShaper/LAO_KEY": number;
    public static "java/awt/font/NumericShaper/TIBETAN_KEY": number;
    public static "java/awt/font/NumericShaper/MYANMAR_KEY": number;
    public static "java/awt/font/NumericShaper/ETHIOPIC_KEY": number;
    public static "java/awt/font/NumericShaper/KHMER_KEY": number;
    public static "java/awt/font/NumericShaper/MONGOLIAN_KEY": number;
    public static "java/awt/font/NumericShaper/NUM_KEYS": number;
    public static "java/awt/font/NumericShaper/CONTEXTUAL_MASK": number;
    public static "java/awt/font/NumericShaper/bases": JVMArray<number>;
    public static "java/awt/font/NumericShaper/contexts": JVMArray<number>;
    public static "java/awt/font/NumericShaper/ctCache": number;
    public static "java/awt/font/NumericShaper/ctCacheLimit": number;
    public "java/awt/font/NumericShaper/currentRange": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper/strongTable": JVMArray<number>;
    public "java/awt/font/NumericShaper/stCache": number;
    public static "getContextKey(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper/getContextKey(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rangeForCodePoint(I)Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper$Range) => void): void;
    public "java/awt/font/NumericShaper/rangeForCodePoint(I)Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper$Range) => void): void;
    public "isStrongDirectional(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper/isStrongDirectional(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getKeyFromMask(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper/getKeyFromMask(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getShaper(I)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "java/awt/font/NumericShaper/getShaper(I)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "getShaper(Ljava/awt/font/NumericShaper$Range;)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "java/awt/font/NumericShaper/getShaper(Ljava/awt/font/NumericShaper$Range;)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "getContextualShaper(I)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "java/awt/font/NumericShaper/getContextualShaper(I)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "getContextualShaper(Ljava/util/Set;)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "java/awt/font/NumericShaper/getContextualShaper(Ljava/util/Set;)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "getContextualShaper(II)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "java/awt/font/NumericShaper/getContextualShaper(II)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "getContextualShaper(Ljava/util/Set;Ljava/awt/font/NumericShaper$Range;)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [java_util_Set, java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public static "java/awt/font/NumericShaper/getContextualShaper(Ljava/util/Set;Ljava/awt/font/NumericShaper$Range;)Ljava/awt/font/NumericShaper;"(thread: JVMThread, args: [java_util_Set, java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/font/NumericShaper$Range;Ljava/util/Set;)V"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/<init>(Ljava/awt/font/NumericShaper$Range;Ljava/util/Set;)V"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "shape([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/shape([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "shape([CIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/shape([CIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "shape([CIILjava/awt/font/NumericShaper$Range;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/shape([CIILjava/awt/font/NumericShaper$Range;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkParams([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/checkParams([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isContextual()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper/isContextual()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRanges()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper/getRanges()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRangeSet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/awt/font/NumericShaper/getRangeSet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "shapeNonContextually([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/shapeNonContextually([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "shapeContextually([CIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/shapeContextually([CIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "shapeContextually([CIILjava/awt/font/NumericShaper$Range;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/shapeContextually([CIILjava/awt/font/NumericShaper$Range;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/font/NumericShaper/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getHighBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper/getHighBit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "search(I[III)I"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper/search(I[III)I"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/font/NumericShaper/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_font_NumericShaper$Range extends java_lang_Enum {
    public static "java/awt/font/NumericShaper$Range/EUROPEAN": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/ARABIC": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/EASTERN_ARABIC": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/DEVANAGARI": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/BENGALI": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/GURMUKHI": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/GUJARATI": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/ORIYA": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/TAMIL": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/TELUGU": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/KANNADA": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/MALAYALAM": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/THAI": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/LAO": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/TIBETAN": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/MYANMAR": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/ETHIOPIC": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/KHMER": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/MONGOLIAN": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/NKO": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/MYANMAR_SHAN": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/LIMBU": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/NEW_TAI_LUE": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/BALINESE": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/SUNDANESE": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/LEPCHA": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/OL_CHIKI": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/VAI": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/SAURASHTRA": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/KAYAH_LI": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/CHAM": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/TAI_THAM_HORA": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/TAI_THAM_THAM": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/JAVANESE": java_awt_font_NumericShaper$Range;
    public static "java/awt/font/NumericShaper$Range/MEETEI_MAYEK": java_awt_font_NumericShaper$Range;
    public "java/awt/font/NumericShaper$Range/base": number;
    public "java/awt/font/NumericShaper$Range/start": number;
    public "java/awt/font/NumericShaper$Range/end": number;
    public static "java/awt/font/NumericShaper$Range/$VALUES": JVMArray<java_awt_font_NumericShaper$Range>;
    public static "values()[Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_font_NumericShaper$Range>) => void): void;
    public static "java/awt/font/NumericShaper$Range/values()[Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_font_NumericShaper$Range>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper$Range) => void): void;
    public static "java/awt/font/NumericShaper$Range/valueOf(Ljava/lang/String;)Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper$Range) => void): void;
    public static "toRangeIndex(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper$Range/toRangeIndex(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "indexToRange(I)Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper$Range) => void): void;
    public static "java/awt/font/NumericShaper$Range/indexToRange(I)Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper$Range) => void): void;
    public static "toRangeMask(Ljava/util/Set;)I"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper$Range/toRangeMask(Ljava/util/Set;)I"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "maskToRangeSet(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "java/awt/font/NumericShaper$Range/maskToRangeSet(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "<init>(Ljava/lang/String;IIII)V"(thread: JVMThread, args: [java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper$Range/<init>(Ljava/lang/String;IIII)V"(thread: JVMThread, args: [java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDigitBase()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper$Range/getDigitBase()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumericBase()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper$Range/getNumericBase()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inRange(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper$Range/inRange(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/String;IIIILjava/awt/font/NumericShaper$1;)V"(thread: JVMThread, args: [java_lang_String, number, number, number, number, java_awt_font_NumericShaper$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper$Range/<init>(Ljava/lang/String;IIIILjava/awt/font/NumericShaper$1;)V"(thread: JVMThread, args: [java_lang_String, number, number, number, number, java_awt_font_NumericShaper$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$100(Ljava/awt/font/NumericShaper$Range;I)Z"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper$Range/access$100(Ljava/awt/font/NumericShaper$Range;I)Z"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper$Range/access$200(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$300(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper$Range/access$300(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$400(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper$Range/access$400(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper$Range/access$500(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$600(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "java/awt/font/NumericShaper$Range/access$600(I)Ljava/util/Set;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public static "access$700(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper$Range/access$700(Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$800(I)Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper$Range) => void): void;
    public static "java/awt/font/NumericShaper$Range/access$800(I)Ljava/awt/font/NumericShaper$Range;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_NumericShaper$Range) => void): void;
    public static "access$900(Ljava/util/Set;)I"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/awt/font/NumericShaper$Range/access$900(Ljava/util/Set;)I"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/awt/font/NumericShaper$Range/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_font_NumericShaper$1 extends java_lang_Object implements java_util_Comparator {
    public "java/awt/font/NumericShaper$1/this$0": java_awt_font_NumericShaper;
    public "<init>(Ljava/awt/font/NumericShaper;)V"(thread: JVMThread, args: [java_awt_font_NumericShaper], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/font/NumericShaper$1/<init>(Ljava/awt/font/NumericShaper;)V"(thread: JVMThread, args: [java_awt_font_NumericShaper], cb?: (e?: java_lang_Throwable) => void): void;
    public "compare(Ljava/awt/font/NumericShaper$Range;Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range, java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper$1/compare(Ljava/awt/font/NumericShaper$Range;Ljava/awt/font/NumericShaper$Range;)I"(thread: JVMThread, args: [java_awt_font_NumericShaper$Range, java_awt_font_NumericShaper$Range], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compare(Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/font/NumericShaper$1/compare(Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "reversed()Ljava/util/Comparator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparing(Ljava/util/Comparator;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparing(Ljava/util/function/Function;Ljava/util/Comparator;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_Function, java_util_Comparator], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparing(Ljava/util/function/Function;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparingInt(Ljava/util/function/ToIntFunction;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_ToIntFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparingLong(Ljava/util/function/ToLongFunction;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_ToLongFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "thenComparingDouble(Ljava/util/function/ToDoubleFunction;)Ljava/util/Comparator;"(thread: JVMThread, args: [java_util_function_ToDoubleFunction], cb?: (e?: java_lang_Throwable, rv?: java_util_Comparator) => void): void;
    "lambda$thenComparing$36697e65$1(Ljava/util/Comparator;Ljava/lang/Object;Ljava/lang/Object;)I"(thread: JVMThread, args: [java_util_Comparator, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_font_Font2DHandle extends java_lang_Object {
    public "sun/font/Font2DHandle/font2D": sun_font_Font2D;
    public "<init>(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/Font2DHandle/<init>(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_awt_peer_FontPeer extends java_lang_Object {
  }
  export class sun_java2d_loops_FontInfo extends java_lang_Object implements java_lang_Cloneable {
    public "sun/java2d/loops/FontInfo/font": java_awt_Font;
    public "sun/java2d/loops/FontInfo/font2D": sun_font_Font2D;
    public "sun/java2d/loops/FontInfo/fontStrike": sun_font_FontStrike;
    public "sun/java2d/loops/FontInfo/devTx": JVMArray<number>;
    public "sun/java2d/loops/FontInfo/glyphTx": JVMArray<number>;
    public "sun/java2d/loops/FontInfo/pixelHeight": number;
    public "sun/java2d/loops/FontInfo/originX": number;
    public "sun/java2d/loops/FontInfo/originY": number;
    public "sun/java2d/loops/FontInfo/aaHint": number;
    public "sun/java2d/loops/FontInfo/lcdRGBOrder": number;
    public "sun/java2d/loops/FontInfo/lcdSubPixPos": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/loops/FontInfo/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mtx([D)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/java2d/loops/FontInfo/mtx([D)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/java2d/loops/FontInfo/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/java2d/loops/FontInfo/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_java2d_pipe_Region extends java_lang_Object {
    public static "sun/java2d/pipe/Region/INIT_SIZE": number;
    public static "sun/java2d/pipe/Region/GROW_SIZE": number;
    public static "sun/java2d/pipe/Region/EMPTY_REGION": sun_java2d_pipe_Region;
    public static "sun/java2d/pipe/Region/WHOLE_REGION": sun_java2d_pipe_Region;
    public "sun/java2d/pipe/Region/lox": number;
    public "sun/java2d/pipe/Region/loy": number;
    public "sun/java2d/pipe/Region/hix": number;
    public "sun/java2d/pipe/Region/hiy": number;
    public "sun/java2d/pipe/Region/endIndex": number;
    public "sun/java2d/pipe/Region/bands": JVMArray<number>;
    public static "sun/java2d/pipe/Region/INCLUDE_A": number;
    public static "sun/java2d/pipe/Region/INCLUDE_B": number;
    public static "sun/java2d/pipe/Region/INCLUDE_COMMON": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/Region/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "dimAdd(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/pipe/Region/dimAdd(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "clipAdd(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/pipe/Region/clipAdd(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "clipScale(ID)I"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/java2d/pipe/Region/clipScale(ID)I"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII[II)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/<init>(IIII[II)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance(Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Shape, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "sun/java2d/pipe/Region/getInstance(Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Shape, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "getInstance(Lsun/java2d/pipe/Region;Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region, java_awt_Shape, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "sun/java2d/pipe/Region/getInstance(Lsun/java2d/pipe/Region;Ljava/awt/Shape;Ljava/awt/geom/AffineTransform;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region, java_awt_Shape, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "getInstance(Lsun/java2d/pipe/Region;ZLjava/awt/Shape;Ljava/awt/geom/AffineTransform;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region, number, java_awt_Shape, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "sun/java2d/pipe/Region/getInstance(Lsun/java2d/pipe/Region;ZLjava/awt/Shape;Ljava/awt/geom/AffineTransform;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region, number, java_awt_Shape, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "getInstance(IIII[I)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "sun/java2d/pipe/Region/getInstance(IIII[I)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "getInstance(Ljava/awt/Rectangle;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "sun/java2d/pipe/Region/getInstance(Ljava/awt/Rectangle;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "getInstanceXYWH(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "sun/java2d/pipe/Region/getInstanceXYWH(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "getInstance([I)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "sun/java2d/pipe/Region/getInstance([I)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "getInstanceXYXY(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public static "sun/java2d/pipe/Region/getInstanceXYXY(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "setOutputArea(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/setOutputArea(Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOutputAreaXYWH(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/setOutputAreaXYWH(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOutputArea([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/setOutputArea([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOutputAreaXYXY(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/setOutputAreaXYXY(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "appendSpans(Lsun/java2d/pipe/SpanIterator;)V"(thread: JVMThread, args: [sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/appendSpans(Lsun/java2d/pipe/SpanIterator;)V"(thread: JVMThread, args: [sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "getScaledRegion(DD)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getScaledRegion(DD)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getTranslatedRegion(II)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getTranslatedRegion(II)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getSafeTranslatedRegion(II)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getSafeTranslatedRegion(II)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getIntersection(Ljava/awt/Rectangle;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getIntersection(Ljava/awt/Rectangle;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getIntersectionXYWH(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getIntersectionXYWH(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getIntersectionXYXY(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getIntersectionXYXY(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getIntersection(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getIntersection(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getUnion(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getUnion(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getDifference(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getDifference(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getExclusiveOr(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getExclusiveOr(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "filterSpans(Lsun/java2d/pipe/Region;Lsun/java2d/pipe/Region;I)V"(thread: JVMThread, args: [sun_java2d_pipe_Region, sun_java2d_pipe_Region, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/filterSpans(Lsun/java2d/pipe/Region;Lsun/java2d/pipe/Region;I)V"(thread: JVMThread, args: [sun_java2d_pipe_Region, sun_java2d_pipe_Region, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBoundsIntersection(Ljava/awt/Rectangle;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getBoundsIntersection(Ljava/awt/Rectangle;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [java_awt_Rectangle], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getBoundsIntersectionXYWH(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getBoundsIntersectionXYWH(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getBoundsIntersectionXYXY(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getBoundsIntersectionXYXY(IIII)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "getBoundsIntersection(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "sun/java2d/pipe/Region/getBoundsIntersection(Lsun/java2d/pipe/Region;)Lsun/java2d/pipe/Region;"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_Region) => void): void;
    public "appendSpan([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/appendSpan([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "needSpace(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/needSpace(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "endRow([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/endRow([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "calcBBox()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/calcBBox()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLoX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/getLoX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLoY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/getLoY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHiX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/getHiX()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHiY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/getHiY()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/getHeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isRectangular()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/isRectangular()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/contains(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInsideXYWH(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/isInsideXYWH(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInsideXYXY(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/isInsideXYXY(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInsideQuickCheck(Lsun/java2d/pipe/Region;)Z"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/isInsideQuickCheck(Lsun/java2d/pipe/Region;)Z"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersectsQuickCheckXYXY(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/intersectsQuickCheckXYXY(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intersectsQuickCheck(Lsun/java2d/pipe/Region;)Z"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/intersectsQuickCheck(Lsun/java2d/pipe/Region;)Z"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "encompasses(Lsun/java2d/pipe/Region;)Z"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/encompasses(Lsun/java2d/pipe/Region;)Z"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "encompassesXYWH(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/encompassesXYWH(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "encompassesXYXY(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/encompassesXYXY(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/getBounds([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "clipBoxToBounds([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/Region/clipBoxToBounds([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getIterator()Lsun/java2d/pipe/RegionIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_RegionIterator) => void): void;
    public "sun/java2d/pipe/Region/getIterator()Lsun/java2d/pipe/RegionIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_RegionIterator) => void): void;
    public "getSpanIterator()Lsun/java2d/pipe/SpanIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_SpanIterator) => void): void;
    public "sun/java2d/pipe/Region/getSpanIterator()Lsun/java2d/pipe/SpanIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_SpanIterator) => void): void;
    public "getSpanIterator([I)Lsun/java2d/pipe/SpanIterator;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_SpanIterator) => void): void;
    public "sun/java2d/pipe/Region/getSpanIterator([I)Lsun/java2d/pipe/SpanIterator;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_SpanIterator) => void): void;
    public "filter(Lsun/java2d/pipe/SpanIterator;)Lsun/java2d/pipe/SpanIterator;"(thread: JVMThread, args: [sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_SpanIterator) => void): void;
    public "sun/java2d/pipe/Region/filter(Lsun/java2d/pipe/SpanIterator;)Lsun/java2d/pipe/SpanIterator;"(thread: JVMThread, args: [sun_java2d_pipe_SpanIterator], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_SpanIterator) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/java2d/pipe/Region/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/Region/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/java2d/pipe/Region/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_java2d_pipe_RegionIterator extends java_lang_Object {
    public "sun/java2d/pipe/RegionIterator/region": sun_java2d_pipe_Region;
    public "sun/java2d/pipe/RegionIterator/curIndex": number;
    public "sun/java2d/pipe/RegionIterator/numXbands": number;
    public "<init>(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/RegionIterator/<init>(Lsun/java2d/pipe/Region;)V"(thread: JVMThread, args: [sun_java2d_pipe_Region], cb?: (e?: java_lang_Throwable) => void): void;
    public "createCopy()Lsun/java2d/pipe/RegionIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_RegionIterator) => void): void;
    public "sun/java2d/pipe/RegionIterator/createCopy()Lsun/java2d/pipe/RegionIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_java2d_pipe_RegionIterator) => void): void;
    public "copyStateFrom(Lsun/java2d/pipe/RegionIterator;)V"(thread: JVMThread, args: [sun_java2d_pipe_RegionIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/RegionIterator/copyStateFrom(Lsun/java2d/pipe/RegionIterator;)V"(thread: JVMThread, args: [sun_java2d_pipe_RegionIterator], cb?: (e?: java_lang_Throwable) => void): void;
    public "nextYRange([I)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/RegionIterator/nextYRange([I)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextXBand([I)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/java2d/pipe/RegionIterator/nextXBand([I)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class sun_font_X11TextRenderer extends sun_java2d_pipe_GlyphListPipe {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/X11TextRenderer/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawGlyphVector(Lsun/java2d/SunGraphics2D;Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/X11TextRenderer/drawGlyphVector(Lsun/java2d/SunGraphics2D;Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "doDrawGlyphList(JJLsun/java2d/pipe/Region;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [Long, any, Long, any, sun_java2d_pipe_Region, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/X11TextRenderer/doDrawGlyphList(JJLsun/java2d/pipe/Region;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [Long, any, Long, any, sun_java2d_pipe_Region, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawGlyphList(Lsun/java2d/SunGraphics2D;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/X11TextRenderer/drawGlyphList(Lsun/java2d/SunGraphics2D;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "traceWrap()Lsun/font/X11TextRenderer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_X11TextRenderer) => void): void;
    public "sun/font/X11TextRenderer/traceWrap()Lsun/font/X11TextRenderer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_X11TextRenderer) => void): void;
  }
  export class sun_java2d_pipe_GlyphListPipe extends java_lang_Object implements sun_java2d_pipe_TextPipe {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/GlyphListPipe/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawString(Lsun/java2d/SunGraphics2D;Ljava/lang/String;DD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_lang_String, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/GlyphListPipe/drawString(Lsun/java2d/SunGraphics2D;Ljava/lang/String;DD)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_lang_String, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawChars(Lsun/java2d/SunGraphics2D;[CIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/GlyphListPipe/drawChars(Lsun/java2d/SunGraphics2D;[CIIII)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawGlyphVector(Lsun/java2d/SunGraphics2D;Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/GlyphListPipe/drawGlyphVector(Lsun/java2d/SunGraphics2D;Ljava/awt/font/GlyphVector;FF)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, java_awt_font_GlyphVector, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawGlyphList(Lsun/java2d/SunGraphics2D;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/GlyphListPipe/drawGlyphList(Lsun/java2d/SunGraphics2D;Lsun/font/GlyphList;)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_font_GlyphList], cb?: (e?: java_lang_Throwable) => void): void;
    public "drawGlyphList(Lsun/java2d/SunGraphics2D;Lsun/font/GlyphList;I)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_font_GlyphList, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/java2d/pipe/GlyphListPipe/drawGlyphList(Lsun/java2d/SunGraphics2D;Lsun/font/GlyphList;I)V"(thread: JVMThread, args: [sun_java2d_SunGraphics2D, sun_font_GlyphList, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_FontConfigManager$FcCompFont extends java_lang_Object {
    public "sun/font/FontConfigManager$FcCompFont/fcName": java_lang_String;
    public "sun/font/FontConfigManager$FcCompFont/fcFamily": java_lang_String;
    public "sun/font/FontConfigManager$FcCompFont/jdkName": java_lang_String;
    public "sun/font/FontConfigManager$FcCompFont/style": number;
    public "sun/font/FontConfigManager$FcCompFont/firstFont": sun_font_FontConfigManager$FontConfigFont;
    public "sun/font/FontConfigManager$FcCompFont/allFonts": JVMArray<sun_font_FontConfigManager$FontConfigFont>;
    public "sun/font/FontConfigManager$FcCompFont/compFont": sun_font_CompositeFont;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontConfigManager$FcCompFont/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_CompositeFont extends sun_font_Font2D {
    public "sun/font/CompositeFont/deferredInitialisation": JVMArray<number>;
    public "sun/font/CompositeFont/componentFileNames": JVMArray<java_lang_String>;
    public "sun/font/CompositeFont/componentNames": JVMArray<java_lang_String>;
    public "sun/font/CompositeFont/components": JVMArray<sun_font_PhysicalFont>;
    public "sun/font/CompositeFont/numSlots": number;
    public "sun/font/CompositeFont/numMetricsSlots": number;
    public "sun/font/CompositeFont/exclusionRanges": JVMArray<number>;
    public "sun/font/CompositeFont/maxIndices": JVMArray<number>;
    public "sun/font/CompositeFont/numGlyphs": number;
    public "sun/font/CompositeFont/localeSlot": number;
    public "sun/font/CompositeFont/isStdComposite": number;
    public "<init>(Ljava/lang/String;[Ljava/lang/String;[Ljava/lang/String;I[I[IZLsun/font/SunFontManager;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, JVMArray<java_lang_String>, number, JVMArray<number>, JVMArray<number>, number, sun_font_SunFontManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CompositeFont/<init>(Ljava/lang/String;[Ljava/lang/String;[Ljava/lang/String;I[I[IZLsun/font/SunFontManager;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, JVMArray<java_lang_String>, number, JVMArray<number>, JVMArray<number>, number, sun_font_SunFontManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/PhysicalFont;Lsun/font/CompositeFont;)V"(thread: JVMThread, args: [sun_font_PhysicalFont, sun_font_CompositeFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CompositeFont/<init>(Lsun/font/PhysicalFont;Lsun/font/CompositeFont;)V"(thread: JVMThread, args: [sun_font_PhysicalFont, sun_font_CompositeFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "doDeferredInitialisation(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CompositeFont/doDeferredInitialisation(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceComponentFont(Lsun/font/PhysicalFont;Lsun/font/PhysicalFont;)V"(thread: JVMThread, args: [sun_font_PhysicalFont, sun_font_PhysicalFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CompositeFont/replaceComponentFont(Lsun/font/PhysicalFont;Lsun/font/PhysicalFont;)V"(thread: JVMThread, args: [sun_font_PhysicalFont, sun_font_PhysicalFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "isExcludedChar(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFont/isExcludedChar(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getStyleMetrics(F[FI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CompositeFont/getStyleMetrics(F[FI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumSlots()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFont/getNumSlots()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSlotFont(I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/CompositeFont/getSlotFont(I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "createStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/CompositeFont/createStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "isStdComposite()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFont/isStdComposite()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getValidatedGlyphCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFont/getValidatedGlyphCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMapper()Lsun/font/CharToGlyphMapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CharToGlyphMapper) => void): void;
    public "sun/font/CompositeFont/getMapper()Lsun/font/CharToGlyphMapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CharToGlyphMapper) => void): void;
    public "hasSupplementaryChars()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFont/hasSupplementaryChars()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFont/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFont/getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDisplay(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFont/canDisplay(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "useAAForPtSize(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFont/useAAForPtSize(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/CompositeFont/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_font_CharToGlyphMapper extends java_lang_Object {
    public static "sun/font/CharToGlyphMapper/HI_SURROGATE_START": number;
    public static "sun/font/CharToGlyphMapper/HI_SURROGATE_END": number;
    public static "sun/font/CharToGlyphMapper/LO_SURROGATE_START": number;
    public static "sun/font/CharToGlyphMapper/LO_SURROGATE_END": number;
    public static "sun/font/CharToGlyphMapper/UNINITIALIZED_GLYPH": number;
    public static "sun/font/CharToGlyphMapper/INVISIBLE_GLYPH_ID": number;
    public static "sun/font/CharToGlyphMapper/INVISIBLE_GLYPHS": number;
    public "sun/font/CharToGlyphMapper/missingGlyph": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CharToGlyphMapper/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CharToGlyphMapper/getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDisplay(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CharToGlyphMapper/canDisplay(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDisplay(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CharToGlyphMapper/canDisplay(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charToGlyph(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CharToGlyphMapper/charToGlyph(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charToGlyph(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CharToGlyphMapper/charToGlyph(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CharToGlyphMapper/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charsToGlyphs(I[C[I)V"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CharToGlyphMapper/charsToGlyphs(I[C[I)V"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "charsToGlyphsNS(I[C[I)Z"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CharToGlyphMapper/charsToGlyphsNS(I[C[I)Z"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charsToGlyphs(I[I[I)V"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CharToGlyphMapper/charsToGlyphs(I[I[I)V"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_FontStrikeDesc extends java_lang_Object {
    public static "sun/font/FontStrikeDesc/AA_ON": number;
    public static "sun/font/FontStrikeDesc/AA_LCD_H": number;
    public static "sun/font/FontStrikeDesc/AA_LCD_V": number;
    public static "sun/font/FontStrikeDesc/FRAC_METRICS_ON": number;
    public static "sun/font/FontStrikeDesc/FRAC_METRICS_SP": number;
    public "sun/font/FontStrikeDesc/devTx": java_awt_geom_AffineTransform;
    public "sun/font/FontStrikeDesc/glyphTx": java_awt_geom_AffineTransform;
    public "sun/font/FontStrikeDesc/style": number;
    public "sun/font/FontStrikeDesc/aaHint": number;
    public "sun/font/FontStrikeDesc/fmHint": number;
    public "sun/font/FontStrikeDesc/hashCode": number;
    public "sun/font/FontStrikeDesc/valuemask": number;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontStrikeDesc/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontStrikeDesc/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrikeDesc/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getAAHintIntVal(Ljava/lang/Object;Lsun/font/Font2D;I)I"(thread: JVMThread, args: [java_lang_Object, sun_font_Font2D, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/FontStrikeDesc/getAAHintIntVal(Ljava/lang/Object;Lsun/font/Font2D;I)I"(thread: JVMThread, args: [java_lang_Object, sun_font_Font2D, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getAAHintIntVal(Lsun/font/Font2D;Ljava/awt/Font;Ljava/awt/font/FontRenderContext;)I"(thread: JVMThread, args: [sun_font_Font2D, java_awt_Font, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/FontStrikeDesc/getAAHintIntVal(Lsun/font/Font2D;Ljava/awt/Font;Ljava/awt/font/FontRenderContext;)I"(thread: JVMThread, args: [sun_font_Font2D, java_awt_Font, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFMHintIntVal(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/FontStrikeDesc/getFMHintIntVal(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;III)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_geom_AffineTransform, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrikeDesc/<init>(Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;III)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_geom_AffineTransform, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrikeDesc/<init>(Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/FontStrikeDesc/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_font_PhysicalFont extends sun_font_Font2D {
    public "sun/font/PhysicalFont/platName": java_lang_String;
    public "sun/font/PhysicalFont/nativeNames": java_lang_Object;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/PhysicalFont/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/PhysicalFont/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/PhysicalFont/<init>(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/PhysicalFont/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/PhysicalFont/getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/PhysicalFont/getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/PhysicalFont/getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/PhysicalFont/getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/PhysicalFont/getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/PhysicalFont/getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/PhysicalFont/getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/PhysicalFont/getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
  }
  export class sun_font_FontConfigManager$FontConfigFont extends java_lang_Object {
    public "sun/font/FontConfigManager$FontConfigFont/familyName": java_lang_String;
    public "sun/font/FontConfigManager$FontConfigFont/styleStr": java_lang_String;
    public "sun/font/FontConfigManager$FontConfigFont/fullName": java_lang_String;
    public "sun/font/FontConfigManager$FontConfigFont/fontFile": java_lang_String;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontConfigManager$FontConfigFont/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_FontConfigManager$FontConfigInfo extends java_lang_Object {
    public "sun/font/FontConfigManager$FontConfigInfo/fcVersion": number;
    public "sun/font/FontConfigManager$FontConfigInfo/cacheDirs": JVMArray<java_lang_String>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontConfigManager$FontConfigInfo/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_FontConfigManager extends java_lang_Object {
    public static "sun/font/FontConfigManager/fontConfigFailed": number;
    public static "sun/font/FontConfigManager/fcInfo": sun_font_FontConfigManager$FontConfigInfo;
    public static "sun/font/FontConfigManager/fontConfigNames": JVMArray<java_lang_String>;
    public "sun/font/FontConfigManager/fontConfigFonts": JVMArray<sun_font_FontConfigManager$FcCompFont>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontConfigManager/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getFontConfigAAHint()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/font/FontConfigManager/getFontConfigAAHint()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "getFontConfigAAHint(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/font/FontConfigManager/getFontConfigAAHint(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "getFCLocaleStr()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/font/FontConfigManager/getFCLocaleStr()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getFontConfigVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/FontConfigManager/getFontConfigVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "initFontConfigFonts(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontConfigManager/initFontConfigFonts(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerFromFcInfo(Lsun/font/FontConfigManager$FcCompFont;)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [sun_font_FontConfigManager$FcCompFont], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/FontConfigManager/registerFromFcInfo(Lsun/font/FontConfigManager$FcCompFont;)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [sun_font_FontConfigManager$FcCompFont], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "getFontConfigFont(Ljava/lang/String;I)Lsun/font/CompositeFont;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_CompositeFont) => void): void;
    public "sun/font/FontConfigManager/getFontConfigFont(Ljava/lang/String;I)Lsun/font/CompositeFont;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_CompositeFont) => void): void;
    public "getFontConfigFonts()[Lsun/font/FontConfigManager$FcCompFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_FontConfigManager$FcCompFont>) => void): void;
    public "sun/font/FontConfigManager/getFontConfigFonts()[Lsun/font/FontConfigManager$FcCompFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_FontConfigManager$FcCompFont>) => void): void;
    public static "getFontConfig(Ljava/lang/String;Lsun/font/FontConfigManager$FontConfigInfo;[Lsun/font/FontConfigManager$FcCompFont;Z)V"(thread: JVMThread, args: [java_lang_String, sun_font_FontConfigManager$FontConfigInfo, JVMArray<sun_font_FontConfigManager$FcCompFont>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FontConfigManager/getFontConfig(Ljava/lang/String;Lsun/font/FontConfigManager$FontConfigInfo;[Lsun/font/FontConfigManager$FcCompFont;Z)V"(thread: JVMThread, args: [java_lang_String, sun_font_FontConfigManager$FontConfigInfo, JVMArray<sun_font_FontConfigManager$FcCompFont>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "populateFontConfig([Lsun/font/FontConfigManager$FcCompFont;)V"(thread: JVMThread, args: [JVMArray<sun_font_FontConfigManager$FcCompFont>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontConfigManager/populateFontConfig([Lsun/font/FontConfigManager$FcCompFont;)V"(thread: JVMThread, args: [JVMArray<sun_font_FontConfigManager$FcCompFont>], cb?: (e?: java_lang_Throwable) => void): void;
    public "loadFontConfig()[Lsun/font/FontConfigManager$FcCompFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_FontConfigManager$FcCompFont>) => void): void;
    public "sun/font/FontConfigManager/loadFontConfig()[Lsun/font/FontConfigManager$FcCompFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_FontConfigManager$FcCompFont>) => void): void;
    public "getFontConfigInfo()Lsun/font/FontConfigManager$FontConfigInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontConfigManager$FontConfigInfo) => void): void;
    public "sun/font/FontConfigManager/getFontConfigInfo()Lsun/font/FontConfigManager$FontConfigInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontConfigManager$FontConfigInfo) => void): void;
    public static "getFontConfigAASettings(Ljava/lang/String;Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/FontConfigManager/getFontConfigAASettings(Ljava/lang/String;Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FontConfigManager/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_SunFontManager extends java_lang_Object implements sun_java2d_FontSupport, sun_font_FontManagerForSGE {
    public static "sun/font/SunFontManager/FONTFORMAT_NONE": number;
    public static "sun/font/SunFontManager/FONTFORMAT_TRUETYPE": number;
    public static "sun/font/SunFontManager/FONTFORMAT_TYPE1": number;
    public static "sun/font/SunFontManager/FONTFORMAT_T2K": number;
    public static "sun/font/SunFontManager/FONTFORMAT_TTC": number;
    public static "sun/font/SunFontManager/FONTFORMAT_COMPOSITE": number;
    public static "sun/font/SunFontManager/FONTFORMAT_NATIVE": number;
    public static "sun/font/SunFontManager/CHANNELPOOLSIZE": number;
    public "sun/font/SunFontManager/fontFileCache": JVMArray<sun_font_FileFont>;
    public "sun/font/SunFontManager/lastPoolIndex": number;
    public "sun/font/SunFontManager/maxCompFont": number;
    public "sun/font/SunFontManager/compFonts": JVMArray<sun_font_CompositeFont>;
    public "sun/font/SunFontManager/compositeFonts": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/SunFontManager/physicalFonts": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/SunFontManager/registeredFonts": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/SunFontManager/fullNameToFont": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/SunFontManager/localeFullNamesToFont": java_util_HashMap;
    public "sun/font/SunFontManager/defaultPhysicalFont": sun_font_PhysicalFont;
    public static "sun/font/SunFontManager/longAddresses": number;
    public "sun/font/SunFontManager/loaded1dot0Fonts": number;
    public "sun/font/SunFontManager/loadedAllFonts": number;
    public "sun/font/SunFontManager/loadedAllFontFiles": number;
    public "sun/font/SunFontManager/jreFontMap": java_util_HashMap;
    public "sun/font/SunFontManager/jreLucidaFontFiles": java_util_HashSet;
    public "sun/font/SunFontManager/jreOtherFontFiles": JVMArray<java_lang_String>;
    public "sun/font/SunFontManager/noOtherJREFontFiles": number;
    public static "sun/font/SunFontManager/lucidaFontName": java_lang_String;
    public static "sun/font/SunFontManager/jreLibDirName": java_lang_String;
    public static "sun/font/SunFontManager/jreFontDirName": java_lang_String;
    public static "sun/font/SunFontManager/missingFontFiles": java_util_HashSet;
    public "sun/font/SunFontManager/defaultFontName": java_lang_String;
    public "sun/font/SunFontManager/defaultFontFileName": java_lang_String;
    public "sun/font/SunFontManager/registeredFontFiles": java_util_HashSet;
    public "sun/font/SunFontManager/badFonts": java_util_ArrayList;
    public "sun/font/SunFontManager/fontPath": java_lang_String;
    public "sun/font/SunFontManager/fontConfig": sun_awt_FontConfiguration;
    public "sun/font/SunFontManager/discoveredAllFonts": number;
    public static "sun/font/SunFontManager/ttFilter": java_io_FilenameFilter;
    public static "sun/font/SunFontManager/t1Filter": java_io_FilenameFilter;
    public "sun/font/SunFontManager/allFonts": JVMArray<java_awt_Font>;
    public "sun/font/SunFontManager/allFamilies": JVMArray<java_lang_String>;
    public "sun/font/SunFontManager/lastDefaultLocale": java_util_Locale;
    public static "sun/font/SunFontManager/noType1Font": number;
    public static "sun/font/SunFontManager/STR_ARRAY": JVMArray<java_lang_String>;
    public "sun/font/SunFontManager/usePlatformFontMetrics": number;
    public "sun/font/SunFontManager/deferredFontFiles": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/SunFontManager/initialisedFonts": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/SunFontManager/fontToFileMap": java_util_HashMap;
    public "sun/font/SunFontManager/fontToFamilyNameMap": java_util_HashMap;
    public "sun/font/SunFontManager/familyToFontListMap": java_util_HashMap;
    public "sun/font/SunFontManager/pathDirs": JVMArray<java_lang_String>;
    public "sun/font/SunFontManager/haveCheckedUnreferencedFontFiles": number;
    public static "sun/font/SunFontManager/platformFontMap": java_util_HashMap;
    public "sun/font/SunFontManager/fontNameCache": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/SunFontManager/fileCloser": java_lang_Thread;
    public "sun/font/SunFontManager/tmpFontFiles": java_util_Vector;
    public static "sun/font/SunFontManager/altJAFontKey": java_lang_Object;
    public static "sun/font/SunFontManager/localeFontKey": java_lang_Object;
    public static "sun/font/SunFontManager/proportionalFontKey": java_lang_Object;
    public "sun/font/SunFontManager/_usingPerAppContextComposites": number;
    public "sun/font/SunFontManager/_usingAlternateComposites": number;
    public static "sun/font/SunFontManager/gAltJAFont": number;
    public "sun/font/SunFontManager/gLocalePref": number;
    public "sun/font/SunFontManager/gPropPref": number;
    public static "sun/font/SunFontManager/installedNames": java_util_HashSet;
    public static "sun/font/SunFontManager/regFamilyKey": java_lang_Object;
    public static "sun/font/SunFontManager/regFullNameKey": java_lang_Object;
    public "sun/font/SunFontManager/createdByFamilyName": java_util_Hashtable;
    public "sun/font/SunFontManager/createdByFullName": java_util_Hashtable;
    public "sun/font/SunFontManager/fontsAreRegistered": number;
    public "sun/font/SunFontManager/fontsAreRegisteredPerAppContext": number;
    public static "sun/font/SunFontManager/systemLocale": java_util_Locale;
    public static "getInstance()Lsun/font/SunFontManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_SunFontManager) => void): void;
    public static "sun/font/SunFontManager/getInstance()Lsun/font/SunFontManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_SunFontManager) => void): void;
    public "getTrueTypeFilter()Ljava/io/FilenameFilter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FilenameFilter) => void): void;
    public "sun/font/SunFontManager/getTrueTypeFilter()Ljava/io/FilenameFilter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FilenameFilter) => void): void;
    public "getType1Filter()Ljava/io/FilenameFilter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FilenameFilter) => void): void;
    public "sun/font/SunFontManager/getType1Filter()Ljava/io/FilenameFilter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FilenameFilter) => void): void;
    public "usingPerAppContextComposites()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/usingPerAppContextComposites()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "initJREFontMap()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/initJREFontMap()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEUDCFont()Lsun/font/TrueTypeFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_TrueTypeFont) => void): void;
    public "sun/font/SunFontManager/getEUDCFont()Lsun/font/TrueTypeFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_TrueTypeFont) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/SunFontManager/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNewComposite(Ljava/lang/String;ILsun/font/Font2DHandle;)Lsun/font/Font2DHandle;"(thread: JVMThread, args: [java_lang_String, number, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2DHandle) => void): void;
    public "sun/font/SunFontManager/getNewComposite(Ljava/lang/String;ILsun/font/Font2DHandle;)Lsun/font/Font2DHandle;"(thread: JVMThread, args: [java_lang_String, number, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2DHandle) => void): void;
    public "registerCompositeFont(Ljava/lang/String;[Ljava/lang/String;[Ljava/lang/String;I[I[IZ)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, JVMArray<java_lang_String>, number, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerCompositeFont(Ljava/lang/String;[Ljava/lang/String;[Ljava/lang/String;I[I[IZ)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, JVMArray<java_lang_String>, number, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "registerCompositeFont(Ljava/lang/String;[Ljava/lang/String;[Ljava/lang/String;I[I[IZLjava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, JVMArray<java_lang_String>, number, JVMArray<number>, JVMArray<number>, number, java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/SunFontManager/registerCompositeFont(Ljava/lang/String;[Ljava/lang/String;[Ljava/lang/String;I[I[IZLjava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, JVMArray<java_lang_String>, number, JVMArray<number>, JVMArray<number>, number, java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "addCompositeToFontList(Lsun/font/CompositeFont;I)V"(thread: JVMThread, args: [sun_font_CompositeFont, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/addCompositeToFontList(Lsun/font/CompositeFont;I)V"(thread: JVMThread, args: [sun_font_CompositeFont, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addToFontList(Lsun/font/PhysicalFont;I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [sun_font_PhysicalFont, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/SunFontManager/addToFontList(Lsun/font/PhysicalFont;I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [sun_font_PhysicalFont, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "getRegisteredFonts()[Lsun/font/Font2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_Font2D>) => void): void;
    public "sun/font/SunFontManager/getRegisteredFonts()[Lsun/font/Font2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_Font2D>) => void): void;
    public "getPhysicalFonts()[Lsun/font/PhysicalFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_PhysicalFont>) => void): void;
    public "sun/font/SunFontManager/getPhysicalFonts()[Lsun/font/PhysicalFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_PhysicalFont>) => void): void;
    public "initialiseDeferredFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/initialiseDeferredFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerDeferredJREFonts(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerDeferredJREFonts(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isDeferredFont(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/isDeferredFont(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "findJREDeferredFont(Ljava/lang/String;I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/SunFontManager/findJREDeferredFont(Ljava/lang/String;I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "findOtherDeferredFont(Ljava/lang/String;I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/SunFontManager/findOtherDeferredFont(Ljava/lang/String;I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "findDeferredFont(Ljava/lang/String;I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/SunFontManager/findDeferredFont(Ljava/lang/String;I)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "registerDeferredFont(Ljava/lang/String;Ljava/lang/String;[Ljava/lang/String;IZI)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<java_lang_String>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerDeferredFont(Ljava/lang/String;Ljava/lang/String;[Ljava/lang/String;IZI)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<java_lang_String>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "initialiseDeferredFont(Ljava/lang/String;)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/SunFontManager/initialiseDeferredFont(Ljava/lang/String;)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "isRegisteredFontFile(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/isRegisteredFontFile(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRegisteredFontFile(Ljava/lang/String;)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/SunFontManager/getRegisteredFontFile(Ljava/lang/String;)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "registerFontFile(Ljava/lang/String;[Ljava/lang/String;IZI)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/SunFontManager/registerFontFile(Ljava/lang/String;[Ljava/lang/String;IZI)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "registerFonts([Ljava/lang/String;[[Ljava/lang/String;IIZIZ)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<JVMArray<java_lang_String>>, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerFonts([Ljava/lang/String;[[Ljava/lang/String;IIZIZ)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<JVMArray<java_lang_String>>, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDefaultPhysicalFont()Lsun/font/PhysicalFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/SunFontManager/getDefaultPhysicalFont()Lsun/font/PhysicalFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "getDefaultLogicalFont(I)Lsun/font/Font2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/SunFontManager/getDefaultLogicalFont(I)Lsun/font/Font2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public static "dotStyleStr(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/font/SunFontManager/dotStyleStr(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "populateFontFileNameMap(Ljava/util/HashMap;Ljava/util/HashMap;Ljava/util/HashMap;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_HashMap, java_util_HashMap, java_util_HashMap, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/populateFontFileNameMap(Ljava/util/HashMap;Ljava/util/HashMap;Ljava/util/HashMap;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_HashMap, java_util_HashMap, java_util_HashMap, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontFilesFromPath(Z)[Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/SunFontManager/getFontFilesFromPath(Z)[Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "resolveWindowsFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/resolveWindowsFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkForUnreferencedFontFiles()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/checkForUnreferencedFontFiles()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveFontFiles(Ljava/util/HashSet;Ljava/util/ArrayList;)V"(thread: JVMThread, args: [java_util_HashSet, java_util_ArrayList], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/resolveFontFiles(Ljava/util/HashSet;Ljava/util/ArrayList;)V"(thread: JVMThread, args: [java_util_HashSet, java_util_ArrayList], cb?: (e?: java_lang_Throwable) => void): void;
    public "populateHardcodedFileNameMap()Ljava/util/HashMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap) => void): void;
    public "sun/font/SunFontManager/populateHardcodedFileNameMap()Ljava/util/HashMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap) => void): void;
    public "findFontFromPlatformMap(Ljava/lang/String;I)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/SunFontManager/findFontFromPlatformMap(Ljava/lang/String;I)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "getFullNameToFileMap()Ljava/util/HashMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap) => void): void;
    public "sun/font/SunFontManager/getFullNameToFileMap()Ljava/util/HashMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap) => void): void;
    public "logPlatformFontInfo()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/logPlatformFontInfo()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontNamesFromPlatform()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/SunFontManager/getFontNamesFromPlatform()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "gotFontsFromPlatform()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/gotFontsFromPlatform()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFileNameForFontName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/SunFontManager/getFileNameForFontName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "registerFontFile(Ljava/lang/String;)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/SunFontManager/registerFontFile(Ljava/lang/String;)Lsun/font/PhysicalFont;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "registerOtherFontFiles(Ljava/util/HashSet;)V"(thread: JVMThread, args: [java_util_HashSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerOtherFontFiles(Ljava/util/HashSet;)V"(thread: JVMThread, args: [java_util_HashSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFamilyNamesFromPlatform(Ljava/util/TreeMap;Ljava/util/Locale;)Z"(thread: JVMThread, args: [java_util_TreeMap, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/getFamilyNamesFromPlatform(Ljava/util/TreeMap;Ljava/util/Locale;)Z"(thread: JVMThread, args: [java_util_TreeMap, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPathName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/SunFontManager/getPathName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "findFontFromPlatform(Ljava/lang/String;I)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/SunFontManager/findFontFromPlatform(Ljava/lang/String;I)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "findFont2D(Ljava/lang/String;II)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/SunFontManager/findFont2D(Ljava/lang/String;II)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "usePlatformFontMetrics()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/usePlatformFontMetrics()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumFonts()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/getNumFonts()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "fontSupportsEncoding(Ljava/awt/Font;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_awt_Font, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/SunFontManager/fontSupportsEncoding(Ljava/awt/Font;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_awt_Font, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFontPath(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/SunFontManager/getFontPath(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "createFont2D(Ljava/io/File;IZLsun/font/CreatedFontTracker;)Lsun/font/Font2D;"(thread: JVMThread, args: [java_io_File, number, number, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/SunFontManager/createFont2D(Ljava/io/File;IZLsun/font/CreatedFontTracker;)Lsun/font/Font2D;"(thread: JVMThread, args: [java_io_File, number, number, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "getFullNameByFileName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/SunFontManager/getFullNameByFileName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "deRegisterBadFont(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/deRegisterBadFont(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceFont(Lsun/font/PhysicalFont;Lsun/font/PhysicalFont;)V"(thread: JVMThread, args: [sun_font_PhysicalFont, sun_font_PhysicalFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/replaceFont(Lsun/font/PhysicalFont;Lsun/font/PhysicalFont;)V"(thread: JVMThread, args: [sun_font_PhysicalFont, sun_font_PhysicalFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "loadLocaleNames()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/loadLocaleNames()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "findFont2DAllLocales(Ljava/lang/String;I)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/SunFontManager/findFont2DAllLocales(Ljava/lang/String;I)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "maybeUsingAlternateCompositeFonts()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/maybeUsingAlternateCompositeFonts()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "usingAlternateCompositeFonts()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/usingAlternateCompositeFonts()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "maybeMultiAppContext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/SunFontManager/maybeMultiAppContext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "useAlternateFontforJALocales()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/useAlternateFontforJALocales()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "usingAlternateFontforJALocales()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/usingAlternateFontforJALocales()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "preferLocaleFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/preferLocaleFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "preferProportionalFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/preferProportionalFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstalledNames()Ljava/util/HashSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashSet) => void): void;
    public static "sun/font/SunFontManager/getInstalledNames()Ljava/util/HashSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashSet) => void): void;
    public "registerFont(Ljava/awt/Font;)Z"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/registerFont(Ljava/awt/Font;)Z"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeFromCache(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/removeFromCache(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCreatedFontFamilyNames()Ljava/util/TreeMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap) => void): void;
    public "sun/font/SunFontManager/getCreatedFontFamilyNames()Ljava/util/TreeMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap) => void): void;
    public "getCreatedFonts()[Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Font>) => void): void;
    public "sun/font/SunFontManager/getCreatedFonts()[Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Font>) => void): void;
    public "getPlatformFontDirs(Z)[Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/SunFontManager/getPlatformFontDirs(Z)[Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getDefaultPlatformFont()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/SunFontManager/getDefaultPlatformFont()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "addDirFonts(Ljava/lang/String;Ljava/io/File;Ljava/io/FilenameFilter;IZIZZ)V"(thread: JVMThread, args: [java_lang_String, java_io_File, java_io_FilenameFilter, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/addDirFonts(Ljava/lang/String;Ljava/io/File;Ljava/io/FilenameFilter;IZIZZ)V"(thread: JVMThread, args: [java_lang_String, java_io_File, java_io_FilenameFilter, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNativeNames(Ljava/lang/String;Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/SunFontManager/getNativeNames(Ljava/lang/String;Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getFileNameFromPlatformName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/SunFontManager/getFileNameFromPlatformName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFontConfiguration()Lsun/awt/FontConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public "sun/font/SunFontManager/getFontConfiguration()Lsun/awt/FontConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public "getPlatformFontPath(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/SunFontManager/getPlatformFontPath(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "isOpenJDK()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/SunFontManager/isOpenJDK()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "loadFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/loadFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerFontDirs(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerFontDirs(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerFontsOnPath(Ljava/lang/String;ZIZZ)V"(thread: JVMThread, args: [java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerFontsOnPath(Ljava/lang/String;ZIZZ)V"(thread: JVMThread, args: [java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerFontsInDir(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerFontsInDir(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerFontsInDir(Ljava/lang/String;ZIZZ)V"(thread: JVMThread, args: [java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerFontsInDir(Ljava/lang/String;ZIZZ)V"(thread: JVMThread, args: [java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerFontDir(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerFontDir(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDefaultFontFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/SunFontManager/getDefaultFontFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "initDefaultFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/initDefaultFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "useAbsoluteFontFileNames()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/useAbsoluteFontFileNames()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createFontConfiguration()Lsun/awt/FontConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public "sun/font/SunFontManager/createFontConfiguration()Lsun/awt/FontConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public "createFontConfiguration(ZZ)Lsun/awt/FontConfiguration;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public "sun/font/SunFontManager/createFontConfiguration(ZZ)Lsun/awt/FontConfiguration;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public "getDefaultFontFaceName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/SunFontManager/getDefaultFontFaceName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "loadFontFiles()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/loadFontFiles()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "initCompositeFonts(Lsun/awt/FontConfiguration;Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [sun_awt_FontConfiguration, java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/initCompositeFonts(Lsun/awt/FontConfiguration;Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [sun_awt_FontConfiguration, java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "addFontToPlatformFontPath(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/addFontToPlatformFontPath(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerFontFile(Ljava/lang/String;[Ljava/lang/String;IZ)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerFontFile(Ljava/lang/String;[Ljava/lang/String;IZ)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "registerPlatformFontsUsedByFontConfiguration()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/registerPlatformFontsUsedByFontConfiguration()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addToMissingFontFileList(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/addToMissingFontFileList(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isNameForRegisteredFile(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/SunFontManager/isNameForRegisteredFile(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createCompositeFonts(Ljava/util/concurrent/ConcurrentHashMap;ZZ)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/createCompositeFonts(Ljava/util/concurrent/ConcurrentHashMap;ZZ)V"(thread: JVMThread, args: [java_util_concurrent_ConcurrentHashMap, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAllInstalledFonts()[Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Font>) => void): void;
    public "sun/font/SunFontManager/getAllInstalledFonts()[Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Font>) => void): void;
    public "getInstalledFontFamilyNames(Ljava/util/Locale;)[Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/SunFontManager/getInstalledFontFamilyNames(Ljava/util/Locale;)[Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "addNativeFontFamilyNames(Ljava/util/TreeMap;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_TreeMap, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/addNativeFontFamilyNames(Ljava/util/TreeMap;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_TreeMap, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "register1dot0Fonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/register1dot0Fonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getJREFontFamilyNames(Ljava/util/TreeMap;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_TreeMap, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/getJREFontFamilyNames(Ljava/util/TreeMap;Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_TreeMap, java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getSystemStartupLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public static "sun/font/SunFontManager/getSystemStartupLocale()Ljava/util/Locale;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Locale) => void): void;
    public "addToPool(Lsun/font/FileFont;)V"(thread: JVMThread, args: [sun_font_FileFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunFontManager/addToPool(Lsun/font/FileFont;)V"(thread: JVMThread, args: [sun_font_FileFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontConfigFUIR(Ljava/lang/String;II)Ljavax/swing/plaf/FontUIResource;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: javax_swing_plaf_FontUIResource) => void): void;
    public "sun/font/SunFontManager/getFontConfigFUIR(Ljava/lang/String;II)Ljavax/swing/plaf/FontUIResource;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: javax_swing_plaf_FontUIResource) => void): void;
    public "lambda$createFont2D$0(Ljava/lang/Runnable;)Ljava/lang/Void;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "sun/font/SunFontManager/lambda$createFont2D$0(Ljava/lang/Runnable;)Ljava/lang/Void;"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "access$200()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/SunFontManager/access$200()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$302(Lsun/font/SunFontManager;Ljava/util/ArrayList;)Ljava/util/ArrayList;"(thread: JVMThread, args: [sun_font_SunFontManager, java_util_ArrayList], cb?: (e?: java_lang_Throwable, rv?: java_util_ArrayList) => void): void;
    public static "sun/font/SunFontManager/access$302(Lsun/font/SunFontManager;Ljava/util/ArrayList;)Ljava/util/ArrayList;"(thread: JVMThread, args: [sun_font_SunFontManager, java_util_ArrayList], cb?: (e?: java_lang_Throwable, rv?: java_util_ArrayList) => void): void;
    public static "access$300(Lsun/font/SunFontManager;)Ljava/util/ArrayList;"(thread: JVMThread, args: [sun_font_SunFontManager], cb?: (e?: java_lang_Throwable, rv?: java_util_ArrayList) => void): void;
    public static "sun/font/SunFontManager/access$300(Lsun/font/SunFontManager;)Ljava/util/ArrayList;"(thread: JVMThread, args: [sun_font_SunFontManager], cb?: (e?: java_lang_Throwable, rv?: java_util_ArrayList) => void): void;
    public static "access$402(Lsun/font/SunFontManager;Lsun/awt/FontConfiguration;)Lsun/awt/FontConfiguration;"(thread: JVMThread, args: [sun_font_SunFontManager, sun_awt_FontConfiguration], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public static "sun/font/SunFontManager/access$402(Lsun/font/SunFontManager;Lsun/awt/FontConfiguration;)Lsun/awt/FontConfiguration;"(thread: JVMThread, args: [sun_font_SunFontManager, sun_awt_FontConfiguration], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public static "access$502(Lsun/font/SunFontManager;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [sun_font_SunFontManager, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/font/SunFontManager/access$502(Lsun/font/SunFontManager;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [sun_font_SunFontManager, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$602(Lsun/font/SunFontManager;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [sun_font_SunFontManager, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/font/SunFontManager/access$602(Lsun/font/SunFontManager;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [sun_font_SunFontManager, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$400(Lsun/font/SunFontManager;)Lsun/awt/FontConfiguration;"(thread: JVMThread, args: [sun_font_SunFontManager], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public static "sun/font/SunFontManager/access$400(Lsun/font/SunFontManager;)Lsun/awt/FontConfiguration;"(thread: JVMThread, args: [sun_font_SunFontManager], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public static "access$700(Lsun/font/SunFontManager;Lsun/awt/FontConfiguration;Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [sun_font_SunFontManager, sun_awt_FontConfiguration, java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/SunFontManager/access$700(Lsun/font/SunFontManager;Lsun/awt/FontConfiguration;Ljava/util/concurrent/ConcurrentHashMap;)V"(thread: JVMThread, args: [sun_font_SunFontManager, sun_awt_FontConfiguration, java_util_concurrent_ConcurrentHashMap], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$900(Lsun/font/SunFontManager;)[Ljava/lang/String;"(thread: JVMThread, args: [sun_font_SunFontManager], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "sun/font/SunFontManager/access$900(Lsun/font/SunFontManager;)[Ljava/lang/String;"(thread: JVMThread, args: [sun_font_SunFontManager], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "access$1000()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "sun/font/SunFontManager/access$1000()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "access$1100(Lsun/font/SunFontManager;Ljava/lang/String;ZIZZ)V"(thread: JVMThread, args: [sun_font_SunFontManager, java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/SunFontManager/access$1100(Lsun/font/SunFontManager;Ljava/lang/String;ZIZZ)V"(thread: JVMThread, args: [sun_font_SunFontManager, java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$1202(Lsun/font/SunFontManager;Z)Z"(thread: JVMThread, args: [sun_font_SunFontManager, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/SunFontManager/access$1202(Lsun/font/SunFontManager;Z)Z"(thread: JVMThread, args: [sun_font_SunFontManager, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/SunFontManager/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class javax_swing_plaf_FontUIResource extends java_awt_Font implements javax_swing_plaf_UIResource {
    public "<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/plaf/FontUIResource/<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
    public "javax/swing/plaf/FontUIResource/<init>(Ljava/awt/Font;)V"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface javax_swing_plaf_UIResource extends java_lang_Object {
  }
  export class sun_font_TrueTypeFont extends sun_font_FileFont {
    public static "sun/font/TrueTypeFont/cmapTag": number;
    public static "sun/font/TrueTypeFont/glyfTag": number;
    public static "sun/font/TrueTypeFont/headTag": number;
    public static "sun/font/TrueTypeFont/hheaTag": number;
    public static "sun/font/TrueTypeFont/hmtxTag": number;
    public static "sun/font/TrueTypeFont/locaTag": number;
    public static "sun/font/TrueTypeFont/maxpTag": number;
    public static "sun/font/TrueTypeFont/nameTag": number;
    public static "sun/font/TrueTypeFont/postTag": number;
    public static "sun/font/TrueTypeFont/os_2Tag": number;
    public static "sun/font/TrueTypeFont/GDEFTag": number;
    public static "sun/font/TrueTypeFont/GPOSTag": number;
    public static "sun/font/TrueTypeFont/GSUBTag": number;
    public static "sun/font/TrueTypeFont/mortTag": number;
    public static "sun/font/TrueTypeFont/fdscTag": number;
    public static "sun/font/TrueTypeFont/fvarTag": number;
    public static "sun/font/TrueTypeFont/featTag": number;
    public static "sun/font/TrueTypeFont/EBLCTag": number;
    public static "sun/font/TrueTypeFont/gaspTag": number;
    public static "sun/font/TrueTypeFont/ttcfTag": number;
    public static "sun/font/TrueTypeFont/v1ttTag": number;
    public static "sun/font/TrueTypeFont/trueTag": number;
    public static "sun/font/TrueTypeFont/ottoTag": number;
    public static "sun/font/TrueTypeFont/MS_PLATFORM_ID": number;
    public static "sun/font/TrueTypeFont/ENGLISH_LOCALE_ID": number;
    public static "sun/font/TrueTypeFont/FAMILY_NAME_ID": number;
    public static "sun/font/TrueTypeFont/FULL_NAME_ID": number;
    public static "sun/font/TrueTypeFont/POSTSCRIPT_NAME_ID": number;
    public static "sun/font/TrueTypeFont/US_LCID": number;
    public static "sun/font/TrueTypeFont/lcidMap": java_util_Map;
    public "sun/font/TrueTypeFont/disposerRecord": sun_font_TrueTypeFont$TTDisposerRecord;
    public "sun/font/TrueTypeFont/fontIndex": number;
    public "sun/font/TrueTypeFont/directoryCount": number;
    public "sun/font/TrueTypeFont/directoryOffset": number;
    public "sun/font/TrueTypeFont/numTables": number;
    public "sun/font/TrueTypeFont/tableDirectory": JVMArray<sun_font_TrueTypeFont$DirectoryEntry>;
    public "sun/font/TrueTypeFont/supportsJA": number;
    public "sun/font/TrueTypeFont/supportsCJK": number;
    public "sun/font/TrueTypeFont/nameLocale": java_util_Locale;
    public "sun/font/TrueTypeFont/localeFamilyName": java_lang_String;
    public "sun/font/TrueTypeFont/localeFullName": java_lang_String;
    public static "sun/font/TrueTypeFont/TTCHEADERSIZE": number;
    public static "sun/font/TrueTypeFont/DIRECTORYHEADERSIZE": number;
    public static "sun/font/TrueTypeFont/DIRECTORYENTRYSIZE": number;
    public static "sun/font/TrueTypeFont/encoding_mapping": JVMArray<java_lang_String>;
    public static "sun/font/TrueTypeFont/languages": JVMArray<JVMArray<java_lang_String>>;
    public static "sun/font/TrueTypeFont/codePages": JVMArray<java_lang_String>;
    public static "sun/font/TrueTypeFont/defaultCodePage": java_lang_String;
    public static "sun/font/TrueTypeFont/reserved_bits1": number;
    public static "sun/font/TrueTypeFont/reserved_bits2": number;
    public "sun/font/TrueTypeFont/fontWidth": number;
    public "sun/font/TrueTypeFont/fontWeight": number;
    public static "sun/font/TrueTypeFont/fsSelectionItalicBit": number;
    public static "sun/font/TrueTypeFont/fsSelectionBoldBit": number;
    public static "sun/font/TrueTypeFont/fsSelectionRegularBit": number;
    public "sun/font/TrueTypeFont/stSize": number;
    public "sun/font/TrueTypeFont/stPos": number;
    public "sun/font/TrueTypeFont/ulSize": number;
    public "sun/font/TrueTypeFont/ulPos": number;
    public "sun/font/TrueTypeFont/gaspTable": JVMArray<number>;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;IZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/<init>(Ljava/lang/String;Ljava/lang/Object;IZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkUseNatives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/checkUseNatives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "open()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "sun/font/TrueTypeFont/open()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readBlock(Ljava/nio/ByteBuffer;II)I"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/readBlock(Ljava/nio/ByteBuffer;II)I"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readBlock(II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/font/TrueTypeFont/readBlock(II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "readBytes(II)[B"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/TrueTypeFont/readBytes(II)[B"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "verify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/verify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "init(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/init(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getCodePage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/font/TrueTypeFont/getCodePage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "supportsEncoding(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/supportsEncoding(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setCJKSupport(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/setCJKSupport(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "supportsJA()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/supportsJA()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTableBuffer(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/font/TrueTypeFont/getTableBuffer(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "getLayoutTableCache()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/TrueTypeFont/getLayoutTableCache()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTableBytes(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/TrueTypeFont/getTableBytes(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getTableSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/getTableSize(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTableOffset(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/getTableOffset(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDirectoryEntry(I)Lsun/font/TrueTypeFont$DirectoryEntry;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_TrueTypeFont$DirectoryEntry) => void): void;
    public "sun/font/TrueTypeFont/getDirectoryEntry(I)Lsun/font/TrueTypeFont$DirectoryEntry;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_TrueTypeFont$DirectoryEntry) => void): void;
    public "useEmbeddedBitmapsForSize(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/useEmbeddedBitmapsForSize(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFullName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/TrueTypeFont/getFullName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setStyle()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/setStyle()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/getWeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setStyle(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/setStyle(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "setStrikethroughMetrics(Ljava/nio/ByteBuffer;I)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/setStrikethroughMetrics(Ljava/nio/ByteBuffer;I)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setUnderlineMetrics(Ljava/nio/ByteBuffer;I)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/setUnderlineMetrics(Ljava/nio/ByteBuffer;I)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getStyleMetrics(F[FI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/getStyleMetrics(F[FI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "makeString([BIS)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/TrueTypeFont/makeString([BIS)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "initNames()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/initNames()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lookupName(SI)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/TrueTypeFont/lookupName(SI)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFontCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/getFontCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getScaler()Lsun/font/FontScaler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontScaler) => void): void;
    public "sun/font/TrueTypeFont/getScaler()Lsun/font/FontScaler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontScaler) => void): void;
    public "getPostscriptName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/TrueTypeFont/getPostscriptName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFontName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/TrueTypeFont/getFontName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "addLCIDMapEntry(Ljava/util/Map;Ljava/lang/String;S)V"(thread: JVMThread, args: [java_util_Map, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/TrueTypeFont/addLCIDMapEntry(Ljava/util/Map;Ljava/lang/String;S)V"(thread: JVMThread, args: [java_util_Map, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createLCIDMap()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/TrueTypeFont/createLCIDMap()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getLCIDFromLocale(Ljava/util/Locale;)S"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/TrueTypeFont/getLCIDFromLocale(Ljava/util/Locale;)S"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFamilyName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/TrueTypeFont/getFamilyName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getMapper()Lsun/font/CharToGlyphMapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CharToGlyphMapper) => void): void;
    public "sun/font/TrueTypeFont/getMapper()Lsun/font/CharToGlyphMapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CharToGlyphMapper) => void): void;
    public "initAllNames(ILjava/util/HashSet;)V"(thread: JVMThread, args: [number, java_util_HashSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont/initAllNames(ILjava/util/HashSet;)V"(thread: JVMThread, args: [number, java_util_HashSet], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAllFamilyNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/TrueTypeFont/getAllFamilyNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getAllFullNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/TrueTypeFont/getAllFullNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/TrueTypeFont/getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getGaspTable()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/TrueTypeFont/getGaspTable()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "useAAForPtSize(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/useAAForPtSize(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasSupplementaryChars()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/TrueTypeFont/hasSupplementaryChars()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/TrueTypeFont/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/TrueTypeFont/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_FontScaler extends java_lang_Object implements sun_java2d_DisposerRecord {
    public static "sun/font/FontScaler/nullScaler": sun_font_FontScaler;
    public static "sun/font/FontScaler/scalerConstructor": java_lang_reflect_Constructor;
    public "sun/font/FontScaler/font": java_lang_ref_WeakReference;
    public "sun/font/FontScaler/nativeScaler": Long;
    public "sun/font/FontScaler/disposed": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontScaler/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getScaler(Lsun/font/Font2D;IZI)Lsun/font/FontScaler;"(thread: JVMThread, args: [sun_font_Font2D, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontScaler) => void): void;
    public static "sun/font/FontScaler/getScaler(Lsun/font/Font2D;IZI)Lsun/font/FontScaler;"(thread: JVMThread, args: [sun_font_Font2D, number, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontScaler) => void): void;
    public static "getNullScaler()Lsun/font/FontScaler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontScaler) => void): void;
    public static "sun/font/FontScaler/getNullScaler()Lsun/font/FontScaler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontScaler) => void): void;
    public "getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/FontScaler/getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontScaler/getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontScaler/getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FontScaler/getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/FontScaler/getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FontScaler/getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FontScaler/getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontScaler/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontScaler/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontScaler/getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphCode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontScaler/getGlyphCode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLayoutTableCache()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FontScaler/getLayoutTableCache()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/FontScaler/getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FontScaler/getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "createScalerContext([DIIFFZ)J"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FontScaler/createScalerContext([DIIFFZ)J"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "invalidateScalerContext(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontScaler/invalidateScalerContext(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FontScaler/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_TrueTypeFont$DirectoryEntry extends java_lang_Object {
    public "sun/font/TrueTypeFont$DirectoryEntry/tag": number;
    public "sun/font/TrueTypeFont$DirectoryEntry/offset": number;
    public "sun/font/TrueTypeFont$DirectoryEntry/length": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont$DirectoryEntry/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_TrueTypeFont$TTDisposerRecord extends java_lang_Object implements sun_java2d_DisposerRecord {
    public "sun/font/TrueTypeFont$TTDisposerRecord/channel": java_nio_channels_FileChannel;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont$TTDisposerRecord/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont$TTDisposerRecord/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/TrueTypeFont$1;)V"(thread: JVMThread, args: [sun_font_TrueTypeFont$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont$TTDisposerRecord/<init>(Lsun/font/TrueTypeFont$1;)V"(thread: JVMThread, args: [sun_font_TrueTypeFont$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_TrueTypeFont$1 extends java_lang_Object implements java_security_PrivilegedAction {
    public "sun/font/TrueTypeFont$1/this$0": sun_font_TrueTypeFont;
    public "<init>(Lsun/font/TrueTypeFont;)V"(thread: JVMThread, args: [sun_font_TrueTypeFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TrueTypeFont$1/<init>(Lsun/font/TrueTypeFont;)V"(thread: JVMThread, args: [sun_font_TrueTypeFont], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/TrueTypeFont$1/run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_awt_FontConfiguration extends java_lang_Object {
    public static "sun/awt/FontConfiguration/osVersion": java_lang_String;
    public static "sun/awt/FontConfiguration/osName": java_lang_String;
    public static "sun/awt/FontConfiguration/encoding": java_lang_String;
    public static "sun/awt/FontConfiguration/startupLocale": java_util_Locale;
    public static "sun/awt/FontConfiguration/localeMap": java_util_Hashtable;
    public static "sun/awt/FontConfiguration/fontConfig": sun_awt_FontConfiguration;
    public static "sun/awt/FontConfiguration/logger": sun_util_logging_PlatformLogger;
    public static "sun/awt/FontConfiguration/isProperties": number;
    public "sun/awt/FontConfiguration/fontManager": sun_font_SunFontManager;
    public "sun/awt/FontConfiguration/preferLocaleFonts": number;
    public "sun/awt/FontConfiguration/preferPropFonts": number;
    public "sun/awt/FontConfiguration/fontConfigFile": java_io_File;
    public "sun/awt/FontConfiguration/foundOsSpecificFile": number;
    public "sun/awt/FontConfiguration/inited": number;
    public "sun/awt/FontConfiguration/javaLib": java_lang_String;
    public static "sun/awt/FontConfiguration/stringIDNum": number;
    public static "sun/awt/FontConfiguration/stringIDs": JVMArray<number>;
    public static "sun/awt/FontConfiguration/stringTable": java_lang_StringBuilder;
    public static "sun/awt/FontConfiguration/verbose": number;
    public "sun/awt/FontConfiguration/initELC": number;
    public "sun/awt/FontConfiguration/initLocale": java_util_Locale;
    public "sun/awt/FontConfiguration/initEncoding": java_lang_String;
    public "sun/awt/FontConfiguration/alphabeticSuffix": java_lang_String;
    public "sun/awt/FontConfiguration/compFontNameIDs": JVMArray<JVMArray<JVMArray<number>>>;
    public "sun/awt/FontConfiguration/compExclusions": JVMArray<JVMArray<JVMArray<number>>>;
    public "sun/awt/FontConfiguration/compCoreNum": JVMArray<number>;
    public "sun/awt/FontConfiguration/coreFontNameIDs": java_util_Set;
    public "sun/awt/FontConfiguration/fallbackFontNameIDs": java_util_Set;
    public static "sun/awt/FontConfiguration/NUM_FONTS": number;
    public static "sun/awt/FontConfiguration/NUM_STYLES": number;
    public static "sun/awt/FontConfiguration/fontNames": JVMArray<java_lang_String>;
    public static "sun/awt/FontConfiguration/publicFontNames": JVMArray<java_lang_String>;
    public static "sun/awt/FontConfiguration/styleNames": JVMArray<java_lang_String>;
    public static "sun/awt/FontConfiguration/installedFallbackFontFiles": JVMArray<java_lang_String>;
    public "sun/awt/FontConfiguration/reorderMap": java_util_HashMap;
    public "sun/awt/FontConfiguration/charsetRegistry": java_util_Hashtable;
    public "sun/awt/FontConfiguration/fontDescriptors": JVMArray<JVMArray<JVMArray<sun_awt_FontDescriptor>>>;
    public "sun/awt/FontConfiguration/existsMap": java_util_HashMap;
    public "sun/awt/FontConfiguration/numCoreFonts": number;
    public "sun/awt/FontConfiguration/componentFonts": JVMArray<java_lang_String>;
    public "sun/awt/FontConfiguration/filenamesMap": java_util_HashMap;
    public "sun/awt/FontConfiguration/coreFontFileNames": java_util_HashSet;
    public static "sun/awt/FontConfiguration/HEAD_LENGTH": number;
    public static "sun/awt/FontConfiguration/INDEX_scriptIDs": number;
    public static "sun/awt/FontConfiguration/INDEX_scriptFonts": number;
    public static "sun/awt/FontConfiguration/INDEX_elcIDs": number;
    public static "sun/awt/FontConfiguration/INDEX_sequences": number;
    public static "sun/awt/FontConfiguration/INDEX_fontfileNameIDs": number;
    public static "sun/awt/FontConfiguration/INDEX_componentFontNameIDs": number;
    public static "sun/awt/FontConfiguration/INDEX_filenames": number;
    public static "sun/awt/FontConfiguration/INDEX_awtfontpaths": number;
    public static "sun/awt/FontConfiguration/INDEX_exclusions": number;
    public static "sun/awt/FontConfiguration/INDEX_proportionals": number;
    public static "sun/awt/FontConfiguration/INDEX_scriptFontsMotif": number;
    public static "sun/awt/FontConfiguration/INDEX_alphabeticSuffix": number;
    public static "sun/awt/FontConfiguration/INDEX_stringIDs": number;
    public static "sun/awt/FontConfiguration/INDEX_stringTable": number;
    public static "sun/awt/FontConfiguration/INDEX_TABLEEND": number;
    public static "sun/awt/FontConfiguration/INDEX_fallbackScripts": number;
    public static "sun/awt/FontConfiguration/INDEX_appendedfontpath": number;
    public static "sun/awt/FontConfiguration/INDEX_version": number;
    public static "sun/awt/FontConfiguration/head": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_scriptIDs": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_scriptFonts": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_elcIDs": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_sequences": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_fontfileNameIDs": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_componentFontNameIDs": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_filenames": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_awtfontpaths": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_exclusions": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_proportionals": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_scriptFontsMotif": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_alphabeticSuffix": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_stringIDs": JVMArray<number>;
    public static "sun/awt/FontConfiguration/table_stringTable": JVMArray<number>;
    public "sun/awt/FontConfiguration/reorderScripts": java_util_HashMap;
    public static "sun/awt/FontConfiguration/stringCache": JVMArray<java_lang_String>;
    public static "sun/awt/FontConfiguration/EMPTY_INT_ARRAY": JVMArray<number>;
    public static "sun/awt/FontConfiguration/EMPTY_STRING_ARRAY": JVMArray<java_lang_String>;
    public static "sun/awt/FontConfiguration/EMPTY_SHORT_ARRAY": JVMArray<number>;
    public static "sun/awt/FontConfiguration/UNDEFINED_COMPONENT_FONT": java_lang_String;
    public static "sun/awt/FontConfiguration/$assertionsDisabled": number;
    public "<init>(Lsun/font/SunFontManager;)V"(thread: JVMThread, args: [sun_font_SunFontManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/<init>(Lsun/font/SunFontManager;)V"(thread: JVMThread, args: [sun_font_SunFontManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "init()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontConfiguration/init()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Lsun/font/SunFontManager;ZZ)V"(thread: JVMThread, args: [sun_font_SunFontManager, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/<init>(Lsun/font/SunFontManager;ZZ)V"(thread: JVMThread, args: [sun_font_SunFontManager, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOsNameAndVersion()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/setOsNameAndVersion()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setEncoding()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/setEncoding()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "foundOsSpecificFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontConfiguration/foundOsSpecificFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fontFilesArePresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontConfiguration/fontFilesArePresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "findFontConfigFile()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/findFontConfigFile()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readFontConfigFile(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/readFontConfigFile(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInstalledFallbackFonts(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/getInstalledFallbackFonts(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "findImpl(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "sun/awt/FontConfiguration/findImpl(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "findFontConfigFile(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "sun/awt/FontConfiguration/findFontConfigFile(Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "loadBinary(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontConfiguration/loadBinary(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "saveBinary(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontConfiguration/saveBinary(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "loadProperties(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontConfiguration/loadProperties(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "initFontConfig()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/initFontConfig()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInitELC()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontConfiguration/getInitELC()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "initAllComponentFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/initAllComponentFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "remapLocaleMap(IISS)S"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontConfiguration/remapLocaleMap(IISS)S"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hasMonoToPropMap()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/hasMonoToPropMap()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remapProportional(IS)S"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontConfiguration/remapProportional(IS)S"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLogicalFontFamilyName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/isLogicalFontFamilyName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLogicalFontFamilyNameLC(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/isLogicalFontFamilyNameLC(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLogicalFontStyleName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/isLogicalFontStyleName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLogicalFontFaceName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/isLogicalFontFaceName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLogicalFontFaceNameLC(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/isLogicalFontFaceNameLC(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFontIndex(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/getFontIndex(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getStyleIndex(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/getStyleIndex(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getArrayIndex([Ljava/lang/String;Ljava/lang/String;)I"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/getArrayIndex([Ljava/lang/String;Ljava/lang/String;)I"(thread: JVMThread, args: [JVMArray<java_lang_String>, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getStyleIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/getStyleIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFontName(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/awt/FontConfiguration/getFontName(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getStyleName(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/awt/FontConfiguration/getStyleName(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getLogicalFontFaceName(Ljava/lang/String;I)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/awt/FontConfiguration/getLogicalFontFaceName(Ljava/lang/String;I)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getStyleString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/awt/FontConfiguration/getStyleString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFallbackFamilyName(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/getFallbackFamilyName(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCompatibilityFamilyName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/getCompatibilityFamilyName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "mapFileName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/mapFileName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "initReorderMap()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/initReorderMap()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shuffle([Ljava/lang/String;II)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/shuffle([Ljava/lang/String;II)V"(thread: JVMThread, args: [JVMArray<java_lang_String>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "willReorderForStartupLocale()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/willReorderForStartupLocale()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getReorderSequence()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "sun/awt/FontConfiguration/getReorderSequence()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "reorderSequenceForLocale([Ljava/lang/String;)V"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/reorderSequenceForLocale([Ljava/lang/String;)V"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "splitSequence(Ljava/lang/String;)Ljava/util/Vector;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    public static "sun/awt/FontConfiguration/splitSequence(Ljava/lang/String;)Ljava/util/Vector;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    public "split(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/awt/FontConfiguration/split(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getFontDescriptors(Ljava/lang/String;I)[Lsun/awt/FontDescriptor;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_FontDescriptor>) => void): void;
    public "sun/awt/FontConfiguration/getFontDescriptors(Ljava/lang/String;I)[Lsun/awt/FontDescriptor;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_FontDescriptor>) => void): void;
    public "getFontDescriptors(II)[Lsun/awt/FontDescriptor;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_FontDescriptor>) => void): void;
    public "sun/awt/FontConfiguration/getFontDescriptors(II)[Lsun/awt/FontDescriptor;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_FontDescriptor>) => void): void;
    public "buildFontDescriptors(II)[Lsun/awt/FontDescriptor;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_FontDescriptor>) => void): void;
    public "sun/awt/FontConfiguration/buildFontDescriptors(II)[Lsun/awt/FontDescriptor;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_awt_FontDescriptor>) => void): void;
    public "makeAWTFontName(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/makeAWTFontName(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getEncoding(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/getEncoding(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFontCharsetEncoder(Ljava/lang/String;Ljava/lang/String;)Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "sun/awt/FontConfiguration/getFontCharsetEncoder(Ljava/lang/String;Ljava/lang/String;)Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "getDefaultFontCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "sun/awt/FontConfiguration/getDefaultFontCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "getAWTFontPathSet()Ljava/util/HashSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashSet) => void): void;
    public "sun/awt/FontConfiguration/getAWTFontPathSet()Ljava/util/HashSet;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_HashSet) => void): void;
    public "get2DCompositeFontInfo()[Lsun/font/CompositeFontDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_CompositeFontDescriptor>) => void): void;
    public "sun/awt/FontConfiguration/get2DCompositeFontInfo()[Lsun/font/CompositeFontDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_CompositeFontDescriptor>) => void): void;
    public "getFaceNameFromComponentFontName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/getFaceNameFromComponentFontName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFileNameFromComponentFontName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/getFileNameFromComponentFontName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "needToSearchForFile(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontConfiguration/needToSearchForFile(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumberCoreFonts()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontConfiguration/getNumberCoreFonts()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPlatformFontNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/awt/FontConfiguration/getPlatformFontNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getFileNameFromPlatformName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/getFileNameFromPlatformName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getExtraFontPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/getExtraFontPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontConfiguration/getVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getFontConfiguration()Lsun/awt/FontConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public static "sun/awt/FontConfiguration/getFontConfiguration()Lsun/awt/FontConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
    public "setFontConfiguration()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontConfiguration/setFontConfiguration()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sanityCheck()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontConfiguration/sanityCheck()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isEmpty([S)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/isEmpty([S)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "dump()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontConfiguration/dump()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getComponentFontID(SII)S"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/getComponentFontID(SII)S"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getComponentFontIDMotif(SII)S"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/getComponentFontIDMotif(SII)S"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getExclusionRanges(S)[I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/getExclusionRanges(S)[I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "contains([SSI)Z"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/contains([SSI)Z"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getComponentFontName(S)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/awt/FontConfiguration/getComponentFontName(S)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getComponentFileName(S)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/awt/FontConfiguration/getComponentFileName(S)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getComponentFileID(S)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/getComponentFileID(S)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getScriptName(S)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/awt/FontConfiguration/getScriptName(S)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCoreScripts(I)[S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/awt/FontConfiguration/getCoreScripts(I)[S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getFallbackScripts()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/getFallbackScripts()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "printTable([SI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontConfiguration/printTable([SI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "readShortTable(Ljava/io/DataInputStream;I)[S"(thread: JVMThread, args: [java_io_DataInputStream, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/readShortTable(Ljava/io/DataInputStream;I)[S"(thread: JVMThread, args: [java_io_DataInputStream, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "writeShortTable(Ljava/io/DataOutputStream;[S)V"(thread: JVMThread, args: [java_io_DataOutputStream, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontConfiguration/writeShortTable(Ljava/io/DataOutputStream;[S)V"(thread: JVMThread, args: [java_io_DataOutputStream, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "toList(Ljava/util/HashMap;)[S"(thread: JVMThread, args: [java_util_HashMap], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/toList(Ljava/util/HashMap;)[S"(thread: JVMThread, args: [java_util_HashMap], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getString(S)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/awt/FontConfiguration/getString(S)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getShortArray(S)[S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/getShortArray(S)[S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getStringID(Ljava/lang/String;)S"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/getStringID(Ljava/lang/String;)S"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getShortArrayID([S)S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/getShortArrayID([S)S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$002([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$002([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$000()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$000()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$102([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$102([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$200(Ljava/util/HashMap;)[S"(thread: JVMThread, args: [java_util_HashMap], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$200(Ljava/util/HashMap;)[S"(thread: JVMThread, args: [java_util_HashMap], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$100()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$100()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$302([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$302([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$300()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$300()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$402([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$402([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$400()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$400()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$502([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$502([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$500()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$500()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$602([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$602([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$600()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$600()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$702([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$702([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$700()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$700()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$802([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$802([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$800()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$800()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$902([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$902([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$900()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$900()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1000(Ljava/lang/String;)S"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/access$1000(Ljava/lang/String;)S"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1102([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1102([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1100()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1100()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1202([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1202([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1200()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1200()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1300()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1300()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1402([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1402([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1400()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1400()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1500([S)S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/access$1500([S)S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1602([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1602([S)[S"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1700()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/awt/FontConfiguration/access$1700()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1800()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1800()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1600()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1600()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$1902([C)[C"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$1902([C)[C"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$2000()Ljava/lang/StringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public static "sun/awt/FontConfiguration/access$2000()Ljava/lang/StringBuilder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuilder) => void): void;
    public static "access$2102([Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "sun/awt/FontConfiguration/access$2102([Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "access$2200()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/awt/FontConfiguration/access$2200()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$2300()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "sun/awt/FontConfiguration/access$2300()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "access$2500()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "sun/awt/FontConfiguration/access$2500()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "access$2600(Ljava/lang/String;)Ljava/util/Vector;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    public static "sun/awt/FontConfiguration/access$2600(Ljava/lang/String;)Ljava/util/Vector;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontConfiguration/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_DataInputStream extends java_io_FilterInputStream implements java_io_DataInput {
    public "java/io/DataInputStream/bytearr": JVMArray<number>;
    public "java/io/DataInputStream/chararr": JVMArray<number>;
    public "java/io/DataInputStream/readBuffer": JVMArray<number>;
    public "java/io/DataInputStream/lineBuffer": JVMArray<number>;
    public "<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataInputStream/<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataInputStream/readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/DataInputStream/readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/DataInputStream/readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/DataInputStream/readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/DataInputStream/readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/DataInputStream/readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "readUTF(Ljava/io/DataInput;)Ljava/lang/String;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/DataInputStream/readUTF(Ljava/io/DataInput;)Ljava/lang/String;"(thread: JVMThread, args: [java_io_DataInput], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_font_CompositeFontDescriptor extends java_lang_Object {
    public "sun/font/CompositeFontDescriptor/faceName": java_lang_String;
    public "sun/font/CompositeFontDescriptor/coreComponentCount": number;
    public "sun/font/CompositeFontDescriptor/componentFaceNames": JVMArray<java_lang_String>;
    public "sun/font/CompositeFontDescriptor/componentFileNames": JVMArray<java_lang_String>;
    public "sun/font/CompositeFontDescriptor/exclusionRanges": JVMArray<number>;
    public "sun/font/CompositeFontDescriptor/exclusionRangeLimits": JVMArray<number>;
    public "<init>(Ljava/lang/String;I[Ljava/lang/String;[Ljava/lang/String;[I[I)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_String>, JVMArray<java_lang_String>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/CompositeFontDescriptor/<init>(Ljava/lang/String;I[Ljava/lang/String;[Ljava/lang/String;[I[I)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_String>, JVMArray<java_lang_String>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFaceName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/CompositeFontDescriptor/getFaceName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCoreComponentCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/CompositeFontDescriptor/getCoreComponentCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getComponentFaceNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/CompositeFontDescriptor/getComponentFaceNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getComponentFileNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/font/CompositeFontDescriptor/getComponentFileNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getExclusionRanges()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/CompositeFontDescriptor/getExclusionRanges()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getExclusionRangeLimits()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/CompositeFontDescriptor/getExclusionRangeLimits()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export class sun_awt_FontDescriptor extends java_lang_Object implements java_lang_Cloneable {
    public "sun/awt/FontDescriptor/nativeName": java_lang_String;
    public "sun/awt/FontDescriptor/encoder": java_nio_charset_CharsetEncoder;
    public "sun/awt/FontDescriptor/charsetName": java_lang_String;
    public "sun/awt/FontDescriptor/exclusionRanges": JVMArray<number>;
    public "sun/awt/FontDescriptor/unicodeEncoder": java_nio_charset_CharsetEncoder;
    public "sun/awt/FontDescriptor/useUnicode": number;
    public static "sun/awt/FontDescriptor/isLE": number;
    public "<init>(Ljava/lang/String;Ljava/nio/charset/CharsetEncoder;[I)V"(thread: JVMThread, args: [java_lang_String, java_nio_charset_CharsetEncoder, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/awt/FontDescriptor/<init>(Ljava/lang/String;Ljava/nio/charset/CharsetEncoder;[I)V"(thread: JVMThread, args: [java_lang_String, java_nio_charset_CharsetEncoder, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNativeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontDescriptor/getNativeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFontCharsetEncoder()Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "sun/awt/FontDescriptor/getFontCharsetEncoder()Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "getFontCharsetName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontDescriptor/getFontCharsetName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getExclusionRanges()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/awt/FontDescriptor/getExclusionRanges()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "isExcluded(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontDescriptor/isExcluded(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/awt/FontDescriptor/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontDescriptor/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "useUnicode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/awt/FontDescriptor/useUnicode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/awt/FontDescriptor/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_FileFont extends sun_font_PhysicalFont {
    public "sun/font/FileFont/useJavaRasterizer": number;
    public "sun/font/FileFont/fileSize": number;
    public "sun/font/FileFont/scaler": sun_font_FontScaler;
    public "sun/font/FileFont/checkedNatives": number;
    public "sun/font/FileFont/useNatives": number;
    public "sun/font/FileFont/nativeFonts": JVMArray<sun_font_NativeFont>;
    public "sun/font/FileFont/glyphToCharMap": JVMArray<number>;
    public "<init>(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFont/<init>(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "createStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/FileFont/createStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "checkUseNatives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FileFont/checkUseNatives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFont/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readBlock(II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "sun/font/FileFont/readBlock(II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "canDoStyle(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FileFont/canDoStyle(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFileToRemove(Ljava/io/File;Lsun/font/CreatedFontTracker;)V"(thread: JVMThread, args: [java_io_File, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFont/setFileToRemove(Ljava/io/File;Lsun/font/CreatedFontTracker;)V"(thread: JVMThread, args: [java_io_File, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setFileToRemove(Ljava/lang/Object;Ljava/io/File;Lsun/font/CreatedFontTracker;)V"(thread: JVMThread, args: [java_lang_Object, java_io_File, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FileFont/setFileToRemove(Ljava/lang/Object;Ljava/io/File;Lsun/font/CreatedFontTracker;)V"(thread: JVMThread, args: [java_lang_Object, java_io_File, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable) => void): void;
    public "deregisterFontAndClearStrikeCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFont/deregisterFontAndClearStrikeCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/FileFont/getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FileFont/getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFont/getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FileFont/getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/FileFont/getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FileFont/getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FileFont/getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getScaler()Lsun/font/FontScaler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontScaler) => void): void;
    public "sun/font/FileFont/getScaler()Lsun/font/FontScaler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontScaler) => void): void;
    public "getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FileFont/getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getPublicFileName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/FileFont/getPublicFileName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface sun_font_FontManagerForSGE extends java_lang_Object, sun_font_FontManager {
    "getCreatedFonts()[Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Font>) => void): void;
    "getCreatedFontFamilyNames()Ljava/util/TreeMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TreeMap) => void): void;
    "getAllInstalledFonts()[Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_Font>) => void): void;
    "getInstalledFontFamilyNames(Ljava/util/Locale;)[Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    "useAlternateFontforJALocales()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "registerFont(Ljava/awt/Font;)Z"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "deRegisterBadFont(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
    "findFont2D(Ljava/lang/String;II)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    "createFont2D(Ljava/io/File;IZLsun/font/CreatedFontTracker;)Lsun/font/Font2D;"(thread: JVMThread, args: [java_io_File, number, number, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    "usingPerAppContextComposites()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getNewComposite(Ljava/lang/String;ILsun/font/Font2DHandle;)Lsun/font/Font2DHandle;"(thread: JVMThread, args: [java_lang_String, number, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2DHandle) => void): void;
    "preferLocaleFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "preferProportionalFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_font_FontManager extends java_lang_Object {
    "registerFont(Ljava/awt/Font;)Z"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "deRegisterBadFont(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
    "findFont2D(Ljava/lang/String;II)Lsun/font/Font2D;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    "createFont2D(Ljava/io/File;IZLsun/font/CreatedFontTracker;)Lsun/font/Font2D;"(thread: JVMThread, args: [java_io_File, number, number, sun_font_CreatedFontTracker], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    "usingPerAppContextComposites()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getNewComposite(Ljava/lang/String;ILsun/font/Font2DHandle;)Lsun/font/Font2DHandle;"(thread: JVMThread, args: [java_lang_String, number, sun_font_Font2DHandle], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2DHandle) => void): void;
    "preferLocaleFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "preferProportionalFonts()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_FontSupport extends java_lang_Object {
    "getFontConfiguration()Lsun/awt/FontConfiguration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_awt_FontConfiguration) => void): void;
  }
  export class sun_font_StrikeCache extends java_lang_Object {
    public static "sun/font/StrikeCache/unsafe": sun_misc_Unsafe;
    public static "sun/font/StrikeCache/refQueue": java_lang_ref_ReferenceQueue;
    public static "sun/font/StrikeCache/disposeListeners": java_util_ArrayList;
    public static "sun/font/StrikeCache/MINSTRIKES": number;
    public static "sun/font/StrikeCache/recentStrikeIndex": number;
    public static "sun/font/StrikeCache/recentStrikes": JVMArray<sun_font_FontStrike>;
    public static "sun/font/StrikeCache/cacheRefTypeWeak": number;
    public static "sun/font/StrikeCache/nativeAddressSize": number;
    public static "sun/font/StrikeCache/glyphInfoSize": number;
    public static "sun/font/StrikeCache/xAdvanceOffset": number;
    public static "sun/font/StrikeCache/yAdvanceOffset": number;
    public static "sun/font/StrikeCache/boundsOffset": number;
    public static "sun/font/StrikeCache/widthOffset": number;
    public static "sun/font/StrikeCache/heightOffset": number;
    public static "sun/font/StrikeCache/rowBytesOffset": number;
    public static "sun/font/StrikeCache/topLeftXOffset": number;
    public static "sun/font/StrikeCache/topLeftYOffset": number;
    public static "sun/font/StrikeCache/pixelDataOffset": number;
    public static "sun/font/StrikeCache/cacheCellOffset": number;
    public static "sun/font/StrikeCache/managedOffset": number;
    public static "sun/font/StrikeCache/invisibleGlyphPtr": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StrikeCache/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getGlyphCacheDescription([J)V"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/getGlyphCacheDescription([J)V"(thread: JVMThread, args: [JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "refStrike(Lsun/font/FontStrike;)V"(thread: JVMThread, args: [sun_font_FontStrike], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/refStrike(Lsun/font/FontStrike;)V"(thread: JVMThread, args: [sun_font_FontStrike], cb?: (e?: java_lang_Throwable) => void): void;
    public static "doDispose(Lsun/font/FontStrikeDisposer;)V"(thread: JVMThread, args: [sun_font_FontStrikeDisposer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/doDispose(Lsun/font/FontStrikeDisposer;)V"(thread: JVMThread, args: [sun_font_FontStrikeDisposer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "longAddresses()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/StrikeCache/longAddresses()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "disposeStrike(Lsun/font/FontStrikeDisposer;)V"(thread: JVMThread, args: [sun_font_FontStrikeDisposer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/disposeStrike(Lsun/font/FontStrikeDisposer;)V"(thread: JVMThread, args: [sun_font_FontStrikeDisposer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "freeIntPointer(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/freeIntPointer(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "freeLongPointer(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/freeLongPointer(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "freeIntMemory([IJ)V"(thread: JVMThread, args: [JVMArray<number>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/freeIntMemory([IJ)V"(thread: JVMThread, args: [JVMArray<number>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "freeLongMemory([JJ)V"(thread: JVMThread, args: [JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/freeLongMemory([JJ)V"(thread: JVMThread, args: [JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "freeCachedIntMemory([IJ)V"(thread: JVMThread, args: [JVMArray<number>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/freeCachedIntMemory([IJ)V"(thread: JVMThread, args: [JVMArray<number>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "freeCachedLongMemory([JJ)V"(thread: JVMThread, args: [JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/freeCachedLongMemory([JJ)V"(thread: JVMThread, args: [JVMArray<Long>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "addGlyphDisposedListener(Lsun/font/GlyphDisposedListener;)V"(thread: JVMThread, args: [sun_font_GlyphDisposedListener], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/addGlyphDisposedListener(Lsun/font/GlyphDisposedListener;)V"(thread: JVMThread, args: [sun_font_GlyphDisposedListener], cb?: (e?: java_lang_Throwable) => void): void;
    public static "notifyDisposeListeners(Ljava/util/ArrayList;)V"(thread: JVMThread, args: [java_util_ArrayList], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/notifyDisposeListeners(Ljava/util/ArrayList;)V"(thread: JVMThread, args: [java_util_ArrayList], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getStrikeRef(Lsun/font/FontStrike;)Ljava/lang/ref/Reference;"(thread: JVMThread, args: [sun_font_FontStrike], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public static "sun/font/StrikeCache/getStrikeRef(Lsun/font/FontStrike;)Ljava/lang/ref/Reference;"(thread: JVMThread, args: [sun_font_FontStrike], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public static "getStrikeRef(Lsun/font/FontStrike;Z)Ljava/lang/ref/Reference;"(thread: JVMThread, args: [sun_font_FontStrike, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public static "sun/font/StrikeCache/getStrikeRef(Lsun/font/FontStrike;Z)Ljava/lang/ref/Reference;"(thread: JVMThread, args: [sun_font_FontStrike, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_Reference) => void): void;
    public static "access$000(Lsun/font/FontStrikeDisposer;)V"(thread: JVMThread, args: [sun_font_FontStrikeDisposer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/access$000(Lsun/font/FontStrikeDisposer;)V"(thread: JVMThread, args: [sun_font_FontStrikeDisposer], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StrikeCache/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_font_GlyphDisposedListener extends java_lang_Object {
    "glyphDisposed(Ljava/util/ArrayList;)V"(thread: JVMThread, args: [java_util_ArrayList], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_FontStrikeDisposer extends java_lang_Object implements sun_java2d_DisposerRecord, sun_java2d_Disposer$PollDisposable {
    public "sun/font/FontStrikeDisposer/font2D": sun_font_Font2D;
    public "sun/font/FontStrikeDisposer/desc": sun_font_FontStrikeDesc;
    public "sun/font/FontStrikeDisposer/longGlyphImages": JVMArray<Long>;
    public "sun/font/FontStrikeDisposer/intGlyphImages": JVMArray<number>;
    public "sun/font/FontStrikeDisposer/segIntGlyphImages": JVMArray<JVMArray<number>>;
    public "sun/font/FontStrikeDisposer/segLongGlyphImages": JVMArray<JVMArray<Long>>;
    public "sun/font/FontStrikeDisposer/pScalerContext": Long;
    public "sun/font/FontStrikeDisposer/disposed": number;
    public "sun/font/FontStrikeDisposer/comp": number;
    public "<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J[I)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrikeDisposer/<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J[I)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J[J)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrikeDisposer/<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J[J)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrikeDisposer/<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrikeDisposer/<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrikeDisposer/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_java2d_Disposer$PollDisposable extends java_lang_Object {
  }
  export class sun_font_FileFontStrike extends sun_font_PhysicalStrike {
    public static "sun/font/FileFontStrike/INVISIBLE_GLYPHS": number;
    public "sun/font/FileFontStrike/fileFont": sun_font_FileFont;
    public static "sun/font/FileFontStrike/UNINITIALISED": number;
    public static "sun/font/FileFontStrike/INTARRAY": number;
    public static "sun/font/FileFontStrike/LONGARRAY": number;
    public static "sun/font/FileFontStrike/SEGINTARRAY": number;
    public static "sun/font/FileFontStrike/SEGLONGARRAY": number;
    public "sun/font/FileFontStrike/glyphCacheFormat": number;
    public static "sun/font/FileFontStrike/SEGSHIFT": number;
    public static "sun/font/FileFontStrike/SEGSIZE": number;
    public "sun/font/FileFontStrike/segmentedCache": number;
    public "sun/font/FileFontStrike/segIntGlyphImages": JVMArray<JVMArray<number>>;
    public "sun/font/FileFontStrike/segLongGlyphImages": JVMArray<JVMArray<Long>>;
    public "sun/font/FileFontStrike/horizontalAdvances": JVMArray<number>;
    public "sun/font/FileFontStrike/segHorizontalAdvances": JVMArray<JVMArray<number>>;
    public "sun/font/FileFontStrike/boundsMap": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/FileFontStrike/glyphMetricsMapRef": java_lang_ref_SoftReference;
    public "sun/font/FileFontStrike/invertDevTx": java_awt_geom_AffineTransform;
    public "sun/font/FileFontStrike/useNatives": number;
    public "sun/font/FileFontStrike/nativeStrikes": JVMArray<sun_font_NativeStrike>;
    public "sun/font/FileFontStrike/intPtSize": number;
    public static "sun/font/FileFontStrike/isXPorLater": number;
    public "sun/font/FileFontStrike/outlineMapRef": java_lang_ref_WeakReference;
    public static "initNative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/FileFontStrike/initNative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Lsun/font/FileFont;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_FileFont, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFontStrike/<init>(Lsun/font/FileFont;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_FileFont, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FileFontStrike/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphImageFromNative(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FileFontStrike/getGlyphImageFromNative(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "_getGlyphImageFromWindows(Ljava/lang/String;IIIZ)J"(thread: JVMThread, args: [java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FileFontStrike/_getGlyphImageFromWindows(Ljava/lang/String;IIIZ)J"(thread: JVMThread, args: [java_lang_String, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImageFromWindows(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FileFontStrike/getGlyphImageFromWindows(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImageFromX11(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FileFontStrike/getGlyphImageFromX11(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImagePtr(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FileFontStrike/getGlyphImagePtr(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImagePtrs([I[JI)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFontStrike/getGlyphImagePtrs([I[JI)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSlot0GlyphImagePtrs([I[JI)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FileFontStrike/getSlot0GlyphImagePtrs([I[JI)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCachedGlyphPtr(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FileFontStrike/getCachedGlyphPtr(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setCachedGlyphPtr(IJ)J"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FileFontStrike/setCachedGlyphPtr(IJ)J"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "initGlyphCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFontStrike/initGlyphCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FileFontStrike/getGlyphAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphAdvance(IZ)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FileFontStrike/getGlyphAdvance(IZ)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCodePointAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FileFontStrike/getCodePointAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphImageBounds(ILjava/awt/geom/Point2D$Float;Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D$Float, java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFontStrike/getGlyphImageBounds(ILjava/awt/geom/Point2D$Float;Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D$Float, java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphImageMinX(JI)I"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FileFontStrike/getGlyphImageMinX(JI)I"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFontMetrics()Lsun/font/StrikeMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/FileFontStrike/getFontMetrics()Lsun/font/StrikeMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getGlyphMetrics(I)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/FileFontStrike/getGlyphMetrics(I)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getGlyphMetrics(IZ)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/FileFontStrike/getGlyphMetrics(IZ)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getCharMetrics(C)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/FileFontStrike/getCharMetrics(C)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getGlyphOutlineBounds(I)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/FileFontStrike/getGlyphOutlineBounds(I)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getOutlineBounds(I)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "sun/font/FileFontStrike/getOutlineBounds(I)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getGlyphOutline(IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FileFontStrike/getGlyphOutline(IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphVectorOutline([IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FileFontStrike/getGlyphVectorOutline([IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "adjustPoint(Ljava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FileFontStrike/adjustPoint(Ljava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FileFontStrike/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_PhysicalStrike extends sun_font_FontStrike {
    public static "sun/font/PhysicalStrike/INTMASK": Long;
    public static "sun/font/PhysicalStrike/longAddresses": number;
    public "sun/font/PhysicalStrike/physicalFont": sun_font_PhysicalFont;
    public "sun/font/PhysicalStrike/mapper": sun_font_CharToGlyphMapper;
    public "sun/font/PhysicalStrike/pScalerContext": Long;
    public "sun/font/PhysicalStrike/longGlyphImages": JVMArray<Long>;
    public "sun/font/PhysicalStrike/intGlyphImages": JVMArray<number>;
    public "sun/font/PhysicalStrike/glyphPointMapCache": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/PhysicalStrike/getImageWithAdvance": number;
    public static "sun/font/PhysicalStrike/complexTX": number;
    public "<init>(Lsun/font/PhysicalFont;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_PhysicalFont, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/PhysicalStrike/<init>(Lsun/font/PhysicalFont;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_PhysicalFont, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/PhysicalStrike/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/PhysicalStrike/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFontMetrics()Lsun/font/StrikeMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/PhysicalStrike/getFontMetrics()Lsun/font/StrikeMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getCodePointAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/PhysicalStrike/getCodePointAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCharMetrics(C)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/PhysicalStrike/getCharMetrics(C)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getSlot0GlyphImagePtrs([I[JI)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/PhysicalStrike/getSlot0GlyphImagePtrs([I[JI)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphPoint(II)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/PhysicalStrike/getGlyphPoint(II)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "adjustPoint(Ljava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/PhysicalStrike/adjustPoint(Ljava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/PhysicalStrike/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_GlyphLayout$GVData extends java_lang_Object {
    public "sun/font/GlyphLayout$GVData/_count": number;
    public "sun/font/GlyphLayout$GVData/_flags": number;
    public "sun/font/GlyphLayout$GVData/_glyphs": JVMArray<number>;
    public "sun/font/GlyphLayout$GVData/_positions": JVMArray<number>;
    public "sun/font/GlyphLayout$GVData/_indices": JVMArray<number>;
    public static "sun/font/GlyphLayout$GVData/UNINITIALIZED_FLAGS": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphLayout$GVData/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "init(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphLayout$GVData/init(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "grow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphLayout$GVData/grow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphLayout$GVData/grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "adjustPositions(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphLayout$GVData/adjustPositions(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "createGlyphVector(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;Lsun/font/StandardGlyphVector;)Lsun/font/StandardGlyphVector;"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext, sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector) => void): void;
    public "sun/font/GlyphLayout$GVData/createGlyphVector(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;Lsun/font/StandardGlyphVector;)Lsun/font/StandardGlyphVector;"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext, sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector) => void): void;
  }
  export class sun_font_StandardGlyphVector extends java_awt_font_GlyphVector {
    public "sun/font/StandardGlyphVector/font": java_awt_Font;
    public "sun/font/StandardGlyphVector/frc": java_awt_font_FontRenderContext;
    public "sun/font/StandardGlyphVector/glyphs": JVMArray<number>;
    public "sun/font/StandardGlyphVector/userGlyphs": JVMArray<number>;
    public "sun/font/StandardGlyphVector/positions": JVMArray<number>;
    public "sun/font/StandardGlyphVector/charIndices": JVMArray<number>;
    public "sun/font/StandardGlyphVector/flags": number;
    public static "sun/font/StandardGlyphVector/UNINITIALIZED_FLAGS": number;
    public "sun/font/StandardGlyphVector/gti": sun_font_StandardGlyphVector$GlyphTransformInfo;
    public "sun/font/StandardGlyphVector/ftx": java_awt_geom_AffineTransform;
    public "sun/font/StandardGlyphVector/dtx": java_awt_geom_AffineTransform;
    public "sun/font/StandardGlyphVector/invdtx": java_awt_geom_AffineTransform;
    public "sun/font/StandardGlyphVector/frctx": java_awt_geom_AffineTransform;
    public "sun/font/StandardGlyphVector/font2D": sun_font_Font2D;
    public "sun/font/StandardGlyphVector/fsref": java_lang_ref_SoftReference;
    public "sun/font/StandardGlyphVector/lbcacheRef": java_lang_ref_SoftReference;
    public "sun/font/StandardGlyphVector/vbcacheRef": java_lang_ref_SoftReference;
    public static "sun/font/StandardGlyphVector/FLAG_USES_VERTICAL_BASELINE": number;
    public static "sun/font/StandardGlyphVector/FLAG_USES_VERTICAL_METRICS": number;
    public static "sun/font/StandardGlyphVector/FLAG_USES_ALTERNATE_ORIENTATION": number;
    public static "sun/font/StandardGlyphVector/$assertionsDisabled": number;
    public "<init>(Ljava/awt/Font;Ljava/lang/String;Ljava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, java_lang_String, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/<init>(Ljava/awt/Font;Ljava/lang/String;Ljava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, java_lang_String, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Font;[CLjava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, JVMArray<number>, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/<init>(Ljava/awt/Font;[CLjava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, JVMArray<number>, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Font;[CIILjava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, JVMArray<number>, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/<init>(Ljava/awt/Font;[CIILjava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, JVMArray<number>, number, number, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTracking(Ljava/awt/Font;)F"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector/getTracking(Ljava/awt/Font;)F"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;[I[F[II)V"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext, JVMArray<number>, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/<init>(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;[I[F[II)V"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext, JVMArray<number>, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "initGlyphVector(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;[I[F[II)V"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext, JVMArray<number>, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/initGlyphVector(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;[I[F[II)V"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext, JVMArray<number>, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Font;Ljava/text/CharacterIterator;Ljava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, java_text_CharacterIterator, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/<init>(Ljava/awt/Font;Ljava/text/CharacterIterator;Ljava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, java_text_CharacterIterator, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Font;[ILjava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, JVMArray<number>, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/<init>(Ljava/awt/Font;[ILjava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_Font, JVMArray<number>, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getStandardGV(Ljava/awt/font/GlyphVector;Lsun/java2d/loops/FontInfo;)Lsun/font/StandardGlyphVector;"(thread: JVMThread, args: [java_awt_font_GlyphVector, sun_java2d_loops_FontInfo], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector) => void): void;
    public static "sun/font/StandardGlyphVector/getStandardGV(Ljava/awt/font/GlyphVector;Lsun/java2d/loops/FontInfo;)Lsun/font/StandardGlyphVector;"(thread: JVMThread, args: [java_awt_font_GlyphVector, sun_java2d_loops_FontInfo], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector) => void): void;
    public "getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "sun/font/StandardGlyphVector/getFont()Ljava/awt/Font;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public "getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "sun/font/StandardGlyphVector/getFontRenderContext()Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public "performDefaultLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/performDefaultLayout()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphCodes(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphCodes(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getGlyphCharIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphCharIndex(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphCharIndices(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphCharIndices(II[I)[I"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getLogicalBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "sun/font/StandardGlyphVector/getLogicalBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getVisualBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "sun/font/StandardGlyphVector/getVisualBounds()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getPixelBounds(Ljava/awt/font/FontRenderContext;FF)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "sun/font/StandardGlyphVector/getPixelBounds(Ljava/awt/font/FontRenderContext;FF)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getOutline()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/font/StandardGlyphVector/getOutline()Ljava/awt/Shape;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getOutline(FF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/font/StandardGlyphVector/getOutline(FF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphOutline(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphOutline(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphOutline(IFF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphOutline(IFF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphPosition(I)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphPosition(I)Ljava/awt/geom/Point2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D) => void): void;
    public "setGlyphPosition(ILjava/awt/geom/Point2D;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/setGlyphPosition(ILjava/awt/geom/Point2D;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphTransform(I)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphTransform(I)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "setGlyphTransform(ILjava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [number, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/setGlyphTransform(ILjava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [number, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLayoutFlags()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector/getLayoutFlags()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphPositions(II[F)[F"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphPositions(II[F)[F"(thread: JVMThread, args: [number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getGlyphLogicalBounds(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphLogicalBounds(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphVisualBounds(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphVisualBounds(I)Ljava/awt/Shape;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphPixelBounds(ILjava/awt/font/FontRenderContext;FF)Ljava/awt/Rectangle;"(thread: JVMThread, args: [number, java_awt_font_FontRenderContext, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphPixelBounds(ILjava/awt/font/FontRenderContext;FF)Ljava/awt/Rectangle;"(thread: JVMThread, args: [number, java_awt_font_FontRenderContext, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getGlyphMetrics(I)Ljava/awt/font/GlyphMetrics;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphMetrics) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphMetrics(I)Ljava/awt/font/GlyphMetrics;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphMetrics) => void): void;
    public "getGlyphJustificationInfo(I)Ljava/awt/font/GlyphJustificationInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphJustificationInfo) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphJustificationInfo(I)Ljava/awt/font/GlyphJustificationInfo;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_GlyphJustificationInfo) => void): void;
    public "equals(Ljava/awt/font/GlyphVector;)Z"(thread: JVMThread, args: [java_awt_font_GlyphVector], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector/equals(Ljava/awt/font/GlyphVector;)Z"(thread: JVMThread, args: [java_awt_font_GlyphVector], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "copy()Lsun/font/StandardGlyphVector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector) => void): void;
    public "sun/font/StandardGlyphVector/copy()Lsun/font/StandardGlyphVector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/StandardGlyphVector/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "setGlyphPositions([FIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/setGlyphPositions([FIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setGlyphPositions([F)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/setGlyphPositions([F)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphPositions([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphPositions([F)[F"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getGlyphTransforms(II[Ljava/awt/geom/AffineTransform;)[Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, number, JVMArray<java_awt_geom_AffineTransform>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_geom_AffineTransform>) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphTransforms(II[Ljava/awt/geom/AffineTransform;)[Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number, number, JVMArray<java_awt_geom_AffineTransform>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_geom_AffineTransform>) => void): void;
    public "getGlyphTransforms()[Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_geom_AffineTransform>) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphTransforms()[Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_awt_geom_AffineTransform>) => void): void;
    public "setGlyphTransforms([Ljava/awt/geom/AffineTransform;III)V"(thread: JVMThread, args: [JVMArray<java_awt_geom_AffineTransform>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/setGlyphTransforms([Ljava/awt/geom/AffineTransform;III)V"(thread: JVMThread, args: [JVMArray<java_awt_geom_AffineTransform>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setGlyphTransforms([Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [JVMArray<java_awt_geom_AffineTransform>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/setGlyphTransforms([Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [JVMArray<java_awt_geom_AffineTransform>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphInfo()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphInfo()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "pixellate(Ljava/awt/font/FontRenderContext;Ljava/awt/geom/Point2D;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_font_FontRenderContext, java_awt_geom_Point2D, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/pixellate(Ljava/awt/font/FontRenderContext;Ljava/awt/geom/Point2D;Ljava/awt/Point;)V"(thread: JVMThread, args: [java_awt_font_FontRenderContext, java_awt_geom_Point2D, java_awt_Point], cb?: (e?: java_lang_Throwable) => void): void;
    public "needsPositions([D)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector/needsPositions([D)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setupGlyphImages([J[F[D)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/StandardGlyphVector/setupGlyphImages([J[F[D)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "matchTX([DLjava/awt/geom/AffineTransform;)Z"(thread: JVMThread, args: [JVMArray<number>, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/StandardGlyphVector/matchTX([DLjava/awt/geom/AffineTransform;)Z"(thread: JVMThread, args: [JVMArray<number>, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getNonTranslateTX(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "sun/font/StandardGlyphVector/getNonTranslateTX(Ljava/awt/geom/AffineTransform;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "equalNonTranslateTX(Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;)Z"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/StandardGlyphVector/equalNonTranslateTX(Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;)Z"(thread: JVMThread, args: [java_awt_geom_AffineTransform, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setRenderTransform([D)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/setRenderTransform([D)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDTX(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/setDTX(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFRCTX()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/setFRCTX()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resetDTX(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/resetDTX(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/font/GlyphVector;Ljava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_font_GlyphVector, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/<init>(Ljava/awt/font/GlyphVector;Ljava/awt/font/FontRenderContext;)V"(thread: JVMThread, args: [java_awt_font_GlyphVector, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValidatedGlyphs([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/StandardGlyphVector/getValidatedGlyphs([I)[I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "init(Ljava/awt/Font;[CIILjava/awt/font/FontRenderContext;I)V"(thread: JVMThread, args: [java_awt_Font, JVMArray<number>, number, number, java_awt_font_FontRenderContext, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/init(Ljava/awt/Font;[CIILjava/awt/font/FontRenderContext;I)V"(thread: JVMThread, args: [java_awt_Font, JVMArray<number>, number, number, java_awt_font_FontRenderContext, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "initFontData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/initFontData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "internalGetGlyphPositions(III[F)[F"(thread: JVMThread, args: [number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/StandardGlyphVector/internalGetGlyphPositions(III[F)[F"(thread: JVMThread, args: [number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getGlyphOutlineBounds(I)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphOutlineBounds(I)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getGlyphsOutline(IIFF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphsOutline(IIFF)Ljava/awt/Shape;"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Shape) => void): void;
    public "getGlyphsPixelBounds(Ljava/awt/font/FontRenderContext;FFII)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphsPixelBounds(Ljava/awt/font/FontRenderContext;FFII)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_font_FontRenderContext, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "clearCaches(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/clearCaches(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearCaches()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/clearCaches()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "initPositions()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/initPositions()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addFlags(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/addFlags(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearFlags(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector/clearFlags(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphStrike(I)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public "sun/font/StandardGlyphVector/getGlyphStrike(I)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public "getDefaultStrike()Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public "sun/font/StandardGlyphVector/getDefaultStrike()Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/StandardGlyphVector/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "appendString(Ljava/lang/StringBuffer;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "sun/font/StandardGlyphVector/appendString(Ljava/lang/StringBuffer;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public static "access$000(Lsun/font/StandardGlyphVector;)[I"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/font/StandardGlyphVector/access$000(Lsun/font/StandardGlyphVector;)[I"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$100(Lsun/font/StandardGlyphVector;I)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StandardGlyphVector/access$100(Lsun/font/StandardGlyphVector;I)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$200(Lsun/font/StandardGlyphVector;I)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StandardGlyphVector/access$200(Lsun/font/StandardGlyphVector;I)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$300(Lsun/font/StandardGlyphVector;I)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StandardGlyphVector/access$300(Lsun/font/StandardGlyphVector;I)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$400(Lsun/font/StandardGlyphVector;)[F"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "sun/font/StandardGlyphVector/access$400(Lsun/font/StandardGlyphVector;)[F"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$500(Lsun/font/StandardGlyphVector;)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public static "sun/font/StandardGlyphVector/access$500(Lsun/font/StandardGlyphVector;)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public static "access$600(Lsun/font/StandardGlyphVector;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "sun/font/StandardGlyphVector/access$600(Lsun/font/StandardGlyphVector;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "access$700(Lsun/font/StandardGlyphVector;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "sun/font/StandardGlyphVector/access$700(Lsun/font/StandardGlyphVector;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "access$800(Lsun/font/StandardGlyphVector;)Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public static "sun/font/StandardGlyphVector/access$800(Lsun/font/StandardGlyphVector;)Ljava/awt/font/FontRenderContext;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_font_FontRenderContext) => void): void;
    public static "access$900(Lsun/font/StandardGlyphVector;)Lsun/font/Font2D;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public static "sun/font/StandardGlyphVector/access$900(Lsun/font/StandardGlyphVector;)Lsun/font/Font2D;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public static "access$1000(Lsun/font/StandardGlyphVector;)Ljava/awt/Font;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "sun/font/StandardGlyphVector/access$1000(Lsun/font/StandardGlyphVector;)Ljava/awt/Font;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_Font) => void): void;
    public static "access$1100(Lsun/font/StandardGlyphVector;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "sun/font/StandardGlyphVector/access$1100(Lsun/font/StandardGlyphVector;)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/StandardGlyphVector/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_StandardGlyphVector$GlyphStrike extends java_lang_Object {
    public "sun/font/StandardGlyphVector$GlyphStrike/sgv": sun_font_StandardGlyphVector;
    public "sun/font/StandardGlyphVector$GlyphStrike/strike": sun_font_FontStrike;
    public "sun/font/StandardGlyphVector$GlyphStrike/dx": number;
    public "sun/font/StandardGlyphVector$GlyphStrike/dy": number;
    public static "create(Lsun/font/StandardGlyphVector;Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [sun_font_StandardGlyphVector, java_awt_geom_AffineTransform, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public static "sun/font/StandardGlyphVector$GlyphStrike/create(Lsun/font/StandardGlyphVector;Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [sun_font_StandardGlyphVector, java_awt_geom_AffineTransform, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public "<init>(Lsun/font/StandardGlyphVector;Lsun/font/FontStrike;FF)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, sun_font_FontStrike, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector$GlyphStrike/<init>(Lsun/font/StandardGlyphVector;Lsun/font/FontStrike;FF)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, sun_font_FontStrike, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getADL(Lsun/font/StandardGlyphVector$ADL;)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector$ADL], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector$GlyphStrike/getADL(Lsun/font/StandardGlyphVector$ADL;)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector$ADL], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphPosition(II[F[F)V"(thread: JVMThread, args: [number, number, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector$GlyphStrike/getGlyphPosition(II[F[F)V"(thread: JVMThread, args: [number, number, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "addDefaultGlyphAdvance(ILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector$GlyphStrike/addDefaultGlyphAdvance(ILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphOutlineBounds(IFF)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "sun/font/StandardGlyphVector$GlyphStrike/getGlyphOutlineBounds(IFF)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "appendGlyphOutline(ILjava/awt/geom/GeneralPath;FF)V"(thread: JVMThread, args: [number, java_awt_geom_GeneralPath, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector$GlyphStrike/appendGlyphOutline(ILjava/awt/geom/GeneralPath;FF)V"(thread: JVMThread, args: [number, java_awt_geom_GeneralPath, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_StandardGlyphVector$ADL extends java_lang_Object {
    public "sun/font/StandardGlyphVector$ADL/ascentX": number;
    public "sun/font/StandardGlyphVector$ADL/ascentY": number;
    public "sun/font/StandardGlyphVector$ADL/descentX": number;
    public "sun/font/StandardGlyphVector$ADL/descentY": number;
    public "sun/font/StandardGlyphVector$ADL/leadingX": number;
    public "sun/font/StandardGlyphVector$ADL/leadingY": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector$ADL/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/StandardGlyphVector$ADL/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toStringBuffer(Ljava/lang/StringBuffer;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
    public "sun/font/StandardGlyphVector$ADL/toStringBuffer(Ljava/lang/StringBuffer;)Ljava/lang/StringBuffer;"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: java_lang_StringBuffer) => void): void;
  }
  export class sun_font_StandardGlyphVector$GlyphTransformInfo extends java_lang_Object {
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/sgv": sun_font_StandardGlyphVector;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/indices": JVMArray<number>;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/transforms": JVMArray<number>;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/strikesRef": java_lang_ref_SoftReference;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/haveAllStrikes": number;
    public "<init>(Lsun/font/StandardGlyphVector;)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/<init>(Lsun/font/StandardGlyphVector;)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/StandardGlyphVector;Lsun/font/StandardGlyphVector$GlyphTransformInfo;)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, sun_font_StandardGlyphVector$GlyphTransformInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/<init>(Lsun/font/StandardGlyphVector;Lsun/font/StandardGlyphVector$GlyphTransformInfo;)V"(thread: JVMThread, args: [sun_font_StandardGlyphVector, sun_font_StandardGlyphVector$GlyphTransformInfo], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Lsun/font/StandardGlyphVector$GlyphTransformInfo;)Z"(thread: JVMThread, args: [sun_font_StandardGlyphVector$GlyphTransformInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/equals(Lsun/font/StandardGlyphVector$GlyphTransformInfo;)Z"(thread: JVMThread, args: [sun_font_StandardGlyphVector$GlyphTransformInfo], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setGlyphTransform(ILjava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [number, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/setGlyphTransform(ILjava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [number, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphTransform(I)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/getGlyphTransform(I)Ljava/awt/geom/AffineTransform;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_AffineTransform) => void): void;
    public "transformCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/transformCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setupGlyphImages([J[FLjava/awt/geom/AffineTransform;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<number>, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/setupGlyphImages([J[FLjava/awt/geom/AffineTransform;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<Long>, JVMArray<number>, java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getGlyphsPixelBounds(Ljava/awt/geom/AffineTransform;FFII)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/getGlyphsPixelBounds(Ljava/awt/geom/AffineTransform;FFII)Ljava/awt/Rectangle;"(thread: JVMThread, args: [java_awt_geom_AffineTransform, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_Rectangle) => void): void;
    public "getStrike(I)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/getStrike(I)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public "getAllStrikes()[Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_StandardGlyphVector$GlyphStrike>) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/getAllStrikes()[Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_StandardGlyphVector$GlyphStrike>) => void): void;
    public "getStrikeArray()[Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_StandardGlyphVector$GlyphStrike>) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/getStrikeArray()[Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_StandardGlyphVector$GlyphStrike>) => void): void;
    public "getStrikeAtIndex([Lsun/font/StandardGlyphVector$GlyphStrike;I)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [JVMArray<sun_font_StandardGlyphVector$GlyphStrike>, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
    public "sun/font/StandardGlyphVector$GlyphTransformInfo/getStrikeAtIndex([Lsun/font/StandardGlyphVector$GlyphStrike;I)Lsun/font/StandardGlyphVector$GlyphStrike;"(thread: JVMThread, args: [JVMArray<sun_font_StandardGlyphVector$GlyphStrike>, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_StandardGlyphVector$GlyphStrike) => void): void;
  }
  export class sun_font_FontStrike extends java_lang_Object {
    public "sun/font/FontStrike/disposer": sun_font_FontStrikeDisposer;
    public "sun/font/FontStrike/desc": sun_font_FontStrikeDesc;
    public "sun/font/FontStrike/strikeMetrics": sun_font_StrikeMetrics;
    public "sun/font/FontStrike/algoStyle": number;
    public "sun/font/FontStrike/boldness": number;
    public "sun/font/FontStrike/italic": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrike/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontStrike/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFontMetrics()Lsun/font/StrikeMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/FontStrike/getFontMetrics()Lsun/font/StrikeMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getGlyphImagePtrs([I[JI)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrike/getGlyphImagePtrs([I[JI)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphImagePtr(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FontStrike/getGlyphImagePtr(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImageBounds(ILjava/awt/geom/Point2D$Float;Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D$Float, java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontStrike/getGlyphImageBounds(ILjava/awt/geom/Point2D$Float;Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D$Float, java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphMetrics(I)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/FontStrike/getGlyphMetrics(I)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getCharMetrics(C)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/FontStrike/getCharMetrics(C)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getGlyphAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontStrike/getGlyphAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCodePointAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontStrike/getCodePointAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphOutlineBounds(I)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/FontStrike/getGlyphOutlineBounds(I)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getGlyphOutline(IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FontStrike/getGlyphOutline(IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphVectorOutline([IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FontStrike/getGlyphVectorOutline([IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
  }
  export class sun_font_SunLayoutEngine extends java_lang_Object implements sun_font_GlyphLayout$LayoutEngine, sun_font_GlyphLayout$LayoutEngineFactory {
    public "sun/font/SunLayoutEngine/key": sun_font_GlyphLayout$LayoutEngineKey;
    public static "sun/font/SunLayoutEngine/instance": sun_font_GlyphLayout$LayoutEngineFactory;
    public "sun/font/SunLayoutEngine/cacheref": java_lang_ref_SoftReference;
    public static "initGVIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/SunLayoutEngine/initGVIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "instance()Lsun/font/GlyphLayout$LayoutEngineFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngineFactory) => void): void;
    public static "sun/font/SunLayoutEngine/instance()Lsun/font/GlyphLayout$LayoutEngineFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngineFactory) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunLayoutEngine/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEngine(Lsun/font/Font2D;II)Lsun/font/GlyphLayout$LayoutEngine;"(thread: JVMThread, args: [sun_font_Font2D, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngine) => void): void;
    public "sun/font/SunLayoutEngine/getEngine(Lsun/font/Font2D;II)Lsun/font/GlyphLayout$LayoutEngine;"(thread: JVMThread, args: [sun_font_Font2D, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngine) => void): void;
    public "getEngine(Lsun/font/GlyphLayout$LayoutEngineKey;)Lsun/font/GlyphLayout$LayoutEngine;"(thread: JVMThread, args: [sun_font_GlyphLayout$LayoutEngineKey], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngine) => void): void;
    public "sun/font/SunLayoutEngine/getEngine(Lsun/font/GlyphLayout$LayoutEngineKey;)Lsun/font/GlyphLayout$LayoutEngine;"(thread: JVMThread, args: [sun_font_GlyphLayout$LayoutEngineKey], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngine) => void): void;
    public "<init>(Lsun/font/GlyphLayout$LayoutEngineKey;)V"(thread: JVMThread, args: [sun_font_GlyphLayout$LayoutEngineKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunLayoutEngine/<init>(Lsun/font/GlyphLayout$LayoutEngineKey;)V"(thread: JVMThread, args: [sun_font_GlyphLayout$LayoutEngineKey], cb?: (e?: java_lang_Throwable) => void): void;
    public "layout(Lsun/font/FontStrikeDesc;[FIILsun/font/TextRecord;ILjava/awt/geom/Point2D$Float;Lsun/font/GlyphLayout$GVData;)V"(thread: JVMThread, args: [sun_font_FontStrikeDesc, JVMArray<number>, number, number, sun_font_TextRecord, number, java_awt_geom_Point2D$Float, sun_font_GlyphLayout$GVData], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/SunLayoutEngine/layout(Lsun/font/FontStrikeDesc;[FIILsun/font/TextRecord;ILjava/awt/geom/Point2D$Float;Lsun/font/GlyphLayout$GVData;)V"(thread: JVMThread, args: [sun_font_FontStrikeDesc, JVMArray<number>, number, number, sun_font_TextRecord, number, java_awt_geom_Point2D$Float, sun_font_GlyphLayout$GVData], cb?: (e?: java_lang_Throwable) => void): void;
    public static "nativeLayout(Lsun/font/Font2D;Lsun/font/FontStrike;[FII[CIIIIIIILjava/awt/geom/Point2D$Float;Lsun/font/GlyphLayout$GVData;JJ)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrike, JVMArray<number>, number, number, JVMArray<number>, number, number, number, number, number, number, number, java_awt_geom_Point2D$Float, sun_font_GlyphLayout$GVData, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/SunLayoutEngine/nativeLayout(Lsun/font/Font2D;Lsun/font/FontStrike;[FII[CIIIIIIILjava/awt/geom/Point2D$Float;Lsun/font/GlyphLayout$GVData;JJ)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrike, JVMArray<number>, number, number, JVMArray<number>, number, number, number, number, number, number, number, java_awt_geom_Point2D$Float, sun_font_GlyphLayout$GVData, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/SunLayoutEngine/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_TextRecord extends java_lang_Object {
    public "sun/font/TextRecord/text": JVMArray<number>;
    public "sun/font/TextRecord/start": number;
    public "sun/font/TextRecord/limit": number;
    public "sun/font/TextRecord/min": number;
    public "sun/font/TextRecord/max": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TextRecord/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "init([CIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/TextRecord/init([CIIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_GlyphLayout$LayoutEngineKey extends java_lang_Object {
    public "sun/font/GlyphLayout$LayoutEngineKey/font": sun_font_Font2D;
    public "sun/font/GlyphLayout$LayoutEngineKey/script": number;
    public "sun/font/GlyphLayout$LayoutEngineKey/lang": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphLayout$LayoutEngineKey/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/Font2D;II)V"(thread: JVMThread, args: [sun_font_Font2D, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphLayout$LayoutEngineKey/<init>(Lsun/font/Font2D;II)V"(thread: JVMThread, args: [sun_font_Font2D, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "init(Lsun/font/Font2D;II)V"(thread: JVMThread, args: [sun_font_Font2D, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/GlyphLayout$LayoutEngineKey/init(Lsun/font/Font2D;II)V"(thread: JVMThread, args: [sun_font_Font2D, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "copy()Lsun/font/GlyphLayout$LayoutEngineKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngineKey) => void): void;
    public "sun/font/GlyphLayout$LayoutEngineKey/copy()Lsun/font/GlyphLayout$LayoutEngineKey;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngineKey) => void): void;
    public "font()Lsun/font/Font2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/GlyphLayout$LayoutEngineKey/font()Lsun/font/Font2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "script()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphLayout$LayoutEngineKey/script()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lang()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphLayout$LayoutEngineKey/lang()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphLayout$LayoutEngineKey/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/GlyphLayout$LayoutEngineKey/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface sun_font_GlyphLayout$LayoutEngineFactory extends java_lang_Object {
    "getEngine(Lsun/font/Font2D;II)Lsun/font/GlyphLayout$LayoutEngine;"(thread: JVMThread, args: [sun_font_Font2D, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngine) => void): void;
    "getEngine(Lsun/font/GlyphLayout$LayoutEngineKey;)Lsun/font/GlyphLayout$LayoutEngine;"(thread: JVMThread, args: [sun_font_GlyphLayout$LayoutEngineKey], cb?: (e?: java_lang_Throwable, rv?: sun_font_GlyphLayout$LayoutEngine) => void): void;
  }
  export interface sun_font_GlyphLayout$LayoutEngine extends java_lang_Object {
    "layout(Lsun/font/FontStrikeDesc;[FIILsun/font/TextRecord;ILjava/awt/geom/Point2D$Float;Lsun/font/GlyphLayout$GVData;)V"(thread: JVMThread, args: [sun_font_FontStrikeDesc, JVMArray<number>, number, number, sun_font_TextRecord, number, java_awt_geom_Point2D$Float, sun_font_GlyphLayout$GVData], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_NullFontScaler extends sun_font_FontScaler {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NullFontScaler/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/Font2D;IZI)V"(thread: JVMThread, args: [sun_font_Font2D, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NullFontScaler/<init>(Lsun/font/Font2D;IZI)V"(thread: JVMThread, args: [sun_font_Font2D, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/NullFontScaler/getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NullFontScaler/getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NullFontScaler/getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/NullFontScaler/getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/NullFontScaler/getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/NullFontScaler/getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getLayoutTableCache()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NullFontScaler/getLayoutTableCache()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "createScalerContext([DIIFFZ)J"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NullFontScaler/createScalerContext([DIIFFZ)J"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "invalidateScalerContext(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NullFontScaler/invalidateScalerContext(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NullFontScaler/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NullFontScaler/getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphCode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NullFontScaler/getGlyphCode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NullFontScaler/getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/NullFontScaler/getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public static "getNullScalerContext()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "sun/font/NullFontScaler/getNullScalerContext()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NullFontScaler/getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class sun_font_NativeStrikeDisposer extends sun_font_FontStrikeDisposer {
    public "sun/font/NativeStrikeDisposer/pNativeScalerContext": Long;
    public "<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J[I)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrikeDisposer/<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J[I)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J[J)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrikeDisposer/<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J[J)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any, JVMArray<Long>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrikeDisposer/<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;J)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrikeDisposer/<init>(Lsun/font/Font2D;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_Font2D, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrikeDisposer/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "freeNativeScalerContext(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrikeDisposer/freeNativeScalerContext(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_NativeStrike extends sun_font_PhysicalStrike {
    public "sun/font/NativeStrike/nativeFont": sun_font_NativeFont;
    public "sun/font/NativeStrike/numGlyphs": number;
    public "sun/font/NativeStrike/invertDevTx": java_awt_geom_AffineTransform;
    public "sun/font/NativeStrike/fontTx": java_awt_geom_AffineTransform;
    public "getNativePointSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NativeStrike/getNativePointSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Lsun/font/NativeFont;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_NativeFont, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrike/<init>(Lsun/font/NativeFont;Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_NativeFont, sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "usingIntGlyphImages()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NativeStrike/usingIntGlyphImages()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLongGlyphImages()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "sun/font/NativeStrike/getLongGlyphImages()[J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<Long>) => void): void;
    public "<init>(Lsun/font/NativeFont;Lsun/font/FontStrikeDesc;Z)V"(thread: JVMThread, args: [sun_font_NativeFont, sun_font_FontStrikeDesc, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrike/<init>(Lsun/font/NativeFont;Lsun/font/FontStrikeDesc;Z)V"(thread: JVMThread, args: [sun_font_NativeFont, sun_font_FontStrikeDesc, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontMetrics()Lsun/font/StrikeMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/NativeStrike/getFontMetrics()Lsun/font/StrikeMetrics;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "createScalerContext([BID)J"(thread: JVMThread, args: [JVMArray<number>, number, number, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NativeStrike/createScalerContext([BID)J"(thread: JVMThread, args: [JVMArray<number>, number, number, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getMaxGlyph(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NativeStrike/getMaxGlyph(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createNullScalerContext()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NativeStrike/createNullScalerContext()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImagePtrs([I[JI)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrike/getGlyphImagePtrs([I[JI)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<Long>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphImagePtr(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NativeStrike/getGlyphImagePtr(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImagePtrNoCache(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NativeStrike/getGlyphImagePtrNoCache(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImageBounds(ILjava/awt/geom/Point2D$Float;Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D$Float, java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeStrike/getGlyphImageBounds(ILjava/awt/geom/Point2D$Float;Ljava/awt/Rectangle;)V"(thread: JVMThread, args: [number, java_awt_geom_Point2D$Float, java_awt_Rectangle], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphMetrics(I)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/NativeStrike/getGlyphMetrics(I)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getGlyphAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NativeStrike/getGlyphAdvance(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphOutlineBounds(I)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/NativeStrike/getGlyphOutlineBounds(I)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getGlyphOutline(IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/NativeStrike/getGlyphOutline(IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphVectorOutline([IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/NativeStrike/getGlyphVectorOutline([IFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
  }
  export class sun_font_NativeFont extends sun_font_PhysicalFont {
    public "sun/font/NativeFont/encoding": java_lang_String;
    public "sun/font/NativeFont/numGlyphs": number;
    public "sun/font/NativeFont/isBitmapDelegate": number;
    public "sun/font/NativeFont/delegateFont": sun_font_PhysicalFont;
    public "<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeFont/<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "initNames()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeFont/initNames()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "hasExternalBitmaps(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/NativeFont/hasExternalBitmaps(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "fontExists(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/NativeFont/fontExists(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "haveBitmapFonts([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/NativeFont/haveBitmapFonts([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "fontExists([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/font/NativeFont/fontExists([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMapper()Lsun/font/CharToGlyphMapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CharToGlyphMapper) => void): void;
    public "sun/font/NativeFont/getMapper()Lsun/font/CharToGlyphMapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CharToGlyphMapper) => void): void;
    public "createStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/NativeFont/createStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "getMaxCharBounds(Ljava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "sun/font/NativeFont/getMaxCharBounds(Ljava/awt/font/FontRenderContext;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/NativeFont/getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NativeFont/getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/NativeFont/getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/NativeFont/getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NativeFont/getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphImageNoDefault(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/NativeFont/getGlyphImageNoDefault(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/NativeFont/getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/NativeFont/getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "countGlyphs([BI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NativeFont/countGlyphs([BI)I"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/NativeFont/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDelegateFont()Lsun/font/PhysicalFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "sun/font/NativeFont/getDelegateFont()Lsun/font/PhysicalFont;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_PhysicalFont) => void): void;
    public "getPlatformNameBytes(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/NativeFont/getPlatformNameBytes(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/NativeFont/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_geom_GeneralPath extends java_awt_geom_Path2D$Float {
    public static "java/awt/geom/GeneralPath/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/GeneralPath/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/GeneralPath/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/GeneralPath/<init>(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/GeneralPath/<init>(Ljava/awt/Shape;)V"(thread: JVMThread, args: [java_awt_Shape], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I[BI[FI)V"(thread: JVMThread, args: [number, JVMArray<number>, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/GeneralPath/<init>(I[BI[FI)V"(thread: JVMThread, args: [number, JVMArray<number>, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_awt_geom_Rectangle2D$Float extends java_awt_geom_Rectangle2D implements java_io_Serializable {
    public "java/awt/geom/Rectangle2D$Float/x": number;
    public "java/awt/geom/Rectangle2D$Float/y": number;
    public "java/awt/geom/Rectangle2D$Float/width": number;
    public "java/awt/geom/Rectangle2D$Float/height": number;
    public static "java/awt/geom/Rectangle2D$Float/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D$Float/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D$Float/<init>(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D$Float/getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D$Float/getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D$Float/getWidth()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D$Float/getHeight()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D$Float/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setRect(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D$Float/setRect(FFFF)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRect(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D$Float/setRect(DDDD)V"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setRect(Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Rectangle2D$Float/setRect(Ljava/awt/geom/Rectangle2D;)V"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "outcode(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Rectangle2D$Float/outcode(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/geom/Rectangle2D$Float/getBounds2D()Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "createIntersection(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/geom/Rectangle2D$Float/createIntersection(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "createUnion(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "java/awt/geom/Rectangle2D$Float/createUnion(Ljava/awt/geom/Rectangle2D;)Ljava/awt/geom/Rectangle2D;"(thread: JVMThread, args: [java_awt_geom_Rectangle2D], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/geom/Rectangle2D$Float/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_awt_geom_Point2D$Float extends java_awt_geom_Point2D implements java_io_Serializable {
    public "java/awt/geom/Point2D$Float/x": number;
    public "java/awt/geom/Point2D$Float/y": number;
    public static "java/awt/geom/Point2D$Float/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D$Float/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D$Float/<init>(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D$Float/getX()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/awt/geom/Point2D$Float/getY()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLocation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D$Float/setLocation(DD)V"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocation(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/awt/geom/Point2D$Float/setLocation(FF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/awt/geom/Point2D$Float/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_font_StrikeMetrics extends java_lang_Object {
    public "sun/font/StrikeMetrics/ascentX": number;
    public "sun/font/StrikeMetrics/ascentY": number;
    public "sun/font/StrikeMetrics/descentX": number;
    public "sun/font/StrikeMetrics/descentY": number;
    public "sun/font/StrikeMetrics/baselineX": number;
    public "sun/font/StrikeMetrics/baselineY": number;
    public "sun/font/StrikeMetrics/leadingX": number;
    public "sun/font/StrikeMetrics/leadingY": number;
    public "sun/font/StrikeMetrics/maxAdvanceX": number;
    public "sun/font/StrikeMetrics/maxAdvanceY": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StrikeMetrics/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(FFFFFFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StrikeMetrics/<init>(FFFFFFFFFF)V"(thread: JVMThread, args: [number, number, number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAscent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StrikeMetrics/getAscent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDescent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StrikeMetrics/getDescent()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLeading()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StrikeMetrics/getLeading()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMaxAdvance()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/StrikeMetrics/getMaxAdvance()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "merge(Lsun/font/StrikeMetrics;)V"(thread: JVMThread, args: [sun_font_StrikeMetrics], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StrikeMetrics/merge(Lsun/font/StrikeMetrics;)V"(thread: JVMThread, args: [sun_font_StrikeMetrics], cb?: (e?: java_lang_Throwable) => void): void;
    public "convertToUserSpace(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/StrikeMetrics/convertToUserSpace(Ljava/awt/geom/AffineTransform;)V"(thread: JVMThread, args: [java_awt_geom_AffineTransform], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/StrikeMetrics/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_font_Font2D extends java_lang_Object {
    public static "sun/font/Font2D/FONT_CONFIG_RANK": number;
    public static "sun/font/Font2D/JRE_RANK": number;
    public static "sun/font/Font2D/TTF_RANK": number;
    public static "sun/font/Font2D/TYPE1_RANK": number;
    public static "sun/font/Font2D/NATIVE_RANK": number;
    public static "sun/font/Font2D/UNKNOWN_RANK": number;
    public static "sun/font/Font2D/DEFAULT_RANK": number;
    public static "sun/font/Font2D/boldNames": JVMArray<java_lang_String>;
    public static "sun/font/Font2D/italicNames": JVMArray<java_lang_String>;
    public static "sun/font/Font2D/boldItalicNames": JVMArray<java_lang_String>;
    public static "sun/font/Font2D/DEFAULT_FRC": java_awt_font_FontRenderContext;
    public "sun/font/Font2D/handle": sun_font_Font2DHandle;
    public "sun/font/Font2D/familyName": java_lang_String;
    public "sun/font/Font2D/fullName": java_lang_String;
    public "sun/font/Font2D/style": number;
    public "sun/font/Font2D/family": sun_font_FontFamily;
    public "sun/font/Font2D/fontRank": number;
    public "sun/font/Font2D/mapper": sun_font_CharToGlyphMapper;
    public "sun/font/Font2D/strikeCache": java_util_concurrent_ConcurrentHashMap;
    public "sun/font/Font2D/lastFontStrike": java_lang_ref_Reference;
    public static "sun/font/Font2D/FWIDTH_NORMAL": number;
    public static "sun/font/Font2D/FWEIGHT_NORMAL": number;
    public static "sun/font/Font2D/FWEIGHT_BOLD": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/Font2D/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getStyle()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/getStyle()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setStyle()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/Font2D/setStyle()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/getWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getWeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/getWeight()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getRank()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/getRank()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setRank(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/Font2D/setRank(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMapper()Lsun/font/CharToGlyphMapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CharToGlyphMapper) => void): void;
    public "sun/font/Font2D/getMapper()Lsun/font/CharToGlyphMapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_font_CharToGlyphMapper) => void): void;
    public "getValidatedGlyphCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/getValidatedGlyphCode(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/Font2D/createStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "getStrike(Ljava/awt/Font;)Lsun/font/FontStrike;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/Font2D/getStrike(Ljava/awt/Font;)Lsun/font/FontStrike;"(thread: JVMThread, args: [java_awt_Font], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "getStrike(Ljava/awt/Font;Ljava/awt/geom/AffineTransform;II)Lsun/font/FontStrike;"(thread: JVMThread, args: [java_awt_Font, java_awt_geom_AffineTransform, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/Font2D/getStrike(Ljava/awt/Font;Ljava/awt/geom/AffineTransform;II)Lsun/font/FontStrike;"(thread: JVMThread, args: [java_awt_Font, java_awt_geom_AffineTransform, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "getStrike(Ljava/awt/Font;Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;II)Lsun/font/FontStrike;"(thread: JVMThread, args: [java_awt_Font, java_awt_geom_AffineTransform, java_awt_geom_AffineTransform, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/Font2D/getStrike(Ljava/awt/Font;Ljava/awt/geom/AffineTransform;Ljava/awt/geom/AffineTransform;II)Lsun/font/FontStrike;"(thread: JVMThread, args: [java_awt_Font, java_awt_geom_AffineTransform, java_awt_geom_AffineTransform, number, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "getStrike(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;)Lsun/font/FontStrike;"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/Font2D/getStrike(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;)Lsun/font/FontStrike;"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "getStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/Font2D/getStrike(Lsun/font/FontStrikeDesc;)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "getStrike(Lsun/font/FontStrikeDesc;Z)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "sun/font/Font2D/getStrike(Lsun/font/FontStrikeDesc;Z)Lsun/font/FontStrike;"(thread: JVMThread, args: [sun_font_FontStrikeDesc, number], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontStrike) => void): void;
    public "removeFromCache(Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/Font2D/removeFromCache(Lsun/font/FontStrikeDesc;)V"(thread: JVMThread, args: [sun_font_FontStrikeDesc], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontMetrics(Ljava/awt/Font;Ljava/awt/geom/AffineTransform;Ljava/lang/Object;Ljava/lang/Object;[F)V"(thread: JVMThread, args: [java_awt_Font, java_awt_geom_AffineTransform, java_lang_Object, java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/Font2D/getFontMetrics(Ljava/awt/Font;Ljava/awt/geom/AffineTransform;Ljava/lang/Object;Ljava/lang/Object;[F)V"(thread: JVMThread, args: [java_awt_Font, java_awt_geom_AffineTransform, java_lang_Object, java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getStyleMetrics(F[FI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/Font2D/getStyleMetrics(F[FI)V"(thread: JVMThread, args: [number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontMetrics(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;[F)V"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/Font2D/getFontMetrics(Ljava/awt/Font;Ljava/awt/font/FontRenderContext;[F)V"(thread: JVMThread, args: [java_awt_Font, java_awt_font_FontRenderContext, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTableBytes(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/font/Font2D/getTableBytes(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/Font2D/getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "supportsEncoding(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/supportsEncoding(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDoStyle(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/canDoStyle(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "useAAForPtSize(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/useAAForPtSize(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasSupplementaryChars()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/hasSupplementaryChars()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPostscriptName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/Font2D/getPostscriptName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFontName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/Font2D/getFontName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFamilyName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/Font2D/getFamilyName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charToGlyph(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/charToGlyph(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDisplay(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/canDisplay(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canDisplay(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/canDisplay(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBaselineFor(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/getBaselineFor(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getItalicAngle(Ljava/awt/Font;Ljava/awt/geom/AffineTransform;Ljava/lang/Object;Ljava/lang/Object;)F"(thread: JVMThread, args: [java_awt_Font, java_awt_geom_AffineTransform, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/Font2D/getItalicAngle(Ljava/awt/Font;Ljava/awt/geom/AffineTransform;Ljava/lang/Object;Ljava/lang/Object;)F"(thread: JVMThread, args: [java_awt_Font, java_awt_geom_AffineTransform, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/Font2D/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_FontFamily extends java_lang_Object {
    public static "sun/font/FontFamily/familyNameMap": java_util_concurrent_ConcurrentHashMap;
    public static "sun/font/FontFamily/allLocaleNames": java_util_HashMap;
    public "sun/font/FontFamily/familyName": java_lang_String;
    public "sun/font/FontFamily/plain": sun_font_Font2D;
    public "sun/font/FontFamily/bold": sun_font_Font2D;
    public "sun/font/FontFamily/italic": sun_font_Font2D;
    public "sun/font/FontFamily/bolditalic": sun_font_Font2D;
    public "sun/font/FontFamily/logicalFont": number;
    public "sun/font/FontFamily/familyRank": number;
    public "sun/font/FontFamily/familyWidth": number;
    public static "getFamily(Ljava/lang/String;)Lsun/font/FontFamily;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontFamily) => void): void;
    public static "sun/font/FontFamily/getFamily(Ljava/lang/String;)Lsun/font/FontFamily;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontFamily) => void): void;
    public static "getAllFamilyNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "sun/font/FontFamily/getAllFamilyNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "remove(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FontFamily/remove(Lsun/font/Font2D;)V"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;ZI)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontFamily/<init>(Ljava/lang/String;ZI)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontFamily/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFamilyName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/FontFamily/getFamilyName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRank()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontFamily/getRank()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFromSameSource(Lsun/font/Font2D;)Z"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontFamily/isFromSameSource(Lsun/font/Font2D;)Z"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "preferredWidth(Lsun/font/Font2D;)Z"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontFamily/preferredWidth(Lsun/font/Font2D;)Z"(thread: JVMThread, args: [sun_font_Font2D], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "closerWeight(Lsun/font/Font2D;Lsun/font/Font2D;I)Z"(thread: JVMThread, args: [sun_font_Font2D, sun_font_Font2D, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FontFamily/closerWeight(Lsun/font/Font2D;Lsun/font/Font2D;I)Z"(thread: JVMThread, args: [sun_font_Font2D, sun_font_Font2D, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setFont(Lsun/font/Font2D;I)V"(thread: JVMThread, args: [sun_font_Font2D, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FontFamily/setFont(Lsun/font/Font2D;I)V"(thread: JVMThread, args: [sun_font_Font2D, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontWithExactStyleMatch(I)Lsun/font/Font2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/FontFamily/getFontWithExactStyleMatch(I)Lsun/font/Font2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "getFont(I)Lsun/font/Font2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/FontFamily/getFont(I)Lsun/font/Font2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "getClosestStyle(I)Lsun/font/Font2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public "sun/font/FontFamily/getClosestStyle(I)Lsun/font/Font2D;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_font_Font2D) => void): void;
    public static "addLocaleNames(Lsun/font/FontFamily;[Ljava/lang/String;)V"(thread: JVMThread, args: [sun_font_FontFamily, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FontFamily/addLocaleNames(Lsun/font/FontFamily;[Ljava/lang/String;)V"(thread: JVMThread, args: [sun_font_FontFamily, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getLocaleFamily(Ljava/lang/String;)Lsun/font/FontFamily;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontFamily) => void): void;
    public static "sun/font/FontFamily/getLocaleFamily(Ljava/lang/String;)Lsun/font/FontFamily;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_font_FontFamily) => void): void;
    public static "getAllFontFamilies()[Lsun/font/FontFamily;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_FontFamily>) => void): void;
    public static "sun/font/FontFamily/getAllFontFamilies()[Lsun/font/FontFamily;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_font_FontFamily>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/font/FontFamily/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FontFamily/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_font_FreetypeFontScaler extends sun_font_FontScaler {
    public static "sun/font/FreetypeFontScaler/TRUETYPE_FONT": number;
    public static "sun/font/FreetypeFontScaler/TYPE1_FONT": number;
    public static "initIDs(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FreetypeFontScaler/initIDs(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "invalidateScaler()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FreetypeFontScaler/invalidateScaler()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/font/Font2D;IZI)V"(thread: JVMThread, args: [sun_font_Font2D, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FreetypeFontScaler/<init>(Lsun/font/Font2D;IZI)V"(thread: JVMThread, args: [sun_font_Font2D, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/FreetypeFontScaler/getFontMetrics(J)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphAdvance(JI)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphMetrics(JILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphImage(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphOutlineBounds(JI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphOutline(JIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphVectorOutline(J[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getLayoutTableCache()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FreetypeFontScaler/getLayoutTableCache()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FreetypeFontScaler/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FreetypeFontScaler/getNumGlyphs()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FreetypeFontScaler/getMissingGlyphCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphCode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphCode(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphPoint(JII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FreetypeFontScaler/getUnitsPerEm()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "createScalerContext([DIIFFZ)J"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FreetypeFontScaler/createScalerContext([DIIFFZ)J"(thread: JVMThread, args: [JVMArray<number>, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "initNativeScaler(Lsun/font/Font2D;IIZI)J"(thread: JVMThread, args: [sun_font_Font2D, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FreetypeFontScaler/initNativeScaler(Lsun/font/Font2D;IIZI)J"(thread: JVMThread, args: [sun_font_Font2D, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFontMetricsNative(Lsun/font/Font2D;JJ)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "sun/font/FreetypeFontScaler/getFontMetricsNative(Lsun/font/Font2D;JJ)Lsun/font/StrikeMetrics;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: sun_font_StrikeMetrics) => void): void;
    public "getGlyphAdvanceNative(Lsun/font/Font2D;JJI)F"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphAdvanceNative(Lsun/font/Font2D;JJI)F"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGlyphMetricsNative(Lsun/font/Font2D;JJILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphMetricsNative(Lsun/font/Font2D;JJILjava/awt/geom/Point2D$Float;)V"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number, java_awt_geom_Point2D$Float], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphImageNative(Lsun/font/Font2D;JJI)J"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphImageNative(Lsun/font/Font2D;JJI)J"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getGlyphOutlineBoundsNative(Lsun/font/Font2D;JJI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphOutlineBoundsNative(Lsun/font/Font2D;JJI)Ljava/awt/geom/Rectangle2D$Float;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Rectangle2D$Float) => void): void;
    public "getGlyphOutlineNative(Lsun/font/Font2D;JJIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphOutlineNative(Lsun/font/Font2D;JJIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphVectorOutlineNative(Lsun/font/Font2D;JJ[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphVectorOutlineNative(Lsun/font/Font2D;JJ[IIFF)Ljava/awt/geom/GeneralPath;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_GeneralPath) => void): void;
    public "getGlyphPointNative(Lsun/font/Font2D;JJII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphPointNative(Lsun/font/Font2D;JJII)Ljava/awt/geom/Point2D$Float;"(thread: JVMThread, args: [sun_font_Font2D, Long, any, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: java_awt_geom_Point2D$Float) => void): void;
    public "getLayoutTableCacheNative(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FreetypeFontScaler/getLayoutTableCacheNative(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "disposeNativeScaler(Lsun/font/Font2D;J)V"(thread: JVMThread, args: [sun_font_Font2D, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FreetypeFontScaler/disposeNativeScaler(Lsun/font/Font2D;J)V"(thread: JVMThread, args: [sun_font_Font2D, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGlyphCodeNative(Lsun/font/Font2D;JC)I"(thread: JVMThread, args: [sun_font_Font2D, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FreetypeFontScaler/getGlyphCodeNative(Lsun/font/Font2D;JC)I"(thread: JVMThread, args: [sun_font_Font2D, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumGlyphsNative(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FreetypeFontScaler/getNumGlyphsNative(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMissingGlyphCodeNative(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/font/FreetypeFontScaler/getMissingGlyphCodeNative(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getUnitsPerEMNative(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FreetypeFontScaler/getUnitsPerEMNative(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "createScalerContextNative(J[DIIFF)J"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/font/FreetypeFontScaler/createScalerContextNative(J[DIIFF)J"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "invalidateScalerContext(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/font/FreetypeFontScaler/invalidateScalerContext(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/font/FreetypeFontScaler/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_atomic_AtomicLong extends java_lang_Number implements java_io_Serializable {
    public static "java/util/concurrent/atomic/AtomicLong/serialVersionUID": Long;
    public static "java/util/concurrent/atomic/AtomicLong/unsafe": sun_misc_Unsafe;
    public static "java/util/concurrent/atomic/AtomicLong/valueOffset": Long;
    public static "java/util/concurrent/atomic/AtomicLong/VM_SUPPORTS_LONG_CAS": number;
    public "java/util/concurrent/atomic/AtomicLong/value": Long;
    public static "VMSupportsCS8()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/atomic/AtomicLong/VMSupportsCS8()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "get()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/get()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "set(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/set(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "lazySet(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/lazySet(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAndSet(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/getAndSet(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compareAndSet(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/compareAndSet(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "weakCompareAndSet(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/weakCompareAndSet(JJ)Z"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAndIncrement()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/getAndIncrement()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getAndDecrement()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/getAndDecrement()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getAndAdd(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/getAndAdd(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "incrementAndGet()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/incrementAndGet()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "decrementAndGet()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/decrementAndGet()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "addAndGet(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/addAndGet(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getAndUpdate(Ljava/util/function/LongUnaryOperator;)J"(thread: JVMThread, args: [java_util_function_LongUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/getAndUpdate(Ljava/util/function/LongUnaryOperator;)J"(thread: JVMThread, args: [java_util_function_LongUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "updateAndGet(Ljava/util/function/LongUnaryOperator;)J"(thread: JVMThread, args: [java_util_function_LongUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/updateAndGet(Ljava/util/function/LongUnaryOperator;)J"(thread: JVMThread, args: [java_util_function_LongUnaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getAndAccumulate(JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/getAndAccumulate(JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "accumulateAndGet(JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/accumulateAndGet(JLjava/util/function/LongBinaryOperator;)J"(thread: JVMThread, args: [Long, any, java_util_function_LongBinaryOperator], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/atomic/AtomicLong/doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/atomic/AtomicLong/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_jar_JarFile extends java_util_zip_ZipFile {
    public "java/util/jar/JarFile/manRef": java_lang_ref_SoftReference;
    public "java/util/jar/JarFile/manEntry": java_util_jar_JarEntry;
    public "java/util/jar/JarFile/jv": java_util_jar_JarVerifier;
    public "java/util/jar/JarFile/jvInitialized": number;
    public "java/util/jar/JarFile/verify": number;
    public "java/util/jar/JarFile/hasClassPathAttribute": number;
    public "java/util/jar/JarFile/hasCheckedSpecialAttributes": number;
    public static "java/util/jar/JarFile/MANIFEST_NAME": java_lang_String;
    public static "java/util/jar/JarFile/CLASSPATH_CHARS": JVMArray<number>;
    public static "java/util/jar/JarFile/CLASSPATH_LASTOCC": JVMArray<number>;
    public static "java/util/jar/JarFile/CLASSPATH_OPTOSFT": JVMArray<number>;
    public static "java/util/jar/JarFile/javaHome": java_lang_String;
    public static "java/util/jar/JarFile/jarNames": JVMArray<java_lang_String>;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;Z)V"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/<init>(Ljava/io/File;Z)V"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;ZI)V"(thread: JVMThread, args: [java_io_File, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/<init>(Ljava/io/File;ZI)V"(thread: JVMThread, args: [java_io_File, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getManifest()Ljava/util/jar/Manifest;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public "java/util/jar/JarFile/getManifest()Ljava/util/jar/Manifest;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public "getManifestFromReference()Ljava/util/jar/Manifest;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public "java/util/jar/JarFile/getManifestFromReference()Ljava/util/jar/Manifest;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public "getMetaInfEntryNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/jar/JarFile/getMetaInfEntryNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getJarEntry(Ljava/lang/String;)Ljava/util/jar/JarEntry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarEntry) => void): void;
    public "java/util/jar/JarFile/getJarEntry(Ljava/lang/String;)Ljava/util/jar/JarEntry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarEntry) => void): void;
    public "getEntry(Ljava/lang/String;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "java/util/jar/JarFile/getEntry(Ljava/lang/String;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "entries()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/jar/JarFile/entries()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    public "java/util/jar/JarFile/stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    public "maybeInstantiateVerifier()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/maybeInstantiateVerifier()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "initializeVerifier()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/initializeVerifier()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBytes(Ljava/util/zip/ZipEntry;)[B"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/util/jar/JarFile/getBytes(Ljava/util/zip/ZipEntry;)[B"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getInputStream(Ljava/util/zip/ZipEntry;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/util/jar/JarFile/getInputStream(Ljava/util/zip/ZipEntry;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getManEntry()Ljava/util/jar/JarEntry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarEntry) => void): void;
    public "java/util/jar/JarFile/getManEntry()Ljava/util/jar/JarEntry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarEntry) => void): void;
    public "hasClassPathAttribute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/JarFile/hasClassPathAttribute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "match([C[B[I[I)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/JarFile/match([C[B[I[I)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkForSpecialAttributes()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/checkForSpecialAttributes()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isKnownNotToHaveSpecialAttributes()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/JarFile/isKnownNotToHaveSpecialAttributes()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureInitialization()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/ensureInitialization()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "newEntry(Ljava/util/zip/ZipEntry;)Ljava/util/jar/JarEntry;"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarEntry) => void): void;
    public "java/util/jar/JarFile/newEntry(Ljava/util/zip/ZipEntry;)Ljava/util/jar/JarEntry;"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarEntry) => void): void;
    public "entryNames([Ljava/security/CodeSource;)Ljava/util/Enumeration;"(thread: JVMThread, args: [JVMArray<java_security_CodeSource>], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/jar/JarFile/entryNames([Ljava/security/CodeSource;)Ljava/util/Enumeration;"(thread: JVMThread, args: [JVMArray<java_security_CodeSource>], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "entries2()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/jar/JarFile/entries2()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getCodeSources(Ljava/net/URL;)[Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSource>) => void): void;
    public "java/util/jar/JarFile/getCodeSources(Ljava/net/URL;)[Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSource>) => void): void;
    public "unsignedEntryNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/jar/JarFile/unsignedEntryNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getCodeSource(Ljava/net/URL;Ljava/lang/String;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "java/util/jar/JarFile/getCodeSource(Ljava/net/URL;Ljava/lang/String;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "setEagerValidation(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarFile/setEagerValidation(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getManifestDigests()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/jar/JarFile/getManifestDigests()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "access$001(Ljava/util/jar/JarFile;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_util_jar_JarFile], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "java/util/jar/JarFile/access$001(Ljava/util/jar/JarFile;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_util_jar_JarFile], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "access$200(Ljava/util/jar/JarFile;)V"(thread: JVMThread, args: [java_util_jar_JarFile], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/jar/JarFile/access$200(Ljava/util/jar/JarFile;)V"(thread: JVMThread, args: [java_util_jar_JarFile], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$300(Ljava/util/jar/JarFile;)Ljava/util/jar/JarVerifier;"(thread: JVMThread, args: [java_util_jar_JarFile], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarVerifier) => void): void;
    public static "java/util/jar/JarFile/access$300(Ljava/util/jar/JarFile;)Ljava/util/jar/JarVerifier;"(thread: JVMThread, args: [java_util_jar_JarFile], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarVerifier) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/jar/JarFile/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_CodeSource extends java_lang_Object implements java_io_Serializable {
    public static "java/security/CodeSource/serialVersionUID": Long;
    public "java/security/CodeSource/location": java_net_URL;
    public "java/security/CodeSource/signers": JVMArray<java_security_CodeSigner>;
    public "java/security/CodeSource/certs": JVMArray<java_security_cert_Certificate>;
    public "java/security/CodeSource/sp": java_net_SocketPermission;
    public "java/security/CodeSource/factory": java_security_cert_CertificateFactory;
    public "<init>(Ljava/net/URL;[Ljava/security/cert/Certificate;)V"(thread: JVMThread, args: [java_net_URL, JVMArray<java_security_cert_Certificate>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/CodeSource/<init>(Ljava/net/URL;[Ljava/security/cert/Certificate;)V"(thread: JVMThread, args: [java_net_URL, JVMArray<java_security_cert_Certificate>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/URL;[Ljava/security/CodeSigner;)V"(thread: JVMThread, args: [java_net_URL, JVMArray<java_security_CodeSigner>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/CodeSource/<init>(Ljava/net/URL;[Ljava/security/CodeSigner;)V"(thread: JVMThread, args: [java_net_URL, JVMArray<java_security_CodeSigner>], cb?: (e?: java_lang_Throwable) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/CodeSource/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/CodeSource/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLocation()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "java/security/CodeSource/getLocation()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "getCertificates()[Ljava/security/cert/Certificate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "java/security/CodeSource/getCertificates()[Ljava/security/cert/Certificate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "getCodeSigners()[Ljava/security/CodeSigner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "java/security/CodeSource/getCodeSigners()[Ljava/security/CodeSigner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "implies(Ljava/security/CodeSource;)Z"(thread: JVMThread, args: [java_security_CodeSource], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/CodeSource/implies(Ljava/security/CodeSource;)Z"(thread: JVMThread, args: [java_security_CodeSource], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "matchCerts(Ljava/security/CodeSource;Z)Z"(thread: JVMThread, args: [java_security_CodeSource, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/CodeSource/matchCerts(Ljava/security/CodeSource;Z)Z"(thread: JVMThread, args: [java_security_CodeSource, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "matchLocation(Ljava/security/CodeSource;)Z"(thread: JVMThread, args: [java_security_CodeSource], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/CodeSource/matchLocation(Ljava/security/CodeSource;)Z"(thread: JVMThread, args: [java_security_CodeSource], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/CodeSource/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/CodeSource/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/CodeSource/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "convertCertArrayToSignerArray([Ljava/security/cert/Certificate;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [JVMArray<java_security_cert_Certificate>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "java/security/CodeSource/convertCertArrayToSignerArray([Ljava/security/cert/Certificate;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [JVMArray<java_security_cert_Certificate>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
  }
  export class java_security_cert_CertificateFactory extends java_lang_Object {
    public "java/security/cert/CertificateFactory/type": java_lang_String;
    public "java/security/cert/CertificateFactory/provider": java_security_Provider;
    public "java/security/cert/CertificateFactory/certFacSpi": java_security_cert_CertificateFactorySpi;
    public "<init>(Ljava/security/cert/CertificateFactorySpi;Ljava/security/Provider;Ljava/lang/String;)V"(thread: JVMThread, args: [java_security_cert_CertificateFactorySpi, java_security_Provider, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/cert/CertificateFactory/<init>(Ljava/security/cert/CertificateFactorySpi;Ljava/security/Provider;Ljava/lang/String;)V"(thread: JVMThread, args: [java_security_cert_CertificateFactorySpi, java_security_Provider, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance(Ljava/lang/String;)Ljava/security/cert/CertificateFactory;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertificateFactory) => void): void;
    public static "java/security/cert/CertificateFactory/getInstance(Ljava/lang/String;)Ljava/security/cert/CertificateFactory;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertificateFactory) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/lang/String;)Ljava/security/cert/CertificateFactory;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertificateFactory) => void): void;
    public static "java/security/cert/CertificateFactory/getInstance(Ljava/lang/String;Ljava/lang/String;)Ljava/security/cert/CertificateFactory;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertificateFactory) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/security/Provider;)Ljava/security/cert/CertificateFactory;"(thread: JVMThread, args: [java_lang_String, java_security_Provider], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertificateFactory) => void): void;
    public static "java/security/cert/CertificateFactory/getInstance(Ljava/lang/String;Ljava/security/Provider;)Ljava/security/cert/CertificateFactory;"(thread: JVMThread, args: [java_lang_String, java_security_Provider], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertificateFactory) => void): void;
    public "getProvider()Ljava/security/Provider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public "java/security/cert/CertificateFactory/getProvider()Ljava/security/Provider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public "getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/cert/CertificateFactory/getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "generateCertificate(Ljava/io/InputStream;)Ljava/security/cert/Certificate;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_Certificate) => void): void;
    public "java/security/cert/CertificateFactory/generateCertificate(Ljava/io/InputStream;)Ljava/security/cert/Certificate;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_Certificate) => void): void;
    public "getCertPathEncodings()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/security/cert/CertificateFactory/getCertPathEncodings()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "generateCertPath(Ljava/io/InputStream;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "java/security/cert/CertificateFactory/generateCertPath(Ljava/io/InputStream;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "generateCertPath(Ljava/io/InputStream;Ljava/lang/String;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_io_InputStream, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "java/security/cert/CertificateFactory/generateCertPath(Ljava/io/InputStream;Ljava/lang/String;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_io_InputStream, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "generateCertPath(Ljava/util/List;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "java/security/cert/CertificateFactory/generateCertPath(Ljava/util/List;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "generateCertificates(Ljava/io/InputStream;)Ljava/util/Collection;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/security/cert/CertificateFactory/generateCertificates(Ljava/io/InputStream;)Ljava/util/Collection;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "generateCRL(Ljava/io/InputStream;)Ljava/security/cert/CRL;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CRL) => void): void;
    public "java/security/cert/CertificateFactory/generateCRL(Ljava/io/InputStream;)Ljava/security/cert/CRL;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CRL) => void): void;
    public "generateCRLs(Ljava/io/InputStream;)Ljava/util/Collection;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/security/cert/CertificateFactory/generateCRLs(Ljava/io/InputStream;)Ljava/util/Collection;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
  }
  export class java_security_cert_CRL extends java_lang_Object {
    public "java/security/cert/CRL/type": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/cert/CRL/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/cert/CRL/getType()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/cert/CRL/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isRevoked(Ljava/security/cert/Certificate;)Z"(thread: JVMThread, args: [java_security_cert_Certificate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/cert/CRL/isRevoked(Ljava/security/cert/Certificate;)Z"(thread: JVMThread, args: [java_security_cert_Certificate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_security_cert_CertificateFactorySpi extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/cert/CertificateFactorySpi/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineGenerateCertificate(Ljava/io/InputStream;)Ljava/security/cert/Certificate;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_Certificate) => void): void;
    public "java/security/cert/CertificateFactorySpi/engineGenerateCertificate(Ljava/io/InputStream;)Ljava/security/cert/Certificate;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_Certificate) => void): void;
    public "engineGenerateCertPath(Ljava/io/InputStream;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "java/security/cert/CertificateFactorySpi/engineGenerateCertPath(Ljava/io/InputStream;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "engineGenerateCertPath(Ljava/io/InputStream;Ljava/lang/String;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_io_InputStream, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "java/security/cert/CertificateFactorySpi/engineGenerateCertPath(Ljava/io/InputStream;Ljava/lang/String;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_io_InputStream, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "engineGenerateCertPath(Ljava/util/List;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "java/security/cert/CertificateFactorySpi/engineGenerateCertPath(Ljava/util/List;)Ljava/security/cert/CertPath;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CertPath) => void): void;
    public "engineGetCertPathEncodings()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/security/cert/CertificateFactorySpi/engineGetCertPathEncodings()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "engineGenerateCertificates(Ljava/io/InputStream;)Ljava/util/Collection;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/security/cert/CertificateFactorySpi/engineGenerateCertificates(Ljava/io/InputStream;)Ljava/util/Collection;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "engineGenerateCRL(Ljava/io/InputStream;)Ljava/security/cert/CRL;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CRL) => void): void;
    public "java/security/cert/CertificateFactorySpi/engineGenerateCRL(Ljava/io/InputStream;)Ljava/security/cert/CRL;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_security_cert_CRL) => void): void;
    public "engineGenerateCRLs(Ljava/io/InputStream;)Ljava/util/Collection;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/security/cert/CertificateFactorySpi/engineGenerateCRLs(Ljava/io/InputStream;)Ljava/util/Collection;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
  }
  export class java_net_SocketPermission extends java_security_Permission implements java_io_Serializable {
    public static "java/net/SocketPermission/serialVersionUID": Long;
    public static "java/net/SocketPermission/CONNECT": number;
    public static "java/net/SocketPermission/LISTEN": number;
    public static "java/net/SocketPermission/ACCEPT": number;
    public static "java/net/SocketPermission/RESOLVE": number;
    public static "java/net/SocketPermission/NONE": number;
    public static "java/net/SocketPermission/ALL": number;
    public static "java/net/SocketPermission/PORT_MIN": number;
    public static "java/net/SocketPermission/PORT_MAX": number;
    public static "java/net/SocketPermission/PRIV_PORT_MAX": number;
    public static "java/net/SocketPermission/DEF_EPH_LOW": number;
    public "java/net/SocketPermission/mask": number;
    public "java/net/SocketPermission/actions": java_lang_String;
    public "java/net/SocketPermission/hostname": java_lang_String;
    public "java/net/SocketPermission/cname": java_lang_String;
    public "java/net/SocketPermission/addresses": JVMArray<java_net_InetAddress>;
    public "java/net/SocketPermission/wildcard": number;
    public "java/net/SocketPermission/init_with_ip": number;
    public "java/net/SocketPermission/invalid": number;
    public "java/net/SocketPermission/portrange": JVMArray<number>;
    public "java/net/SocketPermission/defaultDeny": number;
    public "java/net/SocketPermission/untrusted": number;
    public "java/net/SocketPermission/trusted": number;
    public static "java/net/SocketPermission/trustNameService": number;
    public static "java/net/SocketPermission/debug": sun_security_util_Debug;
    public static "java/net/SocketPermission/debugInit": number;
    public "java/net/SocketPermission/cdomain": java_lang_String;
    public "java/net/SocketPermission/hdomain": java_lang_String;
    public static "getDebug()Lsun/security/util/Debug;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_Debug) => void): void;
    public static "java/net/SocketPermission/getDebug()Lsun/security/util/Debug;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_Debug) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketPermission/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketPermission/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDeny()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketPermission/setDeny()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getHost(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/SocketPermission/getHost(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "parsePort(Ljava/lang/String;)[I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/net/SocketPermission/parsePort(Ljava/lang/String;)[I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "includesEphemerals()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/includesEphemerals()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "init(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketPermission/init(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getMask(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/SocketPermission/getMask(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isUntrusted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/isUntrusted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCanonName()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketPermission/getCanonName()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "match(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/match(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "authorized(Ljava/lang/String;[B)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/authorized(Ljava/lang/String;[B)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "authorizedIPv4(Ljava/lang/String;[B)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/authorizedIPv4(Ljava/lang/String;[B)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "authorizedIPv6(Ljava/lang/String;[B)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/authorizedIPv6(Ljava/lang/String;[B)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getIP()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketPermission/getIP()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "impliesIgnoreMask(Ljava/net/SocketPermission;)Z"(thread: JVMThread, args: [java_net_SocketPermission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/impliesIgnoreMask(Ljava/net/SocketPermission;)Z"(thread: JVMThread, args: [java_net_SocketPermission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareHostnames(Ljava/net/SocketPermission;)Z"(thread: JVMThread, args: [java_net_SocketPermission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/compareHostnames(Ljava/net/SocketPermission;)Z"(thread: JVMThread, args: [java_net_SocketPermission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getMask()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketPermission/getMask()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getActions(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/SocketPermission/getActions(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getActions()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/SocketPermission/getActions()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "newPermissionCollection()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public "java/net/SocketPermission/newPermissionCollection()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketPermission/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketPermission/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initEphemeralPorts(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/SocketPermission/initEphemeralPorts(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "inRange(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/SocketPermission/inRange(IIII)Z"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/SocketPermission/access$000(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/SocketPermission/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_zip_ZipEntry extends java_lang_Object implements java_util_zip_ZipConstants, java_lang_Cloneable {
    public "java/util/zip/ZipEntry/name": java_lang_String;
    public "java/util/zip/ZipEntry/xdostime": Long;
    public "java/util/zip/ZipEntry/mtime": java_nio_file_attribute_FileTime;
    public "java/util/zip/ZipEntry/atime": java_nio_file_attribute_FileTime;
    public "java/util/zip/ZipEntry/ctime": java_nio_file_attribute_FileTime;
    public "java/util/zip/ZipEntry/crc": Long;
    public "java/util/zip/ZipEntry/size": Long;
    public "java/util/zip/ZipEntry/csize": Long;
    public "java/util/zip/ZipEntry/method": number;
    public "java/util/zip/ZipEntry/flag": number;
    public "java/util/zip/ZipEntry/extra": JVMArray<number>;
    public "java/util/zip/ZipEntry/comment": java_lang_String;
    public static "java/util/zip/ZipEntry/STORED": number;
    public static "java/util/zip/ZipEntry/DEFLATED": number;
    public static "java/util/zip/ZipEntry/DOSTIME_BEFORE_1980": Long;
    public static "java/util/zip/ZipEntry/UPPER_DOSTIME_BOUND": Long;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/zip/ZipEntry;)V"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/<init>(Ljava/util/zip/ZipEntry;)V"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/zip/ZipEntry/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setTime(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/setTime(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/ZipEntry/getTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setLastModifiedTime(Ljava/nio/file/attribute/FileTime;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_nio_file_attribute_FileTime], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "java/util/zip/ZipEntry/setLastModifiedTime(Ljava/nio/file/attribute/FileTime;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_nio_file_attribute_FileTime], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "getLastModifiedTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "java/util/zip/ZipEntry/getLastModifiedTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "setLastAccessTime(Ljava/nio/file/attribute/FileTime;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_nio_file_attribute_FileTime], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "java/util/zip/ZipEntry/setLastAccessTime(Ljava/nio/file/attribute/FileTime;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_nio_file_attribute_FileTime], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "getLastAccessTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "java/util/zip/ZipEntry/getLastAccessTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "setCreationTime(Ljava/nio/file/attribute/FileTime;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_nio_file_attribute_FileTime], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "java/util/zip/ZipEntry/setCreationTime(Ljava/nio/file/attribute/FileTime;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_nio_file_attribute_FileTime], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "getCreationTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "java/util/zip/ZipEntry/getCreationTime()Ljava/nio/file/attribute/FileTime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_attribute_FileTime) => void): void;
    public "setSize(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/setSize(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/ZipEntry/getSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getCompressedSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/ZipEntry/getCompressedSize()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setCompressedSize(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/setCompressedSize(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCrc(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/setCrc(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCrc()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/ZipEntry/getCrc()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setMethod(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/setMethod(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMethod()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/ZipEntry/getMethod()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setExtra([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/setExtra([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setExtra0([BZ)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/setExtra0([BZ)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getExtra()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/util/zip/ZipEntry/getExtra()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "setComment(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipEntry/setComment(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComment()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/zip/ZipEntry/getComment()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/ZipEntry/isDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/zip/ZipEntry/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/ZipEntry/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/zip/ZipEntry/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export interface java_util_zip_ZipConstants extends java_lang_Object {
  }
  export class java_util_jar_JarVerifier extends java_lang_Object {
    public static "java/util/jar/JarVerifier/debug": sun_security_util_Debug;
    public "java/util/jar/JarVerifier/verifiedSigners": java_util_Hashtable;
    public "java/util/jar/JarVerifier/sigFileSigners": java_util_Hashtable;
    public "java/util/jar/JarVerifier/sigFileData": java_util_Hashtable;
    public "java/util/jar/JarVerifier/pendingBlocks": java_util_ArrayList;
    public "java/util/jar/JarVerifier/signerCache": java_util_ArrayList;
    public "java/util/jar/JarVerifier/parsingBlockOrSF": number;
    public "java/util/jar/JarVerifier/parsingMeta": number;
    public "java/util/jar/JarVerifier/anyToVerify": number;
    public "java/util/jar/JarVerifier/baos": java_io_ByteArrayOutputStream;
    public "java/util/jar/JarVerifier/manDig": sun_security_util_ManifestDigester;
    public "java/util/jar/JarVerifier/manifestRawBytes": JVMArray<number>;
    public "java/util/jar/JarVerifier/eagerValidation": number;
    public "java/util/jar/JarVerifier/csdomain": java_lang_Object;
    public "java/util/jar/JarVerifier/manifestDigests": java_util_List;
    public "java/util/jar/JarVerifier/urlToCodeSourceMap": java_util_Map;
    public "java/util/jar/JarVerifier/signerToCodeSource": java_util_Map;
    public "java/util/jar/JarVerifier/lastURL": java_net_URL;
    public "java/util/jar/JarVerifier/lastURLMap": java_util_Map;
    public "java/util/jar/JarVerifier/emptySigner": JVMArray<java_security_CodeSigner>;
    public "java/util/jar/JarVerifier/signerMap": java_util_Map;
    public "java/util/jar/JarVerifier/emptyEnumeration": java_util_Enumeration;
    public "java/util/jar/JarVerifier/jarCodeSigners": java_util_List;
    public "<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarVerifier/<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "beginEntry(Ljava/util/jar/JarEntry;Lsun/security/util/ManifestEntryVerifier;)V"(thread: JVMThread, args: [java_util_jar_JarEntry, sun_security_util_ManifestEntryVerifier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarVerifier/beginEntry(Ljava/util/jar/JarEntry;Lsun/security/util/ManifestEntryVerifier;)V"(thread: JVMThread, args: [java_util_jar_JarEntry, sun_security_util_ManifestEntryVerifier], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(ILsun/security/util/ManifestEntryVerifier;)V"(thread: JVMThread, args: [number, sun_security_util_ManifestEntryVerifier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarVerifier/update(ILsun/security/util/ManifestEntryVerifier;)V"(thread: JVMThread, args: [number, sun_security_util_ManifestEntryVerifier], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(I[BIILsun/security/util/ManifestEntryVerifier;)V"(thread: JVMThread, args: [number, JVMArray<number>, number, number, sun_security_util_ManifestEntryVerifier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarVerifier/update(I[BIILsun/security/util/ManifestEntryVerifier;)V"(thread: JVMThread, args: [number, JVMArray<number>, number, number, sun_security_util_ManifestEntryVerifier], cb?: (e?: java_lang_Throwable) => void): void;
    public "processEntry(Lsun/security/util/ManifestEntryVerifier;)V"(thread: JVMThread, args: [sun_security_util_ManifestEntryVerifier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarVerifier/processEntry(Lsun/security/util/ManifestEntryVerifier;)V"(thread: JVMThread, args: [sun_security_util_ManifestEntryVerifier], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCerts(Ljava/lang/String;)[Ljava/security/cert/Certificate;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "java/util/jar/JarVerifier/getCerts(Ljava/lang/String;)[Ljava/security/cert/Certificate;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "getCerts(Ljava/util/jar/JarFile;Ljava/util/jar/JarEntry;)[Ljava/security/cert/Certificate;"(thread: JVMThread, args: [java_util_jar_JarFile, java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "java/util/jar/JarVerifier/getCerts(Ljava/util/jar/JarFile;Ljava/util/jar/JarEntry;)[Ljava/security/cert/Certificate;"(thread: JVMThread, args: [java_util_jar_JarFile, java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "getCodeSigners(Ljava/lang/String;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "java/util/jar/JarVerifier/getCodeSigners(Ljava/lang/String;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "getCodeSigners(Ljava/util/jar/JarFile;Ljava/util/jar/JarEntry;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [java_util_jar_JarFile, java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "java/util/jar/JarVerifier/getCodeSigners(Ljava/util/jar/JarFile;Ljava/util/jar/JarEntry;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [java_util_jar_JarFile, java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public static "mapSignersToCertArray([Ljava/security/CodeSigner;)[Ljava/security/cert/Certificate;"(thread: JVMThread, args: [JVMArray<java_security_CodeSigner>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public static "java/util/jar/JarVerifier/mapSignersToCertArray([Ljava/security/CodeSigner;)[Ljava/security/cert/Certificate;"(thread: JVMThread, args: [JVMArray<java_security_CodeSigner>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "nothingToVerify()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/jar/JarVerifier/nothingToVerify()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doneWithMeta()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarVerifier/doneWithMeta()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mapSignersToCodeSource(Ljava/net/URL;[Ljava/security/CodeSigner;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, JVMArray<java_security_CodeSigner>], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "java/util/jar/JarVerifier/mapSignersToCodeSource(Ljava/net/URL;[Ljava/security/CodeSigner;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, JVMArray<java_security_CodeSigner>], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "mapSignersToCodeSources(Ljava/net/URL;Ljava/util/List;Z)[Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, java_util_List, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSource>) => void): void;
    public "java/util/jar/JarVerifier/mapSignersToCodeSources(Ljava/net/URL;Ljava/util/List;Z)[Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, java_util_List, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSource>) => void): void;
    public "findMatchingSigners(Ljava/security/CodeSource;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [java_security_CodeSource], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "java/util/jar/JarVerifier/findMatchingSigners(Ljava/security/CodeSource;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [java_security_CodeSource], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "signerMap()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/util/jar/JarVerifier/signerMap()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "entryNames(Ljava/util/jar/JarFile;[Ljava/security/CodeSource;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_util_jar_JarFile, JVMArray<java_security_CodeSource>], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/jar/JarVerifier/entryNames(Ljava/util/jar/JarFile;[Ljava/security/CodeSource;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_util_jar_JarFile, JVMArray<java_security_CodeSource>], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "entries2(Ljava/util/jar/JarFile;Ljava/util/Enumeration;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_util_jar_JarFile, java_util_Enumeration], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/jar/JarVerifier/entries2(Ljava/util/jar/JarFile;Ljava/util/Enumeration;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_util_jar_JarFile, java_util_Enumeration], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "isSigningRelated(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/jar/JarVerifier/isSigningRelated(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unsignedEntryNames(Ljava/util/jar/JarFile;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_util_jar_JarFile], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/jar/JarVerifier/unsignedEntryNames(Ljava/util/jar/JarFile;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_util_jar_JarFile], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getJarCodeSigners()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/jar/JarVerifier/getJarCodeSigners()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getCodeSources(Ljava/util/jar/JarFile;Ljava/net/URL;)[Ljava/security/CodeSource;"(thread: JVMThread, args: [java_util_jar_JarFile, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSource>) => void): void;
    public "java/util/jar/JarVerifier/getCodeSources(Ljava/util/jar/JarFile;Ljava/net/URL;)[Ljava/security/CodeSource;"(thread: JVMThread, args: [java_util_jar_JarFile, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSource>) => void): void;
    public "getCodeSource(Ljava/net/URL;Ljava/lang/String;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "java/util/jar/JarVerifier/getCodeSource(Ljava/net/URL;Ljava/lang/String;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "getCodeSource(Ljava/net/URL;Ljava/util/jar/JarFile;Ljava/util/jar/JarEntry;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, java_util_jar_JarFile, java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "java/util/jar/JarVerifier/getCodeSource(Ljava/net/URL;Ljava/util/jar/JarFile;Ljava/util/jar/JarEntry;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL, java_util_jar_JarFile, java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "setEagerValidation(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarVerifier/setEagerValidation(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getManifestDigests()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/jar/JarVerifier/getManifestDigests()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "getUnsignedCS(Ljava/net/URL;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public static "java/util/jar/JarVerifier/getUnsignedCS(Ljava/net/URL;)Ljava/security/CodeSource;"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/jar/JarVerifier/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_security_util_ManifestEntryVerifier extends java_lang_Object {
    public static "sun/security/util/ManifestEntryVerifier/debug": sun_security_util_Debug;
    public "sun/security/util/ManifestEntryVerifier/createdDigests": java_util_HashMap;
    public "sun/security/util/ManifestEntryVerifier/digests": java_util_ArrayList;
    public "sun/security/util/ManifestEntryVerifier/manifestHashes": java_util_ArrayList;
    public "sun/security/util/ManifestEntryVerifier/name": java_lang_String;
    public "sun/security/util/ManifestEntryVerifier/man": java_util_jar_Manifest;
    public "sun/security/util/ManifestEntryVerifier/skip": number;
    public "sun/security/util/ManifestEntryVerifier/entry": java_util_jar_JarEntry;
    public "sun/security/util/ManifestEntryVerifier/signers": JVMArray<java_security_CodeSigner>;
    public static "sun/security/util/ManifestEntryVerifier/hexc": JVMArray<number>;
    public "<init>(Ljava/util/jar/Manifest;)V"(thread: JVMThread, args: [java_util_jar_Manifest], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/ManifestEntryVerifier/<init>(Ljava/util/jar/Manifest;)V"(thread: JVMThread, args: [java_util_jar_Manifest], cb?: (e?: java_lang_Throwable) => void): void;
    public "setEntry(Ljava/lang/String;Ljava/util/jar/JarEntry;)V"(thread: JVMThread, args: [java_lang_String, java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/ManifestEntryVerifier/setEntry(Ljava/lang/String;Ljava/util/jar/JarEntry;)V"(thread: JVMThread, args: [java_lang_String, java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/ManifestEntryVerifier/update(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "update([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/ManifestEntryVerifier/update([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEntry()Ljava/util/jar/JarEntry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarEntry) => void): void;
    public "sun/security/util/ManifestEntryVerifier/getEntry()Ljava/util/jar/JarEntry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_JarEntry) => void): void;
    public "verify(Ljava/util/Hashtable;Ljava/util/Hashtable;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [java_util_Hashtable, java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "sun/security/util/ManifestEntryVerifier/verify(Ljava/util/Hashtable;Ljava/util/Hashtable;)[Ljava/security/CodeSigner;"(thread: JVMThread, args: [java_util_Hashtable, java_util_Hashtable], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public static "toHex([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "sun/security/util/ManifestEntryVerifier/toHex([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/security/util/ManifestEntryVerifier/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_security_util_ManifestDigester extends java_lang_Object {
    public static "sun/security/util/ManifestDigester/MF_MAIN_ATTRS": java_lang_String;
    public "sun/security/util/ManifestDigester/rawBytes": JVMArray<number>;
    public "sun/security/util/ManifestDigester/entries": java_util_HashMap;
    public "findSection(ILsun/security/util/ManifestDigester$Position;)Z"(thread: JVMThread, args: [number, sun_security_util_ManifestDigester$Position], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/security/util/ManifestDigester/findSection(ILsun/security/util/ManifestDigester$Position;)Z"(thread: JVMThread, args: [number, sun_security_util_ManifestDigester$Position], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/ManifestDigester/<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "isNameAttr([BI)Z"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/security/util/ManifestDigester/isNameAttr([BI)Z"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;Z)Lsun/security/util/ManifestDigester$Entry;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_ManifestDigester$Entry) => void): void;
    public "sun/security/util/ManifestDigester/get(Ljava/lang/String;Z)Lsun/security/util/ManifestDigester$Entry;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_ManifestDigester$Entry) => void): void;
    public "manifestDigest(Ljava/security/MessageDigest;)[B"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/security/util/ManifestDigester/manifestDigest(Ljava/security/MessageDigest;)[B"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export class java_security_MessageDigest extends java_security_MessageDigestSpi {
    public static "java/security/MessageDigest/pdebug": sun_security_util_Debug;
    public static "java/security/MessageDigest/skipDebug": number;
    public "java/security/MessageDigest/algorithm": java_lang_String;
    public static "java/security/MessageDigest/INITIAL": number;
    public static "java/security/MessageDigest/IN_PROGRESS": number;
    public "java/security/MessageDigest/state": number;
    public "java/security/MessageDigest/provider": java_security_Provider;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigest/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getInstance(Ljava/lang/String;)Ljava/security/MessageDigest;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_MessageDigest) => void): void;
    public static "java/security/MessageDigest/getInstance(Ljava/lang/String;)Ljava/security/MessageDigest;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_MessageDigest) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/lang/String;)Ljava/security/MessageDigest;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_MessageDigest) => void): void;
    public static "java/security/MessageDigest/getInstance(Ljava/lang/String;Ljava/lang/String;)Ljava/security/MessageDigest;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_security_MessageDigest) => void): void;
    public static "getInstance(Ljava/lang/String;Ljava/security/Provider;)Ljava/security/MessageDigest;"(thread: JVMThread, args: [java_lang_String, java_security_Provider], cb?: (e?: java_lang_Throwable, rv?: java_security_MessageDigest) => void): void;
    public static "java/security/MessageDigest/getInstance(Ljava/lang/String;Ljava/security/Provider;)Ljava/security/MessageDigest;"(thread: JVMThread, args: [java_lang_String, java_security_Provider], cb?: (e?: java_lang_Throwable, rv?: java_security_MessageDigest) => void): void;
    public "getProvider()Ljava/security/Provider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public "java/security/MessageDigest/getProvider()Ljava/security/Provider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public "update(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigest/update(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "update([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigest/update([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "update([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigest/update([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigest/update(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "digest()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/security/MessageDigest/digest()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "digest([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/MessageDigest/digest([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "digest([B)[B"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/security/MessageDigest/digest([B)[B"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/MessageDigest/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "isEqual([B[B)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/security/MessageDigest/isEqual([B[B)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigest/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAlgorithm()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/MessageDigest/getAlgorithm()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDigestLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/MessageDigest/getDigestLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/MessageDigest/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$000(Ljava/security/MessageDigest;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/security/MessageDigest/access$000(Ljava/security/MessageDigest;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$102(Ljava/security/MessageDigest;Ljava/security/Provider;)Ljava/security/Provider;"(thread: JVMThread, args: [java_security_MessageDigest, java_security_Provider], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public static "java/security/MessageDigest/access$102(Ljava/security/MessageDigest;Ljava/security/Provider;)Ljava/security/Provider;"(thread: JVMThread, args: [java_security_MessageDigest, java_security_Provider], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public static "access$100(Ljava/security/MessageDigest;)Ljava/security/Provider;"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public static "java/security/MessageDigest/access$100(Ljava/security/MessageDigest;)Ljava/security/Provider;"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: java_security_Provider) => void): void;
    public static "access$202(Ljava/security/MessageDigest;I)I"(thread: JVMThread, args: [java_security_MessageDigest, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/security/MessageDigest/access$202(Ljava/security/MessageDigest;I)I"(thread: JVMThread, args: [java_security_MessageDigest, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Ljava/security/MessageDigest;)I"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/security/MessageDigest/access$200(Ljava/security/MessageDigest;)I"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/security/MessageDigest/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_MessageDigestSpi extends java_lang_Object {
    public "java/security/MessageDigestSpi/tempArray": JVMArray<number>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigestSpi/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineGetDigestLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/MessageDigestSpi/engineGetDigestLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "engineUpdate(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigestSpi/engineUpdate(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineUpdate([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigestSpi/engineUpdate([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineUpdate(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigestSpi/engineUpdate(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineDigest()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/security/MessageDigestSpi/engineDigest()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "engineDigest([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/MessageDigestSpi/engineDigest([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "engineReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/MessageDigestSpi/engineReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/security/MessageDigestSpi/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class sun_security_util_ManifestDigester$Entry extends java_lang_Object {
    public "sun/security/util/ManifestDigester$Entry/offset": number;
    public "sun/security/util/ManifestDigester$Entry/length": number;
    public "sun/security/util/ManifestDigester$Entry/lengthWithBlankLine": number;
    public "sun/security/util/ManifestDigester$Entry/rawBytes": JVMArray<number>;
    public "sun/security/util/ManifestDigester$Entry/oldStyle": number;
    public "<init>(III[B)V"(thread: JVMThread, args: [number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/ManifestDigester$Entry/<init>(III[B)V"(thread: JVMThread, args: [number, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "digest(Ljava/security/MessageDigest;)[B"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/security/util/ManifestDigester$Entry/digest(Ljava/security/MessageDigest;)[B"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "doOldStyle(Ljava/security/MessageDigest;[BII)V"(thread: JVMThread, args: [java_security_MessageDigest, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/ManifestDigester$Entry/doOldStyle(Ljava/security/MessageDigest;[BII)V"(thread: JVMThread, args: [java_security_MessageDigest, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "digestWorkaround(Ljava/security/MessageDigest;)[B"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/security/util/ManifestDigester$Entry/digestWorkaround(Ljava/security/MessageDigest;)[B"(thread: JVMThread, args: [java_security_MessageDigest], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export class sun_security_util_ManifestDigester$Position extends java_lang_Object {
    public "sun/security/util/ManifestDigester$Position/endOfFirstLine": number;
    public "sun/security/util/ManifestDigester$Position/endOfSection": number;
    public "sun/security/util/ManifestDigester$Position/startOfNext": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/security/util/ManifestDigester$Position/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ByteArrayOutputStream extends java_io_OutputStream {
    public "java/io/ByteArrayOutputStream/buf": JVMArray<number>;
    public "java/io/ByteArrayOutputStream/count": number;
    public static "java/io/ByteArrayOutputStream/MAX_ARRAY_SIZE": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ByteArrayOutputStream/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ByteArrayOutputStream/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ByteArrayOutputStream/ensureCapacity(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ByteArrayOutputStream/grow(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ByteArrayOutputStream/hugeCapacity(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ByteArrayOutputStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ByteArrayOutputStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeTo(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ByteArrayOutputStream/writeTo(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ByteArrayOutputStream/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/io/ByteArrayOutputStream/toByteArray()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ByteArrayOutputStream/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ByteArrayOutputStream/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ByteArrayOutputStream/toString(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ByteArrayOutputStream/toString(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ByteArrayOutputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_jar_JarEntry extends java_util_zip_ZipEntry {
    public "java/util/jar/JarEntry/attr": java_util_jar_Attributes;
    public "java/util/jar/JarEntry/certs": JVMArray<java_security_cert_Certificate>;
    public "java/util/jar/JarEntry/signers": JVMArray<java_security_CodeSigner>;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarEntry/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/zip/ZipEntry;)V"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarEntry/<init>(Ljava/util/zip/ZipEntry;)V"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/jar/JarEntry;)V"(thread: JVMThread, args: [java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/jar/JarEntry/<init>(Ljava/util/jar/JarEntry;)V"(thread: JVMThread, args: [java_util_jar_JarEntry], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAttributes()Ljava/util/jar/Attributes;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Attributes) => void): void;
    public "java/util/jar/JarEntry/getAttributes()Ljava/util/jar/Attributes;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Attributes) => void): void;
    public "getCertificates()[Ljava/security/cert/Certificate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "java/util/jar/JarEntry/getCertificates()[Ljava/security/cert/Certificate;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_cert_Certificate>) => void): void;
    public "getCodeSigners()[Ljava/security/CodeSigner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
    public "java/util/jar/JarEntry/getCodeSigners()[Ljava/security/CodeSigner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_CodeSigner>) => void): void;
  }
  export class java_util_logging_FileHandler extends java_util_logging_StreamHandler {
    public "java/util/logging/FileHandler/meter": java_util_logging_FileHandler$MeteredStream;
    public "java/util/logging/FileHandler/append": number;
    public "java/util/logging/FileHandler/limit": number;
    public "java/util/logging/FileHandler/count": number;
    public "java/util/logging/FileHandler/pattern": java_lang_String;
    public "java/util/logging/FileHandler/lockFileName": java_lang_String;
    public "java/util/logging/FileHandler/lockFileChannel": java_nio_channels_FileChannel;
    public "java/util/logging/FileHandler/files": JVMArray<java_io_File>;
    public static "java/util/logging/FileHandler/MAX_LOCKS": number;
    public static "java/util/logging/FileHandler/locks": java_util_Set;
    public "open(Ljava/io/File;Z)V"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/open(Ljava/io/File;Z)V"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "configure()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/configure()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;IIZ)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/<init>(Ljava/lang/String;IIZ)V"(thread: JVMThread, args: [java_lang_String, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isParentWritable(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/FileHandler/isParentWritable(Ljava/nio/file/Path;)Z"(thread: JVMThread, args: [java_nio_file_Path], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "openFiles()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/openFiles()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "generate(Ljava/lang/String;II)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "java/util/logging/FileHandler/generate(Ljava/lang/String;II)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "rotate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/rotate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "publish(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/publish(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isSetUID()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/logging/FileHandler/isSetUID()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Ljava/util/logging/FileHandler;)V"(thread: JVMThread, args: [java_util_logging_FileHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/FileHandler/access$100(Ljava/util/logging/FileHandler;)V"(thread: JVMThread, args: [java_util_logging_FileHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/FileHandler/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_LogRecord extends java_lang_Object implements java_io_Serializable {
    public static "java/util/logging/LogRecord/globalSequenceNumber": java_util_concurrent_atomic_AtomicLong;
    public static "java/util/logging/LogRecord/MIN_SEQUENTIAL_THREAD_ID": number;
    public static "java/util/logging/LogRecord/nextThreadId": java_util_concurrent_atomic_AtomicInteger;
    public static "java/util/logging/LogRecord/threadIds": java_lang_ThreadLocal;
    public "java/util/logging/LogRecord/level": java_util_logging_Level;
    public "java/util/logging/LogRecord/sequenceNumber": Long;
    public "java/util/logging/LogRecord/sourceClassName": java_lang_String;
    public "java/util/logging/LogRecord/sourceMethodName": java_lang_String;
    public "java/util/logging/LogRecord/message": java_lang_String;
    public "java/util/logging/LogRecord/threadID": number;
    public "java/util/logging/LogRecord/millis": Long;
    public "java/util/logging/LogRecord/thrown": java_lang_Throwable;
    public "java/util/logging/LogRecord/loggerName": java_lang_String;
    public "java/util/logging/LogRecord/resourceBundleName": java_lang_String;
    public "java/util/logging/LogRecord/needToInferCaller": number;
    public "java/util/logging/LogRecord/parameters": JVMArray<java_lang_Object>;
    public "java/util/logging/LogRecord/resourceBundle": java_util_ResourceBundle;
    public static "java/util/logging/LogRecord/serialVersionUID": Long;
    public "defaultThreadID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/LogRecord/defaultThreadID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/util/logging/Level;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/<init>(Ljava/util/logging/Level;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLoggerName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/LogRecord/getLoggerName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setLoggerName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setLoggerName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getResourceBundle()Ljava/util/ResourceBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "java/util/logging/LogRecord/getResourceBundle()Ljava/util/ResourceBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "setResourceBundle(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setResourceBundle(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "getResourceBundleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/LogRecord/getResourceBundleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setResourceBundleName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setResourceBundleName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLevel()Ljava/util/logging/Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "java/util/logging/LogRecord/getLevel()Ljava/util/logging/Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "setLevel(Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setLevel(Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSequenceNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/logging/LogRecord/getSequenceNumber()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setSequenceNumber(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setSequenceNumber(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSourceClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/LogRecord/getSourceClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setSourceClassName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setSourceClassName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSourceMethodName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/LogRecord/getSourceMethodName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setSourceMethodName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setSourceMethodName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/LogRecord/getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setMessage(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setMessage(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getParameters()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/logging/LogRecord/getParameters()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "setParameters([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setParameters([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getThreadID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/LogRecord/getThreadID()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setThreadID(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setThreadID(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMillis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/logging/LogRecord/getMillis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setMillis(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setMillis(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getThrown()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/getThrown()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "setThrown(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/setThrown(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "inferCaller()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogRecord/inferCaller()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isLoggerImplFrame(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/LogRecord/isLoggerImplFrame(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/LogRecord/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_Level extends java_lang_Object implements java_io_Serializable {
    public static "java/util/logging/Level/defaultBundle": java_lang_String;
    public "java/util/logging/Level/name": java_lang_String;
    public "java/util/logging/Level/value": number;
    public "java/util/logging/Level/resourceBundleName": java_lang_String;
    public "java/util/logging/Level/localizedLevelName": java_lang_String;
    public "java/util/logging/Level/cachedLocale": java_util_Locale;
    public static "java/util/logging/Level/OFF": java_util_logging_Level;
    public static "java/util/logging/Level/SEVERE": java_util_logging_Level;
    public static "java/util/logging/Level/WARNING": java_util_logging_Level;
    public static "java/util/logging/Level/INFO": java_util_logging_Level;
    public static "java/util/logging/Level/CONFIG": java_util_logging_Level;
    public static "java/util/logging/Level/FINE": java_util_logging_Level;
    public static "java/util/logging/Level/FINER": java_util_logging_Level;
    public static "java/util/logging/Level/FINEST": java_util_logging_Level;
    public static "java/util/logging/Level/ALL": java_util_logging_Level;
    public static "java/util/logging/Level/serialVersionUID": Long;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Level/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;ILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Level/<init>(Ljava/lang/String;ILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;ILjava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Level/<init>(Ljava/lang/String;ILjava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getResourceBundleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Level/getResourceBundleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Level/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getLocalizedName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Level/getLocalizedName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getLevelName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Level/getLevelName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "computeLocalizedLevelName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Level/computeLocalizedLevelName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCachedLocalizedLevelName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Level/getCachedLocalizedLevelName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getLocalizedLevelName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Level/getLocalizedLevelName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "findLevel(Ljava/lang/String;)Ljava/util/logging/Level;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public static "java/util/logging/Level/findLevel(Ljava/lang/String;)Ljava/util/logging/Level;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Level/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/Level/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/logging/Level/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "parse(Ljava/lang/String;)Ljava/util/logging/Level;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public static "java/util/logging/Level/parse(Ljava/lang/String;)Ljava/util/logging/Level;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/Level/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/Level/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/util/logging/Level;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/logging/Level/access$000(Ljava/util/logging/Level;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$100(Ljava/util/logging/Level;)I"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/logging/Level/access$100(Ljava/util/logging/Level;)I"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Ljava/util/logging/Level;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/logging/Level/access$200(Ljava/util/logging/Level;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "<init>(Ljava/lang/String;ILjava/lang/String;ZLjava/util/logging/Level$1;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, number, java_util_logging_Level$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Level/<init>(Ljava/lang/String;ILjava/lang/String;ZLjava/util/logging/Level$1;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_String, number, java_util_logging_Level$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/Level/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_Level$1 extends java_lang_Object {
  }
  export class java_util_logging_FileHandler$MeteredStream extends java_io_OutputStream {
    public "java/util/logging/FileHandler$MeteredStream/out": java_io_OutputStream;
    public "java/util/logging/FileHandler$MeteredStream/written": number;
    public "java/util/logging/FileHandler$MeteredStream/this$0": java_util_logging_FileHandler;
    public "<init>(Ljava/util/logging/FileHandler;Ljava/io/OutputStream;I)V"(thread: JVMThread, args: [java_util_logging_FileHandler, java_io_OutputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler$MeteredStream/<init>(Ljava/util/logging/FileHandler;Ljava/io/OutputStream;I)V"(thread: JVMThread, args: [java_util_logging_FileHandler, java_io_OutputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler$MeteredStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler$MeteredStream/write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler$MeteredStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler$MeteredStream/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/FileHandler$MeteredStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_StreamHandler extends java_util_logging_Handler {
    public "java/util/logging/StreamHandler/output": java_io_OutputStream;
    public "java/util/logging/StreamHandler/doneHeader": number;
    public "java/util/logging/StreamHandler/writer": java_io_Writer;
    public "configure()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/StreamHandler/configure()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/StreamHandler/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;Ljava/util/logging/Formatter;)V"(thread: JVMThread, args: [java_io_OutputStream, java_util_logging_Formatter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/StreamHandler/<init>(Ljava/io/OutputStream;Ljava/util/logging/Formatter;)V"(thread: JVMThread, args: [java_io_OutputStream, java_util_logging_Formatter], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOutputStream(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/StreamHandler/setOutputStream(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "setEncoding(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/StreamHandler/setEncoding(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "publish(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/StreamHandler/publish(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "isLoggable(Ljava/util/logging/LogRecord;)Z"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/StreamHandler/isLoggable(Ljava/util/logging/LogRecord;)Z"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/StreamHandler/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flushAndClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/StreamHandler/flushAndClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/StreamHandler/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_Formatter extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Formatter/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "format(Ljava/util/logging/LogRecord;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Formatter/format(Ljava/util/logging/LogRecord;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getHead(Ljava/util/logging/Handler;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_Handler], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Formatter/getHead(Ljava/util/logging/Handler;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_Handler], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTail(Ljava/util/logging/Handler;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_Handler], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Formatter/getTail(Ljava/util/logging/Handler;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_Handler], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "formatMessage(Ljava/util/logging/LogRecord;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Formatter/formatMessage(Ljava/util/logging/LogRecord;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_util_logging_Handler extends java_lang_Object {
    public static "java/util/logging/Handler/offValue": number;
    public "java/util/logging/Handler/manager": java_util_logging_LogManager;
    public "java/util/logging/Handler/filter": java_util_logging_Filter;
    public "java/util/logging/Handler/formatter": java_util_logging_Formatter;
    public "java/util/logging/Handler/logLevel": java_util_logging_Level;
    public "java/util/logging/Handler/errorManager": java_util_logging_ErrorManager;
    public "java/util/logging/Handler/encoding": java_lang_String;
    public "java/util/logging/Handler/sealed": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "publish(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/publish(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFormatter(Ljava/util/logging/Formatter;)V"(thread: JVMThread, args: [java_util_logging_Formatter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/setFormatter(Ljava/util/logging/Formatter;)V"(thread: JVMThread, args: [java_util_logging_Formatter], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFormatter()Ljava/util/logging/Formatter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Formatter) => void): void;
    public "java/util/logging/Handler/getFormatter()Ljava/util/logging/Formatter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Formatter) => void): void;
    public "setEncoding(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/setEncoding(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEncoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Handler/getEncoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setFilter(Ljava/util/logging/Filter;)V"(thread: JVMThread, args: [java_util_logging_Filter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/setFilter(Ljava/util/logging/Filter;)V"(thread: JVMThread, args: [java_util_logging_Filter], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFilter()Ljava/util/logging/Filter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Filter) => void): void;
    public "java/util/logging/Handler/getFilter()Ljava/util/logging/Filter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Filter) => void): void;
    public "setErrorManager(Ljava/util/logging/ErrorManager;)V"(thread: JVMThread, args: [java_util_logging_ErrorManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/setErrorManager(Ljava/util/logging/ErrorManager;)V"(thread: JVMThread, args: [java_util_logging_ErrorManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "getErrorManager()Ljava/util/logging/ErrorManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_ErrorManager) => void): void;
    public "java/util/logging/Handler/getErrorManager()Ljava/util/logging/ErrorManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_ErrorManager) => void): void;
    public "reportError(Ljava/lang/String;Ljava/lang/Exception;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_Exception, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/reportError(Ljava/lang/String;Ljava/lang/Exception;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_Exception, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLevel(Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/setLevel(Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLevel()Ljava/util/logging/Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "java/util/logging/Handler/getLevel()Ljava/util/logging/Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "isLoggable(Ljava/util/logging/LogRecord;)Z"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/Handler/isLoggable(Ljava/util/logging/LogRecord;)Z"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkPermission()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Handler/checkPermission()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/Handler/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_ErrorManager extends java_lang_Object {
    public "java/util/logging/ErrorManager/reported": number;
    public static "java/util/logging/ErrorManager/GENERIC_FAILURE": number;
    public static "java/util/logging/ErrorManager/WRITE_FAILURE": number;
    public static "java/util/logging/ErrorManager/FLUSH_FAILURE": number;
    public static "java/util/logging/ErrorManager/CLOSE_FAILURE": number;
    public static "java/util/logging/ErrorManager/OPEN_FAILURE": number;
    public static "java/util/logging/ErrorManager/FORMAT_FAILURE": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/ErrorManager/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "error(Ljava/lang/String;Ljava/lang/Exception;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_Exception, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/ErrorManager/error(Ljava/lang/String;Ljava/lang/Exception;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_Exception, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_logging_Filter extends java_lang_Object {
    "isLoggable(Ljava/util/logging/LogRecord;)Z"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_logging_LogManager extends java_lang_Object {
    public static "java/util/logging/LogManager/manager": java_util_logging_LogManager;
    public "java/util/logging/LogManager/props": java_util_Properties;
    public static "java/util/logging/LogManager/defaultLevel": java_util_logging_Level;
    public "java/util/logging/LogManager/listenerMap": java_util_Map;
    public "java/util/logging/LogManager/systemContext": java_util_logging_LogManager$LoggerContext;
    public "java/util/logging/LogManager/userContext": java_util_logging_LogManager$LoggerContext;
    public "java/util/logging/LogManager/rootLogger": java_util_logging_Logger;
    public "java/util/logging/LogManager/readPrimordialConfiguration": number;
    public "java/util/logging/LogManager/initializedGlobalHandlers": number;
    public "java/util/logging/LogManager/deathImminent": number;
    public "java/util/logging/LogManager/initializedCalled": number;
    public "java/util/logging/LogManager/initializationDone": number;
    public "java/util/logging/LogManager/contextsMap": java_util_WeakHashMap;
    public "java/util/logging/LogManager/loggerRefQueue": java_lang_ref_ReferenceQueue;
    public static "java/util/logging/LogManager/MAX_ITERATIONS": number;
    public "java/util/logging/LogManager/controlPermission": java_security_Permission;
    public static "java/util/logging/LogManager/loggingMXBean": java_util_logging_LoggingMXBean;
    public static "java/util/logging/LogManager/LOGGING_MXBEAN_NAME": java_lang_String;
    public static "java/util/logging/LogManager/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Void;)V"(thread: JVMThread, args: [java_lang_Void], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/<init>(Ljava/lang/Void;)V"(thread: JVMThread, args: [java_lang_Void], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkSubclassPermissions()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "java/util/logging/LogManager/checkSubclassPermissions()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "ensureLogManagerInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/ensureLogManagerInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getLogManager()Ljava/util/logging/LogManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager) => void): void;
    public static "java/util/logging/LogManager/getLogManager()Ljava/util/logging/LogManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager) => void): void;
    public "readPrimordialConfiguration()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/readPrimordialConfiguration()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/addPropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/removePropertyChangeListener(Ljava/beans/PropertyChangeListener;)V"(thread: JVMThread, args: [java_beans_PropertyChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUserContext()Ljava/util/logging/LogManager$LoggerContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager$LoggerContext) => void): void;
    public "java/util/logging/LogManager/getUserContext()Ljava/util/logging/LogManager$LoggerContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager$LoggerContext) => void): void;
    public "getSystemContext()Ljava/util/logging/LogManager$LoggerContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager$LoggerContext) => void): void;
    public "java/util/logging/LogManager/getSystemContext()Ljava/util/logging/LogManager$LoggerContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager$LoggerContext) => void): void;
    public "contexts()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/logging/LogManager/contexts()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "demandLogger(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Class;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "java/util/logging/LogManager/demandLogger(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Class;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "demandSystemLogger(Ljava/lang/String;Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "java/util/logging/LogManager/demandSystemLogger(Ljava/lang/String;Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "loadLoggerHandlers(Ljava/util/logging/Logger;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/loadLoggerHandlers(Ljava/util/logging/Logger;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "drainLoggerRefQueueBounded()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/drainLoggerRefQueueBounded()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addLogger(Ljava/util/logging/Logger;)Z"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/LogManager/addLogger(Ljava/util/logging/Logger;)Z"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "doSetLevel(Ljava/util/logging/Logger;Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/LogManager/doSetLevel(Ljava/util/logging/Logger;Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public static "doSetParent(Ljava/util/logging/Logger;Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/LogManager/doSetParent(Ljava/util/logging/Logger;Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "java/util/logging/LogManager/getLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "getLoggerNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/logging/LogManager/getLoggerNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "readConfiguration()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/readConfiguration()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resetLogger(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/resetLogger(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "parseClassNames(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/logging/LogManager/parseClassNames(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "readConfiguration(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/readConfiguration(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/LogManager/getProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getStringProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/LogManager/getStringProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getIntProperty(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/LogManager/getIntProperty(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBooleanProperty(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/LogManager/getBooleanProperty(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLevelProperty(Ljava/lang/String;Ljava/util/logging/Level;)Ljava/util/logging/Level;"(thread: JVMThread, args: [java_lang_String, java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "java/util/logging/LogManager/getLevelProperty(Ljava/lang/String;Ljava/util/logging/Level;)Ljava/util/logging/Level;"(thread: JVMThread, args: [java_lang_String, java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "getFilterProperty(Ljava/lang/String;Ljava/util/logging/Filter;)Ljava/util/logging/Filter;"(thread: JVMThread, args: [java_lang_String, java_util_logging_Filter], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Filter) => void): void;
    public "java/util/logging/LogManager/getFilterProperty(Ljava/lang/String;Ljava/util/logging/Filter;)Ljava/util/logging/Filter;"(thread: JVMThread, args: [java_lang_String, java_util_logging_Filter], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Filter) => void): void;
    public "getFormatterProperty(Ljava/lang/String;Ljava/util/logging/Formatter;)Ljava/util/logging/Formatter;"(thread: JVMThread, args: [java_lang_String, java_util_logging_Formatter], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Formatter) => void): void;
    public "java/util/logging/LogManager/getFormatterProperty(Ljava/lang/String;Ljava/util/logging/Formatter;)Ljava/util/logging/Formatter;"(thread: JVMThread, args: [java_lang_String, java_util_logging_Formatter], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Formatter) => void): void;
    public "initializeGlobalHandlers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/initializeGlobalHandlers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkPermission()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/checkPermission()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/checkAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLevelsOnExistingLoggers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager/setLevelsOnExistingLoggers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getLoggingMXBean()Ljava/util/logging/LoggingMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LoggingMXBean) => void): void;
    public static "java/util/logging/LogManager/getLoggingMXBean()Ljava/util/logging/LoggingMXBean;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LoggingMXBean) => void): void;
    public static "access$100()Ljava/util/logging/LogManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager) => void): void;
    public static "java/util/logging/LogManager/access$100()Ljava/util/logging/LogManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager) => void): void;
    public static "access$202(Ljava/util/logging/LogManager;Z)Z"(thread: JVMThread, args: [java_util_logging_LogManager, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/logging/LogManager/access$202(Ljava/util/logging/LogManager;Z)Z"(thread: JVMThread, args: [java_util_logging_LogManager, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$302(Ljava/util/logging/LogManager;Z)Z"(thread: JVMThread, args: [java_util_logging_LogManager, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/logging/LogManager/access$302(Ljava/util/logging/LogManager;Z)Z"(thread: JVMThread, args: [java_util_logging_LogManager, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500(Ljava/util/logging/LogManager;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "java/util/logging/LogManager/access$500(Ljava/util/logging/LogManager;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "access$600(Ljava/util/logging/LogManager;)Z"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/logging/LogManager/access$600(Ljava/util/logging/LogManager;)Z"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$700(Ljava/util/logging/LogManager;)Z"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/logging/LogManager/access$700(Ljava/util/logging/LogManager;)Z"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$800(Ljava/util/logging/LogManager;)V"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/LogManager/access$800(Ljava/util/logging/LogManager;)V"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$502(Ljava/util/logging/LogManager;Ljava/util/logging/Logger;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_util_logging_LogManager, java_util_logging_Logger], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "java/util/logging/LogManager/access$502(Ljava/util/logging/LogManager;Ljava/util/logging/Logger;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_util_logging_LogManager, java_util_logging_Logger], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "access$1000()Ljava/util/logging/Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public static "java/util/logging/LogManager/access$1000()Ljava/util/logging/Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public static "access$1100(Ljava/util/logging/Logger;Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/LogManager/access$1100(Ljava/util/logging/Logger;Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$1200(Ljava/util/logging/Logger;Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/LogManager/access$1200(Ljava/util/logging/Logger;Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$1300(Ljava/util/logging/LogManager;Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_LogManager, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/util/logging/LogManager/access$1300(Ljava/util/logging/LogManager;Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_util_logging_LogManager, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "access$1400(Ljava/util/logging/LogManager;)Ljava/lang/ref/ReferenceQueue;"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_ReferenceQueue) => void): void;
    public static "java/util/logging/LogManager/access$1400(Ljava/util/logging/LogManager;)Ljava/lang/ref/ReferenceQueue;"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable, rv?: java_lang_ref_ReferenceQueue) => void): void;
    public static "access$1500(Ljava/util/logging/LogManager;)V"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/LogManager/access$1500(Ljava/util/logging/LogManager;)V"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/LogManager/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_logging_LoggingMXBean extends java_lang_Object {
    "getLoggerNames()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    "getLoggerLevel(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "setLoggerLevel(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    "getParentLoggerName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_util_logging_Logger extends java_lang_Object {
    public static "java/util/logging/Logger/emptyHandlers": JVMArray<java_util_logging_Handler>;
    public static "java/util/logging/Logger/offValue": number;
    public static "java/util/logging/Logger/SYSTEM_LOGGER_RB_NAME": java_lang_String;
    public static "java/util/logging/Logger/SYSTEM_BUNDLE": java_util_logging_Logger$LoggerBundle;
    public static "java/util/logging/Logger/NO_RESOURCE_BUNDLE": java_util_logging_Logger$LoggerBundle;
    public "java/util/logging/Logger/manager": java_util_logging_LogManager;
    public "java/util/logging/Logger/name": java_lang_String;
    public "java/util/logging/Logger/handlers": java_util_concurrent_CopyOnWriteArrayList;
    public "java/util/logging/Logger/loggerBundle": java_util_logging_Logger$LoggerBundle;
    public "java/util/logging/Logger/useParentHandlers": number;
    public "java/util/logging/Logger/filter": java_util_logging_Filter;
    public "java/util/logging/Logger/anonymous": number;
    public "java/util/logging/Logger/catalog": java_util_ResourceBundle;
    public "java/util/logging/Logger/catalogName": java_lang_String;
    public "java/util/logging/Logger/catalogLocale": java_util_Locale;
    public static "java/util/logging/Logger/treeLock": java_lang_Object;
    public "java/util/logging/Logger/parent": java_util_logging_Logger;
    public "java/util/logging/Logger/kids": java_util_ArrayList;
    public "java/util/logging/Logger/levelObject": java_util_logging_Level;
    public "java/util/logging/Logger/levelValue": number;
    public "java/util/logging/Logger/callersClassLoaderRef": java_lang_ref_WeakReference;
    public "java/util/logging/Logger/isSystemLogger": number;
    public static "java/util/logging/Logger/GLOBAL_LOGGER_NAME": java_lang_String;
    public static "java/util/logging/Logger/global": java_util_logging_Logger;
    public static "java/util/logging/Logger/$assertionsDisabled": number;
    public static "getGlobal()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "java/util/logging/Logger/getGlobal()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Class;Ljava/util/logging/LogManager;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Class, java_util_logging_LogManager, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Class;Ljava/util/logging/LogManager;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Class, java_util_logging_LogManager, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setCallersClassLoaderRef(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/setCallersClassLoaderRef(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCallersClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/util/logging/Logger/getCallersClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLogManager(Ljava/util/logging/LogManager;)V"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/setLogManager(Ljava/util/logging/LogManager;)V"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkPermission()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/checkPermission()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "demandLogger(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Class;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "java/util/logging/Logger/demandLogger(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Class;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "getLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "java/util/logging/Logger/getLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "getLogger(Ljava/lang/String;Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "java/util/logging/Logger/getLogger(Ljava/lang/String;Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "getPlatformLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "java/util/logging/Logger/getPlatformLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "getAnonymousLogger()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "java/util/logging/Logger/getAnonymousLogger()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "getAnonymousLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public static "java/util/logging/Logger/getAnonymousLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "getResourceBundle()Ljava/util/ResourceBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "java/util/logging/Logger/getResourceBundle()Ljava/util/ResourceBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "getResourceBundleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Logger/getResourceBundleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setFilter(Ljava/util/logging/Filter;)V"(thread: JVMThread, args: [java_util_logging_Filter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/setFilter(Ljava/util/logging/Filter;)V"(thread: JVMThread, args: [java_util_logging_Filter], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFilter()Ljava/util/logging/Filter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Filter) => void): void;
    public "java/util/logging/Logger/getFilter()Ljava/util/logging/Filter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Filter) => void): void;
    public "log(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/log(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLog(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/doLog(Ljava/util/logging/LogRecord;)V"(thread: JVMThread, args: [java_util_logging_LogRecord], cb?: (e?: java_lang_Throwable) => void): void;
    public "log(Ljava/util/logging/Level;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/log(Ljava/util/logging/Level;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "log(Ljava/util/logging/Level;Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_logging_Level, java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/log(Ljava/util/logging/Level;Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_logging_Level, java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "log(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/log(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "log(Ljava/util/logging/Level;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/log(Ljava/util/logging/Level;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "log(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/log(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "log(Ljava/util/logging/Level;Ljava/lang/Throwable;Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_Throwable, java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/log(Ljava/util/logging/Level;Ljava/lang/Throwable;Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_Throwable, java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Throwable;Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_Throwable, java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logp(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Throwable;Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_Throwable, java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLog(Ljava/util/logging/LogRecord;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_LogRecord, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/doLog(Ljava/util/logging/LogRecord;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_LogRecord, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "doLog(Ljava/util/logging/LogRecord;Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_logging_LogRecord, java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/doLog(Ljava/util/logging/LogRecord;Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_logging_LogRecord, java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/util/ResourceBundle;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_util_ResourceBundle, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/util/ResourceBundle;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_util_ResourceBundle, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/util/ResourceBundle;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_util_ResourceBundle, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/logrb(Ljava/util/logging/Level;Ljava/lang/String;Ljava/lang/String;Ljava/util/ResourceBundle;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_util_logging_Level, java_lang_String, java_lang_String, java_util_ResourceBundle, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "entering(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/entering(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "entering(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/entering(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "entering(Ljava/lang/String;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/entering(Ljava/lang/String;Ljava/lang/String;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "exiting(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/exiting(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "exiting(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/exiting(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "throwing(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/throwing(Ljava/lang/String;Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "severe(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/severe(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "warning(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/warning(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "info(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/info(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "config(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/config(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "fine(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/fine(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "finer(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/finer(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "finest(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/finest(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "severe(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/severe(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "warning(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/warning(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "info(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/info(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "config(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/config(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "fine(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/fine(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "finer(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/finer(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "finest(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/finest(Ljava/util/function/Supplier;)V"(thread: JVMThread, args: [java_util_function_Supplier], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLevel(Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/setLevel(Ljava/util/logging/Level;)V"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable) => void): void;
    public "isLevelInitialized()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/Logger/isLevelInitialized()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLevel()Ljava/util/logging/Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "java/util/logging/Logger/getLevel()Ljava/util/logging/Level;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Level) => void): void;
    public "isLoggable(Ljava/util/logging/Level;)Z"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/Logger/isLoggable(Ljava/util/logging/Level;)Z"(thread: JVMThread, args: [java_util_logging_Level], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/logging/Logger/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "addHandler(Ljava/util/logging/Handler;)V"(thread: JVMThread, args: [java_util_logging_Handler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/addHandler(Ljava/util/logging/Handler;)V"(thread: JVMThread, args: [java_util_logging_Handler], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeHandler(Ljava/util/logging/Handler;)V"(thread: JVMThread, args: [java_util_logging_Handler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/removeHandler(Ljava/util/logging/Handler;)V"(thread: JVMThread, args: [java_util_logging_Handler], cb?: (e?: java_lang_Throwable) => void): void;
    public "getHandlers()[Ljava/util/logging/Handler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_logging_Handler>) => void): void;
    public "java/util/logging/Logger/getHandlers()[Ljava/util/logging/Handler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_logging_Handler>) => void): void;
    public "accessCheckedHandlers()[Ljava/util/logging/Handler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_logging_Handler>) => void): void;
    public "java/util/logging/Logger/accessCheckedHandlers()[Ljava/util/logging/Handler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_logging_Handler>) => void): void;
    public "setUseParentHandlers(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/setUseParentHandlers(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUseParentHandlers()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/Logger/getUseParentHandlers()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "findSystemResourceBundle(Ljava/util/Locale;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public static "java/util/logging/Logger/findSystemResourceBundle(Ljava/util/Locale;)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "findResourceBundle(Ljava/lang/String;Z)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "java/util/logging/Logger/findResourceBundle(Ljava/lang/String;Z)Ljava/util/ResourceBundle;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "setupResourceInfo(Ljava/lang/String;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/setupResourceInfo(Ljava/lang/String;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "setResourceBundle(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/setResourceBundle(Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "getParent()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "java/util/logging/Logger/getParent()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "setParent(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/setParent(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "doSetParent(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/doSetParent(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeChildLogger(Ljava/util/logging/LogManager$LoggerWeakRef;)V"(thread: JVMThread, args: [java_util_logging_LogManager$LoggerWeakRef], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/removeChildLogger(Ljava/util/logging/LogManager$LoggerWeakRef;)V"(thread: JVMThread, args: [java_util_logging_LogManager$LoggerWeakRef], cb?: (e?: java_lang_Throwable) => void): void;
    public "updateEffectiveLevel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger/updateEffectiveLevel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEffectiveLoggerBundle()Ljava/util/logging/Logger$LoggerBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger$LoggerBundle) => void): void;
    public "java/util/logging/Logger/getEffectiveLoggerBundle()Ljava/util/logging/Logger$LoggerBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger$LoggerBundle) => void): void;
    public static "access$000()Ljava/util/logging/Logger$LoggerBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger$LoggerBundle) => void): void;
    public static "java/util/logging/Logger/access$000()Ljava/util/logging/Logger$LoggerBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger$LoggerBundle) => void): void;
    public static "access$100()Ljava/util/logging/Logger$LoggerBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger$LoggerBundle) => void): void;
    public static "java/util/logging/Logger/access$100()Ljava/util/logging/Logger$LoggerBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger$LoggerBundle) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/Logger/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_LogManager$LoggerWeakRef extends java_lang_ref_WeakReference {
    public "java/util/logging/LogManager$LoggerWeakRef/name": java_lang_String;
    public "java/util/logging/LogManager$LoggerWeakRef/node": java_util_logging_LogManager$LogNode;
    public "java/util/logging/LogManager$LoggerWeakRef/parentRef": java_lang_ref_WeakReference;
    public "java/util/logging/LogManager$LoggerWeakRef/disposed": number;
    public "java/util/logging/LogManager$LoggerWeakRef/this$0": java_util_logging_LogManager;
    public "<init>(Ljava/util/logging/LogManager;Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_LogManager, java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerWeakRef/<init>(Ljava/util/logging/LogManager;Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_LogManager, java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerWeakRef/dispose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setNode(Ljava/util/logging/LogManager$LogNode;)V"(thread: JVMThread, args: [java_util_logging_LogManager$LogNode], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerWeakRef/setNode(Ljava/util/logging/LogManager$LogNode;)V"(thread: JVMThread, args: [java_util_logging_LogManager$LogNode], cb?: (e?: java_lang_Throwable) => void): void;
    public "setParentRef(Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerWeakRef/setParentRef(Ljava/lang/ref/WeakReference;)V"(thread: JVMThread, args: [java_lang_ref_WeakReference], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_LogManager$LogNode extends java_lang_Object {
    public "java/util/logging/LogManager$LogNode/children": java_util_HashMap;
    public "java/util/logging/LogManager$LogNode/loggerRef": java_util_logging_LogManager$LoggerWeakRef;
    public "java/util/logging/LogManager$LogNode/parent": java_util_logging_LogManager$LogNode;
    public "java/util/logging/LogManager$LogNode/context": java_util_logging_LogManager$LoggerContext;
    public "<init>(Ljava/util/logging/LogManager$LogNode;Ljava/util/logging/LogManager$LoggerContext;)V"(thread: JVMThread, args: [java_util_logging_LogManager$LogNode, java_util_logging_LogManager$LoggerContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LogNode/<init>(Ljava/util/logging/LogManager$LogNode;Ljava/util/logging/LogManager$LoggerContext;)V"(thread: JVMThread, args: [java_util_logging_LogManager$LogNode, java_util_logging_LogManager$LoggerContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "walkAndSetParent(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LogNode/walkAndSetParent(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_concurrent_CopyOnWriteArrayList extends java_lang_Object implements java_util_List, java_util_RandomAccess, java_lang_Cloneable, java_io_Serializable {
    public static "java/util/concurrent/CopyOnWriteArrayList/serialVersionUID": Long;
    public "java/util/concurrent/CopyOnWriteArrayList/lock": java_util_concurrent_locks_ReentrantLock;
    public "java/util/concurrent/CopyOnWriteArrayList/array": JVMArray<java_lang_Object>;
    public static "java/util/concurrent/CopyOnWriteArrayList/UNSAFE": sun_misc_Unsafe;
    public static "java/util/concurrent/CopyOnWriteArrayList/lockOffset": Long;
    public "getArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/getArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "setArray([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/setArray([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/<init>(Ljava/util/Collection;)V"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/<init>([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "eq(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/CopyOnWriteArrayList/eq(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "indexOf(Ljava/lang/Object;[Ljava/lang/Object;II)I"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/CopyOnWriteArrayList/indexOf(Ljava/lang/Object;[Ljava/lang/Object;II)I"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "lastIndexOf(Ljava/lang/Object;[Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/concurrent/CopyOnWriteArrayList/lastIndexOf(Ljava/lang/Object;[Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/contains(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/indexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/indexOf(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/lastIndexOf(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/lastIndexOf(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/toArray()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/toArray([Ljava/lang/Object;)[Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "get([Ljava/lang/Object;I)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/get([Ljava/lang/Object;I)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/get(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/set(ILjava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/add(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/add(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/remove(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/remove(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "remove(Ljava/lang/Object;[Ljava/lang/Object;I)Z"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/remove(Ljava/lang/Object;[Ljava/lang/Object;I)Z"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/removeRange(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addIfAbsent(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/addIfAbsent(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addIfAbsent(Ljava/lang/Object;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/addIfAbsent(Ljava/lang/Object;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/containsAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/removeAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/retainAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAllAbsent(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/addAllAbsent(Ljava/util/Collection;)I"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/addAll(Ljava/util/Collection;)Z"(thread: JVMThread, args: [java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/addAll(ILjava/util/Collection;)Z"(thread: JVMThread, args: [number, java_util_Collection], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    public "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/iterator()Ljava/util/Iterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Iterator) => void): void;
    public "listIterator()Ljava/util/ListIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/listIterator()Ljava/util/ListIterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/listIterator(I)Ljava/util/ListIterator;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_ListIterator) => void): void;
    public "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    public "subList(II)Ljava/util/List;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/subList(II)Ljava/util/List;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "resetLock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/concurrent/CopyOnWriteArrayList/resetLock()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/concurrent/CopyOnWriteArrayList/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "parallelStream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    "replaceAll(Ljava/util/function/UnaryOperator;)V"(thread: JVMThread, args: [java_util_function_UnaryOperator], cb?: (e?: java_lang_Throwable) => void): void;
    "sort(Ljava/util/Comparator;)V"(thread: JVMThread, args: [java_util_Comparator], cb?: (e?: java_lang_Throwable) => void): void;
    "spliterator()Ljava/util/Spliterator;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator) => void): void;
    "removeIf(Ljava/util/function/Predicate;)Z"(thread: JVMThread, args: [java_util_function_Predicate], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "forEach(Ljava/util/function/Consumer;)V"(thread: JVMThread, args: [java_util_function_Consumer], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_Logger$LoggerBundle extends java_lang_Object {
    public "java/util/logging/Logger$LoggerBundle/resourceBundleName": java_lang_String;
    public "java/util/logging/Logger$LoggerBundle/userBundle": java_util_ResourceBundle;
    public "<init>(Ljava/lang/String;Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_lang_String, java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger$LoggerBundle/<init>(Ljava/lang/String;Ljava/util/ResourceBundle;)V"(thread: JVMThread, args: [java_lang_String, java_util_ResourceBundle], cb?: (e?: java_lang_Throwable) => void): void;
    public "isSystemBundle()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/Logger$LoggerBundle/isSystemBundle()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "get(Ljava/lang/String;Ljava/util/ResourceBundle;)Ljava/util/logging/Logger$LoggerBundle;"(thread: JVMThread, args: [java_lang_String, java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger$LoggerBundle) => void): void;
    public static "java/util/logging/Logger$LoggerBundle/get(Ljava/lang/String;Ljava/util/ResourceBundle;)Ljava/util/logging/Logger$LoggerBundle;"(thread: JVMThread, args: [java_lang_String, java_util_ResourceBundle], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger$LoggerBundle) => void): void;
    public "<init>(Ljava/lang/String;Ljava/util/ResourceBundle;Ljava/util/logging/Logger$1;)V"(thread: JVMThread, args: [java_lang_String, java_util_ResourceBundle, java_util_logging_Logger$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger$LoggerBundle/<init>(Ljava/lang/String;Ljava/util/ResourceBundle;Ljava/util/logging/Logger$1;)V"(thread: JVMThread, args: [java_lang_String, java_util_ResourceBundle, java_util_logging_Logger$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_Logger$1 extends java_lang_Object implements java_security_PrivilegedAction {
    public "java/util/logging/Logger$1/val$locale": java_util_Locale;
    public "<init>(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/Logger$1/<init>(Ljava/util/Locale;)V"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()Ljava/util/ResourceBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "java/util/logging/Logger$1/run()Ljava/util/ResourceBundle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_ResourceBundle) => void): void;
    public "run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/logging/Logger$1/run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_logging_LogManager$LoggerContext extends java_lang_Object {
    public "java/util/logging/LogManager$LoggerContext/namedLoggers": java_util_Hashtable;
    public "java/util/logging/LogManager$LoggerContext/root": java_util_logging_LogManager$LogNode;
    public static "java/util/logging/LogManager$LoggerContext/$assertionsDisabled": number;
    public "java/util/logging/LogManager$LoggerContext/this$0": java_util_logging_LogManager;
    public "<init>(Ljava/util/logging/LogManager;)V"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerContext/<init>(Ljava/util/logging/LogManager;)V"(thread: JVMThread, args: [java_util_logging_LogManager], cb?: (e?: java_lang_Throwable) => void): void;
    public "requiresDefaultLoggers()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/LogManager$LoggerContext/requiresDefaultLoggers()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOwner()Ljava/util/logging/LogManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager) => void): void;
    public "java/util/logging/LogManager$LoggerContext/getOwner()Ljava/util/logging/LogManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager) => void): void;
    public "getRootLogger()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "java/util/logging/LogManager$LoggerContext/getRootLogger()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "getGlobalLogger()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "java/util/logging/LogManager$LoggerContext/getGlobalLogger()Ljava/util/logging/Logger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "demandLogger(Ljava/lang/String;Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "java/util/logging/LogManager$LoggerContext/demandLogger(Ljava/lang/String;Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "ensureInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerContext/ensureInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "findLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "java/util/logging/LogManager$LoggerContext/findLogger(Ljava/lang/String;)Ljava/util/logging/Logger;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_Logger) => void): void;
    public "ensureAllDefaultLoggers(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerContext/ensureAllDefaultLoggers(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureDefaultLogger(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerContext/ensureDefaultLogger(Ljava/util/logging/Logger;)V"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable) => void): void;
    public "addLocalLogger(Ljava/util/logging/Logger;)Z"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/LogManager$LoggerContext/addLocalLogger(Ljava/util/logging/Logger;)Z"(thread: JVMThread, args: [java_util_logging_Logger], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addLocalLogger(Ljava/util/logging/Logger;Z)Z"(thread: JVMThread, args: [java_util_logging_Logger, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/logging/LogManager$LoggerContext/addLocalLogger(Ljava/util/logging/Logger;Z)Z"(thread: JVMThread, args: [java_util_logging_Logger, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeLoggerRef(Ljava/lang/String;Ljava/util/logging/LogManager$LoggerWeakRef;)V"(thread: JVMThread, args: [java_lang_String, java_util_logging_LogManager$LoggerWeakRef], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerContext/removeLoggerRef(Ljava/lang/String;Ljava/util/logging/LogManager$LoggerWeakRef;)V"(thread: JVMThread, args: [java_lang_String, java_util_logging_LogManager$LoggerWeakRef], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLoggerNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/logging/LogManager$LoggerContext/getLoggerNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "processParentHandlers(Ljava/util/logging/Logger;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerContext/processParentHandlers(Ljava/util/logging/Logger;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_logging_Logger, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNode(Ljava/lang/String;)Ljava/util/logging/LogManager$LogNode;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager$LogNode) => void): void;
    public "java/util/logging/LogManager$LoggerContext/getNode(Ljava/lang/String;)Ljava/util/logging/LogManager$LogNode;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager$LogNode) => void): void;
    public "<init>(Ljava/util/logging/LogManager;Ljava/util/logging/LogManager$1;)V"(thread: JVMThread, args: [java_util_logging_LogManager, java_util_logging_LogManager$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$LoggerContext/<init>(Ljava/util/logging/LogManager;Ljava/util/logging/LogManager$1;)V"(thread: JVMThread, args: [java_util_logging_LogManager, java_util_logging_LogManager$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/logging/LogManager$LoggerContext/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_logging_LogManager$1 extends java_lang_Object implements java_security_PrivilegedAction {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/logging/LogManager$1/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()Ljava/util/logging/LogManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager) => void): void;
    public "java/util/logging/LogManager$1/run()Ljava/util/logging/LogManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_logging_LogManager) => void): void;
    public "run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/logging/LogManager$1/run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_prefs_FileSystemPreferences extends java_util_prefs_AbstractPreferences {
    public static "java/util/prefs/FileSystemPreferences/SYNC_INTERVAL": number;
    public static "java/util/prefs/FileSystemPreferences/systemRootDir": java_io_File;
    public static "java/util/prefs/FileSystemPreferences/isSystemRootWritable": number;
    public static "java/util/prefs/FileSystemPreferences/userRootDir": java_io_File;
    public static "java/util/prefs/FileSystemPreferences/isUserRootWritable": number;
    public static "java/util/prefs/FileSystemPreferences/userRoot": java_util_prefs_Preferences;
    public static "java/util/prefs/FileSystemPreferences/systemRoot": java_util_prefs_Preferences;
    public static "java/util/prefs/FileSystemPreferences/USER_READ_WRITE": number;
    public static "java/util/prefs/FileSystemPreferences/USER_RW_ALL_READ": number;
    public static "java/util/prefs/FileSystemPreferences/USER_RWX_ALL_RX": number;
    public static "java/util/prefs/FileSystemPreferences/USER_RWX": number;
    public static "java/util/prefs/FileSystemPreferences/userLockFile": java_io_File;
    public static "java/util/prefs/FileSystemPreferences/systemLockFile": java_io_File;
    public static "java/util/prefs/FileSystemPreferences/userRootLockHandle": number;
    public static "java/util/prefs/FileSystemPreferences/systemRootLockHandle": number;
    public "java/util/prefs/FileSystemPreferences/dir": java_io_File;
    public "java/util/prefs/FileSystemPreferences/prefsFile": java_io_File;
    public "java/util/prefs/FileSystemPreferences/tmpFile": java_io_File;
    public static "java/util/prefs/FileSystemPreferences/userRootModFile": java_io_File;
    public static "java/util/prefs/FileSystemPreferences/isUserRootModified": number;
    public static "java/util/prefs/FileSystemPreferences/userRootModTime": Long;
    public static "java/util/prefs/FileSystemPreferences/systemRootModFile": java_io_File;
    public static "java/util/prefs/FileSystemPreferences/isSystemRootModified": number;
    public static "java/util/prefs/FileSystemPreferences/systemRootModTime": Long;
    public "java/util/prefs/FileSystemPreferences/prefsCache": java_util_Map;
    public "java/util/prefs/FileSystemPreferences/lastSyncTime": Long;
    public static "java/util/prefs/FileSystemPreferences/EAGAIN": number;
    public static "java/util/prefs/FileSystemPreferences/EACCES": number;
    public static "java/util/prefs/FileSystemPreferences/LOCK_HANDLE": number;
    public static "java/util/prefs/FileSystemPreferences/ERROR_CODE": number;
    public "java/util/prefs/FileSystemPreferences/changeLog": java_util_List;
    public "java/util/prefs/FileSystemPreferences/nodeCreate": java_util_prefs_FileSystemPreferences$NodeCreate;
    public static "java/util/prefs/FileSystemPreferences/syncTimer": java_util_Timer;
    public "java/util/prefs/FileSystemPreferences/isUserNode": number;
    public static "java/util/prefs/FileSystemPreferences/EMPTY_STRING_ARRAY": JVMArray<java_lang_String>;
    public static "java/util/prefs/FileSystemPreferences/INIT_SLEEP_TIME": number;
    public static "java/util/prefs/FileSystemPreferences/MAX_ATTEMPTS": number;
    public static "getLogger()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "java/util/prefs/FileSystemPreferences/getLogger()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "getUserRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "java/util/prefs/FileSystemPreferences/getUserRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "setupUserRoot()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/FileSystemPreferences/setupUserRoot()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getSystemRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "java/util/prefs/FileSystemPreferences/getSystemRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "setupSystemRoot()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/FileSystemPreferences/setupSystemRoot()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "replayChanges()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/replayChanges()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "syncWorld()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/FileSystemPreferences/syncWorld()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/prefs/FileSystemPreferences;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/<init>(Ljava/util/prefs/FileSystemPreferences;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isUserNode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/FileSystemPreferences/isUserNode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putSpi(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/putSpi(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSpi(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/FileSystemPreferences/getSpi(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "removeSpi(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/removeSpi(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "initCacheIfNecessary()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/initCacheIfNecessary()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "loadCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/loadCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBackCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/writeBackCache()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "keysSpi()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/prefs/FileSystemPreferences/keysSpi()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "childrenNamesSpi()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/prefs/FileSystemPreferences/childrenNamesSpi()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "childSpi(Ljava/lang/String;)Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_AbstractPreferences) => void): void;
    public "java/util/prefs/FileSystemPreferences/childSpi(Ljava/lang/String;)Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_AbstractPreferences) => void): void;
    public "removeNode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/removeNode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNodeSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/removeNodeSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "syncSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/syncSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "syncSpiPrivileged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/syncSpiPrivileged()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flushSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/flushSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isDirChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/prefs/FileSystemPreferences/isDirChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "dirName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/prefs/FileSystemPreferences/dirName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "byteArray(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/util/prefs/FileSystemPreferences/byteArray(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "nodeName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/prefs/FileSystemPreferences/nodeName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "lockFile(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/FileSystemPreferences/lockFile(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkLockFile0ErrorCode(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/checkLockFile0ErrorCode(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "lockFile0(Ljava/lang/String;IZ)[I"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/util/prefs/FileSystemPreferences/lockFile0(Ljava/lang/String;IZ)[I"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "unlockFile0(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/prefs/FileSystemPreferences/unlockFile0(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "chmod(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/prefs/FileSystemPreferences/chmod(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "unlockFile()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences/unlockFile()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$002(Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$002(Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$000()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$000()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$100(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$100(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$200()Lsun/util/logging/PlatformLogger;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_util_logging_PlatformLogger) => void): void;
    public static "access$302(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$302(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$402(Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$402(Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$400()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$400()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$502(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$502(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$602(Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$602(Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$600()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$600()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$702(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$702(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$802(Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$802(Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$800()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$800()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$700()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$700()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$902(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$902(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1100(Ljava/util/prefs/FileSystemPreferences;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$1100(Ljava/util/prefs/FileSystemPreferences;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "access$1200()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$1200()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$1300()Ljava/util/Timer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Timer) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$1300()Ljava/util/Timer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Timer) => void): void;
    public static "access$1400(Ljava/util/prefs/FileSystemPreferences;)Ljava/io/File;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$1400(Ljava/util/prefs/FileSystemPreferences;)Ljava/io/File;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$1600(Ljava/util/prefs/FileSystemPreferences;)Ljava/io/File;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$1600(Ljava/util/prefs/FileSystemPreferences;)Ljava/io/File;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$1102(Ljava/util/prefs/FileSystemPreferences;Ljava/util/Map;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$1102(Ljava/util/prefs/FileSystemPreferences;Ljava/util/Map;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, java_util_Map], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "access$1702(Ljava/util/prefs/FileSystemPreferences;J)J"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$1702(Ljava/util/prefs/FileSystemPreferences;J)J"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1800(Ljava/util/prefs/FileSystemPreferences;)Ljava/io/File;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$1800(Ljava/util/prefs/FileSystemPreferences;)Ljava/io/File;"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "access$1900(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$1900(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$2000()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$2000()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "access$2102(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$2102(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$500()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$2202(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$2202(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$900()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$900()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$2300(Ljava/util/prefs/FileSystemPreferences;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/FileSystemPreferences/access$2300(Ljava/util/prefs/FileSystemPreferences;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/FileSystemPreferences/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_Timer extends java_lang_Object {
    public "java/util/Timer/queue": java_util_TaskQueue;
    public "java/util/Timer/thread": java_util_TimerThread;
    public "java/util/Timer/threadReaper": java_lang_Object;
    public static "java/util/Timer/nextSerialNumber": java_util_concurrent_atomic_AtomicInteger;
    public static "serialNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Timer/serialNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "schedule(Ljava/util/TimerTask;J)V"(thread: JVMThread, args: [java_util_TimerTask, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/schedule(Ljava/util/TimerTask;J)V"(thread: JVMThread, args: [java_util_TimerTask, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "schedule(Ljava/util/TimerTask;Ljava/util/Date;)V"(thread: JVMThread, args: [java_util_TimerTask, java_util_Date], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/schedule(Ljava/util/TimerTask;Ljava/util/Date;)V"(thread: JVMThread, args: [java_util_TimerTask, java_util_Date], cb?: (e?: java_lang_Throwable) => void): void;
    public "schedule(Ljava/util/TimerTask;JJ)V"(thread: JVMThread, args: [java_util_TimerTask, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/schedule(Ljava/util/TimerTask;JJ)V"(thread: JVMThread, args: [java_util_TimerTask, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "schedule(Ljava/util/TimerTask;Ljava/util/Date;J)V"(thread: JVMThread, args: [java_util_TimerTask, java_util_Date, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/schedule(Ljava/util/TimerTask;Ljava/util/Date;J)V"(thread: JVMThread, args: [java_util_TimerTask, java_util_Date, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "scheduleAtFixedRate(Ljava/util/TimerTask;JJ)V"(thread: JVMThread, args: [java_util_TimerTask, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/scheduleAtFixedRate(Ljava/util/TimerTask;JJ)V"(thread: JVMThread, args: [java_util_TimerTask, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "scheduleAtFixedRate(Ljava/util/TimerTask;Ljava/util/Date;J)V"(thread: JVMThread, args: [java_util_TimerTask, java_util_Date, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/scheduleAtFixedRate(Ljava/util/TimerTask;Ljava/util/Date;J)V"(thread: JVMThread, args: [java_util_TimerTask, java_util_Date, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "sched(Ljava/util/TimerTask;JJ)V"(thread: JVMThread, args: [java_util_TimerTask, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/sched(Ljava/util/TimerTask;JJ)V"(thread: JVMThread, args: [java_util_TimerTask, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "cancel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Timer/cancel()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "purge()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Timer/purge()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/util/Timer;)Ljava/util/TaskQueue;"(thread: JVMThread, args: [java_util_Timer], cb?: (e?: java_lang_Throwable, rv?: java_util_TaskQueue) => void): void;
    public static "java/util/Timer/access$000(Ljava/util/Timer;)Ljava/util/TaskQueue;"(thread: JVMThread, args: [java_util_Timer], cb?: (e?: java_lang_Throwable, rv?: java_util_TaskQueue) => void): void;
    public static "access$100(Ljava/util/Timer;)Ljava/util/TimerThread;"(thread: JVMThread, args: [java_util_Timer], cb?: (e?: java_lang_Throwable, rv?: java_util_TimerThread) => void): void;
    public static "java/util/Timer/access$100(Ljava/util/Timer;)Ljava/util/TimerThread;"(thread: JVMThread, args: [java_util_Timer], cb?: (e?: java_lang_Throwable, rv?: java_util_TimerThread) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Timer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_TimerTask extends java_lang_Object implements java_lang_Runnable {
    public "java/util/TimerTask/lock": java_lang_Object;
    public "java/util/TimerTask/state": number;
    public static "java/util/TimerTask/VIRGIN": number;
    public static "java/util/TimerTask/SCHEDULED": number;
    public static "java/util/TimerTask/EXECUTED": number;
    public static "java/util/TimerTask/CANCELLED": number;
    public "java/util/TimerTask/nextExecutionTime": Long;
    public "java/util/TimerTask/period": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TimerTask/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TimerTask/run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "cancel()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimerTask/cancel()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "scheduledExecutionTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/TimerTask/scheduledExecutionTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
  }
  export class java_util_TimerThread extends java_lang_Thread {
    public "java/util/TimerThread/newTasksMayBeScheduled": number;
    public "java/util/TimerThread/queue": java_util_TaskQueue;
    public "<init>(Ljava/util/TaskQueue;)V"(thread: JVMThread, args: [java_util_TaskQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TimerThread/<init>(Ljava/util/TaskQueue;)V"(thread: JVMThread, args: [java_util_TaskQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TimerThread/run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mainLoop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TimerThread/mainLoop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_TaskQueue extends java_lang_Object {
    public "java/util/TaskQueue/queue": JVMArray<java_util_TimerTask>;
    public "java/util/TaskQueue/size": number;
    public static "java/util/TaskQueue/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TaskQueue/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TaskQueue/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/util/TimerTask;)V"(thread: JVMThread, args: [java_util_TimerTask], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TaskQueue/add(Ljava/util/TimerTask;)V"(thread: JVMThread, args: [java_util_TimerTask], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMin()Ljava/util/TimerTask;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimerTask) => void): void;
    public "java/util/TaskQueue/getMin()Ljava/util/TimerTask;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimerTask) => void): void;
    public "get(I)Ljava/util/TimerTask;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_TimerTask) => void): void;
    public "java/util/TaskQueue/get(I)Ljava/util/TimerTask;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_TimerTask) => void): void;
    public "removeMin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TaskQueue/removeMin()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "quickRemove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TaskQueue/quickRemove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "rescheduleMin(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TaskQueue/rescheduleMin(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TaskQueue/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TaskQueue/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "fixUp(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TaskQueue/fixUp(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "fixDown(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TaskQueue/fixDown(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "heapify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TaskQueue/heapify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/TaskQueue/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_prefs_FileSystemPreferences$NodeCreate extends java_util_prefs_FileSystemPreferences$Change {
    public "java/util/prefs/FileSystemPreferences$NodeCreate/this$0": java_util_prefs_FileSystemPreferences;
    public "<init>(Ljava/util/prefs/FileSystemPreferences;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences$NodeCreate/<init>(Ljava/util/prefs/FileSystemPreferences;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "replay()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences$NodeCreate/replay()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/prefs/FileSystemPreferences;Ljava/util/prefs/FileSystemPreferences$1;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, java_util_prefs_FileSystemPreferences$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences$NodeCreate/<init>(Ljava/util/prefs/FileSystemPreferences;Ljava/util/prefs/FileSystemPreferences$1;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, java_util_prefs_FileSystemPreferences$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_prefs_FileSystemPreferences$1 extends java_lang_Object implements java_security_PrivilegedAction {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences$1/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "java/util/prefs/FileSystemPreferences$1/run()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/prefs/FileSystemPreferences$1/run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_util_prefs_FileSystemPreferences$Change extends java_lang_Object {
    public "java/util/prefs/FileSystemPreferences$Change/this$0": java_util_prefs_FileSystemPreferences;
    public "<init>(Ljava/util/prefs/FileSystemPreferences;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences$Change/<init>(Ljava/util/prefs/FileSystemPreferences;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "replay()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences$Change/replay()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/prefs/FileSystemPreferences;Ljava/util/prefs/FileSystemPreferences$1;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, java_util_prefs_FileSystemPreferences$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/FileSystemPreferences$Change/<init>(Ljava/util/prefs/FileSystemPreferences;Ljava/util/prefs/FileSystemPreferences$1;)V"(thread: JVMThread, args: [java_util_prefs_FileSystemPreferences, java_util_prefs_FileSystemPreferences$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_prefs_Preferences extends java_lang_Object {
    public static "java/util/prefs/Preferences/factory": java_util_prefs_PreferencesFactory;
    public static "java/util/prefs/Preferences/MAX_KEY_LENGTH": number;
    public static "java/util/prefs/Preferences/MAX_VALUE_LENGTH": number;
    public static "java/util/prefs/Preferences/MAX_NAME_LENGTH": number;
    public static "java/util/prefs/Preferences/prefsPerm": java_security_Permission;
    public static "factory()Ljava/util/prefs/PreferencesFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_PreferencesFactory) => void): void;
    public static "java/util/prefs/Preferences/factory()Ljava/util/prefs/PreferencesFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_PreferencesFactory) => void): void;
    public static "factory1()Ljava/util/prefs/PreferencesFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_PreferencesFactory) => void): void;
    public static "java/util/prefs/Preferences/factory1()Ljava/util/prefs/PreferencesFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_PreferencesFactory) => void): void;
    public static "userNodeForPackage(Ljava/lang/Class;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "java/util/prefs/Preferences/userNodeForPackage(Ljava/lang/Class;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "systemNodeForPackage(Ljava/lang/Class;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "java/util/prefs/Preferences/systemNodeForPackage(Ljava/lang/Class;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "nodeName(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/prefs/Preferences/nodeName(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "userRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "java/util/prefs/Preferences/userRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "systemRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public static "java/util/prefs/Preferences/systemRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/put(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/Preferences/get(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "putInt(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/putInt(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/Preferences/getInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putLong(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/putLong(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLong(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/prefs/Preferences/getLong(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putBoolean(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/putBoolean(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBoolean(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/Preferences/getBoolean(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloat(Ljava/lang/String;F)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/putFloat(Ljava/lang/String;F)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFloat(Ljava/lang/String;F)F"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/Preferences/getFloat(Ljava/lang/String;F)F"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDouble(Ljava/lang/String;D)V"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/putDouble(Ljava/lang/String;D)V"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDouble(Ljava/lang/String;D)D"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/Preferences/getDouble(Ljava/lang/String;D)D"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putByteArray(Ljava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/putByteArray(Ljava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getByteArray(Ljava/lang/String;[B)[B"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/util/prefs/Preferences/getByteArray(Ljava/lang/String;[B)[B"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "keys()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/prefs/Preferences/keys()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "childrenNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/prefs/Preferences/childrenNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "parent()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "java/util/prefs/Preferences/parent()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "node(Ljava/lang/String;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "java/util/prefs/Preferences/node(Ljava/lang/String;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "nodeExists(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/Preferences/nodeExists(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeNode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/removeNode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/Preferences/name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "absolutePath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/Preferences/absolutePath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isUserNode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/Preferences/isUserNode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/Preferences/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addPreferenceChangeListener(Ljava/util/prefs/PreferenceChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_PreferenceChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/addPreferenceChangeListener(Ljava/util/prefs/PreferenceChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_PreferenceChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePreferenceChangeListener(Ljava/util/prefs/PreferenceChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_PreferenceChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/removePreferenceChangeListener(Ljava/util/prefs/PreferenceChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_PreferenceChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "addNodeChangeListener(Ljava/util/prefs/NodeChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/addNodeChangeListener(Ljava/util/prefs/NodeChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNodeChangeListener(Ljava/util/prefs/NodeChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/removeNodeChangeListener(Ljava/util/prefs/NodeChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "exportNode(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/exportNode(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "exportSubtree(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/Preferences/exportSubtree(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "importPreferences(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/Preferences/importPreferences(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000()Ljava/util/prefs/PreferencesFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_PreferencesFactory) => void): void;
    public static "java/util/prefs/Preferences/access$000()Ljava/util/prefs/PreferencesFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_PreferencesFactory) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/Preferences/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_prefs_NodeChangeListener extends java_lang_Object, java_util_EventListener {
    "childAdded(Ljava/util/prefs/NodeChangeEvent;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeEvent], cb?: (e?: java_lang_Throwable) => void): void;
    "childRemoved(Ljava/util/prefs/NodeChangeEvent;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_prefs_NodeChangeEvent extends java_util_EventObject {
    public "java/util/prefs/NodeChangeEvent/child": java_util_prefs_Preferences;
    public static "java/util/prefs/NodeChangeEvent/serialVersionUID": Long;
    public "<init>(Ljava/util/prefs/Preferences;Ljava/util/prefs/Preferences;)V"(thread: JVMThread, args: [java_util_prefs_Preferences, java_util_prefs_Preferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/NodeChangeEvent/<init>(Ljava/util/prefs/Preferences;Ljava/util/prefs/Preferences;)V"(thread: JVMThread, args: [java_util_prefs_Preferences, java_util_prefs_Preferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "getParent()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "java/util/prefs/NodeChangeEvent/getParent()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "getChild()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "java/util/prefs/NodeChangeEvent/getChild()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/NodeChangeEvent/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/NodeChangeEvent/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_prefs_PreferenceChangeListener extends java_lang_Object, java_util_EventListener {
    "preferenceChange(Ljava/util/prefs/PreferenceChangeEvent;)V"(thread: JVMThread, args: [java_util_prefs_PreferenceChangeEvent], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_prefs_PreferenceChangeEvent extends java_util_EventObject {
    public "java/util/prefs/PreferenceChangeEvent/key": java_lang_String;
    public "java/util/prefs/PreferenceChangeEvent/newValue": java_lang_String;
    public static "java/util/prefs/PreferenceChangeEvent/serialVersionUID": Long;
    public "<init>(Ljava/util/prefs/Preferences;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_prefs_Preferences, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/PreferenceChangeEvent/<init>(Ljava/util/prefs/Preferences;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_prefs_Preferences, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getNode()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "java/util/prefs/PreferenceChangeEvent/getNode()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "getKey()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/PreferenceChangeEvent/getKey()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getNewValue()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/PreferenceChangeEvent/getNewValue()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/PreferenceChangeEvent/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/PreferenceChangeEvent/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_prefs_PreferencesFactory extends java_lang_Object {
    "systemRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    "userRoot()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
  }
  export class java_util_prefs_AbstractPreferences extends java_util_prefs_Preferences {
    public "java/util/prefs/AbstractPreferences/name": java_lang_String;
    public "java/util/prefs/AbstractPreferences/absolutePath": java_lang_String;
    public "java/util/prefs/AbstractPreferences/parent": java_util_prefs_AbstractPreferences;
    public "java/util/prefs/AbstractPreferences/root": java_util_prefs_AbstractPreferences;
    public "java/util/prefs/AbstractPreferences/newNode": number;
    public "java/util/prefs/AbstractPreferences/kidCache": java_util_Map;
    public "java/util/prefs/AbstractPreferences/removed": number;
    public "java/util/prefs/AbstractPreferences/prefListeners": JVMArray<java_util_prefs_PreferenceChangeListener>;
    public "java/util/prefs/AbstractPreferences/nodeListeners": JVMArray<java_util_prefs_NodeChangeListener>;
    public "java/util/prefs/AbstractPreferences/lock": java_lang_Object;
    public static "java/util/prefs/AbstractPreferences/EMPTY_STRING_ARRAY": JVMArray<java_lang_String>;
    public static "java/util/prefs/AbstractPreferences/EMPTY_ABSTRACT_PREFS_ARRAY": JVMArray<java_util_prefs_AbstractPreferences>;
    public static "java/util/prefs/AbstractPreferences/eventQueue": java_util_List;
    public static "java/util/prefs/AbstractPreferences/eventDispatchThread": java_lang_Thread;
    public "<init>(Ljava/util/prefs/AbstractPreferences;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_prefs_AbstractPreferences, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/<init>(Ljava/util/prefs/AbstractPreferences;Ljava/lang/String;)V"(thread: JVMThread, args: [java_util_prefs_AbstractPreferences, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/put(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/AbstractPreferences/get(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/remove(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "putInt(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/putInt(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/AbstractPreferences/getInt(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putLong(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/putLong(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLong(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/prefs/AbstractPreferences/getLong(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putBoolean(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/putBoolean(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBoolean(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/AbstractPreferences/getBoolean(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloat(Ljava/lang/String;F)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/putFloat(Ljava/lang/String;F)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFloat(Ljava/lang/String;F)F"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/AbstractPreferences/getFloat(Ljava/lang/String;F)F"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDouble(Ljava/lang/String;D)V"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/putDouble(Ljava/lang/String;D)V"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDouble(Ljava/lang/String;D)D"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/AbstractPreferences/getDouble(Ljava/lang/String;D)D"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putByteArray(Ljava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/putByteArray(Ljava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getByteArray(Ljava/lang/String;[B)[B"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/util/prefs/AbstractPreferences/getByteArray(Ljava/lang/String;[B)[B"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "keys()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/prefs/AbstractPreferences/keys()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "childrenNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/prefs/AbstractPreferences/childrenNames()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "cachedChildren()[Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_prefs_AbstractPreferences>) => void): void;
    public "java/util/prefs/AbstractPreferences/cachedChildren()[Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_prefs_AbstractPreferences>) => void): void;
    public "parent()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "java/util/prefs/AbstractPreferences/parent()Ljava/util/prefs/Preferences;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "node(Ljava/lang/String;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "java/util/prefs/AbstractPreferences/node(Ljava/lang/String;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "node(Ljava/util/StringTokenizer;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_util_StringTokenizer], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "java/util/prefs/AbstractPreferences/node(Ljava/util/StringTokenizer;)Ljava/util/prefs/Preferences;"(thread: JVMThread, args: [java_util_StringTokenizer], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_Preferences) => void): void;
    public "nodeExists(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/AbstractPreferences/nodeExists(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nodeExists(Ljava/util/StringTokenizer;)Z"(thread: JVMThread, args: [java_util_StringTokenizer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/AbstractPreferences/nodeExists(Ljava/util/StringTokenizer;)Z"(thread: JVMThread, args: [java_util_StringTokenizer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "removeNode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/removeNode()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNode2()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/removeNode2()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/AbstractPreferences/name()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "absolutePath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/AbstractPreferences/absolutePath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isUserNode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/AbstractPreferences/isUserNode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addPreferenceChangeListener(Ljava/util/prefs/PreferenceChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_PreferenceChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/addPreferenceChangeListener(Ljava/util/prefs/PreferenceChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_PreferenceChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removePreferenceChangeListener(Ljava/util/prefs/PreferenceChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_PreferenceChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/removePreferenceChangeListener(Ljava/util/prefs/PreferenceChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_PreferenceChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "addNodeChangeListener(Ljava/util/prefs/NodeChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/addNodeChangeListener(Ljava/util/prefs/NodeChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNodeChangeListener(Ljava/util/prefs/NodeChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/removeNodeChangeListener(Ljava/util/prefs/NodeChangeListener;)V"(thread: JVMThread, args: [java_util_prefs_NodeChangeListener], cb?: (e?: java_lang_Throwable) => void): void;
    public "putSpi(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/putSpi(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSpi(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/AbstractPreferences/getSpi(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "removeSpi(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/removeSpi(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeNodeSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/removeNodeSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "keysSpi()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/prefs/AbstractPreferences/keysSpi()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "childrenNamesSpi()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/util/prefs/AbstractPreferences/childrenNamesSpi()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getChild(Ljava/lang/String;)Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_AbstractPreferences) => void): void;
    public "java/util/prefs/AbstractPreferences/getChild(Ljava/lang/String;)Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_AbstractPreferences) => void): void;
    public "childSpi(Ljava/lang/String;)Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_AbstractPreferences) => void): void;
    public "java/util/prefs/AbstractPreferences/childSpi(Ljava/lang/String;)Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_AbstractPreferences) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/prefs/AbstractPreferences/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sync2()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/sync2()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "syncSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/syncSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush2()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/flush2()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flushSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/flushSpi()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isRemoved()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/prefs/AbstractPreferences/isRemoved()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "startEventDispatchThreadIfNecessary()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/AbstractPreferences/startEventDispatchThreadIfNecessary()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "prefListeners()[Ljava/util/prefs/PreferenceChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_prefs_PreferenceChangeListener>) => void): void;
    public "java/util/prefs/AbstractPreferences/prefListeners()[Ljava/util/prefs/PreferenceChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_prefs_PreferenceChangeListener>) => void): void;
    public "nodeListeners()[Ljava/util/prefs/NodeChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_prefs_NodeChangeListener>) => void): void;
    public "java/util/prefs/AbstractPreferences/nodeListeners()[Ljava/util/prefs/NodeChangeListener;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_util_prefs_NodeChangeListener>) => void): void;
    public "enqueuePreferenceChangeEvent(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/enqueuePreferenceChangeEvent(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "enqueueNodeAddedEvent(Ljava/util/prefs/Preferences;)V"(thread: JVMThread, args: [java_util_prefs_Preferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/enqueueNodeAddedEvent(Ljava/util/prefs/Preferences;)V"(thread: JVMThread, args: [java_util_prefs_Preferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "enqueueNodeRemovedEvent(Ljava/util/prefs/Preferences;)V"(thread: JVMThread, args: [java_util_prefs_Preferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/enqueueNodeRemovedEvent(Ljava/util/prefs/Preferences;)V"(thread: JVMThread, args: [java_util_prefs_Preferences], cb?: (e?: java_lang_Throwable) => void): void;
    public "exportNode(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/exportNode(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "exportSubtree(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/prefs/AbstractPreferences/exportSubtree(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/util/prefs/AbstractPreferences;)Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: [java_util_prefs_AbstractPreferences], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_AbstractPreferences) => void): void;
    public static "java/util/prefs/AbstractPreferences/access$000(Ljava/util/prefs/AbstractPreferences;)Ljava/util/prefs/AbstractPreferences;"(thread: JVMThread, args: [java_util_prefs_AbstractPreferences], cb?: (e?: java_lang_Throwable, rv?: java_util_prefs_AbstractPreferences) => void): void;
    public static "access$100()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "java/util/prefs/AbstractPreferences/access$100()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/prefs/AbstractPreferences/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_StringTokenizer extends java_lang_Object implements java_util_Enumeration {
    public "java/util/StringTokenizer/currentPosition": number;
    public "java/util/StringTokenizer/newPosition": number;
    public "java/util/StringTokenizer/maxPosition": number;
    public "java/util/StringTokenizer/str": java_lang_String;
    public "java/util/StringTokenizer/delimiters": java_lang_String;
    public "java/util/StringTokenizer/retDelims": number;
    public "java/util/StringTokenizer/delimsChanged": number;
    public "java/util/StringTokenizer/maxDelimCodePoint": number;
    public "java/util/StringTokenizer/hasSurrogates": number;
    public "java/util/StringTokenizer/delimiterCodePoints": JVMArray<number>;
    public "setMaxDelimCodePoint()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/StringTokenizer/setMaxDelimCodePoint()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/StringTokenizer/<init>(Ljava/lang/String;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/StringTokenizer/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/StringTokenizer/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "skipDelimiters(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/StringTokenizer/skipDelimiters(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "scanToken(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/StringTokenizer/scanToken(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDelimiter(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/StringTokenizer/isDelimiter(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasMoreTokens()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/StringTokenizer/hasMoreTokens()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextToken()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/StringTokenizer/nextToken()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "nextToken(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/StringTokenizer/nextToken(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hasMoreElements()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/StringTokenizer/hasMoreElements()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nextElement()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/StringTokenizer/nextElement()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "countTokens()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/StringTokenizer/countTokens()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_util_zip_ZipFile extends java_lang_Object implements java_util_zip_ZipConstants, java_io_Closeable {
    public "java/util/zip/ZipFile/jzfile": Long;
    public "java/util/zip/ZipFile/name": java_lang_String;
    public "java/util/zip/ZipFile/total": number;
    public "java/util/zip/ZipFile/locsig": number;
    public "java/util/zip/ZipFile/closeRequested": number;
    public static "java/util/zip/ZipFile/STORED": number;
    public static "java/util/zip/ZipFile/DEFLATED": number;
    public static "java/util/zip/ZipFile/OPEN_READ": number;
    public static "java/util/zip/ZipFile/OPEN_DELETE": number;
    public static "java/util/zip/ZipFile/usemmap": number;
    public "java/util/zip/ZipFile/zc": java_util_zip_ZipCoder;
    public "java/util/zip/ZipFile/streams": java_util_Map;
    public "java/util/zip/ZipFile/inflaterCache": java_util_Deque;
    public static "java/util/zip/ZipFile/JZENTRY_NAME": number;
    public static "java/util/zip/ZipFile/JZENTRY_EXTRA": number;
    public static "java/util/zip/ZipFile/JZENTRY_COMMENT": number;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/ZipFile/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;I)V"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/<init>(Ljava/io/File;I)V"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;ILjava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_io_File, number, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/<init>(Ljava/io/File;ILjava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_io_File, number, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_lang_String, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/<init>(Ljava/lang/String;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_lang_String, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_io_File, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/<init>(Ljava/io/File;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_io_File, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "getComment()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/zip/ZipFile/getComment()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getEntry(Ljava/lang/String;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "java/util/zip/ZipFile/getEntry(Ljava/lang/String;)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public static "getEntry(J[BZ)J"(thread: JVMThread, args: [Long, any, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/getEntry(J[BZ)J"(thread: JVMThread, args: [Long, any, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "freeEntry(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/ZipFile/freeEntry(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInputStream(Ljava/util/zip/ZipEntry;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/util/zip/ZipFile/getInputStream(Ljava/util/zip/ZipEntry;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_util_zip_ZipEntry], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getInflater()Ljava/util/zip/Inflater;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_Inflater) => void): void;
    public "java/util/zip/ZipFile/getInflater()Ljava/util/zip/Inflater;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_Inflater) => void): void;
    public "releaseInflater(Ljava/util/zip/Inflater;)V"(thread: JVMThread, args: [java_util_zip_Inflater], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/releaseInflater(Ljava/util/zip/Inflater;)V"(thread: JVMThread, args: [java_util_zip_Inflater], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/zip/ZipFile/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "entries()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/zip/ZipFile/entries()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    public "java/util/zip/ZipFile/stream()Ljava/util/stream/Stream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_Stream) => void): void;
    public "getZipEntry(Ljava/lang/String;J)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public "java/util/zip/ZipFile/getZipEntry(Ljava/lang/String;J)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public static "getNextEntry(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/getNextEntry(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/ZipFile/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "close(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/ZipFile/close(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureOpenOrZipException()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipFile/ensureOpenOrZipException()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "startsWithLocHeader()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/ZipFile/startsWithLocHeader()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "open(Ljava/lang/String;IJZ)J"(thread: JVMThread, args: [java_lang_String, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/open(Ljava/lang/String;IJZ)J"(thread: JVMThread, args: [java_lang_String, number, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getTotal(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/ZipFile/getTotal(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "startsWithLOC(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/ZipFile/startsWithLOC(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "read(JJJ[BII)I"(thread: JVMThread, args: [Long, any, Long, any, Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/ZipFile/read(JJJ[BII)I"(thread: JVMThread, args: [Long, any, Long, any, Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getEntryTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/getEntryTime(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getEntryCrc(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/getEntryCrc(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getEntryCSize(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/getEntryCSize(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getEntrySize(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/getEntrySize(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getEntryMethod(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/ZipFile/getEntryMethod(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getEntryFlag(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/ZipFile/getEntryFlag(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getCommentBytes(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/util/zip/ZipFile/getCommentBytes(J)[B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getEntryBytes(JI)[B"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/util/zip/ZipFile/getEntryBytes(JI)[B"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getZipMessage(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/zip/ZipFile/getZipMessage(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$000(Ljava/util/zip/ZipFile;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/util/zip/ZipFile/access$000(Ljava/util/zip/ZipFile;)Ljava/util/Map;"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "access$100(Ljava/util/zip/ZipFile;Ljava/util/zip/Inflater;)V"(thread: JVMThread, args: [java_util_zip_ZipFile, java_util_zip_Inflater], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/ZipFile/access$100(Ljava/util/zip/ZipFile;Ljava/util/zip/Inflater;)V"(thread: JVMThread, args: [java_util_zip_ZipFile, java_util_zip_Inflater], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$200(Ljava/util/zip/ZipFile;)V"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/ZipFile/access$200(Ljava/util/zip/ZipFile;)V"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$300(Ljava/util/zip/ZipFile;)I"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/ZipFile/access$300(Ljava/util/zip/ZipFile;)I"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$400(Ljava/util/zip/ZipFile;)J"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/access$400(Ljava/util/zip/ZipFile;)J"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$500(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/access$500(JI)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$600(Ljava/util/zip/ZipFile;)Z"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/ZipFile/access$600(Ljava/util/zip/ZipFile;)Z"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$700(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/zip/ZipFile/access$700(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$800(Ljava/util/zip/ZipFile;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/zip/ZipFile/access$800(Ljava/util/zip/ZipFile;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$900(Ljava/util/zip/ZipFile;Ljava/lang/String;J)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_util_zip_ZipFile, java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public static "java/util/zip/ZipFile/access$900(Ljava/util/zip/ZipFile;Ljava/lang/String;J)Ljava/util/zip/ZipEntry;"(thread: JVMThread, args: [java_util_zip_ZipFile, java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipEntry) => void): void;
    public static "access$1000(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/ZipFile/access$1000(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$1100(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/access$1100(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1200(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/ZipFile/access$1200(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$1300(Ljava/util/zip/ZipFile;)V"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/ZipFile/access$1300(Ljava/util/zip/ZipFile;)V"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$1400(JJJ[BII)I"(thread: JVMThread, args: [Long, any, Long, any, Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/ZipFile/access$1400(JJJ[BII)I"(thread: JVMThread, args: [Long, any, Long, any, Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1500(Ljava/util/zip/ZipFile;)Z"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/ZipFile/access$1500(Ljava/util/zip/ZipFile;)Z"(thread: JVMThread, args: [java_util_zip_ZipFile], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/ZipFile/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_zip_ZipCoder extends java_lang_Object {
    public "java/util/zip/ZipCoder/cs": java_nio_charset_Charset;
    public "java/util/zip/ZipCoder/dec": java_nio_charset_CharsetDecoder;
    public "java/util/zip/ZipCoder/enc": java_nio_charset_CharsetEncoder;
    public "java/util/zip/ZipCoder/isUTF8": number;
    public "java/util/zip/ZipCoder/utf8": java_util_zip_ZipCoder;
    public "toString([BI)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/zip/ZipCoder/toString([BI)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/zip/ZipCoder/toString([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getBytes(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/util/zip/ZipCoder/getBytes(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getBytesUTF8(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/util/zip/ZipCoder/getBytesUTF8(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "toStringUTF8([BI)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/zip/ZipCoder/toStringUTF8([BI)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isUTF8()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/ZipCoder/isUTF8()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZipCoder/<init>(Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public static "get(Ljava/nio/charset/Charset;)Ljava/util/zip/ZipCoder;"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipCoder) => void): void;
    public static "java/util/zip/ZipCoder/get(Ljava/nio/charset/Charset;)Ljava/util/zip/ZipCoder;"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: java_util_zip_ZipCoder) => void): void;
    public "decoder()Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "java/util/zip/ZipCoder/decoder()Ljava/nio/charset/CharsetDecoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetDecoder) => void): void;
    public "encoder()Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
    public "java/util/zip/ZipCoder/encoder()Ljava/nio/charset/CharsetEncoder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_CharsetEncoder) => void): void;
  }
  export class java_util_zip_Inflater extends java_lang_Object {
    public "java/util/zip/Inflater/zsRef": java_util_zip_ZStreamRef;
    public "java/util/zip/Inflater/buf": JVMArray<number>;
    public "java/util/zip/Inflater/off": number;
    public "java/util/zip/Inflater/len": number;
    public "java/util/zip/Inflater/finished": number;
    public "java/util/zip/Inflater/needDict": number;
    public "java/util/zip/Inflater/bytesRead": Long;
    public "java/util/zip/Inflater/bytesWritten": Long;
    public static "java/util/zip/Inflater/defaultBuf": JVMArray<number>;
    public static "java/util/zip/Inflater/$assertionsDisabled": number;
    public "<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/<init>(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setInput([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/setInput([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setInput([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/setInput([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDictionary([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/setDictionary([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDictionary([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/setDictionary([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRemaining()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/getRemaining()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "needsInput()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/needsInput()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "needsDictionary()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/needsDictionary()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finished()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/finished()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inflate([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/inflate([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inflate([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/inflate([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAdler()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/getAdler()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTotalIn()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/getTotalIn()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBytesRead()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/Inflater/getBytesRead()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTotalOut()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/getTotalOut()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBytesWritten()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/Inflater/getBytesWritten()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Inflater/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ended()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/ended()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Inflater/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init(Z)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/Inflater/init(Z)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "setDictionary(J[BII)V"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Inflater/setDictionary(J[BII)V"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "inflateBytes(J[BII)I"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Inflater/inflateBytes(J[BII)I"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getAdler(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/Inflater/getAdler(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "reset(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Inflater/reset(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "end(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Inflater/end(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Inflater/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_zip_ZStreamRef extends java_lang_Object {
    public "java/util/zip/ZStreamRef/address": Long;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZStreamRef/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "address()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/ZStreamRef/address()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/ZStreamRef/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_zip_CRC32 extends java_lang_Object implements java_util_zip_Checksum {
    public "java/util/zip/CRC32/crc": number;
    public static "java/util/zip/CRC32/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/CRC32/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/CRC32/update(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "update([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/CRC32/update([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "update([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/CRC32/update([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/CRC32/update(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/CRC32/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/CRC32/getValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "update(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/CRC32/update(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "updateBytes(I[BII)I"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/CRC32/updateBytes(I[BII)I"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "updateByteBuffer(IJII)I"(thread: JVMThread, args: [number, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/CRC32/updateByteBuffer(IJII)I"(thread: JVMThread, args: [number, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/CRC32/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_util_zip_Checksum extends java_lang_Object {
    "update(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    "update([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    "getValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_zip_Deflater extends java_lang_Object {
    public "java/util/zip/Deflater/zsRef": java_util_zip_ZStreamRef;
    public "java/util/zip/Deflater/buf": JVMArray<number>;
    public "java/util/zip/Deflater/off": number;
    public "java/util/zip/Deflater/len": number;
    public "java/util/zip/Deflater/level": number;
    public "java/util/zip/Deflater/strategy": number;
    public "java/util/zip/Deflater/setParams": number;
    public "java/util/zip/Deflater/finish": number;
    public "java/util/zip/Deflater/finished": number;
    public "java/util/zip/Deflater/bytesRead": Long;
    public "java/util/zip/Deflater/bytesWritten": Long;
    public static "java/util/zip/Deflater/DEFLATED": number;
    public static "java/util/zip/Deflater/NO_COMPRESSION": number;
    public static "java/util/zip/Deflater/BEST_SPEED": number;
    public static "java/util/zip/Deflater/BEST_COMPRESSION": number;
    public static "java/util/zip/Deflater/DEFAULT_COMPRESSION": number;
    public static "java/util/zip/Deflater/FILTERED": number;
    public static "java/util/zip/Deflater/HUFFMAN_ONLY": number;
    public static "java/util/zip/Deflater/DEFAULT_STRATEGY": number;
    public static "java/util/zip/Deflater/NO_FLUSH": number;
    public static "java/util/zip/Deflater/SYNC_FLUSH": number;
    public static "java/util/zip/Deflater/FULL_FLUSH": number;
    public static "java/util/zip/Deflater/$assertionsDisabled": number;
    public "<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/<init>(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setInput([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/setInput([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setInput([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/setInput([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDictionary([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/setDictionary([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDictionary([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/setDictionary([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setStrategy(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/setStrategy(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLevel(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/setLevel(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "needsInput()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Deflater/needsInput()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finish()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/finish()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finished()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Deflater/finished()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "deflate([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Deflater/deflate([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "deflate([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Deflater/deflate([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "deflate([BIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Deflater/deflate([BIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAdler()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Deflater/getAdler()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTotalIn()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Deflater/getTotalIn()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBytesRead()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/Deflater/getBytesRead()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getTotalOut()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Deflater/getTotalOut()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBytesWritten()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/Deflater/getBytesWritten()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/end()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Deflater/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Deflater/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init(IIZ)J"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/util/zip/Deflater/init(IIZ)J"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "setDictionary(J[BII)V"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Deflater/setDictionary(J[BII)V"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "deflateBytes(J[BIII)I"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/zip/Deflater/deflateBytes(J[BIII)I"(thread: JVMThread, args: [Long, any, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getAdler(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/Deflater/getAdler(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "reset(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Deflater/reset(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "end(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Deflater/end(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Deflater/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_zip_Adler32 extends java_lang_Object implements java_util_zip_Checksum {
    public "java/util/zip/Adler32/adler": number;
    public static "java/util/zip/Adler32/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Adler32/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Adler32/update(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "update([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Adler32/update([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "update([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Adler32/update([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "update(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Adler32/update(Ljava/nio/ByteBuffer;)V"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/zip/Adler32/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/util/zip/Adler32/getValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "update(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/Adler32/update(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "updateBytes(I[BII)I"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/Adler32/updateBytes(I[BII)I"(thread: JVMThread, args: [number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "updateByteBuffer(IJII)I"(thread: JVMThread, args: [number, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/zip/Adler32/updateByteBuffer(IJII)I"(thread: JVMThread, args: [number, Long, any, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/zip/Adler32/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_TimeZone extends java_lang_Object implements java_io_Serializable, java_lang_Cloneable {
    public static "java/util/TimeZone/SHORT": number;
    public static "java/util/TimeZone/LONG": number;
    public static "java/util/TimeZone/ONE_MINUTE": number;
    public static "java/util/TimeZone/ONE_HOUR": number;
    public static "java/util/TimeZone/ONE_DAY": number;
    public static "java/util/TimeZone/serialVersionUID": Long;
    public static "java/util/TimeZone/NO_TIMEZONE": java_util_TimeZone;
    public "java/util/TimeZone/ID": java_lang_String;
    public static "java/util/TimeZone/defaultTimeZone": java_util_TimeZone;
    public static "java/util/TimeZone/GMT_ID": java_lang_String;
    public static "java/util/TimeZone/GMT_ID_LENGTH": number;
    public static "java/util/TimeZone/mainAppContextDefault": java_util_TimeZone;
    public static "java/util/TimeZone/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TimeZone/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOffset(IIIIII)I"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimeZone/getOffset(IIIIII)I"(thread: JVMThread, args: [number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOffset(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimeZone/getOffset(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getOffsets(J[I)I"(thread: JVMThread, args: [Long, any, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimeZone/getOffsets(J[I)I"(thread: JVMThread, args: [Long, any, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setRawOffset(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TimeZone/setRawOffset(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getRawOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimeZone/getRawOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getID()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/TimeZone/getID()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setID(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/TimeZone/setID(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDisplayName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/TimeZone/getDisplayName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/TimeZone/getDisplayName(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayName(ZI)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/TimeZone/getDisplayName(ZI)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDisplayName(ZILjava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [number, number, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/TimeZone/getDisplayName(ZILjava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [number, number, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getDisplayNames(Ljava/lang/String;Ljava/util/Locale;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/util/TimeZone/getDisplayNames(Ljava/lang/String;Ljava/util/Locale;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getDSTSavings()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimeZone/getDSTSavings()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "useDaylightTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimeZone/useDaylightTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "observesDaylightTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimeZone/observesDaylightTime()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inDaylightTime(Ljava/util/Date;)Z"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimeZone/inDaylightTime(Ljava/util/Date;)Z"(thread: JVMThread, args: [java_util_Date], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getTimeZone(Ljava/lang/String;)Ljava/util/TimeZone;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "java/util/TimeZone/getTimeZone(Ljava/lang/String;)Ljava/util/TimeZone;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "getTimeZone(Ljava/time/ZoneId;)Ljava/util/TimeZone;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "java/util/TimeZone/getTimeZone(Ljava/time/ZoneId;)Ljava/util/TimeZone;"(thread: JVMThread, args: [java_time_ZoneId], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public "toZoneId()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public "java/util/TimeZone/toZoneId()Ljava/time/ZoneId;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_time_ZoneId) => void): void;
    public static "getTimeZone(Ljava/lang/String;Z)Ljava/util/TimeZone;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "java/util/TimeZone/getTimeZone(Ljava/lang/String;Z)Ljava/util/TimeZone;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "getAvailableIDs(I)[Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/util/TimeZone/getAvailableIDs(I)[Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "getAvailableIDs()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/util/TimeZone/getAvailableIDs()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "getSystemTimeZoneID(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/TimeZone/getSystemTimeZoneID(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getSystemGMTOffsetID()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/util/TimeZone/getSystemGMTOffsetID()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getDefault()Ljava/util/TimeZone;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "java/util/TimeZone/getDefault()Ljava/util/TimeZone;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "getDefaultRef()Ljava/util/TimeZone;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "java/util/TimeZone/getDefaultRef()Ljava/util/TimeZone;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "setDefaultZone()Ljava/util/TimeZone;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "java/util/TimeZone/setDefaultZone()Ljava/util/TimeZone;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "setDefault(Ljava/util/TimeZone;)V"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/TimeZone/setDefault(Ljava/util/TimeZone;)V"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable) => void): void;
    public "hasSameRules(Ljava/util/TimeZone;)Z"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/TimeZone/hasSameRules(Ljava/util/TimeZone;)Z"(thread: JVMThread, args: [java_util_TimeZone], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/TimeZone/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "parseCustomTimeZone(Ljava/lang/String;)Ljava/util/TimeZone;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "java/util/TimeZone/parseCustomTimeZone(Ljava/lang/String;)Ljava/util/TimeZone;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_TimeZone) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/TimeZone/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_security_PrivilegedExceptionAction extends java_lang_Object {
    "run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_security_AccessControlContext extends java_lang_Object {
    public "java/security/AccessControlContext/context": JVMArray<java_security_ProtectionDomain>;
    public "java/security/AccessControlContext/isPrivileged": number;
    public "java/security/AccessControlContext/isAuthorized": number;
    public "java/security/AccessControlContext/privilegedContext": java_security_AccessControlContext;
    public "java/security/AccessControlContext/combiner": java_security_DomainCombiner;
    public "java/security/AccessControlContext/permissions": JVMArray<java_security_Permission>;
    public "java/security/AccessControlContext/parent": java_security_AccessControlContext;
    public "java/security/AccessControlContext/isWrapped": number;
    public "java/security/AccessControlContext/isLimited": number;
    public "java/security/AccessControlContext/limitedContext": JVMArray<java_security_ProtectionDomain>;
    public static "java/security/AccessControlContext/debugInit": number;
    public static "java/security/AccessControlContext/debug": sun_security_util_Debug;
    public static "getDebug()Lsun/security/util/Debug;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_Debug) => void): void;
    public static "java/security/AccessControlContext/getDebug()Lsun/security/util/Debug;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_security_util_Debug) => void): void;
    public "<init>([Ljava/security/ProtectionDomain;)V"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessControlContext/<init>([Ljava/security/ProtectionDomain;)V"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/security/AccessControlContext;Ljava/security/DomainCombiner;)V"(thread: JVMThread, args: [java_security_AccessControlContext, java_security_DomainCombiner], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessControlContext/<init>(Ljava/security/AccessControlContext;Ljava/security/DomainCombiner;)V"(thread: JVMThread, args: [java_security_AccessControlContext, java_security_DomainCombiner], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/security/AccessControlContext;Ljava/security/DomainCombiner;Z)V"(thread: JVMThread, args: [java_security_AccessControlContext, java_security_DomainCombiner, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessControlContext/<init>(Ljava/security/AccessControlContext;Ljava/security/DomainCombiner;Z)V"(thread: JVMThread, args: [java_security_AccessControlContext, java_security_DomainCombiner, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/security/ProtectionDomain;Ljava/security/DomainCombiner;Ljava/security/AccessControlContext;Ljava/security/AccessControlContext;[Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_ProtectionDomain, java_security_DomainCombiner, java_security_AccessControlContext, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessControlContext/<init>(Ljava/security/ProtectionDomain;Ljava/security/DomainCombiner;Ljava/security/AccessControlContext;Ljava/security/AccessControlContext;[Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_ProtectionDomain, java_security_DomainCombiner, java_security_AccessControlContext, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Ljava/security/ProtectionDomain;Z)V"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessControlContext/<init>([Ljava/security/ProtectionDomain;Z)V"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Ljava/security/ProtectionDomain;Ljava/security/AccessControlContext;)V"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessControlContext/<init>([Ljava/security/ProtectionDomain;Ljava/security/AccessControlContext;)V"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "getContext()[Ljava/security/ProtectionDomain;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_ProtectionDomain>) => void): void;
    public "java/security/AccessControlContext/getContext()[Ljava/security/ProtectionDomain;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_ProtectionDomain>) => void): void;
    public "isPrivileged()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/AccessControlContext/isPrivileged()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAssignedCombiner()Ljava/security/DomainCombiner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_DomainCombiner) => void): void;
    public "java/security/AccessControlContext/getAssignedCombiner()Ljava/security/DomainCombiner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_DomainCombiner) => void): void;
    public "getDomainCombiner()Ljava/security/DomainCombiner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_DomainCombiner) => void): void;
    public "java/security/AccessControlContext/getDomainCombiner()Ljava/security/DomainCombiner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_DomainCombiner) => void): void;
    public "getCombiner()Ljava/security/DomainCombiner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_DomainCombiner) => void): void;
    public "java/security/AccessControlContext/getCombiner()Ljava/security/DomainCombiner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_DomainCombiner) => void): void;
    public "isAuthorized()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/AccessControlContext/isAuthorized()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkPermission(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessControlContext/checkPermission(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkPermission2(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessControlContext/checkPermission2(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
    public "optimize()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public "java/security/AccessControlContext/optimize()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "combine([Ljava/security/ProtectionDomain;[Ljava/security/ProtectionDomain;)[Ljava/security/ProtectionDomain;"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, JVMArray<java_security_ProtectionDomain>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_ProtectionDomain>) => void): void;
    public static "java/security/AccessControlContext/combine([Ljava/security/ProtectionDomain;[Ljava/security/ProtectionDomain;)[Ljava/security/ProtectionDomain;"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, JVMArray<java_security_ProtectionDomain>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_ProtectionDomain>) => void): void;
    public "calculateFields(Ljava/security/AccessControlContext;Ljava/security/AccessControlContext;[Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_AccessControlContext, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessControlContext/calculateFields(Ljava/security/AccessControlContext;Ljava/security/AccessControlContext;[Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_AccessControlContext, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/AccessControlContext/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equalContext(Ljava/security/AccessControlContext;)Z"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/AccessControlContext/equalContext(Ljava/security/AccessControlContext;)Z"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equalPDs([Ljava/security/ProtectionDomain;[Ljava/security/ProtectionDomain;)Z"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, JVMArray<java_security_ProtectionDomain>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/AccessControlContext/equalPDs([Ljava/security/ProtectionDomain;[Ljava/security/ProtectionDomain;)Z"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, JVMArray<java_security_ProtectionDomain>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equalLimitedContext(Ljava/security/AccessControlContext;)Z"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/AccessControlContext/equalLimitedContext(Ljava/security/AccessControlContext;)Z"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getNextPC(Ljava/security/AccessControlContext;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "java/security/AccessControlContext/getNextPC(Ljava/security/AccessControlContext;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "containsAllPDs([Ljava/security/ProtectionDomain;[Ljava/security/ProtectionDomain;)Z"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, JVMArray<java_security_ProtectionDomain>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/security/AccessControlContext/containsAllPDs([Ljava/security/ProtectionDomain;[Ljava/security/ProtectionDomain;)Z"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, JVMArray<java_security_ProtectionDomain>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "containsAllLimits(Ljava/security/AccessControlContext;)Z"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/AccessControlContext/containsAllLimits(Ljava/security/AccessControlContext;)Z"(thread: JVMThread, args: [java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/AccessControlContext/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/security/AccessControlContext/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_security_DomainCombiner extends java_lang_Object {
    "combine([Ljava/security/ProtectionDomain;[Ljava/security/ProtectionDomain;)[Ljava/security/ProtectionDomain;"(thread: JVMThread, args: [JVMArray<java_security_ProtectionDomain>, JVMArray<java_security_ProtectionDomain>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_ProtectionDomain>) => void): void;
  }
  export interface java_security_PrivilegedAction extends java_lang_Object {
    "run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_security_AccessController extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/AccessController/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "doPrivileged(Ljava/security/PrivilegedAction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivileged(Ljava/security/PrivilegedAction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doPrivilegedWithCombiner(Ljava/security/PrivilegedAction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivilegedWithCombiner(Ljava/security/PrivilegedAction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doPrivileged(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivileged(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doPrivileged(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivileged(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doPrivilegedWithCombiner(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivilegedWithCombiner(Ljava/security/PrivilegedAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedAction, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doPrivileged(Ljava/security/PrivilegedExceptionAction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivileged(Ljava/security/PrivilegedExceptionAction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doPrivilegedWithCombiner(Ljava/security/PrivilegedExceptionAction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivilegedWithCombiner(Ljava/security/PrivilegedExceptionAction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "preserveCombiner(Ljava/security/DomainCombiner;Ljava/lang/Class;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [java_security_DomainCombiner, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "java/security/AccessController/preserveCombiner(Ljava/security/DomainCombiner;Ljava/lang/Class;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [java_security_DomainCombiner, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "createWrapper(Ljava/security/DomainCombiner;Ljava/lang/Class;Ljava/security/AccessControlContext;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [java_security_DomainCombiner, java_lang_Class, java_security_AccessControlContext, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "java/security/AccessController/createWrapper(Ljava/security/DomainCombiner;Ljava/lang/Class;Ljava/security/AccessControlContext;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/security/AccessControlContext;"(thread: JVMThread, args: [java_security_DomainCombiner, java_lang_Class, java_security_AccessControlContext, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "getCallerPD(Ljava/lang/Class;)Ljava/security/ProtectionDomain;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_security_ProtectionDomain) => void): void;
    public static "java/security/AccessController/getCallerPD(Ljava/lang/Class;)Ljava/security/ProtectionDomain;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_security_ProtectionDomain) => void): void;
    public static "doPrivileged(Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivileged(Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doPrivileged(Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivileged(Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "doPrivilegedWithCombiner(Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/security/AccessController/doPrivilegedWithCombiner(Ljava/security/PrivilegedExceptionAction;Ljava/security/AccessControlContext;[Ljava/security/Permission;)Ljava/lang/Object;"(thread: JVMThread, args: [java_security_PrivilegedExceptionAction, java_security_AccessControlContext, JVMArray<java_security_Permission>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "getStackAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "java/security/AccessController/getStackAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "getInheritedAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "java/security/AccessController/getInheritedAccessControlContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "getContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "java/security/AccessController/getContext()Ljava/security/AccessControlContext;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_AccessControlContext) => void): void;
    public static "checkPermission(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/security/AccessController/checkPermission(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_Bits extends java_lang_Object {
    public static "java/nio/Bits/unsafe": sun_misc_Unsafe;
    public static "java/nio/Bits/byteOrder": java_nio_ByteOrder;
    public static "java/nio/Bits/pageSize": number;
    public static "java/nio/Bits/unaligned": number;
    public static "java/nio/Bits/unalignedKnown": number;
    public static "java/nio/Bits/maxMemory": Long;
    public static "java/nio/Bits/reservedMemory": java_util_concurrent_atomic_AtomicLong;
    public static "java/nio/Bits/totalCapacity": java_util_concurrent_atomic_AtomicLong;
    public static "java/nio/Bits/count": java_util_concurrent_atomic_AtomicLong;
    public static "java/nio/Bits/memoryLimitSet": number;
    public static "java/nio/Bits/MAX_SLEEPS": number;
    public static "java/nio/Bits/JNI_COPY_TO_ARRAY_THRESHOLD": number;
    public static "java/nio/Bits/JNI_COPY_FROM_ARRAY_THRESHOLD": number;
    public static "java/nio/Bits/UNSAFE_COPY_THRESHOLD": Long;
    public static "java/nio/Bits/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/Bits/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "swap(S)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/swap(S)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "swap(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/swap(C)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "swap(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/swap(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "swap(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/nio/Bits/swap(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "makeChar(BB)C"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/makeChar(BB)C"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getCharL(Ljava/nio/ByteBuffer;I)C"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getCharL(Ljava/nio/ByteBuffer;I)C"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getCharL(J)C"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getCharL(J)C"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getCharB(Ljava/nio/ByteBuffer;I)C"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getCharB(Ljava/nio/ByteBuffer;I)C"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getCharB(J)C"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getCharB(J)C"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getChar(Ljava/nio/ByteBuffer;IZ)C"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getChar(Ljava/nio/ByteBuffer;IZ)C"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getChar(JZ)C"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getChar(JZ)C"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "char1(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/char1(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "char0(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/char0(C)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "putCharL(Ljava/nio/ByteBuffer;IC)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putCharL(Ljava/nio/ByteBuffer;IC)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putCharL(JC)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putCharL(JC)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putCharB(Ljava/nio/ByteBuffer;IC)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putCharB(Ljava/nio/ByteBuffer;IC)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putCharB(JC)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putCharB(JC)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putChar(Ljava/nio/ByteBuffer;ICZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putChar(Ljava/nio/ByteBuffer;ICZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putChar(JCZ)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putChar(JCZ)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "makeShort(BB)S"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/makeShort(BB)S"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getShortL(Ljava/nio/ByteBuffer;I)S"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getShortL(Ljava/nio/ByteBuffer;I)S"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getShortL(J)S"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getShortL(J)S"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getShortB(Ljava/nio/ByteBuffer;I)S"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getShortB(Ljava/nio/ByteBuffer;I)S"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getShortB(J)S"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getShortB(J)S"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getShort(Ljava/nio/ByteBuffer;IZ)S"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getShort(Ljava/nio/ByteBuffer;IZ)S"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getShort(JZ)S"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getShort(JZ)S"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "short1(S)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/short1(S)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "short0(S)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/short0(S)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "putShortL(Ljava/nio/ByteBuffer;IS)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putShortL(Ljava/nio/ByteBuffer;IS)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putShortL(JS)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putShortL(JS)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putShortB(Ljava/nio/ByteBuffer;IS)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putShortB(Ljava/nio/ByteBuffer;IS)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putShortB(JS)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putShortB(JS)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putShort(Ljava/nio/ByteBuffer;ISZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putShort(Ljava/nio/ByteBuffer;ISZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putShort(JSZ)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putShort(JSZ)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "makeInt(BBBB)I"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/makeInt(BBBB)I"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getIntL(Ljava/nio/ByteBuffer;I)I"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getIntL(Ljava/nio/ByteBuffer;I)I"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getIntL(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getIntL(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getIntB(Ljava/nio/ByteBuffer;I)I"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getIntB(Ljava/nio/ByteBuffer;I)I"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getIntB(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getIntB(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getInt(Ljava/nio/ByteBuffer;IZ)I"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getInt(Ljava/nio/ByteBuffer;IZ)I"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getInt(JZ)I"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getInt(JZ)I"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "int3(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/int3(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "int2(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/int2(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "int1(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/int1(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "int0(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/int0(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "putIntL(Ljava/nio/ByteBuffer;II)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putIntL(Ljava/nio/ByteBuffer;II)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putIntL(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putIntL(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putIntB(Ljava/nio/ByteBuffer;II)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putIntB(Ljava/nio/ByteBuffer;II)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putIntB(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putIntB(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putInt(Ljava/nio/ByteBuffer;IIZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putInt(Ljava/nio/ByteBuffer;IIZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putInt(JIZ)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putInt(JIZ)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "makeLong(BBBBBBBB)J"(thread: JVMThread, args: [number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/nio/Bits/makeLong(BBBBBBBB)J"(thread: JVMThread, args: [number, number, number, number, number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getLongL(Ljava/nio/ByteBuffer;I)J"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/nio/Bits/getLongL(Ljava/nio/ByteBuffer;I)J"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getLongL(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/nio/Bits/getLongL(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getLongB(Ljava/nio/ByteBuffer;I)J"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/nio/Bits/getLongB(Ljava/nio/ByteBuffer;I)J"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getLongB(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/nio/Bits/getLongB(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getLong(Ljava/nio/ByteBuffer;IZ)J"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/nio/Bits/getLong(Ljava/nio/ByteBuffer;IZ)J"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getLong(JZ)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/nio/Bits/getLong(JZ)J"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "long7(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/long7(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "long6(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/long6(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "long5(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/long5(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "long4(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/long4(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "long3(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/long3(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "long2(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/long2(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "long1(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/long1(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "long0(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/long0(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "putLongL(Ljava/nio/ByteBuffer;IJ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putLongL(Ljava/nio/ByteBuffer;IJ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putLongL(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putLongL(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putLongB(Ljava/nio/ByteBuffer;IJ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putLongB(Ljava/nio/ByteBuffer;IJ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putLongB(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putLongB(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putLong(Ljava/nio/ByteBuffer;IJZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putLong(Ljava/nio/ByteBuffer;IJZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putLong(JJZ)V"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putLong(JJZ)V"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getFloatL(Ljava/nio/ByteBuffer;I)F"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getFloatL(Ljava/nio/ByteBuffer;I)F"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFloatL(J)F"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getFloatL(J)F"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFloatB(Ljava/nio/ByteBuffer;I)F"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getFloatB(Ljava/nio/ByteBuffer;I)F"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFloatB(J)F"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getFloatB(J)F"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFloat(Ljava/nio/ByteBuffer;IZ)F"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getFloat(Ljava/nio/ByteBuffer;IZ)F"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getFloat(JZ)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getFloat(JZ)F"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "putFloatL(Ljava/nio/ByteBuffer;IF)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putFloatL(Ljava/nio/ByteBuffer;IF)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putFloatL(JF)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putFloatL(JF)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putFloatB(Ljava/nio/ByteBuffer;IF)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putFloatB(Ljava/nio/ByteBuffer;IF)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putFloatB(JF)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putFloatB(JF)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putFloat(Ljava/nio/ByteBuffer;IFZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putFloat(Ljava/nio/ByteBuffer;IFZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putFloat(JFZ)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putFloat(JFZ)V"(thread: JVMThread, args: [Long, any, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getDoubleL(Ljava/nio/ByteBuffer;I)D"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getDoubleL(Ljava/nio/ByteBuffer;I)D"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDoubleL(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getDoubleL(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDoubleB(Ljava/nio/ByteBuffer;I)D"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getDoubleB(Ljava/nio/ByteBuffer;I)D"(thread: JVMThread, args: [java_nio_ByteBuffer, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDoubleB(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getDoubleB(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDouble(Ljava/nio/ByteBuffer;IZ)D"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getDouble(Ljava/nio/ByteBuffer;IZ)D"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDouble(JZ)D"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/getDouble(JZ)D"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "putDoubleL(Ljava/nio/ByteBuffer;ID)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putDoubleL(Ljava/nio/ByteBuffer;ID)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putDoubleL(JD)V"(thread: JVMThread, args: [Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putDoubleL(JD)V"(thread: JVMThread, args: [Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putDoubleB(Ljava/nio/ByteBuffer;ID)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putDoubleB(Ljava/nio/ByteBuffer;ID)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putDoubleB(JD)V"(thread: JVMThread, args: [Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putDoubleB(JD)V"(thread: JVMThread, args: [Long, any, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putDouble(Ljava/nio/ByteBuffer;IDZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putDouble(Ljava/nio/ByteBuffer;IDZ)V"(thread: JVMThread, args: [java_nio_ByteBuffer, number, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "putDouble(JDZ)V"(thread: JVMThread, args: [Long, any, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/putDouble(JDZ)V"(thread: JVMThread, args: [Long, any, number, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "_get(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/_get(J)B"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "_put(JB)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/_put(JB)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "unsafe()Lsun/misc/Unsafe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Unsafe) => void): void;
    public static "java/nio/Bits/unsafe()Lsun/misc/Unsafe;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Unsafe) => void): void;
    public static "byteOrder()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public static "java/nio/Bits/byteOrder()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public static "pageSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/pageSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "pageCount(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/pageCount(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "unaligned()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/unaligned()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "reserveMemory(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/reserveMemory(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "tryReserveMemory(JI)Z"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/Bits/tryReserveMemory(JI)Z"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "unreserveMemory(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/unreserveMemory(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyFromArray(Ljava/lang/Object;JJJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyFromArray(Ljava/lang/Object;JJJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyToArray(JLjava/lang/Object;JJJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyToArray(JLjava/lang/Object;JJJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyFromCharArray(Ljava/lang/Object;JJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyFromCharArray(Ljava/lang/Object;JJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyToCharArray(JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyToCharArray(JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyFromShortArray(Ljava/lang/Object;JJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyFromShortArray(Ljava/lang/Object;JJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyToShortArray(JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyToShortArray(JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyFromIntArray(Ljava/lang/Object;JJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyFromIntArray(Ljava/lang/Object;JJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyToIntArray(JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyToIntArray(JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyFromLongArray(Ljava/lang/Object;JJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyFromLongArray(Ljava/lang/Object;JJJ)V"(thread: JVMThread, args: [java_lang_Object, Long, any, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "copyToLongArray(JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/copyToLongArray(JLjava/lang/Object;JJ)V"(thread: JVMThread, args: [Long, any, java_lang_Object, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000()Ljava/util/concurrent/atomic/AtomicLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicLong) => void): void;
    public static "java/nio/Bits/access$000()Ljava/util/concurrent/atomic/AtomicLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicLong) => void): void;
    public static "access$100()Ljava/util/concurrent/atomic/AtomicLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicLong) => void): void;
    public static "java/nio/Bits/access$100()Ljava/util/concurrent/atomic/AtomicLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicLong) => void): void;
    public static "access$200()Ljava/util/concurrent/atomic/AtomicLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicLong) => void): void;
    public static "java/nio/Bits/access$200()Ljava/util/concurrent/atomic/AtomicLong;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_atomic_AtomicLong) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/Bits/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_MappedByteBuffer extends java_nio_ByteBuffer {
    public "java/nio/MappedByteBuffer/fd": java_io_FileDescriptor;
    public static "java/nio/MappedByteBuffer/unused": number;
    public "<init>(IIIILjava/io/FileDescriptor;)V"(thread: JVMThread, args: [number, number, number, number, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/MappedByteBuffer/<init>(IIIILjava/io/FileDescriptor;)V"(thread: JVMThread, args: [number, number, number, number, java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/MappedByteBuffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkMapped()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/MappedByteBuffer/checkMapped()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mappingOffset()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/MappedByteBuffer/mappingOffset()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "mappingAddress(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/MappedByteBuffer/mappingAddress(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "mappingLength(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/MappedByteBuffer/mappingLength(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "isLoaded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/MappedByteBuffer/isLoaded()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "load()Ljava/nio/MappedByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_MappedByteBuffer) => void): void;
    public "java/nio/MappedByteBuffer/load()Ljava/nio/MappedByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_MappedByteBuffer) => void): void;
    public "force()Ljava/nio/MappedByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_MappedByteBuffer) => void): void;
    public "java/nio/MappedByteBuffer/force()Ljava/nio/MappedByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_MappedByteBuffer) => void): void;
    public "isLoaded0(JJI)Z"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/MappedByteBuffer/isLoaded0(JJI)Z"(thread: JVMThread, args: [Long, any, Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "load0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/MappedByteBuffer/load0(JJ)V"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "force0(Ljava/io/FileDescriptor;JJ)V"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/MappedByteBuffer/force0(Ljava/io/FileDescriptor;JJ)V"(thread: JVMThread, args: [java_io_FileDescriptor, Long, any, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_InetAddressImplFactory extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddressImplFactory/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "create()Ljava/net/InetAddressImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddressImpl) => void): void;
    public static "java/net/InetAddressImplFactory/create()Ljava/net/InetAddressImpl;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddressImpl) => void): void;
    public static "isIPv6Supported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/InetAddressImplFactory/isIPv6Supported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface java_net_InetAddressImpl extends java_lang_Object {
    "getLocalHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "lookupAllHostAddr(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    "getHostByAddr([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "anyLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    "loopbackAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    "isReachable(Ljava/net/InetAddress;ILjava/net/NetworkInterface;I)Z"(thread: JVMThread, args: [java_net_InetAddress, number, java_net_NetworkInterface, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_net_Inet4AddressImpl extends java_lang_Object implements java_net_InetAddressImpl {
    public "java/net/Inet4AddressImpl/anyLocalAddress": java_net_InetAddress;
    public "java/net/Inet4AddressImpl/loopbackAddress": java_net_InetAddress;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet4AddressImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocalHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/Inet4AddressImpl/getLocalHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "lookupAllHostAddr(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public "java/net/Inet4AddressImpl/lookupAllHostAddr(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public "getHostByAddr([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/Inet4AddressImpl/getHostByAddr([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isReachable0([BI[BI)Z"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4AddressImpl/isReachable0([BI[BI)Z"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "anyLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/Inet4AddressImpl/anyLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "loopbackAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/Inet4AddressImpl/loopbackAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "isReachable(Ljava/net/InetAddress;ILjava/net/NetworkInterface;I)Z"(thread: JVMThread, args: [java_net_InetAddress, number, java_net_NetworkInterface, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4AddressImpl/isReachable(Ljava/net/InetAddress;ILjava/net/NetworkInterface;I)Z"(thread: JVMThread, args: [java_net_InetAddress, number, java_net_NetworkInterface, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_net_Inet4Address extends java_net_InetAddress {
    public static "java/net/Inet4Address/INADDRSZ": number;
    public static "java/net/Inet4Address/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet4Address/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet4Address/<init>(Ljava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet4Address/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/Inet4Address/writeReplace()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isMulticastAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isMulticastAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAnyLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isAnyLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLoopbackAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isLoopbackAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLinkLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isLinkLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSiteLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isSiteLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCGlobal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isMCGlobal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCNodeLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isMCNodeLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCLinkLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isMCLinkLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCSiteLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isMCSiteLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCOrgLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/isMCOrgLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAddress()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/net/Inet4Address/getAddress()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getHostAddress()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/Inet4Address/getHostAddress()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet4Address/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "numericToTextFormat([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/Inet4Address/numericToTextFormat([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/Inet4Address/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/Inet4Address/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_Inet6Address extends java_net_InetAddress {
    public static "java/net/Inet6Address/INADDRSZ": number;
    public "java/net/Inet6Address/cached_scope_id": number;
    public "java/net/Inet6Address/holder6": java_net_Inet6Address$Inet6AddressHolder;
    public static "java/net/Inet6Address/serialVersionUID": Long;
    public static "java/net/Inet6Address/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/net/Inet6Address/FIELDS_OFFSET": Long;
    public static "java/net/Inet6Address/UNSAFE": sun_misc_Unsafe;
    public static "java/net/Inet6Address/INT16SZ": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;[BI)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address/<init>(Ljava/lang/String;[BI)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address/<init>(Ljava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;[BLjava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address/<init>(Ljava/lang/String;[BLjava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;[BLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address/<init>(Ljava/lang/String;[BLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getByAddress(Ljava/lang/String;[BLjava/net/NetworkInterface;)Ljava/net/Inet6Address;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: java_net_Inet6Address) => void): void;
    public static "java/net/Inet6Address/getByAddress(Ljava/lang/String;[BLjava/net/NetworkInterface;)Ljava/net/Inet6Address;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: java_net_Inet6Address) => void): void;
    public static "getByAddress(Ljava/lang/String;[BI)Ljava/net/Inet6Address;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_net_Inet6Address) => void): void;
    public static "java/net/Inet6Address/getByAddress(Ljava/lang/String;[BI)Ljava/net/Inet6Address;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: java_net_Inet6Address) => void): void;
    public "initstr(Ljava/lang/String;[BLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address/initstr(Ljava/lang/String;[BLjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "initif(Ljava/lang/String;[BLjava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address/initif(Ljava/lang/String;[BLjava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isDifferentLocalAddressType([B[B)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/Inet6Address/isDifferentLocalAddressType([B[B)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "deriveNumericScope([BLjava/net/NetworkInterface;)I"(thread: JVMThread, args: [JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/Inet6Address/deriveNumericScope([BLjava/net/NetworkInterface;)I"(thread: JVMThread, args: [JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "deriveNumericScope(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/deriveNumericScope(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "isMulticastAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isMulticastAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAnyLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isAnyLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLoopbackAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isLoopbackAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLinkLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isLinkLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLinkLocalAddress([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/Inet6Address/isLinkLocalAddress([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSiteLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isSiteLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isSiteLocalAddress([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/Inet6Address/isSiteLocalAddress([B)Z"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCGlobal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isMCGlobal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCNodeLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isMCNodeLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCLinkLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isMCLinkLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCSiteLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isMCSiteLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCOrgLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isMCOrgLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAddress()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/net/Inet6Address/getAddress()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getScopeId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/getScopeId()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getScopedInterface()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public "java/net/Inet6Address/getScopedInterface()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public "getHostAddress()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/Inet6Address/getHostAddress()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isIPv4CompatibleAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address/isIPv4CompatibleAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "numericToTextFormat([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/Inet6Address/numericToTextFormat([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/Inet6Address/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000([BLjava/net/NetworkInterface;)I"(thread: JVMThread, args: [JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/Inet6Address/access$000([BLjava/net/NetworkInterface;)I"(thread: JVMThread, args: [JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/Inet6Address/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_Inet6Address$Inet6AddressHolder extends java_lang_Object {
    public "java/net/Inet6Address$Inet6AddressHolder/ipaddress": JVMArray<number>;
    public "java/net/Inet6Address$Inet6AddressHolder/scope_id": number;
    public "java/net/Inet6Address$Inet6AddressHolder/scope_id_set": number;
    public "java/net/Inet6Address$Inet6AddressHolder/scope_ifname": java_net_NetworkInterface;
    public "java/net/Inet6Address$Inet6AddressHolder/scope_ifname_set": number;
    public "java/net/Inet6Address$Inet6AddressHolder/this$0": java_net_Inet6Address;
    public "<init>(Ljava/net/Inet6Address;)V"(thread: JVMThread, args: [java_net_Inet6Address], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/<init>(Ljava/net/Inet6Address;)V"(thread: JVMThread, args: [java_net_Inet6Address], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/Inet6Address;[BIZLjava/net/NetworkInterface;Z)V"(thread: JVMThread, args: [java_net_Inet6Address, JVMArray<number>, number, number, java_net_NetworkInterface, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/<init>(Ljava/net/Inet6Address;[BIZLjava/net/NetworkInterface;Z)V"(thread: JVMThread, args: [java_net_Inet6Address, JVMArray<number>, number, number, java_net_NetworkInterface, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setAddr([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/setAddr([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "init([BI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/init([BI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "init([BLjava/net/NetworkInterface;)V"(thread: JVMThread, args: [JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/init([BLjava/net/NetworkInterface;)V"(thread: JVMThread, args: [JVMArray<number>, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "getHostAddress()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/getHostAddress()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isIPv4CompatibleAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isIPv4CompatibleAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMulticastAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isMulticastAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAnyLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isAnyLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLoopbackAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isLoopbackAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLinkLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isLinkLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSiteLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isSiteLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCGlobal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isMCGlobal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCNodeLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isMCNodeLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCLinkLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isMCLinkLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCSiteLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isMCSiteLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCOrgLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/isMCOrgLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/net/Inet6Address;Ljava/net/Inet6Address$1;)V"(thread: JVMThread, args: [java_net_Inet6Address, java_net_Inet6Address$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/<init>(Ljava/net/Inet6Address;Ljava/net/Inet6Address$1;)V"(thread: JVMThread, args: [java_net_Inet6Address, java_net_Inet6Address$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/Inet6Address;[BIZLjava/net/NetworkInterface;ZLjava/net/Inet6Address$1;)V"(thread: JVMThread, args: [java_net_Inet6Address, JVMArray<number>, number, number, java_net_NetworkInterface, number, java_net_Inet6Address$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6Address$Inet6AddressHolder/<init>(Ljava/net/Inet6Address;[BIZLjava/net/NetworkInterface;ZLjava/net/Inet6Address$1;)V"(thread: JVMThread, args: [java_net_Inet6Address, JVMArray<number>, number, number, java_net_NetworkInterface, number, java_net_Inet6Address$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_Inet6Address$1 extends java_lang_Object {
  }
  export class java_net_SocketImpl extends java_lang_Object implements java_net_SocketOptions {
    public "java/net/SocketImpl/socket": java_net_Socket;
    public "java/net/SocketImpl/serverSocket": java_net_ServerSocket;
    public "java/net/SocketImpl/fd": java_io_FileDescriptor;
    public "java/net/SocketImpl/address": java_net_InetAddress;
    public "java/net/SocketImpl/port": number;
    public "java/net/SocketImpl/localport": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "create(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/create(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/connect(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/connect(Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/bind(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "listen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/listen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "accept(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/accept(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/net/SocketImpl/getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "java/net/SocketImpl/getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketImpl/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownInput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/shutdownInput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownOutput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/shutdownOutput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFileDescriptor()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "java/net/SocketImpl/getFileDescriptor()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "getInetAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/SocketImpl/getInetAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketImpl/getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "supportsUrgentData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketImpl/supportsUrgentData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/sendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketImpl/getLocalPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setSocket(Ljava/net/Socket;)V"(thread: JVMThread, args: [java_net_Socket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/setSocket(Ljava/net/Socket;)V"(thread: JVMThread, args: [java_net_Socket], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSocket()Ljava/net/Socket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Socket) => void): void;
    public "java/net/SocketImpl/getSocket()Ljava/net/Socket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_Socket) => void): void;
    public "setServerSocket(Ljava/net/ServerSocket;)V"(thread: JVMThread, args: [java_net_ServerSocket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/setServerSocket(Ljava/net/ServerSocket;)V"(thread: JVMThread, args: [java_net_ServerSocket], cb?: (e?: java_lang_Throwable) => void): void;
    public "getServerSocket()Ljava/net/ServerSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ServerSocket) => void): void;
    public "java/net/SocketImpl/getServerSocket()Ljava/net/ServerSocket;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_ServerSocket) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/SocketImpl/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPerformancePreferences(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/setPerformancePreferences(III)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/SocketImpl/getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "setOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "getOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_net_PlainSocketImpl extends java_net_AbstractPlainSocketImpl {
    public $is_shutdown: boolean;
    public $ws: Interfaces.IWebsock;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/<init>(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/PlainSocketImpl/getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "socketSetOption(IZLjava/lang/Object;)V"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketSetOption(IZLjava/lang/Object;)V"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketCreate(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketCreate(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketConnect(Ljava/net/InetAddress;II)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketConnect(Ljava/net/InetAddress;II)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketBind(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketBind(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketListen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketListen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketAccept(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketAccept(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/PlainSocketImpl/socketAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "socketClose0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketClose0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketShutdown(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketShutdown(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initProto()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/PlainSocketImpl/initProto()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketSetOption0(IZLjava/lang/Object;)V"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketSetOption0(IZLjava/lang/Object;)V"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketGetOption(ILjava/lang/Object;)I"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/PlainSocketImpl/socketGetOption(ILjava/lang/Object;)I"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "socketSendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainSocketImpl/socketSendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/PlainSocketImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_AbstractPlainSocketImpl extends java_net_SocketImpl {
    public "java/net/AbstractPlainSocketImpl/timeout": number;
    public "java/net/AbstractPlainSocketImpl/trafficClass": number;
    public "java/net/AbstractPlainSocketImpl/shut_rd": number;
    public "java/net/AbstractPlainSocketImpl/shut_wr": number;
    public "java/net/AbstractPlainSocketImpl/socketInputStream": java_net_SocketInputStream;
    public "java/net/AbstractPlainSocketImpl/socketOutputStream": java_net_SocketOutputStream;
    public "java/net/AbstractPlainSocketImpl/fdUseCount": number;
    public "java/net/AbstractPlainSocketImpl/fdLock": java_lang_Object;
    public "java/net/AbstractPlainSocketImpl/closePending": number;
    public "java/net/AbstractPlainSocketImpl/CONNECTION_NOT_RESET": number;
    public "java/net/AbstractPlainSocketImpl/CONNECTION_RESET_PENDING": number;
    public "java/net/AbstractPlainSocketImpl/CONNECTION_RESET": number;
    public "java/net/AbstractPlainSocketImpl/resetState": number;
    public "java/net/AbstractPlainSocketImpl/resetLock": java_lang_Object;
    public "java/net/AbstractPlainSocketImpl/stream": number;
    public static "java/net/AbstractPlainSocketImpl/SHUT_RD": number;
    public static "java/net/AbstractPlainSocketImpl/SHUT_WR": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "create(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/create(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/connect(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/connect(Ljava/net/SocketAddress;I)V"(thread: JVMThread, args: [java_net_SocketAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "connectToAddress(Ljava/net/InetAddress;II)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/connectToAddress(Ljava/net/InetAddress;II)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/setOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/AbstractPlainSocketImpl/getOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "doConnect(Ljava/net/InetAddress;II)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/doConnect(Ljava/net/InetAddress;II)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/bind(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "listen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/listen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "accept(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/accept(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/net/AbstractPlainSocketImpl/getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "setInputStream(Ljava/net/SocketInputStream;)V"(thread: JVMThread, args: [java_net_SocketInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/setInputStream(Ljava/net/SocketInputStream;)V"(thread: JVMThread, args: [java_net_SocketInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "java/net/AbstractPlainSocketImpl/getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "setFileDescriptor(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/setFileDescriptor(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "setAddress(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/setAddress(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/setPort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLocalPort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/setLocalPort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainSocketImpl/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownInput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/shutdownInput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "shutdownOutput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/shutdownOutput()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "supportsUrgentData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainSocketImpl/supportsUrgentData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/sendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "acquireFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "java/net/AbstractPlainSocketImpl/acquireFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "releaseFD()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/releaseFD()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isConnectionReset()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainSocketImpl/isConnectionReset()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConnectionResetPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainSocketImpl/isConnectionResetPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setConnectionReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/setConnectionReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setConnectionResetPending()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/setConnectionResetPending()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isClosedOrPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainSocketImpl/isClosedOrPending()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainSocketImpl/getTimeout()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "socketPreClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketPreClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketCreate(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketCreate(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketConnect(Ljava/net/InetAddress;II)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketConnect(Ljava/net/InetAddress;II)V"(thread: JVMThread, args: [java_net_InetAddress, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketBind(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketBind(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketListen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketListen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketAccept(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketAccept(Ljava/net/SocketImpl;)V"(thread: JVMThread, args: [java_net_SocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "socketClose0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketClose0(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketShutdown(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketShutdown(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketSetOption(IZLjava/lang/Object;)V"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketSetOption(IZLjava/lang/Object;)V"(thread: JVMThread, args: [number, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketGetOption(ILjava/lang/Object;)I"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketGetOption(ILjava/lang/Object;)I"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "socketSendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainSocketImpl/socketSendUrgentData(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/AbstractPlainSocketImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_DatagramPacket extends java_lang_Object {
    public "java/net/DatagramPacket/buf": JVMArray<number>;
    public "java/net/DatagramPacket/offset": number;
    public "java/net/DatagramPacket/length": number;
    public "java/net/DatagramPacket/bufLength": number;
    public "java/net/DatagramPacket/address": java_net_InetAddress;
    public "java/net/DatagramPacket/port": number;
    public "<init>([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/<init>([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/<init>([BI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BIILjava/net/InetAddress;I)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/<init>([BIILjava/net/InetAddress;I)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BIILjava/net/SocketAddress;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/<init>([BIILjava/net/SocketAddress;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BILjava/net/InetAddress;I)V"(thread: JVMThread, args: [JVMArray<number>, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/<init>([BILjava/net/InetAddress;I)V"(thread: JVMThread, args: [JVMArray<number>, number, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BILjava/net/SocketAddress;)V"(thread: JVMThread, args: [JVMArray<number>, number, java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/<init>([BILjava/net/SocketAddress;)V"(thread: JVMThread, args: [JVMArray<number>, number, java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/DatagramPacket/getAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramPacket/getPort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getData()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/net/DatagramPacket/getData()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramPacket/getOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/DatagramPacket/getLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setData([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/setData([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setAddress(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/setAddress(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/setPort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setSocketAddress(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/setSocketAddress(Ljava/net/SocketAddress;)V"(thread: JVMThread, args: [java_net_SocketAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "java/net/DatagramPacket/getSocketAddress()Ljava/net/SocketAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_SocketAddress) => void): void;
    public "setData([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/setData([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLength(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/DatagramPacket/setLength(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/DatagramPacket/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/DatagramPacket/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_PlainDatagramSocketImpl extends java_net_AbstractPlainDatagramSocketImpl {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOption(Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/setOption(Ljava/net/SocketOption;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_net_SocketOption, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/PlainDatagramSocketImpl/getOption(Ljava/net/SocketOption;)Ljava/lang/Object;"(thread: JVMThread, args: [java_net_SocketOption], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "socketSetOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/socketSetOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind0(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/bind0(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "send(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/send(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "peek(Ljava/net/InetAddress;)I"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/PlainDatagramSocketImpl/peek(Ljava/net/InetAddress;)I"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peekData(Ljava/net/DatagramPacket;)I"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/PlainDatagramSocketImpl/peekData(Ljava/net/DatagramPacket;)I"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "receive0(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/receive0(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTimeToLive(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/setTimeToLive(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTimeToLive()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/PlainDatagramSocketImpl/getTimeToLive()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setTTL(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/setTTL(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTTL()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/PlainDatagramSocketImpl/getTTL()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "leave(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/leave(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "datagramSocketCreate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/datagramSocketCreate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "datagramSocketClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/datagramSocketClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketSetOption0(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/socketSetOption0(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketGetOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/PlainDatagramSocketImpl/socketGetOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "connect0(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/connect0(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "disconnect0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/PlainDatagramSocketImpl/disconnect0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "dataAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/PlainDatagramSocketImpl/dataAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/PlainDatagramSocketImpl/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/PlainDatagramSocketImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_AbstractPlainDatagramSocketImpl extends java_net_DatagramSocketImpl {
    public "java/net/AbstractPlainDatagramSocketImpl/timeout": number;
    public "java/net/AbstractPlainDatagramSocketImpl/connected": number;
    public "java/net/AbstractPlainDatagramSocketImpl/trafficClass": number;
    public "java/net/AbstractPlainDatagramSocketImpl/connectedAddress": java_net_InetAddress;
    public "java/net/AbstractPlainDatagramSocketImpl/connectedPort": number;
    public static "java/net/AbstractPlainDatagramSocketImpl/os": java_lang_String;
    public static "java/net/AbstractPlainDatagramSocketImpl/connectDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "create()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/create()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/bind(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "bind0(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/bind0(ILjava/net/InetAddress;)V"(thread: JVMThread, args: [number, java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "send(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/send(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/connect(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "disconnect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/disconnect()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "peek(Ljava/net/InetAddress;)I"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/peek(Ljava/net/InetAddress;)I"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peekData(Ljava/net/DatagramPacket;)I"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/peekData(Ljava/net/DatagramPacket;)I"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "receive(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/receive(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "receive0(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/receive0(Ljava/net/DatagramPacket;)V"(thread: JVMThread, args: [java_net_DatagramPacket], cb?: (e?: java_lang_Throwable) => void): void;
    public "setTimeToLive(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/setTimeToLive(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTimeToLive()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/getTimeToLive()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setTTL(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/setTTL(B)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTTL()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/getTTL()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "join(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/join(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "leave(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/leave(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "joinGroup(Ljava/net/SocketAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/joinGroup(Ljava/net/SocketAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/join(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "leaveGroup(Ljava/net/SocketAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/leaveGroup(Ljava/net/SocketAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_SocketAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "leave(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/leave(Ljava/net/InetAddress;Ljava/net/NetworkInterface;)V"(thread: JVMThread, args: [java_net_InetAddress, java_net_NetworkInterface], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isClosed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/isClosed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/setOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/getOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "datagramSocketCreate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/datagramSocketCreate()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "datagramSocketClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/datagramSocketClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketSetOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/socketSetOption(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketGetOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/socketGetOption(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "connect0(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/connect0(Ljava/net/InetAddress;I)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "disconnect0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/disconnect0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "nativeConnectDisabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/nativeConnectDisabled()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "dataAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/AbstractPlainDatagramSocketImpl/dataAvailable()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/AbstractPlainDatagramSocketImpl/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_NetworkInterface extends java_lang_Object {
    public "java/net/NetworkInterface/name": java_lang_String;
    public "java/net/NetworkInterface/displayName": java_lang_String;
    public "java/net/NetworkInterface/index": number;
    public "java/net/NetworkInterface/addrs": JVMArray<java_net_InetAddress>;
    public "java/net/NetworkInterface/bindings": JVMArray<java_net_InterfaceAddress>;
    public "java/net/NetworkInterface/childs": JVMArray<java_net_NetworkInterface>;
    public "java/net/NetworkInterface/parent": java_net_NetworkInterface;
    public "java/net/NetworkInterface/virtual": number;
    public static "java/net/NetworkInterface/defaultInterface": java_net_NetworkInterface;
    public static "java/net/NetworkInterface/defaultIndex": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/NetworkInterface/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I[Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_net_InetAddress>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/NetworkInterface/<init>(Ljava/lang/String;I[Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_net_InetAddress>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/NetworkInterface/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getInetAddresses()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/net/NetworkInterface/getInetAddresses()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getInterfaceAddresses()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/net/NetworkInterface/getInterfaceAddresses()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getSubInterfaces()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/net/NetworkInterface/getSubInterfaces()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "getParent()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public "java/net/NetworkInterface/getParent()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public "getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/NetworkInterface/getIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDisplayName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/NetworkInterface/getDisplayName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getByName(Ljava/lang/String;)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "java/net/NetworkInterface/getByName(Ljava/lang/String;)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "getByIndex(I)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "java/net/NetworkInterface/getByIndex(I)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "getByInetAddress(Ljava/net/InetAddress;)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "java/net/NetworkInterface/getByInetAddress(Ljava/net/InetAddress;)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "getNetworkInterfaces()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "java/net/NetworkInterface/getNetworkInterfaces()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "getAll()[Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_NetworkInterface>) => void): void;
    public static "java/net/NetworkInterface/getAll()[Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_NetworkInterface>) => void): void;
    public static "getByName0(Ljava/lang/String;)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "java/net/NetworkInterface/getByName0(Ljava/lang/String;)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "getByIndex0(I)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "java/net/NetworkInterface/getByIndex0(I)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "getByInetAddress0(Ljava/net/InetAddress;)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "java/net/NetworkInterface/getByInetAddress0(Ljava/net/InetAddress;)Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public "isUp()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/NetworkInterface/isUp()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLoopback()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/NetworkInterface/isLoopback()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPointToPoint()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/NetworkInterface/isPointToPoint()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "supportsMulticast()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/NetworkInterface/supportsMulticast()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHardwareAddress()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/net/NetworkInterface/getHardwareAddress()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getMTU()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/NetworkInterface/getMTU()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isVirtual()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/NetworkInterface/isVirtual()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isUp0(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/NetworkInterface/isUp0(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isLoopback0(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/NetworkInterface/isLoopback0(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "supportsMulticast0(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/NetworkInterface/supportsMulticast0(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isP2P0(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/NetworkInterface/isP2P0(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getMacAddr0([BLjava/lang/String;I)[B"(thread: JVMThread, args: [JVMArray<number>, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/net/NetworkInterface/getMacAddr0([BLjava/lang/String;I)[B"(thread: JVMThread, args: [JVMArray<number>, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getMTU0(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/NetworkInterface/getMTU0(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/NetworkInterface/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/NetworkInterface/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/NetworkInterface/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/NetworkInterface/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getDefault()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "java/net/NetworkInterface/getDefault()Ljava/net/NetworkInterface;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_NetworkInterface) => void): void;
    public static "access$000(Ljava/net/NetworkInterface;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "java/net/NetworkInterface/access$000(Ljava/net/NetworkInterface;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "access$100(Ljava/net/NetworkInterface;)[Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_NetworkInterface>) => void): void;
    public static "java/net/NetworkInterface/access$100(Ljava/net/NetworkInterface;)[Ljava/net/NetworkInterface;"(thread: JVMThread, args: [java_net_NetworkInterface], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_NetworkInterface>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/NetworkInterface/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_InterfaceAddress extends java_lang_Object {
    public "java/net/InterfaceAddress/address": java_net_InetAddress;
    public "java/net/InterfaceAddress/broadcast": java_net_Inet4Address;
    public "java/net/InterfaceAddress/maskLength": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InterfaceAddress/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/InterfaceAddress/getAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getBroadcast()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/InterfaceAddress/getBroadcast()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "getNetworkPrefixLength()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InterfaceAddress/getNetworkPrefixLength()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InterfaceAddress/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InterfaceAddress/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InterfaceAddress/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_net_SocketOutputStream extends java_io_FileOutputStream {
    public "java/net/SocketOutputStream/impl": java_net_AbstractPlainSocketImpl;
    public "java/net/SocketOutputStream/temp": JVMArray<number>;
    public "java/net/SocketOutputStream/socket": java_net_Socket;
    public "java/net/SocketOutputStream/closing": number;
    public "<init>(Ljava/net/AbstractPlainSocketImpl;)V"(thread: JVMThread, args: [java_net_AbstractPlainSocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketOutputStream/<init>(Ljava/net/AbstractPlainSocketImpl;)V"(thread: JVMThread, args: [java_net_AbstractPlainSocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "java/net/SocketOutputStream/getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "socketWrite0(Ljava/io/FileDescriptor;[BII)V"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketOutputStream/socketWrite0(Ljava/io/FileDescriptor;[BII)V"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "socketWrite([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketOutputStream/socketWrite([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketOutputStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketOutputStream/write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketOutputStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketOutputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketOutputStream/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/SocketOutputStream/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/SocketOutputStream/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_SocketInputStream extends java_io_FileInputStream {
    public "java/net/SocketInputStream/eof": number;
    public "java/net/SocketInputStream/impl": java_net_AbstractPlainSocketImpl;
    public "java/net/SocketInputStream/temp": JVMArray<number>;
    public "java/net/SocketInputStream/socket": java_net_Socket;
    public "java/net/SocketInputStream/closing": number;
    public "<init>(Ljava/net/AbstractPlainSocketImpl;)V"(thread: JVMThread, args: [java_net_AbstractPlainSocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketInputStream/<init>(Ljava/net/AbstractPlainSocketImpl;)V"(thread: JVMThread, args: [java_net_AbstractPlainSocketImpl], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "java/net/SocketInputStream/getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "socketRead0(Ljava/io/FileDescriptor;[BIII)I"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketInputStream/socketRead0(Ljava/io/FileDescriptor;[BIII)I"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "socketRead(Ljava/io/FileDescriptor;[BIII)I"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketInputStream/socketRead(Ljava/io/FileDescriptor;[BIII)I"(thread: JVMThread, args: [java_io_FileDescriptor, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketInputStream/read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketInputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketInputStream/read([BIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketInputStream/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/net/SocketInputStream/skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/SocketInputStream/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketInputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setEOF(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketInputStream/setEOF(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/SocketInputStream/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/SocketInputStream/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/SocketInputStream/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_InetAddress extends java_lang_Object implements java_io_Serializable {
    public static "java/net/InetAddress/IPv4": number;
    public static "java/net/InetAddress/IPv6": number;
    public static "java/net/InetAddress/preferIPv6Address": number;
    public "java/net/InetAddress/holder": java_net_InetAddress$InetAddressHolder;
    public static "java/net/InetAddress/nameServices": java_util_List;
    public "java/net/InetAddress/canonicalHostName": java_lang_String;
    public static "java/net/InetAddress/serialVersionUID": Long;
    public static "java/net/InetAddress/addressCache": java_net_InetAddress$Cache;
    public static "java/net/InetAddress/negativeCache": java_net_InetAddress$Cache;
    public static "java/net/InetAddress/addressCacheInit": number;
    public static "java/net/InetAddress/unknown_array": JVMArray<java_net_InetAddress>;
    public static "java/net/InetAddress/impl": java_net_InetAddressImpl;
    public static "java/net/InetAddress/lookupTable": java_util_HashMap;
    public static "java/net/InetAddress/cachedLocalHost": java_net_InetAddress;
    public static "java/net/InetAddress/cacheTime": Long;
    public static "java/net/InetAddress/maxCacheTime": Long;
    public static "java/net/InetAddress/cacheLock": java_lang_Object;
    public static "java/net/InetAddress/FIELDS_OFFSET": Long;
    public static "java/net/InetAddress/UNSAFE": sun_misc_Unsafe;
    public static "java/net/InetAddress/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/net/InetAddress/$assertionsDisabled": number;
    public "holder()Ljava/net/InetAddress$InetAddressHolder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress$InetAddressHolder) => void): void;
    public "java/net/InetAddress/holder()Ljava/net/InetAddress$InetAddressHolder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress$InetAddressHolder) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/net/InetAddress/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isMulticastAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isMulticastAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAnyLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isAnyLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLoopbackAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isLoopbackAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLinkLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isLinkLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSiteLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isSiteLocalAddress()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCGlobal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isMCGlobal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCNodeLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isMCNodeLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCLinkLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isMCLinkLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCSiteLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isMCSiteLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMCOrgLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isMCOrgLocal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isReachable(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isReachable(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isReachable(Ljava/net/NetworkInterface;II)Z"(thread: JVMThread, args: [java_net_NetworkInterface, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/isReachable(Ljava/net/NetworkInterface;II)Z"(thread: JVMThread, args: [java_net_NetworkInterface, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetAddress/getHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getHostName(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetAddress/getHostName(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCanonicalHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetAddress/getCanonicalHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getHostFromNameService(Ljava/net/InetAddress;Z)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/net/InetAddress/getHostFromNameService(Ljava/net/InetAddress;Z)Ljava/lang/String;"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAddress()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/net/InetAddress/getAddress()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getHostAddress()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetAddress/getHostAddress()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetAddress/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "cacheInitIfNeeded()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/InetAddress/cacheInitIfNeeded()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "cacheAddresses(Ljava/lang/String;[Ljava/net/InetAddress;Z)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_net_InetAddress>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/InetAddress/cacheAddresses(Ljava/lang/String;[Ljava/net/InetAddress;Z)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_net_InetAddress>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getCachedAddresses(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "java/net/InetAddress/getCachedAddresses(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "createNSProvider(Ljava/lang/String;)Lsun/net/spi/nameservice/NameService;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_net_spi_nameservice_NameService) => void): void;
    public static "java/net/InetAddress/createNSProvider(Ljava/lang/String;)Lsun/net/spi/nameservice/NameService;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_net_spi_nameservice_NameService) => void): void;
    public static "getByAddress(Ljava/lang/String;[B)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "java/net/InetAddress/getByAddress(Ljava/lang/String;[B)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "getByName(Ljava/lang/String;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "java/net/InetAddress/getByName(Ljava/lang/String;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "getByName(Ljava/lang/String;Ljava/net/InetAddress;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "java/net/InetAddress/getByName(Ljava/lang/String;Ljava/net/InetAddress;)Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "getAllByName(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "java/net/InetAddress/getAllByName(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "getAllByName(Ljava/lang/String;Ljava/net/InetAddress;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "java/net/InetAddress/getAllByName(Ljava/lang/String;Ljava/net/InetAddress;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "getLoopbackAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "java/net/InetAddress/getLoopbackAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "checkNumericZone(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/net/InetAddress/checkNumericZone(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getAllByName0(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "java/net/InetAddress/getAllByName0(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "getAllByName0(Ljava/lang/String;Z)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "java/net/InetAddress/getAllByName0(Ljava/lang/String;Z)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "getAllByName0(Ljava/lang/String;Ljava/net/InetAddress;Z)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "java/net/InetAddress/getAllByName0(Ljava/lang/String;Ljava/net/InetAddress;Z)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "getAddressesFromNameService(Ljava/lang/String;Ljava/net/InetAddress;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "java/net/InetAddress/getAddressesFromNameService(Ljava/lang/String;Ljava/net/InetAddress;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String, java_net_InetAddress], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "checkLookupTable(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "java/net/InetAddress/checkLookupTable(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public static "updateLookupTable(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/InetAddress/updateLookupTable(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getByAddress([B)Ljava/net/InetAddress;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "java/net/InetAddress/getByAddress([B)Ljava/net/InetAddress;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "getLocalHost()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "java/net/InetAddress/getLocalHost()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/InetAddress/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "anyLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "java/net/InetAddress/anyLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public static "loadImpl(Ljava/lang/String;)Ljava/net/InetAddressImpl;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddressImpl) => void): void;
    public static "java/net/InetAddress/loadImpl(Ljava/lang/String;)Ljava/net/InetAddressImpl;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddressImpl) => void): void;
    public "readObjectNoData(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress/readObjectNoData(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/InetAddress/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_net_spi_nameservice_NameService extends java_lang_Object {
    "lookupAllHostAddr(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    "getHostByAddr([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_net_InetAddress$Cache extends java_lang_Object {
    public "java/net/InetAddress$Cache/cache": java_util_LinkedHashMap;
    public "java/net/InetAddress$Cache/type": java_net_InetAddress$Cache$Type;
    public "<init>(Ljava/net/InetAddress$Cache$Type;)V"(thread: JVMThread, args: [java_net_InetAddress$Cache$Type], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress$Cache/<init>(Ljava/net/InetAddress$Cache$Type;)V"(thread: JVMThread, args: [java_net_InetAddress$Cache$Type], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPolicy()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress$Cache/getPolicy()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(Ljava/lang/String;[Ljava/net/InetAddress;)Ljava/net/InetAddress$Cache;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_net_InetAddress>], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress$Cache) => void): void;
    public "java/net/InetAddress$Cache/put(Ljava/lang/String;[Ljava/net/InetAddress;)Ljava/net/InetAddress$Cache;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_net_InetAddress>], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress$Cache) => void): void;
    public "get(Ljava/lang/String;)Ljava/net/InetAddress$CacheEntry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress$CacheEntry) => void): void;
    public "java/net/InetAddress$Cache/get(Ljava/lang/String;)Ljava/net/InetAddress$CacheEntry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress$CacheEntry) => void): void;
  }
  export class java_net_InetAddress$CacheEntry extends java_lang_Object {
    public "java/net/InetAddress$CacheEntry/addresses": JVMArray<java_net_InetAddress>;
    public "java/net/InetAddress$CacheEntry/expiration": Long;
    public "<init>([Ljava/net/InetAddress;J)V"(thread: JVMThread, args: [JVMArray<java_net_InetAddress>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress$CacheEntry/<init>([Ljava/net/InetAddress;J)V"(thread: JVMThread, args: [JVMArray<java_net_InetAddress>, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_net_InetAddress$Cache$Type extends java_lang_Enum {
    public static "java/net/InetAddress$Cache$Type/Positive": java_net_InetAddress$Cache$Type;
    public static "java/net/InetAddress$Cache$Type/Negative": java_net_InetAddress$Cache$Type;
    public static "java/net/InetAddress$Cache$Type/$VALUES": JVMArray<java_net_InetAddress$Cache$Type>;
    public static "values()[Ljava/net/InetAddress$Cache$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress$Cache$Type>) => void): void;
    public static "java/net/InetAddress$Cache$Type/values()[Ljava/net/InetAddress$Cache$Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress$Cache$Type>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/net/InetAddress$Cache$Type;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress$Cache$Type) => void): void;
    public static "java/net/InetAddress$Cache$Type/valueOf(Ljava/lang/String;)Ljava/net/InetAddress$Cache$Type;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress$Cache$Type) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress$Cache$Type/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/net/InetAddress$Cache$Type/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_LinkedHashMap extends java_util_HashMap implements java_util_Map {
    public static "java/util/LinkedHashMap/serialVersionUID": Long;
    public "java/util/LinkedHashMap/head": java_util_LinkedHashMap$Entry;
    public "java/util/LinkedHashMap/tail": java_util_LinkedHashMap$Entry;
    public "java/util/LinkedHashMap/accessOrder": number;
    public "linkNodeLast(Ljava/util/LinkedHashMap$Entry;)V"(thread: JVMThread, args: [java_util_LinkedHashMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/linkNodeLast(Ljava/util/LinkedHashMap$Entry;)V"(thread: JVMThread, args: [java_util_LinkedHashMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "transferLinks(Ljava/util/LinkedHashMap$Entry;Ljava/util/LinkedHashMap$Entry;)V"(thread: JVMThread, args: [java_util_LinkedHashMap$Entry, java_util_LinkedHashMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/transferLinks(Ljava/util/LinkedHashMap$Entry;Ljava/util/LinkedHashMap$Entry;)V"(thread: JVMThread, args: [java_util_LinkedHashMap$Entry, java_util_LinkedHashMap$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "reinitialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/reinitialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "newNode(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "java/util/LinkedHashMap/newNode(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "replacementNode(Ljava/util/HashMap$Node;Ljava/util/HashMap$Node;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [java_util_HashMap$Node, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "java/util/LinkedHashMap/replacementNode(Ljava/util/HashMap$Node;Ljava/util/HashMap$Node;)Ljava/util/HashMap$Node;"(thread: JVMThread, args: [java_util_HashMap$Node, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$Node) => void): void;
    public "newTreeNode(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "java/util/LinkedHashMap/newTreeNode(ILjava/lang/Object;Ljava/lang/Object;Ljava/util/HashMap$Node;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [number, java_lang_Object, java_lang_Object, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "replacementTreeNode(Ljava/util/HashMap$Node;Ljava/util/HashMap$Node;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$Node, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "java/util/LinkedHashMap/replacementTreeNode(Ljava/util/HashMap$Node;Ljava/util/HashMap$Node;)Ljava/util/HashMap$TreeNode;"(thread: JVMThread, args: [java_util_HashMap$Node, java_util_HashMap$Node], cb?: (e?: java_lang_Throwable, rv?: java_util_HashMap$TreeNode) => void): void;
    public "afterNodeRemoval(Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/afterNodeRemoval(Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "afterNodeInsertion(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/afterNodeInsertion(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "afterNodeAccess(Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/afterNodeAccess(Ljava/util/HashMap$Node;)V"(thread: JVMThread, args: [java_util_HashMap$Node], cb?: (e?: java_lang_Throwable) => void): void;
    public "internalWriteEntries(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/internalWriteEntries(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/<init>(Ljava/util/Map;)V"(thread: JVMThread, args: [java_util_Map], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IFZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/<init>(IFZ)V"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedHashMap/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedHashMap/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/LinkedHashMap/getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeEldestEntry(Ljava/util/Map$Entry;)Z"(thread: JVMThread, args: [java_util_Map$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/LinkedHashMap/removeEldestEntry(Ljava/util/Map$Entry;)Z"(thread: JVMThread, args: [java_util_Map$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/LinkedHashMap/keySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "java/util/LinkedHashMap/values()Ljava/util/Collection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Collection) => void): void;
    public "entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/LinkedHashMap/entrySet()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/LinkedHashMap/replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "getOrDefault(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "forEach(Ljava/util/function/BiConsumer;)V"(thread: JVMThread, args: [java_util_function_BiConsumer], cb?: (e?: java_lang_Throwable) => void): void;
    "replaceAll(Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    "putIfAbsent(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "remove(Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "replace(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfAbsent(Ljava/lang/Object;Ljava/util/function/Function;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_Function], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "computeIfPresent(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "compute(Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "merge(Ljava/lang/Object;Ljava/lang/Object;Ljava/util/function/BiFunction;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_net_InetAddress$InetAddressHolder extends java_lang_Object {
    public "java/net/InetAddress$InetAddressHolder/originalHostName": java_lang_String;
    public "java/net/InetAddress$InetAddressHolder/hostName": java_lang_String;
    public "java/net/InetAddress$InetAddressHolder/address": number;
    public "java/net/InetAddress$InetAddressHolder/family": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress$InetAddressHolder/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress$InetAddressHolder/<init>(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "init(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/InetAddress$InetAddressHolder/init(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetAddress$InetAddressHolder/getHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getOriginalHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/InetAddress$InetAddressHolder/getOriginalHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAddress()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress$InetAddressHolder/getAddress()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFamily()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/InetAddress$InetAddressHolder/getFamily()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_net_Inet6AddressImpl extends java_lang_Object implements java_net_InetAddressImpl {
    public "java/net/Inet6AddressImpl/anyLocalAddress": java_net_InetAddress;
    public "java/net/Inet6AddressImpl/loopbackAddress": java_net_InetAddress;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/net/Inet6AddressImpl/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocalHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/Inet6AddressImpl/getLocalHostName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "lookupAllHostAddr(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public "java/net/Inet6AddressImpl/lookupAllHostAddr(Ljava/lang/String;)[Ljava/net/InetAddress;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_net_InetAddress>) => void): void;
    public "getHostByAddr([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/net/Inet6AddressImpl/getHostByAddr([B)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isReachable0([BII[BII)Z"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6AddressImpl/isReachable0([BII[BII)Z"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isReachable(Ljava/net/InetAddress;ILjava/net/NetworkInterface;I)Z"(thread: JVMThread, args: [java_net_InetAddress, number, java_net_NetworkInterface, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/net/Inet6AddressImpl/isReachable(Ljava/net/InetAddress;ILjava/net/NetworkInterface;I)Z"(thread: JVMThread, args: [java_net_InetAddress, number, java_net_NetworkInterface, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "anyLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/Inet6AddressImpl/anyLocalAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "loopbackAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
    public "java/net/Inet6AddressImpl/loopbackAddress()Ljava/net/InetAddress;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_InetAddress) => void): void;
  }
  export class java_lang_invoke_MethodHandle extends java_lang_Object {
    public "java/lang/invoke/MethodHandle/type": java_lang_invoke_MethodType;
    public "java/lang/invoke/MethodHandle/form": java_lang_invoke_LambdaForm;
    public "java/lang/invoke/MethodHandle/asTypeCache": java_lang_invoke_MethodHandle;
    public "java/lang/invoke/MethodHandle/customizationCount": number;
    public static "java/lang/invoke/MethodHandle/FORM_OFFSET": Long;
    public static "java/lang/invoke/MethodHandle/$assertionsDisabled": number;
    public "type()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodHandle/type()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandle/<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "invokeExact([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodHandle/invokeExact([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invoke([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodHandle/invoke([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invokeBasic([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodHandle/invokeBasic([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "linkToVirtual([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/MethodHandle/linkToVirtual([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "linkToStatic([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/MethodHandle/linkToStatic([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "linkToSpecial([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/MethodHandle/linkToSpecial([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "linkToInterface([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/MethodHandle/linkToInterface([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invokeWithArguments([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodHandle/invokeWithArguments([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invokeWithArguments(Ljava/util/List;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodHandle/invokeWithArguments(Ljava/util/List;)Ljava/lang/Object;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "asType(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/asType(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "asTypeCached(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/asTypeCached(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "asTypeUncached(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/asTypeUncached(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "asSpreader(Ljava/lang/Class;I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/asSpreader(Ljava/lang/Class;I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "asSpreaderChecks(Ljava/lang/Class;I)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodHandle/asSpreaderChecks(Ljava/lang/Class;I)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "spreadArrayChecks(Ljava/lang/Class;I)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandle/spreadArrayChecks(Ljava/lang/Class;I)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "asCollector(Ljava/lang/Class;I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/asCollector(Ljava/lang/Class;I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "asCollectorChecks(Ljava/lang/Class;I)Z"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodHandle/asCollectorChecks(Ljava/lang/Class;I)Z"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "asVarargsCollector(Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/asVarargsCollector(Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "isVarargsCollector()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodHandle/isVarargsCollector()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "asFixedArity()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/asFixedArity()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "bindTo(Ljava/lang/Object;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/bindTo(Ljava/lang/Object;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MethodHandle/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "standardString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MethodHandle/standardString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "debugString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MethodHandle/debugString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "bindArgumentL(ILjava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/bindArgumentL(ILjava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "setVarargs(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/setVarargs(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "viewAsType(Ljava/lang/invoke/MethodType;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/viewAsType(Ljava/lang/invoke/MethodType;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "viewAsTypeChecks(Ljava/lang/invoke/MethodType;Z)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodHandle/viewAsTypeChecks(Ljava/lang/invoke/MethodType;Z)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "internalForm()Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/MethodHandle/internalForm()Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "internalMemberName()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MethodHandle/internalMemberName()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "internalCallerClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MethodHandle/internalCallerClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "intrinsicName()Ljava/lang/invoke/MethodHandleImpl$Intrinsic;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandleImpl$Intrinsic) => void): void;
    public "java/lang/invoke/MethodHandle/intrinsicName()Ljava/lang/invoke/MethodHandleImpl$Intrinsic;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandleImpl$Intrinsic) => void): void;
    public "withInternalMemberName(Ljava/lang/invoke/MemberName;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/withInternalMemberName(Ljava/lang/invoke/MemberName;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "isInvokeSpecial()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodHandle/isInvokeSpecial()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "internalValues()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodHandle/internalValues()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "internalProperties()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodHandle/internalProperties()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "copyWith(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/copyWith(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "rebind()Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/MethodHandle/rebind()Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "updateForm(Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandle/updateForm(Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "customize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandle/customize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_MethodHandleImpl$Intrinsic extends java_lang_Enum {
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/SELECT_ALTERNATIVE": java_lang_invoke_MethodHandleImpl$Intrinsic;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/GUARD_WITH_CATCH": java_lang_invoke_MethodHandleImpl$Intrinsic;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/NEW_ARRAY": java_lang_invoke_MethodHandleImpl$Intrinsic;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/ARRAY_LOAD": java_lang_invoke_MethodHandleImpl$Intrinsic;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/ARRAY_STORE": java_lang_invoke_MethodHandleImpl$Intrinsic;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/IDENTITY": java_lang_invoke_MethodHandleImpl$Intrinsic;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/ZERO": java_lang_invoke_MethodHandleImpl$Intrinsic;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/NONE": java_lang_invoke_MethodHandleImpl$Intrinsic;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/$VALUES": JVMArray<java_lang_invoke_MethodHandleImpl$Intrinsic>;
    public static "values()[Ljava/lang/invoke/MethodHandleImpl$Intrinsic;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_MethodHandleImpl$Intrinsic>) => void): void;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/values()[Ljava/lang/invoke/MethodHandleImpl$Intrinsic;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_MethodHandleImpl$Intrinsic>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/invoke/MethodHandleImpl$Intrinsic;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandleImpl$Intrinsic) => void): void;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/valueOf(Ljava/lang/String;)Ljava/lang/invoke/MethodHandleImpl$Intrinsic;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandleImpl$Intrinsic) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandleImpl$Intrinsic/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandleImpl$Intrinsic/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_BoundMethodHandle extends java_lang_invoke_MethodHandle {
    public static "java/lang/invoke/BoundMethodHandle/FIELD_COUNT_THRESHOLD": number;
    public static "java/lang/invoke/BoundMethodHandle/FORM_EXPRESSION_THRESHOLD": number;
    public static "java/lang/invoke/BoundMethodHandle/LOOKUP": java_lang_invoke_MethodHandles$Lookup;
    public static "java/lang/invoke/BoundMethodHandle/SPECIES_DATA": java_lang_invoke_BoundMethodHandle$SpeciesData;
    public static "java/lang/invoke/BoundMethodHandle/SPECIES_DATA_CACHE": JVMArray<java_lang_invoke_BoundMethodHandle$SpeciesData>;
    public static "java/lang/invoke/BoundMethodHandle/$assertionsDisabled": number;
    public "<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/BoundMethodHandle/<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public static "bindSingle(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/invoke/LambdaForm$BasicType;Ljava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_invoke_LambdaForm$BasicType, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/bindSingle(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/invoke/LambdaForm$BasicType;Ljava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_invoke_LambdaForm$BasicType, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "editor()Ljava/lang/invoke/LambdaFormEditor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor) => void): void;
    public "java/lang/invoke/BoundMethodHandle/editor()Ljava/lang/invoke/LambdaFormEditor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor) => void): void;
    public static "bindSingle(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/bindSingle(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentL(ILjava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/bindArgumentL(ILjava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentI(II)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/bindArgumentI(II)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentJ(IJ)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/bindArgumentJ(IJ)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentF(IF)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/bindArgumentF(IF)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentD(ID)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/bindArgumentD(ID)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "rebind()Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/rebind()Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "tooComplex()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/BoundMethodHandle/tooComplex()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "makeReinvoker(Ljava/lang/invoke/MethodHandle;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/makeReinvoker(Ljava/lang/invoke/MethodHandle;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "speciesData()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "java/lang/invoke/BoundMethodHandle/speciesData()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "speciesData(Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/speciesData(Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "fieldCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/BoundMethodHandle/fieldCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "internalProperties()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/BoundMethodHandle/internalProperties()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "internalValues()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/BoundMethodHandle/internalValues()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "arg(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/BoundMethodHandle/arg(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "copyWith(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/copyWith(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "copyWithExtendL(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/copyWithExtendL(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "copyWithExtendI(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;I)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/copyWithExtendI(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;I)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "copyWithExtendJ(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;J)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/copyWithExtendJ(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;J)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "copyWithExtendF(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;F)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/copyWithExtendF(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;F)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "copyWithExtendD(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;D)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/copyWithExtendD(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;D)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public static "getSpeciesData(Ljava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/getSpeciesData(Ljava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "checkCache(ILjava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/checkCache(ILjava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "speciesData_L()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/speciesData_L()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "speciesData_LL()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/speciesData_LL()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "speciesData_LLL()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/speciesData_LLL()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "speciesData_LLLL()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/speciesData_LLLL()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "speciesData_LLLLL()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/speciesData_LLLLL()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "copyWith(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle/copyWith(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public static "access$300()Ljava/lang/invoke/MethodHandles$Lookup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandles$Lookup) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/access$300()Ljava/lang/invoke/MethodHandles$Lookup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandles$Lookup) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/BoundMethodHandle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_LambdaFormEditor extends java_lang_Object {
    public "java/lang/invoke/LambdaFormEditor/lambdaForm": java_lang_invoke_LambdaForm;
    public static "java/lang/invoke/LambdaFormEditor/MIN_CACHE_ARRAY_SIZE": number;
    public static "java/lang/invoke/LambdaFormEditor/MAX_CACHE_ARRAY_SIZE": number;
    public static "java/lang/invoke/LambdaFormEditor/$assertionsDisabled": number;
    public "<init>(Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormEditor/<init>(Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public static "lambdaFormEditor(Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/LambdaFormEditor;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor) => void): void;
    public static "java/lang/invoke/LambdaFormEditor/lambdaFormEditor(Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/LambdaFormEditor;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor) => void): void;
    public "getInCache(Ljava/lang/invoke/LambdaFormEditor$Transform;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/getInCache(Ljava/lang/invoke/LambdaFormEditor$Transform;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "putInCache(Ljava/lang/invoke/LambdaFormEditor$Transform;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/putInCache(Ljava/lang/invoke/LambdaFormEditor$Transform;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "buffer()Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "java/lang/invoke/LambdaFormEditor/buffer()Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "oldSpeciesData()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "java/lang/invoke/LambdaFormEditor/oldSpeciesData()Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "newSpeciesData(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "java/lang/invoke/LambdaFormEditor/newSpeciesData(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "bindArgumentL(Ljava/lang/invoke/BoundMethodHandle;ILjava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/LambdaFormEditor/bindArgumentL(Ljava/lang/invoke/BoundMethodHandle;ILjava/lang/Object;)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentI(Ljava/lang/invoke/BoundMethodHandle;II)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/LambdaFormEditor/bindArgumentI(Ljava/lang/invoke/BoundMethodHandle;II)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentJ(Ljava/lang/invoke/BoundMethodHandle;IJ)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/LambdaFormEditor/bindArgumentJ(Ljava/lang/invoke/BoundMethodHandle;IJ)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentF(Ljava/lang/invoke/BoundMethodHandle;IF)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/LambdaFormEditor/bindArgumentF(Ljava/lang/invoke/BoundMethodHandle;IF)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentD(Ljava/lang/invoke/BoundMethodHandle;ID)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/LambdaFormEditor/bindArgumentD(Ljava/lang/invoke/BoundMethodHandle;ID)Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "bindArgumentType(Ljava/lang/invoke/BoundMethodHandle;ILjava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/LambdaFormEditor/bindArgumentType(Ljava/lang/invoke/BoundMethodHandle;ILjava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle, number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "bindArgumentForm(I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/bindArgumentForm(I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "addArgumentForm(ILjava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/addArgumentForm(ILjava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "dupArgumentForm(II)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/dupArgumentForm(II)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "spreadArgumentsForm(ILjava/lang/Class;I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/spreadArgumentsForm(ILjava/lang/Class;I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "collectArgumentsForm(ILjava/lang/invoke/MethodType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/collectArgumentsForm(ILjava/lang/invoke/MethodType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "collectArgumentArrayForm(ILjava/lang/invoke/MethodHandle;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/collectArgumentArrayForm(ILjava/lang/invoke/MethodHandle;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "filterArgumentForm(ILjava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/filterArgumentForm(ILjava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "makeArgumentCombinationForm(ILjava/lang/invoke/MethodType;ZZ)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_MethodType, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/makeArgumentCombinationForm(ILjava/lang/invoke/MethodType;ZZ)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_MethodType, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "filterReturnForm(Ljava/lang/invoke/LambdaForm$BasicType;Z)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/filterReturnForm(Ljava/lang/invoke/LambdaForm$BasicType;Z)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "foldArgumentsForm(IZLjava/lang/invoke/MethodType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, number, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/foldArgumentsForm(IZLjava/lang/invoke/MethodType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, number, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "permuteArgumentsForm(I[I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormEditor/permuteArgumentsForm(I[I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "permutedTypesMatch([I[Ljava/lang/invoke/LambdaForm$BasicType;[Ljava/lang/invoke/LambdaForm$Name;I)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<java_lang_invoke_LambdaForm$BasicType>, JVMArray<java_lang_invoke_LambdaForm$Name>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaFormEditor/permutedTypesMatch([I[Ljava/lang/invoke/LambdaForm$BasicType;[Ljava/lang/invoke/LambdaForm$Name;I)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<java_lang_invoke_LambdaForm$BasicType>, JVMArray<java_lang_invoke_LambdaForm$Name>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaFormEditor/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_LambdaForm$Name extends java_lang_Object {
    public "java/lang/invoke/LambdaForm$Name/type": java_lang_invoke_LambdaForm$BasicType;
    public "java/lang/invoke/LambdaForm$Name/index": number;
    public "java/lang/invoke/LambdaForm$Name/function": java_lang_invoke_LambdaForm$NamedFunction;
    public "java/lang/invoke/LambdaForm$Name/constraint": java_lang_Object;
    public "java/lang/invoke/LambdaForm$Name/arguments": JVMArray<java_lang_Object>;
    public static "java/lang/invoke/LambdaForm$Name/$assertionsDisabled": number;
    public "<init>(ILjava/lang/invoke/LambdaForm$BasicType;Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType, java_lang_invoke_LambdaForm$NamedFunction, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/<init>(ILjava/lang/invoke/LambdaForm$BasicType;Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType, java_lang_invoke_LambdaForm$NamedFunction, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/<init>(Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/<init>(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/MethodType;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/<init>(Ljava/lang/invoke/MethodType;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/MemberName;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/<init>(Ljava/lang/invoke/MemberName;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$NamedFunction, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/<init>(Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$NamedFunction, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ILjava/lang/invoke/LambdaForm$BasicType;)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/<init>(ILjava/lang/invoke/LambdaForm$BasicType;)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/LambdaForm$BasicType;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/<init>(Ljava/lang/invoke/LambdaForm$BasicType;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable) => void): void;
    public "type()Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "java/lang/invoke/LambdaForm$Name/type()Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "index()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/index()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "initIndex(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/initIndex(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "typeChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/typeChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "resolve()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/resolve()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "newIndex(I)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "java/lang/invoke/LambdaForm$Name/newIndex(I)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "cloneWithIndex(I)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "java/lang/invoke/LambdaForm$Name/cloneWithIndex(I)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "withConstraint(Ljava/lang/Object;)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "java/lang/invoke/LambdaForm$Name/withConstraint(Ljava/lang/Object;)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "replaceName(Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "java/lang/invoke/LambdaForm$Name/replaceName(Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "replaceNames([Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;II)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$Name>, JVMArray<java_lang_invoke_LambdaForm$Name>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "java/lang/invoke/LambdaForm$Name/replaceNames([Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;II)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$Name>, JVMArray<java_lang_invoke_LambdaForm$Name>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "internArguments()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$Name/internArguments()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isParam()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/isParam()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConstantZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/isConstantZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/LambdaForm$Name/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "debugString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/LambdaForm$Name/debugString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/LambdaForm$Name/paramString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "exprString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/LambdaForm$Name/exprString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "typesMatch(Ljava/lang/invoke/LambdaForm$BasicType;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$Name/typesMatch(Ljava/lang/invoke/LambdaForm$BasicType;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastUseIndex(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/lastUseIndex(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "useCount(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/useCount(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/invoke/LambdaForm$Name;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/contains(Ljava/lang/invoke/LambdaForm$Name;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/invoke/LambdaForm$Name;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/equals(Ljava/lang/invoke/LambdaForm$Name;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$Name/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/lang/invoke/LambdaForm$Name;)S"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$Name/access$000(Ljava/lang/invoke/LambdaForm$Name;)S"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm$Name/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_LambdaForm$NamedFunction extends java_lang_Object {
    public "java/lang/invoke/LambdaForm$NamedFunction/member": java_lang_invoke_MemberName;
    public "java/lang/invoke/LambdaForm$NamedFunction/resolvedHandle": java_lang_invoke_MethodHandle;
    public "java/lang/invoke/LambdaForm$NamedFunction/invoker": java_lang_invoke_MethodHandle;
    public static "java/lang/invoke/LambdaForm$NamedFunction/INVOKER_METHOD_TYPE": java_lang_invoke_MethodType;
    public static "java/lang/invoke/LambdaForm$NamedFunction/$assertionsDisabled": number;
    public "<init>(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/<init>(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/MemberName;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/<init>(Ljava/lang/invoke/MemberName;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/MethodType;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/<init>(Ljava/lang/invoke/MethodType;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable) => void): void;
    public "isInvokeBasic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/isInvokeBasic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/<init>(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/reflect/Field;)V"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/<init>(Ljava/lang/reflect/Field;)V"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/<init>(Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolvedHandle()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/resolvedHandle()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "resolve()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/resolve()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initializeInvokers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/initializeInvokers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "invoke__V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke__V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_L_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_L_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_LL_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_LL_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_LLL_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_LLL_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_LLLL_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_LLLL_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_LLLLL_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_LLLLL_V(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke__L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke__L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_L_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_L_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_LL_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_LL_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_LLL_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_LLL_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_LLLL_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_LLLL_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "invoke_LLLLL_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/invoke_LLLLL_L(Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "arityCheck(ILjava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [number, java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/arityCheck(ILjava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [number, java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "arityCheck(ILjava/lang/Class;Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/arityCheck(ILjava/lang/Class;Ljava/lang/invoke/MethodHandle;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MethodHandle, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "computeInvoker(Ljava/lang/invoke/MethodTypeForm;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodTypeForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/computeInvoker(Ljava/lang/invoke/MethodTypeForm;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodTypeForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "invokeWithArguments([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/invokeWithArguments([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invokeWithArgumentsTracing([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/invokeWithArgumentsTracing([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/invoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public static "checkArgumentTypes([Ljava/lang/Object;Ljava/lang/invoke/MethodType;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/checkArgumentTypes([Ljava/lang/Object;Ljava/lang/invoke/MethodType;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "methodType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/methodType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "member()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/member()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "assertMemberIsConsistent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/assertMemberIsConsistent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "memberDeclaringClassOrNull()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/memberDeclaringClassOrNull()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "returnType()Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/returnType()Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "parameterType(I)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/parameterType(I)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "arity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/arity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isIdentity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/isIdentity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConstantZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/isConstantZero()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intrinsicName()Ljava/lang/invoke/MethodHandleImpl$Intrinsic;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandleImpl$Intrinsic) => void): void;
    public "java/lang/invoke/LambdaForm$NamedFunction/intrinsicName()Ljava/lang/invoke/MethodHandleImpl$Intrinsic;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandleImpl$Intrinsic) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm$NamedFunction/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_MethodTypeForm extends java_lang_Object {
    public "java/lang/invoke/MethodTypeForm/argToSlotTable": JVMArray<number>;
    public "java/lang/invoke/MethodTypeForm/slotToArgTable": JVMArray<number>;
    public "java/lang/invoke/MethodTypeForm/argCounts": Long;
    public "java/lang/invoke/MethodTypeForm/primCounts": Long;
    public "java/lang/invoke/MethodTypeForm/erasedType": java_lang_invoke_MethodType;
    public "java/lang/invoke/MethodTypeForm/basicType": java_lang_invoke_MethodType;
    public "java/lang/invoke/MethodTypeForm/methodHandles": JVMArray<java_lang_ref_SoftReference>;
    public static "java/lang/invoke/MethodTypeForm/MH_BASIC_INV": number;
    public static "java/lang/invoke/MethodTypeForm/MH_NF_INV": number;
    public static "java/lang/invoke/MethodTypeForm/MH_UNINIT_CS": number;
    public static "java/lang/invoke/MethodTypeForm/MH_LIMIT": number;
    public "java/lang/invoke/MethodTypeForm/lambdaForms": JVMArray<java_lang_ref_SoftReference>;
    public static "java/lang/invoke/MethodTypeForm/LF_INVVIRTUAL": number;
    public static "java/lang/invoke/MethodTypeForm/LF_INVSTATIC": number;
    public static "java/lang/invoke/MethodTypeForm/LF_INVSPECIAL": number;
    public static "java/lang/invoke/MethodTypeForm/LF_NEWINVSPECIAL": number;
    public static "java/lang/invoke/MethodTypeForm/LF_INVINTERFACE": number;
    public static "java/lang/invoke/MethodTypeForm/LF_INVSTATIC_INIT": number;
    public static "java/lang/invoke/MethodTypeForm/LF_INTERPRET": number;
    public static "java/lang/invoke/MethodTypeForm/LF_REBIND": number;
    public static "java/lang/invoke/MethodTypeForm/LF_DELEGATE": number;
    public static "java/lang/invoke/MethodTypeForm/LF_DELEGATE_BLOCK_INLINING": number;
    public static "java/lang/invoke/MethodTypeForm/LF_EX_LINKER": number;
    public static "java/lang/invoke/MethodTypeForm/LF_EX_INVOKER": number;
    public static "java/lang/invoke/MethodTypeForm/LF_GEN_LINKER": number;
    public static "java/lang/invoke/MethodTypeForm/LF_GEN_INVOKER": number;
    public static "java/lang/invoke/MethodTypeForm/LF_CS_LINKER": number;
    public static "java/lang/invoke/MethodTypeForm/LF_MH_LINKER": number;
    public static "java/lang/invoke/MethodTypeForm/LF_GWC": number;
    public static "java/lang/invoke/MethodTypeForm/LF_GWT": number;
    public static "java/lang/invoke/MethodTypeForm/LF_LIMIT": number;
    public static "java/lang/invoke/MethodTypeForm/NO_CHANGE": number;
    public static "java/lang/invoke/MethodTypeForm/ERASE": number;
    public static "java/lang/invoke/MethodTypeForm/WRAP": number;
    public static "java/lang/invoke/MethodTypeForm/UNWRAP": number;
    public static "java/lang/invoke/MethodTypeForm/INTS": number;
    public static "java/lang/invoke/MethodTypeForm/LONGS": number;
    public static "java/lang/invoke/MethodTypeForm/RAW_RETURN": number;
    public static "java/lang/invoke/MethodTypeForm/$assertionsDisabled": number;
    public "erasedType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodTypeForm/erasedType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "basicType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodTypeForm/basicType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "assertIsBasicType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/assertIsBasicType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "cachedMethodHandle(I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodTypeForm/cachedMethodHandle(I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "setCachedMethodHandle(ILjava/lang/invoke/MethodHandle;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodTypeForm/setCachedMethodHandle(ILjava/lang/invoke/MethodHandle;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "cachedLambdaForm(I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/MethodTypeForm/cachedLambdaForm(I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "setCachedLambdaForm(ILjava/lang/invoke/LambdaForm;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/MethodTypeForm/setCachedLambdaForm(ILjava/lang/invoke/LambdaForm;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "<init>(Ljava/lang/invoke/MethodType;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodTypeForm/<init>(Ljava/lang/invoke/MethodType;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable) => void): void;
    public static "pack(IIII)J"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/MethodTypeForm/pack(IIII)J"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "unpack(JI)C"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodTypeForm/unpack(JI)C"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/parameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parameterSlotCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/parameterSlotCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "returnCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/returnCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "returnSlotCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/returnSlotCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "primitiveParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/primitiveParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longPrimitiveParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/longPrimitiveParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "primitiveReturnCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/primitiveReturnCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longPrimitiveReturnCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/longPrimitiveReturnCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasPrimitives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/hasPrimitives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasNonVoidPrimitives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/hasNonVoidPrimitives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasLongPrimitives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/hasLongPrimitives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parameterToArgSlot(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/parameterToArgSlot(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "argSlotToParameter(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodTypeForm/argSlotToParameter(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "findForm(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodTypeForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodTypeForm) => void): void;
    public static "java/lang/invoke/MethodTypeForm/findForm(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodTypeForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodTypeForm) => void): void;
    public static "canonicalize(Ljava/lang/invoke/MethodType;II)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodTypeForm/canonicalize(Ljava/lang/invoke/MethodType;II)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "canonicalize(Ljava/lang/Class;I)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/invoke/MethodTypeForm/canonicalize(Ljava/lang/Class;I)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "canonicalizeAll([Ljava/lang/Class;I)[Ljava/lang/Class;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public static "java/lang/invoke/MethodTypeForm/canonicalizeAll([Ljava/lang/Class;I)[Ljava/lang/Class;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MethodTypeForm/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodTypeForm/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_LambdaFormBuffer extends java_lang_Object {
    public "java/lang/invoke/LambdaFormBuffer/arity": number;
    public "java/lang/invoke/LambdaFormBuffer/length": number;
    public "java/lang/invoke/LambdaFormBuffer/names": JVMArray<java_lang_invoke_LambdaForm$Name>;
    public "java/lang/invoke/LambdaFormBuffer/originalNames": JVMArray<java_lang_invoke_LambdaForm$Name>;
    public "java/lang/invoke/LambdaFormBuffer/flags": number;
    public "java/lang/invoke/LambdaFormBuffer/firstChange": number;
    public "java/lang/invoke/LambdaFormBuffer/resultName": java_lang_invoke_LambdaForm$Name;
    public "java/lang/invoke/LambdaFormBuffer/debugName": java_lang_String;
    public "java/lang/invoke/LambdaFormBuffer/dups": java_util_ArrayList;
    public static "java/lang/invoke/LambdaFormBuffer/F_TRANS": number;
    public static "java/lang/invoke/LambdaFormBuffer/F_OWNED": number;
    public static "java/lang/invoke/LambdaFormBuffer/$assertionsDisabled": number;
    public "<init>(Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/<init>(Ljava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "lambdaForm()Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/lambdaForm()Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "name(I)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/name(I)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "nameArray()[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/nameArray()[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public "resultIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/resultIndex()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setNames([Ljava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/setNames([Ljava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable) => void): void;
    public "verifyArity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/verifyArity()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "verifyFirstChange()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/verifyFirstChange()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "indexOf(Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/invoke/LambdaForm$NamedFunction;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$NamedFunction, JVMArray<java_lang_invoke_LambdaForm$NamedFunction>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaFormBuffer/indexOf(Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/invoke/LambdaForm$NamedFunction;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$NamedFunction, JVMArray<java_lang_invoke_LambdaForm$NamedFunction>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "indexOf(Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name, JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaFormBuffer/indexOf(Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name, JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inTrans()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/inTrans()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ownedCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/ownedCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "growNames(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/growNames(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "lastIndexOf(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/lastIndexOf(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "noteDuplicate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/noteDuplicate(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearDuplicatesAndNulls()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/clearDuplicatesAndNulls()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "startEdit()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/startEdit()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "changeName(ILjava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/changeName(ILjava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable) => void): void;
    public "setResult(Ljava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/setResult(Ljava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable) => void): void;
    public "endEdit()Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/endEdit()Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "copyNamesInto([Ljava/lang/invoke/LambdaForm$Name;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/copyNamesInto([Ljava/lang/invoke/LambdaForm$Name;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public "replaceFunctions([Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/Object;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$NamedFunction>, JVMArray<java_lang_invoke_LambdaForm$NamedFunction>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/replaceFunctions([Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/invoke/LambdaForm$NamedFunction;[Ljava/lang/Object;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$NamedFunction>, JVMArray<java_lang_invoke_LambdaForm$NamedFunction>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "replaceName(ILjava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/replaceName(ILjava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable) => void): void;
    public "renameParameter(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/renameParameter(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "replaceParameterByNewExpression(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/replaceParameterByNewExpression(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "replaceParameterByCopy(II)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/replaceParameterByCopy(II)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "insertName(ILjava/lang/invoke/LambdaForm$Name;Z)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/insertName(ILjava/lang/invoke/LambdaForm$Name;Z)V"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "insertExpression(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/insertExpression(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "insertParameter(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public "java/lang/invoke/LambdaFormBuffer/insertParameter(ILjava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaFormBuffer;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormBuffer) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaFormBuffer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_LambdaFormEditor$Transform extends java_lang_ref_SoftReference {
    public "java/lang/invoke/LambdaFormEditor$Transform/packedBytes": Long;
    public "java/lang/invoke/LambdaFormEditor$Transform/fullBytes": JVMArray<number>;
    public static "java/lang/invoke/LambdaFormEditor$Transform/STRESS_TEST": number;
    public static "java/lang/invoke/LambdaFormEditor$Transform/PACKED_BYTE_SIZE": number;
    public static "java/lang/invoke/LambdaFormEditor$Transform/PACKED_BYTE_MASK": number;
    public static "java/lang/invoke/LambdaFormEditor$Transform/PACKED_BYTE_MAX_LENGTH": number;
    public static "java/lang/invoke/LambdaFormEditor$Transform/NO_BYTES": JVMArray<number>;
    public static "java/lang/invoke/LambdaFormEditor$Transform/$assertionsDisabled": number;
    public static "packedBytes([B)J"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/packedBytes([B)J"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "packedBytes(II)J"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/packedBytes(II)J"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "packedBytes(III)J"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/packedBytes(III)J"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "packedBytes(IIII)J"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/packedBytes(IIII)J"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "inRange(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/inRange(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "fullBytes([I)[B"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/fullBytes([I)[B"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "byteAt(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/byteAt(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "kind()Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform$Kind) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/kind()Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform$Kind) => void): void;
    public "<init>(J[BLjava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [Long, any, JVMArray<number>, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/<init>(J[BLjava/lang/invoke/LambdaForm;)V"(thread: JVMThread, args: [Long, any, JVMArray<number>, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "bval(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/bval(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "bval(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;)B"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/bval(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;)B"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;I)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;I)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;II)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;II)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;III)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;III)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;[I)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;[I)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;I[B)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;I[B)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;II[B)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/of(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;II[B)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "ofBothArrays(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;[I[B)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/ofBothArrays(Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;[I[B)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform$Kind, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public "withResult(Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/withResult(Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/LambdaFormEditor$Transform;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/invoke/LambdaFormEditor$Transform;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/equals(Ljava/lang/invoke/LambdaFormEditor$Transform;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaFormEditor$Transform], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_LambdaFormEditor$Transform$Kind extends java_lang_Enum {
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/NO_KIND": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/BIND_ARG": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/ADD_ARG": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/DUP_ARG": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/SPREAD_ARGS": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/FILTER_ARG": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/FILTER_RETURN": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/FILTER_RETURN_TO_ZERO": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/COLLECT_ARGS": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/COLLECT_ARGS_TO_VOID": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/COLLECT_ARGS_TO_ARRAY": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/FOLD_ARGS": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/FOLD_ARGS_TO_VOID": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/PERMUTE_ARGS": java_lang_invoke_LambdaFormEditor$Transform$Kind;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/$VALUES": JVMArray<java_lang_invoke_LambdaFormEditor$Transform$Kind>;
    public static "values()[Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaFormEditor$Transform$Kind>) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/values()[Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaFormEditor$Transform$Kind>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform$Kind) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/valueOf(Ljava/lang/String;)Ljava/lang/invoke/LambdaFormEditor$Transform$Kind;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor$Transform$Kind) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaFormEditor$Transform$Kind/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaFormEditor$Transform$Kind/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_LambdaForm$BasicType extends java_lang_Enum {
    public static "java/lang/invoke/LambdaForm$BasicType/L_TYPE": java_lang_invoke_LambdaForm$BasicType;
    public static "java/lang/invoke/LambdaForm$BasicType/I_TYPE": java_lang_invoke_LambdaForm$BasicType;
    public static "java/lang/invoke/LambdaForm$BasicType/J_TYPE": java_lang_invoke_LambdaForm$BasicType;
    public static "java/lang/invoke/LambdaForm$BasicType/F_TYPE": java_lang_invoke_LambdaForm$BasicType;
    public static "java/lang/invoke/LambdaForm$BasicType/D_TYPE": java_lang_invoke_LambdaForm$BasicType;
    public static "java/lang/invoke/LambdaForm$BasicType/V_TYPE": java_lang_invoke_LambdaForm$BasicType;
    public static "java/lang/invoke/LambdaForm$BasicType/ALL_TYPES": JVMArray<java_lang_invoke_LambdaForm$BasicType>;
    public static "java/lang/invoke/LambdaForm$BasicType/ARG_TYPES": JVMArray<java_lang_invoke_LambdaForm$BasicType>;
    public static "java/lang/invoke/LambdaForm$BasicType/ARG_TYPE_LIMIT": number;
    public static "java/lang/invoke/LambdaForm$BasicType/TYPE_LIMIT": number;
    public "java/lang/invoke/LambdaForm$BasicType/btChar": number;
    public "java/lang/invoke/LambdaForm$BasicType/btClass": java_lang_Class;
    public "java/lang/invoke/LambdaForm$BasicType/btWrapper": sun_invoke_util_Wrapper;
    public static "java/lang/invoke/LambdaForm$BasicType/$VALUES": JVMArray<java_lang_invoke_LambdaForm$BasicType>;
    public static "java/lang/invoke/LambdaForm$BasicType/$assertionsDisabled": number;
    public static "values()[Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$BasicType>) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/values()[Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$BasicType>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/valueOf(Ljava/lang/String;)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "<init>(Ljava/lang/String;ICLjava/lang/Class;Lsun/invoke/util/Wrapper;)V"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_Class, sun_invoke_util_Wrapper], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm$BasicType/<init>(Ljava/lang/String;ICLjava/lang/Class;Lsun/invoke/util/Wrapper;)V"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_Class, sun_invoke_util_Wrapper], cb?: (e?: java_lang_Throwable) => void): void;
    public "basicTypeChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$BasicType/basicTypeChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "basicTypeClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/LambdaForm$BasicType/basicTypeClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "basicTypeWrapper()Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public "java/lang/invoke/LambdaForm$BasicType/basicTypeWrapper()Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public "basicTypeSlots()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm$BasicType/basicTypeSlots()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "basicType(B)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/basicType(B)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "basicType(C)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/basicType(C)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "basicType(Lsun/invoke/util/Wrapper;)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [sun_invoke_util_Wrapper], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/basicType(Lsun/invoke/util/Wrapper;)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [sun_invoke_util_Wrapper], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "basicType(Ljava/lang/Class;)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/basicType(Ljava/lang/Class;)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "basicTypeChar(Ljava/lang/Class;)C"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/basicTypeChar(Ljava/lang/Class;)C"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "basicTypes(Ljava/util/List;)[Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$BasicType>) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/basicTypes(Ljava/util/List;)[Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$BasicType>) => void): void;
    public static "basicTypes(Ljava/lang/String;)[Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$BasicType>) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/basicTypes(Ljava/lang/String;)[Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$BasicType>) => void): void;
    public static "basicTypesOrd([Ljava/lang/invoke/LambdaForm$BasicType;)[B"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$BasicType>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/basicTypesOrd([Ljava/lang/invoke/LambdaForm$BasicType;)[B"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$BasicType>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "isBasicTypeChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/isBasicTypeChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isArgBasicTypeChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/isArgBasicTypeChar(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkBasicType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/checkBasicType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/access$100(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "access$300(Ljava/lang/invoke/LambdaForm$BasicType;)C"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/access$300(Ljava/lang/invoke/LambdaForm$BasicType;)C"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm$BasicType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_invoke_util_Wrapper extends java_lang_Enum {
    public static "sun/invoke/util/Wrapper/BOOLEAN": sun_invoke_util_Wrapper;
    public static "sun/invoke/util/Wrapper/BYTE": sun_invoke_util_Wrapper;
    public static "sun/invoke/util/Wrapper/SHORT": sun_invoke_util_Wrapper;
    public static "sun/invoke/util/Wrapper/CHAR": sun_invoke_util_Wrapper;
    public static "sun/invoke/util/Wrapper/INT": sun_invoke_util_Wrapper;
    public static "sun/invoke/util/Wrapper/LONG": sun_invoke_util_Wrapper;
    public static "sun/invoke/util/Wrapper/FLOAT": sun_invoke_util_Wrapper;
    public static "sun/invoke/util/Wrapper/DOUBLE": sun_invoke_util_Wrapper;
    public static "sun/invoke/util/Wrapper/OBJECT": sun_invoke_util_Wrapper;
    public static "sun/invoke/util/Wrapper/VOID": sun_invoke_util_Wrapper;
    public "sun/invoke/util/Wrapper/wrapperType": java_lang_Class;
    public "sun/invoke/util/Wrapper/primitiveType": java_lang_Class;
    public "sun/invoke/util/Wrapper/basicTypeChar": number;
    public "sun/invoke/util/Wrapper/zero": java_lang_Object;
    public "sun/invoke/util/Wrapper/emptyArray": java_lang_Object;
    public "sun/invoke/util/Wrapper/format": number;
    public "sun/invoke/util/Wrapper/wrapperSimpleName": java_lang_String;
    public "sun/invoke/util/Wrapper/primitiveSimpleName": java_lang_String;
    public static "sun/invoke/util/Wrapper/FROM_PRIM": JVMArray<sun_invoke_util_Wrapper>;
    public static "sun/invoke/util/Wrapper/FROM_WRAP": JVMArray<sun_invoke_util_Wrapper>;
    public static "sun/invoke/util/Wrapper/FROM_CHAR": JVMArray<sun_invoke_util_Wrapper>;
    public static "sun/invoke/util/Wrapper/$VALUES": JVMArray<sun_invoke_util_Wrapper>;
    public static "sun/invoke/util/Wrapper/$assertionsDisabled": number;
    public static "values()[Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_invoke_util_Wrapper>) => void): void;
    public static "sun/invoke/util/Wrapper/values()[Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_invoke_util_Wrapper>) => void): void;
    public static "valueOf(Ljava/lang/String;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "sun/invoke/util/Wrapper/valueOf(Ljava/lang/String;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public "<init>(Ljava/lang/String;ILjava/lang/Class;Ljava/lang/Class;CLjava/lang/Object;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_Class, java_lang_Class, number, java_lang_Object, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/invoke/util/Wrapper/<init>(Ljava/lang/String;ILjava/lang/Class;Ljava/lang/Class;CLjava/lang/Object;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_Class, java_lang_Class, number, java_lang_Object, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "detailString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/invoke/util/Wrapper/detailString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "bitWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/bitWidth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "stackSlots()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/stackSlots()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSingleWord()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isSingleWord()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDoubleWord()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isDoubleWord()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isNumeric()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isNumeric()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isIntegral()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isIntegral()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSubwordOrInt()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isSubwordOrInt()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSigned()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isSigned()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isUnsigned()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isUnsigned()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFloating()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isFloating()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isOther()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isOther()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConvertibleFrom(Lsun/invoke/util/Wrapper;)Z"(thread: JVMThread, args: [sun_invoke_util_Wrapper], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/isConvertibleFrom(Lsun/invoke/util/Wrapper;)Z"(thread: JVMThread, args: [sun_invoke_util_Wrapper], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkConvertibleFrom()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/invoke/util/Wrapper/checkConvertibleFrom()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "zero()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/invoke/util/Wrapper/zero()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "zero(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/invoke/util/Wrapper/zero(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "forPrimitiveType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "sun/invoke/util/Wrapper/forPrimitiveType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "findPrimitiveType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "sun/invoke/util/Wrapper/findPrimitiveType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "forWrapperType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "sun/invoke/util/Wrapper/forWrapperType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "findWrapperType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "sun/invoke/util/Wrapper/findWrapperType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "forBasicType(C)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "sun/invoke/util/Wrapper/forBasicType(C)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "forBasicType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "sun/invoke/util/Wrapper/forBasicType(Ljava/lang/Class;)Lsun/invoke/util/Wrapper;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_invoke_util_Wrapper) => void): void;
    public static "hashPrim(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/invoke/util/Wrapper/hashPrim(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hashWrap(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/invoke/util/Wrapper/hashWrap(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hashChar(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/invoke/util/Wrapper/hashChar(C)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "primitiveType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/invoke/util/Wrapper/primitiveType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "wrapperType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/invoke/util/Wrapper/wrapperType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "wrapperType(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/invoke/util/Wrapper/wrapperType(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "newClassCastException(Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/ClassCastException;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassCastException) => void): void;
    public static "sun/invoke/util/Wrapper/newClassCastException(Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/ClassCastException;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassCastException) => void): void;
    public static "asWrapperType(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "sun/invoke/util/Wrapper/asWrapperType(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "asPrimitiveType(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "sun/invoke/util/Wrapper/asPrimitiveType(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "isWrapperType(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/invoke/util/Wrapper/isWrapperType(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isPrimitiveType(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/invoke/util/Wrapper/isPrimitiveType(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "basicTypeChar(Ljava/lang/Class;)C"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/invoke/util/Wrapper/basicTypeChar(Ljava/lang/Class;)C"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "basicTypeChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/invoke/util/Wrapper/basicTypeChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "wrapperSimpleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/invoke/util/Wrapper/wrapperSimpleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "primitiveSimpleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/invoke/util/Wrapper/primitiveSimpleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "cast(Ljava/lang/Object;Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/invoke/util/Wrapper/cast(Ljava/lang/Object;Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "convert(Ljava/lang/Object;Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/invoke/util/Wrapper/convert(Ljava/lang/Object;Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "convert(Ljava/lang/Object;Ljava/lang/Class;Z)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/invoke/util/Wrapper/convert(Ljava/lang/Object;Ljava/lang/Class;Z)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "forceType(Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "sun/invoke/util/Wrapper/forceType(Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "wrap(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/invoke/util/Wrapper/wrap(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "wrap(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/invoke/util/Wrapper/wrap(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "numberValue(Ljava/lang/Object;)Ljava/lang/Number;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Number) => void): void;
    public static "sun/invoke/util/Wrapper/numberValue(Ljava/lang/Object;)Ljava/lang/Number;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Number) => void): void;
    public static "boolValue(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/invoke/util/Wrapper/boolValue(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "newIllegalArgumentException(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/RuntimeException;"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_RuntimeException) => void): void;
    public static "sun/invoke/util/Wrapper/newIllegalArgumentException(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/RuntimeException;"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_RuntimeException) => void): void;
    public static "newIllegalArgumentException(Ljava/lang/String;)Ljava/lang/RuntimeException;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_RuntimeException) => void): void;
    public static "sun/invoke/util/Wrapper/newIllegalArgumentException(Ljava/lang/String;)Ljava/lang/RuntimeException;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_RuntimeException) => void): void;
    public "makeArray(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/invoke/util/Wrapper/makeArray(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "arrayType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/invoke/util/Wrapper/arrayType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "copyArrayUnboxing([Ljava/lang/Object;ILjava/lang/Object;II)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>, number, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/invoke/util/Wrapper/copyArrayUnboxing([Ljava/lang/Object;ILjava/lang/Object;II)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>, number, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyArrayBoxing(Ljava/lang/Object;I[Ljava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, JVMArray<java_lang_Object>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/invoke/util/Wrapper/copyArrayBoxing(Ljava/lang/Object;I[Ljava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, JVMArray<java_lang_Object>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/invoke/util/Wrapper/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ClassCastException extends java_lang_RuntimeException {
    public static "java/lang/ClassCastException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassCastException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassCastException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_BoundMethodHandle$SpeciesData extends java_lang_Object {
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/typeChars": java_lang_String;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/typeCodes": JVMArray<java_lang_invoke_LambdaForm$BasicType>;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/clazz": java_lang_Class;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/constructor": JVMArray<java_lang_invoke_MethodHandle>;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/getters": JVMArray<java_lang_invoke_MethodHandle>;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/nominalGetters": JVMArray<java_lang_invoke_LambdaForm$NamedFunction>;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/extensions": JVMArray<java_lang_invoke_BoundMethodHandle$SpeciesData>;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/EMPTY": java_lang_invoke_BoundMethodHandle$SpeciesData;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/CACHE": java_util_HashMap;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/INIT_DONE": number;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/$assertionsDisabled": number;
    public "fieldCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/fieldCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fieldType(I)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/fieldType(I)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "fieldTypeChar(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/fieldTypeChar(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "fieldSignature()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/fieldSignature()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "fieldHolder()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/fieldHolder()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getterFunction(I)Ljava/lang/invoke/LambdaForm$NamedFunction;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$NamedFunction) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/getterFunction(I)Ljava/lang/invoke/LambdaForm$NamedFunction;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$NamedFunction) => void): void;
    public "getterFunctions()[Ljava/lang/invoke/LambdaForm$NamedFunction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$NamedFunction>) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/getterFunctions()[Ljava/lang/invoke/LambdaForm$NamedFunction;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$NamedFunction>) => void): void;
    public "getterHandles()[Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_MethodHandle>) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/getterHandles()[Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_MethodHandle>) => void): void;
    public "constructor()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/constructor()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/<init>(Ljava/lang/String;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "initForBootstrap()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/initForBootstrap()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isPlaceholder()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/isPlaceholder()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "extendWith(B)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/extendWith(B)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "extendWith(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public "java/lang/invoke/BoundMethodHandle$SpeciesData/extendWith(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "get(Ljava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/get(Ljava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "getForClass(Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/getForClass(Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "lookupCache(Ljava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/lookupCache(Ljava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "updateCache(Ljava/lang/String;Ljava/lang/invoke/BoundMethodHandle$SpeciesData;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_BoundMethodHandle$SpeciesData], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/updateCache(Ljava/lang/String;Ljava/lang/invoke/BoundMethodHandle$SpeciesData;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_BoundMethodHandle$SpeciesData], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "access$000(Ljava/lang/invoke/BoundMethodHandle$SpeciesData;)[Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle$SpeciesData], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_MethodHandle>) => void): void;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/access$000(Ljava/lang/invoke/BoundMethodHandle$SpeciesData;)[Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle$SpeciesData], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_MethodHandle>) => void): void;
    public static "access$100(Ljava/lang/invoke/BoundMethodHandle$SpeciesData;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle$SpeciesData], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/access$100(Ljava/lang/invoke/BoundMethodHandle$SpeciesData;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_invoke_BoundMethodHandle$SpeciesData], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$200(Ljava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/access$200(Ljava/lang/String;)Ljava/lang/invoke/BoundMethodHandle$SpeciesData;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle$SpeciesData) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/BoundMethodHandle$SpeciesData/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_MethodHandles$Lookup extends java_lang_Object {
    public "java/lang/invoke/MethodHandles$Lookup/lookupClass": java_lang_Class;
    public "java/lang/invoke/MethodHandles$Lookup/allowedModes": number;
    public static "java/lang/invoke/MethodHandles$Lookup/PUBLIC": number;
    public static "java/lang/invoke/MethodHandles$Lookup/PRIVATE": number;
    public static "java/lang/invoke/MethodHandles$Lookup/PROTECTED": number;
    public static "java/lang/invoke/MethodHandles$Lookup/PACKAGE": number;
    public static "java/lang/invoke/MethodHandles$Lookup/ALL_MODES": number;
    public static "java/lang/invoke/MethodHandles$Lookup/TRUSTED": number;
    public static "java/lang/invoke/MethodHandles$Lookup/PUBLIC_LOOKUP": java_lang_invoke_MethodHandles$Lookup;
    public static "java/lang/invoke/MethodHandles$Lookup/IMPL_LOOKUP": java_lang_invoke_MethodHandles$Lookup;
    public static "java/lang/invoke/MethodHandles$Lookup/ALLOW_NESTMATE_ACCESS": number;
    public static "java/lang/invoke/MethodHandles$Lookup/LOOKASIDE_TABLE": java_util_concurrent_ConcurrentHashMap;
    public static "java/lang/invoke/MethodHandles$Lookup/$assertionsDisabled": number;
    public static "fixmods(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandles$Lookup/fixmods(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lookupClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/lookupClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "lookupClassOrNull()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/lookupClassOrNull()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "lookupModes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/lookupModes()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Class;I)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/<init>(Ljava/lang/Class;I)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "in(Ljava/lang/Class;)Ljava/lang/invoke/MethodHandles$Lookup;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandles$Lookup) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/in(Ljava/lang/Class;)Ljava/lang/invoke/MethodHandles$Lookup;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandles$Lookup) => void): void;
    public static "checkUnprivilegedlookupClass(Ljava/lang/Class;I)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandles$Lookup/checkUnprivilegedlookupClass(Ljava/lang/Class;I)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "findStatic(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findStatic(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "findVirtual(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findVirtual(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "findVirtualForMH(Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findVirtualForMH(Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "findConstructor(Ljava/lang/Class;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findConstructor(Ljava/lang/Class;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "findSpecial(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_invoke_MethodType, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findSpecial(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_invoke_MethodType, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "findGetter(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findGetter(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "findSetter(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findSetter(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "findStaticGetter(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findStaticGetter(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "findStaticSetter(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findStaticSetter(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "bind(Ljava/lang/Object;Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Object, java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/bind(Ljava/lang/Object;Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Object, java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "unreflect(Ljava/lang/reflect/Method;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/unreflect(Ljava/lang/reflect/Method;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "unreflectForMH(Ljava/lang/reflect/Method;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/unreflectForMH(Ljava/lang/reflect/Method;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "unreflectSpecial(Ljava/lang/reflect/Method;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/unreflectSpecial(Ljava/lang/reflect/Method;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "unreflectConstructor(Ljava/lang/reflect/Constructor;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/unreflectConstructor(Ljava/lang/reflect/Constructor;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "unreflectGetter(Ljava/lang/reflect/Field;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/unreflectGetter(Ljava/lang/reflect/Field;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "unreflectField(Ljava/lang/reflect/Field;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Field, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/unreflectField(Ljava/lang/reflect/Field;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Field, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "unreflectSetter(Ljava/lang/reflect/Field;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/unreflectSetter(Ljava/lang/reflect/Field;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "revealDirect(Ljava/lang/invoke/MethodHandle;)Ljava/lang/invoke/MethodHandleInfo;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandleInfo) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/revealDirect(Ljava/lang/invoke/MethodHandle;)Ljava/lang/invoke/MethodHandleInfo;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandleInfo) => void): void;
    public "resolveOrFail(BLjava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/resolveOrFail(BLjava/lang/Class;Ljava/lang/String;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "resolveOrFail(BLjava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/resolveOrFail(BLjava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "resolveOrFail(BLjava/lang/invoke/MemberName;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/resolveOrFail(BLjava/lang/invoke/MemberName;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "checkSymbolicClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/checkSymbolicClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkMethodName(BLjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/checkMethodName(BLjava/lang/String;)V"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "findBoundCallerClass(Ljava/lang/invoke/MemberName;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/findBoundCallerClass(Ljava/lang/invoke/MemberName;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "hasPrivateAccess()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/hasPrivateAccess()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkSecurityManager(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/checkSecurityManager(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkMethod(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/checkMethod(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkField(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/checkField(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAccess(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/checkAccess(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "accessFailedMessage(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/accessFailedMessage(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "checkSpecialCaller(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/checkSpecialCaller(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "restrictProtectedReceiver(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/restrictProtectedReceiver(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "restrictReceiver(Ljava/lang/invoke/MemberName;Ljava/lang/invoke/DirectMethodHandle;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_invoke_DirectMethodHandle, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/restrictReceiver(Ljava/lang/invoke/MemberName;Ljava/lang/invoke/DirectMethodHandle;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_invoke_DirectMethodHandle, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectMethod(BLjava/lang/Class;Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectMethod(BLjava/lang/Class;Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectMethodNoRestrict(BLjava/lang/Class;Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectMethodNoRestrict(BLjava/lang/Class;Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectMethodNoSecurityManager(BLjava/lang/Class;Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectMethodNoSecurityManager(BLjava/lang/Class;Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectMethodCommon(BLjava/lang/Class;Ljava/lang/invoke/MemberName;ZZLjava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, number, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectMethodCommon(BLjava/lang/Class;Ljava/lang/invoke/MemberName;ZZLjava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, number, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "maybeBindCaller(Ljava/lang/invoke/MemberName;Ljava/lang/invoke/MethodHandle;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_invoke_MethodHandle, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/maybeBindCaller(Ljava/lang/invoke/MemberName;Ljava/lang/invoke/MethodHandle;Ljava/lang/Class;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_invoke_MethodHandle, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectField(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectField(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectFieldNoSecurityManager(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectFieldNoSecurityManager(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectFieldCommon(BLjava/lang/Class;Ljava/lang/invoke/MemberName;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectFieldCommon(BLjava/lang/Class;Ljava/lang/invoke/MemberName;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectConstructor(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectConstructor(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectConstructorNoSecurityManager(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectConstructorNoSecurityManager(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "getDirectConstructorCommon(Ljava/lang/Class;Ljava/lang/invoke/MemberName;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectConstructorCommon(Ljava/lang/Class;Ljava/lang/invoke/MemberName;Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "linkMethodHandleConstant(BLjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/linkMethodHandleConstant(BLjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "canBeCached(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/canBeCached(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDirectMethodForConstant(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/MethodHandles$Lookup/getDirectMethodForConstant(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandles$Lookup/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_DirectMethodHandle extends java_lang_invoke_MethodHandle {
    public "java/lang/invoke/DirectMethodHandle/member": java_lang_invoke_MemberName;
    public static "java/lang/invoke/DirectMethodHandle/IMPL_NAMES": java_lang_invoke_MemberName$Factory;
    public static "java/lang/invoke/DirectMethodHandle/AF_GETFIELD": number;
    public static "java/lang/invoke/DirectMethodHandle/AF_PUTFIELD": number;
    public static "java/lang/invoke/DirectMethodHandle/AF_GETSTATIC": number;
    public static "java/lang/invoke/DirectMethodHandle/AF_PUTSTATIC": number;
    public static "java/lang/invoke/DirectMethodHandle/AF_GETSTATIC_INIT": number;
    public static "java/lang/invoke/DirectMethodHandle/AF_PUTSTATIC_INIT": number;
    public static "java/lang/invoke/DirectMethodHandle/AF_LIMIT": number;
    public static "java/lang/invoke/DirectMethodHandle/FT_LAST_WRAPPER": number;
    public static "java/lang/invoke/DirectMethodHandle/FT_UNCHECKED_REF": number;
    public static "java/lang/invoke/DirectMethodHandle/FT_CHECKED_REF": number;
    public static "java/lang/invoke/DirectMethodHandle/FT_LIMIT": number;
    public static "java/lang/invoke/DirectMethodHandle/ACCESSOR_FORMS": JVMArray<java_lang_invoke_LambdaForm>;
    public static "java/lang/invoke/DirectMethodHandle/$assertionsDisabled": number;
    public "<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/DirectMethodHandle/<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/make(BLjava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [number, java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "make(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/make(Ljava/lang/Class;Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "make(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/make(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "make(Ljava/lang/reflect/Method;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/make(Ljava/lang/reflect/Method;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "make(Ljava/lang/reflect/Field;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/make(Ljava/lang/reflect/Field;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "makeAllocator(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/makeAllocator(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/DirectMethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_DirectMethodHandle) => void): void;
    public "rebind()Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "java/lang/invoke/DirectMethodHandle/rebind()Ljava/lang/invoke/BoundMethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_BoundMethodHandle) => void): void;
    public "copyWith(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/DirectMethodHandle/copyWith(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "internalProperties()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/DirectMethodHandle/internalProperties()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "internalMemberName()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/DirectMethodHandle/internalMemberName()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "preparedLambdaForm(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/preparedLambdaForm(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "preparedLambdaForm(Ljava/lang/invoke/MethodType;I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/preparedLambdaForm(Ljava/lang/invoke/MethodType;I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "makePreparedLambdaForm(Ljava/lang/invoke/MethodType;I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/makePreparedLambdaForm(Ljava/lang/invoke/MethodType;I)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "findDirectMethodHandle(Ljava/lang/invoke/LambdaForm$Name;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/findDirectMethodHandle(Ljava/lang/invoke/LambdaForm$Name;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "maybeCompile(Ljava/lang/invoke/LambdaForm;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/maybeCompile(Ljava/lang/invoke/LambdaForm;Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_invoke_LambdaForm, java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public static "internalMemberName(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/internalMemberName(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "internalMemberNameEnsureInit(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/internalMemberNameEnsureInit(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "shouldBeInitialized(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/shouldBeInitialized(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "ensureInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/DirectMethodHandle/ensureInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkInitialized(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/checkInitialized(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ensureInitialized(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/ensureInitialized(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "constructorMethod(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/constructorMethod(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "allocateInstance(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/allocateInstance(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "fieldOffset(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/fieldOffset(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "checkBase(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/checkBase(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "nullCheck(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/nullCheck(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "staticBase(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/staticBase(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "staticOffset(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/staticOffset(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "checkCast(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/checkCast(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "checkCast(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/DirectMethodHandle/checkCast(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "afIndex(BZI)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/afIndex(BZI)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ftypeKind(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/ftypeKind(Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "preparedFieldLambdaForm(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/preparedFieldLambdaForm(Ljava/lang/invoke/MemberName;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "preparedFieldLambdaForm(BZLjava/lang/Class;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/preparedFieldLambdaForm(BZLjava/lang/Class;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "makePreparedFieldLambdaForm(BZI)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/makePreparedFieldLambdaForm(BZI)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "internalProperties()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/DirectMethodHandle/internalProperties()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/invoke/MemberName;Ljava/lang/invoke/DirectMethodHandle$1;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_invoke_MemberName, java_lang_invoke_DirectMethodHandle$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/DirectMethodHandle/<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/LambdaForm;Ljava/lang/invoke/MemberName;Ljava/lang/invoke/DirectMethodHandle$1;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_LambdaForm, java_lang_invoke_MemberName, java_lang_invoke_DirectMethodHandle$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/DirectMethodHandle/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_DirectMethodHandle$1 extends java_lang_Object {
  }
  export class java_lang_invoke_MemberName$Factory extends java_lang_Object {
    public static "java/lang/invoke/MemberName$Factory/INSTANCE": java_lang_invoke_MemberName$Factory;
    public static "java/lang/invoke/MemberName$Factory/ALLOWED_FLAGS": number;
    public static "java/lang/invoke/MemberName$Factory/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName$Factory/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMembers(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Object;ILjava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Object, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/lang/invoke/MemberName$Factory/getMembers(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Object;ILjava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Object, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "resolve(BLjava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName$Factory/resolve(BLjava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "resolveOrFail(BLjava/lang/invoke/MemberName;Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_invoke_MemberName, java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName$Factory/resolveOrFail(BLjava/lang/invoke/MemberName;Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_invoke_MemberName, java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "resolveOrNull(BLjava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName$Factory/resolveOrNull(BLjava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "getMethods(Ljava/lang/Class;ZLjava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/lang/invoke/MemberName$Factory/getMethods(Ljava/lang/Class;ZLjava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getMethods(Ljava/lang/Class;ZLjava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_String, java_lang_invoke_MethodType, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/lang/invoke/MemberName$Factory/getMethods(Ljava/lang/Class;ZLjava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_String, java_lang_invoke_MethodType, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getConstructors(Ljava/lang/Class;Ljava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/lang/invoke/MemberName$Factory/getConstructors(Ljava/lang/Class;Ljava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getFields(Ljava/lang/Class;ZLjava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/lang/invoke/MemberName$Factory/getFields(Ljava/lang/Class;ZLjava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getFields(Ljava/lang/Class;ZLjava/lang/String;Ljava/lang/Class;Ljava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_String, java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/lang/invoke/MemberName$Factory/getFields(Ljava/lang/Class;ZLjava/lang/String;Ljava/lang/Class;Ljava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_String, java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "getNestedTypes(Ljava/lang/Class;ZLjava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/lang/invoke/MemberName$Factory/getNestedTypes(Ljava/lang/Class;ZLjava/lang/Class;)Ljava/util/List;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public static "newMemberBuffer(I)[Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_MemberName>) => void): void;
    public static "java/lang/invoke/MemberName$Factory/newMemberBuffer(I)[Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_MemberName>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MemberName$Factory/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_lang_invoke_MethodHandleInfo extends java_lang_Object {
    "getReferenceKind()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getMethodType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    "reflectAs(Ljava/lang/Class;Ljava/lang/invoke/MethodHandles$Lookup;)Ljava/lang/reflect/Member;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MethodHandles$Lookup], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Member) => void): void;
    "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "isVarArgs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_lang_invoke_LambdaForm extends java_lang_Object {
    public "java/lang/invoke/LambdaForm/arity": number;
    public "java/lang/invoke/LambdaForm/result": number;
    public "java/lang/invoke/LambdaForm/forceInline": number;
    public "java/lang/invoke/LambdaForm/customized": java_lang_invoke_MethodHandle;
    public "java/lang/invoke/LambdaForm/names": JVMArray<java_lang_invoke_LambdaForm$Name>;
    public "java/lang/invoke/LambdaForm/debugName": java_lang_String;
    public "java/lang/invoke/LambdaForm/vmentry": java_lang_invoke_MemberName;
    public "java/lang/invoke/LambdaForm/isCompiled": number;
    public "java/lang/invoke/LambdaForm/transformCache": java_lang_Object;
    public static "java/lang/invoke/LambdaForm/VOID_RESULT": number;
    public static "java/lang/invoke/LambdaForm/LAST_RESULT": number;
    public static "java/lang/invoke/LambdaForm/USE_PREDEFINED_INTERPRET_METHODS": number;
    public static "java/lang/invoke/LambdaForm/COMPILE_THRESHOLD": number;
    public "java/lang/invoke/LambdaForm/invocationCounter": number;
    public static "java/lang/invoke/LambdaForm/INTERNED_ARGUMENT_LIMIT": number;
    public static "java/lang/invoke/LambdaForm/INTERNED_ARGUMENTS": JVMArray<JVMArray<java_lang_invoke_LambdaForm$Name>>;
    public static "java/lang/invoke/LambdaForm/IMPL_NAMES": java_lang_invoke_MemberName$Factory;
    public static "java/lang/invoke/LambdaForm/LF_identityForm": JVMArray<java_lang_invoke_LambdaForm>;
    public static "java/lang/invoke/LambdaForm/LF_zeroForm": JVMArray<java_lang_invoke_LambdaForm>;
    public static "java/lang/invoke/LambdaForm/NF_identity": JVMArray<java_lang_invoke_LambdaForm$NamedFunction>;
    public static "java/lang/invoke/LambdaForm/NF_zero": JVMArray<java_lang_invoke_LambdaForm$NamedFunction>;
    public static "java/lang/invoke/LambdaForm/DEBUG_NAME_COUNTERS": java_util_HashMap;
    public static "java/lang/invoke/LambdaForm/TRACE_INTERPRETER": number;
    public static "java/lang/invoke/LambdaForm/$assertionsDisabled": number;
    public "<init>(Ljava/lang/String;I[Ljava/lang/invoke/LambdaForm$Name;I)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_invoke_LambdaForm$Name>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm/<init>(Ljava/lang/String;I[Ljava/lang/invoke/LambdaForm$Name;I)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_invoke_LambdaForm$Name>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I[Ljava/lang/invoke/LambdaForm$Name;IZLjava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_invoke_LambdaForm$Name>, number, number, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm/<init>(Ljava/lang/String;I[Ljava/lang/invoke/LambdaForm$Name;IZLjava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_invoke_LambdaForm$Name>, number, number, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I[Ljava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm/<init>(Ljava/lang/String;I[Ljava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;I[Ljava/lang/invoke/LambdaForm$Name;Z)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_invoke_LambdaForm$Name>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm/<init>(Ljava/lang/String;I[Ljava/lang/invoke/LambdaForm$Name;Z)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_invoke_LambdaForm$Name>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;[Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_invoke_LambdaForm$Name>, JVMArray<java_lang_invoke_LambdaForm$Name>, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm/<init>(Ljava/lang/String;[Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/invoke/LambdaForm$Name;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_invoke_LambdaForm$Name>, JVMArray<java_lang_invoke_LambdaForm$Name>, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;[Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/invoke/LambdaForm$Name;Z)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_invoke_LambdaForm$Name>, JVMArray<java_lang_invoke_LambdaForm$Name>, java_lang_invoke_LambdaForm$Name, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm/<init>(Ljava/lang/String;[Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/invoke/LambdaForm$Name;Z)V"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_invoke_LambdaForm$Name>, JVMArray<java_lang_invoke_LambdaForm$Name>, java_lang_invoke_LambdaForm$Name, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "buildNames([Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/invoke/LambdaForm$Name;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$Name>, JVMArray<java_lang_invoke_LambdaForm$Name>, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "java/lang/invoke/LambdaForm/buildNames([Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/invoke/LambdaForm$Name;Ljava/lang/invoke/LambdaForm$Name;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [JVMArray<java_lang_invoke_LambdaForm$Name>, JVMArray<java_lang_invoke_LambdaForm$Name>, java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "buildEmptyNames(ILjava/lang/String;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "java/lang/invoke/LambdaForm/buildEmptyNames(ILjava/lang/String;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "fixResult(I[Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [number, JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/fixResult(I[Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [number, JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "fixDebugName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/invoke/LambdaForm/fixDebugName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "namesOK(I[Ljava/lang/invoke/LambdaForm$Name;)Z"(thread: JVMThread, args: [number, JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/namesOK(I[Ljava/lang/invoke/LambdaForm$Name;)Z"(thread: JVMThread, args: [number, JVMArray<java_lang_invoke_LambdaForm$Name>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "customize(Ljava/lang/invoke/MethodHandle;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaForm/customize(Ljava/lang/invoke/MethodHandle;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "uncustomize()Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaForm/uncustomize()Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "normalize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/normalize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nameRefsAreLegal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/nameRefsAreLegal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "returnType()Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "java/lang/invoke/LambdaForm/returnType()Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "parameterType(I)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "java/lang/invoke/LambdaForm/parameterType(I)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public "parameter(I)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "java/lang/invoke/LambdaForm/parameter(I)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public "parameterConstraint(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/LambdaForm/parameterConstraint(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "arity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/arity()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "expressionCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/expressionCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "methodType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/LambdaForm/methodType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "basicTypeSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/LambdaForm/basicTypeSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "signatureArity(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/signatureArity(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "signatureReturn(Ljava/lang/String;)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "java/lang/invoke/LambdaForm/signatureReturn(Ljava/lang/String;)Ljava/lang/invoke/LambdaForm$BasicType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$BasicType) => void): void;
    public static "isValidSignature(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/isValidSignature(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "signatureType(Ljava/lang/String;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/LambdaForm/signatureType(Ljava/lang/String;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "prepare()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm/prepare()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "compileToBytecode()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/LambdaForm/compileToBytecode()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "computeInitialPreparedForms()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm/computeInitialPreparedForms()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "interpret_L(Ljava/lang/invoke/MethodHandle;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm/interpret_L(Ljava/lang/invoke/MethodHandle;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "interpret_L(Ljava/lang/invoke/MethodHandle;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm/interpret_L(Ljava/lang/invoke/MethodHandle;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "interpret_L(Ljava/lang/invoke/MethodHandle;Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm/interpret_L(Ljava/lang/invoke/MethodHandle;Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "getPreparedForm(Ljava/lang/String;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/LambdaForm/getPreparedForm(Ljava/lang/String;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "argumentTypesMatch(Ljava/lang/String;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/argumentTypesMatch(Ljava/lang/String;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "valueMatches(Ljava/lang/invoke/LambdaForm$BasicType;Ljava/lang/Class;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType, java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/valueMatches(Ljava/lang/invoke/LambdaForm$BasicType;Ljava/lang/Class;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType, java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "returnTypesMatch(Ljava/lang/String;[Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/returnTypesMatch(Ljava/lang/String;[Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkInt(Ljava/lang/Class;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/checkInt(Ljava/lang/Class;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkRef(Ljava/lang/Class;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/checkRef(Ljava/lang/Class;Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "interpretWithArguments([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/LambdaForm/interpretWithArguments([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "interpretName(Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/LambdaForm/interpretName(Ljava/lang/invoke/LambdaForm$Name;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "checkInvocationCounter()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/LambdaForm/checkInvocationCounter()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interpretWithArgumentsTracing([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/LambdaForm/interpretWithArgumentsTracing([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "traceInterpreter(Ljava/lang/String;Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm/traceInterpreter(Ljava/lang/String;Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "traceInterpreter(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm/traceInterpreter(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "arityCheck([Ljava/lang/Object;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/arityCheck([Ljava/lang/Object;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "resultCheck([Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/resultCheck([Ljava/lang/Object;Ljava/lang/Object;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/LambdaForm/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/invoke/LambdaForm;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/equals(Ljava/lang/invoke/LambdaForm;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "editor()Ljava/lang/invoke/LambdaFormEditor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor) => void): void;
    public "java/lang/invoke/LambdaForm/editor()Ljava/lang/invoke/LambdaFormEditor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaFormEditor) => void): void;
    public "contains(Ljava/lang/invoke/LambdaForm$Name;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/contains(Ljava/lang/invoke/LambdaForm$Name;)Z"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "addArguments(I[Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, JVMArray<java_lang_invoke_LambdaForm$BasicType>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaForm/addArguments(I[Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, JVMArray<java_lang_invoke_LambdaForm$BasicType>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "addArguments(ILjava/util/List;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaForm/addArguments(ILjava/util/List;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "permuteArguments(I[I[Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<java_lang_invoke_LambdaForm$BasicType>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public "java/lang/invoke/LambdaForm/permuteArguments(I[I[Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<java_lang_invoke_LambdaForm$BasicType>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "permutedTypesMatch([I[Ljava/lang/invoke/LambdaForm$BasicType;[Ljava/lang/invoke/LambdaForm$Name;I)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<java_lang_invoke_LambdaForm$BasicType>, JVMArray<java_lang_invoke_LambdaForm$Name>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/permutedTypesMatch([I[Ljava/lang/invoke/LambdaForm$BasicType;[Ljava/lang/invoke/LambdaForm$Name;I)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<java_lang_invoke_LambdaForm$BasicType>, JVMArray<java_lang_invoke_LambdaForm$Name>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "basicTypeSignature(Ljava/lang/invoke/MethodType;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/invoke/LambdaForm/basicTypeSignature(Ljava/lang/invoke/MethodType;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "shortenSignature(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/invoke/LambdaForm/shortenSignature(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "lastUseIndex(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/lastUseIndex(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "useCount(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/LambdaForm/useCount(Ljava/lang/invoke/LambdaForm$Name;)I"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "argument(IC)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public static "java/lang/invoke/LambdaForm/argument(IC)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public static "argument(ILjava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public static "java/lang/invoke/LambdaForm/argument(ILjava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public static "internArgument(Ljava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public static "java/lang/invoke/LambdaForm/internArgument(Ljava/lang/invoke/LambdaForm$Name;)Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$Name], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$Name) => void): void;
    public static "arguments(ILjava/lang/String;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "java/lang/invoke/LambdaForm/arguments(ILjava/lang/String;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "arguments(I[C)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "java/lang/invoke/LambdaForm/arguments(I[C)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "arguments(ILjava/util/List;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_util_List], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "java/lang/invoke/LambdaForm/arguments(ILjava/util/List;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_util_List], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "arguments(I[Ljava/lang/Class;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "java/lang/invoke/LambdaForm/arguments(I[Ljava/lang/Class;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "arguments(ILjava/lang/invoke/MethodType;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "java/lang/invoke/LambdaForm/arguments(ILjava/lang/invoke/MethodType;)[Ljava/lang/invoke/LambdaForm$Name;"(thread: JVMThread, args: [number, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_invoke_LambdaForm$Name>) => void): void;
    public static "identityForm(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/LambdaForm/identityForm(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "zeroForm(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/LambdaForm/zeroForm(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "identity(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm$NamedFunction;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$NamedFunction) => void): void;
    public static "java/lang/invoke/LambdaForm/identity(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm$NamedFunction;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$NamedFunction) => void): void;
    public static "constantZero(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm$NamedFunction;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$NamedFunction) => void): void;
    public static "java/lang/invoke/LambdaForm/constantZero(Ljava/lang/invoke/LambdaForm$BasicType;)Ljava/lang/invoke/LambdaForm$NamedFunction;"(thread: JVMThread, args: [java_lang_invoke_LambdaForm$BasicType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm$NamedFunction) => void): void;
    public static "createIdentityForms()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm/createIdentityForms()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "identity_I(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/identity_I(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "identity_J(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/LambdaForm/identity_J(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "identity_F(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/identity_F(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "identity_D(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/identity_D(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "identity_L(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm/identity_L(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "identity_V()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm/identity_V()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "zero_I()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/zero_I()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "zero_J()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/LambdaForm/zero_J()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "zero_F()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/zero_F()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "zero_D()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/zero_D()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "zero_L()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/LambdaForm/zero_L()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "zero_V()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm/zero_V()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$200()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/LambdaForm/access$200()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/LambdaForm/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_MethodType extends java_lang_Object implements java_io_Serializable {
    public toString(): string;
    public static "java/lang/invoke/MethodType/serialVersionUID": Long;
    public "java/lang/invoke/MethodType/rtype": java_lang_Class;
    public "java/lang/invoke/MethodType/ptypes": JVMArray<java_lang_Class>;
    public "java/lang/invoke/MethodType/form": java_lang_invoke_MethodTypeForm;
    public "java/lang/invoke/MethodType/wrapAlt": java_lang_invoke_MethodType;
    public "java/lang/invoke/MethodType/invokers": java_lang_invoke_Invokers;
    public "java/lang/invoke/MethodType/methodDescriptor": java_lang_String;
    public static "java/lang/invoke/MethodType/MAX_JVM_ARITY": number;
    public static "java/lang/invoke/MethodType/MAX_MH_ARITY": number;
    public static "java/lang/invoke/MethodType/MAX_MH_INVOKER_ARITY": number;
    public static "java/lang/invoke/MethodType/internTable": java_lang_invoke_MethodType$ConcurrentWeakInternSet;
    public static "java/lang/invoke/MethodType/NO_PTYPES": JVMArray<java_lang_Class>;
    public static "java/lang/invoke/MethodType/objectOnlyTypes": JVMArray<java_lang_invoke_MethodType>;
    public static "java/lang/invoke/MethodType/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/lang/invoke/MethodType/rtypeOffset": Long;
    public static "java/lang/invoke/MethodType/ptypesOffset": Long;
    public static "java/lang/invoke/MethodType/$assertionsDisabled": number;
    public "<init>(Ljava/lang/Class;[Ljava/lang/Class;Z)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodType/<init>(Ljava/lang/Class;[Ljava/lang/Class;Z)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([Ljava/lang/Class;Ljava/lang/Class;)V"(thread: JVMThread, args: [JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodType/<init>([Ljava/lang/Class;Ljava/lang/Class;)V"(thread: JVMThread, args: [JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "form()Ljava/lang/invoke/MethodTypeForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodTypeForm) => void): void;
    public "java/lang/invoke/MethodType/form()Ljava/lang/invoke/MethodTypeForm;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodTypeForm) => void): void;
    public "rtype()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MethodType/rtype()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "ptypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/invoke/MethodType/ptypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "setForm(Ljava/lang/invoke/MethodTypeForm;)V"(thread: JVMThread, args: [java_lang_invoke_MethodTypeForm], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodType/setForm(Ljava/lang/invoke/MethodTypeForm;)V"(thread: JVMThread, args: [java_lang_invoke_MethodTypeForm], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkRtype(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodType/checkRtype(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkPtype(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodType/checkPtype(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkPtypes([Ljava/lang/Class;)I"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodType/checkPtypes([Ljava/lang/Class;)I"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkSlotCount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodType/checkSlotCount(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "newIndexOutOfBoundsException(Ljava/lang/Object;)Ljava/lang/IndexOutOfBoundsException;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_IndexOutOfBoundsException) => void): void;
    public static "java/lang/invoke/MethodType/newIndexOutOfBoundsException(Ljava/lang/Object;)Ljava/lang/IndexOutOfBoundsException;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_IndexOutOfBoundsException) => void): void;
    public static "methodType(Ljava/lang/Class;[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/methodType(Ljava/lang/Class;[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "methodType(Ljava/lang/Class;Ljava/util/List;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/methodType(Ljava/lang/Class;Ljava/util/List;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "listToArray(Ljava/util/List;)[Ljava/lang/Class;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public static "java/lang/invoke/MethodType/listToArray(Ljava/util/List;)[Ljava/lang/Class;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public static "methodType(Ljava/lang/Class;Ljava/lang/Class;[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/methodType(Ljava/lang/Class;Ljava/lang/Class;[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "methodType(Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/methodType(Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "methodType(Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/methodType(Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "methodType(Ljava/lang/Class;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/methodType(Ljava/lang/Class;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "makeImpl(Ljava/lang/Class;[Ljava/lang/Class;Z)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/makeImpl(Ljava/lang/Class;[Ljava/lang/Class;Z)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "genericMethodType(IZ)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/genericMethodType(IZ)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "genericMethodType(I)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/genericMethodType(I)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "changeParameterType(ILjava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/changeParameterType(ILjava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "insertParameterTypes(I[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/insertParameterTypes(I[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "appendParameterTypes([Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/appendParameterTypes([Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "insertParameterTypes(ILjava/util/List;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/insertParameterTypes(ILjava/util/List;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "appendParameterTypes(Ljava/util/List;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/appendParameterTypes(Ljava/util/List;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "replaceParameterTypes(II[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, number, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/replaceParameterTypes(II[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, number, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "asSpreaderType(Ljava/lang/Class;I)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/asSpreaderType(Ljava/lang/Class;I)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "leadingReferenceParameter()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MethodType/leadingReferenceParameter()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "asCollectorType(Ljava/lang/Class;I)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/asCollectorType(Ljava/lang/Class;I)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "dropParameterTypes(II)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/dropParameterTypes(II)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "changeReturnType(Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/changeReturnType(Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "hasPrimitives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/hasPrimitives()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasWrappers()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/hasWrappers()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "erase()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/erase()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "basicType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/basicType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "invokerType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/invokerType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "generic()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/generic()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "isGeneric()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/isGeneric()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "wrap()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/wrap()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "unwrap()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MethodType/unwrap()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "wrapWithPrims(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/wrapWithPrims(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "unwrapWithNoPrims(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/unwrapWithNoPrims(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "parameterType(I)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MethodType/parameterType(I)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "parameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/parameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "returnType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MethodType/returnType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "parameterList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "java/lang/invoke/MethodType/parameterList()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "lastParameterType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MethodType/lastParameterType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "parameterArray()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/invoke/MethodType/parameterArray()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/invoke/MethodType;)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/equals(Ljava/lang/invoke/MethodType;)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MethodType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isViewableAs(Ljava/lang/invoke/MethodType;Z)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/isViewableAs(Ljava/lang/invoke/MethodType;Z)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parametersAreViewableAs(Ljava/lang/invoke/MethodType;Z)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/parametersAreViewableAs(Ljava/lang/invoke/MethodType;Z)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConvertibleTo(Ljava/lang/invoke/MethodType;)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/isConvertibleTo(Ljava/lang/invoke/MethodType;)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "explicitCastEquivalentToAsType(Ljava/lang/invoke/MethodType;)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/explicitCastEquivalentToAsType(Ljava/lang/invoke/MethodType;)Z"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "explicitCastEquivalentToAsType(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodType/explicitCastEquivalentToAsType(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canConvertParameters([Ljava/lang/Class;[Ljava/lang/Class;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Class>, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/canConvertParameters([Ljava/lang/Class;[Ljava/lang/Class;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Class>, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "canConvert(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodType/canConvert(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parameterSlotCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/parameterSlotCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "invokers()Ljava/lang/invoke/Invokers;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_Invokers) => void): void;
    public "java/lang/invoke/MethodType/invokers()Ljava/lang/invoke/Invokers;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_Invokers) => void): void;
    public "parameterSlotDepth(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/parameterSlotDepth(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "returnSlotCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MethodType/returnSlotCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "fromMethodDescriptorString(Ljava/lang/String;Ljava/lang/ClassLoader;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_String, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodType/fromMethodDescriptorString(Ljava/lang/String;Ljava/lang/ClassLoader;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_String, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "toMethodDescriptorString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MethodType/toMethodDescriptorString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toFieldDescriptorString(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/invoke/MethodType/toFieldDescriptorString(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodType/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodType/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodType/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "MethodType_init(Ljava/lang/Class;[Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodType/MethodType_init(Ljava/lang/Class;[Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable) => void): void;
    public "readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodType/readResolve()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_IndexOutOfBoundsException extends java_lang_RuntimeException {
    public static "java/lang/IndexOutOfBoundsException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IndexOutOfBoundsException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IndexOutOfBoundsException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_MethodType$ConcurrentWeakInternSet extends java_lang_Object {
    public "java/lang/invoke/MethodType$ConcurrentWeakInternSet/map": java_util_concurrent_ConcurrentMap;
    public "java/lang/invoke/MethodType$ConcurrentWeakInternSet/stale": java_lang_ref_ReferenceQueue;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodType$ConcurrentWeakInternSet/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodType$ConcurrentWeakInternSet/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "add(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MethodType$ConcurrentWeakInternSet/add(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "expungeStaleElements()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodType$ConcurrentWeakInternSet/expungeStaleElements()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_Invokers extends java_lang_Object {
    public "java/lang/invoke/Invokers/targetType": java_lang_invoke_MethodType;
    public "java/lang/invoke/Invokers/invokers": JVMArray<java_lang_invoke_MethodHandle>;
    public static "java/lang/invoke/Invokers/INV_EXACT": number;
    public static "java/lang/invoke/Invokers/INV_GENERIC": number;
    public static "java/lang/invoke/Invokers/INV_BASIC": number;
    public static "java/lang/invoke/Invokers/INV_LIMIT": number;
    public static "java/lang/invoke/Invokers/MH_LINKER_ARG_APPENDED": number;
    public static "java/lang/invoke/Invokers/NF_checkExactType": java_lang_invoke_LambdaForm$NamedFunction;
    public static "java/lang/invoke/Invokers/NF_checkGenericType": java_lang_invoke_LambdaForm$NamedFunction;
    public static "java/lang/invoke/Invokers/NF_getCallSiteTarget": java_lang_invoke_LambdaForm$NamedFunction;
    public static "java/lang/invoke/Invokers/NF_checkCustomized": java_lang_invoke_LambdaForm$NamedFunction;
    public static "java/lang/invoke/Invokers/$assertionsDisabled": number;
    public "<init>(Ljava/lang/invoke/MethodType;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/Invokers/<init>(Ljava/lang/invoke/MethodType;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable) => void): void;
    public "exactInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/Invokers/exactInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "genericInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/Invokers/genericInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "basicInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/Invokers/basicInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "cachedInvoker(I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/Invokers/cachedInvoker(I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "setCachedInvoker(ILjava/lang/invoke/MethodHandle;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/Invokers/setCachedInvoker(ILjava/lang/invoke/MethodHandle;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "makeExactOrGeneralInvoker(Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/Invokers/makeExactOrGeneralInvoker(Z)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "maybeCompileToBytecode(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/Invokers/maybeCompileToBytecode(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "invokeBasicMethod(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/Invokers/invokeBasicMethod(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "checkInvoker(Ljava/lang/invoke/MethodHandle;)Z"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/Invokers/checkInvoker(Ljava/lang/invoke/MethodHandle;)Z"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "spreadInvoker(I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/Invokers/spreadInvoker(I)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public static "impliedRestargType(Ljava/lang/invoke/MethodType;I)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/invoke/Invokers/impliedRestargType(Ljava/lang/invoke/MethodType;I)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/Invokers/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "methodHandleInvokeLinkerMethod(Ljava/lang/String;Ljava/lang/invoke/MethodType;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_MethodType, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/Invokers/methodHandleInvokeLinkerMethod(Ljava/lang/String;Ljava/lang/invoke/MethodType;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_MethodType, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "invokeHandleForm(Ljava/lang/invoke/MethodType;ZI)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/Invokers/invokeHandleForm(Ljava/lang/invoke/MethodType;ZI)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "newWrongMethodTypeException(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/WrongMethodTypeException;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_WrongMethodTypeException) => void): void;
    public static "java/lang/invoke/Invokers/newWrongMethodTypeException(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/WrongMethodTypeException;"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_WrongMethodTypeException) => void): void;
    public static "checkExactType(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/Invokers/checkExactType(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkGenericType(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/Invokers/checkGenericType(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "linkToCallSiteMethod(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/Invokers/linkToCallSiteMethod(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "linkToTargetMethod(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/Invokers/linkToTargetMethod(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "callSiteForm(Ljava/lang/invoke/MethodType;Z)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "java/lang/invoke/Invokers/callSiteForm(Ljava/lang/invoke/MethodType;Z)Ljava/lang/invoke/LambdaForm;"(thread: JVMThread, args: [java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_LambdaForm) => void): void;
    public static "getCallSiteTarget(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/Invokers/getCallSiteTarget(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "checkCustomized(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/Invokers/checkCustomized(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "maybeCustomize(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/Invokers/maybeCustomize(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/Invokers/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_WrongMethodTypeException extends java_lang_RuntimeException {
    public static "java/lang/invoke/WrongMethodTypeException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/WrongMethodTypeException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/WrongMethodTypeException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/WrongMethodTypeException/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/WrongMethodTypeException/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_CallSite extends java_lang_Object {
    public "java/lang/invoke/CallSite/target": java_lang_invoke_MethodHandle;
    public static "java/lang/invoke/CallSite/GET_TARGET": java_lang_invoke_MethodHandle;
    public static "java/lang/invoke/CallSite/THROW_UCS": java_lang_invoke_MethodHandle;
    public static "java/lang/invoke/CallSite/TARGET_OFFSET": Long;
    public "<init>(Ljava/lang/invoke/MethodType;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/CallSite/<init>(Ljava/lang/invoke/MethodType;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/CallSite/<init>(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/CallSite/<init>(Ljava/lang/invoke/MethodType;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodType, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "type()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/CallSite/type()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "getTarget()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/CallSite/getTarget()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "setTarget(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/CallSite/setTarget(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkTargetChange(Ljava/lang/invoke/MethodHandle;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/CallSite/checkTargetChange(Ljava/lang/invoke/MethodHandle;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "wrongTargetType(Ljava/lang/invoke/MethodHandle;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/WrongMethodTypeException;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_WrongMethodTypeException) => void): void;
    public static "java/lang/invoke/CallSite/wrongTargetType(Ljava/lang/invoke/MethodHandle;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/WrongMethodTypeException;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_WrongMethodTypeException) => void): void;
    public "dynamicInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/CallSite/dynamicInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "makeDynamicInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/CallSite/makeDynamicInvoker()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public static "uninitializedCallSite([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/CallSite/uninitializedCallSite([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "makeUninitializedCallSite(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/CallSite/makeUninitializedCallSite(Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "setTargetNormal(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/CallSite/setTargetNormal(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTargetVolatile()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "java/lang/invoke/CallSite/getTargetVolatile()Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public "setTargetVolatile(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/CallSite/setTargetVolatile(Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "makeSite(Ljava/lang/invoke/MethodHandle;Ljava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Object;Ljava/lang/Class;)Ljava/lang/invoke/CallSite;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_String, java_lang_invoke_MethodType, java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_CallSite) => void): void;
    public static "java/lang/invoke/CallSite/makeSite(Ljava/lang/invoke/MethodHandle;Ljava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Object;Ljava/lang/Class;)Ljava/lang/invoke/CallSite;"(thread: JVMThread, args: [java_lang_invoke_MethodHandle, java_lang_String, java_lang_invoke_MethodType, java_lang_Object, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_CallSite) => void): void;
    public static "maybeReBox(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/CallSite/maybeReBox(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "maybeReBoxElements([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/CallSite/maybeReBoxElements([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/CallSite/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_MemberName extends java_lang_Object implements java_lang_reflect_Member, java_lang_Cloneable {
    public vmtarget: (thread: JVMThread, descriptor: string, args: any[], cb?: (e?: JVMTypes.java_lang_Throwable, rv?: any) => void) => void;
    public vmindex: number;
    public "java/lang/invoke/MemberName/clazz": java_lang_Class;
    public "java/lang/invoke/MemberName/name": java_lang_String;
    public "java/lang/invoke/MemberName/type": java_lang_Object;
    public "java/lang/invoke/MemberName/flags": number;
    public "java/lang/invoke/MemberName/resolution": java_lang_Object;
    public static "java/lang/invoke/MemberName/MH_INVOKE_MODS": number;
    public static "java/lang/invoke/MemberName/BRIDGE": number;
    public static "java/lang/invoke/MemberName/VARARGS": number;
    public static "java/lang/invoke/MemberName/SYNTHETIC": number;
    public static "java/lang/invoke/MemberName/ANNOTATION": number;
    public static "java/lang/invoke/MemberName/ENUM": number;
    public static "java/lang/invoke/MemberName/CONSTRUCTOR_NAME": java_lang_String;
    public static "java/lang/invoke/MemberName/RECOGNIZED_MODIFIERS": number;
    public static "java/lang/invoke/MemberName/IS_METHOD": number;
    public static "java/lang/invoke/MemberName/IS_CONSTRUCTOR": number;
    public static "java/lang/invoke/MemberName/IS_FIELD": number;
    public static "java/lang/invoke/MemberName/IS_TYPE": number;
    public static "java/lang/invoke/MemberName/CALLER_SENSITIVE": number;
    public static "java/lang/invoke/MemberName/ALL_ACCESS": number;
    public static "java/lang/invoke/MemberName/ALL_KINDS": number;
    public static "java/lang/invoke/MemberName/IS_INVOCABLE": number;
    public static "java/lang/invoke/MemberName/IS_FIELD_OR_METHOD": number;
    public static "java/lang/invoke/MemberName/SEARCH_ALL_SUPERS": number;
    public static "java/lang/invoke/MemberName/$assertionsDisabled": number;
    public "getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MemberName/getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/lang/invoke/MemberName/getClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MemberName/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getMethodOrFieldType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MemberName/getMethodOrFieldType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "getMethodType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MemberName/getMethodType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "getInvocationType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "java/lang/invoke/MemberName/getInvocationType()Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public "getParameterTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/invoke/MemberName/getParameterTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getReturnType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MemberName/getReturnType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getFieldType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/invoke/MemberName/getFieldType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getType()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MemberName/getType()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MemberName/getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getReferenceKind()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/getReferenceKind()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "referenceKindIsConsistent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/referenceKindIsConsistent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isObjectPublicMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isObjectPublicMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "referenceKindIsConsistentWith(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/referenceKindIsConsistentWith(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "staticIsConsistent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/staticIsConsistent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "vminfoIsConsistent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/vminfoIsConsistent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "changeReferenceKind(BB)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName/changeReferenceKind(BB)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "testFlags(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/testFlags(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "testAllFlags(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/testAllFlags(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "testAnyFlags(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/testAnyFlags(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMethodHandleInvoke()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isMethodHandleInvoke()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isMethodHandleInvokeName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MemberName/isMethodHandleInvokeName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isStatic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isStatic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPublic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isPublic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPrivate()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isPrivate()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isProtected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isProtected()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFinal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isFinal()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canBeStaticallyBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/canBeStaticallyBound()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isVolatile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isVolatile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAbstract()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isAbstract()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isNative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isNative()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isBridge()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isBridge()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isVarargs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isVarargs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInvocable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isInvocable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFieldOrMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isFieldOrMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConstructor()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isConstructor()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isField()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isField()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isType()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPackage()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isPackage()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isCallerSensitive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isCallerSensitive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAccessibleFrom(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isAccessibleFrom(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "init(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/init(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "expandFromVM()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/expandFromVM()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "flagsMods(IIB)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MemberName/flagsMods(IIB)I"(thread: JVMThread, args: [number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/reflect/Method;Z)V"(thread: JVMThread, args: [java_lang_reflect_Method, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>(Ljava/lang/reflect/Method;Z)V"(thread: JVMThread, args: [java_lang_reflect_Method, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "asSpecial()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName/asSpecial()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "asConstructor()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName/asConstructor()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "asNormalOriginal()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName/asNormalOriginal()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "<init>(Ljava/lang/reflect/Constructor;)V"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>(Ljava/lang/reflect/Constructor;)V"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/reflect/Field;)V"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>(Ljava/lang/reflect/Field;)V"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/reflect/Field;Z)V"(thread: JVMThread, args: [java_lang_reflect_Field, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>(Ljava/lang/reflect/Field;Z)V"(thread: JVMThread, args: [java_lang_reflect_Field, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isGetter()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isGetter()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSetter()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isSetter()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "asSetter()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName/asSetter()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "makeMethodHandleInvoke(Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/MemberName/makeMethodHandleInvoke(Ljava/lang/String;Ljava/lang/invoke/MethodType;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_MethodType], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "makeMethodHandleInvoke(Ljava/lang/String;Ljava/lang/invoke/MethodType;I)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/MemberName/makeMethodHandleInvoke(Ljava/lang/String;Ljava/lang/invoke/MethodType;I)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_String, java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName/clone()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "getDefinition()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "java/lang/invoke/MemberName/getDefinition()Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/equals(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;B)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;B)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;B)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/invoke/MethodType;B)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_invoke_MethodType, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(BLjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Class, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/<init>(BLjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Class, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "hasReceiverTypeDispatch()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/hasReceiverTypeDispatch()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isResolved()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/invoke/MemberName/isResolved()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "initResolved(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/initResolved(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkForTypeAlias()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MemberName/checkForTypeAlias()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MemberName/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getName(Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/invoke/MemberName/getName(Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "makeAccessException(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/IllegalAccessException;"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_IllegalAccessException) => void): void;
    public "java/lang/invoke/MemberName/makeAccessException(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/IllegalAccessException;"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_IllegalAccessException) => void): void;
    public "message()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/invoke/MemberName/message()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "makeAccessException()Ljava/lang/ReflectiveOperationException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ReflectiveOperationException) => void): void;
    public "java/lang/invoke/MemberName/makeAccessException()Ljava/lang/ReflectiveOperationException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ReflectiveOperationException) => void): void;
    public static "getFactory()Ljava/lang/invoke/MemberName$Factory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName$Factory) => void): void;
    public static "java/lang/invoke/MemberName/getFactory()Ljava/lang/invoke/MemberName$Factory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName$Factory) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/invoke/MemberName/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$002(Ljava/lang/invoke/MemberName;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/MemberName/access$002(Ljava/lang/invoke/MemberName;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$100(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MemberName/access$100(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Ljava/lang/invoke/MemberName;Z)V"(thread: JVMThread, args: [java_lang_invoke_MemberName, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MemberName/access$200(Ljava/lang/invoke/MemberName;Z)V"(thread: JVMThread, args: [java_lang_invoke_MemberName, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$300(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MemberName/access$300(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MemberName/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ReflectiveOperationException extends java_lang_Exception {
    public static "java/lang/ReflectiveOperationException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ReflectiveOperationException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ReflectiveOperationException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ReflectiveOperationException/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ReflectiveOperationException/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_IllegalAccessException extends java_lang_ReflectiveOperationException {
    public static "java/lang/IllegalAccessException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalAccessException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/IllegalAccessException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_MethodHandleNatives extends java_lang_Object {
    public static "java/lang/invoke/MethodHandleNatives/COUNT_GWT": number;
    public static "java/lang/invoke/MethodHandleNatives/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandleNatives/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "init(Ljava/lang/invoke/MemberName;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/init(Ljava/lang/invoke/MemberName;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "expand(Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/expand(Ljava/lang/invoke/MemberName;)V"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable) => void): void;
    public static "resolve(Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/resolve(Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "getMembers(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/String;ILjava/lang/Class;I[Ljava/lang/invoke/MemberName;)I"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_String, number, java_lang_Class, number, JVMArray<java_lang_invoke_MemberName>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/getMembers(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/String;ILjava/lang/Class;I[Ljava/lang/invoke/MemberName;)I"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_String, number, java_lang_Class, number, JVMArray<java_lang_invoke_MemberName>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "objectFieldOffset(Ljava/lang/invoke/MemberName;)J"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/objectFieldOffset(Ljava/lang/invoke/MemberName;)J"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "staticFieldOffset(Ljava/lang/invoke/MemberName;)J"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/staticFieldOffset(Ljava/lang/invoke/MemberName;)J"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "staticFieldBase(Ljava/lang/invoke/MemberName;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/staticFieldBase(Ljava/lang/invoke/MemberName;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "getMemberVMInfo(Ljava/lang/invoke/MemberName;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/getMemberVMInfo(Ljava/lang/invoke/MemberName;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "getConstant(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/getConstant(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "setCallSiteTargetNormal(Ljava/lang/invoke/CallSite;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_CallSite, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/setCallSiteTargetNormal(Ljava/lang/invoke/CallSite;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_CallSite, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setCallSiteTargetVolatile(Ljava/lang/invoke/CallSite;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_CallSite, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/setCallSiteTargetVolatile(Ljava/lang/invoke/CallSite;Ljava/lang/invoke/MethodHandle;)V"(thread: JVMThread, args: [java_lang_invoke_CallSite, java_lang_invoke_MethodHandle], cb?: (e?: java_lang_Throwable) => void): void;
    public static "registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "refKindIsValid(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindIsValid(I)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "refKindIsField(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindIsField(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "refKindIsGetter(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindIsGetter(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "refKindIsSetter(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindIsSetter(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "refKindIsMethod(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindIsMethod(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "refKindIsConstructor(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindIsConstructor(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "refKindHasReceiver(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindHasReceiver(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "refKindIsStatic(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindIsStatic(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "refKindDoesDispatch(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindDoesDispatch(B)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "refKindName(B)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/refKindName(B)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getNamedCon(I[Ljava/lang/Object;)I"(thread: JVMThread, args: [number, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/getNamedCon(I[Ljava/lang/Object;)I"(thread: JVMThread, args: [number, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "verifyConstants()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/verifyConstants()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "linkCallSite(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object, java_lang_Object, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/linkCallSite(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object, java_lang_Object, java_lang_Object, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "linkCallSiteImpl(Ljava/lang/Class;Ljava/lang/invoke/MethodHandle;Ljava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MethodHandle, java_lang_String, java_lang_invoke_MethodType, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/linkCallSiteImpl(Ljava/lang/Class;Ljava/lang/invoke/MethodHandle;Ljava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MethodHandle, java_lang_String, java_lang_invoke_MethodType, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "linkCallSiteTracing(Ljava/lang/Class;Ljava/lang/invoke/MethodHandle;Ljava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MethodHandle, java_lang_String, java_lang_invoke_MethodType, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/linkCallSiteTracing(Ljava/lang/Class;Ljava/lang/invoke/MethodHandle;Ljava/lang/String;Ljava/lang/invoke/MethodType;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, java_lang_invoke_MethodHandle, java_lang_String, java_lang_invoke_MethodType, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "findMethodHandleType(Ljava/lang/Class;[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/findMethodHandleType(Ljava/lang/Class;[Ljava/lang/Class;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "linkMethod(Ljava/lang/Class;ILjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class, java_lang_String, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/linkMethod(Ljava/lang/Class;ILjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class, java_lang_String, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "linkMethodImpl(Ljava/lang/Class;ILjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class, java_lang_String, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/linkMethodImpl(Ljava/lang/Class;ILjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class, java_lang_String, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "fixMethodType(Ljava/lang/Class;Ljava/lang/Object;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/fixMethodType(Ljava/lang/Class;Ljava/lang/Object;)Ljava/lang/invoke/MethodType;"(thread: JVMThread, args: [java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodType) => void): void;
    public static "linkMethodTracing(Ljava/lang/Class;ILjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class, java_lang_String, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/linkMethodTracing(Ljava/lang/Class;ILjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/invoke/MemberName;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class, java_lang_String, java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MemberName) => void): void;
    public static "linkMethodHandleConstant(Ljava/lang/Class;ILjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/linkMethodHandleConstant(Ljava/lang/Class;ILjava/lang/Class;Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/invoke/MethodHandle;"(thread: JVMThread, args: [java_lang_Class, number, java_lang_Class, java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_invoke_MethodHandle) => void): void;
    public static "initCauseFrom(Ljava/lang/Error;Ljava/lang/Exception;)Ljava/lang/Error;"(thread: JVMThread, args: [java_lang_Error, java_lang_Exception], cb?: (e?: java_lang_Throwable, rv?: java_lang_Error) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/initCauseFrom(Ljava/lang/Error;Ljava/lang/Exception;)Ljava/lang/Error;"(thread: JVMThread, args: [java_lang_Error, java_lang_Exception], cb?: (e?: java_lang_Throwable, rv?: java_lang_Error) => void): void;
    public static "isCallerSensitive(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/isCallerSensitive(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "canBeCalledVirtual(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/canBeCalledVirtual(Ljava/lang/invoke/MemberName;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "canBeCalledVirtual(Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/canBeCalledVirtual(Ljava/lang/invoke/MemberName;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_invoke_MemberName, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/invoke/MethodHandleNatives/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Error extends java_lang_Throwable {
    public static "java/lang/Error/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Error/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Error/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Error/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Error/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;ZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Error/<init>(Ljava/lang/String;Ljava/lang/Throwable;ZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ref_Reference extends java_lang_Object {
    public "java/lang/ref/Reference/referent": java_lang_Object;
    public "java/lang/ref/Reference/queue": java_lang_ref_ReferenceQueue;
    public "java/lang/ref/Reference/next": java_lang_ref_Reference;
    public "java/lang/ref/Reference/discovered": java_lang_ref_Reference;
    public static "java/lang/ref/Reference/lock": java_lang_ref_Reference$Lock;
    public static "java/lang/ref/Reference/pending": java_lang_ref_Reference;
    public static "tryHandlePending(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ref/Reference/tryHandlePending(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ref/Reference/get()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/Reference/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isEnqueued()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ref/Reference/isEnqueued()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "enqueue()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ref/Reference/enqueue()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/Reference/<init>(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_ref_ReferenceQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/Reference/<init>(Ljava/lang/Object;Ljava/lang/ref/ReferenceQueue;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_ref_ReferenceQueue], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ref/Reference/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ref_Reference$Lock extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/Reference$Lock/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ref/Reference$1;)V"(thread: JVMThread, args: [java_lang_ref_Reference$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/Reference$Lock/<init>(Ljava/lang/ref/Reference$1;)V"(thread: JVMThread, args: [java_lang_ref_Reference$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ref_Reference$1 extends java_lang_Object implements sun_misc_JavaLangRefAccess {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ref/Reference$1/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "tryHandlePendingReference()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ref/Reference$1/tryHandlePendingReference()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export interface sun_misc_JavaLangRefAccess extends java_lang_Object {
    "tryHandlePendingReference()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_lang_reflect_Parameter extends java_lang_Object implements java_lang_reflect_AnnotatedElement {
    public "java/lang/reflect/Parameter/name": java_lang_String;
    public "java/lang/reflect/Parameter/modifiers": number;
    public "java/lang/reflect/Parameter/executable": java_lang_reflect_Executable;
    public "java/lang/reflect/Parameter/index": number;
    public "java/lang/reflect/Parameter/parameterTypeCache": java_lang_reflect_Type;
    public "java/lang/reflect/Parameter/parameterClassCache": java_lang_Class;
    public "java/lang/reflect/Parameter/declaredAnnotations": java_util_Map;
    public "<init>(Ljava/lang/String;ILjava/lang/reflect/Executable;I)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_reflect_Executable, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Parameter/<init>(Ljava/lang/String;ILjava/lang/reflect/Executable;I)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_reflect_Executable, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Parameter/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Parameter/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isNamePresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Parameter/isNamePresent()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Parameter/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDeclaringExecutable()Ljava/lang/reflect/Executable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Executable) => void): void;
    public "java/lang/reflect/Parameter/getDeclaringExecutable()Ljava/lang/reflect/Executable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Executable) => void): void;
    public "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Parameter/getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Parameter/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRealName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Parameter/getRealName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getParameterizedType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "java/lang/reflect/Parameter/getParameterizedType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "getType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/reflect/Parameter/getType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getAnnotatedType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "java/lang/reflect/Parameter/getAnnotatedType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "isImplicit()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Parameter/isImplicit()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Parameter/isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isVarArgs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Parameter/isVarArgs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/reflect/Parameter/getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Parameter/getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Parameter/getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/reflect/Parameter/getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Parameter/getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Parameter/getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "declaredAnnotations()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/lang/reflect/Parameter/declaredAnnotations()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
  }
  export interface java_lang_reflect_AnnotatedType extends java_lang_Object, java_lang_reflect_AnnotatedElement {
    "getType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
  }
  export interface java_lang_reflect_Type extends java_lang_Object {
    "getTypeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_lang_reflect_AnnotatedElement extends java_lang_Object {
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
  }
  export class java_lang_reflect_Executable extends java_lang_reflect_AccessibleObject implements java_lang_reflect_Member, java_lang_reflect_GenericDeclaration {
    public "java/lang/reflect/Executable/hasRealParameterData": number;
    public "java/lang/reflect/Executable/parameters": JVMArray<java_lang_reflect_Parameter>;
    public "java/lang/reflect/Executable/declaredAnnotations": java_util_Map;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Executable/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAnnotationBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/reflect/Executable/getAnnotationBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getRoot()Ljava/lang/reflect/Executable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Executable) => void): void;
    public "java/lang/reflect/Executable/getRoot()Ljava/lang/reflect/Executable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Executable) => void): void;
    public "hasGenericInformation()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Executable/hasGenericInformation()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGenericInfo()Lsun/reflect/generics/repository/ConstructorRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ConstructorRepository) => void): void;
    public "java/lang/reflect/Executable/getGenericInfo()Lsun/reflect/generics/repository/ConstructorRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ConstructorRepository) => void): void;
    public "equalParamTypes([Ljava/lang/Class;[Ljava/lang/Class;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Class>, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Executable/equalParamTypes([Ljava/lang/Class;[Ljava/lang/Class;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Class>, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "parseParameterAnnotations([B)[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "java/lang/reflect/Executable/parseParameterAnnotations([B)[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "separateWithCommas([Ljava/lang/Class;Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [JVMArray<java_lang_Class>, java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Executable/separateWithCommas([Ljava/lang/Class;Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [JVMArray<java_lang_Class>, java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "printModifiersIfNonzero(Ljava/lang/StringBuilder;IZ)V"(thread: JVMThread, args: [java_lang_StringBuilder, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Executable/printModifiersIfNonzero(Ljava/lang/StringBuilder;IZ)V"(thread: JVMThread, args: [java_lang_StringBuilder, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sharedToString(IZ[Ljava/lang/Class;[Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [number, number, JVMArray<java_lang_Class>, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Executable/sharedToString(IZ[Ljava/lang/Class;[Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [number, number, JVMArray<java_lang_Class>, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "specificToStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Executable/specificToStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "sharedToGenericString(IZ)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Executable/sharedToGenericString(IZ)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "specificToGenericStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Executable/specificToGenericStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/reflect/Executable/getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Executable/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Executable/getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    public "java/lang/reflect/Executable/getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    public "getParameterTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/reflect/Executable/getParameterTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Executable/getParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGenericParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "java/lang/reflect/Executable/getGenericParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "getAllGenericParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "java/lang/reflect/Executable/getAllGenericParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "getParameters()[Ljava/lang/reflect/Parameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Parameter>) => void): void;
    public "java/lang/reflect/Executable/getParameters()[Ljava/lang/reflect/Parameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Parameter>) => void): void;
    public "synthesizeAllParams()[Ljava/lang/reflect/Parameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Parameter>) => void): void;
    public "java/lang/reflect/Executable/synthesizeAllParams()[Ljava/lang/reflect/Parameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Parameter>) => void): void;
    public "verifyParameters([Ljava/lang/reflect/Parameter;)V"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Parameter>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Executable/verifyParameters([Ljava/lang/reflect/Parameter;)V"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Parameter>], cb?: (e?: java_lang_Throwable) => void): void;
    public "privateGetParameters()[Ljava/lang/reflect/Parameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Parameter>) => void): void;
    public "java/lang/reflect/Executable/privateGetParameters()[Ljava/lang/reflect/Parameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Parameter>) => void): void;
    public "hasRealParameterData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Executable/hasRealParameterData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getParameters0()[Ljava/lang/reflect/Parameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Parameter>) => void): void;
    public "java/lang/reflect/Executable/getParameters0()[Ljava/lang/reflect/Parameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Parameter>) => void): void;
    public "getTypeAnnotationBytes0()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/reflect/Executable/getTypeAnnotationBytes0()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getTypeAnnotationBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/reflect/Executable/getTypeAnnotationBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getExceptionTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/reflect/Executable/getExceptionTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getGenericExceptionTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "java/lang/reflect/Executable/getGenericExceptionTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Executable/toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isVarArgs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Executable/isVarArgs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Executable/isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getParameterAnnotations()[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "java/lang/reflect/Executable/getParameterAnnotations()[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "sharedGetParameterAnnotations([Ljava/lang/Class;[B)[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "java/lang/reflect/Executable/sharedGetParameterAnnotations([Ljava/lang/Class;[B)[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "handleParameterNumberMismatch(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Executable/handleParameterNumberMismatch(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/reflect/Executable/getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Executable/getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Executable/getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "declaredAnnotations()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/lang/reflect/Executable/declaredAnnotations()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getAnnotatedReturnType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "java/lang/reflect/Executable/getAnnotatedReturnType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "getAnnotatedReturnType0(Ljava/lang/reflect/Type;)Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: [java_lang_reflect_Type], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "java/lang/reflect/Executable/getAnnotatedReturnType0(Ljava/lang/reflect/Type;)Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: [java_lang_reflect_Type], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "getAnnotatedReceiverType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "java/lang/reflect/Executable/getAnnotatedReceiverType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "getAnnotatedParameterTypes()[Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_AnnotatedType>) => void): void;
    public "java/lang/reflect/Executable/getAnnotatedParameterTypes()[Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_AnnotatedType>) => void): void;
    public "getAnnotatedExceptionTypes()[Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_AnnotatedType>) => void): void;
    public "java/lang/reflect/Executable/getAnnotatedExceptionTypes()[Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_AnnotatedType>) => void): void;
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
  }
  export interface java_lang_reflect_TypeVariable extends java_lang_Object, java_lang_reflect_Type, java_lang_reflect_AnnotatedElement {
    "getBounds()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    "getGenericDeclaration()Ljava/lang/reflect/GenericDeclaration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_GenericDeclaration) => void): void;
    "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getAnnotatedBounds()[Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_AnnotatedType>) => void): void;
    "getTypeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
  }
  export class sun_reflect_generics_repository_ConstructorRepository extends sun_reflect_generics_repository_GenericDeclRepository {
    public "sun/reflect/generics/repository/ConstructorRepository/paramTypes": JVMArray<java_lang_reflect_Type>;
    public "sun/reflect/generics/repository/ConstructorRepository/exceptionTypes": JVMArray<java_lang_reflect_Type>;
    public "<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/repository/ConstructorRepository/<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "parse(Ljava/lang/String;)Lsun/reflect/generics/tree/MethodTypeSignature;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_MethodTypeSignature) => void): void;
    public "sun/reflect/generics/repository/ConstructorRepository/parse(Ljava/lang/String;)Lsun/reflect/generics/tree/MethodTypeSignature;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_MethodTypeSignature) => void): void;
    public static "make(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/repository/ConstructorRepository;"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ConstructorRepository) => void): void;
    public static "sun/reflect/generics/repository/ConstructorRepository/make(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/repository/ConstructorRepository;"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ConstructorRepository) => void): void;
    public "getParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "sun/reflect/generics/repository/ConstructorRepository/getParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "getExceptionTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "sun/reflect/generics/repository/ConstructorRepository/getExceptionTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "parse(Ljava/lang/String;)Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
    public "sun/reflect/generics/repository/ConstructorRepository/parse(Ljava/lang/String;)Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
  }
  export interface sun_reflect_generics_tree_Tree extends java_lang_Object {
  }
  export class sun_reflect_generics_tree_MethodTypeSignature extends java_lang_Object implements sun_reflect_generics_tree_Signature {
    public "sun/reflect/generics/tree/MethodTypeSignature/formalTypeParams": JVMArray<sun_reflect_generics_tree_FormalTypeParameter>;
    public "sun/reflect/generics/tree/MethodTypeSignature/parameterTypes": JVMArray<sun_reflect_generics_tree_TypeSignature>;
    public "sun/reflect/generics/tree/MethodTypeSignature/returnType": sun_reflect_generics_tree_ReturnType;
    public "sun/reflect/generics/tree/MethodTypeSignature/exceptionTypes": JVMArray<sun_reflect_generics_tree_FieldTypeSignature>;
    public "<init>([Lsun/reflect/generics/tree/FormalTypeParameter;[Lsun/reflect/generics/tree/TypeSignature;Lsun/reflect/generics/tree/ReturnType;[Lsun/reflect/generics/tree/FieldTypeSignature;)V"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FormalTypeParameter>, JVMArray<sun_reflect_generics_tree_TypeSignature>, sun_reflect_generics_tree_ReturnType, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/MethodTypeSignature/<init>([Lsun/reflect/generics/tree/FormalTypeParameter;[Lsun/reflect/generics/tree/TypeSignature;Lsun/reflect/generics/tree/ReturnType;[Lsun/reflect/generics/tree/FieldTypeSignature;)V"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FormalTypeParameter>, JVMArray<sun_reflect_generics_tree_TypeSignature>, sun_reflect_generics_tree_ReturnType, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make([Lsun/reflect/generics/tree/FormalTypeParameter;[Lsun/reflect/generics/tree/TypeSignature;Lsun/reflect/generics/tree/ReturnType;[Lsun/reflect/generics/tree/FieldTypeSignature;)Lsun/reflect/generics/tree/MethodTypeSignature;"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FormalTypeParameter>, JVMArray<sun_reflect_generics_tree_TypeSignature>, sun_reflect_generics_tree_ReturnType, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_MethodTypeSignature) => void): void;
    public static "sun/reflect/generics/tree/MethodTypeSignature/make([Lsun/reflect/generics/tree/FormalTypeParameter;[Lsun/reflect/generics/tree/TypeSignature;Lsun/reflect/generics/tree/ReturnType;[Lsun/reflect/generics/tree/FieldTypeSignature;)Lsun/reflect/generics/tree/MethodTypeSignature;"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FormalTypeParameter>, JVMArray<sun_reflect_generics_tree_TypeSignature>, sun_reflect_generics_tree_ReturnType, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_MethodTypeSignature) => void): void;
    public "getFormalTypeParameters()[Lsun/reflect/generics/tree/FormalTypeParameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FormalTypeParameter>) => void): void;
    public "sun/reflect/generics/tree/MethodTypeSignature/getFormalTypeParameters()[Lsun/reflect/generics/tree/FormalTypeParameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FormalTypeParameter>) => void): void;
    public "getParameterTypes()[Lsun/reflect/generics/tree/TypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_TypeSignature>) => void): void;
    public "sun/reflect/generics/tree/MethodTypeSignature/getParameterTypes()[Lsun/reflect/generics/tree/TypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_TypeSignature>) => void): void;
    public "getReturnType()Lsun/reflect/generics/tree/ReturnType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ReturnType) => void): void;
    public "sun/reflect/generics/tree/MethodTypeSignature/getReturnType()Lsun/reflect/generics/tree/ReturnType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ReturnType) => void): void;
    public "getExceptionTypes()[Lsun/reflect/generics/tree/FieldTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FieldTypeSignature>) => void): void;
    public "sun/reflect/generics/tree/MethodTypeSignature/getExceptionTypes()[Lsun/reflect/generics/tree/FieldTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FieldTypeSignature>) => void): void;
    public "accept(Lsun/reflect/generics/visitor/Visitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_Visitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/MethodTypeSignature/accept(Lsun/reflect/generics/visitor/Visitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_Visitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_generics_visitor_Visitor extends java_lang_Object, sun_reflect_generics_visitor_TypeTreeVisitor {
    "visitClassSignature(Lsun/reflect/generics/tree/ClassSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ClassSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitMethodTypeSignature(Lsun/reflect/generics/tree/MethodTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_MethodTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "getResult()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "visitFormalTypeParameter(Lsun/reflect/generics/tree/FormalTypeParameter;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_FormalTypeParameter], cb?: (e?: java_lang_Throwable) => void): void;
    "visitClassTypeSignature(Lsun/reflect/generics/tree/ClassTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ClassTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitArrayTypeSignature(Lsun/reflect/generics/tree/ArrayTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ArrayTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitTypeVariableSignature(Lsun/reflect/generics/tree/TypeVariableSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_TypeVariableSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitWildcard(Lsun/reflect/generics/tree/Wildcard;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_Wildcard], cb?: (e?: java_lang_Throwable) => void): void;
    "visitSimpleClassTypeSignature(Lsun/reflect/generics/tree/SimpleClassTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_SimpleClassTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitBottomSignature(Lsun/reflect/generics/tree/BottomSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_BottomSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitByteSignature(Lsun/reflect/generics/tree/ByteSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ByteSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitBooleanSignature(Lsun/reflect/generics/tree/BooleanSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_BooleanSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitShortSignature(Lsun/reflect/generics/tree/ShortSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ShortSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitCharSignature(Lsun/reflect/generics/tree/CharSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_CharSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitIntSignature(Lsun/reflect/generics/tree/IntSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_IntSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitLongSignature(Lsun/reflect/generics/tree/LongSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_LongSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitFloatSignature(Lsun/reflect/generics/tree/FloatSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_FloatSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitDoubleSignature(Lsun/reflect/generics/tree/DoubleSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_DoubleSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitVoidDescriptor(Lsun/reflect/generics/tree/VoidDescriptor;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_VoidDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_VoidDescriptor extends java_lang_Object implements sun_reflect_generics_tree_ReturnType {
    public static "sun/reflect/generics/tree/VoidDescriptor/singleton": sun_reflect_generics_tree_VoidDescriptor;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/VoidDescriptor/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/VoidDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_VoidDescriptor) => void): void;
    public static "sun/reflect/generics/tree/VoidDescriptor/make()Lsun/reflect/generics/tree/VoidDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_VoidDescriptor) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/VoidDescriptor/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/VoidDescriptor/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_DoubleSignature extends java_lang_Object implements sun_reflect_generics_tree_BaseType {
    public static "sun/reflect/generics/tree/DoubleSignature/singleton": sun_reflect_generics_tree_DoubleSignature;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/DoubleSignature/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/DoubleSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_DoubleSignature) => void): void;
    public static "sun/reflect/generics/tree/DoubleSignature/make()Lsun/reflect/generics/tree/DoubleSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_DoubleSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/DoubleSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/DoubleSignature/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_generics_tree_BaseType extends java_lang_Object, sun_reflect_generics_tree_TypeSignature {
    "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_FloatSignature extends java_lang_Object implements sun_reflect_generics_tree_BaseType {
    public static "sun/reflect/generics/tree/FloatSignature/singleton": sun_reflect_generics_tree_FloatSignature;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/FloatSignature/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/FloatSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_FloatSignature) => void): void;
    public static "sun/reflect/generics/tree/FloatSignature/make()Lsun/reflect/generics/tree/FloatSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_FloatSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/FloatSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/FloatSignature/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_LongSignature extends java_lang_Object implements sun_reflect_generics_tree_BaseType {
    public static "sun/reflect/generics/tree/LongSignature/singleton": sun_reflect_generics_tree_LongSignature;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/LongSignature/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/LongSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_LongSignature) => void): void;
    public static "sun/reflect/generics/tree/LongSignature/make()Lsun/reflect/generics/tree/LongSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_LongSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/LongSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/LongSignature/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_IntSignature extends java_lang_Object implements sun_reflect_generics_tree_BaseType {
    public static "sun/reflect/generics/tree/IntSignature/singleton": sun_reflect_generics_tree_IntSignature;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/IntSignature/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/IntSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_IntSignature) => void): void;
    public static "sun/reflect/generics/tree/IntSignature/make()Lsun/reflect/generics/tree/IntSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_IntSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/IntSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/IntSignature/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_CharSignature extends java_lang_Object implements sun_reflect_generics_tree_BaseType {
    public static "sun/reflect/generics/tree/CharSignature/singleton": sun_reflect_generics_tree_CharSignature;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/CharSignature/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/CharSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_CharSignature) => void): void;
    public static "sun/reflect/generics/tree/CharSignature/make()Lsun/reflect/generics/tree/CharSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_CharSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/CharSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/CharSignature/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_ShortSignature extends java_lang_Object implements sun_reflect_generics_tree_BaseType {
    public static "sun/reflect/generics/tree/ShortSignature/singleton": sun_reflect_generics_tree_ShortSignature;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ShortSignature/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/ShortSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ShortSignature) => void): void;
    public static "sun/reflect/generics/tree/ShortSignature/make()Lsun/reflect/generics/tree/ShortSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ShortSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ShortSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/ShortSignature/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_BooleanSignature extends java_lang_Object implements sun_reflect_generics_tree_BaseType {
    public static "sun/reflect/generics/tree/BooleanSignature/singleton": sun_reflect_generics_tree_BooleanSignature;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/BooleanSignature/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/BooleanSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_BooleanSignature) => void): void;
    public static "sun/reflect/generics/tree/BooleanSignature/make()Lsun/reflect/generics/tree/BooleanSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_BooleanSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/BooleanSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/BooleanSignature/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_ByteSignature extends java_lang_Object implements sun_reflect_generics_tree_BaseType {
    public static "sun/reflect/generics/tree/ByteSignature/singleton": sun_reflect_generics_tree_ByteSignature;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ByteSignature/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/ByteSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ByteSignature) => void): void;
    public static "sun/reflect/generics/tree/ByteSignature/make()Lsun/reflect/generics/tree/ByteSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ByteSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ByteSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/ByteSignature/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_BottomSignature extends java_lang_Object implements sun_reflect_generics_tree_FieldTypeSignature {
    public static "sun/reflect/generics/tree/BottomSignature/singleton": sun_reflect_generics_tree_BottomSignature;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/BottomSignature/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make()Lsun/reflect/generics/tree/BottomSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_BottomSignature) => void): void;
    public static "sun/reflect/generics/tree/BottomSignature/make()Lsun/reflect/generics/tree/BottomSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_BottomSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/BottomSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/BottomSignature/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_SimpleClassTypeSignature extends java_lang_Object implements sun_reflect_generics_tree_FieldTypeSignature {
    public "sun/reflect/generics/tree/SimpleClassTypeSignature/dollar": number;
    public "sun/reflect/generics/tree/SimpleClassTypeSignature/name": java_lang_String;
    public "sun/reflect/generics/tree/SimpleClassTypeSignature/typeArgs": JVMArray<sun_reflect_generics_tree_TypeArgument>;
    public "<init>(Ljava/lang/String;Z[Lsun/reflect/generics/tree/TypeArgument;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<sun_reflect_generics_tree_TypeArgument>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/SimpleClassTypeSignature/<init>(Ljava/lang/String;Z[Lsun/reflect/generics/tree/TypeArgument;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<sun_reflect_generics_tree_TypeArgument>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make(Ljava/lang/String;Z[Lsun/reflect/generics/tree/TypeArgument;)Lsun/reflect/generics/tree/SimpleClassTypeSignature;"(thread: JVMThread, args: [java_lang_String, number, JVMArray<sun_reflect_generics_tree_TypeArgument>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_SimpleClassTypeSignature) => void): void;
    public static "sun/reflect/generics/tree/SimpleClassTypeSignature/make(Ljava/lang/String;Z[Lsun/reflect/generics/tree/TypeArgument;)Lsun/reflect/generics/tree/SimpleClassTypeSignature;"(thread: JVMThread, args: [java_lang_String, number, JVMArray<sun_reflect_generics_tree_TypeArgument>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_SimpleClassTypeSignature) => void): void;
    public "getDollar()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/generics/tree/SimpleClassTypeSignature/getDollar()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/reflect/generics/tree/SimpleClassTypeSignature/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTypeArguments()[Lsun/reflect/generics/tree/TypeArgument;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_TypeArgument>) => void): void;
    public "sun/reflect/generics/tree/SimpleClassTypeSignature/getTypeArguments()[Lsun/reflect/generics/tree/TypeArgument;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_TypeArgument>) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/SimpleClassTypeSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_generics_tree_TypeArgument extends java_lang_Object, sun_reflect_generics_tree_TypeTree {
    "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_generics_tree_TypeTree extends java_lang_Object, sun_reflect_generics_tree_Tree {
    "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_Wildcard extends java_lang_Object implements sun_reflect_generics_tree_TypeArgument {
    public "sun/reflect/generics/tree/Wildcard/upperBounds": JVMArray<sun_reflect_generics_tree_FieldTypeSignature>;
    public "sun/reflect/generics/tree/Wildcard/lowerBounds": JVMArray<sun_reflect_generics_tree_FieldTypeSignature>;
    public static "sun/reflect/generics/tree/Wildcard/emptyBounds": JVMArray<sun_reflect_generics_tree_FieldTypeSignature>;
    public "<init>([Lsun/reflect/generics/tree/FieldTypeSignature;[Lsun/reflect/generics/tree/FieldTypeSignature;)V"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FieldTypeSignature>, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/Wildcard/<init>([Lsun/reflect/generics/tree/FieldTypeSignature;[Lsun/reflect/generics/tree/FieldTypeSignature;)V"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FieldTypeSignature>, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make([Lsun/reflect/generics/tree/FieldTypeSignature;[Lsun/reflect/generics/tree/FieldTypeSignature;)Lsun/reflect/generics/tree/Wildcard;"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FieldTypeSignature>, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Wildcard) => void): void;
    public static "sun/reflect/generics/tree/Wildcard/make([Lsun/reflect/generics/tree/FieldTypeSignature;[Lsun/reflect/generics/tree/FieldTypeSignature;)Lsun/reflect/generics/tree/Wildcard;"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FieldTypeSignature>, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Wildcard) => void): void;
    public "getUpperBounds()[Lsun/reflect/generics/tree/FieldTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FieldTypeSignature>) => void): void;
    public "sun/reflect/generics/tree/Wildcard/getUpperBounds()[Lsun/reflect/generics/tree/FieldTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FieldTypeSignature>) => void): void;
    public "getLowerBounds()[Lsun/reflect/generics/tree/FieldTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FieldTypeSignature>) => void): void;
    public "sun/reflect/generics/tree/Wildcard/getLowerBounds()[Lsun/reflect/generics/tree/FieldTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FieldTypeSignature>) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/Wildcard/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/tree/Wildcard/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_TypeVariableSignature extends java_lang_Object implements sun_reflect_generics_tree_FieldTypeSignature {
    public "sun/reflect/generics/tree/TypeVariableSignature/identifier": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/TypeVariableSignature/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make(Ljava/lang/String;)Lsun/reflect/generics/tree/TypeVariableSignature;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_TypeVariableSignature) => void): void;
    public static "sun/reflect/generics/tree/TypeVariableSignature/make(Ljava/lang/String;)Lsun/reflect/generics/tree/TypeVariableSignature;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_TypeVariableSignature) => void): void;
    public "getIdentifier()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/reflect/generics/tree/TypeVariableSignature/getIdentifier()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/TypeVariableSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_ArrayTypeSignature extends java_lang_Object implements sun_reflect_generics_tree_FieldTypeSignature {
    public "sun/reflect/generics/tree/ArrayTypeSignature/componentType": sun_reflect_generics_tree_TypeSignature;
    public "<init>(Lsun/reflect/generics/tree/TypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_TypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ArrayTypeSignature/<init>(Lsun/reflect/generics/tree/TypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_TypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make(Lsun/reflect/generics/tree/TypeSignature;)Lsun/reflect/generics/tree/ArrayTypeSignature;"(thread: JVMThread, args: [sun_reflect_generics_tree_TypeSignature], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ArrayTypeSignature) => void): void;
    public static "sun/reflect/generics/tree/ArrayTypeSignature/make(Lsun/reflect/generics/tree/TypeSignature;)Lsun/reflect/generics/tree/ArrayTypeSignature;"(thread: JVMThread, args: [sun_reflect_generics_tree_TypeSignature], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ArrayTypeSignature) => void): void;
    public "getComponentType()Lsun/reflect/generics/tree/TypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_TypeSignature) => void): void;
    public "sun/reflect/generics/tree/ArrayTypeSignature/getComponentType()Lsun/reflect/generics/tree/TypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_TypeSignature) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ArrayTypeSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_ClassTypeSignature extends java_lang_Object implements sun_reflect_generics_tree_FieldTypeSignature {
    public "sun/reflect/generics/tree/ClassTypeSignature/path": java_util_List;
    public "<init>(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ClassTypeSignature/<init>(Ljava/util/List;)V"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make(Ljava/util/List;)Lsun/reflect/generics/tree/ClassTypeSignature;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ClassTypeSignature) => void): void;
    public static "sun/reflect/generics/tree/ClassTypeSignature/make(Ljava/util/List;)Lsun/reflect/generics/tree/ClassTypeSignature;"(thread: JVMThread, args: [java_util_List], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ClassTypeSignature) => void): void;
    public "getPath()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "sun/reflect/generics/tree/ClassTypeSignature/getPath()Ljava/util/List;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_List) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ClassTypeSignature/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_ClassSignature extends java_lang_Object implements sun_reflect_generics_tree_Signature {
    public "sun/reflect/generics/tree/ClassSignature/formalTypeParams": JVMArray<sun_reflect_generics_tree_FormalTypeParameter>;
    public "sun/reflect/generics/tree/ClassSignature/superclass": sun_reflect_generics_tree_ClassTypeSignature;
    public "sun/reflect/generics/tree/ClassSignature/superInterfaces": JVMArray<sun_reflect_generics_tree_ClassTypeSignature>;
    public "<init>([Lsun/reflect/generics/tree/FormalTypeParameter;Lsun/reflect/generics/tree/ClassTypeSignature;[Lsun/reflect/generics/tree/ClassTypeSignature;)V"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FormalTypeParameter>, sun_reflect_generics_tree_ClassTypeSignature, JVMArray<sun_reflect_generics_tree_ClassTypeSignature>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ClassSignature/<init>([Lsun/reflect/generics/tree/FormalTypeParameter;Lsun/reflect/generics/tree/ClassTypeSignature;[Lsun/reflect/generics/tree/ClassTypeSignature;)V"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FormalTypeParameter>, sun_reflect_generics_tree_ClassTypeSignature, JVMArray<sun_reflect_generics_tree_ClassTypeSignature>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make([Lsun/reflect/generics/tree/FormalTypeParameter;Lsun/reflect/generics/tree/ClassTypeSignature;[Lsun/reflect/generics/tree/ClassTypeSignature;)Lsun/reflect/generics/tree/ClassSignature;"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FormalTypeParameter>, sun_reflect_generics_tree_ClassTypeSignature, JVMArray<sun_reflect_generics_tree_ClassTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ClassSignature) => void): void;
    public static "sun/reflect/generics/tree/ClassSignature/make([Lsun/reflect/generics/tree/FormalTypeParameter;Lsun/reflect/generics/tree/ClassTypeSignature;[Lsun/reflect/generics/tree/ClassTypeSignature;)Lsun/reflect/generics/tree/ClassSignature;"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FormalTypeParameter>, sun_reflect_generics_tree_ClassTypeSignature, JVMArray<sun_reflect_generics_tree_ClassTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ClassSignature) => void): void;
    public "getFormalTypeParameters()[Lsun/reflect/generics/tree/FormalTypeParameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FormalTypeParameter>) => void): void;
    public "sun/reflect/generics/tree/ClassSignature/getFormalTypeParameters()[Lsun/reflect/generics/tree/FormalTypeParameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FormalTypeParameter>) => void): void;
    public "getSuperclass()Lsun/reflect/generics/tree/ClassTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ClassTypeSignature) => void): void;
    public "sun/reflect/generics/tree/ClassSignature/getSuperclass()Lsun/reflect/generics/tree/ClassTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ClassTypeSignature) => void): void;
    public "getSuperInterfaces()[Lsun/reflect/generics/tree/ClassTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_ClassTypeSignature>) => void): void;
    public "sun/reflect/generics/tree/ClassSignature/getSuperInterfaces()[Lsun/reflect/generics/tree/ClassTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_ClassTypeSignature>) => void): void;
    public "accept(Lsun/reflect/generics/visitor/Visitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_Visitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/ClassSignature/accept(Lsun/reflect/generics/visitor/Visitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_Visitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_generics_visitor_TypeTreeVisitor extends java_lang_Object {
    "getResult()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "visitFormalTypeParameter(Lsun/reflect/generics/tree/FormalTypeParameter;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_FormalTypeParameter], cb?: (e?: java_lang_Throwable) => void): void;
    "visitClassTypeSignature(Lsun/reflect/generics/tree/ClassTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ClassTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitArrayTypeSignature(Lsun/reflect/generics/tree/ArrayTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ArrayTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitTypeVariableSignature(Lsun/reflect/generics/tree/TypeVariableSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_TypeVariableSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitWildcard(Lsun/reflect/generics/tree/Wildcard;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_Wildcard], cb?: (e?: java_lang_Throwable) => void): void;
    "visitSimpleClassTypeSignature(Lsun/reflect/generics/tree/SimpleClassTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_SimpleClassTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitBottomSignature(Lsun/reflect/generics/tree/BottomSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_BottomSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitByteSignature(Lsun/reflect/generics/tree/ByteSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ByteSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitBooleanSignature(Lsun/reflect/generics/tree/BooleanSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_BooleanSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitShortSignature(Lsun/reflect/generics/tree/ShortSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ShortSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitCharSignature(Lsun/reflect/generics/tree/CharSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_CharSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitIntSignature(Lsun/reflect/generics/tree/IntSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_IntSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitLongSignature(Lsun/reflect/generics/tree/LongSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_LongSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitFloatSignature(Lsun/reflect/generics/tree/FloatSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_FloatSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitDoubleSignature(Lsun/reflect/generics/tree/DoubleSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_DoubleSignature], cb?: (e?: java_lang_Throwable) => void): void;
    "visitVoidDescriptor(Lsun/reflect/generics/tree/VoidDescriptor;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_VoidDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_generics_tree_FieldTypeSignature extends java_lang_Object, sun_reflect_generics_tree_BaseType, sun_reflect_generics_tree_TypeSignature, sun_reflect_generics_tree_TypeArgument {
    "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_generics_tree_ReturnType extends java_lang_Object, sun_reflect_generics_tree_TypeTree {
    "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_generics_tree_TypeSignature extends java_lang_Object, sun_reflect_generics_tree_ReturnType {
    "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_tree_FormalTypeParameter extends java_lang_Object implements sun_reflect_generics_tree_TypeTree {
    public "sun/reflect/generics/tree/FormalTypeParameter/name": java_lang_String;
    public "sun/reflect/generics/tree/FormalTypeParameter/bounds": JVMArray<sun_reflect_generics_tree_FieldTypeSignature>;
    public "<init>(Ljava/lang/String;[Lsun/reflect/generics/tree/FieldTypeSignature;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/FormalTypeParameter/<init>(Ljava/lang/String;[Lsun/reflect/generics/tree/FieldTypeSignature;)V"(thread: JVMThread, args: [java_lang_String, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make(Ljava/lang/String;[Lsun/reflect/generics/tree/FieldTypeSignature;)Lsun/reflect/generics/tree/FormalTypeParameter;"(thread: JVMThread, args: [java_lang_String, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_FormalTypeParameter) => void): void;
    public static "sun/reflect/generics/tree/FormalTypeParameter/make(Ljava/lang/String;[Lsun/reflect/generics/tree/FieldTypeSignature;)Lsun/reflect/generics/tree/FormalTypeParameter;"(thread: JVMThread, args: [java_lang_String, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_FormalTypeParameter) => void): void;
    public "getBounds()[Lsun/reflect/generics/tree/FieldTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FieldTypeSignature>) => void): void;
    public "sun/reflect/generics/tree/FormalTypeParameter/getBounds()[Lsun/reflect/generics/tree/FieldTypeSignature;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FieldTypeSignature>) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/reflect/generics/tree/FormalTypeParameter/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/tree/FormalTypeParameter/accept(Lsun/reflect/generics/visitor/TypeTreeVisitor;)V"(thread: JVMThread, args: [sun_reflect_generics_visitor_TypeTreeVisitor], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_generics_tree_Signature extends java_lang_Object, sun_reflect_generics_tree_Tree {
    "getFormalTypeParameters()[Lsun/reflect/generics/tree/FormalTypeParameter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<sun_reflect_generics_tree_FormalTypeParameter>) => void): void;
  }
  export interface sun_reflect_generics_factory_GenericsFactory extends java_lang_Object {
    "makeTypeVariable(Ljava/lang/String;[Lsun/reflect/generics/tree/FieldTypeSignature;)Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: [java_lang_String, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_TypeVariable) => void): void;
    "makeParameterizedType(Ljava/lang/reflect/Type;[Ljava/lang/reflect/Type;Ljava/lang/reflect/Type;)Ljava/lang/reflect/ParameterizedType;"(thread: JVMThread, args: [java_lang_reflect_Type, JVMArray<java_lang_reflect_Type>, java_lang_reflect_Type], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_ParameterizedType) => void): void;
    "findTypeVariable(Ljava/lang/String;)Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_TypeVariable) => void): void;
    "makeWildcard([Lsun/reflect/generics/tree/FieldTypeSignature;[Lsun/reflect/generics/tree/FieldTypeSignature;)Ljava/lang/reflect/WildcardType;"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_FieldTypeSignature>, JVMArray<sun_reflect_generics_tree_FieldTypeSignature>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_WildcardType) => void): void;
    "makeNamedType(Ljava/lang/String;)Ljava/lang/reflect/Type;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeArrayType(Ljava/lang/reflect/Type;)Ljava/lang/reflect/Type;"(thread: JVMThread, args: [java_lang_reflect_Type], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeByte()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeBool()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeShort()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeChar()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeInt()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeLong()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeFloat()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeDouble()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "makeVoid()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
  }
  export interface java_lang_reflect_WildcardType extends java_lang_Object, java_lang_reflect_Type {
    "getUpperBounds()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    "getLowerBounds()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    "getTypeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export interface java_lang_reflect_ParameterizedType extends java_lang_Object, java_lang_reflect_Type {
    "getActualTypeArguments()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    "getRawType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "getOwnerType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    "getTypeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class sun_reflect_generics_repository_GenericDeclRepository extends sun_reflect_generics_repository_AbstractRepository {
    public "sun/reflect/generics/repository/GenericDeclRepository/typeParams": JVMArray<java_lang_reflect_TypeVariable>;
    public "<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/repository/GenericDeclRepository/<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    public "sun/reflect/generics/repository/GenericDeclRepository/getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
  }
  export class sun_reflect_generics_repository_AbstractRepository extends java_lang_Object {
    public "sun/reflect/generics/repository/AbstractRepository/factory": sun_reflect_generics_factory_GenericsFactory;
    public "sun/reflect/generics/repository/AbstractRepository/tree": sun_reflect_generics_tree_Tree;
    public "getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "sun/reflect/generics/repository/AbstractRepository/getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "getTree()Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
    public "sun/reflect/generics/repository/AbstractRepository/getTree()Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
    public "getReifier()Lsun/reflect/generics/visitor/Reifier;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_visitor_Reifier) => void): void;
    public "sun/reflect/generics/repository/AbstractRepository/getReifier()Lsun/reflect/generics/visitor/Reifier;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_visitor_Reifier) => void): void;
    public "<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/repository/AbstractRepository/<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "parse(Ljava/lang/String;)Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
    public "sun/reflect/generics/repository/AbstractRepository/parse(Ljava/lang/String;)Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
  }
  export class sun_reflect_generics_visitor_Reifier extends java_lang_Object implements sun_reflect_generics_visitor_TypeTreeVisitor {
    public "sun/reflect/generics/visitor/Reifier/resultType": java_lang_reflect_Type;
    public "sun/reflect/generics/visitor/Reifier/factory": sun_reflect_generics_factory_GenericsFactory;
    public static "sun/reflect/generics/visitor/Reifier/$assertionsDisabled": number;
    public "<init>(Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/<init>(Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "sun/reflect/generics/visitor/Reifier/getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public static "make(Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/visitor/Reifier;"(thread: JVMThread, args: [sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_visitor_Reifier) => void): void;
    public static "sun/reflect/generics/visitor/Reifier/make(Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/visitor/Reifier;"(thread: JVMThread, args: [sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_visitor_Reifier) => void): void;
    public "reifyTypeArguments([Lsun/reflect/generics/tree/TypeArgument;)[Ljava/lang/reflect/Type;"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_TypeArgument>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "sun/reflect/generics/visitor/Reifier/reifyTypeArguments([Lsun/reflect/generics/tree/TypeArgument;)[Ljava/lang/reflect/Type;"(thread: JVMThread, args: [JVMArray<sun_reflect_generics_tree_TypeArgument>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "getResult()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "sun/reflect/generics/visitor/Reifier/getResult()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "visitFormalTypeParameter(Lsun/reflect/generics/tree/FormalTypeParameter;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_FormalTypeParameter], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitFormalTypeParameter(Lsun/reflect/generics/tree/FormalTypeParameter;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_FormalTypeParameter], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitClassTypeSignature(Lsun/reflect/generics/tree/ClassTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ClassTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitClassTypeSignature(Lsun/reflect/generics/tree/ClassTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ClassTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitArrayTypeSignature(Lsun/reflect/generics/tree/ArrayTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ArrayTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitArrayTypeSignature(Lsun/reflect/generics/tree/ArrayTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ArrayTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitTypeVariableSignature(Lsun/reflect/generics/tree/TypeVariableSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_TypeVariableSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitTypeVariableSignature(Lsun/reflect/generics/tree/TypeVariableSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_TypeVariableSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitWildcard(Lsun/reflect/generics/tree/Wildcard;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_Wildcard], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitWildcard(Lsun/reflect/generics/tree/Wildcard;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_Wildcard], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitSimpleClassTypeSignature(Lsun/reflect/generics/tree/SimpleClassTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_SimpleClassTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitSimpleClassTypeSignature(Lsun/reflect/generics/tree/SimpleClassTypeSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_SimpleClassTypeSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitBottomSignature(Lsun/reflect/generics/tree/BottomSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_BottomSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitBottomSignature(Lsun/reflect/generics/tree/BottomSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_BottomSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitByteSignature(Lsun/reflect/generics/tree/ByteSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ByteSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitByteSignature(Lsun/reflect/generics/tree/ByteSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ByteSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitBooleanSignature(Lsun/reflect/generics/tree/BooleanSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_BooleanSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitBooleanSignature(Lsun/reflect/generics/tree/BooleanSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_BooleanSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitShortSignature(Lsun/reflect/generics/tree/ShortSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ShortSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitShortSignature(Lsun/reflect/generics/tree/ShortSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_ShortSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitCharSignature(Lsun/reflect/generics/tree/CharSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_CharSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitCharSignature(Lsun/reflect/generics/tree/CharSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_CharSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitIntSignature(Lsun/reflect/generics/tree/IntSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_IntSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitIntSignature(Lsun/reflect/generics/tree/IntSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_IntSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitLongSignature(Lsun/reflect/generics/tree/LongSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_LongSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitLongSignature(Lsun/reflect/generics/tree/LongSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_LongSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitFloatSignature(Lsun/reflect/generics/tree/FloatSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_FloatSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitFloatSignature(Lsun/reflect/generics/tree/FloatSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_FloatSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitDoubleSignature(Lsun/reflect/generics/tree/DoubleSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_DoubleSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitDoubleSignature(Lsun/reflect/generics/tree/DoubleSignature;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_DoubleSignature], cb?: (e?: java_lang_Throwable) => void): void;
    public "visitVoidDescriptor(Lsun/reflect/generics/tree/VoidDescriptor;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_VoidDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/visitor/Reifier/visitVoidDescriptor(Lsun/reflect/generics/tree/VoidDescriptor;)V"(thread: JVMThread, args: [sun_reflect_generics_tree_VoidDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "getResult()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "sun/reflect/generics/visitor/Reifier/getResult()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/visitor/Reifier/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_lang_reflect_GenericDeclaration extends java_lang_Object, java_lang_reflect_AnnotatedElement {
    "getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
  }
  export class java_lang_reflect_AccessibleObject extends java_lang_Object implements java_lang_reflect_AnnotatedElement {
    public static "java/lang/reflect/AccessibleObject/ACCESS_PERMISSION": java_security_Permission;
    public "java/lang/reflect/AccessibleObject/override": number;
    public static "java/lang/reflect/AccessibleObject/reflectionFactory": sun_reflect_ReflectionFactory;
    public "java/lang/reflect/AccessibleObject/securityCheckCache": java_lang_Object;
    public static "setAccessible([Ljava/lang/reflect/AccessibleObject;Z)V"(thread: JVMThread, args: [JVMArray<java_lang_reflect_AccessibleObject>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/AccessibleObject/setAccessible([Ljava/lang/reflect/AccessibleObject;Z)V"(thread: JVMThread, args: [JVMArray<java_lang_reflect_AccessibleObject>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setAccessible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/AccessibleObject/setAccessible(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setAccessible0(Ljava/lang/reflect/AccessibleObject;Z)V"(thread: JVMThread, args: [java_lang_reflect_AccessibleObject, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/AccessibleObject/setAccessible0(Ljava/lang/reflect/AccessibleObject;Z)V"(thread: JVMThread, args: [java_lang_reflect_AccessibleObject, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAccessible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/AccessibleObject/isAccessible()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/AccessibleObject/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/reflect/AccessibleObject/getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/AccessibleObject/isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/AccessibleObject/getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/AccessibleObject/getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/reflect/AccessibleObject/getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/AccessibleObject/getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/AccessibleObject/getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "checkAccess(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/AccessibleObject/checkAccess(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "slowCheckMemberAccess(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/Object;ILjava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_Object, number, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/AccessibleObject/slowCheckMemberAccess(Ljava/lang/Class;Ljava/lang/Class;Ljava/lang/Object;ILjava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Class, java_lang_Object, number, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/AccessibleObject/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
  }
  export class sun_reflect_ReflectionFactory extends java_lang_Object {
    public static "sun/reflect/ReflectionFactory/initted": number;
    public static "sun/reflect/ReflectionFactory/reflectionFactoryAccessPerm": java_security_Permission;
    public static "sun/reflect/ReflectionFactory/soleInstance": sun_reflect_ReflectionFactory;
    public static "sun/reflect/ReflectionFactory/langReflectAccess": sun_reflect_LangReflectAccess;
    public static "sun/reflect/ReflectionFactory/noInflation": number;
    public static "sun/reflect/ReflectionFactory/inflationThreshold": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/ReflectionFactory/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getReflectionFactory()Lsun/reflect/ReflectionFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ReflectionFactory) => void): void;
    public static "sun/reflect/ReflectionFactory/getReflectionFactory()Lsun/reflect/ReflectionFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ReflectionFactory) => void): void;
    public "setLangReflectAccess(Lsun/reflect/LangReflectAccess;)V"(thread: JVMThread, args: [sun_reflect_LangReflectAccess], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/ReflectionFactory/setLangReflectAccess(Lsun/reflect/LangReflectAccess;)V"(thread: JVMThread, args: [sun_reflect_LangReflectAccess], cb?: (e?: java_lang_Throwable) => void): void;
    public "newFieldAccessor(Ljava/lang/reflect/Field;Z)Lsun/reflect/FieldAccessor;"(thread: JVMThread, args: [java_lang_reflect_Field, number], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_FieldAccessor) => void): void;
    public "sun/reflect/ReflectionFactory/newFieldAccessor(Ljava/lang/reflect/Field;Z)Lsun/reflect/FieldAccessor;"(thread: JVMThread, args: [java_lang_reflect_Field, number], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_FieldAccessor) => void): void;
    public "newMethodAccessor(Ljava/lang/reflect/Method;)Lsun/reflect/MethodAccessor;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_MethodAccessor) => void): void;
    public "sun/reflect/ReflectionFactory/newMethodAccessor(Ljava/lang/reflect/Method;)Lsun/reflect/MethodAccessor;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_MethodAccessor) => void): void;
    public "newConstructorAccessor(Ljava/lang/reflect/Constructor;)Lsun/reflect/ConstructorAccessor;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstructorAccessor) => void): void;
    public "sun/reflect/ReflectionFactory/newConstructorAccessor(Ljava/lang/reflect/Constructor;)Lsun/reflect/ConstructorAccessor;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstructorAccessor) => void): void;
    public "newField(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;IILjava/lang/String;[B)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class, number, number, java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "sun/reflect/ReflectionFactory/newField(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;IILjava/lang/String;[B)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class, number, number, java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "newMethod(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B[B)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "sun/reflect/ReflectionFactory/newMethod(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B[B)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "newConstructor(Ljava/lang/Class;[Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "sun/reflect/ReflectionFactory/newConstructor(Ljava/lang/Class;[Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "getMethodAccessor(Ljava/lang/reflect/Method;)Lsun/reflect/MethodAccessor;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_MethodAccessor) => void): void;
    public "sun/reflect/ReflectionFactory/getMethodAccessor(Ljava/lang/reflect/Method;)Lsun/reflect/MethodAccessor;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_MethodAccessor) => void): void;
    public "setMethodAccessor(Ljava/lang/reflect/Method;Lsun/reflect/MethodAccessor;)V"(thread: JVMThread, args: [java_lang_reflect_Method, sun_reflect_MethodAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/ReflectionFactory/setMethodAccessor(Ljava/lang/reflect/Method;Lsun/reflect/MethodAccessor;)V"(thread: JVMThread, args: [java_lang_reflect_Method, sun_reflect_MethodAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "getConstructorAccessor(Ljava/lang/reflect/Constructor;)Lsun/reflect/ConstructorAccessor;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstructorAccessor) => void): void;
    public "sun/reflect/ReflectionFactory/getConstructorAccessor(Ljava/lang/reflect/Constructor;)Lsun/reflect/ConstructorAccessor;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstructorAccessor) => void): void;
    public "setConstructorAccessor(Ljava/lang/reflect/Constructor;Lsun/reflect/ConstructorAccessor;)V"(thread: JVMThread, args: [java_lang_reflect_Constructor, sun_reflect_ConstructorAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/ReflectionFactory/setConstructorAccessor(Ljava/lang/reflect/Constructor;Lsun/reflect/ConstructorAccessor;)V"(thread: JVMThread, args: [java_lang_reflect_Constructor, sun_reflect_ConstructorAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "copyMethod(Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "sun/reflect/ReflectionFactory/copyMethod(Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "copyField(Ljava/lang/reflect/Field;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "sun/reflect/ReflectionFactory/copyField(Ljava/lang/reflect/Field;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "copyConstructor(Ljava/lang/reflect/Constructor;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "sun/reflect/ReflectionFactory/copyConstructor(Ljava/lang/reflect/Constructor;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "getExecutableTypeAnnotationBytes(Ljava/lang/reflect/Executable;)[B"(thread: JVMThread, args: [java_lang_reflect_Executable], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "sun/reflect/ReflectionFactory/getExecutableTypeAnnotationBytes(Ljava/lang/reflect/Executable;)[B"(thread: JVMThread, args: [java_lang_reflect_Executable], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "newConstructorForSerialization(Ljava/lang/Class;Ljava/lang/reflect/Constructor;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class, java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "sun/reflect/ReflectionFactory/newConstructorForSerialization(Ljava/lang/Class;Ljava/lang/reflect/Constructor;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class, java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "inflationThreshold()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/ReflectionFactory/inflationThreshold()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "checkInitted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/ReflectionFactory/checkInitted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "langReflectAccess()Lsun/reflect/LangReflectAccess;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_LangReflectAccess) => void): void;
    public static "sun/reflect/ReflectionFactory/langReflectAccess()Lsun/reflect/LangReflectAccess;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_LangReflectAccess) => void): void;
    public static "access$002(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/ReflectionFactory/access$002(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$102(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/ReflectionFactory/access$102(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$202(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sun/reflect/ReflectionFactory/access$202(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/ReflectionFactory/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_FieldAccessor extends java_lang_Object {
    "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "getBoolean(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getByte(Ljava/lang/Object;)B"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getChar(Ljava/lang/Object;)C"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getShort(Ljava/lang/Object;)S"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getInt(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getLong(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "getFloat(Ljava/lang/Object;)F"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getDouble(Ljava/lang/Object;)D"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "set(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    "setBoolean(Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setByte(Ljava/lang/Object;B)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setChar(Ljava/lang/Object;C)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setShort(Ljava/lang/Object;S)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setInt(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setLong(Ljava/lang/Object;J)V"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    "setFloat(Ljava/lang/Object;F)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    "setDouble(Ljava/lang/Object;D)V"(thread: JVMThread, args: [java_lang_Object, number, any], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_reflect_LangReflectAccess extends java_lang_Object {
    "newField(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;IILjava/lang/String;[B)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class, number, number, java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    "newMethod(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B[B)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    "newConstructor(Ljava/lang/Class;[Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    "getMethodAccessor(Ljava/lang/reflect/Method;)Lsun/reflect/MethodAccessor;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_MethodAccessor) => void): void;
    "setMethodAccessor(Ljava/lang/reflect/Method;Lsun/reflect/MethodAccessor;)V"(thread: JVMThread, args: [java_lang_reflect_Method, sun_reflect_MethodAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    "getConstructorAccessor(Ljava/lang/reflect/Constructor;)Lsun/reflect/ConstructorAccessor;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstructorAccessor) => void): void;
    "setConstructorAccessor(Ljava/lang/reflect/Constructor;Lsun/reflect/ConstructorAccessor;)V"(thread: JVMThread, args: [java_lang_reflect_Constructor, sun_reflect_ConstructorAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    "getExecutableTypeAnnotationBytes(Ljava/lang/reflect/Executable;)[B"(thread: JVMThread, args: [java_lang_reflect_Executable], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    "getConstructorSlot(Ljava/lang/reflect/Constructor;)I"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getConstructorSignature(Ljava/lang/reflect/Constructor;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "getConstructorAnnotations(Ljava/lang/reflect/Constructor;)[B"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    "getConstructorParameterAnnotations(Ljava/lang/reflect/Constructor;)[B"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    "copyMethod(Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    "copyField(Ljava/lang/reflect/Field;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_reflect_Field], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    "copyConstructor(Ljava/lang/reflect/Constructor;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
  }
  export class java_lang_reflect_Array extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Array/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "newInstance(Ljava/lang/Class;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/reflect/Array/newInstance(Ljava/lang/Class;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "newInstance(Ljava/lang/Class;[I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/reflect/Array/newInstance(Ljava/lang/Class;[I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "getLength(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/reflect/Array/getLength(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "get(Ljava/lang/Object;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/reflect/Array/get(Ljava/lang/Object;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "getBoolean(Ljava/lang/Object;I)Z"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/reflect/Array/getBoolean(Ljava/lang/Object;I)Z"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getByte(Ljava/lang/Object;I)B"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/reflect/Array/getByte(Ljava/lang/Object;I)B"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getChar(Ljava/lang/Object;I)C"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/reflect/Array/getChar(Ljava/lang/Object;I)C"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getShort(Ljava/lang/Object;I)S"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/reflect/Array/getShort(Ljava/lang/Object;I)S"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getInt(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/reflect/Array/getInt(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getLong(Ljava/lang/Object;I)J"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/reflect/Array/getLong(Ljava/lang/Object;I)J"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "getFloat(Ljava/lang/Object;I)F"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/reflect/Array/getFloat(Ljava/lang/Object;I)F"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getDouble(Ljava/lang/Object;I)D"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/reflect/Array/getDouble(Ljava/lang/Object;I)D"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "set(Ljava/lang/Object;ILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Array/set(Ljava/lang/Object;ILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setBoolean(Ljava/lang/Object;IZ)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Array/setBoolean(Ljava/lang/Object;IZ)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setByte(Ljava/lang/Object;IB)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Array/setByte(Ljava/lang/Object;IB)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setChar(Ljava/lang/Object;IC)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Array/setChar(Ljava/lang/Object;IC)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setShort(Ljava/lang/Object;IS)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Array/setShort(Ljava/lang/Object;IS)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setInt(Ljava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Array/setInt(Ljava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setLong(Ljava/lang/Object;IJ)V"(thread: JVMThread, args: [java_lang_Object, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Array/setLong(Ljava/lang/Object;IJ)V"(thread: JVMThread, args: [java_lang_Object, number, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setFloat(Ljava/lang/Object;IF)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Array/setFloat(Ljava/lang/Object;IF)V"(thread: JVMThread, args: [java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setDouble(Ljava/lang/Object;ID)V"(thread: JVMThread, args: [java_lang_Object, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Array/setDouble(Ljava/lang/Object;ID)V"(thread: JVMThread, args: [java_lang_Object, number, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "newArray(Ljava/lang/Class;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/reflect/Array/newArray(Ljava/lang/Class;I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "multiNewArray(Ljava/lang/Class;[I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/reflect/Array/multiNewArray(Ljava/lang/Class;[I)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_lang_reflect_Proxy extends java_lang_Object implements java_io_Serializable {
    public static "java/lang/reflect/Proxy/serialVersionUID": Long;
    public static "java/lang/reflect/Proxy/constructorParams": JVMArray<java_lang_Class>;
    public static "java/lang/reflect/Proxy/proxyClassCache": java_lang_reflect_WeakCache;
    public "java/lang/reflect/Proxy/h": java_lang_reflect_InvocationHandler;
    public static "java/lang/reflect/Proxy/key0": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Proxy/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/reflect/InvocationHandler;)V"(thread: JVMThread, args: [java_lang_reflect_InvocationHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Proxy/<init>(Ljava/lang/reflect/InvocationHandler;)V"(thread: JVMThread, args: [java_lang_reflect_InvocationHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getProxyClass(Ljava/lang/ClassLoader;[Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/reflect/Proxy/getProxyClass(Ljava/lang/ClassLoader;[Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "checkProxyAccess(Ljava/lang/Class;Ljava/lang/ClassLoader;[Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_ClassLoader, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Proxy/checkProxyAccess(Ljava/lang/Class;Ljava/lang/ClassLoader;[Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_ClassLoader, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getProxyClass0(Ljava/lang/ClassLoader;[Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/reflect/Proxy/getProxyClass0(Ljava/lang/ClassLoader;[Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "newProxyInstance(Ljava/lang/ClassLoader;[Ljava/lang/Class;Ljava/lang/reflect/InvocationHandler;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_ClassLoader, JVMArray<java_lang_Class>, java_lang_reflect_InvocationHandler], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/reflect/Proxy/newProxyInstance(Ljava/lang/ClassLoader;[Ljava/lang/Class;Ljava/lang/reflect/InvocationHandler;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_ClassLoader, JVMArray<java_lang_Class>, java_lang_reflect_InvocationHandler], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "checkNewProxyPermission(Ljava/lang/Class;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Proxy/checkNewProxyPermission(Ljava/lang/Class;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isProxyClass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/reflect/Proxy/isProxyClass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getInvocationHandler(Ljava/lang/Object;)Ljava/lang/reflect/InvocationHandler;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_InvocationHandler) => void): void;
    public static "java/lang/reflect/Proxy/getInvocationHandler(Ljava/lang/Object;)Ljava/lang/reflect/InvocationHandler;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_InvocationHandler) => void): void;
    public static "defineClass0(Ljava/lang/ClassLoader;Ljava/lang/String;[BII)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/reflect/Proxy/defineClass0(Ljava/lang/ClassLoader;Ljava/lang/String;[BII)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "access$200()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/reflect/Proxy/access$200()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$300(Ljava/lang/ClassLoader;Ljava/lang/String;[BII)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/reflect/Proxy/access$300(Ljava/lang/ClassLoader;Ljava/lang/String;[BII)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/reflect/Proxy/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_lang_reflect_InvocationHandler extends java_lang_Object {
    "invoke(Ljava/lang/Object;Ljava/lang/reflect/Method;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_reflect_Method, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_lang_reflect_WeakCache extends java_lang_Object {
    public "java/lang/reflect/WeakCache/refQueue": java_lang_ref_ReferenceQueue;
    public "java/lang/reflect/WeakCache/map": java_util_concurrent_ConcurrentMap;
    public "java/lang/reflect/WeakCache/reverseMap": java_util_concurrent_ConcurrentMap;
    public "java/lang/reflect/WeakCache/subKeyFactory": java_util_function_BiFunction;
    public "java/lang/reflect/WeakCache/valueFactory": java_util_function_BiFunction;
    public "<init>(Ljava/util/function/BiFunction;Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/WeakCache/<init>(Ljava/util/function/BiFunction;Ljava/util/function/BiFunction;)V"(thread: JVMThread, args: [java_util_function_BiFunction, java_util_function_BiFunction], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/reflect/WeakCache/get(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/WeakCache/containsValue(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/WeakCache/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "expungeStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/WeakCache/expungeStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/lang/reflect/WeakCache;)Ljava/util/function/BiFunction;"(thread: JVMThread, args: [java_lang_reflect_WeakCache], cb?: (e?: java_lang_Throwable, rv?: java_util_function_BiFunction) => void): void;
    public static "java/lang/reflect/WeakCache/access$000(Ljava/lang/reflect/WeakCache;)Ljava/util/function/BiFunction;"(thread: JVMThread, args: [java_lang_reflect_WeakCache], cb?: (e?: java_lang_Throwable, rv?: java_util_function_BiFunction) => void): void;
    public static "access$100(Ljava/lang/reflect/WeakCache;)Ljava/util/concurrent/ConcurrentMap;"(thread: JVMThread, args: [java_lang_reflect_WeakCache], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentMap) => void): void;
    public static "java/lang/reflect/WeakCache/access$100(Ljava/lang/reflect/WeakCache;)Ljava/util/concurrent/ConcurrentMap;"(thread: JVMThread, args: [java_lang_reflect_WeakCache], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_ConcurrentMap) => void): void;
  }
  export class java_lang_reflect_Constructor extends java_lang_reflect_Executable {
    public "java/lang/reflect/Constructor/clazz": java_lang_Class;
    public "java/lang/reflect/Constructor/slot": number;
    public "java/lang/reflect/Constructor/parameterTypes": JVMArray<java_lang_Class>;
    public "java/lang/reflect/Constructor/exceptionTypes": JVMArray<java_lang_Class>;
    public "java/lang/reflect/Constructor/modifiers": number;
    public "java/lang/reflect/Constructor/signature": java_lang_String;
    public "java/lang/reflect/Constructor/genericInfo": sun_reflect_generics_repository_ConstructorRepository;
    public "java/lang/reflect/Constructor/annotations": JVMArray<number>;
    public "java/lang/reflect/Constructor/parameterAnnotations": JVMArray<number>;
    public "java/lang/reflect/Constructor/constructorAccessor": sun_reflect_ConstructorAccessor;
    public "java/lang/reflect/Constructor/root": java_lang_reflect_Constructor;
    public "getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "java/lang/reflect/Constructor/getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "getGenericInfo()Lsun/reflect/generics/repository/ConstructorRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ConstructorRepository) => void): void;
    public "java/lang/reflect/Constructor/getGenericInfo()Lsun/reflect/generics/repository/ConstructorRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ConstructorRepository) => void): void;
    public "getRoot()Ljava/lang/reflect/Executable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Executable) => void): void;
    public "java/lang/reflect/Constructor/getRoot()Ljava/lang/reflect/Executable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Executable) => void): void;
    public "<init>(Ljava/lang/Class;[Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Constructor/<init>(Ljava/lang/Class;[Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Class>, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "copy()Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "java/lang/reflect/Constructor/copy()Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "hasGenericInformation()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Constructor/hasGenericInformation()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAnnotationBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/reflect/Constructor/getAnnotationBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/reflect/Constructor/getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Constructor/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Constructor/getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    public "java/lang/reflect/Constructor/getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    public "getParameterTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/reflect/Constructor/getParameterTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Constructor/getParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGenericParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "java/lang/reflect/Constructor/getGenericParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "getExceptionTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/reflect/Constructor/getExceptionTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getGenericExceptionTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "java/lang/reflect/Constructor/getGenericExceptionTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Constructor/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Constructor/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Constructor/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "specificToStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Constructor/specificToStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Constructor/toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "specificToGenericStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Constructor/specificToGenericStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "newInstance([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/reflect/Constructor/newInstance([Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isVarArgs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Constructor/isVarArgs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Constructor/isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "acquireConstructorAccessor()Lsun/reflect/ConstructorAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstructorAccessor) => void): void;
    public "java/lang/reflect/Constructor/acquireConstructorAccessor()Lsun/reflect/ConstructorAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstructorAccessor) => void): void;
    public "getConstructorAccessor()Lsun/reflect/ConstructorAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstructorAccessor) => void): void;
    public "java/lang/reflect/Constructor/getConstructorAccessor()Lsun/reflect/ConstructorAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstructorAccessor) => void): void;
    public "setConstructorAccessor(Lsun/reflect/ConstructorAccessor;)V"(thread: JVMThread, args: [sun_reflect_ConstructorAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Constructor/setConstructorAccessor(Lsun/reflect/ConstructorAccessor;)V"(thread: JVMThread, args: [sun_reflect_ConstructorAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSlot()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Constructor/getSlot()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Constructor/getSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getRawAnnotations()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/reflect/Constructor/getRawAnnotations()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getRawParameterAnnotations()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/reflect/Constructor/getRawParameterAnnotations()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/reflect/Constructor/getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Constructor/getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getParameterAnnotations()[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "java/lang/reflect/Constructor/getParameterAnnotations()[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "handleParameterNumberMismatch(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Constructor/handleParameterNumberMismatch(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAnnotatedReturnType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "java/lang/reflect/Constructor/getAnnotatedReturnType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "getAnnotatedReceiverType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "java/lang/reflect/Constructor/getAnnotatedReceiverType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
  }
  export class java_lang_reflect_Method extends java_lang_reflect_Executable {
    public "java/lang/reflect/Method/clazz": java_lang_Class;
    public "java/lang/reflect/Method/slot": number;
    public "java/lang/reflect/Method/name": java_lang_String;
    public "java/lang/reflect/Method/returnType": java_lang_Class;
    public "java/lang/reflect/Method/parameterTypes": JVMArray<java_lang_Class>;
    public "java/lang/reflect/Method/exceptionTypes": JVMArray<java_lang_Class>;
    public "java/lang/reflect/Method/modifiers": number;
    public "java/lang/reflect/Method/signature": java_lang_String;
    public "java/lang/reflect/Method/genericInfo": sun_reflect_generics_repository_MethodRepository;
    public "java/lang/reflect/Method/annotations": JVMArray<number>;
    public "java/lang/reflect/Method/parameterAnnotations": JVMArray<number>;
    public "java/lang/reflect/Method/annotationDefault": JVMArray<number>;
    public "java/lang/reflect/Method/methodAccessor": sun_reflect_MethodAccessor;
    public "java/lang/reflect/Method/root": java_lang_reflect_Method;
    public "getGenericSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Method/getGenericSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "java/lang/reflect/Method/getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "getGenericInfo()Lsun/reflect/generics/repository/MethodRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_MethodRepository) => void): void;
    public "java/lang/reflect/Method/getGenericInfo()Lsun/reflect/generics/repository/MethodRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_MethodRepository) => void): void;
    public "<init>(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B[B)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Method/<init>(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;[Ljava/lang/Class;IILjava/lang/String;[B[B[B)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class, JVMArray<java_lang_Class>, number, number, java_lang_String, JVMArray<number>, JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "copy()Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "java/lang/reflect/Method/copy()Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "getRoot()Ljava/lang/reflect/Executable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Executable) => void): void;
    public "java/lang/reflect/Method/getRoot()Ljava/lang/reflect/Executable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Executable) => void): void;
    public "hasGenericInformation()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Method/hasGenericInformation()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAnnotationBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/reflect/Method/getAnnotationBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/reflect/Method/getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Method/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Method/getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    public "java/lang/reflect/Method/getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    public "getReturnType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/reflect/Method/getReturnType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getGenericReturnType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "java/lang/reflect/Method/getGenericReturnType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "getParameterTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/reflect/Method/getParameterTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Method/getParameterCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getGenericParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "java/lang/reflect/Method/getGenericParameterTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "getExceptionTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/reflect/Method/getExceptionTypes()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getGenericExceptionTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "java/lang/reflect/Method/getGenericExceptionTypes()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Method/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Method/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Method/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "specificToStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Method/specificToStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Method/toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "specificToGenericStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Method/specificToGenericStringHeader(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "invoke(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/reflect/Method/invoke(Ljava/lang/Object;[Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isBridge()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Method/isBridge()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isVarArgs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Method/isVarArgs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Method/isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDefault()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Method/isDefault()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "acquireMethodAccessor()Lsun/reflect/MethodAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_MethodAccessor) => void): void;
    public "java/lang/reflect/Method/acquireMethodAccessor()Lsun/reflect/MethodAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_MethodAccessor) => void): void;
    public "getMethodAccessor()Lsun/reflect/MethodAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_MethodAccessor) => void): void;
    public "java/lang/reflect/Method/getMethodAccessor()Lsun/reflect/MethodAccessor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_MethodAccessor) => void): void;
    public "setMethodAccessor(Lsun/reflect/MethodAccessor;)V"(thread: JVMThread, args: [sun_reflect_MethodAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Method/setMethodAccessor(Lsun/reflect/MethodAccessor;)V"(thread: JVMThread, args: [sun_reflect_MethodAccessor], cb?: (e?: java_lang_Throwable) => void): void;
    public "getDefaultValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/reflect/Method/getDefaultValue()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/reflect/Method/getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Method/getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getParameterAnnotations()[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "java/lang/reflect/Method/getParameterAnnotations()[[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_annotation_Annotation>>) => void): void;
    public "getAnnotatedReturnType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "java/lang/reflect/Method/getAnnotatedReturnType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "handleParameterNumberMismatch(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Method/handleParameterNumberMismatch(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getGenericInfo()Lsun/reflect/generics/repository/ConstructorRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ConstructorRepository) => void): void;
    public "java/lang/reflect/Method/getGenericInfo()Lsun/reflect/generics/repository/ConstructorRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ConstructorRepository) => void): void;
  }
  export class sun_reflect_generics_repository_MethodRepository extends sun_reflect_generics_repository_ConstructorRepository {
    public "sun/reflect/generics/repository/MethodRepository/returnType": java_lang_reflect_Type;
    public "<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/repository/MethodRepository/<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public static "make(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/repository/MethodRepository;"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_MethodRepository) => void): void;
    public static "sun/reflect/generics/repository/MethodRepository/make(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/repository/MethodRepository;"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_MethodRepository) => void): void;
    public "getReturnType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "sun/reflect/generics/repository/MethodRepository/getReturnType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
  }
  export class java_lang_reflect_Field extends java_lang_reflect_AccessibleObject implements java_lang_reflect_Member {
    public "java/lang/reflect/Field/clazz": java_lang_Class;
    public "java/lang/reflect/Field/slot": number;
    public "java/lang/reflect/Field/name": java_lang_String;
    public "java/lang/reflect/Field/type": java_lang_Class;
    public "java/lang/reflect/Field/modifiers": number;
    public "java/lang/reflect/Field/signature": java_lang_String;
    public "java/lang/reflect/Field/genericInfo": sun_reflect_generics_repository_FieldRepository;
    public "java/lang/reflect/Field/annotations": JVMArray<number>;
    public "java/lang/reflect/Field/fieldAccessor": sun_reflect_FieldAccessor;
    public "java/lang/reflect/Field/overrideFieldAccessor": sun_reflect_FieldAccessor;
    public "java/lang/reflect/Field/root": java_lang_reflect_Field;
    public "java/lang/reflect/Field/declaredAnnotations": java_util_Map;
    public "getGenericSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Field/getGenericSignature()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "java/lang/reflect/Field/getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "getGenericInfo()Lsun/reflect/generics/repository/FieldRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_FieldRepository) => void): void;
    public "java/lang/reflect/Field/getGenericInfo()Lsun/reflect/generics/repository/FieldRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_FieldRepository) => void): void;
    public "<init>(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;IILjava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class, number, number, java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/<init>(Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Class;IILjava/lang/String;[B)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, java_lang_Class, number, number, java_lang_String, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "copy()Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "java/lang/reflect/Field/copy()Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/reflect/Field/getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Field/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEnumConstant()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/isEnumConstant()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/reflect/Field/getType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getGenericType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "java/lang/reflect/Field/getGenericType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Field/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/reflect/Field/toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/reflect/Field/get(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getBoolean(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/getBoolean(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getByte(Ljava/lang/Object;)B"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/getByte(Ljava/lang/Object;)B"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getChar(Ljava/lang/Object;)C"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/getChar(Ljava/lang/Object;)C"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getShort(Ljava/lang/Object;)S"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/getShort(Ljava/lang/Object;)S"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInt(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/getInt(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLong(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/reflect/Field/getLong(Ljava/lang/Object;)J"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFloat(Ljava/lang/Object;)F"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/getFloat(Ljava/lang/Object;)F"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDouble(Ljava/lang/Object;)D"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/reflect/Field/getDouble(Ljava/lang/Object;)D"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "set(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/set(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBoolean(Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/setBoolean(Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setByte(Ljava/lang/Object;B)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/setByte(Ljava/lang/Object;B)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setChar(Ljava/lang/Object;C)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/setChar(Ljava/lang/Object;C)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setShort(Ljava/lang/Object;S)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/setShort(Ljava/lang/Object;S)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setInt(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/setInt(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setLong(Ljava/lang/Object;J)V"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/setLong(Ljava/lang/Object;J)V"(thread: JVMThread, args: [java_lang_Object, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "setFloat(Ljava/lang/Object;F)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/setFloat(Ljava/lang/Object;F)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDouble(Ljava/lang/Object;D)V"(thread: JVMThread, args: [java_lang_Object, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/setDouble(Ljava/lang/Object;D)V"(thread: JVMThread, args: [java_lang_Object, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFieldAccessor(Ljava/lang/Object;)Lsun/reflect/FieldAccessor;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_FieldAccessor) => void): void;
    public "java/lang/reflect/Field/getFieldAccessor(Ljava/lang/Object;)Lsun/reflect/FieldAccessor;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_FieldAccessor) => void): void;
    public "acquireFieldAccessor(Z)Lsun/reflect/FieldAccessor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_FieldAccessor) => void): void;
    public "java/lang/reflect/Field/acquireFieldAccessor(Z)Lsun/reflect/FieldAccessor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_FieldAccessor) => void): void;
    public "getFieldAccessor(Z)Lsun/reflect/FieldAccessor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_FieldAccessor) => void): void;
    public "java/lang/reflect/Field/getFieldAccessor(Z)Lsun/reflect/FieldAccessor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_FieldAccessor) => void): void;
    public "setFieldAccessor(Lsun/reflect/FieldAccessor;Z)V"(thread: JVMThread, args: [sun_reflect_FieldAccessor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/Field/setFieldAccessor(Lsun/reflect/FieldAccessor;Z)V"(thread: JVMThread, args: [sun_reflect_FieldAccessor, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/reflect/Field/getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Field/getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/reflect/Field/getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "declaredAnnotations()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/lang/reflect/Field/declaredAnnotations()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getTypeAnnotationBytes0()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/reflect/Field/getTypeAnnotationBytes0()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getAnnotatedType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "java/lang/reflect/Field/getAnnotatedType()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
  }
  export class sun_reflect_generics_repository_FieldRepository extends sun_reflect_generics_repository_AbstractRepository {
    public "sun/reflect/generics/repository/FieldRepository/genericType": java_lang_reflect_Type;
    public "<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/repository/FieldRepository/<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "parse(Ljava/lang/String;)Lsun/reflect/generics/tree/TypeSignature;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_TypeSignature) => void): void;
    public "sun/reflect/generics/repository/FieldRepository/parse(Ljava/lang/String;)Lsun/reflect/generics/tree/TypeSignature;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_TypeSignature) => void): void;
    public static "make(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/repository/FieldRepository;"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_FieldRepository) => void): void;
    public static "sun/reflect/generics/repository/FieldRepository/make(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/repository/FieldRepository;"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_FieldRepository) => void): void;
    public "getGenericType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "sun/reflect/generics/repository/FieldRepository/getGenericType()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "parse(Ljava/lang/String;)Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
    public "sun/reflect/generics/repository/FieldRepository/parse(Ljava/lang/String;)Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
  }
  export class sun_reflect_ConstantPool extends java_lang_Object {
    public "sun/reflect/ConstantPool/constantPoolOop": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/ConstantPool/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/ConstantPool/getSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getClassAt(I)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/reflect/ConstantPool/getClassAt(I)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getClassAtIfLoaded(I)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/reflect/ConstantPool/getClassAtIfLoaded(I)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getMethodAt(I)Ljava/lang/reflect/Member;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Member) => void): void;
    public "sun/reflect/ConstantPool/getMethodAt(I)Ljava/lang/reflect/Member;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Member) => void): void;
    public "getMethodAtIfLoaded(I)Ljava/lang/reflect/Member;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Member) => void): void;
    public "sun/reflect/ConstantPool/getMethodAtIfLoaded(I)Ljava/lang/reflect/Member;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Member) => void): void;
    public "getFieldAt(I)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "sun/reflect/ConstantPool/getFieldAt(I)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "getFieldAtIfLoaded(I)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "sun/reflect/ConstantPool/getFieldAtIfLoaded(I)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "getMemberRefInfoAt(I)[Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/reflect/ConstantPool/getMemberRefInfoAt(I)[Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getIntAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/ConstantPool/getIntAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLongAt(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/reflect/ConstantPool/getLongAt(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFloatAt(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/ConstantPool/getFloatAt(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDoubleAt(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/ConstantPool/getDoubleAt(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getStringAt(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/reflect/ConstantPool/getStringAt(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getUTF8At(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/reflect/ConstantPool/getUTF8At(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSize0(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/ConstantPool/getSize0(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getClassAt0(Ljava/lang/Object;I)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/reflect/ConstantPool/getClassAt0(Ljava/lang/Object;I)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getClassAtIfLoaded0(Ljava/lang/Object;I)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "sun/reflect/ConstantPool/getClassAtIfLoaded0(Ljava/lang/Object;I)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getMethodAt0(Ljava/lang/Object;I)Ljava/lang/reflect/Member;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Member) => void): void;
    public "sun/reflect/ConstantPool/getMethodAt0(Ljava/lang/Object;I)Ljava/lang/reflect/Member;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Member) => void): void;
    public "getMethodAtIfLoaded0(Ljava/lang/Object;I)Ljava/lang/reflect/Member;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Member) => void): void;
    public "sun/reflect/ConstantPool/getMethodAtIfLoaded0(Ljava/lang/Object;I)Ljava/lang/reflect/Member;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Member) => void): void;
    public "getFieldAt0(Ljava/lang/Object;I)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "sun/reflect/ConstantPool/getFieldAt0(Ljava/lang/Object;I)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "getFieldAtIfLoaded0(Ljava/lang/Object;I)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "sun/reflect/ConstantPool/getFieldAtIfLoaded0(Ljava/lang/Object;I)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "getMemberRefInfoAt0(Ljava/lang/Object;I)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "sun/reflect/ConstantPool/getMemberRefInfoAt0(Ljava/lang/Object;I)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "getIntAt0(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/ConstantPool/getIntAt0(Ljava/lang/Object;I)I"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLongAt0(Ljava/lang/Object;I)J"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "sun/reflect/ConstantPool/getLongAt0(Ljava/lang/Object;I)J"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFloatAt0(Ljava/lang/Object;I)F"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/ConstantPool/getFloatAt0(Ljava/lang/Object;I)F"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDoubleAt0(Ljava/lang/Object;I)D"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/ConstantPool/getDoubleAt0(Ljava/lang/Object;I)D"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getStringAt0(Ljava/lang/Object;I)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/reflect/ConstantPool/getStringAt0(Ljava/lang/Object;I)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getUTF8At0(Ljava/lang/Object;I)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/reflect/ConstantPool/getUTF8At0(Ljava/lang/Object;I)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/ConstantPool/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_AssertionStatusDirectives extends java_lang_Object {
    public "java/lang/AssertionStatusDirectives/classes": JVMArray<java_lang_String>;
    public "java/lang/AssertionStatusDirectives/classEnabled": JVMArray<number>;
    public "java/lang/AssertionStatusDirectives/packages": JVMArray<java_lang_String>;
    public "java/lang/AssertionStatusDirectives/packageEnabled": JVMArray<number>;
    public "java/lang/AssertionStatusDirectives/deflt": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/AssertionStatusDirectives/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_ByteBuffer extends java_nio_Buffer implements java_lang_Comparable {
    public "java/nio/ByteBuffer/hb": JVMArray<number>;
    public "java/nio/ByteBuffer/offset": number;
    public "java/nio/ByteBuffer/isReadOnly": number;
    public "java/nio/ByteBuffer/bigEndian": number;
    public "java/nio/ByteBuffer/nativeByteOrder": number;
    public "<init>(IIII[BI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/ByteBuffer/<init>(IIII[BI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/ByteBuffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "allocateDirect(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public static "java/nio/ByteBuffer/allocateDirect(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public static "allocate(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public static "java/nio/ByteBuffer/allocate(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public static "wrap([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public static "java/nio/ByteBuffer/wrap([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public static "wrap([B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public static "java/nio/ByteBuffer/wrap([B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "slice()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/slice()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "duplicate()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/duplicate()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asReadOnlyBuffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/asReadOnlyBuffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "get()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/get()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/put(B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "get(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/get(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(IB)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/put(IB)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "get([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/get([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "get([B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/get([B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "put(Ljava/nio/ByteBuffer;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/put(Ljava/nio/ByteBuffer;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "put([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/put([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "put([B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/put([B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "array()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/nio/ByteBuffer/array()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compact()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/compact()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/ByteBuffer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "equals(BB)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/ByteBuffer/equals(BB)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/compareTo(Ljava/nio/ByteBuffer;)I"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(BB)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/ByteBuffer/compare(BB)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "java/nio/ByteBuffer/order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "order(Ljava/nio/ByteOrder;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_ByteOrder], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/order(Ljava/nio/ByteOrder;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_ByteOrder], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "_get(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/_get(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "_put(IB)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/ByteBuffer/_put(IB)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putChar(C)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putChar(C)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "getChar(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getChar(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putChar(IC)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putChar(IC)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asCharBuffer()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/ByteBuffer/asCharBuffer()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "getShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putShort(S)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putShort(S)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "getShort(I)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getShort(I)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putShort(IS)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putShort(IS)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asShortBuffer()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ByteBuffer/asShortBuffer()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "getInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putInt(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putInt(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putInt(II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putInt(II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asIntBuffer()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/ByteBuffer/asIntBuffer()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "getLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/ByteBuffer/getLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putLong(J)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putLong(J)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "getLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/ByteBuffer/getLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putLong(IJ)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putLong(IJ)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asLongBuffer()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/ByteBuffer/asLongBuffer()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "getFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloat(F)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putFloat(F)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "getFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloat(IF)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putFloat(IF)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asFloatBuffer()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/ByteBuffer/asFloatBuffer()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "getDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDouble(D)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putDouble(D)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "getDouble(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/getDouble(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDouble(ID)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/ByteBuffer/putDouble(ID)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asDoubleBuffer()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/ByteBuffer/asDoubleBuffer()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/ByteBuffer/array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ByteBuffer/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_DoubleBuffer extends java_nio_Buffer implements java_lang_Comparable {
    public "java/nio/DoubleBuffer/hb": JVMArray<number>;
    public "java/nio/DoubleBuffer/offset": number;
    public "java/nio/DoubleBuffer/isReadOnly": number;
    public "<init>(IIII[DI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/DoubleBuffer/<init>(IIII[DI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/DoubleBuffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "allocate(I)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public static "java/nio/DoubleBuffer/allocate(I)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public static "wrap([DII)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public static "java/nio/DoubleBuffer/wrap([DII)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public static "wrap([D)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public static "java/nio/DoubleBuffer/wrap([D)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "slice()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/slice()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "duplicate()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/duplicate()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "asReadOnlyBuffer()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/asReadOnlyBuffer()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "get()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DoubleBuffer/get()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(D)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/put(D)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "get(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DoubleBuffer/get(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(ID)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/put(ID)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "get([DII)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/get([DII)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "get([D)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/get([D)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "put(Ljava/nio/DoubleBuffer;)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [java_nio_DoubleBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/put(Ljava/nio/DoubleBuffer;)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [java_nio_DoubleBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "put([DII)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/put([DII)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "put([D)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/put([D)Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DoubleBuffer/hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "array()[D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/nio/DoubleBuffer/array()[D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DoubleBuffer/arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compact()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DoubleBuffer/compact()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DoubleBuffer/isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/DoubleBuffer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DoubleBuffer/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DoubleBuffer/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "equals(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/DoubleBuffer/equals(DD)Z"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/nio/DoubleBuffer;)I"(thread: JVMThread, args: [java_nio_DoubleBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DoubleBuffer/compareTo(Ljava/nio/DoubleBuffer;)I"(thread: JVMThread, args: [java_nio_DoubleBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/DoubleBuffer/compare(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "java/nio/DoubleBuffer/order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/DoubleBuffer/array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DoubleBuffer/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_FloatBuffer extends java_nio_Buffer implements java_lang_Comparable {
    public "java/nio/FloatBuffer/hb": JVMArray<number>;
    public "java/nio/FloatBuffer/offset": number;
    public "java/nio/FloatBuffer/isReadOnly": number;
    public "<init>(IIII[FI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/FloatBuffer/<init>(IIII[FI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/FloatBuffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "allocate(I)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public static "java/nio/FloatBuffer/allocate(I)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public static "wrap([FII)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public static "java/nio/FloatBuffer/wrap([FII)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public static "wrap([F)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public static "java/nio/FloatBuffer/wrap([F)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "slice()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/slice()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "duplicate()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/duplicate()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "asReadOnlyBuffer()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/asReadOnlyBuffer()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "get()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/FloatBuffer/get()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(F)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/put(F)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "get(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/FloatBuffer/get(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(IF)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/put(IF)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "get([FII)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/get([FII)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "get([F)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/get([F)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "put(Ljava/nio/FloatBuffer;)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [java_nio_FloatBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/put(Ljava/nio/FloatBuffer;)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [java_nio_FloatBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "put([FII)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/put([FII)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "put([F)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/put([F)Ljava/nio/FloatBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/FloatBuffer/hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "array()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/nio/FloatBuffer/array()[F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/FloatBuffer/arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compact()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/FloatBuffer/compact()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/FloatBuffer/isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/FloatBuffer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/FloatBuffer/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/FloatBuffer/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "equals(FF)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/FloatBuffer/equals(FF)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/nio/FloatBuffer;)I"(thread: JVMThread, args: [java_nio_FloatBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/FloatBuffer/compareTo(Ljava/nio/FloatBuffer;)I"(thread: JVMThread, args: [java_nio_FloatBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(FF)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/FloatBuffer/compare(FF)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "java/nio/FloatBuffer/order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/FloatBuffer/array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/FloatBuffer/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_IntBuffer extends java_nio_Buffer implements java_lang_Comparable {
    public "java/nio/IntBuffer/hb": JVMArray<number>;
    public "java/nio/IntBuffer/offset": number;
    public "java/nio/IntBuffer/isReadOnly": number;
    public "<init>(IIII[II)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/IntBuffer/<init>(IIII[II)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/IntBuffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "allocate(I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public static "java/nio/IntBuffer/allocate(I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public static "wrap([III)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public static "java/nio/IntBuffer/wrap([III)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public static "wrap([I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public static "java/nio/IntBuffer/wrap([I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "slice()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/slice()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "duplicate()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/duplicate()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "asReadOnlyBuffer()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/asReadOnlyBuffer()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "get()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/IntBuffer/get()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/put(I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "get(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/IntBuffer/get(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(II)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/put(II)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "get([III)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/get([III)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "get([I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/get([I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "put(Ljava/nio/IntBuffer;)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [java_nio_IntBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/put(Ljava/nio/IntBuffer;)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [java_nio_IntBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "put([III)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/put([III)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "put([I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/put([I)Ljava/nio/IntBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/IntBuffer/hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "array()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/nio/IntBuffer/array()[I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/IntBuffer/arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compact()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/IntBuffer/compact()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/IntBuffer/isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/IntBuffer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/IntBuffer/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/IntBuffer/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "equals(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/IntBuffer/equals(II)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/nio/IntBuffer;)I"(thread: JVMThread, args: [java_nio_IntBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/IntBuffer/compareTo(Ljava/nio/IntBuffer;)I"(thread: JVMThread, args: [java_nio_IntBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/IntBuffer/compare(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "java/nio/IntBuffer/order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/IntBuffer/array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/IntBuffer/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_nio_ShortBuffer extends java_nio_Buffer implements java_lang_Comparable {
    public "java/nio/ShortBuffer/hb": JVMArray<number>;
    public "java/nio/ShortBuffer/offset": number;
    public "java/nio/ShortBuffer/isReadOnly": number;
    public "<init>(IIII[SI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/ShortBuffer/<init>(IIII[SI)V"(thread: JVMThread, args: [number, number, number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/ShortBuffer/<init>(IIII)V"(thread: JVMThread, args: [number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "allocate(I)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public static "java/nio/ShortBuffer/allocate(I)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public static "wrap([SII)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public static "java/nio/ShortBuffer/wrap([SII)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public static "wrap([S)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public static "java/nio/ShortBuffer/wrap([S)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "slice()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/slice()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "duplicate()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/duplicate()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "asReadOnlyBuffer()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/asReadOnlyBuffer()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "get()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ShortBuffer/get()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(S)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/put(S)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "get(I)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ShortBuffer/get(I)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "put(IS)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/put(IS)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "get([SII)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/get([SII)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "get([S)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/get([S)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "put(Ljava/nio/ShortBuffer;)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [java_nio_ShortBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/put(Ljava/nio/ShortBuffer;)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [java_nio_ShortBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "put([SII)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/put([SII)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "put([S)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/put([S)Ljava/nio/ShortBuffer;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ShortBuffer/hasArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "array()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/nio/ShortBuffer/array()[S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ShortBuffer/arrayOffset()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compact()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/ShortBuffer/compact()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ShortBuffer/isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/ShortBuffer/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ShortBuffer/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ShortBuffer/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "equals(SS)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/ShortBuffer/equals(SS)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/nio/ShortBuffer;)I"(thread: JVMThread, args: [java_nio_ShortBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ShortBuffer/compareTo(Ljava/nio/ShortBuffer;)I"(thread: JVMThread, args: [java_nio_ShortBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(SS)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/nio/ShortBuffer/compare(SS)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "java/nio/ShortBuffer/order()Ljava/nio/ByteOrder;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteOrder) => void): void;
    public "array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/ShortBuffer/array()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/ShortBuffer/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_security_ProtectionDomain extends java_lang_Object {
    public "java/security/ProtectionDomain/codesource": java_security_CodeSource;
    public "java/security/ProtectionDomain/classloader": java_lang_ClassLoader;
    public "java/security/ProtectionDomain/principals": JVMArray<java_security_Principal>;
    public "java/security/ProtectionDomain/permissions": java_security_PermissionCollection;
    public "java/security/ProtectionDomain/hasAllPerm": number;
    public "java/security/ProtectionDomain/staticPermissions": number;
    public "java/security/ProtectionDomain/key": java_security_ProtectionDomain$Key;
    public static "java/security/ProtectionDomain/debug": sun_security_util_Debug;
    public "<init>(Ljava/security/CodeSource;Ljava/security/PermissionCollection;)V"(thread: JVMThread, args: [java_security_CodeSource, java_security_PermissionCollection], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/ProtectionDomain/<init>(Ljava/security/CodeSource;Ljava/security/PermissionCollection;)V"(thread: JVMThread, args: [java_security_CodeSource, java_security_PermissionCollection], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/security/CodeSource;Ljava/security/PermissionCollection;Ljava/lang/ClassLoader;[Ljava/security/Principal;)V"(thread: JVMThread, args: [java_security_CodeSource, java_security_PermissionCollection, java_lang_ClassLoader, JVMArray<java_security_Principal>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/ProtectionDomain/<init>(Ljava/security/CodeSource;Ljava/security/PermissionCollection;Ljava/lang/ClassLoader;[Ljava/security/Principal;)V"(thread: JVMThread, args: [java_security_CodeSource, java_security_PermissionCollection, java_lang_ClassLoader, JVMArray<java_security_Principal>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCodeSource()Ljava/security/CodeSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "java/security/ProtectionDomain/getCodeSource()Ljava/security/CodeSource;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_CodeSource) => void): void;
    public "getClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/security/ProtectionDomain/getClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "getPrincipals()[Ljava/security/Principal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_Principal>) => void): void;
    public "java/security/ProtectionDomain/getPrincipals()[Ljava/security/Principal;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_security_Principal>) => void): void;
    public "getPermissions()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public "java/security/ProtectionDomain/getPermissions()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public "implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/ProtectionDomain/implies(Ljava/security/Permission;)Z"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "impliesCreateAccessControlContext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/security/ProtectionDomain/impliesCreateAccessControlContext()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/ProtectionDomain/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "seeAllp()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/security/ProtectionDomain/seeAllp()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mergePermissions()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public "java/security/ProtectionDomain/mergePermissions()Ljava/security/PermissionCollection;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_PermissionCollection) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/security/ProtectionDomain/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_ProtectionDomain$Key extends java_lang_Object {
    public "java/security/ProtectionDomain$Key/this$0": java_security_ProtectionDomain;
    public "<init>(Ljava/security/ProtectionDomain;)V"(thread: JVMThread, args: [java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/ProtectionDomain$Key/<init>(Ljava/security/ProtectionDomain;)V"(thread: JVMThread, args: [java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_Properties extends java_util_Hashtable {
    public static "java/util/Properties/serialVersionUID": Long;
    public "java/util/Properties/defaults": java_util_Properties;
    public static "java/util/Properties/hexDigit": JVMArray<number>;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Properties;)V"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/<init>(Ljava/util/Properties;)V"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable) => void): void;
    public "setProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/util/Properties/setProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "load(Ljava/io/Reader;)V"(thread: JVMThread, args: [java_io_Reader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/load(Ljava/io/Reader;)V"(thread: JVMThread, args: [java_io_Reader], cb?: (e?: java_lang_Throwable) => void): void;
    public "load(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/load(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "load0(Ljava/util/Properties$LineReader;)V"(thread: JVMThread, args: [java_util_Properties$LineReader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/load0(Ljava/util/Properties$LineReader;)V"(thread: JVMThread, args: [java_util_Properties$LineReader], cb?: (e?: java_lang_Throwable) => void): void;
    public "loadConvert([CII[C)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Properties/loadConvert([CII[C)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "saveConvert(Ljava/lang/String;ZZ)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Properties/saveConvert(Ljava/lang/String;ZZ)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "writeComments(Ljava/io/BufferedWriter;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_BufferedWriter, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Properties/writeComments(Ljava/io/BufferedWriter;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_BufferedWriter, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "save(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/save(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "store(Ljava/io/Writer;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_Writer, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/store(Ljava/io/Writer;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_Writer, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "store(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/store(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "store0(Ljava/io/BufferedWriter;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_io_BufferedWriter, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/store0(Ljava/io/BufferedWriter;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_io_BufferedWriter, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "loadFromXML(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/loadFromXML(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "storeToXML(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/storeToXML(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "storeToXML(Ljava/io/OutputStream;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/storeToXML(Ljava/io/OutputStream;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Properties/getProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/util/Properties/getProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "propertyNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/util/Properties/propertyNames()Ljava/util/Enumeration;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "stringPropertyNames()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "java/util/Properties/stringPropertyNames()Ljava/util/Set;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Set) => void): void;
    public "list(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/list(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "list(Ljava/io/PrintWriter;)V"(thread: JVMThread, args: [java_io_PrintWriter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/list(Ljava/io/PrintWriter;)V"(thread: JVMThread, args: [java_io_PrintWriter], cb?: (e?: java_lang_Throwable) => void): void;
    public "enumerate(Ljava/util/Hashtable;)V"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/enumerate(Ljava/util/Hashtable;)V"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public "enumerateStringProperties(Ljava/util/Hashtable;)V"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties/enumerateStringProperties(Ljava/util/Hashtable;)V"(thread: JVMThread, args: [java_util_Hashtable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "toHex(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/util/Properties/toHex(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/util/Properties/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_BufferedWriter extends java_io_Writer {
    public "java/io/BufferedWriter/out": java_io_Writer;
    public "java/io/BufferedWriter/cb": JVMArray<number>;
    public "java/io/BufferedWriter/nChars": number;
    public "java/io/BufferedWriter/nextChar": number;
    public static "java/io/BufferedWriter/defaultCharBufferSize": number;
    public "java/io/BufferedWriter/lineSeparator": java_lang_String;
    public "<init>(Ljava/io/Writer;)V"(thread: JVMThread, args: [java_io_Writer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/<init>(Ljava/io/Writer;)V"(thread: JVMThread, args: [java_io_Writer], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/Writer;I)V"(thread: JVMThread, args: [java_io_Writer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/<init>(Ljava/io/Writer;I)V"(thread: JVMThread, args: [java_io_Writer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flushBuffer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/flushBuffer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "min(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/BufferedWriter/min(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "newLine()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/newLine()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/BufferedWriter/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/BufferedWriter/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_util_Properties$LineReader extends java_lang_Object {
    public "java/util/Properties$LineReader/inByteBuf": JVMArray<number>;
    public "java/util/Properties$LineReader/inCharBuf": JVMArray<number>;
    public "java/util/Properties$LineReader/lineBuf": JVMArray<number>;
    public "java/util/Properties$LineReader/inLimit": number;
    public "java/util/Properties$LineReader/inOff": number;
    public "java/util/Properties$LineReader/inStream": java_io_InputStream;
    public "java/util/Properties$LineReader/reader": java_io_Reader;
    public "java/util/Properties$LineReader/this$0": java_util_Properties;
    public "<init>(Ljava/util/Properties;Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_util_Properties, java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties$LineReader/<init>(Ljava/util/Properties;Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_util_Properties, java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/util/Properties;Ljava/io/Reader;)V"(thread: JVMThread, args: [java_util_Properties, java_io_Reader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/util/Properties$LineReader/<init>(Ljava/util/Properties;Ljava/io/Reader;)V"(thread: JVMThread, args: [java_util_Properties, java_io_Reader], cb?: (e?: java_lang_Throwable) => void): void;
    public "readLine()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/util/Properties$LineReader/readLine()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_io_PrintStream extends java_io_FilterOutputStream implements java_lang_Appendable, java_io_Closeable {
    public "java/io/PrintStream/autoFlush": number;
    public "java/io/PrintStream/trouble": number;
    public "java/io/PrintStream/formatter": java_util_Formatter;
    public "java/io/PrintStream/textOut": java_io_BufferedWriter;
    public "java/io/PrintStream/charOut": java_io_OutputStreamWriter;
    public "java/io/PrintStream/closing": number;
    public static "requireNonNull(Ljava/lang/Object;Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/io/PrintStream/requireNonNull(Ljava/lang/Object;Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "toCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/io/PrintStream/toCharset(Ljava/lang/String;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public "<init>(ZLjava/io/OutputStream;)V"(thread: JVMThread, args: [number, java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(ZLjava/io/OutputStream;)V"(thread: JVMThread, args: [number, java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ZLjava/io/OutputStream;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [number, java_io_OutputStream, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(ZLjava/io/OutputStream;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [number, java_io_OutputStream, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(ZLjava/nio/charset/Charset;Ljava/io/OutputStream;)V"(thread: JVMThread, args: [number, java_nio_charset_Charset, java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(ZLjava/nio/charset/Charset;Ljava/io/OutputStream;)V"(thread: JVMThread, args: [number, java_nio_charset_Charset, java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;Z)V"(thread: JVMThread, args: [java_io_OutputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(Ljava/io/OutputStream;Z)V"(thread: JVMThread, args: [java_io_OutputStream, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;ZLjava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(Ljava/io/OutputStream;ZLjava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkError()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/PrintStream/checkError()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/setError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/clearError()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/write([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/write(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "newLine()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/newLine()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/print(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/print(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/print(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/print(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/print(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/print(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "print([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/print([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/print(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "print(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/print(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "println()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println(C)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "println([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "println(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/PrintStream/println(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "printf(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "java/io/PrintStream/printf(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "printf(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "java/io/PrintStream/printf(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "java/io/PrintStream/format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "format(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "java/io/PrintStream/format(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "java/io/PrintStream/append(Ljava/lang/CharSequence;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "java/io/PrintStream/append(Ljava/lang/CharSequence;II)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "append(C)Ljava/io/PrintStream;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "java/io/PrintStream/append(C)Ljava/io/PrintStream;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public "append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/io/PrintStream/append(C)Ljava/lang/Appendable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/io/PrintStream/append(Ljava/lang/CharSequence;II)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
    public "java/io/PrintStream/append(Ljava/lang/CharSequence;)Ljava/lang/Appendable;"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_Appendable) => void): void;
  }
  export class java_io_OutputStreamWriter extends java_io_Writer {
    public "java/io/OutputStreamWriter/se": sun_nio_cs_StreamEncoder;
    public "<init>(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/<init>(Ljava/io/OutputStream;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_io_OutputStream, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/<init>(Ljava/io/OutputStream;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_io_OutputStream, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;Ljava/nio/charset/CharsetEncoder;)V"(thread: JVMThread, args: [java_io_OutputStream, java_nio_charset_CharsetEncoder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/<init>(Ljava/io/OutputStream;Ljava/nio/charset/CharsetEncoder;)V"(thread: JVMThread, args: [java_io_OutputStream, java_nio_charset_CharsetEncoder], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEncoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/OutputStreamWriter/getEncoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "flushBuffer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/flushBuffer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/OutputStreamWriter/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_cs_StreamEncoder extends java_io_Writer {
    public static "sun/nio/cs/StreamEncoder/DEFAULT_BYTE_BUFFER_SIZE": number;
    public "sun/nio/cs/StreamEncoder/isOpen": number;
    public "sun/nio/cs/StreamEncoder/cs": java_nio_charset_Charset;
    public "sun/nio/cs/StreamEncoder/encoder": java_nio_charset_CharsetEncoder;
    public "sun/nio/cs/StreamEncoder/bb": java_nio_ByteBuffer;
    public "sun/nio/cs/StreamEncoder/out": java_io_OutputStream;
    public "sun/nio/cs/StreamEncoder/ch": java_nio_channels_WritableByteChannel;
    public "sun/nio/cs/StreamEncoder/haveLeftoverChar": number;
    public "sun/nio/cs/StreamEncoder/leftoverChar": number;
    public "sun/nio/cs/StreamEncoder/lcb": java_nio_CharBuffer;
    public static "sun/nio/cs/StreamEncoder/$assertionsDisabled": number;
    public "ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/ensureOpen()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "forOutputStreamWriter(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/lang/String;)Lsun/nio/cs/StreamEncoder;"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_cs_StreamEncoder) => void): void;
    public static "sun/nio/cs/StreamEncoder/forOutputStreamWriter(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/lang/String;)Lsun/nio/cs/StreamEncoder;"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_nio_cs_StreamEncoder) => void): void;
    public static "forOutputStreamWriter(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/nio/charset/Charset;)Lsun/nio/cs/StreamEncoder;"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: sun_nio_cs_StreamEncoder) => void): void;
    public static "sun/nio/cs/StreamEncoder/forOutputStreamWriter(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/nio/charset/Charset;)Lsun/nio/cs/StreamEncoder;"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: sun_nio_cs_StreamEncoder) => void): void;
    public static "forOutputStreamWriter(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/nio/charset/CharsetEncoder;)Lsun/nio/cs/StreamEncoder;"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_nio_charset_CharsetEncoder], cb?: (e?: java_lang_Throwable, rv?: sun_nio_cs_StreamEncoder) => void): void;
    public static "sun/nio/cs/StreamEncoder/forOutputStreamWriter(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/nio/charset/CharsetEncoder;)Lsun/nio/cs/StreamEncoder;"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_nio_charset_CharsetEncoder], cb?: (e?: java_lang_Throwable, rv?: sun_nio_cs_StreamEncoder) => void): void;
    public static "forEncoder(Ljava/nio/channels/WritableByteChannel;Ljava/nio/charset/CharsetEncoder;I)Lsun/nio/cs/StreamEncoder;"(thread: JVMThread, args: [java_nio_channels_WritableByteChannel, java_nio_charset_CharsetEncoder, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_cs_StreamEncoder) => void): void;
    public static "sun/nio/cs/StreamEncoder/forEncoder(Ljava/nio/channels/WritableByteChannel;Ljava/nio/charset/CharsetEncoder;I)Lsun/nio/cs/StreamEncoder;"(thread: JVMThread, args: [java_nio_channels_WritableByteChannel, java_nio_charset_CharsetEncoder, number], cb?: (e?: java_lang_Throwable, rv?: sun_nio_cs_StreamEncoder) => void): void;
    public "getEncoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/cs/StreamEncoder/getEncoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "flushBuffer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/flushBuffer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/write([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/write(Ljava/lang/String;II)V"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/cs/StreamEncoder/isOpen()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/<init>(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/nio/charset/Charset;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/nio/charset/CharsetEncoder;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_nio_charset_CharsetEncoder], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/<init>(Ljava/io/OutputStream;Ljava/lang/Object;Ljava/nio/charset/CharsetEncoder;)V"(thread: JVMThread, args: [java_io_OutputStream, java_lang_Object, java_nio_charset_CharsetEncoder], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/nio/channels/WritableByteChannel;Ljava/nio/charset/CharsetEncoder;I)V"(thread: JVMThread, args: [java_nio_channels_WritableByteChannel, java_nio_charset_CharsetEncoder, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/<init>(Ljava/nio/channels/WritableByteChannel;Ljava/nio/charset/CharsetEncoder;I)V"(thread: JVMThread, args: [java_nio_channels_WritableByteChannel, java_nio_charset_CharsetEncoder, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBytes()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/writeBytes()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "flushLeftoverChar(Ljava/nio/CharBuffer;Z)V"(thread: JVMThread, args: [java_nio_CharBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/flushLeftoverChar(Ljava/nio/CharBuffer;Z)V"(thread: JVMThread, args: [java_nio_CharBuffer, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "implWrite([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/implWrite([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "implFlushBuffer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/implFlushBuffer()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implFlush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/implFlush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/cs/StreamEncoder/implClose()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "encodingName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/cs/StreamEncoder/encodingName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/nio/cs/StreamEncoder/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_InputStream extends java_lang_Object implements java_io_Closeable {
    public static "java/io/InputStream/MAX_SKIP_BUFFER_SIZE": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/InputStream/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/InputStream/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/InputStream/read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/InputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/InputStream/skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/InputStream/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/InputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "mark(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/InputStream/mark(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/InputStream/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "markSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/InputStream/markSupported()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_lang_System extends java_lang_Object {
    public static "java/lang/System/in": java_io_InputStream;
    public static "java/lang/System/out": java_io_PrintStream;
    public static "java/lang/System/err": java_io_PrintStream;
    public static "java/lang/System/security": java_lang_SecurityManager;
    public static "java/lang/System/cons": java_io_Console;
    public static "java/lang/System/props": java_util_Properties;
    public static "java/lang/System/lineSeparator": java_lang_String;
    public static "registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/System/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setIn(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setIn(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setOut(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setOut(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setErr(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setErr(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "console()Ljava/io/Console;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public static "java/lang/System/console()Ljava/io/Console;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public static "inheritedChannel()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public static "java/lang/System/inheritedChannel()Ljava/nio/channels/Channel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_Channel) => void): void;
    public static "checkIO()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/checkIO()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setIn0(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setIn0(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setOut0(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setOut0(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setErr0(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setErr0(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setSecurityManager(Ljava/lang/SecurityManager;)V"(thread: JVMThread, args: [java_lang_SecurityManager], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setSecurityManager(Ljava/lang/SecurityManager;)V"(thread: JVMThread, args: [java_lang_SecurityManager], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setSecurityManager0(Ljava/lang/SecurityManager;)V"(thread: JVMThread, args: [java_lang_SecurityManager], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setSecurityManager0(Ljava/lang/SecurityManager;)V"(thread: JVMThread, args: [java_lang_SecurityManager], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getSecurityManager()Ljava/lang/SecurityManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_SecurityManager) => void): void;
    public static "java/lang/System/getSecurityManager()Ljava/lang/SecurityManager;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_SecurityManager) => void): void;
    public static "currentTimeMillis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/System/currentTimeMillis()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "nanoTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/System/nanoTime()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "arraycopy(Ljava/lang/Object;ILjava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/arraycopy(Ljava/lang/Object;ILjava/lang/Object;II)V"(thread: JVMThread, args: [java_lang_Object, number, java_lang_Object, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "identityHashCode(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/System/identityHashCode(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initProperties(Ljava/util/Properties;)Ljava/util/Properties;"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "java/lang/System/initProperties(Ljava/util/Properties;)Ljava/util/Properties;"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "getProperties()Ljava/util/Properties;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "java/lang/System/getProperties()Ljava/util/Properties;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Properties) => void): void;
    public static "lineSeparator()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/System/lineSeparator()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "setProperties(Ljava/util/Properties;)V"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setProperties(Ljava/util/Properties;)V"(thread: JVMThread, args: [java_util_Properties], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/System/getProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/System/getProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "setProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/System/setProperty(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "clearProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/System/clearProperty(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "checkKey(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/checkKey(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getenv(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/System/getenv(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getenv()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/lang/System/getenv()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "exit(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/exit(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "gc()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/gc()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "runFinalization()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/runFinalization()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "runFinalizersOnExit(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/runFinalizersOnExit(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "load(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/load(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "loadLibrary(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/loadLibrary(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "mapLibraryName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/System/mapLibraryName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "newPrintStream(Ljava/io/FileOutputStream;Ljava/lang/String;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_io_FileOutputStream, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public static "java/lang/System/newPrintStream(Ljava/io/FileOutputStream;Ljava/lang/String;)Ljava/io/PrintStream;"(thread: JVMThread, args: [java_io_FileOutputStream, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintStream) => void): void;
    public static "initializeSystemClass()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/initializeSystemClass()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setJavaLangAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/setJavaLangAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/System/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ClassLoader extends java_lang_Object {
    public $loader: ClassLoader;
    public "java/lang/ClassLoader/parent": java_lang_ClassLoader;
    public "java/lang/ClassLoader/parallelLockMap": java_util_concurrent_ConcurrentHashMap;
    public "java/lang/ClassLoader/package2certs": java_util_Map;
    public static "java/lang/ClassLoader/nocerts": JVMArray<java_security_cert_Certificate>;
    public "java/lang/ClassLoader/classes": java_util_Vector;
    public "java/lang/ClassLoader/defaultDomain": java_security_ProtectionDomain;
    public "java/lang/ClassLoader/domains": java_util_Set;
    public "java/lang/ClassLoader/packages": java_util_HashMap;
    public static "java/lang/ClassLoader/scl": java_lang_ClassLoader;
    public static "java/lang/ClassLoader/sclSet": number;
    public static "java/lang/ClassLoader/loadedLibraryNames": java_util_Vector;
    public static "java/lang/ClassLoader/systemNativeLibraries": java_util_Vector;
    public "java/lang/ClassLoader/nativeLibraries": java_util_Vector;
    public static "java/lang/ClassLoader/nativeLibraryContext": java_util_Stack;
    public static "java/lang/ClassLoader/usr_paths": JVMArray<java_lang_String>;
    public static "java/lang/ClassLoader/sys_paths": JVMArray<java_lang_String>;
    public "java/lang/ClassLoader/assertionLock": java_lang_Object;
    public "java/lang/ClassLoader/defaultAssertionStatus": number;
    public "java/lang/ClassLoader/packageAssertionStatus": java_util_Map;
    public "java/lang/ClassLoader/classAssertionStatus": java_util_Map;
    public static "registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ClassLoader/registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/addClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkCreateClassLoader()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "java/lang/ClassLoader/checkCreateClassLoader()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "<init>(Ljava/lang/Void;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_Void, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/<init>(Ljava/lang/Void;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_Void, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/<init>(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "loadClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/loadClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "loadClass(Ljava/lang/String;Z)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/loadClass(Ljava/lang/String;Z)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getClassLoadingLock(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ClassLoader/getClassLoadingLock(Ljava/lang/String;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "loadClassInternal(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/loadClassInternal(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "checkPackageAccess(Ljava/lang/Class;Ljava/security/ProtectionDomain;)V"(thread: JVMThread, args: [java_lang_Class, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/checkPackageAccess(Ljava/lang/Class;Ljava/security/ProtectionDomain;)V"(thread: JVMThread, args: [java_lang_Class, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable) => void): void;
    public "findClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/findClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "defineClass([BII)Ljava/lang/Class;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/defineClass([BII)Ljava/lang/Class;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "defineClass(Ljava/lang/String;[BII)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/defineClass(Ljava/lang/String;[BII)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "preDefineClass(Ljava/lang/String;Ljava/security/ProtectionDomain;)Ljava/security/ProtectionDomain;"(thread: JVMThread, args: [java_lang_String, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_security_ProtectionDomain) => void): void;
    public "java/lang/ClassLoader/preDefineClass(Ljava/lang/String;Ljava/security/ProtectionDomain;)Ljava/security/ProtectionDomain;"(thread: JVMThread, args: [java_lang_String, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_security_ProtectionDomain) => void): void;
    public "defineClassSourceLocation(Ljava/security/ProtectionDomain;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/ClassLoader/defineClassSourceLocation(Ljava/security/ProtectionDomain;)Ljava/lang/String;"(thread: JVMThread, args: [java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "postDefineClass(Ljava/lang/Class;Ljava/security/ProtectionDomain;)V"(thread: JVMThread, args: [java_lang_Class, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/postDefineClass(Ljava/lang/Class;Ljava/security/ProtectionDomain;)V"(thread: JVMThread, args: [java_lang_Class, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable) => void): void;
    public "defineClass(Ljava/lang/String;[BIILjava/security/ProtectionDomain;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/defineClass(Ljava/lang/String;[BIILjava/security/ProtectionDomain;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "defineClass(Ljava/lang/String;Ljava/nio/ByteBuffer;Ljava/security/ProtectionDomain;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, java_nio_ByteBuffer, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/defineClass(Ljava/lang/String;Ljava/nio/ByteBuffer;Ljava/security/ProtectionDomain;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, java_nio_ByteBuffer, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "defineClass0(Ljava/lang/String;[BIILjava/security/ProtectionDomain;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/defineClass0(Ljava/lang/String;[BIILjava/security/ProtectionDomain;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number, java_security_ProtectionDomain], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "defineClass1(Ljava/lang/String;[BIILjava/security/ProtectionDomain;Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number, java_security_ProtectionDomain, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/defineClass1(Ljava/lang/String;[BIILjava/security/ProtectionDomain;Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, JVMArray<number>, number, number, java_security_ProtectionDomain, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "defineClass2(Ljava/lang/String;Ljava/nio/ByteBuffer;IILjava/security/ProtectionDomain;Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, java_nio_ByteBuffer, number, number, java_security_ProtectionDomain, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/defineClass2(Ljava/lang/String;Ljava/nio/ByteBuffer;IILjava/security/ProtectionDomain;Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, java_nio_ByteBuffer, number, number, java_security_ProtectionDomain, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "checkName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ClassLoader/checkName(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkCerts(Ljava/lang/String;Ljava/security/CodeSource;)V"(thread: JVMThread, args: [java_lang_String, java_security_CodeSource], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/checkCerts(Ljava/lang/String;Ljava/security/CodeSource;)V"(thread: JVMThread, args: [java_lang_String, java_security_CodeSource], cb?: (e?: java_lang_Throwable) => void): void;
    public "compareCerts([Ljava/security/cert/Certificate;[Ljava/security/cert/Certificate;)Z"(thread: JVMThread, args: [JVMArray<java_security_cert_Certificate>, JVMArray<java_security_cert_Certificate>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ClassLoader/compareCerts([Ljava/security/cert/Certificate;[Ljava/security/cert/Certificate;)Z"(thread: JVMThread, args: [JVMArray<java_security_cert_Certificate>, JVMArray<java_security_cert_Certificate>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "resolveClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/resolveClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveClass0(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/resolveClass0(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "findSystemClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/findSystemClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "findBootstrapClassOrNull(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/findBootstrapClassOrNull(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "findBootstrapClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/findBootstrapClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "findLoadedClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/findLoadedClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "findLoadedClass0(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/ClassLoader/findLoadedClass0(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "setSigners(Ljava/lang/Class;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/setSigners(Ljava/lang/Class;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Class, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "java/lang/ClassLoader/getResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "getResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/lang/ClassLoader/getResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "findResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "java/lang/ClassLoader/findResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "findResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public "java/lang/ClassLoader/findResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "registerAsParallelCapable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ClassLoader/registerAsParallelCapable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getSystemResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public static "java/lang/ClassLoader/getSystemResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public static "getSystemResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "java/lang/ClassLoader/getSystemResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "getBootstrapResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public static "java/lang/ClassLoader/getBootstrapResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public static "getBootstrapResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "java/lang/ClassLoader/getBootstrapResources(Ljava/lang/String;)Ljava/util/Enumeration;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_Enumeration) => void): void;
    public static "getBootstrapClassPath()Lsun/misc/URLClassPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath) => void): void;
    public static "java/lang/ClassLoader/getBootstrapClassPath()Lsun/misc/URLClassPath;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_URLClassPath) => void): void;
    public "getResourceAsStream(Ljava/lang/String;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/lang/ClassLoader/getResourceAsStream(Ljava/lang/String;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public static "getSystemResourceAsStream(Ljava/lang/String;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public static "java/lang/ClassLoader/getSystemResourceAsStream(Ljava/lang/String;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getParent()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/lang/ClassLoader/getParent()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "getSystemClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "java/lang/ClassLoader/getSystemClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "initSystemClassLoader()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ClassLoader/initSystemClassLoader()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAncestor(Ljava/lang/ClassLoader;)Z"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ClassLoader/isAncestor(Ljava/lang/ClassLoader;)Z"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "needsClassLoaderPermissionCheck(Ljava/lang/ClassLoader;Ljava/lang/ClassLoader;)Z"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ClassLoader/needsClassLoaderPermissionCheck(Ljava/lang/ClassLoader;Ljava/lang/ClassLoader;)Z"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getClassLoader(Ljava/lang/Class;)Ljava/lang/ClassLoader;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "java/lang/ClassLoader/getClassLoader(Ljava/lang/Class;)Ljava/lang/ClassLoader;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "checkClassLoaderPermission(Ljava/lang/ClassLoader;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ClassLoader/checkClassLoaderPermission(Ljava/lang/ClassLoader;Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "definePackage(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/net/URL;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public "java/lang/ClassLoader/definePackage(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/net/URL;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_net_URL], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public "getPackage(Ljava/lang/String;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public "java/lang/ClassLoader/getPackage(Ljava/lang/String;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public "getPackages()[Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Package>) => void): void;
    public "java/lang/ClassLoader/getPackages()[Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Package>) => void): void;
    public "findLibrary(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/ClassLoader/findLibrary(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "initializePath(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/lang/ClassLoader/initializePath(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "loadLibrary(Ljava/lang/Class;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ClassLoader/loadLibrary(Ljava/lang/Class;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "findBuiltinLib(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/ClassLoader/findBuiltinLib(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "loadLibrary0(Ljava/lang/Class;Ljava/io/File;)Z"(thread: JVMThread, args: [java_lang_Class, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ClassLoader/loadLibrary0(Ljava/lang/Class;Ljava/io/File;)Z"(thread: JVMThread, args: [java_lang_Class, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "findNative(Ljava/lang/ClassLoader;Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/ClassLoader/findNative(Ljava/lang/ClassLoader;Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_ClassLoader, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setDefaultAssertionStatus(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/setDefaultAssertionStatus(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPackageAssertionStatus(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/setPackageAssertionStatus(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setClassAssertionStatus(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/setClassAssertionStatus(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "clearAssertionStatus()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/clearAssertionStatus()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "desiredAssertionStatus(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ClassLoader/desiredAssertionStatus(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "initializeJavaAssertionMaps()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader/initializeJavaAssertionMaps()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "retrieveDirectives()Ljava/lang/AssertionStatusDirectives;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AssertionStatusDirectives) => void): void;
    public static "java/lang/ClassLoader/retrieveDirectives()Ljava/lang/AssertionStatusDirectives;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_AssertionStatusDirectives) => void): void;
    public static "access$000()Ljava/util/Vector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    public static "java/lang/ClassLoader/access$000()Ljava/util/Vector;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Vector) => void): void;
    public static "access$100()Ljava/util/Stack;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Stack) => void): void;
    public static "java/lang/ClassLoader/access$100()Ljava/util/Stack;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Stack) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ClassLoader/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_SecurityManager extends java_lang_Object {
    public "java/lang/SecurityManager/inCheck": number;
    public "java/lang/SecurityManager/initialized": number;
    public static "java/lang/SecurityManager/rootGroup": java_lang_ThreadGroup;
    public static "java/lang/SecurityManager/packageAccessValid": number;
    public static "java/lang/SecurityManager/packageAccess": JVMArray<java_lang_String>;
    public static "java/lang/SecurityManager/packageAccessLock": java_lang_Object;
    public static "java/lang/SecurityManager/packageDefinitionValid": number;
    public static "java/lang/SecurityManager/packageDefinition": JVMArray<java_lang_String>;
    public static "java/lang/SecurityManager/packageDefinitionLock": java_lang_Object;
    public "hasAllPermission()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/SecurityManager/hasAllPermission()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInCheck()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/SecurityManager/getInCheck()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getClassContext()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/SecurityManager/getClassContext()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "currentClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/lang/SecurityManager/currentClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "currentClassLoader0()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/lang/SecurityManager/currentClassLoader0()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "currentLoadedClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/SecurityManager/currentLoadedClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "classDepth(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/SecurityManager/classDepth(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "classLoaderDepth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/SecurityManager/classLoaderDepth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "classLoaderDepth0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/SecurityManager/classLoaderDepth0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inClass(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/SecurityManager/inClass(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "inClassLoader()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/SecurityManager/inClassLoader()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSecurityContext()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/SecurityManager/getSecurityContext()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "checkPermission(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkPermission(Ljava/security/Permission;)V"(thread: JVMThread, args: [java_security_Permission], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkPermission(Ljava/security/Permission;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_security_Permission, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkPermission(Ljava/security/Permission;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_security_Permission, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkCreateClassLoader()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkCreateClassLoader()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getRootGroup()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public static "java/lang/SecurityManager/getRootGroup()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public "checkAccess(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkAccess(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAccess(Ljava/lang/ThreadGroup;)V"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkAccess(Ljava/lang/ThreadGroup;)V"(thread: JVMThread, args: [java_lang_ThreadGroup], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkExit(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkExit(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkExec(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkExec(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkLink(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkLink(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkRead(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkRead(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkRead(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkRead(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkRead(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkRead(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkWrite(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkWrite(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkWrite(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkWrite(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkDelete(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkDelete(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkConnect(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkConnect(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkConnect(Ljava/lang/String;ILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkConnect(Ljava/lang/String;ILjava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkListen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkListen(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAccept(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkAccept(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkMulticast(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkMulticast(Ljava/net/InetAddress;)V"(thread: JVMThread, args: [java_net_InetAddress], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkMulticast(Ljava/net/InetAddress;B)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkMulticast(Ljava/net/InetAddress;B)V"(thread: JVMThread, args: [java_net_InetAddress, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkPropertiesAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkPropertiesAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkPropertyAccess(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkPropertyAccess(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkTopLevelWindow(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/SecurityManager/checkTopLevelWindow(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkPrintJobAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkPrintJobAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkSystemClipboardAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkSystemClipboardAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAwtEventQueueAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkAwtEventQueueAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getPackages(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/lang/SecurityManager/getPackages(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "checkPackageAccess(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkPackageAccess(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkPackageDefinition(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkPackageDefinition(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkSetFactory()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkSetFactory()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkMemberAccess(Ljava/lang/Class;I)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkMemberAccess(Ljava/lang/Class;I)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkSecurityAccess(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/SecurityManager/checkSecurityAccess(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "currentLoadedClass0()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/SecurityManager/currentLoadedClass0()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getThreadGroup()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public "java/lang/SecurityManager/getThreadGroup()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/SecurityManager/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_StackTraceElement extends java_lang_Object implements java_io_Serializable {
    public "java/lang/StackTraceElement/declaringClass": java_lang_String;
    public "java/lang/StackTraceElement/methodName": java_lang_String;
    public "java/lang/StackTraceElement/fileName": java_lang_String;
    public "java/lang/StackTraceElement/lineNumber": number;
    public static "java/lang/StackTraceElement/serialVersionUID": Long;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StackTraceElement/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFileName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StackTraceElement/getFileName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getLineNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StackTraceElement/getLineNumber()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StackTraceElement/getClassName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getMethodName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StackTraceElement/getMethodName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isNativeMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StackTraceElement/isNativeMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/StackTraceElement/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StackTraceElement/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/StackTraceElement/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_lang_Thread extends java_lang_Object implements java_lang_Runnable {
    public $thread: JVMThread;
    public "java/lang/Thread/name": JVMArray<number>;
    public "java/lang/Thread/priority": number;
    public "java/lang/Thread/threadQ": java_lang_Thread;
    public "java/lang/Thread/eetop": Long;
    public "java/lang/Thread/single_step": number;
    public "java/lang/Thread/daemon": number;
    public "java/lang/Thread/stillborn": number;
    public "java/lang/Thread/target": java_lang_Runnable;
    public "java/lang/Thread/group": java_lang_ThreadGroup;
    public "java/lang/Thread/contextClassLoader": java_lang_ClassLoader;
    public "java/lang/Thread/inheritedAccessControlContext": java_security_AccessControlContext;
    public static "java/lang/Thread/threadInitNumber": number;
    public "java/lang/Thread/threadLocals": java_lang_ThreadLocal$ThreadLocalMap;
    public "java/lang/Thread/inheritableThreadLocals": java_lang_ThreadLocal$ThreadLocalMap;
    public "java/lang/Thread/stackSize": Long;
    public "java/lang/Thread/nativeParkEventPointer": Long;
    public "java/lang/Thread/tid": Long;
    public static "java/lang/Thread/threadSeqNumber": Long;
    public "java/lang/Thread/threadStatus": number;
    public "java/lang/Thread/parkBlocker": java_lang_Object;
    public "java/lang/Thread/blocker": sun_nio_ch_Interruptible;
    public "java/lang/Thread/blockerLock": java_lang_Object;
    public static "java/lang/Thread/MIN_PRIORITY": number;
    public static "java/lang/Thread/NORM_PRIORITY": number;
    public static "java/lang/Thread/MAX_PRIORITY": number;
    public static "java/lang/Thread/EMPTY_STACK_TRACE": JVMArray<java_lang_StackTraceElement>;
    public static "java/lang/Thread/SUBCLASS_IMPLEMENTATION_PERMISSION": java_lang_RuntimePermission;
    public "java/lang/Thread/uncaughtExceptionHandler": java_lang_Thread$UncaughtExceptionHandler;
    public static "java/lang/Thread/defaultUncaughtExceptionHandler": java_lang_Thread$UncaughtExceptionHandler;
    public "java/lang/Thread/threadLocalRandomSeed": Long;
    public "java/lang/Thread/threadLocalRandomProbe": number;
    public "java/lang/Thread/threadLocalRandomSecondarySeed": number;
    public static "registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Thread/registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "nextThreadNum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Thread/nextThreadNum()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "nextThreadID()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Thread/nextThreadID()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "blockedOn(Lsun/nio/ch/Interruptible;)V"(thread: JVMThread, args: [sun_nio_ch_Interruptible], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/blockedOn(Lsun/nio/ch/Interruptible;)V"(thread: JVMThread, args: [sun_nio_ch_Interruptible], cb?: (e?: java_lang_Throwable) => void): void;
    public static "currentThread()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "java/lang/Thread/currentThread()Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "yield()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Thread/yield()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sleep(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Thread/sleep(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sleep(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Thread/sleep(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "init(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable, java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/init(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable, java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "init(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;JLjava/security/AccessControlContext;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable, java_lang_String, Long, any, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/init(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;JLjava/security/AccessControlContext;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable, java_lang_String, Long, any, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/Thread/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/<init>(Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Runnable;Ljava/security/AccessControlContext;)V"(thread: JVMThread, args: [java_lang_Runnable, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/<init>(Ljava/lang/Runnable;Ljava/security/AccessControlContext;)V"(thread: JVMThread, args: [java_lang_Runnable, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/<init>(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ThreadGroup;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/<init>(Ljava/lang/ThreadGroup;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Runnable;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Runnable, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/<init>(Ljava/lang/Runnable;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Runnable, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/<init>(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable, java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/<init>(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable, java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "start()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/start()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "start0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/start0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "exit()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/exit()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "stop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/stop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "stop(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/stop(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "interrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/interrupt()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "interrupted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Thread/interrupted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInterrupted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Thread/isInterrupted()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInterrupted(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Thread/isInterrupted(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "destroy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/destroy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "isAlive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Thread/isAlive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "suspend()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/suspend()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/resume()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPriority(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/setPriority(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getPriority()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Thread/getPriority()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/setName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Thread/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getThreadGroup()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public "java/lang/Thread/getThreadGroup()Ljava/lang/ThreadGroup;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ThreadGroup) => void): void;
    public static "activeCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Thread/activeCount()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "enumerate([Ljava/lang/Thread;)I"(thread: JVMThread, args: [JVMArray<java_lang_Thread>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Thread/enumerate([Ljava/lang/Thread;)I"(thread: JVMThread, args: [JVMArray<java_lang_Thread>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "countStackFrames()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Thread/countStackFrames()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "join(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/join(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "join(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/join(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "join()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/join()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "dumpStack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Thread/dumpStack()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setDaemon(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/setDaemon(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isDaemon()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Thread/isDaemon()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/checkAccess()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Thread/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getContextClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/lang/Thread/getContextClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "setContextClassLoader(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/setContextClassLoader(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public static "holdsLock(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Thread/holdsLock(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getStackTrace()[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_StackTraceElement>) => void): void;
    public "java/lang/Thread/getStackTrace()[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_StackTraceElement>) => void): void;
    public static "getAllStackTraces()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/lang/Thread/getAllStackTraces()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "isCCLOverridden(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Thread/isCCLOverridden(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "auditSubclass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Thread/auditSubclass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "dumpThreads([Ljava/lang/Thread;)[[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: [JVMArray<java_lang_Thread>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_StackTraceElement>>) => void): void;
    public static "java/lang/Thread/dumpThreads([Ljava/lang/Thread;)[[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: [JVMArray<java_lang_Thread>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<java_lang_StackTraceElement>>) => void): void;
    public static "getThreads()[Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread>) => void): void;
    public static "java/lang/Thread/getThreads()[Ljava/lang/Thread;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Thread>) => void): void;
    public "getId()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/Thread/getId()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getState()Ljava/lang/Thread$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$State) => void): void;
    public "java/lang/Thread/getState()Ljava/lang/Thread$State;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$State) => void): void;
    public static "setDefaultUncaughtExceptionHandler(Ljava/lang/Thread$UncaughtExceptionHandler;)V"(thread: JVMThread, args: [java_lang_Thread$UncaughtExceptionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Thread/setDefaultUncaughtExceptionHandler(Ljava/lang/Thread$UncaughtExceptionHandler;)V"(thread: JVMThread, args: [java_lang_Thread$UncaughtExceptionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getDefaultUncaughtExceptionHandler()Ljava/lang/Thread$UncaughtExceptionHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$UncaughtExceptionHandler) => void): void;
    public static "java/lang/Thread/getDefaultUncaughtExceptionHandler()Ljava/lang/Thread$UncaughtExceptionHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$UncaughtExceptionHandler) => void): void;
    public "getUncaughtExceptionHandler()Ljava/lang/Thread$UncaughtExceptionHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$UncaughtExceptionHandler) => void): void;
    public "java/lang/Thread/getUncaughtExceptionHandler()Ljava/lang/Thread$UncaughtExceptionHandler;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread$UncaughtExceptionHandler) => void): void;
    public "setUncaughtExceptionHandler(Ljava/lang/Thread$UncaughtExceptionHandler;)V"(thread: JVMThread, args: [java_lang_Thread$UncaughtExceptionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/setUncaughtExceptionHandler(Ljava/lang/Thread$UncaughtExceptionHandler;)V"(thread: JVMThread, args: [java_lang_Thread$UncaughtExceptionHandler], cb?: (e?: java_lang_Throwable) => void): void;
    public "dispatchUncaughtException(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/dispatchUncaughtException(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "processQueue(Ljava/lang/ref/ReferenceQueue;Ljava/util/concurrent/ConcurrentMap;)V"(thread: JVMThread, args: [java_lang_ref_ReferenceQueue, java_util_concurrent_ConcurrentMap], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Thread/processQueue(Ljava/lang/ref/ReferenceQueue;Ljava/util/concurrent/ConcurrentMap;)V"(thread: JVMThread, args: [java_lang_ref_ReferenceQueue, java_util_concurrent_ConcurrentMap], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPriority0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/setPriority0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "stop0(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/stop0(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "suspend0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/suspend0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "resume0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/resume0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "interrupt0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/interrupt0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setNativeName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Thread/setNativeName(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Thread/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Float extends java_lang_Number implements java_lang_Comparable {
    public unbox(): number;
    public static box(val: number): java_lang_Float;
    public static "java/lang/Float/POSITIVE_INFINITY": number;
    public static "java/lang/Float/NEGATIVE_INFINITY": number;
    public static "java/lang/Float/NaN": number;
    public static "java/lang/Float/MAX_VALUE": number;
    public static "java/lang/Float/MIN_NORMAL": number;
    public static "java/lang/Float/MIN_VALUE": number;
    public static "java/lang/Float/MAX_EXPONENT": number;
    public static "java/lang/Float/MIN_EXPONENT": number;
    public static "java/lang/Float/SIZE": number;
    public static "java/lang/Float/BYTES": number;
    public static "java/lang/Float/TYPE": java_lang_Class;
    public "java/lang/Float/value": number;
    public static "java/lang/Float/serialVersionUID": Long;
    public static "toString(F)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Float/toString(F)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toHexString(F)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Float/toHexString(F)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/Float;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Float) => void): void;
    public static "java/lang/Float/valueOf(Ljava/lang/String;)Ljava/lang/Float;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Float) => void): void;
    public static "valueOf(F)Ljava/lang/Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Float) => void): void;
    public static "java/lang/Float/valueOf(F)Ljava/lang/Float;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Float) => void): void;
    public static "parseFloat(Ljava/lang/String;)F"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/parseFloat(Ljava/lang/String;)F"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isNaN(F)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/isNaN(F)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isInfinite(F)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/isInfinite(F)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isFinite(F)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/isFinite(F)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Float/<init>(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Float/<init>(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Float/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isNaN()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/isNaN()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInfinite()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/isInfinite()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Float/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/Float/longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hashCode(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/hashCode(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "floatToIntBits(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/floatToIntBits(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "floatToRawIntBits(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/floatToRawIntBits(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "intBitsToFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/intBitsToFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Float;)I"(thread: JVMThread, args: [java_lang_Float], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/compareTo(Ljava/lang/Float;)I"(thread: JVMThread, args: [java_lang_Float], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(FF)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/compare(FF)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sum(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/sum(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "max(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/max(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "min(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Float/min(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Float/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Float/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Double extends java_lang_Number implements java_lang_Comparable {
    public unbox(): number;
    public static box(val: number): java_lang_Double;
    public static "java/lang/Double/POSITIVE_INFINITY": number;
    public static "java/lang/Double/NEGATIVE_INFINITY": number;
    public static "java/lang/Double/NaN": number;
    public static "java/lang/Double/MAX_VALUE": number;
    public static "java/lang/Double/MIN_NORMAL": number;
    public static "java/lang/Double/MIN_VALUE": number;
    public static "java/lang/Double/MAX_EXPONENT": number;
    public static "java/lang/Double/MIN_EXPONENT": number;
    public static "java/lang/Double/SIZE": number;
    public static "java/lang/Double/BYTES": number;
    public static "java/lang/Double/TYPE": java_lang_Class;
    public "java/lang/Double/value": number;
    public static "java/lang/Double/serialVersionUID": Long;
    public static "toString(D)Ljava/lang/String;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Double/toString(D)Ljava/lang/String;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "toHexString(D)Ljava/lang/String;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Double/toHexString(D)Ljava/lang/String;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/Double;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Double) => void): void;
    public static "java/lang/Double/valueOf(Ljava/lang/String;)Ljava/lang/Double;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Double) => void): void;
    public static "valueOf(D)Ljava/lang/Double;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Double) => void): void;
    public static "java/lang/Double/valueOf(D)Ljava/lang/Double;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_Double) => void): void;
    public static "parseDouble(Ljava/lang/String;)D"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/parseDouble(Ljava/lang/String;)D"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isNaN(D)Z"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/isNaN(D)Z"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isInfinite(D)Z"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/isInfinite(D)Z"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "isFinite(D)Z"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/isFinite(D)Z"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Double/<init>(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Double/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "isNaN()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/isNaN()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInfinite()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/isInfinite()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Double/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/byteValue()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/shortValue()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/intValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/Double/longValue()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/floatValue()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/doubleValue()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hashCode(D)I"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/hashCode(D)I"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "doubleToLongBits(D)J"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Double/doubleToLongBits(D)J"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "doubleToRawLongBits(D)J"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/Double/doubleToRawLongBits(D)J"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "longBitsToDouble(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/longBitsToDouble(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Double;)I"(thread: JVMThread, args: [java_lang_Double], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/compareTo(Ljava/lang/Double;)I"(thread: JVMThread, args: [java_lang_Double], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compare(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/compare(DD)I"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sum(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/sum(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "max(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/max(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "min(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Double/min(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Double/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Double/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Runtime extends java_lang_Object {
    public static "java/lang/Runtime/currentRuntime": java_lang_Runtime;
    public static "getRuntime()Ljava/lang/Runtime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runtime) => void): void;
    public static "java/lang/Runtime/getRuntime()Ljava/lang/Runtime;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Runtime) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "exit(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/exit(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "addShutdownHook(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/addShutdownHook(Ljava/lang/Thread;)V"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeShutdownHook(Ljava/lang/Thread;)Z"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Runtime/removeShutdownHook(Ljava/lang/Thread;)Z"(thread: JVMThread, args: [java_lang_Thread], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "halt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/halt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "runFinalizersOnExit(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Runtime/runFinalizersOnExit(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "exec(Ljava/lang/String;)Ljava/lang/Process;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "java/lang/Runtime/exec(Ljava/lang/String;)Ljava/lang/Process;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "exec(Ljava/lang/String;[Ljava/lang/String;)Ljava/lang/Process;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "java/lang/Runtime/exec(Ljava/lang/String;[Ljava/lang/String;)Ljava/lang/Process;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "exec(Ljava/lang/String;[Ljava/lang/String;Ljava/io/File;)Ljava/lang/Process;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "java/lang/Runtime/exec(Ljava/lang/String;[Ljava/lang/String;Ljava/io/File;)Ljava/lang/Process;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_String>, java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "exec([Ljava/lang/String;)Ljava/lang/Process;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "java/lang/Runtime/exec([Ljava/lang/String;)Ljava/lang/Process;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "exec([Ljava/lang/String;[Ljava/lang/String;)Ljava/lang/Process;"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "java/lang/Runtime/exec([Ljava/lang/String;[Ljava/lang/String;)Ljava/lang/Process;"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "exec([Ljava/lang/String;[Ljava/lang/String;Ljava/io/File;)Ljava/lang/Process;"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<java_lang_String>, java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "java/lang/Runtime/exec([Ljava/lang/String;[Ljava/lang/String;Ljava/io/File;)Ljava/lang/Process;"(thread: JVMThread, args: [JVMArray<java_lang_String>, JVMArray<java_lang_String>, java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "availableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Runtime/availableProcessors()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "freeMemory()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/Runtime/freeMemory()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "totalMemory()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/Runtime/totalMemory()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "maxMemory()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/Runtime/maxMemory()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "gc()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/gc()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "runFinalization0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Runtime/runFinalization0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "runFinalization()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/runFinalization()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "traceInstructions(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/traceInstructions(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "traceMethodCalls(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/traceMethodCalls(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "load(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/load(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "load0(Ljava/lang/Class;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/load0(Ljava/lang/Class;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "loadLibrary(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/loadLibrary(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "loadLibrary0(Ljava/lang/Class;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Runtime/loadLibrary0(Ljava/lang/Class;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getLocalizedInputStream(Ljava/io/InputStream;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/lang/Runtime/getLocalizedInputStream(Ljava/io/InputStream;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getLocalizedOutputStream(Ljava/io/OutputStream;)Ljava/io/OutputStream;"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "java/lang/Runtime/getLocalizedOutputStream(Ljava/io/OutputStream;)Ljava/io/OutputStream;"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Runtime/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Process extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Process/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "java/lang/Process/getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/lang/Process/getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getErrorStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/lang/Process/getErrorStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "waitFor()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Process/waitFor()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "waitFor(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Process/waitFor(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "exitValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Process/exitValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "destroy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Process/destroy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "destroyForcibly()Ljava/lang/Process;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "java/lang/Process/destroyForcibly()Ljava/lang/Process;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "isAlive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Process/isAlive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_lang_ClassLoader$NativeLibrary extends java_lang_Object {
    public "java/lang/ClassLoader$NativeLibrary/handle": Long;
    public "java/lang/ClassLoader$NativeLibrary/jniVersion": number;
    public "java/lang/ClassLoader$NativeLibrary/fromClass": java_lang_Class;
    public "java/lang/ClassLoader$NativeLibrary/name": java_lang_String;
    public "java/lang/ClassLoader$NativeLibrary/isBuiltin": number;
    public "java/lang/ClassLoader$NativeLibrary/loaded": number;
    public "load(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader$NativeLibrary/load(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "find(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/ClassLoader$NativeLibrary/find(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "unload(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader$NativeLibrary/unload(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Class;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader$NativeLibrary/<init>(Ljava/lang/Class;Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_Class, java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassLoader$NativeLibrary/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getFromClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/ClassLoader$NativeLibrary/getFromClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "access$200(Ljava/lang/ClassLoader$NativeLibrary;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader$NativeLibrary], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/ClassLoader$NativeLibrary/access$200(Ljava/lang/ClassLoader$NativeLibrary;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_ClassLoader$NativeLibrary], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
  }
  export class java_lang_Terminator extends java_lang_Object {
    public static "java/lang/Terminator/handler": sun_misc_SignalHandler;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Terminator/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Terminator/setup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "teardown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Terminator/teardown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Terminator/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Object {
    public ref: number;
    public $monitor: Monitor;
    public getClass(): ClassData;
    public getMonitor(): Monitor;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Object/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Object/registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Object/getClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Object/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Object/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/Object/clone()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Object/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "notify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Object/notify()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "notifyAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Object/notifyAll()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "wait(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Object/wait(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "wait(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Object/wait(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "wait()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Object/wait()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Object/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Object/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Compiler extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Compiler/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Compiler/initialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Compiler/registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "compileClass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Compiler/compileClass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "compileClasses(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Compiler/compileClasses(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "command(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/lang/Compiler/command(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "enable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Compiler/enable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "disable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Compiler/disable()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Compiler/access$000()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Compiler/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ProcessEnvironment extends java_lang_Object {
    public static "java/lang/ProcessEnvironment/theEnvironment": java_util_HashMap;
    public static "java/lang/ProcessEnvironment/theUnmodifiableEnvironment": java_util_Map;
    public static "java/lang/ProcessEnvironment/MIN_NAME_LENGTH": number;
    public static "getenv(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/ProcessEnvironment/getenv(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getenv()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/lang/ProcessEnvironment/getenv()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "environment()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/lang/ProcessEnvironment/environment()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "emptyEnvironment(I)Ljava/util/Map;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/lang/ProcessEnvironment/emptyEnvironment(I)Ljava/util/Map;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "environ()[[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<number>>) => void): void;
    public static "java/lang/ProcessEnvironment/environ()[[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<JVMArray<number>>) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ProcessEnvironment/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "validateVariable(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ProcessEnvironment/validateVariable(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "validateValue(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ProcessEnvironment/validateValue(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "toEnvironmentBlock(Ljava/util/Map;[I)[B"(thread: JVMThread, args: [java_util_Map, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/lang/ProcessEnvironment/toEnvironmentBlock(Ljava/util/Map;[I)[B"(thread: JVMThread, args: [java_util_Map, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "arrayCompare([B[B)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ProcessEnvironment/arrayCompare([B[B)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "arrayEquals([B[B)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ProcessEnvironment/arrayEquals([B[B)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "arrayHash([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ProcessEnvironment/arrayHash([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000([B[B)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ProcessEnvironment/access$000([B[B)Z"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ProcessEnvironment/access$100([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ProcessEnvironment/access$200(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$300([B[B)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ProcessEnvironment/access$300([B[B)I"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$400(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ProcessEnvironment/access$400(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ProcessEnvironment/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Shutdown extends java_lang_Object {
    public static "java/lang/Shutdown/RUNNING": number;
    public static "java/lang/Shutdown/HOOKS": number;
    public static "java/lang/Shutdown/FINALIZERS": number;
    public static "java/lang/Shutdown/state": number;
    public static "java/lang/Shutdown/runFinalizersOnExit": number;
    public static "java/lang/Shutdown/MAX_SYSTEM_HOOKS": number;
    public static "java/lang/Shutdown/hooks": JVMArray<java_lang_Runnable>;
    public static "java/lang/Shutdown/currentRunningHook": number;
    public static "java/lang/Shutdown/lock": java_lang_Object;
    public static "java/lang/Shutdown/haltLock": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Shutdown/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "setRunFinalizersOnExit(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/setRunFinalizersOnExit(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "add(IZLjava/lang/Runnable;)V"(thread: JVMThread, args: [number, number, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/add(IZLjava/lang/Runnable;)V"(thread: JVMThread, args: [number, number, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "runHooks()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/runHooks()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "halt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/halt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "halt0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/halt0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "runAllFinalizers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/runAllFinalizers()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sequence()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/sequence()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "exit(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/exit(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/shutdown()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Shutdown/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Package extends java_lang_Object implements java_lang_reflect_AnnotatedElement {
    public static "java/lang/Package/pkgs": java_util_Map;
    public static "java/lang/Package/urls": java_util_Map;
    public static "java/lang/Package/mans": java_util_Map;
    public "java/lang/Package/pkgName": java_lang_String;
    public "java/lang/Package/specTitle": java_lang_String;
    public "java/lang/Package/specVersion": java_lang_String;
    public "java/lang/Package/specVendor": java_lang_String;
    public "java/lang/Package/implTitle": java_lang_String;
    public "java/lang/Package/implVersion": java_lang_String;
    public "java/lang/Package/implVendor": java_lang_String;
    public "java/lang/Package/sealBase": java_net_URL;
    public "java/lang/Package/loader": java_lang_ClassLoader;
    public "java/lang/Package/packageInfo": java_lang_Class;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Package/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSpecificationTitle()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Package/getSpecificationTitle()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSpecificationVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Package/getSpecificationVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSpecificationVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Package/getSpecificationVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getImplementationTitle()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Package/getImplementationTitle()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getImplementationVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Package/getImplementationVersion()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getImplementationVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Package/getImplementationVendor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isSealed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Package/isSealed()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSealed(Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Package/isSealed(Ljava/net/URL;)Z"(thread: JVMThread, args: [java_net_URL], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isCompatibleWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Package/isCompatibleWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getPackage(Ljava/lang/String;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public static "java/lang/Package/getPackage(Ljava/lang/String;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public static "getPackages()[Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Package>) => void): void;
    public static "java/lang/Package/getPackages()[Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Package>) => void): void;
    public static "getPackage(Ljava/lang/Class;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public static "java/lang/Package/getPackage(Ljava/lang/Class;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Package/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Package/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getPackageInfo()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Package/getPackageInfo()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/Package/getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Package/isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/Package/getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/Package/getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/Package/getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/Package/getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/Package/getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/net/URL;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_net_URL, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Package/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;Ljava/net/URL;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_lang_String, java_net_URL, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/util/jar/Manifest;Ljava/net/URL;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_util_jar_Manifest, java_net_URL, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Package/<init>(Ljava/lang/String;Ljava/util/jar/Manifest;Ljava/net/URL;Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_String, java_util_jar_Manifest, java_net_URL, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getSystemPackage(Ljava/lang/String;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public static "java/lang/Package/getSystemPackage(Ljava/lang/String;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public static "getSystemPackages()[Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Package>) => void): void;
    public static "java/lang/Package/getSystemPackages()[Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Package>) => void): void;
    public static "defineSystemPackage(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public static "java/lang/Package/defineSystemPackage(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/Package;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public static "loadManifest(Ljava/lang/String;)Ljava/util/jar/Manifest;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public static "java/lang/Package/loadManifest(Ljava/lang/String;)Ljava/util/jar/Manifest;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public static "getSystemPackage0(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Package/getSystemPackage0(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getSystemPackages0()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "java/lang/Package/getSystemPackages0()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "access$000()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/lang/Package/access$000()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "access$100(Ljava/lang/String;)Ljava/util/jar/Manifest;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public static "java/lang/Package/access$100(Ljava/lang/String;)Ljava/util/jar/Manifest;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_util_jar_Manifest) => void): void;
    public static "access$200()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/lang/Package/access$200()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "<init>(Ljava/lang/String;Ljava/util/jar/Manifest;Ljava/net/URL;Ljava/lang/ClassLoader;Ljava/lang/Package$1;)V"(thread: JVMThread, args: [java_lang_String, java_util_jar_Manifest, java_net_URL, java_lang_ClassLoader, java_lang_Package$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Package/<init>(Ljava/lang/String;Ljava/util/jar/Manifest;Ljava/net/URL;Ljava/lang/ClassLoader;Ljava/lang/Package$1;)V"(thread: JVMThread, args: [java_lang_String, java_util_jar_Manifest, java_net_URL, java_lang_ClassLoader, java_lang_Package$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$400()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "java/lang/Package/access$400()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Package/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
  }
  export class java_lang_Package$1 extends java_lang_Object implements java_security_PrivilegedAction {
    public "java/lang/Package$1/val$iname": java_lang_String;
    public "java/lang/Package$1/val$fn": java_lang_String;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Package$1/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public "java/lang/Package$1/run()Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public "run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/Package$1/run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_lang_UNIXProcess extends java_lang_Process {
    public static "java/lang/UNIXProcess/fdAccess": sun_misc_JavaIOFileDescriptorAccess;
    public "java/lang/UNIXProcess/pid": number;
    public "java/lang/UNIXProcess/exitcode": number;
    public "java/lang/UNIXProcess/hasExited": number;
    public "java/lang/UNIXProcess/stdin": java_io_OutputStream;
    public "java/lang/UNIXProcess/stdout": java_io_InputStream;
    public "java/lang/UNIXProcess/stderr": java_io_InputStream;
    public "java/lang/UNIXProcess/stdout_inner_stream": java_lang_UNIXProcess$DeferredCloseInputStream;
    public static "java/lang/UNIXProcess/platform": java_lang_UNIXProcess$Platform;
    public static "java/lang/UNIXProcess/launchMechanism": java_lang_UNIXProcess$LaunchMechanism;
    public static "java/lang/UNIXProcess/helperpath": JVMArray<number>;
    public static "java/lang/UNIXProcess/processReaperExecutor": java_util_concurrent_Executor;
    public static "toCString(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/lang/UNIXProcess/toCString(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "waitForProcessExit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess/waitForProcessExit(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "forkAndExec(I[B[B[BI[BI[B[IZ)I"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>, JVMArray<number>, number, JVMArray<number>, number, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess/forkAndExec(I[B[B[BI[BI[B[IZ)I"(thread: JVMThread, args: [number, JVMArray<number>, JVMArray<number>, JVMArray<number>, number, JVMArray<number>, number, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>([B[BI[BI[B[IZ)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, JVMArray<number>, number, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess/<init>([B[BI[BI[B[IZ)V"(thread: JVMThread, args: [JVMArray<number>, JVMArray<number>, number, JVMArray<number>, number, JVMArray<number>, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "newFileDescriptor(I)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public static "java/lang/UNIXProcess/newFileDescriptor(I)Ljava/io/FileDescriptor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "initStreams([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess/initStreams([I)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "java/lang/UNIXProcess/getOutputStream()Ljava/io/OutputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_OutputStream) => void): void;
    public "getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/lang/UNIXProcess/getInputStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getErrorStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/lang/UNIXProcess/getErrorStream()Ljava/io/InputStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "waitFor()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess/waitFor()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "waitFor(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess/waitFor(JLjava/util/concurrent/TimeUnit;)Z"(thread: JVMThread, args: [Long, any, java_util_concurrent_TimeUnit], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "exitValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess/exitValue()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "destroyProcess(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/UNIXProcess/destroyProcess(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "destroy(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess/destroy(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "destroy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess/destroy()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "destroyForcibly()Ljava/lang/Process;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "java/lang/UNIXProcess/destroyForcibly()Ljava/lang/Process;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Process) => void): void;
    public "isAlive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess/isAlive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/UNIXProcess/init()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lambda$initStreams$5()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess/lambda$initStreams$5()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lambda$initStreams$4()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess/lambda$initStreams$4()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lambda$initStreams$3()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess/lambda$initStreams$3()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lambda$new$2([I)Ljava/lang/Void;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "java/lang/UNIXProcess/lambda$new$2([I)Ljava/lang/Void;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public static "lambda$static$1()Ljava/util/concurrent/Executor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Executor) => void): void;
    public static "java/lang/UNIXProcess/lambda$static$1()Ljava/util/concurrent/Executor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_concurrent_Executor) => void): void;
    public static "lambda$null$0(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "java/lang/UNIXProcess/lambda$null$0(Ljava/lang/ThreadGroup;Ljava/lang/Runnable;)Ljava/lang/Thread;"(thread: JVMThread, args: [java_lang_ThreadGroup, java_lang_Runnable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Thread) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/UNIXProcess/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_UNIXProcess$LaunchMechanism extends java_lang_Enum {
    public static "java/lang/UNIXProcess$LaunchMechanism/FORK": java_lang_UNIXProcess$LaunchMechanism;
    public static "java/lang/UNIXProcess$LaunchMechanism/POSIX_SPAWN": java_lang_UNIXProcess$LaunchMechanism;
    public static "java/lang/UNIXProcess$LaunchMechanism/VFORK": java_lang_UNIXProcess$LaunchMechanism;
    public static "java/lang/UNIXProcess$LaunchMechanism/$VALUES": JVMArray<java_lang_UNIXProcess$LaunchMechanism>;
    public static "values()[Ljava/lang/UNIXProcess$LaunchMechanism;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_UNIXProcess$LaunchMechanism>) => void): void;
    public static "java/lang/UNIXProcess$LaunchMechanism/values()[Ljava/lang/UNIXProcess$LaunchMechanism;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_UNIXProcess$LaunchMechanism>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/UNIXProcess$LaunchMechanism;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$LaunchMechanism) => void): void;
    public static "java/lang/UNIXProcess$LaunchMechanism/valueOf(Ljava/lang/String;)Ljava/lang/UNIXProcess$LaunchMechanism;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$LaunchMechanism) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess$LaunchMechanism/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/UNIXProcess$LaunchMechanism/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_UNIXProcess$Platform extends java_lang_Enum {
    public static "java/lang/UNIXProcess$Platform/LINUX": java_lang_UNIXProcess$Platform;
    public static "java/lang/UNIXProcess$Platform/BSD": java_lang_UNIXProcess$Platform;
    public static "java/lang/UNIXProcess$Platform/SOLARIS": java_lang_UNIXProcess$Platform;
    public static "java/lang/UNIXProcess$Platform/AIX": java_lang_UNIXProcess$Platform;
    public "java/lang/UNIXProcess$Platform/defaultLaunchMechanism": java_lang_UNIXProcess$LaunchMechanism;
    public "java/lang/UNIXProcess$Platform/validLaunchMechanisms": java_util_Set;
    public static "java/lang/UNIXProcess$Platform/$VALUES": JVMArray<java_lang_UNIXProcess$Platform>;
    public static "values()[Ljava/lang/UNIXProcess$Platform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_UNIXProcess$Platform>) => void): void;
    public static "java/lang/UNIXProcess$Platform/values()[Ljava/lang/UNIXProcess$Platform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_UNIXProcess$Platform>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/UNIXProcess$Platform;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$Platform) => void): void;
    public static "java/lang/UNIXProcess$Platform/valueOf(Ljava/lang/String;)Ljava/lang/UNIXProcess$Platform;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$Platform) => void): void;
    public "<init>(Ljava/lang/String;I[Ljava/lang/UNIXProcess$LaunchMechanism;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_UNIXProcess$LaunchMechanism>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess$Platform/<init>(Ljava/lang/String;I[Ljava/lang/UNIXProcess$LaunchMechanism;)V"(thread: JVMThread, args: [java_lang_String, number, JVMArray<java_lang_UNIXProcess$LaunchMechanism>], cb?: (e?: java_lang_Throwable) => void): void;
    public "helperPath(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/UNIXProcess$Platform/helperPath(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "helperPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/UNIXProcess$Platform/helperPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "launchMechanism()Ljava/lang/UNIXProcess$LaunchMechanism;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$LaunchMechanism) => void): void;
    public "java/lang/UNIXProcess$Platform/launchMechanism()Ljava/lang/UNIXProcess$LaunchMechanism;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$LaunchMechanism) => void): void;
    public static "get()Ljava/lang/UNIXProcess$Platform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$Platform) => void): void;
    public static "java/lang/UNIXProcess$Platform/get()Ljava/lang/UNIXProcess$Platform;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$Platform) => void): void;
    public static "lambda$get$2()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/UNIXProcess$Platform/lambda$get$2()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "lambda$launchMechanism$1()Ljava/lang/UNIXProcess$LaunchMechanism;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$LaunchMechanism) => void): void;
    public "java/lang/UNIXProcess$Platform/lambda$launchMechanism$1()Ljava/lang/UNIXProcess$LaunchMechanism;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_UNIXProcess$LaunchMechanism) => void): void;
    public "lambda$helperPath$0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/UNIXProcess$Platform/lambda$helperPath$0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/UNIXProcess$Platform/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_UNIXProcess$DeferredCloseInputStream extends java_io_FileInputStream {
    public "java/lang/UNIXProcess$DeferredCloseInputStream/lock": java_lang_Object;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/closePending": number;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/useCount": number;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/streamToClose": java_io_InputStream;
    public "<init>(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/<init>(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "raise()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/raise()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lower()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/lower()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "closeDeferred(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/closeDeferred(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/UNIXProcess$DeferredCloseInputStream/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/lang/UNIXProcess$DeferredCloseInputStream;Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_lang_UNIXProcess$DeferredCloseInputStream, java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/UNIXProcess$DeferredCloseInputStream/access$000(Ljava/lang/UNIXProcess$DeferredCloseInputStream;Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_lang_UNIXProcess$DeferredCloseInputStream, java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_StrictMath extends java_lang_Object {
    public static "java/lang/StrictMath/E": number;
    public static "java/lang/StrictMath/PI": number;
    public static "java/lang/StrictMath/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/StrictMath/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sin(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/sin(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "cos(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/cos(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "tan(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/tan(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "asin(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/asin(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "acos(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/acos(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "atan(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/atan(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toRadians(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/toRadians(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "toDegrees(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/toDegrees(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "exp(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/exp(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "log(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/log(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "log10(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/log10(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sqrt(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/sqrt(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "cbrt(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/cbrt(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "IEEEremainder(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/IEEEremainder(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ceil(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/ceil(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "floor(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/floor(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "floorOrCeil(DDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/floorOrCeil(DDDD)D"(thread: JVMThread, args: [number, any, number, any, number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "rint(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/rint(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "atan2(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/atan2(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "pow(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/pow(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "round(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/round(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "round(D)J"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/StrictMath/round(D)J"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "random()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/random()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "addExact(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/addExact(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "addExact(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/StrictMath/addExact(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "subtractExact(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/subtractExact(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "subtractExact(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/StrictMath/subtractExact(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "multiplyExact(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/multiplyExact(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "multiplyExact(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/StrictMath/multiplyExact(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "toIntExact(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/toIntExact(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "floorDiv(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/floorDiv(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "floorDiv(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/StrictMath/floorDiv(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "floorMod(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/floorMod(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "floorMod(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/StrictMath/floorMod(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "abs(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/abs(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "abs(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/StrictMath/abs(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "abs(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/abs(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "abs(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/abs(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "max(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/max(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "max(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/StrictMath/max(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "max(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/max(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "max(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/max(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "min(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/min(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "min(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/lang/StrictMath/min(JJ)J"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "min(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/min(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "min(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/min(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ulp(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/ulp(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "ulp(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/ulp(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "signum(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/signum(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "signum(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/signum(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "sinh(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/sinh(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "cosh(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/cosh(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "tanh(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/tanh(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "hypot(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/hypot(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "expm1(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/expm1(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "log1p(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/log1p(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "copySign(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/copySign(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "copySign(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/copySign(FF)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getExponent(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/getExponent(F)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getExponent(D)I"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/getExponent(D)I"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "nextAfter(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/nextAfter(DD)D"(thread: JVMThread, args: [number, any, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "nextAfter(FD)F"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/nextAfter(FD)F"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "nextUp(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/nextUp(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "nextUp(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/nextUp(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "nextDown(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/nextDown(D)D"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "nextDown(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/nextDown(F)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "scalb(DI)D"(thread: JVMThread, args: [number, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/scalb(DI)D"(thread: JVMThread, args: [number, any, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "scalb(FI)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/StrictMath/scalb(FI)F"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/StrictMath/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_FileInputStream extends java_io_InputStream {
    public "java/io/FileInputStream/fd": java_io_FileDescriptor;
    public "java/io/FileInputStream/path": java_lang_String;
    public "java/io/FileInputStream/channel": java_nio_channels_FileChannel;
    public "java/io/FileInputStream/closeLock": java_lang_Object;
    public "java/io/FileInputStream/closed": number;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileInputStream/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileInputStream/<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileInputStream/<init>(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "open0(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileInputStream/open0(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "open(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileInputStream/open(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileInputStream/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileInputStream/read0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readBytes([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileInputStream/readBytes([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileInputStream/read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileInputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/FileInputStream/skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileInputStream/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileInputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "java/io/FileInputStream/getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "java/io/FileInputStream/getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/FileInputStream/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileInputStream/close0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileInputStream/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/io/FileInputStream;)V"(thread: JVMThread, args: [java_io_FileInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/FileInputStream/access$000(Ljava/io/FileInputStream;)V"(thread: JVMThread, args: [java_io_FileInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/FileInputStream/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_FileDescriptor extends java_lang_Object {
    public "java/io/FileDescriptor/fd": number;
    public "java/io/FileDescriptor/parent": java_io_Closeable;
    public "java/io/FileDescriptor/otherParents": java_util_List;
    public "java/io/FileDescriptor/closed": number;
    public static "java/io/FileDescriptor/in": java_io_FileDescriptor;
    public static "java/io/FileDescriptor/out": java_io_FileDescriptor;
    public static "java/io/FileDescriptor/err": java_io_FileDescriptor;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileDescriptor/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileDescriptor/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "valid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileDescriptor/valid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileDescriptor/sync()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/FileDescriptor/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "attach(Ljava/io/Closeable;)V"(thread: JVMThread, args: [java_io_Closeable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileDescriptor/attach(Ljava/io/Closeable;)V"(thread: JVMThread, args: [java_io_Closeable], cb?: (e?: java_lang_Throwable) => void): void;
    public "closeAll(Ljava/io/Closeable;)V"(thread: JVMThread, args: [java_io_Closeable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileDescriptor/closeAll(Ljava/io/Closeable;)V"(thread: JVMThread, args: [java_io_Closeable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$002(Ljava/io/FileDescriptor;I)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/FileDescriptor/access$002(Ljava/io/FileDescriptor;I)I"(thread: JVMThread, args: [java_io_FileDescriptor, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$000(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/FileDescriptor/access$000(Ljava/io/FileDescriptor;)I"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/FileDescriptor/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_FileOutputStream extends java_io_OutputStream {
    public "java/io/FileOutputStream/fd": java_io_FileDescriptor;
    public "java/io/FileOutputStream/append": number;
    public "java/io/FileOutputStream/channel": java_nio_channels_FileChannel;
    public "java/io/FileOutputStream/path": java_lang_String;
    public "java/io/FileOutputStream/closeLock": java_lang_Object;
    public "java/io/FileOutputStream/closed": number;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/<init>(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/<init>(Ljava/io/File;)V"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;Z)V"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/<init>(Ljava/io/File;Z)V"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/<init>(Ljava/io/FileDescriptor;)V"(thread: JVMThread, args: [java_io_FileDescriptor], cb?: (e?: java_lang_Throwable) => void): void;
    public "open0(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/open0(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "open(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/open(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/write(IZ)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBytes([BIIZ)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/writeBytes([BIIZ)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "java/io/FileOutputStream/getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "java/io/FileOutputStream/getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/finalize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileOutputStream/close0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/FileOutputStream/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/io/FileOutputStream;)V"(thread: JVMThread, args: [java_io_FileOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/FileOutputStream/access$000(Ljava/io/FileOutputStream;)V"(thread: JVMThread, args: [java_io_FileOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/FileOutputStream/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_File extends java_lang_Object implements java_io_Serializable, java_lang_Comparable {
    public static "java/io/File/fs": java_io_FileSystem;
    public "java/io/File/path": java_lang_String;
    public "java/io/File/status": java_io_File$PathStatus;
    public "java/io/File/prefixLength": number;
    public static "java/io/File/separatorChar": number;
    public static "java/io/File/separator": java_lang_String;
    public static "java/io/File/pathSeparatorChar": number;
    public static "java/io/File/pathSeparator": java_lang_String;
    public static "java/io/File/PATH_OFFSET": Long;
    public static "java/io/File/PREFIX_LENGTH_OFFSET": Long;
    public static "java/io/File/UNSAFE": sun_misc_Unsafe;
    public static "java/io/File/serialVersionUID": Long;
    public "java/io/File/filePath": java_nio_file_Path;
    public static "java/io/File/$assertionsDisabled": number;
    public "isInvalid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/isInvalid()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPrefixLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/getPrefixLength()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/io/File;)V"(thread: JVMThread, args: [java_lang_String, java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File/<init>(Ljava/lang/String;Ljava/io/File;)V"(thread: JVMThread, args: [java_lang_String, java_io_File], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File/<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File/<init>(Ljava/net/URI;)V"(thread: JVMThread, args: [java_net_URI], cb?: (e?: java_lang_Throwable) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/File/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getParent()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/File/getParent()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getParentFile()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "java/io/File/getParentFile()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "getPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/File/getPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isAbsolute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/isAbsolute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAbsolutePath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/File/getAbsolutePath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getAbsoluteFile()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "java/io/File/getAbsoluteFile()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "getCanonicalPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/File/getCanonicalPath()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCanonicalFile()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "java/io/File/getCanonicalFile()Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "slashify(Ljava/lang/String;Z)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/File/slashify(Ljava/lang/String;Z)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "java/io/File/toURL()Ljava/net/URL;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "toURI()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "java/io/File/toURI()Ljava/net/URI;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_net_URI) => void): void;
    public "canRead()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/canRead()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canWrite()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/canWrite()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "exists()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/exists()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/isDirectory()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/isFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isHidden()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/isHidden()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastModified()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/File/lastModified()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "length()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/File/length()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "createNewFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/createNewFile()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "delete()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/delete()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "deleteOnExit()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File/deleteOnExit()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "list()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/io/File/list()[Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "list(Ljava/io/FilenameFilter;)[Ljava/lang/String;"(thread: JVMThread, args: [java_io_FilenameFilter], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/io/File/list(Ljava/io/FilenameFilter;)[Ljava/lang/String;"(thread: JVMThread, args: [java_io_FilenameFilter], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "listFiles()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "java/io/File/listFiles()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "listFiles(Ljava/io/FilenameFilter;)[Ljava/io/File;"(thread: JVMThread, args: [java_io_FilenameFilter], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "java/io/File/listFiles(Ljava/io/FilenameFilter;)[Ljava/io/File;"(thread: JVMThread, args: [java_io_FilenameFilter], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "listFiles(Ljava/io/FileFilter;)[Ljava/io/File;"(thread: JVMThread, args: [java_io_FileFilter], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "java/io/File/listFiles(Ljava/io/FileFilter;)[Ljava/io/File;"(thread: JVMThread, args: [java_io_FileFilter], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "mkdir()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/mkdir()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "mkdirs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/mkdirs()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "renameTo(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/renameTo(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLastModified(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/setLastModified(J)Z"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/setReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setWritable(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/setWritable(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setWritable(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/setWritable(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReadable(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/setReadable(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReadable(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/setReadable(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setExecutable(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/setExecutable(ZZ)Z"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setExecutable(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/setExecutable(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "canExecute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/canExecute()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "listRoots()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public static "java/io/File/listRoots()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "getTotalSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/File/getTotalSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getFreeSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/File/getFreeSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getUsableSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/File/getUsableSpace()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "createTempFile(Ljava/lang/String;Ljava/lang/String;Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/io/File/createTempFile(Ljava/lang/String;Ljava/lang/String;Ljava/io/File;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "createTempFile(Ljava/lang/String;Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public static "java/io/File/createTempFile(Ljava/lang/String;Ljava/lang/String;)Ljava/io/File;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File) => void): void;
    public "compareTo(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/compareTo(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/File/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "toPath()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "java/io/File/toPath()Ljava/nio/file/Path;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_Path) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/File/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/File/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_io_FileFilter extends java_lang_Object {
    "accept(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_io_File$PathStatus extends java_lang_Enum {
    public static "java/io/File$PathStatus/INVALID": java_io_File$PathStatus;
    public static "java/io/File$PathStatus/CHECKED": java_io_File$PathStatus;
    public static "java/io/File$PathStatus/$VALUES": JVMArray<java_io_File$PathStatus>;
    public static "values()[Ljava/io/File$PathStatus;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File$PathStatus>) => void): void;
    public static "java/io/File$PathStatus/values()[Ljava/io/File$PathStatus;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File$PathStatus>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/io/File$PathStatus;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File$PathStatus) => void): void;
    public static "java/io/File$PathStatus/valueOf(Ljava/lang/String;)Ljava/io/File$PathStatus;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_File$PathStatus) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/File$PathStatus/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/File$PathStatus/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_FileSystem extends java_lang_Object {
    public static "java/io/FileSystem/BA_EXISTS": number;
    public static "java/io/FileSystem/BA_REGULAR": number;
    public static "java/io/FileSystem/BA_DIRECTORY": number;
    public static "java/io/FileSystem/BA_HIDDEN": number;
    public static "java/io/FileSystem/ACCESS_READ": number;
    public static "java/io/FileSystem/ACCESS_WRITE": number;
    public static "java/io/FileSystem/ACCESS_EXECUTE": number;
    public static "java/io/FileSystem/SPACE_TOTAL": number;
    public static "java/io/FileSystem/SPACE_FREE": number;
    public static "java/io/FileSystem/SPACE_USABLE": number;
    public static "java/io/FileSystem/useCanonCaches": number;
    public static "java/io/FileSystem/useCanonPrefixCache": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/FileSystem/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/getSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPathSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/getPathSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "normalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/FileSystem/normalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "prefixLength(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/prefixLength(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "resolve(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/FileSystem/resolve(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDefaultParent()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/FileSystem/getDefaultParent()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "fromURIPath(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/FileSystem/fromURIPath(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isAbsolute(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/isAbsolute(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "resolve(Ljava/io/File;)Ljava/lang/String;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/FileSystem/resolve(Ljava/io/File;)Ljava/lang/String;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "canonicalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/FileSystem/canonicalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getBooleanAttributes(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/getBooleanAttributes(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkAccess(Ljava/io/File;I)Z"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/checkAccess(Ljava/io/File;I)Z"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setPermission(Ljava/io/File;IZZ)Z"(thread: JVMThread, args: [java_io_File, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/setPermission(Ljava/io/File;IZZ)Z"(thread: JVMThread, args: [java_io_File, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLastModifiedTime(Ljava/io/File;)J"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/FileSystem/getLastModifiedTime(Ljava/io/File;)J"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLength(Ljava/io/File;)J"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/FileSystem/getLength(Ljava/io/File;)J"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "createFileExclusively(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/createFileExclusively(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "delete(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/delete(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "list(Ljava/io/File;)[Ljava/lang/String;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/io/FileSystem/list(Ljava/io/File;)[Ljava/lang/String;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "createDirectory(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/createDirectory(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rename(Ljava/io/File;Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/rename(Ljava/io/File;Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLastModifiedTime(Ljava/io/File;J)Z"(thread: JVMThread, args: [java_io_File, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/setLastModifiedTime(Ljava/io/File;J)Z"(thread: JVMThread, args: [java_io_File, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReadOnly(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/setReadOnly(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "listRoots()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "java/io/FileSystem/listRoots()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "getSpace(Ljava/io/File;I)J"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/FileSystem/getSpace(Ljava/io/File;I)J"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compare(Ljava/io/File;Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/compare(Ljava/io/File;Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/FileSystem/hashCode(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getBooleanProperty(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/FileSystem/getBooleanProperty(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/FileSystem/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_UnixFileSystem extends java_io_FileSystem {
    public "java/io/UnixFileSystem/slash": number;
    public "java/io/UnixFileSystem/colon": number;
    public "java/io/UnixFileSystem/javaHome": java_lang_String;
    public "java/io/UnixFileSystem/cache": java_io_ExpiringCache;
    public "java/io/UnixFileSystem/javaHomePrefixCache": java_io_ExpiringCache;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/UnixFileSystem/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/getSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPathSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/getPathSeparator()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "normalize(Ljava/lang/String;II)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/UnixFileSystem/normalize(Ljava/lang/String;II)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "normalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/UnixFileSystem/normalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "prefixLength(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/prefixLength(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "resolve(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/UnixFileSystem/resolve(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDefaultParent()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/UnixFileSystem/getDefaultParent()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "fromURIPath(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/UnixFileSystem/fromURIPath(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isAbsolute(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/isAbsolute(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "resolve(Ljava/io/File;)Ljava/lang/String;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/UnixFileSystem/resolve(Ljava/io/File;)Ljava/lang/String;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "canonicalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/UnixFileSystem/canonicalize(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "canonicalize0(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/UnixFileSystem/canonicalize0(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "parentOrNull(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/UnixFileSystem/parentOrNull(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getBooleanAttributes0(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/getBooleanAttributes0(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBooleanAttributes(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/getBooleanAttributes(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "checkAccess(Ljava/io/File;I)Z"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/checkAccess(Ljava/io/File;I)Z"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getLastModifiedTime(Ljava/io/File;)J"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/UnixFileSystem/getLastModifiedTime(Ljava/io/File;)J"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLength(Ljava/io/File;)J"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/UnixFileSystem/getLength(Ljava/io/File;)J"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setPermission(Ljava/io/File;IZZ)Z"(thread: JVMThread, args: [java_io_File, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/setPermission(Ljava/io/File;IZZ)Z"(thread: JVMThread, args: [java_io_File, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "createFileExclusively(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/createFileExclusively(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "delete(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/delete(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "delete0(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/delete0(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "list(Ljava/io/File;)[Ljava/lang/String;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/io/UnixFileSystem/list(Ljava/io/File;)[Ljava/lang/String;"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "createDirectory(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/createDirectory(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rename(Ljava/io/File;Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/rename(Ljava/io/File;Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "rename0(Ljava/io/File;Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/rename0(Ljava/io/File;Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setLastModifiedTime(Ljava/io/File;J)Z"(thread: JVMThread, args: [java_io_File, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/setLastModifiedTime(Ljava/io/File;J)Z"(thread: JVMThread, args: [java_io_File, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setReadOnly(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/setReadOnly(Ljava/io/File;)Z"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "listRoots()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "java/io/UnixFileSystem/listRoots()[Ljava/io/File;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_File>) => void): void;
    public "getSpace(Ljava/io/File;I)J"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/UnixFileSystem/getSpace(Ljava/io/File;I)J"(thread: JVMThread, args: [java_io_File, number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "compare(Ljava/io/File;Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/compare(Ljava/io/File;Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File, java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/UnixFileSystem/hashCode(Ljava/io/File;)I"(thread: JVMThread, args: [java_io_File], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/UnixFileSystem/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/UnixFileSystem/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ExpiringCache extends java_lang_Object {
    public "java/io/ExpiringCache/millisUntilExpiration": Long;
    public "java/io/ExpiringCache/map": java_util_Map;
    public "java/io/ExpiringCache/queryCount": number;
    public "java/io/ExpiringCache/queryOverflow": number;
    public "java/io/ExpiringCache/MAX_ENTRIES": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ExpiringCache/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ExpiringCache/<init>(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ExpiringCache/get(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "put(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ExpiringCache/put(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ExpiringCache/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "entryFor(Ljava/lang/String;)Ljava/io/ExpiringCache$Entry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_ExpiringCache$Entry) => void): void;
    public "java/io/ExpiringCache/entryFor(Ljava/lang/String;)Ljava/io/ExpiringCache$Entry;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_ExpiringCache$Entry) => void): void;
    public "cleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ExpiringCache/cleanup()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/io/ExpiringCache;)I"(thread: JVMThread, args: [java_io_ExpiringCache], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ExpiringCache/access$000(Ljava/io/ExpiringCache;)I"(thread: JVMThread, args: [java_io_ExpiringCache], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_io_ExpiringCache$Entry extends java_lang_Object {
    public "java/io/ExpiringCache$Entry/timestamp": Long;
    public "java/io/ExpiringCache$Entry/val": java_lang_String;
    public "<init>(JLjava/lang/String;)V"(thread: JVMThread, args: [Long, any, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ExpiringCache$Entry/<init>(JLjava/lang/String;)V"(thread: JVMThread, args: [Long, any, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "timestamp()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/ExpiringCache$Entry/timestamp()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setTimestamp(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ExpiringCache$Entry/setTimestamp(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "val()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ExpiringCache$Entry/val()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "setVal(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ExpiringCache$Entry/setVal(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_RandomAccessFile extends java_lang_Object implements java_io_DataOutput, java_io_DataInput, java_io_Closeable {
    public "java/io/RandomAccessFile/fd": java_io_FileDescriptor;
    public "java/io/RandomAccessFile/channel": java_nio_channels_FileChannel;
    public "java/io/RandomAccessFile/rw": number;
    public "java/io/RandomAccessFile/path": java_lang_String;
    public "java/io/RandomAccessFile/closeLock": java_lang_Object;
    public "java/io/RandomAccessFile/closed": number;
    public static "java/io/RandomAccessFile/O_RDONLY": number;
    public static "java/io/RandomAccessFile/O_RDWR": number;
    public static "java/io/RandomAccessFile/O_SYNC": number;
    public static "java/io/RandomAccessFile/O_DSYNC": number;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/<init>(Ljava/io/File;Ljava/lang/String;)V"(thread: JVMThread, args: [java_io_File, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "java/io/RandomAccessFile/getFD()Ljava/io/FileDescriptor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_FileDescriptor) => void): void;
    public "getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "java/io/RandomAccessFile/getChannel()Ljava/nio/channels/FileChannel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_channels_FileChannel) => void): void;
    public "open0(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/open0(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "open(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/open(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/read0()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readBytes([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readBytes([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/read([B)I"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/write0(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBytes([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeBytes([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFilePointer()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/RandomAccessFile/getFilePointer()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "seek(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/seek(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "seek0(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/seek0(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "length()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/RandomAccessFile/length()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "setLength(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/setLength(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/RandomAccessFile/readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/RandomAccessFile/readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/RandomAccessFile/readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/RandomAccessFile/readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public static "initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/RandomAccessFile/initIDs()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/RandomAccessFile/close0()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/io/RandomAccessFile;)V"(thread: JVMThread, args: [java_io_RandomAccessFile], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/RandomAccessFile/access$000(Ljava/io/RandomAccessFile;)V"(thread: JVMThread, args: [java_io_RandomAccessFile], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/RandomAccessFile/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectInputStream extends java_io_InputStream implements java_io_ObjectInput, java_io_ObjectStreamConstants {
    public static "java/io/ObjectInputStream/NULL_HANDLE": number;
    public static "java/io/ObjectInputStream/unsharedMarker": java_lang_Object;
    public static "java/io/ObjectInputStream/primClasses": java_util_HashMap;
    public "java/io/ObjectInputStream/bin": java_io_ObjectInputStream$BlockDataInputStream;
    public "java/io/ObjectInputStream/vlist": java_io_ObjectInputStream$ValidationList;
    public "java/io/ObjectInputStream/depth": number;
    public "java/io/ObjectInputStream/closed": number;
    public "java/io/ObjectInputStream/handles": java_io_ObjectInputStream$HandleTable;
    public "java/io/ObjectInputStream/passHandle": number;
    public "java/io/ObjectInputStream/defaultDataEnd": number;
    public "java/io/ObjectInputStream/primVals": JVMArray<number>;
    public "java/io/ObjectInputStream/enableOverride": number;
    public "java/io/ObjectInputStream/enableResolve": number;
    public "java/io/ObjectInputStream/curContext": java_io_SerialCallbackContext;
    public "<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/readObject()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readObjectOverride()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/readObjectOverride()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readUnshared()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/readUnshared()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "defaultReadObject()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/defaultReadObject()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readFields()Ljava/io/ObjectInputStream$GetField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream$GetField) => void): void;
    public "java/io/ObjectInputStream/readFields()Ljava/io/ObjectInputStream$GetField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream$GetField) => void): void;
    public "registerValidation(Ljava/io/ObjectInputValidation;I)V"(thread: JVMThread, args: [java_io_ObjectInputValidation, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/registerValidation(Ljava/io/ObjectInputValidation;I)V"(thread: JVMThread, args: [java_io_ObjectInputValidation, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveClass(Ljava/io/ObjectStreamClass;)Ljava/lang/Class;"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/io/ObjectInputStream/resolveClass(Ljava/io/ObjectStreamClass;)Ljava/lang/Class;"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "resolveProxyClass([Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/io/ObjectInputStream/resolveProxyClass([Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [JVMArray<java_lang_String>], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "resolveObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/resolveObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "enableResolveObject(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/enableResolveObject(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readStreamHeader()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/readStreamHeader()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readClassDescriptor()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "java/io/ObjectInputStream/readClassDescriptor()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/ObjectInputStream/readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectInputStream/readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectInputStream/readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "verifySubclass()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/verifySubclass()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "auditSubclass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectInputStream/auditSubclass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readObject0(Z)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/readObject0(Z)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "checkResolve(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/checkResolve(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readTypeString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectInputStream/readTypeString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readNull()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/readNull()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readHandle(Z)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/readHandle(Z)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readClass(Z)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/io/ObjectInputStream/readClass(Z)Ljava/lang/Class;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "readClassDesc(Z)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "java/io/ObjectInputStream/readClassDesc(Z)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "isCustomSubclass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream/isCustomSubclass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readProxyDesc(Z)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "java/io/ObjectInputStream/readProxyDesc(Z)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "readNonProxyDesc(Z)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "java/io/ObjectInputStream/readNonProxyDesc(Z)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "readString(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectInputStream/readString(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readArray(Z)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/readArray(Z)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readEnum(Z)Ljava/lang/Enum;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Enum) => void): void;
    public "java/io/ObjectInputStream/readEnum(Z)Ljava/lang/Enum;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Enum) => void): void;
    public "readOrdinaryObject(Z)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream/readOrdinaryObject(Z)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "readExternalData(Ljava/io/Externalizable;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_Externalizable, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/readExternalData(Ljava/io/Externalizable;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_Externalizable, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "readSerialData(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/readSerialData(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "skipCustomData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/skipCustomData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "defaultReadFields(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/defaultReadFields(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "readFatalException()Ljava/io/IOException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_IOException) => void): void;
    public "java/io/ObjectInputStream/readFatalException()Ljava/io/IOException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_IOException) => void): void;
    public "handleReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream/handleReset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "bytesToFloats([BI[FII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectInputStream/bytesToFloats([BI[FII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "bytesToDoubles([BI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectInputStream/bytesToDoubles([BI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "latestUserDefinedLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "java/io/ObjectInputStream/latestUserDefinedLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public static "cloneArray(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/io/ObjectInputStream/cloneArray(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$000(Ljava/io/ObjectInputStream;)I"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectInputStream/access$000(Ljava/io/ObjectInputStream;)I"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$100(Ljava/io/ObjectInputStream;)Ljava/io/ObjectInputStream$HandleTable;"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream$HandleTable) => void): void;
    public static "java/io/ObjectInputStream/access$100(Ljava/io/ObjectInputStream;)Ljava/io/ObjectInputStream$HandleTable;"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream$HandleTable) => void): void;
    public static "access$200(Ljava/io/ObjectInputStream;)Ljava/io/ObjectInputStream$BlockDataInputStream;"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream$BlockDataInputStream) => void): void;
    public static "java/io/ObjectInputStream/access$200(Ljava/io/ObjectInputStream;)Ljava/io/ObjectInputStream$BlockDataInputStream;"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream$BlockDataInputStream) => void): void;
    public static "access$300(Ljava/io/ObjectInputStream;Z)Ljava/lang/Object;"(thread: JVMThread, args: [java_io_ObjectInputStream, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/io/ObjectInputStream/access$300(Ljava/io/ObjectInputStream;Z)Ljava/lang/Object;"(thread: JVMThread, args: [java_io_ObjectInputStream, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$002(Ljava/io/ObjectInputStream;I)I"(thread: JVMThread, args: [java_io_ObjectInputStream, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectInputStream/access$002(Ljava/io/ObjectInputStream;I)I"(thread: JVMThread, args: [java_io_ObjectInputStream, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500(Ljava/io/ObjectInputStream;)Z"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectInputStream/access$500(Ljava/io/ObjectInputStream;)Z"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$600(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectInputStream/access$600(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$700([BI[FII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectInputStream/access$700([BI[FII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$800([BI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectInputStream/access$800([BI[DII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectInputStream/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_io_ObjectInputValidation extends java_lang_Object {
    "validateObject()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectInputStream$GetField extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$GetField/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjectStreamClass()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "java/io/ObjectInputStream$GetField/getObjectStreamClass()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "defaulted(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$GetField/defaulted(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$GetField/get(Ljava/lang/String;Z)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;B)B"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$GetField/get(Ljava/lang/String;B)B"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;C)C"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$GetField/get(Ljava/lang/String;C)C"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;S)S"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$GetField/get(Ljava/lang/String;S)S"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$GetField/get(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/ObjectInputStream$GetField/get(Ljava/lang/String;J)J"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "get(Ljava/lang/String;F)F"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$GetField/get(Ljava/lang/String;F)F"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;D)D"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$GetField/get(Ljava/lang/String;D)D"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream$GetField/get(Ljava/lang/String;Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_io_SerialCallbackContext extends java_lang_Object {
    public "java/io/SerialCallbackContext/obj": java_lang_Object;
    public "java/io/SerialCallbackContext/desc": java_io_ObjectStreamClass;
    public "java/io/SerialCallbackContext/thread": java_lang_Thread;
    public "<init>(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/SerialCallbackContext/<init>(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObj()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/SerialCallbackContext/getObj()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getDesc()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "java/io/SerialCallbackContext/getDesc()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "check()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/SerialCallbackContext/check()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkAndSetUsed()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/SerialCallbackContext/checkAndSetUsed()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "setUsed()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/SerialCallbackContext/setUsed()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectInputStream$HandleTable extends java_lang_Object {
    public static "java/io/ObjectInputStream$HandleTable/STATUS_OK": number;
    public static "java/io/ObjectInputStream$HandleTable/STATUS_UNKNOWN": number;
    public static "java/io/ObjectInputStream$HandleTable/STATUS_EXCEPTION": number;
    public "java/io/ObjectInputStream$HandleTable/status": JVMArray<number>;
    public "java/io/ObjectInputStream$HandleTable/entries": JVMArray<java_lang_Object>;
    public "java/io/ObjectInputStream$HandleTable/deps": JVMArray<java_io_ObjectInputStream$HandleTable$HandleList>;
    public "java/io/ObjectInputStream$HandleTable/lowDep": number;
    public "java/io/ObjectInputStream$HandleTable/size": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$HandleTable/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "assign(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$HandleTable/assign(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "markDependency(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$HandleTable/markDependency(II)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "markException(ILjava/lang/ClassNotFoundException;)V"(thread: JVMThread, args: [number, java_lang_ClassNotFoundException], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$HandleTable/markException(ILjava/lang/ClassNotFoundException;)V"(thread: JVMThread, args: [number, java_lang_ClassNotFoundException], cb?: (e?: java_lang_Throwable) => void): void;
    public "finish(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$HandleTable/finish(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "setObject(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$HandleTable/setObject(ILjava/lang/Object;)V"(thread: JVMThread, args: [number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "lookupObject(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectInputStream$HandleTable/lookupObject(I)Ljava/lang/Object;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "lookupException(I)Ljava/lang/ClassNotFoundException;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassNotFoundException) => void): void;
    public "java/io/ObjectInputStream$HandleTable/lookupException(I)Ljava/lang/ClassNotFoundException;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassNotFoundException) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$HandleTable/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$HandleTable/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "grow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$HandleTable/grow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ClassNotFoundException extends java_lang_ReflectiveOperationException {
    public static "java/lang/ClassNotFoundException/serialVersionUID": Long;
    public "java/lang/ClassNotFoundException/ex": java_lang_Throwable;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassNotFoundException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassNotFoundException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassNotFoundException/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "getException()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/ClassNotFoundException/getException()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/ClassNotFoundException/getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectInputStream$HandleTable$HandleList extends java_lang_Object {
    public "java/io/ObjectInputStream$HandleTable$HandleList/list": JVMArray<number>;
    public "java/io/ObjectInputStream$HandleTable$HandleList/size": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$HandleTable$HandleList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "add(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$HandleTable$HandleList/add(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "get(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$HandleTable$HandleList/get(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$HandleTable$HandleList/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_io_ObjectInputStream$ValidationList extends java_lang_Object {
    public "java/io/ObjectInputStream$ValidationList/list": java_io_ObjectInputStream$ValidationList$Callback;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$ValidationList/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "register(Ljava/io/ObjectInputValidation;I)V"(thread: JVMThread, args: [java_io_ObjectInputValidation, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$ValidationList/register(Ljava/io/ObjectInputValidation;I)V"(thread: JVMThread, args: [java_io_ObjectInputValidation, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "doCallbacks()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$ValidationList/doCallbacks()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$ValidationList/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$400(Ljava/io/ObjectInputStream$ValidationList;)Ljava/io/ObjectInputStream$ValidationList$Callback;"(thread: JVMThread, args: [java_io_ObjectInputStream$ValidationList], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream$ValidationList$Callback) => void): void;
    public static "java/io/ObjectInputStream$ValidationList/access$400(Ljava/io/ObjectInputStream$ValidationList;)Ljava/io/ObjectInputStream$ValidationList$Callback;"(thread: JVMThread, args: [java_io_ObjectInputStream$ValidationList], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectInputStream$ValidationList$Callback) => void): void;
  }
  export class java_io_ObjectInputStream$ValidationList$Callback extends java_lang_Object {
    public "java/io/ObjectInputStream$ValidationList$Callback/obj": java_io_ObjectInputValidation;
    public "java/io/ObjectInputStream$ValidationList$Callback/priority": number;
    public "java/io/ObjectInputStream$ValidationList$Callback/next": java_io_ObjectInputStream$ValidationList$Callback;
    public "java/io/ObjectInputStream$ValidationList$Callback/acc": java_security_AccessControlContext;
    public "<init>(Ljava/io/ObjectInputValidation;ILjava/io/ObjectInputStream$ValidationList$Callback;Ljava/security/AccessControlContext;)V"(thread: JVMThread, args: [java_io_ObjectInputValidation, number, java_io_ObjectInputStream$ValidationList$Callback, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$ValidationList$Callback/<init>(Ljava/io/ObjectInputValidation;ILjava/io/ObjectInputStream$ValidationList$Callback;Ljava/security/AccessControlContext;)V"(thread: JVMThread, args: [java_io_ObjectInputValidation, number, java_io_ObjectInputStream$ValidationList$Callback, java_security_AccessControlContext], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectInputStream$BlockDataInputStream extends java_io_InputStream implements java_io_DataInput {
    public static "java/io/ObjectInputStream$BlockDataInputStream/MAX_BLOCK_SIZE": number;
    public static "java/io/ObjectInputStream$BlockDataInputStream/MAX_HEADER_SIZE": number;
    public static "java/io/ObjectInputStream$BlockDataInputStream/CHAR_BUF_SIZE": number;
    public static "java/io/ObjectInputStream$BlockDataInputStream/HEADER_BLOCKED": number;
    public "java/io/ObjectInputStream$BlockDataInputStream/buf": JVMArray<number>;
    public "java/io/ObjectInputStream$BlockDataInputStream/hbuf": JVMArray<number>;
    public "java/io/ObjectInputStream$BlockDataInputStream/cbuf": JVMArray<number>;
    public "java/io/ObjectInputStream$BlockDataInputStream/blkmode": number;
    public "java/io/ObjectInputStream$BlockDataInputStream/pos": number;
    public "java/io/ObjectInputStream$BlockDataInputStream/end": number;
    public "java/io/ObjectInputStream$BlockDataInputStream/unread": number;
    public "java/io/ObjectInputStream$BlockDataInputStream/in": java_io_ObjectInputStream$PeekInputStream;
    public "java/io/ObjectInputStream$BlockDataInputStream/din": java_io_DataInputStream;
    public "java/io/ObjectInputStream$BlockDataInputStream/this$0": java_io_ObjectInputStream;
    public "<init>(Ljava/io/ObjectInputStream;Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream, java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/<init>(Ljava/io/ObjectInputStream;Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream, java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBlockDataMode(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/setBlockDataMode(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBlockDataMode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/getBlockDataMode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skipBlockData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/skipBlockData()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "readBlockHeader(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readBlockHeader(Z)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "refill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/refill()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "currentBlockRemaining()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/currentBlockRemaining()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peek()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/peek()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "peekByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/peekByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "read([BIIZ)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/read([BIIZ)I"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readFully([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readFully([BIIZ)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readFully([BIIZ)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/skipBytes(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readBoolean()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readByte()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readUnsignedByte()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readUnsignedShort()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readBooleans([ZII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readBooleans([ZII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readChars([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readChars([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readShorts([SII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readShorts([SII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readInts([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readInts([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readFloats([FII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readFloats([FII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readLongs([JII)V"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readLongs([JII)V"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readDoubles([DII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readDoubles([DII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "readLongUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readLongUTF()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readUTFBody(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readUTFBody(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readUTFSpan(Ljava/lang/StringBuilder;J)J"(thread: JVMThread, args: [java_lang_StringBuilder, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readUTFSpan(Ljava/lang/StringBuilder;J)J"(thread: JVMThread, args: [java_lang_StringBuilder, Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "readUTFChar(Ljava/lang/StringBuilder;J)I"(thread: JVMThread, args: [java_lang_StringBuilder, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$BlockDataInputStream/readUTFChar(Ljava/lang/StringBuilder;J)I"(thread: JVMThread, args: [java_lang_StringBuilder, Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_io_ObjectInputStream$PeekInputStream extends java_io_InputStream {
    public "java/io/ObjectInputStream$PeekInputStream/in": java_io_InputStream;
    public "java/io/ObjectInputStream$PeekInputStream/peekb": number;
    public "<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$PeekInputStream/<init>(Ljava/io/InputStream;)V"(thread: JVMThread, args: [java_io_InputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "peek()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$PeekInputStream/peek()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$PeekInputStream/read()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$PeekInputStream/read([BII)I"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$PeekInputStream/readFully([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/ObjectInputStream$PeekInputStream/skip(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectInputStream$PeekInputStream/available()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectInputStream$PeekInputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface java_io_ObjectStreamConstants extends java_lang_Object {
  }
  export class java_lang_Class extends java_lang_Object implements java_io_Serializable, java_lang_reflect_GenericDeclaration, java_lang_reflect_Type, java_lang_reflect_AnnotatedElement {
    public $cls: ClassData;
    public signers: JVMTypes.java_lang_Object[];
    public static "java/lang/Class/ANNOTATION": number;
    public static "java/lang/Class/ENUM": number;
    public static "java/lang/Class/SYNTHETIC": number;
    public "java/lang/Class/cachedConstructor": java_lang_reflect_Constructor;
    public "java/lang/Class/newInstanceCallerCache": java_lang_Class;
    public "java/lang/Class/name": java_lang_String;
    public "java/lang/Class/classLoader": java_lang_ClassLoader;
    public static "java/lang/Class/allPermDomain": java_security_ProtectionDomain;
    public static "java/lang/Class/useCaches": number;
    public "java/lang/Class/reflectionData": java_lang_ref_SoftReference;
    public "java/lang/Class/classRedefinedCount": number;
    public "java/lang/Class/genericInfo": sun_reflect_generics_repository_ClassRepository;
    public static "java/lang/Class/serialVersionUID": Long;
    public static "java/lang/Class/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/lang/Class/reflectionFactory": sun_reflect_ReflectionFactory;
    public static "java/lang/Class/initted": number;
    public "java/lang/Class/enumConstants": JVMArray<java_lang_Object>;
    public "java/lang/Class/enumConstantDirectory": java_util_Map;
    public "java/lang/Class/annotationData": java_lang_Class$AnnotationData;
    public "java/lang/Class/annotationType": sun_reflect_annotation_AnnotationType;
    public "java/lang/Class/classValueMap": java_lang_ClassValue$ClassValueMap;
    public static "registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Class/registerNatives()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class/<init>(Ljava/lang/ClassLoader;)V"(thread: JVMThread, args: [java_lang_ClassLoader], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/toGenericString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "forName(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/Class/forName(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "forName(Ljava/lang/String;ZLjava/lang/ClassLoader;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, number, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/Class/forName(Ljava/lang/String;ZLjava/lang/ClassLoader;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, number, java_lang_ClassLoader], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "forName0(Ljava/lang/String;ZLjava/lang/ClassLoader;Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, number, java_lang_ClassLoader, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/Class/forName0(Ljava/lang/String;ZLjava/lang/ClassLoader;Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String, number, java_lang_ClassLoader, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "newInstance()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/Class/newInstance()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isInstance(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isInstance(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAssignableFrom(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isAssignableFrom(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInterface()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isInterface()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isArray()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isPrimitive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isPrimitive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isAnnotation()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isAnnotation()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isSynthetic()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getName0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/getName0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/lang/Class/getClassLoader()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "getClassLoader0()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "java/lang/Class/getClassLoader0()Ljava/lang/ClassLoader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassLoader) => void): void;
    public "getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    public "java/lang/Class/getTypeParameters()[Ljava/lang/reflect/TypeVariable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_TypeVariable>) => void): void;
    public "getSuperclass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Class/getSuperclass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getGenericSuperclass()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "java/lang/Class/getGenericSuperclass()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "getPackage()Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public "java/lang/Class/getPackage()Ljava/lang/Package;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Package) => void): void;
    public "getInterfaces()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/Class/getInterfaces()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getInterfaces0()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/Class/getInterfaces0()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getGenericInterfaces()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "java/lang/Class/getGenericInterfaces()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "getComponentType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Class/getComponentType()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/getModifiers()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSigners()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/lang/Class/getSigners()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "setSigners([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class/setSigners([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEnclosingMethod()Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "java/lang/Class/getEnclosingMethod()Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "getEnclosingMethod0()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/lang/Class/getEnclosingMethod0()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "getEnclosingMethodInfo()Ljava/lang/Class$EnclosingMethodInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$EnclosingMethodInfo) => void): void;
    public "java/lang/Class/getEnclosingMethodInfo()Ljava/lang/Class$EnclosingMethodInfo;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$EnclosingMethodInfo) => void): void;
    public static "toClass(Ljava/lang/reflect/Type;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_reflect_Type], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/Class/toClass(Ljava/lang/reflect/Type;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_reflect_Type], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getEnclosingConstructor()Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "java/lang/Class/getEnclosingConstructor()Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Class/getDeclaringClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getDeclaringClass0()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Class/getDeclaringClass0()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getEnclosingClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Class/getEnclosingClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getSimpleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/getSimpleName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getTypeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/getTypeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "isAsciiDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Class/isAsciiDigit(C)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getCanonicalName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/getCanonicalName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isAnonymousClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isAnonymousClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isLocalClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isLocalClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMemberClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isMemberClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getSimpleBinaryName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/getSimpleBinaryName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "isLocalOrAnonymousClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isLocalOrAnonymousClass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getClasses()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/Class/getClasses()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getFields()[Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public "java/lang/Class/getFields()[Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public "getMethods()[Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "java/lang/Class/getMethods()[Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "getConstructors()[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public "java/lang/Class/getConstructors()[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public "getField(Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "java/lang/Class/getField(Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "getMethod(Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "java/lang/Class/getMethod(Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "getConstructor([Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "java/lang/Class/getConstructor([Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "getDeclaredClasses()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/Class/getDeclaredClasses()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "getDeclaredFields()[Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public "java/lang/Class/getDeclaredFields()[Ljava/lang/reflect/Field;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public "getDeclaredMethods()[Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "java/lang/Class/getDeclaredMethods()[Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "getDeclaredConstructors()[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public "java/lang/Class/getDeclaredConstructors()[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public "getDeclaredField(Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "java/lang/Class/getDeclaredField(Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "getDeclaredMethod(Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "java/lang/Class/getDeclaredMethod(Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "getDeclaredConstructor([Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "java/lang/Class/getDeclaredConstructor([Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "getResourceAsStream(Ljava/lang/String;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "java/lang/Class/getResourceAsStream(Ljava/lang/String;)Ljava/io/InputStream;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_InputStream) => void): void;
    public "getResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "java/lang/Class/getResource(Ljava/lang/String;)Ljava/net/URL;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_net_URL) => void): void;
    public "getProtectionDomain()Ljava/security/ProtectionDomain;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_ProtectionDomain) => void): void;
    public "java/lang/Class/getProtectionDomain()Ljava/security/ProtectionDomain;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_ProtectionDomain) => void): void;
    public "getProtectionDomain0()Ljava/security/ProtectionDomain;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_ProtectionDomain) => void): void;
    public "java/lang/Class/getProtectionDomain0()Ljava/security/ProtectionDomain;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_security_ProtectionDomain) => void): void;
    public static "getPrimitiveClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/lang/Class/getPrimitiveClass(Ljava/lang/String;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "checkMemberAccess(ILjava/lang/Class;Z)V"(thread: JVMThread, args: [number, java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class/checkMemberAccess(ILjava/lang/Class;Z)V"(thread: JVMThread, args: [number, java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkPackageAccess(Ljava/lang/ClassLoader;Z)V"(thread: JVMThread, args: [java_lang_ClassLoader, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class/checkPackageAccess(Ljava/lang/ClassLoader;Z)V"(thread: JVMThread, args: [java_lang_ClassLoader, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "resolveName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/resolveName(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "reflectionData()Ljava/lang/Class$ReflectionData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$ReflectionData) => void): void;
    public "java/lang/Class/reflectionData()Ljava/lang/Class$ReflectionData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$ReflectionData) => void): void;
    public "newReflectionData(Ljava/lang/ref/SoftReference;I)Ljava/lang/Class$ReflectionData;"(thread: JVMThread, args: [java_lang_ref_SoftReference, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$ReflectionData) => void): void;
    public "java/lang/Class/newReflectionData(Ljava/lang/ref/SoftReference;I)Ljava/lang/Class$ReflectionData;"(thread: JVMThread, args: [java_lang_ref_SoftReference, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$ReflectionData) => void): void;
    public "getGenericSignature0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/getGenericSignature0()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "java/lang/Class/getFactory()Lsun/reflect/generics/factory/GenericsFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_factory_GenericsFactory) => void): void;
    public "getGenericInfo()Lsun/reflect/generics/repository/ClassRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ClassRepository) => void): void;
    public "java/lang/Class/getGenericInfo()Lsun/reflect/generics/repository/ClassRepository;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ClassRepository) => void): void;
    public "getRawAnnotations()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/Class/getRawAnnotations()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getRawTypeAnnotations()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/Class/getRawTypeAnnotations()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "getExecutableTypeAnnotationBytes(Ljava/lang/reflect/Executable;)[B"(thread: JVMThread, args: [java_lang_reflect_Executable], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/lang/Class/getExecutableTypeAnnotationBytes(Ljava/lang/reflect/Executable;)[B"(thread: JVMThread, args: [java_lang_reflect_Executable], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getConstantPool()Lsun/reflect/ConstantPool;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstantPool) => void): void;
    public "java/lang/Class/getConstantPool()Lsun/reflect/ConstantPool;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ConstantPool) => void): void;
    public "privateGetDeclaredFields(Z)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public "java/lang/Class/privateGetDeclaredFields(Z)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public "privateGetPublicFields(Ljava/util/Set;)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public "java/lang/Class/privateGetPublicFields(Ljava/util/Set;)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_util_Set], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public static "addAll(Ljava/util/Collection;[Ljava/lang/reflect/Field;)V"(thread: JVMThread, args: [java_util_Collection, JVMArray<java_lang_reflect_Field>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Class/addAll(Ljava/util/Collection;[Ljava/lang/reflect/Field;)V"(thread: JVMThread, args: [java_util_Collection, JVMArray<java_lang_reflect_Field>], cb?: (e?: java_lang_Throwable) => void): void;
    public "privateGetDeclaredConstructors(Z)[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public "java/lang/Class/privateGetDeclaredConstructors(Z)[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public "privateGetDeclaredMethods(Z)[Ljava/lang/reflect/Method;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "java/lang/Class/privateGetDeclaredMethods(Z)[Ljava/lang/reflect/Method;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "privateGetPublicMethods()[Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "java/lang/Class/privateGetPublicMethods()[Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public static "searchFields([Ljava/lang/reflect/Field;Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Field>, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public static "java/lang/Class/searchFields([Ljava/lang/reflect/Field;Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Field>, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "getField0(Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public "java/lang/Class/getField0(Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public static "searchMethods([Ljava/lang/reflect/Method;Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Method>, java_lang_String, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/lang/Class/searchMethods([Ljava/lang/reflect/Method;Ljava/lang/String;[Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Method>, java_lang_String, JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "getMethod0(Ljava/lang/String;[Ljava/lang/Class;Z)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "java/lang/Class/getMethod0(Ljava/lang/String;[Ljava/lang/Class;Z)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "privateGetMethodRecursive(Ljava/lang/String;[Ljava/lang/Class;ZLjava/lang/Class$MethodArray;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Class>, number, java_lang_Class$MethodArray], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "java/lang/Class/privateGetMethodRecursive(Ljava/lang/String;[Ljava/lang/Class;ZLjava/lang/Class$MethodArray;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Class>, number, java_lang_Class$MethodArray], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "getConstructor0([Ljava/lang/Class;I)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public "java/lang/Class/getConstructor0([Ljava/lang/Class;I)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "arrayContentsEq([Ljava/lang/Object;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Class/arrayContentsEq([Ljava/lang/Object;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "copyFields([Ljava/lang/reflect/Field;)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Field>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public static "java/lang/Class/copyFields([Ljava/lang/reflect/Field;)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Field>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public static "copyMethods([Ljava/lang/reflect/Method;)[Ljava/lang/reflect/Method;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Method>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public static "java/lang/Class/copyMethods([Ljava/lang/reflect/Method;)[Ljava/lang/reflect/Method;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Method>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public static "copyConstructors([Ljava/lang/reflect/Constructor;)[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Constructor>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public static "java/lang/Class/copyConstructors([Ljava/lang/reflect/Constructor;)[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Constructor>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public "getDeclaredFields0(Z)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public "java/lang/Class/getDeclaredFields0(Z)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public "getDeclaredMethods0(Z)[Ljava/lang/reflect/Method;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "java/lang/Class/getDeclaredMethods0(Z)[Ljava/lang/reflect/Method;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "getDeclaredConstructors0(Z)[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public "java/lang/Class/getDeclaredConstructors0(Z)[Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Constructor>) => void): void;
    public "getDeclaredClasses0()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public "java/lang/Class/getDeclaredClasses0()[Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Class>) => void): void;
    public static "argumentTypesToString([Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/Class/argumentTypesToString([Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_Class>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "desiredAssertionStatus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/desiredAssertionStatus()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "desiredAssertionStatus0(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Class/desiredAssertionStatus0(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEnum()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isEnum()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getReflectionFactory()Lsun/reflect/ReflectionFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ReflectionFactory) => void): void;
    public static "java/lang/Class/getReflectionFactory()Lsun/reflect/ReflectionFactory;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_ReflectionFactory) => void): void;
    public static "checkInitted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Class/checkInitted()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getEnumConstants()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/lang/Class/getEnumConstants()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "getEnumConstantsShared()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "java/lang/Class/getEnumConstantsShared()[Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Object>) => void): void;
    public "enumConstantDirectory()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/lang/Class/enumConstantDirectory()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "cast(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/Class/cast(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "cannotCastMsg(Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class/cannotCastMsg(Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "asSubclass(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Class/asSubclass(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/Class/getAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/Class/getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/Class/getAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "java/lang/Class/getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    public "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/Class/getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "java/lang/Class/getDeclaredAnnotations()[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    public "annotationData()Ljava/lang/Class$AnnotationData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$AnnotationData) => void): void;
    public "java/lang/Class/annotationData()Ljava/lang/Class$AnnotationData;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$AnnotationData) => void): void;
    public "createAnnotationData(I)Ljava/lang/Class$AnnotationData;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$AnnotationData) => void): void;
    public "java/lang/Class/createAnnotationData(I)Ljava/lang/Class$AnnotationData;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class$AnnotationData) => void): void;
    public "casAnnotationType(Lsun/reflect/annotation/AnnotationType;Lsun/reflect/annotation/AnnotationType;)Z"(thread: JVMThread, args: [sun_reflect_annotation_AnnotationType, sun_reflect_annotation_AnnotationType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class/casAnnotationType(Lsun/reflect/annotation/AnnotationType;Lsun/reflect/annotation/AnnotationType;)Z"(thread: JVMThread, args: [sun_reflect_annotation_AnnotationType, sun_reflect_annotation_AnnotationType], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getAnnotationType()Lsun/reflect/annotation/AnnotationType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_annotation_AnnotationType) => void): void;
    public "java/lang/Class/getAnnotationType()Lsun/reflect/annotation/AnnotationType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_annotation_AnnotationType) => void): void;
    public "getDeclaredAnnotationMap()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "java/lang/Class/getDeclaredAnnotationMap()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "getAnnotatedSuperclass()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "java/lang/Class/getAnnotatedSuperclass()Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_AnnotatedType) => void): void;
    public "getAnnotatedInterfaces()[Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_AnnotatedType>) => void): void;
    public "java/lang/Class/getAnnotatedInterfaces()[Ljava/lang/reflect/AnnotatedType;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_AnnotatedType>) => void): void;
    public static "access$100(Ljava/lang/Class;Z)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public static "java/lang/Class/access$100(Ljava/lang/Class;Z)[Ljava/lang/reflect/Field;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Field>) => void): void;
    public static "access$200([Ljava/lang/reflect/Field;Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Field>, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public static "java/lang/Class/access$200([Ljava/lang/reflect/Field;Ljava/lang/String;)Ljava/lang/reflect/Field;"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Field>, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Field) => void): void;
    public static "access$300([Ljava/lang/Object;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Class/access$300([Ljava/lang/Object;[Ljava/lang/Object;)Z"(thread: JVMThread, args: [JVMArray<java_lang_Object>, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$402(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Class/access$402(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$502(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Class/access$502(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Class/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getTypeName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    "isAnnotationPresent(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    "getAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
    "getDeclaredAnnotation(Ljava/lang/Class;)Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_Annotation) => void): void;
    "getDeclaredAnnotationsByType(Ljava/lang/Class;)[Ljava/lang/annotation/Annotation;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_Annotation>) => void): void;
  }
  export class java_lang_Class$MethodArray extends java_lang_Object {
    public "java/lang/Class$MethodArray/methods": JVMArray<java_lang_reflect_Method>;
    public "java/lang/Class$MethodArray/length": number;
    public "java/lang/Class$MethodArray/defaults": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "hasDefaults()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class$MethodArray/hasDefaults()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "add(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/add(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAll([Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Method>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/addAll([Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Method>], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAll(Ljava/lang/Class$MethodArray;)V"(thread: JVMThread, args: [java_lang_Class$MethodArray], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/addAll(Ljava/lang/Class$MethodArray;)V"(thread: JVMThread, args: [java_lang_Class$MethodArray], cb?: (e?: java_lang_Throwable) => void): void;
    public "addIfNotPresent(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/addIfNotPresent(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "addAllIfNotPresent(Ljava/lang/Class$MethodArray;)V"(thread: JVMThread, args: [java_lang_Class$MethodArray], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/addAllIfNotPresent(Ljava/lang/Class$MethodArray;)V"(thread: JVMThread, args: [java_lang_Class$MethodArray], cb?: (e?: java_lang_Throwable) => void): void;
    public "addInterfaceMethods([Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Method>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/addInterfaceMethods([Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [JVMArray<java_lang_reflect_Method>], cb?: (e?: java_lang_Throwable) => void): void;
    public "length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class$MethodArray/length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(I)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "java/lang/Class$MethodArray/get(I)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "getFirst()Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "java/lang/Class$MethodArray/getFirst()Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public "removeByNameAndDescriptor(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/removeByNameAndDescriptor(Ljava/lang/reflect/Method;)V"(thread: JVMThread, args: [java_lang_reflect_Method], cb?: (e?: java_lang_Throwable) => void): void;
    public "remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/remove(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "matchesNameAndDescriptor(Ljava/lang/reflect/Method;Ljava/lang/reflect/Method;)Z"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class$MethodArray/matchesNameAndDescriptor(Ljava/lang/reflect/Method;Ljava/lang/reflect/Method;)Z"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compactAndTrim()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/compactAndTrim()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeLessSpecifics()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$MethodArray/removeLessSpecifics()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getArray()[Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public "java/lang/Class$MethodArray/getArray()[Ljava/lang/reflect/Method;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Method>) => void): void;
    public static "hasMoreSpecificClass(Ljava/lang/reflect/Method;Ljava/lang/reflect/Method;)Z"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/Class$MethodArray/hasMoreSpecificClass(Ljava/lang/reflect/Method;Ljava/lang/reflect/Method;)Z"(thread: JVMThread, args: [java_lang_reflect_Method, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_lang_Class$ReflectionData extends java_lang_Object {
    public "java/lang/Class$ReflectionData/declaredFields": JVMArray<java_lang_reflect_Field>;
    public "java/lang/Class$ReflectionData/publicFields": JVMArray<java_lang_reflect_Field>;
    public "java/lang/Class$ReflectionData/declaredMethods": JVMArray<java_lang_reflect_Method>;
    public "java/lang/Class$ReflectionData/publicMethods": JVMArray<java_lang_reflect_Method>;
    public "java/lang/Class$ReflectionData/declaredConstructors": JVMArray<java_lang_reflect_Constructor>;
    public "java/lang/Class$ReflectionData/publicConstructors": JVMArray<java_lang_reflect_Constructor>;
    public "java/lang/Class$ReflectionData/declaredPublicFields": JVMArray<java_lang_reflect_Field>;
    public "java/lang/Class$ReflectionData/declaredPublicMethods": JVMArray<java_lang_reflect_Method>;
    public "java/lang/Class$ReflectionData/interfaces": JVMArray<java_lang_Class>;
    public "java/lang/Class$ReflectionData/redefinedCount": number;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$ReflectionData/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Class$EnclosingMethodInfo extends java_lang_Object {
    public "java/lang/Class$EnclosingMethodInfo/enclosingClass": java_lang_Class;
    public "java/lang/Class$EnclosingMethodInfo/name": java_lang_String;
    public "java/lang/Class$EnclosingMethodInfo/descriptor": java_lang_String;
    public static "java/lang/Class$EnclosingMethodInfo/$assertionsDisabled": number;
    public "<init>([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$EnclosingMethodInfo/<init>([Ljava/lang/Object;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "isPartial()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class$EnclosingMethodInfo/isPartial()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isConstructor()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class$EnclosingMethodInfo/isConstructor()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Class$EnclosingMethodInfo/isMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getEnclosingClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/lang/Class$EnclosingMethodInfo/getEnclosingClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class$EnclosingMethodInfo/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getDescriptor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Class$EnclosingMethodInfo/getDescriptor()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "<init>([Ljava/lang/Object;Ljava/lang/Class$1;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_Class$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$EnclosingMethodInfo/<init>([Ljava/lang/Object;Ljava/lang/Class$1;)V"(thread: JVMThread, args: [JVMArray<java_lang_Object>, java_lang_Class$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Class$EnclosingMethodInfo/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Class$1 extends java_lang_Object implements java_security_PrivilegedAction {
    public "java/lang/Class$1/val$c": java_lang_reflect_Constructor;
    public "java/lang/Class$1/this$0": java_lang_Class;
    public "<init>(Ljava/lang/Class;Ljava/lang/reflect/Constructor;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$1/<init>(Ljava/lang/Class;Ljava/lang/reflect/Constructor;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "java/lang/Class$1/run()Ljava/lang/Void;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Void) => void): void;
    public "run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/Class$1/run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_lang_ClassValue$ClassValueMap extends java_util_WeakHashMap {
    public "java/lang/ClassValue$ClassValueMap/type": java_lang_Class;
    public "java/lang/ClassValue$ClassValueMap/cacheArray": JVMArray<java_lang_ClassValue$Entry>;
    public "java/lang/ClassValue$ClassValueMap/cacheLoad": number;
    public "java/lang/ClassValue$ClassValueMap/cacheLoadLimit": number;
    public static "java/lang/ClassValue$ClassValueMap/INITIAL_ENTRIES": number;
    public static "java/lang/ClassValue$ClassValueMap/CACHE_LOAD_LIMIT": number;
    public static "java/lang/ClassValue$ClassValueMap/PROBE_LIMIT": number;
    public static "java/lang/ClassValue$ClassValueMap/$assertionsDisabled": number;
    public "<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "getCache()[Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_ClassValue$Entry>) => void): void;
    public "java/lang/ClassValue$ClassValueMap/getCache()[Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_ClassValue$Entry>) => void): void;
    public "startEntry(Ljava/lang/ClassValue;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "java/lang/ClassValue$ClassValueMap/startEntry(Ljava/lang/ClassValue;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "finishEntry(Ljava/lang/ClassValue;Ljava/lang/ClassValue$Entry;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue, java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "java/lang/ClassValue$ClassValueMap/finishEntry(Ljava/lang/ClassValue;Ljava/lang/ClassValue$Entry;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue, java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "removeEntry(Ljava/lang/ClassValue;)V"(thread: JVMThread, args: [java_lang_ClassValue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/removeEntry(Ljava/lang/ClassValue;)V"(thread: JVMThread, args: [java_lang_ClassValue], cb?: (e?: java_lang_Throwable) => void): void;
    public "changeEntry(Ljava/lang/ClassValue;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ClassValue, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/changeEntry(Ljava/lang/ClassValue;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ClassValue, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "loadFromCache([Ljava/lang/ClassValue$Entry;I)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "java/lang/ClassValue$ClassValueMap/loadFromCache([Ljava/lang/ClassValue$Entry;I)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "probeHomeLocation([Ljava/lang/ClassValue$Entry;Ljava/lang/ClassValue;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, java_lang_ClassValue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "java/lang/ClassValue$ClassValueMap/probeHomeLocation([Ljava/lang/ClassValue$Entry;Ljava/lang/ClassValue;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, java_lang_ClassValue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "probeBackupLocations([Ljava/lang/ClassValue$Entry;Ljava/lang/ClassValue;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, java_lang_ClassValue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "java/lang/ClassValue$ClassValueMap/probeBackupLocations([Ljava/lang/ClassValue$Entry;Ljava/lang/ClassValue;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, java_lang_ClassValue], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "entryDislocation([Ljava/lang/ClassValue$Entry;ILjava/lang/ClassValue$Entry;)I"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number, java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/ClassValue$ClassValueMap/entryDislocation([Ljava/lang/ClassValue$Entry;ILjava/lang/ClassValue$Entry;)I"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number, java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sizeCache(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/sizeCache(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkCacheLoad()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/checkCacheLoad()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "reduceCacheLoad()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/reduceCacheLoad()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeStaleEntries([Ljava/lang/ClassValue$Entry;II)V"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/removeStaleEntries([Ljava/lang/ClassValue$Entry;II)V"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "findReplacement([Ljava/lang/ClassValue$Entry;I)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "java/lang/ClassValue$ClassValueMap/findReplacement([Ljava/lang/ClassValue$Entry;I)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "removeStaleEntries(Ljava/lang/ClassValue;)V"(thread: JVMThread, args: [java_lang_ClassValue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/removeStaleEntries(Ljava/lang/ClassValue;)V"(thread: JVMThread, args: [java_lang_ClassValue], cb?: (e?: java_lang_Throwable) => void): void;
    public "removeStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/removeStaleEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "addToCache(Ljava/lang/ClassValue$Entry;)V"(thread: JVMThread, args: [java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/addToCache(Ljava/lang/ClassValue$Entry;)V"(thread: JVMThread, args: [java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "addToCache(Ljava/lang/ClassValue;Ljava/lang/ClassValue$Entry;)V"(thread: JVMThread, args: [java_lang_ClassValue, java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$ClassValueMap/addToCache(Ljava/lang/ClassValue;Ljava/lang/ClassValue$Entry;)V"(thread: JVMThread, args: [java_lang_ClassValue, java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable) => void): void;
    public "placeInCache([Ljava/lang/ClassValue$Entry;ILjava/lang/ClassValue$Entry;Z)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number, java_lang_ClassValue$Entry, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "java/lang/ClassValue$ClassValueMap/placeInCache([Ljava/lang/ClassValue$Entry;ILjava/lang/ClassValue$Entry;Z)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, number, java_lang_ClassValue$Entry, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "overwrittenEntry(Ljava/lang/ClassValue$Entry;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "java/lang/ClassValue$ClassValueMap/overwrittenEntry(Ljava/lang/ClassValue$Entry;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ClassValue$ClassValueMap/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ClassValue extends java_lang_Object {
    public static "java/lang/ClassValue/EMPTY_CACHE": JVMArray<java_lang_ClassValue$Entry>;
    public "java/lang/ClassValue/hashCodeForCache": number;
    public static "java/lang/ClassValue/nextHashCode": java_util_concurrent_atomic_AtomicInteger;
    public static "java/lang/ClassValue/HASH_INCREMENT": number;
    public static "java/lang/ClassValue/HASH_MASK": number;
    public "java/lang/ClassValue/identity": java_lang_ClassValue$Identity;
    public "java/lang/ClassValue/version": java_lang_ClassValue$Version;
    public static "java/lang/ClassValue/CRITICAL_SECTION": java_lang_Object;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "computeValue(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ClassValue/computeValue(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "get(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ClassValue/get(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "remove(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue/remove(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/Class;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue/put(Ljava/lang/Class;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getCacheCarefully(Ljava/lang/Class;)[Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_ClassValue$Entry>) => void): void;
    public static "java/lang/ClassValue/getCacheCarefully(Ljava/lang/Class;)[Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_ClassValue$Entry>) => void): void;
    public "getFromBackup([Ljava/lang/ClassValue$Entry;Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ClassValue/getFromBackup([Ljava/lang/ClassValue$Entry;Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [JVMArray<java_lang_ClassValue$Entry>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "castEntry(Ljava/lang/ClassValue$Entry;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "java/lang/ClassValue/castEntry(Ljava/lang/ClassValue$Entry;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "getFromHashMap(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ClassValue/getFromHashMap(Ljava/lang/Class;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "match(Ljava/lang/ClassValue$Entry;)Z"(thread: JVMThread, args: [java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ClassValue/match(Ljava/lang/ClassValue$Entry;)Z"(thread: JVMThread, args: [java_lang_ClassValue$Entry], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "version()Ljava/lang/ClassValue$Version;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Version) => void): void;
    public "java/lang/ClassValue/version()Ljava/lang/ClassValue$Version;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Version) => void): void;
    public "bumpVersion()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue/bumpVersion()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getMap(Ljava/lang/Class;)Ljava/lang/ClassValue$ClassValueMap;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$ClassValueMap) => void): void;
    public static "java/lang/ClassValue/getMap(Ljava/lang/Class;)Ljava/lang/ClassValue$ClassValueMap;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$ClassValueMap) => void): void;
    public static "initializeMap(Ljava/lang/Class;)Ljava/lang/ClassValue$ClassValueMap;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$ClassValueMap) => void): void;
    public static "java/lang/ClassValue/initializeMap(Ljava/lang/Class;)Ljava/lang/ClassValue$ClassValueMap;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$ClassValueMap) => void): void;
    public static "makeEntry(Ljava/lang/ClassValue$Version;Ljava/lang/Object;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue$Version, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "java/lang/ClassValue/makeEntry(Ljava/lang/ClassValue$Version;Ljava/lang/Object;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue$Version, java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ClassValue/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ClassValue$Version extends java_lang_Object {
    public "java/lang/ClassValue$Version/classValue": java_lang_ClassValue;
    public "java/lang/ClassValue$Version/promise": java_lang_ClassValue$Entry;
    public "<init>(Ljava/lang/ClassValue;)V"(thread: JVMThread, args: [java_lang_ClassValue], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$Version/<init>(Ljava/lang/ClassValue;)V"(thread: JVMThread, args: [java_lang_ClassValue], cb?: (e?: java_lang_Throwable) => void): void;
    public "classValue()Ljava/lang/ClassValue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue) => void): void;
    public "java/lang/ClassValue$Version/classValue()Ljava/lang/ClassValue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue) => void): void;
    public "promise()Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "java/lang/ClassValue$Version/promise()Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "isLive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ClassValue$Version/isLive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_lang_ClassValue$Identity extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$Identity/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_ClassValue$Entry extends java_lang_ref_WeakReference {
    public "java/lang/ClassValue$Entry/value": java_lang_Object;
    public static "java/lang/ClassValue$Entry/DEAD_ENTRY": java_lang_ClassValue$Entry;
    public static "java/lang/ClassValue$Entry/$assertionsDisabled": number;
    public "<init>(Ljava/lang/ClassValue$Version;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ClassValue$Version, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$Entry/<init>(Ljava/lang/ClassValue$Version;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_ClassValue$Version, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "assertNotPromise()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$Entry/assertNotPromise()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/ClassValue$Version;)V"(thread: JVMThread, args: [java_lang_ClassValue$Version], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ClassValue$Entry/<init>(Ljava/lang/ClassValue$Version;)V"(thread: JVMThread, args: [java_lang_ClassValue$Version], cb?: (e?: java_lang_Throwable) => void): void;
    public "value()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/ClassValue$Entry/value()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "isPromise()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ClassValue$Entry/isPromise()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "version()Ljava/lang/ClassValue$Version;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Version) => void): void;
    public "java/lang/ClassValue$Entry/version()Ljava/lang/ClassValue$Version;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Version) => void): void;
    public "classValueOrNull()Ljava/lang/ClassValue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue) => void): void;
    public "java/lang/ClassValue$Entry/classValueOrNull()Ljava/lang/ClassValue;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue) => void): void;
    public "isLive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/ClassValue$Entry/isLive()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "refreshVersion(Ljava/lang/ClassValue$Version;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue$Version], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public "java/lang/ClassValue$Entry/refreshVersion(Ljava/lang/ClassValue$Version;)Ljava/lang/ClassValue$Entry;"(thread: JVMThread, args: [java_lang_ClassValue$Version], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassValue$Entry) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/ClassValue$Entry/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_annotation_AnnotationType extends java_lang_Object {
    public "sun/reflect/annotation/AnnotationType/memberTypes": java_util_Map;
    public "sun/reflect/annotation/AnnotationType/memberDefaults": java_util_Map;
    public "sun/reflect/annotation/AnnotationType/members": java_util_Map;
    public "sun/reflect/annotation/AnnotationType/retention": java_lang_annotation_RetentionPolicy;
    public "sun/reflect/annotation/AnnotationType/inherited": number;
    public static "sun/reflect/annotation/AnnotationType/$assertionsDisabled": number;
    public static "getInstance(Ljava/lang/Class;)Lsun/reflect/annotation/AnnotationType;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_annotation_AnnotationType) => void): void;
    public static "sun/reflect/annotation/AnnotationType/getInstance(Ljava/lang/Class;)Lsun/reflect/annotation/AnnotationType;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_annotation_AnnotationType) => void): void;
    public "<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/annotation/AnnotationType/<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public static "invocationHandlerReturnType(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "sun/reflect/annotation/AnnotationType/invocationHandlerReturnType(Ljava/lang/Class;)Ljava/lang/Class;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "memberTypes()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "sun/reflect/annotation/AnnotationType/memberTypes()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "members()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "sun/reflect/annotation/AnnotationType/members()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "memberDefaults()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "sun/reflect/annotation/AnnotationType/memberDefaults()Ljava/util/Map;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Map) => void): void;
    public "retention()Ljava/lang/annotation/RetentionPolicy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_RetentionPolicy) => void): void;
    public "sun/reflect/annotation/AnnotationType/retention()Ljava/lang/annotation/RetentionPolicy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_RetentionPolicy) => void): void;
    public "isInherited()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/reflect/annotation/AnnotationType/isInherited()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/reflect/annotation/AnnotationType/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/annotation/AnnotationType/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_annotation_RetentionPolicy extends java_lang_Enum {
    public static "java/lang/annotation/RetentionPolicy/SOURCE": java_lang_annotation_RetentionPolicy;
    public static "java/lang/annotation/RetentionPolicy/CLASS": java_lang_annotation_RetentionPolicy;
    public static "java/lang/annotation/RetentionPolicy/RUNTIME": java_lang_annotation_RetentionPolicy;
    public static "java/lang/annotation/RetentionPolicy/$VALUES": JVMArray<java_lang_annotation_RetentionPolicy>;
    public static "values()[Ljava/lang/annotation/RetentionPolicy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_RetentionPolicy>) => void): void;
    public static "java/lang/annotation/RetentionPolicy/values()[Ljava/lang/annotation/RetentionPolicy;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_annotation_RetentionPolicy>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ljava/lang/annotation/RetentionPolicy;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_RetentionPolicy) => void): void;
    public static "java/lang/annotation/RetentionPolicy/valueOf(Ljava/lang/String;)Ljava/lang/annotation/RetentionPolicy;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_annotation_RetentionPolicy) => void): void;
    public "<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/annotation/RetentionPolicy/<init>(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/annotation/RetentionPolicy/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Class$AnnotationData extends java_lang_Object {
    public "java/lang/Class$AnnotationData/annotations": java_util_Map;
    public "java/lang/Class$AnnotationData/declaredAnnotations": java_util_Map;
    public "java/lang/Class$AnnotationData/redefinedCount": number;
    public "<init>(Ljava/util/Map;Ljava/util/Map;I)V"(thread: JVMThread, args: [java_util_Map, java_util_Map, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Class$AnnotationData/<init>(Ljava/util/Map;Ljava/util/Map;I)V"(thread: JVMThread, args: [java_util_Map, java_util_Map, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_reflect_generics_repository_ClassRepository extends sun_reflect_generics_repository_GenericDeclRepository {
    public static "sun/reflect/generics/repository/ClassRepository/NONE": sun_reflect_generics_repository_ClassRepository;
    public "sun/reflect/generics/repository/ClassRepository/superclass": java_lang_reflect_Type;
    public "sun/reflect/generics/repository/ClassRepository/superInterfaces": JVMArray<java_lang_reflect_Type>;
    public "<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/reflect/generics/repository/ClassRepository/<init>(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)V"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable) => void): void;
    public "parse(Ljava/lang/String;)Lsun/reflect/generics/tree/ClassSignature;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ClassSignature) => void): void;
    public "sun/reflect/generics/repository/ClassRepository/parse(Ljava/lang/String;)Lsun/reflect/generics/tree/ClassSignature;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_ClassSignature) => void): void;
    public static "make(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/repository/ClassRepository;"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ClassRepository) => void): void;
    public static "sun/reflect/generics/repository/ClassRepository/make(Ljava/lang/String;Lsun/reflect/generics/factory/GenericsFactory;)Lsun/reflect/generics/repository/ClassRepository;"(thread: JVMThread, args: [java_lang_String, sun_reflect_generics_factory_GenericsFactory], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_repository_ClassRepository) => void): void;
    public "getSuperclass()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "sun/reflect/generics/repository/ClassRepository/getSuperclass()Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Type) => void): void;
    public "getSuperInterfaces()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "sun/reflect/generics/repository/ClassRepository/getSuperInterfaces()[Ljava/lang/reflect/Type;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_reflect_Type>) => void): void;
    public "parse(Ljava/lang/String;)Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
    public "sun/reflect/generics/repository/ClassRepository/parse(Ljava/lang/String;)Lsun/reflect/generics/tree/Tree;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: sun_reflect_generics_tree_Tree) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "sun/reflect/generics/repository/ClassRepository/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectStreamClass extends java_lang_Object implements java_io_Serializable {
    public static "java/io/ObjectStreamClass/NO_FIELDS": JVMArray<java_io_ObjectStreamField>;
    public static "java/io/ObjectStreamClass/serialVersionUID": Long;
    public static "java/io/ObjectStreamClass/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/io/ObjectStreamClass/reflFactory": sun_reflect_ReflectionFactory;
    public "java/io/ObjectStreamClass/cl": java_lang_Class;
    public "java/io/ObjectStreamClass/name": java_lang_String;
    public "java/io/ObjectStreamClass/suid": java_lang_Long;
    public "java/io/ObjectStreamClass/isProxy": number;
    public "java/io/ObjectStreamClass/isEnum": number;
    public "java/io/ObjectStreamClass/serializable": number;
    public "java/io/ObjectStreamClass/externalizable": number;
    public "java/io/ObjectStreamClass/hasWriteObjectData": number;
    public "java/io/ObjectStreamClass/hasBlockExternalData": number;
    public "java/io/ObjectStreamClass/resolveEx": java_lang_ClassNotFoundException;
    public "java/io/ObjectStreamClass/deserializeEx": java_io_ObjectStreamClass$ExceptionInfo;
    public "java/io/ObjectStreamClass/serializeEx": java_io_ObjectStreamClass$ExceptionInfo;
    public "java/io/ObjectStreamClass/defaultSerializeEx": java_io_ObjectStreamClass$ExceptionInfo;
    public "java/io/ObjectStreamClass/fields": JVMArray<java_io_ObjectStreamField>;
    public "java/io/ObjectStreamClass/primDataSize": number;
    public "java/io/ObjectStreamClass/numObjFields": number;
    public "java/io/ObjectStreamClass/fieldRefl": java_io_ObjectStreamClass$FieldReflector;
    public "java/io/ObjectStreamClass/dataLayout": JVMArray<java_io_ObjectStreamClass$ClassDataSlot>;
    public "java/io/ObjectStreamClass/cons": java_lang_reflect_Constructor;
    public "java/io/ObjectStreamClass/writeObjectMethod": java_lang_reflect_Method;
    public "java/io/ObjectStreamClass/readObjectMethod": java_lang_reflect_Method;
    public "java/io/ObjectStreamClass/readObjectNoDataMethod": java_lang_reflect_Method;
    public "java/io/ObjectStreamClass/writeReplaceMethod": java_lang_reflect_Method;
    public "java/io/ObjectStreamClass/readResolveMethod": java_lang_reflect_Method;
    public "java/io/ObjectStreamClass/localDesc": java_io_ObjectStreamClass;
    public "java/io/ObjectStreamClass/superDesc": java_io_ObjectStreamClass;
    public "java/io/ObjectStreamClass/initialized": number;
    public static "initNative()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectStreamClass/initNative()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "lookup(Ljava/lang/Class;)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public static "java/io/ObjectStreamClass/lookup(Ljava/lang/Class;)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public static "lookupAny(Ljava/lang/Class;)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public static "java/io/ObjectStreamClass/lookupAny(Ljava/lang/Class;)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectStreamClass/getName()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getSerialVersionUID()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/ObjectStreamClass/getSerialVersionUID()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "forClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "java/io/ObjectStreamClass/forClass()Ljava/lang/Class;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public "getFields()[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public "java/io/ObjectStreamClass/getFields()[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public "getField(Ljava/lang/String;)Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamField) => void): void;
    public "java/io/ObjectStreamClass/getField(Ljava/lang/String;)Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamField) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectStreamClass/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "lookup(Ljava/lang/Class;Z)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public static "java/io/ObjectStreamClass/lookup(Ljava/lang/Class;Z)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/<init>(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "initProxy(Ljava/lang/Class;Ljava/lang/ClassNotFoundException;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_ClassNotFoundException, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/initProxy(Ljava/lang/Class;Ljava/lang/ClassNotFoundException;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Class, java_lang_ClassNotFoundException, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "initNonProxy(Ljava/io/ObjectStreamClass;Ljava/lang/Class;Ljava/lang/ClassNotFoundException;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_Class, java_lang_ClassNotFoundException, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/initNonProxy(Ljava/io/ObjectStreamClass;Ljava/lang/Class;Ljava/lang/ClassNotFoundException;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_Class, java_lang_ClassNotFoundException, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "readNonProxy(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/readNonProxy(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeNonProxy(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/writeNonProxy(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "getResolveException()Ljava/lang/ClassNotFoundException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassNotFoundException) => void): void;
    public "java/io/ObjectStreamClass/getResolveException()Ljava/lang/ClassNotFoundException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_ClassNotFoundException) => void): void;
    public "requireInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/requireInitialized()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkDeserialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/checkDeserialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkSerialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/checkSerialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "checkDefaultSerialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/checkDefaultSerialize()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSuperDesc()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "java/io/ObjectStreamClass/getSuperDesc()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "getLocalDesc()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "java/io/ObjectStreamClass/getLocalDesc()Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "getFields(Z)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public "java/io/ObjectStreamClass/getFields(Z)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public "getField(Ljava/lang/String;Ljava/lang/Class;)Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamField) => void): void;
    public "java/io/ObjectStreamClass/getField(Ljava/lang/String;Ljava/lang/Class;)Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamField) => void): void;
    public "isProxy()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/isProxy()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEnum()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/isEnum()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isExternalizable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/isExternalizable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isSerializable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/isSerializable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasBlockExternalData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/hasBlockExternalData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasWriteObjectData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/hasWriteObjectData()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isInstantiable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/isInstantiable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasWriteObjectMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/hasWriteObjectMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasReadObjectMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/hasReadObjectMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasReadObjectNoDataMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/hasReadObjectNoDataMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasWriteReplaceMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/hasWriteReplaceMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hasReadResolveMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/hasReadResolveMethod()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "newInstance()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectStreamClass/newInstance()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invokeWriteObject(Ljava/lang/Object;Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/invokeWriteObject(Ljava/lang/Object;Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "invokeReadObject(Ljava/lang/Object;Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/invokeReadObject(Ljava/lang/Object;Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "invokeReadObjectNoData(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/invokeReadObjectNoData(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "invokeWriteReplace(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectStreamClass/invokeWriteReplace(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "invokeReadResolve(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectStreamClass/invokeReadResolve(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "getClassDataLayout()[Ljava/io/ObjectStreamClass$ClassDataSlot;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamClass$ClassDataSlot>) => void): void;
    public "java/io/ObjectStreamClass/getClassDataLayout()[Ljava/io/ObjectStreamClass$ClassDataSlot;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamClass$ClassDataSlot>) => void): void;
    public "getClassDataLayout0()[Ljava/io/ObjectStreamClass$ClassDataSlot;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamClass$ClassDataSlot>) => void): void;
    public "java/io/ObjectStreamClass/getClassDataLayout0()[Ljava/io/ObjectStreamClass$ClassDataSlot;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamClass$ClassDataSlot>) => void): void;
    public "getPrimDataSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/getPrimDataSize()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getNumObjFields()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectStreamClass/getNumObjFields()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getPrimFieldValues(Ljava/lang/Object;[B)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/getPrimFieldValues(Ljava/lang/Object;[B)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPrimFieldValues(Ljava/lang/Object;[B)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/setPrimFieldValues(Ljava/lang/Object;[B)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjFieldValues(Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/getObjFieldValues(Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setObjFieldValues(Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/setObjFieldValues(Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "computeFieldOffsets()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass/computeFieldOffsets()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getVariantFor(Ljava/lang/Class;)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public "java/io/ObjectStreamClass/getVariantFor(Ljava/lang/Class;)Ljava/io/ObjectStreamClass;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass) => void): void;
    public static "getExternalizableConstructor(Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "java/io/ObjectStreamClass/getExternalizableConstructor(Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "getSerializableConstructor(Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "java/io/ObjectStreamClass/getSerializableConstructor(Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "getInheritableMethod(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/getInheritableMethod(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "getPrivateMethod(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/getPrivateMethod(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "packageEquals(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectStreamClass/packageEquals(Ljava/lang/Class;Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getPackageName(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/ObjectStreamClass/getPackageName(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "classNamesEqual(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectStreamClass/classNamesEqual(Ljava/lang/String;Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getClassSignature(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/ObjectStreamClass/getClassSignature(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "getMethodSignature([Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/ObjectStreamClass/getMethodSignature([Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "throwMiscException(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectStreamClass/throwMiscException(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public static "getSerialFields(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "java/io/ObjectStreamClass/getSerialFields(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "getDeclaredSerialFields(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "java/io/ObjectStreamClass/getDeclaredSerialFields(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "getDefaultSerialFields(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "java/io/ObjectStreamClass/getDefaultSerialFields(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "getDeclaredSUID(Ljava/lang/Class;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/io/ObjectStreamClass/getDeclaredSUID(Ljava/lang/Class;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "computeDefaultSUID(Ljava/lang/Class;)J"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/io/ObjectStreamClass/computeDefaultSUID(Ljava/lang/Class;)J"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "hasStaticInitializer(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectStreamClass/hasStaticInitializer(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "getReflector([Ljava/io/ObjectStreamField;Ljava/io/ObjectStreamClass;)Ljava/io/ObjectStreamClass$FieldReflector;"(thread: JVMThread, args: [JVMArray<java_io_ObjectStreamField>, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass$FieldReflector) => void): void;
    public static "java/io/ObjectStreamClass/getReflector([Ljava/io/ObjectStreamField;Ljava/io/ObjectStreamClass;)Ljava/io/ObjectStreamClass$FieldReflector;"(thread: JVMThread, args: [JVMArray<java_io_ObjectStreamField>, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass$FieldReflector) => void): void;
    public static "matchFields([Ljava/io/ObjectStreamField;Ljava/io/ObjectStreamClass;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [JVMArray<java_io_ObjectStreamField>, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "java/io/ObjectStreamClass/matchFields([Ljava/io/ObjectStreamField;Ljava/io/ObjectStreamClass;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [JVMArray<java_io_ObjectStreamField>, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "processQueue(Ljava/lang/ref/ReferenceQueue;Ljava/util/concurrent/ConcurrentMap;)V"(thread: JVMThread, args: [java_lang_ref_ReferenceQueue, java_util_concurrent_ConcurrentMap], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectStreamClass/processQueue(Ljava/lang/ref/ReferenceQueue;Ljava/util/concurrent/ConcurrentMap;)V"(thread: JVMThread, args: [java_lang_ref_ReferenceQueue, java_util_concurrent_ConcurrentMap], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/io/ObjectStreamClass;)Ljava/lang/Class;"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "java/io/ObjectStreamClass/access$000(Ljava/io/ObjectStreamClass;)Ljava/lang/Class;"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: java_lang_Class) => void): void;
    public static "access$100(Ljava/lang/Class;)J"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "java/io/ObjectStreamClass/access$100(Ljava/lang/Class;)J"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public static "access$400(Ljava/io/ObjectStreamClass;)Z"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectStreamClass/access$400(Ljava/io/ObjectStreamClass;)Z"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$502(Ljava/io/ObjectStreamClass;Ljava/lang/Long;)Ljava/lang/Long;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_Long], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/io/ObjectStreamClass/access$502(Ljava/io/ObjectStreamClass;Ljava/lang/Long;)Ljava/lang/Long;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_Long], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "access$602(Ljava/io/ObjectStreamClass;[Ljava/io/ObjectStreamField;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_io_ObjectStreamClass, JVMArray<java_io_ObjectStreamField>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "java/io/ObjectStreamClass/access$602(Ljava/io/ObjectStreamClass;[Ljava/io/ObjectStreamField;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_io_ObjectStreamClass, JVMArray<java_io_ObjectStreamField>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "access$700(Ljava/lang/Class;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "java/io/ObjectStreamClass/access$700(Ljava/lang/Class;)Ljava/lang/Long;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_Long) => void): void;
    public static "access$800(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "java/io/ObjectStreamClass/access$800(Ljava/lang/Class;)[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public static "access$900(Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectStreamClass/access$900(Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$1002(Ljava/io/ObjectStreamClass;Ljava/io/ObjectStreamClass$ExceptionInfo;)Ljava/io/ObjectStreamClass$ExceptionInfo;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_io_ObjectStreamClass$ExceptionInfo], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass$ExceptionInfo) => void): void;
    public static "java/io/ObjectStreamClass/access$1002(Ljava/io/ObjectStreamClass;Ljava/io/ObjectStreamClass$ExceptionInfo;)Ljava/io/ObjectStreamClass$ExceptionInfo;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_io_ObjectStreamClass$ExceptionInfo], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass$ExceptionInfo) => void): void;
    public static "access$1102(Ljava/io/ObjectStreamClass;Ljava/io/ObjectStreamClass$ExceptionInfo;)Ljava/io/ObjectStreamClass$ExceptionInfo;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_io_ObjectStreamClass$ExceptionInfo], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass$ExceptionInfo) => void): void;
    public static "java/io/ObjectStreamClass/access$1102(Ljava/io/ObjectStreamClass;Ljava/io/ObjectStreamClass$ExceptionInfo;)Ljava/io/ObjectStreamClass$ExceptionInfo;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_io_ObjectStreamClass$ExceptionInfo], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectStreamClass$ExceptionInfo) => void): void;
    public static "access$1200(Ljava/io/ObjectStreamClass;)Z"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectStreamClass/access$1200(Ljava/io/ObjectStreamClass;)Z"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1302(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Constructor;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "java/io/ObjectStreamClass/access$1302(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Constructor;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Constructor], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "access$1400(Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "java/io/ObjectStreamClass/access$1400(Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "access$1500(Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "java/io/ObjectStreamClass/access$1500(Ljava/lang/Class;)Ljava/lang/reflect/Constructor;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Constructor) => void): void;
    public static "access$1602(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/access$1602(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "access$1700(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/access$1700(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "access$1802(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/access$1802(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "access$1902(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/access$1902(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "access$2002(Ljava/io/ObjectStreamClass;Z)Z"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectStreamClass/access$2002(Ljava/io/ObjectStreamClass;Z)Z"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$1600(Ljava/io/ObjectStreamClass;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/access$1600(Ljava/io/ObjectStreamClass;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "access$2102(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/access$2102(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "access$2200(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/access$2200(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_lang_Class, java_lang_String, JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "access$2302(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "java/io/ObjectStreamClass/access$2302(Ljava/io/ObjectStreamClass;Ljava/lang/reflect/Method;)Ljava/lang/reflect/Method;"(thread: JVMThread, args: [java_io_ObjectStreamClass, java_lang_reflect_Method], cb?: (e?: java_lang_Throwable, rv?: java_lang_reflect_Method) => void): void;
    public static "access$2400(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/ObjectStreamClass/access$2400(Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "access$2500([Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/ObjectStreamClass/access$2500([Ljava/lang/Class;Ljava/lang/Class;)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<java_lang_Class>, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectStreamClass/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectStreamClass$ClassDataSlot extends java_lang_Object {
    public "java/io/ObjectStreamClass$ClassDataSlot/desc": java_io_ObjectStreamClass;
    public "java/io/ObjectStreamClass$ClassDataSlot/hasData": number;
    public "<init>(Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass$ClassDataSlot/<init>(Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectStreamClass$FieldReflector extends java_lang_Object {
    public static "java/io/ObjectStreamClass$FieldReflector/unsafe": sun_misc_Unsafe;
    public "java/io/ObjectStreamClass$FieldReflector/fields": JVMArray<java_io_ObjectStreamField>;
    public "java/io/ObjectStreamClass$FieldReflector/numPrimFields": number;
    public "java/io/ObjectStreamClass$FieldReflector/readKeys": JVMArray<Long>;
    public "java/io/ObjectStreamClass$FieldReflector/writeKeys": JVMArray<Long>;
    public "java/io/ObjectStreamClass$FieldReflector/offsets": JVMArray<number>;
    public "java/io/ObjectStreamClass$FieldReflector/typeCodes": JVMArray<number>;
    public "java/io/ObjectStreamClass$FieldReflector/types": JVMArray<java_lang_Class>;
    public "<init>([Ljava/io/ObjectStreamField;)V"(thread: JVMThread, args: [JVMArray<java_io_ObjectStreamField>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass$FieldReflector/<init>([Ljava/io/ObjectStreamField;)V"(thread: JVMThread, args: [JVMArray<java_io_ObjectStreamField>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFields()[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public "java/io/ObjectStreamClass$FieldReflector/getFields()[Ljava/io/ObjectStreamField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_io_ObjectStreamField>) => void): void;
    public "getPrimFieldValues(Ljava/lang/Object;[B)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass$FieldReflector/getPrimFieldValues(Ljava/lang/Object;[B)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setPrimFieldValues(Ljava/lang/Object;[B)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass$FieldReflector/setPrimFieldValues(Ljava/lang/Object;[B)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getObjFieldValues(Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass$FieldReflector/getObjFieldValues(Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "setObjFieldValues(Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass$FieldReflector/setObjFieldValues(Ljava/lang/Object;[Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectStreamClass$FieldReflector/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectStreamClass$ExceptionInfo extends java_lang_Object {
    public "java/io/ObjectStreamClass$ExceptionInfo/className": java_lang_String;
    public "java/io/ObjectStreamClass$ExceptionInfo/message": java_lang_String;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamClass$ExceptionInfo/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "newInvalidClassException()Ljava/io/InvalidClassException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InvalidClassException) => void): void;
    public "java/io/ObjectStreamClass$ExceptionInfo/newInvalidClassException()Ljava/io/InvalidClassException;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_InvalidClassException) => void): void;
  }
  export class java_io_InvalidClassException extends java_io_ObjectStreamException {
    public static "java/io/InvalidClassException/serialVersionUID": Long;
    public "java/io/InvalidClassException/classname": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/InvalidClassException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/InvalidClassException/<init>(Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/InvalidClassException/getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_io_ObjectStreamException extends java_io_IOException {
    public static "java/io/ObjectStreamException/serialVersionUID": Long;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectStreamException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectOutputStream extends java_io_OutputStream implements java_io_ObjectOutput, java_io_ObjectStreamConstants {
    public "java/io/ObjectOutputStream/bout": java_io_ObjectOutputStream$BlockDataOutputStream;
    public "java/io/ObjectOutputStream/handles": java_io_ObjectOutputStream$HandleTable;
    public "java/io/ObjectOutputStream/subs": java_io_ObjectOutputStream$ReplaceTable;
    public "java/io/ObjectOutputStream/protocol": number;
    public "java/io/ObjectOutputStream/depth": number;
    public "java/io/ObjectOutputStream/primVals": JVMArray<number>;
    public "java/io/ObjectOutputStream/enableOverride": number;
    public "java/io/ObjectOutputStream/enableReplace": number;
    public "java/io/ObjectOutputStream/curContext": java_io_SerialCallbackContext;
    public "java/io/ObjectOutputStream/curPut": java_io_ObjectOutputStream$PutFieldImpl;
    public "java/io/ObjectOutputStream/debugInfoStack": java_io_ObjectOutputStream$DebugTraceInfoStack;
    public static "java/io/ObjectOutputStream/extendedDebugInfo": number;
    public "<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "useProtocolVersion(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/useProtocolVersion(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeObject(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObjectOverride(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeObjectOverride(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeUnshared(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeUnshared(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "defaultWriteObject()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/defaultWriteObject()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "putFields()Ljava/io/ObjectOutputStream$PutField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectOutputStream$PutField) => void): void;
    public "java/io/ObjectOutputStream/putFields()Ljava/io/ObjectOutputStream$PutField;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectOutputStream$PutField) => void): void;
    public "writeFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/reset()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "annotateClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/annotateClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "annotateProxyClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/annotateProxyClass(Ljava/lang/Class;)V"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable) => void): void;
    public "replaceObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectOutputStream/replaceObject(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "enableReplaceObject(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream/enableReplaceObject(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeStreamHeader()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeStreamHeader()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeClassDescriptor(Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeClassDescriptor(Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "drain()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/drain()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getProtocolVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream/getProtocolVersion()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeTypeString(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeTypeString(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "verifySubclass()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/verifySubclass()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "auditSubclass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectOutputStream/auditSubclass(Ljava/lang/Class;)Z"(thread: JVMThread, args: [java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject0(Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeObject0(Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeNull()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeNull()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeHandle(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeHandle(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeClass(Ljava/lang/Class;Z)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeClass(Ljava/lang/Class;Z)V"(thread: JVMThread, args: [java_lang_Class, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeClassDesc(Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeClassDesc(Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "isCustomSubclass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream/isCustomSubclass()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "writeProxyDesc(Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeProxyDesc(Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeNonProxyDesc(Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeNonProxyDesc(Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeString(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeString(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeArray(Ljava/lang/Object;Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeArray(Ljava/lang/Object;Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeEnum(Ljava/lang/Enum;Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_lang_Enum, java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeEnum(Ljava/lang/Enum;Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_lang_Enum, java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeOrdinaryObject(Ljava/lang/Object;Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeOrdinaryObject(Ljava/lang/Object;Ljava/io/ObjectStreamClass;Z)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeExternalData(Ljava/io/Externalizable;)V"(thread: JVMThread, args: [java_io_Externalizable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeExternalData(Ljava/io/Externalizable;)V"(thread: JVMThread, args: [java_io_Externalizable], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeSerialData(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeSerialData(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "defaultWriteFields(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/defaultWriteFields(Ljava/lang/Object;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_lang_Object, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeFatalException(Ljava/io/IOException;)V"(thread: JVMThread, args: [java_io_IOException], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream/writeFatalException(Ljava/io/IOException;)V"(thread: JVMThread, args: [java_io_IOException], cb?: (e?: java_lang_Throwable) => void): void;
    public static "floatsToBytes([FI[BII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectOutputStream/floatsToBytes([FI[BII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "doublesToBytes([DI[BII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectOutputStream/doublesToBytes([DI[BII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/io/ObjectOutputStream;)Ljava/io/ObjectOutputStream$BlockDataOutputStream;"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectOutputStream$BlockDataOutputStream) => void): void;
    public static "java/io/ObjectOutputStream/access$000(Ljava/io/ObjectOutputStream;)Ljava/io/ObjectOutputStream$BlockDataOutputStream;"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectOutputStream$BlockDataOutputStream) => void): void;
    public static "access$100()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/ObjectOutputStream/access$100()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$200(Ljava/io/ObjectOutputStream;)Ljava/io/ObjectOutputStream$DebugTraceInfoStack;"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectOutputStream$DebugTraceInfoStack) => void): void;
    public static "java/io/ObjectOutputStream/access$200(Ljava/io/ObjectOutputStream;)Ljava/io/ObjectOutputStream$DebugTraceInfoStack;"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable, rv?: java_io_ObjectOutputStream$DebugTraceInfoStack) => void): void;
    public static "access$300(Ljava/io/ObjectOutputStream;Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_io_ObjectOutputStream, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectOutputStream/access$300(Ljava/io/ObjectOutputStream;Ljava/lang/Object;Z)V"(thread: JVMThread, args: [java_io_ObjectOutputStream, java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$400([FI[BII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectOutputStream/access$400([FI[BII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$500([DI[BII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectOutputStream/access$500([DI[BII)V"(thread: JVMThread, args: [JVMArray<number>, number, JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/ObjectOutputStream/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectOutputStream$PutField extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/put(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;B)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/put(Ljava/lang/String;B)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;C)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/put(Ljava/lang/String;C)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;S)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/put(Ljava/lang/String;S)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/put(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/put(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;F)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/put(Ljava/lang/String;F)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;D)V"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/put(Ljava/lang/String;D)V"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/put(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutField/write(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectOutputStream$DebugTraceInfoStack extends java_lang_Object {
    public "java/io/ObjectOutputStream$DebugTraceInfoStack/stack": java_util_List;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$DebugTraceInfoStack/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$DebugTraceInfoStack/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "pop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$DebugTraceInfoStack/pop()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "push(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$DebugTraceInfoStack/push(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/ObjectOutputStream$DebugTraceInfoStack/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_io_ObjectOutputStream$PutFieldImpl extends java_io_ObjectOutputStream$PutField {
    public "java/io/ObjectOutputStream$PutFieldImpl/desc": java_io_ObjectStreamClass;
    public "java/io/ObjectOutputStream$PutFieldImpl/primVals": JVMArray<number>;
    public "java/io/ObjectOutputStream$PutFieldImpl/objVals": JVMArray<java_lang_Object>;
    public "java/io/ObjectOutputStream$PutFieldImpl/this$0": java_io_ObjectOutputStream;
    public "<init>(Ljava/io/ObjectOutputStream;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/<init>(Ljava/io/ObjectOutputStream;Ljava/io/ObjectStreamClass;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream, java_io_ObjectStreamClass], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/put(Ljava/lang/String;Z)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;B)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/put(Ljava/lang/String;B)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;C)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/put(Ljava/lang/String;C)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;S)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/put(Ljava/lang/String;S)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/put(Ljava/lang/String;I)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;F)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/put(Ljava/lang/String;F)V"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/put(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;D)V"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/put(Ljava/lang/String;D)V"(thread: JVMThread, args: [java_lang_String, number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "put(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/put(Ljava/lang/String;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "write(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/write(Ljava/io/ObjectOutput;)V"(thread: JVMThread, args: [java_io_ObjectOutput], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/writeFields()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFieldOffset(Ljava/lang/String;Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream$PutFieldImpl/getFieldOffset(Ljava/lang/String;Ljava/lang/Class;)I"(thread: JVMThread, args: [java_lang_String, java_lang_Class], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_io_ObjectOutputStream$ReplaceTable extends java_lang_Object {
    public "java/io/ObjectOutputStream$ReplaceTable/htab": java_io_ObjectOutputStream$HandleTable;
    public "java/io/ObjectOutputStream$ReplaceTable/reps": JVMArray<java_lang_Object>;
    public "<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$ReplaceTable/<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "assign(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$ReplaceTable/assign(Ljava/lang/Object;Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "lookup(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/io/ObjectOutputStream$ReplaceTable/lookup(Ljava/lang/Object;)Ljava/lang/Object;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$ReplaceTable/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream$ReplaceTable/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "grow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$ReplaceTable/grow()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_ObjectOutputStream$HandleTable extends java_lang_Object {
    public "java/io/ObjectOutputStream$HandleTable/size": number;
    public "java/io/ObjectOutputStream$HandleTable/threshold": number;
    public "java/io/ObjectOutputStream$HandleTable/loadFactor": number;
    public "java/io/ObjectOutputStream$HandleTable/spine": JVMArray<number>;
    public "java/io/ObjectOutputStream$HandleTable/next": JVMArray<number>;
    public "java/io/ObjectOutputStream$HandleTable/objs": JVMArray<java_lang_Object>;
    public "<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$HandleTable/<init>(IF)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "assign(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream$HandleTable/assign(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lookup(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream$HandleTable/lookup(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$HandleTable/clear()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream$HandleTable/size()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "insert(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$HandleTable/insert(Ljava/lang/Object;I)V"(thread: JVMThread, args: [java_lang_Object, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "growSpine()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$HandleTable/growSpine()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "growEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$HandleTable/growEntries()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "hash(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream$HandleTable/hash(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
  }
  export class java_io_ObjectOutputStream$BlockDataOutputStream extends java_io_OutputStream implements java_io_DataOutput {
    public static "java/io/ObjectOutputStream$BlockDataOutputStream/MAX_BLOCK_SIZE": number;
    public static "java/io/ObjectOutputStream$BlockDataOutputStream/MAX_HEADER_SIZE": number;
    public static "java/io/ObjectOutputStream$BlockDataOutputStream/CHAR_BUF_SIZE": number;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/buf": JVMArray<number>;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/hbuf": JVMArray<number>;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/cbuf": JVMArray<number>;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/blkmode": number;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/pos": number;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/out": java_io_OutputStream;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/dout": java_io_DataOutputStream;
    public "<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/<init>(Ljava/io/OutputStream;)V"(thread: JVMThread, args: [java_io_OutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "setBlockDataMode(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/setBlockDataMode(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getBlockDataMode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/getBlockDataMode()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/write(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/write([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/write([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/close()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "write([BIIZ)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/write([BIIZ)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "drain()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/drain()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBlockHeader(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeBlockHeader(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeBoolean(Z)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeByte(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeChar(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeShort(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeInt(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeFloat(F)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeLong(J)V"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeDouble(D)V"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeBytes(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeChars(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeBooleans([ZII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeBooleans([ZII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeChars([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeChars([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeShorts([SII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeShorts([SII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeInts([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeInts([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeFloats([FII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeFloats([FII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeLongs([JII)V"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeLongs([JII)V"(thread: JVMThread, args: [JVMArray<Long>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeDoubles([DII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeDoubles([DII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getUTFLength(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/getUTFLength(Ljava/lang/String;)J"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "writeUTF(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeUTF(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeLongUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeLongUTF(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeLongUTF(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeLongUTF(Ljava/lang/String;J)V"(thread: JVMThread, args: [java_lang_String, Long, any], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeUTFBody(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/ObjectOutputStream$BlockDataOutputStream/writeUTFBody(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_Console extends java_lang_Object implements java_io_Flushable {
    public "java/io/Console/readLock": java_lang_Object;
    public "java/io/Console/writeLock": java_lang_Object;
    public "java/io/Console/reader": java_io_Reader;
    public "java/io/Console/out": java_io_Writer;
    public "java/io/Console/pw": java_io_PrintWriter;
    public "java/io/Console/formatter": java_util_Formatter;
    public "java/io/Console/cs": java_nio_charset_Charset;
    public "java/io/Console/rcb": JVMArray<number>;
    public static "java/io/Console/echoOff": number;
    public static "java/io/Console/cons": java_io_Console;
    public static "java/io/Console/$assertionsDisabled": number;
    public "writer()Ljava/io/PrintWriter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "java/io/Console/writer()Ljava/io/PrintWriter;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_PrintWriter) => void): void;
    public "reader()Ljava/io/Reader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_Reader) => void): void;
    public "java/io/Console/reader()Ljava/io/Reader;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_Reader) => void): void;
    public "format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/Console;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public "java/io/Console/format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/Console;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public "printf(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/Console;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public "java/io/Console/printf(Ljava/lang/String;[Ljava/lang/Object;)Ljava/io/Console;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public "readLine(Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/Console/readLine(Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/io/Console/readLine()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "readPassword(Ljava/lang/String;[Ljava/lang/Object;)[C"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/io/Console/readPassword(Ljava/lang/String;[Ljava/lang/Object;)[C"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "readPassword()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/io/Console/readPassword()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Console/flush()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "encoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/io/Console/encoding()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "echo(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/Console/echo(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "readline(Z)[C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/io/Console/readline(Z)[C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "grow()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/io/Console/grow()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "istty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/Console/istty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Console/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$000(Ljava/io/Console;)Ljava/lang/Object;"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "java/io/Console/access$000(Ljava/io/Console;)Ljava/lang/Object;"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public static "access$100(Ljava/io/Console;)[C"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/io/Console/access$100(Ljava/io/Console;)[C"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$200(Ljava/io/Console;)[C"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "java/io/Console/access$200(Ljava/io/Console;)[C"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "access$300()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/Console/access$300()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$400(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/Console/access$400(Z)Z"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$500()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/io/Console/access$500()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "access$600()Ljava/io/Console;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public static "java/io/Console/access$600()Ljava/io/Console;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public static "access$602(Ljava/io/Console;)Ljava/io/Console;"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public static "java/io/Console/access$602(Ljava/io/Console;)Ljava/io/Console;"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: java_io_Console) => void): void;
    public "<init>(Ljava/io/Console$1;)V"(thread: JVMThread, args: [java_io_Console$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Console/<init>(Ljava/io/Console$1;)V"(thread: JVMThread, args: [java_io_Console$1], cb?: (e?: java_lang_Throwable) => void): void;
    public static "access$800(Ljava/io/Console;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "java/io/Console/access$800(Ljava/io/Console;)Ljava/nio/charset/Charset;"(thread: JVMThread, args: [java_io_Console], cb?: (e?: java_lang_Throwable, rv?: java_nio_charset_Charset) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/io/Console/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_io_Console$1 extends java_lang_Object implements java_lang_Runnable {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Console$1/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/io/Console$1/run()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class doppio_security_NodePRNG extends java_security_SecureRandomSpi {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/security/NodePRNG/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isAvailable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "doppio/security/NodePRNG/isAvailable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "engineSetSeed([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/security/NodePRNG/engineSetSeed([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineNextBytes([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/security/NodePRNG/engineNextBytes([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineGenerateSeed(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "doppio/security/NodePRNG/engineGenerateSeed(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export class java_security_SecureRandomSpi extends java_lang_Object implements java_io_Serializable {
    public static "java/security/SecureRandomSpi/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/SecureRandomSpi/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineSetSeed([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/SecureRandomSpi/engineSetSeed([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineNextBytes([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/SecureRandomSpi/engineNextBytes([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineGenerateSeed(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/security/SecureRandomSpi/engineGenerateSeed(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export class doppio_security_BrowserPRNG extends java_security_SecureRandomSpi {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/security/BrowserPRNG/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "isAvailable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "doppio/security/BrowserPRNG/isAvailable()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "engineSetSeed([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/security/BrowserPRNG/engineSetSeed([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineNextBytes([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/security/BrowserPRNG/engineNextBytes([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "engineGenerateSeed(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "doppio/security/BrowserPRNG/engineGenerateSeed(I)[B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
  }
  export class java_lang_String extends java_lang_Object implements java_io_Serializable, java_lang_Comparable, java_lang_CharSequence {
    public toString(): string;
    public "java/lang/String/value": JVMArray<number>;
    public "java/lang/String/hash": number;
    public static "java/lang/String/serialVersionUID": Long;
    public static "java/lang/String/serialPersistentFields": JVMArray<java_io_ObjectStreamField>;
    public static "java/lang/String/CASE_INSENSITIVE_ORDER": java_util_Comparator;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([C)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([CII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([III)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([BIII)V"(thread: JVMThread, args: [JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([BI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "checkBounds([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/String/checkBounds([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BIILjava/lang/String;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([BIILjava/lang/String;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BIILjava/nio/charset/Charset;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([BIILjava/nio/charset/Charset;)V"(thread: JVMThread, args: [JVMArray<number>, number, number, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BLjava/lang/String;)V"(thread: JVMThread, args: [JVMArray<number>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([BLjava/lang/String;)V"(thread: JVMThread, args: [JVMArray<number>, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BLjava/nio/charset/Charset;)V"(thread: JVMThread, args: [JVMArray<number>, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([BLjava/nio/charset/Charset;)V"(thread: JVMThread, args: [JVMArray<number>, java_nio_charset_Charset], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([BII)V"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([B)V"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/StringBuffer;)V"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>(Ljava/lang/StringBuffer;)V"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>(Ljava/lang/StringBuilder;)V"(thread: JVMThread, args: [java_lang_StringBuilder], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>([CZ)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/<init>([CZ)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/length()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/isEmpty()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/charAt(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/codePointAt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointBefore(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/codePointBefore(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "codePointCount(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/codePointCount(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "offsetByCodePoints(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/offsetByCodePoints(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getChars([CI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/getChars([CI)V"(thread: JVMThread, args: [JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChars(II[CI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/getChars(II[CI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBytes(II[BI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/String/getBytes(II[BI)V"(thread: JVMThread, args: [number, number, JVMArray<number>, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getBytes(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/String/getBytes(Ljava/lang/String;)[B"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getBytes(Ljava/nio/charset/Charset;)[B"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/String/getBytes(Ljava/nio/charset/Charset;)[B"(thread: JVMThread, args: [java_nio_charset_Charset], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "getBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/String/getBytes()[B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/equals(Ljava/lang/Object;)Z"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contentEquals(Ljava/lang/StringBuffer;)Z"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/contentEquals(Ljava/lang/StringBuffer;)Z"(thread: JVMThread, args: [java_lang_StringBuffer], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "nonSyncContentEquals(Ljava/lang/AbstractStringBuilder;)Z"(thread: JVMThread, args: [java_lang_AbstractStringBuilder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/nonSyncContentEquals(Ljava/lang/AbstractStringBuilder;)Z"(thread: JVMThread, args: [java_lang_AbstractStringBuilder], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contentEquals(Ljava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/contentEquals(Ljava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "equalsIgnoreCase(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/equalsIgnoreCase(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareTo(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/compareTo(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "compareToIgnoreCase(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/compareToIgnoreCase(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "regionMatches(ILjava/lang/String;II)Z"(thread: JVMThread, args: [number, java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/regionMatches(ILjava/lang/String;II)Z"(thread: JVMThread, args: [number, java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "regionMatches(ZILjava/lang/String;II)Z"(thread: JVMThread, args: [number, number, java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/regionMatches(ZILjava/lang/String;II)Z"(thread: JVMThread, args: [number, number, java_lang_String, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "startsWith(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/startsWith(Ljava/lang/String;I)Z"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "startsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/startsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "endsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/endsWith(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/hashCode()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/indexOf(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/indexOf(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOfSupplementary(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/indexOfSupplementary(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/lastIndexOf(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/lastIndexOf(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOfSupplementary(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/lastIndexOfSupplementary(II)I"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/indexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "indexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/indexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "indexOf([CIILjava/lang/String;I)I"(thread: JVMThread, args: [JVMArray<number>, number, number, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/String/indexOf([CIILjava/lang/String;I)I"(thread: JVMThread, args: [JVMArray<number>, number, number, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "indexOf([CII[CIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/String/indexOf([CII[CIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/lastIndexOf(Ljava/lang/String;)I"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "lastIndexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/lastIndexOf(Ljava/lang/String;I)I"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "lastIndexOf([CIILjava/lang/String;I)I"(thread: JVMThread, args: [JVMArray<number>, number, number, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/String/lastIndexOf([CIILjava/lang/String;I)I"(thread: JVMThread, args: [JVMArray<number>, number, number, java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "lastIndexOf([CII[CIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "java/lang/String/lastIndexOf([CII[CIII)I"(thread: JVMThread, args: [JVMArray<number>, number, number, JVMArray<number>, number, number, number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "substring(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/substring(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "substring(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/substring(II)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "java/lang/String/subSequence(II)Ljava/lang/CharSequence;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_CharSequence) => void): void;
    public "concat(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/concat(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "replace(CC)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/replace(CC)Ljava/lang/String;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "matches(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/matches(Ljava/lang/String;)Z"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "contains(Ljava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/contains(Ljava/lang/CharSequence;)Z"(thread: JVMThread, args: [java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "replaceFirst(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/replaceFirst(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "replaceAll(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/replaceAll(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "replace(Ljava/lang/CharSequence;Ljava/lang/CharSequence;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/replace(Ljava/lang/CharSequence;Ljava/lang/CharSequence;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_CharSequence], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "split(Ljava/lang/String;I)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/lang/String/split(Ljava/lang/String;I)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, number], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "split(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public "java/lang/String/split(Ljava/lang/String;)[Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_String>) => void): void;
    public static "join(Ljava/lang/CharSequence;[Ljava/lang/CharSequence;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence, JVMArray<java_lang_CharSequence>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/join(Ljava/lang/CharSequence;[Ljava/lang/CharSequence;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence, JVMArray<java_lang_CharSequence>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "join(Ljava/lang/CharSequence;Ljava/lang/Iterable;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_Iterable], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/join(Ljava/lang/CharSequence;Ljava/lang/Iterable;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_CharSequence, java_lang_Iterable], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toLowerCase(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/toLowerCase(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toLowerCase()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/toLowerCase()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toUpperCase(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/toUpperCase(Ljava/util/Locale;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toUpperCase()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/toUpperCase()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "trim()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/trim()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "toCharArray()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public "java/lang/String/toCharArray()[C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<number>) => void): void;
    public static "format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/format(Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "format(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/format(Ljava/util/Locale;Ljava/lang/String;[Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_util_Locale, java_lang_String, JVMArray<java_lang_Object>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf(Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/valueOf(Ljava/lang/Object;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf([C)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/valueOf([C)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf([CII)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/valueOf([CII)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "copyValueOf([CII)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/copyValueOf([CII)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "copyValueOf([C)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/copyValueOf([C)Ljava/lang/String;"(thread: JVMThread, args: [JVMArray<number>], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/valueOf(Z)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf(C)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/valueOf(C)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/valueOf(I)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/valueOf(J)Ljava/lang/String;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf(F)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/valueOf(F)Ljava/lang/String;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "valueOf(D)Ljava/lang/String;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "java/lang/String/valueOf(D)Ljava/lang/String;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "intern()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/String/intern()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/String/compareTo(Ljava/lang/Object;)I"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/String/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    "chars()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "codePoints()Ljava/util/stream/IntStream;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_stream_IntStream) => void): void;
    "lambda$codePoints$1()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
    "lambda$chars$0()Ljava/util/Spliterator$OfInt;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_Spliterator$OfInt) => void): void;
  }
  export class doppio_JavaScript extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/JavaScript/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "eval(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "doppio/JavaScript/eval(Ljava/lang/String;)Ljava/lang/String;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class doppio_Debug$LogLevel extends java_lang_Enum {
    public static "doppio/Debug$LogLevel/ERROR": doppio_Debug$LogLevel;
    public static "doppio/Debug$LogLevel/DEBUG": doppio_Debug$LogLevel;
    public static "doppio/Debug$LogLevel/TRACE": doppio_Debug$LogLevel;
    public static "doppio/Debug$LogLevel/VTRACE": doppio_Debug$LogLevel;
    public "doppio/Debug$LogLevel/level": number;
    public "doppio/Debug$LogLevel/name": java_lang_String;
    public static "doppio/Debug$LogLevel/$VALUES": JVMArray<doppio_Debug$LogLevel>;
    public static "values()[Ldoppio/Debug$LogLevel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<doppio_Debug$LogLevel>) => void): void;
    public static "doppio/Debug$LogLevel/values()[Ldoppio/Debug$LogLevel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<doppio_Debug$LogLevel>) => void): void;
    public static "valueOf(Ljava/lang/String;)Ldoppio/Debug$LogLevel;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: doppio_Debug$LogLevel) => void): void;
    public static "doppio/Debug$LogLevel/valueOf(Ljava/lang/String;)Ldoppio/Debug$LogLevel;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: doppio_Debug$LogLevel) => void): void;
    public "<init>(Ljava/lang/String;IILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/Debug$LogLevel/<init>(Ljava/lang/String;IILjava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, number, number, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "doppio/Debug$LogLevel/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "doppio/Debug$LogLevel/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class doppio_Debug extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/Debug/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "SetLogLevel(Ldoppio/Debug$LogLevel;)V"(thread: JVMThread, args: [doppio_Debug$LogLevel], cb?: (e?: java_lang_Throwable) => void): void;
    public static "doppio/Debug/SetLogLevel(Ldoppio/Debug$LogLevel;)V"(thread: JVMThread, args: [doppio_Debug$LogLevel], cb?: (e?: java_lang_Throwable) => void): void;
    public static "GetLogLevel()Ldoppio/Debug$LogLevel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: doppio_Debug$LogLevel) => void): void;
    public static "doppio/Debug/GetLogLevel()Ldoppio/Debug$LogLevel;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: doppio_Debug$LogLevel) => void): void;
  }
  export class doppio_security_DoppioProvider extends java_security_Provider {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "doppio/security/DoppioProvider/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "doppio/security/DoppioProvider/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_ProviderException extends java_lang_RuntimeException {
    public static "java/security/ProviderException/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/ProviderException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/ProviderException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/ProviderException/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/ProviderException/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_security_PrivilegedActionException extends java_lang_Exception {
    public static "java/security/PrivilegedActionException/serialVersionUID": Long;
    public "java/security/PrivilegedActionException/exception": java_lang_Exception;
    public "<init>(Ljava/lang/Exception;)V"(thread: JVMThread, args: [java_lang_Exception], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/security/PrivilegedActionException/<init>(Ljava/lang/Exception;)V"(thread: JVMThread, args: [java_lang_Exception], cb?: (e?: java_lang_Throwable) => void): void;
    public "getException()Ljava/lang/Exception;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Exception) => void): void;
    public "java/security/PrivilegedActionException/getException()Ljava/lang/Exception;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Exception) => void): void;
    public "getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/security/PrivilegedActionException/getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/security/PrivilegedActionException/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_nio_DirectByteBuffer extends java_nio_MappedByteBuffer implements sun_nio_ch_DirectBuffer {
    public static "java/nio/DirectByteBuffer/unsafe": sun_misc_Unsafe;
    public static "java/nio/DirectByteBuffer/arrayBaseOffset": Long;
    public static "java/nio/DirectByteBuffer/unaligned": number;
    public "java/nio/DirectByteBuffer/att": java_lang_Object;
    public "java/nio/DirectByteBuffer/cleaner": sun_misc_Cleaner;
    public static "java/nio/DirectByteBuffer/$assertionsDisabled": number;
    public "attachment()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/DirectByteBuffer/attachment()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "cleaner()Lsun/misc/Cleaner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Cleaner) => void): void;
    public "java/nio/DirectByteBuffer/cleaner()Lsun/misc/Cleaner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Cleaner) => void): void;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/DirectByteBuffer/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JILjava/lang/Object;)V"(thread: JVMThread, args: [Long, any, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/DirectByteBuffer/<init>(JILjava/lang/Object;)V"(thread: JVMThread, args: [Long, any, number, java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/DirectByteBuffer/<init>(JI)V"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(IJLjava/io/FileDescriptor;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [number, Long, any, java_io_FileDescriptor, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/DirectByteBuffer/<init>(IJLjava/io/FileDescriptor;Ljava/lang/Runnable;)V"(thread: JVMThread, args: [number, Long, any, java_io_FileDescriptor, java_lang_Runnable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Lsun/nio/ch/DirectBuffer;IIIII)V"(thread: JVMThread, args: [sun_nio_ch_DirectBuffer, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/DirectByteBuffer/<init>(Lsun/nio/ch/DirectBuffer;IIIII)V"(thread: JVMThread, args: [sun_nio_ch_DirectBuffer, number, number, number, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "slice()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/slice()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "duplicate()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/duplicate()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asReadOnlyBuffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/asReadOnlyBuffer()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "address()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/DirectByteBuffer/address()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "ix(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/DirectByteBuffer/ix(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "get()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/get()B"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/get(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "get([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/get([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "put(B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/put(B)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "put(IB)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/put(IB)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "put(Ljava/nio/ByteBuffer;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/put(Ljava/nio/ByteBuffer;)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [java_nio_ByteBuffer], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "put([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/put([BII)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [JVMArray<number>, number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "compact()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/compact()Ljava/nio/ByteBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/isDirect()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/isReadOnly()Z"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "_get(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/_get(I)B"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "_put(IB)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/DirectByteBuffer/_put(IB)V"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getChar(J)C"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getChar(J)C"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getChar()C"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getChar(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getChar(I)C"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putChar(JC)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putChar(JC)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putChar(C)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putChar(C)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putChar(IC)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putChar(IC)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asCharBuffer()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "java/nio/DirectByteBuffer/asCharBuffer()Ljava/nio/CharBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_CharBuffer) => void): void;
    public "getShort(J)S"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getShort(J)S"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getShort()S"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getShort(I)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getShort(I)S"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putShort(JS)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putShort(JS)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putShort(S)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putShort(S)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putShort(IS)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putShort(IS)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asShortBuffer()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "java/nio/DirectByteBuffer/asShortBuffer()Ljava/nio/ShortBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_ShortBuffer) => void): void;
    public "getInt(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getInt(J)I"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getInt()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getInt(I)I"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putInt(JI)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putInt(JI)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putInt(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putInt(I)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putInt(II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putInt(II)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asIntBuffer()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "java/nio/DirectByteBuffer/asIntBuffer()Ljava/nio/IntBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_IntBuffer) => void): void;
    public "getLong(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/DirectByteBuffer/getLong(J)J"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/DirectByteBuffer/getLong()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "getLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "java/nio/DirectByteBuffer/getLong(I)J"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    public "putLong(JJ)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putLong(JJ)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putLong(J)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putLong(J)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putLong(IJ)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putLong(IJ)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, Long, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asLongBuffer()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "java/nio/DirectByteBuffer/asLongBuffer()Ljava/nio/LongBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_LongBuffer) => void): void;
    public "getFloat(J)F"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getFloat(J)F"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getFloat()F"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getFloat(I)F"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putFloat(JF)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putFloat(JF)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putFloat(F)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putFloat(F)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putFloat(IF)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putFloat(IF)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asFloatBuffer()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "java/nio/DirectByteBuffer/asFloatBuffer()Ljava/nio/FloatBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_FloatBuffer) => void): void;
    public "getDouble(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getDouble(J)D"(thread: JVMThread, args: [Long, any], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getDouble()D"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getDouble(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/nio/DirectByteBuffer/getDouble(I)D"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "putDouble(JD)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putDouble(JD)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [Long, any, number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putDouble(D)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putDouble(D)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "putDouble(ID)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "java/nio/DirectByteBuffer/putDouble(ID)Ljava/nio/ByteBuffer;"(thread: JVMThread, args: [number, number, any], cb?: (e?: java_lang_Throwable, rv?: java_nio_ByteBuffer) => void): void;
    public "asDoubleBuffer()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public "java/nio/DirectByteBuffer/asDoubleBuffer()Ljava/nio/DoubleBuffer;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_DoubleBuffer) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/nio/DirectByteBuffer/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export interface sun_nio_ch_DirectBuffer extends java_lang_Object {
    "address()J"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: Long) => void): void;
    "attachment()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    "cleaner()Lsun/misc/Cleaner;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: sun_misc_Cleaner) => void): void;
  }
  export class java_lang_reflect_InvocationTargetException extends java_lang_ReflectiveOperationException {
    public static "java/lang/reflect/InvocationTargetException/serialVersionUID": Long;
    public "java/lang/reflect/InvocationTargetException/target": java_lang_Throwable;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/InvocationTargetException/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/InvocationTargetException/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Throwable, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/InvocationTargetException/<init>(Ljava/lang/Throwable;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_Throwable, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getTargetException()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/InvocationTargetException/getTargetException()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/reflect/InvocationTargetException/getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
  }
  export class java_lang_invoke_MethodHandleNatives$Constants extends java_lang_Object {
    public static "java/lang/invoke/MethodHandleNatives$Constants/GC_COUNT_GWT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/GC_LAMBDA_SUPPORT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/MN_IS_METHOD": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/MN_IS_CONSTRUCTOR": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/MN_IS_FIELD": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/MN_IS_TYPE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/MN_CALLER_SENSITIVE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/MN_REFERENCE_KIND_SHIFT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/MN_REFERENCE_KIND_MASK": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/MN_SEARCH_SUPERCLASSES": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/MN_SEARCH_INTERFACES": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_BOOLEAN": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_CHAR": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_FLOAT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_DOUBLE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_BYTE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_SHORT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_INT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_LONG": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_OBJECT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_VOID": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/T_ILLEGAL": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_Utf8": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_Integer": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_Float": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_Long": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_Double": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_Class": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_String": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_Fieldref": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_Methodref": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_InterfaceMethodref": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_NameAndType": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_MethodHandle": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_MethodType": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_InvokeDynamic": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/CONSTANT_LIMIT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_PUBLIC": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_PRIVATE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_PROTECTED": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_STATIC": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_FINAL": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_SYNCHRONIZED": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_VOLATILE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_TRANSIENT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_NATIVE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_INTERFACE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_ABSTRACT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_STRICT": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_SYNTHETIC": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_ANNOTATION": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_ENUM": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_SUPER": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_BRIDGE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/ACC_VARARGS": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_NONE": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_getField": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_getStatic": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_putField": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_putStatic": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_invokeVirtual": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_invokeStatic": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_invokeSpecial": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_newInvokeSpecial": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_invokeInterface": number;
    public static "java/lang/invoke/MethodHandleNatives$Constants/REF_LIMIT": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/invoke/MethodHandleNatives$Constants/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_charset_Charset$3 extends java_lang_Object implements java_security_PrivilegedAction {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/charset/Charset$3/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "run()Ljava/util/SortedMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public "java/nio/charset/Charset$3/run()Ljava/util/SortedMap;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_util_SortedMap) => void): void;
    public "run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/nio/charset/Charset$3/run()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
  }
  export class java_lang_ExceptionInInitializerError extends java_lang_LinkageError {
    public static "java/lang/ExceptionInInitializerError/serialVersionUID": Long;
    public "java/lang/ExceptionInInitializerError/exception": java_lang_Throwable;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ExceptionInInitializerError/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ExceptionInInitializerError/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/ExceptionInInitializerError/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getException()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/ExceptionInInitializerError/getException()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/ExceptionInInitializerError/getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
  }
  export class java_lang_LinkageError extends java_lang_Error {
    public static "java/lang/LinkageError/serialVersionUID": Long;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/LinkageError/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/LinkageError/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/LinkageError/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class sun_nio_fs_UnixException extends java_lang_Exception {
    public static "sun/nio/fs/UnixException/serialVersionUID": Long;
    public "sun/nio/fs/UnixException/errno": number;
    public "sun/nio/fs/UnixException/msg": java_lang_String;
    public "<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixException/<init>(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "errno()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "sun/nio/fs/UnixException/errno()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "setError(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixException/setError(I)V"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable) => void): void;
    public "errorString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixException/errorString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "sun/nio/fs/UnixException/getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "translateToIOException(Ljava/lang/String;Ljava/lang/String;)Ljava/io/IOException;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_IOException) => void): void;
    public "sun/nio/fs/UnixException/translateToIOException(Ljava/lang/String;Ljava/lang/String;)Ljava/io/IOException;"(thread: JVMThread, args: [java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_io_IOException) => void): void;
    public "rethrowAsIOException(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixException/rethrowAsIOException(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "rethrowAsIOException(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixException/rethrowAsIOException(Lsun/nio/fs/UnixPath;Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath, sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public "rethrowAsIOException(Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixException/rethrowAsIOException(Lsun/nio/fs/UnixPath;)V"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable) => void): void;
    public "asIOException(Lsun/nio/fs/UnixPath;)Ljava/io/IOException;"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: java_io_IOException) => void): void;
    public "sun/nio/fs/UnixException/asIOException(Lsun/nio/fs/UnixPath;)Ljava/io/IOException;"(thread: JVMThread, args: [sun_nio_fs_UnixPath], cb?: (e?: java_lang_Throwable, rv?: java_io_IOException) => void): void;
  }
  export class sun_nio_fs_DefaultFileSystemProvider extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/DefaultFileSystemProvider/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "createProvider(Ljava/lang/String;)Ljava/nio/file/spi/FileSystemProvider;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileSystemProvider) => void): void;
    public static "sun/nio/fs/DefaultFileSystemProvider/createProvider(Ljava/lang/String;)Ljava/nio/file/spi/FileSystemProvider;"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileSystemProvider) => void): void;
    public static "create()Ljava/nio/file/spi/FileSystemProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileSystemProvider) => void): void;
    public static "sun/nio/fs/DefaultFileSystemProvider/create()Ljava/nio/file/spi/FileSystemProvider;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_nio_file_spi_FileSystemProvider) => void): void;
  }
  export class sun_nio_fs_UnixConstants extends java_lang_Object {
    public static "sun/nio/fs/UnixConstants/O_RDONLY": number;
    public static "sun/nio/fs/UnixConstants/O_WRONLY": number;
    public static "sun/nio/fs/UnixConstants/O_RDWR": number;
    public static "sun/nio/fs/UnixConstants/O_APPEND": number;
    public static "sun/nio/fs/UnixConstants/O_CREAT": number;
    public static "sun/nio/fs/UnixConstants/O_EXCL": number;
    public static "sun/nio/fs/UnixConstants/O_TRUNC": number;
    public static "sun/nio/fs/UnixConstants/O_SYNC": number;
    public static "sun/nio/fs/UnixConstants/O_DSYNC": number;
    public static "sun/nio/fs/UnixConstants/O_NOFOLLOW": number;
    public static "sun/nio/fs/UnixConstants/S_IAMB": number;
    public static "sun/nio/fs/UnixConstants/S_IRUSR": number;
    public static "sun/nio/fs/UnixConstants/S_IWUSR": number;
    public static "sun/nio/fs/UnixConstants/S_IXUSR": number;
    public static "sun/nio/fs/UnixConstants/S_IRGRP": number;
    public static "sun/nio/fs/UnixConstants/S_IWGRP": number;
    public static "sun/nio/fs/UnixConstants/S_IXGRP": number;
    public static "sun/nio/fs/UnixConstants/S_IROTH": number;
    public static "sun/nio/fs/UnixConstants/S_IWOTH": number;
    public static "sun/nio/fs/UnixConstants/S_IXOTH": number;
    public static "sun/nio/fs/UnixConstants/S_IFMT": number;
    public static "sun/nio/fs/UnixConstants/S_IFREG": number;
    public static "sun/nio/fs/UnixConstants/S_IFDIR": number;
    public static "sun/nio/fs/UnixConstants/S_IFLNK": number;
    public static "sun/nio/fs/UnixConstants/S_IFCHR": number;
    public static "sun/nio/fs/UnixConstants/S_IFBLK": number;
    public static "sun/nio/fs/UnixConstants/S_IFIFO": number;
    public static "sun/nio/fs/UnixConstants/R_OK": number;
    public static "sun/nio/fs/UnixConstants/W_OK": number;
    public static "sun/nio/fs/UnixConstants/X_OK": number;
    public static "sun/nio/fs/UnixConstants/F_OK": number;
    public static "sun/nio/fs/UnixConstants/ENOENT": number;
    public static "sun/nio/fs/UnixConstants/EACCES": number;
    public static "sun/nio/fs/UnixConstants/EEXIST": number;
    public static "sun/nio/fs/UnixConstants/ENOTDIR": number;
    public static "sun/nio/fs/UnixConstants/EINVAL": number;
    public static "sun/nio/fs/UnixConstants/EXDEV": number;
    public static "sun/nio/fs/UnixConstants/EISDIR": number;
    public static "sun/nio/fs/UnixConstants/ENOTEMPTY": number;
    public static "sun/nio/fs/UnixConstants/ENOSPC": number;
    public static "sun/nio/fs/UnixConstants/EAGAIN": number;
    public static "sun/nio/fs/UnixConstants/ENOSYS": number;
    public static "sun/nio/fs/UnixConstants/ELOOP": number;
    public static "sun/nio/fs/UnixConstants/EROFS": number;
    public static "sun/nio/fs/UnixConstants/ENODATA": number;
    public static "sun/nio/fs/UnixConstants/ERANGE": number;
    public static "sun/nio/fs/UnixConstants/EMFILE": number;
    public static "sun/nio/fs/UnixConstants/AT_SYMLINK_NOFOLLOW": number;
    public static "sun/nio/fs/UnixConstants/AT_REMOVEDIR": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "sun/nio/fs/UnixConstants/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_file_FileAlreadyExistsException extends java_nio_file_FileSystemException {
    public static "java/nio/file/FileAlreadyExistsException/serialVersionUID": Long;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/FileAlreadyExistsException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/FileAlreadyExistsException/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_nio_file_FileSystemException extends java_io_IOException {
    public static "java/nio/file/FileSystemException/serialVersionUID": Long;
    public "java/nio/file/FileSystemException/file": java_lang_String;
    public "java/nio/file/FileSystemException/other": java_lang_String;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/FileSystemException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/FileSystemException/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "getFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/FileSystemException/getFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getOtherFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/FileSystemException/getOtherFile()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getReason()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/FileSystemException/getReason()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/nio/file/FileSystemException/getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
  }
  export class java_nio_file_NoSuchFileException extends java_nio_file_FileSystemException {
    public static "java/nio/file/NoSuchFileException/serialVersionUID": Long;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/NoSuchFileException/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/nio/file/NoSuchFileException/<init>(Ljava/lang/String;Ljava/lang/String;Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String, java_lang_String, java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Throwable extends java_lang_Object implements java_io_Serializable {
    public static "java/lang/Throwable/serialVersionUID": Long;
    public "java/lang/Throwable/backtrace": java_lang_Object;
    public "java/lang/Throwable/detailMessage": java_lang_String;
    public static "java/lang/Throwable/UNASSIGNED_STACK": JVMArray<java_lang_StackTraceElement>;
    public "java/lang/Throwable/cause": java_lang_Throwable;
    public "java/lang/Throwable/stackTrace": JVMArray<java_lang_StackTraceElement>;
    public static "java/lang/Throwable/SUPPRESSED_SENTINEL": java_util_List;
    public "java/lang/Throwable/suppressedExceptions": java_util_List;
    public static "java/lang/Throwable/NULL_CAUSE_MESSAGE": java_lang_String;
    public static "java/lang/Throwable/SELF_SUPPRESSION_MESSAGE": java_lang_String;
    public static "java/lang/Throwable/CAUSE_CAPTION": java_lang_String;
    public static "java/lang/Throwable/SUPPRESSED_CAPTION": java_lang_String;
    public static "java/lang/Throwable/EMPTY_THROWABLE_ARRAY": JVMArray<java_lang_Throwable>;
    public static "java/lang/Throwable/$assertionsDisabled": number;
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/<init>(Ljava/lang/String;)V"(thread: JVMThread, args: [java_lang_String], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/<init>(Ljava/lang/String;Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/<init>(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/String;Ljava/lang/Throwable;ZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/<init>(Ljava/lang/String;Ljava/lang/Throwable;ZZ)V"(thread: JVMThread, args: [java_lang_String, java_lang_Throwable, number, number], cb?: (e?: java_lang_Throwable) => void): void;
    public "getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Throwable/getMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getLocalizedMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Throwable/getLocalizedMessage()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/getCause()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "initCause(Ljava/lang/Throwable;)Ljava/lang/Throwable;"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/initCause(Ljava/lang/Throwable;)Ljava/lang/Throwable;"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "java/lang/Throwable/toString()Ljava/lang/String;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_String) => void): void;
    public "printStackTrace()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/printStackTrace()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "printStackTrace(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/printStackTrace(Ljava/io/PrintStream;)V"(thread: JVMThread, args: [java_io_PrintStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "printStackTrace(Ljava/lang/Throwable$PrintStreamOrWriter;)V"(thread: JVMThread, args: [java_lang_Throwable$PrintStreamOrWriter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/printStackTrace(Ljava/lang/Throwable$PrintStreamOrWriter;)V"(thread: JVMThread, args: [java_lang_Throwable$PrintStreamOrWriter], cb?: (e?: java_lang_Throwable) => void): void;
    public "printEnclosedStackTrace(Ljava/lang/Throwable$PrintStreamOrWriter;[Ljava/lang/StackTraceElement;Ljava/lang/String;Ljava/lang/String;Ljava/util/Set;)V"(thread: JVMThread, args: [java_lang_Throwable$PrintStreamOrWriter, JVMArray<java_lang_StackTraceElement>, java_lang_String, java_lang_String, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/printEnclosedStackTrace(Ljava/lang/Throwable$PrintStreamOrWriter;[Ljava/lang/StackTraceElement;Ljava/lang/String;Ljava/lang/String;Ljava/util/Set;)V"(thread: JVMThread, args: [java_lang_Throwable$PrintStreamOrWriter, JVMArray<java_lang_StackTraceElement>, java_lang_String, java_lang_String, java_util_Set], cb?: (e?: java_lang_Throwable) => void): void;
    public "printStackTrace(Ljava/io/PrintWriter;)V"(thread: JVMThread, args: [java_io_PrintWriter], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/printStackTrace(Ljava/io/PrintWriter;)V"(thread: JVMThread, args: [java_io_PrintWriter], cb?: (e?: java_lang_Throwable) => void): void;
    public "fillInStackTrace()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/fillInStackTrace()Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "fillInStackTrace(I)Ljava/lang/Throwable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/fillInStackTrace(I)Ljava/lang/Throwable;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_Throwable) => void): void;
    public "getStackTrace()[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_StackTraceElement>) => void): void;
    public "java/lang/Throwable/getStackTrace()[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_StackTraceElement>) => void): void;
    public "getOurStackTrace()[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_StackTraceElement>) => void): void;
    public "java/lang/Throwable/getOurStackTrace()[Ljava/lang/StackTraceElement;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_StackTraceElement>) => void): void;
    public "setStackTrace([Ljava/lang/StackTraceElement;)V"(thread: JVMThread, args: [JVMArray<java_lang_StackTraceElement>], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/setStackTrace([Ljava/lang/StackTraceElement;)V"(thread: JVMThread, args: [JVMArray<java_lang_StackTraceElement>], cb?: (e?: java_lang_Throwable) => void): void;
    public "getStackTraceDepth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "java/lang/Throwable/getStackTraceDepth()I"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: number) => void): void;
    public "getStackTraceElement(I)Ljava/lang/StackTraceElement;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StackTraceElement) => void): void;
    public "java/lang/Throwable/getStackTraceElement(I)Ljava/lang/StackTraceElement;"(thread: JVMThread, args: [number], cb?: (e?: java_lang_Throwable, rv?: java_lang_StackTraceElement) => void): void;
    public "readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/readObject(Ljava/io/ObjectInputStream;)V"(thread: JVMThread, args: [java_io_ObjectInputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/writeObject(Ljava/io/ObjectOutputStream;)V"(thread: JVMThread, args: [java_io_ObjectOutputStream], cb?: (e?: java_lang_Throwable) => void): void;
    public "addSuppressed(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable/addSuppressed(Ljava/lang/Throwable;)V"(thread: JVMThread, args: [java_lang_Throwable], cb?: (e?: java_lang_Throwable) => void): void;
    public "getSuppressed()[Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Throwable>) => void): void;
    public "java/lang/Throwable/getSuppressed()[Ljava/lang/Throwable;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: JVMArray<java_lang_Throwable>) => void): void;
    public static "<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public static "java/lang/Throwable/<clinit>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Throwable$PrintStreamOrWriter extends java_lang_Object {
    public "<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable$PrintStreamOrWriter/<init>()V"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable) => void): void;
    public "lock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "java/lang/Throwable$PrintStreamOrWriter/lock()Ljava/lang/Object;"(thread: JVMThread, args: {}[], cb?: (e?: java_lang_Throwable, rv?: java_lang_Object) => void): void;
    public "println(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable$PrintStreamOrWriter/println(Ljava/lang/Object;)V"(thread: JVMThread, args: [java_lang_Object], cb?: (e?: java_lang_Throwable) => void): void;
    public "<init>(Ljava/lang/Throwable$1;)V"(thread: JVMThread, args: [java_lang_Throwable$1], cb?: (e?: java_lang_Throwable) => void): void;
    public "java/lang/Throwable$PrintStreamOrWriter/<init>(Ljava/lang/Throwable$1;)V"(thread: JVMThread, args: [java_lang_Throwable$1], cb?: (e?: java_lang_Throwable) => void): void;
  }
  export class java_lang_Throwable$1 extends java_lang_Object {
  }
}
export = JVMTypes;
